Package: gpbiometrics
Type: Package
Title: Process and Report Gazepoint Biometrics Data
Version: 0.1.0
Authors@R: person(given = "Stefanos", family = "Balaskas", email = "s.balaskas@ac.upatras.gr", role = c("aut", "cre"))
Description: Imports, validates, quality-checks, preprocesses, summarises, synchronises,
        models, plots, and reports Gazepoint Biometrics exports. The package focuses
        on Gazepoint-specific biometric channels such as GSR/EDA, heart rate,
        interbeat intervals, pulse signals, engagement dial, TTL markers, and
        synchronisation fields that can be combined with Gazepoint GP3 and GP3 HD
        eye-tracking workflows.
License: MIT + file LICENSE
Encoding: UTF-8
Language: en-US
VignetteBuilder: knitr
Suggests: ggplot2, knitr, reticulate, rmarkdown, shiny, testthat (>=
        3.0.0)
Config/testthat/edition: 3
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-06-28 01:00:37 UTC; Stefanos-PC
Author: Stefanos Balaskas [aut, cre]
Maintainer: Stefanos Balaskas <s.balaskas@ac.upatras.gr>
Repository: CRAN
Date/Publication: 2026-07-04 07:10:15 UTC

# Generated by roxygen2: do not edit by hand

export(align_gazepoint_biometrics_to_ttl)
export(analyze_gazepoint_ac_susceptance)
export(analyze_gazepoint_cardiorespiratory_causality)
export(analyze_gazepoint_skin_potential)
export(assess_gazepoint_hrp_waveform_quality)
export(audit_gazepoint_biometric_missingness)
export(audit_gazepoint_biometric_sampling)
export(audit_gazepoint_biometric_sync_drift)
export(audit_gazepoint_distributional_drift)
export(audit_gazepoint_eda_artifacts)
export(audit_gazepoint_engagement_dial)
export(audit_gazepoint_gsr_quality)
export(audit_gazepoint_gsr_units)
export(audit_gazepoint_hr_quality)
export(audit_gazepoint_ibi_quality)
export(audit_gazepoint_signal_activity)
export(audit_gazepoint_stabilization_period)
export(audit_gazepoint_time_resets)
export(baseline_correct_gazepoint_gsr)
export(baseline_correct_gazepoint_hr)
export(baseline_correct_gazepoint_pupil)
export(calculate_gazepoint_rsa)
export(check_gazepoint_biometric_columns)
export(check_gazepoint_plot_contract)
export(chunk_gazepoint_biometrics)
export(classify_gazepoint_eda_response_pattern)
export(classify_gazepoint_scr_intervals)
export(compare_gazepoint_hr_ibi_consistency)
export(convert_gazepoint_gsr_to_conductance)
export(correct_gazepoint_eda_temperature)
export(create_gazepoint_biometrics_checklist)
export(create_gazepoint_biometrics_feature_inventory)
export(create_gazepoint_biometrics_methods_text)
export(create_gazepoint_biometrics_report)
export(create_gazepoint_biometrics_report_tables)
export(create_gazepoint_eda_analysis_pipeline)
export(create_gazepoint_preregistration_template)
export(decompose_gazepoint_eda)
export(denoise_gazepoint_eda_autoencoder)
export(denoise_gazepoint_eda_wavelet)
export(denoise_gazepoint_ppg_autoencoder)
export(denoise_gazepoint_quantization_noise)
export(detect_active_biometric_channels)
export(detect_gazepoint_biometric_schema)
export(detect_gazepoint_biometric_timebase)
export(detect_gazepoint_doubly_stochastic_changepoints)
export(detect_gazepoint_scr_events)
export(detect_gazepoint_scr_peaks)
export(detect_gazepoint_time_columns)
export(diagnose_gazepoint_biometrics_workflow)
export(estimate_gazepoint_signal_lag)
export(export_gazepoint_biometrics_report_bundle)
export(export_gazepoint_rhrv_input)
export(extract_gazepoint_beats_kmeans)
export(extract_gazepoint_bilateral_eda_asymmetry)
export(extract_gazepoint_eda_complexity)
export(extract_gazepoint_eda_spectral_power)
export(extract_gazepoint_eda_tvsymp)
export(extract_gazepoint_edr_pca)
export(extract_gazepoint_hrv_asymmetry)
export(extract_gazepoint_hrv_features)
export(extract_gazepoint_hrv_fragmentation)
export(extract_gazepoint_hrv_fuzzy_csi)
export(extract_gazepoint_hrv_geometric)
export(extract_gazepoint_hrv_nonlinear)
export(extract_gazepoint_hrv_rcmse)
export(extract_gazepoint_hrv_rqa)
export(extract_gazepoint_pdr_signals)
export(extract_gazepoint_respiration_ceemdan)
export(extract_gazepoint_scr_recovery_times)
export(extract_gazepoint_ttl_events)
export(filter_gazepoint_ibi_implausible)
export(flag_gazepoint_artifacts_svm)
export(flag_gazepoint_biometric_dropouts)
export(flag_gazepoint_mad_artifacts)
export(flag_kleckner_eda_artifacts)
export(format_gazepoint_biometrics_feature_inventory)
export(fuse_gazepoint_respiration_kalman)
export(get_gazepoint_plot_data)
export(get_gazepoint_plot_settings)
export(import_gazepoint_biometric_folder)
export(import_gazepoint_biometrics)
export(import_gazepoint_data_summary)
export(import_gazepoint_lsl_xdf)
export(join_gazepoint_biometrics_to_gp3tools)
export(join_gazepoint_biometrics_to_master)
export(model_gazepoint_eda_point_process)
export(model_gazepoint_hr_point_process)
export(model_gazepoint_hrv_ipfm)
export(optimize_gazepoint_cvxeda_tau)
export(plot_gazepoint_aoi_biometrics)
export(plot_gazepoint_biometric_quality)
export(plot_gazepoint_biometric_report_dashboard)
export(plot_gazepoint_biometric_signals)
export(plot_gazepoint_eda_decomposition)
export(plot_gazepoint_eda_gram)
export(plot_gazepoint_multimodal_timeline)
export(plot_gazepoint_saccade_main_sequence)
export(plot_gazepoint_scr_events)
export(plot_gazepoint_scr_specification_curve)
export(plot_gazepoint_signal_activity)
export(plot_gazepoint_time_resets)
export(prepare_gazepoint_aoi_biometrics_model_data)
export(prepare_gazepoint_artifact_svm_features)
export(prepare_gazepoint_biometrics_lme_data)
export(prepare_gazepoint_ctsi_input)
export(prepare_gazepoint_cvxeda_input)
export(prepare_gazepoint_ledalab_input)
export(prepare_gazepoint_multimodal_model_data)
export(prepare_gazepoint_neurokit_eda_input)
export(prepare_gazepoint_pspm_dcm_input)
export(prepare_gazepoint_pspm_input)
export(prepare_gazepoint_pyppg_input)
export(prepare_gazepoint_rhrv_input)
export(prepare_gazepoint_scr_hurdle_model_data)
export(recommend_gazepoint_biometric_exclusions)
export(regress_gazepoint_pupil_luminance)
export(run_gazepoint_automated_statistics)
export(run_gazepoint_biometrics_real_data_readiness)
export(run_gazepoint_biometrics_workflow)
export(run_gazepoint_eda_analysis_pipeline)
export(run_gazepoint_neurokit_eda_crosscheck)
export(run_gazepoint_online_design_optimization)
export(run_gazepoint_scr_multiverse)
export(run_gazepoint_scr_threshold_sensitivity)
export(run_gpbiometrics_shiny)
export(run_gpbiometrics_shiny_annotator)
export(screen_gazepoint_eda_nonresponders)
export(simulate_gazepoint_biometrics)
export(smooth_gazepoint_biometrics)
export(standardise_gazepoint_adaptive_ema)
export(standardise_gazepoint_biometric_names)
export(standardise_gazepoint_biometrics_within_unit)
export(standardise_gazepoint_plot_contract)
export(standardise_gazepoint_range_correction)
export(standardise_gazepoint_zscore)
export(standardize_gazepoint_adaptive_ema)
export(standardize_gazepoint_biometrics_within_unit)
export(standardize_gazepoint_plot_contracts)
export(standardize_gazepoint_range_correction)
export(standardize_gazepoint_zscore)
export(summarise_gazepoint_aoi_biometrics)
export(summarise_gazepoint_biometric_validity)
export(summarise_gazepoint_biometrics_feature_inventory)
export(summarise_gazepoint_biometrics_workflow)
export(summarise_gazepoint_dial_windows)
export(summarise_gazepoint_engagement_windows)
export(summarise_gazepoint_full_biometric_windows)
export(summarise_gazepoint_gsr_tonic_phasic)
export(summarise_gazepoint_gsr_windows)
export(summarise_gazepoint_hr_windows)
export(summarise_gazepoint_hrv_features)
export(summarise_gazepoint_ibi_hrv_windows)
export(summarise_gazepoint_ibi_windows)
export(summarise_gazepoint_multimodal_windows)
export(summarise_gazepoint_scr_event_windows)
export(sync_gazepoint_biometrics_with_gaze)
export(test_gazepoint_hrv_nonlinearity)
export(validate_gazepoint_biometrics)
export(write_gazepoint_biometrics_report_tables)

# gpbiometrics

[![R-CMD-check](https://github.com/stefanosbalaskas/gpbiometrics/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/stefanosbalaskas/gpbiometrics/actions/workflows/R-CMD-check.yaml)
[![DOI](https://zenodo.org/badge/1279519183.svg)](https://doi.org/10.5281/zenodo.20836724)

`gpbiometrics` provides reproducible tools for importing, checking,
preprocessing, summarising, modelling, and reporting Gazepoint
Biometrics and Gazepoint GP3 biometric exports. It is designed for
researchers working with electrodermal activity, pulse/heart-rate
channels, interbeat intervals, TTL markers, stimulus/event timing, and
multimodal Gazepoint workflows.

The package focuses on transparent preprocessing, quality control,
analysis-ready tables, reporting outputs, and conservative physiological
interpretation. It does not infer emotion, stress, cognition,
preference, health status, or diagnosis directly from biometric signals.

## When to use gpbiometrics

Use `gpbiometrics` when you need to:

- import Gazepoint biometric exports from a file or folder;
- inspect signal availability, missingness, time ordering, and TTL
  markers;
- prepare EDA/GSR, SCR, HR, IBI, HRV, PPG, pupil, and multimodal
  features;
- align biometric signals with experimental windows or TTL events;
- create model-ready tables for GLM, mixed-model, or Bayesian workflows;
- export reproducible report bundles and readiness summaries;
- run optional advanced signal-processing and interoperability checks.

## Installation

``` r
# Install from the local package folder during development
# install.packages("devtools")
devtools::load_all()
```

After release, installation can use the package source or repository
location selected by the maintainer.

## Quick start

``` r
library(gpbiometrics)

folder <- "path/to/gazepoint_exports"

dat <- import_gazepoint_biometric_folder(folder)

readiness <- run_gazepoint_biometrics_real_data_readiness(dat)

workflow <- run_gazepoint_biometrics_workflow(
  path = folder,
  include_all_gaze = TRUE,
  include_fixations = FALSE,
  include_other_csv = FALSE,
  expected_sampling_rate_hz = 60
)

summary <- summarise_gazepoint_biometrics_workflow(workflow)
summary
```

For private real data, keep source exports and generated smoke-test
outputs outside the package repository.

## Built-in synthetic kiosk demo

`gpbiometrics` includes a public, fully synthetic Gazepoint-like demo
dataset for examples, vignettes, and reproducible workflow checks. The
dataset simulates a public-service touchscreen kiosk task and is not
derived from real participants.

The synthetic design contains 36 participants, four kiosk tasks per
participant, 60 Hz sampling, and 69,120 rows. The task design crosses
interface complexity (`simple` vs. `dense`) with feedback clarity
(`clear` vs. `ambiguous`). The exports include gaze, AOI labels, pupil
diameter, GSR/EDA, HR, IBI, pulse waveform, engagement dial, TTL
markers, and task metadata.

``` r
demo_dir <- system.file(
  "extdata",
  "gazepoint_biometrics_kiosk_demo_exports",
  package = "gpbiometrics"
)

workflow <- run_gazepoint_biometrics_workflow(
  path = demo_dir,
  include_all_gaze = TRUE,
  include_fixations = FALSE,
  include_other_csv = FALSE,
  expected_sampling_rate_hz = 60
)

summarise_gazepoint_biometrics_workflow(workflow)
```

The synthetic kiosk data are intended for package demonstration only.
They should not be interpreted as real physiology, emotion, stress,
cognition, health status, or diagnosis.

## Main workflow

A typical workflow is:

``` text
Import exports
  -> inspect schema and signal availability
  -> audit quality, timing, TTL markers, and missingness
  -> preprocess EDA / pulse / IBI / pupil channels
  -> extract SCR, HRV, respiration, AOI, or multimodal features
  -> create analysis-ready windows or model tables
  -> export tables, plots, and report bundles
```

The package is intentionally modular: users can run the full workflow or
call individual helpers for a specific signal, window, or
quality-control task.

## Core function map

| Task | Main helpers |
|----|----|
| Import Gazepoint biometric files | `import_gazepoint_biometrics()`, `import_gazepoint_biometric_folder()`, `import_gazepoint_data_summary()` |
| Diagnose workflow readiness | `diagnose_gazepoint_biometrics_workflow()`, `run_gazepoint_biometrics_real_data_readiness()` |
| Run full workflow | `run_gazepoint_biometrics_workflow()`, `summarise_gazepoint_biometrics_workflow()` |
| Export report outputs | `create_gazepoint_biometrics_report()`, `export_gazepoint_biometrics_report_bundle()` |
| Inspect package coverage | `create_gazepoint_biometrics_feature_inventory()`, `gpbiometrics_feature_inventory_table()` |
| Plot report dashboard | `plot_gazepoint_biometric_report_dashboard()` |

## EDA, GSR, and SCR tools

`gpbiometrics` includes tools for EDA/GSR preprocessing, artifact
checks, SCR detection, threshold sensitivity, baseline correction,
response windows, multiverse checks, and report-ready summaries.

Representative helpers include:

``` r
audit_gazepoint_gsr_units()
standardise_gazepoint_adaptive_ema()
classify_gazepoint_scr_intervals()
flag_kleckner_eda_artifacts()
run_gazepoint_scr_multiverse()
extract_gazepoint_scr_recovery_times()
correct_gazepoint_eda_temperature()
```

Specialised EDA helpers are also available for bilateral EDA asymmetry,
skin-potential recordings, AC EDA admittance/susceptance, EDA-gram-style
visualisation, cvxEDA/PsPM/Ledalab/cvxEDA-style bridges, and CTSI input
preparation. These are optional method-specific helpers and should only
be used when the required signal type is actually present.

## Pulse, IBI, HR, HRV, and respiration tools

The package supports pulse and heartbeat workflows, including IBI/HR
summaries, HRV features, nonlinear descriptors, respiration proxies,
point-process summaries, and advanced signal-processing bridges.

Representative helpers include:

``` r
extract_gazepoint_hrv_features()
summarise_gazepoint_hrv_features()
prepare_gazepoint_rhrv_input()
extract_gazepoint_hrv_nonlinear()
extract_gazepoint_hrv_rqa()
extract_gazepoint_hrv_geometric()
extract_gazepoint_hrv_rcmse()
extract_gazepoint_hrv_fuzzy_csi()
model_gazepoint_hrv_ipfm()
extract_gazepoint_beats_kmeans()
```

Respiration-related helpers include PPG-derived respiration, ECG-derived
respiration PCA bridges, CEEMDAN-style respiration extraction, RSA
proxies, and Kalman fusion of multiple respiration proxy streams. These
outputs are signal-derived respiratory proxies, not replacements for
direct respiratory-belt recordings unless externally validated.

## Timing, TTL, windows, and multimodal alignment

Gazepoint biometric exports are often used together with stimulus
timing, TTL markers, AOIs, and event windows. `gpbiometrics` provides
tools for alignment, chunking, and model-ready table creation.

``` r
extract_gazepoint_ttl_events()
align_gazepoint_biometrics_to_ttl()
chunk_gazepoint_biometrics()
prepare_gazepoint_biometrics_lme_data()
plot_gazepoint_multimodal_timeline()
```

## Quality control and reporting

The package emphasises explicit quality-control outputs rather than
silent deletion or hidden preprocessing. Readiness checks and report
bundles help users document what was inspected, what was retained, and
what requires caution.

``` r
run_gazepoint_biometrics_real_data_readiness()
export_gazepoint_biometrics_report_bundle()
create_gazepoint_biometrics_report_tables()
plot_gazepoint_biometric_report_dashboard()
create_gazepoint_preregistration_template()
```

## Advanced optional methods

Advanced helpers are included for specialised workflows such as:

- artifact SVM feature preparation;
- optional autoencoder denoising bridges;
- wavelet and adaptive smoothing;
- entropy, RQA, FuzzyEn, RCMSE, and Lorenz/Poincare-style HRV
  descriptors;
- point-process summaries for EDA and heartbeat timing;
- Granger-style cardiorespiratory directionality;
- EDA-gram-style visualisation;
- AC EDA, skin potential, and bilateral EDA modality support;
- online design-optimisation decision support;
- surrogate nonlinearity testing;
- synthetic biometric signal simulation.

These tools are intended for advanced users who understand the data
requirements and assumptions of each method. Several functions are
interoperability bridges or dependency-light approximations, not
complete reimplementations of external modelling frameworks.

## Interpretation guardrails

Biometric signals require conservative interpretation.

- EDA/GSR/SCR features describe electrodermal activity and
  arousal-related physiology; they do not directly identify emotion,
  stress, preference, cognition, or diagnosis.
- HR, IBI, HRV, PPG, and respiration-proxy features describe
  cardiovascular or signal-derived dynamics; they do not directly
  identify psychological or clinical states.
- Pupil features are strongly affected by luminance and visual context;
  luminance-adjusted residuals are not proof of cognitive-load-only
  effects.
- Eye-tracking or AOI timing features describe visual allocation and
  timing; they do not directly prove attention quality, scrutiny, or
  comprehension.
- Advanced models, bridges, and automated statistics should be reviewed
  against the experimental design before confirmatory use.

## Feature inventory

The package includes a programmatic inventory of available functions.

``` r
inventory <- create_gazepoint_biometrics_feature_inventory()
formatted_inventory <- format_gazepoint_biometrics_feature_inventory(inventory)
inventory_summary <- summarise_gazepoint_biometrics_feature_inventory(formatted_inventory)

inventory$overview
inventory_summary$domain_summary
head(formatted_inventory)
```

This is useful for checking which workflows are available and for
documenting package coverage in reports or manuscripts.

## Private real-data workflow

For private Gazepoint exports, keep raw data outside the package
repository.

``` r
private_folder <- "path/to/private_gazepoint_exports"
private_output <- file.path(tempdir(), "gpbiometrics_real_check")

workflow <- run_gazepoint_biometrics_workflow(
  path = private_folder,
  include_all_gaze = TRUE,
  include_fixations = FALSE,
  include_other_csv = FALSE,
  expected_sampling_rate_hz = 60
)

summarise_gazepoint_biometrics_workflow(workflow)

export_gazepoint_biometrics_report_bundle(
  workflow,
  output_dir = private_output
)
```

Do not commit private Gazepoint exports or private smoke-test outputs.

## Validation status

Current local validation after the advanced development branches:

``` text
devtools::test()
# FAIL 0 | WARN 0 | SKIP 0 | PASS 1649

devtools::check()
# 0 errors | 0 warnings | 0 notes
```

The recurring Quarto/TMPDIR message can appear during local checks. It
is treated as harmless only when the final `R CMD check` summary reports
0 errors, 0 warnings, and 0 notes.

## Citation

If you use `gpbiometrics`, cite the package version, repository or
archive DOI if available, and the specific Gazepoint hardware/software
context used in your study. Also cite the original methodological
sources for specialised analyses such as cvxEDA, Ledalab, PsPM, RHRV,
entropy, point-process, or signal-decomposition methods when those
helpers are used.

If you use `gpbiometrics`, please cite the package with:

``` r
citation("gpbiometrics")
```

The archived software release is available at DOI:
`10.5281/zenodo.20836725`.

Synthetic Gazepoint Biometrics kiosk demo exports

This folder contains fully synthetic Gazepoint-like all-gaze exports for gpbiometrics examples.
The data are artificial and are not derived from real participants.

Scenario:
Participants complete public-service touchscreen kiosk tasks.
The synthetic design crosses interface complexity (simple/dense) and feedback clarity (clear/ambiguous).

Files:
- synthetic_kiosk_p001_all_gaze.csv ... synthetic_kiosk_p036_all_gaze.csv
- synthetic_kiosk_trial_design.csv

Signals include gaze, AOI labels, pupil diameter, GSR/EDA, HR, IBI, pulse waveform, engagement dial, and TTL markers.

Interpretation:
Known synthetic signal patterns are included for demonstration only.
Do not interpret these data as real physiology, emotion, stress, cognition, health status, or diagnosis.

#' Create a Gazepoint biometrics preregistration template
#'
#' Creates a cautious preregistration template for Gazepoint Biometrics EDA/GSR
#' workflows.
#'
#' @param study_title Study title.
#' @param signal_standardization Standardization plan.
#' @param artifact_rules Artifact-rule description.
#' @param eda_min_us Minimum conductance threshold.
#' @param eda_max_us Maximum conductance threshold.
#' @param rapid_change_threshold Maximum absolute percent change per second.
#' @param output_file Optional path to write the template as a text file.
#'
#' @return A character string.
#' @export
create_gazepoint_preregistration_template <- function(study_title = "Gazepoint biometrics study",
                                                      signal_standardization = c("within_participant_z", "range_correction", "none"),
                                                      artifact_rules = c("kleckner_style", "custom", "none"),
                                                      eda_min_us = 0.01,
                                                      eda_max_us = 100,
                                                      rapid_change_threshold = 20,
                                                      output_file = NULL) {
  signal_standardization <- match.arg(signal_standardization)
  artifact_rules <- match.arg(artifact_rules)

  standardization_text <- switch(
    signal_standardization,
    within_participant_z = paste(
      "Biometric responses will be standardised within participant using z-scoring:",
      "z = (x - participant mean) / participant standard deviation.",
      "This transformation is intended to support within-participant comparison and will be reported as removing between-participant level and scale differences."
    ),
    range_correction = paste(
      "Biometric responses will be range-corrected within participant:",
      "(x - participant minimum) / (participant maximum - participant minimum).",
      "This transformation is intended to express responses as a proportion of each participant's observed signal range."
    ),
    none = "No participant-level signal standardisation will be applied before the primary model unless specified in a sensitivity analysis."
  )

  artifact_text <- switch(
    artifact_rules,
    kleckner_style = paste(
      "EDA artifacts will be flagged using transparent Kleckner-style heuristic rules:",
      "non-finite samples, conductance values outside the planned physiological range, rapid absolute percent changes per second,",
      "and transition padding around bad samples.",
      "The planned conductance range is",
      paste0("[", eda_min_us, ", ", eda_max_us, "] microsiemens"),
      "and the planned rapid-change threshold is",
      paste0(rapid_change_threshold, "% per second.")
    ),
    custom = "EDA artifacts will be flagged using study-specific rules defined before analysis and reported with thresholds and exclusion counts.",
    none = "No automated EDA artifact rule will be applied before analysis; manual or visual quality checks will be reported if used."
  )

  text <- paste(
    "# Preregistration template:",
    study_title,
    "",
    "## Data source",
    "Raw Gazepoint Biometrics exports will be imported using gpbiometrics. The analysis will report imported file counts, detected biometric channels, detected time columns, detected timebase, and available TTL/event markers.",
    "",
    "## Quality control",
    "Before statistical analysis, the data will be audited for missingness, inactive or all-zero channels, grouped counter resets, row-order problems, EDA artifacts, and usable signal coverage.",
    "",
    "## EDA preprocessing",
    "GSR/EDA conductance conversion will be applied only when the unit assumptions are explicit. The analysis will report smoothing, baseline-correction, SCR peak-detection, and event-window settings.",
    "",
    "## Standardisation plan",
    standardization_text,
    "",
    "## Artifact rules",
    artifact_text,
    "",
    "## SCR interval windows",
    "If SCR latency-window analysis is used, responses will be classified into FIR, SIR, and TIR windows using prespecified latency boundaries. These labels will be treated as latency descriptors, not emotion or diagnosis labels.",
    "",
    "## Statistical modelling",
    "SCR occurrence/amplitude and continuous SCL/EDA outcomes will be analysed using model-ready tables prepared by gpbiometrics. Hurdle or mixed-effects models will be fitted outside gpbiometrics in appropriate modelling packages.",
    "",
    "## Interpretation guardrail",
    "EDA/GSR, heart-rate, pupil, and gaze measures will not be interpreted as direct evidence of emotion, valence, stress, trust, preference, cognition, or diagnosis without converging design, behavioural, and self-report evidence.",
    sep = "\n"
  )

  if (!is.null(output_file)) {
    writeLines(text, con = output_file, useBytes = TRUE)
  }

  text
}

#' Launch a lightweight gpbiometrics Shiny dashboard
#'
#' Launches an optional local Shiny interface for inspecting a Gazepoint
#' Biometrics CSV file. Shiny is optional and is not required to use the package.
#'
#' @return A Shiny application object, launched for interactive use.
#' @export
run_gpbiometrics_shiny <- function() {
  if (!requireNamespace("shiny", quietly = TRUE)) {
    stop(
      "The optional package `shiny` is required. Install it with install.packages('shiny').",
      call. = FALSE
    )
  }

  ui <- shiny::fluidPage(
    shiny::titlePanel("gpbiometrics local QC dashboard"),
    shiny::sidebarLayout(
      shiny::sidebarPanel(
        shiny::fileInput("file", "Upload Gazepoint CSV", accept = c(".csv", ".txt")),
        shiny::textInput("time_col", "Time column", value = "CNT"),
        shiny::textInput("signal_cols", "Signal columns, comma-separated", value = "GSR_US, HR, IBI, DIAL"),
        shiny::actionButton("run", "Run QC")
      ),
      shiny::mainPanel(
        shiny::h4("Preview"),
        shiny::tableOutput("preview"),
        shiny::h4("Signal activity"),
        shiny::verbatimTextOutput("activity"),
        shiny::h4("Notes"),
        shiny::verbatimTextOutput("notes")
      )
    )
  )

  server <- function(input, output, session) {
    dat <- shiny::reactive({
      req <- input$file

      if (is.null(req)) {
        return(NULL)
      }

      utils::read.csv(req$datapath, stringsAsFactors = FALSE, check.names = FALSE)
    })

    output$preview <- shiny::renderTable({
      x <- dat()
      if (is.null(x)) {
        return(NULL)
      }
      utils::head(x)
    })

    qc <- shiny::eventReactive(input$run, {
      x <- dat()

      if (is.null(x)) {
        return("Upload a file first.")
      }

      signal_cols <- trimws(strsplit(input$signal_cols, ",")[[1]])
      signal_cols <- intersect(signal_cols, names(x))

      if (length(signal_cols) == 0) {
        return("No requested signal columns were found.")
      }

      audit <- audit_gazepoint_signal_activity(
        x,
        signal_cols = signal_cols
      )

      utils::capture.output(print(audit))
    })

    output$activity <- shiny::renderText({
      paste(qc(), collapse = "\n")
    })

    output$notes <- shiny::renderText({
      paste(
        "This GUI is a lightweight local helper.",
        "It supports quick inspection only and does not infer emotion, valence, stress, trust, preference, cognition, or diagnosis.",
        sep = "\n"
      )
    })
  }

  shiny::shinyApp(ui = ui, server = server)
}

test_that("baseline_correct_gazepoint_pupil applies subtractive correction", {
  dat <- data.frame(
    participant = "p1",
    trial = 1,
    time = c(-240, -220, -200, 0, 100, 200),
    pupil = c(3, 3.2, 3.4, 4, 4.2, 4.4)
  )

  out <- baseline_correct_gazepoint_pupil(
    dat,
    pupil_col = "pupil",
    time_col = "time",
    trial_cols = c("participant", "trial"),
    baseline_window = c(-240, -200)
  )

  expect_true("pupil_baseline_corrected" %in% names(out))
  expect_equal(out$pupil_baseline_corrected[4], 4 - 3.2, tolerance = 1e-10)

  summary <- attr(out, "pupil_baseline_summary")
  expect_equal(summary$status, "pupil_baseline_correction_complete")
})

test_that("plot_gazepoint_saccade_main_sequence returns plotted data invisibly", {
  dat <- data.frame(
    amplitude_deg = c(1, 2, 3, 4, 5),
    peak_velocity_deg_s = c(100, 180, 250, 300, 340)
  )

  out <- plot_gazepoint_saccade_main_sequence(
    dat,
    amplitude_col = "amplitude_deg",
    peak_velocity_col = "peak_velocity_deg_s",
    add_smoother = FALSE
  )

  expect_true(is.list(out))
  expect_equal(nrow(out$data), 5)
})

test_that("create_gazepoint_preregistration_template returns cautious text", {
  txt <- create_gazepoint_preregistration_template(
    study_title = "Test study",
    signal_standardization = "within_participant_z",
    artifact_rules = "kleckner_style"
  )

  expect_true(grepl("Test study", txt))
  expect_true(grepl("z =", txt, fixed = TRUE))
  expect_true(grepl("Kleckner-style", txt))
  expect_true(grepl("does not", txt) || grepl("will not", txt))
})

test_that("run_gpbiometrics_shiny errors clearly without shiny or returns app", {
  if (!requireNamespace("shiny", quietly = TRUE)) {
    expect_error(run_gpbiometrics_shiny(), "shiny")
  } else {
    app <- run_gpbiometrics_shiny()
    expect_true(inherits(app, "shiny.appobj"))
  }
})

#' Extract Gazepoint TTL marker events
#'
#' Extracts TTL marker events from Gazepoint Biometrics exports. The function
#' can return either rows where TTL marker values change or all nonzero TTL
#' rows. By default, rows are retained only when the TTL validity column is
#' present and greater than zero. This avoids treating invalid placeholder TTL
#' values as experimental events.
#'
#' @param data A data frame or a path to a Gazepoint CSV export.
#' @param ttl_columns TTL marker columns. If `NULL`, the function uses all
#'   available columns from `TTL0` to `TTL6`.
#' @param group_columns Optional grouping columns within which TTL changes are
#'   detected, such as `source_participant`, `USER`, or `MEDIA_ID`.
#' @param validity_column Optional TTL validity column. Defaults to `"TTLV"`.
#' @param require_validity Logical. Should rows be retained only when
#'   `validity_column` is present and greater than zero? Defaults to `TRUE`.
#' @param mode Event extraction mode. `"changes"` returns rows where TTL values
#'   change. `"nonzero"` returns all rows with nonzero TTL values.
#' @param include_initial Should the first valid, non-missing TTL value within
#'   each group be treated as an event when `mode = "changes"`?
#'
#' @return A data frame of TTL events.
#'
#' @export
extract_gazepoint_ttl_events <- function(data,
                                         ttl_columns = NULL,
                                         group_columns = NULL,
                                         validity_column = "TTLV",
                                         require_validity = TRUE,
                                         mode = c("changes", "nonzero"),
                                         include_initial = TRUE) {
  dat <- coerce_gazepoint_biometrics_data(data)
  mode <- match.arg(mode)

  if (is.null(ttl_columns)) {
    ttl_columns <- intersect(paste0("TTL", 0:6), names(dat))
  }

  if (length(ttl_columns) == 0L) {
    stop("No TTL columns were found in `data`.", call. = FALSE)
  }

  missing_ttl <- setdiff(ttl_columns, names(dat))

  if (length(missing_ttl) > 0L) {
    stop(
      "`ttl_columns` were not found in `data`: ",
      paste(missing_ttl, collapse = ", "),
      call. = FALSE
    )
  }

  if (!is.null(group_columns)) {
    missing_groups <- setdiff(group_columns, names(dat))

    if (length(missing_groups) > 0L) {
      stop(
        "`group_columns` were not found in `data`: ",
        paste(missing_groups, collapse = ", "),
        call. = FALSE
      )
    }

    group_key <- make_group_key(dat, group_columns)
  } else {
    group_key <- rep("all", nrow(dat))
  }

  time_columns <- intersect(c("CNT", "TIME", "TIME_TICK"), names(dat))

  validity_present <- !is.null(validity_column) &&
    length(validity_column) == 1L &&
    validity_column %in% names(dat)

  rows <- list()
  row_counter <- 0L

  for (key in unique(group_key)) {
    group_rows <- which(group_key == key)

    group_validity <- if (validity_present) {
      as_numeric_safe(dat[[validity_column]][group_rows])
    } else {
      rep(NA_real_, length(group_rows))
    }

    valid_mask <- if (isTRUE(require_validity)) {
      validity_present & !is.na(group_validity) & group_validity > 0
    } else {
      rep(TRUE, length(group_rows))
    }

    for (ttl_column in ttl_columns) {
      ttl_values <- as_numeric_safe(dat[[ttl_column]][group_rows])

      event_positions <- detect_ttl_event_positions(
        ttl_values = ttl_values,
        valid_mask = valid_mask,
        mode = mode,
        include_initial = include_initial
      )

      if (length(event_positions) == 0L) {
        next
      }

      for (event_position in event_positions) {
        source_row <- group_rows[event_position]
        row_counter <- row_counter + 1L

        previous_position <- find_previous_valid_ttl_position(
          ttl_values = ttl_values,
          valid_mask = valid_mask,
          current_position = event_position
        )

        previous_value <- if (!is.na(previous_position)) {
          ttl_values[previous_position]
        } else {
          NA_real_
        }

        base <- data.frame(
          row_index = source_row,
          event_order = row_counter,
          ttl_channel = ttl_column,
          ttl_value = ttl_values[event_position],
          previous_ttl_value = previous_value,
          stringsAsFactors = FALSE
        )

        if (!is.null(group_columns)) {
          base <- cbind(
            dat[source_row, group_columns, drop = FALSE],
            base,
            stringsAsFactors = FALSE
          )
        }

        if (length(time_columns) > 0L) {
          base <- cbind(
            base,
            dat[source_row, time_columns, drop = FALSE],
            stringsAsFactors = FALSE
          )
        }

        if (validity_present) {
          base$ttl_validity <- as_numeric_safe(dat[[validity_column]][source_row])
        } else {
          base$ttl_validity <- NA_real_
        }

        rows[[length(rows) + 1L]] <- base
      }
    }
  }

  if (length(rows) == 0L) {
    return(empty_ttl_events(group_columns, time_columns))
  }

  out <- do.call(rbind, rows)
  rownames(out) <- NULL
  out
}


detect_ttl_event_positions <- function(ttl_values,
                                       valid_mask,
                                       mode,
                                       include_initial) {
  present <- !is.na(ttl_values) & valid_mask

  if (mode == "nonzero") {
    return(which(present & ttl_values != 0))
  }

  event <- rep(FALSE, length(ttl_values))

  if (length(ttl_values) == 0L) {
    return(integer(0))
  }

  previous_valid_value <- NA_real_

  for (i in seq_along(ttl_values)) {
    if (!present[i]) {
      next
    }

    if (is.na(previous_valid_value)) {
      event[i] <- isTRUE(include_initial)
    } else {
      event[i] <- ttl_values[i] != previous_valid_value
    }

    previous_valid_value <- ttl_values[i]
  }

  which(event)
}


find_previous_valid_ttl_position <- function(ttl_values,
                                             valid_mask,
                                             current_position) {
  if (current_position <= 1L) {
    return(NA_integer_)
  }

  previous_positions <- seq_len(current_position - 1L)
  previous_positions <- previous_positions[
    valid_mask[previous_positions] & !is.na(ttl_values[previous_positions])
  ]

  if (length(previous_positions) == 0L) {
    return(NA_integer_)
  }

  previous_positions[length(previous_positions)]
}


empty_ttl_events <- function(group_columns,
                             time_columns) {
  out <- data.frame(
    row_index = integer(0),
    event_order = integer(0),
    ttl_channel = character(0),
    ttl_value = numeric(0),
    previous_ttl_value = numeric(0),
    stringsAsFactors = FALSE
  )

  if (!is.null(group_columns) && length(group_columns) > 0L) {
    groups <- as.data.frame(
      stats::setNames(
        replicate(length(group_columns), character(0), simplify = FALSE),
        group_columns
      ),
      stringsAsFactors = FALSE
    )

    out <- cbind(groups, out, stringsAsFactors = FALSE)
  }

  if (length(time_columns) > 0L) {
    times <- as.data.frame(
      stats::setNames(
        replicate(length(time_columns), numeric(0), simplify = FALSE),
        time_columns
      ),
      stringsAsFactors = FALSE
    )

    out <- cbind(out, times, stringsAsFactors = FALSE)
  }

  out$ttl_validity <- numeric(0)
  out
}

test_that("extract_gazepoint_ttl_events extracts TTL changes", {
  dat <- data.frame(
    source_participant = c("User 1", "User 1", "User 1", "User 1"),
    MEDIA_ID = c(0, 0, 0, 0),
    CNT = c(1, 2, 3, 4),
    TTL0 = c(1007, 1007, 1008, 1008),
    TTLV = c(1, 1, 1, 1)
  )

  out <- extract_gazepoint_ttl_events(
    dat,
    group_columns = c("source_participant", "MEDIA_ID")
  )

  expect_equal(nrow(out), 2)
  expect_equal(out$ttl_channel, c("TTL0", "TTL0"))
  expect_equal(out$ttl_value, c(1007, 1008))
  expect_equal(out$CNT, c(1, 3))
  expect_equal(out$ttl_validity, c(1, 1))
})


test_that("extract_gazepoint_ttl_events can omit initial TTL values", {
  dat <- data.frame(
    CNT = c(1, 2, 3, 4),
    TTL0 = c(1007, 1007, 1008, 1008),
    TTLV = c(1, 1, 1, 1)
  )

  out <- extract_gazepoint_ttl_events(
    dat,
    include_initial = FALSE
  )

  expect_equal(nrow(out), 1)
  expect_equal(out$ttl_value, 1008)
  expect_equal(out$previous_ttl_value, 1007)
})


test_that("extract_gazepoint_ttl_events supports nonzero mode", {
  dat <- data.frame(
    CNT = c(1, 2, 3, 4),
    TTL0 = c(0, 1007, 0, 1008),
    TTLV = c(1, 1, 1, 1)
  )

  out <- extract_gazepoint_ttl_events(
    dat,
    mode = "nonzero"
  )

  expect_equal(nrow(out), 2)
  expect_equal(out$ttl_value, c(1007, 1008))
  expect_equal(out$CNT, c(2, 4))
})


test_that("extract_gazepoint_ttl_events separates groups", {
  dat <- data.frame(
    source_participant = c("User 1", "User 1", "User 2", "User 2"),
    CNT = c(1, 2, 1, 2),
    TTL0 = c(1007, 1008, 1007, 1008),
    TTLV = c(1, 1, 1, 1)
  )

  out <- extract_gazepoint_ttl_events(
    dat,
    group_columns = "source_participant",
    include_initial = FALSE
  )

  expect_equal(nrow(out), 2)
  expect_true(all(c("User 1", "User 2") %in% out$source_participant))
  expect_equal(out$ttl_value, c(1008, 1008))
})


test_that("extract_gazepoint_ttl_events filters invalid TTL rows by default", {
  dat <- data.frame(
    CNT = c(1, 2, 3, 4),
    TTL0 = c(1007, 1008, 1009, 1010),
    TTLV = c(0, 0, 1, 1)
  )

  out <- extract_gazepoint_ttl_events(dat)

  expect_equal(nrow(out), 2)
  expect_equal(out$ttl_value, c(1009, 1010))
  expect_equal(out$ttl_validity, c(1, 1))
})


test_that("extract_gazepoint_ttl_events can include invalid TTL rows when requested", {
  dat <- data.frame(
    CNT = c(1, 2, 3, 4),
    TTL0 = c(1007, 1008, 1009, 1010),
    TTLV = c(0, 0, 1, 1)
  )

  out <- extract_gazepoint_ttl_events(
    dat,
    require_validity = FALSE
  )

  expect_equal(nrow(out), 4)
  expect_equal(out$ttl_value, c(1007, 1008, 1009, 1010))
})


test_that("extract_gazepoint_ttl_events rejects missing TTL columns", {
  dat <- data.frame(
    CNT = c(1, 2),
    GSR_US = c(1.1, 1.2)
  )

  expect_error(
    extract_gazepoint_ttl_events(dat),
    "No TTL columns"
  )
})


test_that("extract_gazepoint_ttl_events returns empty table when no events exist", {
  dat <- data.frame(
    source_participant = c("User 1", "User 1"),
    CNT = c(1, 2),
    TTL0 = c(0, 0),
    TTLV = c(1, 1)
  )

  out <- extract_gazepoint_ttl_events(
    dat,
    group_columns = "source_participant",
    mode = "nonzero"
  )

  expect_true(is.data.frame(out))
  expect_equal(nrow(out), 0)
  expect_true("source_participant" %in% names(out))
  expect_true("ttl_channel" %in% names(out))
})

#' Detect Gazepoint SCR peaks
#'
#' Detects candidate skin conductance responses (SCRs) from Gazepoint EDA/GSR
#' signals. The helper prefers a phasic channel such as `GSR_US_PHASIC` when
#' available, and otherwise falls back to a conductance-like signal such as
#' `GSR_US`. It returns explicit onset, peak, amplitude, rise-time, and
#' recovery-time fields for downstream event-window summaries and statistical
#' modelling.
#'
#' This is a conservative R-native peak detector. It is not a replacement for
#' full model-based EDA decomposition tools such as Ledalab, PsPM, or cvxEDA.
#'
#' @param data A data frame containing Gazepoint biometric rows.
#' @param signal_col Optional conductance-like signal column, typically
#'   `GSR_US`. Used when `phasic_col` is absent or unavailable.
#' @param phasic_col Optional phasic EDA signal column, typically
#'   `GSR_US_PHASIC`.
#' @param time_col Optional time/counter column. If `NULL`, common Gazepoint
#'   time columns are detected automatically.
#' @param group_cols Optional grouping columns. If `NULL`, available
#'   source/participant/media/trial-like columns are used.
#' @param prefer_vendor_phasic Logical. If `TRUE`, prefer `GSR_US_PHASIC` when
#'   available.
#' @param amplitude_min Minimum trough-to-peak amplitude required for a detected
#'   SCR.
#' @param recovery_fraction Fraction of the peak amplitude used to define
#'   recovery. The default `.5` estimates half-recovery.
#' @param smooth_width Optional odd integer moving-average width. Use `1` for no
#'   smoothing.
#' @param min_peak_distance Minimum distance, in rows, allowed between retained
#'   candidate peaks within each group. The default `1` preserves all local
#'   maxima. Larger values reduce repeated detection of closely spaced local
#'   maxima within a sustained SCR-like response.
#'
#' @return A list with `overview`, `peaks`, `group_summary`, `signal_summary`,
#'   and `settings`.
#' @export
detect_gazepoint_scr_peaks <- function(data,
                                       signal_col = NULL,
                                       phasic_col = NULL,
                                       time_col = NULL,
                                       group_cols = NULL,
                                       prefer_vendor_phasic = TRUE,
                                       amplitude_min = 0.01,
                                       recovery_fraction = 0.5,
                                       smooth_width = 1,
                                       min_peak_distance = 1) {
  if (!is.data.frame(data)) {
    stop("`data` must be a data frame.", call. = FALSE)
  }

  if (!is.logical(prefer_vendor_phasic) || length(prefer_vendor_phasic) != 1) {
    stop("`prefer_vendor_phasic` must be TRUE or FALSE.", call. = FALSE)
  }

  if (!is.numeric(amplitude_min) || length(amplitude_min) != 1 || amplitude_min < 0) {
    stop("`amplitude_min` must be a non-negative number.", call. = FALSE)
  }

  if (!is.numeric(recovery_fraction) ||
      length(recovery_fraction) != 1 ||
      recovery_fraction <= 0 ||
      recovery_fraction >= 1) {
    stop("`recovery_fraction` must be a number between 0 and 1.", call. = FALSE)
  }

  if (!is.numeric(smooth_width) || length(smooth_width) != 1 || smooth_width < 1) {
    stop("`smooth_width` must be a positive integer.", call. = FALSE)
  }

  if (!is.numeric(min_peak_distance) ||
      length(min_peak_distance) != 1 ||
      min_peak_distance < 1) {
    stop("`min_peak_distance` must be a positive integer.", call. = FALSE)
  }

  smooth_width <- as.integer(round(smooth_width))
  min_peak_distance <- as.integer(round(min_peak_distance))

  dat <- as.data.frame(data, stringsAsFactors = FALSE)
  dat$.gpbiometrics_row_id <- seq_len(nrow(dat))

  names_dat <- names(dat)

  source_signal <- gpbiometrics_scr_resolve_signal_col(
    names_dat = names_dat,
    signal_col = signal_col,
    phasic_col = phasic_col,
    prefer_vendor_phasic = prefer_vendor_phasic
  )

  if (is.null(source_signal) || !source_signal %in% names_dat) {
    stop(
      "No usable SCR signal column was found. Supply `phasic_col` or `signal_col`.",
      call. = FALSE
    )
  }

  if (is.null(time_col)) {
    time_col <- gpbiometrics_scr_first_existing(
      names_dat,
      c(
        "time_ms", "timestamp_ms", "timestamp",
        "TIME", "Time", "time",
        "CNT", "cnt"
      )
    )
  }

  if (!is.null(time_col) && !time_col %in% names_dat) {
    stop("`time_col` was not found in `data`.", call. = FALSE)
  }

  group_cols <- gpbiometrics_scr_resolve_group_cols(names_dat, group_cols)

  missing_group_cols <- setdiff(group_cols, names_dat)

  if (length(missing_group_cols) > 0) {
    stop(
      "`group_cols` not found in `data`: ",
      paste(missing_group_cols, collapse = ", "),
      call. = FALSE
    )
  }

  signal_raw <- suppressWarnings(as.numeric(dat[[source_signal]]))

  if (all(is.na(signal_raw))) {
    stop("The selected SCR signal column must be numeric or numeric-coercible.", call. = FALSE)
  }

  time_value <- if (!is.null(time_col)) {
    suppressWarnings(as.numeric(dat[[time_col]]))
  } else {
    seq_len(nrow(dat))
  }

  group_id <- gpbiometrics_scr_group_id(dat, group_cols)
  group_indices <- split(seq_len(nrow(dat)), group_id, drop = TRUE)

  peaks_list <- list()
  group_summary_list <- list()

  for (group_name in names(group_indices)) {
    idx <- group_indices[[group_name]]
    x_raw <- signal_raw[idx]
    t <- time_value[idx]

    x <- gpbiometrics_scr_smooth_signal(x_raw, smooth_width = smooth_width)

    candidate_peak_index <- gpbiometrics_scr_candidate_peaks(x)
    candidate_peak_index <- gpbiometrics_scr_filter_candidate_peaks(
      x = x,
      candidate_peak_index = candidate_peak_index,
      min_peak_distance = min_peak_distance
    )
    candidate_count <- length(candidate_peak_index)

    detected_for_group <- list()
    below_threshold_count <- 0L
    incomplete_recovery_count <- 0L

    if (candidate_count > 0) {
      for (candidate_i in seq_along(candidate_peak_index)) {
        peak_local <- candidate_peak_index[candidate_i]

        peak <- gpbiometrics_scr_score_peak(
          x = x,
          t = t,
          peak_local = peak_local,
          amplitude_min = amplitude_min,
          recovery_fraction = recovery_fraction
        )

        if (is.null(peak)) {
          below_threshold_count <- below_threshold_count + 1L
          next
        }

        if (identical(peak$status, "detected_incomplete_recovery")) {
          incomplete_recovery_count <- incomplete_recovery_count + 1L
        }

        peak_row <- data.frame(
          group_id = group_name,
          peak_id = length(detected_for_group) + 1L,
          source_signal = source_signal,
          onset_row_id = dat$.gpbiometrics_row_id[idx[peak$onset_index]],
          peak_row_id = dat$.gpbiometrics_row_id[idx[peak$peak_index]],
          recovery_row_id = if (is.na(peak$recovery_index)) {
            NA_integer_
          } else {
            dat$.gpbiometrics_row_id[idx[peak$recovery_index]]
          },
          onset_index = peak$onset_index,
          peak_index = peak$peak_index,
          recovery_index = peak$recovery_index,
          onset_time = peak$onset_time,
          peak_time = peak$peak_time,
          recovery_time = peak$recovery_time,
          onset_value = peak$onset_value,
          peak_value = peak$peak_value,
          recovery_value = peak$recovery_value,
          amplitude = peak$amplitude,
          rise_time = peak$rise_time,
          recovery_time_after_peak = peak$recovery_time_after_peak,
          status = peak$status,
          stringsAsFactors = FALSE
        )

        if (length(group_cols) > 0) {
          group_values <- dat[idx[1], group_cols, drop = FALSE]
          peak_row <- cbind(group_values, peak_row)
        }

        detected_for_group[[length(detected_for_group) + 1L]] <- peak_row
      }
    }

    if (length(detected_for_group) > 0) {
      peaks_list <- c(peaks_list, detected_for_group)
    }

    finite_x <- x[is.finite(x)]
    group_status <- if (length(finite_x) < 3) {
      "insufficient_signal"
    } else if (length(detected_for_group) == 0 && candidate_count == 0) {
      "no_candidate_peaks"
    } else if (length(detected_for_group) == 0) {
      "no_peaks_above_threshold"
    } else if (incomplete_recovery_count > 0) {
      "peaks_detected_with_incomplete_recovery"
    } else {
      "peaks_detected"
    }

    group_row <- data.frame(
      group_id = group_name,
      rows = length(idx),
      finite_signal_rows = length(finite_x),
      candidate_peaks = candidate_count,
      detected_peaks = length(detected_for_group),
      below_threshold_peaks = below_threshold_count,
      incomplete_recovery_peaks = incomplete_recovery_count,
      signal_min = if (length(finite_x) > 0) min(finite_x, na.rm = TRUE) else NA_real_,
      signal_max = if (length(finite_x) > 0) max(finite_x, na.rm = TRUE) else NA_real_,
      signal_sd = if (length(finite_x) > 1) stats::sd(finite_x, na.rm = TRUE) else NA_real_,
      status = group_status,
      stringsAsFactors = FALSE
    )

    if (length(group_cols) > 0) {
      group_values <- dat[idx[1], group_cols, drop = FALSE]
      group_row <- cbind(group_values, group_row)
    }

    group_summary_list[[length(group_summary_list) + 1L]] <- group_row
  }

  peaks <- if (length(peaks_list) > 0) {
    out <- do.call(rbind, peaks_list)
    rownames(out) <- NULL
    out
  } else {
    gpbiometrics_scr_empty_peaks(group_cols)
  }

  group_summary <- do.call(rbind, group_summary_list)
  rownames(group_summary) <- NULL

  signal_summary <- data.frame(
    source_signal = source_signal,
    input_rows = nrow(dat),
    finite_signal_rows = sum(is.finite(signal_raw)),
    missing_signal_rows = sum(is.na(signal_raw)),
    signal_min = if (any(is.finite(signal_raw))) min(signal_raw, na.rm = TRUE) else NA_real_,
    signal_max = if (any(is.finite(signal_raw))) max(signal_raw, na.rm = TRUE) else NA_real_,
    signal_sd = if (sum(is.finite(signal_raw)) > 1) stats::sd(signal_raw, na.rm = TRUE) else NA_real_,
    stringsAsFactors = FALSE
  )

  total_detected <- nrow(peaks)
  total_candidates <- sum(group_summary$candidate_peaks, na.rm = TRUE)

  status <- if (total_detected > 0) {
    "peaks_detected"
  } else if (total_candidates > 0) {
    "candidate_peaks_below_threshold"
  } else {
    "no_candidate_peaks"
  }

  overview <- data.frame(
    input_rows = nrow(dat),
    source_signal = source_signal,
    time_col = if (is.null(time_col)) NA_character_ else time_col,
    group_count = length(group_indices),
    candidate_peaks = total_candidates,
    detected_peaks = total_detected,
    amplitude_min = amplitude_min,
    recovery_fraction = recovery_fraction,
    status = status,
    stringsAsFactors = FALSE
  )

  structure(
    list(
      overview = overview,
      peaks = peaks,
      group_summary = group_summary,
      signal_summary = signal_summary,
      settings = list(
        signal_col = signal_col,
        phasic_col = phasic_col,
        source_signal = source_signal,
        time_col = time_col,
        group_cols = group_cols,
        prefer_vendor_phasic = prefer_vendor_phasic,
        amplitude_min = amplitude_min,
        recovery_fraction = recovery_fraction,
        smooth_width = smooth_width,
        min_peak_distance = min_peak_distance,
        interpretation_notes = c(
          "SCR peaks are electrodermal response features, not emotional-valence labels.",
          "This simple detector is intended for transparent Gazepoint-native QC and feature extraction.",
          "For dense or strongly overlapping event designs, external deconvolution cross-checks such as NeuroKit2, Ledalab, PsPM, or cvxEDA may be appropriate."
        )
      )
    ),
    class = c("gazepoint_scr_peak_detection", "list")
  )
}

gpbiometrics_scr_first_existing <- function(names_dat, candidates) {
  exact <- candidates[candidates %in% names_dat]

  if (length(exact) > 0) {
    return(exact[1])
  }

  lower_names <- tolower(names_dat)
  lower_candidates <- tolower(candidates)
  idx <- match(lower_candidates, lower_names)
  idx <- idx[!is.na(idx)]

  if (length(idx) > 0) {
    return(names_dat[idx[1]])
  }

  NULL
}

gpbiometrics_scr_resolve_signal_col <- function(names_dat,
                                                signal_col,
                                                phasic_col,
                                                prefer_vendor_phasic) {
  if (!is.null(phasic_col)) {
    if (!phasic_col %in% names_dat) {
      stop("`phasic_col` was not found in `data`.", call. = FALSE)
    }
    return(phasic_col)
  }

  if (isTRUE(prefer_vendor_phasic)) {
    vendor_phasic <- gpbiometrics_scr_first_existing(
      names_dat,
      c("GSR_US_PHASIC", "gsr_us_phasic")
    )

    if (!is.null(vendor_phasic)) {
      return(vendor_phasic)
    }
  }

  if (!is.null(signal_col)) {
    if (!signal_col %in% names_dat) {
      stop("`signal_col` was not found in `data`.", call. = FALSE)
    }
    return(signal_col)
  }

  gpbiometrics_scr_first_existing(
    names_dat,
    c(
      "GSR_US", "gsr_us",
      "GSR_US_TONIC", "gsr_us_tonic",
      "GSR", "gsr"
    )
  )
}

gpbiometrics_scr_resolve_group_cols <- function(names_dat, group_cols) {
  if (!is.null(group_cols)) {
    return(unique(group_cols))
  }

  candidates <- c(
    "source_file",
    "source_participant",
    "USER",
    "USER_FILE",
    "participant",
    "subject",
    "subject_id",
    "MEDIA_ID",
    "MEDIA_NAME",
    "media_id",
    "media_name",
    "trial",
    "trial_id",
    "trial_global"
  )

  unique(candidates[candidates %in% names_dat])
}

gpbiometrics_scr_group_id <- function(dat, group_cols) {
  if (length(group_cols) == 0) {
    return(rep("all", nrow(dat)))
  }

  group_dat <- dat[group_cols]

  group_dat[] <- lapply(group_dat, function(x) {
    x_chr <- as.character(x)
    x_chr[is.na(x_chr)] <- "<NA>"
    x_chr
  })

  apply(group_dat, 1, paste, collapse = "||")
}

gpbiometrics_scr_smooth_signal <- function(x, smooth_width) {
  x <- suppressWarnings(as.numeric(x))

  if (smooth_width <= 1 || length(x) < smooth_width) {
    return(x)
  }

  if (smooth_width %% 2 == 0) {
    smooth_width <- smooth_width + 1L
  }

  filt <- rep(1 / smooth_width, smooth_width)
  smoothed <- as.numeric(stats::filter(x, filt, sides = 2))

  replace <- is.finite(smoothed)
  x[replace] <- smoothed[replace]

  x
}

gpbiometrics_scr_candidate_peaks <- function(x) {
  n <- length(x)

  if (n < 3) {
    return(integer())
  }

  out <- integer()

  for (i in 2:(n - 1)) {
    if (!is.finite(x[i - 1]) || !is.finite(x[i]) || !is.finite(x[i + 1])) {
      next
    }

    if (x[i] > x[i - 1] && x[i] >= x[i + 1]) {
      out <- c(out, i)
    }
  }

  out
}

gpbiometrics_scr_score_peak <- function(x,
                                        t,
                                        peak_local,
                                        amplitude_min,
                                        recovery_fraction) {
  onset <- peak_local

  while (onset > 1 &&
         is.finite(x[onset - 1]) &&
         is.finite(x[onset]) &&
         x[onset - 1] <= x[onset]) {
    onset <- onset - 1L
  }

  onset_value <- x[onset]
  peak_value <- x[peak_local]

  amplitude <- peak_value - onset_value

  if (!is.finite(amplitude) || amplitude < amplitude_min) {
    return(NULL)
  }

  recovery_target <- peak_value - recovery_fraction * amplitude

  recovery <- NA_integer_

  if (peak_local < length(x)) {
    after_peak <- (peak_local + 1L):length(x)
    recovery_candidates <- after_peak[
      is.finite(x[after_peak]) & x[after_peak] <= recovery_target
    ]

    if (length(recovery_candidates) > 0) {
      recovery <- recovery_candidates[1]
    }
  }

  onset_time <- t[onset]
  peak_time <- t[peak_local]
  recovery_time <- if (is.na(recovery)) NA_real_ else t[recovery]

  rise_time <- if (is.finite(onset_time) && is.finite(peak_time)) {
    peak_time - onset_time
  } else {
    NA_real_
  }

  recovery_time_after_peak <- if (is.finite(peak_time) && is.finite(recovery_time)) {
    recovery_time - peak_time
  } else {
    NA_real_
  }

  list(
    onset_index = onset,
    peak_index = peak_local,
    recovery_index = recovery,
    onset_time = onset_time,
    peak_time = peak_time,
    recovery_time = recovery_time,
    onset_value = onset_value,
    peak_value = peak_value,
    recovery_value = if (is.na(recovery)) NA_real_ else x[recovery],
    amplitude = amplitude,
    rise_time = rise_time,
    recovery_time_after_peak = recovery_time_after_peak,
    status = if (is.na(recovery)) {
      "detected_incomplete_recovery"
    } else {
      "detected"
    }
  )
}

gpbiometrics_scr_empty_peaks <- function(group_cols) {
  out <- data.frame(stringsAsFactors = FALSE)

  if (length(group_cols) > 0) {
    for (col in group_cols) {
      out[[col]] <- character()
    }
  }

  out$group_id <- character()
  out$peak_id <- integer()
  out$source_signal <- character()
  out$onset_row_id <- integer()
  out$peak_row_id <- integer()
  out$recovery_row_id <- integer()
  out$onset_index <- integer()
  out$peak_index <- integer()
  out$recovery_index <- integer()
  out$onset_time <- numeric()
  out$peak_time <- numeric()
  out$recovery_time <- numeric()
  out$onset_value <- numeric()
  out$peak_value <- numeric()
  out$recovery_value <- numeric()
  out$amplitude <- numeric()
  out$rise_time <- numeric()
  out$recovery_time_after_peak <- numeric()
  out$status <- character()

  out
}

gpbiometrics_scr_filter_candidate_peaks <- function(x,
                                                    candidate_peak_index,
                                                    min_peak_distance) {
  if (length(candidate_peak_index) <= 1 || min_peak_distance <= 1) {
    return(candidate_peak_index)
  }

  peak_values <- x[candidate_peak_index]
  order_by_height <- order(peak_values, decreasing = TRUE, na.last = NA)

  selected <- integer()

  for (candidate in candidate_peak_index[order_by_height]) {
    if (length(selected) == 0) {
      selected <- candidate
      next
    }

    if (all(abs(candidate - selected) >= min_peak_distance)) {
      selected <- c(selected, candidate)
    }
  }

  sort(selected)
}

test_that("detect_gazepoint_scr_peaks detects a simple SCR-like peak", {
  dat <- data.frame(
    source_file = "simple.csv",
    CNT = seq_len(21),
    GSR_US_PHASIC = c(
      rep(0, 5),
      0.02, 0.05, 0.10, 0.06, 0.02, 0,
      rep(0, 10)
    )
  )

  res <- detect_gazepoint_scr_peaks(
    dat,
    phasic_col = "GSR_US_PHASIC",
    time_col = "CNT",
    group_cols = "source_file",
    amplitude_min = 0.03
  )

  expect_s3_class(res, "gazepoint_scr_peak_detection")
  expect_equal(res$overview$status, "peaks_detected")
  expect_equal(res$overview$detected_peaks, 1)
  expect_equal(res$peaks$amplitude, 0.10)
  expect_equal(res$peaks$status, "detected")
})

test_that("detect_gazepoint_scr_peaks prefers GSR_US_PHASIC when available", {
  dat <- data.frame(
    CNT = seq_len(12),
    GSR_US = c(1, 1.01, 1.02, 1.03, 1.04, 1.05, 1.04, 1.03, 1.02, 1.01, 1, 1),
    GSR_US_PHASIC = c(0, 0, 0.01, 0.04, 0.08, 0.03, 0, 0, 0, 0, 0, 0)
  )

  res <- detect_gazepoint_scr_peaks(
    dat,
    time_col = "CNT",
    amplitude_min = 0.02
  )

  expect_equal(res$overview$source_signal, "GSR_US_PHASIC")
  expect_equal(res$settings$source_signal, "GSR_US_PHASIC")
  expect_equal(res$overview$detected_peaks, 1)
})

test_that("detect_gazepoint_scr_peaks falls back to GSR_US when phasic is absent", {
  dat <- data.frame(
    CNT = seq_len(12),
    GSR_US = c(1, 1.01, 1.02, 1.08, 1.02, 1.00, 1, 1, 1, 1, 1, 1)
  )

  res <- detect_gazepoint_scr_peaks(
    dat,
    time_col = "CNT",
    amplitude_min = 0.03
  )

  expect_equal(res$overview$source_signal, "GSR_US")
  expect_equal(res$overview$detected_peaks, 1)
})

test_that("detect_gazepoint_scr_peaks ignores below-threshold peaks", {
  dat <- data.frame(
    CNT = seq_len(12),
    GSR_US_PHASIC = c(0, 0, 0.005, 0.01, 0.005, 0, 0, 0, 0, 0, 0, 0)
  )

  res <- detect_gazepoint_scr_peaks(
    dat,
    phasic_col = "GSR_US_PHASIC",
    time_col = "CNT",
    amplitude_min = 0.05
  )

  expect_equal(res$overview$status, "candidate_peaks_below_threshold")
  expect_equal(res$overview$detected_peaks, 0)
  expect_equal(nrow(res$peaks), 0)
})

test_that("detect_gazepoint_scr_peaks handles grouped data independently", {
  dat <- data.frame(
    source_file = rep(c("a.csv", "b.csv"), each = 12),
    CNT = rep(seq_len(12), 2),
    GSR_US_PHASIC = c(
      c(0, 0, 0.01, 0.07, 0.02, 0, 0, 0, 0, 0, 0, 0),
      c(0, 0, 0.02, 0.09, 0.03, 0, 0, 0, 0, 0, 0, 0)
    )
  )

  res <- detect_gazepoint_scr_peaks(
    dat,
    phasic_col = "GSR_US_PHASIC",
    time_col = "CNT",
    group_cols = "source_file",
    amplitude_min = 0.03
  )

  expect_equal(res$overview$detected_peaks, 2)
  expect_equal(nrow(res$group_summary), 2)
  expect_true(all(res$group_summary$detected_peaks == 1))
})

test_that("detect_gazepoint_scr_peaks reports incomplete recovery", {
  dat <- data.frame(
    CNT = seq_len(8),
    GSR_US_PHASIC = c(0, 0.02, 0.05, 0.10, 0.09, 0.08, 0.07, 0.06)
  )

  res <- detect_gazepoint_scr_peaks(
    dat,
    phasic_col = "GSR_US_PHASIC",
    time_col = "CNT",
    amplitude_min = 0.03
  )

  expect_equal(res$overview$detected_peaks, 1)
  expect_equal(res$peaks$status, "detected_incomplete_recovery")
  expect_true(is.na(res$peaks$recovery_time_after_peak))
})

test_that("detect_gazepoint_scr_peaks supports negative phasic baselines", {
  dat <- data.frame(
    CNT = seq_len(12),
    GSR_US_PHASIC = c(-0.03, -0.02, 0, 0.05, 0.10, 0.03, -0.01, -0.02, -0.02, -0.02, -0.02, -0.02)
  )

  res <- detect_gazepoint_scr_peaks(
    dat,
    phasic_col = "GSR_US_PHASIC",
    time_col = "CNT",
    amplitude_min = 0.05
  )

  expect_equal(res$overview$detected_peaks, 1)
  expect_true(res$peaks$amplitude >= 0.10)
})

test_that("detect_gazepoint_scr_peaks can enforce a minimum distance between peaks", {
  dat <- data.frame(
    CNT = seq_len(15),
    GSR_US_PHASIC = c(
      0, 0, 0.01, 0.08, 0.02,
      0.04, 0.12, 0.03, 0, 0,
      0.01, 0.09, 0.02, 0, 0
    )
  )

  res_no_distance <- detect_gazepoint_scr_peaks(
    dat,
    phasic_col = "GSR_US_PHASIC",
    time_col = "CNT",
    amplitude_min = 0.03,
    min_peak_distance = 1
  )

  res_with_distance <- detect_gazepoint_scr_peaks(
    dat,
    phasic_col = "GSR_US_PHASIC",
    time_col = "CNT",
    amplitude_min = 0.03,
    min_peak_distance = 5
  )

  expect_true(res_no_distance$overview$detected_peaks > res_with_distance$overview$detected_peaks)
  expect_equal(res_with_distance$overview$detected_peaks, 2)
  expect_true(all(diff(res_with_distance$peaks$peak_index) >= 5))
})

test_that("detect_gazepoint_scr_peaks errors for nonnumeric signal", {
  dat <- data.frame(
    CNT = seq_len(5),
    GSR_US_PHASIC = letters[1:5]
  )

  expect_error(
    detect_gazepoint_scr_peaks(dat, phasic_col = "GSR_US_PHASIC"),
    "numeric or numeric-coercible"
  )
})

test_that("summarise_gazepoint_scr_event_windows links peaks to supplied events", {
  events <- data.frame(
    participant = "P1",
    event_time = 10,
    condition = "stimulus"
  )

  peaks <- data.frame(
    participant = "P1",
    peak_id = 1,
    peak_time = 12,
    onset_time = 11,
    amplitude = 0.05,
    rise_time = 1,
    recovery_time_after_peak = 3,
    status = "detected"
  )

  res <- summarise_gazepoint_scr_event_windows(
    scr_peaks = peaks,
    events = events,
    event_time_col = "event_time",
    event_label_col = "condition",
    group_cols = "participant",
    analysis_window = c(0, 6),
    response_window = c(1, 4)
  )

  expect_s3_class(res, "gazepoint_scr_event_window_summary")
  expect_equal(res$overview$status, "scr_event_windows_summarised")
  expect_equal(res$event_table$response_flag, 1)
  expect_equal(res$event_table$scr_amplitude, 0.05)
  expect_equal(res$event_table$scr_latency, 2)
  expect_equal(res$event_table$event_label, "stimulus")
})

test_that("summarise_gazepoint_scr_event_windows returns no response when peak is outside response window", {
  events <- data.frame(
    participant = "P1",
    event_time = 10
  )

  peaks <- data.frame(
    participant = "P1",
    peak_id = 1,
    peak_time = 15,
    onset_time = 14,
    amplitude = 0.05,
    rise_time = 1,
    recovery_time_after_peak = 3,
    status = "detected"
  )

  res <- summarise_gazepoint_scr_event_windows(
    scr_peaks = peaks,
    events = events,
    event_time_col = "event_time",
    group_cols = "participant",
    analysis_window = c(0, 6),
    response_window = c(1, 4)
  )

  expect_equal(res$overview$status, "warn_no_scr_responses")
  expect_equal(res$event_table$response_flag, 0)
  expect_equal(res$event_table$n_candidate_peaks, 1)
  expect_equal(res$event_table$n_response_window_peaks, 0)
  expect_equal(res$event_table$event_status, "no_peaks_in_response_window")
})

test_that("summarise_gazepoint_scr_event_windows selects largest amplitude by default", {
  events <- data.frame(
    participant = "P1",
    event_time = 10
  )

  peaks <- data.frame(
    participant = "P1",
    peak_id = c(1, 2),
    peak_time = c(12, 13),
    onset_time = c(11, 12),
    amplitude = c(0.03, 0.08),
    rise_time = c(1, 1),
    recovery_time_after_peak = c(2, 2),
    status = c("detected", "detected")
  )

  res <- summarise_gazepoint_scr_event_windows(
    scr_peaks = peaks,
    events = events,
    event_time_col = "event_time",
    group_cols = "participant",
    analysis_window = c(0, 6),
    response_window = c(1, 4)
  )

  expect_equal(res$event_table$selected_peak_id, "2")
  expect_equal(res$event_table$scr_amplitude, 0.08)
})

test_that("summarise_gazepoint_scr_event_windows can select first peak", {
  events <- data.frame(
    participant = "P1",
    event_time = 10
  )

  peaks <- data.frame(
    participant = "P1",
    peak_id = c(1, 2),
    peak_time = c(12, 13),
    onset_time = c(11, 12),
    amplitude = c(0.03, 0.08),
    rise_time = c(1, 1),
    recovery_time_after_peak = c(2, 2),
    status = c("detected", "detected")
  )

  res <- summarise_gazepoint_scr_event_windows(
    scr_peaks = peaks,
    events = events,
    event_time_col = "event_time",
    group_cols = "participant",
    analysis_window = c(0, 6),
    response_window = c(1, 4),
    peak_selection = "first_peak"
  )

  expect_equal(res$event_table$selected_peak_id, "1")
  expect_equal(res$event_table$scr_amplitude, 0.03)
})

test_that("summarise_gazepoint_scr_event_windows respects group matching", {
  events <- data.frame(
    participant = c("P1", "P2"),
    event_time = c(10, 10)
  )

  peaks <- data.frame(
    participant = c("P1", "P2"),
    peak_id = c(1, 1),
    peak_time = c(12, 20),
    onset_time = c(11, 19),
    amplitude = c(0.05, 0.10),
    rise_time = c(1, 1),
    recovery_time_after_peak = c(2, 2),
    status = c("detected", "detected")
  )

  res <- summarise_gazepoint_scr_event_windows(
    scr_peaks = peaks,
    events = events,
    event_time_col = "event_time",
    group_cols = "participant",
    analysis_window = c(0, 6),
    response_window = c(1, 4)
  )

  expect_equal(res$overview$event_count, 2)
  expect_equal(res$overview$response_events, 1)
  expect_equal(res$event_table$response_flag[res$event_table$participant == "P1"], 1)
  expect_equal(res$event_table$response_flag[res$event_table$participant == "P2"], 0)
})

test_that("summarise_gazepoint_scr_event_windows accepts peak-detection objects", {
  dat <- data.frame(
    participant = "P1",
    CNT = seq_len(20),
    GSR_US_PHASIC = c(rep(0, 10), 0.02, 0.08, 0.02, rep(0, 7))
  )

  peaks <- detect_gazepoint_scr_peaks(
    dat,
    phasic_col = "GSR_US_PHASIC",
    time_col = "CNT",
    group_cols = "participant",
    amplitude_min = 0.03
  )

  events <- data.frame(
    participant = "P1",
    event_time = 10
  )

  res <- summarise_gazepoint_scr_event_windows(
    scr_peaks = peaks,
    events = events,
    event_time_col = "event_time",
    group_cols = "participant",
    analysis_window = c(0, 6),
    response_window = c(1, 4)
  )

  expect_equal(res$overview$response_events, 1)
  expect_true(res$event_table$scr_amplitude > 0)
})

test_that("summarise_gazepoint_scr_event_windows derives rising TTL events", {
  dat <- data.frame(
    participant = "P1",
    CNT = seq_len(10),
    TTL1 = c(0, 1, 1, 0, 0, 1, 0, 0, 0, 0)
  )

  peaks <- data.frame(
    participant = "P1",
    peak_id = c(1, 2),
    peak_time = c(4, 8),
    onset_time = c(3, 7),
    amplitude = c(0.05, 0.07),
    rise_time = c(1, 1),
    recovery_time_after_peak = c(2, 2),
    status = c("detected", "detected")
  )

  res <- summarise_gazepoint_scr_event_windows(
    data = dat,
    scr_peaks = peaks,
    time_col = "CNT",
    group_cols = "participant",
    ttl_cols = "TTL1",
    analysis_window = c(0, 4),
    response_window = c(1, 3)
  )

  expect_equal(res$overview$event_count, 2)
  expect_equal(res$overview$response_events, 2)
  expect_equal(res$events$event_time, c(2, 6))
})

test_that("summarise_gazepoint_scr_event_windows handles no TTL events", {
  dat <- data.frame(
    participant = "P1",
    CNT = seq_len(10),
    TTL1 = rep(0, 10)
  )

  peaks <- data.frame(
    participant = "P1",
    peak_id = 1,
    peak_time = 4,
    onset_time = 3,
    amplitude = 0.05,
    rise_time = 1,
    recovery_time_after_peak = 2,
    status = "detected"
  )

  res <- summarise_gazepoint_scr_event_windows(
    data = dat,
    scr_peaks = peaks,
    time_col = "CNT",
    group_cols = "participant",
    ttl_cols = "TTL1",
    analysis_window = c(0, 4),
    response_window = c(1, 3)
  )

  expect_equal(res$overview$status, "fail_no_events")
  expect_equal(nrow(res$event_table), 0)
})

test_that("summarise_gazepoint_scr_event_windows can collapse simultaneous TTL events", {
  dat <- data.frame(
    participant = "P1",
    CNT = seq_len(10),
    TTL0 = c(0, 1, 1, 0, 0, 1, 0, 0, 0, 0),
    TTL1 = c(0, 1, 1, 0, 0, 1, 0, 0, 0, 0)
  )

  peaks <- data.frame(
    participant = "P1",
    peak_id = c(1, 2),
    peak_time = c(4, 8),
    onset_time = c(3, 7),
    amplitude = c(0.05, 0.07),
    rise_time = c(1, 1),
    recovery_time_after_peak = c(2, 2),
    status = c("detected", "detected")
  )

  res_uncollapsed <- summarise_gazepoint_scr_event_windows(
    data = dat,
    scr_peaks = peaks,
    time_col = "CNT",
    group_cols = "participant",
    ttl_cols = c("TTL0", "TTL1"),
    analysis_window = c(0, 4),
    response_window = c(1, 3),
    collapse_simultaneous_events = FALSE
  )

  res_collapsed <- summarise_gazepoint_scr_event_windows(
    data = dat,
    scr_peaks = peaks,
    time_col = "CNT",
    group_cols = "participant",
    ttl_cols = c("TTL0", "TTL1"),
    analysis_window = c(0, 4),
    response_window = c(1, 3),
    collapse_simultaneous_events = TRUE
  )

  expect_equal(res_uncollapsed$overview$event_count, 4)
  expect_equal(res_collapsed$overview$event_count, 2)
  expect_equal(res_collapsed$overview$response_events, 2)
  expect_true(all(res_collapsed$events$collapsed_event_count == 2))
  expect_true(all(res_collapsed$events$event_label == "TTL0+TTL1"))
})

test_that("summarise_gazepoint_scr_event_windows validates nested windows", {
  events <- data.frame(
    event_time = 10
  )

  peaks <- data.frame(
    peak_id = 1,
    peak_time = 12,
    onset_time = 11,
    amplitude = 0.05,
    rise_time = 1,
    recovery_time_after_peak = 2,
    status = "detected"
  )

  expect_error(
    summarise_gazepoint_scr_event_windows(
      scr_peaks = peaks,
      events = events,
      event_time_col = "event_time",
      analysis_window = c(0, 4),
      response_window = c(-1, 5)
    ),
    "`response_window` must fall inside `analysis_window`"
  )
})

test_that("summarise_gazepoint_scr_event_windows validates collapse_simultaneous_events", {
  events <- data.frame(
    event_time = 10
  )

  peaks <- data.frame(
    peak_id = 1,
    peak_time = 12,
    amplitude = 0.05
  )

  expect_error(
    summarise_gazepoint_scr_event_windows(
      scr_peaks = peaks,
      events = events,
      event_time_col = "event_time",
      collapse_simultaneous_events = NA
    ),
    "`collapse_simultaneous_events` must be TRUE or FALSE"
  )
})

#' Prepare EDA artifact-classifier segment features
#'
#' Creates segment-level features that can be passed to a user-supplied
#' artifact classifier such as an SVM. No pretrained classifier is bundled.
#'
#' @param dat A data frame.
#' @param eda_col EDA/conductance column.
#' @param time_col Optional time column.
#' @param group_cols Optional grouping columns.
#' @param segment_seconds Segment length in seconds when `time_col` is supplied.
#' @param samples_per_segment Segment length in samples when no usable time
#'   column or sampling rate is available.
#' @param sampling_rate Optional sampling rate in Hz.
#'
#' @return A segment-level feature data frame.
#' @export
prepare_gazepoint_artifact_svm_features <- function(dat,
                                                    eda_col = "GSR_US",
                                                    time_col = NULL,
                                                    group_cols = NULL,
                                                    segment_seconds = 5,
                                                    samples_per_segment = NULL,
                                                    sampling_rate = NULL) {
  if (!is.data.frame(dat)) {
    stop("`dat` must be a data frame.", call. = FALSE)
  }

  if (!eda_col %in% names(dat)) {
    stop("Column `", eda_col, "` was not found in `dat`.", call. = FALSE)
  }

  if (!is.numeric(dat[[eda_col]])) {
    stop("`eda_col` must identify a numeric column.", call. = FALSE)
  }

  if (!is.null(time_col) && !time_col %in% names(dat)) {
    stop("Column `", time_col, "` was not found in `dat`.", call. = FALSE)
  }

  if (is.null(group_cols)) {
    group_cols <- character()
  }

  missing_groups <- setdiff(group_cols, names(dat))
  if (length(missing_groups) > 0) {
    stop("Missing `group_cols`: ", paste(missing_groups, collapse = ", "), call. = FALSE)
  }

  groups <- gpbiometrics_svm_split(dat, group_cols)

  rows <- list()
  row_id <- 1L

  for (group_id in names(groups)) {
    idx <- groups[[group_id]]

    if (!is.null(time_col)) {
      idx <- idx[order(dat[[time_col]][idx])]
    }

    x <- dat[[eda_col]][idx]

    fs <- if (!is.null(time_col)) {
      gpbiometrics_svm_sampling_rate(dat[[time_col]][idx], sampling_rate)
    } else {
      sampling_rate
    }

    n_per_segment <- if (!is.null(samples_per_segment)) {
      samples_per_segment
    } else if (is.finite(fs) && fs > 0) {
      max(2, round(segment_seconds * fs))
    } else {
      5
    }

    starts <- seq(1, length(idx), by = n_per_segment)

    for (seg_i in seq_along(starts)) {
      seg_idx_local <- starts[seg_i]:min(starts[seg_i] + n_per_segment - 1, length(idx))
      seg_idx <- idx[seg_idx_local]
      seg_x <- dat[[eda_col]][seg_idx]
      finite_x <- seg_x[is.finite(seg_x)]

      if (length(finite_x) < 2) {
        status <- "insufficient_segment_data"
        feat <- rep(NA_real_, 10)
        names(feat) <- c(
          "mean_signal", "sd_signal", "min_signal", "max_signal", "range_signal",
          "median_abs_diff", "max_abs_diff", "slope", "zero_crossing_diff",
          "detail_energy"
        )
      } else {
        d <- diff(finite_x)
        tt <- seq_along(finite_x)
        slope <- tryCatch(stats::coef(stats::lm(finite_x ~ tt))[2], error = function(e) NA_real_)
        detail <- gpbiometrics_svm_detail_energy(finite_x)

        feat <- c(
          mean_signal = mean(finite_x),
          sd_signal = stats::sd(finite_x),
          min_signal = min(finite_x),
          max_signal = max(finite_x),
          range_signal = diff(range(finite_x)),
          median_abs_diff = stats::median(abs(d)),
          max_abs_diff = max(abs(d)),
          slope = unname(slope),
          zero_crossing_diff = sum(diff(sign(d - mean(d))) != 0, na.rm = TRUE),
          detail_energy = detail
        )
        status <- "svm_features_prepared"
      }

      time_start <- if (!is.null(time_col)) dat[[time_col]][seg_idx[1]] else NA_real_
      time_end <- if (!is.null(time_col)) dat[[time_col]][seg_idx[length(seg_idx)]] else NA_real_

      rows[[row_id]] <- data.frame(
        group_id = group_id,
        segment_id = paste0(group_id, "_segment_", seg_i),
        segment_index = seg_i,
        start_row = min(seg_idx),
        end_row = max(seg_idx),
        start_time = time_start,
        end_time = time_end,
        n_samples = length(seg_idx),
        n_finite = length(finite_x),
        as.data.frame(as.list(feat)),
        status = status,
        stringsAsFactors = FALSE,
        check.names = FALSE
      )

      row_id <- row_id + 1L
    }
  }

  features <- do.call(rbind, rows)
  rownames(features) <- NULL

  attr(features, "svm_feature_settings") <- list(
    eda_col = eda_col,
    time_col = time_col,
    group_cols = group_cols,
    segment_seconds = segment_seconds,
    samples_per_segment = samples_per_segment,
    sampling_rate = sampling_rate
  )

  class(features) <- unique(c("gazepoint_artifact_svm_features", class(features)))
  features
}

#' Flag EDA artifacts with a user-supplied SVM-style model
#'
#' Applies a user-supplied model or prediction function to segment-level
#' artifact features. No pretrained model is bundled, so this function avoids
#' pretending to reproduce any proprietary or externally trained classifier.
#'
#' @param x Either raw EDA data or output from
#'   `prepare_gazepoint_artifact_svm_features()`.
#' @param model Optional model object or function. If `NULL`, features are
#'   returned with missing artifact predictions.
#' @param feature_cols Optional feature columns used by the model.
#' @param probability_threshold Threshold for artifact probability.
#' @param ... Passed to `prepare_gazepoint_artifact_svm_features()` when `x`
#'   is raw data.
#'
#' @return A data frame with artifact probabilities/classes where available.
#' @export
flag_gazepoint_artifacts_svm <- function(x,
                                         model = NULL,
                                         feature_cols = NULL,
                                         probability_threshold = 0.5,
                                         ...) {
  if (!is.data.frame(x)) {
    stop("`x` must be a data frame.", call. = FALSE)
  }

  if (inherits(x, "gazepoint_artifact_svm_features")) {
    features <- x
  } else {
    features <- prepare_gazepoint_artifact_svm_features(x, ...)
  }

  default_features <- c(
    "mean_signal", "sd_signal", "min_signal", "max_signal", "range_signal",
    "median_abs_diff", "max_abs_diff", "slope", "zero_crossing_diff",
    "detail_energy"
  )

  if (is.null(feature_cols)) {
    feature_cols <- intersect(default_features, names(features))
  }

  missing_features <- setdiff(feature_cols, names(features))
  if (length(missing_features) > 0) {
    stop("Missing `feature_cols`: ", paste(missing_features, collapse = ", "), call. = FALSE)
  }

  out <- features
  out$artifact_probability <- NA_real_
  out$artifact_svm <- NA
  out$artifact_svm_status <- "no_model_supplied"

  if (is.null(model)) {
    attr(out, "svm_artifact_overview") <- data.frame(
      segment_rows = nrow(out),
      model_supplied = FALSE,
      flagged_segments = NA_integer_,
      status = "svm_features_prepared_no_model_supplied",
      interpretation = paste(
        "SVM artifact features were prepared but no classifier was supplied.",
        "Use a validated user-supplied model to create artifact labels."
      ),
      stringsAsFactors = FALSE
    )

    class(out) <- unique(c("gazepoint_artifact_svm_flags", class(out)))
    return(out)
  }

  newdata <- out[feature_cols]
  pred <- gpbiometrics_svm_predict(model, newdata)

  if (is.numeric(pred)) {
    prob <- as.numeric(pred)
    out$artifact_probability <- prob
    out$artifact_svm <- prob >= probability_threshold
    out$artifact_svm_status <- "predicted_from_numeric_probability"
  } else {
    pred_chr <- as.character(pred)
    artifact_label <- tolower(pred_chr) %in% c("artifact", "art", "1", "true", "bad")
    out$artifact_svm <- artifact_label
    out$artifact_svm_status <- "predicted_from_class_label"
  }

  attr(out, "svm_artifact_overview") <- data.frame(
    segment_rows = nrow(out),
    model_supplied = TRUE,
    flagged_segments = sum(out$artifact_svm %in% TRUE, na.rm = TRUE),
    status = "svm_artifact_flags_created",
    interpretation = paste(
      "Artifact labels come from the supplied model.",
      "Validity depends on the model training data, features, scaling, and target device context."
    ),
    stringsAsFactors = FALSE
  )

  class(out) <- unique(c("gazepoint_artifact_svm_flags", class(out)))
  out
}

gpbiometrics_svm_split <- function(dat, group_cols) {
  if (length(group_cols) == 0) {
    return(list(all_rows = seq_len(nrow(dat))))
  }

  gf <- dat[group_cols]
  gf[] <- lapply(gf, function(x) {
    x <- as.character(x)
    x[is.na(x)] <- "<NA>"
    x
  })

  split(seq_len(nrow(dat)), do.call(paste, c(gf, sep = " | ")))
}

gpbiometrics_svm_sampling_rate <- function(time_values, sampling_rate = NULL) {
  if (!is.null(sampling_rate)) {
    return(sampling_rate)
  }

  time_values <- time_values[is.finite(time_values)]

  if (length(time_values) < 3) {
    return(NA_real_)
  }

  dt <- diff(time_values)
  dt <- dt[is.finite(dt) & dt > 0]

  if (length(dt) == 0) {
    return(NA_real_)
  }

  median_dt <- stats::median(dt)

  if (median_dt > 10) {
    1000 / median_dt
  } else {
    1 / median_dt
  }
}

gpbiometrics_svm_detail_energy <- function(x) {
  if (length(x) < 2) {
    return(NA_real_)
  }

  if (length(x) %% 2 == 1) {
    x <- x[-length(x)]
  }

  detail <- (x[seq(1, length(x), by = 2)] - x[seq(2, length(x), by = 2)]) / sqrt(2)
  mean(detail^2, na.rm = TRUE)
}

gpbiometrics_svm_predict <- function(model, newdata) {
  if (is.function(model)) {
    return(model(newdata))
  }

  pred <- tryCatch(
    stats::predict(model, newdata = newdata, type = "response"),
    error = function(e) NULL
  )

  if (is.null(pred)) {
    pred <- tryCatch(
      stats::predict(model, newdata = newdata),
      error = function(e) {
        stop("Could not obtain predictions from `model`: ", conditionMessage(e), call. = FALSE)
      }
    )
  }

  pred
}

test_that("audit_gazepoint_gsr_units flags conductance-like and resistance-like signals", {
  conductance <- data.frame(GSR_US = c(0.8, 1.2, 2.5, 4.0))
  out_c <- audit_gazepoint_gsr_units(conductance, gsr_col = "GSR_US")

  expect_s3_class(out_c, "gazepoint_gsr_unit_audit")
  expect_equal(out_c$overview$likely_unit, "conductance_microSiemens")

  resistance <- data.frame(GSR = c(500000, 1000000, 1500000, 2000000))
  out_r <- audit_gazepoint_gsr_units(resistance, gsr_col = "GSR", convert = TRUE)

  expect_equal(out_r$overview$likely_unit, "resistance_or_impedance_ohms")
  expect_true("GSR_converted_us" %in% names(out_r$data))
  expect_true(all(is.finite(out_r$data$GSR_converted_us)))
})

test_that("standardise_gazepoint_adaptive_ema adds normalized local signal columns", {
  dat <- data.frame(
    participant = rep(c("p1", "p2"), each = 30),
    time = rep(seq_len(30), 2),
    GSR_US = c(seq(1, 2, length.out = 30), seq(2, 4, length.out = 30))
  )

  out <- standardise_gazepoint_adaptive_ema(
    dat,
    signal_col = "GSR_US",
    group_cols = "participant",
    time_col = "time",
    alpha = 0.2
  )

  expect_s3_class(out, "gazepoint_adaptive_ema_normalised")
  expect_true("GSR_US_adaptive_ema" %in% names(out))
  expect_true("GSR_US_ema_center" %in% names(out))
  expect_equal(attr(out, "adaptive_ema_overview")$status, "adaptive_ema_normalization_complete")
})

test_that("run_gazepoint_scr_multiverse scores multiple specifications", {
  time <- rep(seq(-1, 5, by = 0.5), 4)
  trial <- rep(rep(1:2, each = length(seq(-1, 5, by = 0.5))), 2)
  participant <- rep(c("p1", "p2"), each = 2 * length(seq(-1, 5, by = 0.5)))
  condition <- rep(c("control", "treatment", "control", "treatment"), each = length(seq(-1, 5, by = 0.5)))

  signal <- 1 + ifelse(time >= 1 & time <= 3, 0.08, 0) + rnorm(length(time), sd = 0.005)

  dat <- data.frame(
    participant = participant,
    trial = paste(participant, trial, sep = "_"),
    condition = condition,
    time = time,
    GSR_US = signal
  )

  out <- run_gazepoint_scr_multiverse(
    dat,
    signal_col = "GSR_US",
    time_col = "time",
    trial_cols = c("participant", "trial"),
    condition_col = "condition",
    latency_windows = list(c(1, 3), c(1, 4)),
    thresholds = c(0.01, 0.05),
    baseline_methods = c("median", "none")
  )

  expect_s3_class(out, "gazepoint_scr_multiverse")
  expect_equal(out$overview$specification_count, 8)
  expect_true(nrow(out$scored_trials) > 0)
  expect_true(all(c("response_amplitude", "response_present") %in% names(out$scored_trials)))
})

test_that("prepare_gazepoint_artifact_svm_features and SVM bridge work with supplied function", {
  dat <- data.frame(
    participant = "p1",
    time = seq(0, 19),
    GSR_US = c(rep(1, 10), rep(10, 10))
  )

  features <- prepare_gazepoint_artifact_svm_features(
    dat,
    eda_col = "GSR_US",
    time_col = "time",
    group_cols = "participant",
    segment_seconds = 5,
    sampling_rate = 1
  )

  expect_s3_class(features, "gazepoint_artifact_svm_features")
  expect_true("detail_energy" %in% names(features))

  model_fun <- function(newdata) {
    as.numeric(newdata$mean_signal > 5)
  }

  flags <- flag_gazepoint_artifacts_svm(features, model = model_fun)

  expect_s3_class(flags, "gazepoint_artifact_svm_flags")
  expect_true("artifact_svm" %in% names(flags))
  expect_true(any(flags$artifact_svm %in% TRUE))
})

test_that("flag_gazepoint_artifacts_svm returns features without model", {
  dat <- data.frame(
    time = seq(0, 9),
    GSR_US = seq(1, 2, length.out = 10)
  )

  flags <- flag_gazepoint_artifacts_svm(
    dat,
    eda_col = "GSR_US",
    time_col = "time",
    segment_seconds = 5,
    sampling_rate = 1
  )

  expect_s3_class(flags, "gazepoint_artifact_svm_flags")
  expect_equal(attr(flags, "svm_artifact_overview")$status, "svm_features_prepared_no_model_supplied")
})

test_that("import_gazepoint_lsl_xdf fails clearly for missing files", {
  expect_error(
    import_gazepoint_lsl_xdf("missing_file.xdf"),
    "File does not exist"
  )
})

#' Flag MAD-based EDA wearable artifacts
#'
#' Flags dependency-light, subject-specific EDA artifact categories using robust
#' median absolute deviation (MAD) logic. The categories are heuristic QC
#' labels: step artifacts, needle artifacts, flatline artifacts, and wall
#' artifacts.
#'
#' @param dat A data frame.
#' @param eda_col Numeric EDA/conductance column.
#' @param time_col Optional time column for ordering within group.
#' @param group_cols Optional grouping columns.
#' @param mad_multiplier MAD multiplier used for robust thresholding.
#' @param flatline_tolerance Maximum absolute sample-to-sample change treated
#'   as flatline.
#' @param flatline_min_run Minimum consecutive flatline samples.
#' @param wall_abs_change Optional absolute change threshold for wall artifacts.
#' @param output_prefix Prefix for output columns.
#'
#' @return A data frame with artifact flags and artifact-summary attributes.
#' @export
flag_gazepoint_mad_artifacts <- function(dat,
                                         eda_col = "GSR_US",
                                         time_col = NULL,
                                         group_cols = NULL,
                                         mad_multiplier = 8,
                                         flatline_tolerance = 1e-6,
                                         flatline_min_run = 5,
                                         wall_abs_change = NULL,
                                         output_prefix = "mad") {
  if (!is.data.frame(dat)) {
    stop("`dat` must be a data frame.", call. = FALSE)
  }

  if (!eda_col %in% names(dat)) {
    stop("Column `", eda_col, "` was not found in `dat`.", call. = FALSE)
  }

  if (!is.numeric(dat[[eda_col]])) {
    stop("`eda_col` must identify a numeric column.", call. = FALSE)
  }

  if (!is.null(time_col) && !time_col %in% names(dat)) {
    stop("Column `", time_col, "` was not found in `dat`.", call. = FALSE)
  }

  if (is.null(group_cols)) {
    group_cols <- character()
  }

  missing_groups <- setdiff(group_cols, names(dat))
  if (length(missing_groups) > 0) {
    stop("Missing `group_cols`: ", paste(missing_groups, collapse = ", "), call. = FALSE)
  }

  out <- dat

  step_col <- paste0(output_prefix, "_step_artifact")
  needle_col <- paste0(output_prefix, "_needle_artifact")
  flatline_col <- paste0(output_prefix, "_flatline_artifact")
  wall_col <- paste0(output_prefix, "_wall_artifact")
  any_col <- paste0(output_prefix, "_artifact")
  type_col <- paste0(output_prefix, "_artifact_type")
  status_col <- paste0(output_prefix, "_artifact_status")

  out[[step_col]] <- FALSE
  out[[needle_col]] <- FALSE
  out[[flatline_col]] <- FALSE
  out[[wall_col]] <- FALSE
  out[[any_col]] <- FALSE
  out[[type_col]] <- "none"
  out[[status_col]] <- "not_processed"

  groups <- gpbiometrics_mad_split_indices(out, group_cols)

  summary_rows <- lapply(names(groups), function(group_id) {
    idx <- groups[[group_id]]

    if (!is.null(time_col)) {
      idx <- idx[order(out[[time_col]][idx])]
    }

    x <- out[[eda_col]][idx]
    finite_x <- x[is.finite(x)]

    if (length(finite_x) < 5) {
      out[[status_col]][idx] <<- "insufficient_finite_samples"

      return(data.frame(
        group_id = group_id,
        n_rows = length(idx),
        n_finite = length(finite_x),
        robust_mad = NA_real_,
        artifact_rows = NA_integer_,
        status = "insufficient_finite_samples",
        stringsAsFactors = FALSE
      ))
    }

    robust_center <- stats::median(finite_x)
    robust_mad <- stats::median(abs(finite_x - robust_center))

    if (!is.finite(robust_mad) || robust_mad == 0) {
      robust_mad <- stats::mad(finite_x, constant = 1, na.rm = TRUE)
    }

    if (!is.finite(robust_mad) || robust_mad == 0) {
      robust_mad <- .Machine$double.eps
    }

    dx <- c(NA_real_, diff(x))
    abs_dx <- abs(dx)

    wall_threshold <- if (!is.null(wall_abs_change)) {
      wall_abs_change
    } else {
      mad_multiplier * robust_mad
    }

    wall <- is.finite(abs_dx) & abs_dx > wall_threshold

    local_deviation <- abs(x - robust_center)
    needle <- is.finite(local_deviation) &
      local_deviation > mad_multiplier * robust_mad &
      c(FALSE, abs_dx[-1] > mad_multiplier * robust_mad)

    step <- gpbiometrics_mad_step_flags(
      x = x,
      mad = robust_mad,
      mad_multiplier = mad_multiplier
    )

    flatline <- gpbiometrics_mad_flatline_flags(
      x = x,
      tolerance = flatline_tolerance,
      min_run = flatline_min_run
    )

    any_art <- step | needle | flatline | wall
    artifact_type <- rep("none", length(idx))
    artifact_type[wall] <- "wall"
    artifact_type[flatline] <- "flatline"
    artifact_type[needle] <- "needle"
    artifact_type[step] <- "step"
    artifact_type[rowSums(cbind(step, needle, flatline, wall), na.rm = TRUE) > 1] <- "multiple"

    out[[step_col]][idx] <<- step
    out[[needle_col]][idx] <<- needle
    out[[flatline_col]][idx] <<- flatline
    out[[wall_col]][idx] <<- wall
    out[[any_col]][idx] <<- any_art
    out[[type_col]][idx] <<- artifact_type
    out[[status_col]][idx] <<- "mad_artifact_typology_applied"

    data.frame(
      group_id = group_id,
      n_rows = length(idx),
      n_finite = length(finite_x),
      robust_mad = robust_mad,
      step_rows = sum(step, na.rm = TRUE),
      needle_rows = sum(needle, na.rm = TRUE),
      flatline_rows = sum(flatline, na.rm = TRUE),
      wall_rows = sum(wall, na.rm = TRUE),
      artifact_rows = sum(any_art, na.rm = TRUE),
      status = "mad_artifact_typology_applied",
      stringsAsFactors = FALSE
    )
  })

  summary_table <- do.call(rbind, summary_rows)
  rownames(summary_table) <- NULL

  overview <- data.frame(
    input_rows = nrow(dat),
    group_count = length(groups),
    artifact_rows = sum(out[[any_col]], na.rm = TRUE),
    step_rows = sum(out[[step_col]], na.rm = TRUE),
    needle_rows = sum(out[[needle_col]], na.rm = TRUE),
    flatline_rows = sum(out[[flatline_col]], na.rm = TRUE),
    wall_rows = sum(out[[wall_col]], na.rm = TRUE),
    status = "mad_artifact_typology_complete",
    interpretation = paste(
      "MAD artifact categories are robust heuristic QC labels for wearable EDA signals.",
      "They do not infer psychological state and should be inspected alongside raw signals."
    ),
    stringsAsFactors = FALSE
  )

  attr(out, "mad_artifact_overview") <- overview
  attr(out, "mad_artifact_summary") <- summary_table
  attr(out, "mad_artifact_settings") <- list(
    eda_col = eda_col,
    time_col = time_col,
    group_cols = group_cols,
    mad_multiplier = mad_multiplier,
    flatline_tolerance = flatline_tolerance,
    flatline_min_run = flatline_min_run,
    wall_abs_change = wall_abs_change,
    output_prefix = output_prefix
  )

  class(out) <- unique(c("gazepoint_mad_artifact_flags", class(out)))
  out
}

gpbiometrics_mad_split_indices <- function(dat, group_cols) {
  if (length(group_cols) == 0) {
    return(list(all_rows = seq_len(nrow(dat))))
  }

  gf <- dat[group_cols]
  gf[] <- lapply(gf, function(x) {
    x <- as.character(x)
    x[is.na(x)] <- "<NA>"
    x
  })

  split(seq_len(nrow(dat)), do.call(paste, c(gf, sep = " | ")))
}

gpbiometrics_mad_flatline_flags <- function(x,
                                            tolerance = 1e-6,
                                            min_run = 5) {
  flat <- rep(FALSE, length(x))

  if (length(x) < min_run) {
    return(flat)
  }

  small_change <- c(FALSE, abs(diff(x)) <= tolerance)
  small_change[!is.finite(small_change)] <- FALSE

  run <- rle(small_change)
  ends <- cumsum(run$lengths)
  starts <- ends - run$lengths + 1

  for (i in seq_along(run$values)) {
    if (isTRUE(run$values[i]) && run$lengths[i] >= min_run) {
      flat[starts[i]:ends[i]] <- TRUE
    }
  }

  flat
}

gpbiometrics_mad_step_flags <- function(x,
                                        mad,
                                        mad_multiplier = 8) {
  step <- rep(FALSE, length(x))

  if (length(x) < 8 || !is.finite(mad) || mad <= 0) {
    return(step)
  }

  threshold <- mad_multiplier * mad

  for (i in 4:(length(x) - 4)) {
    before <- x[(i - 3):(i - 1)]
    after <- x[i:(i + 3)]

    if (sum(is.finite(before)) >= 2 && sum(is.finite(after)) >= 2) {
      shift <- abs(stats::median(after, na.rm = TRUE) - stats::median(before, na.rm = TRUE))

      if (is.finite(shift) && shift > threshold) {
        step[i:(i + 3)] <- TRUE
      }
    }
  }

  step
}

#' Plot Gazepoint EDA decomposition channels
#'
#' Plots available EDA/GSR decomposition channels, typically `GSR_US`,
#' `GSR_US_TONIC`, and `GSR_US_PHASIC`, as a ggplot object.
#'
#' @param data A Gazepoint biometric data frame or list containing a data frame.
#' @param time_col Optional time/counter column.
#' @param signal_cols Optional signal columns to plot.
#' @param group_cols Optional grouping columns used for facets.
#' @param standardise Logical. If `TRUE`, standardise each signal to z-scores.
#' @param max_points Maximum number of rows retained after simple downsampling.
#' @param title Optional plot title.
#'
#' @return A ggplot object with plot data stored in attributes.
#' @export
plot_gazepoint_eda_decomposition <- function(data,
                                             time_col = NULL,
                                             signal_cols = NULL,
                                             group_cols = NULL,
                                             standardise = FALSE,
                                             max_points = 5000,
                                             title = NULL) {
  gpbiometrics_require_ggplot2()

  dat <- gpbiometrics_eda_plot_extract_data(data)

  if (is.null(time_col)) {
    time_col <- gpbiometrics_eda_plot_first_existing(
      names(dat),
      c(
        "CNT", "cnt", "time", "Time", "TIME",
        "timestamp", "timestamp_ms", "time_ms"
      )
    )
  }

  if (is.null(time_col) || !time_col %in% names(dat)) {
    stop("No usable time column was found. Supply `time_col`.", call. = FALSE)
  }

  if (is.null(signal_cols)) {
    signal_cols <- gpbiometrics_eda_plot_signal_cols(names(dat))
  }

  signal_cols <- signal_cols[signal_cols %in% names(dat)]

  if (length(signal_cols) == 0) {
    stop("No EDA/GSR signal columns were found. Supply `signal_cols`.", call. = FALSE)
  }

  if (is.null(group_cols)) {
    group_cols <- intersect(
      c(
        "source_file",
        "source_participant",
        "participant",
        "subject",
        "MEDIA_ID",
        "MEDIA_NAME",
        "trial",
        "trial_id"
      ),
      names(dat)
    )
  }

  plot_data <- gpbiometrics_eda_plot_long_data(
    dat = dat,
    time_col = time_col,
    signal_cols = signal_cols,
    group_cols = group_cols,
    standardise = standardise,
    max_points = max_points
  )

  y_label <- if (isTRUE(standardise)) {
    "Standardised signal value"
  } else {
    "Signal value"
  }

  p <- ggplot2::ggplot(
    plot_data,
    ggplot2::aes(
      x = .data_time,
      y = .data_value,
      colour = .data_signal
    )
  ) +
    ggplot2::geom_line(na.rm = TRUE) +
    ggplot2::labs(
      title = if (is.null(title)) "Gazepoint EDA decomposition" else title,
      x = time_col,
      y = y_label,
      colour = "Signal"
    ) +
    ggplot2::theme_minimal()

  if (length(unique(plot_data$.data_group)) > 1) {
    p <- p + ggplot2::facet_wrap(
      stats::as.formula("~ .data_group"),
      scales = "free_x"
    )
  }

  settings <- list(
    time_col = time_col,
    signal_cols = signal_cols,
    group_cols = group_cols,
    standardise = standardise,
    max_points = max_points,
    title = title,
    plot_type = "eda_decomposition",
    interpretation_notes = c(
      "The plot is intended for EDA signal inspection and reporting.",
      "EDA/GSR reflects sympathetic arousal-related activity and should not be interpreted as emotional valence."
    )
  )

  standardise_gazepoint_plot_contract(
    plot = p,
    plot_data = plot_data,
    settings = settings,
    interpretation_notes = settings$interpretation_notes,
    plot_type = "eda_decomposition"
  )
}

#' Plot Gazepoint SCR events on an EDA signal
#'
#' Plots an EDA/GSR signal with detected SCR peak markers and optional event
#' onsets from SCR event-window summaries or event tables.
#'
#' @param data Gazepoint biometric data frame.
#' @param scr_peaks A `gazepoint_scr_peak_detection` object or peak data frame.
#' @param event_windows Optional `gazepoint_scr_event_window_summary` object or
#'   event-window data frame.
#' @param events Optional event table used when `event_windows` is not supplied.
#' @param time_col Optional time/counter column.
#' @param signal_col Optional signal column to plot.
#' @param phasic_col Optional preferred phasic signal column.
#' @param group_cols Optional grouping columns used for facets and matching.
#' @param show_events Logical. If `TRUE`, show event onsets when available.
#' @param max_points Maximum number of signal rows retained after downsampling.
#' @param title Optional plot title.
#'
#' @return A ggplot object with plot data stored in attributes.
#' @export
plot_gazepoint_scr_events <- function(data,
                                      scr_peaks,
                                      event_windows = NULL,
                                      events = NULL,
                                      time_col = NULL,
                                      signal_col = NULL,
                                      phasic_col = NULL,
                                      group_cols = NULL,
                                      show_events = TRUE,
                                      max_points = 5000,
                                      title = NULL) {
  gpbiometrics_require_ggplot2()

  if (!is.data.frame(data)) {
    stop("`data` must be a data frame.", call. = FALSE)
  }

  dat <- as.data.frame(data, stringsAsFactors = FALSE)

  peaks <- gpbiometrics_scr_plot_extract_peaks(scr_peaks)

  if (is.null(time_col)) {
    time_col <- gpbiometrics_eda_plot_first_existing(
      names(dat),
      c(
        "CNT", "cnt", "time", "Time", "TIME",
        "timestamp", "timestamp_ms", "time_ms"
      )
    )
  }

  if (is.null(time_col) || !time_col %in% names(dat)) {
    stop("No usable time column was found. Supply `time_col`.", call. = FALSE)
  }

  if (is.null(signal_col)) {
    if (!is.null(phasic_col) && phasic_col %in% names(dat)) {
      signal_col <- phasic_col
    } else if (inherits(scr_peaks, "gazepoint_scr_peak_detection") &&
               !is.null(scr_peaks$overview$source_signal) &&
               scr_peaks$overview$source_signal[1] %in% names(dat)) {
      signal_col <- scr_peaks$overview$source_signal[1]
    } else {
      signal_col <- gpbiometrics_eda_plot_first_existing(
        names(dat),
        c("GSR_US_PHASIC", "GSR_US", "GSR", "EDA", "eda_phasic")
      )
    }
  }

  if (is.null(signal_col) || !signal_col %in% names(dat)) {
    stop("No usable EDA/GSR signal column was found. Supply `signal_col`.", call. = FALSE)
  }

  if (is.null(group_cols)) {
    group_cols <- intersect(
      c(
        "source_file",
        "source_participant",
        "participant",
        "subject",
        "MEDIA_ID",
        "MEDIA_NAME",
        "trial",
        "trial_id"
      ),
      names(dat)
    )
  }

  plot_data <- gpbiometrics_eda_plot_long_data(
    dat = dat,
    time_col = time_col,
    signal_cols = signal_col,
    group_cols = group_cols,
    standardise = FALSE,
    max_points = max_points
  )

  peak_plot_data <- gpbiometrics_scr_plot_peak_data(
    peaks = peaks,
    dat = dat,
    time_col = time_col,
    signal_col = signal_col,
    group_cols = group_cols
  )

  event_plot_data <- if (isTRUE(show_events)) {
    gpbiometrics_scr_plot_event_data(
      event_windows = event_windows,
      events = events,
      group_cols = group_cols
    )
  } else {
    data.frame()
  }

  p <- ggplot2::ggplot(
    plot_data,
    ggplot2::aes(
      x = .data_time,
      y = .data_value
    )
  ) +
    ggplot2::geom_line(na.rm = TRUE) +
    ggplot2::labs(
      title = if (is.null(title)) "Gazepoint SCR events" else title,
      x = time_col,
      y = signal_col
    ) +
    ggplot2::theme_minimal()

  if (nrow(event_plot_data) > 0) {
    p <- p + ggplot2::geom_vline(
      data = event_plot_data,
      ggplot2::aes(xintercept = .event_time),
      inherit.aes = FALSE,
      linetype = "dashed",
      alpha = 0.5
    )
  }

  if (nrow(peak_plot_data) > 0) {
    p <- p + ggplot2::geom_point(
      data = peak_plot_data,
      ggplot2::aes(
        x = .data_time,
        y = .data_value
      ),
      inherit.aes = FALSE,
      na.rm = TRUE
    )
  }

  if (length(unique(plot_data$.data_group)) > 1) {
    p <- p + ggplot2::facet_wrap(
      stats::as.formula("~ .data_group"),
      scales = "free_x"
    )
  }

  settings <- list(
    time_col = time_col,
    signal_col = signal_col,
    group_cols = group_cols,
    show_events = show_events,
    max_points = max_points,
    title = title,
    event_windows = !is.null(event_windows),
    events = !is.null(events),
    scr_peaks = !is.null(scr_peaks),
    plot_type = "scr_events",
    interpretation_notes = c(
      "SCR peak markers show detected electrodermal response features.",
      "Visual inspection should be used as a QC aid, not as emotional-valence inference."
    )
  )

  p <- standardise_gazepoint_plot_contract(
    plot = p,
    plot_data = plot_data,
    settings = settings,
    interpretation_notes = settings$interpretation_notes,
    plot_type = "scr_events"
  )

  attr(p, "peak_data") <- peak_plot_data
  attr(p, "event_data") <- event_plot_data

  p
}

gpbiometrics_require_ggplot2 <- function() {
  if (!requireNamespace("ggplot2", quietly = TRUE)) {
    stop("Package `ggplot2` is required for this plotting helper.", call. = FALSE)
  }

  invisible(TRUE)
}

gpbiometrics_eda_plot_extract_data <- function(data) {
  if (is.data.frame(data)) {
    return(as.data.frame(data, stringsAsFactors = FALSE))
  }

  if (is.list(data)) {
    candidates <- c(
      "data",
      "decomposition",
      "decomposed_data",
      "signals",
      "aligned_data"
    )

    for (nm in candidates) {
      if (!is.null(data[[nm]]) && is.data.frame(data[[nm]])) {
        return(as.data.frame(data[[nm]], stringsAsFactors = FALSE))
      }
    }

    data_frames <- vapply(data, is.data.frame, logical(1))

    if (any(data_frames)) {
      return(as.data.frame(data[[which(data_frames)[1]]], stringsAsFactors = FALSE))
    }
  }

  stop("`data` must be a data frame or a list containing a data frame.", call. = FALSE)
}

gpbiometrics_eda_plot_first_existing <- function(names_dat, candidates) {
  exact <- candidates[candidates %in% names_dat]

  if (length(exact) > 0) {
    return(exact[1])
  }

  lower_names <- tolower(names_dat)
  lower_candidates <- tolower(candidates)
  idx <- match(lower_candidates, lower_names)
  idx <- idx[!is.na(idx)]

  if (length(idx) > 0) {
    return(names_dat[idx[1]])
  }

  NULL
}

gpbiometrics_eda_plot_signal_cols <- function(names_dat) {
  candidates <- c(
    "GSR_US",
    "GSR_US_TONIC",
    "GSR_US_PHASIC",
    "GSR",
    "EDA",
    "eda",
    "eda_clean",
    "eda_tonic",
    "eda_phasic"
  )

  unique(candidates[candidates %in% names_dat])
}

gpbiometrics_eda_plot_group_id <- function(dat, group_cols) {
  if (length(group_cols) == 0 || !all(group_cols %in% names(dat))) {
    return(rep("all", nrow(dat)))
  }

  group_dat <- dat[group_cols]

  group_dat[] <- lapply(group_dat, function(x) {
    x_chr <- as.character(x)
    x_chr[is.na(x_chr)] <- "<NA>"
    x_chr
  })

  apply(group_dat, 1, paste, collapse = "||")
}

gpbiometrics_eda_plot_downsample <- function(dat, max_points) {
  if (!is.numeric(max_points) ||
      length(max_points) != 1 ||
      !is.finite(max_points) ||
      max_points < 1) {
    stop("`max_points` must be a positive finite number.", call. = FALSE)
  }

  if (nrow(dat) <= max_points) {
    return(dat)
  }

  idx <- unique(round(seq(1, nrow(dat), length.out = max_points)))
  dat[idx, , drop = FALSE]
}

gpbiometrics_eda_plot_long_data <- function(dat,
                                            time_col,
                                            signal_cols,
                                            group_cols,
                                            standardise,
                                            max_points) {
  dat <- gpbiometrics_eda_plot_downsample(dat, max_points)

  time_value <- suppressWarnings(as.numeric(dat[[time_col]]))
  group_id <- gpbiometrics_eda_plot_group_id(dat, group_cols)

  out <- lapply(signal_cols, function(signal_col) {
    value <- suppressWarnings(as.numeric(dat[[signal_col]]))

    if (isTRUE(standardise)) {
      finite_value <- value[is.finite(value)]
      value_mean <- if (length(finite_value) > 0) {
        mean(finite_value, na.rm = TRUE)
      } else {
        NA_real_
      }
      value_sd <- if (length(finite_value) > 1) {
        stats::sd(finite_value, na.rm = TRUE)
      } else {
        NA_real_
      }

      value <- if (is.finite(value_sd) && value_sd > 0) {
        (value - value_mean) / value_sd
      } else {
        ifelse(is.na(value), NA_real_, 0)
      }
    }

    data.frame(
      .data_time = time_value,
      .data_value = value,
      .data_signal = signal_col,
      .data_group = group_id,
      stringsAsFactors = FALSE
    )
  })

  out <- do.call(rbind, out)
  rownames(out) <- NULL

  out
}

gpbiometrics_scr_plot_extract_peaks <- function(scr_peaks) {
  if (inherits(scr_peaks, "gazepoint_scr_peak_detection") &&
      !is.null(scr_peaks$peaks)) {
    return(as.data.frame(scr_peaks$peaks, stringsAsFactors = FALSE))
  }

  if (is.data.frame(scr_peaks)) {
    return(as.data.frame(scr_peaks, stringsAsFactors = FALSE))
  }

  stop("`scr_peaks` must be an SCR peak-detection object or a data frame.", call. = FALSE)
}

gpbiometrics_scr_plot_peak_data <- function(peaks,
                                            dat,
                                            time_col,
                                            signal_col,
                                            group_cols) {
  if (nrow(peaks) == 0 || !"peak_time" %in% names(peaks)) {
    return(data.frame())
  }

  dat_group <- gpbiometrics_eda_plot_group_id(dat, group_cols)
  peak_group <- gpbiometrics_eda_plot_group_id(peaks, group_cols)

  if ("event_group_id" %in% names(peaks) && length(group_cols) > 0) {
    peak_group <- as.character(peaks$event_group_id)
  } else if ("group_id" %in% names(peaks) && length(group_cols) == 0) {
    peak_group <- as.character(peaks$group_id)
  }

  time_value <- suppressWarnings(as.numeric(dat[[time_col]]))
  signal_value <- suppressWarnings(as.numeric(dat[[signal_col]]))
  peak_time <- suppressWarnings(as.numeric(peaks$peak_time))

  out <- lapply(seq_len(nrow(peaks)), function(i) {
    g <- peak_group[i]

    candidate <- which(dat_group == g & is.finite(time_value))

    if (length(candidate) == 0 || !is.finite(peak_time[i])) {
      peak_value <- NA_real_
    } else {
      nearest <- candidate[which.min(abs(time_value[candidate] - peak_time[i]))]
      peak_value <- signal_value[nearest]
    }

    data.frame(
      .data_time = peak_time[i],
      .data_value = peak_value,
      .data_group = g,
      stringsAsFactors = FALSE
    )
  })

  out <- do.call(rbind, out)
  rownames(out) <- NULL

  out
}

gpbiometrics_scr_plot_event_data <- function(event_windows = NULL,
                                             events = NULL,
                                             group_cols = NULL) {
  event_table <- NULL

  if (!is.null(event_windows)) {
    if (inherits(event_windows, "gazepoint_scr_event_window_summary") &&
        !is.null(event_windows$event_table)) {
      event_table <- as.data.frame(event_windows$event_table, stringsAsFactors = FALSE)
    } else if (is.data.frame(event_windows)) {
      event_table <- as.data.frame(event_windows, stringsAsFactors = FALSE)
    }
  } else if (!is.null(events) && is.data.frame(events)) {
    event_table <- as.data.frame(events, stringsAsFactors = FALSE)
  }

  if (is.null(event_table) || nrow(event_table) == 0) {
    return(data.frame())
  }

  event_time_col <- gpbiometrics_eda_plot_first_existing(
    names(event_table),
    c(
      "event_time",
      "onset_time",
      "stimulus_time",
      "time",
      "Time",
      "CNT"
    )
  )

  if (is.null(event_time_col)) {
    return(data.frame())
  }

  event_group <- if (length(group_cols) > 0 &&
                     all(group_cols %in% names(event_table))) {
    gpbiometrics_eda_plot_group_id(event_table, group_cols)
  } else if ("event_group_id" %in% names(event_table)) {
    as.character(event_table$event_group_id)
  } else {
    gpbiometrics_eda_plot_group_id(event_table, group_cols)
  }

  out <- data.frame(
    .event_time = suppressWarnings(as.numeric(event_table[[event_time_col]])),
    .data_group = event_group,
    stringsAsFactors = FALSE
  )

  out[is.finite(out$.event_time), , drop = FALSE]
}

test_that("plot_gazepoint_eda_decomposition returns a ggplot object", {
  skip_if_not_installed("ggplot2")

  dat <- data.frame(
    participant = "P1",
    CNT = seq_len(20),
    GSR_US = seq(1, 2, length.out = 20),
    GSR_US_TONIC = seq(1, 1.5, length.out = 20),
    GSR_US_PHASIC = sin(seq(0, pi, length.out = 20))
  )

  p <- plot_gazepoint_eda_decomposition(
    dat,
    time_col = "CNT",
    signal_cols = c("GSR_US", "GSR_US_TONIC", "GSR_US_PHASIC"),
    group_cols = "participant",
    standardise = TRUE
  )

  expect_s3_class(p, "ggplot")
  expect_true(is.data.frame(attr(p, "plot_data")))
  expect_equal(
    unique(attr(p, "plot_data")$.data_signal),
    c("GSR_US", "GSR_US_TONIC", "GSR_US_PHASIC")
  )
})

test_that("plot_gazepoint_eda_decomposition handles constant channels", {
  skip_if_not_installed("ggplot2")

  dat <- data.frame(
    CNT = seq_len(10),
    GSR_US = rep(0, 10),
    GSR_US_PHASIC = rep(0, 10)
  )

  p <- plot_gazepoint_eda_decomposition(
    dat,
    time_col = "CNT",
    standardise = TRUE
  )

  expect_s3_class(p, "ggplot")
  expect_true(all(attr(p, "plot_data")$.data_value == 0))
})

test_that("plot_gazepoint_scr_events overlays SCR peak markers", {
  skip_if_not_installed("ggplot2")

  dat <- data.frame(
    participant = "P1",
    CNT = seq_len(20),
    GSR_US_PHASIC = c(rep(0, 5), 0.02, 0.08, 0.02, rep(0, 12))
  )

  peaks <- data.frame(
    participant = "P1",
    peak_time = 7,
    amplitude = 0.08,
    status = "detected"
  )

  events <- data.frame(
    participant = "P1",
    event_time = 5
  )

  p <- plot_gazepoint_scr_events(
    dat,
    scr_peaks = peaks,
    events = events,
    time_col = "CNT",
    signal_col = "GSR_US_PHASIC",
    group_cols = "participant"
  )

  expect_s3_class(p, "ggplot")
  expect_true(is.data.frame(attr(p, "plot_data")))
  expect_true(is.data.frame(attr(p, "peak_data")))
  expect_true(is.data.frame(attr(p, "event_data")))
  expect_equal(nrow(attr(p, "peak_data")), 1)
  expect_equal(nrow(attr(p, "event_data")), 1)
})

test_that("plot_gazepoint_scr_events accepts peak-detection objects", {
  skip_if_not_installed("ggplot2")

  dat <- data.frame(
    participant = "P1",
    CNT = seq_len(20),
    GSR_US_PHASIC = c(rep(0, 5), 0.02, 0.08, 0.02, rep(0, 12))
  )

  peaks <- detect_gazepoint_scr_peaks(
    dat,
    phasic_col = "GSR_US_PHASIC",
    time_col = "CNT",
    group_cols = "participant",
    amplitude_min = 0.03
  )

  p <- plot_gazepoint_scr_events(
    dat,
    scr_peaks = peaks,
    time_col = "CNT",
    group_cols = "participant"
  )

  expect_s3_class(p, "ggplot")
  expect_true(nrow(attr(p, "peak_data")) >= 1)
})

#' Extract heart-rate fragmentation features
#'
#' Computes dependency-light heart-rate fragmentation descriptors from IBI/RR
#' intervals. Metrics include percentage of inflection points (PIP), inverse
#' average segment length (IALS), percentage of short segments (PSS), percentage
#' of alternation segments (PAS), and long/short segment summaries.
#'
#' These are fragmentation descriptors of interbeat interval dynamics. They
#' should not be interpreted as clinical diagnoses or direct autonomic-state
#' labels by themselves.
#'
#' @param dat A data frame containing IBI/RR intervals.
#' @param ibi_col Numeric IBI/RR interval column.
#' @param group_cols Optional grouping columns.
#' @param zero_tolerance Absolute change below which interval differences are
#'   treated as zero.
#' @param short_segment_length Maximum segment length counted as short.
#'
#' @return A list with `overview`, `features`, and `settings`.
#' @export
extract_gazepoint_hrv_fragmentation <- function(dat,
                                                ibi_col = "IBI",
                                                group_cols = NULL,
                                                zero_tolerance = 0,
                                                short_segment_length = 3) {
  if (!is.data.frame(dat)) {
    stop("`dat` must be a data frame.", call. = FALSE)
  }

  if (!ibi_col %in% names(dat)) {
    stop("Column `", ibi_col, "` was not found in `dat`.", call. = FALSE)
  }

  if (!is.numeric(dat[[ibi_col]])) {
    stop("`ibi_col` must identify a numeric column.", call. = FALSE)
  }

  if (is.null(group_cols)) {
    group_cols <- character()
  }

  missing_groups <- setdiff(group_cols, names(dat))
  if (length(missing_groups) > 0) {
    stop("Missing `group_cols`: ", paste(missing_groups, collapse = ", "), call. = FALSE)
  }

  groups <- gpbiometrics_hrf_split(dat, group_cols)

  rows <- lapply(names(groups), function(group_id) {
    idx <- groups[[group_id]]
    base <- gpbiometrics_hrf_group_values(dat, idx, group_cols, group_id)

    x <- dat[[ibi_col]][idx]
    x <- x[is.finite(x) & x > 0]

    if (length(x) < 5) {
      return(data.frame(
        base,
        group_id = group_id,
        n_intervals = length(x),
        n_differences = max(length(x) - 1, 0),
        percentage_inflection_points = NA_real_,
        pip = NA_real_,
        inverse_average_segment_length = NA_real_,
        ials = NA_real_,
        percentage_short_segments = NA_real_,
        pss = NA_real_,
        percentage_alternation_segments = NA_real_,
        pas = NA_real_,
        mean_segment_length = NA_real_,
        median_segment_length = NA_real_,
        longest_segment = NA_real_,
        status = "insufficient_intervals",
        stringsAsFactors = FALSE,
        check.names = FALSE
      ))
    }

    dx <- diff(x)
    sx <- sign(dx)
    sx[abs(dx) <= zero_tolerance] <- 0
    sx <- sx[sx != 0]

    if (length(sx) < 2) {
      return(data.frame(
        base,
        group_id = group_id,
        n_intervals = length(x),
        n_differences = length(dx),
        percentage_inflection_points = NA_real_,
        pip = NA_real_,
        inverse_average_segment_length = NA_real_,
        ials = NA_real_,
        percentage_short_segments = NA_real_,
        pss = NA_real_,
        percentage_alternation_segments = NA_real_,
        pas = NA_real_,
        mean_segment_length = NA_real_,
        median_segment_length = NA_real_,
        longest_segment = NA_real_,
        status = "insufficient_nonzero_differences",
        stringsAsFactors = FALSE,
        check.names = FALSE
      ))
    }

    inflections <- sum(sx[-1] * sx[-length(sx)] < 0)
    pip <- 100 * inflections / (length(sx) - 1)

    runs <- rle(sx)
    segment_lengths <- runs$lengths

    mean_segment_length <- mean(segment_lengths)
    median_segment_length <- stats::median(segment_lengths)
    longest_segment <- max(segment_lengths)

    ials <- if (is.finite(mean_segment_length) && mean_segment_length > 0) {
      1 / mean_segment_length
    } else {
      NA_real_
    }

    pss <- 100 * sum(segment_lengths <= short_segment_length) / sum(segment_lengths)
    pas <- 100 * sum(segment_lengths == 1) / sum(segment_lengths)

    data.frame(
      base,
      group_id = group_id,
      n_intervals = length(x),
      n_differences = length(dx),
      percentage_inflection_points = pip,
      pip = pip,
      inverse_average_segment_length = ials,
      ials = ials,
      percentage_short_segments = pss,
      pss = pss,
      percentage_alternation_segments = pas,
      pas = pas,
      mean_segment_length = mean_segment_length,
      median_segment_length = median_segment_length,
      longest_segment = longest_segment,
      status = "hrv_fragmentation_extracted",
      stringsAsFactors = FALSE,
      check.names = FALSE
    )
  })

  features <- do.call(rbind, rows)
  rownames(features) <- NULL

  overview <- data.frame(
    group_count = length(groups),
    feature_rows = nrow(features),
    successful_groups = sum(features$status == "hrv_fragmentation_extracted"),
    problem_groups = sum(features$status != "hrv_fragmentation_extracted"),
    status = if (all(features$status == "hrv_fragmentation_extracted")) {
      "hrv_fragmentation_extracted"
    } else if (any(features$status == "hrv_fragmentation_extracted")) {
      "hrv_fragmentation_partial"
    } else {
      "hrv_fragmentation_failed"
    },
    interpretation = paste(
      "Heart-rate fragmentation features describe rapid sign changes and segment structure in IBI/RR dynamics.",
      "They are not diagnostic labels and do not directly infer vagal tone, cardiovascular disease, emotion, or cognition."
    ),
    stringsAsFactors = FALSE
  )

  structure(
    list(
      overview = overview,
      features = features,
      settings = list(
        ibi_col = ibi_col,
        group_cols = group_cols,
        zero_tolerance = zero_tolerance,
        short_segment_length = short_segment_length
      )
    ),
    class = c("gazepoint_hrv_fragmentation", "list")
  )
}

#' Extract heart-rate asymmetry features
#'
#' Computes dependency-light heart-rate asymmetry descriptors from IBI/RR
#' intervals, including acceleration/deceleration proportions, signed run
#' summaries, and Guzik-style squared-difference asymmetry.
#'
#' Positive IBI/RR differences are treated as decelerations because the heart
#' period lengthens. Negative IBI/RR differences are treated as accelerations.
#'
#' @param dat A data frame containing IBI/RR intervals.
#' @param ibi_col Numeric IBI/RR interval column.
#' @param group_cols Optional grouping columns.
#' @param zero_tolerance Absolute change below which interval differences are
#'   treated as zero.
#'
#' @return A list with `overview`, `features`, `run_table`, and `settings`.
#' @export
extract_gazepoint_hrv_asymmetry <- function(dat,
                                            ibi_col = "IBI",
                                            group_cols = NULL,
                                            zero_tolerance = 0) {
  if (!is.data.frame(dat)) {
    stop("`dat` must be a data frame.", call. = FALSE)
  }

  if (!ibi_col %in% names(dat)) {
    stop("Column `", ibi_col, "` was not found in `dat`.", call. = FALSE)
  }

  if (!is.numeric(dat[[ibi_col]])) {
    stop("`ibi_col` must identify a numeric column.", call. = FALSE)
  }

  if (is.null(group_cols)) {
    group_cols <- character()
  }

  missing_groups <- setdiff(group_cols, names(dat))
  if (length(missing_groups) > 0) {
    stop("Missing `group_cols`: ", paste(missing_groups, collapse = ", "), call. = FALSE)
  }

  groups <- gpbiometrics_hrf_split(dat, group_cols)

  feature_rows <- list()
  run_rows <- list()
  run_id <- 1L

  for (group_id in names(groups)) {
    idx <- groups[[group_id]]
    base <- gpbiometrics_hrf_group_values(dat, idx, group_cols, group_id)

    x <- dat[[ibi_col]][idx]
    x <- x[is.finite(x) & x > 0]

    if (length(x) < 5) {
      feature_rows[[group_id]] <- data.frame(
        base,
        group_id = group_id,
        n_intervals = length(x),
        n_differences = max(length(x) - 1, 0),
        acceleration_count = NA_integer_,
        deceleration_count = NA_integer_,
        acceleration_proportion = NA_real_,
        deceleration_proportion = NA_real_,
        acceleration_run_count = NA_integer_,
        deceleration_run_count = NA_integer_,
        mean_acceleration_run_length = NA_real_,
        mean_deceleration_run_length = NA_real_,
        longest_acceleration_run = NA_real_,
        longest_deceleration_run = NA_real_,
        guzik_index = NA_real_,
        porta_index = NA_real_,
        asymmetry_balance = NA_real_,
        status = "insufficient_intervals",
        stringsAsFactors = FALSE,
        check.names = FALSE
      )
      next
    }

    dx <- diff(x)
    dx[abs(dx) <= zero_tolerance] <- 0

    direction <- ifelse(dx > 0, "deceleration",
                        ifelse(dx < 0, "acceleration", "no_change")
    )

    nonzero <- direction != "no_change"

    if (sum(nonzero) < 2) {
      feature_rows[[group_id]] <- data.frame(
        base,
        group_id = group_id,
        n_intervals = length(x),
        n_differences = length(dx),
        acceleration_count = sum(direction == "acceleration"),
        deceleration_count = sum(direction == "deceleration"),
        acceleration_proportion = NA_real_,
        deceleration_proportion = NA_real_,
        acceleration_run_count = NA_integer_,
        deceleration_run_count = NA_integer_,
        mean_acceleration_run_length = NA_real_,
        mean_deceleration_run_length = NA_real_,
        longest_acceleration_run = NA_real_,
        longest_deceleration_run = NA_real_,
        guzik_index = NA_real_,
        porta_index = NA_real_,
        asymmetry_balance = NA_real_,
        status = "insufficient_nonzero_differences",
        stringsAsFactors = FALSE,
        check.names = FALSE
      )
      next
    }

    d_nonzero <- direction[nonzero]
    dx_nonzero <- dx[nonzero]

    runs <- rle(d_nonzero)

    for (i in seq_along(runs$values)) {
      run_rows[[run_id]] <- data.frame(
        group_id = group_id,
        run_index = i,
        run_type = runs$values[i],
        run_length = runs$lengths[i],
        stringsAsFactors = FALSE
      )
      run_id <- run_id + 1L
    }

    acc_runs <- runs$lengths[runs$values == "acceleration"]
    dec_runs <- runs$lengths[runs$values == "deceleration"]

    acc_count <- sum(d_nonzero == "acceleration")
    dec_count <- sum(d_nonzero == "deceleration")
    total_nonzero <- length(d_nonzero)

    positive_energy <- sum(dx_nonzero[dx_nonzero > 0]^2)
    negative_energy <- sum(dx_nonzero[dx_nonzero < 0]^2)
    total_energy <- positive_energy + negative_energy

    guzik_index <- if (total_energy > 0) {
      100 * positive_energy / total_energy
    } else {
      NA_real_
    }

    porta_index <- 100 * dec_count / total_nonzero

    feature_rows[[group_id]] <- data.frame(
      base,
      group_id = group_id,
      n_intervals = length(x),
      n_differences = length(dx),
      acceleration_count = acc_count,
      deceleration_count = dec_count,
      acceleration_proportion = acc_count / total_nonzero,
      deceleration_proportion = dec_count / total_nonzero,
      acceleration_run_count = length(acc_runs),
      deceleration_run_count = length(dec_runs),
      mean_acceleration_run_length = if (length(acc_runs) > 0) mean(acc_runs) else NA_real_,
      mean_deceleration_run_length = if (length(dec_runs) > 0) mean(dec_runs) else NA_real_,
      longest_acceleration_run = if (length(acc_runs) > 0) max(acc_runs) else NA_real_,
      longest_deceleration_run = if (length(dec_runs) > 0) max(dec_runs) else NA_real_,
      guzik_index = guzik_index,
      porta_index = porta_index,
      asymmetry_balance = (dec_count - acc_count) / total_nonzero,
      status = "hrv_asymmetry_extracted",
      stringsAsFactors = FALSE,
      check.names = FALSE
    )
  }

  features <- do.call(rbind, feature_rows)
  rownames(features) <- NULL

  run_table <- if (length(run_rows) > 0) {
    do.call(rbind, run_rows)
  } else {
    data.frame()
  }

  overview <- data.frame(
    group_count = length(groups),
    feature_rows = nrow(features),
    run_rows = nrow(run_table),
    successful_groups = sum(features$status == "hrv_asymmetry_extracted"),
    problem_groups = sum(features$status != "hrv_asymmetry_extracted"),
    status = if (all(features$status == "hrv_asymmetry_extracted")) {
      "hrv_asymmetry_extracted"
    } else if (any(features$status == "hrv_asymmetry_extracted")) {
      "hrv_asymmetry_partial"
    } else {
      "hrv_asymmetry_failed"
    },
    interpretation = paste(
      "Heart-rate asymmetry features summarise unequal acceleration and deceleration dynamics.",
      "They do not directly infer health status, emotion, cognition, or diagnosis."
    ),
    stringsAsFactors = FALSE
  )

  structure(
    list(
      overview = overview,
      features = features,
      run_table = run_table,
      settings = list(
        ibi_col = ibi_col,
        group_cols = group_cols,
        zero_tolerance = zero_tolerance
      )
    ),
    class = c("gazepoint_hrv_asymmetry", "list")
  )
}

gpbiometrics_hrf_split <- function(dat, group_cols) {
  if (length(group_cols) == 0) {
    return(list(all_rows = seq_len(nrow(dat))))
  }

  gf <- dat[group_cols]
  gf[] <- lapply(gf, function(x) {
    x <- as.character(x)
    x[is.na(x)] <- "<NA>"
    x
  })

  split(seq_len(nrow(dat)), do.call(paste, c(gf, sep = " | ")))
}

gpbiometrics_hrf_group_values <- function(dat, idx, group_cols, group_id) {
  if (length(group_cols) == 0) {
    return(data.frame(unit_label = group_id, stringsAsFactors = FALSE))
  }

  values <- lapply(group_cols, function(nm) as.character(dat[[nm]][idx[1]]))
  names(values) <- group_cols
  as.data.frame(values, stringsAsFactors = FALSE, optional = TRUE)
}

test_that("extract_gazepoint_hrv_fragmentation returns fragmentation features", {
  dat <- data.frame(
    participant = "p1",
    IBI = c(0.80, 0.82, 0.79, 0.83, 0.78, 0.84, 0.81, 0.85, 0.80, 0.86)
  )

  out <- extract_gazepoint_hrv_fragmentation(
    dat,
    ibi_col = "IBI",
    group_cols = "participant"
  )

  expect_s3_class(out, "gazepoint_hrv_fragmentation")
  expect_true("pip" %in% names(out$features))
  expect_true("ials" %in% names(out$features))
  expect_true(is.finite(out$features$pip))
})

test_that("extract_gazepoint_hrv_asymmetry returns run features", {
  dat <- data.frame(
    participant = "p1",
    IBI = c(0.80, 0.82, 0.84, 0.81, 0.79, 0.83, 0.85, 0.82, 0.80, 0.86)
  )

  out <- extract_gazepoint_hrv_asymmetry(
    dat,
    ibi_col = "IBI",
    group_cols = "participant"
  )

  expect_s3_class(out, "gazepoint_hrv_asymmetry")
  expect_true("guzik_index" %in% names(out$features))
  expect_true("porta_index" %in% names(out$features))
  expect_true(nrow(out$run_table) > 0)
})

test_that("model_gazepoint_eda_point_process summarises EDA events", {
  time <- seq(0, 60, by = 0.5)
  eda <- 1 + rnorm(length(time), sd = 0.005)
  events <- rep(0, length(time))
  events[time %in% c(10, 20, 35, 50)] <- 1

  dat <- data.frame(
    participant = "p1",
    time = time,
    GSR_US = eda,
    event = events
  )

  out <- model_gazepoint_eda_point_process(
    dat,
    eda_col = "GSR_US",
    time_col = "time",
    group_cols = "participant",
    event_indicator_col = "event"
  )

  expect_s3_class(out, "gazepoint_eda_point_process")
  expect_equal(nrow(out$event_table), 4)
  expect_true("inverse_gaussian_mu" %in% names(out$process_summary))
})

test_that("model_gazepoint_hr_point_process summarises heartbeat intervals", {
  dat <- data.frame(
    participant = "p1",
    IBI = rep(0.8, 30) + rnorm(30, sd = 0.02)
  )

  out <- model_gazepoint_hr_point_process(
    dat,
    ibi_col = "IBI",
    group_cols = "participant"
  )

  expect_s3_class(out, "gazepoint_hr_point_process")
  expect_true(nrow(out$beat_table) > 0)
  expect_true("inverse_gaussian_lambda" %in% names(out$process_summary))
})

test_that("analyze_gazepoint_cardiorespiratory_causality returns Granger-style summaries", {
  set.seed(1)

  n <- 120
  resp <- sin(seq(0, 8 * pi, length.out = n))
  cardiac <- c(rep(0, 3), resp[1:(n - 3)]) + rnorm(n, sd = 0.1)

  dat <- data.frame(
    participant = "p1",
    time = seq_len(n),
    resp = resp,
    cardiac = cardiac
  )

  out <- analyze_gazepoint_cardiorespiratory_causality(
    dat,
    respiration_col = "resp",
    cardiac_col = "cardiac",
    time_col = "time",
    group_cols = "participant",
    lag_order = 3,
    min_rows = 40
  )

  expect_s3_class(out, "gazepoint_cardiorespiratory_causality")
  expect_true("respiration_to_cardiac_p" %in% names(out$causality_summary))
  expect_equal(out$overview$status, "cardiorespiratory_directionality_estimated")
})

test_that("prepare_gazepoint_ctsi_input creates signal and config tables", {
  dat <- data.frame(
    participant = "p1",
    time = seq(0, 20, by = 0.5),
    GSR_US = 1 + sin(seq(0, 20, by = 0.5)) * 0.05,
    onset = rep(c(5, 10, NA), length.out = 41),
    condition = rep(c("A", "B", "none"), length.out = 41)
  )

  out <- prepare_gazepoint_ctsi_input(
    dat,
    eda_col = "GSR_US",
    time_col = "time",
    group_cols = "participant",
    event_onset_col = "onset",
    event_name_col = "condition",
    sampling_rate = 2
  )

  expect_s3_class(out, "gazepoint_ctsi_input")
  expect_true(nrow(out$signal_table) > 0)
  expect_true(nrow(out$ctsi_config) > 0)
  expect_equal(out$overview$status, "ctsi_input_prepared")
})

#' Extract HRV recurrence quantification analysis features
#'
#' Computes dependency-light recurrence quantification analysis (RQA) features
#' from IBI/RR intervals. This is intended as a compact nonlinear HRV summary
#' and not as a clinical diagnostic tool.
#'
#' @param dat A data frame.
#' @param ibi_col Numeric IBI/RR interval column.
#' @param group_cols Optional grouping columns.
#' @param embedding_dimension Embedding dimension for phase-space reconstruction.
#' @param delay Delay used in embedding.
#' @param radius Radius for recurrence threshold. If `NULL`, uses
#'   `radius_multiplier * SD`.
#' @param radius_multiplier Multiplier used when `radius = NULL`.
#' @param min_line_length Minimum diagonal/vertical line length.
#'
#' @return A list with `overview`, `features`, and `settings`.
#' @export
extract_gazepoint_hrv_rqa <- function(dat,
                                      ibi_col = "IBI",
                                      group_cols = NULL,
                                      embedding_dimension = 2,
                                      delay = 1,
                                      radius = NULL,
                                      radius_multiplier = 0.2,
                                      min_line_length = 2) {
  if (!is.data.frame(dat)) {
    stop("`dat` must be a data frame.", call. = FALSE)
  }

  if (!ibi_col %in% names(dat)) {
    stop("Column `", ibi_col, "` was not found in `dat`.", call. = FALSE)
  }

  if (!is.numeric(dat[[ibi_col]])) {
    stop("`ibi_col` must identify a numeric column.", call. = FALSE)
  }

  if (is.null(group_cols)) {
    group_cols <- character()
  }

  missing_groups <- setdiff(group_cols, names(dat))
  if (length(missing_groups) > 0) {
    stop("Missing `group_cols`: ", paste(missing_groups, collapse = ", "), call. = FALSE)
  }

  groups <- gpbiometrics_hrv_rqa_split(dat, group_cols)

  rows <- lapply(names(groups), function(group_id) {
    idx <- groups[[group_id]]
    base <- gpbiometrics_hrv_rqa_values(dat, idx, group_cols, group_id)

    x <- dat[[ibi_col]][idx]
    x <- x[is.finite(x) & x > 0]

    if (length(x) < (embedding_dimension + 2) * delay + 5) {
      return(data.frame(
        base,
        group_id = group_id,
        n_intervals = length(x),
        recurrence_rate = NA_real_,
        determinism = NA_real_,
        laminarity = NA_real_,
        trapping_time = NA_real_,
        diagonal_entropy = NA_real_,
        mean_diagonal_length = NA_real_,
        longest_diagonal = NA_real_,
        status = "insufficient_intervals",
        stringsAsFactors = FALSE,
        check.names = FALSE
      ))
    }

    embedded <- gpbiometrics_hrv_rqa_embed(
      x,
      dimension = embedding_dimension,
      delay = delay
    )

    rad <- if (!is.null(radius)) {
      radius
    } else {
      radius_multiplier * stats::sd(as.numeric(embedded), na.rm = TRUE)
    }

    if (!is.finite(rad) || rad <= 0) {
      rad <- .Machine$double.eps
    }

    rec <- gpbiometrics_hrv_rqa_matrix(embedded, radius = rad)

    diag_lengths <- gpbiometrics_hrv_rqa_diagonal_lengths(rec)
    vert_lengths <- gpbiometrics_hrv_rqa_vertical_lengths(rec)

    recurrence_points <- sum(rec)
    total_points <- length(rec)

    diag_recurrent <- sum(diag_lengths[diag_lengths >= min_line_length], na.rm = TRUE)
    vert_recurrent <- sum(vert_lengths[vert_lengths >= min_line_length], na.rm = TRUE)

    recurrence_rate <- recurrence_points / total_points
    determinism <- if (recurrence_points > 0) diag_recurrent / recurrence_points else NA_real_
    laminarity <- if (recurrence_points > 0) vert_recurrent / recurrence_points else NA_real_
    trapping_time <- if (any(vert_lengths >= min_line_length)) {
      mean(vert_lengths[vert_lengths >= min_line_length])
    } else {
      NA_real_
    }

    diagonal_entropy <- gpbiometrics_hrv_rqa_entropy(
      diag_lengths[diag_lengths >= min_line_length]
    )

    mean_diagonal_length <- if (any(diag_lengths >= min_line_length)) {
      mean(diag_lengths[diag_lengths >= min_line_length])
    } else {
      NA_real_
    }

    longest_diagonal <- if (length(diag_lengths) > 0) {
      max(diag_lengths)
    } else {
      NA_real_
    }

    data.frame(
      base,
      group_id = group_id,
      n_intervals = length(x),
      recurrence_rate = recurrence_rate,
      determinism = determinism,
      laminarity = laminarity,
      trapping_time = trapping_time,
      diagonal_entropy = diagonal_entropy,
      mean_diagonal_length = mean_diagonal_length,
      longest_diagonal = longest_diagonal,
      status = "hrv_rqa_extracted",
      stringsAsFactors = FALSE,
      check.names = FALSE
    )
  })

  features <- do.call(rbind, rows)
  rownames(features) <- NULL

  overview <- data.frame(
    group_count = length(groups),
    feature_rows = nrow(features),
    successful_groups = sum(features$status == "hrv_rqa_extracted"),
    problem_groups = sum(features$status != "hrv_rqa_extracted"),
    status = if (all(features$status == "hrv_rqa_extracted")) {
      "hrv_rqa_extracted"
    } else if (any(features$status == "hrv_rqa_extracted")) {
      "hrv_rqa_partial"
    } else {
      "hrv_rqa_failed"
    },
    interpretation = paste(
      "RQA features describe recurrence structure in IBI/RR dynamics.",
      "They are nonlinear descriptors and do not infer health status or diagnosis by themselves."
    ),
    stringsAsFactors = FALSE
  )

  structure(
    list(
      overview = overview,
      features = features,
      settings = list(
        ibi_col = ibi_col,
        group_cols = group_cols,
        embedding_dimension = embedding_dimension,
        delay = delay,
        radius = radius,
        radius_multiplier = radius_multiplier,
        min_line_length = min_line_length
      )
    ),
    class = c("gazepoint_hrv_rqa", "list")
  )
}

#' Extract geometric HRV features
#'
#' Computes dependency-light geometric HRV descriptors, including the HRV
#' triangular index and an approximate TINN-style triangular interpolation width.
#'
#' @param dat A data frame.
#' @param ibi_col Numeric IBI/RR interval column.
#' @param group_cols Optional grouping columns.
#' @param bin_width Histogram bin width in the same units as `ibi_col`.
#'
#' @return A list with `overview`, `features`, and `settings`.
#' @export
extract_gazepoint_hrv_geometric <- function(dat,
                                            ibi_col = "IBI",
                                            group_cols = NULL,
                                            bin_width = NULL) {
  if (!is.data.frame(dat)) {
    stop("`dat` must be a data frame.", call. = FALSE)
  }

  if (!ibi_col %in% names(dat)) {
    stop("Column `", ibi_col, "` was not found in `dat`.", call. = FALSE)
  }

  if (!is.numeric(dat[[ibi_col]])) {
    stop("`ibi_col` must identify a numeric column.", call. = FALSE)
  }

  if (is.null(group_cols)) {
    group_cols <- character()
  }

  missing_groups <- setdiff(group_cols, names(dat))
  if (length(missing_groups) > 0) {
    stop("Missing `group_cols`: ", paste(missing_groups, collapse = ", "), call. = FALSE)
  }

  groups <- gpbiometrics_hrv_rqa_split(dat, group_cols)

  rows <- lapply(names(groups), function(group_id) {
    idx <- groups[[group_id]]
    base <- gpbiometrics_hrv_rqa_values(dat, idx, group_cols, group_id)

    x <- dat[[ibi_col]][idx]
    x <- x[is.finite(x) & x > 0]

    if (length(x) < 5) {
      return(data.frame(
        base,
        group_id = group_id,
        n_intervals = length(x),
        bin_width = NA_real_,
        hrv_triangular_index = NA_real_,
        tinn = NA_real_,
        histogram_peak_count = NA_real_,
        status = "insufficient_intervals",
        stringsAsFactors = FALSE,
        check.names = FALSE
      ))
    }

    bw <- bin_width

    if (is.null(bw)) {
      bw <- if (stats::median(x, na.rm = TRUE) > 10) 7.8125 else 0.0078125
    }

    breaks <- seq(
      floor(min(x) / bw) * bw,
      ceiling(max(x) / bw) * bw + bw,
      by = bw
    )

    h <- graphics::hist(x, breaks = breaks, plot = FALSE)
    max_count <- max(h$counts)

    hti <- if (max_count > 0) length(x) / max_count else NA_real_

    nonzero <- which(h$counts > 0)

    tinn <- if (length(nonzero) >= 2) {
      h$breaks[max(nonzero) + 1] - h$breaks[min(nonzero)]
    } else {
      NA_real_
    }

    data.frame(
      base,
      group_id = group_id,
      n_intervals = length(x),
      bin_width = bw,
      hrv_triangular_index = hti,
      tinn = tinn,
      histogram_peak_count = max_count,
      status = "hrv_geometric_extracted",
      stringsAsFactors = FALSE,
      check.names = FALSE
    )
  })

  features <- do.call(rbind, rows)
  rownames(features) <- NULL

  overview <- data.frame(
    group_count = length(groups),
    feature_rows = nrow(features),
    successful_groups = sum(features$status == "hrv_geometric_extracted"),
    problem_groups = sum(features$status != "hrv_geometric_extracted"),
    status = if (all(features$status == "hrv_geometric_extracted")) {
      "hrv_geometric_extracted"
    } else if (any(features$status == "hrv_geometric_extracted")) {
      "hrv_geometric_partial"
    } else {
      "hrv_geometric_failed"
    },
    interpretation = paste(
      "Geometric HRV features summarise the distributional shape of IBI/RR intervals.",
      "They are not diagnostic labels by themselves."
    ),
    stringsAsFactors = FALSE
  )

  structure(
    list(
      overview = overview,
      features = features,
      settings = list(
        ibi_col = ibi_col,
        group_cols = group_cols,
        bin_width = bin_width
      )
    ),
    class = c("gazepoint_hrv_geometric", "list")
  )
}

gpbiometrics_hrv_rqa_split <- function(dat, group_cols) {
  if (length(group_cols) == 0) {
    return(list(all_rows = seq_len(nrow(dat))))
  }

  gf <- dat[group_cols]
  gf[] <- lapply(gf, function(x) {
    x <- as.character(x)
    x[is.na(x)] <- "<NA>"
    x
  })

  split(seq_len(nrow(dat)), do.call(paste, c(gf, sep = " | ")))
}

gpbiometrics_hrv_rqa_values <- function(dat, idx, group_cols, group_id) {
  if (length(group_cols) == 0) {
    return(data.frame(unit_label = group_id, stringsAsFactors = FALSE))
  }

  values <- lapply(group_cols, function(nm) as.character(dat[[nm]][idx[1]]))
  names(values) <- group_cols
  as.data.frame(values, stringsAsFactors = FALSE, optional = TRUE)
}

gpbiometrics_hrv_rqa_embed <- function(x, dimension = 2, delay = 1) {
  n <- length(x) - (dimension - 1) * delay

  if (n <= 0) {
    return(matrix(numeric(), nrow = 0, ncol = dimension))
  }

  out <- matrix(NA_real_, nrow = n, ncol = dimension)

  for (j in seq_len(dimension)) {
    out[, j] <- x[(1:n) + (j - 1) * delay]
  }

  out
}

gpbiometrics_hrv_rqa_matrix <- function(embedded, radius) {
  n <- nrow(embedded)
  rec <- matrix(FALSE, nrow = n, ncol = n)

  for (i in seq_len(n)) {
    d <- sqrt(rowSums((t(t(embedded) - embedded[i, ]))^2))
    rec[i, ] <- d <= radius
  }

  rec
}

gpbiometrics_hrv_rqa_diagonal_lengths <- function(rec) {
  n <- nrow(rec)
  lengths <- integer()

  for (offset in seq(-(n - 1), n - 1)) {
    diag_values <- gpbiometrics_hrv_rqa_get_offset_diagonal(rec, offset)

    if (length(diag_values) == 0) {
      next
    }

    lengths <- c(lengths, gpbiometrics_hrv_rqa_run_lengths(diag_values))
  }

  lengths
}

gpbiometrics_hrv_rqa_get_offset_diagonal <- function(rec, offset = 0) {
  n <- nrow(rec)

  if (offset >= 0) {
    i <- seq_len(n - offset)
    j <- i + offset
  } else {
    j <- seq_len(n + offset)
    i <- j - offset
  }

  if (length(i) == 0 || length(j) == 0) {
    return(logical())
  }

  rec[cbind(i, j)]
}

gpbiometrics_hrv_rqa_vertical_lengths <- function(rec) {
  lengths <- integer()

  for (j in seq_len(ncol(rec))) {
    lengths <- c(lengths, gpbiometrics_hrv_rqa_run_lengths(rec[, j]))
  }

  lengths
}

gpbiometrics_hrv_rqa_run_lengths <- function(x) {
  r <- rle(as.logical(x))
  r$lengths[r$values]
}

gpbiometrics_hrv_rqa_entropy <- function(lengths) {
  if (length(lengths) == 0) {
    return(NA_real_)
  }

  tab <- table(lengths)
  p <- as.numeric(tab) / sum(tab)

  -sum(p * log(p))
}

test_that("extract_gazepoint_eda_tvsymp returns time-varying spectral EDA summaries", {
  set.seed(1)

  time <- seq(0, 180, by = 1)
  eda <- sin(2 * pi * 0.12 * time) + rnorm(length(time), sd = 0.05)

  dat <- data.frame(
    participant = "p1",
    time = time,
    GSR_US = eda
  )

  out <- extract_gazepoint_eda_tvsymp(
    dat,
    eda_col = "GSR_US",
    time_col = "time",
    group_cols = "participant",
    sampling_rate = 1,
    window_seconds = 60,
    step_seconds = 30
  )

  expect_s3_class(out, "gazepoint_eda_tvsymp")
  expect_true(nrow(out$tvsymp_timeseries) > 0)
  expect_true("tvsymp_power" %in% names(out$tvsymp_timeseries))
  expect_true("edasympn" %in% names(out$tvsymp_timeseries))
})

test_that("autoencoder denoising bridges apply user-supplied reconstruction functions", {
  dat <- data.frame(
    participant = "p1",
    time = seq_len(32),
    GSR_US = sin(seq(0, 2 * pi, length.out = 32)) + rnorm(32, sd = 0.1),
    HRP = sin(seq(0, 4 * pi, length.out = 32)) + rnorm(32, sd = 0.1)
  )

  model_fun <- function(x) {
    x * 0.5
  }

  eda <- denoise_gazepoint_eda_autoencoder(
    dat,
    eda_col = "GSR_US",
    group_cols = "participant",
    model = model_fun,
    window_samples = 16
  )

  ppg <- denoise_gazepoint_ppg_autoencoder(
    dat,
    ppg_col = "HRP",
    group_cols = "participant",
    model = model_fun,
    window_samples = 16
  )

  expect_s3_class(eda, "gazepoint_autoencoder_denoised")
  expect_s3_class(ppg, "gazepoint_autoencoder_denoised")
  expect_true("GSR_US_autoencoder_denoised" %in% names(eda))
  expect_true("HRP_autoencoder_denoised" %in% names(ppg))
  expect_equal(
    attr(eda, "autoencoder_denoising_overview")$status,
    "autoencoder_reconstruction_complete"
  )
})

test_that("extract_gazepoint_hrv_rqa returns recurrence features", {
  dat <- data.frame(
    participant = "p1",
    IBI = 0.8 + 0.04 * sin(seq(0, 8 * pi, length.out = 80))
  )

  out <- extract_gazepoint_hrv_rqa(
    dat,
    ibi_col = "IBI",
    group_cols = "participant",
    embedding_dimension = 2,
    delay = 1
  )

  expect_s3_class(out, "gazepoint_hrv_rqa")
  expect_true("recurrence_rate" %in% names(out$features))
  expect_true("determinism" %in% names(out$features))
  expect_true(is.finite(out$features$recurrence_rate))
})

test_that("extract_gazepoint_hrv_geometric returns HTI and TINN features", {
  dat <- data.frame(
    participant = "p1",
    IBI = 0.8 + rnorm(100, sd = 0.03)
  )

  out <- extract_gazepoint_hrv_geometric(
    dat,
    ibi_col = "IBI",
    group_cols = "participant"
  )

  expect_s3_class(out, "gazepoint_hrv_geometric")
  expect_true("hrv_triangular_index" %in% names(out$features))
  expect_true("tinn" %in% names(out$features))
  expect_true(is.finite(out$features$hrv_triangular_index))
})

test_that("plot_gazepoint_scr_specification_curve returns plot metadata", {
  specification_summary <- data.frame(
    specification_id = paste0("spec_", 1:5),
    mean_response_amplitude = c(0.02, 0.01, 0.04, 0.03, 0.05),
    response_rate = c(0.2, 0.1, 0.4, 0.3, 0.5)
  )

  out <- plot_gazepoint_scr_specification_curve(
    specification_summary,
    estimate_col = "mean_response_amplitude"
  )

  expect_s3_class(out, "gazepoint_scr_specification_curve_plot")
  expect_true("specification_rank" %in% names(out$plot_data))
  expect_equal(out$plot_data$mean_response_amplitude, sort(specification_summary$mean_response_amplitude))
})

#' Format the gpbiometrics feature inventory for users
#'
#' Adds user-facing labels and interpretation metadata to the package feature
#' inventory. This helper is intentionally non-breaking: it does not replace the
#' core inventory object returned by `create_gazepoint_biometrics_feature_inventory()`.
#'
#' @param inventory Optional inventory object returned by
#'   `create_gazepoint_biometrics_feature_inventory()`. If `NULL`, a fresh
#'   inventory is created.
#' @param include_internal Logical. Passed to
#'   `create_gazepoint_biometrics_feature_inventory()` when `inventory = NULL`.
#' @param sort Logical. If `TRUE`, sort by domain, user level, and function name.
#'
#' @return A data frame with polished user-facing inventory columns.
#' @export
format_gazepoint_biometrics_feature_inventory <- function(inventory = NULL,
                                                          include_internal = FALSE,
                                                          sort = TRUE) {
  if (is.null(inventory)) {
    inventory <- create_gazepoint_biometrics_feature_inventory(
      include_internal = include_internal
    )
  }

  if (is.data.frame(inventory)) {
    tab <- inventory
  } else if (is.list(inventory) && "inventory" %in% names(inventory)) {
    tab <- inventory$inventory
  } else {
    stop(
      "`inventory` must be a data frame or an object returned by `create_gazepoint_biometrics_feature_inventory()`.",
      call. = FALSE
    )
  }

  required <- c("domain", "function_name", "available", "status")
  missing_required <- setdiff(required, names(tab))

  if (length(missing_required) > 0) {
    stop(
      "Inventory table is missing required columns: ",
      paste(missing_required, collapse = ", "),
      call. = FALSE
    )
  }

  out <- tab

  out$domain_label <- gpbiometrics_inventory_domain_label(out$domain)
  out$workflow_stage <- gpbiometrics_inventory_workflow_stage(out$domain)
  out$method_family <- vapply(
    out$function_name,
    gpbiometrics_inventory_method_family,
    character(1)
  )
  out$user_level <- vapply(
    out$function_name,
    gpbiometrics_inventory_user_level,
    character(1)
  )
  out$interpretation_caution <- vapply(
    out$function_name,
    gpbiometrics_inventory_caution,
    character(1)
  )

  out$availability_label <- ifelse(
    isTRUE_VECTOR(out$available),
    "Available",
    "Missing"
  )

  preferred_order <- c(
    "domain",
    "domain_label",
    "workflow_stage",
    "method_family",
    "user_level",
    "function_name",
    "interpretation_caution",
    "expected",
    "user_facing",
    "available",
    "availability_label",
    "status"
  )

  preferred_order <- preferred_order[preferred_order %in% names(out)]
  extra_cols <- setdiff(names(out), preferred_order)

  out <- out[, c(preferred_order, extra_cols), drop = FALSE]

  if (isTRUE(sort)) {
    level_rank <- match(out$user_level, c("Core", "Intermediate", "Advanced"))
    level_rank[is.na(level_rank)] <- 99

    out <- out[order(out$domain, level_rank, out$method_family, out$function_name), , drop = FALSE]
  }

  rownames(out) <- NULL
  out
}

#' Summarise the formatted gpbiometrics feature inventory
#'
#' @param formatted_inventory Optional table returned by
#'   `format_gazepoint_biometrics_feature_inventory()`.
#'
#' @return A list with domain, method-family, and user-level summaries.
#' @export
summarise_gazepoint_biometrics_feature_inventory <- function(formatted_inventory = NULL) {
  if (is.null(formatted_inventory)) {
    formatted_inventory <- format_gazepoint_biometrics_feature_inventory()
  }

  required <- c("domain", "domain_label", "method_family", "user_level", "available")
  missing_required <- setdiff(required, names(formatted_inventory))

  if (length(missing_required) > 0) {
    stop(
      "`formatted_inventory` is missing required columns: ",
      paste(missing_required, collapse = ", "),
      call. = FALSE
    )
  }

  domain_summary <- stats::aggregate(
    available ~ domain + domain_label,
    data = formatted_inventory,
    FUN = function(x) c(feature_count = length(x), available_features = sum(x))
  )

  domain_summary <- data.frame(
    domain = domain_summary$domain,
    domain_label = domain_summary$domain_label,
    feature_count = domain_summary$available[, "feature_count"],
    available_features = domain_summary$available[, "available_features"],
    stringsAsFactors = FALSE
  )

  domain_summary$missing_features <- domain_summary$feature_count -
    domain_summary$available_features
  domain_summary$completion_rate <- domain_summary$available_features /
    domain_summary$feature_count
  domain_summary$status <- ifelse(
    domain_summary$missing_features == 0,
    "complete",
    "incomplete"
  )

  method_summary <- stats::aggregate(
    available ~ method_family,
    data = formatted_inventory,
    FUN = function(x) c(feature_count = length(x), available_features = sum(x))
  )

  method_summary <- data.frame(
    method_family = method_summary$method_family,
    feature_count = method_summary$available[, "feature_count"],
    available_features = method_summary$available[, "available_features"],
    stringsAsFactors = FALSE
  )

  method_summary$missing_features <- method_summary$feature_count -
    method_summary$available_features

  user_level_summary <- stats::aggregate(
    available ~ user_level,
    data = formatted_inventory,
    FUN = function(x) c(feature_count = length(x), available_features = sum(x))
  )

  user_level_summary <- data.frame(
    user_level = user_level_summary$user_level,
    feature_count = user_level_summary$available[, "feature_count"],
    available_features = user_level_summary$available[, "available_features"],
    stringsAsFactors = FALSE
  )

  user_level_summary$missing_features <- user_level_summary$feature_count -
    user_level_summary$available_features

  list(
    overview = data.frame(
      feature_rows = nrow(formatted_inventory),
      domain_count = length(unique(formatted_inventory$domain)),
      method_family_count = length(unique(formatted_inventory$method_family)),
      user_level_count = length(unique(formatted_inventory$user_level)),
      available_features = sum(formatted_inventory$available),
      missing_features = sum(!formatted_inventory$available),
      status = if (all(formatted_inventory$available)) {
        "formatted_inventory_complete"
      } else {
        "formatted_inventory_incomplete"
      },
      stringsAsFactors = FALSE
    ),
    domain_summary = domain_summary,
    method_summary = method_summary,
    user_level_summary = user_level_summary
  )
}

gpbiometrics_inventory_domain_label <- function(domain) {
  labels <- c(
    import_and_schema = "Import and schema",
    quality_and_readiness = "Quality and readiness",
    preprocessing = "Preprocessing and correction",
    eda_scr = "EDA, GSR, and SCR",
    ibi_hr_hrv = "Pulse, IBI, HR, HRV, and respiration",
    ttl_alignment = "TTL and synchronisation",
    aoi_biometrics = "AOI and multimodal summaries",
    modelling_and_windows = "Windows and model-ready data",
    reporting = "Reporting and documentation",
    interoperability = "External interoperability",
    plotting = "Plotting and visual diagnostics"
  )

  unname(ifelse(domain %in% names(labels), labels[domain], domain))
}

gpbiometrics_inventory_workflow_stage <- function(domain) {
  stages <- c(
    import_and_schema = "1. Import",
    quality_and_readiness = "2. Quality control",
    preprocessing = "3. Preprocessing",
    eda_scr = "4. Signal features",
    ibi_hr_hrv = "4. Signal features",
    ttl_alignment = "5. Timing and alignment",
    aoi_biometrics = "6. Multimodal summaries",
    modelling_and_windows = "7. Analysis preparation",
    reporting = "8. Reporting",
    interoperability = "9. External bridges",
    plotting = "10. Visualisation"
  )

  unname(ifelse(domain %in% names(stages), stages[domain], "Other"))
}

gpbiometrics_inventory_method_family <- function(function_name) {
  fn <- tolower(function_name)

  if (grepl("import|schema|column|name|detect_active|simulate", fn)) {
    return("Import, schema, or simulation")
  }

  if (grepl("readiness|diagnose|validate|audit|missing|validity|dropout|exclusion|drift|stabilization", fn)) {
    return("Quality control and readiness")
  }

  if (grepl("standardi|filter|smooth|ema|baseline|correct|regress|denoise|artifact|temperature|luminance|preprocess", fn)) {
    return("Preprocessing and confound control")
  }

  if (grepl("eda|gsr|scr|skin|cvxeda|ledalab|pspm|tvsymp|susceptance|admittance", fn)) {
    return("EDA/GSR/SCR analysis")
  }

  if (grepl("ibi|hrv|heart|hr_|hrp|ppg|pulse|beat|rsa|respiration|ipfm|fuzzy|entropy|rcmse|rqa|lorenz|csi", fn)) {
    return("Cardiovascular and respiration analysis")
  }

  if (grepl("ttl|sync|align|lag|time_reset", fn)) {
    return("Timing and synchronisation")
  }

  if (grepl("aoi|multimodal", fn)) {
    return("AOI and multimodal summaries")
  }

  if (grepl("window|model|lme|hurdle|chunk|online_design|point_process|causality", fn)) {
    return("Model-ready data and advanced modelling")
  }

  if (grepl("report|bundle|checklist|preregistration|pipeline|statistics|export", fn)) {
    return("Reporting, export, and reproducibility")
  }

  if (grepl("plot|dashboard|gram|timeline", fn)) {
    return("Plotting and visual diagnostics")
  }

  "General helper"
}

gpbiometrics_inventory_user_level <- function(function_name) {
  fn <- tolower(function_name)

  advanced_patterns <- paste(
    c(
      "ceemdan",
      "kalman",
      "fuzzy",
      "rcmse",
      "rqa",
      "geometric",
      "nonlinear",
      "ipfm",
      "cvxeda",
      "pspm",
      "ledalab",
      "ctsi",
      "svm",
      "autoencoder",
      "wavelet",
      "tvsymp",
      "point_process",
      "causality",
      "surrogate",
      "doubly_stochastic",
      "spectral",
      "complexity",
      "susceptance",
      "admittance",
      "skin_potential",
      "bilateral",
      "online_design",
      "multiverse",
      "xdf"
    ),
    collapse = "|"
  )

  core_patterns <- paste(
    c(
      "^import_",
      "^detect_gazepoint_biometric",
      "^check_gazepoint_biometric",
      "^validate_",
      "^summarise_",
      "^run_gazepoint_biometrics_workflow",
      "^diagnose_gazepoint_biometrics_workflow",
      "^run_gazepoint_biometrics_real_data_readiness",
      "^create_gazepoint_biometrics_report",
      "^export_gazepoint_biometrics_report_bundle",
      "^create_gazepoint_biometrics_feature_inventory"
    ),
    collapse = "|"
  )

  if (grepl(advanced_patterns, fn)) {
    return("Advanced")
  }

  if (grepl(core_patterns, fn)) {
    return("Core")
  }

  "Intermediate"
}

gpbiometrics_inventory_caution <- function(function_name) {
  fn <- tolower(function_name)

  if (grepl("eda|gsr|scr|skin|cvxeda|ledalab|pspm|tvsymp|susceptance|admittance", fn)) {
    return("EDA/SCR outputs describe electrodermal dynamics; they do not directly infer emotion, stress, cognition, health status, or diagnosis.")
  }

  if (grepl("ibi|hrv|heart|hr_|hrp|ppg|pulse|beat|rsa|respiration|ipfm|fuzzy|entropy|rcmse|rqa|csi", fn)) {
    return("Cardiovascular and respiration outputs are physiological descriptors or proxies; they are not clinical or psychological labels.")
  }

  if (grepl("pupil|luminance", fn)) {
    return("Pupil outputs are strongly affected by luminance and visual context; residuals are not proof of cognitive-load-only effects.")
  }

  if (grepl("aoi|gaze|fixation|saccade", fn)) {
    return("AOI and gaze outputs describe visual timing/allocation; they do not prove comprehension, scrutiny, or preference.")
  }

  if (grepl("statistics|anova|kruskal|posthoc", fn)) {
    return("Automated statistics are exploratory convenience summaries and should be checked against the study design.")
  }

  if (grepl("online_design", fn)) {
    return("Online design output is decision support only and does not autonomously control experiments.")
  }

  if (grepl("import|schema|report|bundle|inventory|plot|dashboard|ttl|sync|align", fn)) {
    return("Use this output to document processing decisions; interpretation depends on the study design and signal quality.")
  }

  "Interpret conservatively and review assumptions before confirmatory use."
}

isTRUE_VECTOR <- function(x) {
  !is.na(x) & x
}

test_that("formatted feature inventory preserves all functions and adds user-facing metadata", {
  inventory <- create_gazepoint_biometrics_feature_inventory()
  formatted <- format_gazepoint_biometrics_feature_inventory(inventory)

  expect_s3_class(formatted, "data.frame")
  expect_equal(nrow(formatted), nrow(inventory$inventory))

  expected_cols <- c(
    "domain",
    "domain_label",
    "workflow_stage",
    "method_family",
    "user_level",
    "function_name",
    "interpretation_caution",
    "available",
    "availability_label",
    "status"
  )

  expect_true(all(expected_cols %in% names(formatted)))
  expect_true(all(formatted$available))
  expect_true(all(formatted$availability_label == "Available"))
  expect_true(all(nchar(formatted$domain_label) > 0))
  expect_true(all(nchar(formatted$method_family) > 0))
  expect_true(all(nchar(formatted$interpretation_caution) > 0))
})

test_that("formatted feature inventory summary reports complete coverage", {
  formatted <- format_gazepoint_biometrics_feature_inventory()
  summary <- summarise_gazepoint_biometrics_feature_inventory(formatted)

  expect_named(
    summary,
    c("overview", "domain_summary", "method_summary", "user_level_summary")
  )

  expect_equal(summary$overview$status, "formatted_inventory_complete")
  expect_equal(summary$overview$feature_rows, nrow(formatted))
  expect_equal(summary$overview$missing_features, 0)
  expect_true(all(summary$domain_summary$status == "complete"))
})

#' Audit Gazepoint GSR/EDA signal quality
#'
#' Audits Gazepoint GSR/EDA columns for missingness, inactive zero rows,
#' validity flags, plausible value ranges, flatlining, and usable sample
#' coverage. When available, `GSR_US` is used by default because it represents
#' skin conductance in microsiemens in Gazepoint exports.
#'
#' @param data A data frame or a path to a Gazepoint CSV export.
#' @param value_column Optional GSR/EDA value column. If `NULL`, `GSR_US` is
#'   used when present, otherwise `GSR`.
#' @param validity_column Optional validity column. Defaults to `"GSRV"`.
#' @param min_value Minimum plausible value.
#' @param max_value Maximum plausible value.
#' @param jump_threshold Optional threshold for detecting large sample-to-sample
#'   jumps.
#'
#' @return A one-row data frame summarising signal quality.
#'
#' @export
audit_gazepoint_gsr_quality <- function(data,
                                        value_column = NULL,
                                        validity_column = "GSRV",
                                        min_value = 0,
                                        max_value = 100,
                                        jump_threshold = NULL) {
  dat <- coerce_gazepoint_biometrics_data(data)

  if (is.null(value_column)) {
    value_column <- choose_first_present(dat, c("GSR_US", "GSR"))
  }

  audit_gazepoint_signal_quality(
    data = dat,
    signal = "gsr_eda",
    value_column = value_column,
    validity_column = validity_column,
    min_value = min_value,
    max_value = max_value,
    jump_threshold = jump_threshold
  )
}


#' Audit Gazepoint heart-rate signal quality
#'
#' Audits Gazepoint heart-rate values for missingness, inactive zero rows,
#' validity flags, plausible value ranges, sudden jumps, flatlining, and usable
#' sample coverage. `HRV` is treated as the heart-rate validity flag, not as a
#' heart-rate-variability metric.
#'
#' @param data A data frame or a path to a Gazepoint CSV export.
#' @param value_column Heart-rate value column. Defaults to `"HR"`.
#' @param validity_column Heart-rate validity column. Defaults to `"HRV"`.
#' @param min_value Minimum plausible heart rate.
#' @param max_value Maximum plausible heart rate.
#' @param jump_threshold Threshold for detecting large sample-to-sample jumps.
#'
#' @return A one-row data frame summarising signal quality.
#'
#' @export
audit_gazepoint_hr_quality <- function(data,
                                       value_column = "HR",
                                       validity_column = "HRV",
                                       min_value = 30,
                                       max_value = 220,
                                       jump_threshold = 25) {
  dat <- coerce_gazepoint_biometrics_data(data)

  audit_gazepoint_signal_quality(
    data = dat,
    signal = "heart_rate",
    value_column = value_column,
    validity_column = validity_column,
    min_value = min_value,
    max_value = max_value,
    jump_threshold = jump_threshold
  )
}


#' Audit Gazepoint engagement-dial signal quality
#'
#' Audits Gazepoint engagement-dial values for missingness, inactive rows,
#' validity flags, plausible range, flatlining, and usable sample coverage.
#'
#' @param data A data frame or a path to a Gazepoint CSV export.
#' @param value_column Engagement-dial value column. Defaults to `"DIAL"`.
#' @param validity_column Engagement-dial validity column. Defaults to
#'   `"DIALV"`.
#' @param min_value Minimum plausible dial value.
#' @param max_value Maximum plausible dial value.
#' @param jump_threshold Optional threshold for detecting large sample-to-sample
#'   jumps.
#'
#' @return A one-row data frame summarising signal quality.
#'
#' @export
audit_gazepoint_engagement_dial <- function(data,
                                            value_column = "DIAL",
                                            validity_column = "DIALV",
                                            min_value = 0,
                                            max_value = 1,
                                            jump_threshold = NULL) {
  dat <- coerce_gazepoint_biometrics_data(data)

  audit_gazepoint_signal_quality(
    data = dat,
    signal = "engagement_dial",
    value_column = value_column,
    validity_column = validity_column,
    min_value = min_value,
    max_value = max_value,
    jump_threshold = jump_threshold
  )
}


audit_gazepoint_signal_quality <- function(data,
                                           signal,
                                           value_column,
                                           validity_column = NULL,
                                           min_value = -Inf,
                                           max_value = Inf,
                                           jump_threshold = NULL) {
  if (is.null(value_column) || length(value_column) != 1L || is.na(value_column)) {
    return(empty_signal_quality(signal, value_column, validity_column, nrow(data)))
  }

  if (!value_column %in% names(data)) {
    return(empty_signal_quality(signal, value_column, validity_column, nrow(data)))
  }

  x <- as_numeric_safe(data[[value_column]])

  missing_rows <- sum(is.na(x))
  zero_rows <- sum(!is.na(x) & x == 0)
  nonzero_rows <- sum(!is.na(x) & x != 0)

  valid <- !is.na(x)

  if (!is.null(validity_column) && validity_column %in% names(data)) {
    v <- as_numeric_safe(data[[validity_column]])
    valid <- valid & !is.na(v) & v > 0
  }

  low <- !is.na(x) & x < min_value
  high <- !is.na(x) & x > max_value

  usable <- valid & !low & !high

  usable_values <- x[usable]

  large_jump_rows <- NA_integer_

  if (!is.null(jump_threshold) && length(usable_values) > 1L) {
    large_jump_rows <- sum(abs(diff(usable_values)) > jump_threshold, na.rm = TRUE)
  }

  flatline <- FALSE

  if (length(usable_values) > 1L) {
    flatline <- length(unique(usable_values)) == 1L
  }

  data.frame(
    signal = signal,
    issue = NA_character_,
    value_column = value_column,
    validity_column = ifelse(
      !is.null(validity_column) && validity_column %in% names(data),
      validity_column,
      NA_character_
    ),
    n_rows = length(x),
    missing_rows = missing_rows,
    missing_pct = safe_pct(missing_rows, length(x)),
    zero_rows = zero_rows,
    zero_pct = safe_pct(zero_rows, length(x)),
    nonzero_rows = nonzero_rows,
    valid_rows = sum(valid),
    invalid_rows = length(x) - sum(valid),
    low_rows = sum(low),
    high_rows = sum(high),
    large_jump_rows = large_jump_rows,
    flatline = flatline,
    usable_rows = length(usable_values),
    usable_pct = safe_pct(length(usable_values), length(x)),
    min_value = ifelse(length(usable_values) > 0L, min(usable_values), NA_real_),
    max_value = ifelse(length(usable_values) > 0L, max(usable_values), NA_real_),
    mean_value = ifelse(length(usable_values) > 0L, mean(usable_values), NA_real_),
    stringsAsFactors = FALSE
  )
}


empty_signal_quality <- function(signal,
                                 value_column,
                                 validity_column,
                                 n_rows) {
  data.frame(
    signal = signal,
    issue = "value_column_missing",
    value_column = ifelse(is.null(value_column), NA_character_, value_column),
    validity_column = ifelse(is.null(validity_column), NA_character_, validity_column),
    n_rows = n_rows,
    missing_rows = NA_integer_,
    missing_pct = NA_real_,
    zero_rows = NA_integer_,
    zero_pct = NA_real_,
    nonzero_rows = NA_integer_,
    valid_rows = NA_integer_,
    invalid_rows = NA_integer_,
    low_rows = NA_integer_,
    high_rows = NA_integer_,
    large_jump_rows = NA_integer_,
    flatline = NA,
    usable_rows = 0L,
    usable_pct = 0,
    min_value = NA_real_,
    max_value = NA_real_,
    mean_value = NA_real_,
    stringsAsFactors = FALSE
  )
}


choose_first_present <- function(data, columns) {
  present <- intersect(columns, names(data))

  if (length(present) == 0L) {
    return(NA_character_)
  }

  present[1]
}

test_that("audit_gazepoint_gsr_quality audits active GSR_US signal", {
  dat <- data.frame(
    GSR_US = c(2.0, 2.1, 2.2, NA, 0),
    GSRV = c(1, 1, 1, 0, 0)
  )

  audit <- audit_gazepoint_gsr_quality(dat)

  expect_equal(audit$signal, "gsr_eda")
  expect_equal(audit$value_column, "GSR_US")
  expect_equal(audit$n_rows, 5)
  expect_equal(audit$missing_rows, 1)
  expect_equal(audit$zero_rows, 1)
  expect_equal(audit$usable_rows, 3)
})


test_that("audit_gazepoint_gsr_quality falls back to GSR when GSR_US is absent", {
  dat <- data.frame(
    GSR = c(500000, 510000, 520000),
    GSRV = c(1, 1, 1)
  )

  audit <- audit_gazepoint_gsr_quality(
    dat,
    min_value = 1,
    max_value = 1000000
  )

  expect_equal(audit$value_column, "GSR")
  expect_equal(audit$usable_rows, 3)
})


test_that("audit_gazepoint_hr_quality flags impossible heart-rate values", {
  dat <- data.frame(
    HR = c(75, 76, 300, 0, NA),
    HRV = c(1, 1, 1, 0, 0)
  )

  audit <- audit_gazepoint_hr_quality(dat)

  expect_equal(audit$signal, "heart_rate")
  expect_equal(audit$n_rows, 5)
  expect_equal(audit$high_rows, 1)
  expect_equal(audit$zero_rows, 1)
  expect_equal(audit$missing_rows, 1)
  expect_equal(audit$usable_rows, 2)
})


test_that("audit_gazepoint_hr_quality detects large jumps", {
  dat <- data.frame(
    HR = c(75, 76, 130, 131),
    HRV = c(1, 1, 1, 1)
  )

  audit <- audit_gazepoint_hr_quality(dat, jump_threshold = 25)

  expect_equal(audit$large_jump_rows, 1)
})


test_that("audit_gazepoint_engagement_dial audits dial range", {
  dat <- data.frame(
    DIAL = c(0.1, 0.2, 1.2, 0, NA),
    DIALV = c(1, 1, 1, 0, 0)
  )

  audit <- audit_gazepoint_engagement_dial(dat)

  expect_equal(audit$signal, "engagement_dial")
  expect_equal(audit$high_rows, 1)
  expect_equal(audit$zero_rows, 1)
  expect_equal(audit$missing_rows, 1)
  expect_equal(audit$usable_rows, 2)
})


test_that("quality audits return empty audit when value column is missing", {
  dat <- data.frame(
    X = c(1, 2, 3)
  )

  audit <- audit_gazepoint_hr_quality(dat)

  expect_equal(audit$signal, "heart_rate")
  expect_equal(audit$issue, "value_column_missing")
  expect_equal(audit$usable_rows, 0)
})
test_that("quality audits summarise usable values rather than inactive zeros", {
  dat <- data.frame(
    GSR_US = c(0, 2, 4),
    GSRV = c(0, 1, 1),
    HR = c(0, 70, 80),
    HRV = c(0, 1, 1)
  )

  gsr <- audit_gazepoint_gsr_quality(dat)
  hr <- audit_gazepoint_hr_quality(dat)

  expect_equal(gsr$usable_rows, 2)
  expect_equal(gsr$min_value, 2)
  expect_equal(gsr$max_value, 4)
  expect_equal(gsr$mean_value, 3)

  expect_equal(hr$usable_rows, 2)
  expect_equal(hr$min_value, 70)
  expect_equal(hr$max_value, 80)
  expect_equal(hr$mean_value, 75)
})

test_that("create_gazepoint_biometrics_report_tables works from workflow object", {
  folder <- tempfile("gazepoint_report_tables_")
  dir.create(folder)

  file_one <- file.path(folder, "User 0_all_gaze.csv")

  writeLines(
    c(
      "USER,MEDIA_ID,CNT,GSR_US,GSRV,HR,HRV,DIAL,DIALV,TTL0,TTLV,",
      "U1,1,1,2.0,1,70,1,0.1,1,1007,1,",
      "U1,1,2,2.2,1,72,1,0.2,1,1008,1,"
    ),
    file_one,
    useBytes = TRUE
  )

  workflow <- run_gazepoint_biometrics_workflow(
    path = folder,
    group_columns = c("USER", "MEDIA_ID")
  )

  tables <- create_gazepoint_biometrics_report_tables(workflow)

  expect_s3_class(tables, "gazepoint_biometrics_report_tables")
  expect_true(is.data.frame(tables$overview))
  expect_true(is.data.frame(tables$diagnostics))
  expect_true(is.data.frame(tables$channels))
  expect_true(is.data.frame(tables$quality))
  expect_true(is.data.frame(tables$sampling))
  expect_true(is.data.frame(tables$window_recommendations))
  expect_true(is.data.frame(tables$participant_recommendations))
  expect_true(is.data.frame(tables$ttl_events))

  expect_true("final_status" %in% names(tables$diagnostics))
  expect_true("signal" %in% names(tables$channels))
  expect_true("usable_pct" %in% names(tables$quality))
  expect_true("time_column" %in% names(tables$sampling))
  expect_true("recommendation" %in% names(tables$window_recommendations))
  expect_true("participant_recommendation" %in% names(tables$participant_recommendations))
  expect_true("ttl_value" %in% names(tables$ttl_events))
})


test_that("create_gazepoint_biometrics_report_tables works from separate components", {
  dat <- data.frame(
    GSR_US = c(1.1, 1.2),
    GSRV = c(1, 1),
    HR = c(70, 72),
    HRV = c(1, 1),
    DIAL = c(1, 1),
    DIALV = c(1, 1),
    CNT = c(1, 2)
  )

  validation <- validate_gazepoint_biometrics(dat)

  quality <- combine_gazepoint_tables(list(
    audit_gazepoint_gsr_quality(dat)
  ))

  sampling <- audit_gazepoint_biometric_sampling(
    dat,
    time_column = "CNT",
    time_unit = "samples"
  )

  diagnostics <- data.frame(
    final_status = "pass",
    diagnostic_reasons = "workflow diagnostics passed",
    validation_issue_count = 0,
    stringsAsFactors = FALSE
  )

  tables <- create_gazepoint_biometrics_report_tables(
    validation = validation,
    quality = quality,
    sampling = sampling,
    diagnostics = diagnostics
  )

  expect_true(is.data.frame(tables$overview))
  expect_true(is.data.frame(tables$diagnostics))
  expect_true(is.data.frame(tables$channels))
  expect_true(is.data.frame(tables$quality))
  expect_true(is.data.frame(tables$sampling))
  expect_true(is.data.frame(tables$window_recommendations))
  expect_true("message" %in% names(tables$window_recommendations))
  expect_equal(tables$diagnostics$final_status, "pass")
})


test_that("create_gazepoint_biometrics_report_tables returns placeholders when optional tables are missing", {
  validation <- validate_gazepoint_biometrics(
    data.frame(
      GSR_US = c(1.1, 1.2),
      GSRV = c(1, 1),
      CNT = c(1, 2)
    )
  )

  tables <- create_gazepoint_biometrics_report_tables(
    validation = validation
  )

  expect_true(is.data.frame(tables$diagnostics))
  expect_true(is.data.frame(tables$sampling))
  expect_true("message" %in% names(tables$diagnostics))
  expect_true("message" %in% names(tables$sampling))
})


test_that("create_gazepoint_biometrics_report_tables limits TTL event rows", {
  ttl_events <- data.frame(
    row_index = 1:5,
    event_order = 1:5,
    ttl_channel = "TTL0",
    ttl_value = 1001:1005,
    previous_ttl_value = c(NA, 1001:1004),
    CNT = 1:5,
    ttl_validity = 1
  )

  tables <- create_gazepoint_biometrics_report_tables(
    ttl_events = ttl_events,
    max_ttl_events = 2
  )

  expect_equal(nrow(tables$ttl_events), 2)
})


test_that("create_gazepoint_biometrics_report_tables rejects invalid workflow object", {
  expect_error(
    create_gazepoint_biometrics_report_tables(workflow = list()),
    "workflow"
  )
})

#' Denoise EDA using dependency-light Haar wavelet shrinkage
#'
#' Applies simple Haar wavelet soft-threshold denoising to EDA signals within
#' optional groups. This is a dependency-light wavelet denoising helper and
#' should not be described as an exact reproduction of stationary-wavelet
#' artifact-removal algorithms.
#'
#' @param dat A data frame containing EDA data.
#' @param eda_col EDA/conductance column.
#' @param group_cols Optional grouping columns.
#' @param output_col Optional output column.
#' @param levels Number of Haar decomposition levels.
#' @param threshold_multiplier Multiplier applied to the robust noise estimate.
#' @param overwrite Logical. If `FALSE`, protect existing output columns.
#'
#' @return A data frame with denoised EDA and denoising attributes.
#' @export
denoise_gazepoint_eda_wavelet <- function(dat,
                                          eda_col = "GSR_US",
                                          group_cols = NULL,
                                          output_col = NULL,
                                          levels = 3,
                                          threshold_multiplier = 1,
                                          overwrite = FALSE) {
  if (!is.data.frame(dat)) stop("`dat` must be a data frame.", call. = FALSE)
  if (!eda_col %in% names(dat)) stop("Column `", eda_col, "` was not found in `dat`.", call. = FALSE)
  if (!is.numeric(dat[[eda_col]])) stop("`eda_col` must identify a numeric column.", call. = FALSE)

  if (is.null(output_col)) output_col <- paste0(eda_col, "_wavelet_denoised")

  if (!isTRUE(overwrite) && output_col %in% names(dat)) {
    stop("Output column `", output_col, "` already exists. Use `overwrite = TRUE`.", call. = FALSE)
  }

  if (!is.numeric(levels) || length(levels) != 1 || levels < 1) {
    stop("`levels` must be a positive integer.", call. = FALSE)
  }

  if (!is.numeric(threshold_multiplier) || length(threshold_multiplier) != 1 || threshold_multiplier <= 0) {
    stop("`threshold_multiplier` must be positive.", call. = FALSE)
  }

  levels <- as.integer(levels)

  if (is.null(group_cols)) group_cols <- character()
  missing_groups <- setdiff(group_cols, names(dat))
  if (length(missing_groups) > 0) {
    stop("Missing `group_cols`: ", paste(missing_groups, collapse = ", "), call. = FALSE)
  }

  out <- dat
  out[[output_col]] <- NA_real_

  groups <- gpbiometrics_wavelet_split_indices(out, group_cols)

  rows <- lapply(names(groups), function(unit_id) {
    idx <- groups[[unit_id]]
    x <- out[[eda_col]][idx]
    finite <- is.finite(x)

    status <- "wavelet_denoised"

    if (sum(finite) < 8) {
      status <- "insufficient_finite_samples"
      return(data.frame(
        unit_id = unit_id,
        n_rows = length(idx),
        n_finite = sum(finite),
        threshold = NA_real_,
        levels_used = NA_integer_,
        status = status,
        stringsAsFactors = FALSE
      ))
    }

    x_fill <- gpbiometrics_wavelet_fill_linear(x)
    den <- gpbiometrics_haar_denoise(x_fill, levels = levels, threshold_multiplier = threshold_multiplier)
    den[!finite] <- NA_real_
    out[[output_col]][idx] <<- den$signal

    data.frame(
      unit_id = unit_id,
      n_rows = length(idx),
      n_finite = sum(finite),
      threshold = den$threshold,
      levels_used = den$levels_used,
      status = status,
      stringsAsFactors = FALSE
    )
  })

  summary_table <- do.call(rbind, rows)
  rownames(summary_table) <- NULL

  overview <- data.frame(
    input_rows = nrow(dat),
    group_count = length(groups),
    successful_groups = sum(summary_table$status == "wavelet_denoised"),
    problem_groups = sum(summary_table$status != "wavelet_denoised"),
    eda_col = eda_col,
    output_col = output_col,
    status = if (all(summary_table$status == "wavelet_denoised")) {
      "eda_wavelet_denoising_complete"
    } else if (any(summary_table$status == "wavelet_denoised")) {
      "eda_wavelet_denoising_partial"
    } else {
      "eda_wavelet_denoising_failed"
    },
    interpretation = paste(
      "Wavelet-denoised EDA is a signal-recovery aid.",
      "It should be compared with raw and artifact-flagged data and does not infer emotion, stress, cognition, or diagnosis."
    ),
    stringsAsFactors = FALSE
  )

  attr(out, "wavelet_denoising_overview") <- overview
  attr(out, "wavelet_denoising_summary") <- summary_table
  attr(out, "wavelet_denoising_settings") <- list(
    eda_col = eda_col,
    group_cols = group_cols,
    output_col = output_col,
    levels = levels,
    threshold_multiplier = threshold_multiplier,
    overwrite = overwrite
  )

  class(out) <- unique(c("gazepoint_eda_wavelet_denoised", class(out)))

  out
}

gpbiometrics_wavelet_split_indices <- function(dat, group_cols) {
  if (length(group_cols) == 0) return(list(all_rows = seq_len(nrow(dat))))
  gf <- dat[group_cols]
  gf[] <- lapply(gf, function(x) {
    x <- as.character(x)
    x[is.na(x)] <- "<NA>"
    x
  })
  split(seq_len(nrow(dat)), do.call(paste, c(gf, sep = " | ")))
}

gpbiometrics_wavelet_fill_linear <- function(x) {
  idx <- seq_along(x)
  finite <- is.finite(x)

  if (all(finite)) return(x)
  if (sum(finite) < 2) return(rep(mean(x[finite], na.rm = TRUE), length(x)))

  stats::approx(idx[finite], x[finite], xout = idx, rule = 2)$y
}

gpbiometrics_haar_denoise <- function(x, levels = 3, threshold_multiplier = 1) {
  original_n <- length(x)
  target_n <- 2^ceiling(log2(original_n))
  padded <- c(x, rep(stats::median(x, na.rm = TRUE), target_n - original_n))

  coeffs <- list()
  approx <- padded
  levels_used <- 0L

  for (lev in seq_len(levels)) {
    if (length(approx) < 2 || length(approx) %% 2 != 0) break

    a <- (approx[seq(1, length(approx), by = 2)] + approx[seq(2, length(approx), by = 2)]) / sqrt(2)
    d <- (approx[seq(1, length(approx), by = 2)] - approx[seq(2, length(approx), by = 2)]) / sqrt(2)

    coeffs[[lev]] <- d
    approx <- a
    levels_used <- lev
  }

  if (length(coeffs) == 0) {
    return(list(signal = x, threshold = NA_real_, levels_used = 0L))
  }

  finest <- coeffs[[1]]
  sigma <- stats::median(abs(finest - stats::median(finest, na.rm = TRUE)), na.rm = TRUE) / 0.6745
  threshold <- threshold_multiplier * sigma * sqrt(2 * log(length(padded)))

  coeffs <- lapply(coeffs, function(d) {
    sign(d) * pmax(abs(d) - threshold, 0)
  })

  recon <- approx

  for (lev in rev(seq_len(levels_used))) {
    d <- coeffs[[lev]]
    up <- numeric(length(d) * 2)
    up[seq(1, length(up), by = 2)] <- (recon + d) / sqrt(2)
    up[seq(2, length(up), by = 2)] <- (recon - d) / sqrt(2)
    recon <- up
  }

  list(signal = recon[seq_len(original_n)], threshold = threshold, levels_used = levels_used)
}

#' Extract FuzzyEn and Lorenz-plot CSI HRV features
#'
#' Computes fuzzy entropy and Lorenz/Poincare-derived cardiac sympathetic index
#' style descriptors from IBI/RR intervals.
#'
#' These outputs are nonlinear/geometric HRV descriptors. They do not infer
#' seizure status, diagnosis, health status, emotion, stress, or cognition.
#'
#' @param dat A data frame.
#' @param ibi_col Numeric IBI/RR interval column.
#' @param group_cols Optional grouping columns.
#' @param m Embedding dimension.
#' @param r_multiplier Tolerance multiplier applied to within-group SD.
#' @param fuzzy_power Fuzzy exponential power.
#' @param min_intervals Minimum intervals per group.
#'
#' @return A list with `overview`, `features`, and `settings`.
#' @export
extract_gazepoint_hrv_fuzzy_csi <- function(dat,
                                            ibi_col = "IBI",
                                            group_cols = NULL,
                                            m = 2,
                                            r_multiplier = 0.2,
                                            fuzzy_power = 2,
                                            min_intervals = 10) {
  if (!is.data.frame(dat)) {
    stop("`dat` must be a data frame.", call. = FALSE)
  }

  if (!ibi_col %in% names(dat)) {
    stop("Column `", ibi_col, "` was not found in `dat`.", call. = FALSE)
  }

  if (!is.numeric(dat[[ibi_col]])) {
    stop("`ibi_col` must identify a numeric column.", call. = FALSE)
  }

  if (is.null(group_cols)) {
    group_cols <- character()
  }

  missing_groups <- setdiff(group_cols, names(dat))
  if (length(missing_groups) > 0) {
    stop("Missing `group_cols`: ", paste(missing_groups, collapse = ", "), call. = FALSE)
  }

  groups <- gpbiometrics_fuzzy_split(dat, group_cols)

  rows <- lapply(names(groups), function(group_id) {
    idx <- groups[[group_id]]
    base <- gpbiometrics_fuzzy_group_values(dat, idx, group_cols, group_id)

    x <- dat[[ibi_col]][idx]
    x <- x[is.finite(x) & x > 0]

    if (length(x) < min_intervals || stats::sd(x) == 0) {
      return(data.frame(
        base,
        group_id = group_id,
        n_intervals = length(x),
        fuzzy_entropy = NA_real_,
        sd1 = NA_real_,
        sd2 = NA_real_,
        csi = NA_real_,
        cvi = NA_real_,
        modified_csi = NA_real_,
        status = "insufficient_intervals",
        stringsAsFactors = FALSE,
        check.names = FALSE
      ))
    }

    r <- r_multiplier * stats::sd(x)
    fuzzy_entropy <- gpbiometrics_fuzzy_entropy(
      x = x,
      m = m,
      r = r,
      fuzzy_power = fuzzy_power
    )

    dx <- diff(x)
    sdnn <- stats::sd(x)
    diff_var <- stats::var(dx, na.rm = TRUE)

    sd1 <- sqrt(diff_var / 2)
    sd2 <- sqrt(max((2 * sdnn^2) - (0.5 * diff_var), 0))

    csi <- if (is.finite(sd1) && sd1 > 0) {
      sd2 / sd1
    } else {
      NA_real_
    }

    cvi <- if (is.finite(sd1) && is.finite(sd2) && sd1 > 0 && sd2 > 0) {
      log10(sd1 * sd2)
    } else {
      NA_real_
    }

    modified_csi <- if (is.finite(sd1) && sd1 > 0 && is.finite(sd2)) {
      sd2^2 / sd1
    } else {
      NA_real_
    }

    data.frame(
      base,
      group_id = group_id,
      n_intervals = length(x),
      fuzzy_entropy = fuzzy_entropy,
      sd1 = sd1,
      sd2 = sd2,
      csi = csi,
      cvi = cvi,
      modified_csi = modified_csi,
      status = "fuzzy_csi_extracted",
      stringsAsFactors = FALSE,
      check.names = FALSE
    )
  })

  features <- do.call(rbind, rows)
  rownames(features) <- NULL

  overview <- data.frame(
    group_count = length(groups),
    feature_rows = nrow(features),
    successful_groups = sum(features$status == "fuzzy_csi_extracted"),
    problem_groups = sum(features$status != "fuzzy_csi_extracted"),
    status = if (all(features$status == "fuzzy_csi_extracted")) {
      "fuzzy_csi_extracted"
    } else if (any(features$status == "fuzzy_csi_extracted")) {
      "fuzzy_csi_partial"
    } else {
      "fuzzy_csi_failed"
    },
    interpretation = paste(
      "FuzzyEn and CSI are nonlinear/geometric HRV descriptors.",
      "They are not diagnostic labels and do not infer clinical or psychological state by themselves."
    ),
    stringsAsFactors = FALSE
  )

  structure(
    list(
      overview = overview,
      features = features,
      settings = list(
        ibi_col = ibi_col,
        group_cols = group_cols,
        m = m,
        r_multiplier = r_multiplier,
        fuzzy_power = fuzzy_power,
        min_intervals = min_intervals
      )
    ),
    class = c("gazepoint_hrv_fuzzy_csi", "list")
  )
}

gpbiometrics_fuzzy_split <- function(dat, group_cols) {
  if (length(group_cols) == 0) {
    return(list(all_rows = seq_len(nrow(dat))))
  }

  gf <- dat[group_cols]
  gf[] <- lapply(gf, function(x) {
    x <- as.character(x)
    x[is.na(x)] <- "<NA>"
    x
  })

  split(seq_len(nrow(dat)), do.call(paste, c(gf, sep = " | ")))
}

gpbiometrics_fuzzy_group_values <- function(dat, idx, group_cols, group_id) {
  if (length(group_cols) == 0) {
    return(data.frame(unit_label = group_id, stringsAsFactors = FALSE))
  }

  values <- lapply(group_cols, function(nm) as.character(dat[[nm]][idx[1]]))
  names(values) <- group_cols
  as.data.frame(values, stringsAsFactors = FALSE, optional = TRUE)
}

gpbiometrics_fuzzy_entropy <- function(x,
                                       m = 2,
                                       r,
                                       fuzzy_power = 2) {
  x <- x[is.finite(x)]

  if (length(x) <= m + 2 || !is.finite(r) || r <= 0) {
    return(NA_real_)
  }

  phi_m <- gpbiometrics_fuzzy_phi(x, m = m, r = r, fuzzy_power = fuzzy_power)
  phi_m1 <- gpbiometrics_fuzzy_phi(x, m = m + 1, r = r, fuzzy_power = fuzzy_power)

  if (!is.finite(phi_m) || !is.finite(phi_m1) || phi_m <= 0 || phi_m1 <= 0) {
    return(NA_real_)
  }

  log(phi_m) - log(phi_m1)
}

gpbiometrics_fuzzy_phi <- function(x,
                                   m,
                                   r,
                                   fuzzy_power = 2) {
  n <- length(x) - m + 1

  if (n <= 1) {
    return(NA_real_)
  }

  emb <- matrix(NA_real_, nrow = n, ncol = m)

  for (j in seq_len(m)) {
    emb[, j] <- x[j:(j + n - 1)]
  }

  emb <- emb - rowMeans(emb)

  values <- numeric()

  for (i in seq_len(n - 1)) {
    for (j in (i + 1):n) {
      d <- max(abs(emb[i, ] - emb[j, ]))
      values <- c(values, exp(-((d^fuzzy_power) / r)))
    }
  }

  if (length(values) == 0) {
    return(NA_real_)
  }

  mean(values, na.rm = TRUE)
}

#' Prepare Gazepoint IBI/RR data for RHRV
#'
#' Backward-compatible alias for [export_gazepoint_rhrv_input()]. This helper
#' prepares analysis-ready inter-beat interval input for optional RHRV workflows.
#' It does not call RHRV and does not derive HRV from Gazepoint's raw `HRV`
#' field. HRV features should be derived from genuine IBI/RR intervals.
#'
#' @inheritParams export_gazepoint_rhrv_input
#'
#' @return A list returned by [export_gazepoint_rhrv_input()].
#' @export
prepare_gazepoint_rhrv_input <- function(data,
                                         ibi_col = "IBI_clean_ms",
                                         group_cols = NULL,
                                         unit = c("auto", "ms", "seconds"),
                                         collapse_repeated_intervals = TRUE,
                                         repeated_tolerance_ms = 1e-8,
                                         min_ibi_ms = 300,
                                         max_ibi_ms = 2000,
                                         output_dir = NULL,
                                         prefix = "gazepoint_rhrv") {
  unit <- match.arg(unit)

  export_gazepoint_rhrv_input(
    data = data,
    ibi_col = ibi_col,
    group_cols = group_cols,
    unit = unit,
    collapse_repeated_intervals = collapse_repeated_intervals,
    repeated_tolerance_ms = repeated_tolerance_ms,
    min_ibi_ms = min_ibi_ms,
    max_ibi_ms = max_ibi_ms,
    output_dir = output_dir,
    prefix = prefix
  )
}

#' Standardize Gazepoint plot return contracts
#'
#' US-spelling compatibility wrapper around
#' [standardise_gazepoint_plot_contract()]. The wrapper accepts either a single
#' ggplot object or a list of ggplot objects. For a list of plots, `plot_data`,
#' `settings`, `interpretation_notes`, and `plot_type` may be supplied either as
#' single values applied to all plots or as same-length lists/vectors applied
#' elementwise.
#'
#' @param plot A ggplot object, or a list of ggplot objects.
#' @param plot_data Optional data frame, or a list of data frames when `plot` is
#'   a list.
#' @param settings Optional settings list, or a list of settings lists when
#'   `plot` is a list.
#' @param interpretation_notes Optional character vector, or a list/character
#'   vector of notes when `plot` is a list.
#' @param plot_type Optional plot-type label, or a character vector/list of
#'   labels when `plot` is a list.
#'
#' @return A standardized ggplot object, or a list of standardized ggplot
#'   objects.
#' @export
standardize_gazepoint_plot_contracts <- function(plot,
                                                 plot_data = NULL,
                                                 settings = list(),
                                                 interpretation_notes = NULL,
                                                 plot_type = NULL) {
  if (inherits(plot, "ggplot")) {
    return(standardise_gazepoint_plot_contract(
      plot = plot,
      plot_data = plot_data,
      settings = settings,
      interpretation_notes = interpretation_notes,
      plot_type = plot_type
    ))
  }

  if (!is.list(plot)) {
    stop("`plot` must be a ggplot object or a list of ggplot objects.", call. = FALSE)
  }

  n_plots <- length(plot)

  if (n_plots == 0) {
    return(list())
  }

  out <- lapply(seq_along(plot), function(i) {
    standardise_gazepoint_plot_contract(
      plot = plot[[i]],
      plot_data = gpbiometrics_alias_select_element(plot_data, i, n_plots),
      settings = gpbiometrics_alias_select_element(settings, i, n_plots),
      interpretation_notes = gpbiometrics_alias_select_element(
        interpretation_notes,
        i,
        n_plots
      ),
      plot_type = gpbiometrics_alias_select_element(plot_type, i, n_plots)
    )
  })

  names(out) <- names(plot)
  out
}

gpbiometrics_alias_select_element <- function(x, i, n) {
  if (is.null(x)) {
    return(NULL)
  }

  if (is.data.frame(x)) {
    return(x)
  }

  if (is.character(x) && length(x) == n) {
    return(x[[i]])
  }

  if (is.list(x) && length(x) == n) {
    return(x[[i]])
  }

  x
}

test_that("prepare_gazepoint_rhrv_input wraps export_gazepoint_rhrv_input", {
  dat <- data.frame(
    participant = c("p1", "p1", "p1", "p2", "p2", "p2"),
    IBI_clean_ms = c(800, 810, 790, 900, 890, 910)
  )

  old <- export_gazepoint_rhrv_input(
    dat,
    ibi_col = "IBI_clean_ms",
    group_cols = "participant"
  )

  new <- prepare_gazepoint_rhrv_input(
    dat,
    ibi_col = "IBI_clean_ms",
    group_cols = "participant"
  )

  expect_named(new, names(old))
  expect_equal(new$overview, old$overview)
  expect_equal(new$beat_table, old$beat_table)
  expect_equal(new$group_summary, old$group_summary)
  expect_equal(new$settings$ibi_col, "IBI_clean_ms")
})

test_that("prepare_gazepoint_rhrv_input preserves conservative IBI filtering", {
  dat <- data.frame(
    participant = "p1",
    IBI_clean_ms = c(250, 800, 1000, 2500)
  )

  out <- prepare_gazepoint_rhrv_input(
    dat,
    ibi_col = "IBI_clean_ms",
    group_cols = "participant",
    min_ibi_ms = 300,
    max_ibi_ms = 2000
  )

  expect_true(all(out$beat_table$ibi_ms >= 300))
  expect_true(all(out$beat_table$ibi_ms <= 2000))
  expect_equal(nrow(out$beat_table), 2)
})

test_that("standardize_gazepoint_plot_contracts handles one plot", {
  skip_if_not_installed("ggplot2")

  dat <- data.frame(x = 1:3, y = c(2, 4, 6))

  p <- ggplot2::ggplot(dat, ggplot2::aes(x = x, y = y)) +
    ggplot2::geom_point()

  out <- standardize_gazepoint_plot_contracts(
    plot = p,
    plot_data = dat,
    settings = list(plot_type = "single_alias_test"),
    interpretation_notes = "Compatibility alias test.",
    plot_type = "single_alias_test"
  )

  expect_s3_class(out, "gazepoint_plot")
  expect_true(isTRUE(attr(out, "gazepoint_plot_contract")))
  expect_equal(attr(out, "plot_type"), "single_alias_test")
  expect_equal(attr(out, "plot_data"), dat)
  expect_equal(attr(out, "interpretation_notes"), "Compatibility alias test.")
})

test_that("standardize_gazepoint_plot_contracts handles a list of plots", {
  skip_if_not_installed("ggplot2")

  dat1 <- data.frame(x = 1:3, y = c(2, 4, 6))
  dat2 <- data.frame(x = 1:3, y = c(3, 6, 9))

  p1 <- ggplot2::ggplot(dat1, ggplot2::aes(x = x, y = y)) +
    ggplot2::geom_point()

  p2 <- ggplot2::ggplot(dat2, ggplot2::aes(x = x, y = y)) +
    ggplot2::geom_line()

  out <- standardize_gazepoint_plot_contracts(
    plot = list(first = p1, second = p2),
    plot_data = list(dat1, dat2),
    settings = list(
      list(plot_type = "first_alias_test"),
      list(plot_type = "second_alias_test")
    ),
    interpretation_notes = c("First plot note.", "Second plot note."),
    plot_type = c("first_alias_test", "second_alias_test")
  )

  expect_type(out, "list")
  expect_length(out, 2)
  expect_named(out, c("first", "second"))

  expect_true(all(vapply(
    out,
    function(x) isTRUE(attr(x, "gazepoint_plot_contract")),
    logical(1)
  )))

  expect_equal(attr(out$first, "plot_type"), "first_alias_test")
  expect_equal(attr(out$second, "plot_type"), "second_alias_test")
  expect_equal(attr(out$first, "plot_data"), dat1)
  expect_equal(attr(out$second, "plot_data"), dat2)
})

test_that("prepare_gazepoint_rhrv_input handles groups with no valid IBI intervals", {
  dat <- data.frame(
    participant = c("p1", "p1", "p2", "p2"),
    IBI_clean_ms = c(800, 810, NA_real_, 2500)
  )

  out <- prepare_gazepoint_rhrv_input(
    dat,
    ibi_col = "IBI_clean_ms",
    group_cols = "participant",
    min_ibi_ms = 300,
    max_ibi_ms = 2000
  )

  expect_s3_class(out, "gazepoint_rhrv_input_export")
  expect_true(is.data.frame(out$beat_table))
  expect_true(is.data.frame(out$group_summary))
  expect_true(is.data.frame(out$manifest))
  expect_true(is.list(out$settings))

  expect_equal(out$settings$ibi_col, "IBI_clean_ms")
  expect_true(all(out$beat_table$ibi_ms >= 300))
  expect_true(all(out$beat_table$ibi_ms <= 2000))

  # The second group has no valid IBI values after filtering.
  # The important regression is that this no longer errors and that no invalid
  # intervals are emitted into the beat table.
  expect_false(any(is.na(out$beat_table$ibi_ms)))
  expect_false(any(out$beat_table$ibi_ms > 2000))
  expect_false(any(out$beat_table$ibi_ms < 300))

  # At least the valid participant should remain represented in exported beats.
  expect_true(any(out$beat_table$group_id == "p1"))
  expect_false(any(out$beat_table$group_id == "p2"))
})

#' Create a gpbiometrics feature inventory
#'
#' Creates a structured inventory of implemented gpbiometrics helper functions.
#' This is useful for reporting, readiness checks, documentation, and package
#' development audits.
#'
#' @param include_internal Logical. If `TRUE`, also checks for non-exported
#'   internal helper names when they are included in the inventory.
#'
#' @return A list with `overview`, `inventory`, `domain_summary`,
#'   `missing_expected`, and `settings`.
#' @export
create_gazepoint_biometrics_feature_inventory <- function(include_internal = FALSE) {
  if (!is.logical(include_internal) ||
      length(include_internal) != 1 ||
      is.na(include_internal)) {
    stop("`include_internal` must be TRUE or FALSE.", call. = FALSE)
  }

  inventory <- gpbiometrics_feature_inventory_table()

  inventory$available <- vapply(
    inventory$function_name,
    gpbiometrics_feature_inventory_exists,
    logical(1),
    include_internal = include_internal
  )

  inventory$status <- ifelse(
    inventory$available,
    "available",
    "missing"
  )

  domain_summary <- gpbiometrics_feature_inventory_domain_summary(inventory)

  missing_expected <- inventory[
    !inventory$available & inventory$expected,
    ,
    drop = FALSE
  ]

  overview <- data.frame(
    feature_rows = nrow(inventory),
    domain_count = length(unique(inventory$domain)),
    available_features = sum(inventory$available, na.rm = TRUE),
    missing_expected_features = nrow(missing_expected),
    include_internal = include_internal,
    status = if (nrow(missing_expected) == 0) {
      "feature_inventory_complete"
    } else {
      "warn_expected_features_missing"
    },
    stringsAsFactors = FALSE
  )

  structure(
    list(
      overview = overview,
      inventory = inventory,
      domain_summary = domain_summary,
      missing_expected = missing_expected,
      settings = list(
        include_internal = include_internal,
        interpretation_notes = c(
          "The inventory checks whether major user-facing gpbiometrics helpers are available.",
          "A missing expected feature indicates a documentation or implementation mismatch.",
          "The inventory supports reporting and readiness-polish workflows; it does not analyse data."
        )
      )
    ),
    class = c("gazepoint_biometrics_feature_inventory", "list")
  )
}

gpbiometrics_feature_inventory_table <- function() {
  expected_by_domain <- list(
    import_and_schema = c(
      "import_gazepoint_biometrics",
      "import_gazepoint_biometric_folder",
      "import_gazepoint_data_summary",
      "check_gazepoint_biometric_columns",
      "import_gazepoint_lsl_xdf",
      "standardise_gazepoint_biometric_names",
      "detect_gazepoint_biometric_schema",
      "detect_gazepoint_time_columns",
      "simulate_gazepoint_biometrics",
      "detect_active_biometric_channels"
    ),


    quality_and_readiness = c(
      "validate_gazepoint_biometrics",
      "summarise_gazepoint_biometric_validity",
      "audit_gazepoint_biometric_missingness",
      "flag_gazepoint_biometric_dropouts",
      "audit_gazepoint_biometric_sampling",
      "audit_gazepoint_signal_activity",
      "audit_gazepoint_time_resets",
      "recommend_gazepoint_biometric_exclusions",
      "run_gazepoint_biometrics_real_data_readiness",
      "audit_gazepoint_distributional_drift",
      "audit_gazepoint_gsr_units",
      "prepare_gazepoint_artifact_svm_features",
      "flag_gazepoint_mad_artifacts",
      "flag_gazepoint_artifacts_svm"
    ),

    preprocessing = c(
      "standardise_gazepoint_zscore",
      "standardize_gazepoint_zscore",
      "standardise_gazepoint_range_correction",
      "standardize_gazepoint_range_correction",
      "baseline_correct_gazepoint_pupil",
      "standardise_gazepoint_adaptive_ema",
      "standardize_gazepoint_adaptive_ema",
      "baseline_correct_gazepoint_gsr",
      "baseline_correct_gazepoint_hr",
      "smooth_gazepoint_biometrics",
      "convert_gazepoint_gsr_to_conductance",
      "decompose_gazepoint_eda",
      "denoise_gazepoint_eda_wavelet",
      "denoise_gazepoint_eda_autoencoder",
      "denoise_gazepoint_ppg_autoencoder",
      "denoise_gazepoint_quantization_noise",
      "correct_gazepoint_eda_temperature",
      "audit_gazepoint_stabilization_period",
      "regress_gazepoint_pupil_luminance",
      "standardize_gazepoint_biometrics_within_unit",
      "standardise_gazepoint_biometrics_within_unit"
    ),

    eda_scr = c(
      "classify_gazepoint_eda_response_pattern",
      "classify_gazepoint_scr_intervals",
      "flag_kleckner_eda_artifacts",
      "audit_gazepoint_gsr_quality",
      "audit_gazepoint_eda_artifacts",
      "detect_gazepoint_scr_events",
      "detect_gazepoint_scr_peaks",
      "summarise_gazepoint_scr_event_windows",
      "prepare_gazepoint_scr_hurdle_model_data",
      "run_gazepoint_scr_threshold_sensitivity",
      "run_gazepoint_scr_multiverse",
      "screen_gazepoint_eda_nonresponders",
      "summarise_gazepoint_gsr_windows",
      "summarise_gazepoint_gsr_tonic_phasic",
      "extract_gazepoint_eda_spectral_power",
      "extract_gazepoint_eda_complexity",
      "extract_gazepoint_eda_tvsymp",
      "plot_gazepoint_eda_decomposition",
      "optimize_gazepoint_cvxeda_tau",
      "model_gazepoint_eda_point_process",
      "extract_gazepoint_bilateral_eda_asymmetry",
      "analyze_gazepoint_skin_potential",
      "analyze_gazepoint_ac_susceptance",
      "detect_gazepoint_doubly_stochastic_changepoints",
      "extract_gazepoint_scr_recovery_times",
      "plot_gazepoint_scr_events"
    ),

    ibi_hr_hrv = c(
      "audit_gazepoint_hr_quality",
      "assess_gazepoint_hrp_waveform_quality",
      "audit_gazepoint_ibi_quality",
      "filter_gazepoint_ibi_implausible",
      "compare_gazepoint_hr_ibi_consistency",
      "summarise_gazepoint_hr_windows",
      "summarise_gazepoint_ibi_windows",
      "summarise_gazepoint_ibi_hrv_windows",
      "extract_gazepoint_hrv_features",
      "extract_gazepoint_pdr_signals",
      "calculate_gazepoint_rsa",
      "test_gazepoint_hrv_nonlinearity",
      "extract_gazepoint_hrv_rqa",
      "extract_gazepoint_hrv_geometric",
      "extract_gazepoint_hrv_fragmentation",
      "extract_gazepoint_hrv_asymmetry",
      "model_gazepoint_hr_point_process",
      "analyze_gazepoint_cardiorespiratory_causality",
      "extract_gazepoint_edr_pca",
      "extract_gazepoint_hrv_rcmse",
      "extract_gazepoint_respiration_ceemdan",
      "fuse_gazepoint_respiration_kalman",
      "extract_gazepoint_hrv_fuzzy_csi",
      "extract_gazepoint_beats_kmeans",
      "model_gazepoint_hrv_ipfm",
      "extract_gazepoint_hrv_nonlinear"
    ),

    ttl_alignment = c(
      "extract_gazepoint_ttl_events",
      "align_gazepoint_biometrics_to_ttl",
      "estimate_gazepoint_signal_lag",
      "audit_gazepoint_biometric_sync_drift",
      "plot_gazepoint_multimodal_timeline"
    ),

    aoi_biometrics = c(
      "summarise_gazepoint_aoi_biometrics",
      "prepare_gazepoint_aoi_biometrics_model_data",
      "plot_gazepoint_aoi_biometrics"
    ),

    modelling_and_windows = c(
      "summarise_gazepoint_engagement_windows",
      "summarise_gazepoint_dial_windows",
      "summarise_gazepoint_multimodal_windows",
      "summarise_gazepoint_full_biometric_windows",
      "sync_gazepoint_biometrics_with_gaze",
      "join_gazepoint_biometrics_to_master",
      "chunk_gazepoint_biometrics",
      "run_gazepoint_online_design_optimization",
      "prepare_gazepoint_multimodal_model_data",
      "prepare_gazepoint_biometrics_lme_data"
    ),

    reporting = c(
      "create_gazepoint_preregistration_template",
      "run_gpbiometrics_shiny",
      "run_gpbiometrics_shiny_annotator",
      "run_gazepoint_biometrics_workflow",
      "summarise_gazepoint_biometrics_workflow",
      "diagnose_gazepoint_biometrics_workflow",
      "create_gazepoint_biometrics_checklist",
      "create_gazepoint_biometrics_methods_text",
      "create_gazepoint_eda_analysis_pipeline",
      "run_gazepoint_eda_analysis_pipeline",
      "create_gazepoint_biometrics_report_tables",
      "write_gazepoint_biometrics_report_tables",
      "create_gazepoint_biometrics_report",
      "run_gazepoint_automated_statistics",
      "export_gazepoint_biometrics_report_bundle"
    ),

    interoperability = c(
      "export_gazepoint_rhrv_input",
      "prepare_gazepoint_rhrv_input",
      "prepare_gazepoint_pyppg_input",
      "prepare_gazepoint_neurokit_eda_input",
      "run_gazepoint_neurokit_eda_crosscheck",
      "prepare_gazepoint_ledalab_input",
      "prepare_gazepoint_pspm_input",
      "prepare_gazepoint_pspm_dcm_input",
      "prepare_gazepoint_ctsi_input",
      "prepare_gazepoint_cvxeda_input"
    ),

    plotting = c(
      "plot_gazepoint_biometric_signals",
      "plot_gazepoint_biometric_quality",
      "plot_gazepoint_eda_decomposition",
      "plot_gazepoint_scr_events",
      "plot_gazepoint_multimodal_timeline",
      "plot_gazepoint_signal_activity",
      "plot_gazepoint_time_resets",
      "plot_gazepoint_biometric_report_dashboard",
      "plot_gazepoint_saccade_main_sequence",
      "standardise_gazepoint_plot_contract",
      "standardize_gazepoint_plot_contracts",
      "check_gazepoint_plot_contract",
      "plot_gazepoint_scr_specification_curve",
      "plot_gazepoint_eda_gram",
      "get_gazepoint_plot_data"
    )


  )

  inventory <- data.frame(
    domain = rep(names(expected_by_domain), lengths(expected_by_domain)),
    function_name = unlist(expected_by_domain, use.names = FALSE),
    expected = TRUE,
    user_facing = TRUE,
    stringsAsFactors = FALSE
  )

  rownames(inventory) <- NULL
  inventory
}

gpbiometrics_feature_inventory_exists <- function(function_name,
                                                  include_internal = FALSE) {
  ns <- tryCatch(
    asNamespace("gpbiometrics"),
    error = function(e) NULL
  )

  if (!is.null(ns) &&
      exists(function_name, envir = ns, mode = "function", inherits = FALSE)) {
    return(TRUE)
  }

  if (exists(function_name, mode = "function", inherits = TRUE)) {
    return(TRUE)
  }

  if (isTRUE(include_internal) && !is.null(ns)) {
    return(exists(function_name, envir = ns, mode = "function", inherits = TRUE))
  }

  FALSE
}

gpbiometrics_feature_inventory_domain_summary <- function(inventory) {
  domains <- unique(inventory$domain)

  out <- lapply(domains, function(domain) {
    d <- inventory[inventory$domain == domain, , drop = FALSE]


    data.frame(
      domain = domain,
      feature_count = nrow(d),
      available_features = sum(d$available, na.rm = TRUE),
      missing_features = sum(!d$available, na.rm = TRUE),
      completion_rate = if (nrow(d) > 0) {
        sum(d$available, na.rm = TRUE) / nrow(d)
      } else {
        NA_real_
      },
      status = if (all(d$available)) {
        "complete"
      } else if (any(d$available)) {
        "partial"
      } else {
        "missing"
      },
      stringsAsFactors = FALSE
    )


  })

  out <- do.call(rbind, out)
  rownames(out) <- NULL

  out
}

test_that("create_gazepoint_biometrics_feature_inventory returns structured outputs", {
  res <- create_gazepoint_biometrics_feature_inventory()

  expect_s3_class(res, "gazepoint_biometrics_feature_inventory")
  expect_true(is.data.frame(res$overview))
  expect_true(is.data.frame(res$inventory))
  expect_true(is.data.frame(res$domain_summary))
  expect_true(is.data.frame(res$missing_expected))
  expect_true(is.list(res$settings))

  expect_true(all(c(
    "domain",
    "function_name",
    "available",
    "status"
  ) %in% names(res$inventory)))

  expect_true(all(c(
    "domain",
    "feature_count",
    "available_features",
    "missing_features",
    "completion_rate",
    "status"
  ) %in% names(res$domain_summary)))
})

test_that("feature inventory includes recent helper domains", {
  res <- create_gazepoint_biometrics_feature_inventory()

  expect_true("aoi_biometrics" %in% res$inventory$domain)
  expect_true("interoperability" %in% res$inventory$domain)
  expect_true("plotting" %in% res$inventory$domain)
  expect_true("eda_scr" %in% res$inventory$domain)
  expect_true("ibi_hr_hrv" %in% res$inventory$domain)

  expect_true("summarise_gazepoint_aoi_biometrics" %in% res$inventory$function_name)
  expect_true("export_gazepoint_rhrv_input" %in% res$inventory$function_name)
  expect_true("standardise_gazepoint_plot_contract" %in% res$inventory$function_name)
})

test_that("feature inventory detects available exported functions", {
  res <- create_gazepoint_biometrics_feature_inventory()

  selected <- res$inventory[
    res$inventory$function_name %in% c(
      "summarise_gazepoint_aoi_biometrics",
      "prepare_gazepoint_aoi_biometrics_model_data",
      "plot_gazepoint_aoi_biometrics",
      "export_gazepoint_rhrv_input",
      "prepare_gazepoint_neurokit_eda_input",
      "run_gazepoint_neurokit_eda_crosscheck",
      "standardise_gazepoint_plot_contract",
      "check_gazepoint_plot_contract"
    ),
    ,
    drop = FALSE
  ]

  expect_equal(nrow(selected), 8)
  expect_true(all(selected$available))
  expect_true(all(selected$status == "available"))
})

test_that("feature inventory domain summary is internally consistent", {
  res <- create_gazepoint_biometrics_feature_inventory()

  expect_equal(
    sum(res$domain_summary$feature_count),
    nrow(res$inventory)
  )

  expect_equal(
    sum(res$domain_summary$available_features),
    sum(res$inventory$available)
  )

  expect_true(all(res$domain_summary$completion_rate >= 0))
  expect_true(all(res$domain_summary$completion_rate <= 1))
})

test_that("feature inventory validates inputs", {
  expect_error(
    create_gazepoint_biometrics_feature_inventory(include_internal = NA),
    "`include_internal` must be TRUE or FALSE"
  )
})

#' Run Gazepoint SCR threshold sensitivity checks
#'
#' Re-runs SCR peak detection across combinations of amplitude thresholds and
#' minimum peak-distance settings. Optionally, it also carries each peak-detection
#' result through SCR event-window summaries so users can see how preprocessing
#' choices affect event-level response rates.
#'
#' @param data A Gazepoint biometric data frame.
#' @param phasic_col Optional phasic EDA signal column, typically
#'   `GSR_US_PHASIC`.
#' @param signal_col Optional conductance-like fallback signal column, typically
#'   `GSR_US`.
#' @param time_col Optional time/counter column.
#' @param group_cols Optional grouping columns.
#' @param amplitude_min_values Numeric vector of SCR amplitude thresholds.
#' @param min_peak_distance_values Numeric vector of minimum peak distances.
#' @param recovery_fraction Fraction used by `detect_gazepoint_scr_peaks()` to
#'   define recovery.
#' @param smooth_width Optional moving-average width passed to
#'   `detect_gazepoint_scr_peaks()`.
#' @param events Optional event table passed to
#'   `summarise_gazepoint_scr_event_windows()`.
#' @param event_time_col Optional event-time column when `events` is supplied.
#' @param event_id_col Optional event identifier column when `events` is supplied.
#' @param event_label_col Optional event label column when `events` is supplied.
#' @param ttl_cols Optional TTL columns for event derivation when `events` is
#'   `NULL`.
#' @param ttl_valid_col Optional TTL validity column.
#' @param event_detection Event-detection rule for TTL columns.
#' @param analysis_window Event-relative analysis window.
#' @param response_window Event-relative response window.
#' @param peak_selection Peak-selection rule for event-window summaries.
#' @param collapse_simultaneous_events Logical. Passed to
#'   `summarise_gazepoint_scr_event_windows()`.
#' @param include_event_windows Logical. If `TRUE`, compute event-window
#'   summaries for each sensitivity setting.
#' @param keep_objects Logical. If `TRUE`, retain peak-detection and
#'   event-window objects in list columns.
#'
#' @return A list with `overview`, `sensitivity_grid`,
#'   `peak_group_summary`, `event_window_summary`, optional `objects`, and
#'   `settings`.
#' @export
run_gazepoint_scr_threshold_sensitivity <- function(data,
                                                    phasic_col = NULL,
                                                    signal_col = NULL,
                                                    time_col = NULL,
                                                    group_cols = NULL,
                                                    amplitude_min_values = c(0.005, 0.01, 0.02, 0.03),
                                                    min_peak_distance_values = c(1, 5, 10, 20, 30),
                                                    recovery_fraction = 0.5,
                                                    smooth_width = 1,
                                                    events = NULL,
                                                    event_time_col = NULL,
                                                    event_id_col = NULL,
                                                    event_label_col = NULL,
                                                    ttl_cols = NULL,
                                                    ttl_valid_col = NULL,
                                                    event_detection = c("rising", "active"),
                                                    analysis_window = c(0, 6),
                                                    response_window = c(1, 4),
                                                    peak_selection = c("largest_amplitude", "first_peak"),
                                                    collapse_simultaneous_events = FALSE,
                                                    include_event_windows = TRUE,
                                                    keep_objects = FALSE) {
  event_detection <- match.arg(event_detection)
  peak_selection <- match.arg(peak_selection)

  if (!is.data.frame(data)) {
    stop("`data` must be a data frame.", call. = FALSE)
  }

  if (!is.numeric(amplitude_min_values) ||
      length(amplitude_min_values) == 0 ||
      any(!is.finite(amplitude_min_values)) ||
      any(amplitude_min_values < 0)) {
    stop("`amplitude_min_values` must be a non-empty non-negative numeric vector.", call. = FALSE)
  }

  if (!is.numeric(min_peak_distance_values) ||
      length(min_peak_distance_values) == 0 ||
      any(!is.finite(min_peak_distance_values)) ||
      any(min_peak_distance_values < 1)) {
    stop("`min_peak_distance_values` must be a non-empty numeric vector with values >= 1.", call. = FALSE)
  }

  if (!is.logical(include_event_windows) ||
      length(include_event_windows) != 1 ||
      is.na(include_event_windows)) {
    stop("`include_event_windows` must be TRUE or FALSE.", call. = FALSE)
  }

  if (!is.logical(keep_objects) ||
      length(keep_objects) != 1 ||
      is.na(keep_objects)) {
    stop("`keep_objects` must be TRUE or FALSE.", call. = FALSE)
  }

  amplitude_min_values <- sort(unique(amplitude_min_values))
  min_peak_distance_values <- sort(unique(as.integer(round(min_peak_distance_values))))

  grid <- expand.grid(
    amplitude_min = amplitude_min_values,
    min_peak_distance = min_peak_distance_values,
    KEEP.OUT.ATTRS = FALSE,
    stringsAsFactors = FALSE
  )

  sensitivity_rows <- list()
  peak_group_rows <- list()
  event_window_rows <- list()
  kept_objects <- list()

  for (i in seq_len(nrow(grid))) {
    amp <- grid$amplitude_min[i]
    dist <- grid$min_peak_distance[i]

    peak_result <- tryCatch(
      detect_gazepoint_scr_peaks(
        data = data,
        signal_col = signal_col,
        phasic_col = phasic_col,
        time_col = time_col,
        group_cols = group_cols,
        amplitude_min = amp,
        recovery_fraction = recovery_fraction,
        smooth_width = smooth_width,
        min_peak_distance = dist
      ),
      error = function(e) e
    )

    if (inherits(peak_result, "error")) {
      sensitivity_rows[[length(sensitivity_rows) + 1L]] <- data.frame(
        amplitude_min = amp,
        min_peak_distance = dist,
        source_signal = NA_character_,
        candidate_peaks = NA_integer_,
        detected_peaks = NA_integer_,
        incomplete_recovery_peaks = NA_integer_,
        groups_with_detected_peaks = NA_integer_,
        event_count = NA_integer_,
        response_events = NA_integer_,
        response_rate = NA_real_,
        event_window_status = NA_character_,
        status = "peak_detection_error",
        error_message = conditionMessage(peak_result),
        stringsAsFactors = FALSE
      )
      next
    }

    peak_group <- peak_result$group_summary
    peak_group$amplitude_min <- amp
    peak_group$min_peak_distance <- dist
    peak_group_rows[[length(peak_group_rows) + 1L]] <- peak_group

    event_result <- NULL
    event_status <- NA_character_
    event_count <- NA_integer_
    response_events <- NA_integer_
    response_rate <- NA_real_

    if (isTRUE(include_event_windows)) {
      event_result <- tryCatch(
        summarise_gazepoint_scr_event_windows(
          data = data,
          scr_peaks = peak_result,
          events = events,
          time_col = time_col,
          event_time_col = event_time_col,
          event_id_col = event_id_col,
          event_label_col = event_label_col,
          group_cols = group_cols,
          ttl_cols = ttl_cols,
          ttl_valid_col = ttl_valid_col,
          event_detection = event_detection,
          analysis_window = analysis_window,
          response_window = response_window,
          peak_selection = peak_selection,
          collapse_simultaneous_events = collapse_simultaneous_events
        ),
        error = function(e) e
      )

      if (inherits(event_result, "error")) {
        event_status <- "event_window_error"
      } else {
        event_status <- event_result$overview$status
        event_count <- event_result$overview$event_count
        response_events <- event_result$overview$response_events
        response_rate <- event_result$overview$response_rate

        event_summary <- event_result$window_qc
        event_summary$amplitude_min <- amp
        event_summary$min_peak_distance <- dist
        event_window_rows[[length(event_window_rows) + 1L]] <- event_summary
      }
    }

    sensitivity_rows[[length(sensitivity_rows) + 1L]] <- data.frame(
      amplitude_min = amp,
      min_peak_distance = dist,
      source_signal = peak_result$overview$source_signal,
      candidate_peaks = peak_result$overview$candidate_peaks,
      detected_peaks = peak_result$overview$detected_peaks,
      incomplete_recovery_peaks = sum(
        peak_result$group_summary$incomplete_recovery_peaks,
        na.rm = TRUE
      ),
      groups_with_detected_peaks = sum(
        peak_result$group_summary$detected_peaks > 0,
        na.rm = TRUE
      ),
      event_count = event_count,
      response_events = response_events,
      response_rate = response_rate,
      event_window_status = event_status,
      status = "sensitivity_completed",
      error_message = NA_character_,
      stringsAsFactors = FALSE
    )

    if (isTRUE(keep_objects)) {
      kept_objects[[paste0("amp_", amp, "_dist_", dist)]] <- list(
        peaks = peak_result,
        event_windows = if (inherits(event_result, "error")) NULL else event_result
      )
    }
  }

  sensitivity_grid <- do.call(rbind, sensitivity_rows)
  rownames(sensitivity_grid) <- NULL

  peak_group_summary <- if (length(peak_group_rows) > 0) {
    out <- do.call(rbind, peak_group_rows)
    rownames(out) <- NULL
    out
  } else {
    data.frame()
  }

  event_window_summary <- if (length(event_window_rows) > 0) {
    out <- do.call(rbind, event_window_rows)
    rownames(out) <- NULL
    out
  } else {
    data.frame()
  }

  completed_rows <- sum(sensitivity_grid$status == "sensitivity_completed", na.rm = TRUE)
  error_rows <- sum(sensitivity_grid$status != "sensitivity_completed", na.rm = TRUE)

  overview <- data.frame(
    grid_rows = nrow(sensitivity_grid),
    completed_rows = completed_rows,
    error_rows = error_rows,
    amplitude_min_count = length(amplitude_min_values),
    min_peak_distance_count = length(min_peak_distance_values),
    include_event_windows = include_event_windows,
    keep_objects = keep_objects,
    status = if (completed_rows == 0) {
      "fail_no_sensitivity_rows_completed"
    } else if (error_rows > 0) {
      "warn_some_sensitivity_rows_failed"
    } else {
      "scr_threshold_sensitivity_completed"
    },
    stringsAsFactors = FALSE
  )

  structure(
    list(
      overview = overview,
      sensitivity_grid = sensitivity_grid,
      peak_group_summary = peak_group_summary,
      event_window_summary = event_window_summary,
      objects = if (isTRUE(keep_objects)) kept_objects else NULL,
      settings = list(
        phasic_col = phasic_col,
        signal_col = signal_col,
        time_col = time_col,
        group_cols = group_cols,
        amplitude_min_values = amplitude_min_values,
        min_peak_distance_values = min_peak_distance_values,
        recovery_fraction = recovery_fraction,
        smooth_width = smooth_width,
        ttl_cols = ttl_cols,
        ttl_valid_col = ttl_valid_col,
        event_detection = event_detection,
        analysis_window = analysis_window,
        response_window = response_window,
        peak_selection = peak_selection,
        collapse_simultaneous_events = collapse_simultaneous_events,
        include_event_windows = include_event_windows,
        keep_objects = keep_objects,
        interpretation_notes = c(
          "Sensitivity summaries show how SCR response counts vary across detection settings.",
          "They are intended for robustness assessment and reporting, not for selecting thresholds after inspecting desired results.",
          "EDA/SCR features should be interpreted as electrodermal response features, not emotional valence."
        )
      )
    ),
    class = c("gazepoint_scr_threshold_sensitivity", "list")
  )
}

test_that("run_gazepoint_scr_threshold_sensitivity creates full parameter grid", {
  dat <- data.frame(
    participant = "P1",
    CNT = seq_len(20),
    GSR_US_PHASIC = c(rep(0, 5), 0.02, 0.08, 0.02, rep(0, 12))
  )

  res <- run_gazepoint_scr_threshold_sensitivity(
    dat,
    phasic_col = "GSR_US_PHASIC",
    time_col = "CNT",
    group_cols = "participant",
    amplitude_min_values = c(0.01, 0.05),
    min_peak_distance_values = c(1, 5),
    include_event_windows = FALSE
  )

  expect_s3_class(res, "gazepoint_scr_threshold_sensitivity")
  expect_equal(res$overview$grid_rows, 4)
  expect_equal(nrow(res$sensitivity_grid), 4)
  expect_true(all(res$sensitivity_grid$status == "sensitivity_completed"))
})

test_that("run_gazepoint_scr_threshold_sensitivity shows stricter amplitude thresholds reduce detections", {
  dat <- data.frame(
    participant = "P1",
    CNT = seq_len(20),
    GSR_US_PHASIC = c(rep(0, 5), 0.02, 0.08, 0.02, rep(0, 12))
  )

  res <- run_gazepoint_scr_threshold_sensitivity(
    dat,
    phasic_col = "GSR_US_PHASIC",
    time_col = "CNT",
    group_cols = "participant",
    amplitude_min_values = c(0.01, 0.10),
    min_peak_distance_values = 1,
    include_event_windows = FALSE
  )

  low <- res$sensitivity_grid$detected_peaks[
    res$sensitivity_grid$amplitude_min == 0.01
  ]

  high <- res$sensitivity_grid$detected_peaks[
    res$sensitivity_grid$amplitude_min == 0.10
  ]

  expect_true(low > high)
})

test_that("run_gazepoint_scr_threshold_sensitivity can include event-window summaries", {
  dat <- data.frame(
    participant = "P1",
    CNT = seq_len(20),
    GSR_US_PHASIC = c(rep(0, 5), 0.02, 0.08, 0.02, rep(0, 12))
  )

  events <- data.frame(
    participant = "P1",
    event_time = 5
  )

  res <- run_gazepoint_scr_threshold_sensitivity(
    dat,
    phasic_col = "GSR_US_PHASIC",
    time_col = "CNT",
    group_cols = "participant",
    amplitude_min_values = c(0.01, 0.05),
    min_peak_distance_values = 1,
    events = events,
    event_time_col = "event_time",
    analysis_window = c(0, 6),
    response_window = c(1, 4),
    include_event_windows = TRUE
  )

  expect_equal(res$overview$grid_rows, 2)
  expect_true(all(res$sensitivity_grid$event_count == 1))
  expect_true(all(res$sensitivity_grid$response_events >= 0))
  expect_true(nrow(res$event_window_summary) >= 1)
})

test_that("run_gazepoint_scr_threshold_sensitivity supports collapsed TTL event windows", {
  dat <- data.frame(
    participant = "P1",
    CNT = seq_len(20),
    TTL0 = c(0, 1, 1, rep(0, 17)),
    TTL1 = c(0, 1, 1, rep(0, 17)),
    GSR_US_PHASIC = c(rep(0, 3), 0.02, 0.08, 0.02, rep(0, 14))
  )

  res <- run_gazepoint_scr_threshold_sensitivity(
    dat,
    phasic_col = "GSR_US_PHASIC",
    time_col = "CNT",
    group_cols = "participant",
    amplitude_min_values = 0.01,
    min_peak_distance_values = 1,
    ttl_cols = c("TTL0", "TTL1"),
    analysis_window = c(0, 6),
    response_window = c(1, 4),
    collapse_simultaneous_events = TRUE,
    include_event_windows = TRUE
  )

  expect_equal(res$sensitivity_grid$event_count, 1)
  expect_equal(res$sensitivity_grid$response_events, 1)
})

test_that("run_gazepoint_scr_threshold_sensitivity can retain objects", {
  dat <- data.frame(
    participant = "P1",
    CNT = seq_len(20),
    GSR_US_PHASIC = c(rep(0, 5), 0.02, 0.08, 0.02, rep(0, 12))
  )

  res <- run_gazepoint_scr_threshold_sensitivity(
    dat,
    phasic_col = "GSR_US_PHASIC",
    time_col = "CNT",
    group_cols = "participant",
    amplitude_min_values = 0.01,
    min_peak_distance_values = 1,
    include_event_windows = FALSE,
    keep_objects = TRUE
  )

  expect_true(is.list(res$objects))
  expect_equal(length(res$objects), 1)
  expect_s3_class(res$objects[[1]]$peaks, "gazepoint_scr_peak_detection")
})

test_that("run_gazepoint_scr_threshold_sensitivity validates inputs", {
  dat <- data.frame(
    CNT = seq_len(5),
    GSR_US_PHASIC = rep(0, 5)
  )

  expect_error(
    run_gazepoint_scr_threshold_sensitivity(
      dat,
      amplitude_min_values = -1
    ),
    "`amplitude_min_values`"
  )

  expect_error(
    run_gazepoint_scr_threshold_sensitivity(
      dat,
      min_peak_distance_values = 0
    ),
    "`min_peak_distance_values`"
  )
})

#' Launch a lightweight gpbiometrics Shiny peak/artifact annotator
#'
#' Launches a local Shiny app for manual EDA peak/artifact annotation. The app
#' lets users upload a Gazepoint CSV, inspect a selected signal, enter manual
#' peak/artifact intervals, and download annotation CSVs.
#'
#' This is an optional local GUI helper. It does not replace automated scoring
#' or infer emotion, stress, cognition, trust, preference, or diagnosis.
#'
#' @return A Shiny application object, launched for interactive use.
#' @export
run_gpbiometrics_shiny_annotator <- function() {
  if (!requireNamespace("shiny", quietly = TRUE)) {
    stop(
      "The optional package `shiny` is required. Install it with install.packages('shiny').",
      call. = FALSE
    )
  }

  ui <- shiny::fluidPage(
    shiny::titlePanel("gpbiometrics EDA annotator"),
    shiny::sidebarLayout(
      shiny::sidebarPanel(
        shiny::fileInput("file", "Upload Gazepoint CSV", accept = c(".csv", ".txt")),
        shiny::textInput("time_col", "Time column", value = "CNT"),
        shiny::textInput("signal_col", "EDA signal column", value = "GSR_US"),
        shiny::numericInput("peak_time", "Manual peak time", value = NA),
        shiny::numericInput("artifact_start", "Artifact start", value = NA),
        shiny::numericInput("artifact_end", "Artifact end", value = NA),
        shiny::textInput("note", "Annotation note", value = ""),
        shiny::actionButton("add_peak", "Add manual peak"),
        shiny::actionButton("add_artifact", "Add artifact interval"),
        shiny::downloadButton("download_annotations", "Download annotations")
      ),
      shiny::mainPanel(
        shiny::plotOutput("eda_plot", click = "plot_click", height = "350px"),
        shiny::h4("Click position"),
        shiny::verbatimTextOutput("click_info"),
        shiny::h4("Annotations"),
        shiny::tableOutput("annotation_table"),
        shiny::h4("Interpretation guardrail"),
        shiny::verbatimTextOutput("guardrail")
      )
    )
  )

  server <- function(input, output, session) {
    dat <- shiny::reactive({
      req <- input$file
      if (is.null(req)) return(NULL)
      utils::read.csv(req$datapath, stringsAsFactors = FALSE, check.names = FALSE)
    })

    annotations <- shiny::reactiveVal(data.frame(
      annotation_type = character(),
      time = numeric(),
      start = numeric(),
      end = numeric(),
      note = character(),
      stringsAsFactors = FALSE
    ))

    output$eda_plot <- shiny::renderPlot({
      x <- dat()
      if (is.null(x)) return(NULL)
      if (!input$time_col %in% names(x) || !input$signal_col %in% names(x)) return(NULL)

      graphics::plot(
        x[[input$time_col]],
        x[[input$signal_col]],
        type = "l",
        xlab = input$time_col,
        ylab = input$signal_col,
        main = "Manual EDA annotation"
      )

      ann <- annotations()

      peaks <- ann[ann$annotation_type == "manual_peak", , drop = FALSE]
      if (nrow(peaks) > 0) {
        graphics::abline(v = peaks$time, lty = 2)
      }

      arts <- ann[ann$annotation_type == "artifact_interval", , drop = FALSE]
      if (nrow(arts) > 0) {
        for (i in seq_len(nrow(arts))) {
          graphics::rect(
            xleft = arts$start[i],
            ybottom = grDevices::extendrange(x[[input$signal_col]])[1],
            xright = arts$end[i],
            ytop = grDevices::extendrange(x[[input$signal_col]])[2],
            density = 12,
            border = NA
          )
        }
      }
    })

    output$click_info <- shiny::renderPrint({
      input$plot_click
    })

    shiny::observeEvent(input$add_peak, {
      ann <- annotations()
      new <- data.frame(
        annotation_type = "manual_peak",
        time = input$peak_time,
        start = NA_real_,
        end = NA_real_,
        note = input$note,
        stringsAsFactors = FALSE
      )
      annotations(rbind(ann, new))
    })

    shiny::observeEvent(input$add_artifact, {
      ann <- annotations()
      new <- data.frame(
        annotation_type = "artifact_interval",
        time = NA_real_,
        start = input$artifact_start,
        end = input$artifact_end,
        note = input$note,
        stringsAsFactors = FALSE
      )
      annotations(rbind(ann, new))
    })

    output$annotation_table <- shiny::renderTable({
      annotations()
    })

    output$download_annotations <- shiny::downloadHandler(
      filename = function() {
        "gpbiometrics_manual_annotations.csv"
      },
      content = function(file) {
        utils::write.csv(annotations(), file, row.names = FALSE)
      }
    )

    output$guardrail <- shiny::renderText({
      paste(
        "Manual annotations are expert review metadata.",
        "They do not infer emotion, stress, cognition, trust, preference, or diagnosis.",
        sep = "\n"
      )
    })
  }

  shiny::shinyApp(ui = ui, server = server)
}

#' Adaptive EMA normalization for non-stationary EDA
#'
#' Applies dependency-light adaptive normalization using an exponential moving
#' average center and robust local scale after IQR-based outlier screening.
#' This preserves local dynamics more than whole-session z-scoring, but it is
#' still a preprocessing transformation and not an emotion/stress classifier.
#'
#' @param dat A data frame.
#' @param signal_col Numeric signal column.
#' @param group_cols Optional grouping columns.
#' @param time_col Optional time column used to order rows within group.
#' @param alpha EMA smoothing parameter in `(0, 1]`.
#' @param iqr_multiplier IQR multiplier for outlier screening.
#' @param suffix Suffix for the normalized output column.
#' @param center_suffix Suffix for the EMA center column.
#' @param scale_suffix Suffix for the EMA scale column.
#' @param min_scale Minimum scale used to avoid division by zero.
#' @param overwrite Logical. If `FALSE`, existing output columns are protected.
#'
#' @return A data frame with adaptive normalized signal columns and attributes.
#' @export
standardise_gazepoint_adaptive_ema <- function(dat,
                                               signal_col = "GSR_US",
                                               group_cols = NULL,
                                               time_col = NULL,
                                               alpha = 0.05,
                                               iqr_multiplier = 1.5,
                                               suffix = "_adaptive_ema",
                                               center_suffix = "_ema_center",
                                               scale_suffix = "_ema_scale",
                                               min_scale = 1e-8,
                                               overwrite = FALSE) {
  if (!is.data.frame(dat)) {
    stop("`dat` must be a data frame.", call. = FALSE)
  }

  if (!signal_col %in% names(dat)) {
    stop("Column `", signal_col, "` was not found in `dat`.", call. = FALSE)
  }

  if (!is.numeric(dat[[signal_col]])) {
    stop("`signal_col` must identify a numeric column.", call. = FALSE)
  }

  if (!is.null(time_col) && !time_col %in% names(dat)) {
    stop("Column `", time_col, "` was not found in `dat`.", call. = FALSE)
  }

  if (is.null(group_cols)) {
    group_cols <- character()
  }

  missing_groups <- setdiff(group_cols, names(dat))
  if (length(missing_groups) > 0) {
    stop("Missing `group_cols`: ", paste(missing_groups, collapse = ", "), call. = FALSE)
  }

  if (!is.numeric(alpha) || length(alpha) != 1 || !is.finite(alpha) || alpha <= 0 || alpha > 1) {
    stop("`alpha` must be a finite number in `(0, 1]`.", call. = FALSE)
  }

  out_col <- paste0(signal_col, suffix)
  center_col <- paste0(signal_col, center_suffix)
  scale_col <- paste0(signal_col, scale_suffix)
  status_col <- paste0(signal_col, "_adaptive_ema_status")

  new_cols <- c(out_col, center_col, scale_col, status_col)

  if (!isTRUE(overwrite) && any(new_cols %in% names(dat))) {
    stop(
      "One or more output columns already exist: ",
      paste(intersect(new_cols, names(dat)), collapse = ", "),
      ". Use `overwrite = TRUE`.",
      call. = FALSE
    )
  }

  out <- dat
  out[[out_col]] <- NA_real_
  out[[center_col]] <- NA_real_
  out[[scale_col]] <- NA_real_
  out[[status_col]] <- "not_processed"

  groups <- gpbiometrics_adaptive_ema_split(out, group_cols)

  summary_rows <- lapply(names(groups), function(unit_id) {
    idx <- groups[[unit_id]]

    if (!is.null(time_col)) {
      idx <- idx[order(out[[time_col]][idx])]
    }

    x <- out[[signal_col]][idx]
    finite <- is.finite(x)

    if (sum(finite) < 5) {
      out[[status_col]][idx] <<- "insufficient_finite_samples"
      return(data.frame(
        unit_id = unit_id,
        n_rows = length(idx),
        n_finite = sum(finite),
        n_outliers = NA_integer_,
        status = "insufficient_finite_samples",
        stringsAsFactors = FALSE
      ))
    }

    q <- stats::quantile(x[finite], probs = c(0.25, 0.75), na.rm = TRUE, names = FALSE)
    iqr <- q[2] - q[1]

    lower <- q[1] - iqr_multiplier * iqr
    upper <- q[2] + iqr_multiplier * iqr

    clean <- x
    outlier <- is.finite(clean) & (clean < lower | clean > upper)
    clean[outlier] <- NA_real_

    clean_filled <- gpbiometrics_adaptive_fill(clean)
    center <- gpbiometrics_adaptive_ema(clean_filled, alpha = alpha)

    abs_dev <- abs(clean_filled - center)
    scale <- gpbiometrics_adaptive_ema(abs_dev, alpha = alpha) * 1.4826
    scale[!is.finite(scale) | scale < min_scale] <- min_scale

    normalized <- (x - center) / scale

    out[[out_col]][idx] <<- normalized
    out[[center_col]][idx] <<- center
    out[[scale_col]][idx] <<- scale
    out[[status_col]][idx] <<- ifelse(outlier, "iqr_outlier_used_for_output_not_center", "adaptive_ema_normalized")

    data.frame(
      unit_id = unit_id,
      n_rows = length(idx),
      n_finite = sum(finite),
      n_outliers = sum(outlier, na.rm = TRUE),
      status = "adaptive_ema_normalized",
      stringsAsFactors = FALSE
    )
  })

  summary_table <- do.call(rbind, summary_rows)
  rownames(summary_table) <- NULL

  overview <- data.frame(
    input_rows = nrow(dat),
    group_count = length(groups),
    successful_groups = sum(summary_table$status == "adaptive_ema_normalized"),
    problem_groups = sum(summary_table$status != "adaptive_ema_normalized"),
    signal_col = signal_col,
    output_col = out_col,
    status = if (all(summary_table$status == "adaptive_ema_normalized")) {
      "adaptive_ema_normalization_complete"
    } else if (any(summary_table$status == "adaptive_ema_normalized")) {
      "adaptive_ema_normalization_partial"
    } else {
      "adaptive_ema_normalization_failed"
    },
    interpretation = paste(
      "Adaptive EMA normalization estimates local signal center and scale after IQR-based outlier screening.",
      "It is a preprocessing transformation and does not infer emotion, stress, cognition, trust, preference, or diagnosis."
    ),
    stringsAsFactors = FALSE
  )

  attr(out, "adaptive_ema_overview") <- overview
  attr(out, "adaptive_ema_summary") <- summary_table
  attr(out, "adaptive_ema_settings") <- list(
    signal_col = signal_col,
    group_cols = group_cols,
    time_col = time_col,
    alpha = alpha,
    iqr_multiplier = iqr_multiplier,
    suffix = suffix,
    center_suffix = center_suffix,
    scale_suffix = scale_suffix,
    min_scale = min_scale,
    overwrite = overwrite
  )

  class(out) <- unique(c("gazepoint_adaptive_ema_normalised", class(out)))
  out
}

#' @rdname standardise_gazepoint_adaptive_ema
#' @export
standardize_gazepoint_adaptive_ema <- standardise_gazepoint_adaptive_ema

gpbiometrics_adaptive_ema_split <- function(dat, group_cols) {
  if (length(group_cols) == 0) {
    return(list(all_rows = seq_len(nrow(dat))))
  }

  gf <- dat[group_cols]
  gf[] <- lapply(gf, function(x) {
    x <- as.character(x)
    x[is.na(x)] <- "<NA>"
    x
  })

  split(seq_len(nrow(dat)), do.call(paste, c(gf, sep = " | ")))
}

gpbiometrics_adaptive_fill <- function(x) {
  idx <- seq_along(x)
  finite <- is.finite(x)

  if (all(finite)) {
    return(x)
  }

  if (sum(finite) == 0) {
    return(rep(0, length(x)))
  }

  if (sum(finite) == 1) {
    return(rep(x[finite][1], length(x)))
  }

  stats::approx(idx[finite], x[finite], xout = idx, rule = 2)$y
}

gpbiometrics_adaptive_ema <- function(x, alpha) {
  out <- numeric(length(x))
  out[1] <- x[1]

  if (length(x) > 1) {
    for (i in 2:length(x)) {
      out[i] <- alpha * x[i] + (1 - alpha) * out[i - 1]
    }
  }

  out
}

#' Summarise Gazepoint IBI-derived HRV windows
#'
#' Summarises interbeat-interval (IBI) values within participant, stimulus,
#' trial, AOI, or other user-defined windows. This function derives simple
#' time-domain variability features from `IBI`. It does not use the Gazepoint
#' `HRV` column as a heart-rate-variability metric, because `HRV` is treated as
#' the Gazepoint heart-rate validity flag.
#'
#' @param data A data frame or a path to a Gazepoint CSV export.
#' @param group_columns Columns defining analysis windows, such as
#'   `c("source_participant", "MEDIA_ID")`.
#' @param ibi_column Interbeat-interval column. Defaults to `"IBI"`.
#' @param validity_column Optional validity column. Defaults to `"HRV"`.
#' @param min_ibi Minimum plausible IBI in seconds.
#' @param max_ibi Maximum plausible IBI in seconds.
#'
#' @return A data frame with one row per window and IBI-derived HRV summaries.
#'
#' @export
summarise_gazepoint_ibi_hrv_windows <- function(data,
                                                group_columns,
                                                ibi_column = "IBI",
                                                validity_column = "HRV",
                                                min_ibi = 0.3,
                                                max_ibi = 2.0) {
  dat <- coerce_gazepoint_biometrics_data(data)

  if (missing(group_columns) || is.null(group_columns) || length(group_columns) == 0L) {
    stop("`group_columns` must define the analysis windows.", call. = FALSE)
  }

  missing_groups <- setdiff(group_columns, names(dat))

  if (length(missing_groups) > 0L) {
    stop(
      "`group_columns` were not found in `data`: ",
      paste(missing_groups, collapse = ", "),
      call. = FALSE
    )
  }

  if (!ibi_column %in% names(dat)) {
    stop("`ibi_column` was not found in `data`: ", ibi_column, call. = FALSE)
  }

  group_key <- make_group_key(dat, group_columns)
  groups <- unique(group_key)

  rows <- lapply(groups, function(key) {
    in_group <- group_key == key

    group_values <- dat[which(in_group)[1], group_columns, drop = FALSE]

    ibi <- as_numeric_safe(dat[[ibi_column]][in_group])

    valid <- !is.na(ibi) & ibi >= min_ibi & ibi <= max_ibi

    validity_present <- !is.null(validity_column) &&
      length(validity_column) == 1L &&
      validity_column %in% names(dat)

    if (validity_present) {
      validity <- as_numeric_safe(dat[[validity_column]][in_group])
      valid <- valid & !is.na(validity) & validity > 0
    }

    usable_ibi <- ibi[valid]

    summary <- summarise_ibi_vector(
      ibi = usable_ibi,
      n_rows = length(ibi),
      ibi_column = ibi_column,
      validity_column = ifelse(validity_present, validity_column, NA_character_)
    )

    cbind(group_values, summary, stringsAsFactors = FALSE)
  })

  out <- do.call(rbind, rows)
  rownames(out) <- NULL

  out
}


summarise_ibi_vector <- function(ibi,
                                 n_rows,
                                 ibi_column,
                                 validity_column) {
  ibi <- ibi[!is.na(ibi)]

  n_usable <- length(ibi)

  diff_ibi <- diff(ibi)

  data.frame(
    ibi_column = ibi_column,
    validity_column = validity_column,
    n_rows = n_rows,
    ibi_usable_rows = n_usable,
    ibi_usable_pct = safe_pct(n_usable, n_rows),
    mean_ibi_sec = ifelse(n_usable > 0L, mean(ibi), NA_real_),
    median_ibi_sec = ifelse(n_usable > 0L, stats::median(ibi), NA_real_),
    sd_ibi_sec = ifelse(n_usable > 1L, stats::sd(ibi), NA_real_),
    mean_hr_from_ibi_bpm = ifelse(n_usable > 0L, mean(60 / ibi), NA_real_),
    sdnn_ms = ifelse(n_usable > 1L, stats::sd(ibi) * 1000, NA_real_),
    rmssd_ms = ifelse(length(diff_ibi) > 0L, sqrt(mean(diff_ibi^2)) * 1000, NA_real_),
    pnn50 = ifelse(length(diff_ibi) > 0L, mean(abs(diff_ibi) > 0.05), NA_real_),
    min_ibi_sec = ifelse(n_usable > 0L, min(ibi), NA_real_),
    max_ibi_sec = ifelse(n_usable > 0L, max(ibi), NA_real_),
    stringsAsFactors = FALSE
  )
}

test_that("summarise_gazepoint_ibi_hrv_windows computes IBI-derived summaries", {
  dat <- data.frame(
    source_participant = c("User 1", "User 1", "User 1", "User 1"),
    MEDIA_ID = c(0, 0, 0, 0),
    IBI = c(1.0, 1.1, 0.9, 1.0),
    HRV = c(1, 1, 1, 1)
  )

  out <- summarise_gazepoint_ibi_hrv_windows(
    dat,
    group_columns = c("source_participant", "MEDIA_ID")
  )

  expect_equal(nrow(out), 1)
  expect_equal(out$ibi_usable_rows, 4)
  expect_equal(out$mean_ibi_sec, 1.0, tolerance = 1e-8)
  expect_equal(out$mean_hr_from_ibi_bpm, mean(60 / c(1.0, 1.1, 0.9, 1.0)), tolerance = 1e-8)
  expect_true("sdnn_ms" %in% names(out))
  expect_true("rmssd_ms" %in% names(out))
  expect_true("pnn50" %in% names(out))
})


test_that("summarise_gazepoint_ibi_hrv_windows excludes invalid and implausible IBI values", {
  dat <- data.frame(
    source_participant = c("User 1", "User 1", "User 1", "User 1"),
    MEDIA_ID = c(0, 0, 0, 0),
    IBI = c(1.0, 0, 3.0, 1.2),
    HRV = c(1, 1, 1, 0)
  )

  out <- summarise_gazepoint_ibi_hrv_windows(
    dat,
    group_columns = c("source_participant", "MEDIA_ID")
  )

  expect_equal(out$ibi_usable_rows, 1)
  expect_equal(out$mean_ibi_sec, 1.0, tolerance = 1e-8)
  expect_equal(out$ibi_usable_pct, 25, tolerance = 1e-8)
})


test_that("summarise_gazepoint_ibi_hrv_windows supports multiple windows", {
  dat <- data.frame(
    source_participant = c("User 1", "User 1", "User 2", "User 2"),
    MEDIA_ID = c(0, 0, 0, 0),
    IBI = c(1.0, 1.1, 0.8, 0.9),
    HRV = c(1, 1, 1, 1)
  )

  out <- summarise_gazepoint_ibi_hrv_windows(
    dat,
    group_columns = c("source_participant", "MEDIA_ID")
  )

  expect_equal(nrow(out), 2)
  expect_true(all(c("User 1", "User 2") %in% out$source_participant))
})


test_that("summarise_gazepoint_ibi_hrv_windows rejects missing IBI column", {
  dat <- data.frame(
    source_participant = "User 1",
    MEDIA_ID = 0,
    HRV = 1
  )

  expect_error(
    summarise_gazepoint_ibi_hrv_windows(
      dat,
      group_columns = c("source_participant", "MEDIA_ID")
    ),
    "ibi_column"
  )
})


test_that("summarise_gazepoint_ibi_hrv_windows rejects missing grouping columns", {
  dat <- data.frame(
    IBI = c(1.0, 1.1),
    HRV = c(1, 1)
  )

  expect_error(
    summarise_gazepoint_ibi_hrv_windows(
      dat,
      group_columns = "source_participant"
    ),
    "group_columns"
  )
})

#' Summarise biometric signals by AOI
#'
#' Summarises Gazepoint biometric channels within area-of-interest (AOI) rows.
#' The helper is intended for AOI-linked physiological descriptions, not for
#' inferring emotional valence.
#'
#' @param data A Gazepoint data frame containing AOI labels and biometric signals.
#' @param aoi_col AOI label column.
#' @param signal_cols Biometric signal columns to summarise.
#' @param group_cols Optional grouping columns, for example participant/media.
#' @param time_col Optional time/counter column.
#' @param valid_aoi_values Optional AOI labels to retain.
#' @param drop_missing_aoi Logical. If `TRUE`, rows with missing/blank AOI labels
#'   are excluded.
#' @param min_rows Minimum rows required for a group/AOI/signal summary to be
#'   marked as usable.
#'
#' @return A list with `overview`, `summary`, `signal_summary`, `aoi_summary`,
#'   `data`, and `settings`.
#' @export
summarise_gazepoint_aoi_biometrics <- function(data,
                                               aoi_col = "AOI",
                                               signal_cols = NULL,
                                               group_cols = NULL,
                                               time_col = NULL,
                                               valid_aoi_values = NULL,
                                               drop_missing_aoi = TRUE,
                                               min_rows = 1) {
  if (!is.data.frame(data)) {
    stop("`data` must be a data frame.", call. = FALSE)
  }

  dat <- as.data.frame(data, stringsAsFactors = FALSE)

  if (!aoi_col %in% names(dat)) {
    stop("`aoi_col` was not found in `data`.", call. = FALSE)
  }

  if (is.null(signal_cols)) {
    signal_cols <- gpbiometrics_aoi_biometrics_default_signals(names(dat))
  }

  signal_cols <- unique(signal_cols)
  missing_signals <- setdiff(signal_cols, names(dat))

  if (length(missing_signals) > 0) {
    stop(
      "`signal_cols` were not found in `data`: ",
      paste(missing_signals, collapse = ", "),
      call. = FALSE
    )
  }

  if (length(signal_cols) == 0) {
    stop("No biometric signal columns were supplied or detected.", call. = FALSE)
  }

  if (is.null(group_cols)) {
    group_cols <- gpbiometrics_aoi_biometrics_default_groups(names(dat))
  }

  missing_groups <- setdiff(group_cols, names(dat))

  if (length(missing_groups) > 0) {
    stop(
      "`group_cols` were not found in `data`: ",
      paste(missing_groups, collapse = ", "),
      call. = FALSE
    )
  }

  if (!is.null(time_col) && !time_col %in% names(dat)) {
    stop("`time_col` was not found in `data`.", call. = FALSE)
  }

  if (!is.logical(drop_missing_aoi) ||
      length(drop_missing_aoi) != 1 ||
      is.na(drop_missing_aoi)) {
    stop("`drop_missing_aoi` must be TRUE or FALSE.", call. = FALSE)
  }

  if (!is.numeric(min_rows) ||
      length(min_rows) != 1 ||
      !is.finite(min_rows) ||
      min_rows < 1) {
    stop("`min_rows` must be a single positive finite number.", call. = FALSE)
  }

  dat$.aoi_label <- as.character(dat[[aoi_col]])
  dat$.aoi_label[is.na(dat$.aoi_label)] <- NA_character_
  dat$.aoi_label[trimws(dat$.aoi_label) == ""] <- NA_character_

  if (isTRUE(drop_missing_aoi)) {
    dat <- dat[!is.na(dat$.aoi_label), , drop = FALSE]
  }

  if (!is.null(valid_aoi_values)) {
    dat <- dat[dat$.aoi_label %in% as.character(valid_aoi_values), , drop = FALSE]
  }

  if (nrow(dat) == 0) {
    empty_summary <- data.frame()

    overview <- data.frame(
      input_rows = nrow(data),
      retained_rows = 0L,
      aoi_count = 0L,
      signal_count = length(signal_cols),
      summary_rows = 0L,
      group_count = 0L,
      status = "fail_no_aoi_rows",
      stringsAsFactors = FALSE
    )

    return(structure(
      list(
        overview = overview,
        summary = empty_summary,
        signal_summary = empty_summary,
        aoi_summary = empty_summary,
        data = dat,
        settings = list(
          aoi_col = aoi_col,
          signal_cols = signal_cols,
          group_cols = group_cols,
          time_col = time_col,
          valid_aoi_values = valid_aoi_values,
          drop_missing_aoi = drop_missing_aoi,
          min_rows = min_rows
        )
      ),
      class = c("gazepoint_aoi_biometrics_summary", "list")
    ))
  }

  dat$.group_id <- gpbiometrics_aoi_biometrics_group_id(dat, group_cols)

  summary <- gpbiometrics_aoi_biometrics_make_summary(
    dat = dat,
    group_cols = group_cols,
    signal_cols = signal_cols,
    min_rows = min_rows
  )

  signal_summary <- gpbiometrics_aoi_biometrics_signal_summary(summary)
  aoi_summary <- gpbiometrics_aoi_biometrics_aoi_summary(summary)

  overview <- data.frame(
    input_rows = nrow(data),
    retained_rows = nrow(dat),
    aoi_count = length(unique(dat$.aoi_label)),
    signal_count = length(signal_cols),
    summary_rows = nrow(summary),
    group_count = length(unique(dat$.group_id)),
    usable_summary_rows = sum(summary$summary_status == "usable", na.rm = TRUE),
    low_row_summary_rows = sum(summary$summary_status == "warn_low_rows", na.rm = TRUE),
    status = if (nrow(summary) == 0) {
      "fail_no_aoi_biometric_summaries"
    } else if (any(summary$summary_status == "warn_low_rows", na.rm = TRUE)) {
      "warn_low_rows_in_some_summaries"
    } else {
      "aoi_biometrics_summarised"
    },
    stringsAsFactors = FALSE
  )

  structure(
    list(
      overview = overview,
      summary = summary,
      signal_summary = signal_summary,
      aoi_summary = aoi_summary,
      data = dat,
      settings = list(
        aoi_col = aoi_col,
        signal_cols = signal_cols,
        group_cols = group_cols,
        time_col = time_col,
        valid_aoi_values = valid_aoi_values,
        drop_missing_aoi = drop_missing_aoi,
        min_rows = min_rows,
        interpretation_notes = c(
          "AOI-linked biometrics summarise signals while AOI labels are active.",
          "EDA/GSR and HR summaries should not be interpreted as emotional valence.",
          "AOI dwell and biometric timing should be checked before confirmatory modelling."
        )
      )
    ),
    class = c("gazepoint_aoi_biometrics_summary", "list")
  )
}

#' Prepare AOI-biometric model data
#'
#' Converts AOI-biometric summaries into a modelling-ready table for GLM/LMM/GLMM
#' workflows.
#'
#' @param x A `gazepoint_aoi_biometrics_summary` object or summary data frame.
#' @param outcome_col Outcome column to model.
#' @param predictor_cols Optional predictor columns to retain.
#' @param factor_cols Optional columns converted to factors.
#' @param numeric_cols Optional columns converted to numeric.
#' @param group_cols Optional grouping columns for random-effect formulas.
#' @param drop_missing_outcome Logical. If `TRUE`, rows with missing outcomes are
#'   removed.
#' @param min_rows Optional minimum contributing rows required.
#' @param standardise_outcome Logical. If `TRUE`, add a z-scored outcome column.
#' @param standardise_within Standardization scope used when
#'   `standardise_outcome = TRUE`. Use `"signal"` to z-score within each
#'   biometric signal or `"all"` to z-score across all rows.
#'
#' @return A list with `overview`, `model_data`, `variable_summary`,
#'   `model_formulas`, and `settings`.
#' @export
prepare_gazepoint_aoi_biometrics_model_data <- function(x,
                                                        outcome_col = "mean_value",
                                                        predictor_cols = c("aoi_label", "signal"),
                                                        factor_cols = c("aoi_label", "signal"),
                                                        numeric_cols = NULL,
                                                        group_cols = NULL,
                                                        drop_missing_outcome = TRUE,
                                                        min_rows = NULL,
                                                        standardise_outcome = FALSE,
                                                        standardise_within = c("signal", "all")) {
  standardise_within <- match.arg(standardise_within)

  dat <- gpbiometrics_aoi_biometrics_extract_summary(x)

  if (!outcome_col %in% names(dat)) {
    stop("`outcome_col` was not found in the summary data.", call. = FALSE)
  }

  if (is.null(group_cols)) {
    group_cols <- intersect(
      c(
        "source_participant",
        "participant",
        "subject",
        "source_file",
        "MEDIA_ID",
        "MEDIA_NAME"
      ),
      names(dat)
    )
  }

  keep_cols <- unique(c(
    predictor_cols,
    factor_cols,
    numeric_cols,
    group_cols,
    outcome_col,
    "n_rows",
    "n_finite",
    "summary_status"
  ))

  if (isTRUE(standardise_outcome) &&
      identical(standardise_within, "signal")) {
    if (!"signal" %in% names(dat)) {
      stop(
        "`standardise_within = \"signal\"` requires a `signal` column.",
        call. = FALSE
      )
    }

    keep_cols <- unique(c(keep_cols, "signal"))
  }

  keep_cols <- keep_cols[keep_cols %in% names(dat)]

  model_data <- dat[, keep_cols, drop = FALSE]

  if (!is.null(min_rows)) {
    if (!is.numeric(min_rows) ||
        length(min_rows) != 1 ||
        !is.finite(min_rows) ||
        min_rows < 1) {
      stop("`min_rows` must be NULL or a single positive finite number.", call. = FALSE)
    }

    if ("n_rows" %in% names(model_data)) {
      model_data <- model_data[model_data$n_rows >= min_rows, , drop = FALSE]
    }
  }

  if (isTRUE(drop_missing_outcome)) {
    outcome <- suppressWarnings(as.numeric(model_data[[outcome_col]]))
    model_data <- model_data[is.finite(outcome), , drop = FALSE]
  }

  for (col in factor_cols) {
    if (col %in% names(model_data)) {
      model_data[[col]] <- as.factor(model_data[[col]])
    }
  }

  for (col in numeric_cols) {
    if (col %in% names(model_data)) {
      model_data[[col]] <- suppressWarnings(as.numeric(model_data[[col]]))
    }
  }

  model_data[[outcome_col]] <- suppressWarnings(as.numeric(model_data[[outcome_col]]))

  z_col <- NA_character_

  if (isTRUE(standardise_outcome)) {
    z_col <- paste0(outcome_col, "_z")

    if (identical(standardise_within, "signal")) {
      model_data[[z_col]] <- NA_real_
      signal_key <- as.character(model_data$signal)
      signal_key[is.na(signal_key)] <- "<NA>"

      for (sig in unique(signal_key)) {
        idx <- which(signal_key == sig)
        model_data[[z_col]][idx] <-
          gpbiometrics_aoi_biometrics_standardise(
            model_data[[outcome_col]][idx]
          )
      }
    } else {
      model_data[[z_col]] <-
        gpbiometrics_aoi_biometrics_standardise(
          model_data[[outcome_col]]
        )
    }
  }

  variable_summary <- gpbiometrics_aoi_biometrics_variable_summary(
    model_data = model_data,
    outcome_col = outcome_col,
    predictor_cols = predictor_cols,
    factor_cols = factor_cols,
    group_cols = group_cols
  )

  fixed_terms <- predictor_cols[predictor_cols %in% names(model_data)]
  random_terms <- group_cols[group_cols %in% names(model_data)]

  fixed_part <- if (length(fixed_terms) > 0) {
    paste(fixed_terms, collapse = " + ")
  } else {
    "1"
  }

  random_part <- if (length(random_terms) > 0) {
    paste(paste0("(1 | ", random_terms, ")"), collapse = " + ")
  } else {
    NULL
  }

  rhs <- paste(c(fixed_part, random_part), collapse = " + ")

  model_formulas <- data.frame(
    outcome = outcome_col,
    formula = paste(outcome_col, "~", rhs),
    z_outcome = if (isTRUE(standardise_outcome)) z_col else NA_character_,
    z_formula = if (isTRUE(standardise_outcome)) paste(z_col, "~", rhs) else NA_character_,
    stringsAsFactors = FALSE
  )

  overview <- data.frame(
    input_rows = nrow(dat),
    model_rows = nrow(model_data),
    outcome_col = outcome_col,
    predictor_count = length(fixed_terms),
    factor_count = sum(factor_cols %in% names(model_data)),
    group_count = length(random_terms),
    standardise_outcome = isTRUE(standardise_outcome),
    standardise_within = if (isTRUE(standardise_outcome)) {
      standardise_within
    } else {
      NA_character_
    },
    status = if (nrow(model_data) == 0) {
      "fail_no_model_rows"
    } else {
      "aoi_biometrics_model_data_prepared"
    },
    stringsAsFactors = FALSE
  )

  structure(
    list(
      overview = overview,
      model_data = model_data,
      variable_summary = variable_summary,
      model_formulas = model_formulas,
      settings = list(
        outcome_col = outcome_col,
        predictor_cols = predictor_cols,
        factor_cols = factor_cols,
        numeric_cols = numeric_cols,
        group_cols = group_cols,
        drop_missing_outcome = drop_missing_outcome,
        min_rows = min_rows,
        standardise_outcome = standardise_outcome,
        standardise_within = standardise_within,
        interpretation_notes = c(
          "Prepared data are intended for user-selected GLM/LMM/GLMM workflows.",
          "Model formulas are suggestions and should be adapted to study design.",
          "When multiple biometric signals are modelled together, within-signal standardization is usually safer than global standardization.",
          "Biometric outcomes should be interpreted as physiological signal summaries, not emotional valence."
        )
      )
    ),
    class = c("gazepoint_aoi_biometrics_model_data", "list")
  )
}

#' Plot AOI-linked biometric summaries
#'
#' Plots AOI-biometric summary values as a ggplot object.
#'
#' @param x A `gazepoint_aoi_biometrics_summary`,
#'   `gazepoint_aoi_biometrics_model_data`, or data frame.
#' @param value_col Value column to plot.
#' @param aoi_col AOI label column.
#' @param signal_col Signal label column.
#' @param group_col Optional grouping column.
#' @param plot_type `"boxplot"`, `"point"`, or `"line"`.
#' @param title Optional plot title.
#'
#' @return A ggplot object with plot data stored in attributes.
#' @export
plot_gazepoint_aoi_biometrics <- function(x,
                                          value_col = "mean_value",
                                          aoi_col = "aoi_label",
                                          signal_col = "signal",
                                          group_col = NULL,
                                          plot_type = c("boxplot", "point", "line"),
                                          title = NULL) {
  plot_type <- match.arg(plot_type)

  gpbiometrics_require_ggplot2()

  dat <- gpbiometrics_aoi_biometrics_extract_plot_data(x)

  required <- c(value_col, aoi_col, signal_col)
  missing_required <- setdiff(required, names(dat))

  if (length(missing_required) > 0) {
    stop(
      "Required plotting columns were not found: ",
      paste(missing_required, collapse = ", "),
      call. = FALSE
    )
  }

  dat$.plot_value <- suppressWarnings(as.numeric(dat[[value_col]]))
  dat$.plot_aoi <- as.factor(dat[[aoi_col]])
  dat$.plot_signal <- as.factor(dat[[signal_col]])

  if (!is.null(group_col)) {
    if (!group_col %in% names(dat)) {
      stop("`group_col` was not found in the plotting data.", call. = FALSE)
    }

    dat$.plot_group <- as.factor(dat[[group_col]])
  } else {
    dat$.plot_group <- factor("all")
  }

  dat <- dat[is.finite(dat$.plot_value), , drop = FALSE]

  p <- ggplot2::ggplot(
    dat,
    ggplot2::aes(
      x = .plot_aoi,
      y = .plot_value
    )
  ) +
    ggplot2::labs(
      title = if (is.null(title)) "AOI-linked biometric summaries" else title,
      x = aoi_col,
      y = value_col
    ) +
    ggplot2::theme_minimal()

  if (identical(plot_type, "boxplot")) {
    p <- p + ggplot2::geom_boxplot(na.rm = TRUE)
  } else if (identical(plot_type, "point")) {
    p <- p + ggplot2::geom_point(
      ggplot2::aes(group = .plot_group),
      position = ggplot2::position_jitter(width = 0.08, height = 0),
      na.rm = TRUE
    )
  } else {
    p <- p + ggplot2::geom_line(
      ggplot2::aes(group = .plot_group),
      na.rm = TRUE
    ) +
      ggplot2::geom_point(na.rm = TRUE)
  }

  if (length(unique(dat$.plot_signal)) > 1) {
    p <- p + ggplot2::facet_wrap(
      stats::as.formula("~ .plot_signal"),
      scales = "free_y"
    )
  }

  settings <- list(
    value_col = value_col,
    aoi_col = aoi_col,
    signal_col = signal_col,
    group_col = group_col,
    plot_type = plot_type,
    interpretation_notes = c(
      "The plot displays biometric summaries by AOI.",
      "Biometric differences by AOI should not be interpreted as emotional valence without additional evidence."
    )
  )

  standardise_gazepoint_plot_contract(
    plot = p,
    plot_data = dat,
    settings = settings,
    interpretation_notes = settings$interpretation_notes,
    plot_type = plot_type
  )
}

gpbiometrics_aoi_biometrics_default_signals <- function(names_dat) {
  candidates <- c(
    "GSR_US",
    "GSR_US_TONIC",
    "GSR_US_PHASIC",
    "GSR",
    "HR",
    "HRP",
    "IBI",
    "IBI_clean_ms",
    "DIAL"
  )

  unique(candidates[candidates %in% names_dat])
}

gpbiometrics_aoi_biometrics_default_groups <- function(names_dat) {
  candidates <- c(
    "source_file",
    "source_participant",
    "participant",
    "subject",
    "MEDIA_ID",
    "MEDIA_NAME",
    "trial",
    "trial_id",
    "trial_global"
  )

  unique(candidates[candidates %in% names_dat])
}

gpbiometrics_aoi_biometrics_group_id <- function(dat, group_cols) {
  if (length(group_cols) == 0) {
    return(rep("all", nrow(dat)))
  }

  group_dat <- dat[group_cols]

  group_dat[] <- lapply(group_dat, function(x) {
    x_chr <- as.character(x)
    x_chr[is.na(x_chr)] <- "<NA>"
    x_chr
  })

  apply(group_dat, 1, paste, collapse = "||")
}

gpbiometrics_aoi_biometrics_make_summary <- function(dat,
                                                     group_cols,
                                                     signal_cols,
                                                     min_rows) {
  split_keys <- c(".group_id", ".aoi_label")
  groups <- split(dat, interaction(dat[split_keys], drop = TRUE), drop = TRUE)

  out <- list()

  counter <- 1L

  for (group_name in names(groups)) {
    d <- groups[[group_name]]

    for (signal in signal_cols) {
      value <- suppressWarnings(as.numeric(d[[signal]]))
      finite_value <- value[is.finite(value)]

      row <- data.frame(
        aoi_label = d$.aoi_label[1],
        group_id = d$.group_id[1],
        signal = signal,
        n_rows = nrow(d),
        n_finite = length(finite_value),
        missing_rows = sum(!is.finite(value)),
        missing_prop = mean(!is.finite(value)),
        mean_value = if (length(finite_value) > 0) mean(finite_value, na.rm = TRUE) else NA_real_,
        median_value = if (length(finite_value) > 0) stats::median(finite_value, na.rm = TRUE) else NA_real_,
        sd_value = if (length(finite_value) > 1) stats::sd(finite_value, na.rm = TRUE) else NA_real_,
        min_value = if (length(finite_value) > 0) min(finite_value, na.rm = TRUE) else NA_real_,
        max_value = if (length(finite_value) > 0) max(finite_value, na.rm = TRUE) else NA_real_,
        first_value = if (length(finite_value) > 0) finite_value[1] else NA_real_,
        last_value = if (length(finite_value) > 0) finite_value[length(finite_value)] else NA_real_,
        delta_value = if (length(finite_value) > 1) {
          finite_value[length(finite_value)] - finite_value[1]
        } else {
          NA_real_
        },
        auc_value = if (length(finite_value) > 0) sum(finite_value, na.rm = TRUE) else NA_real_,
        summary_status = if (nrow(d) < min_rows || length(finite_value) < min_rows) {
          "warn_low_rows"
        } else {
          "usable"
        },
        stringsAsFactors = FALSE
      )

      if (length(group_cols) > 0 && all(group_cols %in% names(d))) {
        row <- cbind(d[1, group_cols, drop = FALSE], row)
      }

      out[[counter]] <- row
      counter <- counter + 1L
    }
  }

  out <- do.call(rbind, out)
  rownames(out) <- NULL

  out
}

gpbiometrics_aoi_biometrics_signal_summary <- function(summary) {
  if (nrow(summary) == 0) {
    return(data.frame())
  }

  signals <- unique(summary$signal)

  out <- lapply(signals, function(signal) {
    d <- summary[summary$signal == signal, , drop = FALSE]
    data.frame(
      signal = signal,
      summary_rows = nrow(d),
      usable_rows = sum(d$summary_status == "usable", na.rm = TRUE),
      aoi_count = length(unique(d$aoi_label)),
      mean_of_means = mean(d$mean_value, na.rm = TRUE),
      median_of_means = stats::median(d$mean_value, na.rm = TRUE),
      stringsAsFactors = FALSE
    )
  })

  out <- do.call(rbind, out)
  rownames(out) <- NULL

  out
}

gpbiometrics_aoi_biometrics_aoi_summary <- function(summary) {
  if (nrow(summary) == 0) {
    return(data.frame())
  }

  aois <- unique(summary$aoi_label)

  out <- lapply(aois, function(aoi) {
    d <- summary[summary$aoi_label == aoi, , drop = FALSE]
    data.frame(
      aoi_label = aoi,
      summary_rows = nrow(d),
      usable_rows = sum(d$summary_status == "usable", na.rm = TRUE),
      signal_count = length(unique(d$signal)),
      total_rows_contributing = sum(d$n_rows, na.rm = TRUE),
      stringsAsFactors = FALSE
    )
  })

  out <- do.call(rbind, out)
  rownames(out) <- NULL

  out
}

gpbiometrics_aoi_biometrics_extract_summary <- function(x) {
  if (inherits(x, "gazepoint_aoi_biometrics_summary") &&
      !is.null(x$summary)) {
    return(as.data.frame(x$summary, stringsAsFactors = FALSE))
  }

  if (is.data.frame(x)) {
    return(as.data.frame(x, stringsAsFactors = FALSE))
  }

  stop("`x` must be an AOI-biometric summary object or a data frame.", call. = FALSE)
}

gpbiometrics_aoi_biometrics_extract_plot_data <- function(x) {
  if (inherits(x, "gazepoint_aoi_biometrics_summary") &&
      !is.null(x$summary)) {
    return(as.data.frame(x$summary, stringsAsFactors = FALSE))
  }

  if (inherits(x, "gazepoint_aoi_biometrics_model_data") &&
      !is.null(x$model_data)) {
    return(as.data.frame(x$model_data, stringsAsFactors = FALSE))
  }

  if (is.data.frame(x)) {
    return(as.data.frame(x, stringsAsFactors = FALSE))
  }

  stop("`x` must be an AOI-biometric object or a data frame.", call. = FALSE)
}

gpbiometrics_aoi_biometrics_variable_summary <- function(model_data,
                                                         outcome_col,
                                                         predictor_cols,
                                                         factor_cols,
                                                         group_cols) {
  if (nrow(model_data) == 0) {
    return(data.frame())
  }

  cols <- unique(c(outcome_col, predictor_cols, factor_cols, group_cols))
  cols <- cols[cols %in% names(model_data)]

  out <- lapply(cols, function(col) {
    x <- model_data[[col]]
    is_num <- is.numeric(x)

    data.frame(
      variable = col,
      class = paste(class(x), collapse = "/"),
      n = length(x),
      missing = sum(is.na(x)),
      unique_values = length(unique(x[!is.na(x)])),
      mean = if (is_num) mean(x, na.rm = TRUE) else NA_real_,
      sd = if (is_num) stats::sd(x, na.rm = TRUE) else NA_real_,
      stringsAsFactors = FALSE
    )
  })

  out <- do.call(rbind, out)
  rownames(out) <- NULL

  out
}

gpbiometrics_aoi_biometrics_standardise <- function(x) {
  x <- suppressWarnings(as.numeric(x))
  x_mean <- mean(x, na.rm = TRUE)
  x_sd <- stats::sd(x, na.rm = TRUE)

  if (is.finite(x_sd) && x_sd > 0) {
    (x - x_mean) / x_sd
  } else {
    ifelse(is.na(x), NA_real_, 0)
  }
}

test_that("summarise_gazepoint_aoi_biometrics summarises signals by AOI", {
  dat <- data.frame(
    participant = c("P1", "P1", "P1", "P1", "P2", "P2"),
    MEDIA_ID = c(1, 1, 1, 1, 1, 1),
    AOI = c("claim", "claim", "logo", "logo", "claim", "logo"),
    GSR_US = c(1, 2, 3, 4, 5, 6),
    HR = c(70, 72, 74, 76, 78, 80)
  )

  res <- summarise_gazepoint_aoi_biometrics(
    dat,
    aoi_col = "AOI",
    signal_cols = c("GSR_US", "HR"),
    group_cols = c("participant", "MEDIA_ID"),
    min_rows = 1
  )

  expect_s3_class(res, "gazepoint_aoi_biometrics_summary")
  expect_equal(res$overview$signal_count, 2)
  expect_true(nrow(res$summary) > 0)
  expect_true(all(c("mean_value", "median_value", "signal", "aoi_label") %in% names(res$summary)))
})

test_that("summarise_gazepoint_aoi_biometrics can filter AOI labels", {
  dat <- data.frame(
    participant = "P1",
    AOI = c("claim", "claim", "logo"),
    GSR_US = c(1, 2, 3)
  )

  res <- summarise_gazepoint_aoi_biometrics(
    dat,
    aoi_col = "AOI",
    signal_cols = "GSR_US",
    group_cols = "participant",
    valid_aoi_values = "claim"
  )

  expect_equal(unique(res$summary$aoi_label), "claim")
  expect_equal(res$overview$aoi_count, 1)
})

test_that("summarise_gazepoint_aoi_biometrics warns for low-row summaries", {
  dat <- data.frame(
    participant = "P1",
    AOI = c("claim", "logo"),
    GSR_US = c(1, 2)
  )

  res <- summarise_gazepoint_aoi_biometrics(
    dat,
    aoi_col = "AOI",
    signal_cols = "GSR_US",
    group_cols = "participant",
    min_rows = 2
  )

  expect_equal(res$overview$status, "warn_low_rows_in_some_summaries")
  expect_true(any(res$summary$summary_status == "warn_low_rows"))
})

test_that("summarise_gazepoint_aoi_biometrics handles no retained AOI rows", {
  dat <- data.frame(
    participant = "P1",
    AOI = c(NA, ""),
    GSR_US = c(1, 2)
  )

  res <- summarise_gazepoint_aoi_biometrics(
    dat,
    aoi_col = "AOI",
    signal_cols = "GSR_US",
    group_cols = "participant"
  )

  expect_equal(res$overview$status, "fail_no_aoi_rows")
  expect_equal(nrow(res$summary), 0)
})

test_that("prepare_gazepoint_aoi_biometrics_model_data prepares model data", {
  dat <- data.frame(
    participant = c("P1", "P1", "P2", "P2"),
    AOI = c("claim", "logo", "claim", "logo"),
    GSR_US = c(1, 2, 3, 4)
  )

  summary <- summarise_gazepoint_aoi_biometrics(
    dat,
    aoi_col = "AOI",
    signal_cols = "GSR_US",
    group_cols = "participant"
  )

  model_data <- prepare_gazepoint_aoi_biometrics_model_data(
    summary,
    outcome_col = "mean_value",
    predictor_cols = c("aoi_label", "signal"),
    factor_cols = c("aoi_label", "signal"),
    group_cols = "participant",
    standardise_outcome = TRUE,
    standardise_within = "signal"
  )

  expect_s3_class(model_data, "gazepoint_aoi_biometrics_model_data")
  expect_equal(model_data$overview$status, "aoi_biometrics_model_data_prepared")
  expect_equal(model_data$overview$standardise_within, "signal")
  expect_true("mean_value_z" %in% names(model_data$model_data))
  expect_true(is.factor(model_data$model_data$aoi_label))
})

test_that("prepare_gazepoint_aoi_biometrics_model_data standardises within signal", {
  dat <- data.frame(
    participant = c("P1", "P1", "P2", "P2"),
    AOI = c("claim", "logo", "claim", "logo"),
    GSR_US = c(1, 2, 3, 4),
    HR = c(70, 72, 74, 76)
  )

  summary <- summarise_gazepoint_aoi_biometrics(
    dat,
    aoi_col = "AOI",
    signal_cols = c("GSR_US", "HR"),
    group_cols = "participant"
  )

  model_data <- prepare_gazepoint_aoi_biometrics_model_data(
    summary,
    outcome_col = "mean_value",
    predictor_cols = c("aoi_label", "signal"),
    factor_cols = c("aoi_label", "signal"),
    group_cols = "participant",
    standardise_outcome = TRUE,
    standardise_within = "signal"
  )

  by_signal_mean <- tapply(
    model_data$model_data$mean_value_z,
    model_data$model_data$signal,
    mean,
    na.rm = TRUE
  )

  expect_equal(model_data$overview$standardise_within, "signal")
  expect_true(all(abs(by_signal_mean) < 1e-10))
})

test_that("prepare_gazepoint_aoi_biometrics_model_data can standardise across all rows", {
  dat <- data.frame(
    participant = c("P1", "P1", "P2", "P2"),
    AOI = c("claim", "logo", "claim", "logo"),
    GSR_US = c(1, 2, 3, 4),
    HR = c(70, 72, 74, 76)
  )

  summary <- summarise_gazepoint_aoi_biometrics(
    dat,
    aoi_col = "AOI",
    signal_cols = c("GSR_US", "HR"),
    group_cols = "participant"
  )

  model_data <- prepare_gazepoint_aoi_biometrics_model_data(
    summary,
    outcome_col = "mean_value",
    predictor_cols = c("aoi_label", "signal"),
    factor_cols = c("aoi_label", "signal"),
    group_cols = "participant",
    standardise_outcome = TRUE,
    standardise_within = "all"
  )

  expect_equal(model_data$overview$standardise_within, "all")
  expect_true(abs(mean(model_data$model_data$mean_value_z, na.rm = TRUE)) < 1e-10)
})

test_that("prepare_gazepoint_aoi_biometrics_model_data can filter by min_rows", {
  dat <- data.frame(
    participant = "P1",
    AOI = c("claim", "claim", "logo"),
    GSR_US = c(1, 2, 3)
  )

  summary <- summarise_gazepoint_aoi_biometrics(
    dat,
    aoi_col = "AOI",
    signal_cols = "GSR_US",
    group_cols = "participant"
  )

  model_data <- prepare_gazepoint_aoi_biometrics_model_data(
    summary,
    outcome_col = "mean_value",
    predictor_cols = "aoi_label",
    min_rows = 2
  )

  expect_true(all(model_data$model_data$n_rows >= 2))
})

test_that("plot_gazepoint_aoi_biometrics returns ggplot objects", {
  skip_if_not_installed("ggplot2")

  dat <- data.frame(
    participant = c("P1", "P1", "P2", "P2"),
    AOI = c("claim", "logo", "claim", "logo"),
    GSR_US = c(1, 2, 3, 4),
    HR = c(70, 72, 74, 76)
  )

  summary <- summarise_gazepoint_aoi_biometrics(
    dat,
    aoi_col = "AOI",
    signal_cols = c("GSR_US", "HR"),
    group_cols = "participant"
  )

  p <- plot_gazepoint_aoi_biometrics(
    summary,
    value_col = "mean_value",
    plot_type = "point",
    group_col = "participant"
  )

  expect_s3_class(p, "ggplot")
  expect_true(is.data.frame(attr(p, "plot_data")))
})

test_that("plot_gazepoint_aoi_biometrics supports model-data objects", {
  skip_if_not_installed("ggplot2")

  dat <- data.frame(
    participant = c("P1", "P1", "P2", "P2"),
    AOI = c("claim", "logo", "claim", "logo"),
    GSR_US = c(1, 2, 3, 4)
  )

  summary <- summarise_gazepoint_aoi_biometrics(
    dat,
    aoi_col = "AOI",
    signal_cols = "GSR_US",
    group_cols = "participant"
  )

  model_data <- prepare_gazepoint_aoi_biometrics_model_data(
    summary,
    standardise_outcome = TRUE,
    standardise_within = "signal"
  )

  p <- plot_gazepoint_aoi_biometrics(
    model_data,
    value_col = "mean_value_z",
    plot_type = "boxplot"
  )

  expect_s3_class(p, "ggplot")
})

test_that("AOI-biometric helpers validate inputs", {
  dat <- data.frame(
    AOI = "claim",
    GSR_US = 1
  )

  expect_error(
    summarise_gazepoint_aoi_biometrics(dat, aoi_col = "missing"),
    "`aoi_col`"
  )

  expect_error(
    summarise_gazepoint_aoi_biometrics(dat, signal_cols = "missing"),
    "`signal_cols`"
  )

  expect_error(
    prepare_gazepoint_aoi_biometrics_model_data(dat, outcome_col = "missing"),
    "`outcome_col`"
  )

  expect_error(
    prepare_gazepoint_aoi_biometrics_model_data(
      data.frame(
        aoi_label = "claim",
        mean_value = 1
      ),
      outcome_col = "mean_value",
      standardise_outcome = TRUE,
      standardise_within = "signal"
    ),
    "`standardise_within = \"signal\"` requires a `signal` column"
  )

  expect_error(
    plot_gazepoint_aoi_biometrics(dat, value_col = "missing"),
    "Required plotting columns"
  )
})

#' Plot an EDA-gram-style time-frequency representation
#'
#' Creates a dependency-light EDA-gram-style representation using sliding-window
#' spectral power. This is inspired by EDA-gram visualisations, but it does not
#' implement a full sparse dictionary decomposition unless such a model is
#' supplied externally.
#'
#' @param dat A data frame containing EDA data.
#' @param eda_col Numeric EDA/conductance column.
#' @param time_col Numeric time column.
#' @param group_cols Optional grouping columns.
#' @param group_id_to_plot Optional group ID to plot. If `NULL`, plots the first
#'   available group.
#' @param sampling_rate Optional sampling rate in Hz. If `NULL`, estimated from
#'   `time_col`.
#' @param window_seconds Sliding-window length.
#' @param step_seconds Sliding-window step.
#' @param frequency_range Frequency range shown in Hz.
#' @param frequency_bins Number of frequency bins.
#' @param log_power Logical. If `TRUE`, plot log1p power.
#' @param plot Logical. If `TRUE`, draw the EDA-gram.
#' @param main Plot title.
#'
#' @return Invisibly returns a list with `overview`, `gram_table`,
#'   `plot_matrix`, and `settings`.
#' @export
plot_gazepoint_eda_gram <- function(dat,
                                    eda_col = "GSR_US",
                                    time_col = "CNT",
                                    group_cols = NULL,
                                    group_id_to_plot = NULL,
                                    sampling_rate = NULL,
                                    window_seconds = 30,
                                    step_seconds = 5,
                                    frequency_range = c(0.01, 0.50),
                                    frequency_bins = 64,
                                    log_power = TRUE,
                                    plot = TRUE,
                                    main = "EDA-gram") {
  if (!is.data.frame(dat)) {
    stop("`dat` must be a data frame.", call. = FALSE)
  }

  if (!eda_col %in% names(dat)) {
    stop("Column `", eda_col, "` was not found in `dat`.", call. = FALSE)
  }

  if (!time_col %in% names(dat)) {
    stop("Column `", time_col, "` was not found in `dat`.", call. = FALSE)
  }

  if (!is.numeric(dat[[eda_col]])) {
    stop("`eda_col` must identify a numeric column.", call. = FALSE)
  }

  if (!is.numeric(dat[[time_col]])) {
    stop("`time_col` must identify a numeric column.", call. = FALSE)
  }

  if (is.null(group_cols)) {
    group_cols <- character()
  }

  missing_groups <- setdiff(group_cols, names(dat))
  if (length(missing_groups) > 0) {
    stop("Missing `group_cols`: ", paste(missing_groups, collapse = ", "), call. = FALSE)
  }

  frequency_grid <- seq(
    frequency_range[1],
    frequency_range[2],
    length.out = frequency_bins
  )

  groups <- gpbiometrics_edagram_split(dat, group_cols)

  gram_rows <- list()
  matrix_by_group <- list()
  row_id <- 1L

  for (group_id in names(groups)) {
    idx <- groups[[group_id]]
    idx <- idx[order(dat[[time_col]][idx])]

    time <- dat[[time_col]][idx]
    eda <- dat[[eda_col]][idx]

    fs <- gpbiometrics_edagram_sampling_rate(time, sampling_rate)

    if (!is.finite(fs) || fs <= 0 || sum(is.finite(eda)) < 8) {
      next
    }

    min_time <- min(time, na.rm = TRUE)
    max_time <- max(time, na.rm = TRUE)

    starts <- seq(min_time, max_time - window_seconds, by = step_seconds)

    if (length(starts) == 0) {
      starts <- min_time
    }

    group_matrix <- matrix(
      NA_real_,
      nrow = length(starts),
      ncol = length(frequency_grid)
    )

    for (i in seq_along(starts)) {
      w_start <- starts[i]
      w_end <- w_start + window_seconds
      in_window <- is.finite(time) & time >= w_start & time <= w_end
      x <- eda[in_window]

      if (sum(is.finite(x)) < 8) {
        next
      }

      x <- gpbiometrics_edagram_fill(x)
      x <- x - mean(x, na.rm = TRUE)

      spec <- stats::spec.pgram(
        x,
        taper = 0.1,
        plot = FALSE,
        demean = TRUE,
        detrend = TRUE,
        fast = TRUE
      )

      freq <- spec$freq * fs
      power <- spec$spec

      interp_power <- stats::approx(
        x = freq,
        y = power,
        xout = frequency_grid,
        rule = 2
      )$y

      if (isTRUE(log_power)) {
        interp_power <- log1p(interp_power)
      }

      group_matrix[i, ] <- interp_power

      for (j in seq_along(frequency_grid)) {
        gram_rows[[row_id]] <- data.frame(
          group_id = group_id,
          window_index = i,
          window_start = w_start,
          window_end = w_end,
          window_midpoint = mean(c(w_start, w_end)),
          frequency_hz = frequency_grid[j],
          power = interp_power[j],
          status = "eda_gram_cell_created",
          stringsAsFactors = FALSE
        )
        row_id <- row_id + 1L
      }
    }

    matrix_by_group[[group_id]] <- list(
      group_id = group_id,
      window_midpoint = starts + window_seconds / 2,
      frequency_hz = frequency_grid,
      power_matrix = group_matrix
    )
  }

  gram_table <- if (length(gram_rows) > 0) {
    do.call(rbind, gram_rows)
  } else {
    data.frame()
  }

  if (length(matrix_by_group) == 0) {
    stop("Could not create EDA-gram: insufficient finite data or sampling information.", call. = FALSE)
  }

  if (is.null(group_id_to_plot)) {
    group_id_to_plot <- names(matrix_by_group)[1]
  }

  if (!group_id_to_plot %in% names(matrix_by_group)) {
    stop("`group_id_to_plot` was not found among available groups.", call. = FALSE)
  }

  plot_matrix <- matrix_by_group[[group_id_to_plot]]

  if (isTRUE(plot)) {
    graphics::image(
      x = plot_matrix$window_midpoint,
      y = plot_matrix$frequency_hz,
      z = plot_matrix$power_matrix,
      xlab = "Time",
      ylab = "Frequency (Hz)",
      main = main
    )
  }

  overview <- data.frame(
    group_count = length(matrix_by_group),
    gram_rows = nrow(gram_table),
    plotted_group_id = group_id_to_plot,
    status = "eda_gram_created",
    interpretation = paste(
      "This is an EDA-gram-style sliding-window spectral representation.",
      "It is not a full sparse dictionary EDA-gram decomposition and does not infer emotion, stress, cognition, health status, or diagnosis."
    ),
    stringsAsFactors = FALSE
  )

  invisible(
    structure(
      list(
        overview = overview,
        gram_table = gram_table,
        plot_matrix = plot_matrix,
        settings = list(
          eda_col = eda_col,
          time_col = time_col,
          group_cols = group_cols,
          sampling_rate = sampling_rate,
          window_seconds = window_seconds,
          step_seconds = step_seconds,
          frequency_range = frequency_range,
          frequency_bins = frequency_bins,
          log_power = log_power,
          main = main
        )
      ),
      class = c("gazepoint_eda_gram", "list")
    )
  )
}

gpbiometrics_edagram_split <- function(dat, group_cols) {
  if (length(group_cols) == 0) {
    return(list(all_rows = seq_len(nrow(dat))))
  }

  gf <- dat[group_cols]
  gf[] <- lapply(gf, function(x) {
    x <- as.character(x)
    x[is.na(x)] <- "<NA>"
    x
  })

  split(seq_len(nrow(dat)), do.call(paste, c(gf, sep = " | ")))
}

gpbiometrics_edagram_sampling_rate <- function(time, sampling_rate = NULL) {
  if (!is.null(sampling_rate)) {
    return(sampling_rate)
  }

  time <- time[is.finite(time)]

  if (length(time) < 3) {
    return(NA_real_)
  }

  dt <- diff(time)
  dt <- dt[is.finite(dt) & dt > 0]

  if (length(dt) == 0) {
    return(NA_real_)
  }

  median_dt <- stats::median(dt)

  if (median_dt > 10) {
    1000 / median_dt
  } else {
    1 / median_dt
  }
}

gpbiometrics_edagram_fill <- function(x) {
  idx <- seq_along(x)
  finite <- is.finite(x)

  if (all(finite)) {
    return(x)
  }

  if (sum(finite) == 0) {
    return(rep(0, length(x)))
  }

  if (sum(finite) == 1) {
    return(rep(x[finite][1], length(x)))
  }

  stats::approx(idx[finite], x[finite], xout = idx, rule = 2)$y
}

test_that("plot_gazepoint_eda_gram creates EDA-gram output", {
  time <- seq(0, 120, by = 1)
  eda <- sin(2 * pi * 0.1 * time) + rnorm(length(time), sd = 0.02)

  dat <- data.frame(
    participant = "p1",
    time = time,
    GSR_US = eda
  )

  out <- plot_gazepoint_eda_gram(
    dat,
    eda_col = "GSR_US",
    time_col = "time",
    group_cols = "participant",
    sampling_rate = 1,
    window_seconds = 30,
    step_seconds = 15,
    frequency_bins = 16,
    plot = FALSE
  )

  expect_s3_class(out, "gazepoint_eda_gram")
  expect_true(nrow(out$gram_table) > 0)
  expect_true("power" %in% names(out$gram_table))
})

test_that("extract_gazepoint_hrv_rcmse returns scale-wise entropy", {
  dat <- data.frame(
    participant = "p1",
    IBI = 0.8 + 0.04 * sin(seq(0, 10 * pi, length.out = 100)) +
      rnorm(100, sd = 0.005)
  )

  out <- extract_gazepoint_hrv_rcmse(
    dat,
    ibi_col = "IBI",
    group_cols = "participant",
    scales = 1:4,
    min_intervals = 20
  )

  expect_s3_class(out, "gazepoint_hrv_rcmse")
  expect_equal(nrow(out$rcmse_by_scale), 4)
  expect_true("mean_rcmse" %in% names(out$summary))
})

test_that("run_gazepoint_automated_statistics selects exploratory tests", {
  dat <- data.frame(
    condition = rep(c("A", "B", "C"), each = 12),
    feature_1 = c(rnorm(12, 0), rnorm(12, 1), rnorm(12, 1.5)),
    feature_2 = c(rnorm(12, 0), rnorm(12, 0), rnorm(12, 0.2))
  )

  out <- run_gazepoint_automated_statistics(
    dat,
    outcome_cols = c("feature_1", "feature_2"),
    group_col = "condition"
  )

  expect_s3_class(out, "gazepoint_automated_statistics")
  expect_equal(nrow(out$test_table), 2)
  expect_true("p_adjusted" %in% names(out$test_table))
})

test_that("analyze_gazepoint_ac_susceptance handles AC EDA components", {
  dat <- data.frame(
    participant = "p1",
    frequency = rep(c(10, 20), each = 20),
    conductance = rnorm(40, mean = 1, sd = 0.05),
    susceptance = rnorm(40, mean = 0.2, sd = 0.02)
  )

  out <- analyze_gazepoint_ac_susceptance(
    dat,
    conductance_col = "conductance",
    susceptance_col = "susceptance",
    frequency_col = "frequency",
    group_cols = "participant"
  )

  expect_s3_class(out, "gazepoint_ac_susceptance")
  expect_true("ac_eda_admittance_magnitude" %in% names(out$timeseries))
  expect_true(nrow(out$summary) >= 2)
})

test_that("run_gazepoint_online_design_optimization recommends a condition", {
  candidates <- data.frame(
    condition = c("attention", "no_attention", "control"),
    expected_utility = c(0.70, 0.55, 0.40),
    cost = c(0.05, 0.02, 0.01)
  )

  out <- run_gazepoint_online_design_optimization(
    candidates,
    condition_col = "condition",
    utility_col = "expected_utility",
    cost_col = "cost",
    previous_assignments = c("attention", "attention", "control")
  )

  expect_s3_class(out, "gazepoint_online_design_optimization")
  expect_equal(nrow(out$recommendation), 1)
  expect_true("optimization_rank" %in% names(out$ranked_candidates))
})

#' Summarise Gazepoint engagement-dial windows
#'
#' Compatibility wrapper for [summarise_gazepoint_engagement_windows()]. This
#' helper uses the term "dial" for users who refer to Gazepoint engagement-dial
#' or self-reported engagement streams, while delegating the calculation to the
#' canonical engagement-window summariser.
#'
#' @param data A data frame containing Gazepoint Biometrics engagement/dial data.
#' @param ... Additional arguments passed to
#'   [summarise_gazepoint_engagement_windows()].
#' @param dial_col Optional dial/engagement column. When supplied, it is mapped
#'   to the corresponding value-column argument of the underlying helper.
#'
#' @return The output of [summarise_gazepoint_engagement_windows()].
#'
#' @examples
#' df <- data.frame(
#'   USER = rep(c("P1", "P2"), each = 3),
#'   DIAL = c(40, 45, 50, 55, 60, 65)
#' )
#' summarise_gazepoint_dial_windows(df)
#'
#' @export
summarise_gazepoint_dial_windows <- function(data, ..., dial_col = NULL) {
  if (!is.data.frame(data)) {
    stop("`data` must be a data frame.", call. = FALSE)
  }

  if (!is.null(dial_col)) {
    .gpbiom_assert_columns(data, dial_col, "dial_col")
  }

  args <- list(...)

  if (!is.null(dial_col)) {
    args <- .gpbiom_add_alias_argument(
      args = args,
      target_function = summarise_gazepoint_engagement_windows,
      aliases = c(
        "value_column",
        "engagement_col",
        "engagement_column",
        "dial_col",
        "dial_column",
        "value_col",
        "signal_col",
        "signal_column"
      ),
      value = dial_col
    )
  }

  .gpbiom_call_with_data_first(
    target_function = summarise_gazepoint_engagement_windows,
    data = data,
    args = args
  )
}


#' Join Gazepoint Biometrics data to gp3tools-style eye-tracking data
#'
#' Compatibility wrapper for [join_gazepoint_biometrics_to_master()]. This alias
#' is provided for users who work with a gp3tools master table and want an
#' explicit gp3tools-facing function name. The implementation delegates to the
#' canonical biometric-to-master join helper.
#'
#' @param biometrics A data frame containing Gazepoint Biometrics samples or
#'   summaries.
#' @param gp3tools_master A gp3tools-style master eye-tracking data frame.
#' @param ... Additional arguments passed to
#'   [join_gazepoint_biometrics_to_master()], including the required `by`
#'   argument when the underlying join helper requires explicit join columns.
#'
#' @return The output of [join_gazepoint_biometrics_to_master()].
#'
#' @examples
#' biometrics <- data.frame(USER = rep("P1", 3), CNT = 1:3, HR = c(70, 71, 72))
#' master <- data.frame(USER = rep("P1", 3), CNT = 1:3, AOI = c("A", "B", "A"))
#' join_gazepoint_biometrics_to_gp3tools(
#'   biometrics,
#'   master,
#'   by = c("USER", "CNT")
#' )
#'
#' @export
join_gazepoint_biometrics_to_gp3tools <- function(biometrics,
                                                  gp3tools_master,
                                                  ...) {
  if (!is.data.frame(biometrics)) {
    stop("`biometrics` must be a data frame.", call. = FALSE)
  }

  if (!is.data.frame(gp3tools_master)) {
    stop("`gp3tools_master` must be a data frame.", call. = FALSE)
  }

  args <- list(...)

  .gpbiom_call_with_two_data_args(
    target_function = join_gazepoint_biometrics_to_master,
    first_data = biometrics,
    second_data = gp3tools_master,
    args = args
  )
}


.gpbiom_call_with_data_first <- function(target_function, data, args) {
  formal_names <- names(formals(target_function))

  if (length(formal_names) == 0L) {
    stop("The target function does not expose formal arguments.", call. = FALSE)
  }

  data_arg <- formal_names[1L]

  do.call(
    target_function,
    c(stats::setNames(list(data), data_arg), args)
  )
}


.gpbiom_call_with_two_data_args <- function(target_function,
                                            first_data,
                                            second_data,
                                            args) {
  formal_names <- names(formals(target_function))

  if (length(formal_names) < 2L) {
    stop("The target function must expose at least two formal arguments.",
         call. = FALSE)
  }

  data_args <- stats::setNames(
    list(first_data, second_data),
    formal_names[1:2]
  )

  do.call(target_function, c(data_args, args))
}


.gpbiom_add_alias_argument <- function(args,
                                       target_function,
                                       aliases,
                                       value) {
  formal_names <- names(formals(target_function))

  existing_alias <- intersect(aliases, names(args))

  if (length(existing_alias) > 0L) {
    return(args)
  }

  supported_alias <- intersect(aliases, formal_names)

  if (length(supported_alias) > 0L) {
    args[[supported_alias[1L]]] <- value
  } else if ("..." %in% formal_names) {
    args[[aliases[1L]]] <- value
  }

  args
}

test_that("summarise_gazepoint_dial_windows validates data input", {
  expect_error(
    summarise_gazepoint_dial_windows(1:3),
    "`data` must be"
  )
})


test_that("summarise_gazepoint_dial_windows validates explicit dial column", {
  df <- data.frame(
    DIAL = c(40, 45, 50)
  )

  expect_error(
    summarise_gazepoint_dial_windows(df, dial_col = "missing"),
    "not found"
  )
})


test_that("summarise_gazepoint_dial_windows delegates to engagement-window helper", {
  df <- data.frame(
    USER = rep(c("P1", "P2"), each = 3),
    DIAL = c(40, 45, 50, 55, 60, 65)
  )

  direct <- summarise_gazepoint_engagement_windows(df)
  alias <- summarise_gazepoint_dial_windows(df)

  expect_equal(alias, direct)
})


test_that("summarise_gazepoint_dial_windows supports explicit dial_col alias", {
  df <- data.frame(
    USER = rep(c("P1", "P2"), each = 3),
    dial_value = c(40, 45, 50, 55, 60, 65)
  )

  direct <- summarise_gazepoint_engagement_windows(
    df,
    value_column = "dial_value"
  )

  alias <- summarise_gazepoint_dial_windows(
    df,
    dial_col = "dial_value"
  )

  expect_equal(alias, direct)
})


test_that("join_gazepoint_biometrics_to_gp3tools validates biometric input", {
  master <- data.frame(USER = "P1", CNT = 1)

  expect_error(
    join_gazepoint_biometrics_to_gp3tools(1:3, master),
    "`biometrics` must be"
  )
})


test_that("join_gazepoint_biometrics_to_gp3tools validates gp3tools master input", {
  biometrics <- data.frame(USER = "P1", CNT = 1)

  expect_error(
    join_gazepoint_biometrics_to_gp3tools(biometrics, 1:3),
    "`gp3tools_master` must be"
  )
})


test_that("join_gazepoint_biometrics_to_gp3tools delegates to master-join helper", {
  biometrics <- data.frame(
    USER = rep("P1", 3),
    CNT = 1:3,
    HR = c(70, 71, 72)
  )

  master <- data.frame(
    USER = rep("P1", 3),
    CNT = 1:3,
    AOI = c("A", "B", "A")
  )

  direct <- join_gazepoint_biometrics_to_master(
    biometrics,
    master,
    by = c("USER", "CNT")
  )

  alias <- join_gazepoint_biometrics_to_gp3tools(
    biometrics,
    master,
    by = c("USER", "CNT")
  )

  expect_equal(alias, direct)
})


test_that("compatibility wrappers expose stable formal arguments", {
  expect_true("data" %in% names(formals(summarise_gazepoint_dial_windows)))
  expect_true("dial_col" %in% names(formals(summarise_gazepoint_dial_windows)))

  expect_true("biometrics" %in% names(formals(join_gazepoint_biometrics_to_gp3tools)))
  expect_true("gp3tools_master" %in% names(formals(join_gazepoint_biometrics_to_gp3tools)))
})

#' Denoise EDA using a user-supplied autoencoder reconstruction model
#'
#' Applies a user-supplied reconstruction function or model to fixed-length EDA
#' windows. No pretrained neural network is bundled. This function is an
#' interoperability bridge for validated user-supplied autoencoders.
#'
#' @param dat A data frame.
#' @param eda_col Numeric EDA column.
#' @param time_col Optional time column.
#' @param group_cols Optional grouping columns.
#' @param model A user-supplied function or model. If `NULL`, the original
#'   signal is copied and status records that no model was supplied.
#' @param window_samples Window length in samples.
#' @param output_col Optional output column.
#' @param overwrite Logical. If `FALSE`, existing output column is protected.
#'
#' @return A data frame with reconstructed signal and denoising attributes.
#' @export
denoise_gazepoint_eda_autoencoder <- function(dat,
                                              eda_col = "GSR_US",
                                              time_col = NULL,
                                              group_cols = NULL,
                                              model = NULL,
                                              window_samples = 128,
                                              output_col = NULL,
                                              overwrite = FALSE) {
  gpbiometrics_autoencoder_denoise_signal(
    dat = dat,
    signal_col = eda_col,
    signal_type = "EDA",
    time_col = time_col,
    group_cols = group_cols,
    model = model,
    window_samples = window_samples,
    output_col = output_col,
    overwrite = overwrite
  )
}

#' Denoise PPG using a user-supplied autoencoder reconstruction model
#'
#' Applies a user-supplied reconstruction function or model to fixed-length PPG
#' windows. No pretrained neural network is bundled. This function is an
#' interoperability bridge for validated user-supplied autoencoders.
#'
#' @param dat A data frame.
#' @param ppg_col Numeric PPG/pulse column.
#' @param time_col Optional time column.
#' @param group_cols Optional grouping columns.
#' @param model A user-supplied function or model. If `NULL`, the original
#'   signal is copied and status records that no model was supplied.
#' @param window_samples Window length in samples.
#' @param output_col Optional output column.
#' @param overwrite Logical. If `FALSE`, existing output column is protected.
#'
#' @return A data frame with reconstructed signal and denoising attributes.
#' @export
denoise_gazepoint_ppg_autoencoder <- function(dat,
                                              ppg_col = "HRP",
                                              time_col = NULL,
                                              group_cols = NULL,
                                              model = NULL,
                                              window_samples = 128,
                                              output_col = NULL,
                                              overwrite = FALSE) {
  gpbiometrics_autoencoder_denoise_signal(
    dat = dat,
    signal_col = ppg_col,
    signal_type = "PPG",
    time_col = time_col,
    group_cols = group_cols,
    model = model,
    window_samples = window_samples,
    output_col = output_col,
    overwrite = overwrite
  )
}

gpbiometrics_autoencoder_denoise_signal <- function(dat,
                                                    signal_col,
                                                    signal_type,
                                                    time_col = NULL,
                                                    group_cols = NULL,
                                                    model = NULL,
                                                    window_samples = 128,
                                                    output_col = NULL,
                                                    overwrite = FALSE) {
  if (!is.data.frame(dat)) {
    stop("`dat` must be a data frame.", call. = FALSE)
  }

  if (!signal_col %in% names(dat)) {
    stop("Column `", signal_col, "` was not found in `dat`.", call. = FALSE)
  }

  if (!is.numeric(dat[[signal_col]])) {
    stop("`signal_col` must identify a numeric column.", call. = FALSE)
  }

  if (!is.null(time_col) && !time_col %in% names(dat)) {
    stop("Column `", time_col, "` was not found in `dat`.", call. = FALSE)
  }

  if (is.null(group_cols)) {
    group_cols <- character()
  }

  missing_groups <- setdiff(group_cols, names(dat))
  if (length(missing_groups) > 0) {
    stop("Missing `group_cols`: ", paste(missing_groups, collapse = ", "), call. = FALSE)
  }

  if (is.null(output_col)) {
    output_col <- paste0(signal_col, "_autoencoder_denoised")
  }

  if (!isTRUE(overwrite) && output_col %in% names(dat)) {
    stop("Output column `", output_col, "` already exists. Use `overwrite = TRUE`.", call. = FALSE)
  }

  if (!is.numeric(window_samples) ||
      length(window_samples) != 1 ||
      !is.finite(window_samples) ||
      window_samples < 4) {
    stop("`window_samples` must be a positive number >= 4.", call. = FALSE)
  }

  window_samples <- as.integer(window_samples)

  out <- dat
  out[[output_col]] <- NA_real_

  status_col <- paste0(output_col, "_status")
  out[[status_col]] <- "not_processed"

  groups <- gpbiometrics_autoencoder_split(out, group_cols)

  summary_rows <- lapply(names(groups), function(group_id) {
    idx <- groups[[group_id]]

    if (!is.null(time_col)) {
      idx <- idx[order(out[[time_col]][idx])]
    }

    x <- out[[signal_col]][idx]
    finite <- is.finite(x)

    if (sum(finite) < 4) {
      out[[output_col]][idx] <<- NA_real_
      out[[status_col]][idx] <<- "insufficient_finite_samples"

      return(data.frame(
        group_id = group_id,
        n_rows = length(idx),
        n_finite = sum(finite),
        model_supplied = !is.null(model),
        status = "insufficient_finite_samples",
        stringsAsFactors = FALSE
      ))
    }

    x_filled <- gpbiometrics_autoencoder_fill(x)

    reconstructed <- if (is.null(model)) {
      x_filled
    } else {
      gpbiometrics_autoencoder_apply_model(
        model = model,
        x = x_filled,
        window_samples = window_samples
      )
    }

    reconstructed[!finite] <- NA_real_

    out[[output_col]][idx] <<- reconstructed
    out[[status_col]][idx] <<- if (is.null(model)) {
      "copied_no_autoencoder_model_supplied"
    } else {
      "autoencoder_reconstruction_applied"
    }

    data.frame(
      group_id = group_id,
      n_rows = length(idx),
      n_finite = sum(finite),
      model_supplied = !is.null(model),
      status = if (is.null(model)) {
        "copied_no_autoencoder_model_supplied"
      } else {
        "autoencoder_reconstruction_applied"
      },
      stringsAsFactors = FALSE
    )
  })

  summary_table <- do.call(rbind, summary_rows)
  rownames(summary_table) <- NULL

  overview <- data.frame(
    signal_type = signal_type,
    input_rows = nrow(dat),
    group_count = length(groups),
    model_supplied = !is.null(model),
    output_col = output_col,
    status = if (is.null(model)) {
      "autoencoder_bridge_no_model_supplied"
    } else {
      "autoencoder_reconstruction_complete"
    },
    interpretation = paste(
      "Autoencoder denoising requires a validated user-supplied reconstruction model.",
      "No pretrained neural network is bundled and outputs should be inspected against raw signals."
    ),
    stringsAsFactors = FALSE
  )

  attr(out, "autoencoder_denoising_overview") <- overview
  attr(out, "autoencoder_denoising_summary") <- summary_table
  attr(out, "autoencoder_denoising_settings") <- list(
    signal_col = signal_col,
    signal_type = signal_type,
    time_col = time_col,
    group_cols = group_cols,
    model_supplied = !is.null(model),
    window_samples = window_samples,
    output_col = output_col,
    overwrite = overwrite
  )

  class(out) <- unique(c("gazepoint_autoencoder_denoised", class(out)))
  out
}

gpbiometrics_autoencoder_split <- function(dat, group_cols) {
  if (length(group_cols) == 0) {
    return(list(all_rows = seq_len(nrow(dat))))
  }

  gf <- dat[group_cols]
  gf[] <- lapply(gf, function(x) {
    x <- as.character(x)
    x[is.na(x)] <- "<NA>"
    x
  })

  split(seq_len(nrow(dat)), do.call(paste, c(gf, sep = " | ")))
}

gpbiometrics_autoencoder_fill <- function(x) {
  idx <- seq_along(x)
  finite <- is.finite(x)

  if (all(finite)) {
    return(x)
  }

  if (sum(finite) == 0) {
    return(rep(0, length(x)))
  }

  if (sum(finite) == 1) {
    return(rep(x[finite][1], length(x)))
  }

  stats::approx(idx[finite], x[finite], xout = idx, rule = 2)$y
}

gpbiometrics_autoencoder_apply_model <- function(model,
                                                 x,
                                                 window_samples = 128) {
  starts <- seq(1, length(x), by = window_samples)
  recon <- numeric(length(x))

  for (start in starts) {
    end <- min(start + window_samples - 1, length(x))
    segment <- x[start:end]

    padded_length <- window_samples
    padded <- c(segment, rep(segment[length(segment)], padded_length - length(segment)))

    pred <- if (is.function(model)) {
      model(matrix(padded, nrow = 1))
    } else {
      tryCatch(
        stats::predict(model, newdata = matrix(padded, nrow = 1)),
        error = function(e) {
          stop("Could not obtain reconstruction from `model`: ", conditionMessage(e), call. = FALSE)
        }
      )
    }

    pred <- as.numeric(pred)

    if (length(pred) < length(segment)) {
      stop("Autoencoder reconstruction returned fewer values than the input segment.", call. = FALSE)
    }

    recon[start:end] <- pred[seq_along(segment)]
  }

  recon
}

#' Audit Gazepoint EDA/GSR artifacts
#'
#' Flags row-level artifacts in Gazepoint electrodermal activity signals,
#' preferring `GSR_US` conductance when available. The helper detects abrupt
#' jumps, abrupt slopes, flatline runs, zero runs, negative conductance values,
#' and optional out-of-range values. It is a conservative preprocessing/QC
#' helper and does not interpret EDA as emotional valence.
#'
#' @param data A data frame containing Gazepoint biometric rows.
#' @param signal_col Optional EDA/GSR signal column. If `NULL`, the function
#'   prefers `GSR_US` and then falls back to common Gazepoint EDA columns.
#' @param time_col Optional time/counter column. If `NULL`, common Gazepoint
#'   time columns are detected automatically.
#' @param group_cols Optional grouping columns. If `NULL`, available
#'   source/participant/media/trial-like columns are used.
#' @param prefer_gsr_us Logical. If `TRUE`, prefer `GSR_US` when `signal_col`
#'   is not supplied.
#' @param jump_threshold_sd Robust z threshold for absolute signal jumps.
#' @param slope_threshold_sd Robust z threshold for absolute signal slopes.
#' @param flat_run_length Minimum repeated-value run length flagged as flatline.
#' @param zero_run_length Minimum zero-value run length flagged as zero run.
#' @param saturation_min Optional lower bound for acceptable signal values.
#' @param saturation_max Optional upper bound for acceptable signal values.
#' @param negative_allowed Optional logical. If `NULL`, negative values are
#'   allowed for phasic component columns but not for conductance-like columns.
#'
#' @return A list with `overview`, `row_flags`, `artifact_runs`,
#'   `group_summary`, and `settings`.
#' @export
audit_gazepoint_eda_artifacts <- function(data,
                                          signal_col = NULL,
                                          time_col = NULL,
                                          group_cols = NULL,
                                          prefer_gsr_us = TRUE,
                                          jump_threshold_sd = 6,
                                          slope_threshold_sd = 6,
                                          flat_run_length = 20,
                                          zero_run_length = 20,
                                          saturation_min = NULL,
                                          saturation_max = NULL,
                                          negative_allowed = NULL) {
  if (!is.data.frame(data)) {
    stop("`data` must be a data frame.", call. = FALSE)
  }

  if (!is.logical(prefer_gsr_us) || length(prefer_gsr_us) != 1) {
    stop("`prefer_gsr_us` must be TRUE or FALSE.", call. = FALSE)
  }

  if (!is.numeric(jump_threshold_sd) ||
      length(jump_threshold_sd) != 1 ||
      jump_threshold_sd <= 0) {
    stop("`jump_threshold_sd` must be a positive number.", call. = FALSE)
  }

  if (!is.numeric(slope_threshold_sd) ||
      length(slope_threshold_sd) != 1 ||
      slope_threshold_sd <= 0) {
    stop("`slope_threshold_sd` must be a positive number.", call. = FALSE)
  }

  if (!is.numeric(flat_run_length) ||
      length(flat_run_length) != 1 ||
      flat_run_length < 1) {
    stop("`flat_run_length` must be a positive number.", call. = FALSE)
  }

  if (!is.numeric(zero_run_length) ||
      length(zero_run_length) != 1 ||
      zero_run_length < 1) {
    stop("`zero_run_length` must be a positive number.", call. = FALSE)
  }

  if (!is.null(saturation_min) &&
      (!is.numeric(saturation_min) || length(saturation_min) != 1)) {
    stop("`saturation_min` must be NULL or a single number.", call. = FALSE)
  }

  if (!is.null(saturation_max) &&
      (!is.numeric(saturation_max) || length(saturation_max) != 1)) {
    stop("`saturation_max` must be NULL or a single number.", call. = FALSE)
  }

  if (!is.null(negative_allowed) &&
      (!is.logical(negative_allowed) || length(negative_allowed) != 1)) {
    stop("`negative_allowed` must be NULL, TRUE, or FALSE.", call. = FALSE)
  }

  dat <- as.data.frame(data, stringsAsFactors = FALSE)
  dat$.gpbiometrics_row_id <- seq_len(nrow(dat))

  names_dat <- names(dat)

  if (is.null(signal_col)) {
    signal_col <- gpbiometrics_eda_artifact_infer_signal_col(
      names_dat = names_dat,
      prefer_gsr_us = prefer_gsr_us
    )
  }

  if (is.null(signal_col) || !signal_col %in% names_dat) {
    stop("No usable EDA/GSR signal column was found. Supply `signal_col`.", call. = FALSE)
  }

  if (is.null(time_col)) {
    time_col <- gpbiometrics_eda_artifact_first_existing(
      names_dat,
      c(
        "time_ms", "timestamp_ms", "timestamp",
        "TIME", "Time", "time",
        "CNT", "cnt"
      )
    )
  }

  if (!is.null(time_col) && !time_col %in% names_dat) {
    stop("`time_col` was not found in `data`.", call. = FALSE)
  }

  group_cols <- gpbiometrics_eda_artifact_resolve_group_cols(names_dat, group_cols)

  missing_group_cols <- setdiff(group_cols, names_dat)

  if (length(missing_group_cols) > 0) {
    stop(
      "`group_cols` not found in `data`: ",
      paste(missing_group_cols, collapse = ", "),
      call. = FALSE
    )
  }

  if (is.null(negative_allowed)) {
    negative_allowed <- grepl("PHASIC", signal_col, ignore.case = TRUE)
  }

  signal_value <- suppressWarnings(as.numeric(dat[[signal_col]]))

  if (all(is.na(signal_value))) {
    stop("`signal_col` must contain numeric or numeric-coercible values.", call. = FALSE)
  }

  time_value <- if (!is.null(time_col)) {
    suppressWarnings(as.numeric(dat[[time_col]]))
  } else {
    seq_len(nrow(dat))
  }

  group_id <- gpbiometrics_eda_artifact_group_id(dat, group_cols)
  group_indices <- split(seq_len(nrow(dat)), group_id, drop = TRUE)

  row_flags <- data.frame(
    .gpbiometrics_row_id = dat$.gpbiometrics_row_id,
    group_id = group_id,
    signal_col = signal_col,
    signal_value = signal_value,
    time_col = if (is.null(time_col)) NA_character_ else time_col,
    time_value = time_value,
    group_row_index = NA_integer_,
    signal_delta = NA_real_,
    time_delta = NA_real_,
    slope = NA_real_,
    robust_jump_z = NA_real_,
    robust_slope_z = NA_real_,
    flag_nonfinite_signal = !is.finite(signal_value),
    flag_nonfinite_time = !is.finite(time_value),
    flag_negative_conductance = FALSE,
    flag_jump = FALSE,
    flag_slope = FALSE,
    flag_flatline_run = FALSE,
    flag_zero_run = FALSE,
    flag_out_of_bounds = FALSE,
    flag_artifact = FALSE,
    stringsAsFactors = FALSE
  )

  if (length(group_cols) > 0) {
    row_flags <- cbind(dat[group_cols], row_flags)
  }

  for (group_name in names(group_indices)) {
    idx <- group_indices[[group_name]]

    row_flags$group_row_index[idx] <- seq_along(idx)

    x <- signal_value[idx]
    t <- time_value[idx]

    dx <- c(NA_real_, diff(x))
    dt <- c(NA_real_, diff(t))

    slope <- rep(NA_real_, length(idx))
    slope[is.finite(dx) & is.finite(dt) & dt != 0] <- dx[
      is.finite(dx) & is.finite(dt) & dt != 0
    ] / dt[is.finite(dx) & is.finite(dt) & dt != 0]

    row_flags$signal_delta[idx] <- dx
    row_flags$time_delta[idx] <- dt
    row_flags$slope[idx] <- slope

    jump_z <- gpbiometrics_eda_artifact_robust_z(dx)
    slope_z <- gpbiometrics_eda_artifact_robust_z(slope)

    row_flags$robust_jump_z[idx] <- jump_z
    row_flags$robust_slope_z[idx] <- slope_z

    row_flags$flag_jump[idx] <- is.finite(jump_z) & abs(jump_z) >= jump_threshold_sd
    row_flags$flag_slope[idx] <- is.finite(slope_z) & abs(slope_z) >= slope_threshold_sd

    row_flags$flag_flatline_run[idx] <- gpbiometrics_eda_artifact_run_flag(
      x = x,
      min_run = flat_run_length,
      zero_only = FALSE
    )

    row_flags$flag_zero_run[idx] <- gpbiometrics_eda_artifact_run_flag(
      x = x,
      min_run = zero_run_length,
      zero_only = TRUE
    )
  }

  if (isFALSE(negative_allowed)) {
    row_flags$flag_negative_conductance <- is.finite(signal_value) & signal_value < 0
  }

  if (!is.null(saturation_min)) {
    row_flags$flag_out_of_bounds <- row_flags$flag_out_of_bounds |
      (is.finite(signal_value) & signal_value < saturation_min)
  }

  if (!is.null(saturation_max)) {
    row_flags$flag_out_of_bounds <- row_flags$flag_out_of_bounds |
      (is.finite(signal_value) & signal_value > saturation_max)
  }

  row_flags$flag_artifact <- row_flags$flag_nonfinite_signal |
    row_flags$flag_negative_conductance |
    row_flags$flag_jump |
    row_flags$flag_slope |
    row_flags$flag_flatline_run |
    row_flags$flag_zero_run |
    row_flags$flag_out_of_bounds

  group_summary <- gpbiometrics_eda_artifact_group_summary(
    row_flags = row_flags,
    group_cols = group_cols
  )

  artifact_runs <- gpbiometrics_eda_artifact_runs(
    row_flags = row_flags,
    group_cols = group_cols
  )

  artifact_rows <- sum(row_flags$flag_artifact, na.rm = TRUE)
  artifact_prop <- if (nrow(row_flags) > 0) artifact_rows / nrow(row_flags) else NA_real_

  status <- if (artifact_rows == 0) {
    "pass"
  } else if (artifact_prop >= 0.50) {
    "fail_high_artifact_rate"
  } else {
    "warn_artifacts_detected"
  }

  overview <- data.frame(
    input_rows = nrow(dat),
    signal_col = signal_col,
    time_col = if (is.null(time_col)) NA_character_ else time_col,
    group_count = length(group_indices),
    artifact_rows = artifact_rows,
    artifact_prop = artifact_prop,
    nonfinite_signal_rows = sum(row_flags$flag_nonfinite_signal, na.rm = TRUE),
    negative_conductance_rows = sum(row_flags$flag_negative_conductance, na.rm = TRUE),
    jump_rows = sum(row_flags$flag_jump, na.rm = TRUE),
    slope_rows = sum(row_flags$flag_slope, na.rm = TRUE),
    flatline_run_rows = sum(row_flags$flag_flatline_run, na.rm = TRUE),
    zero_run_rows = sum(row_flags$flag_zero_run, na.rm = TRUE),
    out_of_bounds_rows = sum(row_flags$flag_out_of_bounds, na.rm = TRUE),
    status = status,
    stringsAsFactors = FALSE
  )

  structure(
    list(
      overview = overview,
      row_flags = row_flags,
      artifact_runs = artifact_runs,
      group_summary = group_summary,
      settings = list(
        signal_col = signal_col,
        time_col = time_col,
        group_cols = group_cols,
        prefer_gsr_us = prefer_gsr_us,
        jump_threshold_sd = jump_threshold_sd,
        slope_threshold_sd = slope_threshold_sd,
        flat_run_length = flat_run_length,
        zero_run_length = zero_run_length,
        saturation_min = saturation_min,
        saturation_max = saturation_max,
        negative_allowed = negative_allowed,
        interpretation_notes = c(
          "EDA/GSR artifacts are preprocessing and quality-control indicators, not psychological events.",
          "GSR/EDA should be interpreted as electrodermal or sympathetic arousal-related activity, not emotional valence.",
          "Prefer GSR_US conductance when available; use GSR conversion only conservatively when GSR_US is absent."
        )
      )
    ),
    class = c("gazepoint_eda_artifact_audit", "list")
  )
}

gpbiometrics_eda_artifact_first_existing <- function(names_dat, candidates) {
  exact <- candidates[candidates %in% names_dat]

  if (length(exact) > 0) {
    return(exact[1])
  }

  lower_names <- tolower(names_dat)
  lower_candidates <- tolower(candidates)
  idx <- match(lower_candidates, lower_names)
  idx <- idx[!is.na(idx)]

  if (length(idx) > 0) {
    return(names_dat[idx[1]])
  }

  NULL
}

gpbiometrics_eda_artifact_infer_signal_col <- function(names_dat,
                                                       prefer_gsr_us = TRUE) {
  if (isTRUE(prefer_gsr_us)) {
    preferred <- gpbiometrics_eda_artifact_first_existing(
      names_dat,
      c("GSR_US", "gsr_us")
    )

    if (!is.null(preferred)) {
      return(preferred)
    }
  }

  gpbiometrics_eda_artifact_first_existing(
    names_dat,
    c(
      "GSR_US", "gsr_us",
      "GSR_US_TONIC", "gsr_us_tonic",
      "GSR", "gsr",
      "GSR_US_PHASIC", "gsr_us_phasic"
    )
  )
}

gpbiometrics_eda_artifact_resolve_group_cols <- function(names_dat,
                                                         group_cols) {
  if (!is.null(group_cols)) {
    return(unique(group_cols))
  }

  candidates <- c(
    "source_file",
    "source_participant",
    "USER",
    "USER_FILE",
    "participant",
    "subject",
    "subject_id",
    "MEDIA_ID",
    "MEDIA_NAME",
    "media_id",
    "media_name",
    "trial",
    "trial_id",
    "trial_global"
  )

  unique(candidates[candidates %in% names_dat])
}

gpbiometrics_eda_artifact_group_id <- function(dat, group_cols) {
  if (length(group_cols) == 0) {
    return(rep("all", nrow(dat)))
  }

  group_dat <- dat[group_cols]

  group_dat[] <- lapply(group_dat, function(x) {
    x_chr <- as.character(x)
    x_chr[is.na(x_chr)] <- "<NA>"
    x_chr
  })

  apply(group_dat, 1, paste, collapse = "||")
}

gpbiometrics_eda_artifact_robust_z <- function(x) {
  out <- rep(NA_real_, length(x))

  finite <- is.finite(x)

  if (sum(finite) < 3) {
    return(out)
  }

  x_finite <- x[finite]
  center <- stats::median(x_finite, na.rm = TRUE)
  mad_value <- stats::median(abs(x_finite - center), na.rm = TRUE)

  scale_value <- mad_value * 1.4826

  if (!is.finite(scale_value) || scale_value == 0) {
    scale_value <- stats::sd(x_finite, na.rm = TRUE)
  }

  if (!is.finite(scale_value) || scale_value == 0) {
    out[finite] <- 0
    return(out)
  }

  out[finite] <- (x_finite - center) / scale_value
  out
}

gpbiometrics_eda_artifact_run_flag <- function(x,
                                               min_run,
                                               zero_only = FALSE) {
  out <- rep(FALSE, length(x))

  if (length(x) == 0) {
    return(out)
  }

  values <- suppressWarnings(as.numeric(x))
  key <- ifelse(is.na(values), "<NA>", as.character(values))

  if (isTRUE(zero_only)) {
    key[!(is.finite(values) & values == 0)] <- paste0("nonzero_", seq_along(values))[
      !(is.finite(values) & values == 0)
    ]
  }

  run_info <- rle(key)
  run_lengths <- run_info$lengths
  run_values <- run_info$values
  run_ends <- cumsum(run_lengths)
  run_starts <- run_ends - run_lengths + 1L

  for (i in seq_along(run_lengths)) {
    if (run_lengths[i] >= min_run) {
      if (isTRUE(zero_only)) {
        if (identical(run_values[i], "0")) {
          out[run_starts[i]:run_ends[i]] <- TRUE
        }
      } else {
        if (!identical(run_values[i], "<NA>")) {
          out[run_starts[i]:run_ends[i]] <- TRUE
        }
      }
    }
  }

  out
}

gpbiometrics_eda_artifact_group_summary <- function(row_flags,
                                                    group_cols) {
  group_ids <- unique(row_flags$group_id)

  out <- lapply(group_ids, function(group_id_i) {
    d <- row_flags[row_flags$group_id == group_id_i, , drop = FALSE]

    artifact_rows <- sum(d$flag_artifact, na.rm = TRUE)
    artifact_prop <- if (nrow(d) > 0) artifact_rows / nrow(d) else NA_real_

    row <- data.frame(
      group_id = group_id_i,
      rows = nrow(d),
      artifact_rows = artifact_rows,
      artifact_prop = artifact_prop,
      nonfinite_signal_rows = sum(d$flag_nonfinite_signal, na.rm = TRUE),
      negative_conductance_rows = sum(d$flag_negative_conductance, na.rm = TRUE),
      jump_rows = sum(d$flag_jump, na.rm = TRUE),
      slope_rows = sum(d$flag_slope, na.rm = TRUE),
      flatline_run_rows = sum(d$flag_flatline_run, na.rm = TRUE),
      zero_run_rows = sum(d$flag_zero_run, na.rm = TRUE),
      out_of_bounds_rows = sum(d$flag_out_of_bounds, na.rm = TRUE),
      status = if (artifact_rows == 0) {
        "pass"
      } else if (artifact_prop >= 0.50) {
        "fail_high_artifact_rate"
      } else {
        "warn_artifacts_detected"
      },
      stringsAsFactors = FALSE
    )

    if (length(group_cols) > 0) {
      group_values <- d[1, group_cols, drop = FALSE]
      row <- cbind(group_values, row)
    }

    row
  })

  out <- do.call(rbind, out)
  rownames(out) <- NULL

  out
}

gpbiometrics_eda_artifact_runs <- function(row_flags,
                                           group_cols) {
  artifact_types <- c(
    "flag_nonfinite_signal",
    "flag_negative_conductance",
    "flag_jump",
    "flag_slope",
    "flag_flatline_run",
    "flag_zero_run",
    "flag_out_of_bounds"
  )

  out <- list()
  group_ids <- unique(row_flags$group_id)

  for (group_id_i in group_ids) {
    idx_group <- which(row_flags$group_id == group_id_i)

    for (artifact_type in artifact_types) {
      flag <- row_flags[[artifact_type]][idx_group]

      if (!any(flag, na.rm = TRUE)) {
        next
      }

      run_info <- rle(flag)
      run_lengths <- run_info$lengths
      run_values <- run_info$values
      run_ends <- cumsum(run_lengths)
      run_starts <- run_ends - run_lengths + 1L

      artifact_runs <- which(run_values)

      for (run_i in artifact_runs) {
        idx_local <- run_starts[run_i]:run_ends[run_i]
        idx <- idx_group[idx_local]

        row <- data.frame(
          group_id = group_id_i,
          artifact_type = artifact_type,
          start_row_id = row_flags$.gpbiometrics_row_id[idx[1]],
          end_row_id = row_flags$.gpbiometrics_row_id[idx[length(idx)]],
          rows = length(idx),
          start_time = row_flags$time_value[idx[1]],
          end_time = row_flags$time_value[idx[length(idx)]],
          stringsAsFactors = FALSE
        )

        if (length(group_cols) > 0) {
          group_values <- row_flags[idx[1], group_cols, drop = FALSE]
          row <- cbind(group_values, row)
        }

        out[[length(out) + 1L]] <- row
      }
    }
  }

  if (length(out) == 0) {
    empty <- data.frame(
      group_id = character(),
      artifact_type = character(),
      start_row_id = integer(),
      end_row_id = integer(),
      rows = integer(),
      start_time = numeric(),
      end_time = numeric(),
      stringsAsFactors = FALSE
    )

    if (length(group_cols) > 0) {
      for (col in rev(group_cols)) {
        empty <- cbind(stats::setNames(data.frame(character(), stringsAsFactors = FALSE), col), empty)
      }
    }

    return(empty)
  }

  out <- do.call(rbind, out)
  rownames(out) <- NULL

  out
}

test_that("audit_gazepoint_eda_artifacts passes clean conductance data", {
  dat <- data.frame(
    source_file = "clean.csv",
    CNT = seq_len(60),
    GSR_US = seq(1, 2, length.out = 60)
  )

  res <- audit_gazepoint_eda_artifacts(
    dat,
    signal_col = "GSR_US",
    time_col = "CNT",
    group_cols = "source_file",
    flat_run_length = 10,
    zero_run_length = 10
  )

  expect_s3_class(res, "gazepoint_eda_artifact_audit")
  expect_equal(res$overview$status, "pass")
  expect_equal(res$overview$artifact_rows, 0)
  expect_equal(nrow(res$artifact_runs), 0)
})

test_that("audit_gazepoint_eda_artifacts auto-detects GSR_US", {
  dat <- data.frame(
    CNT = seq_len(30),
    GSR = seq(1000000, 900000, length.out = 30),
    GSR_US = seq(1, 1.5, length.out = 30)
  )

  res <- audit_gazepoint_eda_artifacts(dat)

  expect_equal(res$overview$signal_col, "GSR_US")
  expect_equal(res$settings$signal_col, "GSR_US")
})

test_that("audit_gazepoint_eda_artifacts detects jumps and slopes", {
  dat <- data.frame(
    source_file = "jump.csv",
    CNT = seq_len(61),
    GSR_US = c(
      seq(1, 1.5, length.out = 30),
      8,
      seq(1.6, 2.0, length.out = 30)
    )
  )

  res <- audit_gazepoint_eda_artifacts(
    dat,
    signal_col = "GSR_US",
    time_col = "CNT",
    group_cols = "source_file",
    jump_threshold_sd = 4,
    slope_threshold_sd = 4,
    flat_run_length = 10,
    zero_run_length = 10
  )

  expect_equal(res$overview$status, "warn_artifacts_detected")
  expect_true(res$overview$jump_rows > 0)
  expect_true(res$overview$slope_rows > 0)
  expect_true(any(res$row_flags$flag_jump))
  expect_true(any(res$row_flags$flag_slope))
})

test_that("audit_gazepoint_eda_artifacts detects flatline and zero runs", {
  dat <- data.frame(
    source_file = "flat.csv",
    CNT = seq_len(30),
    GSR_US = c(rep(0, 10), seq(1, 2, length.out = 20))
  )

  res <- audit_gazepoint_eda_artifacts(
    dat,
    signal_col = "GSR_US",
    time_col = "CNT",
    group_cols = "source_file",
    flat_run_length = 5,
    zero_run_length = 5
  )

  expect_true(res$overview$flatline_run_rows >= 10)
  expect_true(res$overview$zero_run_rows >= 10)
  expect_true(any(res$artifact_runs$artifact_type == "flag_zero_run"))
})

test_that("audit_gazepoint_eda_artifacts detects negative conductance and bounds", {
  dat <- data.frame(
    CNT = seq_len(6),
    GSR_US = c(1, 1.1, -0.2, 1.2, 9, 1.3)
  )

  res <- audit_gazepoint_eda_artifacts(
    dat,
    signal_col = "GSR_US",
    time_col = "CNT",
    saturation_min = 0,
    saturation_max = 5,
    flat_run_length = 3,
    zero_run_length = 3
  )

  expect_true(res$overview$negative_conductance_rows >= 1)
  expect_true(res$overview$out_of_bounds_rows >= 2)
  expect_true(any(res$row_flags$flag_negative_conductance))
  expect_true(any(res$row_flags$flag_out_of_bounds))
})

test_that("audit_gazepoint_eda_artifacts allows negative phasic values by default", {
  dat <- data.frame(
    CNT = seq_len(10),
    GSR_US_PHASIC = c(-0.1, -0.05, 0, 0.05, 0.1, 0.05, 0, -0.03, -0.02, 0)
  )

  res <- audit_gazepoint_eda_artifacts(
    dat,
    signal_col = "GSR_US_PHASIC",
    time_col = "CNT",
    flat_run_length = 5,
    zero_run_length = 5
  )

  expect_false(any(res$row_flags$flag_negative_conductance))
  expect_true(res$settings$negative_allowed)
})

test_that("audit_gazepoint_eda_artifacts respects groups when computing jumps", {
  dat <- data.frame(
    source_file = rep(c("a.csv", "b.csv"), each = 5),
    CNT = rep(seq_len(5), 2),
    GSR_US = c(seq(1, 2, length.out = 5), seq(10, 11, length.out = 5))
  )

  res <- audit_gazepoint_eda_artifacts(
    dat,
    signal_col = "GSR_US",
    time_col = "CNT",
    group_cols = "source_file",
    jump_threshold_sd = 4,
    slope_threshold_sd = 4,
    flat_run_length = 3,
    zero_run_length = 3
  )

  expect_equal(res$overview$status, "pass")
  expect_equal(res$overview$jump_rows, 0)
  expect_equal(res$overview$slope_rows, 0)
})

test_that("audit_gazepoint_eda_artifacts errors for nonnumeric signal", {
  dat <- data.frame(
    CNT = seq_len(3),
    GSR_US = c("a", "b", "c")
  )

  expect_error(
    audit_gazepoint_eda_artifacts(dat, signal_col = "GSR_US"),
    "`signal_col` must contain numeric"
  )
})

#' Plot an SCR specification curve
#'
#' Plots a specification-curve style display from the output of
#' `run_gazepoint_scr_multiverse()` or from a compatible data frame.
#'
#' @param x Output from `run_gazepoint_scr_multiverse()` or a data frame.
#' @param estimate_col Column to rank and plot. Defaults to
#'   `"mean_response_amplitude"` when available, otherwise `"response_rate"`.
#' @param specification_col Specification identifier column.
#' @param add_zero_line Logical. If `TRUE`, draw a horizontal zero line.
#' @param main Plot title.
#'
#' @return Invisibly returns a list with plot data and settings.
#' @export
plot_gazepoint_scr_specification_curve <- function(x,
                                                   estimate_col = NULL,
                                                   specification_col = "specification_id",
                                                   add_zero_line = TRUE,
                                                   main = "SCR specification curve") {
  if (inherits(x, "gazepoint_scr_multiverse")) {
    dat <- x$specification_summary
  } else if (is.data.frame(x)) {
    dat <- x
  } else {
    stop(
      "`x` must be output from `run_gazepoint_scr_multiverse()` or a data frame.",
      call. = FALSE
    )
  }

  if (!is.data.frame(dat) || nrow(dat) == 0) {
    stop("No specification-summary data available to plot.", call. = FALSE)
  }

  if (is.null(estimate_col)) {
    estimate_col <- if ("mean_response_amplitude" %in% names(dat)) {
      "mean_response_amplitude"
    } else if ("response_rate" %in% names(dat)) {
      "response_rate"
    } else {
      stop(
        "Could not infer `estimate_col`; supply it explicitly.",
        call. = FALSE
      )
    }
  }

  if (!estimate_col %in% names(dat)) {
    stop("Column `", estimate_col, "` was not found.", call. = FALSE)
  }

  if (!specification_col %in% names(dat)) {
    stop("Column `", specification_col, "` was not found.", call. = FALSE)
  }

  plot_data <- dat[is.finite(dat[[estimate_col]]), , drop = FALSE]

  if (nrow(plot_data) == 0) {
    stop("No finite estimates available to plot.", call. = FALSE)
  }

  plot_data <- plot_data[order(plot_data[[estimate_col]]), , drop = FALSE]
  plot_data$specification_rank <- seq_len(nrow(plot_data))

  graphics::plot(
    plot_data$specification_rank,
    plot_data[[estimate_col]],
    type = "b",
    xlab = "Specification rank",
    ylab = estimate_col,
    main = main
  )

  if (isTRUE(add_zero_line)) {
    graphics::abline(h = 0, lty = 2)
  }

  invisible(
    structure(
      list(
        plot_data = plot_data,
        settings = list(
          estimate_col = estimate_col,
          specification_col = specification_col,
          add_zero_line = add_zero_line,
          main = main
        ),
        interpretation = paste(
          "The specification curve ranks estimates across defensible SCR scoring specifications.",
          "It supports sensitivity reporting and does not identify a universally correct specification."
        )
      ),
      class = c("gazepoint_scr_specification_curve_plot", "list")
    )
  )
}

#' Summarise Gazepoint SCR responses in event windows
#'
#' Creates one row per event or TTL marker and links detected SCR peaks to
#' event-relative analysis and response windows. The helper is intended to
#' produce transparent event-level EDA/SCR features for downstream mixed models,
#' hurdle models, and reporting.
#'
#' @param data Optional Gazepoint biometric data frame. Required when `events`
#'   is `NULL` and events should be derived from TTL columns.
#' @param scr_peaks A `gazepoint_scr_peak_detection` object returned by
#'   `detect_gazepoint_scr_peaks()`, or a data frame of detected peaks.
#' @param events Optional event data frame. If `NULL`, events are derived from
#'   TTL columns in `data`.
#' @param time_col Optional time/counter column in `data`. Used for TTL-derived
#'   events.
#' @param event_time_col Optional event-time column in `events`. If `NULL`,
#'   common event-time column names are detected.
#' @param event_id_col Optional event identifier column in `events`.
#' @param event_label_col Optional event label/condition column in `events`.
#' @param group_cols Optional grouping columns used to match events and peaks.
#' @param ttl_cols Optional TTL marker columns used when `events = NULL`.
#' @param ttl_valid_col Optional TTL validity column. If supplied, TTL-derived
#'   events require this column to be non-zero.
#' @param event_detection Event-detection rule for TTL columns. `"rising"`
#'   detects rising edges; `"active"` treats every active TTL row as an event.
#' @param analysis_window Numeric length-two vector giving the event-relative
#'   analysis window in the same units as `time_col` or `event_time_col`.
#' @param response_window Numeric length-two vector giving the event-relative
#'   response window used for the binary SCR response flag.
#' @param amplitude_col Column in the peak table containing SCR amplitude.
#' @param peak_time_col Column in the peak table containing peak time.
#' @param onset_time_col Column in the peak table containing onset time.
#' @param rise_time_col Column in the peak table containing SCR rise time.
#' @param recovery_time_col Column in the peak table containing recovery time
#'   after peak.
#' @param peak_status_col Column in the peak table containing peak status.
#' @param peak_selection How to choose one peak when several peaks fall in the
#'   response window. `"largest_amplitude"` selects the largest response;
#'   `"first_peak"` selects the earliest peak.
#' @param collapse_simultaneous_events Logical. If `TRUE`, events with the same
#'   group and event time are collapsed into one row before matching peaks. This
#'   is useful when Gazepoint TTL0--TTL6 channels mark the same event
#'   simultaneously.
#'
#' @return A list with `overview`, `event_table`, `window_qc`, `events`,
#'   `peaks`, and `settings`.
#' @export
summarise_gazepoint_scr_event_windows <- function(data = NULL,
                                                  scr_peaks,
                                                  events = NULL,
                                                  time_col = NULL,
                                                  event_time_col = NULL,
                                                  event_id_col = NULL,
                                                  event_label_col = NULL,
                                                  group_cols = NULL,
                                                  ttl_cols = NULL,
                                                  ttl_valid_col = NULL,
                                                  event_detection = c("rising", "active"),
                                                  analysis_window = c(0, 6),
                                                  response_window = c(1, 4),
                                                  amplitude_col = "amplitude",
                                                  peak_time_col = "peak_time",
                                                  onset_time_col = "onset_time",
                                                  rise_time_col = "rise_time",
                                                  recovery_time_col = "recovery_time_after_peak",
                                                  peak_status_col = "status",
                                                  peak_selection = c("largest_amplitude", "first_peak"),
                                                  collapse_simultaneous_events = FALSE) {
  event_detection <- match.arg(event_detection)
  peak_selection <- match.arg(peak_selection)

  if (missing(scr_peaks) || is.null(scr_peaks)) {
    stop("`scr_peaks` must be supplied.", call. = FALSE)
  }

  if (!is.numeric(analysis_window) ||
      length(analysis_window) != 2 ||
      any(!is.finite(analysis_window)) ||
      analysis_window[2] < analysis_window[1]) {
    stop("`analysis_window` must be a finite length-two vector with end >= start.", call. = FALSE)
  }

  if (!is.numeric(response_window) ||
      length(response_window) != 2 ||
      any(!is.finite(response_window)) ||
      response_window[2] < response_window[1]) {
    stop("`response_window` must be a finite length-two vector with end >= start.", call. = FALSE)
  }

  if (response_window[1] < analysis_window[1] ||
      response_window[2] > analysis_window[2]) {
    stop("`response_window` must fall inside `analysis_window`.", call. = FALSE)
  }

  if (!is.logical(collapse_simultaneous_events) ||
      length(collapse_simultaneous_events) != 1 ||
      is.na(collapse_simultaneous_events)) {
    stop("`collapse_simultaneous_events` must be TRUE or FALSE.", call. = FALSE)
  }

  peaks <- gpbiometrics_scr_event_extract_peaks(scr_peaks)

  required_peak_cols <- c(amplitude_col, peak_time_col)
  missing_peak_cols <- setdiff(required_peak_cols, names(peaks))

  if (length(missing_peak_cols) > 0) {
    stop(
      "`scr_peaks` is missing required columns: ",
      paste(missing_peak_cols, collapse = ", "),
      call. = FALSE
    )
  }

  if (is.null(events)) {
    if (is.null(data) || !is.data.frame(data)) {
      stop("`data` must be supplied when `events` is NULL.", call. = FALSE)
    }

    data <- as.data.frame(data, stringsAsFactors = FALSE)

    if (is.null(time_col)) {
      time_col <- gpbiometrics_scr_event_first_existing(
        names(data),
        c(
          "time_ms", "timestamp_ms", "timestamp",
          "TIME", "Time", "time",
          "CNT", "cnt"
        )
      )
    }

    if (is.null(time_col) || !time_col %in% names(data)) {
      stop("No usable time column was found. Supply `time_col`.", call. = FALSE)
    }

    if (is.null(group_cols)) {
      group_cols <- gpbiometrics_scr_event_resolve_group_cols(names(data))
    }

    events_standard <- gpbiometrics_scr_event_derive_ttl_events(
      data = data,
      time_col = time_col,
      group_cols = group_cols,
      ttl_cols = ttl_cols,
      ttl_valid_col = ttl_valid_col,
      event_detection = event_detection
    )
  } else {
    events <- as.data.frame(events, stringsAsFactors = FALSE)

    if (is.null(event_time_col)) {
      event_time_col <- gpbiometrics_scr_event_first_existing(
        names(events),
        c(
          "event_time", "onset_time", "stimulus_time",
          "time_ms", "timestamp_ms", "timestamp",
          "TIME", "Time", "time",
          "CNT", "cnt"
        )
      )
    }

    if (is.null(event_time_col) || !event_time_col %in% names(events)) {
      stop("No usable event-time column was found. Supply `event_time_col`.", call. = FALSE)
    }

    if (is.null(group_cols)) {
      group_cols <- gpbiometrics_scr_event_resolve_common_group_cols(events, peaks)
    }

    events_standard <- gpbiometrics_scr_event_standardise_events(
      events = events,
      event_time_col = event_time_col,
      event_id_col = event_id_col,
      event_label_col = event_label_col,
      group_cols = group_cols
    )
  }

  missing_event_group_cols <- setdiff(group_cols, names(events_standard))
  missing_peak_group_cols <- setdiff(group_cols, names(peaks))

  if (length(group_cols) > 0 && length(missing_event_group_cols) > 0) {
    stop(
      "`events` is missing grouping columns: ",
      paste(missing_event_group_cols, collapse = ", "),
      call. = FALSE
    )
  }

  if (length(group_cols) > 0 &&
      length(missing_peak_group_cols) > 0 &&
      !"group_id" %in% names(peaks)) {
    stop(
      "`scr_peaks` is missing grouping columns: ",
      paste(missing_peak_group_cols, collapse = ", "),
      call. = FALSE
    )
  }

  events_standard$event_group_id <- gpbiometrics_scr_event_group_id(
    events_standard,
    group_cols
  )

  if (isTRUE(collapse_simultaneous_events)) {
    events_standard <- gpbiometrics_scr_event_collapse_simultaneous_events(
      events_standard
    )
  }

  peaks$event_group_id <- gpbiometrics_scr_event_group_id(
    peaks,
    group_cols
  )

  event_table <- gpbiometrics_scr_event_match_events_to_peaks(
    events = events_standard,
    peaks = peaks,
    group_cols = group_cols,
    analysis_window = analysis_window,
    response_window = response_window,
    amplitude_col = amplitude_col,
    peak_time_col = peak_time_col,
    onset_time_col = onset_time_col,
    rise_time_col = rise_time_col,
    recovery_time_col = recovery_time_col,
    peak_status_col = peak_status_col,
    peak_selection = peak_selection
  )

  window_qc <- gpbiometrics_scr_event_window_qc(
    event_table = event_table,
    group_cols = group_cols
  )

  event_count <- nrow(event_table)
  response_events <- sum(event_table$response_flag == 1, na.rm = TRUE)
  no_response_events <- sum(event_table$response_flag == 0, na.rm = TRUE)

  status <- if (event_count == 0) {
    "fail_no_events"
  } else if (response_events == 0) {
    "warn_no_scr_responses"
  } else {
    "scr_event_windows_summarised"
  }

  overview <- data.frame(
    event_count = event_count,
    peak_count = nrow(peaks),
    response_events = response_events,
    no_response_events = no_response_events,
    response_rate = if (event_count > 0) response_events / event_count else NA_real_,
    group_count = length(unique(events_standard$event_group_id)),
    analysis_window_start = analysis_window[1],
    analysis_window_end = analysis_window[2],
    response_window_start = response_window[1],
    response_window_end = response_window[2],
    status = status,
    stringsAsFactors = FALSE
  )

  structure(
    list(
      overview = overview,
      event_table = event_table,
      window_qc = window_qc,
      events = events_standard,
      peaks = peaks,
      settings = list(
        time_col = time_col,
        event_time_col = event_time_col,
        event_id_col = event_id_col,
        event_label_col = event_label_col,
        group_cols = group_cols,
        ttl_cols = ttl_cols,
        ttl_valid_col = ttl_valid_col,
        event_detection = event_detection,
        analysis_window = analysis_window,
        response_window = response_window,
        amplitude_col = amplitude_col,
        peak_time_col = peak_time_col,
        onset_time_col = onset_time_col,
        rise_time_col = rise_time_col,
        recovery_time_col = recovery_time_col,
        peak_status_col = peak_status_col,
        peak_selection = peak_selection,
        collapse_simultaneous_events = collapse_simultaneous_events,
        interpretation_notes = c(
          "Event-window SCR summaries are electrodermal response features, not emotion or valence labels.",
          "Latency and amplitude depend on the selected response window, threshold, and peak-detection settings.",
          "For dense or overlapping event designs, sensitivity checks and optional external deconvolution cross-checks should be considered."
        )
      )
    ),
    class = c("gazepoint_scr_event_window_summary", "list")
  )
}

gpbiometrics_scr_event_extract_peaks <- function(scr_peaks) {
  if (inherits(scr_peaks, "gazepoint_scr_peak_detection") &&
      !is.null(scr_peaks$peaks)) {
    return(as.data.frame(scr_peaks$peaks, stringsAsFactors = FALSE))
  }

  if (is.data.frame(scr_peaks)) {
    return(as.data.frame(scr_peaks, stringsAsFactors = FALSE))
  }

  stop("`scr_peaks` must be a peak-detection object or a data frame.", call. = FALSE)
}

gpbiometrics_scr_event_first_existing <- function(names_dat, candidates) {
  exact <- candidates[candidates %in% names_dat]

  if (length(exact) > 0) {
    return(exact[1])
  }

  lower_names <- tolower(names_dat)
  lower_candidates <- tolower(candidates)
  idx <- match(lower_candidates, lower_names)
  idx <- idx[!is.na(idx)]

  if (length(idx) > 0) {
    return(names_dat[idx[1]])
  }

  NULL
}

gpbiometrics_scr_event_resolve_group_cols <- function(names_dat) {
  candidates <- c(
    "source_file",
    "source_participant",
    "USER",
    "USER_FILE",
    "participant",
    "subject",
    "subject_id",
    "MEDIA_ID",
    "MEDIA_NAME",
    "media_id",
    "media_name",
    "trial",
    "trial_id",
    "trial_global"
  )

  unique(candidates[candidates %in% names_dat])
}

gpbiometrics_scr_event_resolve_common_group_cols <- function(events, peaks) {
  candidates <- gpbiometrics_scr_event_resolve_group_cols(
    intersect(names(events), names(peaks))
  )

  candidates[candidates %in% names(events) & candidates %in% names(peaks)]
}

gpbiometrics_scr_event_group_id <- function(dat, group_cols) {
  if (length(group_cols) == 0) {
    if ("group_id" %in% names(dat)) {
      return(as.character(dat$group_id))
    }

    return(rep("all", nrow(dat)))
  }

  if (!all(group_cols %in% names(dat))) {
    if ("group_id" %in% names(dat)) {
      return(as.character(dat$group_id))
    }

    return(rep("all", nrow(dat)))
  }

  group_dat <- dat[group_cols]

  group_dat[] <- lapply(group_dat, function(x) {
    x_chr <- as.character(x)
    x_chr[is.na(x_chr)] <- "<NA>"
    x_chr
  })

  apply(group_dat, 1, paste, collapse = "||")
}

gpbiometrics_scr_event_standardise_events <- function(events,
                                                      event_time_col,
                                                      event_id_col,
                                                      event_label_col,
                                                      group_cols) {
  out <- events

  event_time <- suppressWarnings(as.numeric(out[[event_time_col]]))

  if (all(is.na(event_time))) {
    stop("`event_time_col` must contain numeric or numeric-coercible values.", call. = FALSE)
  }

  out$event_time <- event_time

  if (!is.null(event_id_col) && event_id_col %in% names(out)) {
    out$event_id <- as.character(out[[event_id_col]])
  } else if ("event_id" %in% names(out)) {
    out$event_id <- as.character(out$event_id)
  } else {
    out$event_id <- paste0("event_", seq_len(nrow(out)))
  }

  if (!is.null(event_label_col) && event_label_col %in% names(out)) {
    out$event_label <- as.character(out[[event_label_col]])
  } else if ("event_label" %in% names(out)) {
    out$event_label <- as.character(out$event_label)
  } else if ("condition" %in% names(out)) {
    out$event_label <- as.character(out$condition)
  } else {
    out$event_label <- "event"
  }

  out$source_row_id <- if ("source_row_id" %in% names(out)) {
    out$source_row_id
  } else {
    NA_integer_
  }

  missing_group_cols <- setdiff(group_cols, names(out))

  if (length(missing_group_cols) > 0) {
    stop(
      "`events` is missing grouping columns: ",
      paste(missing_group_cols, collapse = ", "),
      call. = FALSE
    )
  }

  out
}

gpbiometrics_scr_event_derive_ttl_events <- function(data,
                                                     time_col,
                                                     group_cols,
                                                     ttl_cols,
                                                     ttl_valid_col,
                                                     event_detection) {
  names_dat <- names(data)

  if (is.null(ttl_cols)) {
    ttl_cols <- grep("^TTL[0-9]+$", names_dat, value = TRUE)

    if ("ttl_marker" %in% names_dat) {
      ttl_cols <- c(ttl_cols, "ttl_marker")
    }
  }

  ttl_cols <- unique(ttl_cols)

  if (length(ttl_cols) == 0) {
    stop("No TTL columns were found. Supply `events` or `ttl_cols`.", call. = FALSE)
  }

  missing_ttl <- setdiff(ttl_cols, names_dat)

  if (length(missing_ttl) > 0) {
    stop(
      "`ttl_cols` not found in `data`: ",
      paste(missing_ttl, collapse = ", "),
      call. = FALSE
    )
  }

  if (!is.null(ttl_valid_col) && !ttl_valid_col %in% names_dat) {
    stop("`ttl_valid_col` was not found in `data`.", call. = FALSE)
  }

  missing_group_cols <- setdiff(group_cols, names_dat)

  if (length(missing_group_cols) > 0) {
    stop(
      "`data` is missing grouping columns: ",
      paste(missing_group_cols, collapse = ", "),
      call. = FALSE
    )
  }

  data$.gpbiometrics_row_id <- seq_len(nrow(data))
  group_id <- gpbiometrics_scr_event_group_id(data, group_cols)
  group_indices <- split(seq_len(nrow(data)), group_id, drop = TRUE)

  out <- list()

  for (group_name in names(group_indices)) {
    idx <- group_indices[[group_name]]

    valid <- rep(TRUE, length(idx))

    if (!is.null(ttl_valid_col)) {
      valid_raw <- suppressWarnings(as.numeric(data[[ttl_valid_col]][idx]))
      valid <- is.finite(valid_raw) & valid_raw != 0
    }

    for (ttl_col in ttl_cols) {
      ttl_raw <- suppressWarnings(as.numeric(data[[ttl_col]][idx]))
      active <- is.finite(ttl_raw) & ttl_raw != 0 & valid

      event_local <- if (identical(event_detection, "rising")) {
        active & !c(FALSE, utils::head(active, -1))
      } else {
        active
      }

      event_idx <- idx[which(event_local)]

      if (length(event_idx) == 0) {
        next
      }

      for (row_i in event_idx) {
        row <- data.frame(
          event_id = paste0(ttl_col, "_row_", data$.gpbiometrics_row_id[row_i]),
          event_label = ttl_col,
          event_source = "ttl",
          event_value = suppressWarnings(as.numeric(data[[ttl_col]][row_i])),
          event_time = suppressWarnings(as.numeric(data[[time_col]][row_i])),
          source_row_id = data$.gpbiometrics_row_id[row_i],
          stringsAsFactors = FALSE
        )

        if (length(group_cols) > 0) {
          group_values <- data[row_i, group_cols, drop = FALSE]
          row <- cbind(group_values, row)
        }

        out[[length(out) + 1L]] <- row
      }
    }
  }

  if (length(out) == 0) {
    empty <- data.frame(stringsAsFactors = FALSE)

    if (length(group_cols) > 0) {
      for (col in group_cols) {
        empty[[col]] <- character()
      }
    }

    empty$event_id <- character()
    empty$event_label <- character()
    empty$event_source <- character()
    empty$event_value <- numeric()
    empty$event_time <- numeric()
    empty$source_row_id <- integer()

    return(empty)
  }

  out <- do.call(rbind, out)
  rownames(out) <- NULL

  out
}

gpbiometrics_scr_event_collapse_simultaneous_events <- function(events) {
  if (nrow(events) == 0) {
    events$collapsed_event_count <- integer()
    events$collapsed_event_ids <- character()
    events$collapsed_event_labels <- character()
    return(events)
  }

  if (nrow(events) == 1) {
    events$collapsed_event_count <- 1L
    events$collapsed_event_ids <- as.character(events$event_id)
    events$collapsed_event_labels <- as.character(events$event_label)
    return(events)
  }

  event_time <- suppressWarnings(as.numeric(events$event_time))

  collapse_key <- ifelse(
    is.finite(event_time),
    paste(
      events$event_group_id,
      format(event_time, digits = 15, scientific = FALSE),
      sep = "||"
    ),
    paste(events$event_group_id, "nonfinite", seq_len(nrow(events)), sep = "||")
  )

  split_events <- split(events, collapse_key, drop = TRUE)

  collapsed <- lapply(split_events, function(d) {
    row <- d[1, , drop = FALSE]

    event_ids <- unique(as.character(d$event_id))
    event_labels <- unique(as.character(d$event_label))

    event_ids <- event_ids[!is.na(event_ids) & nzchar(event_ids)]
    event_labels <- event_labels[!is.na(event_labels) & nzchar(event_labels)]

    if (length(event_ids) == 0) {
      event_ids <- "event"
    }

    if (length(event_labels) == 0) {
      event_labels <- "event"
    }

    row$event_id <- paste(event_ids, collapse = "+")
    row$event_label <- paste(event_labels, collapse = "+")
    row$collapsed_event_count <- nrow(d)
    row$collapsed_event_ids <- paste(event_ids, collapse = "+")
    row$collapsed_event_labels <- paste(event_labels, collapse = "+")

    if ("event_value" %in% names(row)) {
      event_values <- suppressWarnings(as.numeric(d$event_value))
      finite_values <- event_values[is.finite(event_values)]

      row$event_value <- if (length(finite_values) == 1) {
        finite_values[1]
      } else {
        NA_real_
      }
    }

    row
  })

  out <- do.call(rbind, collapsed)
  rownames(out) <- NULL

  out
}

gpbiometrics_scr_event_match_events_to_peaks <- function(events,
                                                         peaks,
                                                         group_cols,
                                                         analysis_window,
                                                         response_window,
                                                         amplitude_col,
                                                         peak_time_col,
                                                         onset_time_col,
                                                         rise_time_col,
                                                         recovery_time_col,
                                                         peak_status_col,
                                                         peak_selection) {
  out <- list()

  for (i in seq_len(nrow(events))) {
    event <- events[i, , drop = FALSE]
    event_time <- event$event_time
    event_group_id <- event$event_group_id

    analysis_start <- event_time + analysis_window[1]
    analysis_end <- event_time + analysis_window[2]
    response_start <- event_time + response_window[1]
    response_end <- event_time + response_window[2]

    group_peaks <- peaks[
      peaks$event_group_id == event_group_id,
      ,
      drop = FALSE
    ]

    peak_time <- suppressWarnings(as.numeric(group_peaks[[peak_time_col]]))

    in_analysis <- is.finite(peak_time) &
      peak_time >= analysis_start &
      peak_time <= analysis_end

    in_response <- is.finite(peak_time) &
      peak_time >= response_start &
      peak_time <= response_end

    analysis_peaks <- group_peaks[in_analysis, , drop = FALSE]
    response_peaks <- group_peaks[in_response, , drop = FALSE]

    selected <- NULL

    if (nrow(response_peaks) > 0) {
      if (identical(peak_selection, "largest_amplitude")) {
        response_amp <- suppressWarnings(as.numeric(response_peaks[[amplitude_col]]))
        selected <- response_peaks[which.max(response_amp), , drop = FALSE]
      } else {
        response_time <- suppressWarnings(as.numeric(response_peaks[[peak_time_col]]))
        selected <- response_peaks[order(response_time)[1], , drop = FALSE]
      }
    }

    selected_peak_time <- NA_real_
    selected_onset_time <- NA_real_
    selected_amplitude <- NA_real_
    selected_rise_time <- NA_real_
    selected_recovery_time <- NA_real_
    selected_peak_status <- NA_character_
    selected_peak_id <- NA_character_

    if (!is.null(selected) && nrow(selected) > 0) {
      selected_peak_time <- suppressWarnings(as.numeric(selected[[peak_time_col]][1]))
      selected_amplitude <- suppressWarnings(as.numeric(selected[[amplitude_col]][1]))

      if (onset_time_col %in% names(selected)) {
        selected_onset_time <- suppressWarnings(as.numeric(selected[[onset_time_col]][1]))
      }

      if (rise_time_col %in% names(selected)) {
        selected_rise_time <- suppressWarnings(as.numeric(selected[[rise_time_col]][1]))
      }

      if (recovery_time_col %in% names(selected)) {
        selected_recovery_time <- suppressWarnings(as.numeric(selected[[recovery_time_col]][1]))
      }

      if (peak_status_col %in% names(selected)) {
        selected_peak_status <- as.character(selected[[peak_status_col]][1])
      }

      if ("peak_id" %in% names(selected)) {
        selected_peak_id <- as.character(selected$peak_id[1])
      }
    }

    response_flag <- as.integer(!is.null(selected) && nrow(selected) > 0)

    event_status <- if (!is.finite(event_time)) {
      "missing_event_time"
    } else if (nrow(analysis_peaks) == 0) {
      "no_peaks_in_analysis_window"
    } else if (nrow(response_peaks) == 0) {
      "no_peaks_in_response_window"
    } else {
      "response_detected"
    }

    row <- data.frame(
      event_id = event$event_id,
      event_label = event$event_label,
      event_time = event_time,
      event_group_id = event_group_id,
      source_row_id = event$source_row_id,
      window_start = analysis_start,
      window_end = analysis_end,
      response_start = response_start,
      response_end = response_end,
      response_flag = response_flag,
      n_candidate_peaks = nrow(analysis_peaks),
      n_response_window_peaks = nrow(response_peaks),
      selected_peak_id = selected_peak_id,
      selected_peak_time = selected_peak_time,
      selected_onset_time = selected_onset_time,
      scr_latency = selected_peak_time - event_time,
      scr_amplitude = selected_amplitude,
      scr_rise_time = selected_rise_time,
      scr_recovery_time = selected_recovery_time,
      selected_peak_status = selected_peak_status,
      event_status = event_status,
      stringsAsFactors = FALSE
    )

    if (length(group_cols) > 0) {
      group_values <- event[1, group_cols, drop = FALSE]
      row <- cbind(group_values, row)
    }

    out[[length(out) + 1L]] <- row
  }

  if (length(out) == 0) {
    return(gpbiometrics_scr_event_empty_event_table(group_cols))
  }

  out <- do.call(rbind, out)
  rownames(out) <- NULL

  out
}

gpbiometrics_scr_event_empty_event_table <- function(group_cols) {
  out <- data.frame(stringsAsFactors = FALSE)

  if (length(group_cols) > 0) {
    for (col in group_cols) {
      out[[col]] <- character()
    }
  }

  out$event_id <- character()
  out$event_label <- character()
  out$event_time <- numeric()
  out$event_group_id <- character()
  out$source_row_id <- integer()
  out$window_start <- numeric()
  out$window_end <- numeric()
  out$response_start <- numeric()
  out$response_end <- numeric()
  out$response_flag <- integer()
  out$n_candidate_peaks <- integer()
  out$n_response_window_peaks <- integer()
  out$selected_peak_id <- character()
  out$selected_peak_time <- numeric()
  out$selected_onset_time <- numeric()
  out$scr_latency <- numeric()
  out$scr_amplitude <- numeric()
  out$scr_rise_time <- numeric()
  out$scr_recovery_time <- numeric()
  out$selected_peak_status <- character()
  out$event_status <- character()

  out
}

gpbiometrics_scr_event_window_qc <- function(event_table,
                                             group_cols) {
  if (nrow(event_table) == 0) {
    out <- data.frame(
      group_id = character(),
      event_count = integer(),
      response_events = integer(),
      response_rate = numeric(),
      no_response_events = integer(),
      events_with_candidate_peaks = integer(),
      stringsAsFactors = FALSE
    )

    return(out)
  }

  group_id <- event_table$event_group_id
  group_ids <- unique(group_id)

  out <- lapply(group_ids, function(group_id_i) {
    d <- event_table[group_id == group_id_i, , drop = FALSE]

    row <- data.frame(
      event_group_id = group_id_i,
      event_count = nrow(d),
      response_events = sum(d$response_flag == 1, na.rm = TRUE),
      response_rate = mean(d$response_flag == 1, na.rm = TRUE),
      no_response_events = sum(d$response_flag == 0, na.rm = TRUE),
      events_with_candidate_peaks = sum(d$n_candidate_peaks > 0, na.rm = TRUE),
      events_with_response_window_peaks = sum(d$n_response_window_peaks > 0, na.rm = TRUE),
      median_scr_amplitude = stats::median(d$scr_amplitude, na.rm = TRUE),
      median_scr_latency = stats::median(d$scr_latency, na.rm = TRUE),
      stringsAsFactors = FALSE
    )

    if (length(group_cols) > 0) {
      group_values <- d[1, group_cols, drop = FALSE]
      row <- cbind(group_values, row)
    }

    row
  })

  out <- do.call(rbind, out)
  rownames(out) <- NULL

  out
}

test_that("screen_gazepoint_eda_nonresponders screens event-window data", {
  event_table <- data.frame(
    participant = c("P1", "P1", "P2", "P2"),
    response_flag = c(1, 0, 0, 0),
    scr_amplitude = c(0.05, NA, NA, NA)
  )

  res <- screen_gazepoint_eda_nonresponders(
    event_table,
    group_cols = "participant",
    min_events = 2,
    min_response_events = 1,
    min_response_rate = 0.1
  )

  expect_s3_class(res, "gazepoint_eda_nonresponder_screen")
  expect_equal(res$overview$group_count, 2)
  expect_equal(res$overview$candidate_nonresponder_count, 1)
  expect_equal(res$candidate_nonresponders$participant, "P2")
})

test_that("screen_gazepoint_eda_nonresponders accepts SCR event-window objects", {
  events <- data.frame(
    participant = c("P1", "P2"),
    event_time = c(10, 10)
  )

  peaks <- data.frame(
    participant = "P1",
    peak_id = 1,
    peak_time = 12,
    onset_time = 11,
    amplitude = 0.05,
    rise_time = 1,
    recovery_time_after_peak = 3,
    status = "detected"
  )

  windows <- summarise_gazepoint_scr_event_windows(
    scr_peaks = peaks,
    events = events,
    event_time_col = "event_time",
    group_cols = "participant",
    analysis_window = c(0, 6),
    response_window = c(1, 4)
  )

  res <- screen_gazepoint_eda_nonresponders(
    windows,
    group_cols = "participant",
    min_events = 1,
    min_response_events = 1
  )

  expect_equal(res$overview$group_count, 2)
  expect_true("P2" %in% res$candidate_nonresponders$participant)
})

test_that("screen_gazepoint_eda_nonresponders can screen peak-only data", {
  peaks <- data.frame(
    participant = c("P1", "P1"),
    peak_time = c(10, 20),
    amplitude = c(0.05, 0.04),
    status = c("detected", "detected")
  )

  res <- screen_gazepoint_eda_nonresponders(
    peaks,
    group_cols = "participant",
    min_detected_peaks = 3
  )

  expect_equal(res$overview$candidate_nonresponder_count, 1)
  expect_equal(res$candidate_nonresponders$participant, "P1")
})

test_that("screen_gazepoint_eda_nonresponders validates thresholds", {
  dat <- data.frame(
    response_flag = 1,
    scr_amplitude = 0.05
  )

  expect_error(
    screen_gazepoint_eda_nonresponders(dat, min_response_rate = 2),
    "`min_response_rate`"
  )

  expect_error(
    screen_gazepoint_eda_nonresponders(dat, min_events = -1),
    "`min_events`"
  )
})

test_that("prepare_gazepoint_scr_hurdle_model_data prepares response and amplitude datasets", {
  event_table <- data.frame(
    participant = c("P1", "P1", "P2", "P2"),
    condition = c("A", "B", "A", "B"),
    event_id = paste0("e", 1:4),
    response_flag = c(1, 0, 1, 0),
    scr_amplitude = c(0.05, NA, 0.08, NA),
    scr_latency = c(2, NA, 3, NA),
    scr_rise_time = c(1, NA, 1, NA),
    scr_recovery_time = c(4, NA, 5, NA)
  )

  res <- prepare_gazepoint_scr_hurdle_model_data(
    event_table,
    predictor_cols = "condition",
    factor_cols = c("participant", "condition"),
    group_cols = "participant"
  )

  expect_s3_class(res, "gazepoint_scr_hurdle_model_data")
  expect_equal(res$overview$status, "scr_hurdle_model_data_prepared")
  expect_equal(nrow(res$response_model_data), 4)
  expect_equal(nrow(res$amplitude_model_data), 2)
  expect_true(all(res$amplitude_model_data$scr_response_binary == 1))
  expect_true(is.factor(res$response_model_data$condition))
})

test_that("prepare_gazepoint_scr_hurdle_model_data accepts event-window summary objects", {
  events <- data.frame(
    participant = "P1",
    event_time = 10
  )

  peaks <- data.frame(
    participant = "P1",
    peak_id = 1,
    peak_time = 12,
    onset_time = 11,
    amplitude = 0.05,
    rise_time = 1,
    recovery_time_after_peak = 3,
    status = "detected"
  )

  windows <- summarise_gazepoint_scr_event_windows(
    scr_peaks = peaks,
    events = events,
    event_time_col = "event_time",
    group_cols = "participant",
    analysis_window = c(0, 6),
    response_window = c(1, 4)
  )

  res <- prepare_gazepoint_scr_hurdle_model_data(
    windows,
    group_cols = "participant"
  )

  expect_equal(res$overview$input_events, 1)
  expect_equal(res$overview$response_events, 1)
  expect_equal(nrow(res$amplitude_model_data), 1)
})

test_that("prepare_gazepoint_scr_hurdle_model_data applies log transforms", {
  event_table <- data.frame(
    event_id = paste0("e", 1:3),
    response_flag = c(1, 1, 0),
    scr_amplitude = c(0.05, 0.10, NA)
  )

  res <- prepare_gazepoint_scr_hurdle_model_data(
    event_table,
    amplitude_transform = "log",
    amplitude_offset = 1e-6
  )

  expect_equal(nrow(res$amplitude_model_data), 2)
  expect_equal(
    res$amplitude_model_data$scr_amplitude_model,
    log(res$amplitude_model_data$scr_amplitude_raw + 1e-6)
  )
})

test_that("prepare_gazepoint_scr_hurdle_model_data can keep rows with missing predictors when requested", {
  event_table <- data.frame(
    event_id = paste0("e", 1:3),
    condition = c("A", NA, "B"),
    response_flag = c(1, 0, 1),
    scr_amplitude = c(0.05, NA, 0.08)
  )

  res_drop <- prepare_gazepoint_scr_hurdle_model_data(
    event_table,
    predictor_cols = "condition",
    drop_missing_predictors = TRUE
  )

  res_keep <- prepare_gazepoint_scr_hurdle_model_data(
    event_table,
    predictor_cols = "condition",
    drop_missing_predictors = FALSE
  )

  expect_equal(nrow(res_drop$response_model_data), 2)
  expect_equal(nrow(res_keep$response_model_data), 3)
})

test_that("prepare_gazepoint_scr_hurdle_model_data creates model formula text", {
  event_table <- data.frame(
    participant = c("P1", "P2"),
    condition = c("A", "B"),
    event_id = c("e1", "e2"),
    response_flag = c(1, 0),
    scr_amplitude = c(0.05, NA)
  )

  res <- prepare_gazepoint_scr_hurdle_model_data(
    event_table,
    predictor_cols = "condition",
    group_cols = "participant"
  )

  expect_true(any(grepl("scr_response_binary ~ condition", res$model_formulas$formula)))
  expect_true(any(grepl("\\(1 \\| participant\\)", res$model_formulas$formula)))
})

test_that("prepare_gazepoint_scr_hurdle_model_data warns through status when no positive amplitudes exist", {
  event_table <- data.frame(
    event_id = paste0("e", 1:3),
    response_flag = c(0, 0, 0),
    scr_amplitude = c(NA, NA, NA)
  )

  res <- prepare_gazepoint_scr_hurdle_model_data(event_table)

  expect_equal(res$overview$status, "warn_no_positive_amplitude_rows")
  expect_equal(nrow(res$response_model_data), 3)
  expect_equal(nrow(res$amplitude_model_data), 0)
})

test_that("prepare_gazepoint_scr_hurdle_model_data validates required columns", {
  event_table <- data.frame(
    event_id = "e1",
    response_flag = 1
  )

  expect_error(
    prepare_gazepoint_scr_hurdle_model_data(event_table),
    "`amplitude_col` was not found"
  )
})

test_that("prepare_gazepoint_scr_hurdle_model_data validates requested predictors", {
  event_table <- data.frame(
    event_id = "e1",
    response_flag = 1,
    scr_amplitude = 0.05
  )

  expect_error(
    prepare_gazepoint_scr_hurdle_model_data(
      event_table,
      predictor_cols = "condition"
    ),
    "Requested columns were not found"
  )
})

#' Import a Gazepoint Biometrics export
#'
#' Reads a Gazepoint CSV export containing biometric columns such as GSR,
#' heart rate, interbeat interval, pulse signal, engagement dial, and TTL
#' synchronization fields. The function is conservative: it preserves original
#' column names, removes only empty trailing columns, and attaches a basic
#' biometric-column summary as an attribute.
#'
#' @param file Path to a Gazepoint CSV export.
#' @param na Values that should be treated as missing.
#'
#' @return A data frame with Gazepoint export columns preserved. The returned
#'   object has class `"gazepoint_biometrics"` and an attribute named
#'   `"biometric_columns"`.
#'
#' @export
import_gazepoint_biometrics <- function(file, na = c("", "NA", "NaN")) {
  if (missing(file) || length(file) != 1L || !nzchar(file)) {
    stop("`file` must be a single non-empty file path.", call. = FALSE)
  }

  if (!file.exists(file)) {
    stop("File does not exist: ", file, call. = FALSE)
  }

  dat <- utils::read.csv(
    file,
    stringsAsFactors = FALSE,
    check.names = FALSE,
    na.strings = na
  )

  dat <- drop_empty_trailing_columns(dat)

  attr(dat, "biometric_columns") <- check_gazepoint_biometric_columns(dat)
  class(dat) <- c("gazepoint_biometrics", class(dat))

  dat
}


#' Check Gazepoint biometric columns
#'
#' Checks whether a data frame contains known Gazepoint Biometrics columns.
#' This function does not assume that the channels are active. It only checks
#' whether the expected columns are present.
#'
#' @param data A data frame imported from a Gazepoint export.
#'
#' @return A data frame describing expected columns, their signal family,
#'   interpretation, and whether they are present.
#'
#' @export
check_gazepoint_biometric_columns <- function(data) {
  if (!is.data.frame(data)) {
    stop("`data` must be a data frame.", call. = FALSE)
  }

  expected <- data.frame(
    column = c(
      "DIAL", "DIALV",
      "GSR", "GSR_US", "GSR_US_TONIC", "GSR_US_PHASIC", "GSRV",
      "HR", "HRV", "HRP", "IBI",
      paste0("TTL", 0:6), "TTLV",
      "CNT", "TIME", "TIME_TICK", "USER", "USERID", "MEDIA_ID",
      "MEDIA_NAME", "FPOGX", "FPOGY", "FPOGS", "FPOGD", "FPOGID"
    ),
    signal = c(
      "engagement_dial", "engagement_dial",
      rep("gsr_eda", 5),
      rep("heart_rate", 4),
      rep("ttl_marker", 8),
      rep("time_identity_sync", 7),
      rep("fixation_gaze", 5)
    ),
    role = c(
      "dial_value", "dial_validity",
      "gsr_raw_or_resistance", "gsr_conductance_microsiemens",
      "gsr_tonic_component", "gsr_phasic_component", "gsr_validity",
      "heart_rate_bpm", "heart_rate_validity_not_hrv_metric",
      "pulse_signal", "interbeat_interval_seconds",
      rep("ttl_channel", 7), "ttl_validity",
      "sample_counter", "recording_time", "recording_tick",
      "user_label", "user_identifier", "media_identifier",
      "media_name", "fixation_x", "fixation_y",
      "fixation_start_time", "fixation_duration", "fixation_identifier"
    ),
    stringsAsFactors = FALSE
  )

  expected$present <- expected$column %in% names(data)
  expected
}


#' Detect active Gazepoint biometric channels
#'
#' Detects whether GSR/EDA, heart-rate, engagement-dial, and TTL channels are
#' present and whether they appear active. A channel can be present but inactive
#' when validity flags are zero or the signal contains only zeros or missing
#' values. For each signal family, `summary_column` identifies the primary
#' column used for the reported minimum and maximum values.
#'
#' @param data A data frame imported from a Gazepoint export.
#'
#' @return A data frame with one row per signal family.
#'
#' @export
detect_active_biometric_channels <- function(data) {
  if (!is.data.frame(data)) {
    stop("`data` must be a data frame.", call. = FALSE)
  }

  out <- rbind(
    detect_one_channel(
      data = data,
      signal = "gsr_eda",
      value_columns = c("GSR_US", "GSR", "GSR_US_TONIC", "GSR_US_PHASIC"),
      validity_columns = "GSRV"
    ),
    detect_one_channel(
      data = data,
      signal = "heart_rate",
      value_columns = c("HR", "IBI", "HRP"),
      validity_columns = "HRV"
    ),
    detect_one_channel(
      data = data,
      signal = "engagement_dial",
      value_columns = "DIAL",
      validity_columns = "DIALV"
    ),
    detect_one_channel(
      data = data,
      signal = "ttl_marker",
      value_columns = paste0("TTL", 0:6),
      validity_columns = "TTLV"
    )
  )

  rownames(out) <- NULL
  out
}


drop_empty_trailing_columns <- function(data) {
  if (!is.data.frame(data) || ncol(data) == 0L) {
    return(data)
  }

  empty_name <- names(data) == "" | is.na(names(data))

  empty_values <- vapply(
    data,
    function(x) all(is.na(x) | trimws(as.character(x)) == ""),
    logical(1)
  )

  drop <- empty_name & empty_values

  if (any(drop)) {
    data <- data[, !drop, drop = FALSE]
  }

  data
}


detect_one_channel <- function(data, signal, value_columns, validity_columns) {
  present_value_columns <- intersect(value_columns, names(data))
  present_validity_columns <- intersect(validity_columns, names(data))

  present <- length(present_value_columns) > 0L ||
    length(present_validity_columns) > 0L

  if (!present) {
    return(data.frame(
      signal = signal,
      present = FALSE,
      active = FALSE,
      value_columns = NA_character_,
      summary_column = NA_character_,
      validity_columns = NA_character_,
      valid_rows = 0L,
      nonzero_rows = 0L,
      min_value = NA_real_,
      max_value = NA_real_,
      stringsAsFactors = FALSE
    ))
  }

  summary_column <- choose_signal_summary_column(
    signal = signal,
    present_value_columns = present_value_columns
  )

  if (is.na(summary_column)) {
    x <- rep(NA_real_, nrow(data))
  } else {
    x <- as_numeric_safe(data[[summary_column]])
  }

  valid <- !is.na(x)

  if (length(present_validity_columns) > 0L) {
    validity_values <- lapply(present_validity_columns, function(column) {
      as_numeric_safe(data[[column]])
    })

    validity_matrix <- do.call(cbind, validity_values)

    valid <- valid &
      rowSums(!is.na(validity_matrix) & validity_matrix > 0) > 0
  }

  nonzero <- !is.na(x) & x != 0

  active <- any(nonzero & valid)

  usable_values <- x[valid & !is.na(x)]

  data.frame(
    signal = signal,
    present = TRUE,
    active = active,
    value_columns = paste(present_value_columns, collapse = ","),
    summary_column = summary_column,
    validity_columns = ifelse(
      length(present_validity_columns) > 0L,
      paste(present_validity_columns, collapse = ","),
      NA_character_
    ),
    valid_rows = sum(valid),
    nonzero_rows = sum(nonzero),
    min_value = ifelse(length(usable_values) > 0L, min(usable_values), NA_real_),
    max_value = ifelse(length(usable_values) > 0L, max(usable_values), NA_real_),
    stringsAsFactors = FALSE
  )
}


choose_signal_summary_column <- function(signal, present_value_columns) {
  if (length(present_value_columns) == 0L) {
    return(NA_character_)
  }

  priority <- switch(
    signal,
    gsr_eda = c("GSR_US", "GSR_US_TONIC", "GSR_US_PHASIC", "GSR"),
    heart_rate = c("HR", "IBI", "HRP"),
    engagement_dial = c("DIAL"),
    ttl_marker = c("TTL0", "TTL1", "TTL2", "TTL3", "TTL4", "TTL5", "TTL6"),
    present_value_columns
  )

  candidates <- intersect(priority, present_value_columns)

  if (length(candidates) > 0L) {
    return(candidates[1])
  }

  present_value_columns[1]
}


as_numeric_safe <- function(x) {
  suppressWarnings(as.numeric(x))
}

test_that("synthetic kiosk demo exports are available and importable", {
  demo_dir <- system.file(
    "extdata",
    "gazepoint_biometrics_kiosk_demo_exports",
    package = "gpbiometrics"
  )

  expect_true(dir.exists(demo_dir))

  all_gaze_files <- list.files(demo_dir, pattern = "_all_gaze\\.csv$", full.names = TRUE)

  expect_equal(length(all_gaze_files), 36)

  overview_file <- file.path(demo_dir, "synthetic_kiosk_overview.csv")
  trial_design_file <- file.path(demo_dir, "synthetic_kiosk_trial_design.csv")
  readme_file <- file.path(demo_dir, "README.txt")

  expect_true(file.exists(overview_file))
  expect_true(file.exists(trial_design_file))
  expect_true(file.exists(readme_file))

  first_file <- all_gaze_files[1]

  dat <- import_gazepoint_biometrics(first_file)

  expect_s3_class(dat, "data.frame")
  expect_equal(ncol(dat), 67)
  expect_true(nrow(dat) > 0)

  expected_columns <- c(
    "MEDIA_ID", "MEDIA_NAME", "CNT", "TIME", "FPOGX", "FPOGY",
    "AOI", "LPMM", "RPMM", "GSR_US", "HR", "IBI",
    "DIAL", "TTL0", "TTL1", "TTL2", "TTL3", "TTL4", "TTL5", "TTL6",
    "interface_complexity", "feedback_clarity", "synthetic_scenario"
  )

  expect_true(all(expected_columns %in% names(dat)))

  validation <- validate_gazepoint_biometrics(dat)

  expect_equal(validation$overview$issue_count, 0)
  expect_true(validation$overview$active_signal_count >= 3)
})

test_that("check_gazepoint_biometric_columns detects known columns", {
  dat <- data.frame(
    TIME = c(0.01, 0.02),
    GSR = c(500000, 510000),
    GSR_US = c(2.0, 1.96),
    GSRV = c(1, 1),
    HR = c(75, 76),
    HRV = c(1, 1),
    HRP = c(0.1, 0.2),
    IBI = c(0.80, 0.79),
    DIAL = c(0.2, 0.4),
    DIALV = c(1, 1)
  )

  cols <- check_gazepoint_biometric_columns(dat)

  expect_true(cols$present[cols$column == "GSR_US"])
  expect_true(cols$present[cols$column == "HR"])
  expect_true(cols$present[cols$column == "IBI"])
  expect_true(cols$present[cols$column == "DIAL"])
})


test_that("detect_active_biometric_channels identifies active signals", {
  dat <- data.frame(
    GSR_US = c(2.0, 2.1, 2.2),
    GSRV = c(1, 1, 1),
    HR = c(75, 76, 77),
    HRV = c(1, 1, 1),
    DIAL = c(0.1, 0.2, 0.3),
    DIALV = c(1, 1, 1)
  )

  active <- detect_active_biometric_channels(dat)

  expect_true(active$active[active$signal == "gsr_eda"])
  expect_true(active$active[active$signal == "heart_rate"])
  expect_true(active$active[active$signal == "engagement_dial"])
})


test_that("detect_active_biometric_channels identifies inactive signals", {
  dat <- data.frame(
    GSR_US = c(0, 0, 0),
    GSRV = c(0, 0, 0),
    HR = c(0, 0, 0),
    HRV = c(0, 0, 0),
    DIAL = c(0, 0, 0),
    DIALV = c(0, 0, 0)
  )

  active <- detect_active_biometric_channels(dat)

  expect_false(active$active[active$signal == "gsr_eda"])
  expect_false(active$active[active$signal == "heart_rate"])
  expect_false(active$active[active$signal == "engagement_dial"])
})


test_that("import_gazepoint_biometrics imports CSV and drops empty trailing column", {
  tmp <- tempfile(fileext = ".csv")

  writeLines(
    c(
      "TIME,GSR_US,GSRV,HR,HRV,DIAL,DIALV,",
      "0.01,2.0,1,75,1,0.1,1,",
      "0.02,2.1,1,76,1,0.2,1,"
    ),
    tmp,
    useBytes = TRUE
  )

  dat <- import_gazepoint_biometrics(tmp)

  expect_s3_class(dat, "gazepoint_biometrics")
  expect_true("GSR_US" %in% names(dat))
  expect_true("HR" %in% names(dat))
  expect_true("DIAL" %in% names(dat))
  expect_false(any(names(dat) == ""))

  cols <- attr(dat, "biometric_columns")
  expect_true(is.data.frame(cols))
})
test_that("detect_active_biometric_channels reports a clear summary column", {
  dat <- data.frame(
    GSR = c(500000, 510000, 520000),
    GSR_US = c(1.1, 1.2, 1.3),
    GSRV = c(1, 1, 1),
    HR = c(70, 72, 74),
    HRV = c(1, 1, 1),
    DIAL = c(0.1, 0.2, 0.3),
    DIALV = c(1, 1, 1)
  )

  out <- detect_active_biometric_channels(dat)

  gsr <- out[out$signal == "gsr_eda", ]

  expect_equal(gsr$summary_column, "GSR_US")
  expect_equal(gsr$min_value, 1.1)
  expect_equal(gsr$max_value, 1.3)
})

#' Extract respiration proxy using a CEEMDAN-style bridge
#'
#' Extracts respiration-like components from PPG or ECG-derived respiratory
#' proxy signals. If `external_fun` is supplied, it is used as the CEEMDAN
#' backend. Otherwise, the function uses a dependency-light multiscale
#' decomposition fallback and labels the result accordingly.
#'
#' This function does not claim to reproduce full CEEMDAN unless a validated
#' external CEEMDAN function is supplied.
#'
#' @param dat A data frame.
#' @param signal_col Numeric PPG/ECG-derived signal column.
#' @param time_col Numeric time column.
#' @param group_cols Optional grouping columns.
#' @param sampling_rate Optional sampling rate in Hz.
#' @param respiration_band Frequency band in Hz used to select respiration-like
#'   components.
#' @param scales Moving-average scales used by the fallback decomposition.
#' @param external_fun Optional function with arguments `x`, `time`, and
#'   `sampling_rate`, returning either a numeric vector or a list/data frame of
#'   components.
#'
#' @return A list with `overview`, `component_table`, `respiration_timeseries`,
#'   `summary`, and `settings`.
#' @export
extract_gazepoint_respiration_ceemdan <- function(dat,
                                                  signal_col,
                                                  time_col = "CNT",
                                                  group_cols = NULL,
                                                  sampling_rate = NULL,
                                                  respiration_band = c(0.10, 0.60),
                                                  scales = c(5, 15, 30, 60, 120),
                                                  external_fun = NULL) {
  if (!is.data.frame(dat)) {
    stop("`dat` must be a data frame.", call. = FALSE)
  }

  if (!signal_col %in% names(dat)) {
    stop("Column `", signal_col, "` was not found in `dat`.", call. = FALSE)
  }

  if (!time_col %in% names(dat)) {
    stop("Column `", time_col, "` was not found in `dat`.", call. = FALSE)
  }

  if (!is.numeric(dat[[signal_col]])) {
    stop("`signal_col` must identify a numeric column.", call. = FALSE)
  }

  if (!is.numeric(dat[[time_col]])) {
    stop("`time_col` must identify a numeric column.", call. = FALSE)
  }

  if (is.null(group_cols)) {
    group_cols <- character()
  }

  missing_groups <- setdiff(group_cols, names(dat))
  if (length(missing_groups) > 0) {
    stop("Missing `group_cols`: ", paste(missing_groups, collapse = ", "), call. = FALSE)
  }

  groups <- gpbiometrics_ceemdan_split(dat, group_cols)

  component_rows <- list()
  timeseries_rows <- list()
  summary_rows <- list()
  component_id <- 1L

  for (group_id in names(groups)) {
    idx <- groups[[group_id]]
    idx <- idx[order(dat[[time_col]][idx])]

    time <- dat[[time_col]][idx]
    x <- dat[[signal_col]][idx]
    fs <- gpbiometrics_ceemdan_sampling_rate(time, sampling_rate)

    if (!is.finite(fs) || fs <= 0 || sum(is.finite(x)) < 10) {
      summary_rows[[group_id]] <- data.frame(
        group_id = group_id,
        n_rows = length(idx),
        sampling_rate_hz = fs,
        selected_component_count = 0L,
        proxy_respiration_rate_hz = NA_real_,
        proxy_respiration_rate_bpm = NA_real_,
        status = "insufficient_signal_or_sampling_rate",
        stringsAsFactors = FALSE
      )
      next
    }

    x_filled <- gpbiometrics_ceemdan_fill(x)

    decomposition <- gpbiometrics_ceemdan_decompose(
      x = x_filled,
      time = time,
      sampling_rate = fs,
      scales = scales,
      external_fun = external_fun
    )

    components <- decomposition$components
    method <- decomposition$method

    if (ncol(components) == 0) {
      summary_rows[[group_id]] <- data.frame(
        group_id = group_id,
        n_rows = length(idx),
        sampling_rate_hz = fs,
        selected_component_count = 0L,
        proxy_respiration_rate_hz = NA_real_,
        proxy_respiration_rate_bpm = NA_real_,
        status = "no_components_extracted",
        stringsAsFactors = FALSE
      )
      next
    }

    component_rates <- vapply(seq_len(ncol(components)), function(j) {
      gpbiometrics_ceemdan_dominant_frequency(
        signal = components[, j],
        sampling_rate = fs,
        frequency_band = respiration_band
      )
    }, numeric(1))

    selected <- is.finite(component_rates) &
      component_rates >= respiration_band[1] &
      component_rates <= respiration_band[2]

    if (!any(selected)) {
      selected[which.max(stats::var(components, na.rm = TRUE))] <- TRUE
    }

    proxy <- rowMeans(components[, selected, drop = FALSE], na.rm = TRUE)

    proxy_rate <- gpbiometrics_ceemdan_dominant_frequency(
      signal = proxy,
      sampling_rate = fs,
      frequency_band = respiration_band
    )

    for (j in seq_len(ncol(components))) {
      component_rows[[component_id]] <- data.frame(
        group_id = group_id,
        component = colnames(components)[j],
        dominant_frequency_hz = component_rates[j],
        selected_for_respiration = selected[j],
        variance = stats::var(components[, j], na.rm = TRUE),
        method = method,
        stringsAsFactors = FALSE
      )
      component_id <- component_id + 1L
    }

    timeseries_rows[[group_id]] <- data.frame(
      row_index = idx,
      group_id = group_id,
      time = time,
      respiration_proxy = proxy,
      status = "respiration_proxy_extracted",
      stringsAsFactors = FALSE
    )

    summary_rows[[group_id]] <- data.frame(
      group_id = group_id,
      n_rows = length(idx),
      sampling_rate_hz = fs,
      selected_component_count = sum(selected),
      proxy_respiration_rate_hz = proxy_rate,
      proxy_respiration_rate_bpm = proxy_rate * 60,
      status = "respiration_proxy_extracted",
      stringsAsFactors = FALSE
    )
  }

  component_table <- if (length(component_rows) > 0) {
    do.call(rbind, component_rows)
  } else {
    data.frame()
  }

  respiration_timeseries <- if (length(timeseries_rows) > 0) {
    do.call(rbind, timeseries_rows)
  } else {
    data.frame()
  }

  summary <- do.call(rbind, summary_rows)
  rownames(summary) <- NULL

  overview <- data.frame(
    group_count = length(groups),
    component_rows = nrow(component_table),
    timeseries_rows = nrow(respiration_timeseries),
    successful_groups = sum(summary$status == "respiration_proxy_extracted"),
    problem_groups = sum(summary$status != "respiration_proxy_extracted"),
    status = if (all(summary$status == "respiration_proxy_extracted")) {
      "ceemdan_respiration_proxy_complete"
    } else if (any(summary$status == "respiration_proxy_extracted")) {
      "ceemdan_respiration_proxy_partial"
    } else {
      "ceemdan_respiration_proxy_failed"
    },
    interpretation = paste(
      "Respiration output is a signal-derived proxy.",
      "Full CEEMDAN is used only when a validated external function is supplied; otherwise a dependency-light multiscale fallback is used."
    ),
    stringsAsFactors = FALSE
  )

  structure(
    list(
      overview = overview,
      component_table = component_table,
      respiration_timeseries = respiration_timeseries,
      summary = summary,
      settings = list(
        signal_col = signal_col,
        time_col = time_col,
        group_cols = group_cols,
        sampling_rate = sampling_rate,
        respiration_band = respiration_band,
        scales = scales,
        external_fun_supplied = !is.null(external_fun)
      )
    ),
    class = c("gazepoint_respiration_ceemdan", "list")
  )
}

#' Fuse respiration proxies using a Kalman filter
#'
#' Fuses two respiration proxy streams, such as PPG-derived respiration and
#' ECG-derived respiration, using a transparent one-dimensional Kalman filter.
#' This is a linear Kalman fusion helper. It is not an extended Kalman filter
#' unless the user supplies nonlinear state/measurement logic externally.
#'
#' @param dat A data frame.
#' @param primary_col First respiration proxy column.
#' @param secondary_col Second respiration proxy column.
#' @param time_col Optional time column.
#' @param group_cols Optional grouping columns.
#' @param process_var Process variance.
#' @param primary_var Measurement variance for `primary_col`.
#' @param secondary_var Measurement variance for `secondary_col`.
#' @param output_col Output fused respiration column.
#'
#' @return A data frame with fused respiration output and Kalman attributes.
#' @export
fuse_gazepoint_respiration_kalman <- function(dat,
                                              primary_col,
                                              secondary_col,
                                              time_col = NULL,
                                              group_cols = NULL,
                                              process_var = 0.01,
                                              primary_var = 0.05,
                                              secondary_var = 0.05,
                                              output_col = "respiration_kalman_fused") {
  if (!is.data.frame(dat)) {
    stop("`dat` must be a data frame.", call. = FALSE)
  }

  required <- c(primary_col, secondary_col, time_col)
  required <- required[!is.null(required)]

  missing_required <- setdiff(required, names(dat))
  if (length(missing_required) > 0) {
    stop("Missing required columns: ", paste(missing_required, collapse = ", "), call. = FALSE)
  }

  if (!is.numeric(dat[[primary_col]])) {
    stop("`primary_col` must identify a numeric column.", call. = FALSE)
  }

  if (!is.numeric(dat[[secondary_col]])) {
    stop("`secondary_col` must identify a numeric column.", call. = FALSE)
  }

  if (!is.null(time_col) && !is.numeric(dat[[time_col]])) {
    stop("`time_col` must identify a numeric column.", call. = FALSE)
  }

  if (is.null(group_cols)) {
    group_cols <- character()
  }

  missing_groups <- setdiff(group_cols, names(dat))
  if (length(missing_groups) > 0) {
    stop("Missing `group_cols`: ", paste(missing_groups, collapse = ", "), call. = FALSE)
  }

  out <- dat
  out[[output_col]] <- NA_real_
  out[[paste0(output_col, "_variance")]] <- NA_real_
  out[[paste0(output_col, "_status")]] <- "not_processed"

  groups <- gpbiometrics_ceemdan_split(out, group_cols)
  summary_rows <- list()

  for (group_id in names(groups)) {
    idx <- groups[[group_id]]

    if (!is.null(time_col)) {
      idx <- idx[order(out[[time_col]][idx])]
    }

    primary <- out[[primary_col]][idx]
    secondary <- out[[secondary_col]][idx]

    filtered <- gpbiometrics_kalman_fuse_two_streams(
      primary = primary,
      secondary = secondary,
      process_var = process_var,
      primary_var = primary_var,
      secondary_var = secondary_var
    )

    out[[output_col]][idx] <- filtered$state
    out[[paste0(output_col, "_variance")]][idx] <- filtered$variance
    out[[paste0(output_col, "_status")]][idx] <- filtered$status

    summary_rows[[group_id]] <- data.frame(
      group_id = group_id,
      n_rows = length(idx),
      finite_primary = sum(is.finite(primary)),
      finite_secondary = sum(is.finite(secondary)),
      fused_rows = sum(filtered$status == "fused"),
      primary_only_rows = sum(filtered$status == "primary_only"),
      secondary_only_rows = sum(filtered$status == "secondary_only"),
      missing_rows = sum(filtered$status == "missing"),
      status = if (any(filtered$status != "missing")) {
        "kalman_respiration_fusion_complete"
      } else {
        "kalman_respiration_fusion_failed"
      },
      stringsAsFactors = FALSE
    )
  }

  summary <- do.call(rbind, summary_rows)
  rownames(summary) <- NULL

  overview <- data.frame(
    input_rows = nrow(dat),
    output_rows = nrow(out),
    group_count = length(groups),
    successful_groups = sum(summary$status == "kalman_respiration_fusion_complete"),
    problem_groups = sum(summary$status != "kalman_respiration_fusion_complete"),
    status = if (all(summary$status == "kalman_respiration_fusion_complete")) {
      "kalman_respiration_fusion_complete"
    } else if (any(summary$status == "kalman_respiration_fusion_complete")) {
      "kalman_respiration_fusion_partial"
    } else {
      "kalman_respiration_fusion_failed"
    },
    interpretation = paste(
      "The fused series is a Kalman-filtered respiration proxy from two measurement streams.",
      "It is not direct respiratory-belt measurement and does not infer psychological or clinical state."
    ),
    stringsAsFactors = FALSE
  )

  attr(out, "kalman_respiration_overview") <- overview
  attr(out, "kalman_respiration_summary") <- summary
  attr(out, "kalman_respiration_settings") <- list(
    primary_col = primary_col,
    secondary_col = secondary_col,
    time_col = time_col,
    group_cols = group_cols,
    process_var = process_var,
    primary_var = primary_var,
    secondary_var = secondary_var,
    output_col = output_col
  )

  class(out) <- unique(c("gazepoint_respiration_kalman_fused", class(out)))
  out
}

gpbiometrics_ceemdan_split <- function(dat, group_cols) {
  if (length(group_cols) == 0) {
    return(list(all_rows = seq_len(nrow(dat))))
  }

  gf <- dat[group_cols]
  gf[] <- lapply(gf, function(x) {
    x <- as.character(x)
    x[is.na(x)] <- "<NA>"
    x
  })

  split(seq_len(nrow(dat)), do.call(paste, c(gf, sep = " | ")))
}

gpbiometrics_ceemdan_sampling_rate <- function(time, sampling_rate = NULL) {
  if (!is.null(sampling_rate)) {
    return(sampling_rate)
  }

  time <- time[is.finite(time)]

  if (length(time) < 3) {
    return(NA_real_)
  }

  dt <- diff(time)
  dt <- dt[is.finite(dt) & dt > 0]

  if (length(dt) == 0) {
    return(NA_real_)
  }

  median_dt <- stats::median(dt)

  if (median_dt > 10) {
    1000 / median_dt
  } else {
    1 / median_dt
  }
}

gpbiometrics_ceemdan_fill <- function(x) {
  idx <- seq_along(x)
  finite <- is.finite(x)

  if (all(finite)) {
    return(x)
  }

  if (sum(finite) == 0) {
    return(rep(0, length(x)))
  }

  if (sum(finite) == 1) {
    return(rep(x[finite][1], length(x)))
  }

  stats::approx(idx[finite], x[finite], xout = idx, rule = 2)$y
}

gpbiometrics_ceemdan_decompose <- function(x,
                                           time,
                                           sampling_rate,
                                           scales,
                                           external_fun = NULL) {
  if (!is.null(external_fun)) {
    raw <- external_fun(x = x, time = time, sampling_rate = sampling_rate)

    if (is.numeric(raw)) {
      components <- matrix(raw, ncol = 1)
    } else if (is.data.frame(raw)) {
      components <- as.matrix(raw[vapply(raw, is.numeric, logical(1))])
    } else if (is.list(raw)) {
      numeric_items <- raw[vapply(raw, is.numeric, logical(1))]
      components <- do.call(cbind, numeric_items)
    } else {
      stop("`external_fun` must return numeric components.", call. = FALSE)
    }

    if (is.null(colnames(components))) {
      colnames(components) <- paste0("external_component_", seq_len(ncol(components)))
    }

    return(list(components = components, method = "external_ceemdan"))
  }

  components <- list()
  previous_smooth <- x

  for (scale in scales) {
    smooth <- gpbiometrics_ceemdan_moving_average(x, scale)
    component <- previous_smooth - smooth
    components[[paste0("fallback_scale_", scale)]] <- component
    previous_smooth <- smooth
  }

  components[["fallback_residual"]] <- previous_smooth

  list(
    components = as.matrix(as.data.frame(components)),
    method = "multiscale_fallback_not_full_ceemdan"
  )
}

gpbiometrics_ceemdan_moving_average <- function(x, window) {
  window <- max(3L, as.integer(window))

  if (length(x) < window) {
    return(rep(mean(x, na.rm = TRUE), length(x)))
  }

  y <- as.numeric(stats::filter(x, rep(1 / window, window), sides = 2))
  y[!is.finite(y)] <- x[!is.finite(y)]
  y[is.na(y)] <- x[is.na(y)]
  y
}

gpbiometrics_ceemdan_dominant_frequency <- function(signal,
                                                    sampling_rate,
                                                    frequency_band) {
  signal <- signal[is.finite(signal)]

  if (length(signal) < 8 || stats::sd(signal) == 0) {
    return(NA_real_)
  }

  spec <- stats::spec.pgram(
    signal,
    taper = 0.1,
    plot = FALSE,
    demean = TRUE,
    detrend = TRUE,
    fast = TRUE
  )

  freq <- spec$freq * sampling_rate
  power <- spec$spec

  keep <- is.finite(freq) &
    is.finite(power) &
    freq >= frequency_band[1] &
    freq <= frequency_band[2]

  if (!any(keep)) {
    return(NA_real_)
  }

  freq[keep][which.max(power[keep])]
}

gpbiometrics_kalman_fuse_two_streams <- function(primary,
                                                 secondary,
                                                 process_var,
                                                 primary_var,
                                                 secondary_var) {
  n <- length(primary)
  state <- rep(NA_real_, n)
  variance <- rep(NA_real_, n)
  status <- rep("missing", n)

  initial_values <- c(primary, secondary)
  initial_values <- initial_values[is.finite(initial_values)]

  if (length(initial_values) == 0) {
    return(list(state = state, variance = variance, status = status))
  }

  current_state <- mean(initial_values)
  current_var <- stats::var(initial_values)

  if (!is.finite(current_var) || current_var <= 0) {
    current_var <- 1
  }

  for (i in seq_len(n)) {
    current_var <- current_var + process_var

    measurements <- c(primary[i], secondary[i])
    measurement_vars <- c(primary_var, secondary_var)
    valid <- is.finite(measurements)

    if (any(valid)) {
      for (j in which(valid)) {
        kalman_gain <- current_var / (current_var + measurement_vars[j])
        current_state <- current_state + kalman_gain * (measurements[j] - current_state)
        current_var <- (1 - kalman_gain) * current_var
      }

      if (all(valid)) {
        status[i] <- "fused"
      } else if (valid[1]) {
        status[i] <- "primary_only"
      } else {
        status[i] <- "secondary_only"
      }
    }

    state[i] <- current_state
    variance[i] <- current_var
  }

  list(state = state, variance = variance, status = status)
}

test_that("extract_gazepoint_respiration_ceemdan returns respiration proxy", {
  time <- seq(0, 120, by = 0.1)
  signal <- sin(2 * pi * 0.25 * time) + 0.2 * sin(2 * pi * 1.2 * time)

  dat <- data.frame(
    participant = "p1",
    time = time,
    signal = signal
  )

  out <- extract_gazepoint_respiration_ceemdan(
    dat,
    signal_col = "signal",
    time_col = "time",
    group_cols = "participant",
    sampling_rate = 10,
    respiration_band = c(0.1, 0.5)
  )

  expect_s3_class(out, "gazepoint_respiration_ceemdan")
  expect_true(nrow(out$respiration_timeseries) > 0)
  expect_true("proxy_respiration_rate_hz" %in% names(out$summary))
})

test_that("fuse_gazepoint_respiration_kalman fuses two respiration proxies", {
  time <- seq(0, 60, by = 0.5)
  true_resp <- sin(2 * pi * 0.25 * time)

  dat <- data.frame(
    participant = "p1",
    time = time,
    pdr = true_resp + rnorm(length(time), sd = 0.1),
    edr = true_resp + rnorm(length(time), sd = 0.1)
  )

  out <- fuse_gazepoint_respiration_kalman(
    dat,
    primary_col = "pdr",
    secondary_col = "edr",
    time_col = "time",
    group_cols = "participant"
  )

  expect_s3_class(out, "gazepoint_respiration_kalman_fused")
  expect_true("respiration_kalman_fused" %in% names(out))
  expect_equal(
    attr(out, "kalman_respiration_overview")$status,
    "kalman_respiration_fusion_complete"
  )
})

test_that("extract_gazepoint_hrv_fuzzy_csi returns FuzzyEn and CSI", {
  dat <- data.frame(
    participant = "p1",
    IBI = 0.8 + 0.04 * sin(seq(0, 8 * pi, length.out = 80)) +
      rnorm(80, sd = 0.005)
  )

  out <- extract_gazepoint_hrv_fuzzy_csi(
    dat,
    ibi_col = "IBI",
    group_cols = "participant"
  )

  expect_s3_class(out, "gazepoint_hrv_fuzzy_csi")
  expect_true("fuzzy_entropy" %in% names(out$features))
  expect_true("csi" %in% names(out$features))
  expect_true(is.finite(out$features$csi))
})

test_that("detect_gazepoint_doubly_stochastic_changepoints returns score table", {
  time <- seq(0, 120, by = 1)
  signal <- c(rnorm(61, 0, 0.1), rnorm(60, 2, 0.1))

  dat <- data.frame(
    participant = "p1",
    time = time,
    signal = signal
  )

  out <- detect_gazepoint_doubly_stochastic_changepoints(
    dat,
    signal_col = "signal",
    time_col = "time",
    group_cols = "participant",
    window_seconds = 10,
    step_seconds = 5,
    threshold_mad_multiplier = 3
  )

  expect_s3_class(out, "gazepoint_doubly_stochastic_changepoints")
  expect_true(nrow(out$score_table) > 0)
  expect_true("change_score" %in% names(out$score_table))
})

test_that("extract_gazepoint_scr_recovery_times returns rec_t2 and rec_tc", {
  time <- seq(0, 40, by = 0.5)
  response_time <- pmax(0, time - 5)

  eda <- 1 + ifelse(
    time >= 5,
    exp(-response_time / 8) - exp(-response_time / 0.8),
    0
  )

  eda <- eda / max(eda)

  dat <- data.frame(
    participant = "p1",
    time = time,
    GSR_US = eda,
    onset = c(5, rep(NA_real_, length(time) - 1))
  )

  out <- extract_gazepoint_scr_recovery_times(
    dat,
    eda_col = "GSR_US",
    time_col = "time",
    event_onset_col = "onset",
    group_cols = "participant",
    peak_window_s = 10,
    recovery_window_s = 30
  )

  expect_s3_class(out, "gazepoint_scr_recovery_times")
  expect_true("rec_t2" %in% names(out$recovery_table))
  expect_true("rec_tc" %in% names(out$recovery_table))
})

#' Plot Gazepoint biometric signal time series
#'
#' Plots one or more Gazepoint biometric signals using base R graphics and
#' returns the plotted data and signal summary. The helper is intentionally
#' descriptive and does not infer emotional valence, cognition, or HRV from raw
#' biometric columns.
#'
#' @param data A data frame.
#' @param signal_cols Optional character vector of signal columns. If `NULL`,
#'   common Gazepoint biometric signal columns are detected.
#' @param time_col Optional time/order column for the x-axis. If `NULL`, row
#'   number is used.
#' @param group_col Optional grouping column recorded in the returned overview.
#'   The current plotting implementation overlays the selected rows rather than
#'   faceting by group.
#' @param max_points Maximum number of rows to plot. Large data are evenly
#'   downsampled for display only; returned summaries still describe the input
#'   signal columns.
#' @param standardize Logical. Should each signal be z-standardised before
#'   plotting? This is useful when signals are on different scales.
#' @param type Plot type: `"line"`, `"points"`, or `"both"`.
#' @param main Optional plot title.
#' @param xlab Optional x-axis label.
#' @param ylab Optional y-axis label.
#' @param legend Logical. Should a legend be drawn when more than one signal is
#'   plotted?
#' @param plot Logical. If `FALSE`, no plot is drawn and only the plot object is
#'   returned.
#' @param ... Additional arguments passed to [matplot()].
#'
#' @return A list with `overview`, `plot_data`, `signal_summary`, and `settings`.
#'
#' @examples
#' df <- data.frame(
#'   CNT = 1:5,
#'   GSR = c(1, 1.1, 1.2, 1.1, 1),
#'   HR = c(70, 71, 72, 71, 70)
#' )
#' plot_gazepoint_biometric_signals(df, time_col = "CNT", plot = FALSE)
#'
#' @export
plot_gazepoint_biometric_signals <- function(data,
                                             signal_cols = NULL,
                                             time_col = NULL,
                                             group_col = NULL,
                                             max_points = 5000L,
                                             standardize = FALSE,
                                             type = c("line", "points", "both"),
                                             main = NULL,
                                             xlab = NULL,
                                             ylab = NULL,
                                             legend = TRUE,
                                             plot = TRUE,
                                             ...) {
  if (!is.data.frame(data)) {
    stop("`data` must be a data frame.", call. = FALSE)
  }

  type <- match.arg(type)

  if (!is.null(signal_cols)) {
    .gpbiom_assert_columns(data, signal_cols, "signal_cols")
  }

  if (!is.null(time_col)) {
    .gpbiom_assert_columns(data, time_col, "time_col")
    if (length(time_col) != 1L) {
      stop("`time_col` must be NULL or a single column name.", call. = FALSE)
    }
  }

  if (!is.null(group_col)) {
    .gpbiom_assert_columns(data, group_col, "group_col")
    if (length(group_col) != 1L) {
      stop("`group_col` must be NULL or a single column name.", call. = FALSE)
    }
  }

  .gpbiom_assert_positive_integer(max_points, "max_points")

  if (!is.logical(standardize) ||
      length(standardize) != 1L ||
      is.na(standardize)) {
    stop("`standardize` must be TRUE or FALSE.", call. = FALSE)
  }

  if (!is.logical(legend) || length(legend) != 1L || is.na(legend)) {
    stop("`legend` must be TRUE or FALSE.", call. = FALSE)
  }

  if (!is.logical(plot) || length(plot) != 1L || is.na(plot)) {
    stop("`plot` must be TRUE or FALSE.", call. = FALSE)
  }

  if (is.null(signal_cols)) {
    signal_cols <- .gpbiom_detect_signal_columns(data)
  }

  if (length(signal_cols) == 0L) {
    stop("No biometric signal columns were detected. Provide `signal_cols` explicitly.",
         call. = FALSE)
  }

  non_numeric <- signal_cols[!vapply(signal_cols, function(column) {
    is.numeric(data[[column]])
  }, logical(1))]

  if (length(non_numeric) > 0L) {
    stop(
      "All selected `signal_cols` must be numeric. Non-numeric columns: ",
      paste(non_numeric, collapse = ", "),
      call. = FALSE
    )
  }

  order_index <- .gpbiom_plot_order_index(data, time_col)
  plot_index <- .gpbiom_plot_downsample_index(order_index, max_points)

  x <- if (is.null(time_col)) {
    seq_len(nrow(data))[plot_index]
  } else {
    data[[time_col]][plot_index]
  }

  signal_matrix <- as.matrix(data[plot_index, signal_cols, drop = FALSE])

  if (isTRUE(standardize)) {
    signal_matrix <- .gpbiom_plot_standardize_matrix(signal_matrix)
  }

  plot_data <- data.frame(
    .row_id = plot_index,
    .x = x,
    signal_matrix,
    check.names = FALSE
  )

  signal_summary <- .gpbiom_summarise_signal_columns(
    data = data,
    signal_cols = signal_cols,
    active_min_unique = 2L
  )

  overview <- data.frame(
    n_rows = nrow(data),
    plotted_rows = nrow(plot_data),
    signal_column_count = length(signal_cols),
    time_col = ifelse(is.null(time_col), NA_character_, time_col),
    group_col = ifelse(is.null(group_col), NA_character_, group_col),
    group_count = .gpbiom_plot_group_count(data, group_col),
    standardize = standardize,
    plot_created = plot,
    status = if (nrow(plot_data) == 0L) {
      "no_rows_to_plot"
    } else {
      "signal_plot_prepared"
    },
    stringsAsFactors = FALSE
  )

  if (isTRUE(plot)) {
    plot_type <- switch(
      type,
      line = "l",
      points = "p",
      both = "b"
    )

    plot_main <- if (is.null(main)) {
      "Gazepoint biometric signals"
    } else {
      main
    }

    plot_xlab <- if (is.null(xlab)) {
      if (is.null(time_col)) "Row" else time_col
    } else {
      xlab
    }

    plot_ylab <- if (is.null(ylab)) {
      if (isTRUE(standardize)) "Standardised signal" else "Signal value"
    } else {
      ylab
    }

    graphics::matplot(
      x = plot_data$.x,
      y = as.matrix(plot_data[signal_cols]),
      type = plot_type,
      lty = 1,
      pch = 1,
      xlab = plot_xlab,
      ylab = plot_ylab,
      main = plot_main,
      ...
    )

    if (isTRUE(legend) && length(signal_cols) > 1L) {
      graphics::legend(
        "topright",
        legend = signal_cols,
        col = seq_along(signal_cols),
        lty = 1,
        pch = 1,
        bty = "n"
      )
    }
  }

  out <- list(
    overview = overview,
    plot_data = plot_data,
    signal_summary = signal_summary,
    settings = list(
      signal_cols = signal_cols,
      time_col = time_col,
      group_col = group_col,
      max_points = as.integer(max_points),
      standardize = standardize,
      type = type,
      note = paste0(
        "Signal plots describe biometric time-series patterns only; they do ",
        "not establish emotional valence, cognition, or HRV."
      )
    )
  )

  class(out) <- c("gazepoint_biometric_signal_plot", class(out))
  out
}


#' Plot Gazepoint biometric quality indicators
#'
#' Plots and summarises biometric quality indicators such as dropout flags,
#' validity flags, missingness flags, and quality/audit flags. When no explicit
#' quality columns are available, the function can derive missingness indicators
#' from detected biometric signal columns.
#'
#' @param data A data frame.
#' @param quality_cols Optional quality/flag columns. If `NULL`, likely quality
#'   columns are detected from names and types.
#' @param signal_cols Optional signal columns used to derive missingness flags
#'   when no quality columns are detected.
#' @param time_col Optional time/order column recorded in the returned settings.
#' @param group_col Optional grouping column for group-level quality summaries.
#' @param dropout_prefix Prefix used by dropout columns created by
#'   [flag_gazepoint_biometric_dropouts()].
#' @param max_points Maximum number of rows used for row-level returned plot
#'   data.
#' @param main Optional plot title.
#' @param plot Logical. If `FALSE`, no plot is drawn.
#' @param ... Additional arguments passed to [barplot()].
#'
#' @return A list with `overview`, `quality_summary`, `group_summary`,
#'   `plot_data`, and `settings`.
#'
#' @examples
#' df <- data.frame(
#'   CNT = 1:5,
#'   GSR = c(1, NA, 1.2, 1.1, NA),
#'   HR_valid = c(1, 1, 0, 1, 1)
#' )
#' plot_gazepoint_biometric_quality(df, signal_cols = "GSR", plot = FALSE)
#'
#' @export
plot_gazepoint_biometric_quality <- function(data,
                                             quality_cols = NULL,
                                             signal_cols = NULL,
                                             time_col = NULL,
                                             group_col = NULL,
                                             dropout_prefix = "biometric_dropout",
                                             max_points = 5000L,
                                             main = NULL,
                                             plot = TRUE,
                                             ...) {
  if (!is.data.frame(data)) {
    stop("`data` must be a data frame.", call. = FALSE)
  }

  if (!is.null(quality_cols)) {
    .gpbiom_assert_columns(data, quality_cols, "quality_cols")
  }

  if (!is.null(signal_cols)) {
    .gpbiom_assert_columns(data, signal_cols, "signal_cols")
  }

  if (!is.null(time_col)) {
    .gpbiom_assert_columns(data, time_col, "time_col")
    if (length(time_col) != 1L) {
      stop("`time_col` must be NULL or a single column name.", call. = FALSE)
    }
  }

  if (!is.null(group_col)) {
    .gpbiom_assert_columns(data, group_col, "group_col")
    if (length(group_col) != 1L) {
      stop("`group_col` must be NULL or a single column name.", call. = FALSE)
    }
  }

  if (!is.character(dropout_prefix) ||
      length(dropout_prefix) != 1L ||
      is.na(dropout_prefix) ||
      !nzchar(dropout_prefix)) {
    stop("`dropout_prefix` must be a non-empty character string.",
         call. = FALSE)
  }

  .gpbiom_assert_positive_integer(max_points, "max_points")

  if (!is.logical(plot) || length(plot) != 1L || is.na(plot)) {
    stop("`plot` must be TRUE or FALSE.", call. = FALSE)
  }

  if (is.null(quality_cols)) {
    quality_cols <- .gpbiom_detect_quality_columns(data, dropout_prefix)
  }

  derived_from_signals <- FALSE

  if (length(quality_cols) == 0L) {
    if (is.null(signal_cols)) {
      signal_cols <- .gpbiom_detect_signal_columns(data)
    }

    if (length(signal_cols) == 0L) {
      stop("No quality columns or signal columns were detected.",
           call. = FALSE)
    }

    quality_data <- .gpbiom_derive_missingness_quality_data(data, signal_cols)
    quality_cols <- names(quality_data)
    derived_from_signals <- TRUE
  } else {
    quality_data <- data[quality_cols]
  }

  quality_summary <- .gpbiom_quality_summary(
    quality_data = quality_data,
    original_names = quality_cols,
    derived_from_signals = derived_from_signals
  )

  group_summary <- .gpbiom_quality_group_summary(
    data = data,
    quality_data = quality_data,
    group_col = group_col,
    derived_from_signals = derived_from_signals
  )

  order_index <- .gpbiom_plot_order_index(data, time_col)
  plot_index <- .gpbiom_plot_downsample_index(order_index, max_points)

  plot_data <- data.frame(
    .row_id = plot_index,
    quality_data[plot_index, , drop = FALSE],
    check.names = FALSE
  )

  overview <- data.frame(
    n_rows = nrow(data),
    plotted_rows = nrow(plot_data),
    quality_column_count = length(quality_cols),
    group_col = ifelse(is.null(group_col), NA_character_, group_col),
    group_count = .gpbiom_plot_group_count(data, group_col),
    derived_from_signals = derived_from_signals,
    plot_created = plot,
    status = if (nrow(quality_summary) == 0L) {
      "no_quality_summary"
    } else if (any(quality_summary$flag_rate > 0, na.rm = TRUE)) {
      "quality_flags_present"
    } else {
      "no_quality_flags_present"
    },
    stringsAsFactors = FALSE
  )

  if (isTRUE(plot)) {
    plot_main <- if (is.null(main)) {
      "Gazepoint biometric quality indicators"
    } else {
      main
    }

    heights <- quality_summary$flag_rate
    names(heights) <- quality_summary$column

    graphics::barplot(
      heights,
      ylim = c(0, 1),
      ylab = "Flag rate",
      main = plot_main,
      las = 2,
      ...
    )
  }

  out <- list(
    overview = overview,
    quality_summary = quality_summary,
    group_summary = group_summary,
    plot_data = plot_data,
    settings = list(
      quality_cols = quality_cols,
      signal_cols = signal_cols,
      time_col = time_col,
      group_col = group_col,
      dropout_prefix = dropout_prefix,
      max_points = as.integer(max_points),
      derived_from_signals = derived_from_signals,
      note = paste0(
        "Quality plots summarise availability, validity, dropout, or ",
        "missingness indicators; they are not physiological interpretations."
      )
    )
  )

  class(out) <- c("gazepoint_biometric_quality_plot", class(out))
  out
}


.gpbiom_plot_order_index <- function(data, time_col) {
  index <- seq_len(nrow(data))

  if (is.null(time_col)) {
    return(index)
  }

  index[order(data[[time_col]], na.last = TRUE)]
}


.gpbiom_plot_downsample_index <- function(index, max_points) {
  if (length(index) <= max_points) {
    return(index)
  }

  index[unique(round(seq(1, length(index), length.out = max_points)))]
}


.gpbiom_plot_standardize_matrix <- function(x) {
  out <- x

  for (j in seq_len(ncol(out))) {
    values <- out[, j]
    finite <- values[!is.na(values) & is.finite(values)]

    if (length(finite) == 0L) {
      out[, j] <- NA_real_
      next
    }

    center <- mean(finite)
    scale <- stats::sd(finite)

    if (!is.finite(scale) || scale == 0) {
      out[, j] <- values - center
    } else {
      out[, j] <- (values - center) / scale
    }
  }

  out
}


.gpbiom_plot_group_count <- function(data, group_col) {
  if (is.null(group_col)) {
    return(NA_integer_)
  }

  length(unique(data[[group_col]][!is.na(data[[group_col]])]))
}


.gpbiom_detect_quality_columns <- function(data, dropout_prefix) {
  column_names <- names(data)
  cleaned <- .gpbiom_clean_name(column_names)

  quality_name <- grepl(
    paste0(
      "dropout|valid|validity|quality|flag|artifact|artefact|",
      "missing|nonfinite|excluded|exclude|", dropout_prefix
    ),
    cleaned,
    ignore.case = TRUE
  )

  usable_type <- vapply(data, function(column) {
    is.logical(column) || is.numeric(column) || is.character(column) ||
      is.factor(column)
  }, logical(1))

  column_names[quality_name & usable_type]
}


.gpbiom_derive_missingness_quality_data <- function(data, signal_cols) {
  out <- data.frame(row.names = seq_len(nrow(data)))

  for (signal in signal_cols) {
    safe <- .gpbiom_safe_colname(signal)
    values <- data[[signal]]

    flags <- is.na(values)

    if (is.numeric(values)) {
      flags <- flags | !is.finite(values)
    }

    out[[paste0(safe, "_missing")]] <- flags
  }

  out
}


.gpbiom_quality_summary <- function(quality_data,
                                    original_names,
                                    derived_from_signals) {
  if (length(original_names) == 0L || ncol(quality_data) == 0L) {
    return(data.frame(
      column = character(),
      n = integer(),
      n_flagged = integer(),
      flag_rate = numeric(),
      n_missing = integer(),
      missing_rate = numeric(),
      source = character(),
      stringsAsFactors = FALSE
    ))
  }

  rows <- lapply(seq_along(original_names), function(i) {
    column <- names(quality_data)[i]
    values <- quality_data[[i]]
    flags <- .gpbiom_quality_flag_vector(values, column)

    data.frame(
      column = original_names[i],
      n = length(values),
      n_flagged = sum(flags, na.rm = TRUE),
      flag_rate = if (length(values) > 0L) mean(flags, na.rm = TRUE) else NA_real_,
      n_missing = sum(is.na(values)),
      missing_rate = if (length(values) > 0L) mean(is.na(values)) else NA_real_,
      source = if (isTRUE(derived_from_signals)) {
        "derived_signal_missingness"
      } else {
        "quality_column"
      },
      stringsAsFactors = FALSE
    )
  })

  do.call(rbind, rows)
}


.gpbiom_quality_group_summary <- function(data,
                                          quality_data,
                                          group_col,
                                          derived_from_signals) {
  if (is.null(group_col)) {
    return(data.frame(
      group = character(),
      column = character(),
      n = integer(),
      n_flagged = integer(),
      flag_rate = numeric(),
      source = character(),
      stringsAsFactors = FALSE
    ))
  }

  groups <- split(seq_len(nrow(data)), data[[group_col]], drop = TRUE)

  rows <- list()

  for (group_name in names(groups)) {
    index <- groups[[group_name]]

    for (column in names(quality_data)) {
      values <- quality_data[[column]][index]
      flags <- .gpbiom_quality_flag_vector(values, column)

      rows[[length(rows) + 1L]] <- data.frame(
        group = group_name,
        column = column,
        n = length(values),
        n_flagged = sum(flags, na.rm = TRUE),
        flag_rate = if (length(values) > 0L) mean(flags, na.rm = TRUE) else NA_real_,
        source = if (isTRUE(derived_from_signals)) {
          "derived_signal_missingness"
        } else {
          "quality_column"
        },
        stringsAsFactors = FALSE
      )
    }
  }

  if (length(rows) == 0L) {
    return(data.frame(
      group = character(),
      column = character(),
      n = integer(),
      n_flagged = integer(),
      flag_rate = numeric(),
      source = character(),
      stringsAsFactors = FALSE
    ))
  }

  do.call(rbind, rows)
}


.gpbiom_quality_flag_vector <- function(values, column_name) {
  name <- tolower(column_name)

  if (is.logical(values)) {
    return(!is.na(values) & values)
  }

  if (is.numeric(values)) {
    if (grepl("valid", name) && !grepl("invalid", name)) {
      return(!is.na(values) & is.finite(values) & values <= 0)
    }

    return(!is.na(values) & is.finite(values) & values != 0)
  }

  cleaned <- tolower(trimws(as.character(values)))

  if (grepl("valid", name) && !grepl("invalid", name)) {
    return(!is.na(values) & cleaned %in% c(
      "0", "false", "invalid", "no", "n", "bad", "fail", "failed"
    ))
  }

  !is.na(values) & cleaned %in% c(
    "1", "true", "yes", "y", "flag", "flagged", "bad", "fail",
    "failed", "invalid", "artifact", "artefact", "missing", "exclude",
    "excluded"
  )
}

test_that("plot_gazepoint_biometric_signals prepares signal plot data", {
  df <- data.frame(
    CNT = 1:5,
    GSR = c(1, 1.1, 1.2, 1.1, 1),
    HR = c(70, 71, 72, 71, 70)
  )

  out <- plot_gazepoint_biometric_signals(
    df,
    time_col = "CNT",
    plot = FALSE
  )

  expect_s3_class(out, "gazepoint_biometric_signal_plot")
  expect_s3_class(out$overview, "data.frame")
  expect_s3_class(out$plot_data, "data.frame")
  expect_s3_class(out$signal_summary, "data.frame")

  expect_equal(out$overview$n_rows, 5)
  expect_equal(out$overview$plotted_rows, 5)
  expect_equal(out$overview$signal_column_count, 2)
  expect_equal(out$settings$time_col, "CNT")
  expect_true(all(c("GSR", "HR") %in% names(out$plot_data)))
})


test_that("plot_gazepoint_biometric_signals supports explicit signal columns", {
  df <- data.frame(
    time = 1:4,
    custom_signal = c(1, 2, 3, 4)
  )

  out <- plot_gazepoint_biometric_signals(
    df,
    signal_cols = "custom_signal",
    time_col = "time",
    plot = FALSE
  )

  expect_equal(out$settings$signal_cols, "custom_signal")
  expect_equal(out$overview$signal_column_count, 1)
})


test_that("plot_gazepoint_biometric_signals downsamples large data for display", {
  df <- data.frame(
    CNT = 1:100,
    GSR = seq(1, 2, length.out = 100)
  )

  out <- plot_gazepoint_biometric_signals(
    df,
    time_col = "CNT",
    max_points = 10,
    plot = FALSE
  )

  expect_equal(out$overview$n_rows, 100)
  expect_true(out$overview$plotted_rows <= 10)
})


test_that("plot_gazepoint_biometric_signals can standardize signals", {
  df <- data.frame(
    CNT = 1:5,
    GSR = c(1, 2, 3, 4, 5)
  )

  out <- plot_gazepoint_biometric_signals(
    df,
    time_col = "CNT",
    standardize = TRUE,
    plot = FALSE
  )

  expect_equal(round(mean(out$plot_data$GSR), 10), 0)
  expect_equal(round(stats::sd(out$plot_data$GSR), 10), 1)
})


test_that("plot_gazepoint_biometric_signals can draw without error", {
  df <- data.frame(
    CNT = 1:5,
    GSR = c(1, 1.1, 1.2, 1.1, 1),
    HR = c(70, 71, 72, 71, 70)
  )

  path <- tempfile(fileext = ".png")
  grDevices::png(path)
  expect_silent(
    plot_gazepoint_biometric_signals(
      df,
      time_col = "CNT",
      plot = TRUE
    )
  )
  grDevices::dev.off()

  expect_true(file.exists(path))
})


test_that("plot_gazepoint_biometric_signals validates arguments", {
  df <- data.frame(GSR = c(1, 2, 3))

  expect_error(
    plot_gazepoint_biometric_signals(1:3),
    "`data` must be"
  )

  expect_error(
    plot_gazepoint_biometric_signals(df, signal_cols = "missing"),
    "not found"
  )

  expect_error(
    plot_gazepoint_biometric_signals(df, time_col = "missing"),
    "not found"
  )

  expect_error(
    plot_gazepoint_biometric_signals(df, max_points = 0),
    "`max_points`"
  )

  expect_error(
    plot_gazepoint_biometric_signals(df, standardize = NA),
    "`standardize`"
  )

  expect_error(
    plot_gazepoint_biometric_signals(df, plot = NA),
    "`plot`"
  )

  expect_error(
    plot_gazepoint_biometric_signals(data.frame(label = letters[1:3])),
    "No biometric signal"
  )

  expect_error(
    plot_gazepoint_biometric_signals(
      data.frame(GSR = letters[1:3]),
      signal_cols = "GSR"
    ),
    "must be numeric"
  )
})


test_that("plot_gazepoint_biometric_quality summarises explicit quality columns", {
  df <- data.frame(
    CNT = 1:5,
    HR_valid = c(1, 1, 0, 1, 0),
    biometric_dropout_any = c(FALSE, TRUE, FALSE, FALSE, TRUE)
  )

  out <- plot_gazepoint_biometric_quality(
    df,
    quality_cols = c("HR_valid", "biometric_dropout_any"),
    plot = FALSE
  )

  expect_s3_class(out, "gazepoint_biometric_quality_plot")
  expect_s3_class(out$overview, "data.frame")
  expect_s3_class(out$quality_summary, "data.frame")
  expect_s3_class(out$plot_data, "data.frame")

  expect_equal(out$overview$quality_column_count, 2)
  expect_equal(out$overview$status, "quality_flags_present")
  expect_equal(out$quality_summary$n_flagged[out$quality_summary$column == "HR_valid"], 2)
  expect_equal(out$quality_summary$n_flagged[out$quality_summary$column == "biometric_dropout_any"], 2)
})


test_that("plot_gazepoint_biometric_quality can derive missingness from signals", {
  df <- data.frame(
    CNT = 1:5,
    GSR = c(1, NA, 1.2, 1.1, NA)
  )

  out <- plot_gazepoint_biometric_quality(
    df,
    signal_cols = "GSR",
    plot = FALSE
  )

  expect_true(out$overview$derived_from_signals)
  expect_equal(out$quality_summary$n_flagged, 2)
  expect_equal(out$quality_summary$source, "derived_signal_missingness")
})


test_that("plot_gazepoint_biometric_quality supports group summaries", {
  df <- data.frame(
    USER = rep(c("P1", "P2"), each = 3),
    GSR_missing = c(FALSE, TRUE, FALSE, TRUE, TRUE, FALSE)
  )

  out <- plot_gazepoint_biometric_quality(
    df,
    quality_cols = "GSR_missing",
    group_col = "USER",
    plot = FALSE
  )

  expect_equal(nrow(out$group_summary), 2)
  expect_true(all(out$group_summary$group %in% c("P1", "P2")))
  expect_equal(out$overview$group_count, 2)
})


test_that("plot_gazepoint_biometric_quality can draw without error", {
  df <- data.frame(
    CNT = 1:5,
    HR_valid = c(1, 1, 0, 1, 0)
  )

  path <- tempfile(fileext = ".png")
  grDevices::png(path)
  expect_silent(
    plot_gazepoint_biometric_quality(
      df,
      quality_cols = "HR_valid",
      plot = TRUE
    )
  )
  grDevices::dev.off()

  expect_true(file.exists(path))
})


test_that("plot_gazepoint_biometric_quality validates arguments", {
  df <- data.frame(GSR = c(1, 2, 3))

  expect_error(
    plot_gazepoint_biometric_quality(1:3),
    "`data` must be"
  )

  expect_error(
    plot_gazepoint_biometric_quality(df, quality_cols = "missing"),
    "not found"
  )

  expect_error(
    plot_gazepoint_biometric_quality(df, signal_cols = "missing"),
    "not found"
  )

  expect_error(
    plot_gazepoint_biometric_quality(df, group_col = "missing"),
    "not found"
  )

  expect_error(
    plot_gazepoint_biometric_quality(df, dropout_prefix = ""),
    "`dropout_prefix`"
  )

  expect_error(
    plot_gazepoint_biometric_quality(df, max_points = 0),
    "`max_points`"
  )

  expect_error(
    plot_gazepoint_biometric_quality(df, plot = NA),
    "`plot`"
  )

  expect_error(
    plot_gazepoint_biometric_quality(data.frame(label = letters[1:3])),
    "No quality columns or signal columns"
  )
})

#' Summarise full Gazepoint biometric windows
#'
#' Creates a combined window-level summary table containing GSR/EDA,
#' heart-rate, engagement-dial, and IBI-derived HRV summaries. This function is
#' intended for biometric analyses where both continuous physiological values
#' and interbeat-interval variability features are needed.
#'
#' @param data A data frame or a path to a Gazepoint CSV export.
#' @param group_columns Columns defining analysis windows, such as
#'   `c("source_participant", "MEDIA_ID")`.
#' @param include_ibi_hrv Logical. Should IBI-derived HRV summaries be included?
#'
#' @return A data frame with one row per window and prefixed biometric summary
#'   columns.
#'
#' @export
summarise_gazepoint_full_biometric_windows <- function(data,
                                                       group_columns,
                                                       include_ibi_hrv = TRUE) {
  dat <- coerce_gazepoint_biometrics_data(data)

  if (missing(group_columns) || is.null(group_columns) || length(group_columns) == 0L) {
    stop("`group_columns` must define the analysis windows.", call. = FALSE)
  }

  missing_groups <- setdiff(group_columns, names(dat))

  if (length(missing_groups) > 0L) {
    stop(
      "`group_columns` were not found in `data`: ",
      paste(missing_groups, collapse = ", "),
      call. = FALSE
    )
  }

  multimodal <- summarise_gazepoint_multimodal_windows(
    data = dat,
    group_columns = group_columns
  )

  if (!isTRUE(include_ibi_hrv)) {
    return(multimodal)
  }

  if (!"IBI" %in% names(dat)) {
    return(multimodal)
  }

  ibi <- summarise_gazepoint_ibi_hrv_windows(
    data = dat,
    group_columns = group_columns
  )

  ibi <- prefix_ibi_summary(ibi)

  out <- merge(
    multimodal,
    ibi,
    by = group_columns,
    all = TRUE,
    sort = FALSE
  )

  rownames(out) <- NULL
  out
}


prefix_ibi_summary <- function(summary) {
  columns_to_prefix <- setdiff(
    names(summary),
    c("source_participant", "USER", "USERID", "participant", "subject", "MEDIA_ID", "MEDIA_NAME")
  )

  group_like_columns <- names(summary)[
    names(summary) %in% c("source_participant", "USER", "USERID", "participant", "subject", "MEDIA_ID", "MEDIA_NAME")
  ]

  non_group_columns <- setdiff(names(summary), group_like_columns)

  names(summary)[match(non_group_columns, names(summary))] <-
    paste("ibi", non_group_columns, sep = "_")

  summary
}

test_that("summarise_gazepoint_full_biometric_windows combines multimodal and IBI summaries", {
  dat <- data.frame(
    source_participant = c("User 1", "User 1", "User 1"),
    MEDIA_ID = c(0, 0, 0),
    GSR_US = c(1.1, 1.2, 1.3),
    GSRV = c(1, 1, 1),
    HR = c(70, 72, 74),
    HRV = c(1, 1, 1),
    DIAL = c(1, 1, 1),
    DIALV = c(1, 1, 1),
    IBI = c(1.0, 1.1, 0.9)
  )

  out <- summarise_gazepoint_full_biometric_windows(
    dat,
    group_columns = c("source_participant", "MEDIA_ID")
  )

  expect_equal(nrow(out), 1)
  expect_true("gsr_mean_value" %in% names(out))
  expect_true("hr_mean_value" %in% names(out))
  expect_true("dial_mean_value" %in% names(out))
  expect_true("ibi_mean_ibi_sec" %in% names(out))
  expect_true("ibi_rmssd_ms" %in% names(out))
})


test_that("summarise_gazepoint_full_biometric_windows can omit IBI summaries", {
  dat <- data.frame(
    source_participant = c("User 1", "User 1"),
    MEDIA_ID = c(0, 0),
    GSR_US = c(1.1, 1.2),
    GSRV = c(1, 1),
    HR = c(70, 72),
    HRV = c(1, 1),
    DIAL = c(1, 1),
    DIALV = c(1, 1),
    IBI = c(1.0, 1.1)
  )

  out <- summarise_gazepoint_full_biometric_windows(
    dat,
    group_columns = c("source_participant", "MEDIA_ID"),
    include_ibi_hrv = FALSE
  )

  expect_true("gsr_mean_value" %in% names(out))
  expect_false("ibi_mean_ibi_sec" %in% names(out))
})


test_that("summarise_gazepoint_full_biometric_windows works without IBI column", {
  dat <- data.frame(
    source_participant = c("User 1", "User 1"),
    MEDIA_ID = c(0, 0),
    GSR_US = c(1.1, 1.2),
    GSRV = c(1, 1),
    HR = c(70, 72),
    HRV = c(1, 1),
    DIAL = c(1, 1),
    DIALV = c(1, 1)
  )

  out <- summarise_gazepoint_full_biometric_windows(
    dat,
    group_columns = c("source_participant", "MEDIA_ID")
  )

  expect_true("gsr_mean_value" %in% names(out))
  expect_false("ibi_mean_ibi_sec" %in% names(out))
})


test_that("summarise_gazepoint_full_biometric_windows rejects missing grouping columns", {
  dat <- data.frame(
    GSR_US = c(1.1, 1.2),
    GSRV = c(1, 1)
  )

  expect_error(
    summarise_gazepoint_full_biometric_windows(
      dat,
      group_columns = "source_participant"
    ),
    "group_columns"
  )
})

#' Standardise a Gazepoint plot return contract
#'
#' Adds consistent attributes to a ggplot object so plotting helpers can be
#' tested and reused in automated reports.
#'
#' @param plot A ggplot object.
#' @param plot_data Optional data frame used to create the plot.
#' @param settings Optional list of plot settings.
#' @param interpretation_notes Optional character vector of interpretation notes.
#' @param plot_type Optional short plot-type label.
#'
#' @return A ggplot object with standardized attributes.
#' @export
standardise_gazepoint_plot_contract <- function(plot,
                                                plot_data = NULL,
                                                settings = list(),
                                                interpretation_notes = NULL,
                                                plot_type = NULL) {
  gpbiometrics_require_ggplot2()

  if (!inherits(plot, "ggplot")) {
    stop("`plot` must be a ggplot object.", call. = FALSE)
  }

  if (!is.null(plot_data) && !is.data.frame(plot_data)) {
    stop("`plot_data` must be NULL or a data frame.", call. = FALSE)
  }

  if (!is.list(settings)) {
    stop("`settings` must be a list.", call. = FALSE)
  }

  if (!is.null(interpretation_notes) && !is.character(interpretation_notes)) {
    stop("`interpretation_notes` must be NULL or a character vector.", call. = FALSE)
  }

  if (!is.null(plot_type) &&
      (!is.character(plot_type) || length(plot_type) != 1 || is.na(plot_type))) {
    stop("`plot_type` must be NULL or a single character value.", call. = FALSE)
  }

  if (is.null(plot_data) && is.data.frame(attr(plot, "plot_data"))) {
    plot_data <- attr(plot, "plot_data")
  }

  if (length(settings) == 0 && is.list(attr(plot, "settings"))) {
    settings <- attr(plot, "settings")
  }

  if (is.null(interpretation_notes)) {
    existing_notes <- attr(plot, "interpretation_notes")

    if (is.character(existing_notes)) {
      interpretation_notes <- existing_notes
    } else if (is.list(settings) &&
               "interpretation_notes" %in% names(settings) &&
               is.character(settings$interpretation_notes)) {
      interpretation_notes <- settings$interpretation_notes
    } else {
      interpretation_notes <- character()
    }
  }

  if (is.null(plot_type) &&
      is.list(settings) &&
      "plot_type" %in% names(settings) &&
      is.character(settings$plot_type) &&
      length(settings$plot_type) == 1) {
    plot_type <- settings$plot_type
  }

  attr(plot, "plot_data") <- plot_data
  attr(plot, "settings") <- settings
  attr(plot, "interpretation_notes") <- interpretation_notes
  attr(plot, "plot_type") <- plot_type
  attr(plot, "gazepoint_plot_contract") <- TRUE

  if (!inherits(plot, "gazepoint_plot")) {
    class(plot) <- c("gazepoint_plot", class(plot))
  }

  plot
}

#' Check a Gazepoint plot return contract
#'
#' Checks whether an object follows the package's plotting return convention.
#'
#' @param plot A plot object.
#' @param require_plot_data Logical. If `TRUE`, `plot_data` must be present.
#' @param require_settings Logical. If `TRUE`, `settings` must be present.
#'
#' @return A list with `overview`, `checks`, `plot_data`, and `settings`.
#' @export
check_gazepoint_plot_contract <- function(plot,
                                          require_plot_data = TRUE,
                                          require_settings = TRUE) {
  if (!is.logical(require_plot_data) ||
      length(require_plot_data) != 1 ||
      is.na(require_plot_data)) {
    stop("`require_plot_data` must be TRUE or FALSE.", call. = FALSE)
  }

  if (!is.logical(require_settings) ||
      length(require_settings) != 1 ||
      is.na(require_settings)) {
    stop("`require_settings` must be TRUE or FALSE.", call. = FALSE)
  }

  is_ggplot <- inherits(plot, "ggplot")
  is_gazepoint_plot <- inherits(plot, "gazepoint_plot") ||
    isTRUE(attr(plot, "gazepoint_plot_contract"))

  plot_data <- attr(plot, "plot_data")
  settings <- attr(plot, "settings")
  interpretation_notes <- attr(plot, "interpretation_notes")
  plot_type <- attr(plot, "plot_type")

  has_plot_data <- is.data.frame(plot_data)
  has_settings <- is.list(settings)
  has_notes <- is.character(interpretation_notes)
  has_plot_type <- !is.null(plot_type) &&
    is.character(plot_type) &&
    length(plot_type) == 1 &&
    !is.na(plot_type)

  checks <- data.frame(
    check = c(
      "is_ggplot",
      "is_gazepoint_plot",
      "has_plot_data",
      "has_settings",
      "has_interpretation_notes",
      "has_plot_type"
    ),
    passed = c(
      is_ggplot,
      is_gazepoint_plot,
      has_plot_data,
      has_settings,
      has_notes,
      has_plot_type
    ),
    required = c(
      TRUE,
      FALSE,
      require_plot_data,
      require_settings,
      FALSE,
      FALSE
    ),
    stringsAsFactors = FALSE
  )

  required_failed <- checks$required & !checks$passed

  overview <- data.frame(
    is_ggplot = is_ggplot,
    is_gazepoint_plot = is_gazepoint_plot,
    has_plot_data = has_plot_data,
    plot_data_rows = if (has_plot_data) nrow(plot_data) else NA_integer_,
    has_settings = has_settings,
    has_interpretation_notes = has_notes,
    has_plot_type = has_plot_type,
    status = if (any(required_failed)) {
      "fail_plot_contract"
    } else if (!is_gazepoint_plot || !has_notes || !has_plot_type) {
      "warn_partial_plot_contract"
    } else {
      "pass_plot_contract"
    },
    stringsAsFactors = FALSE
  )

  structure(
    list(
      overview = overview,
      checks = checks,
      plot_data = if (has_plot_data) plot_data else data.frame(),
      settings = if (has_settings) settings else list()
    ),
    class = c("gazepoint_plot_contract_check", "list")
  )
}

#' Extract stored plot data
#'
#' @param plot A plot object returned by a gpbiometrics plotting helper.
#'
#' @return A data frame.
#' @export
get_gazepoint_plot_data <- function(plot) {
  plot_data <- attr(plot, "plot_data")

  if (!is.data.frame(plot_data)) {
    stop("No `plot_data` data frame is stored on this plot object.", call. = FALSE)
  }

  plot_data
}

#' Extract stored plot settings
#'
#' @param plot A plot object returned by a gpbiometrics plotting helper.
#'
#' @return A list.
#' @export
get_gazepoint_plot_settings <- function(plot) {
  settings <- attr(plot, "settings")

  if (!is.list(settings)) {
    stop("No `settings` list is stored on this plot object.", call. = FALSE)
  }

  settings
}

test_that("standardise_gazepoint_plot_contract adds plot attributes", {
  skip_if_not_installed("ggplot2")

  dat <- data.frame(
    x = 1:3,
    y = c(2, 3, 4)
  )

  p <- ggplot2::ggplot(dat, ggplot2::aes(x = x, y = y)) +
    ggplot2::geom_point()

  p2 <- standardise_gazepoint_plot_contract(
    plot = p,
    plot_data = dat,
    settings = list(plot_type = "point", x = "x", y = "y"),
    interpretation_notes = "Test plot contract.",
    plot_type = "point"
  )

  expect_s3_class(p2, "gazepoint_plot")
  expect_s3_class(p2, "ggplot")
  expect_true(isTRUE(attr(p2, "gazepoint_plot_contract")))
  expect_true(is.data.frame(attr(p2, "plot_data")))
  expect_true(is.list(attr(p2, "settings")))
  expect_equal(attr(p2, "plot_type"), "point")
  expect_equal(attr(p2, "interpretation_notes"), "Test plot contract.")
})

test_that("check_gazepoint_plot_contract detects complete contracts", {
  skip_if_not_installed("ggplot2")

  dat <- data.frame(
    x = 1:3,
    y = c(2, 3, 4)
  )

  p <- ggplot2::ggplot(dat, ggplot2::aes(x = x, y = y)) +
    ggplot2::geom_point()

  p <- standardise_gazepoint_plot_contract(
    plot = p,
    plot_data = dat,
    settings = list(plot_type = "point"),
    interpretation_notes = "Complete contract.",
    plot_type = "point"
  )

  check <- check_gazepoint_plot_contract(p)

  expect_s3_class(check, "gazepoint_plot_contract_check")
  expect_equal(check$overview$status, "pass_plot_contract")
  expect_true(check$overview$is_ggplot)
  expect_true(check$overview$has_plot_data)
  expect_equal(check$overview$plot_data_rows, 3)
})

test_that("check_gazepoint_plot_contract can warn for partial contracts", {
  skip_if_not_installed("ggplot2")

  dat <- data.frame(
    x = 1:3,
    y = c(2, 3, 4)
  )

  p <- ggplot2::ggplot(dat, ggplot2::aes(x = x, y = y)) +
    ggplot2::geom_point()

  check <- check_gazepoint_plot_contract(
    p,
    require_plot_data = FALSE,
    require_settings = FALSE
  )

  expect_equal(check$overview$status, "warn_partial_plot_contract")
  expect_true(check$overview$is_ggplot)
})

test_that("check_gazepoint_plot_contract fails when required data are missing", {
  skip_if_not_installed("ggplot2")

  dat <- data.frame(
    x = 1:3,
    y = c(2, 3, 4)
  )

  p <- ggplot2::ggplot(dat, ggplot2::aes(x = x, y = y)) +
    ggplot2::geom_point()

  check <- check_gazepoint_plot_contract(p)

  expect_equal(check$overview$status, "fail_plot_contract")
  expect_false(check$overview$has_plot_data)
})

test_that("get_gazepoint_plot_data and settings extract stored attributes", {
  skip_if_not_installed("ggplot2")

  dat <- data.frame(
    x = 1:3,
    y = c(2, 3, 4)
  )

  p <- ggplot2::ggplot(dat, ggplot2::aes(x = x, y = y)) +
    ggplot2::geom_point()

  p <- standardise_gazepoint_plot_contract(
    plot = p,
    plot_data = dat,
    settings = list(plot_type = "point"),
    interpretation_notes = "Stored plot.",
    plot_type = "point"
  )

  expect_equal(get_gazepoint_plot_data(p), dat)
  expect_equal(get_gazepoint_plot_settings(p)$plot_type, "point")
})

test_that("plot-contract helpers validate inputs", {
  skip_if_not_installed("ggplot2")

  dat <- data.frame(
    x = 1:3,
    y = c(2, 3, 4)
  )

  p <- ggplot2::ggplot(dat, ggplot2::aes(x = x, y = y)) +
    ggplot2::geom_point()

  expect_error(
    standardise_gazepoint_plot_contract("not a plot"),
    "`plot` must be a ggplot object"
  )

  expect_error(
    standardise_gazepoint_plot_contract(p, plot_data = list()),
    "`plot_data` must be NULL or a data frame"
  )

  expect_error(
    check_gazepoint_plot_contract(p, require_plot_data = NA),
    "`require_plot_data` must be TRUE or FALSE"
  )

  expect_error(
    get_gazepoint_plot_data(p),
    "No `plot_data` data frame"
  )

  expect_error(
    get_gazepoint_plot_settings(p),
    "No `settings` list"
  )
})

test_that("plot_gazepoint_aoi_biometrics follows the plot contract", {
  skip_if_not_installed("ggplot2")

  dat <- data.frame(
    participant = c("P1", "P1", "P2", "P2"),
    AOI = c("claim", "logo", "claim", "logo"),
    GSR_US = c(1, 2, 3, 4),
    HR = c(70, 72, 74, 76)
  )

  summary <- summarise_gazepoint_aoi_biometrics(
    dat,
    aoi_col = "AOI",
    signal_cols = c("GSR_US", "HR"),
    group_cols = "participant"
  )

  p <- plot_gazepoint_aoi_biometrics(
    summary,
    value_col = "mean_value",
    plot_type = "point",
    group_col = "participant"
  )

  check <- check_gazepoint_plot_contract(p)

  expect_s3_class(p, "gazepoint_plot")
  expect_equal(check$overview$status, "pass_plot_contract")
  expect_true(is.data.frame(get_gazepoint_plot_data(p)))
  expect_true(is.list(get_gazepoint_plot_settings(p)))
})

test_that("plot_gazepoint_eda_decomposition follows the plot contract", {
  skip_if_not_installed("ggplot2")

  dat <- data.frame(
    participant = rep("P1", 10),
    CNT = 0:9,
    GSR_US = seq(1, 2, length.out = 10),
    GSR_US_TONIC = seq(0.9, 1.8, length.out = 10),
    GSR_US_PHASIC = c(0, 0.01, 0.02, 0.01, 0, 0.03, 0.04, 0.02, 0.01, 0)
  )

  p <- plot_gazepoint_eda_decomposition(
    data = dat,
    time_col = "CNT",
    signal_cols = c("GSR_US", "GSR_US_TONIC", "GSR_US_PHASIC"),
    group_cols = "participant",
    standardise = TRUE,
    max_points = 100,
    title = "EDA decomposition test"
  )

  check <- check_gazepoint_plot_contract(p)

  expect_s3_class(p, "gazepoint_plot")
  expect_equal(check$overview$status, "pass_plot_contract")
  expect_true(is.data.frame(get_gazepoint_plot_data(p)))
  expect_equal(attr(p, "plot_type"), "eda_decomposition")
})

test_that("plot_gazepoint_eda_decomposition follows the plot contract", {
  skip_if_not_installed("ggplot2")

  dat <- data.frame(
    participant = rep("P1", 10),
    CNT = 0:9,
    GSR_US = seq(1, 2, length.out = 10),
    GSR_US_TONIC = seq(0.9, 1.8, length.out = 10),
    GSR_US_PHASIC = c(0, 0.01, 0.02, 0.01, 0, 0.03, 0.04, 0.02, 0.01, 0)
  )

  p <- plot_gazepoint_eda_decomposition(
    data = dat,
    time_col = "CNT",
    signal_cols = c("GSR_US", "GSR_US_TONIC", "GSR_US_PHASIC"),
    group_cols = "participant",
    standardise = TRUE,
    max_points = 100,
    title = "EDA decomposition test"
  )

  check <- check_gazepoint_plot_contract(p)

  expect_s3_class(p, "gazepoint_plot")
  expect_equal(check$overview$status, "pass_plot_contract")
  expect_true(is.data.frame(get_gazepoint_plot_data(p)))
  expect_true(is.list(get_gazepoint_plot_settings(p)))
  expect_equal(attr(p, "plot_type"), "eda_decomposition")
})

test_that("plot_gazepoint_scr_events follows the plot contract", {
  skip_if_not_installed("ggplot2")

  dat <- data.frame(
    participant = rep("P1", 20),
    CNT = 0:19,
    GSR_US = c(
      1.00, 1.01, 1.03, 1.08, 1.05,
      1.02, 1.01, 1.00, 1.02, 1.04,
      1.10, 1.08, 1.04, 1.02, 1.01,
      1.00, 1.01, 1.02, 1.01, 1.00
    )
  )

  peaks <- data.frame(
    participant = "P1",
    peak_time = c(3, 10),
    peak_amplitude = c(0.08, 0.10)
  )

  p <- plot_gazepoint_scr_events(
    data = dat,
    scr_peaks = peaks,
    time_col = "CNT",
    signal_col = "GSR_US",
    group_cols = "participant",
    max_points = 100,
    title = "SCR events test"
  )

  check <- check_gazepoint_plot_contract(p)

  expect_s3_class(p, "gazepoint_plot")
  expect_equal(check$overview$status, "pass_plot_contract")
  expect_true(is.data.frame(get_gazepoint_plot_data(p)))
  expect_true(is.list(get_gazepoint_plot_settings(p)))
  expect_equal(attr(p, "plot_type"), "scr_events")
  expect_true(is.data.frame(attr(p, "peak_data")))
  expect_true(is.data.frame(attr(p, "event_data")))
})

test_that("plot_gazepoint_multimodal_timeline follows the plot contract", {
  skip_if_not_installed("ggplot2")

  dat <- data.frame(
    participant = rep("P1", 10),
    MEDIA_ID = rep(1, 10),
    CNT = 0:9,
    GSR_US = seq(1, 2, length.out = 10),
    HR = seq(70, 80, length.out = 10),
    TTL0 = c(1, rep(0, 9))
  )

  p <- plot_gazepoint_multimodal_timeline(
    data = dat,
    time_col = "CNT",
    signal_cols = c("GSR_US", "HR"),
    group_cols = c("participant", "MEDIA_ID"),
    event_col = "TTL0",
    standardise = TRUE,
    show_event_markers = TRUE,
    title = "Multimodal timeline test"
  )

  check <- check_gazepoint_plot_contract(p)

  expect_s3_class(p, "gazepoint_plot")
  expect_equal(check$overview$status, "pass_plot_contract")
  expect_true(is.data.frame(get_gazepoint_plot_data(p)))
  expect_true(is.list(get_gazepoint_plot_settings(p)))
  expect_equal(attr(p, "plot_type"), "multimodal_timeline")
  expect_true(is.data.frame(attr(p, "event_data")))
})

#' Create a Gazepoint Biometrics report
#'
#' Creates a structured, manuscript-oriented report object for Gazepoint
#' Biometrics data, workflow outputs, quality checks, report tables, methods
#' text, and reporting checklists. The report is intentionally conservative:
#' GSR/EDA is described as electrodermal activity/arousal-related signal rather
#' than emotional valence; heart-rate interpretation is tied to baseline/task
#' context; eye-tracking is described as visual attention rather than direct
#' cognition; and raw `HRV` columns are not treated as HRV metrics.
#'
#' The function can also write a lightweight Markdown or HTML file without
#' adding heavy reporting dependencies.
#'
#' @param data Optional biometric data frame.
#' @param workflow Optional workflow object or workflow summary list.
#' @param validation Optional validation object or data frame.
#' @param quality Optional quality-audit object or data frame.
#' @param sampling Optional sampling-audit object or data frame.
#' @param missingness Optional missingness-audit object or data frame.
#' @param exclusions Optional exclusion-recommendation object or data frame.
#' @param report_tables Optional report-table object, data frame, or named list
#'   of data frames.
#' @param methods_text Optional methods text, character vector, or object
#'   returned by `create_gazepoint_biometrics_methods_text()`.
#' @param checklist Optional checklist object or data frame.
#' @param title Report title.
#' @param subtitle Optional report subtitle.
#' @param output_file Optional path to write a report file.
#' @param format Output format when `output_file` is supplied. Supported values
#'   are `"markdown"` and `"html"`.
#' @param include_timestamp Logical. Should a creation timestamp be included?
#' @param overwrite Logical. Should an existing `output_file` be overwritten?
#' @param max_table_rows Maximum number of rows shown per table in the written
#'   report.
#'
#' @return A list of class `"gazepoint_biometrics_report"` with `overview`,
#'   `sections`, `tables`, `objects`, `output_file`, and `settings`.
#'
#' @examples
#' df <- data.frame(
#'   CNT = 1:5,
#'   GSR = c(1, 1.1, 1.2, 1.1, 1),
#'   HR = c(70, 71, 72, 71, 70),
#'   DIAL = c(40, 42, 44, 43, 41)
#' )
#' report <- create_gazepoint_biometrics_report(df)
#' names(report)
#'
#' @export
create_gazepoint_biometrics_report <- function(data = NULL,
                                               workflow = NULL,
                                               validation = NULL,
                                               quality = NULL,
                                               sampling = NULL,
                                               missingness = NULL,
                                               exclusions = NULL,
                                               report_tables = NULL,
                                               methods_text = NULL,
                                               checklist = NULL,
                                               title = "Gazepoint Biometrics report",
                                               subtitle = NULL,
                                               output_file = NULL,
                                               format = c("markdown", "html"),
                                               include_timestamp = FALSE,
                                               overwrite = FALSE,
                                               max_table_rows = 20L) {
  format <- match.arg(format)

  if (!is.null(data) && !is.data.frame(data)) {
    stop("`data` must be NULL or a data frame.", call. = FALSE)
  }

  if (!is.character(title) || length(title) != 1L || is.na(title) ||
      !nzchar(title)) {
    stop("`title` must be a non-empty character string.", call. = FALSE)
  }

  if (!is.null(subtitle) &&
      (!is.character(subtitle) || length(subtitle) != 1L || is.na(subtitle))) {
    stop("`subtitle` must be NULL or a single character string.", call. = FALSE)
  }

  if (!is.null(output_file) &&
      (!is.character(output_file) || length(output_file) != 1L ||
       is.na(output_file) || !nzchar(output_file))) {
    stop("`output_file` must be NULL or a non-empty character string.",
         call. = FALSE)
  }

  if (!is.logical(include_timestamp) ||
      length(include_timestamp) != 1L ||
      is.na(include_timestamp)) {
    stop("`include_timestamp` must be TRUE or FALSE.", call. = FALSE)
  }

  if (!is.logical(overwrite) || length(overwrite) != 1L || is.na(overwrite)) {
    stop("`overwrite` must be TRUE or FALSE.", call. = FALSE)
  }

  .gpbiom_assert_positive_integer(max_table_rows, "max_table_rows")

  if (!is.null(output_file) && file.exists(output_file) && !isTRUE(overwrite)) {
    stop("`output_file` already exists. Use `overwrite = TRUE` to replace it.",
         call. = FALSE)
  }

  derived <- .gpbiom_report_derive_objects(
    data = data,
    workflow = workflow,
    validation = validation,
    quality = quality,
    sampling = sampling,
    missingness = missingness,
    exclusions = exclusions,
    report_tables = report_tables,
    methods_text = methods_text,
    checklist = checklist
  )

  overview <- .gpbiom_report_overview(
    data = data,
    workflow = workflow,
    derived = derived,
    title = title,
    subtitle = subtitle,
    include_timestamp = include_timestamp
  )

  tables <- .gpbiom_report_collect_tables(
    derived = derived,
    report_tables = report_tables
  )

  sections <- .gpbiom_report_sections(
    title = title,
    subtitle = subtitle,
    overview = overview,
    derived = derived,
    tables = tables,
    include_timestamp = include_timestamp,
    max_table_rows = max_table_rows
  )

  output_path <- NA_character_

  if (!is.null(output_file)) {
    text <- .gpbiom_report_render(
      sections = sections,
      format = format
    )

    output_dir <- dirname(output_file)

    if (!identical(output_dir, ".") && !dir.exists(output_dir)) {
      dir.create(output_dir, recursive = TRUE, showWarnings = FALSE)
    }

    writeLines(text, output_file, useBytes = TRUE)
    output_path <- normalizePath(output_file, winslash = "/", mustWork = FALSE)
  }

  out <- list(
    overview = overview,
    sections = sections,
    tables = tables,
    objects = derived,
    output_file = output_path,
    settings = list(
      title = title,
      subtitle = subtitle,
      format = format,
      include_timestamp = include_timestamp,
      overwrite = overwrite,
      max_table_rows = as.integer(max_table_rows),
      cautions = .gpbiom_report_cautions()
    )
  )

  class(out) <- c("gazepoint_biometrics_report", class(out))
  out
}


.gpbiom_report_derive_objects <- function(data,
                                          workflow,
                                          validation,
                                          quality,
                                          sampling,
                                          missingness,
                                          exclusions,
                                          report_tables,
                                          methods_text,
                                          checklist) {
  schema <- NULL
  validity <- NULL

  if (!is.null(data)) {
    schema <- tryCatch(
      detect_gazepoint_biometric_schema(data),
      error = function(e) NULL
    )

    validity <- tryCatch(
      summarise_gazepoint_biometric_validity(data),
      error = function(e) NULL
    )

    if (is.null(sampling)) {
      sampling <- tryCatch(
        audit_gazepoint_biometric_sampling(data),
        error = function(e) NULL
      )
    }

    if (is.null(missingness)) {
      missingness <- tryCatch(
        audit_gazepoint_biometric_missingness(data),
        error = function(e) NULL
      )
    }
  }

  list(
    schema = schema,
    validity = validity,
    workflow = workflow,
    validation = validation,
    quality = quality,
    sampling = sampling,
    missingness = missingness,
    exclusions = exclusions,
    report_tables = report_tables,
    methods_text = methods_text,
    checklist = checklist
  )
}


.gpbiom_report_overview <- function(data,
                                    workflow,
                                    derived,
                                    title,
                                    subtitle,
                                    include_timestamp) {
  schema_overview <- .gpbiom_first_data_frame(derived$schema, "overview")
  validity_overview <- .gpbiom_first_data_frame(derived$validity, "overview")
  workflow_overview <- .gpbiom_report_workflow_overview(workflow)

  n_rows <- if (!is.null(data)) nrow(data) else NA_integer_
  n_columns <- if (!is.null(data)) ncol(data) else NA_integer_

  if (!is.null(schema_overview) && "n_rows" %in% names(schema_overview)) {
    n_rows <- schema_overview$n_rows[1L]
  }

  if (!is.null(schema_overview) && "n_columns" %in% names(schema_overview)) {
    n_columns <- schema_overview$n_columns[1L]
  }

  active_signal_count <- NA_integer_

  if (!is.null(schema_overview) &&
      "active_signal_count" %in% names(schema_overview)) {
    active_signal_count <- schema_overview$active_signal_count[1L]
  } else if (!is.null(validity_overview) &&
             "active_signal_count" %in% names(validity_overview)) {
    active_signal_count <- validity_overview$active_signal_count[1L]
  }

  status <- .gpbiom_report_status(
    schema_overview = schema_overview,
    validity_overview = validity_overview,
    workflow_overview = workflow_overview
  )

  data.frame(
    title = title,
    subtitle = ifelse(is.null(subtitle), NA_character_, subtitle),
    created_at = if (isTRUE(include_timestamp)) {
      format(Sys.time(), "%Y-%m-%d %H:%M:%S %Z")
    } else {
      NA_character_
    },
    n_rows = n_rows,
    n_columns = n_columns,
    active_signal_count = active_signal_count,
    has_data = !is.null(data),
    has_workflow = !is.null(workflow),
    has_methods_text = !is.null(derived$methods_text),
    has_checklist = !is.null(derived$checklist),
    table_count = length(.gpbiom_report_extract_tables(derived$report_tables)),
    status = status,
    stringsAsFactors = FALSE
  )
}


.gpbiom_report_status <- function(schema_overview,
                                  validity_overview,
                                  workflow_overview) {
  statuses <- character()

  if (!is.null(schema_overview) && "status" %in% names(schema_overview)) {
    statuses <- c(statuses, schema_overview$status[1L])
  }

  if (!is.null(validity_overview) && "status" %in% names(validity_overview)) {
    statuses <- c(statuses, validity_overview$status[1L])
  }

  if (!is.null(workflow_overview) && "status" %in% names(workflow_overview)) {
    statuses <- c(statuses, workflow_overview$status[1L])
  }

  if (length(statuses) == 0L) {
    return("report_created_without_diagnostic_status")
  }

  if (any(grepl("fail|error|no_", statuses, ignore.case = TRUE))) {
    return("report_created_with_readiness_concerns")
  }

  if (any(grepl("warn|issue|limited|insufficient|inactive",
                statuses, ignore.case = TRUE))) {
    return("report_created_with_cautions")
  }

  "report_created"
}


.gpbiom_report_workflow_overview <- function(workflow) {
  if (is.null(workflow)) {
    return(NULL)
  }

  if (is.data.frame(workflow)) {
    return(workflow)
  }

  if (is.list(workflow)) {
    direct <- .gpbiom_first_data_frame(workflow, "overview")

    if (!is.null(direct)) {
      return(direct)
    }

    summary_obj <- tryCatch(
      summarise_gazepoint_biometrics_workflow(workflow),
      error = function(e) NULL
    )

    if (is.data.frame(summary_obj)) {
      return(summary_obj)
    }

    if (is.list(summary_obj)) {
      summary_overview <- .gpbiom_first_data_frame(summary_obj, "overview")

      if (!is.null(summary_overview)) {
        return(summary_overview)
      }
    }
  }

  data.frame(
    status = "workflow_object_supplied",
    stringsAsFactors = FALSE
  )
}


.gpbiom_report_collect_tables <- function(derived, report_tables) {
  tables <- list()

  schema_overview <- .gpbiom_first_data_frame(derived$schema, "overview")
  schema_columns <- .gpbiom_first_data_frame(derived$schema, "columns")
  validity_signals <- .gpbiom_first_data_frame(derived$validity, "signals")
  validity_flags <- .gpbiom_first_data_frame(derived$validity, "validity_flags")
  workflow_overview <- .gpbiom_report_workflow_overview(derived$workflow)

  if (!is.null(schema_overview)) {
    tables$schema_overview <- schema_overview
  }

  if (!is.null(schema_columns)) {
    tables$schema_columns <- schema_columns
  }

  if (!is.null(validity_signals)) {
    tables$signal_validity <- validity_signals
  }

  if (!is.null(validity_flags)) {
    tables$validity_flags <- validity_flags
  }

  if (!is.null(workflow_overview)) {
    tables$workflow_overview <- workflow_overview
  }

  user_tables <- .gpbiom_report_extract_tables(report_tables)

  for (name in names(user_tables)) {
    tables[[name]] <- user_tables[[name]]
  }

  tables
}


.gpbiom_report_extract_tables <- function(x) {
  if (is.null(x)) {
    return(list())
  }

  if (is.data.frame(x)) {
    return(list(report_table = x))
  }

  if (is.list(x)) {
    tables <- x[vapply(x, is.data.frame, logical(1))]

    if (length(tables) == 0L) {
      return(list())
    }

    if (is.null(names(tables)) || any(!nzchar(names(tables)))) {
      names(tables) <- paste0("table_", seq_along(tables))
    }

    return(tables)
  }

  list()
}


.gpbiom_first_data_frame <- function(x, name = NULL) {
  if (is.null(x)) {
    return(NULL)
  }

  if (is.data.frame(x)) {
    return(x)
  }

  if (is.list(x) && !is.null(name) && name %in% names(x) &&
      is.data.frame(x[[name]])) {
    return(x[[name]])
  }

  if (is.list(x)) {
    data_frames <- x[vapply(x, is.data.frame, logical(1))]

    if (length(data_frames) > 0L) {
      return(data_frames[[1L]])
    }
  }

  NULL
}


.gpbiom_report_sections <- function(title,
                                    subtitle,
                                    overview,
                                    derived,
                                    tables,
                                    include_timestamp,
                                    max_table_rows) {
  methods_text <- .gpbiom_report_text_from_object(derived$methods_text)
  checklist_text <- .gpbiom_report_checklist_text(derived$checklist)

  sections <- list(
    title = c(
      paste0("# ", title),
      if (!is.null(subtitle)) paste0("\n", subtitle) else character(),
      if (isTRUE(include_timestamp)) {
        paste0("\nCreated: ", overview$created_at[1L])
      } else {
        character()
      }
    ),
    overview = c(
      "## Overview",
      .gpbiom_report_table_to_markdown(overview, max_rows = max_table_rows)
    ),
    signal_summary = c(
      "## Signal availability and schema",
      .gpbiom_report_signal_summary_text(derived),
      .gpbiom_report_optional_table(
        tables$signal_validity,
        max_rows = max_table_rows
      )
    ),
    quality_readiness = c(
      "## Quality and readiness",
      .gpbiom_report_quality_text(derived),
      .gpbiom_report_optional_table(
        .gpbiom_report_quality_table(derived),
        max_rows = max_table_rows
      )
    ),
    methods = c(
      "## Methods text",
      if (length(methods_text) > 0L) {
        methods_text
      } else {
        "_No methods text object was supplied._"
      }
    ),
    checklist = c(
      "## Reporting checklist",
      if (length(checklist_text) > 0L) {
        checklist_text
      } else {
        "_No checklist object was supplied._"
      }
    ),
    tables = c(
      "## Report tables",
      .gpbiom_report_all_tables_text(tables, max_rows = max_table_rows)
    ),
    cautions = c(
      "## Interpretation cautions",
      paste0("- ", .gpbiom_report_cautions())
    )
  )

  sections
}


.gpbiom_report_signal_summary_text <- function(derived) {
  schema_overview <- .gpbiom_first_data_frame(derived$schema, "overview")
  validity_overview <- .gpbiom_first_data_frame(derived$validity, "overview")

  if (is.null(schema_overview) && is.null(validity_overview)) {
    return("_No schema or signal-validity summary was available._")
  }

  lines <- character()

  if (!is.null(schema_overview)) {
    if ("active_signal_count" %in% names(schema_overview)) {
      lines <- c(
        lines,
        paste0(
          "- Active signal groups detected: ",
          schema_overview$active_signal_count[1L],
          "."
        )
      )
    }

    flags <- c(
      gsr_eda = "has_gsr_eda",
      heart_rate = "has_heart_rate",
      ibi = "has_ibi",
      engagement_dial = "has_engagement_dial",
      ttl_marker = "has_ttl_marker"
    )

    present <- names(flags)[vapply(flags, function(column) {
      column %in% names(schema_overview) && isTRUE(schema_overview[[column]][1L])
    }, logical(1))]

    if (length(present) > 0L) {
      lines <- c(lines, paste0("- Present channels: ",
                               paste(present, collapse = ", "), "."))
    }
  }

  if (!is.null(validity_overview) && "status" %in% names(validity_overview)) {
    lines <- c(lines, paste0("- Validity status: ",
                             validity_overview$status[1L], "."))
  }

  if (length(lines) == 0L) {
    lines <- "_Schema object was supplied but did not contain standard overview fields._"
  }

  lines
}


.gpbiom_report_quality_text <- function(derived) {
  objects <- list(
    validation = derived$validation,
    quality = derived$quality,
    sampling = derived$sampling,
    missingness = derived$missingness,
    exclusions = derived$exclusions
  )

  present <- names(objects)[!vapply(objects, is.null, logical(1))]

  if (length(present) == 0L) {
    return("_No separate validation, quality, sampling, missingness, or exclusion object was supplied._")
  }

  paste0("- Supplied quality/readiness objects: ", paste(present, collapse = ", "), ".")
}


.gpbiom_report_quality_table <- function(derived) {
  objects <- list(
    validation = derived$validation,
    quality = derived$quality,
    sampling = derived$sampling,
    missingness = derived$missingness,
    exclusions = derived$exclusions
  )

  rows <- lapply(names(objects), function(name) {
    object <- objects[[name]]

    if (is.null(object)) {
      return(NULL)
    }

    table <- .gpbiom_first_data_frame(object, "overview")

    status <- NA_character_
    n_rows <- NA_integer_

    if (!is.null(table)) {
      n_rows <- nrow(table)

      if ("status" %in% names(table)) {
        status <- as.character(table$status[1L])
      } else if ("final_status" %in% names(table)) {
        status <- as.character(table$final_status[1L])
      }
    }

    data.frame(
      object = name,
      table_rows = n_rows,
      status = status,
      stringsAsFactors = FALSE
    )
  })

  rows <- Filter(Negate(is.null), rows)

  if (length(rows) == 0L) {
    return(NULL)
  }

  do.call(rbind, rows)
}


.gpbiom_report_text_from_object <- function(x) {
  if (is.null(x)) {
    return(character())
  }

  if (is.character(x)) {
    return(x)
  }

  if (is.data.frame(x)) {
    return(.gpbiom_report_table_to_markdown(x, max_rows = nrow(x)))
  }

  if (is.list(x)) {
    character_items <- unlist(x[vapply(x, is.character, logical(1))],
                              use.names = FALSE)

    if (length(character_items) > 0L) {
      return(character_items)
    }

    table <- .gpbiom_first_data_frame(x)

    if (!is.null(table)) {
      return(.gpbiom_report_table_to_markdown(table, max_rows = nrow(table)))
    }
  }

  as.character(utils::capture.output(utils::str(x)))
}


.gpbiom_report_checklist_text <- function(checklist) {
  if (is.null(checklist)) {
    return(character())
  }

  if (is.character(checklist)) {
    return(checklist)
  }

  table <- .gpbiom_first_data_frame(checklist, "checklist")

  if (is.null(table)) {
    table <- .gpbiom_first_data_frame(checklist)
  }

  if (!is.null(table)) {
    return(.gpbiom_report_table_to_markdown(table, max_rows = nrow(table)))
  }

  as.character(utils::capture.output(utils::str(checklist)))
}


.gpbiom_report_optional_table <- function(table, max_rows) {
  if (is.null(table) || !is.data.frame(table) || nrow(table) == 0L) {
    return("_No table available._")
  }

  .gpbiom_report_table_to_markdown(table, max_rows = max_rows)
}


.gpbiom_report_all_tables_text <- function(tables, max_rows) {
  if (length(tables) == 0L) {
    return("_No report tables were available._")
  }

  lines <- character()

  for (name in names(tables)) {
    lines <- c(
      lines,
      paste0("### ", name),
      .gpbiom_report_table_to_markdown(tables[[name]], max_rows = max_rows),
      ""
    )
  }

  lines
}


.gpbiom_report_table_to_markdown <- function(table, max_rows = 20L) {
  if (is.null(table) || !is.data.frame(table)) {
    return("_No table available._")
  }

  if (nrow(table) == 0L) {
    return("_Table has no rows._")
  }

  shown <- utils::head(table, max_rows)
  shown[] <- lapply(shown, .gpbiom_report_cell)

  header <- paste0("| ", paste(names(shown), collapse = " | "), " |")
  separator <- paste0("| ", paste(rep("---", ncol(shown)), collapse = " | "), " |")

  rows <- apply(shown, 1L, function(row) {
    paste0("| ", paste(row, collapse = " | "), " |")
  })

  footer <- if (nrow(table) > max_rows) {
    paste0("\n_Showing ", max_rows, " of ", nrow(table), " rows._")
  } else {
    character()
  }

  c(header, separator, rows, footer)
}


.gpbiom_report_cell <- function(x) {
  out <- as.character(x)
  out[is.na(out)] <- ""
  out <- gsub("\\|", "\\\\|", out)
  out <- gsub("\r?\n", " ", out)
  out
}


.gpbiom_report_render <- function(sections, format) {
  markdown <- unlist(sections, use.names = FALSE)
  markdown <- markdown[nzchar(markdown)]

  if (identical(format, "markdown")) {
    return(markdown)
  }

  escaped <- .gpbiom_html_escape(markdown)

  c(
    "<!doctype html>",
    "<html>",
    "<head>",
    "<meta charset=\"utf-8\">",
    "<title>Gazepoint Biometrics report</title>",
    "</head>",
    "<body>",
    "<pre>",
    escaped,
    "</pre>",
    "</body>",
    "</html>"
  )
}


.gpbiom_html_escape <- function(x) {
  x <- gsub("&", "&amp;", x, fixed = TRUE)
  x <- gsub("<", "&lt;", x, fixed = TRUE)
  x <- gsub(">", "&gt;", x, fixed = TRUE)
  x
}


.gpbiom_report_cautions <- function() {
  c(
    "GSR/EDA should be interpreted as electrodermal activity or arousal-related signal, not emotional valence.",
    "Heart-rate summaries require baseline, task, and artefact context before substantive interpretation.",
    "Raw HRV columns in Gazepoint exports should be treated as validity/vendor flags unless documentation proves otherwise.",
    "IBI-derived HRV-style metrics should be computed only from genuine IBI/RR interval columns.",
    "Eye-tracking measures indicate visual attention and do not directly establish deeper cognition, scrutiny, or evaluation."
  )
}

test_that("create_gazepoint_biometrics_report creates a report from data", {
  df <- data.frame(
    CNT = 1:5,
    GSR = c(1, 1.1, 1.2, 1.1, 1),
    HR = c(70, 71, 72, 71, 70),
    DIAL = c(40, 42, 44, 43, 41)
  )

  report <- create_gazepoint_biometrics_report(df)

  expect_s3_class(report, "gazepoint_biometrics_report")
  expect_s3_class(report$overview, "data.frame")
  expect_type(report$sections, "list")
  expect_type(report$tables, "list")
  expect_type(report$objects, "list")

  expect_equal(report$overview$n_rows, 5)
  expect_equal(report$overview$n_columns, 4)
  expect_true(report$overview$has_data)
  expect_true("schema_overview" %in% names(report$tables))
  expect_true("signal_validity" %in% names(report$tables))
})


test_that("create_gazepoint_biometrics_report can include supplied methods text and checklist", {
  df <- data.frame(
    CNT = 1:3,
    HR = c(70, 71, 72)
  )

  checklist <- data.frame(
    item = c("Sampling reported", "Artefact handling reported"),
    status = c("pass", "warn")
  )

  report <- create_gazepoint_biometrics_report(
    data = df,
    methods_text = "Biometric data were processed using gpbiometrics.",
    checklist = checklist
  )

  methods_section <- paste(report$sections$methods, collapse = "\n")
  checklist_section <- paste(report$sections$checklist, collapse = "\n")

  expect_match(methods_section, "gpbiometrics")
  expect_match(checklist_section, "Sampling reported")
  expect_match(checklist_section, "Artefact handling reported")
  expect_true(report$overview$has_methods_text)
  expect_true(report$overview$has_checklist)
})


test_that("create_gazepoint_biometrics_report collects user-supplied report tables", {
  df <- data.frame(
    CNT = 1:3,
    GSR = c(1, 2, 3)
  )

  extra_tables <- list(
    custom_summary = data.frame(
      metric = "mean_gsr",
      value = 2
    )
  )

  report <- create_gazepoint_biometrics_report(
    data = df,
    report_tables = extra_tables
  )

  expect_true("custom_summary" %in% names(report$tables))
  expect_equal(report$tables$custom_summary$metric, "mean_gsr")
})


test_that("create_gazepoint_biometrics_report supports workflow objects", {
  workflow <- list(
    overview = data.frame(
      n_rows = 10,
      status = "workflow_ok"
    )
  )

  report <- create_gazepoint_biometrics_report(workflow = workflow)

  expect_true(report$overview$has_workflow)
  expect_true("workflow_overview" %in% names(report$tables))
  expect_equal(report$tables$workflow_overview$status, "workflow_ok")
})


test_that("create_gazepoint_biometrics_report writes markdown output", {
  df <- data.frame(
    CNT = 1:3,
    HR = c(70, 71, 72)
  )

  path <- tempfile(fileext = ".md")

  report <- create_gazepoint_biometrics_report(
    data = df,
    output_file = path,
    format = "markdown"
  )

  expect_true(file.exists(path))
  text <- readLines(path, warn = FALSE)

  expect_true(any(grepl("# Gazepoint Biometrics report", text)))
  expect_true(any(grepl("Interpretation cautions", text)))
  expect_equal(report$output_file, normalizePath(path, winslash = "/", mustWork = FALSE))
})


test_that("create_gazepoint_biometrics_report writes html output", {
  df <- data.frame(
    CNT = 1:3,
    HR = c(70, 71, 72)
  )

  path <- tempfile(fileext = ".html")

  report <- create_gazepoint_biometrics_report(
    data = df,
    output_file = path,
    format = "html"
  )

  expect_true(file.exists(path))
  text <- readLines(path, warn = FALSE)

  expect_true(any(grepl("<!doctype html>", text, fixed = TRUE)))
  expect_true(any(grepl("<pre>", text, fixed = TRUE)))
  expect_equal(report$settings$format, "html")
})


test_that("create_gazepoint_biometrics_report protects existing files", {
  path <- tempfile(fileext = ".md")
  writeLines("existing", path)

  expect_error(
    create_gazepoint_biometrics_report(output_file = path),
    "already exists"
  )

  report <- create_gazepoint_biometrics_report(
    output_file = path,
    overwrite = TRUE
  )

  expect_true(file.exists(path))
  expect_s3_class(report, "gazepoint_biometrics_report")
})


test_that("create_gazepoint_biometrics_report validates arguments", {
  expect_error(
    create_gazepoint_biometrics_report(data = 1:3),
    "`data` must be"
  )

  expect_error(
    create_gazepoint_biometrics_report(title = ""),
    "`title`"
  )

  expect_error(
    create_gazepoint_biometrics_report(subtitle = 1),
    "`subtitle`"
  )

  expect_error(
    create_gazepoint_biometrics_report(output_file = ""),
    "`output_file`"
  )

  expect_error(
    create_gazepoint_biometrics_report(include_timestamp = NA),
    "`include_timestamp`"
  )

  expect_error(
    create_gazepoint_biometrics_report(overwrite = NA),
    "`overwrite`"
  )

  expect_error(
    create_gazepoint_biometrics_report(max_table_rows = 0),
    "`max_table_rows`"
  )
})


test_that("create_gazepoint_biometrics_report includes conservative interpretation cautions", {
  report <- create_gazepoint_biometrics_report()

  cautions <- paste(report$settings$cautions, collapse = " ")

  expect_match(cautions, "GSR/EDA")
  expect_match(cautions, "not emotional valence")
  expect_match(cautions, "Raw HRV")
  expect_match(cautions, "genuine IBI/RR")
  expect_match(cautions, "visual attention")
})

#' Prepare Gazepoint HRP/PPG waveform input for pyPPG
#'
#' Prepares a Gazepoint heart-rate pulse waveform column, usually `HRP`, as a
#' lightweight input table for optional pyPPG workflows. This helper does not
#' call Python, does not require pyPPG, and does not derive HRV features. It only
#' prepares waveform values, timing information when available, and conservative
#' group-level summaries for interoperability review.
#'
#' @param data A Gazepoint biometric data frame or a list containing one.
#' @param ppg_col Optional HRP/PPG waveform column. If `NULL`, common Gazepoint
#'   HRP/PPG column names are detected.
#' @param time_col Optional time, timestamp, or sample-counter column.
#' @param group_cols Optional grouping columns.
#' @param sampling_rate Optional sampling rate in Hz. Useful when `time_col`
#'   is a sample counter such as `CNT`.
#' @param time_unit Unit of `time_col`: `"auto"`, `"ms"`, `"seconds"`, or
#'   `"samples"`.
#' @param min_finite_prop Minimum finite waveform proportion used for group
#'   readiness summaries.
#' @param output_dir Optional directory for CSV export. If `NULL`, no files are
#'   written.
#' @param prefix File prefix used when `output_dir` is supplied.
#'
#' @return A list with `overview`, `waveform_table`, `group_summary`,
#'   `manifest`, and `settings`.
#' @export
prepare_gazepoint_pyppg_input <- function(data,
                                          ppg_col = NULL,
                                          time_col = NULL,
                                          group_cols = NULL,
                                          sampling_rate = NULL,
                                          time_unit = c("auto", "ms", "seconds", "samples"),
                                          min_finite_prop = 0.50,
                                          output_dir = NULL,
                                          prefix = "gazepoint_pyppg") {
  time_unit <- match.arg(time_unit)

  dat <- gpbiometrics_pyppg_extract_data(data)
  names_dat <- names(dat)

  if (!is.null(sampling_rate)) {
    gpbiometrics_pyppg_validate_positive_scalar(sampling_rate, "sampling_rate")
  }

  if (!is.numeric(min_finite_prop) ||
      length(min_finite_prop) != 1 ||
      is.na(min_finite_prop) ||
      min_finite_prop < 0 ||
      min_finite_prop > 1) {
    stop("`min_finite_prop` must be a single number between 0 and 1.", call. = FALSE)
  }

  if (is.null(ppg_col)) {
    ppg_col <- gpbiometrics_pyppg_first_existing(
      names_dat,
      c("HRP", "hrp", "PPG", "ppg", "pulse", "pulse_wave", "ppg_signal")
    )
  }

  if (is.null(ppg_col) || !ppg_col %in% names_dat) {
    stop("No usable HRP/PPG waveform column was found. Supply `ppg_col`.", call. = FALSE)
  }

  if (is.null(time_col)) {
    time_col <- gpbiometrics_pyppg_first_existing(
      names_dat,
      c(
        "time_ms", "timestamp_ms", "timestamp",
        "TIME", "Time", "time",
        "CNT", "cnt"
      )
    )
  }

  if (!is.null(time_col) && !time_col %in% names_dat) {
    stop("`time_col` was not found in `data`.", call. = FALSE)
  }

  group_cols <- gpbiometrics_pyppg_resolve_group_cols(names_dat, group_cols)

  missing_group_cols <- setdiff(group_cols, names_dat)

  if (length(missing_group_cols) > 0) {
    stop(
      "`group_cols` not found in `data`: ",
      paste(missing_group_cols, collapse = ", "),
      call. = FALSE
    )
  }

  dat$.gpbiometrics_group_id <- gpbiometrics_pyppg_group_id(dat, group_cols)
  dat$.gpbiometrics_row_id <- seq_len(nrow(dat))

  group_indices <- split(seq_len(nrow(dat)), dat$.gpbiometrics_group_id, drop = TRUE)

  waveform_parts <- list()
  summary_parts <- list()

  for (group_id in names(group_indices)) {
    idx <- group_indices[[group_id]]
    d <- dat[idx, , drop = FALSE]
    group_values <- gpbiometrics_pyppg_group_values(d, group_cols)

    ppg_value <- suppressWarnings(as.numeric(d[[ppg_col]]))

    timing <- gpbiometrics_pyppg_resolve_timing(
      d = d,
      time_col = time_col,
      sampling_rate = sampling_rate,
      time_unit = time_unit
    )

    waveform <- data.frame(
      group_id = group_id,
      sample_index = seq_len(nrow(d)),
      source_row = d$.gpbiometrics_row_id,
      time_raw = timing$time_raw,
      time_s = timing$time_s,
      ppg_signal = ppg_value,
      ppg_col = ppg_col,
      detected_time_unit = timing$detected_time_unit,
      stringsAsFactors = FALSE
    )

    waveform <- gpbiometrics_pyppg_prepend_group_values(group_values, waveform)
    waveform_parts[[length(waveform_parts) + 1L]] <- waveform

    finite_ppg <- ppg_value[is.finite(ppg_value)]
    finite_prop <- if (length(ppg_value) > 0) {
      length(finite_ppg) / length(ppg_value)
    } else {
      NA_real_
    }

    status <- if (length(ppg_value) == 0) {
      "empty_group"
    } else if (!is.finite(finite_prop) || finite_prop < min_finite_prop) {
      "insufficient_finite_waveform"
    } else if (all(is.na(timing$time_s))) {
      "prepared_with_sample_index_only"
    } else {
      "ready_for_pyppg_input"
    }

    summary_row <- data.frame(
      group_id = group_id,
      ppg_col = ppg_col,
      n_rows = nrow(d),
      n_finite = length(finite_ppg),
      finite_prop = finite_prop,
      ppg_min = if (length(finite_ppg)) min(finite_ppg) else NA_real_,
      ppg_median = if (length(finite_ppg)) stats::median(finite_ppg) else NA_real_,
      ppg_max = if (length(finite_ppg)) max(finite_ppg) else NA_real_,
      ppg_sd = if (length(finite_ppg) > 1) stats::sd(finite_ppg) else NA_real_,
      time_col = if (is.null(time_col)) NA_character_ else time_col,
      detected_time_unit = timing$detected_time_unit,
      time_span_s = timing$time_span_s,
      median_time_step_s = timing$median_time_step_s,
      estimated_sampling_rate_hz = timing$estimated_sampling_rate_hz,
      status = status,
      interpretation = gpbiometrics_pyppg_status_interpretation(status),
      stringsAsFactors = FALSE
    )

    summary_parts[[length(summary_parts) + 1L]] <-
      gpbiometrics_pyppg_prepend_group_values(group_values, summary_row)
  }

  waveform_table <- gpbiometrics_pyppg_rbind(waveform_parts)
  group_summary <- gpbiometrics_pyppg_rbind(summary_parts)

  manifest <- gpbiometrics_pyppg_write_outputs(
    waveform_table = waveform_table,
    group_summary = group_summary,
    output_dir = output_dir,
    prefix = prefix
  )

  ready_groups <- sum(group_summary$status %in% c(
    "ready_for_pyppg_input",
    "prepared_with_sample_index_only"
  ), na.rm = TRUE)

  overview_status <- if (nrow(dat) == 0) {
    "empty_input"
  } else if (ready_groups == nrow(group_summary)) {
    "pyppg_input_prepared"
  } else if (ready_groups > 0) {
    "partial_pyppg_input_prepared"
  } else {
    "pyppg_input_not_ready"
  }

  overview <- data.frame(
    input_rows = nrow(dat),
    group_count = nrow(group_summary),
    ready_group_count = as.integer(ready_groups),
    ppg_col = ppg_col,
    time_col = if (is.null(time_col)) NA_character_ else time_col,
    output_file_count = nrow(manifest),
    status = overview_status,
    interpretation = paste(
      "Output is prepared waveform input for optional pyPPG-style workflows.",
      "No Python dependency is invoked and no HRV or physiological diagnosis is inferred."
    ),
    stringsAsFactors = FALSE
  )

  structure(
    list(
      overview = overview,
      waveform_table = waveform_table,
      group_summary = group_summary,
      manifest = manifest,
      settings = list(
        ppg_col = ppg_col,
        time_col = time_col,
        group_cols = group_cols,
        sampling_rate = sampling_rate,
        time_unit = time_unit,
        min_finite_prop = min_finite_prop,
        output_dir = output_dir,
        prefix = prefix
      )
    ),
    class = c("gazepoint_pyppg_input", "list")
  )
}

#' Assess Gazepoint HRP waveform quality
#'
#' Computes descriptive quality-control summaries for a Gazepoint HRP/PPG
#' waveform column. The output is intended for waveform availability,
#' missingness, flatness, and timing-gap review. It does not infer diagnosis,
#' emotion, valence, cognition, preference, or true physiological state.
#'
#' @param data A Gazepoint biometric data frame or a list containing one.
#' @param hrp_col Optional HRP/PPG waveform column. If `NULL`, common column
#'   names are detected.
#' @param time_col Optional time, timestamp, or sample-counter column.
#' @param group_cols Optional grouping columns.
#' @param sampling_rate Optional sampling rate in Hz.
#' @param time_unit Unit of `time_col`: `"auto"`, `"ms"`, `"seconds"`, or
#'   `"samples"`.
#' @param min_rows Minimum rows required per group.
#' @param min_finite_prop Minimum finite waveform proportion required per group.
#' @param max_flat_prop Maximum allowed proportion of near-zero consecutive
#'   differences among finite waveform values.
#' @param flat_tolerance Absolute difference threshold used to identify near-flat
#'   consecutive waveform changes.
#' @param max_gap_multiplier Time gaps larger than this multiple of the median
#'   positive time step are flagged.
#'
#' @return A list with `overview`, `group_quality`, `row_flags`, and `settings`.
#' @export
assess_gazepoint_hrp_waveform_quality <- function(data,
                                                  hrp_col = NULL,
                                                  time_col = NULL,
                                                  group_cols = NULL,
                                                  sampling_rate = NULL,
                                                  time_unit = c("auto", "ms", "seconds", "samples"),
                                                  min_rows = 20,
                                                  min_finite_prop = 0.80,
                                                  max_flat_prop = 0.95,
                                                  flat_tolerance = 1e-8,
                                                  max_gap_multiplier = 3) {
  time_unit <- match.arg(time_unit)

  dat <- gpbiometrics_pyppg_extract_data(data)
  names_dat <- names(dat)

  gpbiometrics_pyppg_validate_positive_scalar(min_rows, "min_rows")
  gpbiometrics_pyppg_validate_proportion(min_finite_prop, "min_finite_prop")
  gpbiometrics_pyppg_validate_proportion(max_flat_prop, "max_flat_prop")
  gpbiometrics_pyppg_validate_nonnegative_scalar(flat_tolerance, "flat_tolerance")
  gpbiometrics_pyppg_validate_positive_scalar(max_gap_multiplier, "max_gap_multiplier")

  if (!is.null(sampling_rate)) {
    gpbiometrics_pyppg_validate_positive_scalar(sampling_rate, "sampling_rate")
  }

  if (is.null(hrp_col)) {
    hrp_col <- gpbiometrics_pyppg_first_existing(
      names_dat,
      c("HRP", "hrp", "PPG", "ppg", "pulse", "pulse_wave", "ppg_signal")
    )
  }

  if (is.null(hrp_col) || !hrp_col %in% names_dat) {
    stop("No usable HRP/PPG waveform column was found. Supply `hrp_col`.", call. = FALSE)
  }

  if (is.null(time_col)) {
    time_col <- gpbiometrics_pyppg_first_existing(
      names_dat,
      c(
        "time_ms", "timestamp_ms", "timestamp",
        "TIME", "Time", "time",
        "CNT", "cnt"
      )
    )
  }

  if (!is.null(time_col) && !time_col %in% names_dat) {
    stop("`time_col` was not found in `data`.", call. = FALSE)
  }

  group_cols <- gpbiometrics_pyppg_resolve_group_cols(names_dat, group_cols)

  missing_group_cols <- setdiff(group_cols, names_dat)

  if (length(missing_group_cols) > 0) {
    stop(
      "`group_cols` not found in `data`: ",
      paste(missing_group_cols, collapse = ", "),
      call. = FALSE
    )
  }

  dat$.gpbiometrics_group_id <- gpbiometrics_pyppg_group_id(dat, group_cols)
  dat$.gpbiometrics_row_id <- seq_len(nrow(dat))

  group_indices <- split(seq_len(nrow(dat)), dat$.gpbiometrics_group_id, drop = TRUE)

  quality_parts <- list()
  flag_parts <- list()

  for (group_id in names(group_indices)) {
    idx <- group_indices[[group_id]]
    d <- dat[idx, , drop = FALSE]
    group_values <- gpbiometrics_pyppg_group_values(d, group_cols)

    hrp_value <- suppressWarnings(as.numeric(d[[hrp_col]]))

    timing <- gpbiometrics_pyppg_resolve_timing(
      d = d,
      time_col = time_col,
      sampling_rate = sampling_rate,
      time_unit = time_unit
    )

    finite_hrp <- hrp_value[is.finite(hrp_value)]
    finite_prop <- if (length(hrp_value) > 0) {
      length(finite_hrp) / length(hrp_value)
    } else {
      NA_real_
    }

    diffs <- diff(finite_hrp)
    zero_diff_prop <- if (length(diffs) > 0) {
      mean(abs(diffs) <= flat_tolerance)
    } else {
      NA_real_
    }

    missing_flags <- !is.finite(hrp_value)

    time_gap_flags <- gpbiometrics_pyppg_large_gap_flags(
      time_s = timing$time_s,
      max_gap_multiplier = max_gap_multiplier
    )

    longest_missing_run <- gpbiometrics_pyppg_longest_true_run(missing_flags)

    status <- gpbiometrics_pyppg_quality_status(
      n_rows = nrow(d),
      finite_prop = finite_prop,
      n_unique_finite = length(unique(finite_hrp)),
      zero_diff_prop = zero_diff_prop,
      n_large_time_gaps = sum(time_gap_flags, na.rm = TRUE),
      min_rows = min_rows,
      min_finite_prop = min_finite_prop,
      max_flat_prop = max_flat_prop
    )

    quality_row <- data.frame(
      group_id = group_id,
      hrp_col = hrp_col,
      n_rows = nrow(d),
      n_finite = length(finite_hrp),
      finite_prop = finite_prop,
      missing_prop = if (length(hrp_value) > 0) mean(!is.finite(hrp_value)) else NA_real_,
      n_unique_finite = length(unique(finite_hrp)),
      hrp_min = if (length(finite_hrp)) min(finite_hrp) else NA_real_,
      hrp_median = if (length(finite_hrp)) stats::median(finite_hrp) else NA_real_,
      hrp_max = if (length(finite_hrp)) max(finite_hrp) else NA_real_,
      hrp_sd = if (length(finite_hrp) > 1) stats::sd(finite_hrp) else NA_real_,
      hrp_iqr = if (length(finite_hrp) > 1) stats::IQR(finite_hrp) else NA_real_,
      median_abs_diff = if (length(diffs) > 0) stats::median(abs(diffs)) else NA_real_,
      zero_diff_prop = zero_diff_prop,
      longest_missing_run = as.integer(longest_missing_run),
      n_large_time_gaps = as.integer(sum(time_gap_flags, na.rm = TRUE)),
      time_col = if (is.null(time_col)) NA_character_ else time_col,
      detected_time_unit = timing$detected_time_unit,
      median_time_step_s = timing$median_time_step_s,
      estimated_sampling_rate_hz = timing$estimated_sampling_rate_hz,
      status = status,
      interpretation = gpbiometrics_pyppg_quality_interpretation(status),
      stringsAsFactors = FALSE
    )

    quality_parts[[length(quality_parts) + 1L]] <-
      gpbiometrics_pyppg_prepend_group_values(group_values, quality_row)

    row_flags <- data.frame(
      group_id = group_id,
      source_row = d$.gpbiometrics_row_id,
      sample_index = seq_len(nrow(d)),
      hrp_value = hrp_value,
      time_s = timing$time_s,
      flag_missing_or_nonfinite_hrp = missing_flags,
      flag_large_time_gap = time_gap_flags,
      stringsAsFactors = FALSE
    )

    flag_parts[[length(flag_parts) + 1L]] <-
      gpbiometrics_pyppg_prepend_group_values(group_values, row_flags)
  }

  group_quality <- gpbiometrics_pyppg_rbind(quality_parts)
  row_flags <- gpbiometrics_pyppg_rbind(flag_parts)

  fail_count <- sum(grepl("^fail_", group_quality$status), na.rm = TRUE)
  review_count <- sum(grepl("^review_", group_quality$status), na.rm = TRUE)
  pass_count <- sum(group_quality$status == "descriptive_quality_pass", na.rm = TRUE)

  final_status <- if (nrow(dat) == 0) {
    "empty_input"
  } else if (fail_count > 0) {
    "fail_review_required"
  } else if (review_count > 0) {
    "review_recommended"
  } else {
    "pass"
  }

  overview <- data.frame(
    input_rows = nrow(dat),
    group_count = nrow(group_quality),
    pass_group_count = as.integer(pass_count),
    review_group_count = as.integer(review_count),
    fail_group_count = as.integer(fail_count),
    hrp_col = hrp_col,
    time_col = if (is.null(time_col)) NA_character_ else time_col,
    status = final_status,
    interpretation = paste(
      "HRP waveform quality is summarized descriptively for QC.",
      "The helper does not infer diagnosis, emotion, valence, cognition, trust, preference, or true physiological state."
    ),
    stringsAsFactors = FALSE
  )

  structure(
    list(
      overview = overview,
      group_quality = group_quality,
      row_flags = row_flags,
      settings = list(
        hrp_col = hrp_col,
        time_col = time_col,
        group_cols = group_cols,
        sampling_rate = sampling_rate,
        time_unit = time_unit,
        min_rows = min_rows,
        min_finite_prop = min_finite_prop,
        max_flat_prop = max_flat_prop,
        flat_tolerance = flat_tolerance,
        max_gap_multiplier = max_gap_multiplier
      )
    ),
    class = c("gazepoint_hrp_waveform_quality", "list")
  )
}

gpbiometrics_pyppg_extract_data <- function(data) {
  if (is.data.frame(data)) {
    return(as.data.frame(data, stringsAsFactors = FALSE))
  }

  if (is.list(data)) {
    candidates <- c(
      "data",
      "biometrics",
      "merged_data",
      "all_data",
      "imported_data",
      "raw_data",
      "combined_data"
    )

    for (nm in candidates) {
      if (!is.null(data[[nm]]) && is.data.frame(data[[nm]])) {
        return(as.data.frame(data[[nm]], stringsAsFactors = FALSE))
      }
    }

    data_frame_items <- vapply(data, is.data.frame, logical(1))

    if (any(data_frame_items)) {
      return(as.data.frame(data[[which(data_frame_items)[1]]], stringsAsFactors = FALSE))
    }
  }

  stop("`data` must be a data frame or a list containing a data frame.", call. = FALSE)
}

gpbiometrics_pyppg_first_existing <- function(names_dat, candidates) {
  exact <- candidates[candidates %in% names_dat]

  if (length(exact) > 0) {
    return(exact[1])
  }

  lower_names <- tolower(names_dat)
  lower_candidates <- tolower(candidates)
  idx <- match(lower_candidates, lower_names)
  idx <- idx[!is.na(idx)]

  if (length(idx) > 0) {
    return(names_dat[idx[1]])
  }

  NULL
}

gpbiometrics_pyppg_resolve_group_cols <- function(names_dat, group_cols) {
  if (!is.null(group_cols)) {
    group_cols <- as.character(group_cols)
    group_cols <- group_cols[!is.na(group_cols) & nzchar(group_cols)]
    return(unique(group_cols))
  }

  candidates <- c(
    "source_file",
    "source_participant",
    "participant",
    "subject",
    "subject_id",
    "USER",
    "USER_FILE",
    "MEDIA_ID",
    "MEDIA_NAME",
    "stimulus",
    "stimulus_id",
    "trial",
    "trial_id",
    "trial_global"
  )

  unique(candidates[candidates %in% names_dat])
}

gpbiometrics_pyppg_group_id <- function(dat, group_cols) {
  if (length(group_cols) == 0) {
    return(rep("all", nrow(dat)))
  }

  group_dat <- dat[group_cols]

  group_dat[] <- lapply(group_dat, function(x) {
    x_chr <- as.character(x)
    x_chr[is.na(x_chr)] <- "<NA>"
    x_chr
  })

  apply(group_dat, 1, paste, collapse = "||")
}

gpbiometrics_pyppg_group_values <- function(dat, group_cols) {
  if (length(group_cols) == 0 || nrow(dat) == 0) {
    return(NULL)
  }

  out <- as.data.frame(dat[1, group_cols, drop = FALSE], stringsAsFactors = FALSE)
  rownames(out) <- NULL
  out
}

gpbiometrics_pyppg_prepend_group_values <- function(group_values, row) {
  row <- as.data.frame(row, stringsAsFactors = FALSE)
  rownames(row) <- NULL

  if (is.null(group_values) || ncol(group_values) == 0) {
    return(row)
  }

  group_values <- as.data.frame(group_values, stringsAsFactors = FALSE)
  rownames(group_values) <- NULL

  group_values <- group_values[rep(1, nrow(row)), , drop = FALSE]
  rownames(group_values) <- NULL

  cbind(group_values, row, stringsAsFactors = FALSE)
}

gpbiometrics_pyppg_rbind <- function(x) {
  x <- x[!vapply(x, is.null, logical(1))]

  if (length(x) == 0) {
    return(data.frame(stringsAsFactors = FALSE))
  }

  out <- do.call(rbind, x)
  rownames(out) <- NULL
  out
}

gpbiometrics_pyppg_resolve_timing <- function(d,
                                              time_col = NULL,
                                              sampling_rate = NULL,
                                              time_unit = "auto") {
  if (is.null(time_col)) {
    time_raw <- seq_len(nrow(d))
    detected_time_unit <- if (is.null(sampling_rate)) {
      "sample_index"
    } else {
      "samples"
    }
  } else {
    time_raw <- suppressWarnings(as.numeric(d[[time_col]]))
    detected_time_unit <- gpbiometrics_pyppg_detect_time_unit(
      time_raw = time_raw,
      time_col = time_col,
      sampling_rate = sampling_rate,
      time_unit = time_unit
    )
  }

  finite_time <- time_raw[is.finite(time_raw)]

  time_s <- rep(NA_real_, length(time_raw))

  if (length(finite_time) > 0) {
    time_origin <- min(finite_time)

    if (identical(detected_time_unit, "ms")) {
      time_s <- (time_raw - time_origin) / 1000
    } else if (identical(detected_time_unit, "seconds")) {
      time_s <- time_raw - time_origin
    } else if (identical(detected_time_unit, "samples") && !is.null(sampling_rate)) {
      time_s <- (time_raw - time_origin) / sampling_rate
    }
  }

  finite_time_s <- time_s[is.finite(time_s)]
  time_span_s <- if (length(finite_time_s) > 1) {
    max(finite_time_s) - min(finite_time_s)
  } else {
    NA_real_
  }

  positive_steps <- diff(sort(unique(finite_time_s)))
  positive_steps <- positive_steps[is.finite(positive_steps) & positive_steps > 0]

  median_time_step_s <- if (length(positive_steps) > 0) {
    stats::median(positive_steps)
  } else {
    NA_real_
  }

  estimated_sampling_rate_hz <- if (!is.null(sampling_rate)) {
    sampling_rate
  } else if (is.finite(median_time_step_s) && median_time_step_s > 0) {
    1 / median_time_step_s
  } else {
    NA_real_
  }

  list(
    time_raw = time_raw,
    time_s = time_s,
    detected_time_unit = detected_time_unit,
    time_span_s = time_span_s,
    median_time_step_s = median_time_step_s,
    estimated_sampling_rate_hz = estimated_sampling_rate_hz
  )
}

gpbiometrics_pyppg_detect_time_unit <- function(time_raw,
                                                time_col = NULL,
                                                sampling_rate = NULL,
                                                time_unit = "auto") {
  if (!identical(time_unit, "auto")) {
    return(time_unit)
  }

  lower_col <- tolower(if (is.null(time_col)) "" else time_col)

  if (grepl("cnt|sample", lower_col)) {
    return("samples")
  }

  if (grepl("ms|millisecond", lower_col)) {
    return("ms")
  }

  if (grepl("sec|second", lower_col)) {
    return("seconds")
  }

  finite_time <- sort(unique(time_raw[is.finite(time_raw)]))

  if (length(finite_time) < 2) {
    return(if (is.null(sampling_rate)) "seconds" else "samples")
  }

  positive_steps <- diff(finite_time)
  positive_steps <- positive_steps[is.finite(positive_steps) & positive_steps > 0]

  if (length(positive_steps) == 0) {
    return(if (is.null(sampling_rate)) "seconds" else "samples")
  }

  median_step <- stats::median(positive_steps)

  if (median_step > 10) {
    "ms"
  } else {
    "seconds"
  }
}

gpbiometrics_pyppg_status_interpretation <- function(status) {
  switch(
    status,
    empty_group = "No rows were available for this group.",
    insufficient_finite_waveform = "The waveform has too few finite values for reliable pyPPG input preparation.",
    prepared_with_sample_index_only = paste(
      "A waveform table was prepared, but time in seconds could not be derived.",
      "Supply `sampling_rate` for sample-counter based pyPPG workflows."
    ),
    ready_for_pyppg_input = paste(
      "Waveform and timing columns were prepared for optional pyPPG-style input.",
      "External pyPPG processing remains optional and is not run by gpbiometrics."
    ),
    "Waveform input status was not recognized."
  )
}

gpbiometrics_pyppg_write_outputs <- function(waveform_table,
                                             group_summary,
                                             output_dir = NULL,
                                             prefix = "gazepoint_pyppg") {
  manifest <- data.frame(
    item = character(),
    path = character(),
    rows = integer(),
    status = character(),
    stringsAsFactors = FALSE
  )

  if (is.null(output_dir)) {
    return(manifest)
  }

  if (!is.character(output_dir) || length(output_dir) != 1 || is.na(output_dir) || !nzchar(output_dir)) {
    stop("`output_dir` must be NULL or a single non-empty character value.", call. = FALSE)
  }

  if (!is.character(prefix) || length(prefix) != 1 || is.na(prefix) || !nzchar(prefix)) {
    stop("`prefix` must be a single non-empty character value.", call. = FALSE)
  }

  dir.create(output_dir, recursive = TRUE, showWarnings = FALSE)

  waveform_path <- file.path(output_dir, paste0(prefix, "_waveform_table.csv"))
  summary_path <- file.path(output_dir, paste0(prefix, "_group_summary.csv"))

  utils::write.csv(waveform_table, waveform_path, row.names = FALSE)
  utils::write.csv(group_summary, summary_path, row.names = FALSE)

  manifest <- rbind(
    manifest,
    data.frame(
      item = "waveform_table",
      path = normalizePath(waveform_path, winslash = "/", mustWork = FALSE),
      rows = nrow(waveform_table),
      status = "written",
      stringsAsFactors = FALSE
    ),
    data.frame(
      item = "group_summary",
      path = normalizePath(summary_path, winslash = "/", mustWork = FALSE),
      rows = nrow(group_summary),
      status = "written",
      stringsAsFactors = FALSE
    )
  )

  rownames(manifest) <- NULL
  manifest
}

gpbiometrics_pyppg_large_gap_flags <- function(time_s, max_gap_multiplier = 3) {
  out <- rep(FALSE, length(time_s))

  finite_time <- time_s[is.finite(time_s)]

  if (length(finite_time) < 3) {
    return(out)
  }

  steps <- diff(time_s)
  positive_steps <- steps[is.finite(steps) & steps > 0]

  if (length(positive_steps) == 0) {
    return(out)
  }

  median_step <- stats::median(positive_steps)

  if (!is.finite(median_step) || median_step <= 0) {
    return(out)
  }

  gap <- c(FALSE, steps > max_gap_multiplier * median_step)
  gap[is.na(gap)] <- FALSE
  gap
}

gpbiometrics_pyppg_longest_true_run <- function(x) {
  x <- as.logical(x)
  x[is.na(x)] <- FALSE

  if (length(x) == 0 || !any(x)) {
    return(0L)
  }

  r <- rle(x)
  as.integer(max(r$lengths[r$values]))
}

gpbiometrics_pyppg_quality_status <- function(n_rows,
                                              finite_prop,
                                              n_unique_finite,
                                              zero_diff_prop,
                                              n_large_time_gaps,
                                              min_rows,
                                              min_finite_prop,
                                              max_flat_prop) {
  if (n_rows < min_rows) {
    return("fail_insufficient_rows")
  }

  if (!is.finite(finite_prop) || finite_prop < min_finite_prop) {
    return("fail_low_finite_signal")
  }

  if (n_unique_finite < 3) {
    return("review_flat_signal")
  }

  if (is.finite(zero_diff_prop) && zero_diff_prop > max_flat_prop) {
    return("review_low_variability")
  }

  if (n_large_time_gaps > 0) {
    return("review_time_gaps")
  }

  "descriptive_quality_pass"
}

gpbiometrics_pyppg_quality_interpretation <- function(status) {
  switch(
    status,
    fail_insufficient_rows = "Too few rows were available for descriptive HRP waveform QC.",
    fail_low_finite_signal = "The HRP waveform has too many missing or nonfinite values for descriptive QC.",
    review_flat_signal = "The HRP waveform has very low uniqueness and should be manually reviewed.",
    review_low_variability = "The HRP waveform has a high proportion of near-flat consecutive changes.",
    review_time_gaps = "The timing trace contains large gaps relative to the local sampling interval.",
    descriptive_quality_pass = paste(
      "Basic descriptive HRP waveform checks passed.",
      "This is a QC summary only and does not establish physiological validity."
    ),
    "HRP waveform quality status was not recognized."
  )
}

gpbiometrics_pyppg_validate_positive_scalar <- function(x, arg) {
  if (!is.numeric(x) || length(x) != 1 || is.na(x) || x <= 0) {
    stop("`", arg, "` must be a single positive number.", call. = FALSE)
  }

  invisible(TRUE)
}

gpbiometrics_pyppg_validate_nonnegative_scalar <- function(x, arg) {
  if (!is.numeric(x) || length(x) != 1 || is.na(x) || x < 0) {
    stop("`", arg, "` must be a single non-negative number.", call. = FALSE)
  }

  invisible(TRUE)
}

gpbiometrics_pyppg_validate_proportion <- function(x, arg) {
  if (!is.numeric(x) || length(x) != 1 || is.na(x) || x < 0 || x > 1) {
    stop("`", arg, "` must be a single number between 0 and 1.", call. = FALSE)
  }

  invisible(TRUE)
}

test_that("prepare_gazepoint_pyppg_input prepares HRP waveform tables", {
  dat <- data.frame(
    participant = rep(c("p1", "p2"), each = 30),
    CNT = rep(1:30, times = 2),
    HRP = c(
      sin(seq(0, 2 * pi, length.out = 30)),
      cos(seq(0, 2 * pi, length.out = 30))
    )
  )

  out <- prepare_gazepoint_pyppg_input(
    dat,
    ppg_col = "HRP",
    time_col = "CNT",
    group_cols = "participant",
    sampling_rate = 60
  )

  expect_s3_class(out, "gazepoint_pyppg_input")
  expect_equal(out$overview$status, "pyppg_input_prepared")
  expect_equal(out$overview$group_count, 2)
  expect_equal(nrow(out$waveform_table), nrow(dat))
  expect_true(all(c("sample_index", "time_s", "ppg_signal") %in% names(out$waveform_table)))
  expect_true(all(is.finite(out$waveform_table$time_s)))
  expect_equal(unique(out$group_summary$status), "ready_for_pyppg_input")
})

test_that("prepare_gazepoint_pyppg_input can prepare sample-index-only output", {
  dat <- data.frame(
    participant = "p1",
    HRP = sin(seq(0, 2 * pi, length.out = 20))
  )

  out <- prepare_gazepoint_pyppg_input(
    dat,
    ppg_col = "HRP",
    group_cols = "participant"
  )

  expect_s3_class(out, "gazepoint_pyppg_input")
  expect_equal(out$overview$status, "pyppg_input_prepared")
  expect_equal(out$group_summary$status, "prepared_with_sample_index_only")
  expect_true(all(is.na(out$waveform_table$time_s)))
})

test_that("prepare_gazepoint_pyppg_input writes optional CSV outputs", {
  dat <- data.frame(
    participant = "p1",
    CNT = 1:20,
    HRP = sin(seq(0, 2 * pi, length.out = 20))
  )

  output_dir <- tempfile("pyppg_export_")

  out <- prepare_gazepoint_pyppg_input(
    dat,
    ppg_col = "HRP",
    time_col = "CNT",
    group_cols = "participant",
    sampling_rate = 60,
    output_dir = output_dir,
    prefix = "test_pyppg"
  )

  expect_equal(nrow(out$manifest), 2)
  expect_true(all(file.exists(out$manifest$path)))
  expect_true(any(grepl("waveform_table", out$manifest$item)))
  expect_true(any(grepl("group_summary", out$manifest$item)))
})

test_that("prepare_gazepoint_pyppg_input errors when no HRP or PPG column is available", {
  dat <- data.frame(
    time_ms = 1:10,
    HR = 70 + seq_len(10)
  )

  expect_error(
    prepare_gazepoint_pyppg_input(dat),
    "No usable HRP/PPG waveform column"
  )
})

test_that("assess_gazepoint_hrp_waveform_quality passes variable waveform data", {
  dat <- data.frame(
    participant = "p1",
    time_ms = seq(0, by = 10, length.out = 80),
    HRP = sin(seq(0, 6 * pi, length.out = 80))
  )

  out <- assess_gazepoint_hrp_waveform_quality(
    dat,
    hrp_col = "HRP",
    time_col = "time_ms",
    group_cols = "participant",
    min_rows = 20
  )

  expect_s3_class(out, "gazepoint_hrp_waveform_quality")
  expect_equal(out$overview$status, "pass")
  expect_equal(out$group_quality$status, "descriptive_quality_pass")
  expect_equal(nrow(out$row_flags), nrow(dat))
  expect_true(all(c(
    "flag_missing_or_nonfinite_hrp",
    "flag_large_time_gap"
  ) %in% names(out$row_flags)))
})

test_that("assess_gazepoint_hrp_waveform_quality flags flat waveform data", {
  dat <- data.frame(
    participant = "p1",
    time_ms = seq(0, by = 10, length.out = 50),
    HRP = rep(1, 50)
  )

  out <- assess_gazepoint_hrp_waveform_quality(
    dat,
    hrp_col = "HRP",
    time_col = "time_ms",
    group_cols = "participant",
    min_rows = 20
  )

  expect_s3_class(out, "gazepoint_hrp_waveform_quality")
  expect_equal(out$overview$status, "review_recommended")
  expect_equal(out$group_quality$status, "review_flat_signal")
})

test_that("assess_gazepoint_hrp_waveform_quality flags low finite signal", {
  dat <- data.frame(
    participant = "p1",
    time_ms = seq(0, by = 10, length.out = 50),
    HRP = c(rep(NA_real_, 30), sin(seq(0, 2 * pi, length.out = 20)))
  )

  out <- assess_gazepoint_hrp_waveform_quality(
    dat,
    hrp_col = "HRP",
    time_col = "time_ms",
    group_cols = "participant",
    min_rows = 20,
    min_finite_prop = 0.80
  )

  expect_equal(out$overview$status, "fail_review_required")
  expect_equal(out$group_quality$status, "fail_low_finite_signal")
  expect_true(any(out$row_flags$flag_missing_or_nonfinite_hrp))
})

test_that("assess_gazepoint_hrp_waveform_quality flags large time gaps", {
  dat <- data.frame(
    participant = "p1",
    time_ms = c(seq(0, by = 10, length.out = 30), seq(1000, by = 10, length.out = 30)),
    HRP = sin(seq(0, 4 * pi, length.out = 60))
  )

  out <- assess_gazepoint_hrp_waveform_quality(
    dat,
    hrp_col = "HRP",
    time_col = "time_ms",
    group_cols = "participant",
    min_rows = 20
  )

  expect_equal(out$overview$status, "review_recommended")
  expect_equal(out$group_quality$status, "review_time_gaps")
  expect_true(any(out$row_flags$flag_large_time_gap))
})

#' Simulate Gazepoint-style biometric signals
#'
#' Generates synthetic Gazepoint-style EDA, PPG, HR, IBI, and TTL-like signals
#' with known ground-truth events. This is intended for teaching, examples, unit
#' tests, and model-validation workflows. It does not generate real participant
#' physiology.
#'
#' @param n_seconds Duration in seconds.
#' @param sampling_rate Sampling rate in Hz.
#' @param participant_id Participant identifier.
#' @param scr_onsets Optional SCR onset times in seconds.
#' @param scr_rate_per_min Expected SCR events per minute when `scr_onsets` is
#'   `NULL`.
#' @param pulse_rate_bpm Mean pulse rate.
#' @param respiration_rate_bpm Respiration-modulation rate.
#' @param eda_noise_sd EDA noise SD.
#' @param ppg_noise_sd PPG noise SD.
#' @param include_ttl Logical. If `TRUE`, adds TTL0 pulses at SCR onsets.
#' @param seed Optional random seed.
#'
#' @return A list with `overview`, `data`, `ground_truth`, and `settings`.
#' @export
simulate_gazepoint_biometrics <- function(n_seconds = 120,
                                          sampling_rate = 60,
                                          participant_id = "sim_p1",
                                          scr_onsets = NULL,
                                          scr_rate_per_min = 4,
                                          pulse_rate_bpm = 72,
                                          respiration_rate_bpm = 15,
                                          eda_noise_sd = 0.01,
                                          ppg_noise_sd = 0.02,
                                          include_ttl = TRUE,
                                          seed = NULL) {
  if (!is.null(seed)) {
    set.seed(seed)
  }

  if (!is.numeric(n_seconds) ||
      length(n_seconds) != 1 ||
      !is.finite(n_seconds) ||
      n_seconds <= 0) {
    stop("`n_seconds` must be a positive number.", call. = FALSE)
  }

  if (!is.numeric(sampling_rate) ||
      length(sampling_rate) != 1 ||
      !is.finite(sampling_rate) ||
      sampling_rate <= 0) {
    stop("`sampling_rate` must be a positive number.", call. = FALSE)
  }

  time <- seq(0, n_seconds, by = 1 / sampling_rate)
  n <- length(time)

  if (is.null(scr_onsets)) {
    expected_events <- max(1, round(n_seconds / 60 * scr_rate_per_min))
    scr_onsets <- sort(stats::runif(expected_events, 5, max(6, n_seconds - 5)))
  }

  scr_onsets <- scr_onsets[is.finite(scr_onsets) & scr_onsets >= 0 & scr_onsets <= n_seconds]

  tonic <- 1 + 0.001 * time
  eda <- tonic

  scr_truth <- data.frame(
    event_id = seq_along(scr_onsets),
    onset = scr_onsets,
    amplitude = stats::runif(length(scr_onsets), 0.03, 0.20),
    tau0 = 3.0,
    tau1 = 0.7,
    stringsAsFactors = FALSE
  )

  for (i in seq_along(scr_onsets)) {
    response_time <- pmax(0, time - scr_onsets[i])
    response <- exp(-response_time / scr_truth$tau0[i]) -
      exp(-response_time / scr_truth$tau1[i])
    response[response < 0] <- 0

    if (max(response) > 0) {
      response <- response / max(response)
    }

    eda <- eda + scr_truth$amplitude[i] * response
  }

  eda <- eda + stats::rnorm(n, sd = eda_noise_sd)

  pulse_interval <- 60 / pulse_rate_bpm
  pulse_times <- seq(0.5, n_seconds, by = pulse_interval)
  pulse_times <- pulse_times + stats::rnorm(length(pulse_times), sd = pulse_interval * 0.03)
  pulse_times <- pulse_times[pulse_times >= 0 & pulse_times <= n_seconds]

  respiration <- 1 + 0.15 * sin(2 * pi * (respiration_rate_bpm / 60) * time)

  ppg <- numeric(n)

  for (pt in pulse_times) {
    ppg <- ppg + exp(-0.5 * ((time - pt) / 0.06)^2)
  }

  ppg <- respiration * ppg
  ppg <- ppg / max(ppg)
  ppg <- ppg + stats::rnorm(n, sd = ppg_noise_sd)

  ibi <- rep(pulse_interval, n)

  if (length(pulse_times) >= 2) {
    pulse_ibi <- diff(pulse_times)
    pulse_mid <- pulse_times[-1]
    ibi <- stats::approx(pulse_mid, pulse_ibi, xout = time, rule = 2)$y
  }

  hr <- 60 / ibi

  ttl0 <- rep(0L, n)

  if (isTRUE(include_ttl) && length(scr_onsets) > 0) {
    ttl_idx <- vapply(scr_onsets, function(onset) {
      which.min(abs(time - onset))
    }, integer(1))

    ttl0[ttl_idx] <- 1L
  }

  dat <- data.frame(
    participant_id = participant_id,
    CNT = time,
    GSR_US = eda,
    HRP = ppg,
    HR = hr,
    IBI = ibi,
    TTL0 = ttl0,
    stringsAsFactors = FALSE
  )

  pulse_truth <- data.frame(
    pulse_id = seq_along(pulse_times),
    peak_time = pulse_times,
    expected_ibi = pulse_interval,
    stringsAsFactors = FALSE
  )

  ground_truth <- list(
    scr_events = scr_truth,
    pulse_peaks = pulse_truth,
    respiration_rate_bpm = respiration_rate_bpm,
    pulse_rate_bpm = pulse_rate_bpm
  )

  overview <- data.frame(
    rows = nrow(dat),
    n_seconds = n_seconds,
    sampling_rate_hz = sampling_rate,
    scr_events = nrow(scr_truth),
    pulse_peaks = nrow(pulse_truth),
    status = "synthetic_gazepoint_biometrics_created",
    interpretation = paste(
      "Synthetic signals are for teaching, examples, tests, and model validation.",
      "They are not real participant physiology."
    ),
    stringsAsFactors = FALSE
  )

  structure(
    list(
      overview = overview,
      data = dat,
      ground_truth = ground_truth,
      settings = list(
        n_seconds = n_seconds,
        sampling_rate = sampling_rate,
        participant_id = participant_id,
        scr_onsets = scr_onsets,
        scr_rate_per_min = scr_rate_per_min,
        pulse_rate_bpm = pulse_rate_bpm,
        respiration_rate_bpm = respiration_rate_bpm,
        eda_noise_sd = eda_noise_sd,
        ppg_noise_sd = ppg_noise_sd,
        include_ttl = include_ttl,
        seed = seed
      )
    ),
    class = c("gazepoint_biometrics_simulation", "list")
  )
}

#' Chunk Gazepoint biometric data into fixed analysis episodes
#'
#' Adds programmatic fixed-duration chunks/episodes to continuous biometric data.
#' This is useful for baseline segmentation, repeated-measures feature
#' extraction, and analyses that do not rely on external TTL markers.
#'
#' @param dat A data frame.
#' @param time_col Numeric time column.
#' @param group_cols Optional grouping columns.
#' @param chunk_seconds Chunk duration in seconds.
#' @param start_time Optional fixed start time. If `NULL`, uses group minimum.
#' @param chunk_col Output chunk identifier column.
#' @param episode_col Output episode label column.
#' @param include_partial Logical. If `FALSE`, partial final chunks are marked
#'   but not assigned as complete chunks.
#'
#' @return A data frame with chunk columns and chunk-summary attributes.
#' @export
chunk_gazepoint_biometrics <- function(dat,
                                       time_col = "CNT",
                                       group_cols = NULL,
                                       chunk_seconds = 60,
                                       start_time = NULL,
                                       chunk_col = "chunk_id",
                                       episode_col = "episode_id",
                                       include_partial = FALSE) {
  if (!is.data.frame(dat)) {
    stop("`dat` must be a data frame.", call. = FALSE)
  }

  if (!time_col %in% names(dat)) {
    stop("Column `", time_col, "` was not found in `dat`.", call. = FALSE)
  }

  if (!is.numeric(dat[[time_col]])) {
    stop("`time_col` must identify a numeric column.", call. = FALSE)
  }

  if (!is.numeric(chunk_seconds) ||
      length(chunk_seconds) != 1 ||
      !is.finite(chunk_seconds) ||
      chunk_seconds <= 0) {
    stop("`chunk_seconds` must be a positive number.", call. = FALSE)
  }

  if (is.null(group_cols)) {
    group_cols <- character()
  }

  missing_groups <- setdiff(group_cols, names(dat))
  if (length(missing_groups) > 0) {
    stop("Missing `group_cols`: ", paste(missing_groups, collapse = ", "), call. = FALSE)
  }

  out <- dat
  out[[chunk_col]] <- NA_integer_
  out[[episode_col]] <- NA_character_
  out$chunk_start <- NA_real_
  out$chunk_end <- NA_real_
  out$chunk_midpoint <- NA_real_
  out$chunk_complete <- FALSE

  groups <- gpbiometrics_chunk_split(out, group_cols)

  summary_rows <- list()
  row_id <- 1L

  for (group_id in names(groups)) {
    idx <- groups[[group_id]]
    time <- out[[time_col]][idx]
    finite <- is.finite(time)

    if (!any(finite)) {
      next
    }

    group_start <- if (!is.null(start_time)) {
      start_time
    } else {
      min(time[finite])
    }

    group_end <- max(time[finite])
    raw_chunk <- floor((time - group_start) / chunk_seconds) + 1L

    raw_chunk[!is.finite(raw_chunk) | raw_chunk < 1] <- NA_integer_

    for (chunk_id in sort(unique(raw_chunk[is.finite(raw_chunk)]))) {
      chunk_idx <- idx[raw_chunk == chunk_id]
      chunk_start <- group_start + (chunk_id - 1) * chunk_seconds
      chunk_end <- chunk_start + chunk_seconds
      chunk_time <- out[[time_col]][chunk_idx]

      complete <- max(chunk_time, na.rm = TRUE) >= chunk_end ||
        isTRUE(include_partial)

      if (!complete && !isTRUE(include_partial)) {
        out[[chunk_col]][chunk_idx] <- NA_integer_
        out[[episode_col]][chunk_idx] <- NA_character_
      } else {
        out[[chunk_col]][chunk_idx] <- chunk_id
        out[[episode_col]][chunk_idx] <- paste0(group_id, "_chunk_", chunk_id)
      }

      out$chunk_start[chunk_idx] <- chunk_start
      out$chunk_end[chunk_idx] <- chunk_end
      out$chunk_midpoint[chunk_idx] <- mean(c(chunk_start, chunk_end))
      out$chunk_complete[chunk_idx] <- complete

      summary_rows[[row_id]] <- data.frame(
        group_id = group_id,
        chunk_id = chunk_id,
        episode_id = paste0(group_id, "_chunk_", chunk_id),
        chunk_start = chunk_start,
        chunk_end = chunk_end,
        chunk_midpoint = mean(c(chunk_start, chunk_end)),
        row_count = length(chunk_idx),
        observed_start = min(chunk_time, na.rm = TRUE),
        observed_end = max(chunk_time, na.rm = TRUE),
        complete = complete,
        assigned = complete || isTRUE(include_partial),
        stringsAsFactors = FALSE
      )

      row_id <- row_id + 1L
    }
  }

  chunk_summary <- if (length(summary_rows) > 0) {
    do.call(rbind, summary_rows)
  } else {
    data.frame()
  }

  rownames(chunk_summary) <- NULL

  overview <- data.frame(
    input_rows = nrow(dat),
    output_rows = nrow(out),
    group_count = length(groups),
    chunk_rows = nrow(chunk_summary),
    assigned_chunks = sum(chunk_summary$assigned, na.rm = TRUE),
    chunk_seconds = chunk_seconds,
    include_partial = include_partial,
    status = "biometric_chunks_created",
    stringsAsFactors = FALSE
  )

  attr(out, "chunk_overview") <- overview
  attr(out, "chunk_summary") <- chunk_summary
  attr(out, "chunk_settings") <- list(
    time_col = time_col,
    group_cols = group_cols,
    chunk_seconds = chunk_seconds,
    start_time = start_time,
    chunk_col = chunk_col,
    episode_col = episode_col,
    include_partial = include_partial
  )

  class(out) <- unique(c("gazepoint_biometric_chunks", class(out)))
  out
}

gpbiometrics_chunk_split <- function(dat, group_cols) {
  if (length(group_cols) == 0) {
    return(list(all_rows = seq_len(nrow(dat))))
  }

  gf <- dat[group_cols]
  gf[] <- lapply(gf, function(x) {
    x <- as.character(x)
    x[is.na(x)] <- "<NA>"
    x
  })

  split(seq_len(nrow(dat)), do.call(paste, c(gf, sep = " | ")))
}

test_that("optimize_gazepoint_cvxeda_tau selects a best tau per group", {
  time <- seq(0, 120, by = 0.5)
  eda <- 1 + 0.01 * sin(2 * pi * 0.05 * time) +
    0.05 * exp(-pmax(0, time - 20) / 3) +
    0.03 * exp(-pmax(0, time - 70) / 3)

  dat <- data.frame(
    participant = "p1",
    time = time,
    GSR_US = eda
  )

  out <- optimize_gazepoint_cvxeda_tau(
    dat,
    eda_col = "GSR_US",
    time_col = "time",
    group_cols = "participant",
    tau0_grid = c(2, 3, 4),
    sampling_rate = 2
  )

  expect_s3_class(out, "gazepoint_cvxeda_tau_optimization")
  expect_equal(nrow(out$best_tau), 1)
  expect_true(out$best_tau$tau0 %in% c(2, 3, 4))
  expect_true("rmse" %in% names(out$optimization_table))
})

test_that("test_gazepoint_hrv_nonlinearity returns surrogate test output", {
  set.seed(1)

  dat <- data.frame(
    participant = "p1",
    IBI = 0.8 + 0.04 * sin(seq(0, 8 * pi, length.out = 80)) +
      stats::rnorm(80, sd = 0.005)
  )

  out <- test_gazepoint_hrv_nonlinearity(
    dat,
    ibi_col = "IBI",
    group_cols = "participant",
    metric = "sample_entropy",
    n_surrogates = 9,
    surrogate_method = "shuffle",
    seed = 1
  )

  expect_s3_class(out, "gazepoint_hrv_nonlinearity_test")
  expect_equal(nrow(out$results), 1)
  expect_equal(nrow(out$surrogate_statistics), 9)
  expect_true("p_two_sided" %in% names(out$results))
})

test_that("simulate_gazepoint_biometrics returns synthetic data and ground truth", {
  out <- simulate_gazepoint_biometrics(
    n_seconds = 20,
    sampling_rate = 20,
    scr_onsets = c(5, 12),
    seed = 1
  )

  expect_s3_class(out, "gazepoint_biometrics_simulation")
  expect_true(nrow(out$data) > 0)
  expect_true("GSR_US" %in% names(out$data))
  expect_true("HRP" %in% names(out$data))
  expect_equal(nrow(out$ground_truth$scr_events), 2)
})

test_that("chunk_gazepoint_biometrics assigns fixed analysis chunks", {
  dat <- data.frame(
    participant = rep(c("p1", "p2"), each = 121),
    CNT = rep(seq(0, 120, by = 1), 2),
    GSR_US = stats::rnorm(242)
  )

  out <- chunk_gazepoint_biometrics(
    dat,
    time_col = "CNT",
    group_cols = "participant",
    chunk_seconds = 60,
    include_partial = TRUE
  )

  expect_s3_class(out, "gazepoint_biometric_chunks")
  expect_true("chunk_id" %in% names(out))
  expect_true("episode_id" %in% names(out))
  expect_true(nrow(attr(out, "chunk_summary")) >= 4)
})

#' Extract bilateral EDA asymmetry features
#'
#' Computes left-right electrodermal activity asymmetry descriptors from two
#' simultaneously recorded EDA channels. The function returns row-level
#' asymmetry time series and group-level summaries.
#'
#' These descriptors quantify bilateral EDA differences only. They do not infer
#' hemisphere activation, amygdala activity, psychopathology, emotion, stress,
#' cognition, health status, or diagnosis.
#'
#' @param dat A data frame.
#' @param left_col Numeric left-side EDA column.
#' @param right_col Numeric right-side EDA column.
#' @param time_col Optional numeric time column for ordering and gradient
#'   calculation.
#' @param group_cols Optional grouping columns.
#' @param output_prefix Prefix used for row-level output columns.
#'
#' @return A list with `overview`, `asymmetry_timeseries`, `summary`, and
#'   `settings`.
#' @export
extract_gazepoint_bilateral_eda_asymmetry <- function(dat,
                                                      left_col,
                                                      right_col,
                                                      time_col = NULL,
                                                      group_cols = NULL,
                                                      output_prefix = "beda") {
  if (!is.data.frame(dat)) {
    stop("`dat` must be a data frame.", call. = FALSE)
  }

  required <- c(left_col, right_col, time_col)
  required <- required[!is.null(required)]

  missing_required <- setdiff(required, names(dat))
  if (length(missing_required) > 0) {
    stop("Missing required columns: ", paste(missing_required, collapse = ", "), call. = FALSE)
  }

  if (!is.numeric(dat[[left_col]])) {
    stop("`left_col` must identify a numeric column.", call. = FALSE)
  }

  if (!is.numeric(dat[[right_col]])) {
    stop("`right_col` must identify a numeric column.", call. = FALSE)
  }

  if (!is.null(time_col) && !is.numeric(dat[[time_col]])) {
    stop("`time_col` must identify a numeric column.", call. = FALSE)
  }

  if (is.null(group_cols)) {
    group_cols <- character()
  }

  missing_groups <- setdiff(group_cols, names(dat))
  if (length(missing_groups) > 0) {
    stop("Missing `group_cols`: ", paste(missing_groups, collapse = ", "), call. = FALSE)
  }

  groups <- gpbiometrics_modality_split(dat, group_cols)

  rows <- list()
  summary_rows <- list()

  diff_col <- paste0(output_prefix, "_left_minus_right")
  abs_col <- paste0(output_prefix, "_absolute_difference")
  norm_col <- paste0(output_prefix, "_normalised_difference")
  log_ratio_col <- paste0(output_prefix, "_log_left_right_ratio")
  gradient_col <- paste0(output_prefix, "_difference_gradient")

  for (group_id in names(groups)) {
    idx <- groups[[group_id]]

    if (!is.null(time_col)) {
      idx <- idx[order(dat[[time_col]][idx])]
    }

    left <- dat[[left_col]][idx]
    right <- dat[[right_col]][idx]
    mean_pair <- rowMeans(cbind(left, right), na.rm = TRUE)

    asym_diff <- left - right
    abs_diff <- abs(asym_diff)

    norm_diff <- ifelse(
      is.finite(mean_pair) & mean_pair != 0,
      asym_diff / mean_pair,
      NA_real_
    )

    log_ratio <- ifelse(
      is.finite(left) & is.finite(right) & left > 0 & right > 0,
      log(left / right),
      NA_real_
    )

    gradient <- rep(NA_real_, length(idx))

    if (!is.null(time_col) && length(idx) >= 2) {
      time <- dat[[time_col]][idx]
      dt <- diff(time)
      dd <- diff(asym_diff)

      gradient[-1] <- ifelse(
        is.finite(dt) & dt != 0,
        dd / dt,
        NA_real_
      )
    }

    group_dat <- data.frame(
      row_index = idx,
      group_id = group_id,
      stringsAsFactors = FALSE
    )

    if (!is.null(time_col)) {
      group_dat[[time_col]] <- dat[[time_col]][idx]
    }

    group_dat[[left_col]] <- left
    group_dat[[right_col]] <- right
    group_dat[[diff_col]] <- asym_diff
    group_dat[[abs_col]] <- abs_diff
    group_dat[[norm_col]] <- norm_diff
    group_dat[[log_ratio_col]] <- log_ratio
    group_dat[[gradient_col]] <- gradient

    rows[[group_id]] <- group_dat

    finite_pair <- is.finite(left) & is.finite(right)

    summary_rows[[group_id]] <- data.frame(
      group_id = group_id,
      n_rows = length(idx),
      n_valid_pairs = sum(finite_pair),
      mean_left = mean(left[finite_pair], na.rm = TRUE),
      mean_right = mean(right[finite_pair], na.rm = TRUE),
      mean_left_minus_right = mean(asym_diff[finite_pair], na.rm = TRUE),
      median_left_minus_right = stats::median(asym_diff[finite_pair], na.rm = TRUE),
      mean_absolute_difference = mean(abs_diff[finite_pair], na.rm = TRUE),
      mean_normalised_difference = mean(norm_diff[finite_pair], na.rm = TRUE),
      mean_log_left_right_ratio = mean(log_ratio[finite_pair], na.rm = TRUE),
      sd_left_minus_right = stats::sd(asym_diff[finite_pair], na.rm = TRUE),
      status = if (sum(finite_pair) > 0) {
        "bilateral_eda_asymmetry_extracted"
      } else {
        "no_valid_bilateral_pairs"
      },
      stringsAsFactors = FALSE
    )
  }

  asymmetry_timeseries <- do.call(rbind, rows)
  rownames(asymmetry_timeseries) <- NULL

  summary <- do.call(rbind, summary_rows)
  rownames(summary) <- NULL

  overview <- data.frame(
    group_count = length(groups),
    timeseries_rows = nrow(asymmetry_timeseries),
    summary_rows = nrow(summary),
    successful_groups = sum(summary$status == "bilateral_eda_asymmetry_extracted"),
    problem_groups = sum(summary$status != "bilateral_eda_asymmetry_extracted"),
    status = if (all(summary$status == "bilateral_eda_asymmetry_extracted")) {
      "bilateral_eda_asymmetry_complete"
    } else if (any(summary$status == "bilateral_eda_asymmetry_extracted")) {
      "bilateral_eda_asymmetry_partial"
    } else {
      "bilateral_eda_asymmetry_failed"
    },
    interpretation = paste(
      "Bilateral EDA asymmetry features quantify left-right signal differences.",
      "They do not infer hemisphere activation, brain-region activity, emotion, stress, cognition, health status, or diagnosis."
    ),
    stringsAsFactors = FALSE
  )

  structure(
    list(
      overview = overview,
      asymmetry_timeseries = asymmetry_timeseries,
      summary = summary,
      settings = list(
        left_col = left_col,
        right_col = right_col,
        time_col = time_col,
        group_cols = group_cols,
        output_prefix = output_prefix
      )
    ),
    class = c("gazepoint_bilateral_eda_asymmetry", "list")
  )
}

#' Add small uniform noise to reduce quantization overlap
#'
#' Adds uniform white noise with magnitude tied to hardware resolution. This is
#' intended only for nonlinear phase-space methods that are sensitive to exact
#' repeated values caused by coarse interval quantization.
#'
#' @param dat A data frame.
#' @param signal_cols Numeric signal columns to jitter.
#' @param resolution Numeric scalar or named numeric vector giving measurement
#'   resolution for each column.
#' @param group_cols Optional grouping columns, retained in settings.
#' @param output_suffix Suffix for jittered columns.
#' @param seed Optional random seed.
#' @param overwrite Logical. If `FALSE`, existing output columns are protected.
#'
#' @return A data frame with jittered columns and quantization-noise attributes.
#' @export
denoise_gazepoint_quantization_noise <- function(dat,
                                                 signal_cols,
                                                 resolution,
                                                 group_cols = NULL,
                                                 output_suffix = "_quantization_jittered",
                                                 seed = NULL,
                                                 overwrite = FALSE) {
  if (!is.data.frame(dat)) {
    stop("`dat` must be a data frame.", call. = FALSE)
  }

  missing_signals <- setdiff(signal_cols, names(dat))
  if (length(missing_signals) > 0) {
    stop("Missing `signal_cols`: ", paste(missing_signals, collapse = ", "), call. = FALSE)
  }

  non_numeric <- signal_cols[!vapply(dat[signal_cols], is.numeric, logical(1))]
  if (length(non_numeric) > 0) {
    stop("These `signal_cols` are not numeric: ", paste(non_numeric, collapse = ", "), call. = FALSE)
  }

  if (is.null(group_cols)) {
    group_cols <- character()
  }

  missing_groups <- setdiff(group_cols, names(dat))
  if (length(missing_groups) > 0) {
    stop("Missing `group_cols`: ", paste(missing_groups, collapse = ", "), call. = FALSE)
  }

  if (!is.numeric(resolution) || length(resolution) < 1 || any(!is.finite(resolution)) || any(resolution <= 0)) {
    stop("`resolution` must contain positive finite numeric values.", call. = FALSE)
  }

  if (!is.null(seed)) {
    set.seed(seed)
  }

  out <- dat
  summary_rows <- list()

  for (signal_col in signal_cols) {
    output_col <- paste0(signal_col, output_suffix)

    if (!isTRUE(overwrite) && output_col %in% names(out)) {
      stop("Output column `", output_col, "` already exists. Use `overwrite = TRUE`.", call. = FALSE)
    }

    col_resolution <- gpbiometrics_quantization_resolution(signal_col, resolution)

    noise <- stats::runif(
      n = nrow(out),
      min = -col_resolution / 2,
      max = col_resolution / 2
    )

    x <- out[[signal_col]]
    y <- x
    y[is.finite(x)] <- x[is.finite(x)] + noise[is.finite(x)]

    out[[output_col]] <- y

    summary_rows[[signal_col]] <- data.frame(
      signal_col = signal_col,
      output_col = output_col,
      resolution = col_resolution,
      noise_min = -col_resolution / 2,
      noise_max = col_resolution / 2,
      finite_rows = sum(is.finite(x)),
      changed_rows = sum(is.finite(x)),
      status = "quantization_jitter_added",
      stringsAsFactors = FALSE
    )
  }

  summary <- do.call(rbind, summary_rows)
  rownames(summary) <- NULL

  overview <- data.frame(
    input_rows = nrow(dat),
    output_rows = nrow(out),
    signal_count = length(signal_cols),
    status = "quantization_noise_reduction_complete",
    interpretation = paste(
      "Uniform jitter is intended only to reduce exact overlap from coarse quantization before nonlinear phase-space analyses.",
      "It does not recover lost physiological information and should not replace raw data for ordinary summaries."
    ),
    stringsAsFactors = FALSE
  )

  attr(out, "quantization_noise_overview") <- overview
  attr(out, "quantization_noise_summary") <- summary
  attr(out, "quantization_noise_settings") <- list(
    signal_cols = signal_cols,
    resolution = resolution,
    group_cols = group_cols,
    output_suffix = output_suffix,
    seed = seed,
    overwrite = overwrite
  )

  class(out) <- unique(c("gazepoint_quantization_noise_adjusted", class(out)))
  out
}

#' Extract ECG-derived respiration using PCA
#'
#' Extracts an ECG-derived respiration proxy from beat-level ECG morphology
#' features using principal component analysis. This function requires
#' ECG-derived morphology columns, such as QRS amplitudes, widths, or sampled
#' beat-shape features. It is not intended for HR, IBI, or PPG-only data.
#'
#' @param dat A data frame.
#' @param ecg_cols Numeric ECG morphology columns.
#' @param time_col Optional time column.
#' @param group_cols Optional grouping columns.
#' @param n_components Number of PCA components to retain.
#' @param scale Logical. If `TRUE`, scale ECG morphology columns before PCA.
#' @param output_prefix Prefix for PCA output columns.
#'
#' @return A list with `overview`, `edr_timeseries`, `component_summary`, and
#'   `settings`.
#' @export
extract_gazepoint_edr_pca <- function(dat,
                                      ecg_cols,
                                      time_col = NULL,
                                      group_cols = NULL,
                                      n_components = 1,
                                      scale = TRUE,
                                      output_prefix = "edr_pca") {
  if (!is.data.frame(dat)) {
    stop("`dat` must be a data frame.", call. = FALSE)
  }

  if (length(ecg_cols) < 2) {
    stop("`ecg_cols` must contain at least two ECG morphology columns.", call. = FALSE)
  }

  missing_ecg <- setdiff(ecg_cols, names(dat))
  if (length(missing_ecg) > 0) {
    stop("Missing `ecg_cols`: ", paste(missing_ecg, collapse = ", "), call. = FALSE)
  }

  non_numeric <- ecg_cols[!vapply(dat[ecg_cols], is.numeric, logical(1))]
  if (length(non_numeric) > 0) {
    stop("These `ecg_cols` are not numeric: ", paste(non_numeric, collapse = ", "), call. = FALSE)
  }

  if (!is.null(time_col) && !time_col %in% names(dat)) {
    stop("Column `", time_col, "` was not found in `dat`.", call. = FALSE)
  }

  if (!is.null(time_col) && !is.numeric(dat[[time_col]])) {
    stop("`time_col` must identify a numeric column.", call. = FALSE)
  }

  if (is.null(group_cols)) {
    group_cols <- character()
  }

  missing_groups <- setdiff(group_cols, names(dat))
  if (length(missing_groups) > 0) {
    stop("Missing `group_cols`: ", paste(missing_groups, collapse = ", "), call. = FALSE)
  }

  groups <- gpbiometrics_modality_split(dat, group_cols)

  rows <- list()
  component_rows <- list()

  for (group_id in names(groups)) {
    idx <- groups[[group_id]]

    if (!is.null(time_col)) {
      idx <- idx[order(dat[[time_col]][idx])]
    }

    x <- dat[idx, ecg_cols, drop = FALSE]
    complete <- stats::complete.cases(x)

    group_out <- data.frame(
      row_index = idx,
      group_id = group_id,
      stringsAsFactors = FALSE
    )

    if (!is.null(time_col)) {
      group_out[[time_col]] <- dat[[time_col]][idx]
    }

    for (component in seq_len(n_components)) {
      group_out[[paste0(output_prefix, "_pc", component)]] <- NA_real_
    }

    if (sum(complete) < max(3, n_components + 1)) {
      group_out[[paste0(output_prefix, "_status")]] <- "insufficient_complete_ecg_morphology"
      rows[[group_id]] <- group_out

      component_rows[[group_id]] <- data.frame(
        group_id = group_id,
        component = seq_len(n_components),
        variance_explained = NA_real_,
        cumulative_variance_explained = NA_real_,
        status = "insufficient_complete_ecg_morphology",
        stringsAsFactors = FALSE
      )

      next
    }

    pca <- stats::prcomp(
      x[complete, , drop = FALSE],
      center = TRUE,
      scale. = scale
    )

    available_components <- min(n_components, ncol(pca$x))

    for (component in seq_len(available_components)) {
      group_out[[paste0(output_prefix, "_pc", component)]][complete] <- pca$x[, component]
    }

    group_out[[paste0(output_prefix, "_status")]] <- ifelse(
      complete,
      "edr_pca_extracted",
      "incomplete_ecg_morphology"
    )

    variance <- pca$sdev^2
    variance_explained <- variance / sum(variance)

    component_rows[[group_id]] <- data.frame(
      group_id = group_id,
      component = seq_len(available_components),
      variance_explained = variance_explained[seq_len(available_components)],
      cumulative_variance_explained = cumsum(variance_explained)[seq_len(available_components)],
      status = "edr_pca_extracted",
      stringsAsFactors = FALSE
    )

    rows[[group_id]] <- group_out
  }

  edr_timeseries <- do.call(rbind, rows)
  rownames(edr_timeseries) <- NULL

  component_summary <- do.call(rbind, component_rows)
  rownames(component_summary) <- NULL

  status_col <- paste0(output_prefix, "_status")

  overview <- data.frame(
    group_count = length(groups),
    timeseries_rows = nrow(edr_timeseries),
    component_rows = nrow(component_summary),
    successful_rows = sum(edr_timeseries[[status_col]] == "edr_pca_extracted"),
    problem_rows = sum(edr_timeseries[[status_col]] != "edr_pca_extracted"),
    status = if (any(edr_timeseries[[status_col]] == "edr_pca_extracted")) {
      "edr_pca_extracted"
    } else {
      "edr_pca_failed"
    },
    interpretation = paste(
      "EDR-PCA outputs are respiration proxy components from ECG morphology columns.",
      "They require ECG-like morphology input and should not be interpreted as direct respiratory-belt measurement."
    ),
    stringsAsFactors = FALSE
  )

  structure(
    list(
      overview = overview,
      edr_timeseries = edr_timeseries,
      component_summary = component_summary,
      settings = list(
        ecg_cols = ecg_cols,
        time_col = time_col,
        group_cols = group_cols,
        n_components = n_components,
        scale = scale,
        output_prefix = output_prefix
      )
    ),
    class = c("gazepoint_edr_pca", "list")
  )
}

#' Analyse endosomatic skin-potential recordings
#'
#' Computes skin-potential level and skin-potential response descriptors from a
#' voltage-like skin-potential column. This is for endosomatic skin-potential
#' recordings, not standard exosomatic skin conductance.
#'
#' @param dat A data frame.
#' @param sp_col Numeric skin-potential column, usually in millivolts.
#' @param time_col Numeric time column.
#' @param group_cols Optional grouping columns.
#' @param response_direction `"both"`, `"positive"`, or `"negative"`.
#' @param response_threshold Optional absolute threshold for response detection.
#'   If `NULL`, a MAD-based derivative threshold is used.
#' @param min_response_distance_s Minimum distance between detected responses.
#'
#' @return A list with `overview`, `level_summary`, `response_table`,
#'   `timeseries`, and `settings`.
#' @export
analyze_gazepoint_skin_potential <- function(dat,
                                             sp_col,
                                             time_col,
                                             group_cols = NULL,
                                             response_direction = c("both", "positive", "negative"),
                                             response_threshold = NULL,
                                             min_response_distance_s = 1) {
  if (!is.data.frame(dat)) {
    stop("`dat` must be a data frame.", call. = FALSE)
  }

  response_direction <- match.arg(response_direction)

  required <- c(sp_col, time_col)
  missing_required <- setdiff(required, names(dat))
  if (length(missing_required) > 0) {
    stop("Missing required columns: ", paste(missing_required, collapse = ", "), call. = FALSE)
  }

  if (!is.numeric(dat[[sp_col]])) {
    stop("`sp_col` must identify a numeric skin-potential column.", call. = FALSE)
  }

  if (!is.numeric(dat[[time_col]])) {
    stop("`time_col` must identify a numeric time column.", call. = FALSE)
  }

  if (is.null(group_cols)) {
    group_cols <- character()
  }

  missing_groups <- setdiff(group_cols, names(dat))
  if (length(missing_groups) > 0) {
    stop("Missing `group_cols`: ", paste(missing_groups, collapse = ", "), call. = FALSE)
  }

  groups <- gpbiometrics_modality_split(dat, group_cols)

  level_rows <- list()
  response_rows <- list()
  timeseries_rows <- list()
  response_id <- 1L

  for (group_id in names(groups)) {
    idx <- groups[[group_id]]
    idx <- idx[order(dat[[time_col]][idx])]

    time <- dat[[time_col]][idx]
    sp <- dat[[sp_col]][idx]

    finite <- is.finite(time) & is.finite(sp)

    spl_center <- if (any(finite)) {
      stats::median(sp[finite], na.rm = TRUE)
    } else {
      NA_real_
    }

    sp_centered <- sp - spl_center
    derivative <- rep(NA_real_, length(sp))

    if (length(sp) >= 2) {
      dt <- diff(time)
      dsp <- diff(sp)
      derivative[-1] <- ifelse(is.finite(dt) & dt != 0, dsp / dt, NA_real_)
    }

    threshold <- response_threshold

    if (is.null(threshold)) {
      mad_derivative <- stats::mad(derivative, constant = 1, na.rm = TRUE)
      if (!is.finite(mad_derivative) || mad_derivative == 0) {
        mad_derivative <- stats::sd(derivative, na.rm = TRUE)
      }
      if (!is.finite(mad_derivative) || mad_derivative == 0) {
        mad_derivative <- NA_real_
      }
      threshold <- 6 * mad_derivative
    }

    candidate <- gpbiometrics_skin_potential_candidates(
      derivative = derivative,
      threshold = threshold,
      direction = response_direction
    )

    event_idx <- gpbiometrics_skin_potential_select_events(
      candidate = candidate,
      time = time,
      amplitude = derivative,
      min_distance_s = min_response_distance_s
    )

    response_flag <- rep(FALSE, length(idx))
    response_flag[event_idx] <- TRUE

    if (length(event_idx) > 0) {
      for (j in seq_along(event_idx)) {
        local_idx <- event_idx[j]
        response_rows[[response_id]] <- data.frame(
          group_id = group_id,
          response_index = j,
          row_index = idx[local_idx],
          response_time = time[local_idx],
          skin_potential = sp[local_idx],
          centered_skin_potential = sp_centered[local_idx],
          derivative = derivative[local_idx],
          response_polarity = if (derivative[local_idx] > 0) {
            "positive"
          } else if (derivative[local_idx] < 0) {
            "negative"
          } else {
            "zero"
          },
          stringsAsFactors = FALSE
        )
        response_id <- response_id + 1L
      }
    }

    timeseries_rows[[group_id]] <- data.frame(
      row_index = idx,
      group_id = group_id,
      time = time,
      skin_potential = sp,
      centered_skin_potential = sp_centered,
      skin_potential_derivative = derivative,
      skin_potential_response = response_flag,
      stringsAsFactors = FALSE
    )

    level_rows[[group_id]] <- data.frame(
      group_id = group_id,
      n_rows = length(idx),
      n_finite = sum(finite),
      mean_spl = mean(sp[finite], na.rm = TRUE),
      median_spl = stats::median(sp[finite], na.rm = TRUE),
      sd_spl = stats::sd(sp[finite], na.rm = TRUE),
      min_spl = min(sp[finite], na.rm = TRUE),
      max_spl = max(sp[finite], na.rm = TRUE),
      response_count = length(event_idx),
      response_rate_per_min = gpbiometrics_skin_potential_rate(time, length(event_idx)),
      threshold_used = threshold,
      status = if (sum(finite) > 0) {
        "skin_potential_analysed"
      } else {
        "no_valid_skin_potential"
      },
      stringsAsFactors = FALSE
    )
  }

  level_summary <- do.call(rbind, level_rows)
  rownames(level_summary) <- NULL

  response_table <- if (length(response_rows) > 0) {
    do.call(rbind, response_rows)
  } else {
    data.frame()
  }

  timeseries <- do.call(rbind, timeseries_rows)
  rownames(timeseries) <- NULL

  overview <- data.frame(
    group_count = length(groups),
    timeseries_rows = nrow(timeseries),
    response_rows = nrow(response_table),
    successful_groups = sum(level_summary$status == "skin_potential_analysed"),
    problem_groups = sum(level_summary$status != "skin_potential_analysed"),
    status = if (all(level_summary$status == "skin_potential_analysed")) {
      "skin_potential_analysis_complete"
    } else if (any(level_summary$status == "skin_potential_analysed")) {
      "skin_potential_analysis_partial"
    } else {
      "skin_potential_analysis_failed"
    },
    interpretation = paste(
      "Skin potential analysis is for endosomatic voltage-like recordings, not exosomatic skin conductance.",
      "Outputs are SPL/SPR descriptors and do not infer emotion, stress, cognition, health status, or diagnosis."
    ),
    stringsAsFactors = FALSE
  )

  structure(
    list(
      overview = overview,
      level_summary = level_summary,
      response_table = response_table,
      timeseries = timeseries,
      settings = list(
        sp_col = sp_col,
        time_col = time_col,
        group_cols = group_cols,
        response_direction = response_direction,
        response_threshold = response_threshold,
        min_response_distance_s = min_response_distance_s
      )
    ),
    class = c("gazepoint_skin_potential_analysis", "list")
  )
}

gpbiometrics_modality_split <- function(dat, group_cols) {
  if (length(group_cols) == 0) {
    return(list(all_rows = seq_len(nrow(dat))))
  }

  gf <- dat[group_cols]
  gf[] <- lapply(gf, function(x) {
    x <- as.character(x)
    x[is.na(x)] <- "<NA>"
    x
  })

  split(seq_len(nrow(dat)), do.call(paste, c(gf, sep = " | ")))
}

gpbiometrics_quantization_resolution <- function(signal_col, resolution) {
  if (length(resolution) == 1 && is.null(names(resolution))) {
    return(as.numeric(resolution[1]))
  }

  if (!is.null(names(resolution)) && signal_col %in% names(resolution)) {
    return(as.numeric(resolution[[signal_col]]))
  }

  if (length(resolution) == 1) {
    return(as.numeric(resolution[1]))
  }

  stop(
    "When `resolution` has multiple values, it must be named and include `",
    signal_col,
    "`.",
    call. = FALSE
  )
}

gpbiometrics_skin_potential_candidates <- function(derivative,
                                                   threshold,
                                                   direction = "both") {
  if (!is.finite(threshold) || threshold <= 0) {
    return(integer())
  }

  if (direction == "positive") {
    return(which(is.finite(derivative) & derivative >= threshold))
  }

  if (direction == "negative") {
    return(which(is.finite(derivative) & derivative <= -threshold))
  }

  which(is.finite(derivative) & abs(derivative) >= threshold)
}

gpbiometrics_skin_potential_select_events <- function(candidate,
                                                      time,
                                                      amplitude,
                                                      min_distance_s = 1) {
  if (length(candidate) == 0) {
    return(integer())
  }

  candidate <- candidate[is.finite(time[candidate])]

  if (length(candidate) == 0) {
    return(integer())
  }

  selected <- candidate[1]

  if (length(candidate) > 1) {
    for (idx in candidate[-1]) {
      last <- selected[length(selected)]

      if ((time[idx] - time[last]) >= min_distance_s) {
        selected <- c(selected, idx)
      } else if (abs(amplitude[idx]) > abs(amplitude[last])) {
        selected[length(selected)] <- idx
      }
    }
  }

  selected
}

gpbiometrics_skin_potential_rate <- function(time, event_count) {
  time <- time[is.finite(time)]

  if (length(time) < 2) {
    return(NA_real_)
  }

  duration <- max(time) - min(time)

  if (!is.finite(duration) || duration <= 0) {
    return(NA_real_)
  }

  event_count / duration * 60
}

test_that("extract_gazepoint_bilateral_eda_asymmetry returns bilateral descriptors", {
  dat <- data.frame(
    participant = "p1",
    time = seq(0, 10, by = 0.5),
    EDA_left = 1 + sin(seq(0, 10, by = 0.5)) * 0.05,
    EDA_right = 1 + cos(seq(0, 10, by = 0.5)) * 0.04
  )

  out <- extract_gazepoint_bilateral_eda_asymmetry(
    dat,
    left_col = "EDA_left",
    right_col = "EDA_right",
    time_col = "time",
    group_cols = "participant"
  )

  expect_s3_class(out, "gazepoint_bilateral_eda_asymmetry")
  expect_true("beda_left_minus_right" %in% names(out$asymmetry_timeseries))
  expect_true("mean_left_minus_right" %in% names(out$summary))
  expect_equal(out$overview$status, "bilateral_eda_asymmetry_complete")
})

test_that("denoise_gazepoint_quantization_noise adds jittered columns", {
  dat <- data.frame(
    participant = "p1",
    IBI = rep(c(0.80, 0.81, 0.82), length.out = 30),
    GSR_US = rep(c(1.00, 1.01, 1.02), length.out = 30)
  )

  out <- denoise_gazepoint_quantization_noise(
    dat,
    signal_cols = c("IBI", "GSR_US"),
    resolution = c(IBI = 0.001, GSR_US = 0.01),
    seed = 1
  )

  expect_s3_class(out, "gazepoint_quantization_noise_adjusted")
  expect_true("IBI_quantization_jittered" %in% names(out))
  expect_true("GSR_US_quantization_jittered" %in% names(out))
  expect_equal(attr(out, "quantization_noise_overview")$status, "quantization_noise_reduction_complete")
})

test_that("extract_gazepoint_edr_pca returns ECG-derived respiration proxy components", {
  time <- seq(0, 20, by = 0.5)
  resp <- sin(2 * pi * 0.2 * time)

  dat <- data.frame(
    participant = "p1",
    time = time,
    qrs_amp = 1 + 0.1 * resp,
    qrs_width = 0.08 + 0.01 * resp,
    qrs_area = 0.5 + 0.05 * resp
  )

  out <- extract_gazepoint_edr_pca(
    dat,
    ecg_cols = c("qrs_amp", "qrs_width", "qrs_area"),
    time_col = "time",
    group_cols = "participant",
    n_components = 1
  )

  expect_s3_class(out, "gazepoint_edr_pca")
  expect_true("edr_pca_pc1" %in% names(out$edr_timeseries))
  expect_true("variance_explained" %in% names(out$component_summary))
  expect_equal(out$overview$status, "edr_pca_extracted")
})

test_that("analyze_gazepoint_skin_potential returns SPL and SPR descriptors", {
  time <- seq(0, 60, by = 0.5)
  sp <- 2 + 0.01 * sin(time)

  sp[which.min(abs(time - 10))] <- sp[which.min(abs(time - 10))] + 0.5
  sp[which.min(abs(time - 30))] <- sp[which.min(abs(time - 30))] - 0.6

  dat <- data.frame(
    participant = "p1",
    time = time,
    SP_mV = sp
  )

  out <- analyze_gazepoint_skin_potential(
    dat,
    sp_col = "SP_mV",
    time_col = "time",
    group_cols = "participant",
    response_direction = "both",
    response_threshold = 0.5,
    min_response_distance_s = 2
  )

  expect_s3_class(out, "gazepoint_skin_potential_analysis")
  expect_true("mean_spl" %in% names(out$level_summary))
  expect_true("skin_potential_response" %in% names(out$timeseries))
  expect_equal(out$overview$status, "skin_potential_analysis_complete")
})

#' Extract PPG-derived respiration proxy signals
#'
#' Extracts dependency-light PPG-derived respiration proxy features from a
#' Gazepoint pulse/PPG waveform. The function estimates respiration-modulated
#' pulse features such as respiratory-induced intensity variability (RIIV),
#' pulse amplitude variability (PAV), pulse width variability (PWV), and
#' pulse-rate variability (PRV).
#'
#' These are proxy respiratory features. They should not be treated as a
#' replacement for a respiration belt or clinical respiratory measurement.
#'
#' @param dat A data frame containing a PPG/pulse waveform.
#' @param ppg_col Numeric PPG/pulse waveform column, often `HRP`.
#' @param time_col Numeric time column.
#' @param group_cols Optional grouping columns.
#' @param sampling_rate Optional sampling rate in Hz. If `NULL`, estimated from
#'   `time_col`.
#' @param min_peak_distance_s Minimum plausible distance between pulse peaks.
#' @param smooth_window Number of samples used for simple moving-average
#'   smoothing before peak detection.
#' @param respiration_band Expected respiration-frequency band in Hz.
#' @param pdr_resample_rate Resampling rate used for spectral estimation of PDR
#'   proxy signals.
#'
#' @return A list with `overview`, `pulse_features`, `pdr_timeseries`,
#'   `pdr_summary`, and `settings`.
#' @export
extract_gazepoint_pdr_signals <- function(dat,
                                          ppg_col = "HRP",
                                          time_col = "CNT",
                                          group_cols = NULL,
                                          sampling_rate = NULL,
                                          min_peak_distance_s = 0.30,
                                          smooth_window = 5,
                                          respiration_band = c(0.10, 0.60),
                                          pdr_resample_rate = 4) {
  if (!is.data.frame(dat)) {
    stop("`dat` must be a data frame.", call. = FALSE)
  }

  if (!ppg_col %in% names(dat)) {
    stop("Column `", ppg_col, "` was not found in `dat`.", call. = FALSE)
  }

  if (!time_col %in% names(dat)) {
    stop("Column `", time_col, "` was not found in `dat`.", call. = FALSE)
  }

  if (!is.numeric(dat[[ppg_col]])) {
    stop("`ppg_col` must identify a numeric column.", call. = FALSE)
  }

  if (!is.numeric(dat[[time_col]])) {
    stop("`time_col` must identify a numeric column.", call. = FALSE)
  }

  if (is.null(group_cols)) {
    group_cols <- character()
  }

  missing_groups <- setdiff(group_cols, names(dat))
  if (length(missing_groups) > 0) {
    stop("Missing `group_cols`: ", paste(missing_groups, collapse = ", "), call. = FALSE)
  }

  if (!is.numeric(respiration_band) ||
      length(respiration_band) != 2 ||
      any(!is.finite(respiration_band)) ||
      respiration_band[1] <= 0 ||
      respiration_band[1] >= respiration_band[2]) {
    stop("`respiration_band` must be a positive numeric vector of length two.", call. = FALSE)
  }

  groups <- gpbiometrics_pdr_split_indices(dat, group_cols)

  pulse_rows <- list()
  summary_rows <- list()
  pulse_row_id <- 1L

  for (group_id in names(groups)) {
    idx <- groups[[group_id]]
    idx <- idx[order(dat[[time_col]][idx])]

    time <- dat[[time_col]][idx]
    ppg <- dat[[ppg_col]][idx]

    fs <- gpbiometrics_pdr_sampling_rate(time, sampling_rate)

    if (!is.finite(fs) || fs <= 0) {
      fs <- NA_real_
    }

    smoothed <- gpbiometrics_pdr_smooth(ppg, smooth_window)
    peaks <- gpbiometrics_pdr_find_peaks(
      smoothed,
      time = time,
      min_peak_distance_s = min_peak_distance_s
    )

    if (length(peaks) < 3) {
      summary_rows[[group_id]] <- data.frame(
        group_id = group_id,
        n_samples = length(idx),
        n_pulses = length(peaks),
        sampling_rate_hz = fs,
        riiv_resp_rate_hz = NA_real_,
        pav_resp_rate_hz = NA_real_,
        pwv_resp_rate_hz = NA_real_,
        prv_resp_rate_hz = NA_real_,
        proxy_resp_rate_hz = NA_real_,
        proxy_resp_rate_bpm = NA_real_,
        status = "insufficient_pulse_peaks",
        stringsAsFactors = FALSE
      )
      next
    }

    for (i in 2:length(peaks)) {
      prev_peak <- peaks[i - 1]
      current_peak <- peaks[i]

      segment <- prev_peak:current_peak
      trough_local <- which.min(smoothed[segment])
      trough <- segment[trough_local]

      prev_prev_peak <- if (i >= 3) peaks[i - 2] else NA_integer_

      pulse_interval_s <- time[current_peak] - time[prev_peak]
      pulse_rate_bpm <- if (is.finite(pulse_interval_s) && pulse_interval_s > 0) {
        60 / pulse_interval_s
      } else {
        NA_real_
      }

      pulse_width_s <- if (is.finite(prev_prev_peak)) {
        time[current_peak] - time[prev_peak]
      } else {
        NA_real_
      }

      pulse_amplitude <- smoothed[current_peak] - smoothed[trough]
      riiv <- smoothed[trough]
      pav <- pulse_amplitude
      pwv <- pulse_width_s
      prv <- pulse_rate_bpm

      pulse_rows[[pulse_row_id]] <- data.frame(
        group_id = group_id,
        pulse_index = i - 1L,
        peak_row = idx[current_peak],
        trough_row = idx[trough],
        peak_time = time[current_peak],
        trough_time = time[trough],
        peak_value = smoothed[current_peak],
        trough_value = smoothed[trough],
        riiv = riiv,
        pav = pav,
        pwv = pwv,
        prv = prv,
        pulse_interval_s = pulse_interval_s,
        pulse_rate_bpm = pulse_rate_bpm,
        stringsAsFactors = FALSE
      )

      pulse_row_id <- pulse_row_id + 1L
    }

    group_pulses <- do.call(rbind, pulse_rows)
    group_pulses <- group_pulses[group_pulses$group_id == group_id, , drop = FALSE]

    if (nrow(group_pulses) == 0) {
      summary_rows[[group_id]] <- data.frame(
        group_id = group_id,
        n_samples = length(idx),
        n_pulses = length(peaks),
        sampling_rate_hz = fs,
        riiv_resp_rate_hz = NA_real_,
        pav_resp_rate_hz = NA_real_,
        pwv_resp_rate_hz = NA_real_,
        prv_resp_rate_hz = NA_real_,
        proxy_resp_rate_hz = NA_real_,
        proxy_resp_rate_bpm = NA_real_,
        status = "pdr_feature_extraction_failed",
        stringsAsFactors = FALSE
      )
      next
    }

    riiv_rate <- gpbiometrics_pdr_rate_from_feature(
      group_pulses$peak_time,
      group_pulses$riiv,
      respiration_band = respiration_band,
      resample_rate = pdr_resample_rate
    )

    pav_rate <- gpbiometrics_pdr_rate_from_feature(
      group_pulses$peak_time,
      group_pulses$pav,
      respiration_band = respiration_band,
      resample_rate = pdr_resample_rate
    )

    pwv_rate <- gpbiometrics_pdr_rate_from_feature(
      group_pulses$peak_time,
      group_pulses$pwv,
      respiration_band = respiration_band,
      resample_rate = pdr_resample_rate
    )

    prv_rate <- gpbiometrics_pdr_rate_from_feature(
      group_pulses$peak_time,
      group_pulses$prv,
      respiration_band = respiration_band,
      resample_rate = pdr_resample_rate
    )

    valid_rates <- c(riiv_rate, pav_rate, pwv_rate, prv_rate)
    proxy_rate <- if (any(is.finite(valid_rates))) {
      stats::median(valid_rates[is.finite(valid_rates)])
    } else {
      NA_real_
    }

    summary_rows[[group_id]] <- data.frame(
      group_id = group_id,
      n_samples = length(idx),
      n_pulses = length(peaks),
      sampling_rate_hz = fs,
      riiv_resp_rate_hz = riiv_rate,
      pav_resp_rate_hz = pav_rate,
      pwv_resp_rate_hz = pwv_rate,
      prv_resp_rate_hz = prv_rate,
      proxy_resp_rate_hz = proxy_rate,
      proxy_resp_rate_bpm = proxy_rate * 60,
      status = if (is.finite(proxy_rate)) {
        "pdr_extracted"
      } else {
        "pdr_rate_not_estimated"
      },
      stringsAsFactors = FALSE
    )
  }

  pulse_features <- if (length(pulse_rows) > 0) {
    do.call(rbind, pulse_rows)
  } else {
    data.frame()
  }

  if (nrow(pulse_features) > 0) {
    pulse_features$resp_proxy <- gpbiometrics_pdr_row_proxy(
      pulse_features[, c("riiv", "pav", "pwv", "prv"), drop = FALSE]
    )

    pdr_timeseries <- pulse_features[, c(
      "group_id", "pulse_index", "peak_time", "riiv", "pav", "pwv", "prv",
      "resp_proxy"
    )]
  } else {
    pdr_timeseries <- data.frame()
  }

  pdr_summary <- do.call(rbind, summary_rows)
  rownames(pdr_summary) <- NULL

  overview <- data.frame(
    group_count = length(groups),
    pulse_feature_rows = nrow(pulse_features),
    pdr_summary_rows = nrow(pdr_summary),
    successful_groups = sum(pdr_summary$status == "pdr_extracted"),
    problem_groups = sum(pdr_summary$status != "pdr_extracted"),
    ppg_col = ppg_col,
    time_col = time_col,
    status = if (all(pdr_summary$status == "pdr_extracted")) {
      "pdr_extraction_complete"
    } else if (any(pdr_summary$status == "pdr_extracted")) {
      "pdr_extraction_partial"
    } else {
      "pdr_extraction_failed"
    },
    interpretation = paste(
      "PPG-derived respiration features are proxy estimates from pulse morphology and timing.",
      "They are not a substitute for direct respiratory-belt measurement."
    ),
    stringsAsFactors = FALSE
  )

  structure(
    list(
      overview = overview,
      pulse_features = pulse_features,
      pdr_timeseries = pdr_timeseries,
      pdr_summary = pdr_summary,
      settings = list(
        ppg_col = ppg_col,
        time_col = time_col,
        group_cols = group_cols,
        sampling_rate = sampling_rate,
        min_peak_distance_s = min_peak_distance_s,
        smooth_window = smooth_window,
        respiration_band = respiration_band,
        pdr_resample_rate = pdr_resample_rate
      )
    ),
    class = c("gazepoint_pdr_signals", "list")
  )
}

#' Calculate respiration-informed RSA proxy features
#'
#' Calculates dependency-light RSA proxy summaries from IBI/RR intervals and,
#' optionally, PPG-derived respiration features from
#' `extract_gazepoint_pdr_signals()`. This provides peak-to-trough and
#' Porges-Bohrer-inspired band-power proxy summaries. These outputs should be
#' interpreted as respiration-informed HRV/RSA features, not direct clinical
#' vagal-tone estimates.
#'
#' @param dat A data frame containing IBI/RR data.
#' @param ibi_col IBI/RR interval column.
#' @param time_col Time column for the IBI/RR observations.
#' @param group_cols Optional grouping columns.
#' @param pdr Optional output from `extract_gazepoint_pdr_signals()`.
#' @param resp_rate_hz Optional fixed respiration rate in Hz.
#' @param respiration_band Default respiration/HF band when no PDR rate is
#'   available.
#' @param resample_rate Resampling rate for spectral RSA proxy calculation.
#'
#' @return A list with `overview`, `rsa_summary`, and `settings`.
#' @export
calculate_gazepoint_rsa <- function(dat,
                                    ibi_col = "IBI",
                                    time_col = "CNT",
                                    group_cols = NULL,
                                    pdr = NULL,
                                    resp_rate_hz = NULL,
                                    respiration_band = c(0.12, 0.40),
                                    resample_rate = 4) {
  if (!is.data.frame(dat)) {
    stop("`dat` must be a data frame.", call. = FALSE)
  }

  if (!ibi_col %in% names(dat)) {
    stop("Column `", ibi_col, "` was not found in `dat`.", call. = FALSE)
  }

  if (!time_col %in% names(dat)) {
    stop("Column `", time_col, "` was not found in `dat`.", call. = FALSE)
  }

  if (!is.numeric(dat[[ibi_col]])) {
    stop("`ibi_col` must identify a numeric column.", call. = FALSE)
  }

  if (!is.numeric(dat[[time_col]])) {
    stop("`time_col` must identify a numeric column.", call. = FALSE)
  }

  if (is.null(group_cols)) {
    group_cols <- character()
  }

  missing_groups <- setdiff(group_cols, names(dat))
  if (length(missing_groups) > 0) {
    stop("Missing `group_cols`: ", paste(missing_groups, collapse = ", "), call. = FALSE)
  }

  groups <- gpbiometrics_pdr_split_indices(dat, group_cols)

  pdr_summary <- NULL
  pdr_timeseries <- NULL

  if (!is.null(pdr)) {
    if (!inherits(pdr, "gazepoint_pdr_signals")) {
      stop("`pdr` must be output from `extract_gazepoint_pdr_signals()`.", call. = FALSE)
    }
    pdr_summary <- pdr$pdr_summary
    pdr_timeseries <- pdr$pdr_timeseries
  }

  rows <- lapply(names(groups), function(group_id) {
    idx <- groups[[group_id]]
    idx <- idx[order(dat[[time_col]][idx])]

    time <- dat[[time_col]][idx]
    ibi <- dat[[ibi_col]][idx]

    finite <- is.finite(time) & is.finite(ibi) & ibi > 0

    time <- time[finite]
    ibi <- ibi[finite]

    group_resp_rate <- gpbiometrics_rsa_group_resp_rate(
      group_id = group_id,
      pdr_summary = pdr_summary,
      resp_rate_hz = resp_rate_hz
    )

    p2t <- gpbiometrics_rsa_peak_to_trough_proxy(
      time = time,
      ibi = ibi,
      group_id = group_id,
      pdr_timeseries = pdr_timeseries
    )

    pb <- gpbiometrics_rsa_pb_proxy(
      time = time,
      ibi = ibi,
      resp_rate_hz = group_resp_rate,
      respiration_band = respiration_band,
      resample_rate = resample_rate
    )

    data.frame(
      group_id = group_id,
      n_intervals = length(ibi),
      resp_rate_hz = group_resp_rate,
      resp_rate_bpm = group_resp_rate * 60,
      rsa_p2t_proxy = p2t,
      rsa_pb_log_power_proxy = pb,
      status = if (length(ibi) >= 5 && (is.finite(p2t) || is.finite(pb))) {
        "rsa_proxy_calculated"
      } else {
        "rsa_proxy_insufficient_information"
      },
      stringsAsFactors = FALSE
    )
  })

  rsa_summary <- do.call(rbind, rows)
  rownames(rsa_summary) <- NULL

  overview <- data.frame(
    group_count = length(groups),
    rsa_rows = nrow(rsa_summary),
    successful_groups = sum(rsa_summary$status == "rsa_proxy_calculated"),
    problem_groups = sum(rsa_summary$status != "rsa_proxy_calculated"),
    status = if (all(rsa_summary$status == "rsa_proxy_calculated")) {
      "rsa_proxy_complete"
    } else if (any(rsa_summary$status == "rsa_proxy_calculated")) {
      "rsa_proxy_partial"
    } else {
      "rsa_proxy_failed"
    },
    interpretation = paste(
      "RSA outputs are respiration-informed HRV proxy summaries.",
      "Without direct respiration measurement they should not be interpreted as definitive vagal-tone estimates."
    ),
    stringsAsFactors = FALSE
  )

  structure(
    list(
      overview = overview,
      rsa_summary = rsa_summary,
      settings = list(
        ibi_col = ibi_col,
        time_col = time_col,
        group_cols = group_cols,
        pdr_supplied = !is.null(pdr),
        resp_rate_hz = resp_rate_hz,
        respiration_band = respiration_band,
        resample_rate = resample_rate
      )
    ),
    class = c("gazepoint_rsa_proxy", "list")
  )
}

gpbiometrics_pdr_split_indices <- function(dat, group_cols) {
  if (length(group_cols) == 0) {
    return(list(all_rows = seq_len(nrow(dat))))
  }

  gf <- dat[group_cols]
  gf[] <- lapply(gf, function(x) {
    x <- as.character(x)
    x[is.na(x)] <- "<NA>"
    x
  })

  split(seq_len(nrow(dat)), do.call(paste, c(gf, sep = " | ")))
}

gpbiometrics_pdr_sampling_rate <- function(time, sampling_rate = NULL) {
  if (!is.null(sampling_rate)) {
    return(sampling_rate)
  }

  time <- time[is.finite(time)]

  if (length(time) < 3) {
    return(NA_real_)
  }

  dt <- diff(time)
  dt <- dt[is.finite(dt) & dt > 0]

  if (length(dt) == 0) {
    return(NA_real_)
  }

  median_dt <- stats::median(dt)

  if (median_dt > 10) {
    1000 / median_dt
  } else {
    1 / median_dt
  }
}

gpbiometrics_pdr_smooth <- function(x, window = 5) {
  if (!is.numeric(window) || length(window) != 1 || window <= 1) {
    return(as.numeric(x))
  }

  window <- as.integer(window)

  if (length(x) < window) {
    return(as.numeric(x))
  }

  y <- as.numeric(stats::filter(x, rep(1 / window, window), sides = 2))
  y[!is.finite(y)] <- x[!is.finite(y)]
  y[is.na(y)] <- x[is.na(y)]
  y
}

gpbiometrics_pdr_find_peaks <- function(x, time, min_peak_distance_s = 0.30) {
  finite <- is.finite(x) & is.finite(time)

  if (sum(finite) < 3) {
    return(integer())
  }

  candidate <- which(
    c(FALSE, x[-c(1, length(x))] > x[-c(length(x) - 1, length(x))] &
        x[-c(1, length(x))] >= x[-c(1, 2)], FALSE)
  )

  candidate <- candidate[is.finite(x[candidate]) & is.finite(time[candidate])]

  if (length(candidate) == 0) {
    return(integer())
  }

  selected <- candidate[1]

  if (length(candidate) > 1) {
    for (i in candidate[-1]) {
      if ((time[i] - time[selected[length(selected)]]) >= min_peak_distance_s) {
        selected <- c(selected, i)
      } else if (x[i] > x[selected[length(selected)]]) {
        selected[length(selected)] <- i
      }
    }
  }

  selected
}

gpbiometrics_pdr_rate_from_feature <- function(time,
                                               feature,
                                               respiration_band = c(0.10, 0.60),
                                               resample_rate = 4) {
  keep <- is.finite(time) & is.finite(feature)

  time <- time[keep]
  feature <- feature[keep]

  if (length(time) < 8 || length(unique(time)) < 8 || stats::sd(feature) == 0) {
    return(NA_real_)
  }

  ord <- order(time)
  time <- time[ord]
  feature <- feature[ord]

  grid <- seq(min(time), max(time), by = 1 / resample_rate)

  if (length(grid) < 16) {
    return(NA_real_)
  }

  y <- stats::approx(time, feature, xout = grid, rule = 2)$y
  y <- y - mean(y, na.rm = TRUE)

  if (stats::sd(y, na.rm = TRUE) == 0) {
    return(NA_real_)
  }

  spec <- stats::spec.pgram(
    y,
    taper = 0.1,
    plot = FALSE,
    demean = TRUE,
    detrend = TRUE,
    fast = TRUE
  )

  freq <- spec$freq * resample_rate
  power <- spec$spec

  in_band <- is.finite(freq) &
    is.finite(power) &
    freq >= respiration_band[1] &
    freq <= respiration_band[2]

  if (!any(in_band)) {
    return(NA_real_)
  }

  freq[in_band][which.max(power[in_band])]
}

gpbiometrics_pdr_row_proxy <- function(feature_df) {
  scaled <- lapply(feature_df, function(x) {
    if (sum(is.finite(x)) < 2 || stats::sd(x, na.rm = TRUE) == 0) {
      return(rep(NA_real_, length(x)))
    }
    as.numeric(scale(x))
  })

  scaled <- as.data.frame(scaled)

  apply(scaled, 1, function(x) {
    if (all(!is.finite(x))) {
      NA_real_
    } else {
      mean(x, na.rm = TRUE)
    }
  })
}

gpbiometrics_rsa_group_resp_rate <- function(group_id,
                                             pdr_summary = NULL,
                                             resp_rate_hz = NULL) {
  if (!is.null(resp_rate_hz) &&
      is.numeric(resp_rate_hz) &&
      length(resp_rate_hz) == 1 &&
      is.finite(resp_rate_hz)) {
    return(resp_rate_hz)
  }

  if (is.null(pdr_summary) || nrow(pdr_summary) == 0) {
    return(NA_real_)
  }

  row <- pdr_summary[pdr_summary$group_id == group_id, , drop = FALSE]

  if (nrow(row) == 0 || !is.finite(row$proxy_resp_rate_hz[1])) {
    return(NA_real_)
  }

  row$proxy_resp_rate_hz[1]
}

gpbiometrics_rsa_peak_to_trough_proxy <- function(time,
                                                  ibi,
                                                  group_id,
                                                  pdr_timeseries = NULL) {
  if (is.null(pdr_timeseries) || nrow(pdr_timeseries) == 0) {
    return(NA_real_)
  }

  pdr_group <- pdr_timeseries[pdr_timeseries$group_id == group_id, , drop = FALSE]

  if (nrow(pdr_group) < 4 || !"resp_proxy" %in% names(pdr_group)) {
    return(NA_real_)
  }

  pdr_group <- pdr_group[order(pdr_group$peak_time), , drop = FALSE]
  resp <- pdr_group$resp_proxy
  resp_time <- pdr_group$peak_time

  peaks <- gpbiometrics_pdr_find_peaks(
    resp,
    time = resp_time,
    min_peak_distance_s = 1
  )

  if (length(peaks) < 2) {
    return(NA_real_)
  }

  cycle_values <- numeric()

  for (i in seq_len(length(peaks) - 1)) {
    start_time <- resp_time[peaks[i]]
    end_time <- resp_time[peaks[i + 1]]

    in_cycle <- is.finite(time) &
      time >= start_time &
      time <= end_time

    if (sum(in_cycle) >= 2) {
      cycle_ibi <- ibi[in_cycle]
      cycle_values <- c(
        cycle_values,
        max(cycle_ibi, na.rm = TRUE) - min(cycle_ibi, na.rm = TRUE)
      )
    }
  }

  if (length(cycle_values) == 0) {
    return(NA_real_)
  }

  mean(cycle_values, na.rm = TRUE)
}

gpbiometrics_rsa_pb_proxy <- function(time,
                                      ibi,
                                      resp_rate_hz = NA_real_,
                                      respiration_band = c(0.12, 0.40),
                                      resample_rate = 4) {
  keep <- is.finite(time) & is.finite(ibi) & ibi > 0

  time <- time[keep]
  ibi <- ibi[keep]

  if (length(time) < 8 || length(unique(time)) < 8 || stats::sd(ibi) == 0) {
    return(NA_real_)
  }

  ord <- order(time)
  time <- time[ord]
  ibi <- ibi[ord]

  grid <- seq(min(time), max(time), by = 1 / resample_rate)

  if (length(grid) < 16) {
    return(NA_real_)
  }

  y <- stats::approx(time, ibi, xout = grid, rule = 2)$y
  y <- y - mean(y, na.rm = TRUE)

  band <- respiration_band

  if (is.finite(resp_rate_hz) && resp_rate_hz > 0) {
    band <- c(
      max(0.05, resp_rate_hz - 0.05),
      min(resample_rate / 2 - 0.01, resp_rate_hz + 0.05)
    )
  }

  spec <- stats::spec.pgram(
    y,
    taper = 0.1,
    plot = FALSE,
    demean = TRUE,
    detrend = TRUE,
    fast = TRUE
  )

  freq <- spec$freq * resample_rate
  power <- spec$spec

  in_band <- is.finite(freq) &
    is.finite(power) &
    freq >= band[1] &
    freq <= band[2]

  if (!any(in_band)) {
    return(NA_real_)
  }

  band_power <- sum(power[in_band], na.rm = TRUE)

  if (!is.finite(band_power) || band_power <= 0) {
    return(NA_real_)
  }

  log(band_power)
}

test_that("extract_gazepoint_pdr_signals returns PDR summaries", {
  set.seed(1)

  time <- seq(0, 60, by = 0.05)
  respiration <- 1 + 0.15 * sin(2 * pi * 0.25 * time)
  ppg <- respiration * sin(2 * pi * 1.2 * time) + 0.02 * rnorm(length(time))

  dat <- data.frame(
    participant = "p1",
    time = time,
    HRP = ppg
  )

  out <- extract_gazepoint_pdr_signals(
    dat,
    ppg_col = "HRP",
    time_col = "time",
    group_cols = "participant",
    sampling_rate = 20,
    min_peak_distance_s = 0.4
  )

  expect_s3_class(out, "gazepoint_pdr_signals")
  expect_true(nrow(out$pulse_features) > 10)
  expect_true("proxy_resp_rate_hz" %in% names(out$pdr_summary))
  expect_true(out$overview$status %in% c("pdr_extraction_complete", "pdr_extraction_partial"))
})

test_that("calculate_gazepoint_rsa returns respiration-informed RSA proxies", {
  set.seed(1)

  time_ppg <- seq(0, 60, by = 0.05)
  respiration <- 1 + 0.15 * sin(2 * pi * 0.25 * time_ppg)
  ppg <- respiration * sin(2 * pi * 1.2 * time_ppg) + 0.02 * rnorm(length(time_ppg))

  ppg_dat <- data.frame(
    participant = "p1",
    time = time_ppg,
    HRP = ppg
  )

  pdr <- extract_gazepoint_pdr_signals(
    ppg_dat,
    ppg_col = "HRP",
    time_col = "time",
    group_cols = "participant",
    sampling_rate = 20,
    min_peak_distance_s = 0.4
  )

  ibi_time <- seq(0, 60, by = 1)
  ibi_dat <- data.frame(
    participant = "p1",
    time = ibi_time,
    IBI = 0.8 + 0.05 * sin(2 * pi * 0.25 * ibi_time)
  )

  rsa <- calculate_gazepoint_rsa(
    ibi_dat,
    ibi_col = "IBI",
    time_col = "time",
    group_cols = "participant",
    pdr = pdr
  )

  expect_s3_class(rsa, "gazepoint_rsa_proxy")
  expect_true("rsa_pb_log_power_proxy" %in% names(rsa$rsa_summary))
  expect_true(rsa$overview$status %in% c("rsa_proxy_complete", "rsa_proxy_partial"))
})

test_that("extended nonlinear HRV features include ApEn, DFA, and MSE", {
  dat <- data.frame(
    participant = "p1",
    IBI = 0.8 + 0.04 * sin(seq(0, 10 * pi, length.out = 80)) + rnorm(80, sd = 0.005)
  )

  out <- extract_gazepoint_hrv_nonlinear(
    dat,
    ibi_col = "IBI",
    group_cols = "participant",
    mse_scales = 1:3
  )

  expect_s3_class(out, "gazepoint_hrv_nonlinear")
  expect_true("approximate_entropy" %in% names(out$features))
  expect_true("dfa_alpha" %in% names(out$features))
  expect_true("mse_mean" %in% names(out$features))
  expect_true("mse_scale_1" %in% names(out$features))
})

test_that("flag_gazepoint_mad_artifacts flags wearable artifact typology", {
  x <- c(
    seq(1, 1.2, length.out = 10),
    rep(1.2, 8),
    5,
    1.3,
    1.31,
    8,
    8.1,
    8.2
  )

  dat <- data.frame(
    participant = "p1",
    time = seq_along(x),
    GSR_US = x
  )

  out <- flag_gazepoint_mad_artifacts(
    dat,
    eda_col = "GSR_US",
    time_col = "time",
    group_cols = "participant",
    mad_multiplier = 4,
    flatline_tolerance = 1e-8,
    flatline_min_run = 4
  )

  expect_s3_class(out, "gazepoint_mad_artifact_flags")
  expect_true("mad_artifact" %in% names(out))
  expect_true(any(out$mad_artifact))
  expect_true(any(out$mad_artifact_type %in% c("flatline", "needle", "step", "wall", "multiple")))
})

test_that("prepare_gazepoint_pspm_dcm_input creates signal and event tables", {
  dat <- data.frame(
    participant = "p1",
    session = "s1",
    time = seq(0, 10, by = 0.5),
    GSR_US = 1 + sin(seq(0, 10, by = 0.5)) * 0.1,
    event_onset = rep(c(2, 6, NA), length.out = 21),
    event_duration = 0,
    condition = rep(c("A", "B", "none"), length.out = 21)
  )

  out <- prepare_gazepoint_pspm_dcm_input(
    dat,
    eda_col = "GSR_US",
    time_col = "time",
    event_onset_col = "event_onset",
    event_duration_col = "event_duration",
    event_name_col = "condition",
    participant_col = "participant",
    session_col = "session"
  )

  expect_s3_class(out, "gazepoint_pspm_dcm_input")
  expect_true(nrow(out$signal_table) > 0)
  expect_true(nrow(out$event_table) > 0)
  expect_equal(out$overview$status, "pspm_dcm_input_prepared")
})

#' Create Gazepoint Biometrics report tables
#'
#' Creates compact report-ready tables from a Gazepoint Biometrics workflow
#' object or from separate workflow components. The function does not write
#' files. It returns cleaned tables that can be printed, exported, or inserted
#' into reports and supplementary materials.
#'
#' @param workflow Optional workflow object produced by
#'   `run_gazepoint_biometrics_workflow()`.
#' @param validation Optional validation object produced by
#'   `validate_gazepoint_biometrics()`.
#' @param quality Optional quality-audit table.
#' @param sampling Optional sampling/timing audit table produced by
#'   `audit_gazepoint_biometric_sampling()`.
#' @param diagnostics Optional diagnostic table produced by
#'   `diagnose_gazepoint_biometrics_workflow()`.
#' @param exclusion_recommendations Optional object produced by
#'   `recommend_gazepoint_biometric_exclusions()`.
#' @param ttl_events Optional TTL event table produced by
#'   `extract_gazepoint_ttl_events()`.
#' @param max_ttl_events Maximum number of TTL events to include in the compact
#'   TTL event table.
#'
#' @return A list of report-ready tables.
#'
#' @export
create_gazepoint_biometrics_report_tables <- function(workflow = NULL,
                                                      validation = NULL,
                                                      quality = NULL,
                                                      sampling = NULL,
                                                      diagnostics = NULL,
                                                      exclusion_recommendations = NULL,
                                                      ttl_events = NULL,
                                                      max_ttl_events = 20) {
  if (!is.null(workflow)) {
    if (!inherits(workflow, "gazepoint_biometrics_workflow")) {
      stop(
        "`workflow` must be produced by run_gazepoint_biometrics_workflow().",
        call. = FALSE
      )
    }

    validation <- workflow$validation
    quality <- workflow$quality
    sampling <- workflow$sampling
    diagnostics <- diagnose_gazepoint_biometrics_workflow(workflow)
    exclusion_recommendations <- workflow$exclusion_recommendations
    ttl_events <- workflow$ttl_events
    overview <- workflow$overview
  } else {
    overview <- NULL
  }

  out <- list(
    overview = create_report_overview_table(overview, validation),
    diagnostics = create_report_diagnostics_table(diagnostics),
    channels = create_report_channel_table(validation),
    quality = create_report_quality_table(quality),
    sampling = create_report_sampling_table(sampling),
    window_recommendations = create_report_window_recommendation_table(
      exclusion_recommendations
    ),
    participant_recommendations = create_report_participant_recommendation_table(
      exclusion_recommendations
    ),
    ttl_events = create_report_ttl_event_table(
      ttl_events,
      max_ttl_events = max_ttl_events
    )
  )

  class(out) <- c("gazepoint_biometrics_report_tables", "list")
  out
}


create_report_overview_table <- function(overview,
                                         validation) {
  if (!is.null(overview) && is.data.frame(overview)) {
    return(overview)
  }

  if (!is.null(validation) && is.list(validation) && is.data.frame(validation$overview)) {
    return(validation$overview)
  }

  data.frame(
    message = "No overview information supplied.",
    stringsAsFactors = FALSE
  )
}


create_report_diagnostics_table <- function(diagnostics) {
  if (is.null(diagnostics) || !is.data.frame(diagnostics)) {
    return(data.frame(
      message = "No workflow diagnostics table supplied.",
      stringsAsFactors = FALSE
    ))
  }

  keep_columns <- intersect(
    c(
      "final_status",
      "diagnostic_reasons",
      "validation_issue_count",
      "active_gsr_eda",
      "active_heart_rate",
      "active_engagement_dial",
      "active_ttl_marker",
      "low_quality_signal_count",
      "sampling_problem_rows",
      "n_windows",
      "keep_windows",
      "review_windows",
      "exclude_windows",
      "exclude_window_pct",
      "review_window_pct",
      "ttl_event_count"
    ),
    names(diagnostics)
  )

  diagnostics[, keep_columns, drop = FALSE]
}


create_report_channel_table <- function(validation) {
  if (is.null(validation) ||
      !is.list(validation) ||
      !is.data.frame(validation$active_channels)) {
    return(data.frame(
      message = "No active-channel table supplied.",
      stringsAsFactors = FALSE
    ))
  }

  channels <- validation$active_channels

  keep_columns <- intersect(
    c(
      "signal",
      "present",
      "active",
      "summary_column",
      "validity_columns",
      "valid_rows",
      "nonzero_rows",
      "min_value",
      "max_value"
    ),
    names(channels)
  )

  channels[, keep_columns, drop = FALSE]
}


create_report_quality_table <- function(quality) {
  if (is.null(quality) || !is.data.frame(quality)) {
    return(data.frame(
      message = "No quality-audit table supplied.",
      stringsAsFactors = FALSE
    ))
  }

  keep_columns <- intersect(
    c(
      "signal",
      "issue",
      "value_column",
      "validity_column",
      "n_rows",
      "zero_rows",
      "zero_pct",
      "valid_rows",
      "invalid_rows",
      "usable_rows",
      "usable_pct",
      "min_value",
      "max_value",
      "mean_value",
      "flatline"
    ),
    names(quality)
  )

  quality[, keep_columns, drop = FALSE]
}


create_report_sampling_table <- function(sampling) {
  if (is.null(sampling) || !is.data.frame(sampling)) {
    return(data.frame(
      message = "No sampling/timing audit table supplied.",
      stringsAsFactors = FALSE
    ))
  }

  keep_columns <- intersect(
    c(
      "source_file",
      "source_participant",
      "USER",
      "USERID",
      "participant",
      "subject",
      "MEDIA_ID",
      "MEDIA_NAME",
      "group",
      "time_column",
      "time_unit",
      "n_rows",
      "missing_time_rows",
      "missing_time_pct",
      "duplicate_time_rows",
      "zero_interval_rows",
      "negative_interval_rows",
      "monotonic_non_decreasing",
      "strictly_increasing",
      "median_interval_seconds",
      "estimated_rate_hz",
      "expected_rate_hz",
      "rate_deviation_hz",
      "rate_status"
    ),
    names(sampling)
  )

  sampling[, keep_columns, drop = FALSE]
}


create_report_window_recommendation_table <- function(exclusion_recommendations) {
  if (is.null(exclusion_recommendations) ||
      !is.list(exclusion_recommendations) ||
      !is.data.frame(exclusion_recommendations$window_recommendations)) {
    return(data.frame(
      message = "No window-level exclusion recommendations supplied.",
      stringsAsFactors = FALSE
    ))
  }

  windows <- exclusion_recommendations$window_recommendations

  keep_columns <- intersect(
    c(
      "source_participant",
      "USER",
      "USERID",
      "participant",
      "subject",
      "MEDIA_ID",
      "MEDIA_NAME",
      "gsr_usable_pct",
      "hr_usable_pct",
      "dial_usable_pct",
      "recommendation",
      "recommendation_reason"
    ),
    names(windows)
  )

  windows[, keep_columns, drop = FALSE]
}


create_report_participant_recommendation_table <- function(exclusion_recommendations) {
  if (is.null(exclusion_recommendations) ||
      !is.list(exclusion_recommendations) ||
      !is.data.frame(exclusion_recommendations$participant_recommendations)) {
    return(data.frame(
      message = "No participant-level exclusion recommendations supplied.",
      stringsAsFactors = FALSE
    ))
  }

  exclusion_recommendations$participant_recommendations
}


create_report_ttl_event_table <- function(ttl_events,
                                          max_ttl_events = 20) {
  if (is.null(ttl_events) || !is.data.frame(ttl_events)) {
    return(data.frame(
      message = "No TTL event table supplied.",
      stringsAsFactors = FALSE
    ))
  }

  keep_columns <- intersect(
    c(
      "source_participant",
      "USER",
      "USERID",
      "MEDIA_ID",
      "MEDIA_NAME",
      "row_index",
      "event_order",
      "ttl_channel",
      "ttl_value",
      "previous_ttl_value",
      "CNT",
      "TIME",
      "TIME_TICK",
      "ttl_validity"
    ),
    names(ttl_events)
  )

  ttl_events <- ttl_events[, keep_columns, drop = FALSE]

  if (!is.null(max_ttl_events) &&
      is.finite(max_ttl_events) &&
      nrow(ttl_events) > max_ttl_events) {
    ttl_events <- ttl_events[seq_len(max_ttl_events), , drop = FALSE]
  }

  ttl_events
}

#' Extract nonlinear HRV features from IBI/RR intervals
#'
#' Computes dependency-light nonlinear HRV descriptors from IBI/RR intervals,
#' including Poincare SD1/SD2, sample entropy, approximate entropy,
#' multiscale entropy, and detrended fluctuation analysis.
#'
#' These are variability and complexity descriptors. They should not be
#' interpreted as direct emotion, cognitive-load, health-status, or diagnostic
#' labels by themselves.
#'
#' @param dat A data frame containing IBI/RR intervals.
#' @param ibi_col IBI/RR interval column.
#' @param group_cols Optional grouping columns.
#' @param min_intervals Minimum finite intervals per group.
#' @param sampen_m Embedding dimension for sample entropy.
#' @param sampen_r_multiplier Tolerance multiplier applied to the within-group SD.
#' @param mse_scales Integer scales used for multiscale entropy.
#'
#' @return A list with `overview`, `features`, and `settings`.
#' @export
extract_gazepoint_hrv_nonlinear <- function(dat,
                                            ibi_col = "IBI",
                                            group_cols = NULL,
                                            min_intervals = 10,
                                            sampen_m = 2,
                                            sampen_r_multiplier = 0.2,
                                            mse_scales = 1:5) {
  if (!is.data.frame(dat)) {
    stop("`dat` must be a data frame.", call. = FALSE)
  }

  if (!ibi_col %in% names(dat)) {
    stop("Column `", ibi_col, "` was not found in `dat`.", call. = FALSE)
  }

  if (!is.numeric(dat[[ibi_col]])) {
    stop("`ibi_col` must identify a numeric column.", call. = FALSE)
  }

  if (!is.numeric(mse_scales) ||
      length(mse_scales) < 1 ||
      any(!is.finite(mse_scales)) ||
      any(mse_scales < 1)) {
    stop("`mse_scales` must be positive integer-like values.", call. = FALSE)
  }

  mse_scales <- sort(unique(as.integer(mse_scales)))

  group_cols <- gpbiometrics_complexity_group_cols(dat, group_cols)
  groups <- gpbiometrics_complexity_split_indices(dat, group_cols)

  rows <- lapply(names(groups), function(unit_id) {
    idx <- groups[[unit_id]]
    base <- gpbiometrics_complexity_unit_values(dat, idx, group_cols, unit_id)

    x <- dat[[ibi_col]][idx]
    x <- x[is.finite(x) & x > 0]

    mse_names <- paste0("mse_scale_", mse_scales)

    if (length(x) < min_intervals) {
      mse_empty <- as.data.frame(as.list(rep(NA_real_, length(mse_names))))
      names(mse_empty) <- mse_names

      return(data.frame(
        base,
        unit_id = unit_id,
        n_intervals = length(x),
        mean_ibi = NA_real_,
        sdnn = NA_real_,
        rmssd = NA_real_,
        sd1 = NA_real_,
        sd2 = NA_real_,
        sd1_sd2_ratio = NA_real_,
        sample_entropy = NA_real_,
        approximate_entropy = NA_real_,
        dfa_alpha = NA_real_,
        mse_mean = NA_real_,
        mse_empty,
        status = "insufficient_intervals",
        stringsAsFactors = FALSE,
        check.names = FALSE
      ))
    }

    diff_x <- diff(x)
    sdnn <- stats::sd(x)
    diff_var <- stats::var(diff_x, na.rm = TRUE)

    rmssd <- sqrt(mean(diff_x^2, na.rm = TRUE))
    sd1 <- sqrt(diff_var / 2)
    sd2 <- sqrt(max((2 * sdnn^2) - (0.5 * diff_var), 0))
    ratio <- if (is.finite(sd2) && sd2 > 0) sd1 / sd2 else NA_real_

    sampen_r <- sampen_r_multiplier * stats::sd(x)

    sampen <- gpbiometrics_sample_entropy(
      x,
      m = sampen_m,
      r = sampen_r
    )

    apen <- gpbiometrics_approximate_entropy(
      x,
      m = sampen_m,
      r = sampen_r
    )

    dfa <- gpbiometrics_dfa_alpha(x)

    mse <- gpbiometrics_multiscale_entropy(
      x,
      scales = mse_scales,
      m = sampen_m,
      r_multiplier = sampen_r_multiplier
    )

    mse_df <- as.data.frame(as.list(mse))
    names(mse_df) <- mse_names

    data.frame(
      base,
      unit_id = unit_id,
      n_intervals = length(x),
      mean_ibi = mean(x),
      sdnn = sdnn,
      rmssd = rmssd,
      sd1 = sd1,
      sd2 = sd2,
      sd1_sd2_ratio = ratio,
      sample_entropy = sampen,
      approximate_entropy = apen,
      dfa_alpha = dfa,
      mse_mean = if (any(is.finite(mse))) mean(mse, na.rm = TRUE) else NA_real_,
      mse_df,
      status = "nonlinear_hrv_extracted",
      stringsAsFactors = FALSE,
      check.names = FALSE
    )
  })

  features <- do.call(rbind, rows)
  rownames(features) <- NULL

  overview <- data.frame(
    group_count = length(groups),
    feature_rows = nrow(features),
    successful_groups = sum(features$status == "nonlinear_hrv_extracted"),
    problem_groups = sum(features$status != "nonlinear_hrv_extracted"),
    ibi_col = ibi_col,
    status = if (all(features$status == "nonlinear_hrv_extracted")) {
      "nonlinear_hrv_extracted"
    } else if (any(features$status == "nonlinear_hrv_extracted")) {
      "nonlinear_hrv_partial"
    } else {
      "nonlinear_hrv_failed"
    },
    interpretation = paste(
      "Nonlinear HRV features describe interval variability, irregularity, and scale-dependent structure.",
      "They do not infer emotion, cognitive load, psychiatric status, health status, or diagnosis by themselves."
    ),
    stringsAsFactors = FALSE
  )

  structure(
    list(
      overview = overview,
      features = features,
      settings = list(
        ibi_col = ibi_col,
        group_cols = group_cols,
        min_intervals = min_intervals,
        sampen_m = sampen_m,
        sampen_r_multiplier = sampen_r_multiplier,
        mse_scales = mse_scales
      )
    ),
    class = c("gazepoint_hrv_nonlinear", "list")
  )
}

#' Extract EDA complexity features
#'
#' Computes dependency-light EDA complexity descriptors, including sample entropy
#' and detrended fluctuation analysis alpha.
#'
#' @param dat A data frame containing EDA data.
#' @param eda_col EDA/conductance column.
#' @param group_cols Optional grouping columns.
#' @param min_samples Minimum finite samples per group.
#' @param sampen_m Embedding dimension for sample entropy.
#' @param sampen_r_multiplier Tolerance multiplier applied to within-group SD.
#'
#' @return A list with `overview`, `features`, and `settings`.
#' @export
extract_gazepoint_eda_complexity <- function(dat,
                                             eda_col = "GSR_US",
                                             group_cols = NULL,
                                             min_samples = 32,
                                             sampen_m = 2,
                                             sampen_r_multiplier = 0.2) {
  if (!is.data.frame(dat)) stop("`dat` must be a data frame.", call. = FALSE)
  if (!eda_col %in% names(dat)) stop("Column `", eda_col, "` was not found in `dat`.", call. = FALSE)
  if (!is.numeric(dat[[eda_col]])) stop("`eda_col` must identify a numeric column.", call. = FALSE)

  group_cols <- gpbiometrics_complexity_group_cols(dat, group_cols)
  groups <- gpbiometrics_complexity_split_indices(dat, group_cols)

  rows <- lapply(names(groups), function(unit_id) {
    idx <- groups[[unit_id]]
    base <- gpbiometrics_complexity_unit_values(dat, idx, group_cols, unit_id)
    x <- dat[[eda_col]][idx]
    x <- x[is.finite(x)]

    if (length(x) < min_samples || stats::sd(x) == 0) {
      return(data.frame(
        base,
        unit_id = unit_id,
        n_samples = length(x),
        signal_sd = if (length(x) > 1) stats::sd(x) else NA_real_,
        sample_entropy = NA_real_,
        dfa_alpha = NA_real_,
        status = "insufficient_or_constant_signal",
        stringsAsFactors = FALSE,
        check.names = FALSE
      ))
    }

    se <- gpbiometrics_sample_entropy(x, m = sampen_m, r = sampen_r_multiplier * stats::sd(x))
    dfa <- gpbiometrics_dfa_alpha(x)

    data.frame(
      base,
      unit_id = unit_id,
      n_samples = length(x),
      signal_sd = stats::sd(x),
      sample_entropy = se,
      dfa_alpha = dfa,
      status = "eda_complexity_extracted",
      stringsAsFactors = FALSE,
      check.names = FALSE
    )
  })

  features <- do.call(rbind, rows)
  rownames(features) <- NULL

  overview <- data.frame(
    group_count = length(groups),
    feature_rows = nrow(features),
    successful_groups = sum(features$status == "eda_complexity_extracted"),
    problem_groups = sum(features$status != "eda_complexity_extracted"),
    eda_col = eda_col,
    status = if (all(features$status == "eda_complexity_extracted")) {
      "eda_complexity_extracted"
    } else if (any(features$status == "eda_complexity_extracted")) {
      "eda_complexity_partial"
    } else {
      "eda_complexity_failed"
    },
    interpretation = paste(
      "EDA complexity features describe nonlinear or scale-dependent signal structure.",
      "They are not direct emotion, stress, cognition, trust, preference, or diagnosis labels."
    ),
    stringsAsFactors = FALSE
  )

  structure(
    list(overview = overview, features = features, settings = list(
      eda_col = eda_col,
      group_cols = group_cols,
      min_samples = min_samples,
      sampen_m = sampen_m,
      sampen_r_multiplier = sampen_r_multiplier
    )),
    class = c("gazepoint_eda_complexity", "list")
  )
}

gpbiometrics_sample_entropy <- function(x, m = 2, r) {
  x <- x[is.finite(x)]

  if (length(x) <= m + 2 || !is.finite(r) || r <= 0) return(NA_real_)

  count_matches <- function(mm) {
    n <- length(x) - mm + 1
    if (n <= 1) return(NA_real_)

    emb <- matrix(NA_real_, nrow = n, ncol = mm)
    for (j in seq_len(mm)) emb[, j] <- x[j:(j + n - 1)]

    count <- 0
    total <- 0

    for (i in seq_len(n - 1)) {
      for (k in (i + 1):n) {
        total <- total + 1
        if (max(abs(emb[i, ] - emb[k, ])) <= r) count <- count + 1
      }
    }

    if (total == 0) return(NA_real_)
    count / total
  }

  a <- count_matches(m + 1)
  b <- count_matches(m)

  if (!is.finite(a) || !is.finite(b) || a <= 0 || b <= 0) return(NA_real_)
  -log(a / b)
}

gpbiometrics_dfa_alpha <- function(x) {
  x <- x[is.finite(x)]

  n <- length(x)
  if (n < 32) return(NA_real_)

  y <- cumsum(x - mean(x))
  scales <- unique(floor(exp(seq(log(4), log(max(8, floor(n / 4))), length.out = 8))))
  scales <- scales[scales >= 4 & scales < n / 2]

  if (length(scales) < 3) return(NA_real_)

  fluct <- vapply(scales, function(s) {
    starts <- seq(1, n - s + 1, by = s)
    rms <- vapply(starts, function(st) {
      seg <- y[st:(st + s - 1)]
      fit <- stats::lm(seg ~ seq_along(seg))
      sqrt(mean(stats::resid(fit)^2))
    }, numeric(1))

    mean(rms, na.rm = TRUE)
  }, numeric(1))

  keep <- is.finite(fluct) & fluct > 0 & is.finite(scales)
  if (sum(keep) < 3) return(NA_real_)

  unname(stats::coef(stats::lm(log(fluct[keep]) ~ log(scales[keep])))[2])
}

gpbiometrics_complexity_group_cols <- function(dat, group_cols = NULL) {
  if (is.null(group_cols)) return(character())
  missing_cols <- setdiff(group_cols, names(dat))
  if (length(missing_cols) > 0) {
    stop("Missing `group_cols`: ", paste(missing_cols, collapse = ", "), call. = FALSE)
  }
  group_cols
}

gpbiometrics_complexity_split_indices <- function(dat, group_cols) {
  if (length(group_cols) == 0) return(list(all_rows = seq_len(nrow(dat))))
  gf <- dat[group_cols]
  gf[] <- lapply(gf, function(x) {
    x <- as.character(x)
    x[is.na(x)] <- "<NA>"
    x
  })
  split(seq_len(nrow(dat)), do.call(paste, c(gf, sep = " | ")))
}

gpbiometrics_complexity_unit_values <- function(dat, idx, group_cols, unit_id) {
  if (length(group_cols) == 0) return(data.frame(unit_label = unit_id, stringsAsFactors = FALSE))
  values <- lapply(group_cols, function(nm) as.character(dat[[nm]][idx[1]]))
  names(values) <- group_cols
  as.data.frame(values, stringsAsFactors = FALSE, optional = TRUE)
}

gpbiometrics_approximate_entropy <- function(x, m = 2, r) {
  x <- x[is.finite(x)]

  if (length(x) <= m + 2 || !is.finite(r) || r <= 0) {
    return(NA_real_)
  }

  phi <- function(mm) {
    n <- length(x) - mm + 1

    if (n <= 1) {
      return(NA_real_)
    }

    emb <- matrix(NA_real_, nrow = n, ncol = mm)

    for (j in seq_len(mm)) {
      emb[, j] <- x[j:(j + n - 1)]
    }

    c_i <- numeric(n)

    for (i in seq_len(n)) {
      dist <- apply(emb, 1, function(row) max(abs(row - emb[i, ])))
      c_i[i] <- mean(dist <= r)
    }

    mean(log(pmax(c_i, .Machine$double.eps)))
  }

  phi_m <- phi(m)
  phi_m1 <- phi(m + 1)

  if (!is.finite(phi_m) || !is.finite(phi_m1)) {
    return(NA_real_)
  }

  phi_m - phi_m1
}

gpbiometrics_multiscale_entropy <- function(x,
                                            scales = 1:5,
                                            m = 2,
                                            r_multiplier = 0.2) {
  x <- x[is.finite(x)]

  vapply(scales, function(scale) {
    coarse <- gpbiometrics_coarse_grain(x, scale)

    if (length(coarse) <= m + 2 || stats::sd(coarse) == 0) {
      return(NA_real_)
    }

    gpbiometrics_sample_entropy(
      coarse,
      m = m,
      r = r_multiplier * stats::sd(coarse)
    )
  }, numeric(1))
}

gpbiometrics_coarse_grain <- function(x, scale) {
  scale <- as.integer(scale)

  if (scale <= 1) {
    return(x)
  }

  n_blocks <- floor(length(x) / scale)

  if (n_blocks < 1) {
    return(numeric())
  }

  vapply(seq_len(n_blocks), function(i) {
    start <- (i - 1) * scale + 1
    end <- i * scale
    mean(x[start:end], na.rm = TRUE)
  }, numeric(1))
}

test_that("extract_gazepoint_eda_spectral_power returns band-power summaries", {
  dat <- data.frame(
    participant = "p1",
    time = seq(0, 127),
    GSR_US = sin(2 * pi * 0.1 * seq(0, 127)) + rnorm(128, sd = 0.01)
  )

  out <- extract_gazepoint_eda_spectral_power(
    dat,
    eda_col = "GSR_US",
    time_col = "time",
    group_cols = "participant",
    sampling_rate = 1
  )

  expect_s3_class(out, "gazepoint_eda_spectral_power")
  expect_equal(out$overview$status, "eda_spectral_power_extracted")
  expect_true(is.finite(out$spectral_summary$band_power))
  expect_equal(out$spectral_summary$band_lower_hz, 0.045)
  expect_equal(out$spectral_summary$band_upper_hz, 0.25)
})

test_that("extract_gazepoint_hrv_nonlinear returns Poincare and entropy features", {
  dat <- data.frame(
    participant = "p1",
    IBI = 0.8 + sin(seq(0, 4 * pi, length.out = 60)) * 0.05
  )

  out <- extract_gazepoint_hrv_nonlinear(
    dat,
    ibi_col = "IBI",
    group_cols = "participant"
  )

  expect_s3_class(out, "gazepoint_hrv_nonlinear")
  expect_equal(out$overview$status, "nonlinear_hrv_extracted")
  expect_true(is.finite(out$features$sd1))
  expect_true(is.finite(out$features$sd2))
})

test_that("extract_gazepoint_eda_complexity returns EDA complexity features", {
  dat <- data.frame(
    participant = "p1",
    GSR_US = sin(seq(0, 8 * pi, length.out = 128)) + rnorm(128, sd = 0.05)
  )

  out <- extract_gazepoint_eda_complexity(
    dat,
    eda_col = "GSR_US",
    group_cols = "participant"
  )

  expect_s3_class(out, "gazepoint_eda_complexity")
  expect_equal(out$overview$status, "eda_complexity_extracted")
  expect_true(is.finite(out$features$dfa_alpha))
})

test_that("denoise_gazepoint_eda_wavelet adds denoised signal", {
  set.seed(1)
  dat <- data.frame(
    participant = "p1",
    GSR_US = sin(seq(0, 4 * pi, length.out = 64)) + rnorm(64, sd = 0.2)
  )

  out <- denoise_gazepoint_eda_wavelet(
    dat,
    eda_col = "GSR_US",
    group_cols = "participant"
  )

  expect_s3_class(out, "gazepoint_eda_wavelet_denoised")
  expect_true("GSR_US_wavelet_denoised" %in% names(out))
  expect_equal(attr(out, "wavelet_denoising_overview")$status, "eda_wavelet_denoising_complete")
})

test_that("audit_gazepoint_distributional_drift returns PSI and KS summaries", {
  dat <- data.frame(
    participant = rep("p1", 100),
    session = rep(c(1, 2), each = 50),
    GSR_US = c(rnorm(50, 1, 0.1), rnorm(50, 2, 0.1))
  )

  out <- audit_gazepoint_distributional_drift(
    dat,
    signal_cols = "GSR_US",
    session_col = "session",
    participant_col = "participant"
  )

  expect_s3_class(out, "gazepoint_distributional_drift")
  expect_true(is.data.frame(out$drift_summary))
  expect_true("psi" %in% names(out$drift_summary))
  expect_true(any(out$drift_summary$comparison_session == "2"))
})

test_that("run_gpbiometrics_shiny_annotator errors clearly without shiny or returns app", {
  if (!requireNamespace("shiny", quietly = TRUE)) {
    expect_error(run_gpbiometrics_shiny_annotator(), "shiny")
  } else {
    app <- run_gpbiometrics_shiny_annotator()
    expect_true(inherits(app, "shiny.appobj"))
  }
})

#' Standardise SCR or SCL using intra-individual z-scoring
#'
#' Adds a within-participant z-scored version of a signal column. This is a
#' lightweight compatibility wrapper around the package's more general
#' within-unit standardisation helper.
#'
#' @param dat A data frame containing SCR, SCL, or another biometric signal.
#' @param signal_col Signal column to standardise.
#' @param group_col Participant or unit column.
#' @param suffix Suffix for the output column.
#' @param min_valid Minimum finite observations required within each group.
#' @param overwrite Logical. If `FALSE`, existing output columns are protected.
#'
#' @return A data frame with an added z-scored column.
#' @export
standardise_gazepoint_zscore <- function(dat,
                                         signal_col = "SCR_Amplitude",
                                         group_col = "source_participant",
                                         suffix = "_Z",
                                         min_valid = 2,
                                         overwrite = FALSE) {
  if (!is.data.frame(dat)) {
    stop("`dat` must be a data frame.", call. = FALSE)
  }

  if (!is.character(signal_col) || length(signal_col) != 1 || is.na(signal_col)) {
    stop("`signal_col` must be a single column name.", call. = FALSE)
  }

  if (!is.character(group_col) || length(group_col) != 1 || is.na(group_col)) {
    stop("`group_col` must be a single column name.", call. = FALSE)
  }

  out <- standardize_gazepoint_biometrics_within_unit(
    data = dat,
    signal_cols = signal_col,
    unit_cols = group_col,
    suffix = suffix,
    center = TRUE,
    scale = TRUE,
    min_valid = min_valid,
    zero_sd_action = "NA",
    overwrite = overwrite
  )

  attr(out, "standardization_method") <- "intra_individual_z_score"
  attr(out, "interpretation") <- paste(
    "Intra-individual z-scoring expresses each observation relative to the participant's own mean and standard deviation.",
    "It supports within-participant comparison but removes between-participant level and scale differences."
  )

  out
}

#' @rdname standardise_gazepoint_zscore
#' @export
standardize_gazepoint_zscore <- function(dat,
                                         signal_col = "SCR_Amplitude",
                                         group_col = "source_participant",
                                         suffix = "_Z",
                                         min_valid = 2,
                                         overwrite = FALSE) {
  standardise_gazepoint_zscore(
    dat = dat,
    signal_col = signal_col,
    group_col = group_col,
    suffix = suffix,
    min_valid = min_valid,
    overwrite = overwrite
  )
}

#' Standardise SCR or SCL using within-participant range correction
#'
#' Adds a range-corrected signal column using `(x - min) / (max - min)` within
#' participant or another grouping unit. This expresses each value as a
#' proportion of the observed within-unit range.
#'
#' @param dat A data frame containing SCR, SCL, or another biometric signal.
#' @param signal_col Signal column to range-correct.
#' @param group_col Participant or unit column.
#' @param suffix Suffix for the output column.
#' @param min_valid Minimum finite observations required within each group.
#' @param zero_range_action What to do when max equals min: `"NA"` or `"zero"`.
#' @param overwrite Logical. If `FALSE`, existing output columns are protected.
#'
#' @return A data frame with an added range-corrected column.
#' @export
standardise_gazepoint_range_correction <- function(dat,
                                                   signal_col,
                                                   group_col = "source_participant",
                                                   suffix = "_Range_Corrected",
                                                   min_valid = 2,
                                                   zero_range_action = c("NA", "zero"),
                                                   overwrite = FALSE) {
  if (!is.data.frame(dat)) {
    stop("`dat` must be a data frame.", call. = FALSE)
  }

  zero_range_action <- match.arg(zero_range_action)

  if (!is.character(signal_col) || length(signal_col) != 1 || is.na(signal_col)) {
    stop("`signal_col` must be a single column name.", call. = FALSE)
  }

  if (!signal_col %in% names(dat)) {
    stop("Column `", signal_col, "` was not found in `dat`.", call. = FALSE)
  }

  if (!is.numeric(dat[[signal_col]])) {
    stop("`signal_col` must identify a numeric column.", call. = FALSE)
  }

  if (!is.character(group_col) || length(group_col) != 1 || is.na(group_col)) {
    stop("`group_col` must be a single column name.", call. = FALSE)
  }

  if (!group_col %in% names(dat)) {
    stop("Column `", group_col, "` was not found in `dat`.", call. = FALSE)
  }

  output_col <- paste0(signal_col, suffix)

  if (!isTRUE(overwrite) && output_col %in% names(dat)) {
    stop(
      "Output column `", output_col, "` already exists. Use `overwrite = TRUE` to replace it.",
      call. = FALSE
    )
  }

  out <- dat
  out[[output_col]] <- NA_real_

  groups <- split(seq_len(nrow(dat)), dat[[group_col]], drop = TRUE)

  parameter_rows <- lapply(names(groups), function(unit_id) {
    idx <- groups[[unit_id]]
    x <- dat[[signal_col]][idx]
    finite_x <- x[is.finite(x)]

    n_finite <- length(finite_x)

    min_val <- NA_real_
    max_val <- NA_real_
    range_val <- NA_real_
    status <- "range_corrected"

    if (n_finite < min_valid) {
      status <- "insufficient_finite_rows"
      out[[output_col]][idx] <<- NA_real_
    } else {
      min_val <- min(finite_x)
      max_val <- max(finite_x)
      range_val <- max_val - min_val

      if (!is.finite(range_val) || range_val == 0) {
        status <- "zero_or_missing_range"

        if (identical(zero_range_action, "zero")) {
          corrected <- rep(NA_real_, length(x))
          corrected[is.finite(x)] <- 0
          out[[output_col]][idx] <<- corrected
        } else {
          out[[output_col]][idx] <<- NA_real_
        }
      } else {
        corrected <- (x - min_val) / range_val
        corrected[!is.finite(corrected)] <- NA_real_
        out[[output_col]][idx] <<- corrected
      }
    }

    data.frame(
      unit_id = as.character(unit_id),
      signal_col = signal_col,
      output_col = output_col,
      n_rows = length(idx),
      n_finite = n_finite,
      min_val = min_val,
      max_val = max_val,
      range_val = range_val,
      status = status,
      stringsAsFactors = FALSE
    )
  })

  parameters <- do.call(rbind, parameter_rows)
  rownames(parameters) <- NULL

  summary <- data.frame(
    input_rows = nrow(dat),
    group_count = length(groups),
    signal_col = signal_col,
    output_col = output_col,
    corrected_groups = sum(parameters$status == "range_corrected"),
    problem_groups = sum(parameters$status != "range_corrected"),
    status = if (all(parameters$status == "range_corrected")) {
      "range_correction_complete"
    } else if (any(parameters$status == "range_corrected")) {
      "range_correction_partial"
    } else {
      "range_correction_failed"
    },
    interpretation = paste(
      "Range correction expresses each value as a proportion of the observed within-unit signal range.",
      "It reduces between-unit range differences but depends strongly on the observed minimum and maximum."
    ),
    stringsAsFactors = FALSE
  )

  attr(out, "range_correction_summary") <- summary
  attr(out, "range_correction_parameters") <- parameters
  attr(out, "range_correction_settings") <- list(
    signal_col = signal_col,
    group_col = group_col,
    suffix = suffix,
    min_valid = min_valid,
    zero_range_action = zero_range_action,
    overwrite = overwrite
  )

  class(out) <- unique(c("gazepoint_range_corrected", class(out)))

  out
}

#' @rdname standardise_gazepoint_range_correction
#' @export
standardize_gazepoint_range_correction <- function(dat,
                                                   signal_col,
                                                   group_col = "source_participant",
                                                   suffix = "_Range_Corrected",
                                                   min_valid = 2,
                                                   zero_range_action = c("NA", "zero"),
                                                   overwrite = FALSE) {
  standardise_gazepoint_range_correction(
    dat = dat,
    signal_col = signal_col,
    group_col = group_col,
    suffix = suffix,
    min_valid = min_valid,
    zero_range_action = zero_range_action,
    overwrite = overwrite
  )
}

test_that("standardise_gazepoint_zscore creates within-participant z scores", {
  dat <- data.frame(
    source_participant = rep(c("p1", "p2"), each = 5),
    SCR_Amplitude = c(1, 2, 3, 4, 5, 10, 20, 30, 40, 50)
  )

  out <- standardise_gazepoint_zscore(dat)

  expect_true("SCR_Amplitude_Z" %in% names(out))

  by_id <- split(out, out$source_participant)

  expect_equal(mean(by_id$p1$SCR_Amplitude_Z), 0, tolerance = 1e-10)
  expect_equal(stats::sd(by_id$p1$SCR_Amplitude_Z), 1, tolerance = 1e-10)
  expect_equal(mean(by_id$p2$SCR_Amplitude_Z), 0, tolerance = 1e-10)
  expect_equal(stats::sd(by_id$p2$SCR_Amplitude_Z), 1, tolerance = 1e-10)
})

test_that("standardise_gazepoint_range_correction rescales within participant", {
  dat <- data.frame(
    source_participant = rep(c("p1", "p2"), each = 3),
    SCR_Amplitude = c(1, 2, 3, 10, 20, 30)
  )

  out <- standardise_gazepoint_range_correction(
    dat,
    signal_col = "SCR_Amplitude"
  )

  expect_true("SCR_Amplitude_Range_Corrected" %in% names(out))

  p1 <- out$SCR_Amplitude_Range_Corrected[out$source_participant == "p1"]
  p2 <- out$SCR_Amplitude_Range_Corrected[out$source_participant == "p2"]

  expect_equal(p1, c(0, 0.5, 1))
  expect_equal(p2, c(0, 0.5, 1))

  summary <- attr(out, "range_correction_summary")
  expect_equal(summary$status, "range_correction_complete")
})

test_that("standardise_gazepoint_range_correction handles zero ranges", {
  dat <- data.frame(
    source_participant = "p1",
    SCR_Amplitude = rep(1, 4)
  )

  out <- standardise_gazepoint_range_correction(
    dat,
    signal_col = "SCR_Amplitude"
  )

  expect_true(all(is.na(out$SCR_Amplitude_Range_Corrected)))
  expect_equal(attr(out, "range_correction_summary")$status, "range_correction_failed")
})

#' Audit IBI/RR interval quality
#'
#' Audits inter-beat interval (IBI) or RR-interval values in Gazepoint
#' Biometrics-style exports. This helper deliberately does not use raw `HRV`
#' columns as HRV metrics. HRV-style summaries should be derived only from
#' genuine IBI/RR interval columns.
#'
#' @param data A data frame.
#' @param ibi_col Optional IBI/RR interval column. If `NULL`, the function
#'   detects recognised IBI/RR-style column names.
#' @param group_cols Optional grouping columns, such as participant, trial,
#'   stimulus, condition, or window labels.
#' @param time_col Optional time/order column used to order samples before
#'   successive-difference checks.
#' @param unit Unit of the IBI values. `"auto"` treats median values below 10 as
#'   seconds and larger values as milliseconds.
#' @param min_ibi_ms Minimum plausible IBI in milliseconds.
#' @param max_ibi_ms Maximum plausible IBI in milliseconds.
#' @param max_jump_ms Maximum plausible absolute change between successive IBI
#'   values within a group.
#'
#' @return A list with `overview`, `samples`, `group_summary`, and `settings`.
#'
#' @examples
#' df <- data.frame(
#'   USER = rep(c("P1", "P2"), each = 4),
#'   IBI = c(800, 810, 790, 805, 900, 910, 905, 920)
#' )
#' audit_gazepoint_ibi_quality(df, group_cols = "USER")
#'
#' @export
audit_gazepoint_ibi_quality <- function(data,
                                        ibi_col = NULL,
                                        group_cols = NULL,
                                        time_col = NULL,
                                        unit = c("auto", "milliseconds", "seconds"),
                                        min_ibi_ms = 300,
                                        max_ibi_ms = 2000,
                                        max_jump_ms = 500) {
  if (!is.data.frame(data)) {
    stop("`data` must be a data frame.", call. = FALSE)
  }

  unit <- match.arg(unit)

  if (!is.null(ibi_col)) {
    .gpbiom_assert_columns(data, ibi_col, "ibi_col")
  }

  if (!is.null(group_cols)) {
    .gpbiom_assert_columns(data, group_cols, "group_cols")
  }

  if (!is.null(time_col)) {
    .gpbiom_assert_columns(data, time_col, "time_col")
  }

  .gpbiom_assert_positive_number(min_ibi_ms, "min_ibi_ms")
  .gpbiom_assert_positive_number(max_ibi_ms, "max_ibi_ms")
  .gpbiom_assert_positive_number(max_jump_ms, "max_jump_ms")

  if (min_ibi_ms >= max_ibi_ms) {
    stop("`min_ibi_ms` must be smaller than `max_ibi_ms`.", call. = FALSE)
  }

  source_col <- if (is.null(ibi_col)) {
    .gpbiom_choose_ibi_column(data)
  } else {
    ibi_col
  }

  if (is.na(source_col)) {
    stop(
      "No IBI/RR interval column was detected. Provide `ibi_col` explicitly.",
      call. = FALSE
    )
  }

  if (!is.numeric(data[[source_col]])) {
    stop("The selected IBI/RR interval column must be numeric.", call. = FALSE)
  }

  resolved_unit <- .gpbiom_resolve_ibi_unit(data[[source_col]], unit)
  ibi_ms <- .gpbiom_ibi_to_ms(data[[source_col]], resolved_unit)

  sample_table <- data.frame(
    row_id = seq_len(nrow(data)),
    group = .gpbiom_group_labels(data, group_cols),
    ibi_raw = data[[source_col]],
    ibi_ms = ibi_ms,
    missing_ibi = is.na(data[[source_col]]),
    nonfinite_ibi = !is.na(data[[source_col]]) & !is.finite(data[[source_col]]),
    nonpositive_ibi = !is.na(ibi_ms) & is.finite(ibi_ms) & ibi_ms <= 0,
    below_min_ibi = !is.na(ibi_ms) & is.finite(ibi_ms) & ibi_ms > 0 &
      ibi_ms < min_ibi_ms,
    above_max_ibi = !is.na(ibi_ms) & is.finite(ibi_ms) &
      ibi_ms > max_ibi_ms,
    large_jump_ibi = FALSE,
    stringsAsFactors = FALSE
  )

  groups <- .gpbiom_group_indices(data, group_cols)

  for (indices in groups) {
    ordered_indices <- .gpbiom_order_indices(data, indices, time_col)
    ordered_ibi <- sample_table$ibi_ms[ordered_indices]

    jump_flags <- .gpbiom_ibi_large_jump_flags(
      values = ordered_ibi,
      max_jump_ms = max_jump_ms
    )

    sample_table$large_jump_ibi[ordered_indices] <- jump_flags
  }

  sample_table$valid_ibi <- !sample_table$missing_ibi &
    !sample_table$nonfinite_ibi &
    !sample_table$nonpositive_ibi &
    !sample_table$below_min_ibi &
    !sample_table$above_max_ibi

  sample_table$any_quality_flag <- sample_table$missing_ibi |
    sample_table$nonfinite_ibi |
    sample_table$nonpositive_ibi |
    sample_table$below_min_ibi |
    sample_table$above_max_ibi |
    sample_table$large_jump_ibi

  sample_table$status <- ifelse(
    sample_table$missing_ibi,
    "missing_ibi",
    ifelse(
      sample_table$nonfinite_ibi,
      "nonfinite_ibi",
      ifelse(
        sample_table$nonpositive_ibi,
        "nonpositive_ibi",
        ifelse(
          sample_table$below_min_ibi,
          "below_min_ibi",
          ifelse(
            sample_table$above_max_ibi,
            "above_max_ibi",
            ifelse(
              sample_table$large_jump_ibi,
              "large_jump_ibi",
              "valid_ibi"
            )
          )
        )
      )
    )
  )

  group_summary <- .gpbiom_ibi_group_summary(
    sample_table = sample_table,
    group_cols = group_cols,
    min_valid_ibi = 2L
  )

  n_valid <- sum(sample_table$valid_ibi, na.rm = TRUE)
  n_quality_flagged <- sum(sample_table$any_quality_flag, na.rm = TRUE)

  overview <- data.frame(
    n_rows = nrow(data),
    ibi_column = source_col,
    unit = resolved_unit,
    group_column_count = length(group_cols),
    n_missing_ibi = sum(sample_table$missing_ibi, na.rm = TRUE),
    n_nonfinite_ibi = sum(sample_table$nonfinite_ibi, na.rm = TRUE),
    n_nonpositive_ibi = sum(sample_table$nonpositive_ibi, na.rm = TRUE),
    n_below_min_ibi = sum(sample_table$below_min_ibi, na.rm = TRUE),
    n_above_max_ibi = sum(sample_table$above_max_ibi, na.rm = TRUE),
    n_large_jump_ibi = sum(sample_table$large_jump_ibi, na.rm = TRUE),
    n_valid_ibi = n_valid,
    valid_ibi_rate = if (nrow(data) > 0L) n_valid / nrow(data) else NA_real_,
    n_quality_flagged = n_quality_flagged,
    quality_flag_rate = if (nrow(data) > 0L) {
      n_quality_flagged / nrow(data)
    } else {
      NA_real_
    },
    status = if (n_valid == 0L) {
      "no_valid_ibi_intervals"
    } else if (n_quality_flagged > 0L) {
      "ibi_quality_issues_detected"
    } else {
      "ibi_quality_ok"
    },
    stringsAsFactors = FALSE
  )

  settings <- list(
    ibi_col = source_col,
    group_cols = group_cols,
    time_col = time_col,
    unit = unit,
    resolved_unit = resolved_unit,
    min_ibi_ms = min_ibi_ms,
    max_ibi_ms = max_ibi_ms,
    max_jump_ms = max_jump_ms,
    note = paste0(
      "IBI quality and HRV-style summaries are based only on the selected ",
      "IBI/RR interval column, not on raw HRV validity/vendor columns."
    )
  )

  list(
    overview = overview,
    samples = sample_table,
    group_summary = group_summary,
    settings = settings
  )
}


#' Summarise IBI/RR windows
#'
#' Computes descriptive IBI/RR interval and simple HRV-style window summaries
#' from genuine inter-beat interval data. The function does not use raw `HRV`
#' columns as HRV metrics. It calculates metrics such as mean IBI, mean
#' instantaneous heart rate, SDNN, RMSSD, pNN20, and pNN50 only from valid IBI/RR
#' intervals.
#'
#' @param data A data frame.
#' @param ibi_col Optional IBI/RR interval column. If `NULL`, the function
#'   detects recognised IBI/RR-style column names.
#' @param group_cols Optional grouping columns defining windows, such as
#'   participant, trial, stimulus, condition, or window labels.
#' @param time_col Optional time/order column used to order IBI values before
#'   successive-difference metrics are computed.
#' @param unit Unit of the IBI values. `"auto"` treats median values below 10 as
#'   seconds and larger values as milliseconds.
#' @param min_ibi_ms Minimum plausible IBI in milliseconds.
#' @param max_ibi_ms Maximum plausible IBI in milliseconds.
#' @param max_jump_ms Maximum plausible absolute change between successive IBI
#'   values within a group.
#' @param exclude_large_jumps Logical. Should intervals flagged as large jumps be
#'   excluded from the window summaries?
#' @param min_valid_ibi Minimum valid IBI count required for a window to be marked
#'   as sufficient.
#'
#' @return A list with `overview`, `windows`, `samples`, and `settings`.
#'
#' @examples
#' df <- data.frame(
#'   USER = rep(c("P1", "P2"), each = 4),
#'   IBI = c(800, 810, 790, 805, 900, 910, 905, 920)
#' )
#' summarise_gazepoint_ibi_windows(df, group_cols = "USER")
#'
#' @export
summarise_gazepoint_ibi_windows <- function(data,
                                            ibi_col = NULL,
                                            group_cols = NULL,
                                            time_col = NULL,
                                            unit = c("auto", "milliseconds", "seconds"),
                                            min_ibi_ms = 300,
                                            max_ibi_ms = 2000,
                                            max_jump_ms = 500,
                                            exclude_large_jumps = TRUE,
                                            min_valid_ibi = 2L) {
  if (!is.data.frame(data)) {
    stop("`data` must be a data frame.", call. = FALSE)
  }

  unit <- match.arg(unit)

  if (!is.null(group_cols)) {
    .gpbiom_assert_columns(data, group_cols, "group_cols")
  }

  if (!is.null(time_col)) {
    .gpbiom_assert_columns(data, time_col, "time_col")
  }

  if (!is.logical(exclude_large_jumps) ||
      length(exclude_large_jumps) != 1L ||
      is.na(exclude_large_jumps)) {
    stop("`exclude_large_jumps` must be TRUE or FALSE.", call. = FALSE)
  }

  .gpbiom_assert_positive_integer(min_valid_ibi, "min_valid_ibi")

  audit <- audit_gazepoint_ibi_quality(
    data = data,
    ibi_col = ibi_col,
    group_cols = group_cols,
    time_col = time_col,
    unit = unit,
    min_ibi_ms = min_ibi_ms,
    max_ibi_ms = max_ibi_ms,
    max_jump_ms = max_jump_ms
  )

  samples <- audit$samples
  samples$analysis_valid_ibi <- samples$valid_ibi

  if (isTRUE(exclude_large_jumps)) {
    samples$analysis_valid_ibi <- samples$analysis_valid_ibi &
      !samples$large_jump_ibi
  }

  windows <- .gpbiom_ibi_window_summary(
    sample_table = samples,
    group_cols = group_cols,
    min_valid_ibi = as.integer(min_valid_ibi)
  )

  n_sufficient <- sum(windows$status == "sufficient_ibi_window", na.rm = TRUE)

  overview <- data.frame(
    n_rows = nrow(data),
    ibi_column = audit$overview$ibi_column,
    unit = audit$overview$unit,
    window_count = nrow(windows),
    sufficient_window_count = n_sufficient,
    insufficient_window_count = nrow(windows) - n_sufficient,
    exclude_large_jumps = exclude_large_jumps,
    min_valid_ibi = as.integer(min_valid_ibi),
    status = if (nrow(windows) == 0L) {
      "no_ibi_windows"
    } else if (n_sufficient == 0L) {
      "no_sufficient_ibi_windows"
    } else if (n_sufficient < nrow(windows)) {
      "some_ibi_windows_insufficient"
    } else {
      "ibi_windows_summarised"
    },
    stringsAsFactors = FALSE
  )

  settings <- list(
    ibi_col = audit$overview$ibi_column,
    group_cols = group_cols,
    time_col = time_col,
    unit = unit,
    resolved_unit = audit$overview$unit,
    min_ibi_ms = min_ibi_ms,
    max_ibi_ms = max_ibi_ms,
    max_jump_ms = max_jump_ms,
    exclude_large_jumps = exclude_large_jumps,
    min_valid_ibi = as.integer(min_valid_ibi),
    note = paste0(
      "Window summaries are derived from genuine IBI/RR intervals only. ",
      "They are not calculated from raw HRV validity/vendor columns."
    )
  )

  list(
    overview = overview,
    windows = windows,
    samples = samples,
    settings = settings
  )
}


.gpbiom_assert_positive_number <- function(x, argument_name) {
  if (!is.numeric(x) ||
      length(x) != 1L ||
      is.na(x) ||
      !is.finite(x) ||
      x <= 0) {
    stop("`", argument_name, "` must be a positive number.", call. = FALSE)
  }

  invisible(TRUE)
}


.gpbiom_choose_ibi_column <- function(data) {
  mapping <- standardise_gazepoint_biometric_names(data, rename = FALSE)
  candidates <- mapping$original_name[mapping$standard_name == "IBI"]

  if (length(candidates) == 0L) {
    return(NA_character_)
  }

  numeric_candidates <- candidates[vapply(candidates, function(column) {
    is.numeric(data[[column]])
  }, logical(1))]

  if (length(numeric_candidates) > 0L) {
    return(numeric_candidates[1L])
  }

  candidates[1L]
}


.gpbiom_resolve_ibi_unit <- function(values, unit) {
  if (identical(unit, "milliseconds")) {
    return("milliseconds")
  }

  if (identical(unit, "seconds")) {
    return("seconds")
  }

  finite <- values[!is.na(values) & is.finite(values) & values > 0]

  if (length(finite) == 0L) {
    return("milliseconds")
  }

  if (stats::median(finite) < 10) {
    "seconds"
  } else {
    "milliseconds"
  }
}


.gpbiom_ibi_to_ms <- function(values, unit) {
  if (identical(unit, "seconds")) {
    return(values * 1000)
  }

  values
}


.gpbiom_group_labels <- function(data, group_cols) {
  if (length(group_cols) == 0L) {
    return(rep("all", nrow(data)))
  }

  as.character(interaction(data[group_cols], drop = TRUE, lex.order = TRUE))
}


.gpbiom_ibi_large_jump_flags <- function(values, max_jump_ms) {
  flags <- rep(FALSE, length(values))

  if (length(values) < 2L) {
    return(flags)
  }

  valid <- !is.na(values) & is.finite(values) & values > 0

  for (i in 2:length(values)) {
    if (valid[i] && valid[i - 1L]) {
      flags[i] <- abs(values[i] - values[i - 1L]) > max_jump_ms
    }
  }

  flags
}


.gpbiom_ibi_group_summary <- function(sample_table,
                                      group_cols,
                                      min_valid_ibi) {
  groups <- split(seq_len(nrow(sample_table)), sample_table$group)

  rows <- lapply(names(groups), function(group_name) {
    rows <- sample_table[groups[[group_name]], , drop = FALSE]
    valid <- rows$ibi_ms[rows$valid_ibi]

    metrics <- .gpbiom_ibi_metrics(valid)

    data.frame(
      group = group_name,
      n_rows = nrow(rows),
      n_valid_ibi = length(valid),
      valid_ibi_rate = if (nrow(rows) > 0L) length(valid) / nrow(rows) else NA_real_,
      n_quality_flagged = sum(rows$any_quality_flag, na.rm = TRUE),
      quality_flag_rate = if (nrow(rows) > 0L) {
        mean(rows$any_quality_flag, na.rm = TRUE)
      } else {
        NA_real_
      },
      mean_ibi_ms = metrics$mean_ibi_ms,
      median_ibi_ms = metrics$median_ibi_ms,
      mean_hr_bpm = metrics$mean_hr_bpm,
      sdnn_ms = metrics$sdnn_ms,
      rmssd_ms = metrics$rmssd_ms,
      pnn50 = metrics$pnn50,
      status = if (length(valid) >= min_valid_ibi) {
        "sufficient_ibi"
      } else {
        "insufficient_ibi"
      },
      stringsAsFactors = FALSE
    )
  })

  do.call(rbind, rows)
}


.gpbiom_ibi_window_summary <- function(sample_table,
                                       group_cols,
                                       min_valid_ibi) {
  groups <- split(seq_len(nrow(sample_table)), sample_table$group)

  rows <- lapply(names(groups), function(group_name) {
    rows <- sample_table[groups[[group_name]], , drop = FALSE]
    valid <- rows$ibi_ms[rows$analysis_valid_ibi]

    metrics <- .gpbiom_ibi_metrics(valid)

    data.frame(
      group = group_name,
      n_rows = nrow(rows),
      n_ibi = sum(!rows$missing_ibi, na.rm = TRUE),
      n_valid_ibi = length(valid),
      valid_ibi_rate = if (nrow(rows) > 0L) length(valid) / nrow(rows) else NA_real_,
      n_excluded_for_quality = sum(!rows$analysis_valid_ibi, na.rm = TRUE),
      duration_s = if (length(valid) > 0L) sum(valid) / 1000 else NA_real_,
      mean_ibi_ms = metrics$mean_ibi_ms,
      median_ibi_ms = metrics$median_ibi_ms,
      min_ibi_ms = metrics$min_ibi_ms,
      max_ibi_ms = metrics$max_ibi_ms,
      mean_hr_bpm = metrics$mean_hr_bpm,
      median_hr_bpm = metrics$median_hr_bpm,
      min_hr_bpm = metrics$min_hr_bpm,
      max_hr_bpm = metrics$max_hr_bpm,
      sdnn_ms = metrics$sdnn_ms,
      rmssd_ms = metrics$rmssd_ms,
      pnn20 = metrics$pnn20,
      pnn50 = metrics$pnn50,
      status = if (length(valid) >= min_valid_ibi) {
        "sufficient_ibi_window"
      } else {
        "insufficient_ibi_window"
      },
      stringsAsFactors = FALSE
    )
  })

  do.call(rbind, rows)
}


.gpbiom_ibi_metrics <- function(ibi_ms) {
  ibi_ms <- ibi_ms[!is.na(ibi_ms) & is.finite(ibi_ms) & ibi_ms > 0]

  hr <- 60000 / ibi_ms
  diffs <- diff(ibi_ms)

  list(
    mean_ibi_ms = if (length(ibi_ms) > 0L) mean(ibi_ms) else NA_real_,
    median_ibi_ms = if (length(ibi_ms) > 0L) stats::median(ibi_ms) else NA_real_,
    min_ibi_ms = if (length(ibi_ms) > 0L) min(ibi_ms) else NA_real_,
    max_ibi_ms = if (length(ibi_ms) > 0L) max(ibi_ms) else NA_real_,
    mean_hr_bpm = if (length(hr) > 0L) mean(hr) else NA_real_,
    median_hr_bpm = if (length(hr) > 0L) stats::median(hr) else NA_real_,
    min_hr_bpm = if (length(hr) > 0L) min(hr) else NA_real_,
    max_hr_bpm = if (length(hr) > 0L) max(hr) else NA_real_,
    sdnn_ms = if (length(ibi_ms) > 1L) stats::sd(ibi_ms) else NA_real_,
    rmssd_ms = if (length(diffs) > 0L) sqrt(mean(diffs^2)) else NA_real_,
    pnn20 = if (length(diffs) > 0L) mean(abs(diffs) > 20) else NA_real_,
    pnn50 = if (length(diffs) > 0L) mean(abs(diffs) > 50) else NA_real_
  )
}

test_that("audit_gazepoint_ibi_quality audits valid millisecond IBI data", {
  df <- data.frame(
    USER = rep(c("P1", "P2"), each = 4),
    IBI = c(800, 810, 790, 805, 900, 910, 905, 920)
  )

  out <- audit_gazepoint_ibi_quality(df, group_cols = "USER")

  expect_type(out, "list")
  expect_s3_class(out$overview, "data.frame")
  expect_s3_class(out$samples, "data.frame")
  expect_s3_class(out$group_summary, "data.frame")

  expect_equal(out$overview$ibi_column, "IBI")
  expect_equal(out$overview$unit, "milliseconds")
  expect_equal(out$overview$n_valid_ibi, 8)
  expect_equal(out$overview$status, "ibi_quality_ok")
  expect_equal(nrow(out$group_summary), 2)
  expect_true(all(out$samples$valid_ibi))
})


test_that("audit_gazepoint_ibi_quality detects seconds automatically", {
  df <- data.frame(
    IBI = c(0.8, 0.81, 0.79, 0.805)
  )

  out <- audit_gazepoint_ibi_quality(df)

  expect_equal(out$overview$unit, "seconds")
  expect_equal(out$samples$ibi_ms, c(800, 810, 790, 805))
  expect_equal(out$overview$status, "ibi_quality_ok")
})


test_that("audit_gazepoint_ibi_quality flags implausible and missing intervals", {
  df <- data.frame(
    IBI = c(800, NA, 0, 250, 2500, Inf, 810)
  )

  out <- audit_gazepoint_ibi_quality(df)

  expect_equal(out$overview$n_missing_ibi, 1)
  expect_equal(out$overview$n_nonpositive_ibi, 1)
  expect_equal(out$overview$n_below_min_ibi, 1)
  expect_equal(out$overview$n_above_max_ibi, 1)
  expect_equal(out$overview$n_nonfinite_ibi, 1)
  expect_equal(out$overview$status, "ibi_quality_issues_detected")

  expect_true("missing_ibi" %in% out$samples$status)
  expect_true("nonpositive_ibi" %in% out$samples$status)
  expect_true("below_min_ibi" %in% out$samples$status)
  expect_true("above_max_ibi" %in% out$samples$status)
  expect_true("nonfinite_ibi" %in% out$samples$status)
})


test_that("audit_gazepoint_ibi_quality flags large jumps within ordered groups", {
  df <- data.frame(
    USER = c("P1", "P1", "P1", "P1"),
    TIME = c(1, 2, 3, 4),
    IBI = c(800, 810, 1500, 1510)
  )

  out <- audit_gazepoint_ibi_quality(
    df,
    group_cols = "USER",
    time_col = "TIME",
    max_jump_ms = 500
  )

  expect_true(out$samples$large_jump_ibi[3])
  expect_false(out$samples$large_jump_ibi[4])
  expect_equal(out$overview$n_large_jump_ibi, 1)
  expect_equal(out$overview$status, "ibi_quality_issues_detected")
})


test_that("audit_gazepoint_ibi_quality does not treat HRV as IBI", {
  df <- data.frame(
    HRV = c(1, 1, 1, 0),
    HR = c(70, 71, 72, 73)
  )

  expect_error(
    audit_gazepoint_ibi_quality(df),
    "No IBI/RR interval column"
  )
})


test_that("audit_gazepoint_ibi_quality validates arguments", {
  df <- data.frame(IBI = c(800, 810, 790))

  expect_error(
    audit_gazepoint_ibi_quality(1:3),
    "`data` must be"
  )

  expect_error(
    audit_gazepoint_ibi_quality(df, ibi_col = "missing"),
    "not found"
  )

  expect_error(
    audit_gazepoint_ibi_quality(data.frame(IBI = letters[1:3])),
    "must be numeric"
  )

  expect_error(
    audit_gazepoint_ibi_quality(df, min_ibi_ms = 0),
    "`min_ibi_ms`"
  )

  expect_error(
    audit_gazepoint_ibi_quality(df, min_ibi_ms = 1000, max_ibi_ms = 500),
    "`min_ibi_ms` must be smaller"
  )
})


test_that("summarise_gazepoint_ibi_windows summarises valid IBI windows", {
  df <- data.frame(
    USER = rep(c("P1", "P2"), each = 4),
    IBI = c(800, 810, 790, 805, 900, 910, 905, 920)
  )

  out <- summarise_gazepoint_ibi_windows(df, group_cols = "USER")

  expect_type(out, "list")
  expect_s3_class(out$overview, "data.frame")
  expect_s3_class(out$windows, "data.frame")
  expect_s3_class(out$samples, "data.frame")

  expect_equal(out$overview$window_count, 2)
  expect_equal(out$overview$sufficient_window_count, 2)
  expect_equal(out$overview$status, "ibi_windows_summarised")
  expect_true(all(out$windows$status == "sufficient_ibi_window"))
  expect_true(all(c("sdnn_ms", "rmssd_ms", "pnn20", "pnn50") %in% names(out$windows)))
})


test_that("summarise_gazepoint_ibi_windows can summarise all rows as one window", {
  df <- data.frame(
    IBI = c(800, 810, 790, 805)
  )

  out <- summarise_gazepoint_ibi_windows(df)

  expect_equal(nrow(out$windows), 1)
  expect_equal(out$windows$group, "all")
  expect_equal(out$windows$n_valid_ibi, 4)
  expect_equal(out$overview$status, "ibi_windows_summarised")
})


test_that("summarise_gazepoint_ibi_windows excludes large jumps by default", {
  df <- data.frame(
    IBI = c(800, 810, 1500, 1510)
  )

  out <- summarise_gazepoint_ibi_windows(
    df,
    max_jump_ms = 500,
    min_valid_ibi = 2
  )

  expect_true(out$samples$large_jump_ibi[3])
  expect_equal(out$windows$n_valid_ibi, 3)
  expect_equal(out$overview$status, "ibi_windows_summarised")
})


test_that("summarise_gazepoint_ibi_windows can keep large jumps when requested", {
  df <- data.frame(
    IBI = c(800, 810, 1500, 1510)
  )

  out <- summarise_gazepoint_ibi_windows(
    df,
    max_jump_ms = 500,
    exclude_large_jumps = FALSE
  )

  expect_equal(out$windows$n_valid_ibi, 4)
})


test_that("summarise_gazepoint_ibi_windows marks insufficient windows", {
  df <- data.frame(
    USER = c("P1", "P1", "P2", "P2"),
    IBI = c(800, NA, 900, NA)
  )

  out <- summarise_gazepoint_ibi_windows(
    df,
    group_cols = "USER",
    min_valid_ibi = 2
  )

  expect_equal(out$overview$sufficient_window_count, 0)
  expect_equal(out$overview$status, "no_sufficient_ibi_windows")
  expect_true(all(out$windows$status == "insufficient_ibi_window"))
})


test_that("summarise_gazepoint_ibi_windows validates arguments", {
  df <- data.frame(IBI = c(800, 810, 790))

  expect_error(
    summarise_gazepoint_ibi_windows(1:3),
    "`data` must be"
  )

  expect_error(
    summarise_gazepoint_ibi_windows(df, group_cols = "missing"),
    "not found"
  )

  expect_error(
    summarise_gazepoint_ibi_windows(df, exclude_large_jumps = NA),
    "`exclude_large_jumps`"
  )

  expect_error(
    summarise_gazepoint_ibi_windows(df, min_valid_ibi = 0),
    "`min_valid_ibi`"
  )
})


test_that("summarise_gazepoint_ibi_windows documents conservative HRV interpretation", {
  df <- data.frame(
    IBI = c(800, 810, 790, 805)
  )

  out <- summarise_gazepoint_ibi_windows(df)

  expect_match(out$settings$note, "genuine IBI/RR intervals")
  expect_match(out$settings$note, "not calculated from raw HRV")
})

#' Audit Gazepoint biometric sampling and timing
#'
#' Audits timing or row-order information in Gazepoint Biometrics exports. The
#' function checks monotonicity, duplicate timestamps, nonpositive intervals,
#' and estimated sampling rate when the selected time column has a real time
#' unit. If only `CNT` is available, the function can still check ordering but
#' does not estimate a sampling rate unless `time_unit` is explicitly meaningful.
#'
#' @param data A data frame or a path to a Gazepoint CSV export.
#' @param group_columns Optional grouping columns within which timing should be
#'   audited, such as `c("source_participant", "MEDIA_ID")`.
#' @param time_column Optional time/order column. If `NULL`, the function uses
#'   the first available column among `TIME`, `TIME_TICK`, and `CNT`.
#' @param time_unit Unit of the selected time column. Use `"seconds"`,
#'   `"milliseconds"`, `"microseconds"`, or `"samples"`. When `"samples"` is
#'   used, sampling-rate estimates are returned as `NA`.
#' @param expected_rate_hz Optional expected sampling rate in Hz.
#' @param tolerance_hz Acceptable absolute deviation from `expected_rate_hz`.
#'
#' @return A data frame with one row per group.
#'
#' @export
audit_gazepoint_biometric_sampling <- function(data,
                                               group_columns = NULL,
                                               time_column = NULL,
                                               time_unit = c(
                                                 "seconds",
                                                 "milliseconds",
                                                 "microseconds",
                                                 "samples"
                                               ),
                                               expected_rate_hz = 60,
                                               tolerance_hz = 5) {
  dat <- coerce_gazepoint_biometrics_data(data)
  time_unit <- match.arg(time_unit)

  if (is.null(time_column)) {
    time_column <- choose_first_present(dat, c("TIME", "TIME_TICK", "CNT"))
  }

  if (is.null(time_column) || length(time_column) != 1L || is.na(time_column)) {
    stop("No timing/order column was found in `data`.", call. = FALSE)
  }

  if (!time_column %in% names(dat)) {
    stop("`time_column` was not found in `data`: ", time_column, call. = FALSE)
  }

  if (!is.null(group_columns)) {
    missing_groups <- setdiff(group_columns, names(dat))

    if (length(missing_groups) > 0L) {
      stop(
        "`group_columns` were not found in `data`: ",
        paste(missing_groups, collapse = ", "),
        call. = FALSE
      )
    }

    group_key <- make_group_key(dat, group_columns)
  } else {
    group_key <- rep("all", nrow(dat))
  }

  rows <- lapply(unique(group_key), function(key) {
    in_group <- group_key == key
    source_rows <- which(in_group)

    group_values <- if (!is.null(group_columns)) {
      dat[source_rows[1], group_columns, drop = FALSE]
    } else {
      data.frame(group = "all", stringsAsFactors = FALSE)
    }

    time_values <- as_numeric_safe(dat[[time_column]][in_group])

    summary <- summarise_sampling_vector(
      time_values = time_values,
      time_column = time_column,
      time_unit = time_unit,
      expected_rate_hz = expected_rate_hz,
      tolerance_hz = tolerance_hz
    )

    cbind(group_values, summary, stringsAsFactors = FALSE)
  })

  out <- do.call(rbind, rows)
  rownames(out) <- NULL
  out
}


summarise_sampling_vector <- function(time_values,
                                      time_column,
                                      time_unit,
                                      expected_rate_hz,
                                      tolerance_hz) {
  non_missing <- !is.na(time_values)
  usable_time <- time_values[non_missing]

  intervals <- diff(usable_time)

  positive_intervals <- intervals[!is.na(intervals) & intervals > 0]

  estimated_rate_hz <- NA_real_
  median_interval_seconds <- NA_real_

  if (length(positive_intervals) > 0L && time_unit != "samples") {
    interval_seconds <- convert_interval_to_seconds(
      positive_intervals,
      time_unit = time_unit
    )

    median_interval_seconds <- stats::median(interval_seconds)

    if (!is.na(median_interval_seconds) && median_interval_seconds > 0) {
      estimated_rate_hz <- 1 / median_interval_seconds
    }
  }

  rate_deviation_hz <- ifelse(
    !is.na(estimated_rate_hz) && !is.null(expected_rate_hz) && !is.na(expected_rate_hz),
    estimated_rate_hz - expected_rate_hz,
    NA_real_
  )

  rate_status <- classify_sampling_rate(
    estimated_rate_hz = estimated_rate_hz,
    expected_rate_hz = expected_rate_hz,
    tolerance_hz = tolerance_hz
  )

  data.frame(
    time_column = time_column,
    time_unit = time_unit,
    n_rows = length(time_values),
    non_missing_time_rows = sum(non_missing),
    missing_time_rows = sum(!non_missing),
    missing_time_pct = safe_pct(sum(!non_missing), length(time_values)),
    duplicate_time_rows = sum(duplicated(usable_time)),
    interval_rows = length(intervals),
    positive_interval_rows = length(positive_intervals),
    zero_interval_rows = sum(!is.na(intervals) & intervals == 0),
    negative_interval_rows = sum(!is.na(intervals) & intervals < 0),
    monotonic_non_decreasing = all(intervals >= 0, na.rm = TRUE),
    strictly_increasing = all(intervals > 0, na.rm = TRUE),
    median_interval_seconds = median_interval_seconds,
    estimated_rate_hz = estimated_rate_hz,
    expected_rate_hz = expected_rate_hz,
    rate_deviation_hz = rate_deviation_hz,
    rate_status = rate_status,
    stringsAsFactors = FALSE
  )
}


convert_interval_to_seconds <- function(intervals,
                                        time_unit) {
  if (time_unit == "seconds") {
    return(intervals)
  }

  if (time_unit == "milliseconds") {
    return(intervals / 1000)
  }

  if (time_unit == "microseconds") {
    return(intervals / 1000000)
  }

  rep(NA_real_, length(intervals))
}


classify_sampling_rate <- function(estimated_rate_hz,
                                   expected_rate_hz,
                                   tolerance_hz) {
  if (is.na(estimated_rate_hz)) {
    return("not_estimated")
  }

  if (is.null(expected_rate_hz) || is.na(expected_rate_hz)) {
    return("estimated")
  }

  if (abs(estimated_rate_hz - expected_rate_hz) <= tolerance_hz) {
    return("within_tolerance")
  }

  "outside_tolerance"
}

test_that("audit_gazepoint_biometric_sampling estimates sampling rate from seconds", {
  dat <- data.frame(
    source_participant = rep("User 1", 4),
    MEDIA_ID = rep(0, 4),
    TIME = c(0, 1 / 60, 2 / 60, 3 / 60)
  )

  out <- audit_gazepoint_biometric_sampling(
    dat,
    group_columns = c("source_participant", "MEDIA_ID"),
    time_column = "TIME",
    time_unit = "seconds",
    expected_rate_hz = 60,
    tolerance_hz = 1
  )

  expect_equal(nrow(out), 1)
  expect_equal(out$estimated_rate_hz, 60, tolerance = 1e-8)
  expect_equal(out$rate_status, "within_tolerance")
  expect_true(out$strictly_increasing)
})


test_that("audit_gazepoint_biometric_sampling supports milliseconds", {
  dat <- data.frame(
    TIME_TICK = c(0, 16.6667, 33.3334, 50.0001)
  )

  out <- audit_gazepoint_biometric_sampling(
    dat,
    time_column = "TIME_TICK",
    time_unit = "milliseconds",
    expected_rate_hz = 60,
    tolerance_hz = 1
  )

  expect_equal(nrow(out), 1)
  expect_equal(out$rate_status, "within_tolerance")
  expect_true(out$estimated_rate_hz > 59)
  expect_true(out$estimated_rate_hz < 61)
})


test_that("audit_gazepoint_biometric_sampling detects duplicate and nonmonotonic rows", {
  dat <- data.frame(
    TIME = c(0, 0.1, 0.1, 0.05)
  )

  out <- audit_gazepoint_biometric_sampling(
    dat,
    time_column = "TIME",
    time_unit = "seconds"
  )

  expect_equal(out$duplicate_time_rows, 1)
  expect_equal(out$zero_interval_rows, 1)
  expect_equal(out$negative_interval_rows, 1)
  expect_false(out$monotonic_non_decreasing)
  expect_false(out$strictly_increasing)
})


test_that("audit_gazepoint_biometric_sampling works with sample/order columns without rate estimation", {
  dat <- data.frame(
    CNT = c(1, 2, 3, 4)
  )

  out <- audit_gazepoint_biometric_sampling(
    dat,
    time_column = "CNT",
    time_unit = "samples"
  )

  expect_equal(out$rate_status, "not_estimated")
  expect_true(is.na(out$estimated_rate_hz))
  expect_true(out$strictly_increasing)
})


test_that("audit_gazepoint_biometric_sampling supports grouped output", {
  dat <- data.frame(
    source_participant = c("User 1", "User 1", "User 2", "User 2"),
    TIME = c(0, 0.1, 0, 0.2)
  )

  out <- audit_gazepoint_biometric_sampling(
    dat,
    group_columns = "source_participant",
    time_column = "TIME",
    time_unit = "seconds"
  )

  expect_equal(nrow(out), 2)
  expect_true(all(c("User 1", "User 2") %in% out$source_participant))
})


test_that("audit_gazepoint_biometric_sampling rejects missing timing column", {
  dat <- data.frame(
    GSR_US = c(1.1, 1.2)
  )

  expect_error(
    audit_gazepoint_biometric_sampling(dat),
    "No timing"
  )
})


test_that("audit_gazepoint_biometric_sampling rejects missing group columns", {
  dat <- data.frame(
    TIME = c(0, 0.1)
  )

  expect_error(
    audit_gazepoint_biometric_sampling(
      dat,
      group_columns = "source_participant"
    ),
    "group_columns"
  )
})

#' Analyse cardiorespiratory Granger-style directionality
#'
#' Computes dependency-light linear Granger-style directionality tests between a
#' respiration proxy and heart-rate or IBI/RR signal. This estimates predictive
#' directionality in a VAR-style model. It does not prove physiological
#' causality from observational data by itself.
#'
#' @param dat A data frame.
#' @param respiration_col Numeric respiration proxy column.
#' @param cardiac_col Numeric cardiac column, such as HR, IBI, or RR.
#' @param time_col Optional time column for ordering.
#' @param group_cols Optional grouping columns.
#' @param lag_order VAR lag order.
#' @param min_rows Minimum complete rows per group.
#' @param standardise Logical. If `TRUE`, z-standardise both series per group.
#'
#' @return A list with `overview`, `causality_summary`, and `settings`.
#' @export
analyze_gazepoint_cardiorespiratory_causality <- function(dat,
                                                          respiration_col,
                                                          cardiac_col,
                                                          time_col = NULL,
                                                          group_cols = NULL,
                                                          lag_order = 3,
                                                          min_rows = 30,
                                                          standardise = TRUE) {
  if (!is.data.frame(dat)) {
    stop("`dat` must be a data frame.", call. = FALSE)
  }

  required <- c(respiration_col, cardiac_col, time_col)
  required <- required[!is.null(required)]

  missing_required <- setdiff(required, names(dat))
  if (length(missing_required) > 0) {
    stop("Missing required columns: ", paste(missing_required, collapse = ", "), call. = FALSE)
  }

  if (!is.numeric(dat[[respiration_col]])) {
    stop("`respiration_col` must identify a numeric column.", call. = FALSE)
  }

  if (!is.numeric(dat[[cardiac_col]])) {
    stop("`cardiac_col` must identify a numeric column.", call. = FALSE)
  }

  if (!is.null(time_col) && !is.numeric(dat[[time_col]])) {
    stop("`time_col` must identify a numeric column.", call. = FALSE)
  }

  if (is.null(group_cols)) {
    group_cols <- character()
  }

  missing_groups <- setdiff(group_cols, names(dat))
  if (length(missing_groups) > 0) {
    stop("Missing `group_cols`: ", paste(missing_groups, collapse = ", "), call. = FALSE)
  }

  groups <- gpbiometrics_cardioresp_split(dat, group_cols)

  rows <- lapply(names(groups), function(group_id) {
    idx <- groups[[group_id]]

    if (!is.null(time_col)) {
      idx <- idx[order(dat[[time_col]][idx])]
    }

    resp <- dat[[respiration_col]][idx]
    cardiac <- dat[[cardiac_col]][idx]

    keep <- is.finite(resp) & is.finite(cardiac)
    resp <- resp[keep]
    cardiac <- cardiac[keep]

    if (length(resp) < min_rows ||
        length(resp) <= lag_order * 3 ||
        stats::sd(resp) == 0 ||
        stats::sd(cardiac) == 0) {
      return(data.frame(
        group_id = group_id,
        n_rows = length(resp),
        lag_order = lag_order,
        respiration_to_cardiac_f = NA_real_,
        respiration_to_cardiac_p = NA_real_,
        respiration_to_cardiac_log_ratio = NA_real_,
        cardiac_to_respiration_f = NA_real_,
        cardiac_to_respiration_p = NA_real_,
        cardiac_to_respiration_log_ratio = NA_real_,
        dominant_direction = NA_character_,
        status = "insufficient_information",
        stringsAsFactors = FALSE
      ))
    }

    if (isTRUE(standardise)) {
      resp <- as.numeric(scale(resp))
      cardiac <- as.numeric(scale(cardiac))
    }

    r_to_c <- gpbiometrics_granger_test(
      target = cardiac,
      driver = resp,
      lag_order = lag_order
    )

    c_to_r <- gpbiometrics_granger_test(
      target = resp,
      driver = cardiac,
      lag_order = lag_order
    )

    dominant <- if (is.finite(r_to_c$log_ratio) && is.finite(c_to_r$log_ratio)) {
      if (r_to_c$log_ratio > c_to_r$log_ratio) {
        "respiration_to_cardiac_stronger"
      } else if (c_to_r$log_ratio > r_to_c$log_ratio) {
        "cardiac_to_respiration_stronger"
      } else {
        "balanced"
      }
    } else {
      NA_character_
    }

    data.frame(
      group_id = group_id,
      n_rows = length(resp),
      lag_order = lag_order,
      respiration_to_cardiac_f = r_to_c$f_value,
      respiration_to_cardiac_p = r_to_c$p_value,
      respiration_to_cardiac_log_ratio = r_to_c$log_ratio,
      cardiac_to_respiration_f = c_to_r$f_value,
      cardiac_to_respiration_p = c_to_r$p_value,
      cardiac_to_respiration_log_ratio = c_to_r$log_ratio,
      dominant_direction = dominant,
      status = "granger_directionality_estimated",
      stringsAsFactors = FALSE
    )
  })

  causality_summary <- do.call(rbind, rows)
  rownames(causality_summary) <- NULL

  overview <- data.frame(
    group_count = length(groups),
    result_rows = nrow(causality_summary),
    successful_groups = sum(causality_summary$status == "granger_directionality_estimated"),
    problem_groups = sum(causality_summary$status != "granger_directionality_estimated"),
    status = if (all(causality_summary$status == "granger_directionality_estimated")) {
      "cardiorespiratory_directionality_estimated"
    } else if (any(causality_summary$status == "granger_directionality_estimated")) {
      "cardiorespiratory_directionality_partial"
    } else {
      "cardiorespiratory_directionality_failed"
    },
    interpretation = paste(
      "These are linear Granger-style predictive-directionality summaries.",
      "They do not prove mechanistic causality and should be interpreted with experimental design, stationarity, lag choice, and measurement quality."
    ),
    stringsAsFactors = FALSE
  )

  structure(
    list(
      overview = overview,
      causality_summary = causality_summary,
      settings = list(
        respiration_col = respiration_col,
        cardiac_col = cardiac_col,
        time_col = time_col,
        group_cols = group_cols,
        lag_order = lag_order,
        min_rows = min_rows,
        standardise = standardise
      )
    ),
    class = c("gazepoint_cardiorespiratory_causality", "list")
  )
}

gpbiometrics_cardioresp_split <- function(dat, group_cols) {
  if (length(group_cols) == 0) {
    return(list(all_rows = seq_len(nrow(dat))))
  }

  gf <- dat[group_cols]
  gf[] <- lapply(gf, function(x) {
    x <- as.character(x)
    x[is.na(x)] <- "<NA>"
    x
  })

  split(seq_len(nrow(dat)), do.call(paste, c(gf, sep = " | ")))
}

gpbiometrics_granger_test <- function(target, driver, lag_order = 3) {
  n <- length(target)

  response <- target[(lag_order + 1):n]

  target_lags <- sapply(seq_len(lag_order), function(lag) {
    target[(lag_order + 1 - lag):(n - lag)]
  })

  driver_lags <- sapply(seq_len(lag_order), function(lag) {
    driver[(lag_order + 1 - lag):(n - lag)]
  })

  target_lags <- as.data.frame(target_lags)
  driver_lags <- as.data.frame(driver_lags)

  names(target_lags) <- paste0("target_lag_", seq_len(lag_order))
  names(driver_lags) <- paste0("driver_lag_", seq_len(lag_order))

  reduced_dat <- data.frame(response = response, target_lags)
  full_dat <- data.frame(response = response, target_lags, driver_lags)

  reduced <- stats::lm(response ~ ., data = reduced_dat)
  full <- stats::lm(response ~ ., data = full_dat)

  rss_reduced <- sum(stats::residuals(reduced)^2)
  rss_full <- sum(stats::residuals(full)^2)

  df1 <- lag_order
  df2 <- stats::df.residual(full)

  f_value <- if (rss_full > 0 && df2 > 0) {
    ((rss_reduced - rss_full) / df1) / (rss_full / df2)
  } else {
    NA_real_
  }

  p_value <- if (is.finite(f_value)) {
    stats::pf(f_value, df1 = df1, df2 = df2, lower.tail = FALSE)
  } else {
    NA_real_
  }

  log_ratio <- if (is.finite(rss_reduced) && is.finite(rss_full) && rss_full > 0) {
    log(rss_reduced / rss_full)
  } else {
    NA_real_
  }

  list(
    f_value = f_value,
    p_value = p_value,
    log_ratio = log_ratio
  )
}

#' Optimise subject-specific cvxEDA slow time constant
#'
#' Performs a dependency-light grid search over the slow Bateman impulse-response
#' time constant used in cvxEDA-style EDA decomposition workflows. The default
#' fast time constant is fixed at 0.7 seconds and the slow time constant is
#' searched between 2 and 4 seconds. This function does not run the original
#' cvxEDA optimisation; it provides a subject-specific tau-selection bridge for
#' downstream cvxEDA-style workflows.
#'
#' @param dat A data frame containing EDA data.
#' @param eda_col Numeric EDA/conductance column.
#' @param time_col Numeric time column.
#' @param group_cols Optional grouping columns, usually participant/session.
#' @param tau0_grid Candidate slow time constants.
#' @param tau1 Fixed fast time constant.
#' @param sampling_rate Optional sampling rate in Hz. If `NULL`, estimated from
#'   `time_col`.
#' @param ridge_lambda Small ridge penalty used in frequency-domain
#'   deconvolution.
#' @param max_irf_seconds Maximum impulse-response duration.
#'
#' @return A list with `overview`, `best_tau`, `optimization_table`, and
#'   `settings`.
#' @export
optimize_gazepoint_cvxeda_tau <- function(dat,
                                          eda_col = "GSR_US",
                                          time_col = "CNT",
                                          group_cols = NULL,
                                          tau0_grid = seq(2, 4, by = 0.25),
                                          tau1 = 0.7,
                                          sampling_rate = NULL,
                                          ridge_lambda = 0.01,
                                          max_irf_seconds = 20) {
  if (!is.data.frame(dat)) {
    stop("`dat` must be a data frame.", call. = FALSE)
  }

  if (!eda_col %in% names(dat)) {
    stop("Column `", eda_col, "` was not found in `dat`.", call. = FALSE)
  }

  if (!time_col %in% names(dat)) {
    stop("Column `", time_col, "` was not found in `dat`.", call. = FALSE)
  }

  if (!is.numeric(dat[[eda_col]])) {
    stop("`eda_col` must identify a numeric column.", call. = FALSE)
  }

  if (!is.numeric(dat[[time_col]])) {
    stop("`time_col` must identify a numeric column.", call. = FALSE)
  }

  if (is.null(group_cols)) {
    group_cols <- character()
  }

  missing_groups <- setdiff(group_cols, names(dat))
  if (length(missing_groups) > 0) {
    stop("Missing `group_cols`: ", paste(missing_groups, collapse = ", "), call. = FALSE)
  }

  tau0_grid <- sort(unique(as.numeric(tau0_grid)))

  if (length(tau0_grid) == 0 ||
      any(!is.finite(tau0_grid)) ||
      any(tau0_grid <= tau1)) {
    stop("`tau0_grid` must contain finite values larger than `tau1`.", call. = FALSE)
  }

  groups <- gpbiometrics_cvxeda_tau_split(dat, group_cols)

  rows <- list()
  row_id <- 1L

  for (group_id in names(groups)) {
    idx <- groups[[group_id]]
    idx <- idx[order(dat[[time_col]][idx])]

    time <- dat[[time_col]][idx]
    eda <- dat[[eda_col]][idx]
    fs <- gpbiometrics_cvxeda_tau_sampling_rate(time, sampling_rate)

    finite_eda <- is.finite(eda)

    if (!is.finite(fs) || fs <= 0 || sum(finite_eda) < 20) {
      for (tau0 in tau0_grid) {
        rows[[row_id]] <- data.frame(
          group_id = group_id,
          tau0 = tau0,
          tau1 = tau1,
          sampling_rate_hz = fs,
          n_samples = length(idx),
          n_finite = sum(finite_eda),
          rmse = NA_real_,
          mae = NA_real_,
          residual_sd = NA_real_,
          correlation = NA_real_,
          status = "insufficient_signal_or_sampling_rate",
          stringsAsFactors = FALSE
        )
        row_id <- row_id + 1L
      }
      next
    }

    y <- gpbiometrics_cvxeda_tau_fill(eda)
    y <- y - mean(y, na.rm = TRUE)

    for (tau0 in tau0_grid) {
      fit <- gpbiometrics_cvxeda_tau_fit(
        y = y,
        sampling_rate = fs,
        tau0 = tau0,
        tau1 = tau1,
        ridge_lambda = ridge_lambda,
        max_irf_seconds = max_irf_seconds
      )

      rows[[row_id]] <- data.frame(
        group_id = group_id,
        tau0 = tau0,
        tau1 = tau1,
        sampling_rate_hz = fs,
        n_samples = length(idx),
        n_finite = sum(finite_eda),
        rmse = fit$rmse,
        mae = fit$mae,
        residual_sd = fit$residual_sd,
        correlation = fit$correlation,
        status = fit$status,
        stringsAsFactors = FALSE
      )

      row_id <- row_id + 1L
    }
  }

  optimization_table <- do.call(rbind, rows)
  rownames(optimization_table) <- NULL

  best_rows <- lapply(split(optimization_table, optimization_table$group_id), function(d) {
    ok <- d[is.finite(d$rmse), , drop = FALSE]

    if (nrow(ok) == 0) {
      return(d[1, , drop = FALSE])
    }

    ok[which.min(ok$rmse), , drop = FALSE]
  })

  best_tau <- do.call(rbind, best_rows)
  rownames(best_tau) <- NULL

  best_tau$status <- ifelse(
    is.finite(best_tau$rmse),
    "best_tau_selected",
    "best_tau_not_selected"
  )

  overview <- data.frame(
    group_count = length(groups),
    candidate_tau_count = length(tau0_grid),
    optimization_rows = nrow(optimization_table),
    successful_groups = sum(best_tau$status == "best_tau_selected"),
    problem_groups = sum(best_tau$status != "best_tau_selected"),
    status = if (all(best_tau$status == "best_tau_selected")) {
      "cvxeda_tau_optimization_complete"
    } else if (any(best_tau$status == "best_tau_selected")) {
      "cvxeda_tau_optimization_partial"
    } else {
      "cvxeda_tau_optimization_failed"
    },
    interpretation = paste(
      "The selected tau0 minimises a dependency-light reconstruction residual.",
      "This is a cvxEDA parameter-selection bridge and not the original cvxEDA convex optimisation."
    ),
    stringsAsFactors = FALSE
  )

  structure(
    list(
      overview = overview,
      best_tau = best_tau,
      optimization_table = optimization_table,
      settings = list(
        eda_col = eda_col,
        time_col = time_col,
        group_cols = group_cols,
        tau0_grid = tau0_grid,
        tau1 = tau1,
        sampling_rate = sampling_rate,
        ridge_lambda = ridge_lambda,
        max_irf_seconds = max_irf_seconds
      )
    ),
    class = c("gazepoint_cvxeda_tau_optimization", "list")
  )
}

gpbiometrics_cvxeda_tau_split <- function(dat, group_cols) {
  if (length(group_cols) == 0) {
    return(list(all_rows = seq_len(nrow(dat))))
  }

  gf <- dat[group_cols]
  gf[] <- lapply(gf, function(x) {
    x <- as.character(x)
    x[is.na(x)] <- "<NA>"
    x
  })

  split(seq_len(nrow(dat)), do.call(paste, c(gf, sep = " | ")))
}

gpbiometrics_cvxeda_tau_sampling_rate <- function(time, sampling_rate = NULL) {
  if (!is.null(sampling_rate)) {
    return(sampling_rate)
  }

  time <- time[is.finite(time)]

  if (length(time) < 3) {
    return(NA_real_)
  }

  dt <- diff(time)
  dt <- dt[is.finite(dt) & dt > 0]

  if (length(dt) == 0) {
    return(NA_real_)
  }

  median_dt <- stats::median(dt)

  if (median_dt > 10) {
    1000 / median_dt
  } else {
    1 / median_dt
  }
}

gpbiometrics_cvxeda_tau_fill <- function(x) {
  idx <- seq_along(x)
  finite <- is.finite(x)

  if (all(finite)) {
    return(x)
  }

  if (sum(finite) == 0) {
    return(rep(0, length(x)))
  }

  if (sum(finite) == 1) {
    return(rep(x[finite][1], length(x)))
  }

  stats::approx(idx[finite], x[finite], xout = idx, rule = 2)$y
}

gpbiometrics_cvxeda_tau_irf <- function(sampling_rate,
                                        tau0,
                                        tau1,
                                        max_irf_seconds = 20) {
  t <- seq(0, max_irf_seconds, by = 1 / sampling_rate)

  h <- exp(-t / tau0) - exp(-t / tau1)
  h[h < 0] <- 0

  if (sum(h) > 0) {
    h <- h / sum(h)
  }

  h
}

gpbiometrics_cvxeda_tau_fit <- function(y,
                                        sampling_rate,
                                        tau0,
                                        tau1,
                                        ridge_lambda = 0.01,
                                        max_irf_seconds = 20) {
  h <- gpbiometrics_cvxeda_tau_irf(
    sampling_rate = sampling_rate,
    tau0 = tau0,
    tau1 = tau1,
    max_irf_seconds = max_irf_seconds
  )

  n <- length(y)
  n_conv <- n + length(h) - 1
  n_fft <- 2^ceiling(log2(n_conv))

  y_pad <- c(y, rep(0, n_fft - n))
  h_pad <- c(h, rep(0, n_fft - length(h)))

  y_fft <- stats::fft(y_pad)
  h_fft <- stats::fft(h_pad)

  driver_fft <- Conj(h_fft) * y_fft / (Mod(h_fft)^2 + ridge_lambda)
  recon_fft <- h_fft * driver_fft

  recon <- Re(stats::fft(recon_fft, inverse = TRUE) / n_fft)[seq_len(n)]

  residual <- y - recon

  data.frame(
    rmse = sqrt(mean(residual^2, na.rm = TRUE)),
    mae = mean(abs(residual), na.rm = TRUE),
    residual_sd = stats::sd(residual, na.rm = TRUE),
    correlation = suppressWarnings(stats::cor(y, recon, use = "pairwise.complete.obs")),
    status = "tau_fit_evaluated",
    stringsAsFactors = FALSE
  )
}

#' Recommend Gazepoint biometric exclusions
#'
#' Creates window-level and participant-level exclusion recommendations from
#' Gazepoint biometric usable-sample summaries. The function does not remove
#' data. It only creates transparent keep, review, or exclude recommendations.
#'
#' @param data A row-level Gazepoint Biometrics data frame, a path to a
#'   Gazepoint CSV export, or an already summarised multimodal window table.
#' @param group_columns Columns defining analysis windows when `data` is
#'   row-level, such as `c("source_participant", "MEDIA_ID")`.
#' @param data_is_window_summary Logical. If `TRUE`, `data` is treated as an
#'   already summarised window table.
#' @param participant_column Optional participant identifier column. If `NULL`,
#'   the function tries `source_participant`, `USER`, and `USERID`.
#' @param gsr_min_usable_pct Minimum acceptable usable percentage for GSR/EDA.
#' @param hr_min_usable_pct Minimum acceptable usable percentage for heart rate.
#' @param dial_min_usable_pct Minimum acceptable usable percentage for
#'   engagement dial.
#' @param require_gsr Should low GSR/EDA coverage cause exclusion rather than
#'   review?
#' @param require_hr Should low heart-rate coverage cause exclusion rather than
#'   review?
#' @param require_dial Should low engagement-dial coverage cause exclusion
#'   rather than review?
#'
#' @return A list with `overview`, `window_recommendations`,
#'   `participant_recommendations`, and `settings`.
#'
#' @export
recommend_gazepoint_biometric_exclusions <- function(data,
                                                     group_columns = NULL,
                                                     data_is_window_summary = FALSE,
                                                     participant_column = NULL,
                                                     gsr_min_usable_pct = 50,
                                                     hr_min_usable_pct = 50,
                                                     dial_min_usable_pct = 50,
                                                     require_gsr = TRUE,
                                                     require_hr = TRUE,
                                                     require_dial = FALSE) {
  if (isTRUE(data_is_window_summary)) {
    windows <- data

    if (!is.data.frame(windows)) {
      stop(
        "`data` must be a data frame when `data_is_window_summary = TRUE`.",
        call. = FALSE
      )
    }
  } else {
    dat <- coerce_gazepoint_biometrics_data(data)

    if (is.null(group_columns) || length(group_columns) == 0L) {
      stop(
        "`group_columns` must be supplied when `data` is row-level.",
        call. = FALSE
      )
    }

    missing_groups <- setdiff(group_columns, names(dat))

    if (length(missing_groups) > 0L) {
      stop(
        "`group_columns` were not found in `data`: ",
        paste(missing_groups, collapse = ", "),
        call. = FALSE
      )
    }

    windows <- summarise_gazepoint_multimodal_windows(
      data = dat,
      group_columns = group_columns
    )
  }

  required_summary_columns <- c(
    "gsr_usable_pct",
    "hr_usable_pct",
    "dial_usable_pct"
  )

  missing_summary_columns <- setdiff(required_summary_columns, names(windows))

  if (length(missing_summary_columns) > 0L) {
    stop(
      "`data` does not look like a multimodal biometric window summary. ",
      "Missing columns: ",
      paste(missing_summary_columns, collapse = ", "),
      call. = FALSE
    )
  }

  if (is.null(participant_column)) {
    participant_column <- infer_biometric_participant_column(windows)
  }

  window_recommendations <- create_biometric_window_recommendations(
    windows = windows,
    gsr_min_usable_pct = gsr_min_usable_pct,
    hr_min_usable_pct = hr_min_usable_pct,
    dial_min_usable_pct = dial_min_usable_pct,
    require_gsr = require_gsr,
    require_hr = require_hr,
    require_dial = require_dial
  )

  participant_recommendations <- create_biometric_participant_recommendations(
    window_recommendations = window_recommendations,
    participant_column = participant_column
  )

  overview <- data.frame(
    n_windows = nrow(window_recommendations),
    keep_windows = sum(window_recommendations$recommendation == "keep"),
    review_windows = sum(window_recommendations$recommendation == "review"),
    exclude_windows = sum(window_recommendations$recommendation == "exclude"),
    participant_column = ifelse(is.na(participant_column), NA_character_, participant_column),
    n_participants = ifelse(
      nrow(participant_recommendations) > 0L,
      nrow(participant_recommendations),
      NA_integer_
    ),
    stringsAsFactors = FALSE
  )

  settings <- data.frame(
    gsr_min_usable_pct = gsr_min_usable_pct,
    hr_min_usable_pct = hr_min_usable_pct,
    dial_min_usable_pct = dial_min_usable_pct,
    require_gsr = require_gsr,
    require_hr = require_hr,
    require_dial = require_dial,
    stringsAsFactors = FALSE
  )

  out <- list(
    overview = overview,
    window_recommendations = window_recommendations,
    participant_recommendations = participant_recommendations,
    settings = settings
  )

  class(out) <- c("gazepoint_biometric_exclusion_recommendations", "list")
  out
}


infer_biometric_participant_column <- function(data) {
  candidates <- c("source_participant", "USER", "USERID", "participant", "subject")

  present <- intersect(candidates, names(data))

  if (length(present) == 0L) {
    return(NA_character_)
  }

  present[1]
}


create_biometric_window_recommendations <- function(windows,
                                                    gsr_min_usable_pct,
                                                    hr_min_usable_pct,
                                                    dial_min_usable_pct,
                                                    require_gsr,
                                                    require_hr,
                                                    require_dial) {
  out <- windows

  out$low_gsr_usable <- is_low_usable_pct(
    out$gsr_usable_pct,
    gsr_min_usable_pct
  )

  out$low_hr_usable <- is_low_usable_pct(
    out$hr_usable_pct,
    hr_min_usable_pct
  )

  out$low_dial_usable <- is_low_usable_pct(
    out$dial_usable_pct,
    dial_min_usable_pct
  )

  required_fail <- rep(FALSE, nrow(out))

  if (isTRUE(require_gsr)) {
    required_fail <- required_fail | out$low_gsr_usable
  }

  if (isTRUE(require_hr)) {
    required_fail <- required_fail | out$low_hr_usable
  }

  if (isTRUE(require_dial)) {
    required_fail <- required_fail | out$low_dial_usable
  }

  optional_warn <- rep(FALSE, nrow(out))

  if (!isTRUE(require_gsr)) {
    optional_warn <- optional_warn | out$low_gsr_usable
  }

  if (!isTRUE(require_hr)) {
    optional_warn <- optional_warn | out$low_hr_usable
  }

  if (!isTRUE(require_dial)) {
    optional_warn <- optional_warn | out$low_dial_usable
  }

  out$recommendation <- ifelse(
    required_fail,
    "exclude",
    ifelse(optional_warn, "review", "keep")
  )

  out$recommendation_reason <- vapply(
    seq_len(nrow(out)),
    function(i) {
      reasons <- character(0)

      if (out$low_gsr_usable[i]) {
        reasons <- c(
          reasons,
          paste0("GSR/EDA usable coverage below ", gsr_min_usable_pct, "%")
        )
      }

      if (out$low_hr_usable[i]) {
        reasons <- c(
          reasons,
          paste0("Heart-rate usable coverage below ", hr_min_usable_pct, "%")
        )
      }

      if (out$low_dial_usable[i]) {
        reasons <- c(
          reasons,
          paste0("Engagement-dial usable coverage below ", dial_min_usable_pct, "%")
        )
      }

      if (length(reasons) == 0L) {
        return("usable biometric coverage acceptable")
      }

      paste(reasons, collapse = "; ")
    },
    character(1)
  )

  out
}


is_low_usable_pct <- function(x, threshold) {
  is.na(x) | x < threshold
}


create_biometric_participant_recommendations <- function(window_recommendations,
                                                         participant_column) {
  if (is.na(participant_column) || !participant_column %in% names(window_recommendations)) {
    return(data.frame(
      participant = character(0),
      n_windows = integer(0),
      keep_windows = integer(0),
      review_windows = integer(0),
      exclude_windows = integer(0),
      exclude_pct = numeric(0),
      participant_recommendation = character(0),
      stringsAsFactors = FALSE
    ))
  }

  participants <- unique(as.character(window_recommendations[[participant_column]]))

  rows <- lapply(participants, function(participant) {
    in_participant <- as.character(window_recommendations[[participant_column]]) == participant

    recommendations <- window_recommendations$recommendation[in_participant]

    n_windows <- length(recommendations)
    keep_windows <- sum(recommendations == "keep")
    review_windows <- sum(recommendations == "review")
    exclude_windows <- sum(recommendations == "exclude")

    participant_recommendation <- if (exclude_windows == n_windows) {
      "exclude"
    } else if (exclude_windows > 0L || review_windows > 0L) {
      "review"
    } else {
      "keep"
    }

    data.frame(
      participant = participant,
      n_windows = n_windows,
      keep_windows = keep_windows,
      review_windows = review_windows,
      exclude_windows = exclude_windows,
      exclude_pct = safe_pct(exclude_windows, n_windows),
      participant_recommendation = participant_recommendation,
      stringsAsFactors = FALSE
    )
  })

  out <- do.call(rbind, rows)
  rownames(out) <- NULL
  out
}

test_that("recommend_gazepoint_biometric_exclusions works from window summaries", {
  windows <- data.frame(
    source_participant = c("User 0", "User 0", "User 1", "User 1"),
    MEDIA_ID = c(0, 1, 0, 1),
    gsr_usable_pct = c(0, 0, 100, 100),
    hr_usable_pct = c(0, 0, 100, 90),
    dial_usable_pct = c(0, 0, 100, 100)
  )

  out <- recommend_gazepoint_biometric_exclusions(
    data = windows,
    data_is_window_summary = TRUE,
    require_gsr = TRUE,
    require_hr = TRUE,
    require_dial = FALSE
  )

  expect_s3_class(out, "gazepoint_biometric_exclusion_recommendations")
  expect_true(is.data.frame(out$overview))
  expect_true(is.data.frame(out$window_recommendations))
  expect_true(is.data.frame(out$participant_recommendations))

  expect_equal(out$overview$n_windows, 4)
  expect_equal(out$overview$exclude_windows, 2)
  expect_equal(out$overview$keep_windows, 2)

  user0 <- out$participant_recommendations[
    out$participant_recommendations$participant == "User 0",
  ]

  user1 <- out$participant_recommendations[
    out$participant_recommendations$participant == "User 1",
  ]

  expect_equal(user0$participant_recommendation, "exclude")
  expect_equal(user1$participant_recommendation, "keep")
})


test_that("recommend_gazepoint_biometric_exclusions can review optional dial coverage", {
  windows <- data.frame(
    source_participant = "User 1",
    MEDIA_ID = 0,
    gsr_usable_pct = 100,
    hr_usable_pct = 100,
    dial_usable_pct = 0
  )

  out <- recommend_gazepoint_biometric_exclusions(
    data = windows,
    data_is_window_summary = TRUE,
    require_gsr = TRUE,
    require_hr = TRUE,
    require_dial = FALSE
  )

  expect_equal(out$window_recommendations$recommendation, "review")
  expect_true(grepl("Engagement-dial", out$window_recommendations$recommendation_reason))
})


test_that("recommend_gazepoint_biometric_exclusions works from row-level data", {
  dat <- data.frame(
    source_participant = c("User 1", "User 1", "User 1"),
    MEDIA_ID = c(0, 0, 0),
    GSR_US = c(1.1, 1.2, 1.3),
    GSRV = c(1, 1, 1),
    HR = c(70, 72, 74),
    HRV = c(1, 1, 1),
    DIAL = c(1, 1, 1),
    DIALV = c(1, 1, 1)
  )

  out <- recommend_gazepoint_biometric_exclusions(
    data = dat,
    group_columns = c("source_participant", "MEDIA_ID")
  )

  expect_equal(out$overview$n_windows, 1)
  expect_equal(out$window_recommendations$recommendation, "keep")
})


test_that("recommend_gazepoint_biometric_exclusions rejects missing summary columns", {
  windows <- data.frame(
    source_participant = "User 1",
    MEDIA_ID = 0,
    gsr_usable_pct = 100
  )

  expect_error(
    recommend_gazepoint_biometric_exclusions(
      data = windows,
      data_is_window_summary = TRUE
    ),
    "Missing columns"
  )
})


test_that("recommend_gazepoint_biometric_exclusions requires group columns for row-level data", {
  dat <- data.frame(
    GSR_US = c(1.1, 1.2),
    GSRV = c(1, 1)
  )

  expect_error(
    recommend_gazepoint_biometric_exclusions(dat),
    "group_columns"
  )
})

test_that("diagnose_gazepoint_biometrics_workflow returns pass for clean workflow", {
  folder <- tempfile("gazepoint_diagnose_")
  dir.create(folder)

  file_one <- file.path(folder, "User 0_all_gaze.csv")

  writeLines(
    c(
      "USER,MEDIA_ID,CNT,GSR_US,GSRV,HR,HRV,DIAL,DIALV,TTL0,TTLV,",
      "U1,1,1,2.0,1,70,1,0.1,1,1007,1,",
      "U1,1,2,2.2,1,72,1,0.2,1,1008,1,"
    ),
    file_one,
    useBytes = TRUE
  )

  workflow <- run_gazepoint_biometrics_workflow(
    path = folder,
    group_columns = c("USER", "MEDIA_ID")
  )

  out <- diagnose_gazepoint_biometrics_workflow(workflow)

  expect_true(is.data.frame(out))
  expect_equal(nrow(out), 1)
  expect_equal(out$final_status, "pass")
  expect_true(out$active_gsr_eda)
  expect_true(out$active_heart_rate)
})


test_that("diagnose_gazepoint_biometrics_workflow fails when excluded windows exceed threshold", {
  windows <- data.frame(
    source_participant = c("User 0", "User 0"),
    MEDIA_ID = c(0, 1),
    gsr_usable_pct = c(0, 0),
    hr_usable_pct = c(0, 0),
    dial_usable_pct = c(0, 0)
  )

  folder <- tempfile("gazepoint_diagnose_")
  dir.create(folder)

  file_one <- file.path(folder, "User 0_all_gaze.csv")

  writeLines(
    c(
      "USER,MEDIA_ID,CNT,GSR_US,GSRV,HR,HRV,DIAL,DIALV,TTL0,TTLV,",
      "U1,1,1,0,0,0,0,0,0,0,0,",
      "U1,1,2,0,0,0,0,0,0,0,0,"
    ),
    file_one,
    useBytes = TRUE
  )

  workflow <- run_gazepoint_biometrics_workflow(
    path = folder,
    group_columns = c("USER", "MEDIA_ID"),
    require_active_signal = FALSE
  )

  workflow$exclusion_recommendations <- recommend_gazepoint_biometric_exclusions(
    windows,
    data_is_window_summary = TRUE
  )

  out <- diagnose_gazepoint_biometrics_workflow(
    workflow,
    require_gsr = FALSE,
    require_hr = FALSE
  )

  expect_equal(out$final_status, "fail")
  expect_true(out$exclude_window_pct > 25)
})

test_that("diagnose_gazepoint_biometrics_workflow reviews low usable quality", {
  folder <- tempfile("gazepoint_diagnose_")
  dir.create(folder)

  file_one <- file.path(folder, "User 0_all_gaze.csv")

  writeLines(
    c(
      "USER,MEDIA_ID,CNT,GSR_US,GSRV,HR,HRV,DIAL,DIALV,TTL0,TTLV,",
      "U1,1,1,0,0,70,1,1,1,1007,1,",
      "U1,1,2,0,0,72,1,1,1,1008,1,",
      "U1,1,3,2.2,1,74,1,1,1,1009,1,"
    ),
    file_one,
    useBytes = TRUE
  )

  workflow <- run_gazepoint_biometrics_workflow(
    path = folder,
    group_columns = c("USER", "MEDIA_ID")
  )

  out <- diagnose_gazepoint_biometrics_workflow(
    workflow,
    max_exclude_window_pct = 100
  )

  expect_equal(out$final_status, "review")
  expect_true(out$low_quality_signal_count >= 1)
  expect_true(grepl("low usable coverage", out$diagnostic_reasons))
})

test_that("diagnose_gazepoint_biometrics_workflow rejects invalid object", {
  expect_error(
    diagnose_gazepoint_biometrics_workflow(list()),
    "workflow"
  )
})

#' Write Gazepoint Biometrics report tables
#'
#' Writes report-ready Gazepoint Biometrics tables to CSV files. The input can
#' be a workflow object produced by `run_gazepoint_biometrics_workflow()`, a
#' report-table object produced by `create_gazepoint_biometrics_report_tables()`,
#' or a named list of data frames.
#'
#' @param tables A Gazepoint Biometrics workflow object, report-table object, or
#'   named list of data frames.
#' @param output_dir Output directory for CSV files.
#' @param prefix Filename prefix.
#' @param overwrite Should existing files be overwritten?
#' @param include_empty_message_tables Should placeholder tables containing only
#'   a `message` column be written?
#'
#' @return A data frame indexing written and skipped files.
#'
#' @export
write_gazepoint_biometrics_report_tables <- function(tables,
                                                     output_dir,
                                                     prefix = "gazepoint_biometrics",
                                                     overwrite = TRUE,
                                                     include_empty_message_tables = FALSE) {
  if (missing(output_dir) || is.null(output_dir) || length(output_dir) != 1L) {
    stop("`output_dir` must be supplied as a single folder path.", call. = FALSE)
  }

  if (inherits(tables, "gazepoint_biometrics_workflow")) {
    tables <- create_gazepoint_biometrics_report_tables(workflow = tables)
  }

  if (is.data.frame(tables) || !is.list(tables)) {
    stop(
      "`tables` must be a workflow object, a report-table object, or a named list of data frames.",
      call. = FALSE
    )
  }

  if (!dir.exists(output_dir)) {
    dir.create(output_dir, recursive = TRUE, showWarnings = FALSE)
  }

  if (!dir.exists(output_dir)) {
    stop("Could not create `output_dir`: ", output_dir, call. = FALSE)
  }

  table_names <- names(tables)

  if (is.null(table_names)) {
    table_names <- paste0("table_", seq_along(tables))
  }

  table_names[table_names == ""] <- paste0("table_", which(table_names == ""))

  index_rows <- lapply(seq_along(tables), function(i) {
    table_name <- table_names[i]
    table <- tables[[i]]

    file_name <- paste0(
      safe_filename_component(prefix),
      "_",
      safe_filename_component(table_name),
      ".csv"
    )

    file_path <- file.path(output_dir, file_name)

    if (!is.data.frame(table)) {
      return(data.frame(
        table = table_name,
        file = file_path,
        n_rows = NA_integer_,
        n_columns = NA_integer_,
        written = FALSE,
        skipped_reason = "not_a_data_frame",
        stringsAsFactors = FALSE
      ))
    }

    is_message_only <- identical(names(table), "message")

    if (is_message_only && !isTRUE(include_empty_message_tables)) {
      return(data.frame(
        table = table_name,
        file = file_path,
        n_rows = nrow(table),
        n_columns = ncol(table),
        written = FALSE,
        skipped_reason = "message_only_table",
        stringsAsFactors = FALSE
      ))
    }

    if (file.exists(file_path) && !isTRUE(overwrite)) {
      return(data.frame(
        table = table_name,
        file = file_path,
        n_rows = nrow(table),
        n_columns = ncol(table),
        written = FALSE,
        skipped_reason = "file_exists",
        stringsAsFactors = FALSE
      ))
    }

    utils::write.csv(
      table,
      file = file_path,
      row.names = FALSE,
      na = ""
    )

    data.frame(
      table = table_name,
      file = file_path,
      n_rows = nrow(table),
      n_columns = ncol(table),
      written = TRUE,
      skipped_reason = NA_character_,
      stringsAsFactors = FALSE
    )
  })

  out <- do.call(rbind, index_rows)
  rownames(out) <- NULL
  out
}


safe_filename_component <- function(x) {
  x <- as.character(x)
  x <- gsub("[^A-Za-z0-9_\\-]+", "_", x)
  x <- gsub("_+", "_", x)
  x <- gsub("^_|_$", "", x)

  ifelse(nchar(x) == 0L, "table", x)
}

test_that("write_gazepoint_biometrics_report_tables writes named data frames", {
  output_dir <- tempfile("gazepoint_report_tables_")

  tables <- list(
    overview = data.frame(n_rows = 2, n_columns = 5),
    quality = data.frame(signal = "gsr_eda", usable_pct = 100),
    missing = data.frame(message = "No table supplied.")
  )

  index <- write_gazepoint_biometrics_report_tables(
    tables = tables,
    output_dir = output_dir,
    prefix = "test_report"
  )

  expect_true(is.data.frame(index))
  expect_equal(nrow(index), 3)
  expect_equal(sum(index$written), 2)
  expect_equal(
    index$skipped_reason[index$table == "missing"],
    "message_only_table"
  )

  written_files <- index$file[index$written]

  expect_true(all(file.exists(written_files)))

  overview <- utils::read.csv(
    index$file[index$table == "overview"],
    stringsAsFactors = FALSE
  )

  expect_equal(overview$n_rows, 2)
})


test_that("write_gazepoint_biometrics_report_tables can write workflow report tables", {
  folder <- tempfile("gazepoint_report_export_")
  dir.create(folder)

  file_one <- file.path(folder, "User 0_all_gaze.csv")

  writeLines(
    c(
      "USER,MEDIA_ID,CNT,GSR_US,GSRV,HR,HRV,DIAL,DIALV,TTL0,TTLV,",
      "U1,1,1,2.0,1,70,1,0.1,1,1007,1,",
      "U1,1,2,2.2,1,72,1,0.2,1,1008,1,"
    ),
    file_one,
    useBytes = TRUE
  )

  workflow <- run_gazepoint_biometrics_workflow(
    path = folder,
    group_columns = c("USER", "MEDIA_ID")
  )

  output_dir <- tempfile("gazepoint_report_tables_")

  index <- write_gazepoint_biometrics_report_tables(
    tables = workflow,
    output_dir = output_dir,
    prefix = "workflow_report"
  )

  expect_true(is.data.frame(index))
  expect_true(any(index$table == "overview"))
  expect_true(any(index$table == "quality"))
  expect_true(any(index$written))
  expect_true(all(file.exists(index$file[index$written])))
})


test_that("write_gazepoint_biometrics_report_tables respects overwrite false", {
  output_dir <- tempfile("gazepoint_report_tables_")

  tables <- list(
    overview = data.frame(n_rows = 2)
  )

  first <- write_gazepoint_biometrics_report_tables(
    tables = tables,
    output_dir = output_dir,
    prefix = "test_report"
  )

  second <- write_gazepoint_biometrics_report_tables(
    tables = tables,
    output_dir = output_dir,
    prefix = "test_report",
    overwrite = FALSE
  )

  expect_true(first$written)
  expect_false(second$written)
  expect_equal(second$skipped_reason, "file_exists")
})


test_that("write_gazepoint_biometrics_report_tables can include message tables", {
  output_dir <- tempfile("gazepoint_report_tables_")

  tables <- list(
    missing = data.frame(message = "No table supplied.")
  )

  index <- write_gazepoint_biometrics_report_tables(
    tables = tables,
    output_dir = output_dir,
    include_empty_message_tables = TRUE
  )

  expect_true(index$written)
  expect_true(file.exists(index$file))
})


test_that("write_gazepoint_biometrics_report_tables rejects invalid input", {
  expect_error(
    write_gazepoint_biometrics_report_tables(
      tables = data.frame(x = 1),
      output_dir = tempfile("gazepoint_report_tables_")
    ),
    "tables"
  )
})

#' Run automated exploratory statistics for Gazepoint feature tables
#'
#' Runs simple exploratory group comparisons for numeric feature columns.
#' The function selects one-way ANOVA when all groups pass Shapiro-Wilk checks
#' and Kruskal-Wallis otherwise. It also performs pairwise post-hoc tests with
#' multiplicity correction.
#'
#' This is an exploratory reporting helper. It is not a substitute for a
#' preregistered statistical model or expert review of study design.
#'
#' @param dat A data frame.
#' @param outcome_cols Numeric outcome columns.
#' @param group_col Grouping/condition column.
#' @param alpha Significance level.
#' @param p_adjust_method P-value adjustment method.
#' @param normality_alpha Alpha used for Shapiro-Wilk normality screening.
#' @param min_group_n Minimum observations per group.
#'
#' @return A list with `overview`, `test_table`, `posthoc_table`,
#'   `normality_table`, and `settings`.
#' @export
run_gazepoint_automated_statistics <- function(dat,
                                               outcome_cols,
                                               group_col,
                                               alpha = 0.05,
                                               p_adjust_method = "holm",
                                               normality_alpha = 0.05,
                                               min_group_n = 3) {
  if (!is.data.frame(dat)) {
    stop("`dat` must be a data frame.", call. = FALSE)
  }

  required <- c(outcome_cols, group_col)
  missing_required <- setdiff(required, names(dat))

  if (length(missing_required) > 0) {
    stop("Missing required columns: ", paste(missing_required, collapse = ", "), call. = FALSE)
  }

  non_numeric <- outcome_cols[!vapply(dat[outcome_cols], is.numeric, logical(1))]
  if (length(non_numeric) > 0) {
    stop("These `outcome_cols` are not numeric: ", paste(non_numeric, collapse = ", "), call. = FALSE)
  }

  test_rows <- list()
  normality_rows <- list()
  posthoc_rows <- list()
  test_id <- 1L
  norm_id <- 1L
  posthoc_id <- 1L

  for (outcome in outcome_cols) {
    d <- dat[, c(outcome, group_col), drop = FALSE]
    names(d) <- c("outcome", "group")
    d <- d[is.finite(d$outcome) & !is.na(d$group), , drop = FALSE]
    d$group <- as.factor(d$group)

    group_counts <- table(d$group)

    if (nrow(d) < min_group_n * 2 || length(group_counts) < 2 || any(group_counts < min_group_n)) {
      test_rows[[test_id]] <- data.frame(
        outcome = outcome,
        test = NA_character_,
        statistic = NA_real_,
        df1 = NA_real_,
        df2 = NA_real_,
        p_value = NA_real_,
        p_adjusted = NA_real_,
        normality_screen_passed = NA,
        status = "insufficient_group_data",
        stringsAsFactors = FALSE
      )
      test_id <- test_id + 1L
      next
    }

    group_levels <- levels(d$group)
    normality_p <- numeric(length(group_levels))

    for (i in seq_along(group_levels)) {
      x <- d$outcome[d$group == group_levels[i]]
      normality_p[i] <- gpbiometrics_auto_stats_shapiro(x)

      normality_rows[[norm_id]] <- data.frame(
        outcome = outcome,
        group = group_levels[i],
        n = length(x),
        shapiro_p = normality_p[i],
        normality_passed = is.finite(normality_p[i]) && normality_p[i] >= normality_alpha,
        stringsAsFactors = FALSE
      )
      norm_id <- norm_id + 1L
    }

    normality_passed <- all(is.finite(normality_p) & normality_p >= normality_alpha)

    if (normality_passed) {
      fit <- stats::aov(outcome ~ group, data = d)
      tab <- summary(fit)[[1]]
      statistic <- unname(tab$`F value`[1])
      p_value <- unname(tab$`Pr(>F)`[1])
      df1 <- unname(tab$Df[1])
      df2 <- unname(tab$Df[2])
      test_name <- "one_way_anova"

      pairwise <- stats::pairwise.t.test(
        x = d$outcome,
        g = d$group,
        p.adjust.method = p_adjust_method,
        pool.sd = FALSE
      )
    } else {
      kt <- stats::kruskal.test(outcome ~ group, data = d)
      statistic <- unname(kt$statistic)
      p_value <- kt$p.value
      df1 <- unname(kt$parameter)
      df2 <- NA_real_
      test_name <- "kruskal_wallis"

      pairwise <- stats::pairwise.wilcox.test(
        x = d$outcome,
        g = d$group,
        p.adjust.method = p_adjust_method,
        exact = FALSE
      )
    }

    test_rows[[test_id]] <- data.frame(
      outcome = outcome,
      test = test_name,
      statistic = statistic,
      df1 = df1,
      df2 = df2,
      p_value = p_value,
      p_adjusted = NA_real_,
      normality_screen_passed = normality_passed,
      status = "test_completed",
      stringsAsFactors = FALSE
    )
    test_id <- test_id + 1L

    posthoc_dat <- gpbiometrics_auto_stats_pairwise_table(pairwise$p.value, outcome)

    if (nrow(posthoc_dat) > 0) {
      for (i in seq_len(nrow(posthoc_dat))) {
        posthoc_rows[[posthoc_id]] <- posthoc_dat[i, , drop = FALSE]
        posthoc_id <- posthoc_id + 1L
      }
    }
  }

  test_table <- do.call(rbind, test_rows)
  rownames(test_table) <- NULL

  if (any(is.finite(test_table$p_value))) {
    ok <- is.finite(test_table$p_value)
    test_table$p_adjusted[ok] <- stats::p.adjust(test_table$p_value[ok], method = p_adjust_method)
  }

  normality_table <- if (length(normality_rows) > 0) {
    do.call(rbind, normality_rows)
  } else {
    data.frame()
  }

  posthoc_table <- if (length(posthoc_rows) > 0) {
    do.call(rbind, posthoc_rows)
  } else {
    data.frame()
  }

  if (nrow(posthoc_table) > 0 && "p_adjusted" %in% names(posthoc_table)) {
    posthoc_table$significant <- is.finite(posthoc_table$p_adjusted) &
      posthoc_table$p_adjusted < alpha
  }

  overview <- data.frame(
    outcome_count = length(outcome_cols),
    completed_tests = sum(test_table$status == "test_completed"),
    insufficient_tests = sum(test_table$status != "test_completed"),
    significant_tests = sum(is.finite(test_table$p_adjusted) & test_table$p_adjusted < alpha),
    posthoc_rows = nrow(posthoc_table),
    status = if (all(test_table$status == "test_completed")) {
      "automated_statistics_complete"
    } else if (any(test_table$status == "test_completed")) {
      "automated_statistics_partial"
    } else {
      "automated_statistics_failed"
    },
    interpretation = paste(
      "Automated statistics are exploratory convenience summaries.",
      "They should be reviewed against the study design and are not a substitute for preregistered modelling."
    ),
    stringsAsFactors = FALSE
  )

  structure(
    list(
      overview = overview,
      test_table = test_table,
      posthoc_table = posthoc_table,
      normality_table = normality_table,
      settings = list(
        outcome_cols = outcome_cols,
        group_col = group_col,
        alpha = alpha,
        p_adjust_method = p_adjust_method,
        normality_alpha = normality_alpha,
        min_group_n = min_group_n
      )
    ),
    class = c("gazepoint_automated_statistics", "list")
  )
}

gpbiometrics_auto_stats_shapiro <- function(x) {
  x <- x[is.finite(x)]

  if (length(x) < 3 || length(x) > 5000 || stats::sd(x) == 0) {
    return(NA_real_)
  }

  stats::shapiro.test(x)$p.value
}

gpbiometrics_auto_stats_pairwise_table <- function(p_matrix, outcome) {
  if (is.null(p_matrix) || length(p_matrix) == 0) {
    return(data.frame())
  }

  rows <- list()
  row_id <- 1L

  for (i in seq_len(nrow(p_matrix))) {
    for (j in seq_len(ncol(p_matrix))) {
      p <- p_matrix[i, j]

      if (is.finite(p)) {
        rows[[row_id]] <- data.frame(
          outcome = outcome,
          group_1 = rownames(p_matrix)[i],
          group_2 = colnames(p_matrix)[j],
          p_adjusted = p,
          stringsAsFactors = FALSE
        )
        row_id <- row_id + 1L
      }
    }
  }

  if (length(rows) == 0) {
    return(data.frame())
  }

  do.call(rbind, rows)
}

#' Standardize biometric signals within participant or other analysis units
#'
#' Adds within-unit standardized biometric columns, usually within participant
#' or participant-by-session/stimulus groups. This is useful when the analysis
#' focuses on relative within-person signal change rather than absolute
#' between-person level differences.
#'
#' The helper is intentionally conservative. It does not run automatically in
#' the main workflow and does not infer emotion, valence, stress, trust,
#' preference, cognition, or diagnosis. Within-unit z-scoring removes
#' between-unit level and scale differences and should therefore be reported
#' explicitly.
#'
#' @param data A data frame containing Gazepoint biometric data.
#' @param signal_cols Character vector of biometric signal columns to
#'   standardize. If `NULL`, common numeric biometric columns are detected.
#' @param unit_cols Character vector defining the unit within which means and
#'   standard deviations are computed. If `NULL`, common participant/session
#'   columns are detected. If no columns are detected, the whole data frame is
#'   treated as one unit.
#' @param reference_col Optional logical or categorical column identifying rows
#'   used to estimate the reference mean and standard deviation. For example,
#'   this can be a baseline-window flag. The resulting parameters are then
#'   applied to all rows in the same unit.
#' @param reference_value Value in `reference_col` that marks reference rows.
#'   Defaults to `TRUE`.
#' @param suffix Suffix for standardized output columns.
#' @param center Logical. If `TRUE`, subtract the within-unit reference mean.
#' @param scale Logical. If `TRUE`, divide by the within-unit reference
#'   standard deviation.
#' @param min_valid Minimum number of finite reference observations required
#'   per unit and signal.
#' @param zero_sd_action What to do when the within-unit standard deviation is
#'   zero or unavailable. `"NA"` returns `NA`; `"zero"` returns zero for finite
#'   centered values.
#' @param overwrite Logical. If `FALSE`, existing output columns are protected.
#'
#' @return A data frame with added standardized columns. Attributes include
#'   `standardization_summary`, `standardization_parameters`, and `settings`.
#' @export
standardize_gazepoint_biometrics_within_unit <- function(data,
                                                         signal_cols = NULL,
                                                         unit_cols = NULL,
                                                         reference_col = NULL,
                                                         reference_value = TRUE,
                                                         suffix = "_z_within",
                                                         center = TRUE,
                                                         scale = TRUE,
                                                         min_valid = 2,
                                                         zero_sd_action = c("NA", "zero"),
                                                         overwrite = FALSE) {
  if (!is.data.frame(data)) {
    stop("`data` must be a data frame.", call. = FALSE)
  }

  zero_sd_action <- match.arg(zero_sd_action)

  gpbiometrics_within_z_validate_logical(center, "center")
  gpbiometrics_within_z_validate_logical(scale, "scale")
  gpbiometrics_within_z_validate_logical(overwrite, "overwrite")

  if (!is.character(suffix) || length(suffix) != 1 || is.na(suffix) || !nzchar(suffix)) {
    stop("`suffix` must be a non-empty character string.", call. = FALSE)
  }

  if (!is.numeric(min_valid) || length(min_valid) != 1 || is.na(min_valid) || min_valid < 1) {
    stop("`min_valid` must be a positive number.", call. = FALSE)
  }

  min_valid <- as.integer(min_valid)

  signal_cols <- gpbiometrics_within_z_resolve_signal_cols(data, signal_cols)
  unit_cols <- gpbiometrics_within_z_resolve_unit_cols(data, unit_cols)

  if (!is.null(reference_col)) {
    if (!is.character(reference_col) || length(reference_col) != 1 || is.na(reference_col)) {
      stop("`reference_col` must be NULL or a single column name.", call. = FALSE)
    }

    if (!reference_col %in% names(data)) {
      stop("Column `", reference_col, "` was not found in `data`.", call. = FALSE)
    }
  }

  output_cols <- paste0(signal_cols, suffix)

  if (!isTRUE(overwrite)) {
    existing <- intersect(output_cols, names(data))

    if (length(existing) > 0) {
      stop(
        "The following output columns already exist: ",
        paste(existing, collapse = ", "),
        ". Use `overwrite = TRUE` to replace them.",
        call. = FALSE
      )
    }
  }

  out <- data

  for (nm in output_cols) {
    out[[nm]] <- NA_real_
  }

  unit_index <- gpbiometrics_within_z_split_indices(data, unit_cols)

  parameter_rows <- list()

  row_id <- 1L

  for (unit_id in names(unit_index)) {
    idx <- unit_index[[unit_id]]
    unit_values <- gpbiometrics_within_z_unit_values(data, idx, unit_cols, unit_id)

    reference_idx <- idx

    if (!is.null(reference_col)) {
      ref_values <- data[[reference_col]][idx]
      keep_ref <- !is.na(ref_values) & ref_values == reference_value
      reference_idx <- idx[keep_ref]
    }

    for (j in seq_along(signal_cols)) {
      signal_col <- signal_cols[j]
      output_col <- output_cols[j]

      x_reference <- data[[signal_col]][reference_idx]
      finite_reference <- is.finite(x_reference)
      n_reference_finite <- sum(finite_reference)

      status <- "standardized"
      reference_mean <- NA_real_
      reference_sd <- NA_real_

      if (n_reference_finite < min_valid) {
        status <- "insufficient_reference_rows"
        out[[output_col]][idx] <- NA_real_
      } else {
        reference_mean <- mean(x_reference[finite_reference], na.rm = TRUE)
        reference_sd <- stats::sd(x_reference[finite_reference], na.rm = TRUE)

        x_all <- data[[signal_col]][idx]

        transformed <- x_all

        if (isTRUE(center)) {
          transformed <- transformed - reference_mean
        }

        if (isTRUE(scale)) {
          if (!is.finite(reference_sd) || reference_sd == 0) {
            status <- "zero_or_missing_sd"

            if (identical(zero_sd_action, "zero")) {
              transformed[is.finite(transformed)] <- 0
              transformed[!is.finite(transformed)] <- NA_real_
            } else {
              transformed[] <- NA_real_
            }
          } else {
            transformed <- transformed / reference_sd
          }
        }

        transformed[!is.finite(transformed)] <- NA_real_
        out[[output_col]][idx] <- transformed
      }

      parameter_rows[[row_id]] <- data.frame(
        unit_values,
        unit_id = unit_id,
        signal_col = signal_col,
        output_col = output_col,
        n_rows = length(idx),
        n_reference_rows = length(reference_idx),
        n_reference_finite = n_reference_finite,
        reference_mean = reference_mean,
        reference_sd = reference_sd,
        center = center,
        scale = scale,
        status = status,
        stringsAsFactors = FALSE,
        check.names = FALSE
      )

      row_id <- row_id + 1L
    }
  }

  parameters <- do.call(rbind, parameter_rows)
  rownames(parameters) <- NULL

  summary <- data.frame(
    input_rows = nrow(data),
    signal_count = length(signal_cols),
    output_count = length(output_cols),
    unit_count = length(unit_index),
    parameter_rows = nrow(parameters),
    standardized_rows = sum(parameters$status == "standardized"),
    problem_rows = sum(parameters$status != "standardized"),
    center = center,
    scale = scale,
    reference_col = if (is.null(reference_col)) NA_character_ else reference_col,
    suffix = suffix,
    status = if (all(parameters$status == "standardized")) {
      "within_unit_standardization_complete"
    } else if (any(parameters$status == "standardized")) {
      "within_unit_standardization_partial"
    } else {
      "within_unit_standardization_failed"
    },
    interpretation = paste(
      "Within-unit standardization rescales biometric signals relative to each unit's own reference distribution.",
      "It supports within-person comparison but removes between-unit level and scale differences.",
      "It does not infer emotion, valence, stress, trust, preference, cognition, or diagnosis."
    ),
    stringsAsFactors = FALSE
  )

  attr(out, "standardization_summary") <- summary
  attr(out, "standardization_parameters") <- parameters
  attr(out, "settings") <- list(
    signal_cols = signal_cols,
    unit_cols = unit_cols,
    reference_col = reference_col,
    reference_value = reference_value,
    suffix = suffix,
    center = center,
    scale = scale,
    min_valid = min_valid,
    zero_sd_action = zero_sd_action,
    overwrite = overwrite
  )

  class(out) <- unique(c("gazepoint_within_unit_standardized", class(out)))

  out
}

#' @rdname standardize_gazepoint_biometrics_within_unit
#' @export
standardise_gazepoint_biometrics_within_unit <- function(data,
                                                         signal_cols = NULL,
                                                         unit_cols = NULL,
                                                         reference_col = NULL,
                                                         reference_value = TRUE,
                                                         suffix = "_z_within",
                                                         center = TRUE,
                                                         scale = TRUE,
                                                         min_valid = 2,
                                                         zero_sd_action = c("NA", "zero"),
                                                         overwrite = FALSE) {
  standardize_gazepoint_biometrics_within_unit(
    data = data,
    signal_cols = signal_cols,
    unit_cols = unit_cols,
    reference_col = reference_col,
    reference_value = reference_value,
    suffix = suffix,
    center = center,
    scale = scale,
    min_valid = min_valid,
    zero_sd_action = zero_sd_action,
    overwrite = overwrite
  )
}

gpbiometrics_within_z_resolve_signal_cols <- function(data, signal_cols = NULL) {
  if (!is.null(signal_cols)) {
    if (!is.character(signal_cols) || length(signal_cols) == 0) {
      stop("`signal_cols` must be NULL or a non-empty character vector.", call. = FALSE)
    }

    missing_cols <- setdiff(signal_cols, names(data))

    if (length(missing_cols) > 0) {
      stop(
        "The following `signal_cols` were not found in `data`: ",
        paste(missing_cols, collapse = ", "),
        call. = FALSE
      )
    }

    non_numeric <- signal_cols[!vapply(data[signal_cols], is.numeric, logical(1))]

    if (length(non_numeric) > 0) {
      stop(
        "The following `signal_cols` are not numeric: ",
        paste(non_numeric, collapse = ", "),
        call. = FALSE
      )
    }

    return(unique(signal_cols))
  }

  candidates <- c(
    "GSR_US",
    "GSR_US_PHASIC",
    "GSR_US_TONIC",
    "GSR",
    "EDA",
    "HR",
    "HRP",
    "IBI",
    "DIAL"
  )

  found <- intersect(candidates, names(data))
  found <- found[vapply(data[found], is.numeric, logical(1))]

  if (length(found) == 0) {
    stop(
      "No common numeric biometric signal columns were detected. Supply `signal_cols` explicitly.",
      call. = FALSE
    )
  }

  unique(found)
}

gpbiometrics_within_z_resolve_unit_cols <- function(data, unit_cols = NULL) {
  if (!is.null(unit_cols)) {
    if (!is.character(unit_cols)) {
      stop("`unit_cols` must be NULL or a character vector.", call. = FALSE)
    }

    missing_cols <- setdiff(unit_cols, names(data))

    if (length(missing_cols) > 0) {
      stop(
        "The following `unit_cols` were not found in `data`: ",
        paste(missing_cols, collapse = ", "),
        call. = FALSE
      )
    }

    return(unique(unit_cols))
  }

  candidates <- c(
    "source_participant",
    "participant",
    "participant_id",
    "subject",
    "subject_id",
    "USER_FILE",
    "source_file",
    "session",
    "session_id"
  )

  unique(intersect(candidates, names(data)))
}

gpbiometrics_within_z_split_indices <- function(data, unit_cols) {
  if (length(unit_cols) == 0) {
    return(list(all_rows = seq_len(nrow(data))))
  }

  unit_frame <- data[unit_cols]
  unit_frame[] <- lapply(unit_frame, function(x) {
    x <- as.character(x)
    x[is.na(x)] <- "<NA>"
    x
  })

  unit_key <- do.call(paste, c(unit_frame, sep = " | "))
  split(seq_len(nrow(data)), unit_key)
}

gpbiometrics_within_z_unit_values <- function(data, idx, unit_cols, unit_id) {
  if (length(unit_cols) == 0) {
    return(data.frame(
      unit_label = unit_id,
      stringsAsFactors = FALSE
    ))
  }

  values <- lapply(unit_cols, function(nm) {
    as.character(data[[nm]][idx[1]])
  })

  names(values) <- unit_cols

  as.data.frame(
    values,
    stringsAsFactors = FALSE,
    optional = TRUE
  )
}

gpbiometrics_within_z_validate_logical <- function(x, name) {
  if (!is.logical(x) || length(x) != 1 || is.na(x)) {
    stop("`", name, "` must be TRUE or FALSE.", call. = FALSE)
  }

  invisible(TRUE)
}

test_that("standardize_gazepoint_biometrics_within_unit standardizes within participant", {
  dat <- data.frame(
    participant = rep(c("p1", "p2"), each = 5),
    GSR_US = c(1, 2, 3, 4, 5, 10, 20, 30, 40, 50),
    HR = c(70, 71, 72, 73, 74, 80, 82, 84, 86, 88)
  )

  out <- standardize_gazepoint_biometrics_within_unit(
    dat,
    signal_cols = c("GSR_US", "HR"),
    unit_cols = "participant"
  )

  expect_s3_class(out, "gazepoint_within_unit_standardized")
  expect_true(all(c("GSR_US_z_within", "HR_z_within") %in% names(out)))

  by_participant <- split(out, out$participant)

  expect_equal(mean(by_participant$p1$GSR_US_z_within), 0, tolerance = 1e-10)
  expect_equal(stats::sd(by_participant$p1$GSR_US_z_within), 1, tolerance = 1e-10)
  expect_equal(mean(by_participant$p2$GSR_US_z_within), 0, tolerance = 1e-10)
  expect_equal(stats::sd(by_participant$p2$GSR_US_z_within), 1, tolerance = 1e-10)

  summary <- attr(out, "standardization_summary")
  parameters <- attr(out, "standardization_parameters")

  expect_equal(summary$status, "within_unit_standardization_complete")
  expect_equal(summary$unit_count, 2)
  expect_equal(summary$signal_count, 2)
  expect_true(is.data.frame(parameters))
  expect_equal(nrow(parameters), 4)
})

test_that("standardize_gazepoint_biometrics_within_unit uses reference rows when supplied", {
  dat <- data.frame(
    participant = rep(c("p1", "p2"), each = 6),
    baseline = rep(c(TRUE, TRUE, TRUE, FALSE, FALSE, FALSE), 2),
    GSR_US = c(1, 2, 3, 4, 5, 6, 10, 20, 30, 40, 50, 60)
  )

  out <- standardize_gazepoint_biometrics_within_unit(
    dat,
    signal_cols = "GSR_US",
    unit_cols = "participant",
    reference_col = "baseline",
    reference_value = TRUE
  )

  baseline_rows <- out$baseline

  p1_base <- out$participant == "p1" & baseline_rows
  p2_base <- out$participant == "p2" & baseline_rows

  expect_equal(mean(out$GSR_US_z_within[p1_base]), 0, tolerance = 1e-10)
  expect_equal(stats::sd(out$GSR_US_z_within[p1_base]), 1, tolerance = 1e-10)
  expect_equal(mean(out$GSR_US_z_within[p2_base]), 0, tolerance = 1e-10)
  expect_equal(stats::sd(out$GSR_US_z_within[p2_base]), 1, tolerance = 1e-10)

  parameters <- attr(out, "standardization_parameters")
  expect_true(all(parameters$n_reference_rows == 3))
})

test_that("standardize_gazepoint_biometrics_within_unit handles zero standard deviation conservatively", {
  dat <- data.frame(
    participant = rep(c("p1", "p2"), each = 4),
    GSR_US = c(1, 1, 1, 1, 1, 2, 3, 4)
  )

  out_na <- standardize_gazepoint_biometrics_within_unit(
    dat,
    signal_cols = "GSR_US",
    unit_cols = "participant",
    zero_sd_action = "NA"
  )

  expect_true(all(is.na(out_na$GSR_US_z_within[out_na$participant == "p1"])))
  expect_false(all(is.na(out_na$GSR_US_z_within[out_na$participant == "p2"])))

  out_zero <- standardize_gazepoint_biometrics_within_unit(
    dat,
    signal_cols = "GSR_US",
    unit_cols = "participant",
    zero_sd_action = "zero"
  )

  expect_true(all(out_zero$GSR_US_z_within[out_zero$participant == "p1"] == 0))

  summary <- attr(out_na, "standardization_summary")
  expect_equal(summary$status, "within_unit_standardization_partial")
})

test_that("standardise_gazepoint_biometrics_within_unit is an alias", {
  dat <- data.frame(
    participant = rep(c("p1", "p2"), each = 4),
    GSR_US = c(1, 2, 3, 4, 10, 20, 30, 40)
  )

  out_us <- standardize_gazepoint_biometrics_within_unit(
    dat,
    signal_cols = "GSR_US",
    unit_cols = "participant"
  )

  out_uk <- standardise_gazepoint_biometrics_within_unit(
    dat,
    signal_cols = "GSR_US",
    unit_cols = "participant"
  )

  expect_equal(out_us$GSR_US_z_within, out_uk$GSR_US_z_within)
})

test_that("standardize_gazepoint_biometrics_within_unit protects existing columns", {
  dat <- data.frame(
    participant = "p1",
    GSR_US = 1:4,
    GSR_US_z_within = 1:4
  )

  expect_error(
    standardize_gazepoint_biometrics_within_unit(
      dat,
      signal_cols = "GSR_US",
      unit_cols = "participant"
    ),
    "already exist"
  )

  out <- standardize_gazepoint_biometrics_within_unit(
    dat,
    signal_cols = "GSR_US",
    unit_cols = "participant",
    overwrite = TRUE
  )

  expect_true("GSR_US_z_within" %in% names(out))
})

test_that("standardize_gazepoint_biometrics_within_unit validates inputs", {
  dat <- data.frame(
    participant = "p1",
    GSR_US = 1:4,
    label = letters[1:4]
  )

  expect_error(
    standardize_gazepoint_biometrics_within_unit(
      dat,
      signal_cols = "missing_signal"
    ),
    "missing_signal"
  )

  expect_error(
    standardize_gazepoint_biometrics_within_unit(
      dat,
      signal_cols = "label"
    ),
    "not numeric"
  )

  expect_error(
    standardize_gazepoint_biometrics_within_unit(
      dat,
      signal_cols = "GSR_US",
      unit_cols = "missing_unit"
    ),
    "missing_unit"
  )
})

#' Export a Gazepoint biometrics report bundle
#'
#' Exports selected report tables, text outputs, optional plot objects, and a
#' manifest to a local output directory. This helper is intended for reproducible
#' reporting. It does not commit files and should not be used to export private
#' real-data outputs into a package repository.
#'
#' @param bundle Optional list-like object containing data frames, text, or plots.
#' @param output_dir Output directory.
#' @param prefix File prefix.
#' @param tables Optional named list of data frames to export as CSV files.
#' @param text Optional named list or character vector of text outputs to export
#'   as TXT files.
#' @param plots Optional named list of ggplot objects to export as PNG files.
#' @param include_readme Logical. Should a README text file be written?
#' @param include_session_info Logical. Should session information be written?
#' @param overwrite Logical. Should existing files be overwritten?
#'
#' @return A list with `overview`, `manifest`, `output_dir`, and `settings`.
#' @export
export_gazepoint_biometrics_report_bundle <- function(bundle = NULL,
                                                      output_dir,
                                                      prefix = "gpbiometrics_report",
                                                      tables = NULL,
                                                      text = NULL,
                                                      plots = NULL,
                                                      include_readme = TRUE,
                                                      include_session_info = TRUE,
                                                      overwrite = FALSE) {
  if (missing(output_dir) || length(output_dir) != 1 || !is.character(output_dir)) {
    stop("`output_dir` must be a single directory path.", call. = FALSE)
  }

  if (!is.character(prefix) || length(prefix) != 1 || !nzchar(prefix)) {
    stop("`prefix` must be a non-empty character string.", call. = FALSE)
  }

  if (!is.logical(include_readme) || length(include_readme) != 1) {
    stop("`include_readme` must be TRUE or FALSE.", call. = FALSE)
  }

  if (!is.logical(include_session_info) || length(include_session_info) != 1) {
    stop("`include_session_info` must be TRUE or FALSE.", call. = FALSE)
  }

  if (!is.logical(overwrite) || length(overwrite) != 1) {
    stop("`overwrite` must be TRUE or FALSE.", call. = FALSE)
  }

  if (!dir.exists(output_dir)) {
    dir.create(output_dir, recursive = TRUE, showWarnings = FALSE)
  }

  if (!dir.exists(output_dir)) {
    stop("Could not create `output_dir`.", call. = FALSE)
  }

  prefix_safe <- gpbiometrics_bundle_safe_name(prefix)

  collected_tables <- list()

  if (!is.null(bundle)) {
    collected_tables <- c(
      collected_tables,
      gpbiometrics_bundle_collect_tables(bundle, parent_name = "bundle")
    )
  }

  if (!is.null(tables)) {
    if (is.data.frame(tables)) {
      tables <- list(table = tables)
    }

    if (!is.list(tables)) {
      stop("`tables` must be a data frame or a named list of data frames.", call. = FALSE)
    }

    collected_tables <- c(
      collected_tables,
      gpbiometrics_bundle_collect_tables(tables, parent_name = "tables")
    )
  }

  collected_text <- list()

  if (!is.null(bundle)) {
    collected_text <- c(
      collected_text,
      gpbiometrics_bundle_collect_text(bundle, parent_name = "bundle")
    )
  }

  if (!is.null(text)) {
    if (is.character(text) && is.null(names(text))) {
      text <- list(text = text)
    }

    if (!is.list(text)) {
      stop("`text` must be a character vector or a named list.", call. = FALSE)
    }

    collected_text <- c(
      collected_text,
      gpbiometrics_bundle_collect_text(text, parent_name = "text")
    )
  }

  collected_plots <- list()

  if (!is.null(bundle)) {
    collected_plots <- c(
      collected_plots,
      gpbiometrics_bundle_collect_plots(bundle, parent_name = "bundle")
    )
  }

  if (!is.null(plots)) {
    if (!is.list(plots)) {
      stop("`plots` must be a named list of plot objects.", call. = FALSE)
    }

    collected_plots <- c(
      collected_plots,
      gpbiometrics_bundle_collect_plots(plots, parent_name = "plots")
    )
  }

  manifest <- data.frame(
    item = character(),
    type = character(),
    path = character(),
    status = character(),
    note = character(),
    stringsAsFactors = FALSE
  )

  for (name in names(collected_tables)) {
    table_i <- collected_tables[[name]]
    file_path <- file.path(
      output_dir,
      paste0(prefix_safe, "_", gpbiometrics_bundle_safe_name(name), ".csv")
    )

    gpbiometrics_bundle_check_overwrite(file_path, overwrite)

    utils::write.csv(table_i, file_path, row.names = FALSE, na = "")

    manifest <- rbind(
      manifest,
      data.frame(
        item = name,
        type = "table_csv",
        path = normalizePath(file_path, winslash = "/", mustWork = FALSE),
        status = "written",
        note = "",
        stringsAsFactors = FALSE
      )
    )
  }

  for (name in names(collected_text)) {
    text_i <- collected_text[[name]]
    file_path <- file.path(
      output_dir,
      paste0(prefix_safe, "_", gpbiometrics_bundle_safe_name(name), ".txt")
    )

    gpbiometrics_bundle_check_overwrite(file_path, overwrite)

    writeLines(as.character(text_i), file_path, useBytes = TRUE)

    manifest <- rbind(
      manifest,
      data.frame(
        item = name,
        type = "text",
        path = normalizePath(file_path, winslash = "/", mustWork = FALSE),
        status = "written",
        note = "",
        stringsAsFactors = FALSE
      )
    )
  }

  if (length(collected_plots) > 0) {
    if (!requireNamespace("ggplot2", quietly = TRUE)) {
      for (name in names(collected_plots)) {
        manifest <- rbind(
          manifest,
          data.frame(
            item = name,
            type = "plot_png",
            path = "",
            status = "skipped",
            note = "Package ggplot2 is not installed.",
            stringsAsFactors = FALSE
          )
        )
      }
    } else {
      for (name in names(collected_plots)) {
        plot_i <- collected_plots[[name]]
        file_path <- file.path(
          output_dir,
          paste0(prefix_safe, "_", gpbiometrics_bundle_safe_name(name), ".png")
        )

        gpbiometrics_bundle_check_overwrite(file_path, overwrite)

        ggplot2::ggsave(
          filename = file_path,
          plot = plot_i,
          width = 8,
          height = 5,
          dpi = 300
        )

        manifest <- rbind(
          manifest,
          data.frame(
            item = name,
            type = "plot_png",
            path = normalizePath(file_path, winslash = "/", mustWork = FALSE),
            status = "written",
            note = "",
            stringsAsFactors = FALSE
          )
        )
      }
    }
  }

  if (isTRUE(include_readme)) {
    file_path <- file.path(output_dir, paste0(prefix_safe, "_README.txt"))
    gpbiometrics_bundle_check_overwrite(file_path, overwrite)

    readme_text <- c(
      "Gazepoint biometrics report bundle",
      "",
      paste0("Created: ", as.character(Sys.time())),
      "",
      "This bundle contains exported package outputs for inspection and reporting.",
      "Private real-data outputs should be stored outside the package repository.",
      "",
      "Included workflow scope may include:",
      "- Import/schema checks and active biometric-channel detection.",
      "- Missingness, dropout, sampling, time-reset, activity, and readiness checks.",
      "- GSR/EDA quality checks, decomposition, SCR peak/event-window summaries, and SCR sensitivity analyses.",
      "- HR/IBI quality checks, implausible-IBI filtering, HR-IBI consistency checks, and IBI-derived HRV summaries.",
      "- TTL/event alignment and multimodal timeline inspection.",
      "- AOI-linked biometric summaries and model-preparation tables.",
      "- Optional interoperability exports for tools such as RHRV or NeuroKit2.",
      "- Contract-standardised ggplot outputs with stored plot data, settings, and interpretation notes.",
      "",
      "Interpretation cautions:",
      "- GSR/EDA should be interpreted as electrodermal/arousal-related signal, not emotional valence.",
      "- SCR detections are signal-processing events and depend on preprocessing, threshold, and event-window settings.",
      "- Raw Gazepoint HRV should be treated as a vendor/validity field unless independently documented.",
      "- HRV features should be derived from genuine IBI/RR intervals after plausibility filtering and window-duration checks.",
      "- AOI-linked biometric summaries describe signals during AOI exposure and do not by themselves establish preference, trust, scrutiny, or cognitive evaluation.",
      "- Eye-tracking measures indicate visual attention, not direct cognition or evaluation.",
      "- Plots are intended for quality control, synchronization checks, and reporting support, not standalone inferential evidence."
    )

    writeLines(readme_text, file_path, useBytes = TRUE)

    manifest <- rbind(
      manifest,
      data.frame(
        item = "README",
        type = "text",
        path = normalizePath(file_path, winslash = "/", mustWork = FALSE),
        status = "written",
        note = "",
        stringsAsFactors = FALSE
      )
    )
  }

  if (isTRUE(include_session_info)) {
    file_path <- file.path(output_dir, paste0(prefix_safe, "_session_info.txt"))
    gpbiometrics_bundle_check_overwrite(file_path, overwrite)

    writeLines(utils::capture.output(utils::sessionInfo()), file_path, useBytes = TRUE)

    manifest <- rbind(
      manifest,
      data.frame(
        item = "session_info",
        type = "text",
        path = normalizePath(file_path, winslash = "/", mustWork = FALSE),
        status = "written",
        note = "",
        stringsAsFactors = FALSE
      )
    )
  }

  manifest_path <- file.path(output_dir, paste0(prefix_safe, "_manifest.csv"))
  gpbiometrics_bundle_check_overwrite(manifest_path, overwrite)

  utils::write.csv(manifest, manifest_path, row.names = FALSE, na = "")

  manifest <- rbind(
    manifest,
    data.frame(
      item = "manifest",
      type = "manifest_csv",
      path = normalizePath(manifest_path, winslash = "/", mustWork = FALSE),
      status = "written",
      note = "",
      stringsAsFactors = FALSE
    )
  )

  written_count <- sum(manifest$status == "written")
  skipped_count <- sum(manifest$status == "skipped")

  status <- if (written_count == 0) {
    "nothing_written"
  } else if (skipped_count > 0) {
    "bundle_exported_with_skips"
  } else {
    "bundle_exported"
  }

  structure(
    list(
      overview = data.frame(
        output_dir = normalizePath(output_dir, winslash = "/", mustWork = FALSE),
        written_files = written_count,
        skipped_items = skipped_count,
        status = status,
        stringsAsFactors = FALSE
      ),
      manifest = manifest,
      output_dir = normalizePath(output_dir, winslash = "/", mustWork = FALSE),
      settings = list(
        prefix = prefix,
        include_readme = include_readme,
        include_session_info = include_session_info,
        overwrite = overwrite
      )
    ),
    class = c("gazepoint_biometrics_report_bundle", "list")
  )
}

#' Run a final real-data readiness gate for Gazepoint biometrics data
#'
#' Provides conservative pass/warn/fail checks before using real Gazepoint
#' Biometrics exports for analysis or reporting. The function checks basic row
#' count, signal availability, missingness, time ordering, TTL availability, and
#' HRV/IBI caution status. It does not certify data quality or infer emotional
#' states.
#'
#' @param data A biometric data frame. If `NULL`, the function tries to extract a
#'   data frame from `workflow_result`.
#' @param workflow_result Optional workflow/list object containing biometric data.
#' @param min_rows Minimum number of rows expected for a usable real-data check.
#' @param min_active_signal_count Minimum number of biometric signal columns with
#'   at least one non-missing/non-zero value.
#' @param max_missing_prop Maximum acceptable missing proportion for detected
#'   signal columns before a warning is raised.
#' @param required_signal_cols Optional signal columns that must be present.
#' @param require_gsr_us_preferred Logical. If `TRUE`, warns when `GSR_US` is
#'   absent but `GSR` is present.
#' @param require_ibi_for_hrv Logical. If `TRUE`, fails when HRV is present but
#'   IBI is absent. If `FALSE`, this condition is reported as a warning.
#' @param time_col Optional time column. If `NULL`, common time columns are
#'   detected automatically.
#' @param ttl_cols Optional TTL marker columns. If `NULL`, `ttl_marker` or
#'   `TTL0`-`TTL6` are detected automatically.
#'
#' @return A list with `overview`, `checks`, `signal_summary`, and `settings`.
#' @export
run_gazepoint_biometrics_real_data_readiness <- function(data = NULL,
                                                         workflow_result = NULL,
                                                         min_rows = 100,
                                                         min_active_signal_count = 1,
                                                         max_missing_prop = 0.50,
                                                         required_signal_cols = NULL,
                                                         require_gsr_us_preferred = TRUE,
                                                         require_ibi_for_hrv = FALSE,
                                                         time_col = NULL,
                                                         ttl_cols = NULL) {
  if (is.null(data)) {
    data <- gpbiometrics_readiness_extract_data(workflow_result)
  }

  if (!is.data.frame(data)) {
    stop(
      "`data` must be a data frame, or `workflow_result` must contain a data frame.",
      call. = FALSE
    )
  }

  if (!is.numeric(min_rows) || length(min_rows) != 1 || min_rows < 1) {
    stop("`min_rows` must be a positive number.", call. = FALSE)
  }

  if (!is.numeric(min_active_signal_count) ||
      length(min_active_signal_count) != 1 ||
      min_active_signal_count < 0) {
    stop("`min_active_signal_count` must be a non-negative number.", call. = FALSE)
  }

  if (!is.numeric(max_missing_prop) ||
      length(max_missing_prop) != 1 ||
      max_missing_prop < 0 ||
      max_missing_prop > 1) {
    stop("`max_missing_prop` must be between 0 and 1.", call. = FALSE)
  }

  if (!is.logical(require_gsr_us_preferred) || length(require_gsr_us_preferred) != 1) {
    stop("`require_gsr_us_preferred` must be TRUE or FALSE.", call. = FALSE)
  }

  if (!is.logical(require_ibi_for_hrv) || length(require_ibi_for_hrv) != 1) {
    stop("`require_ibi_for_hrv` must be TRUE or FALSE.", call. = FALSE)
  }

  dat <- as.data.frame(data, stringsAsFactors = FALSE)
  names_dat <- names(dat)

  signal_cols <- gpbiometrics_readiness_signal_cols(names_dat)
  signal_summary <- gpbiometrics_readiness_signal_summary(dat, signal_cols)

  active_signal_count <- sum(signal_summary$active_signal, na.rm = TRUE)

  if (is.null(time_col)) {
    time_col <- gpbiometrics_readiness_first_existing(
      names_dat,
      c(
        "time_ms", "timestamp_ms", "timestamp",
        "TIME", "Time", "time",
        "CNT", "cnt"
      )
    )
  }

  if (is.null(ttl_cols)) {
    ttl_marker <- gpbiometrics_readiness_first_existing(names_dat, c("ttl_marker"))

    if (!is.null(ttl_marker)) {
      ttl_cols <- ttl_marker
    } else {
      ttl_cols <- grep("^TTL[0-6]$", names_dat, value = TRUE, ignore.case = TRUE)
    }
  }

  checks <- data.frame(
    check = character(),
    status = character(),
    detail = character(),
    stringsAsFactors = FALSE
  )

  checks <- gpbiometrics_readiness_add_check(
    checks,
    check = "row_count",
    status = if (nrow(dat) >= min_rows) "pass" else "fail",
    detail = paste0("Input contains ", nrow(dat), " rows; minimum requested is ", min_rows, ".")
  )

  checks <- gpbiometrics_readiness_add_check(
    checks,
    check = "active_signal_count",
    status = if (active_signal_count >= min_active_signal_count) "pass" else "fail",
    detail = paste0(
      "Detected ",
      active_signal_count,
      " active biometric signal column(s); minimum requested is ",
      min_active_signal_count,
      "."
    )
  )

  if (!is.null(required_signal_cols)) {
    missing_required <- setdiff(required_signal_cols, names_dat)

    checks <- gpbiometrics_readiness_add_check(
      checks,
      check = "required_signal_columns",
      status = if (length(missing_required) == 0) "pass" else "fail",
      detail = if (length(missing_required) == 0) {
        "All requested signal columns are present."
      } else {
        paste("Missing requested signal columns:", paste(missing_required, collapse = ", "))
      }
    )
  }

  high_missing <- signal_summary$signal[
    signal_summary$active_signal &
      is.finite(signal_summary$missing_prop) &
      signal_summary$missing_prop > max_missing_prop
  ]

  checks <- gpbiometrics_readiness_add_check(
    checks,
    check = "signal_missingness",
    status = if (length(high_missing) == 0) "pass" else "warn",
    detail = if (length(high_missing) == 0) {
      paste0("No active signal exceeded missingness threshold ", max_missing_prop, ".")
    } else {
      paste(
        "Active signals exceeding missingness threshold:",
        paste(high_missing, collapse = ", ")
      )
    }
  )

  has_gsr_us <- "GSR_US" %in% names_dat || "gsr_us" %in% names_dat
  has_gsr <- "GSR" %in% names_dat || "gsr" %in% names_dat

  checks <- gpbiometrics_readiness_add_check(
    checks,
    check = "gsr_conductance_channel",
    status = if (has_gsr_us || !has_gsr || !isTRUE(require_gsr_us_preferred)) "pass" else "warn",
    detail = if (has_gsr_us) {
      "GSR_US conductance channel is available."
    } else if (has_gsr) {
      "GSR is present but GSR_US is absent; conversion should be conservative and documented."
    } else {
      "No GSR/GSR_US channel detected."
    }
  )

  has_hrv <- "HRV" %in% names_dat || "hrv" %in% names_dat
  has_ibi <- "IBI" %in% names_dat || "ibi" %in% names_dat

  hrv_status <- "pass"
  hrv_detail <- "No HRV/IBI conflict detected."

  if (has_hrv && !has_ibi) {
    hrv_status <- if (isTRUE(require_ibi_for_hrv)) "fail" else "warn"
    hrv_detail <- paste(
      "HRV column is present but IBI is absent.",
      "Treat raw Gazepoint HRV as a vendor/validity field unless independently documented."
    )
  } else if (has_hrv && has_ibi) {
    hrv_detail <- paste(
      "HRV and IBI are present.",
      "Derived HRV features should use IBI/RR intervals, not the raw HRV field."
    )
  } else if (!has_hrv && has_ibi) {
    hrv_detail <- "IBI is present; HRV features can be derived from genuine IBI/RR intervals after quality checks."
  }

  checks <- gpbiometrics_readiness_add_check(
    checks,
    check = "hrv_ibi_caution",
    status = hrv_status,
    detail = hrv_detail
  )

  time_status <- "warn"
  time_detail <- "No usable time column detected."

  if (!is.null(time_col) && time_col %in% names_dat) {
    time_numeric <- suppressWarnings(as.numeric(dat[[time_col]]))

    if (all(is.na(time_numeric))) {
      time_status <- "fail"
      time_detail <- paste0("Detected time column `", time_col, "` is not numeric/coercible.")
    } else {
      time_group_cols <- intersect(
        c(
          "source_file",
          "source_participant",
          "USER",
          "USER_FILE",
          "participant",
          "subject",
          "MEDIA_ID",
          "MEDIA_NAME",
          "trial",
          "trial_id"
        ),
        names_dat
      )

      if (length(time_group_cols) > 0) {
        time_group_id <- apply(
          dat[time_group_cols],
          1,
          function(z) paste(ifelse(is.na(z), "<NA>", as.character(z)), collapse = "||")
        )
      } else {
        time_group_id <- rep("all", nrow(dat))
      }

      split_rows <- split(seq_len(nrow(dat)), time_group_id, drop = TRUE)

      negative_steps_by_group <- vapply(
        split_rows,
        function(idx) {
          x <- time_numeric[idx]
          x <- x[is.finite(x)]
          if (length(x) < 2) {
            return(0L)
          }
          sum(diff(x) < 0, na.rm = TRUE)
        },
        integer(1)
      )

      negative_steps <- sum(negative_steps_by_group)
      affected_groups <- sum(negative_steps_by_group > 0)

      if (negative_steps == 0) {
        time_status <- "pass"
        time_detail <- paste0(
          "Time column `",
          time_col,
          "` is usable and non-decreasing within ",
          length(split_rows),
          " detected group(s)."
        )
      } else {
        time_status <- "warn"
        time_detail <- paste0(
          "Time column `",
          time_col,
          "` has ",
          negative_steps,
          " negative step(s) within ",
          affected_groups,
          " detected group(s)."
        )
      }
    }
  }

  checks <- gpbiometrics_readiness_add_check(
    checks,
    check = "time_column",
    status = time_status,
    detail = time_detail
  )

  ttl_status <- "info"
  ttl_detail <- "No TTL marker columns detected."

  if (length(ttl_cols) > 0) {
    ttl_cols <- ttl_cols[ttl_cols %in% names_dat]

    if (length(ttl_cols) > 0) {
      ttl_active <- vapply(
        ttl_cols,
        function(col) {
          x <- dat[[col]]
          if (is.logical(x)) {
            return(any(!is.na(x) & x))
          }
          if (is.numeric(x)) {
            return(any(!is.na(x) & x != 0))
          }
          x_chr <- trimws(as.character(x))
          any(!is.na(x_chr) & nzchar(x_chr) & !toupper(x_chr) %in% c("0", "FALSE", "F", "NA", "NAN", "NULL"))
        },
        logical(1)
      )

      ttl_status <- if (any(ttl_active)) "pass" else "warn"
      ttl_detail <- if (any(ttl_active)) {
        paste("Active TTL marker detected in:", paste(ttl_cols[ttl_active], collapse = ", "))
      } else {
        paste("TTL marker columns detected but no active TTL marker found:", paste(ttl_cols, collapse = ", "))
      }
    }
  }

  checks <- gpbiometrics_readiness_add_check(
    checks,
    check = "ttl_markers",
    status = ttl_status,
    detail = ttl_detail
  )

  fail_count <- sum(checks$status == "fail")
  warn_count <- sum(checks$status == "warn")
  pass_count <- sum(checks$status == "pass")
  info_count <- sum(checks$status == "info")

  final_status <- if (fail_count > 0) {
    "fail"
  } else if (warn_count > 0) {
    "warn"
  } else {
    "pass"
  }

  decision <- switch(
    final_status,
    pass = "ready_for_analysis_with_standard_cautions",
    warn = "review_before_analysis",
    fail = "not_ready_for_analysis"
  )

  structure(
    list(
      overview = data.frame(
        input_rows = nrow(dat),
        detected_signal_count = length(signal_cols),
        active_signal_count = active_signal_count,
        pass_checks = pass_count,
        warn_checks = warn_count,
        fail_checks = fail_count,
        info_checks = info_count,
        final_status = final_status,
        decision = decision,
        stringsAsFactors = FALSE
      ),
      checks = checks,
      signal_summary = signal_summary,
      settings = list(
        min_rows = min_rows,
        min_active_signal_count = min_active_signal_count,
        max_missing_prop = max_missing_prop,
        required_signal_cols = required_signal_cols,
        require_gsr_us_preferred = require_gsr_us_preferred,
        require_ibi_for_hrv = require_ibi_for_hrv,
        time_col = time_col,
        ttl_cols = ttl_cols
      )
    ),
    class = c("gazepoint_biometrics_real_data_readiness", "list")
  )
}

gpbiometrics_bundle_safe_name <- function(x) {
  x <- as.character(x)
  x <- gsub("[^A-Za-z0-9_-]+", "_", x)
  x <- gsub("_+", "_", x)
  x <- gsub("^_|_$", "", x)

  if (!nzchar(x)) {
    x <- "item"
  }

  x
}

gpbiometrics_bundle_check_overwrite <- function(path, overwrite) {
  if (file.exists(path) && !isTRUE(overwrite)) {
    stop(
      "File already exists and `overwrite = FALSE`: ",
      path,
      call. = FALSE
    )
  }

  invisible(TRUE)
}

gpbiometrics_bundle_collect_tables <- function(x, parent_name = "item") {
  out <- list()

  if (is.data.frame(x)) {
    out[[parent_name]] <- x
    return(out)
  }

  if (!is.list(x)) {
    return(out)
  }

  nms <- names(x)

  if (is.null(nms)) {
    nms <- paste0("item_", seq_along(x))
  }

  for (i in seq_along(x)) {
    child_name <- paste(parent_name, nms[i], sep = "_")
    child <- x[[i]]

    if (is.data.frame(child)) {
      out[[child_name]] <- child
    } else if (is.list(child) && !inherits(child, "ggplot")) {
      out <- c(out, gpbiometrics_bundle_collect_tables(child, child_name))
    }
  }

  out
}

gpbiometrics_bundle_collect_text <- function(x, parent_name = "item") {
  out <- list()

  if (is.character(x) && length(x) > 0) {
    out[[parent_name]] <- x
    return(out)
  }

  if (!is.list(x)) {
    return(out)
  }

  nms <- names(x)

  if (is.null(nms)) {
    nms <- paste0("item_", seq_along(x))
  }

  for (i in seq_along(x)) {
    child_name <- paste(parent_name, nms[i], sep = "_")
    child <- x[[i]]

    if (is.character(child) && length(child) > 0) {
      out[[child_name]] <- child
    } else if (is.list(child) && !inherits(child, "ggplot")) {
      out <- c(out, gpbiometrics_bundle_collect_text(child, child_name))
    }
  }

  out
}

gpbiometrics_bundle_collect_plots <- function(x, parent_name = "item") {
  out <- list()

  if (inherits(x, "ggplot")) {
    out[[parent_name]] <- x
    return(out)
  }

  if (!is.list(x)) {
    return(out)
  }

  nms <- names(x)

  if (is.null(nms)) {
    nms <- paste0("item_", seq_along(x))
  }

  for (i in seq_along(x)) {
    child_name <- paste(parent_name, nms[i], sep = "_")
    child <- x[[i]]

    if (inherits(child, "ggplot")) {
      out[[child_name]] <- child
    } else if (is.list(child)) {
      out <- c(out, gpbiometrics_bundle_collect_plots(child, child_name))
    }
  }

  out
}

gpbiometrics_readiness_extract_data <- function(workflow_result) {
  if (is.null(workflow_result)) {
    return(NULL)
  }

  if (is.data.frame(workflow_result)) {
    return(workflow_result)
  }

  if (!is.list(workflow_result)) {
    return(NULL)
  }

  candidate_names <- c(
    "biometrics",
    "biometric_data",
    "data",
    "raw_data",
    "imported_data",
    "all_gaze",
    "samples"
  )

  for (candidate in candidate_names) {
    if (!is.null(workflow_result[[candidate]]) &&
        is.data.frame(workflow_result[[candidate]])) {
      return(workflow_result[[candidate]])
    }
  }

  for (item in workflow_result) {
    if (is.data.frame(item)) {
      return(item)
    }
  }

  NULL
}

gpbiometrics_readiness_first_existing <- function(names_dat, candidates) {
  exact <- candidates[candidates %in% names_dat]

  if (length(exact) > 0) {
    return(exact[1])
  }

  lower_names <- tolower(names_dat)
  lower_candidates <- tolower(candidates)
  idx <- match(lower_candidates, lower_names)
  idx <- idx[!is.na(idx)]

  if (length(idx) > 0) {
    return(names_dat[idx[1]])
  }

  NULL
}

gpbiometrics_readiness_signal_cols <- function(names_dat) {
  candidates <- c(
    "DIAL", "DIALV",
    "GSR", "GSR_US", "GSR_US_TONIC", "GSR_US_PHASIC", "GSRV",
    "HR", "HRV", "HRP", "IBI",
    "dial", "dialv",
    "gsr", "gsr_us", "gsr_us_tonic", "gsr_us_phasic", "gsrv",
    "hr", "hrv", "hrp", "ibi"
  )

  unique(candidates[candidates %in% names_dat])
}

gpbiometrics_readiness_signal_summary <- function(dat, signal_cols) {
  out <- data.frame(
    signal = signal_cols,
    present = TRUE,
    numeric_or_coercible = NA,
    missing_count = NA_integer_,
    missing_prop = NA_real_,
    nonzero_count = NA_integer_,
    active_signal = NA,
    stringsAsFactors = FALSE
  )

  if (length(signal_cols) == 0) {
    return(out)
  }

  for (i in seq_along(signal_cols)) {
    signal <- signal_cols[i]
    x <- suppressWarnings(as.numeric(dat[[signal]]))

    numeric_or_coercible <- !all(is.na(x))
    missing_count <- sum(is.na(x))
    missing_prop <- if (length(x) == 0) NA_real_ else mean(is.na(x))
    nonzero_count <- sum(!is.na(x) & x != 0)
    active_signal <- numeric_or_coercible && any(!is.na(x) & x != 0)

    out$numeric_or_coercible[i] <- numeric_or_coercible
    out$missing_count[i] <- missing_count
    out$missing_prop[i] <- missing_prop
    out$nonzero_count[i] <- nonzero_count
    out$active_signal[i] <- active_signal
  }

  out
}

gpbiometrics_readiness_add_check <- function(checks, check, status, detail) {
  rbind(
    checks,
    data.frame(
      check = check,
      status = status,
      detail = detail,
      stringsAsFactors = FALSE
    )
  )
}

test_that("export_gazepoint_biometrics_report_bundle writes tables and manifest", {
  out_dir <- tempfile("gpbiometrics_bundle_")

  tables <- list(
    overview = data.frame(status = "ok", rows = 10),
    quality = data.frame(signal = c("GSR_US", "HR"), missing_prop = c(0, 0.1))
  )

  res <- export_gazepoint_biometrics_report_bundle(
    output_dir = out_dir,
    prefix = "test_bundle",
    tables = tables,
    text = list(methods = c("Methods text.", "Caution text.")),
    include_readme = TRUE,
    include_session_info = FALSE
  )

  expect_s3_class(res, "gazepoint_biometrics_report_bundle")
  expect_equal(res$overview$status, "bundle_exported")
  expect_true(file.exists(file.path(out_dir, "test_bundle_tables_overview.csv")))
  expect_true(file.exists(file.path(out_dir, "test_bundle_tables_quality.csv")))
  expect_true(file.exists(file.path(out_dir, "test_bundle_text_methods.txt")))
  expect_true(file.exists(file.path(out_dir, "test_bundle_README.txt")))
  expect_true(file.exists(file.path(out_dir, "test_bundle_manifest.csv")))
})

test_that("export_gazepoint_biometrics_report_bundle prevents accidental overwrite", {
  out_dir <- tempfile("gpbiometrics_bundle_")
  dir.create(out_dir)

  existing <- file.path(out_dir, "test_bundle_tables_overview.csv")
  writeLines("existing", existing)

  expect_error(
    export_gazepoint_biometrics_report_bundle(
      output_dir = out_dir,
      prefix = "test_bundle",
      tables = list(overview = data.frame(x = 1)),
      include_readme = FALSE,
      include_session_info = FALSE,
      overwrite = FALSE
    ),
    "File already exists"
  )
})

test_that("run_gazepoint_biometrics_real_data_readiness passes adequate synthetic data", {
  dat <- data.frame(
    time_ms = seq(0, 14900, by = 100),
    GSR_US = seq(1, 2, length.out = 150),
    HR = seq(70, 80, length.out = 150),
    IBI = seq(850, 760, length.out = 150),
    TTL0 = c(rep(0, 50), 1, rep(0, 99)),
    TTLV = 1
  )

  res <- run_gazepoint_biometrics_real_data_readiness(
    dat,
    min_rows = 100,
    min_active_signal_count = 2,
    max_missing_prop = 0.25
  )

  expect_s3_class(res, "gazepoint_biometrics_real_data_readiness")
  expect_equal(res$overview$final_status, "pass")
  expect_equal(res$overview$decision, "ready_for_analysis_with_standard_cautions")
  expect_true(any(res$checks$check == "ttl_markers" & res$checks$status == "pass"))
})

test_that("run_gazepoint_biometrics_real_data_readiness warns for GSR-only and HRV-without-IBI", {
  dat <- data.frame(
    time_ms = seq(0, 11900, by = 100),
    GSR = seq(1000000, 500000, length.out = 120),
    HRV = rep(1, 120)
  )

  res <- run_gazepoint_biometrics_real_data_readiness(
    dat,
    min_rows = 100,
    min_active_signal_count = 1
  )

  expect_equal(res$overview$final_status, "warn")
  expect_true(any(res$checks$check == "gsr_conductance_channel" & res$checks$status == "warn"))
  expect_true(any(res$checks$check == "hrv_ibi_caution" & res$checks$status == "warn"))
})

test_that("run_gazepoint_biometrics_real_data_readiness fails for too few rows", {
  dat <- data.frame(
    time_ms = seq(0, 900, by = 100),
    GSR_US = seq(1, 2, length.out = 10)
  )

  res <- run_gazepoint_biometrics_real_data_readiness(
    dat,
    min_rows = 100,
    min_active_signal_count = 1
  )

  expect_equal(res$overview$final_status, "fail")
  expect_true(any(res$checks$check == "row_count" & res$checks$status == "fail"))
})

test_that("run_gazepoint_biometrics_real_data_readiness can extract data from workflow-like list", {
  dat <- data.frame(
    time_ms = seq(0, 9900, by = 100),
    GSR_US = seq(1, 2, length.out = 100),
    HR = seq(70, 75, length.out = 100)
  )

  workflow <- list(
    overview = data.frame(status = "ok"),
    biometrics = dat
  )

  res <- run_gazepoint_biometrics_real_data_readiness(
    workflow_result = workflow,
    min_rows = 100,
    min_active_signal_count = 2
  )

  expect_equal(res$overview$final_status, "pass")
})

test_that("run_gazepoint_biometrics_real_data_readiness checks time order within source groups", {
  dat <- data.frame(
    source_file = rep(c("file_a.csv", "file_b.csv"), each = 100),
    CNT = rep(seq_len(100), 2),
    GSR_US = seq(1, 2, length.out = 200),
    HR = seq(70, 80, length.out = 200),
    TTL0 = c(1, rep(0, 199))
  )

  res <- run_gazepoint_biometrics_real_data_readiness(
    dat,
    min_rows = 100,
    min_active_signal_count = 2,
    time_col = "CNT"
  )

  expect_equal(res$overview$final_status, "pass")
  expect_true(any(res$checks$check == "time_column" & res$checks$status == "pass"))
})

test_that("run_gazepoint_biometrics_real_data_readiness still warns for negative time steps within groups", {
  dat <- data.frame(
    source_file = rep("file_a.csv", 120),
    CNT = c(seq_len(60), seq_len(60)),
    GSR_US = seq(1, 2, length.out = 120),
    HR = seq(70, 80, length.out = 120),
    TTL0 = c(1, rep(0, 119))
  )

  res <- run_gazepoint_biometrics_real_data_readiness(
    dat,
    min_rows = 100,
    min_active_signal_count = 2,
    time_col = "CNT"
  )

  expect_equal(res$overview$final_status, "warn")
  expect_true(any(res$checks$check == "time_column" & res$checks$status == "warn"))
})

#' Run a multiverse of SCR scoring specifications
#'
#' Scores SCR amplitudes across multiple latency windows, thresholds, baseline
#' methods, and response metrics. Optionally applies a user-supplied model
#' function to each specification. This supports transparent sensitivity
#' analysis and specification-curve style reporting.
#'
#' @param dat Sample-level EDA data.
#' @param signal_col Conductance/EDA signal column.
#' @param time_col Time column, preferably relative to stimulus onset. If
#'   `event_time_col` is supplied, relative time is computed as
#'   `time_col - event_time_col`.
#' @param trial_cols Columns identifying trials.
#' @param condition_col Optional experimental condition column.
#' @param participant_col Optional participant column.
#' @param event_time_col Optional event/stimulus onset time column.
#' @param latency_windows List of response windows in seconds.
#' @param thresholds SCR response thresholds.
#' @param baseline_methods Baseline methods: `"median"`, `"mean"`, or `"none"`.
#' @param baseline_window Baseline window in relative seconds.
#' @param response_metrics Response metrics: `"max_minus_baseline"` or
#'   `"peak_to_peak"`.
#' @param model_function Optional function applied to each specification-level
#'   trial summary.
#'
#' @return A list with specification grid, scored trials, optional model results,
#'   and robustness overview.
#' @export
run_gazepoint_scr_multiverse <- function(dat,
                                         signal_col = "GSR_US",
                                         time_col = "time",
                                         trial_cols = NULL,
                                         condition_col = NULL,
                                         participant_col = NULL,
                                         event_time_col = NULL,
                                         latency_windows = list(c(1, 3), c(1, 4), c(1, 5)),
                                         thresholds = c(0.01, 0.05),
                                         baseline_methods = c("median", "mean"),
                                         baseline_window = c(-1, 0),
                                         response_metrics = c("max_minus_baseline"),
                                         model_function = NULL) {
  if (!is.data.frame(dat)) {
    stop("`dat` must be a data frame.", call. = FALSE)
  }

  required <- c(signal_col, time_col, trial_cols, condition_col, participant_col, event_time_col)
  required <- required[!is.null(required)]

  missing_required <- setdiff(required, names(dat))
  if (length(missing_required) > 0) {
    stop("Missing required columns: ", paste(missing_required, collapse = ", "), call. = FALSE)
  }

  if (!is.numeric(dat[[signal_col]])) {
    stop("`signal_col` must identify a numeric column.", call. = FALSE)
  }

  if (!is.numeric(dat[[time_col]])) {
    stop("`time_col` must identify a numeric column.", call. = FALSE)
  }

  baseline_methods <- match.arg(baseline_methods, choices = c("median", "mean", "none"), several.ok = TRUE)
  response_metrics <- match.arg(response_metrics, choices = c("max_minus_baseline", "peak_to_peak"), several.ok = TRUE)

  if (is.null(trial_cols)) {
    dat$.gpbiometrics_trial_id <- "trial_1"
    trial_cols <- ".gpbiometrics_trial_id"
  }

  specification_grid <- expand.grid(
    window_id = seq_along(latency_windows),
    threshold = thresholds,
    baseline_method = baseline_methods,
    response_metric = response_metrics,
    stringsAsFactors = FALSE
  )

  specification_grid$latency_lower <- vapply(latency_windows[specification_grid$window_id], `[`, numeric(1), 1)
  specification_grid$latency_upper <- vapply(latency_windows[specification_grid$window_id], `[`, numeric(1), 2)
  specification_grid$specification_id <- paste0("spec_", seq_len(nrow(specification_grid)))

  trial_groups <- gpbiometrics_scr_multiverse_split(dat, trial_cols)

  scored_rows <- list()
  row_id <- 1L

  for (spec_i in seq_len(nrow(specification_grid))) {
    spec <- specification_grid[spec_i, , drop = FALSE]

    for (trial_id in names(trial_groups)) {
      idx <- trial_groups[[trial_id]]

      rel_time <- dat[[time_col]][idx]

      if (!is.null(event_time_col)) {
        event_time <- dat[[event_time_col]][idx][1]
        rel_time <- rel_time - event_time
      }

      signal <- dat[[signal_col]][idx]

      baseline_idx <- is.finite(rel_time) &
        rel_time >= baseline_window[1] &
        rel_time <= baseline_window[2] &
        is.finite(signal)

      response_idx <- is.finite(rel_time) &
        rel_time >= spec$latency_lower &
        rel_time <= spec$latency_upper &
        is.finite(signal)

      baseline_value <- gpbiometrics_scr_multiverse_baseline(
        signal[baseline_idx],
        method = spec$baseline_method
      )

      if (sum(response_idx) == 0) {
        response_amplitude <- NA_real_
        peak_time <- NA_real_
        status <- "no_response_window_samples"
      } else {
        response_signal <- signal[response_idx]
        response_time <- rel_time[response_idx]

        if (spec$response_metric == "max_minus_baseline") {
          peak_i <- which.max(response_signal)
          response_amplitude <- max(response_signal, na.rm = TRUE) - baseline_value
          peak_time <- response_time[peak_i]
        } else {
          peak_i <- which.max(response_signal)
          response_amplitude <- max(response_signal, na.rm = TRUE) - min(response_signal, na.rm = TRUE)
          peak_time <- response_time[peak_i]
        }

        status <- "scr_scored"
      }

      base_values <- gpbiometrics_scr_multiverse_values(dat, idx, c(trial_cols, participant_col, condition_col))

      scored_rows[[row_id]] <- data.frame(
        specification_id = spec$specification_id,
        window_id = spec$window_id,
        latency_lower = spec$latency_lower,
        latency_upper = spec$latency_upper,
        threshold = spec$threshold,
        baseline_method = spec$baseline_method,
        response_metric = spec$response_metric,
        trial_id = trial_id,
        base_values,
        baseline_value = baseline_value,
        response_amplitude = response_amplitude,
        response_present = is.finite(response_amplitude) && response_amplitude >= spec$threshold,
        peak_time = peak_time,
        status = status,
        stringsAsFactors = FALSE,
        check.names = FALSE
      )

      row_id <- row_id + 1L
    }
  }

  scored_trials <- do.call(rbind, scored_rows)
  rownames(scored_trials) <- NULL

  specification_summary <- stats::aggregate(
    cbind(response_amplitude, response_present) ~ specification_id + latency_lower +
      latency_upper + threshold + baseline_method + response_metric,
    data = scored_trials,
    FUN = function(x) mean(x, na.rm = TRUE)
  )

  names(specification_summary)[names(specification_summary) == "response_amplitude"] <- "mean_response_amplitude"
  names(specification_summary)[names(specification_summary) == "response_present"] <- "response_rate"

  model_results <- NULL

  if (!is.null(model_function)) {
    model_results <- lapply(split(scored_trials, scored_trials$specification_id), function(x) {
      tryCatch(
        model_function(x),
        error = function(e) list(error = conditionMessage(e))
      )
    })
  } else if (!is.null(condition_col) && condition_col %in% names(scored_trials)) {
    model_results <- gpbiometrics_scr_multiverse_default_models(scored_trials, condition_col)
  }

  overview <- data.frame(
    specification_count = nrow(specification_grid),
    trial_count = length(trial_groups),
    scored_rows = nrow(scored_trials),
    successful_rows = sum(scored_trials$status == "scr_scored"),
    problem_rows = sum(scored_trials$status != "scr_scored"),
    status = if (all(scored_trials$status == "scr_scored")) {
      "scr_multiverse_complete"
    } else if (any(scored_trials$status == "scr_scored")) {
      "scr_multiverse_partial"
    } else {
      "scr_multiverse_failed"
    },
    interpretation = paste(
      "The SCR multiverse scores responses across defensible parameter specifications.",
      "It supports transparent sensitivity reporting and does not identify a universally best scoring rule."
    ),
    stringsAsFactors = FALSE
  )

  structure(
    list(
      overview = overview,
      specification_grid = specification_grid,
      scored_trials = scored_trials,
      specification_summary = specification_summary,
      model_results = model_results,
      settings = list(
        signal_col = signal_col,
        time_col = time_col,
        trial_cols = trial_cols,
        condition_col = condition_col,
        participant_col = participant_col,
        event_time_col = event_time_col,
        latency_windows = latency_windows,
        thresholds = thresholds,
        baseline_methods = baseline_methods,
        baseline_window = baseline_window,
        response_metrics = response_metrics
      )
    ),
    class = c("gazepoint_scr_multiverse", "list")
  )
}

gpbiometrics_scr_multiverse_split <- function(dat, trial_cols) {
  gf <- dat[trial_cols]
  gf[] <- lapply(gf, function(x) {
    x <- as.character(x)
    x[is.na(x)] <- "<NA>"
    x
  })

  split(seq_len(nrow(dat)), do.call(paste, c(gf, sep = " | ")))
}

gpbiometrics_scr_multiverse_values <- function(dat, idx, cols) {
  cols <- cols[!is.null(cols)]
  cols <- unique(cols)
  cols <- cols[cols %in% names(dat)]

  if (length(cols) == 0) {
    return(data.frame(dummy = NA_character_)[0])
  }

  vals <- lapply(cols, function(nm) as.character(dat[[nm]][idx[1]]))
  names(vals) <- cols
  as.data.frame(vals, stringsAsFactors = FALSE, optional = TRUE)
}

gpbiometrics_scr_multiverse_baseline <- function(x, method) {
  if (method == "none") {
    return(0)
  }

  x <- x[is.finite(x)]

  if (length(x) == 0) {
    return(0)
  }

  if (method == "median") {
    stats::median(x)
  } else {
    mean(x)
  }
}

gpbiometrics_scr_multiverse_default_models <- function(scored_trials, condition_col) {
  lapply(split(scored_trials, scored_trials$specification_id), function(x) {
    out <- list()

    if (length(unique(x[[condition_col]])) < 2) {
      return(list(status = "condition_has_less_than_two_levels"))
    }

    x[[condition_col]] <- as.factor(x[[condition_col]])

    out$hurdle_part <- tryCatch({
      fit <- stats::glm(response_present ~ x[[condition_col]], data = x, family = stats::binomial())
      coef_table <- stats::coef(summary(fit))
      data.frame(
        term = rownames(coef_table),
        estimate = coef_table[, "Estimate"],
        p_value = coef_table[, "Pr(>|z|)"],
        row.names = NULL,
        stringsAsFactors = FALSE
      )
    }, error = function(e) {
      data.frame(error = conditionMessage(e), stringsAsFactors = FALSE)
    })

    positive <- x[is.finite(x$response_amplitude) & x$response_amplitude > 0, , drop = FALSE]

    out$amplitude_part <- tryCatch({
      if (nrow(positive) < 3 || length(unique(positive[[condition_col]])) < 2) {
        data.frame(status = "insufficient_positive_responses", stringsAsFactors = FALSE)
      } else {
        fit <- stats::lm(log1p(response_amplitude) ~ positive[[condition_col]], data = positive)
        coef_table <- stats::coef(summary(fit))
        data.frame(
          term = rownames(coef_table),
          estimate = coef_table[, "Estimate"],
          p_value = coef_table[, "Pr(>|t|)"],
          row.names = NULL,
          stringsAsFactors = FALSE
        )
      }
    }, error = function(e) {
      data.frame(error = conditionMessage(e), stringsAsFactors = FALSE)
    })

    out$status <- "default_two_part_models_completed"
    out
  })
}

#' Import a folder of Gazepoint Biometrics exports
#'
#' Reads rectangular Gazepoint CSV exports from a folder and combines files that
#' contain at least one known Gazepoint Biometrics column. This function is
#' designed for all-gaze and fixation-style exports. Multi-section Gazepoint
#' Data Summary files should be parsed separately.
#'
#' @param path Folder containing Gazepoint CSV exports.
#' @param pattern Regular expression used to identify candidate CSV files.
#' @param recursive Should subfolders be searched?
#' @param include_fixations Should files with `"fixation"` in the file name be
#'   included?
#' @param include_all_gaze Should files with `"all_gaze"` in the file name be
#'   included?
#' @param include_other_csv Should other CSV files be attempted? The default is
#'   `FALSE` to avoid accidentally trying to parse multi-section
#'   `Data_Summary_export` files as rectangular data.
#' @param na Values that should be treated as missing.
#'
#' @return A data frame with all imported rows combined. The output includes a
#'   `source_file` column and has class `"gazepoint_biometrics_folder"`.
#'
#' @export
import_gazepoint_biometric_folder <- function(path,
                                              pattern = "\\.csv$",
                                              recursive = FALSE,
                                              include_fixations = TRUE,
                                              include_all_gaze = TRUE,
                                              include_other_csv = FALSE,
                                              na = c("", "NA", "NaN")) {
  if (missing(path) || length(path) != 1L || !nzchar(path)) {
    stop("`path` must be a single non-empty folder path.", call. = FALSE)
  }

  if (!dir.exists(path)) {
    stop("Folder does not exist: ", path, call. = FALSE)
  }

  files <- list.files(
    path = path,
    pattern = pattern,
    full.names = TRUE,
    recursive = recursive,
    ignore.case = TRUE
  )

  if (length(files) == 0L) {
    stop("No CSV files were found in: ", path, call. = FALSE)
  }

  candidate_files <- keep_candidate_biometric_files(
    files = files,
    include_fixations = include_fixations,
    include_all_gaze = include_all_gaze,
    include_other_csv = include_other_csv
  )

  if (length(candidate_files) == 0L) {
    stop("No candidate Gazepoint biometric CSV files were found.", call. = FALSE)
  }

  imported <- lapply(candidate_files, function(file) {
    dat <- import_gazepoint_biometrics(file, na = na)

    cols <- check_gazepoint_biometric_columns(dat)

    if (!any(cols$present[cols$signal %in% c(
      "gsr_eda",
      "heart_rate",
      "engagement_dial",
      "ttl_marker"
    )])) {
      return(NULL)
    }

    source_file <- basename(file)

    dat$source_file <- source_file
    dat$source_type <- detect_gazepoint_source_type(source_file)
    dat$source_participant <- detect_gazepoint_source_participant(source_file)

    dat
  })

  imported <- imported[!vapply(imported, is.null, logical(1))]

  if (length(imported) == 0L) {
    stop(
      "CSV files were found, but none contained known Gazepoint Biometrics columns.",
      call. = FALSE
    )
  }

  out <- combine_gazepoint_tables(imported)

  attr(out, "source_files") <- basename(candidate_files)
  attr(out, "biometric_columns") <- check_gazepoint_biometric_columns(out)
  attr(out, "active_channels") <- detect_active_biometric_channels(out)

  class(out) <- c("gazepoint_biometrics_folder", "gazepoint_biometrics", class(out))

  out
}


keep_candidate_biometric_files <- function(files,
                                           include_fixations = TRUE,
                                           include_all_gaze = TRUE,
                                           include_other_csv = FALSE) {
  file_names <- basename(files)
  file_names_lower <- tolower(file_names)

  is_data_summary <- grepl("data_summary", file_names_lower)
  is_fixation <- grepl("fixation", file_names_lower)
  is_all_gaze <- grepl("all_gaze", file_names_lower)

  keep <- rep(FALSE, length(files))

  if (include_fixations) {
    keep <- keep | is_fixation
  }

  if (include_all_gaze) {
    keep <- keep | is_all_gaze
  }

  if (include_other_csv) {
    keep <- keep | (!is_data_summary)
  }

  files[keep & !is_data_summary]
}


combine_gazepoint_tables <- function(tables) {
  if (!is.list(tables) || length(tables) == 0L) {
    stop("`tables` must be a non-empty list of data frames.", call. = FALSE)
  }

  all_names <- unique(unlist(lapply(tables, names), use.names = FALSE))

  aligned <- lapply(tables, function(dat) {
    missing_names <- setdiff(all_names, names(dat))

    for (nm in missing_names) {
      dat[[nm]] <- NA
    }

    dat[all_names]
  })

  out <- do.call(rbind, aligned)
  rownames(out) <- NULL
  out
}

detect_gazepoint_source_type <- function(file_name) {
  file_name_lower <- tolower(file_name)

  if (grepl("all_gaze", file_name_lower)) {
    return("all_gaze")
  }

  if (grepl("fixation", file_name_lower)) {
    return("fixations")
  }

  if (grepl("data_summary", file_name_lower)) {
    return("data_summary")
  }

  "other"
}


detect_gazepoint_source_participant <- function(file_name) {
  x <- basename(file_name)

  x <- sub("\\.csv$", "", x, ignore.case = TRUE)
  x <- sub("_all_gaze$", "", x, ignore.case = TRUE)
  x <- sub("_fixations$", "", x, ignore.case = TRUE)
  x <- sub("_fixation$", "", x, ignore.case = TRUE)

  trimws(x)
}

test_that("import_gazepoint_biometric_folder imports multiple biometric files", {
  folder <- tempfile("gazepoint_folder_")
  dir.create(folder)

  file_one <- file.path(folder, "User 0_all_gaze.csv")
  file_two <- file.path(folder, "User 0_fixations.csv")
  file_summary <- file.path(folder, "Data_Summary_export_test.csv")

  writeLines(
    c(
      "TIME,GSR_US,GSRV,HR,HRV,DIAL,DIALV,",
      "0.01,2.0,1,75,1,0.1,1,",
      "0.02,2.1,1,76,1,0.2,1,"
    ),
    file_one,
    useBytes = TRUE
  )

  writeLines(
    c(
      "TIME,FPOGX,FPOGY,FPOGS,FPOGD,FPOGID,GSR_US,GSRV,HR,HRV,DIAL,DIALV,",
      "0.03,0.5,0.6,0.01,0.20,1,2.2,1,77,1,0.3,1,",
      "0.04,0.6,0.7,0.03,0.30,2,2.3,1,78,1,0.4,1,"
    ),
    file_two,
    useBytes = TRUE
  )

  writeLines(
    c(
      "Gazepoint Analysis,v7.2.0",
      "Processed on,example",
      "AOI Summary"
    ),
    file_summary,
    useBytes = TRUE
  )

  dat <- import_gazepoint_biometric_folder(folder)

  expect_s3_class(dat, "gazepoint_biometrics_folder")
  expect_equal(nrow(dat), 4)
  expect_true("source_file" %in% names(dat))
  expect_true("source_type" %in% names(dat))
  expect_true("source_participant" %in% names(dat))
  expect_true(all(c("GSR_US", "HR", "DIAL") %in% names(dat)))
  expect_false(any(grepl("Data_Summary", dat$source_file)))

  expect_true(all(c("all_gaze", "fixations") %in% unique(dat$source_type)))
  expect_true("User 0" %in% unique(dat$source_participant))

  active <- attr(dat, "active_channels")
  expect_true(is.data.frame(active))
  expect_true(active$active[active$signal == "gsr_eda"])
  expect_true(active$active[active$signal == "heart_rate"])
  expect_true(active$active[active$signal == "engagement_dial"])
})


test_that("import_gazepoint_biometric_folder rejects missing folders", {
  expect_error(
    import_gazepoint_biometric_folder("folder_that_does_not_exist"),
    "Folder does not exist"
  )
})


test_that("import_gazepoint_biometric_folder skips non-biometric CSV files", {
  folder <- tempfile("gazepoint_folder_")
  dir.create(folder)

  file_one <- file.path(folder, "User 0_all_gaze.csv")

  writeLines(
    c(
      "TIME,X,Y,",
      "0.01,1,2,",
      "0.02,3,4,"
    ),
    file_one,
    useBytes = TRUE
  )

  expect_error(
    import_gazepoint_biometric_folder(folder),
    "none contained known Gazepoint Biometrics columns"
  )
})

#' Audit Gazepoint biometric time resets
#'
#' Detects negative time steps, duplicate time steps, non-finite time values, and
#' recording segments within grouped Gazepoint biometric exports. This helper is
#' intended for quality control and synchronization inspection. It does not alter
#' raw values unless `return_reindexed_time = TRUE`, in which case an additional
#' segment-relative time column is added.
#'
#' @param data A data frame containing Gazepoint biometric rows.
#' @param time_col Optional time/counter column. If `NULL`, common Gazepoint time
#'   columns are detected automatically.
#' @param group_cols Optional grouping columns. If `NULL`, available
#'   source/participant/media/trial-like columns are used.
#' @param allow_ties Logical. If `TRUE`, repeated time values are not treated as
#'   non-monotonic.
#' @param split_on_negative_step Logical. If `TRUE`, negative time steps start a
#'   new segment within each group.
#' @param return_reindexed_time Logical. If `TRUE`, adds
#'   `time_reindexed_within_segment`, starting at zero within each detected
#'   segment.
#' @param min_segment_rows Minimum rows expected per segment before a segment is
#'   flagged as short.
#'
#' @return A list with `overview`, `segment_summary`, `row_flags`,
#'   `data_with_segments`, and `settings`.
#' @export
audit_gazepoint_time_resets <- function(data,
                                        time_col = NULL,
                                        group_cols = NULL,
                                        allow_ties = TRUE,
                                        split_on_negative_step = TRUE,
                                        return_reindexed_time = FALSE,
                                        min_segment_rows = 1) {
  if (!is.data.frame(data)) {
    stop("`data` must be a data frame.", call. = FALSE)
  }

  if (!is.logical(allow_ties) || length(allow_ties) != 1) {
    stop("`allow_ties` must be TRUE or FALSE.", call. = FALSE)
  }

  if (!is.logical(split_on_negative_step) || length(split_on_negative_step) != 1) {
    stop("`split_on_negative_step` must be TRUE or FALSE.", call. = FALSE)
  }

  if (!is.logical(return_reindexed_time) || length(return_reindexed_time) != 1) {
    stop("`return_reindexed_time` must be TRUE or FALSE.", call. = FALSE)
  }

  if (!is.numeric(min_segment_rows) || length(min_segment_rows) != 1 || min_segment_rows < 1) {
    stop("`min_segment_rows` must be a positive number.", call. = FALSE)
  }

  dat <- as.data.frame(data, stringsAsFactors = FALSE)
  dat$.gpbiometrics_row_id <- seq_len(nrow(dat))

  names_dat <- names(dat)

  if (is.null(time_col)) {
    time_col <- gpbiometrics_time_qc_first_existing(
      names_dat,
      c(
        "time_ms", "timestamp_ms", "timestamp",
        "TIME", "Time", "time",
        "CNT", "cnt"
      )
    )
  }

  if (is.null(time_col) || !time_col %in% names_dat) {
    stop("No usable time column was found. Supply `time_col`.", call. = FALSE)
  }

  group_cols <- gpbiometrics_time_qc_resolve_group_cols(names_dat, group_cols)

  missing_group_cols <- setdiff(group_cols, names_dat)

  if (length(missing_group_cols) > 0) {
    stop(
      "`group_cols` not found in `data`: ",
      paste(missing_group_cols, collapse = ", "),
      call. = FALSE
    )
  }

  time_value <- suppressWarnings(as.numeric(dat[[time_col]]))

  row_flags <- data.frame(
    .gpbiometrics_row_id = dat$.gpbiometrics_row_id,
    time_col = time_col,
    time_value = time_value,
    group_id = gpbiometrics_time_qc_group_id(dat, group_cols),
    group_row_index = NA_integer_,
    time_delta = NA_real_,
    flag_nonfinite_time = !is.finite(time_value),
    flag_negative_step = FALSE,
    flag_duplicate_time = FALSE,
    flag_nonmonotonic = FALSE,
    reset_segment_index = NA_integer_,
    flag_short_segment = FALSE,
    stringsAsFactors = FALSE
  )

  if (length(group_cols) > 0) {
    row_flags <- cbind(dat[group_cols], row_flags)
  }

  group_indices <- split(seq_len(nrow(dat)), row_flags$group_id, drop = TRUE)

  for (group_name in names(group_indices)) {
    idx <- group_indices[[group_name]]
    row_flags$group_row_index[idx] <- seq_along(idx)

    x <- time_value[idx]
    delta <- c(NA_real_, diff(x))

    negative_step <- is.finite(delta) & delta < 0
    duplicate_step <- is.finite(delta) & delta == 0

    nonmonotonic <- negative_step
    if (!isTRUE(allow_ties)) {
      nonmonotonic <- is.finite(delta) & delta <= 0
    }

    row_flags$time_delta[idx] <- delta
    row_flags$flag_negative_step[idx] <- negative_step
    row_flags$flag_duplicate_time[idx] <- duplicate_step
    row_flags$flag_nonmonotonic[idx] <- nonmonotonic

    if (isTRUE(split_on_negative_step)) {
      row_flags$reset_segment_index[idx] <- cumsum(negative_step) + 1L
    } else {
      row_flags$reset_segment_index[idx] <- 1L
    }
  }

  segment_summary <- gpbiometrics_time_qc_segment_summary(
    dat = dat,
    row_flags = row_flags,
    group_cols = group_cols,
    min_segment_rows = min_segment_rows
  )

  if (nrow(segment_summary) > 0) {
    short_segments <- segment_summary$segment_key[
      segment_summary$rows < min_segment_rows
    ]

    row_flags$flag_short_segment <- paste(
      row_flags$group_id,
      row_flags$reset_segment_index,
      sep = "||segment_"
    ) %in% short_segments
  }

  data_with_segments <- dat
  data_with_segments$time_qc_group_id <- row_flags$group_id
  data_with_segments$time_qc_group_row_index <- row_flags$group_row_index
  data_with_segments$time_qc_delta <- row_flags$time_delta
  data_with_segments$time_qc_negative_step <- row_flags$flag_negative_step
  data_with_segments$time_qc_duplicate_step <- row_flags$flag_duplicate_time
  data_with_segments$time_qc_nonmonotonic <- row_flags$flag_nonmonotonic
  data_with_segments$time_qc_segment_index <- row_flags$reset_segment_index

  if (isTRUE(return_reindexed_time)) {
    data_with_segments$time_reindexed_within_segment <- gpbiometrics_time_qc_reindex_time(
      time_value = time_value,
      group_id = row_flags$group_id,
      segment_index = row_flags$reset_segment_index
    )
  }

  n_negative_steps <- sum(row_flags$flag_negative_step, na.rm = TRUE)
  n_duplicate_steps <- sum(row_flags$flag_duplicate_time, na.rm = TRUE)
  n_nonmonotonic <- sum(row_flags$flag_nonmonotonic, na.rm = TRUE)
  n_nonfinite <- sum(row_flags$flag_nonfinite_time, na.rm = TRUE)
  affected_groups <- length(unique(row_flags$group_id[row_flags$flag_negative_step]))

  status <- if (all(is.na(time_value))) {
    "fail_no_numeric_time"
  } else if (n_negative_steps > 0 || n_nonmonotonic > 0 || n_nonfinite > 0) {
    "warn_time_irregularities_detected"
  } else {
    "pass"
  }

  overview <- data.frame(
    input_rows = nrow(dat),
    time_col = time_col,
    group_count = length(group_indices),
    segment_count = nrow(segment_summary),
    negative_steps = n_negative_steps,
    duplicate_steps = n_duplicate_steps,
    nonmonotonic_steps = n_nonmonotonic,
    nonfinite_time_rows = n_nonfinite,
    affected_groups = affected_groups,
    status = status,
    stringsAsFactors = FALSE
  )

  structure(
    list(
      overview = overview,
      segment_summary = segment_summary,
      row_flags = row_flags,
      data_with_segments = data_with_segments,
      settings = list(
        time_col = time_col,
        group_cols = group_cols,
        allow_ties = allow_ties,
        split_on_negative_step = split_on_negative_step,
        return_reindexed_time = return_reindexed_time,
        min_segment_rows = min_segment_rows
      )
    ),
    class = c("gazepoint_time_reset_audit", "list")
  )
}

#' Audit Gazepoint biometric signal activity
#'
#' Screens biometric signal columns for missingness, all-zero channels, constant
#' values, low variation, and active signal presence within groups. This helper
#' is designed to identify inactive files or channels before event-level EDA,
#' HR, IBI, or multimodal analysis.
#'
#' @param data A data frame containing Gazepoint biometric rows.
#' @param signal_cols Optional signal columns. If `NULL`, common Gazepoint
#'   biometric columns are detected automatically.
#' @param group_cols Optional grouping columns. If `NULL`, available source,
#'   participant, media, or trial columns are used.
#' @param zero_is_inactive Logical. If `TRUE`, all-zero signals are labelled as
#'   inactive.
#' @param min_unique_nonzero Minimum number of distinct non-zero finite values
#'   required for an `"active"` status.
#' @param missing_as_inactive Logical. If `TRUE`, all-missing signals are labelled
#'   as insufficient/inactive.
#'
#' @return A list with `overview`, `signal_by_group`, `inactive_groups`,
#'   `inactive_signals`, and `settings`.
#' @export
audit_gazepoint_signal_activity <- function(data,
                                            signal_cols = NULL,
                                            group_cols = NULL,
                                            zero_is_inactive = TRUE,
                                            min_unique_nonzero = 2,
                                            missing_as_inactive = TRUE) {
  if (!is.data.frame(data)) {
    stop("`data` must be a data frame.", call. = FALSE)
  }

  if (!is.logical(zero_is_inactive) || length(zero_is_inactive) != 1) {
    stop("`zero_is_inactive` must be TRUE or FALSE.", call. = FALSE)
  }

  if (!is.numeric(min_unique_nonzero) ||
      length(min_unique_nonzero) != 1 ||
      min_unique_nonzero < 1) {
    stop("`min_unique_nonzero` must be a positive number.", call. = FALSE)
  }

  if (!is.logical(missing_as_inactive) || length(missing_as_inactive) != 1) {
    stop("`missing_as_inactive` must be TRUE or FALSE.", call. = FALSE)
  }

  dat <- as.data.frame(data, stringsAsFactors = FALSE)
  names_dat <- names(dat)

  if (is.null(signal_cols)) {
    signal_cols <- gpbiometrics_activity_infer_signal_cols(names_dat)
  }

  signal_cols <- unique(signal_cols)

  if (length(signal_cols) == 0) {
    stop("No biometric signal columns were found. Supply `signal_cols`.", call. = FALSE)
  }

  missing_signal_cols <- setdiff(signal_cols, names_dat)

  if (length(missing_signal_cols) > 0) {
    stop(
      "`signal_cols` not found in `data`: ",
      paste(missing_signal_cols, collapse = ", "),
      call. = FALSE
    )
  }

  group_cols <- gpbiometrics_activity_resolve_group_cols(names_dat, group_cols)

  missing_group_cols <- setdiff(group_cols, names_dat)

  if (length(missing_group_cols) > 0) {
    stop(
      "`group_cols` not found in `data`: ",
      paste(missing_group_cols, collapse = ", "),
      call. = FALSE
    )
  }

  group_id <- gpbiometrics_activity_group_id(dat, group_cols)
  group_indices <- split(seq_len(nrow(dat)), group_id, drop = TRUE)

  signal_by_group_parts <- list()

  for (group_name in names(group_indices)) {
    idx <- group_indices[[group_name]]
    group_values <- gpbiometrics_activity_group_values(dat, group_cols, idx)

    for (signal in signal_cols) {
      x_raw <- dat[[signal]][idx]
      x_num <- suppressWarnings(as.numeric(x_raw))

      numeric_or_coercible <- !all(is.na(x_num))
      finite_values <- x_num[is.finite(x_num)]
      nonzero_values <- finite_values[finite_values != 0]

      n <- length(x_num)
      n_missing <- sum(is.na(x_num))
      n_zero <- sum(!is.na(x_num) & x_num == 0)
      n_nonzero <- sum(!is.na(x_num) & x_num != 0)
      n_unique_finite <- length(unique(finite_values))
      n_unique_nonzero <- length(unique(nonzero_values))

      sd_value <- if (length(finite_values) > 1) {
        stats::sd(finite_values, na.rm = TRUE)
      } else {
        NA_real_
      }

      mean_value <- if (length(finite_values) > 0) {
        mean(finite_values, na.rm = TRUE)
      } else {
        NA_real_
      }

      min_value <- if (length(finite_values) > 0) {
        min(finite_values, na.rm = TRUE)
      } else {
        NA_real_
      }

      max_value <- if (length(finite_values) > 0) {
        max(finite_values, na.rm = TRUE)
      } else {
        NA_real_
      }

      status <- gpbiometrics_activity_signal_status(
        numeric_or_coercible = numeric_or_coercible,
        n = n,
        n_missing = n_missing,
        n_zero = n_zero,
        n_nonzero = n_nonzero,
        n_unique_finite = n_unique_finite,
        n_unique_nonzero = n_unique_nonzero,
        zero_is_inactive = zero_is_inactive,
        min_unique_nonzero = min_unique_nonzero,
        missing_as_inactive = missing_as_inactive
      )

      row <- data.frame(
        group_id = group_name,
        signal = signal,
        n = n,
        missing_count = n_missing,
        missing_prop = if (n > 0) n_missing / n else NA_real_,
        zero_count = n_zero,
        zero_prop = if (n > 0) n_zero / n else NA_real_,
        nonzero_count = n_nonzero,
        nonzero_prop = if (n > 0) n_nonzero / n else NA_real_,
        unique_finite = n_unique_finite,
        unique_nonzero = n_unique_nonzero,
        mean = mean_value,
        sd = sd_value,
        min = min_value,
        max = max_value,
        numeric_or_coercible = numeric_or_coercible,
        status = status,
        stringsAsFactors = FALSE
      )

      if (length(group_values) > 0) {
        row <- cbind(group_values, row)
      }

      signal_by_group_parts[[length(signal_by_group_parts) + 1L]] <- row
    }
  }

  signal_by_group <- do.call(rbind, signal_by_group_parts)
  rownames(signal_by_group) <- NULL

  inactive_statuses <- c(
    "inactive_all_zero",
    "inactive_constant",
    "low_variation",
    "insufficient_data",
    "nonnumeric"
  )

  group_summary <- gpbiometrics_activity_group_summary(
    signal_by_group = signal_by_group,
    group_cols = group_cols,
    inactive_statuses = inactive_statuses
  )

  inactive_groups <- group_summary[
    group_summary$active_signal_count == 0,
    ,
    drop = FALSE
  ]

  inactive_signals <- gpbiometrics_activity_signal_summary(
    signal_by_group = signal_by_group,
    inactive_statuses = inactive_statuses
  )

  active_group_count <- sum(group_summary$active_signal_count > 0)
  no_active_group_count <- sum(group_summary$active_signal_count == 0)

  status <- if (active_group_count == 0) {
    "fail_no_active_signals"
  } else if (no_active_group_count > 0) {
    "warn_inactive_groups_detected"
  } else if (any(signal_by_group$status != "active")) {
    "warn_inactive_or_low_variation_signals_detected"
  } else {
    "pass"
  }

  overview <- data.frame(
    input_rows = nrow(dat),
    signal_count = length(signal_cols),
    group_count = length(group_indices),
    active_group_count = active_group_count,
    no_active_group_count = no_active_group_count,
    inactive_signal_rows = sum(signal_by_group$status %in% inactive_statuses),
    status = status,
    stringsAsFactors = FALSE
  )

  structure(
    list(
      overview = overview,
      signal_by_group = signal_by_group,
      inactive_groups = inactive_groups,
      inactive_signals = inactive_signals,
      settings = list(
        signal_cols = signal_cols,
        group_cols = group_cols,
        zero_is_inactive = zero_is_inactive,
        min_unique_nonzero = min_unique_nonzero,
        missing_as_inactive = missing_as_inactive
      )
    ),
    class = c("gazepoint_signal_activity_audit", "list")
  )
}

gpbiometrics_time_qc_first_existing <- function(names_dat, candidates) {
  exact <- candidates[candidates %in% names_dat]

  if (length(exact) > 0) {
    return(exact[1])
  }

  lower_names <- tolower(names_dat)
  lower_candidates <- tolower(candidates)
  idx <- match(lower_candidates, lower_names)
  idx <- idx[!is.na(idx)]

  if (length(idx) > 0) {
    return(names_dat[idx[1]])
  }

  NULL
}

gpbiometrics_time_qc_resolve_group_cols <- function(names_dat, group_cols) {
  if (!is.null(group_cols)) {
    return(unique(group_cols))
  }

  candidates <- c(
    "source_file",
    "source_participant",
    "USER",
    "USER_FILE",
    "participant",
    "subject",
    "subject_id",
    "MEDIA_ID",
    "MEDIA_NAME",
    "media_id",
    "media_name",
    "trial",
    "trial_id",
    "trial_global"
  )

  unique(candidates[candidates %in% names_dat])
}

gpbiometrics_time_qc_group_id <- function(dat, group_cols) {
  if (length(group_cols) == 0) {
    return(rep("all", nrow(dat)))
  }

  group_dat <- dat[group_cols]

  group_dat[] <- lapply(group_dat, function(x) {
    x_chr <- as.character(x)
    x_chr[is.na(x_chr)] <- "<NA>"
    x_chr
  })

  apply(group_dat, 1, paste, collapse = "||")
}

gpbiometrics_time_qc_segment_summary <- function(dat,
                                                 row_flags,
                                                 group_cols,
                                                 min_segment_rows) {
  segment_key <- paste(
    row_flags$group_id,
    row_flags$reset_segment_index,
    sep = "||segment_"
  )

  split_rows <- split(seq_len(nrow(row_flags)), segment_key, drop = TRUE)

  out <- lapply(names(split_rows), function(key) {
    idx <- split_rows[[key]]
    time_values <- row_flags$time_value[idx]
    finite_time <- time_values[is.finite(time_values)]

    row <- data.frame(
      segment_key = key,
      group_id = row_flags$group_id[idx[1]],
      segment_index = row_flags$reset_segment_index[idx[1]],
      start_row_id = row_flags$.gpbiometrics_row_id[idx[1]],
      end_row_id = row_flags$.gpbiometrics_row_id[idx[length(idx)]],
      rows = length(idx),
      start_time = if (length(finite_time) > 0) finite_time[1] else NA_real_,
      end_time = if (length(finite_time) > 0) finite_time[length(finite_time)] else NA_real_,
      duration = if (length(finite_time) > 1) {
        finite_time[length(finite_time)] - finite_time[1]
      } else {
        NA_real_
      },
      nonfinite_time_rows = sum(row_flags$flag_nonfinite_time[idx], na.rm = TRUE),
      negative_steps = sum(row_flags$flag_negative_step[idx], na.rm = TRUE),
      duplicate_steps = sum(row_flags$flag_duplicate_time[idx], na.rm = TRUE),
      nonmonotonic_steps = sum(row_flags$flag_nonmonotonic[idx], na.rm = TRUE),
      short_segment = length(idx) < min_segment_rows,
      stringsAsFactors = FALSE
    )

    if (length(group_cols) > 0) {
      row <- cbind(dat[idx[1], group_cols, drop = FALSE], row)
    }

    row
  })

  out <- do.call(rbind, out)
  rownames(out) <- NULL

  out
}

gpbiometrics_time_qc_reindex_time <- function(time_value, group_id, segment_index) {
  out <- rep(NA_real_, length(time_value))
  segment_key <- paste(group_id, segment_index, sep = "||segment_")
  split_rows <- split(seq_along(time_value), segment_key, drop = TRUE)

  for (key in names(split_rows)) {
    idx <- split_rows[[key]]
    x <- time_value[idx]
    finite_idx <- which(is.finite(x))

    if (length(finite_idx) > 0) {
      first_time <- x[finite_idx[1]]
      out[idx] <- x - first_time
    }
  }

  out
}

gpbiometrics_activity_infer_signal_cols <- function(names_dat) {
  candidates <- c(
    "DIAL", "DIALV",
    "GSR", "GSR_US", "GSR_US_TONIC", "GSR_US_PHASIC", "GSRV",
    "HR", "HRV", "HRP", "IBI",
    "dial", "dialv",
    "gsr", "gsr_us", "gsr_us_tonic", "gsr_us_phasic", "gsrv",
    "hr", "hrv", "hrp", "ibi"
  )

  unique(candidates[candidates %in% names_dat])
}

gpbiometrics_activity_resolve_group_cols <- function(names_dat, group_cols) {
  if (!is.null(group_cols)) {
    return(unique(group_cols))
  }

  candidates <- c(
    "source_file",
    "source_participant",
    "USER",
    "USER_FILE",
    "participant",
    "subject",
    "subject_id",
    "MEDIA_ID",
    "MEDIA_NAME",
    "media_id",
    "media_name",
    "trial",
    "trial_id",
    "trial_global"
  )

  unique(candidates[candidates %in% names_dat])
}

gpbiometrics_activity_group_id <- function(dat, group_cols) {
  if (length(group_cols) == 0) {
    return(rep("all", nrow(dat)))
  }

  group_dat <- dat[group_cols]

  group_dat[] <- lapply(group_dat, function(x) {
    x_chr <- as.character(x)
    x_chr[is.na(x_chr)] <- "<NA>"
    x_chr
  })

  apply(group_dat, 1, paste, collapse = "||")
}

gpbiometrics_activity_group_values <- function(dat, group_cols, idx) {
  if (length(group_cols) == 0) {
    return(data.frame(stringsAsFactors = FALSE))
  }

  dat[idx[1], group_cols, drop = FALSE]
}

gpbiometrics_activity_signal_status <- function(numeric_or_coercible,
                                                n,
                                                n_missing,
                                                n_zero,
                                                n_nonzero,
                                                n_unique_finite,
                                                n_unique_nonzero,
                                                zero_is_inactive,
                                                min_unique_nonzero,
                                                missing_as_inactive) {
  if (!isTRUE(numeric_or_coercible)) {
    return("nonnumeric")
  }

  if (n == 0 || (isTRUE(missing_as_inactive) && n_missing == n)) {
    return("insufficient_data")
  }

  if (isTRUE(zero_is_inactive) && n_nonzero == 0 && n_zero > 0) {
    return("inactive_all_zero")
  }

  if (n_unique_finite <= 1) {
    return("inactive_constant")
  }

  if (n_unique_nonzero < min_unique_nonzero) {
    return("low_variation")
  }

  "active"
}

gpbiometrics_activity_group_summary <- function(signal_by_group,
                                                group_cols,
                                                inactive_statuses) {
  group_ids <- unique(signal_by_group$group_id)

  out <- lapply(group_ids, function(group_id_i) {
    d <- signal_by_group[signal_by_group$group_id == group_id_i, , drop = FALSE]

    row <- data.frame(
      group_id = group_id_i,
      signal_count = nrow(d),
      active_signal_count = sum(d$status == "active"),
      inactive_signal_count = sum(d$status %in% inactive_statuses),
      low_variation_signal_count = sum(d$status == "low_variation"),
      all_zero_signal_count = sum(d$status == "inactive_all_zero"),
      constant_signal_count = sum(d$status == "inactive_constant"),
      group_status = if (sum(d$status == "active") == 0) {
        "no_active_signals"
      } else if (sum(d$status == "active") < nrow(d)) {
        "partial_active_signals"
      } else {
        "all_signals_active"
      },
      stringsAsFactors = FALSE
    )

    if (length(group_cols) > 0) {
      group_values <- d[1, group_cols, drop = FALSE]
      row <- cbind(group_values, row)
    }

    row
  })

  out <- do.call(rbind, out)
  rownames(out) <- NULL

  out
}

gpbiometrics_activity_signal_summary <- function(signal_by_group,
                                                 inactive_statuses) {
  signals <- unique(signal_by_group$signal)

  out <- lapply(signals, function(signal_i) {
    d <- signal_by_group[signal_by_group$signal == signal_i, , drop = FALSE]

    data.frame(
      signal = signal_i,
      group_count = nrow(d),
      active_group_count = sum(d$status == "active"),
      inactive_group_count = sum(d$status %in% inactive_statuses),
      all_zero_group_count = sum(d$status == "inactive_all_zero"),
      constant_group_count = sum(d$status == "inactive_constant"),
      low_variation_group_count = sum(d$status == "low_variation"),
      median_nonzero_prop = stats::median(d$nonzero_prop, na.rm = TRUE),
      median_missing_prop = stats::median(d$missing_prop, na.rm = TRUE),
      signal_status = if (sum(d$status == "active") == 0) {
        "inactive_in_all_groups"
      } else if (sum(d$status == "active") < nrow(d)) {
        "active_in_some_groups"
      } else {
        "active_in_all_groups"
      },
      stringsAsFactors = FALSE
    )
  })

  out <- do.call(rbind, out)
  rownames(out) <- NULL

  out
}

test_that("audit_gazepoint_time_resets passes monotonic time within groups", {
  dat <- data.frame(
    source_file = rep(c("a.csv", "b.csv"), each = 5),
    CNT = rep(seq_len(5), 2),
    GSR_US = seq(1, 2, length.out = 10)
  )

  res <- audit_gazepoint_time_resets(
    dat,
    time_col = "CNT",
    group_cols = "source_file"
  )

  expect_s3_class(res, "gazepoint_time_reset_audit")
  expect_equal(res$overview$status, "pass")
  expect_equal(res$overview$negative_steps, 0)
  expect_equal(res$overview$segment_count, 2)
})

test_that("audit_gazepoint_time_resets detects reset within a group", {
  dat <- data.frame(
    source_file = "a.csv",
    CNT = c(1, 2, 3, 1, 2, 3),
    GSR_US = seq(1, 2, length.out = 6)
  )

  res <- audit_gazepoint_time_resets(
    dat,
    time_col = "CNT",
    group_cols = "source_file",
    return_reindexed_time = TRUE
  )

  expect_equal(res$overview$status, "warn_time_irregularities_detected")
  expect_equal(res$overview$negative_steps, 1)
  expect_equal(res$overview$segment_count, 2)
  expect_true("time_reindexed_within_segment" %in% names(res$data_with_segments))
  expect_equal(
    res$data_with_segments$time_reindexed_within_segment,
    c(0, 1, 2, 0, 1, 2)
  )
})

test_that("audit_gazepoint_time_resets handles duplicate time according to allow_ties", {
  dat <- data.frame(
    CNT = c(1, 2, 2, 3),
    HR = c(70, 71, 72, 73)
  )

  res_ties_allowed <- audit_gazepoint_time_resets(
    dat,
    time_col = "CNT",
    allow_ties = TRUE
  )

  res_ties_not_allowed <- audit_gazepoint_time_resets(
    dat,
    time_col = "CNT",
    allow_ties = FALSE
  )

  expect_equal(res_ties_allowed$overview$status, "pass")
  expect_equal(res_ties_allowed$overview$duplicate_steps, 1)

  expect_equal(res_ties_not_allowed$overview$status, "warn_time_irregularities_detected")
  expect_equal(res_ties_not_allowed$overview$nonmonotonic_steps, 1)
})

test_that("audit_gazepoint_time_resets fails for nonnumeric time", {
  dat <- data.frame(
    time_label = c("a", "b", "c"),
    GSR_US = c(1, 2, 3)
  )

  res <- audit_gazepoint_time_resets(
    dat,
    time_col = "time_label"
  )

  expect_equal(res$overview$status, "fail_no_numeric_time")
  expect_equal(res$overview$nonfinite_time_rows, 3)
})

test_that("audit_gazepoint_signal_activity detects all-zero inactive groups", {
  dat <- data.frame(
    source_file = rep(c("inactive.csv", "active.csv"), each = 5),
    GSR_US = c(rep(0, 5), seq(1, 2, length.out = 5)),
    HR = c(rep(0, 5), seq(70, 75, length.out = 5))
  )

  res <- audit_gazepoint_signal_activity(
    dat,
    signal_cols = c("GSR_US", "HR"),
    group_cols = "source_file"
  )

  expect_s3_class(res, "gazepoint_signal_activity_audit")
  expect_equal(res$overview$status, "warn_inactive_groups_detected")
  expect_equal(res$overview$no_active_group_count, 1)
  expect_true(any(res$inactive_groups$source_file == "inactive.csv"))
  expect_true(any(res$signal_by_group$status == "inactive_all_zero"))
})

test_that("audit_gazepoint_signal_activity distinguishes constant nonzero from active signals", {
  dat <- data.frame(
    participant = rep(c("P1", "P2"), each = 5),
    GSR_US = c(rep(1, 5), seq(1, 2, length.out = 5)),
    HR = c(rep(70, 5), seq(70, 75, length.out = 5))
  )

  res <- audit_gazepoint_signal_activity(
    dat,
    signal_cols = c("GSR_US", "HR"),
    group_cols = "participant"
  )

  p1_status <- unique(res$signal_by_group$status[res$signal_by_group$participant == "P1"])
  p2_status <- unique(res$signal_by_group$status[res$signal_by_group$participant == "P2"])

  expect_true(all(p1_status == "inactive_constant"))
  expect_true(all(p2_status == "active"))
})

test_that("audit_gazepoint_signal_activity detects low variation when nonzero uniqueness is low", {
  dat <- data.frame(
    source_file = "a.csv",
    GSR_US = c(0, 0, 1, 1, 1, 1),
    HR = c(70, 71, 72, 73, 74, 75)
  )

  res <- audit_gazepoint_signal_activity(
    dat,
    signal_cols = c("GSR_US", "HR"),
    group_cols = "source_file",
    min_unique_nonzero = 2
  )

  expect_true(any(res$signal_by_group$signal == "GSR_US" & res$signal_by_group$status == "low_variation"))
  expect_true(any(res$signal_by_group$signal == "HR" & res$signal_by_group$status == "active"))
})

test_that("audit_gazepoint_signal_activity auto-detects Gazepoint biometric signals", {
  dat <- data.frame(
    source_file = "a.csv",
    GSR_US = seq(1, 2, length.out = 10),
    HR = seq(70, 80, length.out = 10),
    unrelated = letters[1:10]
  )

  res <- audit_gazepoint_signal_activity(
    dat,
    group_cols = "source_file"
  )

  expect_true(all(c("GSR_US", "HR") %in% res$settings$signal_cols))
  expect_false("unrelated" %in% res$settings$signal_cols)
  expect_equal(res$overview$status, "pass")
})

#' Extract time-varying spectral EDA features
#'
#' Computes a dependency-light approximation of TVSymp-style time-varying
#' spectral EDA power using sliding-window spectral analysis. The default band
#' is 0.08--0.24 Hz, following the TVSymp literature. This function does not
#' claim exact VFCDM reproduction.
#'
#' @param dat A data frame containing EDA data.
#' @param eda_col Numeric EDA/conductance column.
#' @param time_col Numeric time column.
#' @param group_cols Optional grouping columns.
#' @param sampling_rate Optional sampling rate in Hz. If `NULL`, estimated from
#'   `time_col`.
#' @param band Frequency band in Hz used for TVSymp-style power.
#' @param window_seconds Sliding-window length in seconds.
#' @param step_seconds Sliding-window step in seconds.
#' @param min_valid_fraction Minimum valid fraction per window.
#' @param normalise Logical. If `TRUE`, compute EDASympn-style relative band
#'   power normalised by total positive-frequency power.
#'
#' @return A list with `overview`, `tvsymp_timeseries`, `summary`, and `settings`.
#' @export
extract_gazepoint_eda_tvsymp <- function(dat,
                                         eda_col = "GSR_US",
                                         time_col = "CNT",
                                         group_cols = NULL,
                                         sampling_rate = NULL,
                                         band = c(0.08, 0.24),
                                         window_seconds = 60,
                                         step_seconds = 5,
                                         min_valid_fraction = 0.70,
                                         normalise = TRUE) {
  if (!is.data.frame(dat)) {
    stop("`dat` must be a data frame.", call. = FALSE)
  }

  if (!eda_col %in% names(dat)) {
    stop("Column `", eda_col, "` was not found in `dat`.", call. = FALSE)
  }

  if (!time_col %in% names(dat)) {
    stop("Column `", time_col, "` was not found in `dat`.", call. = FALSE)
  }

  if (!is.numeric(dat[[eda_col]])) {
    stop("`eda_col` must identify a numeric column.", call. = FALSE)
  }

  if (!is.numeric(dat[[time_col]])) {
    stop("`time_col` must identify a numeric column.", call. = FALSE)
  }

  if (is.null(group_cols)) {
    group_cols <- character()
  }

  missing_groups <- setdiff(group_cols, names(dat))
  if (length(missing_groups) > 0) {
    stop("Missing `group_cols`: ", paste(missing_groups, collapse = ", "), call. = FALSE)
  }

  if (!is.numeric(band) ||
      length(band) != 2 ||
      any(!is.finite(band)) ||
      band[1] <= 0 ||
      band[1] >= band[2]) {
    stop("`band` must be a positive numeric vector of length two.", call. = FALSE)
  }

  groups <- gpbiometrics_tvsymp_split_indices(dat, group_cols)

  rows <- list()
  row_id <- 1L

  for (group_id in names(groups)) {
    idx <- groups[[group_id]]
    idx <- idx[order(dat[[time_col]][idx])]

    time <- dat[[time_col]][idx]
    eda <- dat[[eda_col]][idx]

    fs <- gpbiometrics_tvsymp_sampling_rate(time, sampling_rate)

    if (!is.finite(fs) || fs <= 0) {
      rows[[row_id]] <- data.frame(
        group_id = group_id,
        window_index = NA_integer_,
        window_start = NA_real_,
        window_end = NA_real_,
        window_midpoint = NA_real_,
        n_samples = length(idx),
        valid_fraction = mean(is.finite(eda)),
        sampling_rate_hz = fs,
        tvsymp_power = NA_real_,
        edasympn = NA_real_,
        total_power = NA_real_,
        status = "sampling_rate_not_available",
        stringsAsFactors = FALSE
      )
      row_id <- row_id + 1L
      next
    }

    min_time <- min(time, na.rm = TRUE)
    max_time <- max(time, na.rm = TRUE)

    starts <- seq(min_time, max_time - window_seconds, by = step_seconds)

    if (length(starts) == 0) {
      starts <- min_time
    }

    for (i in seq_along(starts)) {
      w_start <- starts[i]
      w_end <- w_start + window_seconds
      in_window <- is.finite(time) & time >= w_start & time <= w_end

      x <- eda[in_window]
      valid_fraction <- mean(is.finite(x))

      if (length(x) < 8 || valid_fraction < min_valid_fraction) {
        rows[[row_id]] <- data.frame(
          group_id = group_id,
          window_index = i,
          window_start = w_start,
          window_end = w_end,
          window_midpoint = mean(c(w_start, w_end)),
          n_samples = length(x),
          valid_fraction = valid_fraction,
          sampling_rate_hz = fs,
          tvsymp_power = NA_real_,
          edasympn = NA_real_,
          total_power = NA_real_,
          status = "insufficient_window_data",
          stringsAsFactors = FALSE
        )
        row_id <- row_id + 1L
        next
      }

      x <- gpbiometrics_tvsymp_fill_linear(x)
      x <- x - mean(x, na.rm = TRUE)

      spec <- stats::spec.pgram(
        x,
        taper = 0.1,
        plot = FALSE,
        demean = TRUE,
        detrend = TRUE,
        fast = TRUE
      )

      freq <- spec$freq * fs
      power <- spec$spec

      keep <- is.finite(freq) & is.finite(power) & freq > 0
      freq <- freq[keep]
      power <- power[keep]

      in_band <- freq >= band[1] & freq <= band[2]
      tvsymp_power <- sum(power[in_band], na.rm = TRUE)
      total_power <- sum(power, na.rm = TRUE)

      edasympn <- if (isTRUE(normalise) && is.finite(total_power) && total_power > 0) {
        tvsymp_power / total_power
      } else {
        NA_real_
      }

      rows[[row_id]] <- data.frame(
        group_id = group_id,
        window_index = i,
        window_start = w_start,
        window_end = w_end,
        window_midpoint = mean(c(w_start, w_end)),
        n_samples = length(x),
        valid_fraction = valid_fraction,
        sampling_rate_hz = fs,
        tvsymp_power = tvsymp_power,
        edasympn = edasympn,
        total_power = total_power,
        status = "tvsymp_extracted",
        stringsAsFactors = FALSE
      )

      row_id <- row_id + 1L
    }
  }

  tvsymp_timeseries <- do.call(rbind, rows)
  rownames(tvsymp_timeseries) <- NULL

  summary <- stats::aggregate(
    cbind(tvsymp_power, edasympn) ~ group_id,
    data = tvsymp_timeseries,
    FUN = function(x) mean(x, na.rm = TRUE)
  )

  names(summary)[names(summary) == "tvsymp_power"] <- "mean_tvsymp_power"
  names(summary)[names(summary) == "edasympn"] <- "mean_edasympn"

  overview <- data.frame(
    group_count = length(groups),
    window_rows = nrow(tvsymp_timeseries),
    successful_windows = sum(tvsymp_timeseries$status == "tvsymp_extracted"),
    problem_windows = sum(tvsymp_timeseries$status != "tvsymp_extracted"),
    band_lower_hz = band[1],
    band_upper_hz = band[2],
    status = if (all(tvsymp_timeseries$status == "tvsymp_extracted")) {
      "tvsymp_extraction_complete"
    } else if (any(tvsymp_timeseries$status == "tvsymp_extracted")) {
      "tvsymp_extraction_partial"
    } else {
      "tvsymp_extraction_failed"
    },
    interpretation = paste(
      "TVSymp-style features describe time-varying EDA spectral power.",
      "This is a sliding-window spectral approximation, not exact VFCDM reproduction.",
      "Outputs should not be interpreted as direct stress, emotion, cognition, health status, or diagnosis."
    ),
    stringsAsFactors = FALSE
  )

  structure(
    list(
      overview = overview,
      tvsymp_timeseries = tvsymp_timeseries,
      summary = summary,
      settings = list(
        eda_col = eda_col,
        time_col = time_col,
        group_cols = group_cols,
        sampling_rate = sampling_rate,
        band = band,
        window_seconds = window_seconds,
        step_seconds = step_seconds,
        min_valid_fraction = min_valid_fraction,
        normalise = normalise
      )
    ),
    class = c("gazepoint_eda_tvsymp", "list")
  )
}

gpbiometrics_tvsymp_split_indices <- function(dat, group_cols) {
  if (length(group_cols) == 0) {
    return(list(all_rows = seq_len(nrow(dat))))
  }

  gf <- dat[group_cols]
  gf[] <- lapply(gf, function(x) {
    x <- as.character(x)
    x[is.na(x)] <- "<NA>"
    x
  })

  split(seq_len(nrow(dat)), do.call(paste, c(gf, sep = " | ")))
}

gpbiometrics_tvsymp_sampling_rate <- function(time, sampling_rate = NULL) {
  if (!is.null(sampling_rate)) {
    return(sampling_rate)
  }

  time <- time[is.finite(time)]

  if (length(time) < 3) {
    return(NA_real_)
  }

  dt <- diff(time)
  dt <- dt[is.finite(dt) & dt > 0]

  if (length(dt) == 0) {
    return(NA_real_)
  }

  median_dt <- stats::median(dt)

  if (median_dt > 10) {
    1000 / median_dt
  } else {
    1 / median_dt
  }
}

gpbiometrics_tvsymp_fill_linear <- function(x) {
  idx <- seq_along(x)
  finite <- is.finite(x)

  if (all(finite)) {
    return(x)
  }

  if (sum(finite) == 0) {
    return(rep(0, length(x)))
  }

  if (sum(finite) == 1) {
    return(rep(x[finite][1], length(x)))
  }

  stats::approx(idx[finite], x[finite], xout = idx, rule = 2)$y
}

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = FALSE
)

## -----------------------------------------------------------------------------
# library(gpbiometrics)

## -----------------------------------------------------------------------------
# demo_dir <- system.file(
#   "extdata",
#   "gazepoint_biometrics_kiosk_demo_exports",
#   package = "gpbiometrics"
# )
# 
# workflow <- run_gazepoint_biometrics_workflow(
#   path = demo_dir,
#   include_all_gaze = TRUE,
#   include_fixations = FALSE,
#   include_other_csv = FALSE,
#   expected_sampling_rate_hz = 60
# )
# 
# summarise_gazepoint_biometrics_workflow(workflow)

## -----------------------------------------------------------------------------
# dat <- import_gazepoint_biometrics("path/to/biometric_export.csv")

## -----------------------------------------------------------------------------
# folder <- "path/to/gazepoint_exports"
# 
# dat <- import_gazepoint_biometric_folder(folder)

## -----------------------------------------------------------------------------
# readiness <- run_gazepoint_biometrics_real_data_readiness(dat)
# 
# readiness$overview
# readiness$checks

## -----------------------------------------------------------------------------
# workflow <- run_gazepoint_biometrics_workflow(
#   path = folder,
#   include_all_gaze = TRUE,
#   include_fixations = FALSE,
#   include_other_csv = FALSE,
#   expected_sampling_rate_hz = 60
# )
# 
# summary <- summarise_gazepoint_biometrics_workflow(workflow)
# summary

## -----------------------------------------------------------------------------
# unit_audit <- audit_gazepoint_gsr_units(dat)
# 
# smoothed <- standardise_gazepoint_adaptive_ema(
#   dat,
#   signal_col = "GSR_US",
#   time_col = "CNT"
# )
# 
# scr_recovery <- extract_gazepoint_scr_recovery_times(
#   dat,
#   eda_col = "GSR_US",
#   time_col = "CNT",
#   event_onset_col = "event_onset"
# )

## -----------------------------------------------------------------------------
# temperature_adjusted <- correct_gazepoint_eda_temperature(
#   dat,
#   eda_col = "GSR_US",
#   temperature_cols = c("ambient_temperature", "body_temperature"),
#   group_cols = "participant"
# )

## -----------------------------------------------------------------------------
# hrv <- extract_gazepoint_hrv_features(
#   dat,
#   ibi_col = "IBI",
#   group_cols = "participant"
# )
# 
# hrv_summary <- summarise_gazepoint_hrv_features(hrv)

## -----------------------------------------------------------------------------
# nonlinear_hrv <- extract_gazepoint_hrv_nonlinear(
#   dat,
#   ibi_col = "IBI",
#   group_cols = "participant"
# )
# 
# rcmse <- extract_gazepoint_hrv_rcmse(
#   dat,
#   ibi_col = "IBI",
#   group_cols = "participant"
# )
# 
# fuzzy_csi <- extract_gazepoint_hrv_fuzzy_csi(
#   dat,
#   ibi_col = "IBI",
#   group_cols = "participant"
# )

## -----------------------------------------------------------------------------
# beats <- extract_gazepoint_beats_kmeans(
#   dat,
#   pulse_col = "HRP",
#   time_col = "CNT",
#   group_cols = "participant"
# )

## -----------------------------------------------------------------------------
# ttl_events <- extract_gazepoint_ttl_events(dat)
# 
# aligned <- align_gazepoint_biometrics_to_ttl(
#   dat,
#   ttl_col = "TTL0",
#   time_col = "CNT"
# )
# 
# model_data <- prepare_gazepoint_biometrics_lme_data(
#   dat,
#   outcome_cols = c("GSR_US", "HR"),
#   group_cols = c("participant", "condition"),
#   time_col = "CNT"
# )

## -----------------------------------------------------------------------------
# bundle <- export_gazepoint_biometrics_report_bundle(
#   workflow,
#   output_dir = file.path(tempdir(), "gpbiometrics_report_bundle")
# )

## -----------------------------------------------------------------------------
# inventory <- create_gazepoint_biometrics_feature_inventory()
# 
# formatted_inventory <- format_gazepoint_biometrics_feature_inventory(
#   inventory
# )
# 
# inventory_summary <- summarise_gazepoint_biometrics_feature_inventory(
#   formatted_inventory
# )
# 
# inventory$overview
# inventory_summary$domain_summary
# head(formatted_inventory)

## ----eval=FALSE---------------------------------------------------------------
# private_folder <- "path/to/private_gazepoint_exports"
# private_output <- file.path(tempdir(), "gpbiometrics_real_check")
# 
# workflow <- run_gazepoint_biometrics_workflow(
#   path = private_folder,
#   include_all_gaze = TRUE,
#   include_fixations = FALSE,
#   include_other_csv = FALSE,
#   expected_sampling_rate_hz = 60
# )
# 
# summarise_gazepoint_biometrics_workflow(workflow)
# 
# export_gazepoint_biometrics_report_bundle(
#   workflow,
#   output_dir = private_output
# )
# 
# summary <- summarise_gazepoint_biometrics_workflow(workflow)
# summary

#' Detect stochastic change points in noisy biometric signals
#'
#' Detects abrupt changes in noisy biological time series using a dependency-light
#' stochastic rolling-window approximation. The score combines adjacent-window
#' changes in mean and variance with a robust adaptive threshold.
#'
#' This is not a full reproduction of any specific doubly stochastic model. It
#' is a transparent approximation for QC and exploratory segmentation.
#'
#' @param dat A data frame.
#' @param signal_col Numeric signal column.
#' @param time_col Numeric time column.
#' @param group_cols Optional grouping columns.
#' @param window_seconds Window length in seconds.
#' @param step_seconds Step size in seconds.
#' @param threshold_mad_multiplier Robust threshold multiplier.
#' @param min_distance_s Minimum distance between detected change points.
#'
#' @return A list with `overview`, `score_table`, `changepoints`, and `settings`.
#' @export
detect_gazepoint_doubly_stochastic_changepoints <- function(dat,
                                                            signal_col,
                                                            time_col = "CNT",
                                                            group_cols = NULL,
                                                            window_seconds = 10,
                                                            step_seconds = 2,
                                                            threshold_mad_multiplier = 6,
                                                            min_distance_s = 5) {
  if (!is.data.frame(dat)) {
    stop("`dat` must be a data frame.", call. = FALSE)
  }

  if (!signal_col %in% names(dat)) {
    stop("Column `", signal_col, "` was not found in `dat`.", call. = FALSE)
  }

  if (!time_col %in% names(dat)) {
    stop("Column `", time_col, "` was not found in `dat`.", call. = FALSE)
  }

  if (!is.numeric(dat[[signal_col]])) {
    stop("`signal_col` must identify a numeric column.", call. = FALSE)
  }

  if (!is.numeric(dat[[time_col]])) {
    stop("`time_col` must identify a numeric column.", call. = FALSE)
  }

  if (is.null(group_cols)) {
    group_cols <- character()
  }

  missing_groups <- setdiff(group_cols, names(dat))
  if (length(missing_groups) > 0) {
    stop("Missing `group_cols`: ", paste(missing_groups, collapse = ", "), call. = FALSE)
  }

  groups <- gpbiometrics_cp_split(dat, group_cols)

  score_rows <- list()
  changepoint_rows <- list()
  score_id <- 1L
  cp_id <- 1L

  for (group_id in names(groups)) {
    idx <- groups[[group_id]]
    idx <- idx[order(dat[[time_col]][idx])]

    time <- dat[[time_col]][idx]
    signal <- dat[[signal_col]][idx]

    starts <- seq(
      min(time, na.rm = TRUE),
      max(time, na.rm = TRUE) - window_seconds,
      by = step_seconds
    )

    if (length(starts) < 3) {
      next
    }

    window_summary <- lapply(seq_along(starts), function(i) {
      start <- starts[i]
      end <- start + window_seconds
      in_window <- is.finite(time) & time >= start & time <= end
      x <- signal[in_window]
      x <- x[is.finite(x)]

      data.frame(
        group_id = group_id,
        window_index = i,
        window_start = start,
        window_end = end,
        window_midpoint = mean(c(start, end)),
        n = length(x),
        mean = if (length(x) > 0) mean(x) else NA_real_,
        variance = if (length(x) > 1) stats::var(x) else NA_real_,
        stringsAsFactors = FALSE
      )
    })

    window_summary <- do.call(rbind, window_summary)

    score <- rep(NA_real_, nrow(window_summary))

    for (i in 2:nrow(window_summary)) {
      m1 <- window_summary$mean[i - 1]
      m2 <- window_summary$mean[i]
      v1 <- window_summary$variance[i - 1]
      v2 <- window_summary$variance[i]

      pooled_sd <- sqrt(mean(c(v1, v2), na.rm = TRUE))
      mean_score <- if (is.finite(pooled_sd) && pooled_sd > 0) {
        abs(m2 - m1) / pooled_sd
      } else {
        0
      }

      variance_score <- if (is.finite(v1) && is.finite(v2) && v1 > 0 && v2 > 0) {
        abs(log(v2 / v1))
      } else {
        0
      }

      score[i] <- mean_score + variance_score
    }

    center <- stats::median(score, na.rm = TRUE)
    mad_score <- stats::mad(score, constant = 1, na.rm = TRUE)

    if (!is.finite(mad_score) || mad_score == 0) {
      mad_score <- .Machine$double.eps
    }

    threshold <- center + threshold_mad_multiplier * mad_score
    is_candidate <- is.finite(score) & score > threshold

    window_summary$change_score <- score
    window_summary$threshold <- threshold
    window_summary$changepoint_candidate <- is_candidate
    window_summary$status <- "changepoint_score_created"

    for (i in seq_len(nrow(window_summary))) {
      score_rows[[score_id]] <- window_summary[i, , drop = FALSE]
      score_id <- score_id + 1L
    }

    selected <- gpbiometrics_cp_select(
      candidate_time = window_summary$window_midpoint[is_candidate],
      candidate_score = window_summary$change_score[is_candidate],
      min_distance_s = min_distance_s
    )

    if (length(selected$time) > 0) {
      for (i in seq_along(selected$time)) {
        changepoint_rows[[cp_id]] <- data.frame(
          group_id = group_id,
          changepoint_index = i,
          changepoint_time = selected$time[i],
          change_score = selected$score[i],
          threshold = threshold,
          status = "changepoint_detected",
          stringsAsFactors = FALSE
        )
        cp_id <- cp_id + 1L
      }
    }
  }

  score_table <- if (length(score_rows) > 0) {
    do.call(rbind, score_rows)
  } else {
    data.frame()
  }

  changepoints <- if (length(changepoint_rows) > 0) {
    do.call(rbind, changepoint_rows)
  } else {
    data.frame()
  }

  overview <- data.frame(
    group_count = length(groups),
    score_rows = nrow(score_table),
    changepoint_rows = nrow(changepoints),
    status = "doubly_stochastic_changepoint_screen_complete",
    interpretation = paste(
      "Detected change points are robust stochastic screening markers for abrupt distributional shifts.",
      "They are not proof of a specific latent state or psychological event."
    ),
    stringsAsFactors = FALSE
  )

  structure(
    list(
      overview = overview,
      score_table = score_table,
      changepoints = changepoints,
      settings = list(
        signal_col = signal_col,
        time_col = time_col,
        group_cols = group_cols,
        window_seconds = window_seconds,
        step_seconds = step_seconds,
        threshold_mad_multiplier = threshold_mad_multiplier,
        min_distance_s = min_distance_s
      )
    ),
    class = c("gazepoint_doubly_stochastic_changepoints", "list")
  )
}

#' Extract SCR recovery times
#'
#' Extracts 50 percent half-recovery time (`rec.t2`) and 63 percent recovery
#' time (`rec.tc`) for skin conductance responses from an EDA waveform and
#' event onsets.
#'
#' @param dat A data frame.
#' @param eda_col Numeric EDA/conductance column.
#' @param time_col Numeric time column.
#' @param event_onset_col Optional event onset column. Finite values are treated
#'   as event onsets.
#' @param group_cols Optional grouping columns.
#' @param pre_onset_baseline_s Baseline window before event onset.
#' @param peak_window_s Window after onset used to find the response peak.
#' @param recovery_window_s Window after peak used to find recovery.
#'
#' @return A list with `overview`, `recovery_table`, and `settings`.
#' @export
extract_gazepoint_scr_recovery_times <- function(dat,
                                                 eda_col = "GSR_US",
                                                 time_col = "CNT",
                                                 event_onset_col = NULL,
                                                 group_cols = NULL,
                                                 pre_onset_baseline_s = 2,
                                                 peak_window_s = 5,
                                                 recovery_window_s = 20) {
  if (!is.data.frame(dat)) {
    stop("`dat` must be a data frame.", call. = FALSE)
  }

  required <- c(eda_col, time_col, event_onset_col)
  required <- required[!is.null(required)]

  missing_required <- setdiff(required, names(dat))
  if (length(missing_required) > 0) {
    stop("Missing required columns: ", paste(missing_required, collapse = ", "), call. = FALSE)
  }

  if (!is.numeric(dat[[eda_col]])) {
    stop("`eda_col` must identify a numeric column.", call. = FALSE)
  }

  if (!is.numeric(dat[[time_col]])) {
    stop("`time_col` must identify a numeric column.", call. = FALSE)
  }

  if (is.null(group_cols)) {
    group_cols <- character()
  }

  missing_groups <- setdiff(group_cols, names(dat))
  if (length(missing_groups) > 0) {
    stop("Missing `group_cols`: ", paste(missing_groups, collapse = ", "), call. = FALSE)
  }

  groups <- gpbiometrics_cp_split(dat, group_cols)
  recovery_rows <- list()
  row_id <- 1L

  for (group_id in names(groups)) {
    idx <- groups[[group_id]]
    idx <- idx[order(dat[[time_col]][idx])]

    time <- dat[[time_col]][idx]
    eda <- dat[[eda_col]][idx]

    onsets <- if (!is.null(event_onset_col)) {
      dat[[event_onset_col]][idx]
    } else {
      gpbiometrics_scr_recovery_detect_onsets(time, eda)
    }

    onsets <- sort(unique(onsets[is.finite(onsets)]))

    if (length(onsets) == 0) {
      recovery_rows[[row_id]] <- data.frame(
        group_id = group_id,
        event_index = NA_integer_,
        onset_time = NA_real_,
        baseline = NA_real_,
        peak_time = NA_real_,
        peak_amplitude = NA_real_,
        rec_t2 = NA_real_,
        rec_tc = NA_real_,
        status = "no_events_available",
        stringsAsFactors = FALSE
      )
      row_id <- row_id + 1L
      next
    }

    for (i in seq_along(onsets)) {
      onset <- onsets[i]

      baseline_idx <- is.finite(time) &
        time >= onset - pre_onset_baseline_s &
        time <= onset

      baseline <- if (sum(baseline_idx) > 0) {
        stats::median(eda[baseline_idx], na.rm = TRUE)
      } else {
        eda[which.min(abs(time - onset))]
      }

      peak_idx <- is.finite(time) &
        time >= onset &
        time <= onset + peak_window_s

      if (sum(peak_idx) == 0 || !is.finite(baseline)) {
        recovery_rows[[row_id]] <- data.frame(
          group_id = group_id,
          event_index = i,
          onset_time = onset,
          baseline = baseline,
          peak_time = NA_real_,
          peak_amplitude = NA_real_,
          rec_t2 = NA_real_,
          rec_tc = NA_real_,
          status = "peak_not_found",
          stringsAsFactors = FALSE
        )
        row_id <- row_id + 1L
        next
      }

      local_peak <- which.max(eda[peak_idx])
      peak_rows <- which(peak_idx)
      peak_row <- peak_rows[local_peak]
      peak_time <- time[peak_row]
      peak_value <- eda[peak_row]
      amplitude <- peak_value - baseline

      if (!is.finite(amplitude) || amplitude <= 0) {
        recovery_rows[[row_id]] <- data.frame(
          group_id = group_id,
          event_index = i,
          onset_time = onset,
          baseline = baseline,
          peak_time = peak_time,
          peak_amplitude = amplitude,
          rec_t2 = NA_real_,
          rec_tc = NA_real_,
          status = "nonpositive_response_amplitude",
          stringsAsFactors = FALSE
        )
        row_id <- row_id + 1L
        next
      }

      recovery_idx <- is.finite(time) &
        time >= peak_time &
        time <= peak_time + recovery_window_s

      rec_t2 <- gpbiometrics_scr_recovery_time(
        time = time[recovery_idx],
        eda = eda[recovery_idx],
        baseline = baseline,
        amplitude = amplitude,
        fraction_recovered = 0.50
      )

      rec_tc <- gpbiometrics_scr_recovery_time(
        time = time[recovery_idx],
        eda = eda[recovery_idx],
        baseline = baseline,
        amplitude = amplitude,
        fraction_recovered = 0.63
      )

      recovery_rows[[row_id]] <- data.frame(
        group_id = group_id,
        event_index = i,
        onset_time = onset,
        baseline = baseline,
        peak_time = peak_time,
        peak_amplitude = amplitude,
        rec_t2 = rec_t2,
        rec_tc = rec_tc,
        status = if (is.finite(rec_t2) || is.finite(rec_tc)) {
          "scr_recovery_extracted"
        } else {
          "recovery_not_observed"
        },
        stringsAsFactors = FALSE
      )

      row_id <- row_id + 1L
    }
  }

  recovery_table <- do.call(rbind, recovery_rows)
  rownames(recovery_table) <- NULL

  overview <- data.frame(
    group_count = length(groups),
    recovery_rows = nrow(recovery_table),
    recovered_rows = sum(recovery_table$status == "scr_recovery_extracted"),
    problem_rows = sum(recovery_table$status != "scr_recovery_extracted"),
    status = if (any(recovery_table$status == "scr_recovery_extracted")) {
      "scr_recovery_times_extracted"
    } else {
      "scr_recovery_times_not_extracted"
    },
    interpretation = paste(
      "SCR recovery times describe conductance decay kinetics after a response peak.",
      "They do not directly infer psychological or clinical state."
    ),
    stringsAsFactors = FALSE
  )

  structure(
    list(
      overview = overview,
      recovery_table = recovery_table,
      settings = list(
        eda_col = eda_col,
        time_col = time_col,
        event_onset_col = event_onset_col,
        group_cols = group_cols,
        pre_onset_baseline_s = pre_onset_baseline_s,
        peak_window_s = peak_window_s,
        recovery_window_s = recovery_window_s
      )
    ),
    class = c("gazepoint_scr_recovery_times", "list")
  )
}

gpbiometrics_cp_split <- function(dat, group_cols) {
  if (length(group_cols) == 0) {
    return(list(all_rows = seq_len(nrow(dat))))
  }

  gf <- dat[group_cols]
  gf[] <- lapply(gf, function(x) {
    x <- as.character(x)
    x[is.na(x)] <- "<NA>"
    x
  })

  split(seq_len(nrow(dat)), do.call(paste, c(gf, sep = " | ")))
}

gpbiometrics_cp_select <- function(candidate_time,
                                   candidate_score,
                                   min_distance_s) {
  keep <- is.finite(candidate_time) & is.finite(candidate_score)

  candidate_time <- candidate_time[keep]
  candidate_score <- candidate_score[keep]

  if (length(candidate_time) == 0) {
    return(list(time = numeric(), score = numeric()))
  }

  ord <- order(candidate_score, decreasing = TRUE)
  candidate_time <- candidate_time[ord]
  candidate_score <- candidate_score[ord]

  selected_time <- numeric()
  selected_score <- numeric()

  for (i in seq_along(candidate_time)) {
    if (length(selected_time) == 0 ||
        all(abs(candidate_time[i] - selected_time) >= min_distance_s)) {
      selected_time <- c(selected_time, candidate_time[i])
      selected_score <- c(selected_score, candidate_score[i])
    }
  }

  ord_time <- order(selected_time)

  list(
    time = selected_time[ord_time],
    score = selected_score[ord_time]
  )
}

gpbiometrics_scr_recovery_time <- function(time,
                                           eda,
                                           baseline,
                                           amplitude,
                                           fraction_recovered) {
  if (length(time) == 0 || !is.finite(amplitude) || amplitude <= 0) {
    return(NA_real_)
  }

  target <- baseline + amplitude * (1 - fraction_recovered)
  hit <- which(is.finite(eda) & eda <= target)

  if (length(hit) == 0) {
    return(NA_real_)
  }

  time[hit[1]] - min(time, na.rm = TRUE)
}

gpbiometrics_scr_recovery_detect_onsets <- function(time, eda) {
  if (length(time) < 5 || length(eda) < 5) {
    return(numeric())
  }

  d <- c(NA_real_, diff(eda))
  threshold <- stats::median(d, na.rm = TRUE) + 6 * stats::mad(d, constant = 1, na.rm = TRUE)

  if (!is.finite(threshold)) {
    return(numeric())
  }

  time[which(is.finite(d) & d > threshold)]
}

#' Analyse AC EDA admittance and susceptance recordings
#'
#' Computes summaries for specialised alternating-current EDA recordings. This
#' function is for true AC admittance/susceptance data, not ordinary DC
#' skin-conductance columns such as `GSR_US`.
#'
#' @param dat A data frame.
#' @param conductance_col Optional real conductance component column.
#' @param susceptance_col Optional imaginary susceptance component column.
#' @param admittance_col Optional admittance magnitude column.
#' @param phase_col Optional phase angle column.
#' @param frequency_col Optional AC frequency column.
#' @param time_col Optional time column.
#' @param group_cols Optional grouping columns.
#'
#' @return A list with `overview`, `timeseries`, `summary`, and `settings`.
#' @export
analyze_gazepoint_ac_susceptance <- function(dat,
                                             conductance_col = NULL,
                                             susceptance_col = NULL,
                                             admittance_col = NULL,
                                             phase_col = NULL,
                                             frequency_col = NULL,
                                             time_col = NULL,
                                             group_cols = NULL) {
  if (!is.data.frame(dat)) {
    stop("`dat` must be a data frame.", call. = FALSE)
  }

  supplied_signal_cols <- c(conductance_col, susceptance_col, admittance_col, phase_col)
  supplied_signal_cols <- supplied_signal_cols[!is.null(supplied_signal_cols)]

  if (length(supplied_signal_cols) == 0) {
    stop(
      "Supply at least one AC EDA signal column, such as `conductance_col`, `susceptance_col`, `admittance_col`, or `phase_col`.",
      call. = FALSE
    )
  }

  required <- c(supplied_signal_cols, frequency_col, time_col)
  required <- required[!is.null(required)]

  missing_required <- setdiff(required, names(dat))

  if (length(missing_required) > 0) {
    stop("Missing required columns: ", paste(missing_required, collapse = ", "), call. = FALSE)
  }

  non_numeric <- required[!vapply(dat[required], is.numeric, logical(1))]
  if (length(non_numeric) > 0) {
    stop("These required columns are not numeric: ", paste(non_numeric, collapse = ", "), call. = FALSE)
  }

  if (is.null(group_cols)) {
    group_cols <- character()
  }

  missing_groups <- setdiff(group_cols, names(dat))
  if (length(missing_groups) > 0) {
    stop("Missing `group_cols`: ", paste(missing_groups, collapse = ", "), call. = FALSE)
  }

  timeseries <- dat

  g <- if (!is.null(conductance_col)) {
    dat[[conductance_col]]
  } else {
    rep(NA_real_, nrow(dat))
  }

  b <- if (!is.null(susceptance_col)) {
    dat[[susceptance_col]]
  } else {
    rep(NA_real_, nrow(dat))
  }

  admittance <- if (!is.null(admittance_col)) {
    dat[[admittance_col]]
  } else if (!all(is.na(g)) && !all(is.na(b))) {
    sqrt(g^2 + b^2)
  } else {
    rep(NA_real_, nrow(dat))
  }

  phase <- if (!is.null(phase_col)) {
    dat[[phase_col]]
  } else if (!all(is.na(g)) && !all(is.na(b))) {
    atan2(b, g)
  } else {
    rep(NA_real_, nrow(dat))
  }

  timeseries$ac_eda_conductance_component <- g
  timeseries$ac_eda_susceptance_component <- b
  timeseries$ac_eda_admittance_magnitude <- admittance
  timeseries$ac_eda_phase_radians <- phase

  if (!is.null(frequency_col)) {
    timeseries$ac_eda_frequency <- dat[[frequency_col]]
  } else {
    timeseries$ac_eda_frequency <- NA_real_
  }

  summary_group_cols <- c(group_cols, if (!is.null(frequency_col)) "ac_eda_frequency" else NULL)

  groups <- if (length(summary_group_cols) == 0) {
    list(all_rows = seq_len(nrow(timeseries)))
  } else {
    gpbiometrics_aceda_split(timeseries, summary_group_cols)
  }

  summary_rows <- lapply(names(groups), function(group_id) {
    idx <- groups[[group_id]]

    data.frame(
      group_id = group_id,
      n_rows = length(idx),
      mean_conductance_component = mean(timeseries$ac_eda_conductance_component[idx], na.rm = TRUE),
      mean_susceptance_component = mean(timeseries$ac_eda_susceptance_component[idx], na.rm = TRUE),
      mean_admittance_magnitude = mean(timeseries$ac_eda_admittance_magnitude[idx], na.rm = TRUE),
      mean_phase_radians = mean(timeseries$ac_eda_phase_radians[idx], na.rm = TRUE),
      sd_admittance_magnitude = stats::sd(timeseries$ac_eda_admittance_magnitude[idx], na.rm = TRUE),
      status = "ac_eda_summary_created",
      stringsAsFactors = FALSE
    )
  })

  summary <- do.call(rbind, summary_rows)
  rownames(summary) <- NULL

  overview <- data.frame(
    input_rows = nrow(dat),
    output_rows = nrow(timeseries),
    summary_rows = nrow(summary),
    status = "ac_eda_susceptance_analysis_complete",
    interpretation = paste(
      "AC EDA outputs describe admittance, susceptance, and phase properties when true AC recordings are available.",
      "They are not compatible with ordinary DC GSR_US alone and do not infer emotion, stress, cognition, health status, or diagnosis."
    ),
    stringsAsFactors = FALSE
  )

  structure(
    list(
      overview = overview,
      timeseries = timeseries,
      summary = summary,
      settings = list(
        conductance_col = conductance_col,
        susceptance_col = susceptance_col,
        admittance_col = admittance_col,
        phase_col = phase_col,
        frequency_col = frequency_col,
        time_col = time_col,
        group_cols = group_cols
      )
    ),
    class = c("gazepoint_ac_susceptance", "list")
  )
}

gpbiometrics_aceda_split <- function(dat, group_cols) {
  gf <- dat[group_cols]
  gf[] <- lapply(gf, function(x) {
    x <- as.character(x)
    x[is.na(x)] <- "<NA>"
    x
  })

  split(seq_len(nrow(dat)), do.call(paste, c(gf, sep = " | ")))
}

#' Diagnose a Gazepoint Biometrics workflow
#'
#' Creates a compact diagnostic readiness summary from a Gazepoint Biometrics
#' workflow object. The function does not change or remove data. It returns a
#' one-row decision table with pass/review/fail status and concise reasons.
#'
#' @param workflow A workflow object produced by
#'   `run_gazepoint_biometrics_workflow()`.
#' @param require_gsr Should GSR/EDA be required for a pass status?
#' @param require_hr Should heart rate be required for a pass status?
#' @param require_dial Should engagement dial be required for a pass status?
#' @param max_exclude_window_pct Maximum acceptable percentage of excluded
#'   windows before the workflow is marked as fail.
#' @param max_review_window_pct Maximum acceptable percentage of review windows
#'   before the workflow is marked as review.
#'
#' @return A one-row data frame with diagnostic status and reasons.
#'
#' @export
diagnose_gazepoint_biometrics_workflow <- function(workflow,
                                                   require_gsr = TRUE,
                                                   require_hr = TRUE,
                                                   require_dial = FALSE,
                                                   max_exclude_window_pct = 25,
                                                   max_review_window_pct = 25) {
  if (!inherits(workflow, "gazepoint_biometrics_workflow")) {
    stop(
      "`workflow` must be produced by run_gazepoint_biometrics_workflow().",
      call. = FALSE
    )
  }

  validation_issues <- nrow(workflow$validation$issues)

  active_channels <- workflow$validation$active_channels

  gsr_active <- is_signal_active(active_channels, "gsr_eda")
  hr_active <- is_signal_active(active_channels, "heart_rate")
  dial_active <- is_signal_active(active_channels, "engagement_dial")
  ttl_active <- is_signal_active(active_channels, "ttl_marker")

  required_channel_failures <- character(0)

  if (isTRUE(require_gsr) && !isTRUE(gsr_active)) {
    required_channel_failures <- c(required_channel_failures, "GSR/EDA inactive")
  }

  if (isTRUE(require_hr) && !isTRUE(hr_active)) {
    required_channel_failures <- c(required_channel_failures, "heart rate inactive")
  }

  if (isTRUE(require_dial) && !isTRUE(dial_active)) {
    required_channel_failures <- c(required_channel_failures, "engagement dial inactive")
  }

  quality <- workflow$quality

  low_quality_signals <- character(0)

  if (is.data.frame(quality) && "signal" %in% names(quality) && "usable_pct" %in% names(quality)) {
    low_quality <- quality[!is.na(quality$usable_pct) & quality$usable_pct < 50, , drop = FALSE]

    if (nrow(low_quality) > 0L) {
      low_quality_signals <- unique(low_quality$signal)
    }
  }

  sampling <- workflow$sampling

  sampling_problem_rows <- 0L

  if (is.data.frame(sampling)) {
    bad_sampling <- rep(FALSE, nrow(sampling))

    if ("duplicate_time_rows" %in% names(sampling)) {
      bad_sampling <- bad_sampling | (!is.na(sampling$duplicate_time_rows) & sampling$duplicate_time_rows > 0)
    }

    if ("zero_interval_rows" %in% names(sampling)) {
      bad_sampling <- bad_sampling | (!is.na(sampling$zero_interval_rows) & sampling$zero_interval_rows > 0)
    }

    if ("negative_interval_rows" %in% names(sampling)) {
      bad_sampling <- bad_sampling | (!is.na(sampling$negative_interval_rows) & sampling$negative_interval_rows > 0)
    }

    if ("rate_status" %in% names(sampling)) {
      bad_sampling <- bad_sampling | sampling$rate_status == "outside_tolerance"
    }

    sampling_problem_rows <- sum(bad_sampling, na.rm = TRUE)
  }

  exclusion_summary <- summarise_workflow_exclusion_status(workflow)

  ttl_event_count <- ifelse(
    !is.null(workflow$ttl_events) && is.data.frame(workflow$ttl_events),
    nrow(workflow$ttl_events),
    NA_integer_
  )

  reasons <- character(0)
  fail_reasons <- character(0)
  review_reasons <- character(0)

  if (validation_issues > 0L) {
    fail_reasons <- c(fail_reasons, paste0(validation_issues, " validation issue(s)"))
  }

  if (length(required_channel_failures) > 0L) {
    fail_reasons <- c(fail_reasons, paste(required_channel_failures, collapse = "; "))
  }

  if (exclusion_summary$exclude_window_pct > max_exclude_window_pct) {
    fail_reasons <- c(
      fail_reasons,
      paste0(
        "excluded windows exceed ",
        max_exclude_window_pct,
        "% threshold"
      )
    )
  }

  if (length(low_quality_signals) > 0L) {
    review_reasons <- c(
      review_reasons,
      paste0(
        "low usable coverage in ",
        paste(low_quality_signals, collapse = ", ")
      )
    )
  }

  if (sampling_problem_rows > 0L) {
    review_reasons <- c(
      review_reasons,
      paste0(sampling_problem_rows, " sampling/timing problem row(s)")
    )
  }

  if (exclusion_summary$review_window_pct > max_review_window_pct) {
    review_reasons <- c(
      review_reasons,
      paste0(
        "review windows exceed ",
        max_review_window_pct,
        "% threshold"
      )
    )
  }

  if (is.na(ttl_event_count)) {
    review_reasons <- c(review_reasons, "TTL events not extracted")
  }

  final_status <- if (length(fail_reasons) > 0L) {
    "fail"
  } else if (length(review_reasons) > 0L) {
    "review"
  } else {
    "pass"
  }

  reasons <- c(fail_reasons, review_reasons)

  if (length(reasons) == 0L) {
    reasons <- "workflow diagnostics passed"
  }

  data.frame(
    final_status = final_status,
    diagnostic_reasons = paste(reasons, collapse = "; "),
    validation_issue_count = validation_issues,
    active_gsr_eda = gsr_active,
    active_heart_rate = hr_active,
    active_engagement_dial = dial_active,
    active_ttl_marker = ttl_active,
    low_quality_signal_count = length(low_quality_signals),
    sampling_problem_rows = sampling_problem_rows,
    n_windows = exclusion_summary$n_windows,
    keep_windows = exclusion_summary$keep_windows,
    review_windows = exclusion_summary$review_windows,
    exclude_windows = exclusion_summary$exclude_windows,
    exclude_window_pct = exclusion_summary$exclude_window_pct,
    review_window_pct = exclusion_summary$review_window_pct,
    ttl_event_count = ttl_event_count,
    stringsAsFactors = FALSE
  )
}


summarise_workflow_exclusion_status <- function(workflow) {
  empty <- data.frame(
    n_windows = NA_integer_,
    keep_windows = NA_integer_,
    review_windows = NA_integer_,
    exclude_windows = NA_integer_,
    exclude_window_pct = NA_real_,
    review_window_pct = NA_real_,
    stringsAsFactors = FALSE
  )

  exclusions <- workflow$exclusion_recommendations

  if (is.null(exclusions) ||
      !is.list(exclusions) ||
      !is.data.frame(exclusions$window_recommendations)) {
    return(empty)
  }

  rec <- exclusions$window_recommendations$recommendation

  n_windows <- length(rec)
  keep_windows <- sum(rec == "keep")
  review_windows <- sum(rec == "review")
  exclude_windows <- sum(rec == "exclude")

  data.frame(
    n_windows = n_windows,
    keep_windows = keep_windows,
    review_windows = review_windows,
    exclude_windows = exclude_windows,
    exclude_window_pct = safe_pct(exclude_windows, n_windows),
    review_window_pct = safe_pct(review_windows, n_windows),
    stringsAsFactors = FALSE
  )
}

#' Extract frequency-domain EDA spectral power
#'
#' Computes power spectral density summaries for an EDA signal, including
#' spectral power in the EDASymp-inspired 0.045--0.25 Hz band. This is a
#' descriptive spectral feature and should not be interpreted as direct stress,
#' emotion, valence, cognition, trust, preference, or diagnosis.
#'
#' @param dat A data frame containing EDA data.
#' @param eda_col EDA/conductance column.
#' @param time_col Optional time column.
#' @param group_cols Optional grouping columns.
#' @param sampling_rate Optional sampling rate in Hz. Required if `time_col`
#'   does not allow sampling-rate estimation.
#' @param band Numeric vector of length two defining the frequency band in Hz.
#' @param min_samples Minimum finite samples per group.
#' @param detrend Logical. If `TRUE`, remove a linear trend before spectral analysis.
#'
#' @return A list with `overview`, `spectral_summary`, `settings`, and
#'   interpretation text.
#' @export
extract_gazepoint_eda_spectral_power <- function(dat,
                                                 eda_col = "GSR_US",
                                                 time_col = NULL,
                                                 group_cols = NULL,
                                                 sampling_rate = NULL,
                                                 band = c(0.045, 0.25),
                                                 min_samples = 32,
                                                 detrend = TRUE) {
  if (!is.data.frame(dat)) {
    stop("`dat` must be a data frame.", call. = FALSE)
  }

  if (!eda_col %in% names(dat)) {
    stop("Column `", eda_col, "` was not found in `dat`.", call. = FALSE)
  }

  if (!is.numeric(dat[[eda_col]])) {
    stop("`eda_col` must identify a numeric column.", call. = FALSE)
  }

  if (!is.null(time_col) && !time_col %in% names(dat)) {
    stop("Column `", time_col, "` was not found in `dat`.", call. = FALSE)
  }

  if (!is.null(group_cols)) {
    missing_groups <- setdiff(group_cols, names(dat))
    if (length(missing_groups) > 0) {
      stop(
        "The following `group_cols` were not found in `dat`: ",
        paste(missing_groups, collapse = ", "),
        call. = FALSE
      )
    }
  } else {
    group_cols <- character()
  }

  if (!is.numeric(band) || length(band) != 2 || any(!is.finite(band)) || band[1] <= 0 || band[1] >= band[2]) {
    stop("`band` must be a positive numeric vector of length two with lower < upper.", call. = FALSE)
  }

  if (!is.numeric(min_samples) || length(min_samples) != 1 || min_samples < 4) {
    stop("`min_samples` must be a number >= 4.", call. = FALSE)
  }

  groups <- gpbiometrics_spectral_split_indices(dat, group_cols)

  rows <- lapply(names(groups), function(unit_id) {
    idx <- groups[[unit_id]]
    x <- dat[[eda_col]][idx]

    time_values <- if (!is.null(time_col)) dat[[time_col]][idx] else NULL
    fs <- gpbiometrics_spectral_sampling_rate(time_values, sampling_rate)

    finite <- is.finite(x)
    n_finite <- sum(finite)

    base <- gpbiometrics_spectral_unit_values(dat, idx, group_cols, unit_id)

    if (n_finite < min_samples || !is.finite(fs) || fs <= 0) {
      return(data.frame(
        base,
        unit_id = unit_id,
        n_rows = length(idx),
        n_finite = n_finite,
        sampling_rate_hz = fs,
        total_power = NA_real_,
        band_power = NA_real_,
        relative_band_power = NA_real_,
        peak_frequency_hz = NA_real_,
        spectral_centroid_hz = NA_real_,
        band_lower_hz = band[1],
        band_upper_hz = band[2],
        status = "insufficient_data",
        stringsAsFactors = FALSE,
        check.names = FALSE
      ))
    }

    y <- x[finite]

    if (isTRUE(detrend) && length(y) >= 3) {
      y <- stats::resid(stats::lm(y ~ seq_along(y)))
    } else {
      y <- y - mean(y, na.rm = TRUE)
    }

    spec <- stats::spec.pgram(
      y,
      spans = NULL,
      taper = 0.1,
      plot = FALSE,
      fast = TRUE,
      demean = TRUE,
      detrend = FALSE
    )

    freq <- spec$freq * fs
    power <- spec$spec

    keep_positive <- is.finite(freq) & is.finite(power) & freq > 0

    freq <- freq[keep_positive]
    power <- power[keep_positive]

    in_band <- freq >= band[1] & freq <= band[2]

    total_power <- sum(power, na.rm = TRUE)
    band_power <- sum(power[in_band], na.rm = TRUE)

    rel_band <- if (is.finite(total_power) && total_power > 0) {
      band_power / total_power
    } else {
      NA_real_
    }

    peak_frequency <- if (length(power) > 0 && any(is.finite(power))) {
      freq[which.max(power)]
    } else {
      NA_real_
    }

    centroid <- if (sum(power, na.rm = TRUE) > 0) {
      sum(freq * power, na.rm = TRUE) / sum(power, na.rm = TRUE)
    } else {
      NA_real_
    }

    data.frame(
      base,
      unit_id = unit_id,
      n_rows = length(idx),
      n_finite = n_finite,
      sampling_rate_hz = fs,
      total_power = total_power,
      band_power = band_power,
      relative_band_power = rel_band,
      peak_frequency_hz = peak_frequency,
      spectral_centroid_hz = centroid,
      band_lower_hz = band[1],
      band_upper_hz = band[2],
      status = "spectral_power_extracted",
      stringsAsFactors = FALSE,
      check.names = FALSE
    )
  })

  spectral_summary <- do.call(rbind, rows)
  rownames(spectral_summary) <- NULL

  overview <- data.frame(
    group_count = length(groups),
    spectral_rows = nrow(spectral_summary),
    successful_groups = sum(spectral_summary$status == "spectral_power_extracted"),
    problem_groups = sum(spectral_summary$status != "spectral_power_extracted"),
    eda_col = eda_col,
    band_lower_hz = band[1],
    band_upper_hz = band[2],
    status = if (all(spectral_summary$status == "spectral_power_extracted")) {
      "eda_spectral_power_extracted"
    } else if (any(spectral_summary$status == "spectral_power_extracted")) {
      "eda_spectral_power_partial"
    } else {
      "eda_spectral_power_failed"
    },
    interpretation = paste(
      "Spectral EDA features quantify frequency-domain signal properties.",
      "Band power in the 0.045--0.25 Hz range is an EDASymp-inspired descriptive feature.",
      "It is not a direct stress, emotion, valence, cognition, trust, preference, or diagnosis measure."
    ),
    stringsAsFactors = FALSE
  )

  structure(
    list(
      overview = overview,
      spectral_summary = spectral_summary,
      settings = list(
        eda_col = eda_col,
        time_col = time_col,
        group_cols = group_cols,
        sampling_rate = sampling_rate,
        band = band,
        min_samples = min_samples,
        detrend = detrend
      )
    ),
    class = c("gazepoint_eda_spectral_power", "list")
  )
}

gpbiometrics_spectral_split_indices <- function(dat, group_cols) {
  if (length(group_cols) == 0) {
    return(list(all_rows = seq_len(nrow(dat))))
  }

  group_frame <- dat[group_cols]
  group_frame[] <- lapply(group_frame, function(x) {
    x <- as.character(x)
    x[is.na(x)] <- "<NA>"
    x
  })

  group_key <- do.call(paste, c(group_frame, sep = " | "))
  split(seq_len(nrow(dat)), group_key)
}

gpbiometrics_spectral_unit_values <- function(dat, idx, group_cols, unit_id) {
  if (length(group_cols) == 0) {
    return(data.frame(unit_label = unit_id, stringsAsFactors = FALSE))
  }

  values <- lapply(group_cols, function(nm) as.character(dat[[nm]][idx[1]]))
  names(values) <- group_cols
  as.data.frame(values, stringsAsFactors = FALSE, optional = TRUE)
}

gpbiometrics_spectral_sampling_rate <- function(time_values = NULL, sampling_rate = NULL) {
  if (!is.null(sampling_rate)) {
    if (!is.numeric(sampling_rate) || length(sampling_rate) != 1 || !is.finite(sampling_rate) || sampling_rate <= 0) {
      stop("`sampling_rate` must be a positive finite number.", call. = FALSE)
    }
    return(as.numeric(sampling_rate))
  }

  if (is.null(time_values)) {
    return(NA_real_)
  }

  time_values <- time_values[is.finite(time_values)]

  if (length(time_values) < 3) {
    return(NA_real_)
  }

  dt <- diff(time_values)
  dt <- dt[is.finite(dt) & dt > 0]

  if (length(dt) == 0) {
    return(NA_real_)
  }

  median_dt <- stats::median(dt)

  if (median_dt > 10) {
    return(1000 / median_dt)
  }

  1 / median_dt
}

#' Plot multimodal Gazepoint biometric timelines
#'
#' Creates a conservative timeline plot for one or more biometric channels. The
#' plot is intended for inspection, synchronization checks, and reporting
#' support. It does not interpret electrodermal activity as emotional valence.
#'
#' @param data A data frame containing biometric samples or aligned biometric
#'   rows.
#' @param time_col Optional time column. If `NULL`, common time columns are
#'   detected automatically.
#' @param signal_cols Optional biometric signal columns. If `NULL`, common
#'   Gazepoint biometric columns are detected automatically.
#' @param group_cols Optional grouping columns used to separate trajectories.
#' @param participant_col,stimulus_col,trial_col Optional common grouping columns
#'   to add to `group_cols`.
#' @param event_time_col Optional column containing event times for vertical
#'   markers.
#' @param event_col Optional event/TTL indicator column used for vertical markers.
#' @param standardise Logical. If `TRUE`, signals are z-scored within channel for
#'   visual comparison.
#' @param show_event_markers Logical. Should event markers be drawn when
#'   available?
#' @param title Optional plot title.
#'
#' @return A ggplot object with the long plotting data stored in the `plot_data`
#'   attribute and settings stored in the `settings` attribute.
#' @export
plot_gazepoint_multimodal_timeline <- function(data,
                                               time_col = NULL,
                                               signal_cols = NULL,
                                               group_cols = NULL,
                                               participant_col = NULL,
                                               stimulus_col = NULL,
                                               trial_col = NULL,
                                               event_time_col = NULL,
                                               event_col = NULL,
                                               standardise = TRUE,
                                               show_event_markers = TRUE,
                                               title = NULL) {
  if (!requireNamespace("ggplot2", quietly = TRUE)) {
    stop("Package `ggplot2` is required for timeline plotting.", call. = FALSE)
  }

  if (!is.data.frame(data)) {
    stop("`data` must be a data frame.", call. = FALSE)
  }

  if (!is.logical(standardise) || length(standardise) != 1 || is.na(standardise)) {
    stop("`standardise` must be TRUE or FALSE.", call. = FALSE)
  }

  if (!is.logical(show_event_markers) ||
      length(show_event_markers) != 1 ||
      is.na(show_event_markers)) {
    stop("`show_event_markers` must be TRUE or FALSE.", call. = FALSE)
  }

  dat <- as.data.frame(data, stringsAsFactors = FALSE)

  if (nrow(dat) == 0) {
    stop("`data` has no rows to plot.", call. = FALSE)
  }

  names_dat <- names(dat)

  if (is.null(time_col)) {
    time_col <- gpbiometrics_timeline_first_existing(
      names_dat,
      c(
        "event_relative_time_ms",
        "time_ms",
        "timestamp_ms",
        "timestamp",
        "TIME",
        "Time",
        "time",
        "CNT",
        "cnt"
      )
    )
  }

  if (is.null(time_col) || !time_col %in% names_dat) {
    stop("No usable time column was found. Supply `time_col`.", call. = FALSE)
  }

  time_value <- suppressWarnings(as.numeric(dat[[time_col]]))

  if (all(is.na(time_value))) {
    stop("`time_col` must contain numeric values.", call. = FALSE)
  }

  if (is.null(signal_cols)) {
    signal_cols <- gpbiometrics_timeline_infer_signal_cols(names_dat)
  }

  signal_cols <- unique(signal_cols)

  if (length(signal_cols) == 0) {
    stop("No biometric signal columns were found. Supply `signal_cols`.", call. = FALSE)
  }

  missing_signal_cols <- setdiff(signal_cols, names_dat)

  if (length(missing_signal_cols) > 0) {
    stop(
      "`signal_cols` not found in `data`: ",
      paste(missing_signal_cols, collapse = ", "),
      call. = FALSE
    )
  }

  numeric_signal <- vapply(
    signal_cols,
    function(col) {
      x <- suppressWarnings(as.numeric(dat[[col]]))
      !all(is.na(x))
    },
    logical(1)
  )

  signal_cols <- signal_cols[numeric_signal]

  if (length(signal_cols) == 0) {
    stop("None of the selected signal columns contain numeric values.", call. = FALSE)
  }

  group_cols <- gpbiometrics_timeline_resolve_group_cols(
    names_dat = names_dat,
    group_cols = group_cols,
    participant_col = participant_col,
    stimulus_col = stimulus_col,
    trial_col = trial_col
  )

  missing_group_cols <- setdiff(group_cols, names_dat)

  if (length(missing_group_cols) > 0) {
    stop(
      "`group_cols` not found in `data`: ",
      paste(missing_group_cols, collapse = ", "),
      call. = FALSE
    )
  }

  panel_group <- gpbiometrics_timeline_group_id(dat, group_cols)

  long_data <- gpbiometrics_timeline_long_data(
    dat = dat,
    time_value = time_value,
    time_col = time_col,
    signal_cols = signal_cols,
    panel_group = panel_group,
    standardise = standardise
  )

  y_label <- if (isTRUE(standardise)) {
    "Standardised signal value"
  } else {
    "Signal value"
  }

  p <- ggplot2::ggplot(
    long_data,
    ggplot2::aes(
      x = .data_time,
      y = .data_value,
      group = interaction(.data_group, .data_signal, drop = TRUE)
    )
  ) +
    ggplot2::geom_line(na.rm = TRUE, alpha = 0.7) +
    ggplot2::facet_wrap(stats::as.formula("~ .data_signal"), scales = "free_y") +
    ggplot2::labs(
      x = time_col,
      y = y_label,
      title = title
    ) +
    ggplot2::theme_minimal()

  event_times <- numeric()
  event_data <- data.frame()

  if (isTRUE(show_event_markers)) {
    event_times <- gpbiometrics_timeline_event_times(
      dat = dat,
      time_value = time_value,
      time_col = time_col,
      event_time_col = event_time_col,
      event_col = event_col
    )

    if (length(event_times) > 0) {
      event_data <- data.frame(.event_time = unique(event_times))
      p <- p +
        ggplot2::geom_vline(
          data = event_data,
          ggplot2::aes(xintercept = .event_time),
          inherit.aes = FALSE,
          linetype = "dashed",
          alpha = 0.35
        )
    }
  }

  settings <- list(
    time_col = time_col,
    signal_cols = signal_cols,
    group_cols = group_cols,
    event_time_col = event_time_col,
    event_col = event_col,
    standardise = standardise,
    show_event_markers = show_event_markers,
    event_times = event_times,
    title = title,
    plot_type = "multimodal_timeline",
    interpretation_notes = c(
      "The plot is intended for multimodal signal inspection, synchronization checks, and reporting support.",
      "EDA/GSR and related biometric traces should not be interpreted as emotional valence.",
      "Standardised traces support visual comparison only and do not replace signal-specific modelling."
    )
  )

  p <- standardise_gazepoint_plot_contract(
    plot = p,
    plot_data = long_data,
    settings = settings,
    interpretation_notes = settings$interpretation_notes,
    plot_type = "multimodal_timeline"
  )

  attr(p, "event_data") <- event_data

  p
}

gpbiometrics_timeline_first_existing <- function(names_dat, candidates) {
  exact <- candidates[candidates %in% names_dat]

  if (length(exact) > 0) {
    return(exact[1])
  }

  lower_names <- tolower(names_dat)
  lower_candidates <- tolower(candidates)
  idx <- match(lower_candidates, lower_names)
  idx <- idx[!is.na(idx)]

  if (length(idx) > 0) {
    return(names_dat[idx[1]])
  }

  NULL
}

gpbiometrics_timeline_infer_signal_cols <- function(names_dat) {
  candidates <- c(
    "GSR_US",
    "gsr_us",
    "GSR_US_TONIC",
    "gsr_us_tonic",
    "GSR_US_PHASIC",
    "gsr_us_phasic",
    "GSR",
    "gsr",
    "HR",
    "hr",
    "IBI",
    "ibi",
    "DIAL",
    "dial"
  )

  candidates[candidates %in% names_dat]
}

gpbiometrics_timeline_resolve_group_cols <- function(names_dat,
                                                     group_cols,
                                                     participant_col,
                                                     stimulus_col,
                                                     trial_col) {
  explicit <- unique(stats::na.omit(c(
    group_cols,
    participant_col,
    stimulus_col,
    trial_col
  )))

  if (length(explicit) > 0) {
    return(explicit)
  }

  participant <- gpbiometrics_timeline_first_existing(
    names_dat,
    c("participant", "subject", "subject_id", "USER", "USER_FILE", "user_file")
  )

  stimulus <- gpbiometrics_timeline_first_existing(
    names_dat,
    c("stimulus", "stimulus_id", "MEDIA_ID", "MEDIA_NAME", "media_id", "media_name")
  )

  trial <- gpbiometrics_timeline_first_existing(
    names_dat,
    c("trial", "trial_id", "TRIAL", "trial_global")
  )

  unique(stats::na.omit(c(participant, stimulus, trial)))
}

gpbiometrics_timeline_group_id <- function(dat, group_cols) {
  if (length(group_cols) == 0) {
    return(rep("all", nrow(dat)))
  }

  group_dat <- dat[group_cols]

  group_dat[] <- lapply(group_dat, function(x) {
    x_chr <- as.character(x)
    x_chr[is.na(x_chr)] <- "<NA>"
    x_chr
  })

  apply(group_dat, 1, paste, collapse = "||")
}

gpbiometrics_timeline_long_data <- function(dat,
                                            time_value,
                                            time_col,
                                            signal_cols,
                                            panel_group,
                                            standardise) {
  parts <- vector("list", length(signal_cols))

  for (i in seq_along(signal_cols)) {
    signal <- signal_cols[i]
    value_raw <- suppressWarnings(as.numeric(dat[[signal]]))

    value_plot <- value_raw

    if (isTRUE(standardise)) {
      value_mean <- mean(value_raw, na.rm = TRUE)
      value_sd <- stats::sd(value_raw, na.rm = TRUE)

      if (is.finite(value_sd) && value_sd > 0) {
        value_plot <- (value_raw - value_mean) / value_sd
      } else {
        value_plot <- ifelse(is.na(value_raw), NA_real_, 0)
      }
    }

    parts[[i]] <- data.frame(
      .data_time = time_value,
      .data_value = value_plot,
      .data_value_raw = value_raw,
      .data_signal = signal,
      .data_group = panel_group,
      .data_row_id = seq_len(nrow(dat)),
      stringsAsFactors = FALSE
    )
  }

  out <- do.call(rbind, parts)
  rownames(out) <- NULL

  out
}

gpbiometrics_timeline_event_times <- function(dat,
                                              time_value,
                                              time_col,
                                              event_time_col,
                                              event_col) {
  event_times <- numeric()

  if (!is.null(event_time_col)) {
    if (!event_time_col %in% names(dat)) {
      stop("`event_time_col` was not found in `data`.", call. = FALSE)
    }

    event_times <- suppressWarnings(as.numeric(dat[[event_time_col]]))
    event_times <- event_times[is.finite(event_times)]
    return(unique(event_times))
  }

  if (!is.null(event_col)) {
    if (!event_col %in% names(dat)) {
      stop("`event_col` was not found in `data`.", call. = FALSE)
    }

    active <- gpbiometrics_timeline_active_event(dat[[event_col]])
    event_times <- time_value[active & is.finite(time_value)]
    return(unique(event_times))
  }

  if (identical(time_col, "event_relative_time_ms") ||
      identical(time_col, "event_relative_time")) {
    return(0)
  }

  numeric()
}

gpbiometrics_timeline_active_event <- function(x) {
  if (is.logical(x)) {
    return(!is.na(x) & x)
  }

  if (is.numeric(x)) {
    return(!is.na(x) & x != 0)
  }

  x_chr <- trimws(as.character(x))
  x_chr[is.na(x_chr)] <- ""

  x_num <- suppressWarnings(as.numeric(x_chr))

  if (sum(!is.na(x_num) & x_chr != "") >= max(1, sum(x_chr != "") / 2)) {
    return(!is.na(x_num) & x_num != 0)
  }

  x_chr != "" & !toupper(x_chr) %in% c("0", "FALSE", "F", "NA", "NAN", "NULL")
}

test_that("plot_gazepoint_multimodal_timeline returns ggplot with inferred signals", {
  skip_if_not_installed("ggplot2")

  dat <- data.frame(
    participant = "P1",
    trial = 1,
    time_ms = seq(0, 900, by = 100),
    GSR_US = seq(1, 2, length.out = 10),
    HR = seq(70, 79),
    IBI = seq(850, 760, length.out = 10)
  )

  p <- plot_gazepoint_multimodal_timeline(
    dat,
    time_col = "time_ms",
    group_cols = c("participant", "trial")
  )

  expect_s3_class(p, "ggplot")
  plot_data <- attr(p, "plot_data")
  settings <- attr(p, "settings")

  expect_true(is.data.frame(plot_data))
  expect_equal(sort(unique(plot_data$.data_signal)), c("GSR_US", "HR", "IBI"))
  expect_equal(settings$time_col, "time_ms")
  expect_true(settings$standardise)
})

test_that("plot_gazepoint_multimodal_timeline supports event-relative time marker", {
  skip_if_not_installed("ggplot2")

  dat <- data.frame(
    event_relative_time_ms = seq(-200, 300, by = 100),
    GSR_US = seq(1, 1.5, length.out = 6),
    HR = seq(70, 75)
  )

  p <- plot_gazepoint_multimodal_timeline(dat)

  settings <- attr(p, "settings")

  expect_s3_class(p, "ggplot")
  expect_equal(settings$time_col, "event_relative_time_ms")
  expect_equal(settings$event_times, 0)
})

test_that("plot_gazepoint_multimodal_timeline supports explicit event column", {
  skip_if_not_installed("ggplot2")

  dat <- data.frame(
    time_ms = seq(0, 400, by = 100),
    marker = c(0, 1, 0, 0, 1),
    GSR_US = seq(1, 1.4, length.out = 5)
  )

  p <- plot_gazepoint_multimodal_timeline(
    dat,
    time_col = "time_ms",
    signal_cols = "GSR_US",
    event_col = "marker"
  )

  settings <- attr(p, "settings")

  expect_s3_class(p, "ggplot")
  expect_equal(settings$event_times, c(100, 400))
})

test_that("plot_gazepoint_multimodal_timeline errors when no signal is available", {
  skip_if_not_installed("ggplot2")

  dat <- data.frame(
    time_ms = 1:3,
    participant = "P1"
  )

  expect_error(
    plot_gazepoint_multimodal_timeline(dat),
    "No biometric signal columns"
  )
})

#' Standardise Gazepoint biometric column names
#'
#' Standardises common Gazepoint Biometrics column-name variants to stable
#' canonical names. The helper is intentionally conservative: it recognises
#' common biometric, timing, marker, participant, and stimulus columns, but it
#' leaves unknown columns unchanged apart from optional snake-case cleaning.
#'
#' @param data A data frame or a character vector of column names.
#' @param style Naming style to return. `"canonical"` returns uppercase
#'   Gazepoint-style names for recognised columns. `"snake"` returns lowercase
#'   snake-case names.
#' @param rename Logical. If `data` is a data frame, should the returned data
#'   frame have standardised names? If `FALSE`, a name-mapping table is returned.
#'
#' @return If `data` is a character vector, a character vector of standardised
#'   names. If `data` is a data frame and `rename = TRUE`, the data frame with
#'   standardised names. If `rename = FALSE`, a data frame mapping original names
#'   to standardised names.
#'
#' @examples
#' standardise_gazepoint_biometric_names(c("time ms", "heart rate", "eda uS"))
#'
#' df <- data.frame(`time ms` = 1:3, `heart rate` = c(70, 72, 71))
#' names(standardise_gazepoint_biometric_names(df))
#'
#' @export
standardise_gazepoint_biometric_names <- function(data,
                                                  style = c("canonical", "snake"),
                                                  rename = TRUE) {
  style <- match.arg(style)

  if (is.data.frame(data)) {
    original_names <- names(data)
  } else if (is.character(data)) {
    original_names <- data
  } else {
    stop("`data` must be a data frame or a character vector of column names.",
         call. = FALSE)
  }

  cleaned <- .gpbiom_clean_name(original_names)
  canonical <- .gpbiom_canonical_name(cleaned)

  if (identical(style, "snake")) {
    standard_names <- tolower(canonical)
  } else {
    standard_names <- canonical
  }

  standard_names <- make.unique(standard_names, sep = "_")

  if (is.character(data)) {
    return(standard_names)
  }

  mapping <- data.frame(
    original_name = original_names,
    standard_name = standard_names,
    changed = !identical(original_names, standard_names) &
      original_names != standard_names,
    stringsAsFactors = FALSE
  )

  if (!isTRUE(rename)) {
    return(mapping)
  }

  names(data) <- standard_names
  data
}


#' Detect Gazepoint biometric time columns
#'
#' Detects likely timing and counter columns in Gazepoint Biometrics exports.
#' The function reports candidate timing columns rather than assuming that any
#' single time variable is always present or always measured in the same unit.
#'
#' @param data A data frame or a character vector of column names.
#'
#' @return A data frame with one row per detected time-related column.
#'
#' @examples
#' detect_gazepoint_time_columns(c("CNT", "TIME_MS", "GSR", "HR"))
#'
#' @export
detect_gazepoint_time_columns <- function(data) {
  if (is.data.frame(data)) {
    original_names <- names(data)
  } else if (is.character(data)) {
    original_names <- data
  } else {
    stop("`data` must be a data frame or a character vector of column names.",
         call. = FALSE)
  }

  if (length(original_names) == 0L) {
    return(.gpbiom_empty_time_columns())
  }

  cleaned <- .gpbiom_clean_name(original_names)
  canonical <- .gpbiom_canonical_name(cleaned)

  role <- rep(NA_character_, length(original_names))
  unit_hint <- rep(NA_character_, length(original_names))
  confidence <- rep(0, length(original_names))
  reason <- rep(NA_character_, length(original_names))

  is_counter <- canonical %in% c("CNT", "SAMPLE", "SAMPLE_INDEX")
  role[is_counter] <- "sample_counter"
  unit_hint[is_counter] <- "samples"
  confidence[is_counter] <- 1
  reason[is_counter] <- "Recognised sample counter column."

  is_seconds <- canonical %in% c("TIME", "TIME_S", "TIMESTAMP_S")
  role[is_seconds] <- "timestamp"
  unit_hint[is_seconds] <- "seconds"
  confidence[is_seconds] <- 0.95
  reason[is_seconds] <- "Recognised time column with seconds-like name."

  is_ms <- canonical %in% c("TIME_MS", "TIMESTAMP_MS")
  role[is_ms] <- "timestamp"
  unit_hint[is_ms] <- "milliseconds"
  confidence[is_ms] <- 0.95
  reason[is_ms] <- "Recognised time column with milliseconds-like name."

  is_tick <- canonical %in% c("TIME_TICK", "TIME_TICKS", "TICK", "TICKS")
  role[is_tick] <- "timestamp"
  unit_hint[is_tick] <- "ticks"
  confidence[is_tick] <- 0.90
  reason[is_tick] <- "Recognised tick-style timing column."

  is_trial_time <- canonical %in% c("TRIAL_TIME", "MEDIA_TIME", "STIMULUS_TIME")
  role[is_trial_time] <- "trial_or_media_time"
  unit_hint[is_trial_time] <- "unknown"
  confidence[is_trial_time] <- 0.80
  reason[is_trial_time] <- "Recognised trial/media-relative timing column."

  generic_time <- is.na(role) & grepl("(^|_)time($|_)", cleaned)
  role[generic_time] <- "candidate_time"
  unit_hint[generic_time] <- "unknown"
  confidence[generic_time] <- 0.60
  reason[generic_time] <- "Column name contains a generic time token."

  detected <- !is.na(role)

  if (!any(detected)) {
    return(.gpbiom_empty_time_columns())
  }

  out <- data.frame(
    column = original_names[detected],
    standard_name = canonical[detected],
    role = role[detected],
    unit_hint = unit_hint[detected],
    confidence = confidence[detected],
    reason = reason[detected],
    stringsAsFactors = FALSE
  )

  out[order(-out$confidence, out$column), , drop = FALSE]
}


#' Detect the likely timebase of Gazepoint biometric data
#'
#' Inspects timing and counter columns and returns a conservative summary of the
#' likely primary timebase. Sampling rate is estimated only when numeric timing
#' information is available and intervals are positive.
#'
#' @param data A data frame.
#' @param time_col Optional explicit timing column.
#' @param counter_col Optional explicit counter column.
#'
#' @return A list with `overview`, `time_columns`, `interval_summary`, and
#'   `warnings`.
#'
#' @examples
#' df <- data.frame(CNT = 1:5, TIME = seq(0, by = 1 / 60, length.out = 5))
#' detect_gazepoint_biometric_timebase(df)
#'
#' @export
detect_gazepoint_biometric_timebase <- function(data,
                                                time_col = NULL,
                                                counter_col = NULL) {
  if (!is.data.frame(data)) {
    stop("`data` must be a data frame.", call. = FALSE)
  }

  time_columns <- detect_gazepoint_time_columns(data)
  warnings <- character()

  if (!is.null(time_col)) {
    if (!time_col %in% names(data)) {
      stop("`time_col` was not found in `data`.", call. = FALSE)
    }
    primary_col <- time_col
  } else {
    primary_col <- .gpbiom_choose_primary_time_column(data, time_columns)
  }

  if (!is.null(counter_col)) {
    if (!counter_col %in% names(data)) {
      stop("`counter_col` was not found in `data`.", call. = FALSE)
    }
    counter <- counter_col
  } else {
    counter <- .gpbiom_choose_counter_column(data, time_columns)
  }

  if (is.na(primary_col)) {
    warnings <- c(warnings, "No usable numeric time or counter column detected.")

    overview <- data.frame(
      n_rows = nrow(data),
      primary_time_column = NA_character_,
      primary_time_role = NA_character_,
      unit = NA_character_,
      median_interval = NA_real_,
      sampling_rate_hz = NA_real_,
      counter_column = ifelse(is.na(counter), NA_character_, counter),
      n_valid_intervals = 0L,
      status = "no_timebase_detected",
      stringsAsFactors = FALSE
    )

    return(list(
      overview = overview,
      time_columns = time_columns,
      interval_summary = .gpbiom_empty_interval_summary(),
      warnings = warnings
    ))
  }

  role <- .gpbiom_time_role(primary_col, time_columns)
  unit <- .gpbiom_time_unit(primary_col, time_columns, data[[primary_col]])

  interval_summary <- .gpbiom_interval_summary(data[[primary_col]], unit)
  sampling_rate <- .gpbiom_sampling_rate(interval_summary$median_interval, unit)

  if (is.na(sampling_rate)) {
    warnings <- c(
      warnings,
      "Sampling rate could not be estimated from the selected timebase."
    )
  }

  status <- if (is.na(sampling_rate)) {
    "timebase_detected_without_rate"
  } else {
    "timebase_detected"
  }

  overview <- data.frame(
    n_rows = nrow(data),
    primary_time_column = primary_col,
    primary_time_role = role,
    unit = unit,
    median_interval = interval_summary$median_interval,
    sampling_rate_hz = sampling_rate,
    counter_column = ifelse(is.na(counter), NA_character_, counter),
    n_valid_intervals = interval_summary$n_valid_intervals,
    status = status,
    stringsAsFactors = FALSE
  )

  list(
    overview = overview,
    time_columns = time_columns,
    interval_summary = interval_summary,
    warnings = warnings
  )
}


#' Detect the schema of Gazepoint biometric data
#'
#' Detects likely biometric, timing, marker, and identifying columns in a
#' Gazepoint Biometrics export. The function is deliberately descriptive. It
#' reports what appears to be present and active, but it does not infer emotion,
#' valence, or HRV from ambiguous raw columns.
#'
#' @param data A data frame.
#'
#' @return A list with `overview`, `columns`, `time_columns`, `timebase`,
#'   `name_map`, and `notes`.
#'
#' @examples
#' df <- data.frame(
#'   CNT = 1:5,
#'   TIME = seq(0, by = 1 / 60, length.out = 5),
#'   GSR = c(100, 101, 102, 101, 100),
#'   HR = c(70, 71, 72, 71, 70),
#'   HRV = c(1, 1, 1, 1, 1)
#' )
#' detect_gazepoint_biometric_schema(df)
#'
#' @export
detect_gazepoint_biometric_schema <- function(data) {
  if (!is.data.frame(data)) {
    stop("`data` must be a data frame.", call. = FALSE)
  }

  name_map <- standardise_gazepoint_biometric_names(data, rename = FALSE)
  standard_names <- name_map$standard_name

  detected_columns <- .gpbiom_schema_column_table(data, standard_names)
  time_columns <- detect_gazepoint_time_columns(names(data))
  timebase <- detect_gazepoint_biometric_timebase(data)

  has_group <- function(group) {
    any(detected_columns$signal_group == group & detected_columns$present)
  }

  active_group <- function(group) {
    any(detected_columns$signal_group == group & detected_columns$active)
  }

  active_signal_groups <- unique(detected_columns$signal_group[
    detected_columns$active &
      detected_columns$signal_group %in%
      c("gsr_eda", "heart_rate", "ibi", "engagement_dial", "ttl_marker")
  ])

  validation_notes <- c(
    "Treat raw HRV columns as validity or vendor flags unless documentation proves they contain HRV metrics.",
    "IBI-derived HRV summaries should be computed only from genuine IBI/RR interval columns.",
    "GSR/EDA units should not be overclaimed unless the export column or study documentation identifies them."
  )

  status <- if (has_group("timing") &&
                any(c(has_group("gsr_eda"), has_group("heart_rate"),
                      has_group("ibi"), has_group("engagement_dial")))) {
    "biometric_schema_detected"
  } else if (has_group("timing")) {
    "timing_detected_without_clear_biometric_signal"
  } else {
    "limited_schema_detected"
  }

  overview <- data.frame(
    n_rows = nrow(data),
    n_columns = ncol(data),
    time_column_count = nrow(time_columns),
    has_counter = any(standard_names == "CNT"),
    has_gsr_eda = has_group("gsr_eda"),
    has_gsr_conductance = any(standard_names == "GSR_US"),
    has_gsr_resistance = any(standard_names == "GSR_OHMS"),
    has_heart_rate = has_group("heart_rate"),
    has_hrv_flag = any(standard_names == "HRV"),
    has_ibi = has_group("ibi"),
    has_engagement_dial = has_group("engagement_dial"),
    has_ttl_marker = has_group("ttl_marker"),
    active_gsr_eda = active_group("gsr_eda"),
    active_heart_rate = active_group("heart_rate"),
    active_ibi = active_group("ibi"),
    active_engagement_dial = active_group("engagement_dial"),
    active_ttl_marker = active_group("ttl_marker"),
    active_signal_count = length(active_signal_groups),
    status = status,
    stringsAsFactors = FALSE
  )

  list(
    overview = overview,
    columns = detected_columns,
    time_columns = time_columns,
    timebase = timebase,
    name_map = name_map,
    notes = validation_notes
  )
}


.gpbiom_clean_name <- function(x) {
  x <- trimws(as.character(x))
  x <- gsub("\u00B5", "u", x, fixed = TRUE)
  x <- gsub("[^A-Za-z0-9]+", "_", x)
  x <- gsub("_+", "_", x)
  x <- gsub("^_|_$", "", x)
  tolower(x)
}


.gpbiom_canonical_name <- function(cleaned) {
  out <- toupper(cleaned)

  map <- list(
    CNT = c("cnt", "counter", "sample", "sample_index", "sample_number",
            "sample_no", "sample_id"),
    TIME = c("time", "timestamp", "time_s", "time_sec", "timestamp_s",
             "timestamp_sec", "recording_time", "recording_time_s"),
    TIME_MS = c("time_ms", "timestamp_ms", "recording_time_ms"),
    TIME_TICK = c("time_tick", "time_ticks", "tick", "ticks"),
    TRIAL_TIME = c("trial_time", "trial_time_s", "time_in_trial"),
    MEDIA_TIME = c("media_time", "stimulus_time", "stimulus_time_s"),

    USER = c("user", "participant", "participant_id", "subject",
             "subject_id", "id"),
    USER_FILE = c("user_file", "file", "filename", "source_file"),
    MEDIA_ID = c("media_id", "stimulus_id"),
    MEDIA_NAME = c("media_name", "stimulus", "stimulus_name",
                   "image", "video"),
    TRIAL = c("trial", "trial_id", "trial_number"),
    CONDITION = c("condition", "group", "experimental_condition"),

    GSR = c("gsr", "eda", "electrodermal_activity", "skin_conductance",
            "skin_response"),
    GSR_US = c("gsr_us", "gsr_u_s", "gsr_microsiemens",
               "gsr_micro_siemens", "eda_us", "eda_u_s",
               "eda_microsiemens", "conductance", "conductance_us",
               "skin_conductance_us"),
    GSR_OHMS = c("gsr_ohm", "gsr_ohms", "eda_ohm", "eda_ohms",
                 "resistance", "resistance_ohm", "resistance_ohms",
                 "skin_resistance", "skin_resistance_ohms"),

    HR = c("hr", "heart_rate", "heartrate", "bpm", "pulse",
           "pulse_rate"),
    HRV = c("hrv", "hr_valid", "hr_validity", "heart_rate_valid",
            "heart_rate_validity"),
    IBI = c("ibi", "rr", "rr_interval", "rr_intervals",
            "interbeat_interval", "inter_beat_interval",
            "interbeat_interval_ms", "rr_ms"),

    ENGAGEMENT = c("engagement", "engagement_dial", "dial",
                   "dial_value", "engagement_value", "rotary",
                   "self_reported_engagement"),

    TTL = c("ttl", "ttl_value", "ttl_signal", "ttl_marker",
            "event_marker", "marker", "trigger", "digital_marker"),
    EVENT = c("event", "event_name", "event_label")
  )

  for (target in names(map)) {
    out[cleaned %in% map[[target]]] <- target
  }
  ttl_numbered <- grepl("^ttl[0-9]+$", cleaned)
  out[ttl_numbered] <- "TTL"

  ttl_validity <- cleaned %in% c("ttlv", "ttl_valid", "ttl_validity")
  out[ttl_validity] <- "TTLV"
  out
}


.gpbiom_empty_time_columns <- function() {
  data.frame(
    column = character(),
    standard_name = character(),
    role = character(),
    unit_hint = character(),
    confidence = numeric(),
    reason = character(),
    stringsAsFactors = FALSE
  )
}


.gpbiom_choose_primary_time_column <- function(data, time_columns) {
  if (nrow(time_columns) == 0L) {
    return(NA_character_)
  }

  candidates <- time_columns

  usable <- vapply(candidates$column, function(column) {
    column %in% names(data) && is.numeric(data[[column]]) &&
      sum(!is.na(data[[column]])) >= 2L
  }, logical(1))

  candidates <- candidates[usable, , drop = FALSE]

  if (nrow(candidates) == 0L) {
    return(NA_character_)
  }

  preferred_roles <- c("timestamp", "trial_or_media_time", "sample_counter",
                       "candidate_time")

  for (role in preferred_roles) {
    role_candidates <- candidates[candidates$role == role, , drop = FALSE]
    if (nrow(role_candidates) > 0L) {
      return(role_candidates$column[which.max(role_candidates$confidence)])
    }
  }

  candidates$column[which.max(candidates$confidence)]
}


.gpbiom_choose_counter_column <- function(data, time_columns) {
  if (nrow(time_columns) == 0L) {
    return(NA_character_)
  }

  counter_rows <- time_columns[time_columns$role == "sample_counter", ,
                               drop = FALSE]

  if (nrow(counter_rows) == 0L) {
    return(NA_character_)
  }

  usable <- vapply(counter_rows$column, function(column) {
    column %in% names(data) && is.numeric(data[[column]])
  }, logical(1))

  if (!any(usable)) {
    return(NA_character_)
  }

  counter_rows$column[which(usable)[1L]]
}


.gpbiom_time_role <- function(column, time_columns) {
  if (nrow(time_columns) == 0L || !column %in% time_columns$column) {
    return("unknown")
  }

  time_columns$role[match(column, time_columns$column)]
}


.gpbiom_time_unit <- function(column, time_columns, values) {
  if (nrow(time_columns) > 0L && column %in% time_columns$column) {
    hint <- time_columns$unit_hint[match(column, time_columns$column)]

    if (!is.na(hint) && !hint %in% c("unknown", "samples")) {
      return(hint)
    }

    if (!is.na(hint) && identical(hint, "samples")) {
      return("samples")
    }
  }

  numeric_values <- values[!is.na(values)]

  if (length(numeric_values) < 2L || !is.numeric(numeric_values)) {
    return("unknown")
  }

  diffs <- diff(sort(unique(numeric_values)))
  diffs <- diffs[is.finite(diffs) & diffs > 0]

  if (length(diffs) == 0L) {
    return("unknown")
  }

  median_diff <- stats::median(diffs)

  if (is.finite(median_diff) && median_diff > 0 && median_diff < 1) {
    "seconds"
  } else if (is.finite(median_diff) && median_diff >= 1 && median_diff <= 1000) {
    "milliseconds"
  } else {
    "unknown"
  }
}


.gpbiom_interval_summary <- function(values, unit) {
  if (!is.numeric(values)) {
    return(.gpbiom_empty_interval_summary())
  }

  values <- values[!is.na(values)]

  if (length(values) < 2L) {
    return(.gpbiom_empty_interval_summary())
  }

  diffs <- diff(values)
  valid <- diffs[is.finite(diffs) & diffs > 0]

  if (length(valid) == 0L) {
    return(.gpbiom_empty_interval_summary())
  }

  data.frame(
    unit = unit,
    n_intervals = length(diffs),
    n_valid_intervals = length(valid),
    n_zero_or_negative_intervals = sum(is.finite(diffs) & diffs <= 0),
    min_interval = min(valid),
    median_interval = stats::median(valid),
    mean_interval = mean(valid),
    max_interval = max(valid),
    stringsAsFactors = FALSE
  )
}


.gpbiom_empty_interval_summary <- function() {
  data.frame(
    unit = NA_character_,
    n_intervals = 0L,
    n_valid_intervals = 0L,
    n_zero_or_negative_intervals = 0L,
    min_interval = NA_real_,
    median_interval = NA_real_,
    mean_interval = NA_real_,
    max_interval = NA_real_,
    stringsAsFactors = FALSE
  )
}


.gpbiom_sampling_rate <- function(median_interval, unit) {
  if (length(median_interval) != 1L ||
      !is.finite(median_interval) ||
      median_interval <= 0) {
    return(NA_real_)
  }

  if (identical(unit, "seconds")) {
    return(1 / median_interval)
  }

  if (identical(unit, "milliseconds")) {
    return(1000 / median_interval)
  }

  NA_real_
}


.gpbiom_schema_column_table <- function(data, standard_names) {
  groups <- .gpbiom_schema_groups()

  rows <- lapply(seq_along(standard_names), function(i) {
    standard <- standard_names[i]
    column <- names(data)[i]
    values <- data[[i]]

    group <- .gpbiom_schema_group_for_standard(standard, groups)

    non_missing <- sum(!is.na(values))
    unique_non_missing <- length(unique(values[!is.na(values)]))
    active <- non_missing > 0L && unique_non_missing > 0L

    note <- .gpbiom_schema_note(standard)

    data.frame(
      column = column,
      standard_name = standard,
      signal_group = group,
      present = TRUE,
      active = active,
      n_non_missing = non_missing,
      n_unique_non_missing = unique_non_missing,
      interpretation_note = note,
      stringsAsFactors = FALSE
    )
  })

  do.call(rbind, rows)
}

.gpbiom_schema_group_for_standard <- function(standard, groups) {
  if (standard %in% names(groups)) {
    return(groups[[standard]])
  }

  if (grepl("^TTL(_[0-9]+)?$", standard)) {
    return("ttl_marker")
  }

  if (grepl("^TTLV(_[0-9]+)?$", standard)) {
    return("ttl_validity_flag")
  }

  "other"
}

.gpbiom_schema_groups <- function() {
  c(
    CNT = "timing",
    SAMPLE = "timing",
    SAMPLE_INDEX = "timing",
    TIME = "timing",
    TIME_S = "timing",
    TIMESTAMP_S = "timing",
    TIME_MS = "timing",
    TIMESTAMP_MS = "timing",
    TIME_TICK = "timing",
    TIME_TICKS = "timing",
    TICK = "timing",
    TICKS = "timing",
    TRIAL_TIME = "timing",
    MEDIA_TIME = "timing",
    STIMULUS_TIME = "timing",

    USER = "identifier",
    USER_FILE = "identifier",
    MEDIA_ID = "stimulus",
    MEDIA_NAME = "stimulus",
    TRIAL = "trial",
    CONDITION = "condition",

    GSR = "gsr_eda",
    GSR_US = "gsr_eda",
    GSR_OHMS = "gsr_eda",
    HR = "heart_rate",
    HRV = "heart_rate_validity_flag",
    IBI = "ibi",
    ENGAGEMENT = "engagement_dial",
    TTL = "ttl_marker",
    TTLV = "ttl_validity_flag",
    EVENT = "event"
  )
}


.gpbiom_schema_note <- function(standard_name) {
  if (identical(standard_name, "HRV")) {
    return("Treat as a validity/vendor flag unless documentation proves this column contains HRV metrics.")
  }

  if (identical(standard_name, "IBI")) {
    return("May support IBI/RR-derived HRV summaries if values are genuine inter-beat intervals.")
  }

  if (standard_name %in% c("GSR", "GSR_US", "GSR_OHMS")) {
    return("GSR/EDA unit interpretation depends on export documentation and column naming.")
  }

  if (identical(standard_name, "ENGAGEMENT")) {
    return("Engagement dial/self-report signal; do not interpret as physiological arousal.")
  }

  NA_character_
}

test_that("standardise_gazepoint_biometric_names standardises character vectors", {
  input <- c("time ms", "heart rate", "eda uS", "rr interval", "engagement dial")
  output <- standardise_gazepoint_biometric_names(input)

  expect_equal(output, c("TIME_MS", "HR", "GSR_US", "IBI", "ENGAGEMENT"))
})


test_that("standardise_gazepoint_biometric_names can rename data frames", {
  df <- data.frame(
    `time ms` = 1:3,
    `heart rate` = c(70, 71, 72),
    check.names = FALSE
  )

  renamed <- standardise_gazepoint_biometric_names(df)

  expect_s3_class(renamed, "data.frame")
  expect_equal(names(renamed), c("TIME_MS", "HR"))
})


test_that("standardise_gazepoint_biometric_names can return a mapping table", {
  df <- data.frame(
    `time ms` = 1:3,
    `heart rate` = c(70, 71, 72),
    check.names = FALSE
  )

  mapping <- standardise_gazepoint_biometric_names(df, rename = FALSE)

  expect_s3_class(mapping, "data.frame")
  expect_equal(mapping$original_name, c("time ms", "heart rate"))
  expect_equal(mapping$standard_name, c("TIME_MS", "HR"))
  expect_true(all(mapping$changed))
})


test_that("standardise_gazepoint_biometric_names supports snake style", {
  output <- standardise_gazepoint_biometric_names(
    c("time ms", "heart rate", "eda uS"),
    style = "snake"
  )

  expect_equal(output, c("time_ms", "hr", "gsr_us"))
})


test_that("standardise_gazepoint_biometric_names makes duplicates unique", {
  output <- standardise_gazepoint_biometric_names(c("GSR", "eda"))

  expect_equal(output, c("GSR", "GSR_1"))
})


test_that("detect_gazepoint_time_columns detects counters and timestamps", {
  out <- detect_gazepoint_time_columns(c("CNT", "TIME", "TIME_MS", "GSR"))

  expect_s3_class(out, "data.frame")
  expect_true(all(c("CNT", "TIME", "TIME_MS") %in% out$column))
  expect_true("sample_counter" %in% out$role)
  expect_true("timestamp" %in% out$role)
  expect_true("seconds" %in% out$unit_hint)
  expect_true("milliseconds" %in% out$unit_hint)
})


test_that("detect_gazepoint_time_columns returns empty table when no time column exists", {
  out <- detect_gazepoint_time_columns(c("GSR", "HR", "ENGAGEMENT"))

  expect_s3_class(out, "data.frame")
  expect_equal(nrow(out), 0)
  expect_equal(
    names(out),
    c("column", "standard_name", "role", "unit_hint", "confidence", "reason")
  )
})


test_that("detect_gazepoint_biometric_timebase estimates rate from seconds", {
  df <- data.frame(
    CNT = 1:6,
    TIME = seq(0, by = 1 / 60, length.out = 6),
    GSR = seq(100, 105)
  )

  out <- detect_gazepoint_biometric_timebase(df)

  expect_type(out, "list")
  expect_equal(out$overview$primary_time_column, "TIME")
  expect_equal(out$overview$unit, "seconds")
  expect_equal(round(out$overview$sampling_rate_hz), 60)
  expect_equal(out$overview$status, "timebase_detected")
})


test_that("detect_gazepoint_biometric_timebase estimates rate from milliseconds", {
  df <- data.frame(
    TIME_MS = seq(0, by = 16.6667, length.out = 6),
    HR = c(70, 70, 71, 71, 72, 72)
  )

  out <- detect_gazepoint_biometric_timebase(df)

  expect_equal(out$overview$primary_time_column, "TIME_MS")
  expect_equal(out$overview$unit, "milliseconds")
  expect_equal(round(out$overview$sampling_rate_hz), 60)
})


test_that("detect_gazepoint_biometric_timebase handles no usable timebase", {
  df <- data.frame(
    GSR = c(1, 2, 3),
    HR = c(70, 71, 72)
  )

  out <- detect_gazepoint_biometric_timebase(df)

  expect_equal(out$overview$status, "no_timebase_detected")
  expect_true(is.na(out$overview$primary_time_column))
  expect_true(length(out$warnings) >= 1)
})


test_that("detect_gazepoint_biometric_schema reports active biometric channels", {
  df <- data.frame(
    CNT = 1:6,
    TIME = seq(0, by = 1 / 60, length.out = 6),
    GSR = c(100, 101, 102, 103, 102, 101),
    HR = c(70, 70, 71, 71, 72, 72),
    HRV = c(1, 1, 1, 1, 1, 1),
    ENGAGEMENT = c(50, 51, 52, 53, 54, 55),
    TTL = c(0, 0, 1, 0, 1, 0)
  )

  out <- detect_gazepoint_biometric_schema(df)

  expect_type(out, "list")
  expect_s3_class(out$overview, "data.frame")
  expect_s3_class(out$columns, "data.frame")
  expect_true(out$overview$has_gsr_eda)
  expect_true(out$overview$has_heart_rate)
  expect_true(out$overview$has_hrv_flag)
  expect_true(out$overview$has_engagement_dial)
  expect_true(out$overview$has_ttl_marker)
  expect_true(out$overview$active_gsr_eda)
  expect_true(out$overview$active_heart_rate)
  expect_true(out$overview$active_engagement_dial)
  expect_true(out$overview$active_ttl_marker)
  expect_equal(out$overview$status, "biometric_schema_detected")
})


test_that("detect_gazepoint_biometric_schema detects numbered TTL marker columns", {
  df <- data.frame(
    CNT = 1:5,
    GSR = c(1, 1.1, 1.2, 1.1, 1),
    TTL0 = c(0, 0, 1, 0, 0),
    TTL1 = c(0, 1, 0, 0, 0),
    TTLV = c(1, 1, 1, 1, 1)
  )

  out <- detect_gazepoint_biometric_schema(df)

  expect_true(out$overview$has_ttl_marker)
  expect_true(out$overview$active_ttl_marker)

  ttl_rows <- out$columns[
    out$columns$column %in% c("TTL0", "TTL1", "TTLV"),
    ,
    drop = FALSE
  ]

  expect_equal(nrow(ttl_rows), 3)
  expect_true("TTL0" %in% ttl_rows$column)
  expect_true("TTL1" %in% ttl_rows$column)
  expect_true("TTLV" %in% ttl_rows$column)

  numbered_ttl_rows <- ttl_rows[ttl_rows$column %in% c("TTL0", "TTL1"), ]

  expect_true(all(grepl("^TTL(_[0-9]+)?$", numbered_ttl_rows$standard_name)))

  expect_true(all(numbered_ttl_rows$signal_group == "ttl_marker"))

  expect_equal(
    ttl_rows$standard_name[ttl_rows$column == "TTLV"],
    "TTLV"
  )

  expect_equal(
    ttl_rows$signal_group[ttl_rows$column == "TTLV"],
    "ttl_validity_flag"
  )
})


test_that("detect_gazepoint_biometric_schema keeps HRV interpretation conservative", {
  df <- data.frame(
    CNT = 1:3,
    TIME = c(0, 0.016, 0.032),
    HR = c(70, 71, 72),
    HRV = c(1, 1, 1)
  )

  out <- detect_gazepoint_biometric_schema(df)

  hrv_row <- out$columns[out$columns$standard_name == "HRV", , drop = FALSE]

  expect_equal(hrv_row$signal_group, "heart_rate_validity_flag")
  expect_match(hrv_row$interpretation_note, "validity/vendor flag")
  expect_true(any(grepl("Treat raw HRV columns", out$notes)))
})


test_that("schema helpers reject unsupported inputs", {
  expect_error(
    standardise_gazepoint_biometric_names(1:3),
    "`data` must be"
  )

  expect_error(
    detect_gazepoint_time_columns(1:3),
    "`data` must be"
  )

  expect_error(
    detect_gazepoint_biometric_timebase(c("CNT", "TIME")),
    "`data` must be"
  )

  expect_error(
    detect_gazepoint_biometric_schema(c("CNT", "TIME")),
    "`data` must be"
  )
})

#' Prepare Gazepoint SCR hurdle-model data
#'
#' Converts SCR event-window summaries into model-ready data for two-part
#' electrodermal-response analyses: a binary response/no-response component and
#' a positive-amplitude component among response events.
#'
#' @param scr_event_windows A `gazepoint_scr_event_window_summary` object
#'   returned by `summarise_gazepoint_scr_event_windows()`, or a data frame
#'   containing event-window SCR summaries.
#' @param response_col Column containing the binary SCR response flag.
#' @param amplitude_col Column containing SCR amplitude.
#' @param latency_col Optional column containing SCR latency.
#' @param rise_time_col Optional column containing SCR rise time.
#' @param recovery_time_col Optional column containing SCR recovery time.
#' @param predictor_cols Optional fixed-effect predictor columns to include in
#'   generated formulas and complete-case checks.
#' @param factor_cols Optional columns to coerce to factors.
#' @param numeric_cols Optional columns to coerce to numeric.
#' @param group_cols Optional grouping columns retained for random effects or
#'   clustered summaries.
#' @param event_id_col Optional event identifier column.
#' @param amplitude_transform Transformation for the positive-amplitude outcome:
#'   `"none"`, `"log"`, or `"log1p"`.
#' @param amplitude_offset Small positive offset used when
#'   `amplitude_transform = "log"`.
#' @param drop_missing_predictors Logical. If `TRUE`, model datasets are
#'   restricted to rows complete on outcome and predictor columns.
#'
#' @return A list with `overview`, `response_model_data`,
#'   `amplitude_model_data`, `variable_summary`, `model_formulas`, and
#'   `settings`.
#' @export
prepare_gazepoint_scr_hurdle_model_data <- function(scr_event_windows,
                                                    response_col = "response_flag",
                                                    amplitude_col = "scr_amplitude",
                                                    latency_col = "scr_latency",
                                                    rise_time_col = "scr_rise_time",
                                                    recovery_time_col = "scr_recovery_time",
                                                    predictor_cols = NULL,
                                                    factor_cols = NULL,
                                                    numeric_cols = NULL,
                                                    group_cols = NULL,
                                                    event_id_col = "event_id",
                                                    amplitude_transform = c("none", "log", "log1p"),
                                                    amplitude_offset = 1e-6,
                                                    drop_missing_predictors = TRUE) {
  amplitude_transform <- match.arg(amplitude_transform)

  if (missing(scr_event_windows) || is.null(scr_event_windows)) {
    stop("`scr_event_windows` must be supplied.", call. = FALSE)
  }

  if (!is.logical(drop_missing_predictors) ||
      length(drop_missing_predictors) != 1 ||
      is.na(drop_missing_predictors)) {
    stop("`drop_missing_predictors` must be TRUE or FALSE.", call. = FALSE)
  }

  if (!is.numeric(amplitude_offset) ||
      length(amplitude_offset) != 1 ||
      !is.finite(amplitude_offset) ||
      amplitude_offset <= 0) {
    stop("`amplitude_offset` must be a single positive finite number.", call. = FALSE)
  }

  event_table <- gpbiometrics_scr_hurdle_extract_event_table(scr_event_windows)

  if (!response_col %in% names(event_table)) {
    stop("`response_col` was not found in the event table.", call. = FALSE)
  }

  if (!amplitude_col %in% names(event_table)) {
    stop("`amplitude_col` was not found in the event table.", call. = FALSE)
  }

  required_requested_cols <- unique(c(
    predictor_cols,
    factor_cols,
    numeric_cols,
    group_cols,
    event_id_col
  ))

  required_requested_cols <- required_requested_cols[
    !is.na(required_requested_cols)
  ]

  missing_requested <- setdiff(required_requested_cols, names(event_table))

  if (length(missing_requested) > 0) {
    stop(
      "Requested columns were not found in the event table: ",
      paste(missing_requested, collapse = ", "),
      call. = FALSE
    )
  }

  dat <- as.data.frame(event_table, stringsAsFactors = FALSE)

  for (col in unique(factor_cols)) {
    dat[[col]] <- as.factor(dat[[col]])
  }

  for (col in unique(c(numeric_cols, response_col, amplitude_col))) {
    if (col %in% names(dat)) {
      dat[[col]] <- suppressWarnings(as.numeric(dat[[col]]))
    }
  }

  response_value <- suppressWarnings(as.numeric(dat[[response_col]]))

  if (all(is.na(response_value))) {
    stop("`response_col` must contain numeric or numeric-coercible values.", call. = FALSE)
  }

  response_binary <- ifelse(is.finite(response_value) & response_value > 0, 1L, 0L)
  response_binary[!is.finite(response_value)] <- NA_integer_

  dat$scr_response_binary <- response_binary

  amplitude_value <- suppressWarnings(as.numeric(dat[[amplitude_col]]))
  dat$scr_amplitude_raw <- amplitude_value

  dat$scr_amplitude_positive <- ifelse(
    dat$scr_response_binary == 1L & is.finite(amplitude_value) & amplitude_value > 0,
    amplitude_value,
    NA_real_
  )

  dat$scr_amplitude_model <- gpbiometrics_scr_hurdle_transform_amplitude(
    x = dat$scr_amplitude_positive,
    transform = amplitude_transform,
    offset = amplitude_offset
  )

  dat$scr_latency_model <- if (latency_col %in% names(dat)) {
    suppressWarnings(as.numeric(dat[[latency_col]]))
  } else {
    NA_real_
  }

  dat$scr_rise_time_model <- if (rise_time_col %in% names(dat)) {
    suppressWarnings(as.numeric(dat[[rise_time_col]]))
  } else {
    NA_real_
  }

  dat$scr_recovery_time_model <- if (recovery_time_col %in% names(dat)) {
    suppressWarnings(as.numeric(dat[[recovery_time_col]]))
  } else {
    NA_real_
  }

  response_complete_cols <- unique(c(
    "scr_response_binary",
    predictor_cols,
    group_cols,
    event_id_col
  ))

  response_complete_cols <- response_complete_cols[response_complete_cols %in% names(dat)]

  amplitude_complete_cols <- unique(c(
    "scr_amplitude_model",
    predictor_cols,
    group_cols,
    event_id_col
  ))

  amplitude_complete_cols <- amplitude_complete_cols[amplitude_complete_cols %in% names(dat)]

  dat$response_model_complete <- stats::complete.cases(dat[response_complete_cols])
  dat$amplitude_model_complete <- stats::complete.cases(dat[amplitude_complete_cols]) &
    dat$scr_response_binary == 1L &
    is.finite(dat$scr_amplitude_model)

  response_model_data <- dat[!is.na(dat$scr_response_binary), , drop = FALSE]

  if (isTRUE(drop_missing_predictors)) {
    response_model_data <- response_model_data[
      response_model_data$response_model_complete,
      ,
      drop = FALSE
    ]
  }

  amplitude_model_data <- dat[
    dat$scr_response_binary == 1L & is.finite(dat$scr_amplitude_model),
    ,
    drop = FALSE
  ]

  if (isTRUE(drop_missing_predictors)) {
    amplitude_model_data <- amplitude_model_data[
      amplitude_model_data$amplitude_model_complete,
      ,
      drop = FALSE
    ]
  }

  variable_summary <- gpbiometrics_scr_hurdle_variable_summary(
    dat = dat,
    predictor_cols = predictor_cols,
    group_cols = group_cols,
    factor_cols = factor_cols,
    numeric_cols = numeric_cols
  )

  model_formulas <- gpbiometrics_scr_hurdle_formulas(
    predictor_cols = predictor_cols,
    group_cols = group_cols
  )

  overview <- data.frame(
    input_events = nrow(dat),
    response_model_rows = nrow(response_model_data),
    amplitude_model_rows = nrow(amplitude_model_data),
    response_events = sum(dat$scr_response_binary == 1L, na.rm = TRUE),
    no_response_events = sum(dat$scr_response_binary == 0L, na.rm = TRUE),
    response_rate = mean(dat$scr_response_binary == 1L, na.rm = TRUE),
    positive_amplitude_events = sum(is.finite(dat$scr_amplitude_positive), na.rm = TRUE),
    predictor_count = length(predictor_cols),
    group_count = length(group_cols),
    amplitude_transform = amplitude_transform,
    status = if (nrow(response_model_data) == 0) {
      "fail_no_response_model_rows"
    } else if (nrow(amplitude_model_data) == 0) {
      "warn_no_positive_amplitude_rows"
    } else {
      "scr_hurdle_model_data_prepared"
    },
    stringsAsFactors = FALSE
  )

  structure(
    list(
      overview = overview,
      response_model_data = response_model_data,
      amplitude_model_data = amplitude_model_data,
      variable_summary = variable_summary,
      model_formulas = model_formulas,
      settings = list(
        response_col = response_col,
        amplitude_col = amplitude_col,
        latency_col = latency_col,
        rise_time_col = rise_time_col,
        recovery_time_col = recovery_time_col,
        predictor_cols = predictor_cols,
        factor_cols = factor_cols,
        numeric_cols = numeric_cols,
        group_cols = group_cols,
        event_id_col = event_id_col,
        amplitude_transform = amplitude_transform,
        amplitude_offset = amplitude_offset,
        drop_missing_predictors = drop_missing_predictors,
        interpretation_notes = c(
          "The response model is a binary SCR response/no-response model.",
          "The amplitude model uses only response events with positive finite SCR amplitudes.",
          "EDA/SCR outcomes indicate electrodermal response features, not emotional valence."
        )
      )
    ),
    class = c("gazepoint_scr_hurdle_model_data", "list")
  )
}

gpbiometrics_scr_hurdle_extract_event_table <- function(scr_event_windows) {
  if (inherits(scr_event_windows, "gazepoint_scr_event_window_summary") &&
      !is.null(scr_event_windows$event_table)) {
    return(as.data.frame(scr_event_windows$event_table, stringsAsFactors = FALSE))
  }

  if (is.data.frame(scr_event_windows)) {
    return(as.data.frame(scr_event_windows, stringsAsFactors = FALSE))
  }

  stop(
    "`scr_event_windows` must be an SCR event-window summary object or a data frame.",
    call. = FALSE
  )
}

gpbiometrics_scr_hurdle_transform_amplitude <- function(x,
                                                        transform,
                                                        offset) {
  if (identical(transform, "none")) {
    return(x)
  }

  if (identical(transform, "log")) {
    return(log(x + offset))
  }

  if (identical(transform, "log1p")) {
    return(log1p(x))
  }

  x
}

gpbiometrics_scr_hurdle_variable_summary <- function(dat,
                                                     predictor_cols,
                                                     group_cols,
                                                     factor_cols,
                                                     numeric_cols) {
  cols <- unique(c(
    "scr_response_binary",
    "scr_amplitude_raw",
    "scr_amplitude_positive",
    "scr_amplitude_model",
    "scr_latency_model",
    "scr_rise_time_model",
    "scr_recovery_time_model",
    predictor_cols,
    group_cols,
    factor_cols,
    numeric_cols
  ))

  cols <- cols[cols %in% names(dat)]

  out <- lapply(cols, function(col) {
    x <- dat[[col]]

    if (is.numeric(x) || is.integer(x)) {
      finite_x <- x[is.finite(x)]

      data.frame(
        variable = col,
        class = paste(class(x), collapse = "/"),
        missing = sum(is.na(x)),
        unique_values = length(unique(x[!is.na(x)])),
        mean = if (length(finite_x) > 0) mean(finite_x, na.rm = TRUE) else NA_real_,
        sd = if (length(finite_x) > 1) stats::sd(finite_x, na.rm = TRUE) else NA_real_,
        min = if (length(finite_x) > 0) min(finite_x, na.rm = TRUE) else NA_real_,
        max = if (length(finite_x) > 0) max(finite_x, na.rm = TRUE) else NA_real_,
        stringsAsFactors = FALSE
      )
    } else {
      data.frame(
        variable = col,
        class = paste(class(x), collapse = "/"),
        missing = sum(is.na(x)),
        unique_values = length(unique(x[!is.na(x)])),
        mean = NA_real_,
        sd = NA_real_,
        min = NA_real_,
        max = NA_real_,
        stringsAsFactors = FALSE
      )
    }
  })

  out <- do.call(rbind, out)
  rownames(out) <- NULL

  out
}

gpbiometrics_scr_hurdle_formulas <- function(predictor_cols,
                                             group_cols) {
  fixed_rhs <- if (length(predictor_cols) > 0) {
    paste(predictor_cols, collapse = " + ")
  } else {
    "1"
  }

  random_rhs <- if (length(group_cols) > 0) {
    paste0("(1 | ", group_cols, ")", collapse = " + ")
  } else {
    character()
  }

  rhs <- paste(c(fixed_rhs, random_rhs), collapse = " + ")

  data.frame(
    model_component = c("response", "amplitude"),
    outcome = c("scr_response_binary", "scr_amplitude_model"),
    formula = c(
      paste("scr_response_binary ~", rhs),
      paste("scr_amplitude_model ~", rhs)
    ),
    suggested_family = c("binomial", "gaussian_or_gamma_positive"),
    stringsAsFactors = FALSE
  )
}

#' Prepare Gazepoint EDA data for PsPM DCM workflows
#'
#' Prepares Gazepoint EDA data and event metadata for downstream PsPM dynamic
#' causal modelling workflows. This function does not run PsPM or invert a DCM
#' model in R. It creates structured input tables and notes for MATLAB/PsPM.
#'
#' @param dat A data frame containing EDA data.
#' @param eda_col Numeric EDA/conductance column.
#' @param time_col Numeric time column.
#' @param event_onset_col Optional event onset column.
#' @param event_duration_col Optional event duration column.
#' @param event_name_col Optional event name/condition column.
#' @param participant_col Optional participant column.
#' @param session_col Optional session column.
#' @param sampling_rate Optional sampling rate in Hz.
#' @param output_dir Optional directory for CSV export.
#' @param prefix File prefix when `output_dir` is supplied.
#'
#' @return A list with `overview`, `signal_table`, `event_table`,
#'   `pspm_notes`, `written_files`, and `settings`.
#' @export
prepare_gazepoint_pspm_dcm_input <- function(dat,
                                             eda_col = "GSR_US",
                                             time_col = "CNT",
                                             event_onset_col = NULL,
                                             event_duration_col = NULL,
                                             event_name_col = NULL,
                                             participant_col = NULL,
                                             session_col = NULL,
                                             sampling_rate = NULL,
                                             output_dir = NULL,
                                             prefix = "gazepoint_pspm_dcm") {
  if (!is.data.frame(dat)) {
    stop("`dat` must be a data frame.", call. = FALSE)
  }

  required <- c(
    eda_col,
    time_col,
    event_onset_col,
    event_duration_col,
    event_name_col,
    participant_col,
    session_col
  )

  required <- required[!is.null(required)]

  missing_required <- setdiff(required, names(dat))

  if (length(missing_required) > 0) {
    stop("Missing required columns: ", paste(missing_required, collapse = ", "), call. = FALSE)
  }

  if (!is.numeric(dat[[eda_col]])) {
    stop("`eda_col` must identify a numeric column.", call. = FALSE)
  }

  if (!is.numeric(dat[[time_col]])) {
    stop("`time_col` must identify a numeric column.", call. = FALSE)
  }

  signal_table <- data.frame(
    time = dat[[time_col]],
    conductance = dat[[eda_col]],
    stringsAsFactors = FALSE
  )

  if (!is.null(participant_col)) {
    signal_table$participant <- dat[[participant_col]]
  }

  if (!is.null(session_col)) {
    signal_table$session <- dat[[session_col]]
  }

  signal_table <- signal_table[is.finite(signal_table$time), , drop = FALSE]
  signal_table <- signal_table[order(signal_table$time), , drop = FALSE]

  if (is.null(sampling_rate)) {
    sampling_rate <- gpbiometrics_pspm_dcm_sampling_rate(signal_table$time)
  }

  event_table <- data.frame()

  if (!is.null(event_onset_col)) {
    event_table <- data.frame(
      onset = dat[[event_onset_col]],
      duration = if (!is.null(event_duration_col)) {
        dat[[event_duration_col]]
      } else {
        0
      },
      name = if (!is.null(event_name_col)) {
        as.character(dat[[event_name_col]])
      } else {
        "event"
      },
      stringsAsFactors = FALSE
    )

    if (!is.null(participant_col)) {
      event_table$participant <- dat[[participant_col]]
    }

    if (!is.null(session_col)) {
      event_table$session <- dat[[session_col]]
    }

    event_table <- unique(event_table[is.finite(event_table$onset), , drop = FALSE])
    event_table <- event_table[order(event_table$onset), , drop = FALSE]
  }

  pspm_notes <- c(
    "This object prepares Gazepoint EDA signal and event tables for downstream PsPM DCM workflows.",
    "It does not run MATLAB, PsPM, Bayesian model inversion, or DCM estimation inside R.",
    "Use the signal table as the observed skin-conductance time series and the event table as model input/onset metadata.",
    "Verify PsPM version, units, sampling rate, filtering, event timing, and participant/session splitting before model inversion.",
    "DCM outputs should be interpreted as model-based estimates of latent sudomotor drivers under the chosen model, not direct psychological labels."
  )

  written_files <- character()

  if (!is.null(output_dir)) {
    if (!dir.exists(output_dir)) {
      dir.create(output_dir, recursive = TRUE, showWarnings = FALSE)
    }

    signal_file <- file.path(output_dir, paste0(prefix, "_signal.csv"))
    utils::write.csv(signal_table, signal_file, row.names = FALSE)
    written_files <- c(written_files, signal_file)

    if (nrow(event_table) > 0) {
      event_file <- file.path(output_dir, paste0(prefix, "_events.csv"))
      utils::write.csv(event_table, event_file, row.names = FALSE)
      written_files <- c(written_files, event_file)
    }

    notes_file <- file.path(output_dir, paste0(prefix, "_notes.txt"))
    writeLines(pspm_notes, notes_file, useBytes = TRUE)
    written_files <- c(written_files, notes_file)
  }

  overview <- data.frame(
    signal_rows = nrow(signal_table),
    event_rows = nrow(event_table),
    sampling_rate_hz = sampling_rate,
    output_written = length(written_files) > 0,
    status = if (nrow(signal_table) > 0) {
      "pspm_dcm_input_prepared"
    } else {
      "pspm_dcm_input_empty_signal"
    },
    interpretation = paste(
      "Prepared tables are an interoperability bridge for PsPM DCM workflows.",
      "They do not constitute DCM estimation or psychological-state inference."
    ),
    stringsAsFactors = FALSE
  )

  structure(
    list(
      overview = overview,
      signal_table = signal_table,
      event_table = event_table,
      pspm_notes = pspm_notes,
      written_files = written_files,
      settings = list(
        eda_col = eda_col,
        time_col = time_col,
        event_onset_col = event_onset_col,
        event_duration_col = event_duration_col,
        event_name_col = event_name_col,
        participant_col = participant_col,
        session_col = session_col,
        sampling_rate = sampling_rate,
        output_dir = output_dir,
        prefix = prefix
      )
    ),
    class = c("gazepoint_pspm_dcm_input", "list")
  )
}

gpbiometrics_pspm_dcm_sampling_rate <- function(time) {
  time <- time[is.finite(time)]

  if (length(time) < 3) {
    return(NA_real_)
  }

  dt <- diff(time)
  dt <- dt[is.finite(dt) & dt > 0]

  if (length(dt) == 0) {
    return(NA_real_)
  }

  median_dt <- stats::median(dt)

  if (median_dt > 10) {
    1000 / median_dt
  } else {
    1 / median_dt
  }
}

#' Summarise Gazepoint GSR/EDA windows
#'
#' Summarises Gazepoint GSR/EDA values within participant, trial, stimulus, AOI,
#' or other user-defined windows. When available, `GSR_US` is used by default
#' because it represents skin conductance in microsiemens in Gazepoint exports.
#'
#' @param data A data frame or a path to a Gazepoint CSV export.
#' @param group_columns Optional grouping columns defining windows, such as
#'   `c("USER", "MEDIA_ID")`.
#' @param value_column Optional GSR/EDA value column. If `NULL`, `GSR_US` is
#'   used when present, otherwise `GSR`.
#' @param validity_column Optional validity column. Defaults to `"GSRV"`.
#' @param exclude_zero Should zero values be excluded from usable summaries?
#'
#' @return A data frame with one row per window.
#'
#' @export
summarise_gazepoint_gsr_windows <- function(data,
                                            group_columns = NULL,
                                            value_column = NULL,
                                            validity_column = "GSRV",
                                            exclude_zero = TRUE) {
  dat <- coerce_gazepoint_biometrics_data(data)

  if (is.null(value_column)) {
    value_column <- choose_first_present(dat, c("GSR_US", "GSR"))
  }

  summarise_gazepoint_signal_windows(
    data = dat,
    signal = "gsr_eda",
    group_columns = group_columns,
    value_column = value_column,
    validity_column = validity_column,
    exclude_zero = exclude_zero
  )
}


#' Summarise Gazepoint heart-rate windows
#'
#' Summarises Gazepoint heart-rate values within participant, trial, stimulus,
#' AOI, or other user-defined windows. `HRV` is treated as a validity flag, not
#' as a heart-rate-variability metric.
#'
#' @param data A data frame or a path to a Gazepoint CSV export.
#' @param group_columns Optional grouping columns defining windows, such as
#'   `c("USER", "MEDIA_ID")`.
#' @param value_column Heart-rate value column. Defaults to `"HR"`.
#' @param validity_column Heart-rate validity column. Defaults to `"HRV"`.
#' @param exclude_zero Should zero values be excluded from usable summaries?
#'
#' @return A data frame with one row per window.
#'
#' @export
summarise_gazepoint_hr_windows <- function(data,
                                           group_columns = NULL,
                                           value_column = "HR",
                                           validity_column = "HRV",
                                           exclude_zero = TRUE) {
  dat <- coerce_gazepoint_biometrics_data(data)

  summarise_gazepoint_signal_windows(
    data = dat,
    signal = "heart_rate",
    group_columns = group_columns,
    value_column = value_column,
    validity_column = validity_column,
    exclude_zero = exclude_zero
  )
}


#' Summarise Gazepoint engagement-dial windows
#'
#' Summarises Gazepoint engagement-dial values within participant, trial,
#' stimulus, AOI, or other user-defined windows.
#'
#' @param data A data frame or a path to a Gazepoint CSV export.
#' @param group_columns Optional grouping columns defining windows, such as
#'   `c("USER", "MEDIA_ID")`.
#' @param value_column Engagement-dial value column. Defaults to `"DIAL"`.
#' @param validity_column Engagement-dial validity column. Defaults to
#'   `"DIALV"`.
#' @param exclude_zero Should zero values be excluded from usable summaries?
#'
#' @return A data frame with one row per window.
#'
#' @export
summarise_gazepoint_engagement_windows <- function(data,
                                                   group_columns = NULL,
                                                   value_column = "DIAL",
                                                   validity_column = "DIALV",
                                                   exclude_zero = FALSE) {
  dat <- coerce_gazepoint_biometrics_data(data)

  summarise_gazepoint_signal_windows(
    data = dat,
    signal = "engagement_dial",
    group_columns = group_columns,
    value_column = value_column,
    validity_column = validity_column,
    exclude_zero = exclude_zero
  )
}


#' Summarise Gazepoint multimodal biometric windows
#'
#' Creates a combined window-level summary table for GSR/EDA, heart rate, and
#' engagement dial. The output is suitable for later merging with eye-tracking
#' summaries from `gp3tools`.
#'
#' @param data A data frame or a path to a Gazepoint CSV export.
#' @param group_columns Optional grouping columns defining windows, such as
#'   `c("USER", "MEDIA_ID")`.
#' @param exclude_zero Should zero values be excluded from GSR and heart-rate
#'   summaries?
#'
#' @return A data frame with one row per window and prefixed biometric summary
#'   columns.
#'
#' @export
summarise_gazepoint_multimodal_windows <- function(data,
                                                   group_columns = NULL,
                                                   exclude_zero = TRUE) {
  dat <- coerce_gazepoint_biometrics_data(data)

  gsr <- summarise_gazepoint_gsr_windows(
    dat,
    group_columns = group_columns,
    exclude_zero = exclude_zero
  )

  hr <- summarise_gazepoint_hr_windows(
    dat,
    group_columns = group_columns,
    exclude_zero = exclude_zero
  )

  dial <- summarise_gazepoint_engagement_windows(
    dat,
    group_columns = group_columns,
    exclude_zero = FALSE
  )

  gsr <- prefix_signal_summary(gsr, prefix = "gsr")
  hr <- prefix_signal_summary(hr, prefix = "hr")
  dial <- prefix_signal_summary(dial, prefix = "dial")

  out <- merge_signal_summaries(
    summaries = list(gsr, hr, dial),
    group_columns = group_columns
  )

  rownames(out) <- NULL
  out
}


summarise_gazepoint_signal_windows <- function(data,
                                               signal,
                                               group_columns = NULL,
                                               value_column,
                                               validity_column = NULL,
                                               exclude_zero = TRUE) {
  if (is.null(value_column) || length(value_column) != 1L || is.na(value_column)) {
    stop("`value_column` could not be determined.", call. = FALSE)
  }

  if (!value_column %in% names(data)) {
    stop("`value_column` was not found in `data`: ", value_column, call. = FALSE)
  }

  if (!is.null(group_columns)) {
    missing_groups <- setdiff(group_columns, names(data))

    if (length(missing_groups) > 0L) {
      stop(
        "`group_columns` were not found in `data`: ",
        paste(missing_groups, collapse = ", "),
        call. = FALSE
      )
    }
  }

  if (is.null(group_columns) || length(group_columns) == 0L) {
    group_keys <- rep("all", nrow(data))
    group_frame <- data.frame(.window = "all", stringsAsFactors = FALSE)
    group_columns_out <- ".window"
  } else {
    group_keys <- make_group_key(data, group_columns)
    group_frame <- unique(data[group_columns])
    group_frame$.group_key <- make_group_key(group_frame, group_columns)
    group_columns_out <- group_columns
  }

  unique_keys <- unique(group_keys)

  summaries <- lapply(unique_keys, function(key) {
    rows <- group_keys == key

    group_values <- if (is.null(group_columns) || length(group_columns) == 0L) {
      data.frame(.window = "all", stringsAsFactors = FALSE)
    } else {
      data[which(rows)[1], group_columns, drop = FALSE]
    }

    x <- as_numeric_safe(data[[value_column]][rows])

    valid <- !is.na(x)

    if (isTRUE(exclude_zero)) {
      valid <- valid & x != 0
    }

    validity_present <- !is.null(validity_column) &&
      length(validity_column) == 1L &&
      validity_column %in% names(data)

    if (validity_present) {
      v <- as_numeric_safe(data[[validity_column]][rows])
      valid <- valid & !is.na(v) & v > 0
    }

    usable_x <- x[valid]

    out <- data.frame(
      signal = signal,
      value_column = value_column,
      validity_column = ifelse(validity_present, validity_column, NA_character_),
      n_rows = length(x),
      usable_rows = length(usable_x),
      usable_pct = safe_pct(length(usable_x), length(x)),
      missing_rows = sum(is.na(x)),
      zero_rows = sum(!is.na(x) & x == 0),
      mean_value = ifelse(length(usable_x) > 0L, mean(usable_x), NA_real_),
      median_value = ifelse(length(usable_x) > 0L, stats::median(usable_x), NA_real_),
      sd_value = ifelse(length(usable_x) > 1L, stats::sd(usable_x), NA_real_),
      min_value = ifelse(length(usable_x) > 0L, min(usable_x), NA_real_),
      max_value = ifelse(length(usable_x) > 0L, max(usable_x), NA_real_),
      first_value = ifelse(length(usable_x) > 0L, usable_x[1], NA_real_),
      last_value = ifelse(length(usable_x) > 0L, usable_x[length(usable_x)], NA_real_),
      change_value = ifelse(length(usable_x) > 0L, usable_x[length(usable_x)] - usable_x[1], NA_real_),
      stringsAsFactors = FALSE
    )

    cbind(group_values, out, stringsAsFactors = FALSE)
  })

  out <- do.call(rbind, summaries)
  rownames(out) <- NULL

  if (identical(group_columns_out, ".window")) {
    names(out)[names(out) == ".window"] <- "window"
  }

  out
}


prefix_signal_summary <- function(summary, prefix) {
  metadata_columns <- c(
    "signal",
    "value_column",
    "validity_column",
    "n_rows",
    "usable_rows",
    "usable_pct",
    "missing_rows",
    "zero_rows",
    "mean_value",
    "median_value",
    "sd_value",
    "min_value",
    "max_value",
    "first_value",
    "last_value",
    "change_value"
  )

  columns_to_prefix <- intersect(metadata_columns, names(summary))
  names(summary)[match(columns_to_prefix, names(summary))] <-
    paste(prefix, columns_to_prefix, sep = "_")

  summary
}


merge_signal_summaries <- function(summaries, group_columns = NULL) {
  if (length(summaries) == 0L) {
    return(data.frame())
  }

  if (is.null(group_columns) || length(group_columns) == 0L) {
    group_columns <- "window"
  }

  out <- summaries[[1]]

  if (length(summaries) == 1L) {
    return(out)
  }

  for (i in 2:length(summaries)) {
    out <- merge(
      out,
      summaries[[i]],
      by = group_columns,
      all = TRUE,
      sort = FALSE
    )
  }

  out
}

test_that("summarise_gazepoint_gsr_windows summarises grouped GSR data", {
  dat <- data.frame(
    USER = c("U1", "U1", "U2", "U2"),
    MEDIA_ID = c(1, 1, 1, 1),
    GSR_US = c(2.0, 2.4, 1.0, 1.2),
    GSRV = c(1, 1, 1, 1)
  )

  out <- summarise_gazepoint_gsr_windows(
    dat,
    group_columns = c("USER", "MEDIA_ID")
  )

  expect_equal(nrow(out), 2)
  expect_true("mean_value" %in% names(out))

  u1 <- out[out$USER == "U1", ]

  expect_equal(u1$usable_rows, 2)
  expect_equal(u1$mean_value, 2.2, tolerance = 1e-8)
  expect_equal(u1$change_value, 0.4, tolerance = 1e-8)
})


test_that("summarise_gazepoint_hr_windows excludes invalid and zero values", {
  dat <- data.frame(
    USER = c("U1", "U1", "U1", "U1"),
    HR = c(70, 72, 0, 90),
    HRV = c(1, 1, 0, 0)
  )

  out <- summarise_gazepoint_hr_windows(
    dat,
    group_columns = "USER"
  )

  expect_equal(out$usable_rows, 2)
  expect_equal(out$mean_value, 71, tolerance = 1e-8)
  expect_equal(out$zero_rows, 1)
})


test_that("summarise_gazepoint_engagement_windows keeps valid zero dial values", {
  dat <- data.frame(
    USER = c("U1", "U1", "U1"),
    DIAL = c(0, 0.5, 1),
    DIALV = c(1, 1, 1)
  )

  out <- summarise_gazepoint_engagement_windows(
    dat,
    group_columns = "USER"
  )

  expect_equal(out$usable_rows, 3)
  expect_equal(out$mean_value, 0.5, tolerance = 1e-8)
})


test_that("summarise_gazepoint_signal_windows supports ungrouped summaries", {
  dat <- data.frame(
    GSR_US = c(2.0, 2.2, 2.4),
    GSRV = c(1, 1, 1)
  )

  out <- summarise_gazepoint_gsr_windows(dat)

  expect_equal(nrow(out), 1)
  expect_equal(out$window, "all")
  expect_equal(out$mean_value, 2.2, tolerance = 1e-8)
})


test_that("summarise_gazepoint_multimodal_windows combines signal summaries", {
  dat <- data.frame(
    USER = c("U1", "U1", "U2", "U2"),
    MEDIA_ID = c(1, 1, 1, 1),
    GSR_US = c(2.0, 2.4, 1.0, 1.2),
    GSRV = c(1, 1, 1, 1),
    HR = c(70, 72, 80, 82),
    HRV = c(1, 1, 1, 1),
    DIAL = c(0.1, 0.2, 0.3, 0.4),
    DIALV = c(1, 1, 1, 1)
  )

  out <- summarise_gazepoint_multimodal_windows(
    dat,
    group_columns = c("USER", "MEDIA_ID")
  )

  expect_equal(nrow(out), 2)
  expect_true("gsr_mean_value" %in% names(out))
  expect_true("hr_mean_value" %in% names(out))
  expect_true("dial_mean_value" %in% names(out))

  u1 <- out[out$USER == "U1", ]

  expect_equal(u1$gsr_mean_value, 2.2, tolerance = 1e-8)
  expect_equal(u1$hr_mean_value, 71, tolerance = 1e-8)
  expect_equal(u1$dial_mean_value, 0.15, tolerance = 1e-8)
})


test_that("window summaries reject missing grouping columns", {
  dat <- data.frame(
    GSR_US = c(2.0, 2.2),
    GSRV = c(1, 1)
  )

  expect_error(
    summarise_gazepoint_gsr_windows(dat, group_columns = "USER"),
    "group_columns"
  )
})

#' Prepare Gazepoint multimodal model data
#'
#' Creates a model-ready table from Gazepoint biometric window summaries and,
#' optionally, eye-tracking summaries produced by `gp3tools` or another
#' workflow. The function is intentionally conservative: it does not fit a
#' model, impute missing values, or remove rows automatically.
#'
#' @param biometrics A data frame containing row-level Gazepoint Biometrics data
#'   or an already summarised biometric window table.
#' @param eye_tracking Optional eye-tracking summary table to merge with the
#'   biometric summaries.
#' @param group_columns Columns defining the analysis unit, such as
#'   `c("USER", "MEDIA_ID")`.
#' @param biometric_is_summarised Logical. If `FALSE`, biometric window
#'   summaries are created using [summarise_gazepoint_multimodal_windows()].
#'   If `TRUE`, `biometrics` is treated as already summarised.
#' @param by Optional merge keys. If `NULL`, `group_columns` are used.
#' @param all Should a full outer join be used when eye-tracking data are
#'   supplied? Defaults to `FALSE`, giving an inner join.
#'
#' @return A data frame with class `"gazepoint_multimodal_model_data"` and a
#'   `"model_data_summary"` attribute.
#'
#' @export
prepare_gazepoint_multimodal_model_data <- function(biometrics,
                                                    eye_tracking = NULL,
                                                    group_columns = NULL,
                                                    biometric_is_summarised = FALSE,
                                                    by = NULL,
                                                    all = FALSE) {
  if (!is.data.frame(biometrics)) {
    biometrics <- coerce_gazepoint_biometrics_data(biometrics)
  }

  if (is.null(group_columns) || length(group_columns) == 0L) {
    stop("`group_columns` must define the model-data analysis unit.", call. = FALSE)
  }

  if (!all(group_columns %in% names(biometrics))) {
    missing_groups <- setdiff(group_columns, names(biometrics))
    stop(
      "`group_columns` were not found in `biometrics`: ",
      paste(missing_groups, collapse = ", "),
      call. = FALSE
    )
  }

  if (isTRUE(biometric_is_summarised)) {
    biometric_summary <- biometrics
  } else {
    biometric_summary <- summarise_gazepoint_multimodal_windows(
      biometrics,
      group_columns = group_columns
    )
  }

  if (is.null(eye_tracking)) {
    out <- biometric_summary
    source <- "biometrics_only"
  } else {
    if (!is.data.frame(eye_tracking)) {
      stop("`eye_tracking` must be a data frame when supplied.", call. = FALSE)
    }

    if (is.null(by)) {
      by <- group_columns
    }

    missing_bio_keys <- setdiff(by, names(biometric_summary))
    missing_eye_keys <- setdiff(by, names(eye_tracking))

    if (length(missing_bio_keys) > 0L) {
      stop(
        "`by` columns were not found in biometric summaries: ",
        paste(missing_bio_keys, collapse = ", "),
        call. = FALSE
      )
    }

    if (length(missing_eye_keys) > 0L) {
      stop(
        "`by` columns were not found in `eye_tracking`: ",
        paste(missing_eye_keys, collapse = ", "),
        call. = FALSE
      )
    }

    out <- merge(
      eye_tracking,
      biometric_summary,
      by = by,
      all = all,
      sort = FALSE
    )

    source <- "eye_tracking_plus_biometrics"
  }

  attr(out, "model_data_summary") <- data.frame(
    source = source,
    n_rows = nrow(out),
    n_columns = ncol(out),
    group_columns = paste(group_columns, collapse = ","),
    has_eye_tracking = !is.null(eye_tracking),
    stringsAsFactors = FALSE
  )

  class(out) <- c("gazepoint_multimodal_model_data", class(out))

  out
}

test_that("prepare_gazepoint_multimodal_model_data summarises biometrics only", {
  biometrics <- data.frame(
    USER = c("U1", "U1", "U2", "U2"),
    MEDIA_ID = c(1, 1, 1, 1),
    GSR_US = c(2.0, 2.4, 1.0, 1.2),
    GSRV = c(1, 1, 1, 1),
    HR = c(70, 72, 80, 82),
    HRV = c(1, 1, 1, 1),
    DIAL = c(0.1, 0.2, 0.3, 0.4),
    DIALV = c(1, 1, 1, 1)
  )

  out <- prepare_gazepoint_multimodal_model_data(
    biometrics = biometrics,
    group_columns = c("USER", "MEDIA_ID")
  )

  expect_s3_class(out, "gazepoint_multimodal_model_data")
  expect_equal(nrow(out), 2)
  expect_true("gsr_mean_value" %in% names(out))
  expect_true("hr_mean_value" %in% names(out))
  expect_true("dial_mean_value" %in% names(out))

  summary <- attr(out, "model_data_summary")
  expect_true(is.data.frame(summary))
  expect_equal(summary$source, "biometrics_only")
})


test_that("prepare_gazepoint_multimodal_model_data merges eye-tracking summaries", {
  biometrics <- data.frame(
    USER = c("U1", "U1", "U2", "U2"),
    MEDIA_ID = c(1, 1, 1, 1),
    GSR_US = c(2.0, 2.4, 1.0, 1.2),
    GSRV = c(1, 1, 1, 1),
    HR = c(70, 72, 80, 82),
    HRV = c(1, 1, 1, 1),
    DIAL = c(0.1, 0.2, 0.3, 0.4),
    DIALV = c(1, 1, 1, 1)
  )

  eye_tracking <- data.frame(
    USER = c("U1", "U2"),
    MEDIA_ID = c(1, 1),
    dwell_time = c(1200, 900),
    fixation_count = c(8, 6)
  )

  out <- prepare_gazepoint_multimodal_model_data(
    biometrics = biometrics,
    eye_tracking = eye_tracking,
    group_columns = c("USER", "MEDIA_ID")
  )

  expect_equal(nrow(out), 2)
  expect_true("dwell_time" %in% names(out))
  expect_true("fixation_count" %in% names(out))
  expect_true("gsr_mean_value" %in% names(out))

  summary <- attr(out, "model_data_summary")
  expect_equal(summary$source, "eye_tracking_plus_biometrics")
  expect_true(summary$has_eye_tracking)
})


test_that("prepare_gazepoint_multimodal_model_data accepts already summarised biometrics", {
  biometric_summary <- data.frame(
    USER = c("U1", "U2"),
    MEDIA_ID = c(1, 1),
    gsr_mean_value = c(2.2, 1.1),
    hr_mean_value = c(71, 81),
    dial_mean_value = c(0.15, 0.35)
  )

  out <- prepare_gazepoint_multimodal_model_data(
    biometrics = biometric_summary,
    group_columns = c("USER", "MEDIA_ID"),
    biometric_is_summarised = TRUE
  )

  expect_equal(nrow(out), 2)
  expect_equal(out$gsr_mean_value[1], 2.2)
})


test_that("prepare_gazepoint_multimodal_model_data rejects missing group columns", {
  biometrics <- data.frame(
    GSR_US = c(2.0, 2.2),
    GSRV = c(1, 1)
  )

  expect_error(
    prepare_gazepoint_multimodal_model_data(
      biometrics = biometrics,
      group_columns = "USER"
    ),
    "group_columns"
  )
})


test_that("prepare_gazepoint_multimodal_model_data rejects missing eye-tracking merge keys", {
  biometrics <- data.frame(
    USER = c("U1", "U1"),
    MEDIA_ID = c(1, 1),
    GSR_US = c(2.0, 2.2),
    GSRV = c(1, 1),
    HR = c(70, 71),
    HRV = c(1, 1),
    DIAL = c(0.1, 0.2),
    DIALV = c(1, 1)
  )

  eye_tracking <- data.frame(
    USER = "U1",
    dwell_time = 1200
  )

  expect_error(
    prepare_gazepoint_multimodal_model_data(
      biometrics = biometrics,
      eye_tracking = eye_tracking,
      group_columns = c("USER", "MEDIA_ID")
    ),
    "eye_tracking"
  )
})

#' Decompose Gazepoint GSR/EDA into tonic and phasic components
#'
#' Creates descriptive tonic and phasic EDA columns from Gazepoint GSR/EDA data.
#' If vendor-provided tonic/phasic columns such as `GSR_US_TONIC` and
#' `GSR_US_PHASIC` are available, they are used by default. Otherwise, a simple
#' rolling-median tonic estimate is used and the phasic component is calculated
#' as signal minus tonic. This helper is intentionally conservative and does not
#' replace specialised biosignal-processing software.
#'
#' @param data A data frame.
#' @param signal_col Optional GSR/EDA signal column. If `NULL`, a likely
#'   conductance-like column is detected.
#' @param tonic_col Optional existing tonic column.
#' @param phasic_col Optional existing phasic column.
#' @param time_col Optional time/order column.
#' @param group_cols Optional grouping columns.
#' @param window_size Rolling-median window size used when existing tonic/phasic
#'   columns are not available. Even values are increased by one.
#' @param output_prefix Prefix for output columns.
#' @param overwrite Logical. Should existing output columns be overwritten?
#'
#' @return A data frame with added tonic, phasic, and method columns. Attributes
#'   include `overview` and `settings`.
#'
#' @examples
#' df <- data.frame(CNT = 1:10, GSR_US = seq(1, 2, length.out = 10))
#' out <- decompose_gazepoint_eda(df, signal_col = "GSR_US", window_size = 3)
#' names(out)
#'
#' @export
decompose_gazepoint_eda <- function(data,
                                    signal_col = NULL,
                                    tonic_col = NULL,
                                    phasic_col = NULL,
                                    time_col = NULL,
                                    group_cols = NULL,
                                    window_size = 31L,
                                    output_prefix = "eda",
                                    overwrite = FALSE) {
  if (!is.data.frame(data)) {
    stop("`data` must be a data frame.", call. = FALSE)
  }

  if (!is.null(signal_col)) {
    .gpbiom_assert_columns(data, signal_col, "signal_col")
    if (length(signal_col) != 1L) {
      stop("`signal_col` must be NULL or a single column name.", call. = FALSE)
    }
  }

  if (!is.null(tonic_col)) {
    .gpbiom_assert_columns(data, tonic_col, "tonic_col")
    if (length(tonic_col) != 1L) {
      stop("`tonic_col` must be NULL or a single column name.", call. = FALSE)
    }
  }

  if (!is.null(phasic_col)) {
    .gpbiom_assert_columns(data, phasic_col, "phasic_col")
    if (length(phasic_col) != 1L) {
      stop("`phasic_col` must be NULL or a single column name.", call. = FALSE)
    }
  }

  if (!is.null(time_col)) {
    .gpbiom_assert_columns(data, time_col, "time_col")
    if (length(time_col) != 1L) {
      stop("`time_col` must be NULL or a single column name.", call. = FALSE)
    }
  }

  if (!is.null(group_cols)) {
    .gpbiom_assert_columns(data, group_cols, "group_cols")
  }

  .gpbiom_assert_positive_integer(window_size, "window_size")

  if (!is.character(output_prefix) ||
      length(output_prefix) != 1L ||
      is.na(output_prefix) ||
      !nzchar(output_prefix)) {
    stop("`output_prefix` must be a non-empty character string.",
         call. = FALSE)
  }

  if (!is.logical(overwrite) || length(overwrite) != 1L || is.na(overwrite)) {
    stop("`overwrite` must be TRUE or FALSE.", call. = FALSE)
  }

  if (is.null(signal_col)) {
    signal_col <- .gpbiom_choose_eda_signal_column(data)
  }

  if (is.na(signal_col) || !signal_col %in% names(data)) {
    stop("No GSR/EDA signal column was detected. Provide `signal_col` explicitly.",
         call. = FALSE)
  }

  if (!is.numeric(data[[signal_col]])) {
    stop("`signal_col` must be numeric.", call. = FALSE)
  }

  existing <- .gpbiom_detect_existing_eda_components(data, tonic_col, phasic_col)

  tonic_out <- paste0(output_prefix, "_tonic")
  phasic_out <- paste0(output_prefix, "_phasic")
  method_out <- paste0(output_prefix, "_decomposition_method")

  new_cols <- c(tonic_out, phasic_out, method_out)
  conflicts <- new_cols[new_cols %in% names(data)]

  if (length(conflicts) > 0L && !isTRUE(overwrite)) {
    stop(
      "Output columns already exist: ",
      paste(conflicts, collapse = ", "),
      ". Use `overwrite = TRUE` to replace them.",
      call. = FALSE
    )
  }

  out <- data
  out[[tonic_out]] <- NA_real_
  out[[phasic_out]] <- NA_real_
  out[[method_out]] <- NA_character_

  groups <- .gpbiom_eda_group_indices(out, group_cols)

  if (!is.na(existing$tonic_col) && !is.na(existing$phasic_col)) {
    out[[tonic_out]] <- as.numeric(out[[existing$tonic_col]])
    out[[phasic_out]] <- as.numeric(out[[existing$phasic_col]])
    out[[method_out]] <- "existing_tonic_phasic_columns"
    method <- "existing_tonic_phasic_columns"
  } else {
    window_size <- as.integer(window_size)
    if (window_size %% 2L == 0L) {
      window_size <- window_size + 1L
    }

    for (group_name in names(groups)) {
      idx <- groups[[group_name]]
      idx <- .gpbiom_eda_order_index(out, idx, time_col)

      signal <- as.numeric(out[[signal_col]][idx])
      tonic <- .gpbiom_eda_rolling_median(signal, window_size = window_size)

      out[[tonic_out]][idx] <- tonic
      out[[phasic_out]][idx] <- signal - tonic
      out[[method_out]][idx] <- "rolling_median_residual"
    }

    method <- "rolling_median_residual"
  }

  overview <- data.frame(
    n_rows = nrow(out),
    signal_col = signal_col,
    tonic_col = tonic_out,
    phasic_col = phasic_out,
    method = method,
    used_existing_components = !is.na(existing$tonic_col) &&
      !is.na(existing$phasic_col),
    group_count = length(groups),
    n_tonic_non_missing = sum(!is.na(out[[tonic_out]])),
    n_phasic_non_missing = sum(!is.na(out[[phasic_out]])),
    status = "eda_decomposition_created",
    stringsAsFactors = FALSE
  )

  attr(out, "overview") <- overview
  attr(out, "settings") <- list(
    signal_col = signal_col,
    input_tonic_col = existing$tonic_col,
    input_phasic_col = existing$phasic_col,
    time_col = time_col,
    group_cols = group_cols,
    window_size = as.integer(window_size),
    output_prefix = output_prefix,
    note = paste0(
      "EDA decomposition is descriptive. Use specialised biosignal software ",
      "for confirmatory SCR/EDA decomposition when required."
    )
  )

  out
}


#' Detect SCR-like events in Gazepoint GSR/EDA data
#'
#' Detects simple SCR-like peaks from a phasic EDA signal. If a phasic column is
#' not supplied, the function first creates a descriptive phasic component using
#' `decompose_gazepoint_eda()`. This helper is intended for exploratory quality
#' control and descriptive summaries, not as a replacement for specialised SCR
#' detection pipelines.
#'
#' @param data A data frame.
#' @param phasic_col Optional phasic EDA column.
#' @param signal_col Optional raw/conductance EDA column used when `phasic_col`
#'   is not supplied.
#' @param time_col Optional time/order column.
#' @param group_cols Optional grouping columns.
#' @param threshold Optional numeric detection threshold. If `NULL`, a robust
#'   group-specific threshold is estimated as median plus three MADs, bounded
#'   below by zero.
#' @param min_peak_distance Minimum distance between retained peaks in samples.
#' @param window_size Rolling-median window size used if decomposition is needed.
#'
#' @return A list with `overview`, `events`, `group_summary`, and `settings`.
#'
#' @examples
#' df <- data.frame(
#'   CNT = 1:20,
#'   GSR_US_PHASIC = c(rep(0, 5), 0.2, 0.8, 0.2, rep(0, 12))
#' )
#' detect_gazepoint_scr_events(df, phasic_col = "GSR_US_PHASIC", time_col = "CNT")
#'
#' @export
detect_gazepoint_scr_events <- function(data,
                                        phasic_col = NULL,
                                        signal_col = NULL,
                                        time_col = NULL,
                                        group_cols = NULL,
                                        threshold = NULL,
                                        min_peak_distance = 10L,
                                        window_size = 31L) {
  if (!is.data.frame(data)) {
    stop("`data` must be a data frame.", call. = FALSE)
  }

  if (!is.null(phasic_col)) {
    .gpbiom_assert_columns(data, phasic_col, "phasic_col")
    if (length(phasic_col) != 1L) {
      stop("`phasic_col` must be NULL or a single column name.", call. = FALSE)
    }
  }

  if (!is.null(signal_col)) {
    .gpbiom_assert_columns(data, signal_col, "signal_col")
    if (length(signal_col) != 1L) {
      stop("`signal_col` must be NULL or a single column name.", call. = FALSE)
    }
  }

  if (!is.null(time_col)) {
    .gpbiom_assert_columns(data, time_col, "time_col")
    if (length(time_col) != 1L) {
      stop("`time_col` must be NULL or a single column name.", call. = FALSE)
    }
  }

  if (!is.null(group_cols)) {
    .gpbiom_assert_columns(data, group_cols, "group_cols")
  }

  if (!is.null(threshold) &&
      (!is.numeric(threshold) ||
       length(threshold) != 1L ||
       is.na(threshold))) {
    stop("`threshold` must be NULL or a single numeric value.", call. = FALSE)
  }

  .gpbiom_assert_positive_integer(min_peak_distance, "min_peak_distance")
  .gpbiom_assert_positive_integer(window_size, "window_size")

  working <- data
  decomposition_used <- FALSE

  if (is.null(phasic_col)) {
    existing <- .gpbiom_detect_existing_eda_components(working, NULL, NULL)

    if (!is.na(existing$phasic_col)) {
      phasic_col <- existing$phasic_col
    } else {
      working <- decompose_gazepoint_eda(
        working,
        signal_col = signal_col,
        time_col = time_col,
        group_cols = group_cols,
        window_size = window_size,
        output_prefix = "scr_detection_eda",
        overwrite = TRUE
      )
      phasic_col <- "scr_detection_eda_phasic"
      decomposition_used <- TRUE
    }
  }

  if (!is.numeric(working[[phasic_col]])) {
    stop("`phasic_col` must be numeric.", call. = FALSE)
  }

  groups <- .gpbiom_eda_group_indices(working, group_cols)

  event_rows <- list()
  summary_rows <- list()
  event_id <- 0L

  for (group_name in names(groups)) {
    idx <- groups[[group_name]]
    idx <- .gpbiom_eda_order_index(working, idx, time_col)

    values <- as.numeric(working[[phasic_col]][idx])
    local_threshold <- if (is.null(threshold)) {
      .gpbiom_scr_default_threshold(values)
    } else {
      threshold
    }

    peaks <- .gpbiom_scr_peak_indices(
      values = values,
      threshold = local_threshold,
      min_peak_distance = min_peak_distance
    )

    if (length(peaks) > 0L) {
      for (peak in peaks) {
        event_id <- event_id + 1L
        row_index <- idx[peak]

        event_rows[[length(event_rows) + 1L]] <- data.frame(
          event_id = event_id,
          group = group_name,
          row_index = row_index,
          time = if (is.null(time_col)) NA_real_ else working[[time_col]][row_index],
          peak_value = values[peak],
          threshold = local_threshold,
          phasic_col = phasic_col,
          detection_method = "local_peak_above_threshold",
          stringsAsFactors = FALSE
        )
      }
    }

    summary_rows[[length(summary_rows) + 1L]] <- data.frame(
      group = group_name,
      n_samples = length(values),
      threshold = local_threshold,
      n_events = length(peaks),
      event_rate_per_1000_samples = if (length(values) > 0L) {
        length(peaks) / length(values) * 1000
      } else {
        NA_real_
      },
      stringsAsFactors = FALSE
    )
  }

  events <- if (length(event_rows) == 0L) {
    .gpbiom_empty_scr_events()
  } else {
    do.call(rbind, event_rows)
  }

  group_summary <- if (length(summary_rows) == 0L) {
    data.frame(
      group = character(),
      n_samples = integer(),
      threshold = numeric(),
      n_events = integer(),
      event_rate_per_1000_samples = numeric(),
      stringsAsFactors = FALSE
    )
  } else {
    do.call(rbind, summary_rows)
  }

  overview <- data.frame(
    n_rows = nrow(data),
    group_count = length(groups),
    phasic_col = phasic_col,
    decomposition_used = decomposition_used,
    threshold = if (is.null(threshold)) NA_real_ else threshold,
    min_peak_distance = as.integer(min_peak_distance),
    n_events = nrow(events),
    status = if (nrow(events) > 0L) {
      "scr_events_detected"
    } else {
      "no_scr_events_detected"
    },
    stringsAsFactors = FALSE
  )

  out <- list(
    overview = overview,
    events = events,
    group_summary = group_summary,
    settings = list(
      phasic_col = phasic_col,
      signal_col = signal_col,
      time_col = time_col,
      group_cols = group_cols,
      threshold = threshold,
      min_peak_distance = as.integer(min_peak_distance),
      window_size = as.integer(window_size),
      note = paste0(
        "SCR events are simple SCR-like local peaks. Use specialised ",
        "biosignal software for confirmatory SCR event detection."
      )
    )
  )

  class(out) <- c("gazepoint_scr_events", class(out))
  out
}


.gpbiom_choose_eda_signal_column <- function(data) {
  preferred <- c(
    "GSR_US", "EDA_US", "GSR", "EDA",
    "GSR_OHMS", "SKIN_CONDUCTANCE"
  )

  existing <- preferred[preferred %in% names(data)]

  numeric_existing <- existing[vapply(existing, function(column) {
    is.numeric(data[[column]])
  }, logical(1))]

  if (length(numeric_existing) > 0L) {
    return(numeric_existing[1L])
  }

  gsr_like <- names(data)[grepl("GSR|EDA|CONDUCTANCE", names(data), ignore.case = TRUE)]

  numeric_gsr_like <- gsr_like[vapply(gsr_like, function(column) {
    is.numeric(data[[column]])
  }, logical(1))]

  if (length(numeric_gsr_like) > 0L) {
    return(numeric_gsr_like[1L])
  }

  NA_character_
}


.gpbiom_detect_existing_eda_components <- function(data,
                                                   tonic_col = NULL,
                                                   phasic_col = NULL) {
  if (is.null(tonic_col)) {
    tonic_candidates <- c(
      "GSR_US_TONIC", "EDA_TONIC", "GSR_TONIC",
      "TONIC", "SCL"
    )
    tonic_col <- tonic_candidates[tonic_candidates %in% names(data)][1L]
  }

  if (is.null(phasic_col)) {
    phasic_candidates <- c(
      "GSR_US_PHASIC", "EDA_PHASIC", "GSR_PHASIC",
      "PHASIC", "SCR"
    )
    phasic_col <- phasic_candidates[phasic_candidates %in% names(data)][1L]
  }

  if (length(tonic_col) == 0L || is.na(tonic_col)) {
    tonic_col <- NA_character_
  }

  if (length(phasic_col) == 0L || is.na(phasic_col)) {
    phasic_col <- NA_character_
  }

  list(tonic_col = tonic_col, phasic_col = phasic_col)
}


.gpbiom_eda_group_indices <- function(data, group_cols = NULL) {
  if (is.null(group_cols) || length(group_cols) == 0L) {
    return(list(all = seq_len(nrow(data))))
  }

  grouping <- interaction(data[group_cols], drop = TRUE, sep = " | ")
  split(seq_len(nrow(data)), grouping)
}


.gpbiom_eda_order_index <- function(data, idx, time_col = NULL) {
  if (is.null(time_col)) {
    return(idx)
  }

  idx[order(data[[time_col]][idx], na.last = TRUE)]
}


.gpbiom_scr_default_threshold <- function(values) {
  finite <- values[is.finite(values)]

  if (length(finite) == 0L) {
    return(Inf)
  }

  center <- stats::median(finite)
  spread <- stats::mad(finite, constant = 1.4826)

  if (!is.finite(spread) || spread == 0) {
    spread <- stats::sd(finite, na.rm = TRUE)
  }

  if (!is.finite(spread) || spread == 0) {
    return(max(0, center))
  }

  max(0, center + 3 * spread)
}


.gpbiom_scr_peak_indices <- function(values,
                                     threshold,
                                     min_peak_distance = 10L) {
  if (length(values) < 3L) {
    return(integer())
  }

  candidate <- which(
    is.finite(values) &
      values >= threshold &
      c(FALSE, values[-c(1L, length(values))] >= values[-c(length(values) - 1L, length(values))], FALSE) &
      c(FALSE, values[-c(1L, length(values))] >= values[-c(1L, 2L)], FALSE)
  )

  candidate <- candidate[candidate > 1L & candidate < length(values)]

  if (length(candidate) == 0L) {
    return(integer())
  }

  candidate <- candidate[order(values[candidate], decreasing = TRUE)]

  kept <- integer()

  for (idx in candidate) {
    if (length(kept) == 0L ||
        all(abs(idx - kept) >= min_peak_distance)) {
      kept <- c(kept, idx)
    }
  }

  sort(kept)
}

.gpbiom_eda_rolling_median <- function(values, window_size = 31L) {
  values <- as.numeric(values)

  if (length(values) == 0L) {
    return(numeric())
  }

  window_size <- as.integer(window_size)

  if (is.na(window_size) || window_size < 1L) {
    stop("`window_size` must be a positive integer.", call. = FALSE)
  }

  if (window_size %% 2L == 0L) {
    window_size <- window_size + 1L
  }

  half_window <- floor(window_size / 2L)
  out <- rep(NA_real_, length(values))

  for (i in seq_along(values)) {
    from <- max(1L, i - half_window)
    to <- min(length(values), i + half_window)
    window_values <- values[from:to]
    window_values <- window_values[is.finite(window_values)]

    if (length(window_values) > 0L) {
      out[i] <- stats::median(window_values)
    }
  }

  out
}

.gpbiom_empty_scr_events <- function() {
  data.frame(
    event_id = integer(),
    group = character(),
    row_index = integer(),
    time = numeric(),
    peak_value = numeric(),
    threshold = numeric(),
    phasic_col = character(),
    detection_method = character(),
    stringsAsFactors = FALSE
  )
}

test_that("decompose_gazepoint_eda uses existing tonic and phasic columns", {
  df <- data.frame(
    CNT = 1:5,
    GSR_US = c(1, 1.1, 1.2, 1.1, 1),
    GSR_US_TONIC = c(1, 1, 1, 1, 1),
    GSR_US_PHASIC = c(0, 0.1, 0.2, 0.1, 0)
  )

  out <- decompose_gazepoint_eda(df, signal_col = "GSR_US")

  expect_s3_class(out, "data.frame")
  expect_true(all(c("eda_tonic", "eda_phasic", "eda_decomposition_method") %in% names(out)))
  expect_equal(out$eda_tonic, df$GSR_US_TONIC)
  expect_equal(out$eda_phasic, df$GSR_US_PHASIC)
  expect_equal(attr(out, "overview")$method, "existing_tonic_phasic_columns")
})


test_that("decompose_gazepoint_eda creates rolling median residual components", {
  df <- data.frame(
    CNT = 1:7,
    GSR_US = c(1, 1, 1, 2, 1, 1, 1)
  )

  out <- decompose_gazepoint_eda(
    df,
    signal_col = "GSR_US",
    time_col = "CNT",
    window_size = 3,
    output_prefix = "test_eda"
  )

  expect_true(all(c("test_eda_tonic", "test_eda_phasic") %in% names(out)))
  expect_equal(attr(out, "overview")$method, "rolling_median_residual")
  expect_equal(nrow(out), 7)
  expect_true(any(out$test_eda_phasic != 0, na.rm = TRUE))
})


test_that("decompose_gazepoint_eda supports grouping", {
  df <- data.frame(
    id = rep(c("P1", "P2"), each = 5),
    CNT = rep(1:5, 2),
    GSR_US = c(1, 1, 2, 1, 1, 2, 2, 3, 2, 2)
  )

  out <- decompose_gazepoint_eda(
    df,
    signal_col = "GSR_US",
    time_col = "CNT",
    group_cols = "id",
    window_size = 3
  )

  expect_equal(attr(out, "overview")$group_count, 2)
  expect_equal(nrow(out), nrow(df))
})


test_that("decompose_gazepoint_eda validates arguments", {
  df <- data.frame(GSR_US = c(1, 2, 3))

  expect_error(
    decompose_gazepoint_eda(1:3),
    "`data` must be"
  )

  expect_error(
    decompose_gazepoint_eda(df, signal_col = "missing"),
    "not found"
  )

  expect_error(
    decompose_gazepoint_eda(data.frame(GSR_US = letters[1:3]), signal_col = "GSR_US"),
    "`signal_col` must be numeric"
  )

  expect_error(
    decompose_gazepoint_eda(df, window_size = 0),
    "`window_size`"
  )

  expect_error(
    decompose_gazepoint_eda(df, output_prefix = ""),
    "`output_prefix`"
  )

  existing <- data.frame(
    GSR_US = c(1, 2, 3),
    eda_tonic = c(1, 1, 1)
  )

  expect_error(
    decompose_gazepoint_eda(existing, signal_col = "GSR_US"),
    "Output columns already exist"
  )
})


test_that("detect_gazepoint_scr_events detects SCR-like peaks from phasic column", {
  df <- data.frame(
    CNT = 1:20,
    GSR_US_PHASIC = c(
      rep(0, 5),
      0.2, 0.8, 0.2,
      rep(0, 4),
      0.3, 0.9, 0.2,
      rep(0, 5)
    )
  )

  out <- detect_gazepoint_scr_events(
    df,
    phasic_col = "GSR_US_PHASIC",
    time_col = "CNT",
    threshold = 0.5,
    min_peak_distance = 3
  )

  expect_s3_class(out, "gazepoint_scr_events")
  expect_equal(out$overview$n_events, 2)
  expect_equal(out$overview$status, "scr_events_detected")
  expect_true(all(out$events$peak_value >= 0.5))
})


test_that("detect_gazepoint_scr_events handles no events", {
  df <- data.frame(
    CNT = 1:10,
    GSR_US_PHASIC = rep(0, 10)
  )

  out <- detect_gazepoint_scr_events(
    df,
    phasic_col = "GSR_US_PHASIC",
    threshold = 0.5
  )

  expect_equal(out$overview$n_events, 0)
  expect_equal(out$overview$status, "no_scr_events_detected")
  expect_equal(nrow(out$events), 0)
})


test_that("detect_gazepoint_scr_events supports grouping", {
  df <- data.frame(
    id = rep(c("P1", "P2"), each = 10),
    CNT = rep(1:10, 2),
    GSR_US_PHASIC = c(
      0, 0, 0.8, 0, 0, 0, 0, 0, 0, 0,
      0, 0, 0, 0, 0, 0.9, 0, 0, 0, 0
    )
  )

  out <- detect_gazepoint_scr_events(
    df,
    phasic_col = "GSR_US_PHASIC",
    time_col = "CNT",
    group_cols = "id",
    threshold = 0.5,
    min_peak_distance = 3
  )

  expect_equal(out$overview$group_count, 2)
  expect_equal(out$overview$n_events, 2)
  expect_equal(nrow(out$group_summary), 2)
})


test_that("detect_gazepoint_scr_events can decompose when no phasic column is supplied", {
  df <- data.frame(
    CNT = 1:20,
    GSR_US = c(rep(1, 5), 1.2, 2, 1.2, rep(1, 12))
  )

  out <- detect_gazepoint_scr_events(
    df,
    signal_col = "GSR_US",
    time_col = "CNT",
    threshold = 0.3,
    min_peak_distance = 3,
    window_size = 3
  )

  expect_true(out$overview$decomposition_used)
  expect_s3_class(out$events, "data.frame")
})


test_that("detect_gazepoint_scr_events validates arguments", {
  df <- data.frame(GSR_US_PHASIC = c(0, 1, 0))

  expect_error(
    detect_gazepoint_scr_events(1:3),
    "`data` must be"
  )

  expect_error(
    detect_gazepoint_scr_events(df, phasic_col = "missing"),
    "not found"
  )

  expect_error(
    detect_gazepoint_scr_events(df, threshold = NA),
    "`threshold`"
  )

  expect_error(
    detect_gazepoint_scr_events(df, min_peak_distance = 0),
    "`min_peak_distance`"
  )

  expect_error(
    detect_gazepoint_scr_events(data.frame(GSR_US_PHASIC = letters[1:3]), phasic_col = "GSR_US_PHASIC"),
    "`phasic_col` must be numeric"
  )
})

#' Summarise time-domain HRV features from Gazepoint IBI/RR intervals
#'
#' Computes conservative time-domain HRV-style features from genuine interbeat
#' interval or RR interval columns. The raw Gazepoint `HRV` column is not used
#' as an HRV metric because it is treated as a validity/vendor flag unless
#' independently documented otherwise.
#'
#' The helper computes descriptive features including mean IBI, mean heart rate
#' derived from IBI, SDNN, RMSSD, and pNN50. It does not compute frequency-domain
#' HRV and does not replace specialised ECG/PPG HRV software.
#'
#' @param data A data frame.
#' @param ibi_col Optional IBI/RR interval column. If `NULL`, a likely IBI/RR
#'   column is detected. The raw `HRV` column is never selected automatically.
#' @param group_cols Optional grouping columns, such as participant, stimulus,
#'   trial, or window.
#' @param time_col Optional time/order column used to order IBI values within
#'   each group before calculating successive-difference features.
#' @param ibi_unit Unit of the IBI/RR column. Use `"auto"`, `"seconds"`, or
#'   `"milliseconds"`.
#' @param min_ibi_ms Minimum plausible IBI in milliseconds.
#' @param max_ibi_ms Maximum plausible IBI in milliseconds.
#' @param min_valid_ibi Minimum number of valid IBI values required before a
#'   group is marked as having computed HRV features.
#'
#' @return A list with `overview`, `features`, and `settings`.
#'
#' @examples
#' df <- data.frame(
#'   participant = "P1",
#'   IBI = c(0.9, 1.0, 1.1, 1.0, 0.95)
#' )
#' summarise_gazepoint_hrv_features(df, group_cols = "participant")
#'
#' @export
summarise_gazepoint_hrv_features <- function(data,
                                             ibi_col = NULL,
                                             group_cols = NULL,
                                             time_col = NULL,
                                             ibi_unit = c("auto", "seconds", "milliseconds"),
                                             min_ibi_ms = 300,
                                             max_ibi_ms = 2000,
                                             min_valid_ibi = 3L) {
  if (!is.data.frame(data)) {
    stop("`data` must be a data frame.", call. = FALSE)
  }

  ibi_unit <- match.arg(ibi_unit)

  if (!is.null(ibi_col)) {
    .gpbiom_hrv_assert_columns(data, ibi_col, "ibi_col")
    if (length(ibi_col) != 1L) {
      stop("`ibi_col` must be NULL or a single column name.", call. = FALSE)
    }
  }

  if (!is.null(group_cols)) {
    .gpbiom_hrv_assert_columns(data, group_cols, "group_cols")
  }

  if (!is.null(time_col)) {
    .gpbiom_hrv_assert_columns(data, time_col, "time_col")
    if (length(time_col) != 1L) {
      stop("`time_col` must be NULL or a single column name.", call. = FALSE)
    }
  }

  .gpbiom_hrv_assert_positive_number(min_ibi_ms, "min_ibi_ms")
  .gpbiom_hrv_assert_positive_number(max_ibi_ms, "max_ibi_ms")
  .gpbiom_hrv_assert_positive_integer(min_valid_ibi, "min_valid_ibi")

  if (min_ibi_ms >= max_ibi_ms) {
    stop("`min_ibi_ms` must be smaller than `max_ibi_ms`.", call. = FALSE)
  }

  if (is.null(ibi_col)) {
    ibi_col <- .gpbiom_hrv_choose_ibi_column(data)
  }

  if (is.na(ibi_col) || !ibi_col %in% names(data)) {
    stop(
      "No IBI/RR interval column was detected. Provide `ibi_col` explicitly. ",
      "The raw `HRV` column is not treated as an HRV metric.",
      call. = FALSE
    )
  }

  if (identical(ibi_col, "HRV")) {
    stop(
      "`HRV` is treated as a validity/vendor flag, not as an HRV metric. ",
      "Use a genuine IBI/RR interval column.",
      call. = FALSE
    )
  }

  if (!is.numeric(data[[ibi_col]])) {
    stop("`ibi_col` must be numeric.", call. = FALSE)
  }

  groups <- .gpbiom_hrv_group_indices(data, group_cols)

  feature_rows <- lapply(names(groups), function(group_name) {
    idx <- groups[[group_name]]
    idx <- .gpbiom_hrv_order_index(data, idx, time_col)

    values <- data[[ibi_col]][idx]
    unit <- .gpbiom_hrv_resolve_unit(values, ibi_unit)
    values_ms <- .gpbiom_hrv_to_ms(values, unit)

    .gpbiom_hrv_feature_row(
      group_name = group_name,
      values_ms = values_ms,
      unit = unit,
      min_ibi_ms = min_ibi_ms,
      max_ibi_ms = max_ibi_ms,
      min_valid_ibi = min_valid_ibi
    )
  })

  features <- if (length(feature_rows) == 0L) {
    .gpbiom_hrv_empty_features()
  } else {
    do.call(rbind, feature_rows)
  }

  overview <- data.frame(
    n_rows = nrow(data),
    ibi_col = ibi_col,
    group_count = length(groups),
    feature_rows = nrow(features),
    groups_with_computed_features = sum(features$status == "hrv_features_computed"),
    groups_with_insufficient_ibi = sum(features$status == "insufficient_valid_ibi"),
    total_valid_ibi = sum(features$n_valid_ibi, na.rm = TRUE),
    status = if (any(features$status == "hrv_features_computed")) {
      "hrv_features_available"
    } else {
      "insufficient_valid_ibi"
    },
    stringsAsFactors = FALSE
  )

  out <- list(
    overview = overview,
    features = features,
    settings = list(
      ibi_col = ibi_col,
      group_cols = group_cols,
      time_col = time_col,
      ibi_unit = ibi_unit,
      min_ibi_ms = min_ibi_ms,
      max_ibi_ms = max_ibi_ms,
      min_valid_ibi = as.integer(min_valid_ibi),
      note = paste0(
        "Time-domain HRV features are derived from IBI/RR intervals only. ",
        "The raw Gazepoint HRV column is treated as a validity/vendor flag."
      )
    )
  )

  class(out) <- c("gazepoint_hrv_features", class(out))
  out
}


.gpbiom_hrv_choose_ibi_column <- function(data) {
  standard_names <- standardise_gazepoint_biometric_names(names(data))

  candidates <- names(data)[standard_names == "IBI"]

  numeric_candidates <- candidates[vapply(candidates, function(column) {
    is.numeric(data[[column]])
  }, logical(1))]

  if (length(numeric_candidates) > 0L) {
    return(numeric_candidates[1L])
  }

  rr_like <- names(data)[grepl(
    "^(IBI|RR|RR_INTERVAL|RR_MS|INTERBEAT|INTER_BEAT)",
    toupper(names(data))
  )]

  rr_like <- rr_like[!toupper(rr_like) %in% "HRV"]

  numeric_rr_like <- rr_like[vapply(rr_like, function(column) {
    is.numeric(data[[column]])
  }, logical(1))]

  if (length(numeric_rr_like) > 0L) {
    return(numeric_rr_like[1L])
  }

  NA_character_
}


.gpbiom_hrv_resolve_unit <- function(values, ibi_unit = "auto") {
  if (!identical(ibi_unit, "auto")) {
    return(ibi_unit)
  }

  finite <- values[is.finite(values)]

  if (length(finite) == 0L) {
    return("unknown")
  }

  median_value <- stats::median(finite)

  if (is.finite(median_value) && median_value > 10) {
    "milliseconds"
  } else {
    "seconds"
  }
}


.gpbiom_hrv_to_ms <- function(values, unit) {
  values <- as.numeric(values)

  if (identical(unit, "milliseconds")) {
    return(values)
  }

  if (identical(unit, "seconds")) {
    return(values * 1000)
  }

  values
}


.gpbiom_hrv_group_indices <- function(data, group_cols = NULL) {
  if (is.null(group_cols) || length(group_cols) == 0L) {
    return(list(all = seq_len(nrow(data))))
  }

  group_data <- data[group_cols]

  group_key <- apply(group_data, 1, function(row) {
    row <- as.character(row)
    row[is.na(row)] <- "<NA>"
    paste(row, collapse = " | ")
  })

  split(seq_len(nrow(data)), group_key, drop = TRUE)
}


.gpbiom_hrv_order_index <- function(data, idx, time_col = NULL) {
  if (is.null(time_col)) {
    return(idx)
  }

  idx[order(data[[time_col]][idx], na.last = TRUE)]
}


.gpbiom_hrv_feature_row <- function(group_name,
                                    values_ms,
                                    unit,
                                    min_ibi_ms,
                                    max_ibi_ms,
                                    min_valid_ibi) {
  n_total <- length(values_ms)
  missing <- is.na(values_ms) | !is.finite(values_ms)
  finite <- !missing

  out_of_range <- finite & (values_ms < min_ibi_ms | values_ms > max_ibi_ms)
  valid <- finite & !out_of_range

  valid_values <- values_ms[valid]

  n_valid <- length(valid_values)

  diffs <- if (n_valid >= 2L) {
    diff(valid_values)
  } else {
    numeric()
  }

  status <- if (n_valid >= min_valid_ibi) {
    "hrv_features_computed"
  } else {
    "insufficient_valid_ibi"
  }

  data.frame(
    group = group_name,
    n_total_ibi = n_total,
    n_missing_ibi = sum(missing),
    n_out_of_range_ibi = sum(out_of_range),
    n_valid_ibi = n_valid,
    valid_ibi_rate = if (n_total > 0L) n_valid / n_total else NA_real_,
    unit_detected = unit,
    mean_ibi_ms = if (n_valid > 0L) mean(valid_values) else NA_real_,
    median_ibi_ms = if (n_valid > 0L) stats::median(valid_values) else NA_real_,
    sdnn_ms = if (n_valid >= 2L) stats::sd(valid_values) else NA_real_,
    rmssd_ms = if (length(diffs) > 0L) sqrt(mean(diffs^2)) else NA_real_,
    pnn50_percent = if (length(diffs) > 0L) mean(abs(diffs) > 50) * 100 else NA_real_,
    mean_hr_bpm_from_ibi = if (n_valid > 0L) 60000 / mean(valid_values) else NA_real_,
    min_ibi_ms = if (n_valid > 0L) min(valid_values) else NA_real_,
    max_ibi_ms = if (n_valid > 0L) max(valid_values) else NA_real_,
    status = status,
    stringsAsFactors = FALSE
  )
}


.gpbiom_hrv_empty_features <- function() {
  data.frame(
    group = character(),
    n_total_ibi = integer(),
    n_missing_ibi = integer(),
    n_out_of_range_ibi = integer(),
    n_valid_ibi = integer(),
    valid_ibi_rate = numeric(),
    unit_detected = character(),
    mean_ibi_ms = numeric(),
    median_ibi_ms = numeric(),
    sdnn_ms = numeric(),
    rmssd_ms = numeric(),
    pnn50_percent = numeric(),
    mean_hr_bpm_from_ibi = numeric(),
    min_ibi_ms = numeric(),
    max_ibi_ms = numeric(),
    status = character(),
    stringsAsFactors = FALSE
  )
}


.gpbiom_hrv_assert_columns <- function(data, columns, argument) {
  missing <- setdiff(columns, names(data))

  if (length(missing) > 0L) {
    stop(
      "`", argument, "` contains columns not found in `data`: ",
      paste(missing, collapse = ", "),
      call. = FALSE
    )
  }

  invisible(TRUE)
}


.gpbiom_hrv_assert_positive_number <- function(x, argument) {
  if (!is.numeric(x) ||
      length(x) != 1L ||
      is.na(x) ||
      !is.finite(x) ||
      x <= 0) {
    stop("`", argument, "` must be a positive number.", call. = FALSE)
  }

  invisible(TRUE)
}


.gpbiom_hrv_assert_positive_integer <- function(x, argument) {
  if (!is.numeric(x) ||
      length(x) != 1L ||
      is.na(x) ||
      !is.finite(x) ||
      x <= 0 ||
      x != as.integer(x)) {
    stop("`", argument, "` must be a positive integer.", call. = FALSE)
  }

  invisible(TRUE)
}

test_that("summarise_gazepoint_hrv_features computes time-domain features from IBI seconds", {
  df <- data.frame(
    participant = "P1",
    time = 1:4,
    IBI = c(1.0, 1.1, 0.9, 1.0)
  )

  out <- summarise_gazepoint_hrv_features(
    df,
    group_cols = "participant",
    time_col = "time"
  )

  expect_s3_class(out, "gazepoint_hrv_features")
  expect_s3_class(out$overview, "data.frame")
  expect_s3_class(out$features, "data.frame")

  expect_equal(out$overview$status, "hrv_features_available")
  expect_equal(out$features$n_valid_ibi, 4)
  expect_equal(out$features$unit_detected, "seconds")
  expect_equal(out$features$mean_ibi_ms, 1000)
  expect_equal(out$features$mean_hr_bpm_from_ibi, 60)
  expect_equal(round(out$features$rmssd_ms, 3), round(sqrt(20000), 3))
  expect_equal(out$features$pnn50_percent, 100)
})


test_that("summarise_gazepoint_hrv_features detects IBI and ignores raw HRV", {
  df <- data.frame(
    HRV = c(1, 1, 1, 1),
    IBI = c(900, 1000, 1100, 1000)
  )

  out <- summarise_gazepoint_hrv_features(df)

  expect_equal(out$settings$ibi_col, "IBI")
  expect_equal(out$features$unit_detected, "milliseconds")
  expect_equal(out$features$mean_ibi_ms, 1000)
})


test_that("summarise_gazepoint_hrv_features supports groups", {
  df <- data.frame(
    participant = rep(c("P1", "P2"), each = 4),
    time = rep(1:4, 2),
    IBI = c(1.0, 1.1, 0.9, 1.0, 0.8, 0.85, 0.9, 0.95)
  )

  out <- summarise_gazepoint_hrv_features(
    df,
    group_cols = "participant",
    time_col = "time"
  )

  expect_equal(out$overview$group_count, 2)
  expect_equal(nrow(out$features), 2)
  expect_true(all(out$features$status == "hrv_features_computed"))
})


test_that("summarise_gazepoint_hrv_features orders values by time", {
  df <- data.frame(
    time = c(3, 1, 2),
    IBI = c(1.2, 1.0, 1.1)
  )

  out <- summarise_gazepoint_hrv_features(
    df,
    time_col = "time",
    min_valid_ibi = 3
  )

  expect_equal(out$features$mean_ibi_ms, 1100)
  expect_equal(out$features$n_valid_ibi, 3)
})


test_that("summarise_gazepoint_hrv_features flags insufficient and invalid IBI", {
  df <- data.frame(
    IBI = c(0.2, 1.0, 3.0, NA)
  )

  out <- summarise_gazepoint_hrv_features(
    df,
    min_ibi_ms = 300,
    max_ibi_ms = 2000,
    min_valid_ibi = 3
  )

  expect_equal(out$features$n_valid_ibi, 1)
  expect_equal(out$features$n_out_of_range_ibi, 2)
  expect_equal(out$features$n_missing_ibi, 1)
  expect_equal(out$features$status, "insufficient_valid_ibi")
  expect_equal(out$overview$status, "insufficient_valid_ibi")
})


test_that("summarise_gazepoint_hrv_features supports explicit millisecond unit", {
  df <- data.frame(
    rr_ms = c(800, 850, 900, 950)
  )

  out <- summarise_gazepoint_hrv_features(
    df,
    ibi_col = "rr_ms",
    ibi_unit = "milliseconds"
  )

  expect_equal(out$features$unit_detected, "milliseconds")
  expect_equal(out$features$mean_ibi_ms, 875)
})


test_that("summarise_gazepoint_hrv_features validates arguments", {
  expect_error(
    summarise_gazepoint_hrv_features(1:3),
    "`data` must be"
  )

  df <- data.frame(IBI = c(1, 1.1, 1.0))

  expect_error(
    summarise_gazepoint_hrv_features(df, ibi_col = "missing"),
    "not found"
  )

  expect_error(
    summarise_gazepoint_hrv_features(data.frame(HRV = c(1, 1, 1))),
    "No IBI/RR interval"
  )

  expect_error(
    summarise_gazepoint_hrv_features(data.frame(HRV = c(1, 1, 1)), ibi_col = "HRV"),
    "validity/vendor flag"
  )

  expect_error(
    summarise_gazepoint_hrv_features(
      data.frame(IBI = letters[1:3]),
      ibi_col = "IBI"
    ),
    "`ibi_col` must be numeric"
  )

  expect_error(
    summarise_gazepoint_hrv_features(df, min_ibi_ms = 0),
    "`min_ibi_ms`"
  )

  expect_error(
    summarise_gazepoint_hrv_features(df, max_ibi_ms = 0),
    "`max_ibi_ms`"
  )

  expect_error(
    summarise_gazepoint_hrv_features(df, min_ibi_ms = 2000, max_ibi_ms = 300),
    "smaller"
  )

  expect_error(
    summarise_gazepoint_hrv_features(df, min_valid_ibi = 0),
    "`min_valid_ibi`"
  )
})

#' Plot Gazepoint biometric signal activity
#'
#' Plots signal-activity summaries produced by
#' [audit_gazepoint_signal_activity()], or computes them from a biometric data
#' frame. The plot is intended for quality-control review of signal
#' availability, missingness, zero activity, and basic activity status. It does
#' not infer emotion, valence, cognition, preference, trust, or physiological
#' diagnosis.
#'
#' @param data A Gazepoint biometric data frame, or an
#'   `audit_gazepoint_signal_activity()` result.
#' @param signal_cols Optional signal columns used when `data` is a raw data
#'   frame.
#' @param group_cols Optional grouping columns used when `data` is a raw data
#'   frame.
#' @param metric Summary metric to plot.
#' @param max_groups Maximum number of groups to display.
#' @param title Optional plot title.
#'
#' @return A ggplot object with the package plot contract attached.
#' @export
plot_gazepoint_signal_activity <- function(data,
                                           signal_cols = NULL,
                                           group_cols = NULL,
                                           metric = c(
                                             "active_signal",
                                             "nonzero_prop",
                                             "missing_prop",
                                             "n_unique_finite"
                                           ),
                                           max_groups = 30,
                                           title = NULL) {
  gpbiometrics_require_ggplot2()

  metric <- match.arg(metric)

  if (!is.numeric(max_groups) || length(max_groups) != 1 ||
      is.na(max_groups) || max_groups < 1) {
    stop("`max_groups` must be a single positive number.", call. = FALSE)
  }

  activity <- gpbiometrics_qc_plot_extract_signal_activity(
    data = data,
    signal_cols = signal_cols,
    group_cols = group_cols
  )

  plot_data <- as.data.frame(activity$signal_by_group, stringsAsFactors = FALSE)

  if (nrow(plot_data) == 0) {
    stop("No signal-activity rows were available for plotting.", call. = FALSE)
  }

  plot_data <- gpbiometrics_qc_plot_ensure_activity_metrics(plot_data)

  if (!metric %in% names(plot_data)) {
    stop("Metric `", metric, "` was not found in the signal-activity summary.", call. = FALSE)
  }

  plot_data$.plot_group <- gpbiometrics_qc_plot_group_label(plot_data)
  plot_data$.plot_value <- gpbiometrics_qc_plot_metric_to_numeric(plot_data[[metric]])
  plot_data$.plot_metric <- metric

  selected_groups <- unique(plot_data$.plot_group)

  if (length(selected_groups) > max_groups) {
    selected_groups <- selected_groups[seq_len(max_groups)]
    plot_data <- plot_data[plot_data$.plot_group %in% selected_groups, , drop = FALSE]
  }

  if (is.null(title)) {
    title <- "Gazepoint biometric signal activity"
  }

  p <- ggplot2::ggplot(
    plot_data,
    ggplot2::aes(x = .data$signal, y = .data$.plot_value)
  ) +
    ggplot2::geom_col() +
    ggplot2::facet_wrap(stats::as.formula("~ .plot_group"), scales = "free_x") +
    ggplot2::labs(
      title = title,
      subtitle = "QC summary only; signal activity is not evidence of emotion, valence, cognition, or preference.",
      x = "Signal",
      y = metric
    ) +
    ggplot2::theme(axis.text.x = ggplot2::element_text(angle = 45, hjust = 1))

  settings <- list(
    plot_type = "signal_activity",
    metric = metric,
    signal_cols = signal_cols,
    group_cols = group_cols,
    max_groups = max_groups,
    interpretation_notes = c(
      "Signal activity plots summarize availability and basic variation.",
      "They do not infer emotion, valence, cognition, trust, preference, or physiological diagnosis."
    )
  )

  standardise_gazepoint_plot_contract(
    plot = p,
    plot_data = plot_data,
    settings = settings,
    interpretation_notes = settings$interpretation_notes,
    plot_type = "signal_activity"
  )
}

#' Plot Gazepoint time resets and time-order flags
#'
#' Plots row-level time/counter progression and flags from
#' [audit_gazepoint_time_resets()], or computes them from a biometric data
#' frame. The plot is intended for synchronization and file-structure QC. It
#' does not establish causal timing or true physiological latency.
#'
#' @param data A Gazepoint biometric data frame, or an
#'   `audit_gazepoint_time_resets()` result.
#' @param time_col Optional time or counter column used when `data` is a raw
#'   data frame.
#' @param group_cols Optional grouping columns used when `data` is a raw data
#'   frame.
#' @param max_groups Maximum number of groups to display.
#' @param title Optional plot title.
#'
#' @return A ggplot object with the package plot contract attached.
#' @export
plot_gazepoint_time_resets <- function(data,
                                       time_col = NULL,
                                       group_cols = NULL,
                                       max_groups = 30,
                                       title = NULL) {
  gpbiometrics_require_ggplot2()

  if (!is.numeric(max_groups) || length(max_groups) != 1 ||
      is.na(max_groups) || max_groups < 1) {
    stop("`max_groups` must be a single positive number.", call. = FALSE)
  }

  reset_audit <- gpbiometrics_qc_plot_extract_time_resets(
    data = data,
    time_col = time_col,
    group_cols = group_cols
  )

  plot_data <- as.data.frame(reset_audit$row_flags, stringsAsFactors = FALSE)

  if (nrow(plot_data) == 0) {
    stop("No time-reset rows were available for plotting.", call. = FALSE)
  }

  if (!"time_value" %in% names(plot_data)) {
    stop("The time-reset audit must contain a `time_value` column.", call. = FALSE)
  }

  if ("group_row_index" %in% names(plot_data)) {
    plot_data$.plot_index <- plot_data$group_row_index
  } else {
    plot_data$.plot_index <- seq_len(nrow(plot_data))
  }

  plot_data$.plot_group <- gpbiometrics_qc_plot_group_label(plot_data)

  issue_cols <- intersect(
    c(
      "flag_nonfinite_time",
      "flag_negative_step",
      "flag_duplicate_time",
      "flag_nonmonotonic",
      "flag_short_segment"
    ),
    names(plot_data)
  )

  if (length(issue_cols) == 0) {
    plot_data$.any_time_issue <- FALSE
  } else {
    issue_data <- plot_data[issue_cols]
    issue_data[] <- lapply(issue_data, function(x) {
      x <- as.logical(x)
      x[is.na(x)] <- FALSE
      x
    })

    plot_data$.any_time_issue <- rowSums(issue_data) > 0
  }

  selected_groups <- unique(plot_data$.plot_group)

  if (length(selected_groups) > max_groups) {
    selected_groups <- selected_groups[seq_len(max_groups)]
    plot_data <- plot_data[plot_data$.plot_group %in% selected_groups, , drop = FALSE]
  }

  if (is.null(title)) {
    title <- "Gazepoint time/counter reset diagnostics"
  }

  p <- ggplot2::ggplot(
    plot_data,
    ggplot2::aes(x = .data$.plot_index, y = .data$time_value)
  ) +
    ggplot2::geom_line(ggplot2::aes(group = .data$.plot_group)) +
    ggplot2::geom_point(ggplot2::aes(shape = .data$.any_time_issue)) +
    ggplot2::facet_wrap(stats::as.formula("~ .plot_group"), scales = "free_x") +
    ggplot2::labs(
      title = title,
      subtitle = "QC summary only; time-order flags do not establish causal timing or physiological latency.",
      x = "Row index within group",
      y = "Time/counter value",
      shape = "Time issue"
    )

  settings <- list(
    plot_type = "time_resets",
    time_col = time_col,
    group_cols = group_cols,
    max_groups = max_groups,
    interpretation_notes = c(
      "Time-reset plots summarize time/counter ordering and reset flags.",
      "They support synchronization QC only and do not establish causal timing or true physiological latency."
    )
  )

  standardise_gazepoint_plot_contract(
    plot = p,
    plot_data = plot_data,
    settings = settings,
    interpretation_notes = settings$interpretation_notes,
    plot_type = "time_resets"
  )
}

#' Create a lightweight Gazepoint biometric QC plot dashboard
#'
#' Creates a lightweight dashboard object containing QC plots for signal activity
#' and time-reset diagnostics. This is a structured list of ggplot objects, not a
#' Shiny application. The dashboard is intended for report preparation and manual
#' QC review.
#'
#' @param data Optional Gazepoint biometric data frame.
#' @param signal_activity Optional result from
#'   [audit_gazepoint_signal_activity()].
#' @param time_resets Optional result from [audit_gazepoint_time_resets()].
#' @param signal_cols Optional signal columns used when computing signal
#'   activity from `data`.
#' @param group_cols Optional grouping columns.
#' @param time_col Optional time or counter column.
#' @param include_signal_activity If `TRUE`, include a signal-activity plot.
#' @param include_time_resets If `TRUE`, include a time-reset plot.
#' @param max_groups Maximum number of groups to display in each plot.
#' @param continue_on_error If `TRUE`, plot failures are recorded in `errors`
#'   rather than stopping the dashboard.
#' @param title_prefix Optional title prefix.
#'
#' @return A list with `overview`, `plots`, `errors`, `inputs`, and `settings`.
#' @export
plot_gazepoint_biometric_report_dashboard <- function(data = NULL,
                                                      signal_activity = NULL,
                                                      time_resets = NULL,
                                                      signal_cols = NULL,
                                                      group_cols = NULL,
                                                      time_col = NULL,
                                                      include_signal_activity = TRUE,
                                                      include_time_resets = TRUE,
                                                      max_groups = 30,
                                                      continue_on_error = TRUE,
                                                      title_prefix = "Gazepoint biometric QC") {
  gpbiometrics_require_ggplot2()

  if (!is.logical(include_signal_activity) ||
      length(include_signal_activity) != 1 ||
      is.na(include_signal_activity)) {
    stop("`include_signal_activity` must be TRUE or FALSE.", call. = FALSE)
  }

  if (!is.logical(include_time_resets) ||
      length(include_time_resets) != 1 ||
      is.na(include_time_resets)) {
    stop("`include_time_resets` must be TRUE or FALSE.", call. = FALSE)
  }

  if (!is.logical(continue_on_error) ||
      length(continue_on_error) != 1 ||
      is.na(continue_on_error)) {
    stop("`continue_on_error` must be TRUE or FALSE.", call. = FALSE)
  }

  if (is.null(data) && is.null(signal_activity) && is.null(time_resets)) {
    stop(
      "Supply `data`, `signal_activity`, `time_resets`, or a combination of these.",
      call. = FALSE
    )
  }

  plots <- list()
  errors <- data.frame(
    plot = character(),
    message = character(),
    stringsAsFactors = FALSE
  )

  if (isTRUE(include_signal_activity)) {
    signal_input <- if (!is.null(signal_activity)) signal_activity else data

    signal_result <- tryCatch(
      plot_gazepoint_signal_activity(
        data = signal_input,
        signal_cols = signal_cols,
        group_cols = group_cols,
        max_groups = max_groups,
        title = paste(title_prefix, "- signal activity")
      ),
      error = function(e) e
    )

    if (inherits(signal_result, "error")) {
      if (!isTRUE(continue_on_error)) {
        stop(conditionMessage(signal_result), call. = FALSE)
      }

      errors <- rbind(
        errors,
        data.frame(
          plot = "signal_activity",
          message = conditionMessage(signal_result),
          stringsAsFactors = FALSE
        )
      )
    } else {
      plots$signal_activity <- signal_result
    }
  }

  if (isTRUE(include_time_resets)) {
    time_input <- if (!is.null(time_resets)) time_resets else data

    time_result <- tryCatch(
      plot_gazepoint_time_resets(
        data = time_input,
        time_col = time_col,
        group_cols = group_cols,
        max_groups = max_groups,
        title = paste(title_prefix, "- time resets")
      ),
      error = function(e) e
    )

    if (inherits(time_result, "error")) {
      if (!isTRUE(continue_on_error)) {
        stop(conditionMessage(time_result), call. = FALSE)
      }

      errors <- rbind(
        errors,
        data.frame(
          plot = "time_resets",
          message = conditionMessage(time_result),
          stringsAsFactors = FALSE
        )
      )
    } else {
      plots$time_resets <- time_result
    }
  }

  plot_contract_ok <- vapply(
    plots,
    function(x) isTRUE(attr(x, "gazepoint_plot_contract")),
    logical(1)
  )

  status <- if (length(plots) == 0) {
    "no_plots_created"
  } else if (nrow(errors) > 0) {
    "partial_dashboard_created"
  } else {
    "dashboard_created"
  }

  overview <- data.frame(
    plot_count = length(plots),
    error_count = nrow(errors),
    all_plot_contracts_ok = if (length(plot_contract_ok) == 0) {
      NA
    } else {
      all(plot_contract_ok)
    },
    status = status,
    interpretation = paste(
      "Dashboard plots are lightweight QC aids for manual review and reporting.",
      "They do not infer emotion, valence, cognition, trust, preference, or physiological diagnosis."
    ),
    stringsAsFactors = FALSE
  )

  structure(
    list(
      overview = overview,
      plots = plots,
      errors = errors,
      inputs = list(
        data_supplied = !is.null(data),
        signal_activity_supplied = !is.null(signal_activity),
        time_resets_supplied = !is.null(time_resets)
      ),
      settings = list(
        signal_cols = signal_cols,
        group_cols = group_cols,
        time_col = time_col,
        include_signal_activity = include_signal_activity,
        include_time_resets = include_time_resets,
        max_groups = max_groups,
        continue_on_error = continue_on_error,
        title_prefix = title_prefix
      )
    ),
    class = c("gazepoint_biometric_plot_dashboard", "list")
  )
}

gpbiometrics_qc_plot_extract_signal_activity <- function(data,
                                                         signal_cols = NULL,
                                                         group_cols = NULL) {
  if (is.list(data) && is.data.frame(data$signal_by_group)) {
    return(data)
  }

  if (!is.data.frame(data)) {
    stop(
      "`data` must be a data frame or an audit_gazepoint_signal_activity() result.",
      call. = FALSE
    )
  }

  audit_gazepoint_signal_activity(
    data = data,
    signal_cols = signal_cols,
    group_cols = group_cols
  )
}

gpbiometrics_qc_plot_extract_time_resets <- function(data,
                                                     time_col = NULL,
                                                     group_cols = NULL) {
  if (is.list(data) && is.data.frame(data$row_flags)) {
    return(data)
  }

  if (!is.data.frame(data)) {
    stop(
      "`data` must be a data frame or an audit_gazepoint_time_resets() result.",
      call. = FALSE
    )
  }

  audit_gazepoint_time_resets(
    data = data,
    time_col = time_col,
    group_cols = group_cols
  )
}

gpbiometrics_qc_plot_group_label <- function(dat) {
  if ("group_id" %in% names(dat)) {
    out <- as.character(dat$group_id)
    out[is.na(out) | !nzchar(out)] <- "all"
    return(out)
  }

  candidates <- intersect(
    c(
      "source_file",
      "source_participant",
      "participant",
      "subject",
      "subject_id",
      "USER",
      "USER_FILE",
      "MEDIA_ID",
      "MEDIA_NAME",
      "stimulus",
      "stimulus_id",
      "trial",
      "trial_id",
      "trial_global",
      "reset_segment_index"
    ),
    names(dat)
  )

  if (length(candidates) == 0) {
    return(rep("all", nrow(dat)))
  }

  group_dat <- dat[candidates]

  group_dat[] <- lapply(group_dat, function(x) {
    x_chr <- as.character(x)
    x_chr[is.na(x_chr)] <- "<NA>"
    x_chr
  })

  apply(group_dat, 1, paste, collapse = "||")
}

gpbiometrics_qc_plot_metric_to_numeric <- function(x) {
  if (is.logical(x)) {
    out <- as.numeric(x)
    out[is.na(out)] <- 0
    return(out)
  }

  if (is.character(x)) {
    lower_x <- tolower(x)
    if (all(lower_x %in% c("true", "false", "active", "inactive", "pass", "fail", NA))) {
      return(as.numeric(lower_x %in% c("true", "active", "pass")))
    }
  }

  suppressWarnings(as.numeric(x))
}

gpbiometrics_qc_plot_ensure_activity_metrics <- function(plot_data) {
  plot_data <- as.data.frame(plot_data, stringsAsFactors = FALSE)

  if (!"missing_prop" %in% names(plot_data) && all(c("n_missing", "n") %in% names(plot_data))) {
    n <- suppressWarnings(as.numeric(plot_data$n))
    n_missing <- suppressWarnings(as.numeric(plot_data$n_missing))
    plot_data$missing_prop <- ifelse(is.finite(n) & n > 0, n_missing / n, NA_real_)
  }

  if (!"nonzero_prop" %in% names(plot_data) && all(c("n_nonzero", "n") %in% names(plot_data))) {
    n <- suppressWarnings(as.numeric(plot_data$n))
    n_nonzero <- suppressWarnings(as.numeric(plot_data$n_nonzero))
    plot_data$nonzero_prop <- ifelse(is.finite(n) & n > 0, n_nonzero / n, NA_real_)
  }

  if (!"active_signal" %in% names(plot_data)) {
    if ("is_active" %in% names(plot_data)) {
      plot_data$active_signal <- as.logical(plot_data$is_active)
    } else if ("active" %in% names(plot_data)) {
      plot_data$active_signal <- as.logical(plot_data$active)
    } else if ("status" %in% names(plot_data)) {
      status <- tolower(as.character(plot_data$status))
      plot_data$active_signal <- !grepl(
        "inactive|insufficient|missing|flat|constant|unavailable|fail",
        status
      )
      plot_data$active_signal[is.na(status)] <- FALSE
    } else if ("signal_status" %in% names(plot_data)) {
      status <- tolower(as.character(plot_data$signal_status))
      plot_data$active_signal <- !grepl(
        "inactive|insufficient|missing|flat|constant|unavailable|fail",
        status
      )
      plot_data$active_signal[is.na(status)] <- FALSE
    } else if ("n_unique_nonzero" %in% names(plot_data)) {
      n_unique_nonzero <- suppressWarnings(as.numeric(plot_data$n_unique_nonzero))
      plot_data$active_signal <- is.finite(n_unique_nonzero) & n_unique_nonzero >= 2
    } else if ("n_unique_finite" %in% names(plot_data)) {
      n_unique_finite <- suppressWarnings(as.numeric(plot_data$n_unique_finite))
      plot_data$active_signal <- is.finite(n_unique_finite) & n_unique_finite >= 2
    } else if ("nonzero_prop" %in% names(plot_data)) {
      nonzero_prop <- suppressWarnings(as.numeric(plot_data$nonzero_prop))
      plot_data$active_signal <- is.finite(nonzero_prop) & nonzero_prop > 0
    } else {
      plot_data$active_signal <- FALSE
    }
  }

  plot_data$active_signal[is.na(plot_data$active_signal)] <- FALSE

  plot_data
}

test_that("plot_gazepoint_signal_activity returns a contracted ggplot", {
  skip_if_not_installed("ggplot2")

  dat <- data.frame(
    participant = rep(c("p1", "p2"), each = 20),
    GSR_US = c(rep(0, 10), seq_len(10), rep(1, 20)),
    HR = c(60 + seq_len(20), rep(70, 20))
  )

  p <- plot_gazepoint_signal_activity(
    dat,
    signal_cols = c("GSR_US", "HR"),
    group_cols = "participant",
    metric = "nonzero_prop"
  )

  expect_s3_class(p, "gazepoint_plot")
  expect_true(isTRUE(attr(p, "gazepoint_plot_contract")))
  expect_equal(attr(p, "plot_type"), "signal_activity")
  expect_true(is.data.frame(attr(p, "plot_data")))
  expect_true(all(c("signal", ".plot_value", ".plot_group") %in% names(attr(p, "plot_data"))))
})

test_that("plot_gazepoint_signal_activity accepts an existing activity audit", {
  skip_if_not_installed("ggplot2")

  dat <- data.frame(
    participant = rep(c("p1", "p2"), each = 10),
    GSR_US = c(rep(0, 10), seq_len(10)),
    HR = 60 + seq_len(20)
  )

  audit <- audit_gazepoint_signal_activity(
    dat,
    signal_cols = c("GSR_US", "HR"),
    group_cols = "participant"
  )

  p <- plot_gazepoint_signal_activity(
    audit,
    metric = "active_signal"
  )

  expect_s3_class(p, "gazepoint_plot")
  expect_equal(attr(p, "plot_type"), "signal_activity")
})

test_that("plot_gazepoint_time_resets returns a contracted ggplot", {
  skip_if_not_installed("ggplot2")

  dat <- data.frame(
    participant = "p1",
    time_ms = c(1, 2, 3, 1, 2, 3),
    GSR_US = seq_len(6)
  )

  p <- plot_gazepoint_time_resets(
    dat,
    time_col = "time_ms",
    group_cols = "participant"
  )

  expect_s3_class(p, "gazepoint_plot")
  expect_true(isTRUE(attr(p, "gazepoint_plot_contract")))
  expect_equal(attr(p, "plot_type"), "time_resets")
  expect_true(is.data.frame(attr(p, "plot_data")))
  expect_true(any(attr(p, "plot_data")$.any_time_issue))
})

test_that("plot_gazepoint_time_resets accepts an existing time-reset audit", {
  skip_if_not_installed("ggplot2")

  dat <- data.frame(
    participant = "p1",
    time_ms = c(1, 2, 3, 4, 5),
    GSR_US = seq_len(5)
  )

  audit <- audit_gazepoint_time_resets(
    dat,
    time_col = "time_ms",
    group_cols = "participant"
  )

  p <- plot_gazepoint_time_resets(audit)

  expect_s3_class(p, "gazepoint_plot")
  expect_equal(attr(p, "plot_type"), "time_resets")
  expect_false(any(attr(p, "plot_data")$.any_time_issue))
})

test_that("plot_gazepoint_biometric_report_dashboard returns a lightweight dashboard", {
  skip_if_not_installed("ggplot2")

  dat <- data.frame(
    participant = rep(c("p1", "p2"), each = 20),
    time_ms = rep(seq_len(20), times = 2),
    GSR_US = c(rep(0, 10), seq_len(10), rep(1, 20)),
    HR = c(60 + seq_len(20), rep(70, 20))
  )

  dashboard <- plot_gazepoint_biometric_report_dashboard(
    data = dat,
    signal_cols = c("GSR_US", "HR"),
    group_cols = "participant",
    time_col = "time_ms"
  )

  expect_s3_class(dashboard, "gazepoint_biometric_plot_dashboard")
  expect_equal(dashboard$overview$status, "dashboard_created")
  expect_equal(dashboard$overview$plot_count, 2)
  expect_named(dashboard$plots, c("signal_activity", "time_resets"))
  expect_equal(nrow(dashboard$errors), 0)

  expect_true(all(vapply(
    dashboard$plots,
    function(x) isTRUE(attr(x, "gazepoint_plot_contract")),
    logical(1)
  )))
})

test_that("plot_gazepoint_biometric_report_dashboard can continue after plot errors", {
  skip_if_not_installed("ggplot2")

  dat <- data.frame(
    participant = "p1",
    time_ms = seq_len(10),
    GSR_US = seq_len(10)
  )

  dashboard <- plot_gazepoint_biometric_report_dashboard(
    data = dat,
    signal_cols = "missing_signal",
    group_cols = "participant",
    time_col = "time_ms",
    continue_on_error = TRUE
  )

  expect_s3_class(dashboard, "gazepoint_biometric_plot_dashboard")
  expect_equal(dashboard$overview$status, "partial_dashboard_created")
  expect_equal(dashboard$overview$plot_count, 1)
  expect_equal(dashboard$overview$error_count, 1)
  expect_true("time_resets" %in% names(dashboard$plots))
  expect_equal(dashboard$errors$plot, "signal_activity")
})

#' Audit distributional drift across sessions or ordered blocks
#'
#' Compares signal distributions across sessions/blocks using baseline-vs-current
#' differences, Kolmogorov-Smirnov tests, and Population Stability Index (PSI).
#'
#' @param dat A data frame containing longitudinal biometric data.
#' @param signal_cols Numeric signal columns to audit.
#' @param session_col Session/block/timepoint column.
#' @param participant_col Optional participant column.
#' @param reference_session Optional reference session. If `NULL`, the first
#'   ordered session is used within each participant/global group.
#' @param bins Number of bins for PSI.
#' @param psi_warn PSI threshold for warning.
#' @param psi_fail PSI threshold for failure.
#'
#' @return A list with `overview`, `drift_summary`, and `settings`.
#' @export
audit_gazepoint_distributional_drift <- function(dat,
                                                 signal_cols,
                                                 session_col = "session",
                                                 participant_col = NULL,
                                                 reference_session = NULL,
                                                 bins = 10,
                                                 psi_warn = 0.10,
                                                 psi_fail = 0.25) {
  if (!is.data.frame(dat)) stop("`dat` must be a data frame.", call. = FALSE)
  if (!session_col %in% names(dat)) stop("Column `", session_col, "` was not found.", call. = FALSE)

  missing_signals <- setdiff(signal_cols, names(dat))
  if (length(missing_signals) > 0) {
    stop("Missing `signal_cols`: ", paste(missing_signals, collapse = ", "), call. = FALSE)
  }

  non_numeric <- signal_cols[!vapply(dat[signal_cols], is.numeric, logical(1))]
  if (length(non_numeric) > 0) {
    stop("Non-numeric `signal_cols`: ", paste(non_numeric, collapse = ", "), call. = FALSE)
  }

  if (!is.null(participant_col) && !participant_col %in% names(dat)) {
    stop("Column `", participant_col, "` was not found.", call. = FALSE)
  }

  groups <- if (is.null(participant_col)) {
    list(all_participants = seq_len(nrow(dat)))
  } else {
    split(seq_len(nrow(dat)), dat[[participant_col]], drop = TRUE)
  }

  rows <- list()
  row_id <- 1L

  for (group_id in names(groups)) {
    idx_group <- groups[[group_id]]
    sessions <- unique(dat[[session_col]][idx_group])
    sessions <- sessions[order(sessions)]

    ref <- if (!is.null(reference_session)) reference_session else sessions[1]

    idx_ref <- idx_group[dat[[session_col]][idx_group] == ref]

    for (sess in sessions) {
      idx_sess <- idx_group[dat[[session_col]][idx_group] == sess]

      for (signal_col in signal_cols) {
        x_ref <- dat[[signal_col]][idx_ref]
        x_cur <- dat[[signal_col]][idx_sess]

        x_ref <- x_ref[is.finite(x_ref)]
        x_cur <- x_cur[is.finite(x_cur)]

        status <- "drift_audited"
        ks_p <- NA_real_
        psi <- NA_real_
        mean_difference <- NA_real_
        sd_ratio <- NA_real_

        if (length(x_ref) < 2 || length(x_cur) < 2) {
          status <- "insufficient_data"
        } else {
          mean_difference <- mean(x_cur) - mean(x_ref)
          sd_ratio <- if (stats::sd(x_ref) > 0) stats::sd(x_cur) / stats::sd(x_ref) else NA_real_
          ks_p <- tryCatch(stats::ks.test(x_ref, x_cur)$p.value, error = function(e) NA_real_)
          psi <- gpbiometrics_psi(x_ref, x_cur, bins = bins)

          if (is.finite(psi) && psi >= psi_fail) {
            status <- "drift_fail"
          } else if (is.finite(psi) && psi >= psi_warn) {
            status <- "drift_warn"
          }
        }

        rows[[row_id]] <- data.frame(
          group_id = as.character(group_id),
          reference_session = as.character(ref),
          comparison_session = as.character(sess),
          signal_col = signal_col,
          n_reference = length(x_ref),
          n_comparison = length(x_cur),
          reference_mean = if (length(x_ref) > 0) mean(x_ref) else NA_real_,
          comparison_mean = if (length(x_cur) > 0) mean(x_cur) else NA_real_,
          mean_difference = mean_difference,
          sd_ratio = sd_ratio,
          ks_p_value = ks_p,
          psi = psi,
          status = status,
          stringsAsFactors = FALSE
        )

        row_id <- row_id + 1L
      }
    }
  }

  drift_summary <- do.call(rbind, rows)
  rownames(drift_summary) <- NULL

  overview <- data.frame(
    drift_rows = nrow(drift_summary),
    signal_count = length(signal_cols),
    participant_groups = length(groups),
    warn_rows = sum(drift_summary$status == "drift_warn"),
    fail_rows = sum(drift_summary$status == "drift_fail"),
    insufficient_rows = sum(drift_summary$status == "insufficient_data"),
    status = if (any(drift_summary$status == "drift_fail")) {
      "distributional_drift_failures_detected"
    } else if (any(drift_summary$status == "drift_warn")) {
      "distributional_drift_warnings_detected"
    } else {
      "distributional_drift_audited"
    },
    interpretation = paste(
      "Distributional drift metrics flag changes in signal distributions across sessions or blocks.",
      "They support longitudinal QC and do not prove that data cannot be pooled without design-specific judgement."
    ),
    stringsAsFactors = FALSE
  )

  structure(
    list(
      overview = overview,
      drift_summary = drift_summary,
      settings = list(
        signal_cols = signal_cols,
        session_col = session_col,
        participant_col = participant_col,
        reference_session = reference_session,
        bins = bins,
        psi_warn = psi_warn,
        psi_fail = psi_fail
      )
    ),
    class = c("gazepoint_distributional_drift", "list")
  )
}

gpbiometrics_psi <- function(reference, current, bins = 10, eps = 1e-6) {
  reference <- reference[is.finite(reference)]
  current <- current[is.finite(current)]

  if (length(reference) < 2 || length(current) < 2) return(NA_real_)

  probs <- seq(0, 1, length.out = bins + 1)
  breaks <- unique(as.numeric(stats::quantile(reference, probs = probs, na.rm = TRUE, type = 7)))

  if (length(breaks) < 3) return(NA_real_)

  breaks[1] <- -Inf
  breaks[length(breaks)] <- Inf

  ref_tab <- table(cut(reference, breaks = breaks, include.lowest = TRUE))
  cur_tab <- table(cut(current, breaks = breaks, include.lowest = TRUE))

  ref_prop <- as.numeric(ref_tab) / sum(ref_tab)
  cur_prop <- as.numeric(cur_tab) / sum(cur_tab)

  ref_prop <- pmax(ref_prop, eps)
  cur_prop <- pmax(cur_prop, eps)

  sum((cur_prop - ref_prop) * log(cur_prop / ref_prop))
}

#' Extract refined composite multiscale entropy from HRV intervals
#'
#' Computes refined composite multiscale entropy (RCMSE) from IBI/RR intervals.
#' RCMSE pools template-match counts across all coarse-grained offsets at each
#' scale, making it more stable than ordinary MSE for shorter physiological
#' time series.
#'
#' @param dat A data frame containing IBI/RR intervals.
#' @param ibi_col Numeric IBI/RR interval column.
#' @param group_cols Optional grouping columns.
#' @param scales Positive integer scales.
#' @param m Embedding dimension.
#' @param r_multiplier Tolerance multiplier applied to SD.
#' @param min_intervals Minimum intervals per group.
#'
#' @return A list with `overview`, `rcmse_by_scale`, `summary`, and `settings`.
#' @export
extract_gazepoint_hrv_rcmse <- function(dat,
                                        ibi_col = "IBI",
                                        group_cols = NULL,
                                        scales = 1:10,
                                        m = 2,
                                        r_multiplier = 0.2,
                                        min_intervals = 20) {
  if (!is.data.frame(dat)) {
    stop("`dat` must be a data frame.", call. = FALSE)
  }

  if (!ibi_col %in% names(dat)) {
    stop("Column `", ibi_col, "` was not found in `dat`.", call. = FALSE)
  }

  if (!is.numeric(dat[[ibi_col]])) {
    stop("`ibi_col` must identify a numeric column.", call. = FALSE)
  }

  if (is.null(group_cols)) {
    group_cols <- character()
  }

  missing_groups <- setdiff(group_cols, names(dat))
  if (length(missing_groups) > 0) {
    stop("Missing `group_cols`: ", paste(missing_groups, collapse = ", "), call. = FALSE)
  }

  scales <- sort(unique(as.integer(scales)))

  if (length(scales) == 0 || any(scales < 1)) {
    stop("`scales` must contain positive integer-like values.", call. = FALSE)
  }

  groups <- gpbiometrics_rcmse_split(dat, group_cols)

  scale_rows <- list()
  summary_rows <- list()
  row_id <- 1L

  for (group_id in names(groups)) {
    idx <- groups[[group_id]]
    x <- dat[[ibi_col]][idx]
    x <- x[is.finite(x) & x > 0]

    if (length(x) < min_intervals || stats::sd(x) == 0) {
      for (scale in scales) {
        scale_rows[[row_id]] <- data.frame(
          group_id = group_id,
          scale = scale,
          rcmse = NA_real_,
          match_count_m = NA_real_,
          match_count_m1 = NA_real_,
          status = "insufficient_intervals",
          stringsAsFactors = FALSE
        )
        row_id <- row_id + 1L
      }

      summary_rows[[group_id]] <- data.frame(
        group_id = group_id,
        n_intervals = length(x),
        mean_rcmse = NA_real_,
        finite_scales = 0L,
        status = "insufficient_intervals",
        stringsAsFactors = FALSE
      )
      next
    }

    r <- r_multiplier * stats::sd(x)

    group_rcmse <- numeric(length(scales))

    for (i in seq_along(scales)) {
      scale <- scales[i]
      value <- gpbiometrics_rcmse_one_scale(
        x = x,
        scale = scale,
        m = m,
        r = r
      )

      group_rcmse[i] <- value$rcmse

      scale_rows[[row_id]] <- data.frame(
        group_id = group_id,
        scale = scale,
        rcmse = value$rcmse,
        match_count_m = value$count_m,
        match_count_m1 = value$count_m1,
        status = if (is.finite(value$rcmse)) {
          "rcmse_extracted"
        } else {
          "rcmse_not_estimated"
        },
        stringsAsFactors = FALSE
      )

      row_id <- row_id + 1L
    }

    summary_rows[[group_id]] <- data.frame(
      group_id = group_id,
      n_intervals = length(x),
      mean_rcmse = if (any(is.finite(group_rcmse))) {
        mean(group_rcmse, na.rm = TRUE)
      } else {
        NA_real_
      },
      finite_scales = sum(is.finite(group_rcmse)),
      status = if (any(is.finite(group_rcmse))) {
        "rcmse_extracted"
      } else {
        "rcmse_not_estimated"
      },
      stringsAsFactors = FALSE
    )
  }

  rcmse_by_scale <- do.call(rbind, scale_rows)
  rownames(rcmse_by_scale) <- NULL

  summary <- do.call(rbind, summary_rows)
  rownames(summary) <- NULL

  overview <- data.frame(
    group_count = length(groups),
    scale_rows = nrow(rcmse_by_scale),
    successful_groups = sum(summary$status == "rcmse_extracted"),
    problem_groups = sum(summary$status != "rcmse_extracted"),
    status = if (all(summary$status == "rcmse_extracted")) {
      "rcmse_extraction_complete"
    } else if (any(summary$status == "rcmse_extracted")) {
      "rcmse_extraction_partial"
    } else {
      "rcmse_extraction_failed"
    },
    interpretation = paste(
      "RCMSE describes multiscale irregularity in IBI/RR intervals.",
      "It is not a direct label for emotion, stress, cognition, health status, or diagnosis."
    ),
    stringsAsFactors = FALSE
  )

  structure(
    list(
      overview = overview,
      rcmse_by_scale = rcmse_by_scale,
      summary = summary,
      settings = list(
        ibi_col = ibi_col,
        group_cols = group_cols,
        scales = scales,
        m = m,
        r_multiplier = r_multiplier,
        min_intervals = min_intervals
      )
    ),
    class = c("gazepoint_hrv_rcmse", "list")
  )
}

gpbiometrics_rcmse_split <- function(dat, group_cols) {
  if (length(group_cols) == 0) {
    return(list(all_rows = seq_len(nrow(dat))))
  }

  gf <- dat[group_cols]
  gf[] <- lapply(gf, function(x) {
    x <- as.character(x)
    x[is.na(x)] <- "<NA>"
    x
  })

  split(seq_len(nrow(dat)), do.call(paste, c(gf, sep = " | ")))
}

gpbiometrics_rcmse_one_scale <- function(x, scale, m, r) {
  count_m <- 0
  count_m1 <- 0

  for (offset in seq_len(scale)) {
    coarse <- gpbiometrics_rcmse_coarse_grain_offset(x, scale, offset)

    if (length(coarse) > m + 2) {
      count_m <- count_m + gpbiometrics_rcmse_match_count(coarse, m, r)
      count_m1 <- count_m1 + gpbiometrics_rcmse_match_count(coarse, m + 1, r)
    }
  }

  rcmse <- if (count_m > 0 && count_m1 > 0) {
    -log(count_m1 / count_m)
  } else {
    NA_real_
  }

  list(
    rcmse = rcmse,
    count_m = count_m,
    count_m1 = count_m1
  )
}

gpbiometrics_rcmse_coarse_grain_offset <- function(x, scale, offset) {
  start <- offset
  values <- numeric()

  while (start <= length(x)) {
    stop_idx <- min(start + scale - 1, length(x))

    if ((stop_idx - start + 1) == scale) {
      values <- c(values, mean(x[start:stop_idx], na.rm = TRUE))
    }

    start <- start + scale
  }

  values
}

gpbiometrics_rcmse_match_count <- function(x, m, r) {
  x <- x[is.finite(x)]

  n <- length(x) - m + 1

  if (n <= 1 || !is.finite(r) || r <= 0) {
    return(0)
  }

  emb <- matrix(NA_real_, nrow = n, ncol = m)

  for (j in seq_len(m)) {
    emb[, j] <- x[j:(j + n - 1)]
  }

  count <- 0

  for (i in seq_len(n - 1)) {
    for (j in (i + 1):n) {
      if (max(abs(emb[i, ] - emb[j, ])) <= r) {
        count <- count + 1
      }
    }
  }

  count
}

#' Prepare Gazepoint biometric summaries for mixed-model analysis
#'
#' Prepares biometric window-level or event-level summaries for downstream
#' mixed-model analysis. This helper does not fit a model. It checks variables,
#' optionally baseline-corrects the selected outcome, optionally scales numeric
#' predictors, converts grouping/factor variables, flags complete cases, and
#' returns a conservative model formula.
#'
#' @param data A data frame containing biometric summary rows.
#' @param outcome_col Name of the outcome column to analyse.
#' @param fixed_effect_cols Optional fixed-effect predictor columns.
#' @param condition_cols Optional condition/design columns to include as fixed
#'   effects.
#' @param covariate_cols Optional covariate columns to include as fixed effects.
#' @param random_effect_cols Optional grouping columns for random intercepts.
#' @param participant_col,stimulus_col,trial_col Optional common grouping columns.
#' @param window_col Optional analysis-window column. Included as a fixed effect
#'   when `include_window = TRUE`.
#' @param baseline_col Optional baseline column.
#' @param baseline_correct Logical. If `TRUE`, creates an outcome column equal to
#'   `outcome_col - baseline_col`.
#' @param factor_cols Optional columns to convert to factors.
#' @param continuous_cols Optional numeric predictor columns to scale when
#'   `scale_continuous = TRUE`.
#' @param scale_continuous Logical. If `TRUE`, creates z-scored versions of
#'   numeric continuous predictors and uses those in the formula.
#' @param include_window Logical. Should `window_col` be included as a fixed
#'   effect?
#' @param drop_missing Logical. Should incomplete model rows be removed from
#'   `model_data`?
#' @param min_rows Minimum number of complete rows required for a `"ready"`
#'   status.
#'
#' @return A list with `overview`, `data`, `model_data`, `model_formula`,
#'   `variable_summary`, and `settings`.
#' @export
prepare_gazepoint_biometrics_lme_data <- function(data,
                                                  outcome_col,
                                                  fixed_effect_cols = NULL,
                                                  condition_cols = NULL,
                                                  covariate_cols = NULL,
                                                  random_effect_cols = NULL,
                                                  participant_col = NULL,
                                                  stimulus_col = NULL,
                                                  trial_col = NULL,
                                                  window_col = NULL,
                                                  baseline_col = NULL,
                                                  baseline_correct = FALSE,
                                                  factor_cols = NULL,
                                                  continuous_cols = NULL,
                                                  scale_continuous = FALSE,
                                                  include_window = TRUE,
                                                  drop_missing = TRUE,
                                                  min_rows = 10) {
  if (!is.data.frame(data)) {
    stop("`data` must be a data frame.", call. = FALSE)
  }

  if (missing(outcome_col) || length(outcome_col) != 1 || !is.character(outcome_col)) {
    stop("`outcome_col` must be a single column name.", call. = FALSE)
  }

  if (!outcome_col %in% names(data)) {
    stop("`outcome_col` was not found in `data`.", call. = FALSE)
  }

  if (!is.logical(baseline_correct) || length(baseline_correct) != 1) {
    stop("`baseline_correct` must be TRUE or FALSE.", call. = FALSE)
  }

  if (!is.logical(scale_continuous) || length(scale_continuous) != 1) {
    stop("`scale_continuous` must be TRUE or FALSE.", call. = FALSE)
  }

  if (!is.logical(include_window) || length(include_window) != 1) {
    stop("`include_window` must be TRUE or FALSE.", call. = FALSE)
  }

  if (!is.logical(drop_missing) || length(drop_missing) != 1) {
    stop("`drop_missing` must be TRUE or FALSE.", call. = FALSE)
  }

  if (!is.numeric(min_rows) || length(min_rows) != 1 || min_rows < 1) {
    stop("`min_rows` must be a positive number.", call. = FALSE)
  }

  dat <- as.data.frame(data, stringsAsFactors = FALSE)
  dat$.gpbiometrics_original_row_id <- seq_len(nrow(dat))

  all_requested <- unique(stats::na.omit(c(
    fixed_effect_cols,
    condition_cols,
    covariate_cols,
    random_effect_cols,
    participant_col,
    stimulus_col,
    trial_col,
    window_col,
    baseline_col,
    factor_cols,
    continuous_cols
  )))

  missing_requested <- setdiff(all_requested, names(dat))

  if (length(missing_requested) > 0) {
    stop(
      "The following requested columns were not found in `data`: ",
      paste(missing_requested, collapse = ", "),
      call. = FALSE
    )
  }

  outcome_numeric <- suppressWarnings(as.numeric(dat[[outcome_col]]))

  if (all(is.na(outcome_numeric))) {
    stop("`outcome_col` must contain numeric values.", call. = FALSE)
  }

  dat[[outcome_col]] <- outcome_numeric
  analysis_outcome_col <- outcome_col

  if (isTRUE(baseline_correct)) {
    if (is.null(baseline_col)) {
      stop("`baseline_col` must be supplied when `baseline_correct = TRUE`.", call. = FALSE)
    }

    baseline_numeric <- suppressWarnings(as.numeric(dat[[baseline_col]]))

    if (all(is.na(baseline_numeric))) {
      stop("`baseline_col` must contain numeric values.", call. = FALSE)
    }

    dat[[baseline_col]] <- baseline_numeric

    analysis_outcome_col <- paste0(outcome_col, "_baseline_corrected")
    dat[[analysis_outcome_col]] <- dat[[outcome_col]] - dat[[baseline_col]]
  }

  inferred_random <- gpbiometrics_lme_infer_random_cols(
    names_dat = names(dat),
    random_effect_cols = random_effect_cols,
    participant_col = participant_col,
    stimulus_col = stimulus_col,
    trial_col = trial_col
  )

  random_effect_cols <- inferred_random

  fixed_terms <- unique(stats::na.omit(c(
    fixed_effect_cols,
    condition_cols,
    covariate_cols,
    if (isTRUE(include_window)) window_col else NULL
  )))

  factor_cols <- unique(stats::na.omit(c(
    factor_cols,
    fixed_terms[!vapply(dat[fixed_terms], is.numeric, logical(1))],
    random_effect_cols
  )))

  for (col in factor_cols) {
    dat[[col]] <- as.factor(dat[[col]])
  }

  if (is.null(continuous_cols)) {
    continuous_cols <- fixed_terms[
      vapply(dat[fixed_terms], is.numeric, logical(1))
    ]
  }

  continuous_cols <- unique(stats::na.omit(continuous_cols))

  scaled_map <- data.frame(
    original_col = character(),
    scaled_col = character(),
    stringsAsFactors = FALSE
  )

  formula_fixed_terms <- fixed_terms

  if (isTRUE(scale_continuous) && length(continuous_cols) > 0) {
    for (col in continuous_cols) {
      x <- suppressWarnings(as.numeric(dat[[col]]))
      x_mean <- mean(x, na.rm = TRUE)
      x_sd <- stats::sd(x, na.rm = TRUE)

      scaled_col <- paste0("z_", make.names(col))

      if (!is.finite(x_sd) || x_sd == 0) {
        dat[[scaled_col]] <- NA_real_
      } else {
        dat[[scaled_col]] <- (x - x_mean) / x_sd
      }

      formula_fixed_terms[formula_fixed_terms == col] <- scaled_col

      scaled_map <- rbind(
        scaled_map,
        data.frame(
          original_col = col,
          scaled_col = scaled_col,
          stringsAsFactors = FALSE
        )
      )
    }
  }

  required_model_cols <- unique(stats::na.omit(c(
    analysis_outcome_col,
    formula_fixed_terms,
    random_effect_cols
  )))

  complete_case <- stats::complete.cases(dat[, required_model_cols, drop = FALSE])
  dat$lme_complete_case <- complete_case

  model_data <- dat

  if (isTRUE(drop_missing)) {
    model_data <- model_data[model_data$lme_complete_case, , drop = FALSE]
  }

  formula_rhs <- "1"

  if (length(formula_fixed_terms) > 0) {
    formula_rhs <- paste(
      vapply(formula_fixed_terms, gpbiometrics_lme_formula_name, character(1)),
      collapse = " + "
    )
  }

  if (length(random_effect_cols) > 0) {
    random_rhs <- paste0(
      "(1 | ",
      vapply(random_effect_cols, gpbiometrics_lme_formula_name, character(1)),
      ")"
    )

    formula_rhs <- paste(c(formula_rhs, random_rhs), collapse = " + ")
  }

  formula_text <- paste(
    gpbiometrics_lme_formula_name(analysis_outcome_col),
    "~",
    formula_rhs
  )

  model_formula <- stats::as.formula(formula_text)

  n_complete <- sum(complete_case)

  status <- if (nrow(dat) == 0) {
    "empty_input"
  } else if (n_complete == 0) {
    "no_complete_model_rows"
  } else if (n_complete < min_rows) {
    "limited_complete_rows"
  } else {
    "ready"
  }

  variable_summary <- gpbiometrics_lme_variable_summary(
    dat = dat,
    variables = unique(stats::na.omit(c(
      analysis_outcome_col,
      outcome_col,
      baseline_col,
      formula_fixed_terms,
      fixed_terms,
      random_effect_cols
    ))),
    analysis_outcome_col = analysis_outcome_col,
    fixed_terms = formula_fixed_terms,
    random_effect_cols = random_effect_cols,
    baseline_col = baseline_col
  )

  structure(
    list(
      overview = data.frame(
        input_rows = nrow(dat),
        complete_model_rows = n_complete,
        model_rows = nrow(model_data),
        outcome_col = outcome_col,
        analysis_outcome_col = analysis_outcome_col,
        fixed_effect_count = length(formula_fixed_terms),
        random_effect_count = length(random_effect_cols),
        status = status,
        stringsAsFactors = FALSE
      ),
      data = dat,
      model_data = model_data,
      model_formula = model_formula,
      variable_summary = variable_summary,
      settings = list(
        outcome_col = outcome_col,
        analysis_outcome_col = analysis_outcome_col,
        fixed_effect_cols = fixed_effect_cols,
        condition_cols = condition_cols,
        covariate_cols = covariate_cols,
        formula_fixed_terms = formula_fixed_terms,
        random_effect_cols = random_effect_cols,
        participant_col = participant_col,
        stimulus_col = stimulus_col,
        trial_col = trial_col,
        window_col = window_col,
        baseline_col = baseline_col,
        baseline_correct = baseline_correct,
        factor_cols = factor_cols,
        continuous_cols = continuous_cols,
        scaled_map = scaled_map,
        scale_continuous = scale_continuous,
        include_window = include_window,
        drop_missing = drop_missing,
        min_rows = min_rows,
        formula_text = formula_text
      )
    ),
    class = c("gazepoint_biometrics_lme_data", "list")
  )
}

gpbiometrics_lme_first_existing <- function(names_dat, candidates) {
  exact <- candidates[candidates %in% names_dat]

  if (length(exact) > 0) {
    return(exact[1])
  }

  lower_names <- tolower(names_dat)
  lower_candidates <- tolower(candidates)
  idx <- match(lower_candidates, lower_names)
  idx <- idx[!is.na(idx)]

  if (length(idx) > 0) {
    return(names_dat[idx[1]])
  }

  NULL
}

gpbiometrics_lme_infer_random_cols <- function(names_dat,
                                               random_effect_cols,
                                               participant_col,
                                               stimulus_col,
                                               trial_col) {
  explicit <- unique(stats::na.omit(c(
    random_effect_cols,
    participant_col,
    stimulus_col,
    trial_col
  )))

  if (length(explicit) > 0) {
    return(explicit)
  }

  participant <- gpbiometrics_lme_first_existing(
    names_dat,
    c("participant", "subject", "subject_id", "USER", "USER_FILE", "user_file")
  )

  stimulus <- gpbiometrics_lme_first_existing(
    names_dat,
    c("stimulus", "stimulus_id", "MEDIA_ID", "MEDIA_NAME", "media_id", "media_name")
  )

  trial <- gpbiometrics_lme_first_existing(
    names_dat,
    c("trial", "trial_id", "TRIAL", "trial_global")
  )

  unique(stats::na.omit(c(participant, stimulus, trial)))
}

gpbiometrics_lme_formula_name <- function(x) {
  syntactic <- make.names(x) == x
  reserved <- x %in% c(
    "if", "else", "repeat", "while", "function", "for",
    "in", "next", "break", "TRUE", "FALSE", "NULL",
    "Inf", "NaN", "NA", "NA_integer_", "NA_real_",
    "NA_complex_", "NA_character_"
  )

  if (isTRUE(syntactic) && !reserved) {
    return(x)
  }

  paste0("`", gsub("`", "\\\\`", x), "`")
}

gpbiometrics_lme_variable_summary <- function(dat,
                                              variables,
                                              analysis_outcome_col,
                                              fixed_terms,
                                              random_effect_cols,
                                              baseline_col) {
  out <- data.frame(
    variable = variables,
    role = "other",
    class = NA_character_,
    missing_count = NA_integer_,
    unique_count = NA_integer_,
    stringsAsFactors = FALSE
  )

  for (i in seq_len(nrow(out))) {
    variable <- out$variable[i]

    out$class[i] <- paste(class(dat[[variable]]), collapse = ";")
    out$missing_count[i] <- sum(is.na(dat[[variable]]))
    out$unique_count[i] <- length(unique(dat[[variable]][!is.na(dat[[variable]])]))

    if (variable == analysis_outcome_col) {
      out$role[i] <- "analysis_outcome"
    } else if (variable %in% fixed_terms) {
      out$role[i] <- "fixed_effect"
    } else if (variable %in% random_effect_cols) {
      out$role[i] <- "random_effect"
    } else if (!is.null(baseline_col) && variable == baseline_col) {
      out$role[i] <- "baseline"
    }
  }

  out
}

test_that("prepare_gazepoint_biometrics_lme_data prepares baseline-corrected model data", {
  dat <- data.frame(
    participant = rep(paste0("P", 1:6), each = 2),
    stimulus = rep(c("ad_a", "ad_b"), 6),
    trial = rep(1:2, 6),
    condition = rep(c("control", "claim"), 6),
    window_label = rep(c("baseline", "task"), 6),
    trial_order = seq_len(12),
    gsr_mean = seq(1, 2.1, length.out = 12),
    gsr_baseline = rep(seq(0.8, 1.3, length.out = 6), each = 2)
  )

  res <- prepare_gazepoint_biometrics_lme_data(
    dat,
    outcome_col = "gsr_mean",
    condition_cols = "condition",
    covariate_cols = "trial_order",
    participant_col = "participant",
    stimulus_col = "stimulus",
    window_col = "window_label",
    baseline_col = "gsr_baseline",
    baseline_correct = TRUE,
    scale_continuous = TRUE,
    min_rows = 4
  )

  expect_s3_class(res, "gazepoint_biometrics_lme_data")
  expect_equal(res$overview$status, "ready")
  expect_true("gsr_mean_baseline_corrected" %in% names(res$model_data))
  expect_true("z_trial_order" %in% names(res$model_data))
  expect_true(inherits(res$model_formula, "formula"))
  expect_equal(nrow(res$model_data), 12)
  expect_true(all(res$model_data$lme_complete_case))
})

test_that("prepare_gazepoint_biometrics_lme_data flags limited complete rows conservatively", {
  dat <- data.frame(
    participant = c("P1", "P1", "P2"),
    condition = c("a", "b", "a"),
    hr_mean = c(70, NA, 75)
  )

  res <- prepare_gazepoint_biometrics_lme_data(
    dat,
    outcome_col = "hr_mean",
    condition_cols = "condition",
    participant_col = "participant",
    min_rows = 3
  )

  expect_equal(res$overview$complete_model_rows, 2)
  expect_equal(res$overview$status, "limited_complete_rows")
  expect_equal(nrow(res$model_data), 2)
})

test_that("prepare_gazepoint_biometrics_lme_data keeps incomplete rows when requested", {
  dat <- data.frame(
    participant = c("P1", "P1", "P2"),
    condition = c("a", "b", "a"),
    hr_mean = c(70, NA, 75)
  )

  res <- prepare_gazepoint_biometrics_lme_data(
    dat,
    outcome_col = "hr_mean",
    condition_cols = "condition",
    participant_col = "participant",
    drop_missing = FALSE,
    min_rows = 1
  )

  expect_equal(nrow(res$model_data), 3)
  expect_equal(sum(res$model_data$lme_complete_case), 2)
})

test_that("prepare_gazepoint_biometrics_lme_data errors for missing outcome", {
  dat <- data.frame(
    participant = "P1",
    condition = "a",
    hr_mean = 70
  )

  expect_error(
    prepare_gazepoint_biometrics_lme_data(
      dat,
      outcome_col = "missing_outcome"
    ),
    "`outcome_col` was not found"
  )
})

#' Baseline-correct Gazepoint pupil size
#'
#' Applies subtractive or divisive baseline correction to pupil-size data within
#' trials or other grouping units.
#'
#' @param dat A data frame containing pupil-size data.
#' @param pupil_col Pupil column. If `NULL`, common Gazepoint pupil columns are
#'   detected.
#' @param time_col Time column.
#' @param stimulus_onset_col Optional stimulus-onset column. If supplied,
#'   baseline windows are interpreted relative to onset.
#' @param trial_cols Trial/grouping columns.
#' @param baseline_window Numeric vector of length two defining the baseline
#'   window relative to stimulus onset.
#' @param baseline_function `"median"` or `"mean"`.
#' @param correction `"subtract"` or `"divide"`.
#' @param suffix Suffix for corrected output column.
#' @param min_baseline_rows Minimum finite baseline rows required.
#' @param overwrite Logical. If `FALSE`, existing output columns are protected.
#'
#' @return A data frame with a corrected pupil column and baseline attributes.
#' @export
baseline_correct_gazepoint_pupil <- function(dat,
                                             pupil_col = NULL,
                                             time_col = "CNT",
                                             stimulus_onset_col = NULL,
                                             trial_cols = NULL,
                                             baseline_window = c(-240, -200),
                                             baseline_function = c("median", "mean"),
                                             correction = c("subtract", "divide"),
                                             suffix = "_baseline_corrected",
                                             min_baseline_rows = 2,
                                             overwrite = FALSE) {
  if (!is.data.frame(dat)) {
    stop("`dat` must be a data frame.", call. = FALSE)
  }

  baseline_function <- match.arg(baseline_function)
  correction <- match.arg(correction)

  pupil_col <- gpbiometrics_pupil_resolve_col(dat, pupil_col)

  if (!time_col %in% names(dat)) {
    stop("Column `", time_col, "` was not found in `dat`.", call. = FALSE)
  }

  if (!is.null(stimulus_onset_col) && !stimulus_onset_col %in% names(dat)) {
    stop("Column `", stimulus_onset_col, "` was not found in `dat`.", call. = FALSE)
  }

  if (is.null(trial_cols)) {
    trial_cols <- intersect(
      c("source_participant", "participant", "USER_FILE", "source_file", "MEDIA_ID", "trial", "trial_id"),
      names(dat)
    )
  }

  missing_trials <- setdiff(trial_cols, names(dat))

  if (length(missing_trials) > 0) {
    stop(
      "The following `trial_cols` were not found in `dat`: ",
      paste(missing_trials, collapse = ", "),
      call. = FALSE
    )
  }

  if (!is.numeric(baseline_window) || length(baseline_window) != 2 || baseline_window[1] >= baseline_window[2]) {
    stop("`baseline_window` must be a numeric vector of length two with start < end.", call. = FALSE)
  }

  output_col <- paste0(pupil_col, suffix)

  if (!isTRUE(overwrite) && output_col %in% names(dat)) {
    stop(
      "Output column `", output_col, "` already exists. Use `overwrite = TRUE` to replace it.",
      call. = FALSE
    )
  }

  out <- dat
  out[[output_col]] <- NA_real_

  groups <- gpbiometrics_pupil_split_indices(out, trial_cols)

  baseline_rows <- lapply(names(groups), function(unit_id) {
    idx <- groups[[unit_id]]

    if (!is.null(stimulus_onset_col)) {
      relative_time <- out[[time_col]][idx] - out[[stimulus_onset_col]][idx]
    } else {
      relative_time <- out[[time_col]][idx]
    }

    baseline_idx <- idx[
      is.finite(relative_time) &
        relative_time >= baseline_window[1] &
        relative_time <= baseline_window[2]
    ]

    baseline_values <- out[[pupil_col]][baseline_idx]
    baseline_values <- baseline_values[is.finite(baseline_values)]

    status <- "baseline_corrected"
    baseline_value <- NA_real_

    if (length(baseline_values) < min_baseline_rows) {
      status <- "insufficient_baseline_rows"
      out[[output_col]][idx] <<- NA_real_
    } else {
      baseline_value <- if (identical(baseline_function, "median")) {
        stats::median(baseline_values)
      } else {
        mean(baseline_values)
      }

      x <- out[[pupil_col]][idx]

      corrected <- if (identical(correction, "subtract")) {
        x - baseline_value
      } else {
        x / baseline_value
      }

      corrected[!is.finite(corrected)] <- NA_real_
      out[[output_col]][idx] <<- corrected
    }

    data.frame(
      unit_id = unit_id,
      pupil_col = pupil_col,
      output_col = output_col,
      n_rows = length(idx),
      n_baseline_rows = length(baseline_idx),
      n_finite_baseline_rows = length(baseline_values),
      baseline_value = baseline_value,
      status = status,
      stringsAsFactors = FALSE
    )
  })

  baseline_table <- do.call(rbind, baseline_rows)
  rownames(baseline_table) <- NULL

  summary <- data.frame(
    input_rows = nrow(out),
    trial_count = length(groups),
    corrected_trials = sum(baseline_table$status == "baseline_corrected"),
    problem_trials = sum(baseline_table$status != "baseline_corrected"),
    pupil_col = pupil_col,
    output_col = output_col,
    correction = correction,
    baseline_function = baseline_function,
    baseline_window = paste(baseline_window, collapse = " to "),
    status = if (all(baseline_table$status == "baseline_corrected")) {
      "pupil_baseline_correction_complete"
    } else if (any(baseline_table$status == "baseline_corrected")) {
      "pupil_baseline_correction_partial"
    } else {
      "pupil_baseline_correction_failed"
    },
    interpretation = paste(
      "Pupil baseline correction expresses pupil size relative to a trial-level reference period.",
      "It does not infer cognition, workload, emotion, preference, or diagnosis by itself."
    ),
    stringsAsFactors = FALSE
  )

  attr(out, "pupil_baseline_summary") <- summary
  attr(out, "pupil_baseline_table") <- baseline_table
  attr(out, "pupil_baseline_settings") <- list(
    pupil_col = pupil_col,
    time_col = time_col,
    stimulus_onset_col = stimulus_onset_col,
    trial_cols = trial_cols,
    baseline_window = baseline_window,
    baseline_function = baseline_function,
    correction = correction,
    suffix = suffix,
    min_baseline_rows = min_baseline_rows,
    overwrite = overwrite
  )

  class(out) <- unique(c("gazepoint_pupil_baseline_corrected", class(out)))

  out
}

gpbiometrics_pupil_resolve_col <- function(dat, pupil_col = NULL) {
  if (!is.null(pupil_col)) {
    if (!pupil_col %in% names(dat)) {
      stop("Column `", pupil_col, "` was not found in `dat`.", call. = FALSE)
    }

    if (!is.numeric(dat[[pupil_col]])) {
      stop("`pupil_col` must identify a numeric column.", call. = FALSE)
    }

    return(pupil_col)
  }

  candidates <- c("Pupil", "pupil", "pupil_size", "PUPIL", "LPMM", "RPMM", "LPD", "RPD")
  found <- intersect(candidates, names(dat))
  found <- found[vapply(dat[found], is.numeric, logical(1))]

  if (length(found) == 0) {
    stop("No common numeric Gazepoint pupil column was detected. Supply `pupil_col`.", call. = FALSE)
  }

  found[1]
}

gpbiometrics_pupil_split_indices <- function(dat, trial_cols) {
  if (length(trial_cols) == 0) {
    return(list(all_rows = seq_len(nrow(dat))))
  }

  group_frame <- dat[trial_cols]
  group_frame[] <- lapply(group_frame, function(x) {
    x <- as.character(x)
    x[is.na(x)] <- "<NA>"
    x
  })

  group_key <- do.call(paste, c(group_frame, sep = " | "))
  split(seq_len(nrow(dat)), group_key)
}

#' Plot Gazepoint saccade main-sequence diagnostics
#'
#' Plots saccade amplitude against peak velocity. The function expects
#' saccade-level amplitude and peak-velocity columns. If raw sample-level data
#' are supplied, users should first derive saccade-level kinematics with a
#' validated fixation/saccade detector.
#'
#' @param dat A saccade-level data frame.
#' @param amplitude_col Saccade amplitude column.
#' @param peak_velocity_col Peak velocity column.
#' @param group_col Optional grouping column.
#' @param log_axes Logical. If `TRUE`, use log10 axes.
#' @param add_smoother Logical. If `TRUE`, add a lowess curve.
#' @param main Plot title.
#'
#' @return Invisibly returns the plotted data and settings.
#' @export
plot_gazepoint_saccade_main_sequence <- function(dat,
                                                 amplitude_col = NULL,
                                                 peak_velocity_col = NULL,
                                                 group_col = NULL,
                                                 log_axes = TRUE,
                                                 add_smoother = TRUE,
                                                 main = "Gazepoint saccade main-sequence diagnostic") {
  if (!is.data.frame(dat)) {
    stop("`dat` must be a data frame.", call. = FALSE)
  }

  amplitude_col <- gpbiometrics_mainseq_resolve_col(
    dat,
    amplitude_col,
    c("saccade_amplitude", "amplitude", "amplitude_deg", "SACC_AMPLITUDE")
  )

  peak_velocity_col <- gpbiometrics_mainseq_resolve_col(
    dat,
    peak_velocity_col,
    c("peak_velocity", "peak_velocity_deg_s", "saccade_peak_velocity", "SACC_PEAK_VELOCITY")
  )

  if (!is.null(group_col) && !group_col %in% names(dat)) {
    stop("Column `", group_col, "` was not found in `dat`.", call. = FALSE)
  }

  plot_dat <- dat[
    is.finite(dat[[amplitude_col]]) &
      is.finite(dat[[peak_velocity_col]]) &
      dat[[amplitude_col]] > 0 &
      dat[[peak_velocity_col]] > 0,
    ,
    drop = FALSE
  ]

  if (nrow(plot_dat) == 0) {
    stop("No finite positive amplitude/peak-velocity rows are available.", call. = FALSE)
  }

  x <- plot_dat[[amplitude_col]]
  y <- plot_dat[[peak_velocity_col]]

  if (isTRUE(log_axes)) {
    x_plot <- log10(x)
    y_plot <- log10(y)
    xlab <- paste0("log10(", amplitude_col, ")")
    ylab <- paste0("log10(", peak_velocity_col, ")")
  } else {
    x_plot <- x
    y_plot <- y
    xlab <- amplitude_col
    ylab <- peak_velocity_col
  }

  graphics::plot(
    x_plot,
    y_plot,
    xlab = xlab,
    ylab = ylab,
    main = main,
    pch = 19,
    cex = 0.7
  )

  if (isTRUE(add_smoother) && nrow(plot_dat) >= 5) {
    fit <- stats::lowess(x_plot, y_plot)
    graphics::lines(fit, lwd = 2)
  }

  out <- list(
    data = plot_dat,
    settings = list(
      amplitude_col = amplitude_col,
      peak_velocity_col = peak_velocity_col,
      group_col = group_col,
      log_axes = log_axes,
      add_smoother = add_smoother
    ),
    interpretation = paste(
      "This is a saccade-kinematic quality diagnostic.",
      "It should be interpreted only when amplitude and peak velocity were derived from valid saccade-level data."
    )
  )

  invisible(out)
}

gpbiometrics_mainseq_resolve_col <- function(dat, supplied, candidates) {
  if (!is.null(supplied)) {
    if (!supplied %in% names(dat)) {
      stop("Column `", supplied, "` was not found in `dat`.", call. = FALSE)
    }

    if (!is.numeric(dat[[supplied]])) {
      stop("Column `", supplied, "` must be numeric.", call. = FALSE)
    }

    return(supplied)
  }

  found <- intersect(candidates, names(dat))
  found <- found[vapply(dat[found], is.numeric, logical(1))]

  if (length(found) == 0) {
    stop(
      "Could not detect a required saccade kinematic column. Supply it explicitly.",
      call. = FALSE
    )
  }

  found[1]
}

#' Filter implausible Gazepoint IBI values
#'
#' Flags and optionally cleans implausible inter-beat interval values. The helper
#' is conservative and does not remove rows; instead, it returns row-level flags
#' and a cleaned IBI column with implausible values set to `NA`.
#'
#' @param data A Gazepoint biometric data frame.
#' @param ibi_col IBI/RR interval column.
#' @param time_col Optional time/counter column.
#' @param group_cols Optional grouping columns.
#' @param validity_col Optional validity column.
#' @param unit Unit of the IBI column: `"auto"`, `"ms"`, or `"seconds"`.
#' @param min_ibi_ms Minimum plausible IBI in milliseconds.
#' @param max_ibi_ms Maximum plausible IBI in milliseconds.
#' @param max_change_ms Maximum plausible absolute adjacent IBI change within
#'   group, in milliseconds.
#' @param max_change_prop Maximum plausible proportional adjacent IBI change
#'   within group.
#' @param output_col Name of the cleaned IBI output column.
#'
#' @return A list with `overview`, `data`, `row_flags`, `group_summary`, and
#'   `settings`.
#' @export
filter_gazepoint_ibi_implausible <- function(data,
                                             ibi_col = "IBI",
                                             time_col = NULL,
                                             group_cols = NULL,
                                             validity_col = NULL,
                                             unit = c("auto", "ms", "seconds"),
                                             min_ibi_ms = 300,
                                             max_ibi_ms = 2000,
                                             max_change_ms = 400,
                                             max_change_prop = 0.30,
                                             output_col = "IBI_clean_ms") {
  unit <- match.arg(unit)

  if (!is.data.frame(data)) {
    stop("`data` must be a data frame.", call. = FALSE)
  }

  data <- as.data.frame(data, stringsAsFactors = FALSE)

  if (!ibi_col %in% names(data)) {
    stop("`ibi_col` was not found in `data`.", call. = FALSE)
  }

  if (!is.null(time_col) && !time_col %in% names(data)) {
    stop("`time_col` was not found in `data`.", call. = FALSE)
  }

  if (!is.null(validity_col) && !validity_col %in% names(data)) {
    stop("`validity_col` was not found in `data`.", call. = FALSE)
  }

  if (is.null(group_cols)) {
    group_cols <- gpbiometrics_ibi_core_resolve_group_cols(names(data))
  }

  missing_groups <- setdiff(group_cols, names(data))

  if (length(missing_groups) > 0) {
    stop(
      "`group_cols` were not found in `data`: ",
      paste(missing_groups, collapse = ", "),
      call. = FALSE
    )
  }

  gpbiometrics_ibi_core_validate_positive_scalar(min_ibi_ms, "min_ibi_ms")
  gpbiometrics_ibi_core_validate_positive_scalar(max_ibi_ms, "max_ibi_ms")
  gpbiometrics_ibi_core_validate_positive_scalar(max_change_ms, "max_change_ms")
  gpbiometrics_ibi_core_validate_positive_scalar(max_change_prop, "max_change_prop")

  if (max_ibi_ms <= min_ibi_ms) {
    stop("`max_ibi_ms` must be greater than `min_ibi_ms`.", call. = FALSE)
  }

  ibi_raw <- suppressWarnings(as.numeric(data[[ibi_col]]))
  detected_unit <- gpbiometrics_ibi_core_detect_unit(ibi_raw, unit)
  ibi_ms <- if (identical(detected_unit, "seconds")) ibi_raw * 1000 else ibi_raw

  group_id <- gpbiometrics_ibi_core_group_id(data, group_cols)

  flag_nonfinite <- !is.finite(ibi_ms)
  flag_nonpositive <- is.finite(ibi_ms) & ibi_ms <= 0
  flag_too_low <- is.finite(ibi_ms) & ibi_ms < min_ibi_ms
  flag_too_high <- is.finite(ibi_ms) & ibi_ms > max_ibi_ms

  flag_invalid_validity <- rep(FALSE, nrow(data))

  if (!is.null(validity_col)) {
    validity <- suppressWarnings(as.numeric(data[[validity_col]]))
    flag_invalid_validity <- !is.finite(validity) | validity == 0
  }

  change_flags <- gpbiometrics_ibi_core_change_flags(
    ibi_ms = ibi_ms,
    group_id = group_id,
    max_change_ms = max_change_ms,
    max_change_prop = max_change_prop
  )

  flag_implausible <- flag_nonfinite |
    flag_nonpositive |
    flag_too_low |
    flag_too_high |
    flag_invalid_validity |
    change_flags$flag_large_absolute_change |
    change_flags$flag_large_relative_change

  clean_ibi_ms <- ibi_ms
  clean_ibi_ms[flag_implausible] <- NA_real_

  out_data <- data
  out_data[[output_col]] <- clean_ibi_ms

  row_flags <- data.frame(
    row_id = seq_len(nrow(data)),
    group_id = group_id,
    ibi_raw = ibi_raw,
    ibi_ms = ibi_ms,
    ibi_clean_ms = clean_ibi_ms,
    flag_nonfinite = flag_nonfinite,
    flag_nonpositive = flag_nonpositive,
    flag_too_low = flag_too_low,
    flag_too_high = flag_too_high,
    flag_invalid_validity = flag_invalid_validity,
    flag_large_absolute_change = change_flags$flag_large_absolute_change,
    flag_large_relative_change = change_flags$flag_large_relative_change,
    flag_implausible = flag_implausible,
    stringsAsFactors = FALSE
  )

  if (length(group_cols) > 0) {
    row_flags <- cbind(data[group_cols], row_flags)
  }

  group_summary <- gpbiometrics_ibi_core_group_summary(
    row_flags = row_flags,
    group_cols = group_cols
  )

  overview <- data.frame(
    input_rows = nrow(data),
    ibi_col = ibi_col,
    output_col = output_col,
    detected_unit = detected_unit,
    implausible_rows = sum(flag_implausible, na.rm = TRUE),
    implausible_prop = mean(flag_implausible, na.rm = TRUE),
    clean_rows = sum(is.finite(clean_ibi_ms), na.rm = TRUE),
    group_count = length(unique(group_id)),
    status = if (all(!is.finite(clean_ibi_ms))) {
      "fail_no_clean_ibi_values"
    } else if (any(flag_implausible, na.rm = TRUE)) {
      "warn_implausible_ibi_detected"
    } else {
      "ibi_values_pass"
    },
    stringsAsFactors = FALSE
  )

  structure(
    list(
      overview = overview,
      data = out_data,
      row_flags = row_flags,
      group_summary = group_summary,
      settings = list(
        ibi_col = ibi_col,
        time_col = time_col,
        group_cols = group_cols,
        validity_col = validity_col,
        unit = unit,
        detected_unit = detected_unit,
        min_ibi_ms = min_ibi_ms,
        max_ibi_ms = max_ibi_ms,
        max_change_ms = max_change_ms,
        max_change_prop = max_change_prop,
        output_col = output_col,
        interpretation_notes = c(
          "IBI filtering flags implausible intervals but does not remove rows automatically.",
          "Thresholds should be reported and may require sensitivity checks.",
          "The Gazepoint HRV column should not be treated as genuine HRV unless independently documented."
        )
      )
    ),
    class = c("gazepoint_ibi_filter", "list")
  )
}

#' Compare Gazepoint HR and IBI-derived heart rate
#'
#' Compares recorded HR against HR derived from genuine IBI/RR intervals using
#' `60000 / IBI_ms`.
#'
#' @param data A Gazepoint biometric data frame or `gazepoint_ibi_filter` object.
#' @param hr_col Heart-rate column in beats per minute.
#' @param ibi_col IBI/RR interval column.
#' @param time_col Optional time/counter column.
#' @param group_cols Optional grouping columns.
#' @param unit Unit of the IBI column: `"auto"`, `"ms"`, or `"seconds"`.
#' @param max_abs_diff_bpm Maximum acceptable absolute HR difference in bpm.
#' @param max_rel_diff_prop Maximum acceptable relative HR difference.
#'
#' @return A list with `overview`, `row_diagnostics`, `group_summary`, and
#'   `settings`.
#' @export
compare_gazepoint_hr_ibi_consistency <- function(data,
                                                 hr_col = "HR",
                                                 ibi_col = "IBI",
                                                 time_col = NULL,
                                                 group_cols = NULL,
                                                 unit = c("auto", "ms", "seconds"),
                                                 max_abs_diff_bpm = 10,
                                                 max_rel_diff_prop = 0.15) {
  unit <- match.arg(unit)

  dat <- gpbiometrics_ibi_core_extract_data(data)

  if (!hr_col %in% names(dat)) {
    stop("`hr_col` was not found in `data`.", call. = FALSE)
  }

  if (!ibi_col %in% names(dat)) {
    stop("`ibi_col` was not found in `data`.", call. = FALSE)
  }

  if (!is.null(time_col) && !time_col %in% names(dat)) {
    stop("`time_col` was not found in `data`.", call. = FALSE)
  }

  if (is.null(group_cols)) {
    group_cols <- gpbiometrics_ibi_core_resolve_group_cols(names(dat))
  }

  missing_groups <- setdiff(group_cols, names(dat))

  if (length(missing_groups) > 0) {
    stop(
      "`group_cols` were not found in `data`: ",
      paste(missing_groups, collapse = ", "),
      call. = FALSE
    )
  }

  gpbiometrics_ibi_core_validate_positive_scalar(max_abs_diff_bpm, "max_abs_diff_bpm")
  gpbiometrics_ibi_core_validate_positive_scalar(max_rel_diff_prop, "max_rel_diff_prop")

  hr_observed <- suppressWarnings(as.numeric(dat[[hr_col]]))
  ibi_raw <- suppressWarnings(as.numeric(dat[[ibi_col]]))
  detected_unit <- gpbiometrics_ibi_core_detect_unit(ibi_raw, unit)
  ibi_ms <- if (identical(detected_unit, "seconds")) ibi_raw * 1000 else ibi_raw

  hr_from_ibi <- ifelse(is.finite(ibi_ms) & ibi_ms > 0, 60000 / ibi_ms, NA_real_)

  abs_diff_bpm <- abs(hr_observed - hr_from_ibi)
  rel_diff_prop <- abs_diff_bpm / hr_from_ibi

  flag_missing_pair <- !is.finite(hr_observed) | !is.finite(hr_from_ibi)
  flag_inconsistent <- !flag_missing_pair &
    (abs_diff_bpm > max_abs_diff_bpm | rel_diff_prop > max_rel_diff_prop)

  group_id <- gpbiometrics_ibi_core_group_id(dat, group_cols)

  row_diagnostics <- data.frame(
    row_id = seq_len(nrow(dat)),
    group_id = group_id,
    hr_observed_bpm = hr_observed,
    ibi_ms = ibi_ms,
    hr_from_ibi_bpm = hr_from_ibi,
    abs_diff_bpm = abs_diff_bpm,
    rel_diff_prop = rel_diff_prop,
    flag_missing_pair = flag_missing_pair,
    flag_inconsistent = flag_inconsistent,
    stringsAsFactors = FALSE
  )

  if (length(group_cols) > 0) {
    row_diagnostics <- cbind(dat[group_cols], row_diagnostics)
  }

  group_summary <- gpbiometrics_hr_ibi_consistency_group_summary(
    row_diagnostics = row_diagnostics,
    group_cols = group_cols
  )

  comparable_rows <- sum(!flag_missing_pair, na.rm = TRUE)
  inconsistent_rows <- sum(flag_inconsistent, na.rm = TRUE)

  overview <- data.frame(
    input_rows = nrow(dat),
    comparable_rows = comparable_rows,
    inconsistent_rows = inconsistent_rows,
    inconsistent_prop = if (comparable_rows > 0) inconsistent_rows / comparable_rows else NA_real_,
    detected_ibi_unit = detected_unit,
    mean_abs_diff_bpm = mean(abs_diff_bpm, na.rm = TRUE),
    median_abs_diff_bpm = stats::median(abs_diff_bpm, na.rm = TRUE),
    group_count = length(unique(group_id)),
    status = if (comparable_rows == 0) {
      "fail_no_comparable_hr_ibi_rows"
    } else if (inconsistent_rows > 0) {
      "warn_hr_ibi_inconsistency_detected"
    } else {
      "hr_ibi_consistency_pass"
    },
    stringsAsFactors = FALSE
  )

  structure(
    list(
      overview = overview,
      row_diagnostics = row_diagnostics,
      group_summary = group_summary,
      settings = list(
        hr_col = hr_col,
        ibi_col = ibi_col,
        time_col = time_col,
        group_cols = group_cols,
        unit = unit,
        detected_ibi_unit = detected_unit,
        max_abs_diff_bpm = max_abs_diff_bpm,
        max_rel_diff_prop = max_rel_diff_prop,
        interpretation_notes = c(
          "HR derived from IBI is computed as 60000 / IBI_ms.",
          "Discrepancies may reflect smoothing, device timing, missing intervals, or artifacts.",
          "Consistency diagnostics should be used as quality checks rather than automatic exclusions."
        )
      )
    ),
    class = c("gazepoint_hr_ibi_consistency", "list")
  )
}

#' Extract time-domain HRV features from Gazepoint IBI intervals
#'
#' Computes simple time-domain HRV features from genuine IBI/RR intervals. This
#' helper does not use the Gazepoint `HRV` column as an HRV outcome.
#'
#' @param data A Gazepoint biometric data frame or `gazepoint_ibi_filter` object.
#' @param ibi_col IBI/RR interval column.
#' @param group_cols Optional grouping columns.
#' @param unit Unit of the IBI column: `"auto"`, `"ms"`, or `"seconds"`.
#' @param min_intervals Minimum clean intervals required per group.
#' @param min_duration_s Minimum IBI-sequence duration in seconds required before
#'   computed HRV features are treated as fully reportable. Groups below this
#'   duration still return features but receive `warn_short_hrv_duration`.
#' @param diff_threshold_ms Threshold for NN50/pNN50.
#' @param collapse_repeated_intervals Logical. If `TRUE`, consecutive repeated
#'   IBI values are collapsed before HRV features are computed. This is useful
#'   for Gazepoint exports where the same IBI value may be repeated across
#'   multiple gaze-sampling rows until a new interval is available.
#' @param repeated_tolerance_ms Numeric tolerance used when identifying
#'   repeated consecutive IBI values.
#'
#' @return A list with `overview`, `features`, `settings`.
#' @export
extract_gazepoint_hrv_features <- function(data,
                                           ibi_col = "IBI_clean_ms",
                                           group_cols = NULL,
                                           unit = c("auto", "ms", "seconds"),
                                           min_intervals = 3,
                                           min_duration_s = 30,
                                           diff_threshold_ms = 50,
                                           collapse_repeated_intervals = TRUE,
                                           repeated_tolerance_ms = 1e-8) {
  unit <- match.arg(unit)

  dat <- gpbiometrics_ibi_core_extract_data(data)

  if (!ibi_col %in% names(dat)) {
    stop("`ibi_col` was not found in `data`.", call. = FALSE)
  }

  if (is.null(group_cols)) {
    group_cols <- gpbiometrics_ibi_core_resolve_group_cols(names(dat))
  }

  missing_groups <- setdiff(group_cols, names(dat))

  if (length(missing_groups) > 0) {
    stop(
      "`group_cols` were not found in `data`: ",
      paste(missing_groups, collapse = ", "),
      call. = FALSE
    )
  }

  gpbiometrics_ibi_core_validate_positive_scalar(min_intervals, "min_intervals")

  if (!is.numeric(min_duration_s) ||
      length(min_duration_s) != 1 ||
      !is.finite(min_duration_s) ||
      min_duration_s < 0) {
    stop("`min_duration_s` must be a single non-negative finite number.", call. = FALSE)
  }

  gpbiometrics_ibi_core_validate_positive_scalar(diff_threshold_ms, "diff_threshold_ms")

  if (!is.logical(collapse_repeated_intervals) ||
      length(collapse_repeated_intervals) != 1 ||
      is.na(collapse_repeated_intervals)) {
    stop("`collapse_repeated_intervals` must be TRUE or FALSE.", call. = FALSE)
  }

  if (!is.numeric(repeated_tolerance_ms) ||
      length(repeated_tolerance_ms) != 1 ||
      !is.finite(repeated_tolerance_ms) ||
      repeated_tolerance_ms < 0) {
    stop("`repeated_tolerance_ms` must be a single non-negative finite number.", call. = FALSE)
  }

  ibi_raw <- suppressWarnings(as.numeric(dat[[ibi_col]]))
  detected_unit <- gpbiometrics_ibi_core_detect_unit(ibi_raw, unit)
  ibi_ms <- if (identical(detected_unit, "seconds")) ibi_raw * 1000 else ibi_raw

  dat$.ibi_hrv_ms <- ibi_ms
  dat$.group_id <- gpbiometrics_ibi_core_group_id(dat, group_cols)

  group_ids <- unique(dat$.group_id)

  features <- lapply(group_ids, function(group_id) {
    d <- dat[dat$.group_id == group_id, , drop = FALSE]
    x_raw <- d$.ibi_hrv_ms
    x_raw <- x_raw[is.finite(x_raw) & x_raw > 0]

    x <- if (isTRUE(collapse_repeated_intervals)) {
      gpbiometrics_hrv_core_collapse_repeated_intervals(
        ibi_ms = x_raw,
        tolerance_ms = repeated_tolerance_ms
      )
    } else {
      x_raw
    }

    row <- gpbiometrics_hrv_core_features(
      ibi_ms = x,
      min_intervals = min_intervals,
      min_duration_s = min_duration_s,
      diff_threshold_ms = diff_threshold_ms
    )

    row$input_interval_rows <- length(x_raw)
    row$used_intervals_after_collapse <- length(x)
    row$collapsed_repeated_intervals <- isTRUE(collapse_repeated_intervals)
    row$group_id <- group_id

    if (length(group_cols) > 0 && all(group_cols %in% names(d))) {
      row <- cbind(d[1, group_cols, drop = FALSE], row)
    }

    row
  })

  features <- do.call(rbind, features)
  rownames(features) <- NULL

  valid_feature_groups <- sum(
    features$feature_status %in% c(
      "features_computed",
      "warn_short_hrv_duration"
    ),
    na.rm = TRUE
  )

  short_duration_groups <- sum(
    features$feature_status == "warn_short_hrv_duration",
    na.rm = TRUE
  )

  overview <- data.frame(
    group_count = nrow(features),
    valid_feature_groups = valid_feature_groups,
    insufficient_interval_groups = sum(features$feature_status == "insufficient_intervals", na.rm = TRUE),
    short_duration_groups = short_duration_groups,
    detected_ibi_unit = detected_unit,
    min_intervals = min_intervals,
    min_duration_s = min_duration_s,
    diff_threshold_ms = diff_threshold_ms,
    status = if (valid_feature_groups == 0) {
      "fail_no_hrv_features_computed"
    } else if (valid_feature_groups < nrow(features)) {
      "warn_some_groups_insufficient_intervals"
    } else if (short_duration_groups > 0) {
      "warn_short_hrv_duration"
    } else {
      "hrv_features_computed"
    },
    stringsAsFactors = FALSE
  )

  structure(
    list(
      overview = overview,
      features = features,
      settings = list(
        ibi_col = ibi_col,
        group_cols = group_cols,
        unit = unit,
        detected_ibi_unit = detected_unit,
        min_intervals = min_intervals,
        min_duration_s = min_duration_s,
        diff_threshold_ms = diff_threshold_ms,
        collapse_repeated_intervals = collapse_repeated_intervals,
        repeated_tolerance_ms = repeated_tolerance_ms,
        interpretation_notes = c(
          "Features are computed from genuine IBI/RR intervals only.",
          "The Gazepoint HRV column is not used as an HRV outcome.",
          "Short windows provide unstable HRV estimates and should be interpreted cautiously."
        )
      )
    ),
    class = c("gazepoint_hrv_feature_extraction", "list")
  )
}

gpbiometrics_ibi_core_extract_data <- function(data) {
  if (inherits(data, "gazepoint_ibi_filter") && !is.null(data$data)) {
    return(as.data.frame(data$data, stringsAsFactors = FALSE))
  }

  if (is.data.frame(data)) {
    return(as.data.frame(data, stringsAsFactors = FALSE))
  }

  stop("`data` must be a data frame or a `gazepoint_ibi_filter` object.", call. = FALSE)
}

gpbiometrics_ibi_core_resolve_group_cols <- function(names_dat) {
  candidates <- c(
    "source_file",
    "source_participant",
    "USER",
    "USER_FILE",
    "participant",
    "subject",
    "subject_id",
    "MEDIA_ID",
    "MEDIA_NAME",
    "media_id",
    "media_name",
    "trial",
    "trial_id",
    "trial_global"
  )

  unique(candidates[candidates %in% names_dat])
}

gpbiometrics_ibi_core_group_id <- function(dat, group_cols) {
  if (length(group_cols) == 0) {
    return(rep("all", nrow(dat)))
  }

  group_dat <- dat[group_cols]

  group_dat[] <- lapply(group_dat, function(x) {
    x_chr <- as.character(x)
    x_chr[is.na(x_chr)] <- "<NA>"
    x_chr
  })

  apply(group_dat, 1, paste, collapse = "||")
}

gpbiometrics_ibi_core_detect_unit <- function(x, unit) {
  if (!identical(unit, "auto")) {
    return(unit)
  }

  finite_x <- x[is.finite(x) & x > 0]

  if (length(finite_x) == 0) {
    return("ms")
  }

  med <- stats::median(finite_x, na.rm = TRUE)

  if (is.finite(med) && med > 0.2 && med < 5) {
    "seconds"
  } else {
    "ms"
  }
}

gpbiometrics_ibi_core_validate_positive_scalar <- function(x, name) {
  if (!is.numeric(x) || length(x) != 1 || !is.finite(x) || x <= 0) {
    stop("`", name, "` must be a single positive finite number.", call. = FALSE)
  }

  invisible(TRUE)
}

gpbiometrics_ibi_core_change_flags <- function(ibi_ms,
                                               group_id,
                                               max_change_ms,
                                               max_change_prop) {
  flag_abs <- rep(FALSE, length(ibi_ms))
  flag_rel <- rep(FALSE, length(ibi_ms))

  groups <- unique(group_id)

  for (g in groups) {
    idx <- which(group_id == g)

    if (length(idx) < 2) {
      next
    }

    x <- ibi_ms[idx]
    prev <- c(NA_real_, utils::head(x, -1))
    diff_abs <- abs(x - prev)
    diff_rel <- diff_abs / prev

    local_abs <- is.finite(diff_abs) & diff_abs > max_change_ms
    local_rel <- is.finite(diff_rel) & diff_rel > max_change_prop

    flag_abs[idx] <- local_abs
    flag_rel[idx] <- local_rel
  }

  list(
    flag_large_absolute_change = flag_abs,
    flag_large_relative_change = flag_rel
  )
}

gpbiometrics_ibi_core_group_summary <- function(row_flags,
                                                group_cols) {
  if (nrow(row_flags) == 0) {
    return(data.frame())
  }

  group_ids <- unique(row_flags$group_id)

  out <- lapply(group_ids, function(group_id) {
    d <- row_flags[row_flags$group_id == group_id, , drop = FALSE]
    clean <- d$ibi_clean_ms[is.finite(d$ibi_clean_ms)]

    row <- data.frame(
      group_id = group_id,
      rows = nrow(d),
      implausible_rows = sum(d$flag_implausible, na.rm = TRUE),
      implausible_prop = mean(d$flag_implausible, na.rm = TRUE),
      clean_rows = length(clean),
      mean_clean_ibi_ms = if (length(clean) > 0) mean(clean, na.rm = TRUE) else NA_real_,
      median_clean_ibi_ms = if (length(clean) > 0) stats::median(clean, na.rm = TRUE) else NA_real_,
      stringsAsFactors = FALSE
    )

    if (length(group_cols) > 0 && all(group_cols %in% names(d))) {
      row <- cbind(d[1, group_cols, drop = FALSE], row)
    }

    row
  })

  out <- do.call(rbind, out)
  rownames(out) <- NULL

  out
}

gpbiometrics_hr_ibi_consistency_group_summary <- function(row_diagnostics,
                                                          group_cols) {
  group_ids <- unique(row_diagnostics$group_id)

  out <- lapply(group_ids, function(group_id) {
    d <- row_diagnostics[row_diagnostics$group_id == group_id, , drop = FALSE]
    comparable <- !d$flag_missing_pair

    row <- data.frame(
      group_id = group_id,
      rows = nrow(d),
      comparable_rows = sum(comparable, na.rm = TRUE),
      inconsistent_rows = sum(d$flag_inconsistent, na.rm = TRUE),
      inconsistent_prop = if (sum(comparable, na.rm = TRUE) > 0) {
        sum(d$flag_inconsistent, na.rm = TRUE) / sum(comparable, na.rm = TRUE)
      } else {
        NA_real_
      },
      median_abs_diff_bpm = stats::median(d$abs_diff_bpm, na.rm = TRUE),
      mean_abs_diff_bpm = mean(d$abs_diff_bpm, na.rm = TRUE),
      stringsAsFactors = FALSE
    )

    if (length(group_cols) > 0 && all(group_cols %in% names(d))) {
      row <- cbind(d[1, group_cols, drop = FALSE], row)
    }

    row
  })

  out <- do.call(rbind, out)
  rownames(out) <- NULL

  out
}

gpbiometrics_hrv_core_collapse_repeated_intervals <- function(ibi_ms,
                                                              tolerance_ms) {
  if (length(ibi_ms) <= 1) {
    return(ibi_ms)
  }

  diff_value <- abs(diff(ibi_ms))
  keep <- c(TRUE, diff_value > tolerance_ms)

  ibi_ms[keep]
}

gpbiometrics_hrv_core_features <- function(ibi_ms,
                                           min_intervals,
                                           min_duration_s,
                                           diff_threshold_ms) {
  n_intervals <- length(ibi_ms)

  if (n_intervals < min_intervals) {
    return(data.frame(
      n_intervals = n_intervals,
      duration_s = if (n_intervals > 0) sum(ibi_ms, na.rm = TRUE) / 1000 else NA_real_,
      min_duration_s = min_duration_s,
      mean_ibi_ms = NA_real_,
      median_ibi_ms = NA_real_,
      mean_hr_bpm = NA_real_,
      sdnn_ms = NA_real_,
      rmssd_ms = NA_real_,
      sdsd_ms = NA_real_,
      nn50 = NA_integer_,
      pnn50 = NA_real_,
      cvnn = NA_real_,
      min_ibi_ms = if (n_intervals > 0) min(ibi_ms, na.rm = TRUE) else NA_real_,
      max_ibi_ms = if (n_intervals > 0) max(ibi_ms, na.rm = TRUE) else NA_real_,
      feature_status = "insufficient_intervals",
      stringsAsFactors = FALSE
    ))
  }

  diff_ibi <- diff(ibi_ms)

  duration_s <- sum(ibi_ms, na.rm = TRUE) / 1000

  data.frame(
    n_intervals = n_intervals,
    duration_s = duration_s,
    min_duration_s = min_duration_s,
    mean_ibi_ms = mean(ibi_ms, na.rm = TRUE),
    median_ibi_ms = stats::median(ibi_ms, na.rm = TRUE),
    mean_hr_bpm = mean(60000 / ibi_ms, na.rm = TRUE),
    sdnn_ms = if (n_intervals > 1) stats::sd(ibi_ms, na.rm = TRUE) else NA_real_,
    rmssd_ms = if (length(diff_ibi) > 0) sqrt(mean(diff_ibi^2, na.rm = TRUE)) else NA_real_,
    sdsd_ms = if (length(diff_ibi) > 1) stats::sd(diff_ibi, na.rm = TRUE) else NA_real_,
    nn50 = sum(abs(diff_ibi) > diff_threshold_ms, na.rm = TRUE),
    pnn50 = if (length(diff_ibi) > 0) {
      sum(abs(diff_ibi) > diff_threshold_ms, na.rm = TRUE) / length(diff_ibi)
    } else {
      NA_real_
    },
    cvnn = stats::sd(ibi_ms, na.rm = TRUE) / mean(ibi_ms, na.rm = TRUE),
    min_ibi_ms = min(ibi_ms, na.rm = TRUE),
    max_ibi_ms = max(ibi_ms, na.rm = TRUE),
    feature_status = if (is.finite(duration_s) && duration_s < min_duration_s) {
      "warn_short_hrv_duration"
    } else {
      "features_computed"
    },
    stringsAsFactors = FALSE
  )
}

test_that("filter_gazepoint_ibi_implausible flags implausible IBI values", {
  dat <- data.frame(
    participant = "P1",
    IBI = c(1000, 1020, 250, 1050, 2500, 1080)
  )

  res <- filter_gazepoint_ibi_implausible(
    dat,
    ibi_col = "IBI",
    group_cols = "participant",
    min_ibi_ms = 300,
    max_ibi_ms = 2000,
    max_change_ms = 500,
    max_change_prop = 0.5
  )

  expect_s3_class(res, "gazepoint_ibi_filter")
  expect_equal(res$overview$detected_unit, "ms")
  expect_true(any(res$row_flags$flag_too_low))
  expect_true(any(res$row_flags$flag_too_high))
  expect_true("IBI_clean_ms" %in% names(res$data))
})

test_that("filter_gazepoint_ibi_implausible detects seconds automatically", {
  dat <- data.frame(
    participant = "P1",
    IBI = c(1.0, 1.1, 0.9)
  )

  res <- filter_gazepoint_ibi_implausible(
    dat,
    ibi_col = "IBI",
    group_cols = "participant"
  )

  expect_equal(res$overview$detected_unit, "seconds")
  expect_equal(res$row_flags$ibi_ms, c(1000, 1100, 900))
})

test_that("filter_gazepoint_ibi_implausible respects validity flags", {
  dat <- data.frame(
    participant = "P1",
    IBI = c(1000, 1000, 1000),
    IBIV = c(1, 0, 1)
  )

  res <- filter_gazepoint_ibi_implausible(
    dat,
    ibi_col = "IBI",
    validity_col = "IBIV",
    group_cols = "participant"
  )

  expect_equal(sum(res$row_flags$flag_invalid_validity), 1)
  expect_equal(sum(is.na(res$data$IBI_clean_ms)), 1)
})

test_that("compare_gazepoint_hr_ibi_consistency detects matching and mismatching rows", {
  dat <- data.frame(
    participant = "P1",
    HR = c(60, 60, 100),
    IBI = c(1000, 1000, 1000)
  )

  res <- compare_gazepoint_hr_ibi_consistency(
    dat,
    hr_col = "HR",
    ibi_col = "IBI",
    group_cols = "participant",
    max_abs_diff_bpm = 10
  )

  expect_s3_class(res, "gazepoint_hr_ibi_consistency")
  expect_equal(res$overview$comparable_rows, 3)
  expect_equal(res$overview$inconsistent_rows, 1)
  expect_true(any(res$row_diagnostics$flag_inconsistent))
})

test_that("compare_gazepoint_hr_ibi_consistency accepts filtered IBI objects", {
  dat <- data.frame(
    participant = "P1",
    HR = c(60, 55),
    IBI = c(1000, 1090)
  )

  filtered <- filter_gazepoint_ibi_implausible(
    dat,
    ibi_col = "IBI",
    group_cols = "participant"
  )

  res <- compare_gazepoint_hr_ibi_consistency(
    filtered,
    hr_col = "HR",
    ibi_col = "IBI_clean_ms",
    group_cols = "participant"
  )

  expect_equal(res$overview$comparable_rows, 2)
})

test_that("extract_gazepoint_hrv_features collapses repeated IBI values by default", {
  dat <- data.frame(
    participant = "P1",
    IBI_clean_ms = c(
      1000, 1000, 1000,
      1020, 1020,
      980, 980,
      1010
    )
  )

  res <- extract_gazepoint_hrv_features(
    dat,
    ibi_col = "IBI_clean_ms",
    group_cols = "participant",
    min_intervals = 3,
    min_duration_s = 0
  )

  expect_equal(res$features$input_interval_rows, 8)
  expect_equal(res$features$used_intervals_after_collapse, 4)
  expect_true(res$features$collapsed_repeated_intervals)
  expect_equal(res$features$n_intervals, 4)
  expect_equal(res$features$feature_status, "features_computed")
})

test_that("extract_gazepoint_hrv_features can retain repeated IBI values", {
  dat <- data.frame(
    participant = "P1",
    IBI_clean_ms = c(
      1000, 1000, 1000,
      1020, 1020,
      980, 980,
      1010
    )
  )

  res <- extract_gazepoint_hrv_features(
    dat,
    ibi_col = "IBI_clean_ms",
    group_cols = "participant",
    min_intervals = 3,
    min_duration_s = 0,
    collapse_repeated_intervals = FALSE
  )

  expect_equal(res$features$input_interval_rows, 8)
  expect_equal(res$features$used_intervals_after_collapse, 8)
  expect_false(res$features$collapsed_repeated_intervals)
  expect_equal(res$features$n_intervals, 8)
  expect_equal(res$features$feature_status, "features_computed")
})

test_that("extract_gazepoint_hrv_features validates repeated interval options", {
  dat <- data.frame(
    participant = "P1",
    IBI_clean_ms = c(1000, 1010, 990)
  )

  expect_error(
    extract_gazepoint_hrv_features(
      dat,
      ibi_col = "IBI_clean_ms",
      collapse_repeated_intervals = NA
    ),
    "`collapse_repeated_intervals` must be TRUE or FALSE"
  )

  expect_error(
    extract_gazepoint_hrv_features(
      dat,
      ibi_col = "IBI_clean_ms",
      repeated_tolerance_ms = -1
    ),
    "`repeated_tolerance_ms` must be a single non-negative finite number"
  )
})

test_that("extract_gazepoint_hrv_features computes time-domain features", {
  dat <- data.frame(
    participant = "P1",
    IBI_clean_ms = c(1000, 1020, 980, 1010, 990)
  )

  res <- extract_gazepoint_hrv_features(
    dat,
    ibi_col = "IBI_clean_ms",
    group_cols = "participant",
    min_intervals = 3,
    min_duration_s = 0
  )

  expect_s3_class(res, "gazepoint_hrv_feature_extraction")
  expect_equal(res$overview$status, "hrv_features_computed")
  expect_equal(res$features$n_intervals, 5)
  expect_true(is.finite(res$features$sdnn_ms))
  expect_true(is.finite(res$features$rmssd_ms))
})

test_that("extract_gazepoint_hrv_features warns for short HRV duration", {
  dat <- data.frame(
    participant = "P1",
    IBI_clean_ms = c(1000, 1020, 980, 1010, 990)
  )

  res <- extract_gazepoint_hrv_features(
    dat,
    ibi_col = "IBI_clean_ms",
    group_cols = "participant",
    min_intervals = 3,
    min_duration_s = 30
  )

  expect_equal(res$overview$status, "warn_short_hrv_duration")
  expect_equal(res$overview$short_duration_groups, 1)
  expect_equal(res$features$feature_status, "warn_short_hrv_duration")
  expect_true(is.finite(res$features$rmssd_ms))
})

test_that("extract_gazepoint_hrv_features reports insufficient intervals", {
  dat <- data.frame(
    participant = "P1",
    IBI_clean_ms = c(1000, 1020)
  )

  res <- extract_gazepoint_hrv_features(
    dat,
    ibi_col = "IBI_clean_ms",
    group_cols = "participant",
    min_intervals = 3
  )

  expect_equal(res$overview$status, "fail_no_hrv_features_computed")
  expect_equal(res$features$feature_status, "insufficient_intervals")
})

test_that("extract_gazepoint_hrv_features works by group", {
  dat <- data.frame(
    participant = c(rep("P1", 4), rep("P2", 4)),
    IBI_clean_ms = c(1000, 1010, 990, 1005, 800, 810, 790, 805)
  )

  res <- extract_gazepoint_hrv_features(
    dat,
    ibi_col = "IBI_clean_ms",
    group_cols = "participant",
    min_intervals = 3,
    min_duration_s = 0
  )

  expect_equal(nrow(res$features), 2)
  expect_true(all(res$features$feature_status == "features_computed"))
})

test_that("IBI/HR/HRV helpers validate inputs", {
  dat <- data.frame(
    IBI = c(1000, 1010)
  )

  expect_error(
    filter_gazepoint_ibi_implausible(dat, ibi_col = "missing"),
    "`ibi_col`"
  )

  expect_error(
    compare_gazepoint_hr_ibi_consistency(dat),
    "`hr_col`"
  )

  expect_error(
    extract_gazepoint_hrv_features(dat, ibi_col = "missing"),
    "`ibi_col`"
  )

  expect_error(
    extract_gazepoint_hrv_features(
      dat,
      ibi_col = "IBI",
      min_duration_s = -1
    ),
    "`min_duration_s` must be a single non-negative finite number"
  )
})

#' Screen Gazepoint EDA/SCR nonresponders
#'
#' Screens groups, usually participants, for low or absent SCR responding using
#' either SCR event-window summaries or SCR peak-detection outputs. The helper
#' reports candidate nonresponders under explicit, user-controlled criteria.
#'
#' @param x A `gazepoint_scr_event_window_summary`,
#'   `gazepoint_scr_peak_detection`, or data frame.
#' @param group_cols Optional grouping columns, usually participant columns.
#' @param response_col Binary response column for event-window data.
#' @param amplitude_col SCR amplitude column.
#' @param min_events Minimum number of events required before event-window
#'   nonresponder classification is made.
#' @param min_response_events Minimum number of response events required.
#' @param min_response_rate Minimum response rate required.
#' @param min_detected_peaks Minimum detected peaks required when only peak data
#'   are available.
#'
#' @return A list with `overview`, `group_summary`, `candidate_nonresponders`,
#'   and `settings`.
#' @export
screen_gazepoint_eda_nonresponders <- function(x,
                                               group_cols = NULL,
                                               response_col = "response_flag",
                                               amplitude_col = "scr_amplitude",
                                               min_events = 1,
                                               min_response_events = 1,
                                               min_response_rate = 0.05,
                                               min_detected_peaks = 1) {
  if (missing(x) || is.null(x)) {
    stop("`x` must be supplied.", call. = FALSE)
  }

  if (!is.numeric(min_events) || length(min_events) != 1 ||
      !is.finite(min_events) || min_events < 0) {
    stop("`min_events` must be a single non-negative finite number.", call. = FALSE)
  }

  if (!is.numeric(min_response_events) ||
      length(min_response_events) != 1 ||
      !is.finite(min_response_events) ||
      min_response_events < 0) {
    stop("`min_response_events` must be a single non-negative finite number.", call. = FALSE)
  }

  if (!is.numeric(min_response_rate) ||
      length(min_response_rate) != 1 ||
      !is.finite(min_response_rate) ||
      min_response_rate < 0 ||
      min_response_rate > 1) {
    stop("`min_response_rate` must be between 0 and 1.", call. = FALSE)
  }

  if (!is.numeric(min_detected_peaks) ||
      length(min_detected_peaks) != 1 ||
      !is.finite(min_detected_peaks) ||
      min_detected_peaks < 0) {
    stop("`min_detected_peaks` must be a single non-negative finite number.", call. = FALSE)
  }

  event_table <- gpbiometrics_eda_nonresponder_extract_events(x)
  peak_table <- gpbiometrics_eda_nonresponder_extract_peaks(x)

  if (is.null(event_table) && is.null(peak_table)) {
    stop("`x` must contain SCR event-window or peak-detection data.", call. = FALSE)
  }

  if (is.null(group_cols)) {
    group_cols <- gpbiometrics_eda_nonresponder_resolve_group_cols(
      event_table = event_table,
      peak_table = peak_table
    )
  }

  event_summary <- gpbiometrics_eda_nonresponder_event_summary(
    event_table = event_table,
    group_cols = group_cols,
    response_col = response_col,
    amplitude_col = amplitude_col
  )

  peak_summary <- gpbiometrics_eda_nonresponder_peak_summary(
    peak_table = peak_table,
    group_cols = group_cols,
    amplitude_col = amplitude_col
  )

  group_summary <- gpbiometrics_eda_nonresponder_merge_summaries(
    event_summary = event_summary,
    peak_summary = peak_summary,
    group_cols = group_cols,
    min_events = min_events,
    min_response_events = min_response_events,
    min_response_rate = min_response_rate,
    min_detected_peaks = min_detected_peaks
  )

  candidate_nonresponders <- group_summary[
    group_summary$candidate_nonresponder %in% TRUE,
    ,
    drop = FALSE
  ]

  overview <- data.frame(
    group_count = nrow(group_summary),
    candidate_nonresponder_count = nrow(candidate_nonresponders),
    event_window_groups = if (nrow(group_summary) > 0) {
      sum(!is.na(group_summary$event_count))
    } else {
      0L
    },
    peak_detection_groups = if (nrow(group_summary) > 0) {
      sum(!is.na(group_summary$detected_peaks))
    } else {
      0L
    },
    min_events = min_events,
    min_response_events = min_response_events,
    min_response_rate = min_response_rate,
    min_detected_peaks = min_detected_peaks,
    status = if (nrow(group_summary) == 0) {
      "fail_no_groups"
    } else if (nrow(candidate_nonresponders) > 0) {
      "candidate_nonresponders_detected"
    } else {
      "no_candidate_nonresponders_detected"
    },
    stringsAsFactors = FALSE
  )

  structure(
    list(
      overview = overview,
      group_summary = group_summary,
      candidate_nonresponders = candidate_nonresponders,
      settings = list(
        group_cols = group_cols,
        response_col = response_col,
        amplitude_col = amplitude_col,
        min_events = min_events,
        min_response_events = min_response_events,
        min_response_rate = min_response_rate,
        min_detected_peaks = min_detected_peaks,
        interpretation_notes = c(
          "Nonresponder screening is threshold-dependent and should be reported with the selected criteria.",
          "Candidate nonresponders should not be automatically excluded without study-specific justification.",
          "EDA/SCR response presence is not an emotional-valence measure."
        )
      )
    ),
    class = c("gazepoint_eda_nonresponder_screen", "list")
  )
}

gpbiometrics_eda_nonresponder_extract_events <- function(x) {
  if (inherits(x, "gazepoint_scr_event_window_summary") &&
      !is.null(x$event_table)) {
    return(as.data.frame(x$event_table, stringsAsFactors = FALSE))
  }

  if (is.list(x) && !is.null(x$event_table) && is.data.frame(x$event_table)) {
    return(as.data.frame(x$event_table, stringsAsFactors = FALSE))
  }

  if (is.data.frame(x) &&
      any(c("response_flag", "scr_response_binary") %in% names(x))) {
    return(as.data.frame(x, stringsAsFactors = FALSE))
  }

  NULL
}

gpbiometrics_eda_nonresponder_extract_peaks <- function(x) {
  if (inherits(x, "gazepoint_scr_peak_detection") && !is.null(x$peaks)) {
    return(as.data.frame(x$peaks, stringsAsFactors = FALSE))
  }

  if (inherits(x, "gazepoint_scr_event_window_summary") && !is.null(x$peaks)) {
    return(as.data.frame(x$peaks, stringsAsFactors = FALSE))
  }

  if (is.list(x) && !is.null(x$peaks) && is.data.frame(x$peaks)) {
    return(as.data.frame(x$peaks, stringsAsFactors = FALSE))
  }

  if (is.data.frame(x) && "peak_time" %in% names(x)) {
    return(as.data.frame(x, stringsAsFactors = FALSE))
  }

  NULL
}

gpbiometrics_eda_nonresponder_resolve_group_cols <- function(event_table,
                                                             peak_table) {
  candidates <- c(
    "source_participant",
    "participant",
    "subject",
    "subject_id",
    "USER",
    "USER_FILE",
    "source_file",
    "MEDIA_ID",
    "MEDIA_NAME",
    "trial",
    "trial_id"
  )

  event_names <- if (is.null(event_table)) character() else names(event_table)
  peak_names <- if (is.null(peak_table)) character() else names(peak_table)

  common <- candidates[candidates %in% event_names & candidates %in% peak_names]

  if (length(common) > 0) {
    return(common)
  }

  event_only <- candidates[candidates %in% event_names]

  if (length(event_only) > 0) {
    return(event_only)
  }

  peak_only <- candidates[candidates %in% peak_names]

  if (length(peak_only) > 0) {
    return(peak_only)
  }

  character()
}

gpbiometrics_eda_nonresponder_group_id <- function(dat, group_cols) {
  if (is.null(dat) || nrow(dat) == 0) {
    return(character())
  }

  if (length(group_cols) == 0) {
    return(rep("all", nrow(dat)))
  }

  if (!all(group_cols %in% names(dat))) {
    return(rep("all", nrow(dat)))
  }

  group_dat <- dat[group_cols]

  group_dat[] <- lapply(group_dat, function(x) {
    x_chr <- as.character(x)
    x_chr[is.na(x_chr)] <- "<NA>"
    x_chr
  })

  apply(group_dat, 1, paste, collapse = "||")
}

gpbiometrics_eda_nonresponder_event_summary <- function(event_table,
                                                        group_cols,
                                                        response_col,
                                                        amplitude_col) {
  if (is.null(event_table) || nrow(event_table) == 0) {
    return(data.frame())
  }

  if (!response_col %in% names(event_table) &&
      "scr_response_binary" %in% names(event_table)) {
    response_col <- "scr_response_binary"
  }

  if (!response_col %in% names(event_table)) {
    return(data.frame())
  }

  event_table$.group_id <- gpbiometrics_eda_nonresponder_group_id(
    event_table,
    group_cols
  )

  group_ids <- unique(event_table$.group_id)

  out <- lapply(group_ids, function(group_id) {
    d <- event_table[event_table$.group_id == group_id, , drop = FALSE]

    response <- suppressWarnings(as.numeric(d[[response_col]]))
    response_binary <- ifelse(is.finite(response) & response > 0, 1L, 0L)
    response_binary[!is.finite(response)] <- NA_integer_

    amplitude <- if (amplitude_col %in% names(d)) {
      suppressWarnings(as.numeric(d[[amplitude_col]]))
    } else {
      rep(NA_real_, nrow(d))
    }

    finite_amp <- amplitude[is.finite(amplitude)]

    row <- data.frame(
      group_id = group_id,
      event_count = nrow(d),
      response_events = sum(response_binary == 1L, na.rm = TRUE),
      no_response_events = sum(response_binary == 0L, na.rm = TRUE),
      response_rate = mean(response_binary == 1L, na.rm = TRUE),
      positive_amplitude_events = sum(is.finite(amplitude) & amplitude > 0, na.rm = TRUE),
      median_response_amplitude = if (length(finite_amp) > 0) {
        stats::median(finite_amp, na.rm = TRUE)
      } else {
        NA_real_
      },
      stringsAsFactors = FALSE
    )

    if (length(group_cols) > 0 && all(group_cols %in% names(d))) {
      row <- cbind(d[1, group_cols, drop = FALSE], row)
    }

    row
  })

  out <- do.call(rbind, out)
  rownames(out) <- NULL

  out
}

gpbiometrics_eda_nonresponder_peak_summary <- function(peak_table,
                                                       group_cols,
                                                       amplitude_col) {
  if (is.null(peak_table) || nrow(peak_table) == 0) {
    return(data.frame())
  }

  peak_table$.group_id <- gpbiometrics_eda_nonresponder_group_id(
    peak_table,
    group_cols
  )

  group_ids <- unique(peak_table$.group_id)

  out <- lapply(group_ids, function(group_id) {
    d <- peak_table[peak_table$.group_id == group_id, , drop = FALSE]

    amplitude <- if (amplitude_col %in% names(d)) {
      suppressWarnings(as.numeric(d[[amplitude_col]]))
    } else if ("amplitude" %in% names(d)) {
      suppressWarnings(as.numeric(d$amplitude))
    } else {
      rep(NA_real_, nrow(d))
    }

    finite_amp <- amplitude[is.finite(amplitude)]

    complete_recovery <- if ("status" %in% names(d)) {
      sum(as.character(d$status) == "detected", na.rm = TRUE)
    } else {
      NA_integer_
    }

    row <- data.frame(
      group_id = group_id,
      detected_peaks = nrow(d),
      complete_recovery_peaks = complete_recovery,
      incomplete_recovery_peaks = if ("status" %in% names(d)) {
        sum(as.character(d$status) == "detected_incomplete_recovery", na.rm = TRUE)
      } else {
        NA_integer_
      },
      median_peak_amplitude = if (length(finite_amp) > 0) {
        stats::median(finite_amp, na.rm = TRUE)
      } else {
        NA_real_
      },
      max_peak_amplitude = if (length(finite_amp) > 0) {
        max(finite_amp, na.rm = TRUE)
      } else {
        NA_real_
      },
      stringsAsFactors = FALSE
    )

    if (length(group_cols) > 0 && all(group_cols %in% names(d))) {
      row <- cbind(d[1, group_cols, drop = FALSE], row)
    }

    row
  })

  out <- do.call(rbind, out)
  rownames(out) <- NULL

  out
}

gpbiometrics_eda_nonresponder_merge_summaries <- function(event_summary,
                                                          peak_summary,
                                                          group_cols,
                                                          min_events,
                                                          min_response_events,
                                                          min_response_rate,
                                                          min_detected_peaks) {
  event_ids <- if (nrow(event_summary) > 0) event_summary$group_id else character()
  peak_ids <- if (nrow(peak_summary) > 0) peak_summary$group_id else character()

  group_ids <- unique(c(event_ids, peak_ids))

  if (length(group_ids) == 0) {
    return(data.frame())
  }

  out <- lapply(group_ids, function(group_id) {
    event_row <- event_summary[event_summary$group_id == group_id, , drop = FALSE]
    peak_row <- peak_summary[peak_summary$group_id == group_id, , drop = FALSE]

    row <- data.frame(group_id = group_id, stringsAsFactors = FALSE)

    if (length(group_cols) > 0) {
      source_row <- if (nrow(event_row) > 0) event_row else peak_row

      if (nrow(source_row) > 0 && all(group_cols %in% names(source_row))) {
        row <- cbind(source_row[1, group_cols, drop = FALSE], row)
      }
    }

    event_count <- if (nrow(event_row) > 0) event_row$event_count else NA_integer_
    response_events <- if (nrow(event_row) > 0) event_row$response_events else NA_integer_
    response_rate <- if (nrow(event_row) > 0) event_row$response_rate else NA_real_
    detected_peaks <- if (nrow(peak_row) > 0) peak_row$detected_peaks else NA_integer_

    classification_basis <- if (is.finite(event_count)) {
      "event_windows"
    } else if (is.finite(detected_peaks)) {
      "peak_detection"
    } else {
      "none"
    }

    candidate_nonresponder <- NA

    classification_status <- "insufficient_information"

    if (identical(classification_basis, "event_windows")) {
      if (event_count < min_events) {
        candidate_nonresponder <- NA
        classification_status <- "insufficient_events"
      } else {
        candidate_nonresponder <- response_events < min_response_events ||
          response_rate < min_response_rate
        classification_status <- if (candidate_nonresponder) {
          "candidate_nonresponder"
        } else {
          "responder_detected"
        }
      }
    } else if (identical(classification_basis, "peak_detection")) {
      candidate_nonresponder <- detected_peaks < min_detected_peaks
      classification_status <- if (candidate_nonresponder) {
        "candidate_nonresponder"
      } else {
        "responder_detected"
      }
    }

    event_cols <- c(
      "event_count",
      "response_events",
      "no_response_events",
      "response_rate",
      "positive_amplitude_events",
      "median_response_amplitude"
    )

    peak_cols <- c(
      "detected_peaks",
      "complete_recovery_peaks",
      "incomplete_recovery_peaks",
      "median_peak_amplitude",
      "max_peak_amplitude"
    )

    for (col in event_cols) {
      row[[col]] <- if (nrow(event_row) > 0 && col %in% names(event_row)) {
        event_row[[col]][1]
      } else {
        NA
      }
    }

    for (col in peak_cols) {
      row[[col]] <- if (nrow(peak_row) > 0 && col %in% names(peak_row)) {
        peak_row[[col]][1]
      } else {
        NA
      }
    }

    row$classification_basis <- classification_basis
    row$candidate_nonresponder <- candidate_nonresponder
    row$class_status <- classification_status

    row
  })

  out <- do.call(rbind, out)
  rownames(out) <- NULL

  out
}

#' Import a Gazepoint Data Summary export
#'
#' Reads a multi-section `Data_Summary_export_*.csv` file produced by Gazepoint
#' Analysis. These files are not ordinary rectangular CSV files. They contain
#' metadata followed by sections such as `AOI Summary` and `AOI Statistics (for
#' each user)`. The latter may include AOI-level biometric summaries such as
#' average dial value, average GSR, average heart rate, average interbeat
#' interval, and pupil diameter.
#'
#' @param file Path to a Gazepoint `Data_Summary_export_*.csv` file.
#'
#' @return A list with `metadata`, `aoi_summary`, and `aoi_statistics` data
#'   frames. The returned object has class `"gazepoint_data_summary"`.
#'
#' @export
import_gazepoint_data_summary <- function(file) {
  if (missing(file) || length(file) != 1L || !nzchar(file)) {
    stop("`file` must be a single non-empty file path.", call. = FALSE)
  }

  if (!file.exists(file)) {
    stop("File does not exist: ", file, call. = FALSE)
  }

  lines <- readLines(file, warn = FALSE, encoding = "UTF-8")

  metadata <- parse_gazepoint_data_summary_metadata(lines, file)

  aoi_summary <- parse_gazepoint_data_summary_section(
    lines = lines,
    section_pattern = "^AOI Summary$",
    source_file = basename(file)
  )

  aoi_statistics <- parse_gazepoint_data_summary_section(
    lines = lines,
    section_pattern = "^AOI Statistics",
    source_file = basename(file)
  )

  out <- list(
    metadata = metadata,
    aoi_summary = aoi_summary,
    aoi_statistics = aoi_statistics
  )

  class(out) <- c("gazepoint_data_summary", "list")
  out
}


parse_gazepoint_data_summary_metadata <- function(lines, file) {
  first_line <- split_first_csv_line(lines[1])
  second_line <- if (length(lines) >= 2L) split_first_csv_line(lines[2]) else c(NA, NA)

  notes <- lines[grepl("^Note:", trimws(lines), ignore.case = TRUE)]
  notes <- trimws(notes)

  data.frame(
    source_file = basename(file),
    software = first_line[1],
    version = first_line[2],
    processed_label = second_line[1],
    processed_on = second_line[2],
    notes = paste(notes, collapse = " | "),
    stringsAsFactors = FALSE
  )
}


parse_gazepoint_data_summary_section <- function(lines,
                                                 section_pattern,
                                                 source_file) {
  trimmed <- trimws(lines)

  section_idx <- grep(section_pattern, trimmed, ignore.case = TRUE)

  if (length(section_idx) == 0L) {
    return(data.frame(source_file = character(0), stringsAsFactors = FALSE))
  }

  section_idx <- section_idx[1]

  header_idx <- next_nonempty_line(trimmed, section_idx + 1L)

  if (is.na(header_idx)) {
    return(data.frame(source_file = character(0), stringsAsFactors = FALSE))
  }

  data_start <- header_idx + 1L
  data_end <- data_start - 1L

  if (data_start <= length(lines)) {
    for (i in data_start:length(lines)) {
      if (!nzchar(trimmed[i])) {
        break
      }

      data_end <- i
    }
  }

  section_lines <- lines[header_idx:data_end]

  if (length(section_lines) == 1L) {
    header <- parse_csv_text(section_lines)
    out <- header[0, , drop = FALSE]
  } else {
    out <- parse_csv_text(section_lines)
  }

  out <- drop_empty_trailing_columns(out)
  out <- convert_numeric_like_columns(out)

  out$source_file <- source_file

  out
}


next_nonempty_line <- function(trimmed_lines, start) {
  if (start > length(trimmed_lines)) {
    return(NA_integer_)
  }

  for (i in start:length(trimmed_lines)) {
    if (nzchar(trimmed_lines[i])) {
      return(i)
    }
  }

  NA_integer_
}


parse_csv_text <- function(lines) {
  text <- paste(lines, collapse = "\n")

  utils::read.csv(
    text = text,
    stringsAsFactors = FALSE,
    check.names = FALSE,
    na.strings = c("", "NA", "NaN")
  )
}


split_first_csv_line <- function(line) {
  if (length(line) == 0L || is.na(line)) {
    return(c(NA_character_, NA_character_))
  }

  parts <- strsplit(line, ",", fixed = TRUE)[[1]]

  if (length(parts) == 1L) {
    return(c(trimws(parts[1]), NA_character_))
  }

  c(trimws(parts[1]), trimws(paste(parts[-1], collapse = ",")))
}


convert_numeric_like_columns <- function(data) {
  if (!is.data.frame(data) || ncol(data) == 0L) {
    return(data)
  }

  for (nm in names(data)) {
    x <- data[[nm]]

    if (!is.character(x)) {
      next
    }

    x_trim <- trimws(x)

    non_missing <- x_trim[!is.na(x_trim) & nzchar(x_trim)]

    if (length(non_missing) == 0L) {
      next
    }

    numeric_x <- suppressWarnings(as.numeric(x_trim))
    numeric_non_missing <- numeric_x[!is.na(x_trim) & nzchar(x_trim)]

    if (all(!is.na(numeric_non_missing))) {
      data[[nm]] <- numeric_x
    }
  }

  data
}

test_that("import_gazepoint_data_summary parses metadata and sections", {
  tmp <- tempfile(fileext = ".csv")

  writeLines(
    c(
      "Gazepoint Analysis,v7.2.0",
      "Processed on,Thu Jun 11 23:50:26 2026",
      "",
      "Note: -1 values indicate an AOI that was never viewed",
      "AOI Summary",
      "Media ID, Media Name,AOI ID,AOI Name,Viewers (#),Ave Time Viewed (sec),",
      "0,NewMedia0,2,AOI 2,2,1.400,",
      "1,NewMedia1,0,AOI 0,2,2.235,",
      "",
      "AOI Statistics (for each user)",
      "Media ID, Media Name,AOI ID,AOI Name,User ID,User Name,Time Viewed (sec),Fixations (#),Ave Dial (0-1),Ave GSR (kOhm),Ave Heart Rate (BPM),Ave Interbeat Interval (s),Ave Left Pupil (mm),Ave Right Pupil (mm),",
      "0,NewMedia0,2,AOI 2,4,User 4,0.937,3,1.000,815880.062,88.442,0.678,7.041,5.707,",
      "0,NewMedia0,2,AOI 2,5,User 5,1.863,7,1.000,362668.094,87.988,0.681,4.605,5.451,"
    ),
    tmp,
    useBytes = TRUE
  )

  out <- import_gazepoint_data_summary(tmp)

  expect_s3_class(out, "gazepoint_data_summary")
  expect_true(is.data.frame(out$metadata))
  expect_true(is.data.frame(out$aoi_summary))
  expect_true(is.data.frame(out$aoi_statistics))

  expect_equal(out$metadata$software, "Gazepoint Analysis")
  expect_equal(out$metadata$version, "v7.2.0")
  expect_equal(nrow(out$aoi_summary), 2)
  expect_equal(nrow(out$aoi_statistics), 2)

  expect_true("Ave Dial (0-1)" %in% names(out$aoi_statistics))
  expect_true("Ave GSR (kOhm)" %in% names(out$aoi_statistics))
  expect_true("Ave Heart Rate (BPM)" %in% names(out$aoi_statistics))
  expect_true("Ave Interbeat Interval (s)" %in% names(out$aoi_statistics))

  expect_equal(out$aoi_statistics[["Ave Heart Rate (BPM)"]][1], 88.442)
  expect_equal(out$aoi_statistics[["Ave Dial (0-1)"]][1], 1)
})


test_that("import_gazepoint_data_summary preserves source file column", {
  tmp <- tempfile("Data_Summary_export_test_", fileext = ".csv")

  writeLines(
    c(
      "Gazepoint Analysis,v7.2.0",
      "Processed on,example",
      "",
      "AOI Summary",
      "Media ID,AOI ID,Viewers (#),",
      "0,2,2,",
      "",
      "AOI Statistics (for each user)",
      "Media ID,AOI ID,User ID,Ave Dial (0-1),",
      "0,2,4,1.000,"
    ),
    tmp,
    useBytes = TRUE
  )

  out <- import_gazepoint_data_summary(tmp)

  expect_true("source_file" %in% names(out$aoi_summary))
  expect_true("source_file" %in% names(out$aoi_statistics))
  expect_equal(out$aoi_summary$source_file[1], basename(tmp))
})


test_that("import_gazepoint_data_summary returns empty section when section is absent", {
  tmp <- tempfile(fileext = ".csv")

  writeLines(
    c(
      "Gazepoint Analysis,v7.2.0",
      "Processed on,example"
    ),
    tmp,
    useBytes = TRUE
  )

  out <- import_gazepoint_data_summary(tmp)

  expect_true(is.data.frame(out$aoi_summary))
  expect_true(is.data.frame(out$aoi_statistics))
  expect_equal(nrow(out$aoi_summary), 0)
  expect_equal(nrow(out$aoi_statistics), 0)
})


test_that("import_gazepoint_data_summary rejects missing file", {
  expect_error(
    import_gazepoint_data_summary("missing_data_summary.csv"),
    "File does not exist"
  )
})

#' gpbiometrics: Process, Audit, Plot, and Report Gazepoint Biometrics Data
#'
#' `gpbiometrics` provides helper functions for importing, checking,
#' preprocessing, summarising, diagnosing, plotting, and reporting Gazepoint
#' Biometrics exports. It is designed for reproducible offline workflows
#' involving physiological, response-channel, event-marker, and timing data
#' exported alongside Gazepoint eye-tracking recordings.
#'
#' @details
#' The package supports common Gazepoint biometric channels, including GSR/EDA,
#' heart rate, interbeat interval or RR-style intervals, engagement dial, and
#' TTL markers. It provides functions for folder-level import, data-summary
#' import, schema and timebase detection, active-channel detection, column
#' validation, missingness audits, signal-quality audits, sampling and timing
#' audits, dropout and flatline flagging, baseline correction, smoothing, window
#' summaries, TTL event extraction, biometric-to-gaze synchronisation, exclusion
#' recommendations, workflow diagnostics, report-ready tables, cautious methods
#' text, lightweight report generation, and base-R diagnostic plotting.
#'
#' Schema and timing helpers include
#' `standardise_gazepoint_biometric_names()`,
#' `detect_gazepoint_biometric_schema()`,
#' `detect_gazepoint_time_columns()`, and
#' `detect_gazepoint_biometric_timebase()`. Signal-availability and quality
#' helpers include `summarise_gazepoint_biometric_validity()`,
#' `flag_gazepoint_biometric_dropouts()`, `audit_gazepoint_gsr_quality()`,
#' `audit_gazepoint_hr_quality()`, `audit_gazepoint_engagement_dial()`, and
#' `audit_gazepoint_ibi_quality()`.
#'
#' Preprocessing and summary helpers include
#' `baseline_correct_gazepoint_gsr()`,
#' `baseline_correct_gazepoint_hr()`, `smooth_gazepoint_biometrics()`,
#' `convert_gazepoint_gsr_to_conductance()`,
#' `summarise_gazepoint_gsr_windows()`,
#' `summarise_gazepoint_hr_windows()`,
#' `summarise_gazepoint_engagement_windows()`,
#' `summarise_gazepoint_dial_windows()`,
#' `summarise_gazepoint_ibi_windows()`,
#' `summarise_gazepoint_gsr_tonic_phasic()`,
#' `summarise_gazepoint_multimodal_windows()`,
#' `summarise_gazepoint_ibi_hrv_windows()`, and
#' `summarise_gazepoint_full_biometric_windows()`.
#'
#' Synchronisation, modelling, workflow, and reporting helpers include
#' `sync_gazepoint_biometrics_with_gaze()`,
#' `join_gazepoint_biometrics_to_master()`,
#' `join_gazepoint_biometrics_to_gp3tools()`,
#' `prepare_gazepoint_multimodal_model_data()`,
#' `run_gazepoint_biometrics_workflow()`,
#' `summarise_gazepoint_biometrics_workflow()`,
#' `diagnose_gazepoint_biometrics_workflow()`,
#' `create_gazepoint_biometrics_checklist()`,
#' `create_gazepoint_biometrics_methods_text()`,
#' `create_gazepoint_biometrics_report_tables()`,
#' `write_gazepoint_biometrics_report_tables()`, and
#' `create_gazepoint_biometrics_report()`. Diagnostic plotting helpers include
#' `plot_gazepoint_biometric_signals()` and
#' `plot_gazepoint_biometric_quality()`.
#'
#' The package treats biometric signals conservatively. GSR/EDA is handled as an
#' electrodermal activity or arousal-related signal rather than emotional
#' valence. Heart-rate summaries should be interpreted relative to baseline,
#' artefact handling, and task context. Raw Gazepoint `HRV` columns should be
#' treated as validity or vendor flags unless independent documentation proves
#' otherwise. IBI-derived HRV-style summaries should be computed only from
#' genuine interbeat interval or RR-style interval columns. Eye-tracking
#' measures, when combined with biometric data, should be interpreted as
#' indicators of visual attention rather than direct evidence of cognition,
#' scrutiny, or evaluation.
#'
#' @section Main workflow:
#'
#' The main workflow function is:
#'
#' `run_gazepoint_biometrics_workflow()`
#'
#' Useful follow-up helpers include:
#'
#' `summarise_gazepoint_biometrics_workflow()`,
#' `diagnose_gazepoint_biometrics_workflow()`,
#' `create_gazepoint_biometrics_report_tables()`,
#' `write_gazepoint_biometrics_report_tables()`, and
#' `create_gazepoint_biometrics_report()`.
#'
#' @section Interpretation caution:
#'
#' `gpbiometrics` is a preprocessing, quality-control, visualisation, and
#' reporting toolkit. It does not classify emotions, mental states, or cognitive
#' processes directly. Researchers should interpret biometric signals in
#' relation to experimental design, baseline periods, stimulus timing, task
#' demands, artefact handling, and complementary behavioural or self-report
#' measures.
#'
#' @docType package
#' @name gpbiometrics-package
#' @aliases gpbiometrics
#' @keywords internal
"_PACKAGE"

#' Export Gazepoint IBI data for RHRV-style workflows
#'
#' Prepares cleaned inter-beat interval data for external HRV workflows. The
#' helper writes simple beat tables with cumulative beat time and IBI/RR interval
#' columns. It does not use the Gazepoint `HRV` column.
#'
#' @param data A Gazepoint biometric data frame or `gazepoint_ibi_filter` object.
#' @param ibi_col IBI/RR interval column.
#' @param group_cols Optional grouping columns.
#' @param unit Unit of the IBI column: `"auto"`, `"ms"`, or `"seconds"`.
#' @param collapse_repeated_intervals Logical. If `TRUE`, consecutive repeated
#'   IBI values are collapsed before export.
#' @param repeated_tolerance_ms Numeric tolerance used when identifying repeated
#'   consecutive IBI values.
#' @param min_ibi_ms Minimum plausible IBI in milliseconds retained for export.
#' @param max_ibi_ms Maximum plausible IBI in milliseconds retained for export.
#' @param output_dir Optional directory where per-group CSV files are written.
#' @param prefix File prefix used when `output_dir` is supplied.
#'
#' @return A list with `overview`, `beat_table`, `group_summary`, `manifest`,
#'   and `settings`.
#' @export
export_gazepoint_rhrv_input <- function(data,
                                        ibi_col = "IBI_clean_ms",
                                        group_cols = NULL,
                                        unit = c("auto", "ms", "seconds"),
                                        collapse_repeated_intervals = TRUE,
                                        repeated_tolerance_ms = 1e-8,
                                        min_ibi_ms = 300,
                                        max_ibi_ms = 2000,
                                        output_dir = NULL,
                                        prefix = "gazepoint_rhrv") {
  unit <- match.arg(unit)

  dat <- gpbiometrics_external_extract_data(data)

  if (!ibi_col %in% names(dat)) {
    stop("`ibi_col` was not found in `data`.", call. = FALSE)
  }

  if (is.null(group_cols)) {
    group_cols <- gpbiometrics_external_default_group_cols(names(dat))
  }

  missing_groups <- setdiff(group_cols, names(dat))

  if (length(missing_groups) > 0) {
    stop(
      "`group_cols` were not found in `data`: ",
      paste(missing_groups, collapse = ", "),
      call. = FALSE
    )
  }

  if (!is.logical(collapse_repeated_intervals) ||
      length(collapse_repeated_intervals) != 1 ||
      is.na(collapse_repeated_intervals)) {
    stop("`collapse_repeated_intervals` must be TRUE or FALSE.", call. = FALSE)
  }

  gpbiometrics_external_validate_nonnegative_scalar(
    repeated_tolerance_ms,
    "repeated_tolerance_ms"
  )
  gpbiometrics_external_validate_positive_scalar(min_ibi_ms, "min_ibi_ms")
  gpbiometrics_external_validate_positive_scalar(max_ibi_ms, "max_ibi_ms")

  if (max_ibi_ms <= min_ibi_ms) {
    stop("`max_ibi_ms` must be greater than `min_ibi_ms`.", call. = FALSE)
  }

  ibi_raw <- suppressWarnings(as.numeric(dat[[ibi_col]]))
  detected_unit <- gpbiometrics_external_detect_ibi_unit(ibi_raw, unit)
  ibi_ms <- if (identical(detected_unit, "seconds")) ibi_raw * 1000 else ibi_raw

  dat$.ibi_export_ms <- ibi_ms
  dat$.group_id <- gpbiometrics_external_group_id(dat, group_cols)

  group_ids <- unique(dat$.group_id)

  beat_tables <- lapply(group_ids, function(group_id) {
    d <- dat[dat$.group_id == group_id, , drop = FALSE]
    x <- d$.ibi_export_ms
    x <- x[is.finite(x) & x >= min_ibi_ms & x <= max_ibi_ms]

    input_interval_rows <- length(x)

    if (isTRUE(collapse_repeated_intervals)) {
      x <- gpbiometrics_external_collapse_repeated_intervals(
        x,
        tolerance_ms = repeated_tolerance_ms
      )
    }

    beat_table <- data.frame(
      group_id = rep(group_id, length(x)),
      beat_index = seq_along(x),
      time_s = cumsum(x) / 1000,
      ibi_ms = x,
      ibi_s = x / 1000,
      stringsAsFactors = FALSE
    )

    beat_table$input_interval_rows <- rep(as.integer(input_interval_rows), nrow(beat_table))
    beat_table$output_interval_rows <- rep(as.integer(length(x)), nrow(beat_table))
    beat_table$used_intervals_after_collapse <- rep(as.integer(length(x)), nrow(beat_table))

    if (length(group_cols) > 0 && nrow(d) > 0 &&
        all(group_cols %in% names(d))) {
      beat_table <- cbind(d[rep(1, nrow(beat_table)), group_cols, drop = FALSE], beat_table)
    }

    beat_table
  })

  beat_table <- do.call(rbind, beat_tables)
  rownames(beat_table) <- NULL

  group_summary <- gpbiometrics_external_rhrv_group_summary(
    beat_table = beat_table,
    group_cols = group_cols
  )

  manifest <- gpbiometrics_external_write_group_files(
    data = beat_table,
    group_cols = group_cols,
    output_dir = output_dir,
    prefix = prefix,
    suffix = "rhrv_input"
  )

  overview <- data.frame(
    input_rows = nrow(dat),
    beat_rows = nrow(beat_table),
    group_count = length(group_ids),
    detected_ibi_unit = detected_unit,
    collapse_repeated_intervals = collapse_repeated_intervals,
    files_written = sum(!is.na(manifest$file_path)),
    status = if (nrow(beat_table) == 0) {
      "fail_no_ibi_rows_for_export"
    } else if (!is.null(output_dir)) {
      "rhrv_input_exported"
    } else {
      "rhrv_input_prepared"
    },
    stringsAsFactors = FALSE
  )

  structure(
    list(
      overview = overview,
      beat_table = beat_table,
      group_summary = group_summary,
      manifest = manifest,
      settings = list(
        ibi_col = ibi_col,
        group_cols = group_cols,
        unit = unit,
        detected_ibi_unit = detected_unit,
        collapse_repeated_intervals = collapse_repeated_intervals,
        repeated_tolerance_ms = repeated_tolerance_ms,
        min_ibi_ms = min_ibi_ms,
        max_ibi_ms = max_ibi_ms,
        output_dir = output_dir,
        prefix = prefix,
        interpretation_notes = c(
          "The export uses genuine IBI/RR intervals only.",
          "The Gazepoint HRV column is not used.",
          "External HRV software may require additional formatting depending on the selected workflow."
        )
      )
    ),
    class = c("gazepoint_rhrv_input_export", "list")
  )
}

#' Prepare Gazepoint EDA input for NeuroKit2-style workflows
#'
#' Prepares EDA/GSR signal tables for optional external NeuroKit2 processing.
#' This helper does not require Python or NeuroKit2.
#'
#' @param data A Gazepoint biometric data frame.
#' @param eda_col EDA/GSR signal column.
#' @param time_col Optional time/counter column.
#' @param group_cols Optional grouping columns.
#' @param sampling_rate Optional sampling rate in Hz.
#' @param output_dir Optional directory where per-group CSV files are written.
#' @param prefix File prefix used when `output_dir` is supplied.
#'
#' @return A list with `overview`, `eda_table`, `group_summary`, `manifest`,
#'   and `settings`.
#' @export
prepare_gazepoint_neurokit_eda_input <- function(data,
                                                 eda_col = "GSR_US",
                                                 time_col = NULL,
                                                 group_cols = NULL,
                                                 sampling_rate = NULL,
                                                 output_dir = NULL,
                                                 prefix = "gazepoint_neurokit_eda") {
  if (!is.data.frame(data)) {
    stop("`data` must be a data frame.", call. = FALSE)
  }

  dat <- as.data.frame(data, stringsAsFactors = FALSE)

  if (!eda_col %in% names(dat)) {
    stop("`eda_col` was not found in `data`.", call. = FALSE)
  }

  if (!is.null(time_col) && !time_col %in% names(dat)) {
    stop("`time_col` was not found in `data`.", call. = FALSE)
  }

  if (is.null(group_cols)) {
    group_cols <- gpbiometrics_external_default_group_cols(names(dat))
  }

  missing_groups <- setdiff(group_cols, names(dat))

  if (length(missing_groups) > 0) {
    stop(
      "`group_cols` were not found in `data`: ",
      paste(missing_groups, collapse = ", "),
      call. = FALSE
    )
  }

  if (!is.null(sampling_rate)) {
    gpbiometrics_external_validate_positive_scalar(sampling_rate, "sampling_rate")
  }

  dat$.group_id <- gpbiometrics_external_group_id(dat, group_cols)
  group_ids <- unique(dat$.group_id)

  eda_tables <- lapply(group_ids, function(group_id) {
    d <- dat[dat$.group_id == group_id, , drop = FALSE]

    eda <- suppressWarnings(as.numeric(d[[eda_col]]))
    sample_index <- seq_along(eda)

    time_raw <- if (!is.null(time_col)) {
      suppressWarnings(as.numeric(d[[time_col]]))
    } else {
      sample_index - 1
    }

    time_s <- gpbiometrics_external_time_seconds(
      time_raw = time_raw,
      sampling_rate = sampling_rate
    )

    out <- data.frame(
      group_id = group_id,
      sample_index = sample_index,
      time_raw = time_raw,
      time_s = time_s,
      eda = eda,
      stringsAsFactors = FALSE
    )

    if (length(group_cols) > 0 && all(group_cols %in% names(d))) {
      out <- cbind(d[, group_cols, drop = FALSE], out)
    }

    out
  })

  eda_table <- do.call(rbind, eda_tables)
  rownames(eda_table) <- NULL

  group_summary <- gpbiometrics_external_eda_group_summary(
    eda_table = eda_table,
    group_cols = group_cols
  )

  manifest <- gpbiometrics_external_write_group_files(
    data = eda_table,
    group_cols = group_cols,
    output_dir = output_dir,
    prefix = prefix,
    suffix = "neurokit_eda_input"
  )

  overview <- data.frame(
    input_rows = nrow(dat),
    eda_rows = nrow(eda_table),
    group_count = length(group_ids),
    eda_col = eda_col,
    sampling_rate = if (is.null(sampling_rate)) NA_real_ else sampling_rate,
    files_written = sum(!is.na(manifest$file_path)),
    status = if (nrow(eda_table) == 0) {
      "fail_no_eda_rows_for_export"
    } else if (!is.null(output_dir)) {
      "neurokit_eda_input_exported"
    } else {
      "neurokit_eda_input_prepared"
    },
    stringsAsFactors = FALSE
  )

  structure(
    list(
      overview = overview,
      eda_table = eda_table,
      group_summary = group_summary,
      manifest = manifest,
      settings = list(
        eda_col = eda_col,
        time_col = time_col,
        group_cols = group_cols,
        sampling_rate = sampling_rate,
        output_dir = output_dir,
        prefix = prefix,
        interpretation_notes = c(
          "This helper prepares EDA/GSR input for optional external NeuroKit2 processing.",
          "EDA/GSR reflects sympathetic arousal-related activity and should not be treated as emotional valence.",
          "External NeuroKit2 execution is optional and not required by gpbiometrics."
        )
      )
    ),
    class = c("gazepoint_neurokit_eda_input", "list")
  )
}

#' Optionally run a NeuroKit2 EDA cross-check
#'
#' Optionally calls Python/NeuroKit2 on prepared EDA input. By default,
#' `execute = FALSE`, so no external dependency is required.
#'
#' @param data A Gazepoint biometric data frame or
#'   `gazepoint_neurokit_eda_input` object.
#' @param eda_col EDA/GSR signal column, used when `data` is a data frame.
#' @param time_col Optional time/counter column.
#' @param group_cols Optional grouping columns.
#' @param sampling_rate Sampling rate in Hz required for NeuroKit2 execution.
#' @param execute Logical. If `FALSE`, only prepare input and return skipped
#'   status.
#' @param python Python executable.
#' @param output_dir Directory for temporary/input/output files.
#' @param prefix File prefix.
#' @param keep_files Logical. If `FALSE`, temporary files produced during
#'   execution may be removed.
#'
#' @return A list with `overview`, `prepared_input`, `crosscheck_summary`,
#'   `manifest`, and `settings`.
#' @export
run_gazepoint_neurokit_eda_crosscheck <- function(data,
                                                  eda_col = "GSR_US",
                                                  time_col = NULL,
                                                  group_cols = NULL,
                                                  sampling_rate = NULL,
                                                  execute = FALSE,
                                                  python = "python",
                                                  output_dir = tempdir(),
                                                  prefix = "gazepoint_neurokit_crosscheck",
                                                  keep_files = FALSE) {
  if (!is.logical(execute) || length(execute) != 1 || is.na(execute)) {
    stop("`execute` must be TRUE or FALSE.", call. = FALSE)
  }

  if (!is.logical(keep_files) || length(keep_files) != 1 || is.na(keep_files)) {
    stop("`keep_files` must be TRUE or FALSE.", call. = FALSE)
  }

  prepared <- if (inherits(data, "gazepoint_neurokit_eda_input")) {
    data
  } else {
    prepare_gazepoint_neurokit_eda_input(
      data = data,
      eda_col = eda_col,
      time_col = time_col,
      group_cols = group_cols,
      sampling_rate = sampling_rate,
      output_dir = NULL,
      prefix = prefix
    )
  }

  if (!isTRUE(execute)) {
    overview <- data.frame(
      input_rows = nrow(prepared$eda_table),
      groups = prepared$overview$group_count,
      executed = FALSE,
      status = "skipped_execute_false",
      message = "Set `execute = TRUE` to run optional Python/NeuroKit2 cross-check.",
      stringsAsFactors = FALSE
    )

    return(structure(
      list(
        overview = overview,
        prepared_input = prepared,
        crosscheck_summary = data.frame(),
        manifest = prepared$manifest,
        settings = list(
          eda_col = eda_col,
          time_col = time_col,
          group_cols = group_cols,
          sampling_rate = sampling_rate,
          execute = execute,
          python = python,
          output_dir = output_dir,
          prefix = prefix,
          keep_files = keep_files
        )
      ),
      class = c("gazepoint_neurokit_eda_crosscheck", "list")
    ))
  }

  if (is.null(sampling_rate) || !is.finite(sampling_rate) || sampling_rate <= 0) {
    stop("`sampling_rate` must be supplied as a positive finite number when `execute = TRUE`.", call. = FALSE)
  }

  python_path <- Sys.which(python)

  if (identical(unname(python_path), "")) {
    overview <- data.frame(
      input_rows = nrow(prepared$eda_table),
      groups = prepared$overview$group_count,
      executed = FALSE,
      status = "skipped_python_not_found",
      message = paste("Python executable not found:", python),
      stringsAsFactors = FALSE
    )

    return(structure(
      list(
        overview = overview,
        prepared_input = prepared,
        crosscheck_summary = data.frame(),
        manifest = prepared$manifest,
        settings = list(
          eda_col = eda_col,
          time_col = time_col,
          group_cols = group_cols,
          sampling_rate = sampling_rate,
          execute = execute,
          python = python,
          output_dir = output_dir,
          prefix = prefix,
          keep_files = keep_files
        )
      ),
      class = c("gazepoint_neurokit_eda_crosscheck", "list")
    ))
  }

  neurokit_available <- gpbiometrics_external_python_has_neurokit2(
    python_path = python_path
  )

  if (!isTRUE(neurokit_available)) {
    overview <- data.frame(
      input_rows = nrow(prepared$eda_table),
      groups = prepared$overview$group_count,
      executed = FALSE,
      status = "skipped_neurokit2_not_available",
      message = "Python was found, but neurokit2 could not be imported.",
      stringsAsFactors = FALSE
    )

    return(structure(
      list(
        overview = overview,
        prepared_input = prepared,
        crosscheck_summary = data.frame(),
        manifest = prepared$manifest,
        settings = list(
          eda_col = eda_col,
          time_col = time_col,
          group_cols = group_cols,
          sampling_rate = sampling_rate,
          execute = execute,
          python = python,
          output_dir = output_dir,
          prefix = prefix,
          keep_files = keep_files
        )
      ),
      class = c("gazepoint_neurokit_eda_crosscheck", "list")
    ))
  }

  dir.create(output_dir, recursive = TRUE, showWarnings = FALSE)

  input_file <- file.path(output_dir, paste0(prefix, "_input.csv"))
  output_file <- file.path(output_dir, paste0(prefix, "_summary.csv"))
  script_file <- file.path(output_dir, paste0(prefix, "_runner.py"))

  utils::write.csv(prepared$eda_table, input_file, row.names = FALSE)

  gpbiometrics_external_write_neurokit_script(
    script_file = script_file,
    input_file = input_file,
    output_file = output_file,
    sampling_rate = sampling_rate
  )

  run_status <- try(
    system2(
      python_path,
      args = script_file,
      stdout = TRUE,
      stderr = TRUE
    ),
    silent = TRUE
  )

  if (inherits(run_status, "try-error") || !file.exists(output_file)) {
    overview <- data.frame(
      input_rows = nrow(prepared$eda_table),
      groups = prepared$overview$group_count,
      executed = FALSE,
      status = "neurokit2_execution_failed",
      message = paste(as.character(run_status), collapse = " "),
      stringsAsFactors = FALSE
    )

    return(structure(
      list(
        overview = overview,
        prepared_input = prepared,
        crosscheck_summary = data.frame(),
        manifest = data.frame(
          input_file = input_file,
          output_file = output_file,
          script_file = script_file,
          stringsAsFactors = FALSE
        ),
        settings = list(
          eda_col = eda_col,
          time_col = time_col,
          group_cols = group_cols,
          sampling_rate = sampling_rate,
          execute = execute,
          python = python,
          output_dir = output_dir,
          prefix = prefix,
          keep_files = keep_files
        )
      ),
      class = c("gazepoint_neurokit_eda_crosscheck", "list")
    ))
  }

  crosscheck_summary <- utils::read.csv(output_file, stringsAsFactors = FALSE)

  if (!isTRUE(keep_files)) {
    unlink(c(input_file, script_file), force = TRUE)
  }

  overview <- data.frame(
    input_rows = nrow(prepared$eda_table),
    groups = prepared$overview$group_count,
    executed = TRUE,
    status = "neurokit2_crosscheck_completed",
    message = "Optional NeuroKit2 EDA cross-check completed.",
    stringsAsFactors = FALSE
  )

  structure(
    list(
      overview = overview,
      prepared_input = prepared,
      crosscheck_summary = crosscheck_summary,
      manifest = data.frame(
        input_file = if (isTRUE(keep_files)) input_file else NA_character_,
        output_file = output_file,
        script_file = if (isTRUE(keep_files)) script_file else NA_character_,
        stringsAsFactors = FALSE
      ),
      settings = list(
        eda_col = eda_col,
        time_col = time_col,
        group_cols = group_cols,
        sampling_rate = sampling_rate,
        execute = execute,
        python = python,
        output_dir = output_dir,
        prefix = prefix,
        keep_files = keep_files
      )
    ),
    class = c("gazepoint_neurokit_eda_crosscheck", "list")
  )
}

gpbiometrics_external_extract_data <- function(data) {
  if (inherits(data, "gazepoint_ibi_filter") && !is.null(data$data)) {
    return(as.data.frame(data$data, stringsAsFactors = FALSE))
  }

  if (is.data.frame(data)) {
    return(as.data.frame(data, stringsAsFactors = FALSE))
  }

  stop("`data` must be a data frame or a supported gpbiometrics object.", call. = FALSE)
}

gpbiometrics_external_default_group_cols <- function(names_dat) {
  candidates <- c(
    "source_file",
    "source_participant",
    "participant",
    "subject",
    "subject_id",
    "MEDIA_ID",
    "MEDIA_NAME",
    "trial",
    "trial_id",
    "trial_global"
  )

  unique(candidates[candidates %in% names_dat])
}

gpbiometrics_external_group_id <- function(dat, group_cols) {
  if (length(group_cols) == 0) {
    return(rep("all", nrow(dat)))
  }

  group_dat <- dat[group_cols]

  group_dat[] <- lapply(group_dat, function(x) {
    x_chr <- as.character(x)
    x_chr[is.na(x_chr)] <- "<NA>"
    x_chr
  })

  apply(group_dat, 1, paste, collapse = "||")
}

gpbiometrics_external_validate_positive_scalar <- function(x, name) {
  if (!is.numeric(x) || length(x) != 1 || !is.finite(x) || x <= 0) {
    stop("`", name, "` must be a single positive finite number.", call. = FALSE)
  }

  invisible(TRUE)
}

gpbiometrics_external_validate_nonnegative_scalar <- function(x, name) {
  if (!is.numeric(x) || length(x) != 1 || !is.finite(x) || x < 0) {
    stop("`", name, "` must be a single non-negative finite number.", call. = FALSE)
  }

  invisible(TRUE)
}

gpbiometrics_external_detect_ibi_unit <- function(x, unit) {
  if (!identical(unit, "auto")) {
    return(unit)
  }

  finite_x <- x[is.finite(x) & x > 0]

  if (length(finite_x) == 0) {
    return("ms")
  }

  med <- stats::median(finite_x, na.rm = TRUE)

  if (is.finite(med) && med > 0.2 && med < 5) {
    "seconds"
  } else {
    "ms"
  }
}

gpbiometrics_external_collapse_repeated_intervals <- function(x,
                                                              tolerance_ms) {
  if (length(x) <= 1) {
    return(x)
  }

  keep <- c(TRUE, abs(diff(x)) > tolerance_ms)
  x[keep]
}

gpbiometrics_external_time_seconds <- function(time_raw,
                                               sampling_rate = NULL) {
  if (!is.null(sampling_rate)) {
    return((seq_along(time_raw) - 1) / sampling_rate)
  }

  finite_time <- time_raw[is.finite(time_raw)]

  if (length(finite_time) < 2) {
    return(seq_along(time_raw) - 1)
  }

  shifted <- time_raw - min(finite_time, na.rm = TRUE)

  if (max(shifted, na.rm = TRUE) > 10000) {
    shifted / 1000
  } else {
    shifted
  }
}

gpbiometrics_external_rhrv_group_summary <- function(beat_table,
                                                     group_cols) {
  if (nrow(beat_table) == 0) {
    return(data.frame())
  }

  group_ids <- unique(beat_table$group_id)

  out <- lapply(group_ids, function(group_id) {
    d <- beat_table[beat_table$group_id == group_id, , drop = FALSE]

    row <- data.frame(
      group_id = group_id,
      beat_rows = nrow(d),
      duration_s = if (nrow(d) > 0) max(d$time_s, na.rm = TRUE) else NA_real_,
      mean_ibi_ms = mean(d$ibi_ms, na.rm = TRUE),
      median_ibi_ms = stats::median(d$ibi_ms, na.rm = TRUE),
      min_ibi_ms = min(d$ibi_ms, na.rm = TRUE),
      max_ibi_ms = max(d$ibi_ms, na.rm = TRUE),
      input_interval_rows = if ("input_interval_rows" %in% names(d)) d$input_interval_rows[1] else NA_integer_,
      used_intervals_after_collapse = if ("used_intervals_after_collapse" %in% names(d)) d$used_intervals_after_collapse[1] else NA_integer_,
      stringsAsFactors = FALSE
    )

    if (length(group_cols) > 0 && all(group_cols %in% names(d))) {
      row <- cbind(d[1, group_cols, drop = FALSE], row)
    }

    row
  })

  out <- do.call(rbind, out)
  rownames(out) <- NULL

  out
}

gpbiometrics_external_eda_group_summary <- function(eda_table,
                                                    group_cols) {
  if (nrow(eda_table) == 0) {
    return(data.frame())
  }

  group_ids <- unique(eda_table$group_id)

  out <- lapply(group_ids, function(group_id) {
    d <- eda_table[eda_table$group_id == group_id, , drop = FALSE]
    finite_eda <- d$eda[is.finite(d$eda)]

    row <- data.frame(
      group_id = group_id,
      rows = nrow(d),
      finite_rows = length(finite_eda),
      missing_rows = sum(!is.finite(d$eda)),
      mean_eda = if (length(finite_eda) > 0) mean(finite_eda, na.rm = TRUE) else NA_real_,
      median_eda = if (length(finite_eda) > 0) stats::median(finite_eda, na.rm = TRUE) else NA_real_,
      min_time_s = min(d$time_s, na.rm = TRUE),
      max_time_s = max(d$time_s, na.rm = TRUE),
      stringsAsFactors = FALSE
    )

    if (length(group_cols) > 0 && all(group_cols %in% names(d))) {
      row <- cbind(d[1, group_cols, drop = FALSE], row)
    }

    row
  })

  out <- do.call(rbind, out)
  rownames(out) <- NULL

  out
}

gpbiometrics_external_write_group_files <- function(data,
                                                    group_cols,
                                                    output_dir,
                                                    prefix,
                                                    suffix) {
  if (is.null(output_dir)) {
    return(data.frame(
      group_id = unique(data$group_id),
      file_path = NA_character_,
      stringsAsFactors = FALSE
    ))
  }

  dir.create(output_dir, recursive = TRUE, showWarnings = FALSE)

  group_ids <- unique(data$group_id)

  manifest <- lapply(seq_along(group_ids), function(i) {
    group_id <- group_ids[i]
    d <- data[data$group_id == group_id, , drop = FALSE]

    file_name <- paste0(
      prefix,
      "_",
      sprintf("%03d", i),
      "_",
      gpbiometrics_external_safe_file_name(group_id),
      "_",
      suffix,
      ".csv"
    )

    file_path <- file.path(output_dir, file_name)
    utils::write.csv(d, file_path, row.names = FALSE)

    row <- data.frame(
      group_id = group_id,
      file_path = file_path,
      stringsAsFactors = FALSE
    )

    if (length(group_cols) > 0 && all(group_cols %in% names(d))) {
      row <- cbind(d[1, group_cols, drop = FALSE], row)
    }

    row
  })

  manifest <- do.call(rbind, manifest)
  rownames(manifest) <- NULL

  manifest
}

gpbiometrics_external_safe_file_name <- function(x) {
  x <- gsub("[^A-Za-z0-9_-]+", "_", x)
  x <- gsub("_+", "_", x)
  x <- gsub("^_|_$", "", x)

  if (nchar(x) == 0) {
    "group"
  } else {
    substr(x, 1, 80)
  }
}

gpbiometrics_external_python_has_neurokit2 <- function(python_path) {
  status <- try(
    system2(
      python_path,
      args = c("-c", "import neurokit2"),
      stdout = TRUE,
      stderr = TRUE
    ),
    silent = TRUE
  )

  !inherits(status, "try-error")
}

gpbiometrics_external_write_neurokit_script <- function(script_file,
                                                        input_file,
                                                        output_file,
                                                        sampling_rate) {
  lines <- c(
    "import pandas as pd",
    "import neurokit2 as nk",
    paste0("input_file = r'''", input_file, "'''"),
    paste0("output_file = r'''", output_file, "'''"),
    paste0("sampling_rate = ", sampling_rate),
    "df = pd.read_csv(input_file)",
    "rows = []",
    "for group_id, g in df.groupby('group_id'):",
    "    eda = pd.to_numeric(g['eda'], errors='coerce').dropna().values",
    "    row = {'group_id': group_id, 'rows': len(g), 'finite_rows': len(eda)}",
    "    try:",
    "        if len(eda) > 3:",
    "            signals, info = nk.eda_process(eda, sampling_rate=sampling_rate)",
    "            row['scr_peaks'] = int(signals['SCR_Peaks'].sum()) if 'SCR_Peaks' in signals else None",
    "            row['eda_clean_mean'] = float(signals['EDA_Clean'].mean()) if 'EDA_Clean' in signals else None",
    "            row['status'] = 'processed'",
    "        else:",
    "            row['scr_peaks'] = None",
    "            row['eda_clean_mean'] = None",
    "            row['status'] = 'too_few_samples'",
    "    except Exception as e:",
    "        row['scr_peaks'] = None",
    "        row['eda_clean_mean'] = None",
    "        row['status'] = 'error: ' + str(e)",
    "    rows.append(row)",
    "pd.DataFrame(rows).to_csv(output_file, index=False)"
  )

  writeLines(lines, script_file)
  invisible(script_file)
}

test_that("export_gazepoint_rhrv_input prepares beat tables", {
  dat <- data.frame(
    participant = "P1",
    IBI_clean_ms = c(1000, 1000, 1020, 1020, 980)
  )

  res <- export_gazepoint_rhrv_input(
    dat,
    ibi_col = "IBI_clean_ms",
    group_cols = "participant",
    collapse_repeated_intervals = TRUE
  )

  expect_s3_class(res, "gazepoint_rhrv_input_export")
  expect_equal(res$overview$status, "rhrv_input_prepared")
  expect_equal(res$overview$beat_rows, 3)
  expect_equal(res$beat_table$ibi_ms, c(1000, 1020, 980))
  expect_true(all(c("time_s", "ibi_ms", "ibi_s") %in% names(res$beat_table)))
})

test_that("export_gazepoint_rhrv_input can retain repeated intervals", {
  dat <- data.frame(
    participant = "P1",
    IBI_clean_ms = c(1000, 1000, 1020)
  )

  res <- export_gazepoint_rhrv_input(
    dat,
    ibi_col = "IBI_clean_ms",
    group_cols = "participant",
    collapse_repeated_intervals = FALSE
  )

  expect_equal(res$overview$beat_rows, 3)
  expect_equal(res$beat_table$used_intervals_after_collapse[1], 3)
})

test_that("export_gazepoint_rhrv_input writes group files", {
  dat <- data.frame(
    participant = c("P1", "P1", "P2", "P2"),
    IBI_clean_ms = c(1000, 1020, 900, 910)
  )

  out_dir <- tempfile("rhrv_export_")

  res <- export_gazepoint_rhrv_input(
    dat,
    ibi_col = "IBI_clean_ms",
    group_cols = "participant",
    output_dir = out_dir
  )

  expect_equal(res$overview$status, "rhrv_input_exported")
  expect_equal(sum(file.exists(res$manifest$file_path)), 2)

  unlink(out_dir, recursive = TRUE, force = TRUE)
})

test_that("prepare_gazepoint_neurokit_eda_input prepares EDA tables", {
  dat <- data.frame(
    participant = "P1",
    CNT = 0:4,
    GSR_US = c(1, 1.1, 1.2, 1.1, 1.0)
  )

  res <- prepare_gazepoint_neurokit_eda_input(
    dat,
    eda_col = "GSR_US",
    time_col = "CNT",
    group_cols = "participant",
    sampling_rate = 10
  )

  expect_s3_class(res, "gazepoint_neurokit_eda_input")
  expect_equal(res$overview$status, "neurokit_eda_input_prepared")
  expect_equal(res$overview$eda_rows, 5)
  expect_equal(res$eda_table$time_s, c(0, 0.1, 0.2, 0.3, 0.4))
})

test_that("prepare_gazepoint_neurokit_eda_input writes group files", {
  dat <- data.frame(
    participant = c("P1", "P1", "P2", "P2"),
    CNT = c(0, 1, 0, 1),
    GSR_US = c(1, 2, 3, 4)
  )

  out_dir <- tempfile("eda_export_")

  res <- prepare_gazepoint_neurokit_eda_input(
    dat,
    eda_col = "GSR_US",
    time_col = "CNT",
    group_cols = "participant",
    output_dir = out_dir
  )

  expect_equal(res$overview$status, "neurokit_eda_input_exported")
  expect_equal(sum(file.exists(res$manifest$file_path)), 2)

  unlink(out_dir, recursive = TRUE, force = TRUE)
})

test_that("run_gazepoint_neurokit_eda_crosscheck skips by default", {
  dat <- data.frame(
    participant = "P1",
    CNT = 0:4,
    GSR_US = c(1, 1.1, 1.2, 1.1, 1.0)
  )

  res <- run_gazepoint_neurokit_eda_crosscheck(
    dat,
    eda_col = "GSR_US",
    time_col = "CNT",
    group_cols = "participant",
    sampling_rate = 10,
    execute = FALSE
  )

  expect_s3_class(res, "gazepoint_neurokit_eda_crosscheck")
  expect_equal(res$overview$status, "skipped_execute_false")
  expect_false(res$overview$executed)
  expect_s3_class(res$prepared_input, "gazepoint_neurokit_eda_input")
})

test_that("run_gazepoint_neurokit_eda_crosscheck accepts prepared input", {
  dat <- data.frame(
    participant = "P1",
    CNT = 0:4,
    GSR_US = c(1, 1.1, 1.2, 1.1, 1.0)
  )

  prepared <- prepare_gazepoint_neurokit_eda_input(
    dat,
    eda_col = "GSR_US",
    time_col = "CNT",
    group_cols = "participant",
    sampling_rate = 10
  )

  res <- run_gazepoint_neurokit_eda_crosscheck(
    prepared,
    sampling_rate = 10,
    execute = FALSE
  )

  expect_equal(res$overview$status, "skipped_execute_false")
  expect_equal(res$overview$input_rows, 5)
})

test_that("external interoperability helpers validate inputs", {
  dat <- data.frame(
    participant = "P1",
    IBI_clean_ms = 1000,
    GSR_US = 1
  )

  expect_error(
    export_gazepoint_rhrv_input(dat, ibi_col = "missing"),
    "`ibi_col`"
  )

  expect_error(
    prepare_gazepoint_neurokit_eda_input(dat, eda_col = "missing"),
    "`eda_col`"
  )

  expect_error(
    run_gazepoint_neurokit_eda_crosscheck(dat, execute = NA),
    "`execute` must be TRUE or FALSE"
  )
})

#' Audit likely GSR/EDA units
#'
#' Checks whether a Gazepoint GSR/EDA column looks more like conductance
#' in microSiemens or resistance/impedance-like values in Ohms. This is a
#' preprocessing safety audit, not a definitive device calibration test.
#'
#' @param dat A data frame.
#' @param gsr_col Name of the GSR/EDA column to audit.
#' @param convert Logical. If `TRUE`, add a conductance-converted column when
#'   the signal is likely resistance/impedance-like.
#' @param output_col Output column used when `convert = TRUE`.
#' @param resistance_to_us_factor Conversion factor. For Ohms to microSiemens,
#'   use `1000000 / resistance`.
#'
#' @return A list with `overview`, `diagnostics`, `recommendation`, and,
#'   when requested, `data`.
#' @export
audit_gazepoint_gsr_units <- function(dat,
                                      gsr_col = "GSR",
                                      convert = FALSE,
                                      output_col = NULL,
                                      resistance_to_us_factor = 1000000) {
  if (!is.data.frame(dat)) {
    stop("`dat` must be a data frame.", call. = FALSE)
  }

  if (!gsr_col %in% names(dat)) {
    stop("Column `", gsr_col, "` was not found in `dat`.", call. = FALSE)
  }

  x <- dat[[gsr_col]]

  if (!is.numeric(x)) {
    stop("`gsr_col` must identify a numeric column.", call. = FALSE)
  }

  finite_x <- x[is.finite(x)]

  if (length(finite_x) == 0) {
    stop("`gsr_col` contains no finite numeric values.", call. = FALSE)
  }

  q <- stats::quantile(
    finite_x,
    probs = c(0, 0.01, 0.25, 0.5, 0.75, 0.99, 1),
    na.rm = TRUE,
    names = FALSE
  )

  names(q) <- c("min", "q01", "q25", "median", "q75", "q99", "max")

  col_lower <- tolower(gsr_col)

  likely_unit <- "ambiguous"
  confidence <- "low"

  if (grepl("_us$|us$|microsiemens|micro_siemens|conductance", col_lower)) {
    likely_unit <- "conductance_microSiemens"
    confidence <- "high_column_name"
  } else if (q["median"] > 1000 || q["q75"] > 1000) {
    likely_unit <- "resistance_or_impedance_ohms"
    confidence <- "high_numeric_range"
  } else if (q["median"] > 0 && q["median"] <= 100 && q["q99"] <= 500) {
    likely_unit <- "conductance_microSiemens"
    confidence <- "moderate_numeric_range"
  } else if (q["median"] > 100 && q["median"] <= 1000) {
    likely_unit <- "ambiguous_large_conductance_or_scaled_signal"
    confidence <- "low_numeric_range"
  }

  threshold_warning <- if (likely_unit == "resistance_or_impedance_ohms") {
    paste(
      "Do not apply SCR thresholds expressed in microSiemens directly to this column.",
      "Convert resistance-like values to conductance first or use a verified conductance column such as GSR_US."
    )
  } else if (likely_unit == "conductance_microSiemens") {
    "SCR thresholds expressed in microSiemens may be appropriate if the column is verified as conductance."
  } else {
    "Verify device/export documentation before applying SCR thresholds."
  }

  diagnostics <- data.frame(
    gsr_col = gsr_col,
    n_rows = length(x),
    n_finite = length(finite_x),
    missing_rate = mean(!is.finite(x)),
    min = q["min"],
    q01 = q["q01"],
    q25 = q["q25"],
    median = q["median"],
    q75 = q["q75"],
    q99 = q["q99"],
    max = q["max"],
    likely_unit = likely_unit,
    confidence = confidence,
    stringsAsFactors = FALSE
  )

  overview <- data.frame(
    gsr_col = gsr_col,
    likely_unit = likely_unit,
    confidence = confidence,
    convert_requested = isTRUE(convert),
    status = if (likely_unit == "resistance_or_impedance_ohms") {
      "unit_warning_resistance_like"
    } else if (likely_unit == "conductance_microSiemens") {
      "unit_audit_conductance_like"
    } else {
      "unit_audit_ambiguous"
    },
    interpretation = paste(
      "This audit uses column names and numerical ranges to flag likely GSR/EDA units.",
      "It is a preprocessing sanity check and does not replace device/export documentation."
    ),
    stringsAsFactors = FALSE
  )

  recommendation <- data.frame(
    gsr_col = gsr_col,
    likely_unit = likely_unit,
    recommendation = threshold_warning,
    stringsAsFactors = FALSE
  )

  result <- list(
    overview = overview,
    diagnostics = diagnostics,
    recommendation = recommendation,
    settings = list(
      gsr_col = gsr_col,
      convert = convert,
      output_col = output_col,
      resistance_to_us_factor = resistance_to_us_factor
    )
  )

  if (isTRUE(convert)) {
    if (is.null(output_col)) {
      output_col <- paste0(gsr_col, "_converted_us")
    }

    out <- dat

    if (likely_unit == "resistance_or_impedance_ohms") {
      out[[output_col]] <- ifelse(
        is.finite(out[[gsr_col]]) & out[[gsr_col]] > 0,
        resistance_to_us_factor / out[[gsr_col]],
        NA_real_
      )
      attr(out, "gsr_unit_conversion") <- "resistance_or_impedance_to_microSiemens"
    } else {
      out[[output_col]] <- out[[gsr_col]]
      attr(out, "gsr_unit_conversion") <- "copied_without_conversion_unit_not_resistance_like"
    }

    result$data <- out
  }

  class(result) <- c("gazepoint_gsr_unit_audit", "list")
  result
}

#' Summarise validity and availability of Gazepoint biometric signals
#'
#' Summarises missingness, finite numeric availability, variability, and optional
#' validity-flag columns for Gazepoint Biometrics data. The helper is descriptive:
#' it reports whether biometric signals appear available and usable, but it does
#' not infer emotion, valence, or HRV from ambiguous raw columns. In particular,
#' raw `HRV` columns are treated as validity/vendor flags unless the user has
#' independent documentation proving otherwise.
#'
#' @param data A data frame.
#' @param signal_cols Optional character vector of biometric signal columns to
#'   summarise. If `NULL`, common Gazepoint biometric signal columns are detected
#'   from the column names.
#' @param validity_cols Optional character vector of validity-flag columns to
#'   summarise. If `NULL`, common validity-like columns are detected, including
#'   `HRV` when present.
#' @param group_cols Optional character vector of grouping columns, such as
#'   participant, stimulus, trial, or condition columns.
#' @param active_min_unique Minimum number of unique finite values required for
#'   a numeric signal to be treated as active.
#'
#' @return A list with `overview`, `signals`, `validity_flags`,
#'   `group_summary`, and `settings`.
#'
#' @examples
#' df <- data.frame(
#'   USER = rep(c("P1", "P2"), each = 4),
#'   GSR = c(1, 2, NA, 4, 2, 2, 2, 2),
#'   HR = c(70, 71, 72, NA, 80, 81, 82, 83),
#'   HRV = c(1, 1, 0, 1, 1, 1, 1, 1)
#' )
#' summarise_gazepoint_biometric_validity(df, group_cols = "USER")
#'
#' @export
summarise_gazepoint_biometric_validity <- function(data,
                                                   signal_cols = NULL,
                                                   validity_cols = NULL,
                                                   group_cols = NULL,
                                                   active_min_unique = 2L) {
  if (!is.data.frame(data)) {
    stop("`data` must be a data frame.", call. = FALSE)
  }

  if (!is.null(signal_cols)) {
    .gpbiom_assert_columns(data, signal_cols, "signal_cols")
  }

  if (!is.null(validity_cols)) {
    .gpbiom_assert_columns(data, validity_cols, "validity_cols")
  }

  if (!is.null(group_cols)) {
    .gpbiom_assert_columns(data, group_cols, "group_cols")
  }

  if (!is.numeric(active_min_unique) ||
      length(active_min_unique) != 1L ||
      is.na(active_min_unique) ||
      active_min_unique < 1) {
    stop("`active_min_unique` must be a positive number.", call. = FALSE)
  }

  active_min_unique <- as.integer(active_min_unique)

  if (is.null(signal_cols)) {
    signal_cols <- .gpbiom_detect_signal_columns(data)
  }

  if (is.null(validity_cols)) {
    validity_cols <- .gpbiom_detect_validity_columns(data)
  }

  signal_summary <- .gpbiom_summarise_signal_columns(
    data = data,
    signal_cols = signal_cols,
    active_min_unique = active_min_unique
  )

  validity_summary <- .gpbiom_summarise_validity_columns(
    data = data,
    validity_cols = validity_cols
  )

  group_summary <- .gpbiom_summarise_signal_groups(
    data = data,
    signal_cols = signal_cols,
    group_cols = group_cols,
    active_min_unique = active_min_unique
  )

  active_signal_count <- sum(signal_summary$status == "active_signal")
  inactive_signal_count <- nrow(signal_summary) - active_signal_count

  overview <- data.frame(
    n_rows = nrow(data),
    n_columns = ncol(data),
    signal_column_count = length(signal_cols),
    active_signal_count = active_signal_count,
    inactive_signal_count = inactive_signal_count,
    validity_flag_column_count = length(validity_cols),
    group_column_count = length(group_cols),
    status = if (length(signal_cols) == 0L) {
      "no_biometric_signal_columns_detected"
    } else if (active_signal_count == 0L) {
      "no_active_biometric_signals_detected"
    } else if (inactive_signal_count > 0L) {
      "some_biometric_signals_inactive_or_limited"
    } else {
      "biometric_signals_available"
    },
    stringsAsFactors = FALSE
  )

  settings <- list(
    signal_cols = signal_cols,
    validity_cols = validity_cols,
    group_cols = group_cols,
    active_min_unique = active_min_unique,
    notes = c(
      "GSR/EDA availability does not identify emotional valence.",
      "Heart-rate availability requires baseline/task context for interpretation.",
      "Raw HRV columns are treated as validity/vendor flags unless independently documented as HRV metrics."
    )
  )

  list(
    overview = overview,
    signals = signal_summary,
    validity_flags = validity_summary,
    group_summary = group_summary,
    settings = settings
  )
}


#' Flag biometric dropouts and flatline periods
#'
#' Flags missing-value runs and sustained flatline runs in Gazepoint biometric
#' signal columns. Missing dropouts are defined as consecutive missing or
#' non-finite numeric samples. Flatline dropouts are defined as consecutive
#' finite numeric samples that remain unchanged within a tolerance.
#'
#' The function adds row-level flags and stores a dropout summary in the returned
#' data frame attributes. It does not remove rows.
#'
#' @param data A data frame.
#' @param signal_cols Optional character vector of biometric signal columns. If
#'   `NULL`, common Gazepoint biometric signal columns are detected.
#' @param group_cols Optional grouping columns. Runs are computed separately
#'   within each group.
#' @param time_col Optional time column used to order rows within each group
#'   before run detection. If `NULL`, the current row order is used.
#' @param min_missing_run Minimum consecutive missing/non-finite samples required
#'   to flag a missing dropout.
#' @param min_flatline_run Minimum consecutive unchanged finite samples required
#'   to flag a flatline dropout.
#' @param constant_tolerance Numeric tolerance used when detecting unchanged
#'   values for flatline runs.
#' @param prefix Prefix for generated dropout columns.
#'
#' @return The input data frame with added logical dropout columns. The attributes
#'   `dropout_summary` and `dropout_settings` contain structured summaries.
#'
#' @examples
#' df <- data.frame(
#'   CNT = 1:8,
#'   GSR = c(1, NA, NA, NA, 2, 2, 2, 3),
#'   HR = c(70, 71, 72, 73, 74, 75, 76, 77)
#' )
#' flag_gazepoint_biometric_dropouts(df, min_missing_run = 3, min_flatline_run = 3)
#'
#' @export
flag_gazepoint_biometric_dropouts <- function(data,
                                              signal_cols = NULL,
                                              group_cols = NULL,
                                              time_col = NULL,
                                              min_missing_run = 5L,
                                              min_flatline_run = 10L,
                                              constant_tolerance = 0,
                                              prefix = "biometric_dropout") {
  if (!is.data.frame(data)) {
    stop("`data` must be a data frame.", call. = FALSE)
  }

  if (!is.null(signal_cols)) {
    .gpbiom_assert_columns(data, signal_cols, "signal_cols")
  }

  if (!is.null(group_cols)) {
    .gpbiom_assert_columns(data, group_cols, "group_cols")
  }

  if (!is.null(time_col)) {
    .gpbiom_assert_columns(data, time_col, "time_col")
  }

  .gpbiom_assert_positive_integer(min_missing_run, "min_missing_run")
  .gpbiom_assert_positive_integer(min_flatline_run, "min_flatline_run")

  if (!is.numeric(constant_tolerance) ||
      length(constant_tolerance) != 1L ||
      is.na(constant_tolerance) ||
      constant_tolerance < 0) {
    stop("`constant_tolerance` must be a non-negative number.", call. = FALSE)
  }

  if (!is.character(prefix) || length(prefix) != 1L || is.na(prefix) ||
      !nzchar(prefix)) {
    stop("`prefix` must be a non-empty character string.", call. = FALSE)
  }

  if (is.null(signal_cols)) {
    signal_cols <- .gpbiom_detect_signal_columns(data)
  }

  out <- data

  if (length(signal_cols) == 0L) {
    any_col <- paste0(prefix, "_any")
    out[[any_col]] <- FALSE

    attr(out, "dropout_summary") <- .gpbiom_empty_dropout_summary()
    attr(out, "dropout_settings") <- list(
      signal_cols = signal_cols,
      group_cols = group_cols,
      time_col = time_col,
      min_missing_run = as.integer(min_missing_run),
      min_flatline_run = as.integer(min_flatline_run),
      constant_tolerance = constant_tolerance,
      prefix = prefix
    )

    return(out)
  }

  row_order_groups <- .gpbiom_group_indices(out, group_cols)

  added_signal_cols <- character()

  for (signal in signal_cols) {
    safe_signal <- .gpbiom_safe_colname(signal)

    missing_col <- paste0(prefix, "_", safe_signal, "_missing")
    flatline_col <- paste0(prefix, "_", safe_signal, "_flatline")
    signal_any_col <- paste0(prefix, "_", safe_signal, "_any")

    out[[missing_col]] <- FALSE
    out[[flatline_col]] <- FALSE

    for (indices in row_order_groups) {
      ordered_indices <- .gpbiom_order_indices(out, indices, time_col)
      values <- out[[signal]][ordered_indices]

      missing_flags <- .gpbiom_missing_run_flags(values, min_missing_run)
      flatline_flags <- .gpbiom_flatline_run_flags(
        values = values,
        min_flatline_run = min_flatline_run,
        constant_tolerance = constant_tolerance
      )

      out[[missing_col]][ordered_indices] <- missing_flags
      out[[flatline_col]][ordered_indices] <- flatline_flags
    }

    out[[signal_any_col]] <- out[[missing_col]] | out[[flatline_col]]
    added_signal_cols <- c(added_signal_cols, signal_any_col)
  }

  any_col <- paste0(prefix, "_any")
  out[[any_col]] <- Reduce(`|`, out[added_signal_cols])

  dropout_summary <- .gpbiom_dropout_summary(
    out = out,
    signal_cols = signal_cols,
    prefix = prefix
  )

  attr(out, "dropout_summary") <- dropout_summary
  attr(out, "dropout_settings") <- list(
    signal_cols = signal_cols,
    group_cols = group_cols,
    time_col = time_col,
    min_missing_run = as.integer(min_missing_run),
    min_flatline_run = as.integer(min_flatline_run),
    constant_tolerance = constant_tolerance,
    prefix = prefix
  )

  out
}


.gpbiom_assert_columns <- function(data, columns, argument_name) {
  if (!is.character(columns) || anyNA(columns)) {
    stop("`", argument_name, "` must be a character vector of column names.",
         call. = FALSE)
  }

  missing <- setdiff(columns, names(data))

  if (length(missing) > 0L) {
    stop(
      "`", argument_name, "` contains columns not found in `data`: ",
      paste(missing, collapse = ", "),
      call. = FALSE
    )
  }

  invisible(TRUE)
}


.gpbiom_assert_positive_integer <- function(x, argument_name) {
  if (!is.numeric(x) ||
      length(x) != 1L ||
      is.na(x) ||
      x < 1 ||
      x != as.integer(x)) {
    stop("`", argument_name, "` must be a positive integer.", call. = FALSE)
  }

  invisible(TRUE)
}


.gpbiom_detect_signal_columns <- function(data) {
  mapping <- standardise_gazepoint_biometric_names(data, rename = FALSE)

  signal_standards <- c(
    "GSR",
    "GSR_US",
    "GSR_OHMS",
    "HR",
    "IBI",
    "ENGAGEMENT"
  )

  mapping$original_name[mapping$standard_name %in% signal_standards]
}


.gpbiom_detect_validity_columns <- function(data) {
  mapping <- standardise_gazepoint_biometric_names(data, rename = FALSE)
  cleaned <- .gpbiom_clean_name(mapping$original_name)

  validity_like <- mapping$standard_name %in% c("HRV") |
    grepl("valid|validity|quality", cleaned)

  mapping$original_name[validity_like]
}


.gpbiom_signal_type <- function(column) {
  standard <- standardise_gazepoint_biometric_names(column)

  if (standard %in% c("GSR", "GSR_US", "GSR_OHMS")) {
    return("gsr_eda")
  }

  if (identical(standard, "HR")) {
    return("heart_rate")
  }

  if (identical(standard, "IBI")) {
    return("ibi")
  }

  if (identical(standard, "ENGAGEMENT")) {
    return("engagement_dial")
  }

  "other"
}


.gpbiom_summarise_signal_columns <- function(data,
                                             signal_cols,
                                             active_min_unique) {
  if (length(signal_cols) == 0L) {
    return(data.frame(
      column = character(),
      signal_type = character(),
      n = integer(),
      n_missing = integer(),
      missing_rate = numeric(),
      n_non_missing = integer(),
      n_finite = integer(),
      finite_rate = numeric(),
      n_unique_finite = integer(),
      mean = numeric(),
      median = numeric(),
      sd = numeric(),
      min = numeric(),
      max = numeric(),
      status = character(),
      stringsAsFactors = FALSE
    ))
  }

  rows <- lapply(signal_cols, function(column) {
    values <- data[[column]]
    n <- length(values)
    missing <- is.na(values)

    if (is.numeric(values)) {
      finite <- is.finite(values)
      finite_values <- values[finite]
      n_unique_finite <- length(unique(finite_values))

      status <- if (length(finite_values) == 0L) {
        "all_missing_or_nonfinite"
      } else if (n_unique_finite < active_min_unique) {
        "constant_or_low_variability_signal"
      } else {
        "active_signal"
      }

      data.frame(
        column = column,
        signal_type = .gpbiom_signal_type(column),
        n = n,
        n_missing = sum(missing),
        missing_rate = if (n > 0L) mean(missing) else NA_real_,
        n_non_missing = sum(!missing),
        n_finite = length(finite_values),
        finite_rate = if (n > 0L) length(finite_values) / n else NA_real_,
        n_unique_finite = n_unique_finite,
        mean = if (length(finite_values) > 0L) mean(finite_values) else NA_real_,
        median = if (length(finite_values) > 0L) stats::median(finite_values) else NA_real_,
        sd = if (length(finite_values) > 1L) stats::sd(finite_values) else NA_real_,
        min = if (length(finite_values) > 0L) min(finite_values) else NA_real_,
        max = if (length(finite_values) > 0L) max(finite_values) else NA_real_,
        status = status,
        stringsAsFactors = FALSE
      )
    } else {
      non_missing_values <- values[!missing]
      n_unique <- length(unique(non_missing_values))

      status <- if (length(non_missing_values) == 0L) {
        "all_missing"
      } else if (n_unique < active_min_unique) {
        "constant_or_low_variability_signal"
      } else {
        "non_numeric_signal_present"
      }

      data.frame(
        column = column,
        signal_type = .gpbiom_signal_type(column),
        n = n,
        n_missing = sum(missing),
        missing_rate = if (n > 0L) mean(missing) else NA_real_,
        n_non_missing = sum(!missing),
        n_finite = NA_integer_,
        finite_rate = NA_real_,
        n_unique_finite = NA_integer_,
        mean = NA_real_,
        median = NA_real_,
        sd = NA_real_,
        min = NA_real_,
        max = NA_real_,
        status = status,
        stringsAsFactors = FALSE
      )
    }
  })

  do.call(rbind, rows)
}


.gpbiom_summarise_validity_columns <- function(data, validity_cols) {
  if (length(validity_cols) == 0L) {
    return(data.frame(
      column = character(),
      standard_name = character(),
      n = integer(),
      n_missing = integer(),
      missing_rate = numeric(),
      n_valid_like = integer(),
      valid_like_rate = numeric(),
      n_invalid_like = integer(),
      invalid_like_rate = numeric(),
      interpretation_note = character(),
      stringsAsFactors = FALSE
    ))
  }

  rows <- lapply(validity_cols, function(column) {
    values <- data[[column]]
    n <- length(values)
    missing <- is.na(values)

    valid_like <- .gpbiom_valid_like(values)
    invalid_like <- .gpbiom_invalid_like(values)

    standard <- standardise_gazepoint_biometric_names(column)

    note <- if (identical(standard, "HRV")) {
      "Treated as a validity/vendor flag, not as an HRV metric."
    } else {
      "Detected as a validity-like column from its name."
    }

    data.frame(
      column = column,
      standard_name = standard,
      n = n,
      n_missing = sum(missing),
      missing_rate = if (n > 0L) mean(missing) else NA_real_,
      n_valid_like = sum(valid_like, na.rm = TRUE),
      valid_like_rate = if (n > 0L) mean(valid_like, na.rm = TRUE) else NA_real_,
      n_invalid_like = sum(invalid_like, na.rm = TRUE),
      invalid_like_rate = if (n > 0L) mean(invalid_like, na.rm = TRUE) else NA_real_,
      interpretation_note = note,
      stringsAsFactors = FALSE
    )
  })

  do.call(rbind, rows)
}


.gpbiom_valid_like <- function(values) {
  if (is.logical(values)) {
    return(!is.na(values) & values)
  }

  if (is.numeric(values)) {
    return(!is.na(values) & is.finite(values) & values > 0)
  }

  cleaned <- tolower(trimws(as.character(values)))

  !is.na(values) & cleaned %in% c(
    "1", "true", "valid", "yes", "y", "ok", "good", "pass"
  )
}


.gpbiom_invalid_like <- function(values) {
  if (is.logical(values)) {
    return(!is.na(values) & !values)
  }

  if (is.numeric(values)) {
    return(!is.na(values) & is.finite(values) & values <= 0)
  }

  cleaned <- tolower(trimws(as.character(values)))

  !is.na(values) & cleaned %in% c(
    "0", "false", "invalid", "no", "n", "bad", "fail"
  )
}


.gpbiom_summarise_signal_groups <- function(data,
                                            signal_cols,
                                            group_cols,
                                            active_min_unique) {
  if (length(signal_cols) == 0L || length(group_cols) == 0L) {
    return(data.frame(
      group = character(),
      signal_column_count = integer(),
      active_signal_count = integer(),
      status = character(),
      stringsAsFactors = FALSE
    ))
  }

  split_key <- interaction(data[group_cols], drop = TRUE, lex.order = TRUE)
  indices <- split(seq_len(nrow(data)), split_key)

  rows <- lapply(names(indices), function(group_name) {
    group_data <- data[indices[[group_name]], , drop = FALSE]
    signal_summary <- .gpbiom_summarise_signal_columns(
      data = group_data,
      signal_cols = signal_cols,
      active_min_unique = active_min_unique
    )

    active_signal_count <- sum(signal_summary$status == "active_signal")

    data.frame(
      group = group_name,
      n_rows = nrow(group_data),
      signal_column_count = length(signal_cols),
      active_signal_count = active_signal_count,
      status = if (active_signal_count == 0L) {
        "no_active_signals_in_group"
      } else if (active_signal_count < length(signal_cols)) {
        "some_signals_inactive_or_limited_in_group"
      } else {
        "signals_available_in_group"
      },
      stringsAsFactors = FALSE
    )
  })

  do.call(rbind, rows)
}


.gpbiom_group_indices <- function(data, group_cols) {
  if (length(group_cols) == 0L) {
    return(list(seq_len(nrow(data))))
  }

  split_key <- interaction(data[group_cols], drop = TRUE, lex.order = TRUE)
  unname(split(seq_len(nrow(data)), split_key))
}


.gpbiom_order_indices <- function(data, indices, time_col) {
  if (is.null(time_col)) {
    return(indices)
  }

  values <- data[[time_col]][indices]
  indices[order(values, na.last = TRUE)]
}


.gpbiom_missing_run_flags <- function(values, min_missing_run) {
  missing <- is.na(values)

  if (is.numeric(values)) {
    missing <- missing | !is.finite(values)
  }

  .gpbiom_run_flags(missing, min_missing_run)
}


.gpbiom_flatline_run_flags <- function(values,
                                       min_flatline_run,
                                       constant_tolerance) {
  if (!is.numeric(values)) {
    return(rep(FALSE, length(values)))
  }

  valid <- !is.na(values) & is.finite(values)

  if (length(values) == 0L || sum(valid) < min_flatline_run) {
    return(rep(FALSE, length(values)))
  }

  flags <- rep(FALSE, length(values))
  run_start <- NA_integer_
  run_length <- 0L

  for (i in seq_along(values)) {
    if (!valid[i]) {
      if (!is.na(run_start) && run_length >= min_flatline_run) {
        flags[run_start:(i - 1L)] <- TRUE
      }
      run_start <- NA_integer_
      run_length <- 0L
      next
    }

    if (is.na(run_start)) {
      run_start <- i
      run_length <- 1L
      next
    }

    previous <- values[i - 1L]

    if (valid[i - 1L] && abs(values[i] - previous) <= constant_tolerance) {
      run_length <- run_length + 1L
    } else {
      if (run_length >= min_flatline_run) {
        flags[run_start:(i - 1L)] <- TRUE
      }
      run_start <- i
      run_length <- 1L
    }
  }

  if (!is.na(run_start) && run_length >= min_flatline_run) {
    flags[run_start:length(values)] <- TRUE
  }

  flags
}


.gpbiom_run_flags <- function(condition, min_run) {
  condition[is.na(condition)] <- FALSE

  if (length(condition) == 0L || !any(condition)) {
    return(rep(FALSE, length(condition)))
  }

  run <- rle(condition)
  flags <- rep(FALSE, length(condition))
  ends <- cumsum(run$lengths)
  starts <- ends - run$lengths + 1L

  flagged_runs <- which(run$values & run$lengths >= min_run)

  for (i in flagged_runs) {
    flags[starts[i]:ends[i]] <- TRUE
  }

  flags
}


.gpbiom_safe_colname <- function(x) {
  cleaned <- gsub("[^A-Za-z0-9]+", "_", as.character(x))
  cleaned <- gsub("_+", "_", cleaned)
  cleaned <- gsub("^_|_$", "", cleaned)

  if (!nzchar(cleaned)) {
    cleaned <- "signal"
  }

  make.names(cleaned)
}


.gpbiom_dropout_summary <- function(out, signal_cols, prefix) {
  rows <- lapply(signal_cols, function(signal) {
    safe_signal <- .gpbiom_safe_colname(signal)

    missing_col <- paste0(prefix, "_", safe_signal, "_missing")
    flatline_col <- paste0(prefix, "_", safe_signal, "_flatline")
    any_col <- paste0(prefix, "_", safe_signal, "_any")

    data.frame(
      column = signal,
      signal_type = .gpbiom_signal_type(signal),
      n = nrow(out),
      n_missing_dropout = sum(out[[missing_col]], na.rm = TRUE),
      missing_dropout_rate = if (nrow(out) > 0L) {
        mean(out[[missing_col]], na.rm = TRUE)
      } else {
        NA_real_
      },
      n_flatline_dropout = sum(out[[flatline_col]], na.rm = TRUE),
      flatline_dropout_rate = if (nrow(out) > 0L) {
        mean(out[[flatline_col]], na.rm = TRUE)
      } else {
        NA_real_
      },
      n_any_dropout = sum(out[[any_col]], na.rm = TRUE),
      any_dropout_rate = if (nrow(out) > 0L) {
        mean(out[[any_col]], na.rm = TRUE)
      } else {
        NA_real_
      },
      stringsAsFactors = FALSE
    )
  })

  do.call(rbind, rows)
}


.gpbiom_empty_dropout_summary <- function() {
  data.frame(
    column = character(),
    signal_type = character(),
    n = integer(),
    n_missing_dropout = integer(),
    missing_dropout_rate = numeric(),
    n_flatline_dropout = integer(),
    flatline_dropout_rate = numeric(),
    n_any_dropout = integer(),
    any_dropout_rate = numeric(),
    stringsAsFactors = FALSE
  )
}

test_that("summarise_gazepoint_biometric_validity detects active signals", {
  df <- data.frame(
    USER = rep(c("P1", "P2"), each = 4),
    GSR = c(1, 2, NA, 4, 2, 3, 4, 5),
    HR = c(70, 71, 72, NA, 80, 81, 82, 83),
    HRV = c(1, 1, 0, 1, 1, 1, 1, 1),
    ENGAGEMENT = c(50, 51, 52, 53, 50, 50, 50, 50)
  )

  out <- summarise_gazepoint_biometric_validity(df, group_cols = "USER")

  expect_type(out, "list")
  expect_s3_class(out$overview, "data.frame")
  expect_s3_class(out$signals, "data.frame")
  expect_s3_class(out$validity_flags, "data.frame")
  expect_s3_class(out$group_summary, "data.frame")

  expect_equal(out$overview$signal_column_count, 3)
  expect_equal(out$overview$validity_flag_column_count, 1)
  expect_true(out$overview$active_signal_count >= 2)
  expect_equal(out$overview$status, "biometric_signals_available")

  expect_true(all(c("GSR", "HR", "ENGAGEMENT") %in% out$signals$column))
  expect_true("HRV" %in% out$validity_flags$column)
})


test_that("summarise_gazepoint_biometric_validity treats HRV conservatively", {
  df <- data.frame(
    HR = c(70, 71, 72, 73),
    HRV = c(1, 1, 0, 1)
  )

  out <- summarise_gazepoint_biometric_validity(df)

  expect_equal(out$validity_flags$standard_name, "HRV")
  expect_match(out$validity_flags$interpretation_note, "validity/vendor flag")
  expect_true(any(grepl("Raw HRV columns", out$settings$notes)))
})


test_that("summarise_gazepoint_biometric_validity handles no detected signals", {
  df <- data.frame(
    USER = c("P1", "P2"),
    CONDITION = c("A", "B")
  )

  out <- summarise_gazepoint_biometric_validity(df)

  expect_equal(out$overview$signal_column_count, 0)
  expect_equal(out$overview$active_signal_count, 0)
  expect_equal(out$overview$status, "no_biometric_signal_columns_detected")
  expect_equal(nrow(out$signals), 0)
})


test_that("summarise_gazepoint_biometric_validity reports constant signals as limited", {
  df <- data.frame(
    GSR = c(1, 1, 1, 1),
    HR = c(70, 70, 70, 70)
  )

  out <- summarise_gazepoint_biometric_validity(df)

  expect_true(all(out$signals$status == "constant_or_low_variability_signal"))
  expect_equal(out$overview$status, "no_active_biometric_signals_detected")
})


test_that("summarise_gazepoint_biometric_validity can use explicit signal columns", {
  df <- data.frame(
    custom_signal = c(1, 2, 3, 4),
    validity_flag = c(1, 1, 0, 1)
  )

  out <- summarise_gazepoint_biometric_validity(
    df,
    signal_cols = "custom_signal",
    validity_cols = "validity_flag"
  )

  expect_equal(out$overview$signal_column_count, 1)
  expect_equal(out$overview$validity_flag_column_count, 1)
  expect_equal(out$signals$column, "custom_signal")
  expect_equal(out$validity_flags$column, "validity_flag")
})


test_that("summarise_gazepoint_biometric_validity validates arguments", {
  df <- data.frame(GSR = 1:3)

  expect_error(
    summarise_gazepoint_biometric_validity(1:3),
    "`data` must be"
  )

  expect_error(
    summarise_gazepoint_biometric_validity(df, signal_cols = "missing"),
    "not found"
  )

  expect_error(
    summarise_gazepoint_biometric_validity(df, active_min_unique = 0),
    "`active_min_unique`"
  )
})


test_that("flag_gazepoint_biometric_dropouts flags missing runs", {
  df <- data.frame(
    CNT = 1:8,
    GSR = c(1, NA, NA, NA, 2, 3, 4, 5),
    HR = c(70, 71, 72, 73, 74, 75, 76, 77)
  )

  out <- flag_gazepoint_biometric_dropouts(
    df,
    signal_cols = c("GSR", "HR"),
    min_missing_run = 3,
    min_flatline_run = 3
  )

  expect_true(all(out$biometric_dropout_GSR_missing[2:4]))
  expect_false(any(out$biometric_dropout_HR_missing))
  expect_true(all(out$biometric_dropout_any[2:4]))

  summary <- attr(out, "dropout_summary")
  expect_s3_class(summary, "data.frame")
  expect_equal(summary$n_missing_dropout[summary$column == "GSR"], 3)
})


test_that("flag_gazepoint_biometric_dropouts flags flatline runs", {
  df <- data.frame(
    CNT = 1:8,
    GSR = c(1, 2, 3, 4, 5, 6, 7, 8),
    HR = c(70, 70, 70, 71, 72, 72, 72, 72)
  )

  out <- flag_gazepoint_biometric_dropouts(
    df,
    signal_cols = c("GSR", "HR"),
    min_missing_run = 3,
    min_flatline_run = 3
  )

  expect_true(all(out$biometric_dropout_HR_flatline[1:3]))
  expect_true(all(out$biometric_dropout_HR_flatline[5:8]))
  expect_false(any(out$biometric_dropout_GSR_flatline))

  summary <- attr(out, "dropout_summary")
  expect_equal(summary$n_flatline_dropout[summary$column == "HR"], 7)
})


test_that("flag_gazepoint_biometric_dropouts respects grouping", {
  df <- data.frame(
    USER = rep(c("P1", "P2"), each = 3),
    CNT = c(1, 2, 3, 1, 2, 3),
    GSR = c(1, NA, NA, 2, NA, 3)
  )

  out <- flag_gazepoint_biometric_dropouts(
    df,
    signal_cols = "GSR",
    group_cols = "USER",
    min_missing_run = 2,
    min_flatline_run = 3
  )

  expect_true(all(out$biometric_dropout_GSR_missing[2:3]))
  expect_false(out$biometric_dropout_GSR_missing[5])
})


test_that("flag_gazepoint_biometric_dropouts can order within group by time", {
  df <- data.frame(
    USER = c("P1", "P1", "P1"),
    TIME = c(3, 1, 2),
    GSR = c(NA, 1, NA)
  )

  out <- flag_gazepoint_biometric_dropouts(
    df,
    signal_cols = "GSR",
    group_cols = "USER",
    time_col = "TIME",
    min_missing_run = 2,
    min_flatline_run = 3
  )

  expect_true(out$biometric_dropout_GSR_missing[1])
  expect_true(out$biometric_dropout_GSR_missing[3])
  expect_false(out$biometric_dropout_GSR_missing[2])
})


test_that("flag_gazepoint_biometric_dropouts handles no detected signals", {
  df <- data.frame(USER = c("P1", "P2"))

  out <- flag_gazepoint_biometric_dropouts(df)

  expect_true("biometric_dropout_any" %in% names(out))
  expect_false(any(out$biometric_dropout_any))
  expect_equal(nrow(attr(out, "dropout_summary")), 0)
})


test_that("flag_gazepoint_biometric_dropouts validates arguments", {
  df <- data.frame(GSR = 1:3)

  expect_error(
    flag_gazepoint_biometric_dropouts(1:3),
    "`data` must be"
  )

  expect_error(
    flag_gazepoint_biometric_dropouts(df, signal_cols = "missing"),
    "not found"
  )

  expect_error(
    flag_gazepoint_biometric_dropouts(df, min_missing_run = 0),
    "`min_missing_run`"
  )

  expect_error(
    flag_gazepoint_biometric_dropouts(df, min_flatline_run = 1.5),
    "`min_flatline_run`"
  )

  expect_error(
    flag_gazepoint_biometric_dropouts(df, constant_tolerance = -1),
    "`constant_tolerance`"
  )
})

#' Test HRV nonlinearity using surrogate data
#'
#' Tests whether a nonlinear HRV statistic differs from surrogate RR/IBI
#' sequences. This is a screening tool for evidence inconsistent with a simple
#' linear stochastic null process. It does not prove deterministic chaos or
#' diagnose any condition.
#'
#' @param dat A data frame containing IBI/RR intervals.
#' @param ibi_col Numeric IBI/RR interval column.
#' @param group_cols Optional grouping columns.
#' @param metric Nonlinear statistic to test.
#' @param n_surrogates Number of surrogate series per group.
#' @param surrogate_method `"phase_randomized"` or `"shuffle"`.
#' @param m Embedding dimension for entropy metrics.
#' @param r_multiplier Tolerance multiplier for entropy metrics.
#' @param statistic_fun Optional custom statistic function accepting numeric x.
#' @param seed Optional random seed.
#'
#' @return A list with `overview`, `results`, `surrogate_statistics`, and
#'   `settings`.
#' @export
test_gazepoint_hrv_nonlinearity <- function(dat,
                                            ibi_col = "IBI",
                                            group_cols = NULL,
                                            metric = c(
                                              "sample_entropy",
                                              "approximate_entropy",
                                              "sd1_sd2_ratio"
                                            ),
                                            n_surrogates = 99,
                                            surrogate_method = c(
                                              "phase_randomized",
                                              "shuffle"
                                            ),
                                            m = 2,
                                            r_multiplier = 0.2,
                                            statistic_fun = NULL,
                                            seed = NULL) {
  if (!is.data.frame(dat)) {
    stop("`dat` must be a data frame.", call. = FALSE)
  }

  if (!ibi_col %in% names(dat)) {
    stop("Column `", ibi_col, "` was not found in `dat`.", call. = FALSE)
  }

  if (!is.numeric(dat[[ibi_col]])) {
    stop("`ibi_col` must identify a numeric column.", call. = FALSE)
  }

  if (is.null(group_cols)) {
    group_cols <- character()
  }

  missing_groups <- setdiff(group_cols, names(dat))
  if (length(missing_groups) > 0) {
    stop("Missing `group_cols`: ", paste(missing_groups, collapse = ", "), call. = FALSE)
  }

  metric <- match.arg(metric)
  surrogate_method <- match.arg(surrogate_method)

  if (!is.numeric(n_surrogates) ||
      length(n_surrogates) != 1 ||
      !is.finite(n_surrogates) ||
      n_surrogates < 1) {
    stop("`n_surrogates` must be a positive integer.", call. = FALSE)
  }

  n_surrogates <- as.integer(n_surrogates)

  if (!is.null(seed)) {
    set.seed(seed)
  }

  groups <- gpbiometrics_nonlinearity_split(dat, group_cols)

  result_rows <- list()
  surrogate_rows <- list()
  surrogate_id <- 1L

  for (group_id in names(groups)) {
    idx <- groups[[group_id]]

    x <- dat[[ibi_col]][idx]
    x <- x[is.finite(x) & x > 0]

    observed <- gpbiometrics_nonlinearity_statistic(
      x = x,
      metric = metric,
      m = m,
      r_multiplier = r_multiplier,
      statistic_fun = statistic_fun
    )

    surrogate_values <- rep(NA_real_, n_surrogates)

    if (length(x) >= 10 && is.finite(observed)) {
      for (i in seq_len(n_surrogates)) {
        sx <- gpbiometrics_nonlinearity_surrogate(
          x,
          method = surrogate_method
        )

        surrogate_values[i] <- gpbiometrics_nonlinearity_statistic(
          x = sx,
          metric = metric,
          m = m,
          r_multiplier = r_multiplier,
          statistic_fun = statistic_fun
        )

        surrogate_rows[[surrogate_id]] <- data.frame(
          group_id = group_id,
          surrogate_index = i,
          surrogate_statistic = surrogate_values[i],
          stringsAsFactors = FALSE
        )

        surrogate_id <- surrogate_id + 1L
      }
    }

    finite_surrogates <- surrogate_values[is.finite(surrogate_values)]

    if (length(finite_surrogates) > 0 && is.finite(observed)) {
      p_greater <- (1 + sum(finite_surrogates >= observed)) /
        (length(finite_surrogates) + 1)

      p_less <- (1 + sum(finite_surrogates <= observed)) /
        (length(finite_surrogates) + 1)

      p_two_sided <- min(1, 2 * min(p_greater, p_less))

      z_score <- if (stats::sd(finite_surrogates) > 0) {
        (observed - mean(finite_surrogates)) / stats::sd(finite_surrogates)
      } else {
        NA_real_
      }

      status <- if (p_two_sided < 0.05) {
        "surrogate_difference_detected"
      } else {
        "surrogate_difference_not_detected"
      }
    } else {
      p_greater <- NA_real_
      p_less <- NA_real_
      p_two_sided <- NA_real_
      z_score <- NA_real_
      status <- "insufficient_information"
    }

    result_rows[[group_id]] <- data.frame(
      group_id = group_id,
      n_intervals = length(x),
      observed_statistic = observed,
      surrogate_mean = if (length(finite_surrogates) > 0) {
        mean(finite_surrogates)
      } else {
        NA_real_
      },
      surrogate_sd = if (length(finite_surrogates) > 1) {
        stats::sd(finite_surrogates)
      } else {
        NA_real_
      },
      p_greater = p_greater,
      p_less = p_less,
      p_two_sided = p_two_sided,
      z_score = z_score,
      status = status,
      stringsAsFactors = FALSE
    )
  }

  results <- do.call(rbind, result_rows)
  rownames(results) <- NULL

  surrogate_statistics <- if (length(surrogate_rows) > 0) {
    do.call(rbind, surrogate_rows)
  } else {
    data.frame()
  }

  overview <- data.frame(
    group_count = length(groups),
    result_rows = nrow(results),
    n_surrogates = n_surrogates,
    metric = metric,
    surrogate_method = surrogate_method,
    significant_groups = sum(results$status == "surrogate_difference_detected"),
    status = if (all(results$status != "insufficient_information")) {
      "surrogate_nonlinearity_test_complete"
    } else if (any(results$status != "insufficient_information")) {
      "surrogate_nonlinearity_test_partial"
    } else {
      "surrogate_nonlinearity_test_failed"
    },
    interpretation = paste(
      "A significant surrogate test indicates that the selected statistic differs from the surrogate null distribution.",
      "It does not prove deterministic chaos, clinical status, emotion, stress, or diagnosis."
    ),
    stringsAsFactors = FALSE
  )

  structure(
    list(
      overview = overview,
      results = results,
      surrogate_statistics = surrogate_statistics,
      settings = list(
        ibi_col = ibi_col,
        group_cols = group_cols,
        metric = metric,
        n_surrogates = n_surrogates,
        surrogate_method = surrogate_method,
        m = m,
        r_multiplier = r_multiplier,
        custom_statistic = !is.null(statistic_fun),
        seed = seed
      )
    ),
    class = c("gazepoint_hrv_nonlinearity_test", "list")
  )
}

gpbiometrics_nonlinearity_split <- function(dat, group_cols) {
  if (length(group_cols) == 0) {
    return(list(all_rows = seq_len(nrow(dat))))
  }

  gf <- dat[group_cols]
  gf[] <- lapply(gf, function(x) {
    x <- as.character(x)
    x[is.na(x)] <- "<NA>"
    x
  })

  split(seq_len(nrow(dat)), do.call(paste, c(gf, sep = " | ")))
}

gpbiometrics_nonlinearity_statistic <- function(x,
                                                metric,
                                                m = 2,
                                                r_multiplier = 0.2,
                                                statistic_fun = NULL) {
  x <- x[is.finite(x)]

  if (length(x) < 10 || stats::sd(x) == 0) {
    return(NA_real_)
  }

  if (!is.null(statistic_fun)) {
    return(as.numeric(statistic_fun(x))[1])
  }

  r <- r_multiplier * stats::sd(x)

  if (metric == "sample_entropy") {
    return(gpbiometrics_sample_entropy(x, m = m, r = r))
  }

  if (metric == "approximate_entropy") {
    return(gpbiometrics_approximate_entropy(x, m = m, r = r))
  }

  if (metric == "sd1_sd2_ratio") {
    dx <- diff(x)
    sdnn <- stats::sd(x)
    diff_var <- stats::var(dx, na.rm = TRUE)
    sd1 <- sqrt(diff_var / 2)
    sd2 <- sqrt(max((2 * sdnn^2) - (0.5 * diff_var), 0))

    if (is.finite(sd2) && sd2 > 0) {
      return(sd1 / sd2)
    }

    return(NA_real_)
  }

  NA_real_
}

gpbiometrics_nonlinearity_surrogate <- function(x,
                                                method = "phase_randomized") {
  x <- x[is.finite(x)]

  if (method == "shuffle") {
    return(sample(x, length(x), replace = FALSE))
  }

  gpbiometrics_phase_randomized_surrogate(x)
}

gpbiometrics_phase_randomized_surrogate <- function(x) {
  n <- length(x)

  if (n < 4) {
    return(sample(x, length(x), replace = FALSE))
  }

  x_centered <- x - mean(x)
  fft_x <- stats::fft(x_centered)

  random_phase <- stats::runif(n, min = 0, max = 2 * pi)

  phase_multiplier <- exp(1i * random_phase)
  phase_multiplier[1] <- 1

  if (n %% 2 == 0) {
    phase_multiplier[(n / 2) + 1] <- 1
    positive <- 2:(n / 2)
    negative <- n:(n / 2 + 2)
  } else {
    positive <- 2:((n + 1) / 2)
    negative <- n:((n + 3) / 2)
  }

  phase_multiplier[negative] <- Conj(phase_multiplier[positive])

  surrogate <- Re(stats::fft(fft_x * phase_multiplier, inverse = TRUE) / n)
  surrogate <- surrogate + mean(x)

  surrogate
}

#' Validate a Gazepoint Biometrics export
#'
#' Performs a conservative validation of a Gazepoint Biometrics table or CSV
#' file. The function checks whether known biometric columns are present,
#' whether biometric channels appear active, whether common time/synchronisation
#' columns are available, and whether obvious structural issues are present.
#'
#' @param data A data frame or a path to a Gazepoint CSV export.
#' @param require_active_signal Logical. If `TRUE`, validation reports a warning
#'   when no active GSR/EDA, heart-rate, or engagement-dial channel is detected.
#'
#' @return A list with `overview`, `columns`, `active_channels`, and `issues`.
#'   The returned object has class `"gazepoint_biometrics_validation"`.
#'
#' @export
validate_gazepoint_biometrics <- function(data, require_active_signal = FALSE) {
  dat <- coerce_gazepoint_biometrics_data(data)

  cols <- check_gazepoint_biometric_columns(dat)
  active <- detect_active_biometric_channels(dat)

  biometric_signals <- c("gsr_eda", "heart_rate", "engagement_dial")
  biometric_or_marker_signals <- c(biometric_signals, "ttl_marker")

  has_known_biometric_columns <- any(
    cols$present[cols$signal %in% biometric_or_marker_signals]
  )

  active_signal_count <- sum(
    active$active[active$signal %in% biometric_signals],
    na.rm = TRUE
  )

  time_columns <- c("CNT", "TIME", "TIME_TICK")
  present_time_columns <- intersect(time_columns, names(dat))

  empty_column_names <- names(dat) == "" | is.na(names(dat))

  issues <- empty_validation_issues()

  if (nrow(dat) == 0L) {
    issues <- add_validation_issue(
      issues,
      issue = "empty_data",
      severity = "error",
      details = "The data contain zero rows."
    )
  }

  if (any(empty_column_names)) {
    issues <- add_validation_issue(
      issues,
      issue = "empty_column_names",
      severity = "warning",
      details = paste0(sum(empty_column_names), " column name(s) are empty or missing.")
    )
  }

  if (!has_known_biometric_columns) {
    issues <- add_validation_issue(
      issues,
      issue = "no_known_biometric_columns",
      severity = "error",
      details = "No known Gazepoint Biometrics columns were detected."
    )
  }

  if (length(present_time_columns) == 0L) {
    issues <- add_validation_issue(
      issues,
      issue = "no_time_columns",
      severity = "warning",
      details = "No common Gazepoint time columns were detected: CNT, TIME, or TIME_TICK."
    )
  }

  if (isTRUE(require_active_signal) && active_signal_count == 0L) {
    issues <- add_validation_issue(
      issues,
      issue = "no_active_biometric_signal",
      severity = "warning",
      details = "Biometric columns are present, but no active GSR/EDA, heart-rate, or engagement-dial signal was detected."
    )
  }

  overview <- data.frame(
    n_rows = nrow(dat),
    n_columns = ncol(dat),
    known_biometric_columns = sum(cols$present[cols$signal %in% biometric_or_marker_signals]),
    active_signal_count = active_signal_count,
    present_time_columns = paste(present_time_columns, collapse = ","),
    issue_count = nrow(issues),
    stringsAsFactors = FALSE
  )

  out <- list(
    overview = overview,
    columns = cols,
    active_channels = active,
    issues = issues
  )

  class(out) <- c("gazepoint_biometrics_validation", "list")
  out
}


#' Audit missingness in Gazepoint biometric channels
#'
#' Summarises missingness and zero values for Gazepoint biometric columns. This
#' is useful because Gazepoint exports may contain biometric columns even when a
#' channel was inactive or invalid during recording.
#'
#' @param data A data frame or a path to a Gazepoint CSV export.
#' @param columns Optional character vector of columns to audit. If `NULL`,
#'   known present Gazepoint biometric, TTL, and validity columns are audited.
#'
#' @return A data frame with one row per audited column.
#'
#' @export
audit_gazepoint_biometric_missingness <- function(data, columns = NULL) {
  dat <- coerce_gazepoint_biometrics_data(data)

  cols <- check_gazepoint_biometric_columns(dat)

  if (is.null(columns)) {
    columns <- cols$column[
      cols$present &
        cols$signal %in% c("gsr_eda", "heart_rate", "engagement_dial", "ttl_marker")
    ]
  }

  if (!is.character(columns)) {
    stop("`columns` must be a character vector or NULL.", call. = FALSE)
  }

  columns <- intersect(columns, names(dat))

  if (length(columns) == 0L) {
    return(data.frame(
      column = character(0),
      signal = character(0),
      role = character(0),
      n_rows = integer(0),
      missing_rows = integer(0),
      missing_pct = numeric(0),
      zero_rows = integer(0),
      zero_pct = numeric(0),
      min_value = numeric(0),
      max_value = numeric(0),
      stringsAsFactors = FALSE
    ))
  }

  out <- lapply(columns, function(column) {
    x <- dat[[column]]
    x_chr <- trimws(as.character(x))
    x_num <- as_numeric_safe(x)

    missing <- is.na(x) | x_chr == ""
    zero <- !is.na(x_num) & x_num == 0

    non_missing_numeric <- x_num[!is.na(x_num)]

    signal <- cols$signal[match(column, cols$column)]
    role <- cols$role[match(column, cols$column)]

    data.frame(
      column = column,
      signal = ifelse(is.na(signal), "unknown", signal),
      role = ifelse(is.na(role), "unknown", role),
      n_rows = nrow(dat),
      missing_rows = sum(missing),
      missing_pct = safe_pct(sum(missing), nrow(dat)),
      zero_rows = sum(zero),
      zero_pct = safe_pct(sum(zero), nrow(dat)),
      min_value = ifelse(
        length(non_missing_numeric) > 0L,
        min(non_missing_numeric, na.rm = TRUE),
        NA_real_
      ),
      max_value = ifelse(
        length(non_missing_numeric) > 0L,
        max(non_missing_numeric, na.rm = TRUE),
        NA_real_
      ),
      stringsAsFactors = FALSE
    )
  })

  out <- do.call(rbind, out)
  rownames(out) <- NULL
  out
}


coerce_gazepoint_biometrics_data <- function(data) {
  if (is.character(data) && length(data) == 1L && nzchar(data)) {
    return(import_gazepoint_biometrics(data))
  }

  if (is.data.frame(data)) {
    return(data)
  }

  stop(
    "`data` must be a data frame or a single path to a Gazepoint CSV export.",
    call. = FALSE
  )
}


empty_validation_issues <- function() {
  data.frame(
    issue = character(0),
    severity = character(0),
    details = character(0),
    stringsAsFactors = FALSE
  )
}


add_validation_issue <- function(issues, issue, severity, details) {
  rbind(
    issues,
    data.frame(
      issue = issue,
      severity = severity,
      details = details,
      stringsAsFactors = FALSE
    )
  )
}


safe_pct <- function(numerator, denominator) {
  if (is.na(denominator) || denominator == 0L) {
    return(NA_real_)
  }

  100 * numerator / denominator
}

test_that("validate_gazepoint_biometrics validates active biometric data", {
  dat <- data.frame(
    TIME = c(0.01, 0.02, 0.03),
    GSR_US = c(2.0, 2.1, 2.2),
    GSRV = c(1, 1, 1),
    HR = c(75, 76, 77),
    HRV = c(1, 1, 1),
    DIAL = c(0.1, 0.2, 0.3),
    DIALV = c(1, 1, 1)
  )

  val <- validate_gazepoint_biometrics(dat, require_active_signal = TRUE)

  expect_s3_class(val, "gazepoint_biometrics_validation")
  expect_equal(val$overview$n_rows, 3)
  expect_equal(val$overview$active_signal_count, 3)
  expect_equal(nrow(val$issues), 0)
})


test_that("validate_gazepoint_biometrics reports inactive biometric data", {
  dat <- data.frame(
    TIME = c(0.01, 0.02, 0.03),
    GSR_US = c(0, 0, 0),
    GSRV = c(0, 0, 0),
    HR = c(0, 0, 0),
    HRV = c(0, 0, 0),
    DIAL = c(0, 0, 0),
    DIALV = c(0, 0, 0)
  )

  val <- validate_gazepoint_biometrics(dat, require_active_signal = TRUE)

  expect_true("no_active_biometric_signal" %in% val$issues$issue)
  expect_equal(val$overview$active_signal_count, 0)
})


test_that("validate_gazepoint_biometrics reports missing biometric columns", {
  dat <- data.frame(
    TIME = c(0.01, 0.02),
    X = c(1, 2),
    Y = c(3, 4)
  )

  val <- validate_gazepoint_biometrics(dat)

  expect_true("no_known_biometric_columns" %in% val$issues$issue)
})


test_that("audit_gazepoint_biometric_missingness summarises columns", {
  dat <- data.frame(
    GSR_US = c(2.0, NA, 0),
    GSRV = c(1, 0, 0),
    HR = c(75, NA, 0),
    HRV = c(1, 0, 0),
    DIAL = c(0.1, NA, 0),
    DIALV = c(1, 0, 0)
  )

  audit <- audit_gazepoint_biometric_missingness(dat)

  expect_true(is.data.frame(audit))
  expect_true("GSR_US" %in% audit$column)
  expect_true("HR" %in% audit$column)
  expect_true("DIAL" %in% audit$column)

  gsr <- audit[audit$column == "GSR_US", ]

  expect_equal(gsr$n_rows, 3)
  expect_equal(gsr$missing_rows, 1)
  expect_equal(gsr$zero_rows, 1)
})


test_that("audit_gazepoint_biometric_missingness returns empty table when no known columns exist", {
  dat <- data.frame(
    X = c(1, 2),
    Y = c(3, 4)
  )

  audit <- audit_gazepoint_biometric_missingness(dat)

  expect_true(is.data.frame(audit))
  expect_equal(nrow(audit), 0)
})

#' Create a Gazepoint Biometrics reporting checklist
#'
#' Creates a compact reporting checklist for Gazepoint Biometrics exports. The
#' checklist summarises detected biometric channels, validation issues, signal
#' quality, missingness, available workflow domains, and interpretation
#' cautions. It is intended to support transparent manuscript reporting and
#' reviewer-facing methods documentation.
#'
#' @param data A data frame or a path to a Gazepoint CSV export.
#' @param require_active_signal Logical. Should inactive biometric channels be
#'   flagged in the validation output?
#'
#' @return A list with `overview`, `channels`, `quality`, `missingness`,
#'   `validation_issues`, `workflow_capabilities`, `feature_inventory`,
#'   `reporting_guidance`, and `interpretation_cautions`.
#'
#' @export
create_gazepoint_biometrics_checklist <- function(data,
                                                  require_active_signal = TRUE) {
  dat <- coerce_gazepoint_biometrics_data(data)

  validation <- validate_gazepoint_biometrics(
    dat,
    require_active_signal = require_active_signal
  )

  channels <- validation$active_channels

  quality <- combine_gazepoint_tables(list(
    audit_gazepoint_gsr_quality(dat),
    audit_gazepoint_hr_quality(dat),
    audit_gazepoint_engagement_dial(dat)
  ))

  missingness <- audit_gazepoint_biometric_missingness(dat)

  feature_inventory <- create_gazepoint_biometrics_feature_inventory()

  workflow_capabilities <- gpbiometrics_checklist_capabilities(
    feature_inventory
  )

  reporting_guidance <- gpbiometrics_checklist_reporting_guidance()

  overview <- data.frame(
    n_rows = nrow(dat),
    n_columns = ncol(dat),
    active_gsr_eda = is_signal_active(channels, "gsr_eda"),
    active_heart_rate = is_signal_active(channels, "heart_rate"),
    active_engagement_dial = is_signal_active(channels, "engagement_dial"),
    active_ttl_marker = is_signal_active(channels, "ttl_marker"),
    validation_issue_count = nrow(validation$issues),
    workflow_domain_count = feature_inventory$overview$domain_count,
    available_workflow_domains = sum(
      workflow_capabilities$status == "complete",
      na.rm = TRUE
    ),
    feature_inventory_status = feature_inventory$overview$status,
    stringsAsFactors = FALSE
  )

  cautions <- data.frame(
    topic = c(
      "GSR/EDA",
      "SCR events",
      "Heart rate",
      "Heart-rate variability",
      "IBI/RR intervals",
      "Engagement dial",
      "AOI-linked biometrics",
      "Pupil diameter",
      "Eye tracking",
      "Interoperability",
      "Plotting and visual inspection"
    ),
    caution = c(
      "GSR/EDA should be interpreted as physiological arousal or electrodermal activity, not emotional valence by itself.",
      "SCR detections are signal-processing events and should be reported with threshold, latency-window, and sensitivity settings.",
      "Heart-rate changes require baseline, task-context, and signal-quality interpretation.",
      "Heart-rate variability should be derived only from appropriate interbeat-interval or pulse information, not from a validity flag.",
      "IBI/RR intervals should be filtered for implausible values, repeated export rows, and short-duration windows before HRV interpretation.",
      "Engagement-dial values are self-reported or user-controlled continuous responses and should not be treated as automatic physiology.",
      "AOI-linked biometric summaries describe signal values during AOI exposure and do not establish emotional valence, preference, or cognitive evaluation by themselves.",
      "Pupil diameter may reflect luminance, arousal, cognitive load, fatigue, or other processes.",
      "Eye-tracking measures indicate visual attention and timing, not direct cognitive evaluation by themselves.",
      "External interoperability helpers prepare or cross-check data; external package outputs should be reported with software versions and settings.",
      "Plots are intended for quality control, synchronization checks, and transparent reporting, not as standalone inferential evidence."
    ),
    stringsAsFactors = FALSE
  )

  out <- list(
    overview = overview,
    channels = channels,
    quality = quality,
    missingness = missingness,
    validation_issues = validation$issues,
    workflow_capabilities = workflow_capabilities,
    feature_inventory = feature_inventory,
    reporting_guidance = reporting_guidance,
    interpretation_cautions = cautions
  )

  class(out) <- c("gazepoint_biometrics_checklist", "list")
  out
}


#' Create Gazepoint Biometrics methods text
#'
#' Creates a compact draft methods paragraph describing Gazepoint Biometrics
#' data processing. The text is intentionally cautious and avoids making
#' emotional or cognitive claims from physiological or eye-tracking measures
#' alone.
#'
#' @param checklist A checklist produced by
#'   `create_gazepoint_biometrics_checklist()`. If `NULL`, `data` must be
#'   supplied.
#' @param data Optional data frame or path to a Gazepoint CSV export used to
#'   create the checklist when `checklist = NULL`.
#' @param include_cautions Logical. Should interpretation cautions be appended?
#'
#' @return A character string containing draft methods text.
#'
#' @export
create_gazepoint_biometrics_methods_text <- function(checklist = NULL,
                                                     data = NULL,
                                                     include_cautions = TRUE) {
  if (is.null(checklist)) {
    if (is.null(data)) {
      stop("Either `checklist` or `data` must be supplied.", call. = FALSE)
    }

    checklist <- create_gazepoint_biometrics_checklist(data)
  }

  if (!inherits(checklist, "gazepoint_biometrics_checklist")) {
    stop(
      "`checklist` must be produced by create_gazepoint_biometrics_checklist().",
      call. = FALSE
    )
  }

  overview <- checklist$overview

  active_signals <- c(
    if (isTRUE(overview$active_gsr_eda)) "GSR/EDA",
    if (isTRUE(overview$active_heart_rate)) "heart rate",
    if (isTRUE(overview$active_engagement_dial)) "engagement dial",
    if (isTRUE(overview$active_ttl_marker)) "TTL markers"
  )

  if (length(active_signals) == 0L) {
    signal_text <- "no active biometric channels"
  } else {
    signal_text <- paste(active_signals, collapse = ", ")
  }

  capability_text <- gpbiometrics_methods_capability_sentence(checklist)

  text <- paste0(
    "Gazepoint Biometrics exports were imported and screened using gpbiometrics. ",
    "The processed table contained ",
    overview$n_rows,
    " rows and ",
    overview$n_columns,
    " columns. ",
    "Detected active channels included ",
    signal_text,
    ". ",
    "Biometric channels were checked for missing values, zero/inactive rows, ",
    "validity flags, implausible values, sudden jumps, flatlining, and usable ",
    "sample coverage. GSR/EDA summaries prioritised the Gazepoint conductance ",
    "column when available, heart-rate summaries treated raw HRV as a vendor or ",
    "validity field rather than a heart-rate-variability metric, and ",
    "engagement-dial summaries were treated as continuous user-controlled ",
    "response values. ",
    capability_text
  )

  if (isTRUE(include_cautions)) {
    text <- paste0(
      text,
      " Physiological and eye-tracking measures were interpreted conservatively: ",
      "GSR/EDA and SCR features were treated as electrodermal or ",
      "arousal-related signal information rather than emotional valence, ",
      "heart-rate and IBI-derived HRV summaries were interpreted relative to ",
      "baseline, signal quality, and window duration, AOI-linked biometric ",
      "summaries were treated as descriptive signal-by-region summaries, and ",
      "eye-tracking measures were treated as indicators of visual attention rather ",
      "than direct evidence of cognitive evaluation."
    )
  }

  text
}


gpbiometrics_checklist_capabilities <- function(feature_inventory) {
  if (!inherits(feature_inventory, "gazepoint_biometrics_feature_inventory")) {
    return(data.frame())
  }

  out <- feature_inventory$domain_summary

  if (!is.data.frame(out) || nrow(out) == 0) {
    return(data.frame())
  }

  out$reporting_relevance <- vapply(
    out$domain,
    gpbiometrics_checklist_domain_relevance,
    character(1)
  )

  out
}


gpbiometrics_checklist_domain_relevance <- function(domain) {
  switch(
    domain,
    import_and_schema = "Import, schema detection, and active-channel reporting.",
    quality_and_readiness = "Missingness, sampling, activity, time-ordering, and readiness checks.",
    preprocessing = "Baseline correction, smoothing, EDA decomposition, and conservative signal conversion.",
    eda_scr = "EDA quality, SCR detection, event-window summaries, nonresponder screening, and SCR sensitivity checks.",
    ibi_hr_hrv = "HR/IBI quality, implausible-IBI filtering, HR-IBI consistency checks, and IBI-derived HRV features.",
    ttl_alignment = "TTL extraction, event alignment, and multimodal timeline inspection.",
    aoi_biometrics = "AOI-linked biometric summaries, model-preparation tables, and AOI-biometric plotting.",
    modelling_and_windows = "Window summaries, multimodal model tables, and LME-ready data preparation.",
    reporting = "Workflow summaries, report tables, methods text, readiness gates, and report bundles.",
    interoperability = "Optional RHRV and NeuroKit2-compatible export or cross-check workflows.",
    plotting = "Contract-standardised ggplot outputs with stored plot data, settings, and interpretation notes.",
    "General package capability."
  )
}


gpbiometrics_checklist_reporting_guidance <- function() {
  data.frame(
    section = c(
      "Data source",
      "Signal quality",
      "EDA/SCR",
      "IBI/HR/HRV",
      "TTL alignment",
      "AOI-linked biometrics",
      "Modelling",
      "Interoperability",
      "Plots",
      "Interpretation"
    ),
    guidance = c(
      "Report Gazepoint export type, imported files, detected channels, row counts, and software/package versions.",
      "Report missingness, inactive channels, validity flags, sampling checks, time ordering, and exclusion criteria.",
      "Report GSR/GSR_US choice, preprocessing, SCR thresholds, response windows, and threshold-sensitivity checks when used.",
      "Report IBI source, plausibility filtering, HR-IBI consistency checks, HRV window duration, and whether HRV was derived from genuine IBI/RR intervals.",
      "Report TTL/event columns, alignment rules, event windows, and whether CNT or timestamp resets were handled within groups.",
      "Report AOI definitions, AOI exposure thresholds, signal summaries, and whether AOI-linked outputs were descriptive or model-based.",
      "Report model family, grouping structure, outcome transformations, standardisation choices, and random-effects structure where applicable.",
      "Report external package names, versions, settings, and whether interoperability helpers only prepared input or executed cross-checks.",
      "Use stored plot data/settings for reproducibility and describe plots as QC/reporting aids rather than inferential evidence.",
      "Avoid inferring emotional valence, preference, trust, scrutiny, or cognitive evaluation from biometric or eye-tracking signals alone."
    ),
    stringsAsFactors = FALSE
  )
}

gpbiometrics_methods_capability_sentence <- function(checklist) {
  if (!is.list(checklist) ||
      !"workflow_capabilities" %in% names(checklist) ||
      !is.data.frame(checklist$workflow_capabilities)) {
    return("")
  }

  capabilities <- checklist$workflow_capabilities

  complete_domains <- capabilities$domain[
    capabilities$status == "complete"
  ]

  if (length(complete_domains) == 0) {
    return("")
  }

  domain_labels <- vapply(
    complete_domains,
    gpbiometrics_methods_domain_label,
    character(1)
  )

  domain_text <- paste(domain_labels, collapse = ", ")

  paste0(
    "The available workflow inventory covered ",
    length(complete_domains),
    " complete processing domains: ",
    domain_text,
    ". These include quality/readiness checks, EDA/SCR event workflows, ",
    "IBI/HR/HRV processing, TTL alignment, AOI-linked biometric summaries, ",
    "analysis-ready model tables, interoperability exports, and ",
    "contract-standardised plots with stored plot data and settings."
  )
}


gpbiometrics_methods_domain_label <- function(domain) {
  switch(
    domain,
    import_and_schema = "import and schema detection",
    quality_and_readiness = "quality and readiness",
    preprocessing = "preprocessing",
    eda_scr = "EDA/SCR",
    ibi_hr_hrv = "IBI/HR/HRV",
    ttl_alignment = "TTL alignment",
    aoi_biometrics = "AOI-linked biometrics",
    modelling_and_windows = "modelling and window summaries",
    reporting = "reporting",
    interoperability = "interoperability",
    plotting = "plotting",
    gsub("_", " ", domain)
  )
}

is_signal_active <- function(channels, signal) {
  if (!is.data.frame(channels) || !"signal" %in% names(channels) ||
      !"active" %in% names(channels)) {
    return(FALSE)
  }

  value <- channels$active[channels$signal == signal]

  if (length(value) == 0L || is.na(value[1])) {
    return(FALSE)
  }

  isTRUE(value[1])
}

test_that("biometrics checklist includes workflow capabilities and guidance", {
  dat <- data.frame(
    GSR_US = c(0, 1.1, 1.2, 1.3),
    HR = c(70, 71, 72, 73),
    IBI = c(0.80, 0.82, 0.81, 0.83),
    DIAL = c(0, 1, 1, 1),
    TTL0 = c(1, 0, 0, 0)
  )

  checklist <- create_gazepoint_biometrics_checklist(dat)

  expect_s3_class(checklist, "gazepoint_biometrics_checklist")
  expect_true("workflow_capabilities" %in% names(checklist))
  expect_true("feature_inventory" %in% names(checklist))
  expect_true("reporting_guidance" %in% names(checklist))

  expect_true(is.data.frame(checklist$workflow_capabilities))
  expect_true(is.data.frame(checklist$reporting_guidance))
  expect_s3_class(
    checklist$feature_inventory,
    "gazepoint_biometrics_feature_inventory"
  )

  expect_true("eda_scr" %in% checklist$workflow_capabilities$domain)
  expect_true("ibi_hr_hrv" %in% checklist$workflow_capabilities$domain)
  expect_true("aoi_biometrics" %in% checklist$workflow_capabilities$domain)
  expect_true("interoperability" %in% checklist$workflow_capabilities$domain)
  expect_true("plotting" %in% checklist$workflow_capabilities$domain)
})

test_that("methods text mentions expanded workflow capabilities cautiously", {
  dat <- data.frame(
    GSR_US = c(0, 1.1, 1.2, 1.3),
    HR = c(70, 71, 72, 73),
    IBI = c(0.80, 0.82, 0.81, 0.83),
    DIAL = c(0, 1, 1, 1),
    TTL0 = c(1, 0, 0, 0)
  )

  checklist <- create_gazepoint_biometrics_checklist(dat)

  text <- create_gazepoint_biometrics_methods_text(
    checklist = checklist,
    include_cautions = TRUE
  )

  expect_true(is.character(text))
  expect_length(text, 1)
  expect_match(text, "EDA/SCR", fixed = TRUE)
  expect_match(text, "IBI/HR/HRV", fixed = TRUE)
  expect_match(text, "AOI-linked biometric", fixed = TRUE)
  expect_match(text, "contract-standardised plots", fixed = TRUE)
  expect_match(text, "emotional valence", fixed = TRUE)
})

test_that("reporting guidance covers key manuscript sections", {
  dat <- data.frame(
    GSR_US = c(0, 1.1, 1.2, 1.3),
    HR = c(70, 71, 72, 73),
    IBI = c(0.80, 0.82, 0.81, 0.83),
    DIAL = c(0, 1, 1, 1),
    TTL0 = c(1, 0, 0, 0)
  )

  checklist <- create_gazepoint_biometrics_checklist(dat)

  expect_true("EDA/SCR" %in% checklist$reporting_guidance$section)
  expect_true("IBI/HR/HRV" %in% checklist$reporting_guidance$section)
  expect_true("AOI-linked biometrics" %in% checklist$reporting_guidance$section)
  expect_true("Interoperability" %in% checklist$reporting_guidance$section)
  expect_true("Plots" %in% checklist$reporting_guidance$section)
})

test_that("report bundle README includes expanded workflow cautions", {
  skip_if_not_installed("ggplot2")

  output_dir <- tempfile("gpbiometrics_bundle_")
  dir.create(output_dir)

  bundle <- export_gazepoint_biometrics_report_bundle(
    output_dir = output_dir,
    prefix = "test_bundle",
    tables = list(example = data.frame(x = 1:2)),
    include_readme = TRUE,
    include_session_info = FALSE,
    overwrite = TRUE
  )

  expect_s3_class(bundle, "gazepoint_biometrics_report_bundle")
  expect_equal(bundle$overview$status, "bundle_exported")

  readme_path <- bundle$manifest$path[
    bundle$manifest$item == "README"
  ]

  expect_true(file.exists(readme_path))

  readme_text <- paste(readLines(readme_path, warn = FALSE), collapse = "\n")

  expect_match(readme_text, "SCR peak/event-window", fixed = TRUE)
  expect_match(readme_text, "IBI-derived HRV", fixed = TRUE)
  expect_match(readme_text, "AOI-linked biometric", fixed = TRUE)
  expect_match(readme_text, "Contract-standardised ggplot", fixed = TRUE)
})

test_that("create_gazepoint_biometrics_checklist returns structured output", {
  dat <- data.frame(
    TIME = c(0.01, 0.02, 0.03),
    GSR_US = c(2.0, 2.1, 2.2),
    GSRV = c(1, 1, 1),
    HR = c(75, 76, 77),
    HRV = c(1, 1, 1),
    DIAL = c(0.1, 0.2, 0.3),
    DIALV = c(1, 1, 1),
    TTL0 = c(1, 1, 1),
    TTLV = c(1, 1, 1)
  )

  checklist <- create_gazepoint_biometrics_checklist(dat)

  expect_s3_class(checklist, "gazepoint_biometrics_checklist")
  expect_true(is.data.frame(checklist$overview))
  expect_true(is.data.frame(checklist$channels))
  expect_true(is.data.frame(checklist$quality))
  expect_true(is.data.frame(checklist$missingness))
  expect_true(is.data.frame(checklist$validation_issues))
  expect_true(is.data.frame(checklist$interpretation_cautions))

  expect_true(checklist$overview$active_gsr_eda)
  expect_true(checklist$overview$active_heart_rate)
  expect_true(checklist$overview$active_engagement_dial)
})


test_that("create_gazepoint_biometrics_methods_text creates cautious methods text", {
  dat <- data.frame(
    GSR_US = c(2.0, 2.1, 2.2),
    GSRV = c(1, 1, 1),
    HR = c(75, 76, 77),
    HRV = c(1, 1, 1),
    DIAL = c(0.1, 0.2, 0.3),
    DIALV = c(1, 1, 1)
  )

  checklist <- create_gazepoint_biometrics_checklist(dat)
  text <- create_gazepoint_biometrics_methods_text(checklist)

  expect_type(text, "character")
  expect_length(text, 1)
  expect_true(grepl("Gazepoint Biometrics", text))
  expect_true(grepl("GSR/EDA", text))
  expect_true(grepl("heart rate", text))
  expect_true(grepl("engagement dial", text))
  expect_true(grepl("conservatively", text))
})


test_that("create_gazepoint_biometrics_methods_text accepts data directly", {
  dat <- data.frame(
    GSR_US = c(2.0, 2.1, 2.2),
    GSRV = c(1, 1, 1)
  )

  text <- create_gazepoint_biometrics_methods_text(data = dat)

  expect_type(text, "character")
  expect_true(grepl("processed table contained", text))
})


test_that("create_gazepoint_biometrics_methods_text requires checklist or data", {
  expect_error(
    create_gazepoint_biometrics_methods_text(),
    "Either `checklist` or `data` must be supplied"
  )
})


test_that("create_gazepoint_biometrics_methods_text rejects invalid checklist", {
  expect_error(
    create_gazepoint_biometrics_methods_text(checklist = list()),
    "must be produced"
  )
})

#' Align Gazepoint biometric samples to TTL events
#'
#' Aligns biometric rows to TTL/event markers and returns event-relative time and
#' sample indices. The helper is conservative: TTL events are detected from
#' rising edges by default, validity flags are used when available, and no
#' physiological interpretation is added.
#'
#' @param data A data frame containing biometric samples.
#' @param ttl_cols Optional TTL marker columns. If `NULL`, the function first
#'   looks for `ttl_marker`, then raw `TTL0`-`TTL6` columns.
#' @param event_col Optional single user-specified event column. If supplied, it
#'   is used instead of automatic TTL-column detection.
#' @param ttl_valid_col Optional TTL validity column. If `NULL`, the function
#'   looks for `ttl_validity_flag` or `TTLV`.
#' @param time_col Optional time column. If `NULL`, common time-column names are
#'   detected automatically when present.
#' @param sample_col Optional sample/counter column. If `NULL`, `CNT`/`cnt` is
#'   used when present; otherwise row order is used.
#' @param group_cols Optional grouping columns. If `NULL`, the function uses
#'   available participant/stimulus/trial-like columns when present.
#' @param participant_col,stimulus_col,trial_col Optional explicit grouping
#'   columns to add to `group_cols`.
#' @param event_value Optional value(s) that define an active event. If `NULL`,
#'   non-zero numeric/logical values and non-empty character values are treated
#'   as active.
#' @param valid_values Values treated as valid in the TTL validity column.
#' @param event_edge Event-detection rule. `"rising"` keeps inactive-to-active
#'   transitions, `"change"` keeps changes among active event values, and
#'   `"active"` keeps every active sample.
#' @param pre_window_ms,post_window_ms Event window in milliseconds when a usable
#'   time column is available.
#' @param pre_window_samples,post_window_samples Event window in samples when no
#'   usable time column is available. If omitted, only the event sample is kept.
#' @param collapse_nearby_ms Optional minimum distance between retained events
#'   within a group, in milliseconds.
#' @param require_valid_ttl If `TRUE`, a detected TTL validity column must be
#'   active for a row to count as an event.
#'
#' @return A list with `overview`, `events`, `aligned_data`, and `settings`.
#' @export
align_gazepoint_biometrics_to_ttl <- function(data,
                                              ttl_cols = NULL,
                                              event_col = NULL,
                                              ttl_valid_col = NULL,
                                              time_col = NULL,
                                              sample_col = NULL,
                                              group_cols = NULL,
                                              participant_col = NULL,
                                              stimulus_col = NULL,
                                              trial_col = NULL,
                                              event_value = NULL,
                                              valid_values = c(TRUE, 1, "1"),
                                              event_edge = c("rising", "change", "active"),
                                              pre_window_ms = 1000,
                                              post_window_ms = 5000,
                                              pre_window_samples = NULL,
                                              post_window_samples = NULL,
                                              collapse_nearby_ms = 0,
                                              require_valid_ttl = TRUE) {
  event_edge <- match.arg(event_edge)

  if (!is.data.frame(data)) {
    stop("`data` must be a data frame.", call. = FALSE)
  }

  if (!is.numeric(pre_window_ms) || length(pre_window_ms) != 1 || pre_window_ms < 0) {
    stop("`pre_window_ms` must be a single non-negative number.", call. = FALSE)
  }

  if (!is.numeric(post_window_ms) || length(post_window_ms) != 1 || post_window_ms < 0) {
    stop("`post_window_ms` must be a single non-negative number.", call. = FALSE)
  }

  if (!is.numeric(collapse_nearby_ms) ||
      length(collapse_nearby_ms) != 1 ||
      collapse_nearby_ms < 0) {
    stop("`collapse_nearby_ms` must be a single non-negative number.", call. = FALSE)
  }

  dat <- as.data.frame(data, stringsAsFactors = FALSE)
  dat$.gpbiometrics_row_id <- seq_len(nrow(dat))

  if (nrow(dat) == 0) {
    events <- ttl_alignment_empty_events(group_cols = character())
    aligned <- ttl_alignment_empty_aligned(dat)

    return(structure(
      list(
        overview = data.frame(
          input_rows = 0L,
          ttl_event_rows = 0L,
          aligned_rows = 0L,
          group_count = 0L,
          status = "empty_input",
          stringsAsFactors = FALSE
        ),
        events = events,
        aligned_data = aligned,
        settings = list(
          ttl_cols = ttl_cols,
          event_col = event_col,
          ttl_valid_col = ttl_valid_col,
          time_col = time_col,
          sample_col = sample_col,
          group_cols = group_cols,
          event_edge = event_edge,
          pre_window_ms = pre_window_ms,
          post_window_ms = post_window_ms,
          pre_window_samples = pre_window_samples,
          post_window_samples = post_window_samples,
          collapse_nearby_ms = collapse_nearby_ms,
          require_valid_ttl = require_valid_ttl
        )
      ),
      class = c("gazepoint_biometrics_ttl_alignment", "list")
    ))
  }

  names_dat <- names(dat)

  if (!is.null(event_col)) {
    if (!event_col %in% names_dat) {
      stop("`event_col` was not found in `data`.", call. = FALSE)
    }
    ttl_cols <- event_col
    event_source <- "user_event_col"
  } else {
    if (is.null(ttl_cols)) {
      ttl_cols <- ttl_alignment_infer_ttl_cols(names_dat)
    }

    if (length(ttl_cols) == 0) {
      stop(
        "No TTL/event columns were found. Supply `ttl_cols` or `event_col`.",
        call. = FALSE
      )
    }

    missing_ttl_cols <- setdiff(ttl_cols, names_dat)
    if (length(missing_ttl_cols) > 0) {
      stop(
        "`ttl_cols` not found in `data`: ",
        paste(missing_ttl_cols, collapse = ", "),
        call. = FALSE
      )
    }

    event_source <- "ttl_cols"
  }

  if (is.null(ttl_valid_col)) {
    ttl_valid_col <- ttl_alignment_first_existing(
      names_dat,
      c("ttl_validity_flag", "TTLV", "ttlv")
    )
  }

  if (!is.null(ttl_valid_col) && !ttl_valid_col %in% names_dat) {
    stop("`ttl_valid_col` was not found in `data`.", call. = FALSE)
  }

  if (is.null(time_col)) {
    time_col <- ttl_alignment_first_existing(
      names_dat,
      c(
        "time_ms", "timestamp_ms", "timestamp",
        "TIME", "Time", "time",
        "recording_time", "sample_time"
      )
    )
  }

  if (!is.null(time_col) && !time_col %in% names_dat) {
    stop("`time_col` was not found in `data`.", call. = FALSE)
  }

  if (is.null(sample_col)) {
    sample_col <- ttl_alignment_first_existing(names_dat, c("CNT", "cnt", "sample", "sample_index"))
  }

  if (!is.null(sample_col) && !sample_col %in% names_dat) {
    stop("`sample_col` was not found in `data`.", call. = FALSE)
  }

  group_cols <- ttl_alignment_resolve_group_cols(
    names_dat = names_dat,
    group_cols = group_cols,
    participant_col = participant_col,
    stimulus_col = stimulus_col,
    trial_col = trial_col
  )

  missing_group_cols <- setdiff(group_cols, names_dat)
  if (length(missing_group_cols) > 0) {
    stop(
      "`group_cols` not found in `data`: ",
      paste(missing_group_cols, collapse = ", "),
      call. = FALSE
    )
  }

  time_ms <- ttl_alignment_time_ms(dat, time_col)

  dat$.gpbiometrics_time_ms <- time_ms
  dat$.gpbiometrics_group_id <- ttl_alignment_group_id(dat, group_cols)

  order_value <- dat$.gpbiometrics_row_id
  if (!all(is.na(dat$.gpbiometrics_time_ms))) {
    order_value <- dat$.gpbiometrics_time_ms
  } else if (!is.null(sample_col)) {
    sample_numeric <- suppressWarnings(as.numeric(dat[[sample_col]]))
    if (!all(is.na(sample_numeric))) {
      order_value <- sample_numeric
    }
  }

  dat$.gpbiometrics_order_value <- order_value
  dat$.gpbiometrics_group_sample_index <- NA_integer_

  group_indices <- split(seq_len(nrow(dat)), dat$.gpbiometrics_group_id, drop = TRUE)

  for (idx in group_indices) {
    idx_ordered <- idx[order(dat$.gpbiometrics_order_value[idx], dat$.gpbiometrics_row_id[idx])]
    dat$.gpbiometrics_group_sample_index[idx_ordered] <- seq_along(idx_ordered)
  }

  active_info <- ttl_alignment_active_info(
    dat = dat,
    ttl_cols = ttl_cols,
    event_value = event_value
  )

  ttl_active <- active_info$active
  event_ttl_column <- active_info$event_ttl_column
  event_ttl_value <- active_info$event_ttl_value

  ttl_valid <- rep(TRUE, nrow(dat))

  if (!is.null(ttl_valid_col) && isTRUE(require_valid_ttl)) {
    ttl_valid <- ttl_alignment_valid(dat[[ttl_valid_col]], valid_values)
  }

  event_active <- ttl_active & ttl_valid

  event_rows <- integer()
  event_group_sequence <- integer()

  for (group_name in names(group_indices)) {
    idx <- group_indices[[group_name]]
    idx <- idx[order(dat$.gpbiometrics_order_value[idx], dat$.gpbiometrics_row_id[idx])]

    active_group <- event_active[idx]
    value_group <- event_ttl_value[idx]

    lag_active <- c(FALSE, utils::head(active_group, -1))
    lag_value <- c(NA_character_, utils::head(value_group, -1))

    is_event <- switch(
      event_edge,
      rising = active_group & !lag_active,
      change = active_group & (!lag_active | value_group != lag_value),
      active = active_group
    )

    candidate_rows <- idx[is_event]

    if (length(candidate_rows) > 1 &&
        collapse_nearby_ms > 0 &&
        !all(is.na(dat$.gpbiometrics_time_ms[candidate_rows]))) {
      keep <- rep(TRUE, length(candidate_rows))
      last_kept_time <- dat$.gpbiometrics_time_ms[candidate_rows[1]]

      for (i in seq_along(candidate_rows)[-1]) {
        current_time <- dat$.gpbiometrics_time_ms[candidate_rows[i]]

        if (!is.na(current_time) &&
            !is.na(last_kept_time) &&
            (current_time - last_kept_time) < collapse_nearby_ms) {
          keep[i] <- FALSE
        } else {
          last_kept_time <- current_time
        }
      }

      candidate_rows <- candidate_rows[keep]
    }

    if (length(candidate_rows) > 0) {
      event_rows <- c(event_rows, candidate_rows)
      event_group_sequence <- c(event_group_sequence, seq_along(candidate_rows))
    }
  }

  if (length(event_rows) == 0) {
    events <- ttl_alignment_empty_events(group_cols)
    aligned <- ttl_alignment_empty_aligned(dat)

    return(structure(
      list(
        overview = data.frame(
          input_rows = nrow(dat),
          ttl_event_rows = 0L,
          aligned_rows = 0L,
          group_count = length(group_indices),
          status = "no_ttl_events_detected",
          stringsAsFactors = FALSE
        ),
        events = events,
        aligned_data = aligned,
        settings = list(
          ttl_cols = ttl_cols,
          event_col = event_col,
          ttl_valid_col = ttl_valid_col,
          time_col = time_col,
          sample_col = sample_col,
          group_cols = group_cols,
          event_source = event_source,
          event_edge = event_edge,
          pre_window_ms = pre_window_ms,
          post_window_ms = post_window_ms,
          pre_window_samples = pre_window_samples,
          post_window_samples = post_window_samples,
          collapse_nearby_ms = collapse_nearby_ms,
          require_valid_ttl = require_valid_ttl
        )
      ),
      class = c("gazepoint_biometrics_ttl_alignment", "list")
    ))
  }

  event_rows <- event_rows[order(
    dat$.gpbiometrics_group_id[event_rows],
    dat$.gpbiometrics_order_value[event_rows],
    dat$.gpbiometrics_row_id[event_rows]
  )]

  event_group_sequence <- stats::ave(
    seq_along(event_rows),
    dat$.gpbiometrics_group_id[event_rows],
    FUN = seq_along
  )

  events <- data.frame(
    ttl_event_id = paste0("ttl_event_", seq_along(event_rows)),
    ttl_event_sequence = as.integer(event_group_sequence),
    event_row_id = dat$.gpbiometrics_row_id[event_rows],
    event_group_id = dat$.gpbiometrics_group_id[event_rows],
    event_group_sample_index = dat$.gpbiometrics_group_sample_index[event_rows],
    event_time_ms = dat$.gpbiometrics_time_ms[event_rows],
    event_ttl_column = event_ttl_column[event_rows],
    event_ttl_value = event_ttl_value[event_rows],
    stringsAsFactors = FALSE
  )

  if (length(group_cols) > 0) {
    events <- cbind(dat[event_rows, group_cols, drop = FALSE], events)
  }

  aligned_parts <- vector("list", nrow(events))

  for (i in seq_len(nrow(events))) {
    event_group_id <- events$event_group_id[i]
    group_rows <- which(dat$.gpbiometrics_group_id == event_group_id)

    event_row <- event_rows[i]
    event_sample_index <- dat$.gpbiometrics_group_sample_index[event_row]
    rel_sample <- dat$.gpbiometrics_group_sample_index[group_rows] - event_sample_index

    event_time <- dat$.gpbiometrics_time_ms[event_row]
    rel_time <- dat$.gpbiometrics_time_ms[group_rows] - event_time

    if (!all(is.na(rel_time)) && !is.na(event_time)) {
      keep_rows <- rel_time >= -pre_window_ms & rel_time <= post_window_ms
      keep_rows[is.na(keep_rows)] <- FALSE
    } else {
      pre_s <- if (is.null(pre_window_samples)) 0 else pre_window_samples
      post_s <- if (is.null(post_window_samples)) 0 else post_window_samples

      if (!is.numeric(pre_s) || length(pre_s) != 1 || pre_s < 0) {
        stop("`pre_window_samples` must be `NULL` or a single non-negative number.", call. = FALSE)
      }

      if (!is.numeric(post_s) || length(post_s) != 1 || post_s < 0) {
        stop("`post_window_samples` must be `NULL` or a single non-negative number.", call. = FALSE)
      }

      keep_rows <- rel_sample >= -pre_s & rel_sample <= post_s
    }

    kept_group_rows <- group_rows[keep_rows]

    if (length(kept_group_rows) == 0) {
      next
    }

    aligned_i <- dat[kept_group_rows, , drop = FALSE]
    aligned_i$ttl_event_id <- events$ttl_event_id[i]
    aligned_i$ttl_event_sequence <- events$ttl_event_sequence[i]
    aligned_i$event_row_id <- events$event_row_id[i]
    aligned_i$event_group_sample_index <- events$event_group_sample_index[i]
    aligned_i$event_time_ms <- events$event_time_ms[i]
    aligned_i$event_ttl_column <- events$event_ttl_column[i]
    aligned_i$event_ttl_value <- events$event_ttl_value[i]
    aligned_i$event_relative_sample_index <- rel_sample[keep_rows]
    aligned_i$event_relative_time_ms <- rel_time[keep_rows]
    aligned_i$event_window_position <- ifelse(
      aligned_i$event_relative_sample_index < 0,
      "pre_event",
      ifelse(aligned_i$event_relative_sample_index == 0, "event", "post_event")
    )
    aligned_i$within_pre_event_window <- aligned_i$event_relative_sample_index < 0
    aligned_i$within_post_event_window <- aligned_i$event_relative_sample_index > 0
    aligned_i$ttl_alignment_status <- "aligned"

    aligned_parts[[i]] <- aligned_i
  }

  aligned <- do.call(rbind, aligned_parts)

  if (is.null(aligned)) {
    aligned <- ttl_alignment_empty_aligned(dat)
    status <- "ttl_events_detected_no_rows_aligned"
  } else {
    rownames(aligned) <- NULL
    status <- "ttl_events_aligned"
  }

  internal_cols <- c(
    ".gpbiometrics_group_id",
    ".gpbiometrics_order_value"
  )

  aligned <- aligned[, setdiff(names(aligned), internal_cols), drop = FALSE]

  events$event_group_id <- NULL

  structure(
    list(
      overview = data.frame(
        input_rows = nrow(dat),
        ttl_event_rows = nrow(events),
        aligned_rows = nrow(aligned),
        group_count = length(group_indices),
        status = status,
        stringsAsFactors = FALSE
      ),
      events = events,
      aligned_data = aligned,
      settings = list(
        ttl_cols = ttl_cols,
        event_col = event_col,
        ttl_valid_col = ttl_valid_col,
        time_col = time_col,
        sample_col = sample_col,
        group_cols = group_cols,
        event_source = event_source,
        event_edge = event_edge,
        pre_window_ms = pre_window_ms,
        post_window_ms = post_window_ms,
        pre_window_samples = pre_window_samples,
        post_window_samples = post_window_samples,
        collapse_nearby_ms = collapse_nearby_ms,
        require_valid_ttl = require_valid_ttl
      )
    ),
    class = c("gazepoint_biometrics_ttl_alignment", "list")
  )
}

ttl_alignment_first_existing <- function(names_dat, candidates) {
  exact <- candidates[candidates %in% names_dat]
  if (length(exact) > 0) {
    return(exact[1])
  }

  lower_names <- tolower(names_dat)
  lower_candidates <- tolower(candidates)
  idx <- match(lower_candidates, lower_names)
  idx <- idx[!is.na(idx)]

  if (length(idx) > 0) {
    return(names_dat[idx[1]])
  }

  NULL
}

ttl_alignment_infer_ttl_cols <- function(names_dat) {
  marker <- ttl_alignment_first_existing(names_dat, c("ttl_marker"))

  if (!is.null(marker)) {
    return(marker)
  }

  raw_ttl <- grep("^TTL[0-6]$", names_dat, value = TRUE, ignore.case = TRUE)
  raw_ttl <- raw_ttl[order(tolower(raw_ttl))]

  raw_ttl
}

ttl_alignment_resolve_group_cols <- function(names_dat,
                                             group_cols,
                                             participant_col,
                                             stimulus_col,
                                             trial_col) {
  explicit_cols <- c(participant_col, stimulus_col, trial_col)
  explicit_cols <- explicit_cols[!is.null(explicit_cols)]

  if (!is.null(group_cols)) {
    return(unique(c(group_cols, explicit_cols)))
  }

  participant <- ttl_alignment_first_existing(
    names_dat,
    c("participant", "subject", "subject_id", "USER", "USER_FILE", "user_file")
  )

  stimulus <- ttl_alignment_first_existing(
    names_dat,
    c("stimulus", "stimulus_id", "MEDIA_ID", "MEDIA_NAME", "media_id", "media_name")
  )

  trial <- ttl_alignment_first_existing(
    names_dat,
    c("trial", "trial_id", "TRIAL", "trial_global")
  )

  unique(stats::na.omit(c(explicit_cols, participant, stimulus, trial)))
}

ttl_alignment_group_id <- function(dat, group_cols) {
  if (length(group_cols) == 0) {
    return(rep("all", nrow(dat)))
  }

  group_dat <- dat[group_cols]
  group_dat[] <- lapply(group_dat, function(x) {
    x_chr <- as.character(x)
    x_chr[is.na(x_chr)] <- "<NA>"
    x_chr
  })

  apply(group_dat, 1, paste, collapse = "||")
}

ttl_alignment_time_ms <- function(dat, time_col) {
  if (is.null(time_col)) {
    return(rep(NA_real_, nrow(dat)))
  }

  time_numeric <- suppressWarnings(as.numeric(dat[[time_col]]))

  if (all(is.na(time_numeric))) {
    return(rep(NA_real_, nrow(dat)))
  }

  diffs <- diff(time_numeric)
  diffs <- diffs[is.finite(diffs) & diffs > 0]

  if (length(diffs) == 0) {
    return(time_numeric)
  }

  median_diff <- stats::median(diffs, na.rm = TRUE)

  # Gazepoint-like seconds often have frame steps around .016 at 60 Hz.
  # Larger numeric steps are kept as milliseconds/input-ms scale.
  if (is.finite(median_diff) && median_diff > 0 && median_diff <= 0.25) {
    return(time_numeric * 1000)
  }

  time_numeric
}

ttl_alignment_active <- function(x, event_value = NULL) {
  if (!is.null(event_value)) {
    return(!is.na(x) & as.character(x) %in% as.character(event_value))
  }

  if (is.logical(x)) {
    return(!is.na(x) & x)
  }

  if (is.numeric(x)) {
    return(!is.na(x) & x != 0)
  }

  x_chr <- trimws(as.character(x))
  x_chr[is.na(x_chr)] <- ""

  x_num <- suppressWarnings(as.numeric(x_chr))
  non_empty <- x_chr != ""

  if (sum(!is.na(x_num) & non_empty) >= max(1, sum(non_empty) / 2)) {
    return(!is.na(x_num) & x_num != 0)
  }

  non_empty & !toupper(x_chr) %in% c("0", "FALSE", "F", "NA", "NAN", "NULL")
}

ttl_alignment_valid <- function(x, valid_values) {
  if (is.null(valid_values)) {
    return(ttl_alignment_active(x))
  }

  !is.na(x) & as.character(x) %in% as.character(valid_values)
}

ttl_alignment_active_info <- function(dat, ttl_cols, event_value) {
  active_mat <- sapply(
    ttl_cols,
    function(col) ttl_alignment_active(dat[[col]], event_value = event_value)
  )

  if (is.null(dim(active_mat))) {
    active_mat <- matrix(active_mat, ncol = 1)
    colnames(active_mat) <- ttl_cols
  }

  active <- rowSums(active_mat, na.rm = TRUE) > 0

  event_ttl_column <- rep(NA_character_, nrow(dat))
  event_ttl_value <- rep(NA_character_, nrow(dat))

  for (i in seq_len(nrow(dat))) {
    active_cols <- ttl_cols[active_mat[i, ]]

    if (length(active_cols) > 0) {
      event_ttl_column[i] <- paste(active_cols, collapse = ";")
      event_ttl_value[i] <- paste(
        paste0(active_cols, "=", as.character(dat[i, active_cols, drop = TRUE])),
        collapse = ";"
      )
    }
  }

  list(
    active = active,
    event_ttl_column = event_ttl_column,
    event_ttl_value = event_ttl_value
  )
}

ttl_alignment_empty_events <- function(group_cols) {
  base <- data.frame(stringsAsFactors = FALSE)

  if (length(group_cols) > 0) {
    for (col in group_cols) {
      base[[col]] <- character()
    }
  }

  base$ttl_event_id <- character()
  base$ttl_event_sequence <- integer()
  base$event_row_id <- integer()
  base$event_group_sample_index <- integer()
  base$event_time_ms <- numeric()
  base$event_ttl_column <- character()
  base$event_ttl_value <- character()

  base
}

ttl_alignment_empty_aligned <- function(dat) {
  aligned <- dat[0, , drop = FALSE]
  aligned$ttl_event_id <- character()
  aligned$ttl_event_sequence <- integer()
  aligned$event_row_id <- integer()
  aligned$event_group_sample_index <- integer()
  aligned$event_time_ms <- numeric()
  aligned$event_ttl_column <- character()
  aligned$event_ttl_value <- character()
  aligned$event_relative_sample_index <- integer()
  aligned$event_relative_time_ms <- numeric()
  aligned$event_window_position <- character()
  aligned$within_pre_event_window <- logical()
  aligned$within_post_event_window <- logical()
  aligned$ttl_alignment_status <- character()

  internal_cols <- c(
    ".gpbiometrics_group_id",
    ".gpbiometrics_order_value"
  )

  aligned[, setdiff(names(aligned), internal_cols), drop = FALSE]
}

test_that("align_gazepoint_biometrics_to_ttl detects standardized TTL rising edges", {
  dat <- data.frame(
    participant = "P1",
    trial = 1,
    time_ms = seq(0, 900, by = 100),
    ttl_marker = c(0, 0, 1, 1, 0, 0, 1, 0, 0, 0),
    ttl_validity_flag = 1,
    GSR_US = seq(1, 2, length.out = 10)
  )

  res <- align_gazepoint_biometrics_to_ttl(
    dat,
    time_col = "time_ms",
    group_cols = c("participant", "trial"),
    pre_window_ms = 100,
    post_window_ms = 200
  )

  expect_s3_class(res, "gazepoint_biometrics_ttl_alignment")
  expect_equal(res$overview$status, "ttl_events_aligned")
  expect_equal(nrow(res$events), 2)
  expect_equal(unique(res$aligned_data$ttl_event_id), c("ttl_event_1", "ttl_event_2"))
  expect_true(0 %in% res$aligned_data$event_relative_sample_index)
  expect_true("pre_event" %in% res$aligned_data$event_window_position)
  expect_true("event" %in% res$aligned_data$event_window_position)
  expect_true("post_event" %in% res$aligned_data$event_window_position)
})

test_that("align_gazepoint_biometrics_to_ttl detects raw TTL0-TTL6 columns", {
  dat <- data.frame(
    subject = "S1",
    MEDIA_ID = "M1",
    time_ms = seq(0, 900, by = 100),
    TTL0 = c(0, 0, 1, 1, 0, 0, 0, 0, 0, 0),
    TTL1 = c(0, 0, 0, 0, 0, 0, 1, 0, 0, 0),
    TTLV = 1,
    HR = seq(70, 79)
  )

  res <- align_gazepoint_biometrics_to_ttl(
    dat,
    time_col = "time_ms",
    pre_window_ms = 100,
    post_window_ms = 100
  )

  expect_equal(nrow(res$events), 2)
  expect_true("TTL0" %in% res$events$event_ttl_column)
  expect_true("TTL1" %in% res$events$event_ttl_column)
  expect_equal(res$settings$ttl_valid_col, "TTLV")
})

test_that("align_gazepoint_biometrics_to_ttl supports a user-specified event column", {
  dat <- data.frame(
    participant = rep(c("P1", "P2"), each = 5),
    trial = rep(1, 10),
    time_ms = rep(seq(0, 400, by = 100), 2),
    marker = c(0, "start", 0, 0, 0, 0, "start", 0, 0, 0),
    GSR_US = seq(1, 2, length.out = 10)
  )

  res <- align_gazepoint_biometrics_to_ttl(
    dat,
    event_col = "marker",
    event_value = "start",
    time_col = "time_ms",
    group_cols = c("participant", "trial"),
    pre_window_ms = 100,
    post_window_ms = 100
  )

  expect_equal(nrow(res$events), 2)
  expect_equal(sort(unique(res$events$participant)), c("P1", "P2"))
  expect_equal(res$settings$event_source, "user_event_col")
})

test_that("align_gazepoint_biometrics_to_ttl uses TTL validity conservatively", {
  dat <- data.frame(
    time_ms = seq(0, 400, by = 100),
    TTL0 = c(0, 1, 0, 0, 0),
    TTLV = c(1, 0, 1, 1, 1),
    GSR_US = seq(1, 1.4, length.out = 5)
  )

  res <- align_gazepoint_biometrics_to_ttl(
    dat,
    time_col = "time_ms",
    pre_window_ms = 100,
    post_window_ms = 100
  )

  expect_equal(res$overview$status, "no_ttl_events_detected")
  expect_equal(nrow(res$events), 0)
  expect_equal(nrow(res$aligned_data), 0)
})

test_that("align_gazepoint_biometrics_to_ttl falls back to sample windows without time", {
  dat <- data.frame(
    participant = "P1",
    marker = c(0, 1, 0, 0, 0),
    HR = c(70, 72, 73, 74, 75)
  )

  res <- align_gazepoint_biometrics_to_ttl(
    dat,
    event_col = "marker",
    group_cols = "participant",
    pre_window_samples = 1,
    post_window_samples = 1
  )

  expect_equal(nrow(res$events), 1)
  expect_equal(
    res$aligned_data$event_relative_sample_index,
    c(-1, 0, 1)
  )
  expect_true(all(is.na(res$aligned_data$event_relative_time_ms)))
})

#' Classify SCRs into FIR, SIR, and TIR latency intervals
#'
#' Classifies extracted SCRs into first-, second-, and third-interval response
#' windows based on response latency after stimulus onset.
#'
#' @param dat A data frame containing SCR events or peaks.
#' @param response_time_col Optional response/peak time column.
#' @param stimulus_onset_col Optional stimulus-onset column. Required when
#'   `latency_col` is not supplied.
#' @param latency_col Optional precomputed latency column.
#' @param output_col Name of the output interval column.
#' @param latency_output_col Name of the latency output column.
#' @param fir Numeric vector of length two defining FIR window in seconds.
#' @param sir Numeric vector of length two defining SIR window in seconds.
#' @param tir Numeric vector of length two defining TIR window in seconds.
#'
#' @return A data frame with interval labels and latency metadata.
#' @export
classify_gazepoint_scr_intervals <- function(dat,
                                             response_time_col = NULL,
                                             stimulus_onset_col = NULL,
                                             latency_col = NULL,
                                             output_col = "scr_interval",
                                             latency_output_col = "scr_latency_s",
                                             fir = c(1, 4),
                                             sir = c(4, 7),
                                             tir = c(7, 10)) {
  if (!is.data.frame(dat)) {
    stop("`dat` must be a data frame.", call. = FALSE)
  }

  gpbiometrics_scr_interval_validate_window(fir, "fir")
  gpbiometrics_scr_interval_validate_window(sir, "sir")
  gpbiometrics_scr_interval_validate_window(tir, "tir")

  if (!is.null(latency_col)) {
    if (!latency_col %in% names(dat)) {
      stop("Column `", latency_col, "` was not found in `dat`.", call. = FALSE)
    }

    latency <- dat[[latency_col]]
  } else {
    if (is.null(response_time_col) || is.null(stimulus_onset_col)) {
      stop(
        "Supply either `latency_col` or both `response_time_col` and `stimulus_onset_col`.",
        call. = FALSE
      )
    }

    if (!response_time_col %in% names(dat)) {
      stop("Column `", response_time_col, "` was not found in `dat`.", call. = FALSE)
    }

    if (!stimulus_onset_col %in% names(dat)) {
      stop("Column `", stimulus_onset_col, "` was not found in `dat`.", call. = FALSE)
    }

    latency <- dat[[response_time_col]] - dat[[stimulus_onset_col]]
  }

  if (!is.numeric(latency)) {
    stop("SCR latency values must be numeric and expressed in seconds.", call. = FALSE)
  }

  out <- dat
  interval <- rep("outside_defined_intervals", nrow(out))
  interval[!is.finite(latency)] <- "missing_latency"

  in_fir <- is.finite(latency) & latency >= fir[1] & latency < fir[2]
  in_sir <- is.finite(latency) & latency >= sir[1] & latency < sir[2]
  in_tir <- is.finite(latency) & latency >= tir[1] & latency <= tir[2]

  interval[in_fir] <- "FIR"
  interval[in_sir] <- "SIR"
  interval[in_tir] <- "TIR"

  out[[latency_output_col]] <- latency
  out[[output_col]] <- interval

  summary <- data.frame(
    input_rows = nrow(out),
    fir_rows = sum(interval == "FIR"),
    sir_rows = sum(interval == "SIR"),
    tir_rows = sum(interval == "TIR"),
    outside_rows = sum(interval == "outside_defined_intervals"),
    missing_latency_rows = sum(interval == "missing_latency"),
    fir_window = paste(fir, collapse = "-"),
    sir_window = paste(sir, collapse = "-"),
    tir_window = paste(tir, collapse = "-"),
    status = "scr_intervals_classified",
    interpretation = paste(
      "FIR, SIR, and TIR labels are latency-window descriptors.",
      "They do not infer emotion, valence, stress, fear, trust, preference, cognition, or diagnosis."
    ),
    stringsAsFactors = FALSE
  )

  attr(out, "scr_interval_summary") <- summary
  attr(out, "scr_interval_settings") <- list(
    response_time_col = response_time_col,
    stimulus_onset_col = stimulus_onset_col,
    latency_col = latency_col,
    output_col = output_col,
    latency_output_col = latency_output_col,
    fir = fir,
    sir = sir,
    tir = tir
  )

  class(out) <- unique(c("gazepoint_scr_intervals", class(out)))

  out
}

gpbiometrics_scr_interval_validate_window <- function(x, name) {
  if (!is.numeric(x) || length(x) != 2 || any(!is.finite(x)) || x[1] >= x[2]) {
    stop("`", name, "` must be a numeric vector of length two with start < end.", call. = FALSE)
  }

  invisible(TRUE)
}

#' Flag EDA artifacts using transparent Kleckner-style heuristics
#'
#' Applies simple transparent EDA artifact flags: non-finite values,
#' physiological range violations, rapid percent change per second, and
#' transitional padding around flagged samples.
#'
#' This helper is Kleckner-style rather than a claim of exact reproduction of
#' every rule in a specific external implementation.
#'
#' @param dat A data frame containing EDA data.
#' @param eda_col Conductance column in microsiemens.
#' @param time_col Optional time column.
#' @param group_cols Optional grouping columns.
#' @param min_us Minimum plausible conductance.
#' @param max_us Maximum plausible conductance.
#' @param max_abs_percent_change_per_second Maximum absolute percent change per
#'   second before flagging.
#' @param transition_padding Number of neighbouring rows to flag around bad
#'   samples within each group.
#' @param output_prefix Prefix for output columns.
#'
#' @return A data frame with artifact flag columns and summary attributes.
#' @export
flag_kleckner_eda_artifacts <- function(dat,
                                        eda_col = "GSR_US",
                                        time_col = NULL,
                                        group_cols = NULL,
                                        min_us = 0.01,
                                        max_us = 100,
                                        max_abs_percent_change_per_second = 20,
                                        transition_padding = 1,
                                        output_prefix = "kleckner") {
  if (!is.data.frame(dat)) {
    stop("`dat` must be a data frame.", call. = FALSE)
  }

  if (!eda_col %in% names(dat)) {
    stop("Column `", eda_col, "` was not found in `dat`.", call. = FALSE)
  }

  if (!is.numeric(dat[[eda_col]])) {
    stop("`eda_col` must identify a numeric conductance column.", call. = FALSE)
  }

  if (!is.null(time_col) && !time_col %in% names(dat)) {
    stop("Column `", time_col, "` was not found in `dat`.", call. = FALSE)
  }

  if (!is.null(group_cols)) {
    missing_groups <- setdiff(group_cols, names(dat))

    if (length(missing_groups) > 0) {
      stop(
        "The following `group_cols` were not found in `dat`: ",
        paste(missing_groups, collapse = ", "),
        call. = FALSE
      )
    }
  } else {
    group_cols <- character()
  }

  if (!is.numeric(transition_padding) || length(transition_padding) != 1 || transition_padding < 0) {
    stop("`transition_padding` must be a non-negative number.", call. = FALSE)
  }

  transition_padding <- as.integer(transition_padding)

  out <- dat

  nonfinite_col <- paste0(output_prefix, "_nonfinite")
  range_col <- paste0(output_prefix, "_range_artifact")
  rapid_col <- paste0(output_prefix, "_rapid_change_artifact")
  transition_col <- paste0(output_prefix, "_transition_artifact")
  final_col <- paste0(output_prefix, "_artifact")
  status_col <- paste0(output_prefix, "_artifact_status")

  for (nm in c(nonfinite_col, range_col, rapid_col, transition_col, final_col)) {
    out[[nm]] <- FALSE
  }

  out[[status_col]] <- "usable"

  groups <- gpbiometrics_kleckner_split_indices(out, group_cols)

  for (idx in groups) {
    x <- out[[eda_col]][idx]

    nonfinite <- !is.finite(x)
    range_bad <- is.finite(x) & (x < min_us | x > max_us)

    rapid_bad <- rep(FALSE, length(idx))

    if (length(idx) >= 2) {
      dx <- diff(x)

      if (!is.null(time_col)) {
        time_values <- out[[time_col]][idx]
        dt <- diff(time_values)
        dt[!is.finite(dt) | dt <= 0] <- NA_real_
      } else {
        dt <- rep(1, length(dx))
      }

      previous <- x[-length(x)]
      percent_change_per_second <- abs((dx / previous) * 100) / dt

      rapid_pair <- is.finite(percent_change_per_second) &
        percent_change_per_second > max_abs_percent_change_per_second

      rapid_bad[-1] <- rapid_pair
    }

    primary_bad <- nonfinite | range_bad | rapid_bad
    transition_bad <- gpbiometrics_kleckner_expand_flags(primary_bad, transition_padding) & !primary_bad

    out[[nonfinite_col]][idx] <- nonfinite
    out[[range_col]][idx] <- range_bad
    out[[rapid_col]][idx] <- rapid_bad
    out[[transition_col]][idx] <- transition_bad
    out[[final_col]][idx] <- primary_bad | transition_bad
  }

  out[[status_col]][out[[transition_col]]] <- "transition_artifact"
  out[[status_col]][out[[rapid_col]]] <- "rapid_change_artifact"
  out[[status_col]][out[[range_col]]] <- "range_artifact"
  out[[status_col]][out[[nonfinite_col]]] <- "nonfinite_artifact"

  summary <- data.frame(
    input_rows = nrow(out),
    artifact_rows = sum(out[[final_col]], na.rm = TRUE),
    artifact_rate = mean(out[[final_col]], na.rm = TRUE),
    nonfinite_rows = sum(out[[nonfinite_col]], na.rm = TRUE),
    range_artifact_rows = sum(out[[range_col]], na.rm = TRUE),
    rapid_change_artifact_rows = sum(out[[rapid_col]], na.rm = TRUE),
    transition_artifact_rows = sum(out[[transition_col]], na.rm = TRUE),
    min_us = min_us,
    max_us = max_us,
    max_abs_percent_change_per_second = max_abs_percent_change_per_second,
    transition_padding = transition_padding,
    status = "kleckner_style_artifacts_flagged",
    interpretation = paste(
      "These are transparent Kleckner-style EDA quality flags.",
      "They indicate potential signal-quality problems and do not infer emotion, valence, stress, cognition, or diagnosis."
    ),
    stringsAsFactors = FALSE
  )

  attr(out, "kleckner_artifact_summary") <- summary
  attr(out, "kleckner_artifact_settings") <- list(
    eda_col = eda_col,
    time_col = time_col,
    group_cols = group_cols,
    min_us = min_us,
    max_us = max_us,
    max_abs_percent_change_per_second = max_abs_percent_change_per_second,
    transition_padding = transition_padding,
    output_prefix = output_prefix
  )

  class(out) <- unique(c("gazepoint_kleckner_eda_artifacts", class(out)))

  out
}

gpbiometrics_kleckner_split_indices <- function(dat, group_cols) {
  if (length(group_cols) == 0) {
    return(list(all_rows = seq_len(nrow(dat))))
  }

  group_frame <- dat[group_cols]
  group_frame[] <- lapply(group_frame, function(x) {
    x <- as.character(x)
    x[is.na(x)] <- "<NA>"
    x
  })

  group_key <- do.call(paste, c(group_frame, sep = " | "))
  split(seq_len(nrow(dat)), group_key)
}

gpbiometrics_kleckner_expand_flags <- function(flags, padding) {
  if (padding <= 0 || !any(flags)) {
    return(flags)
  }

  out <- flags
  bad_idx <- which(flags)

  for (i in bad_idx) {
    lo <- max(1, i - padding)
    hi <- min(length(flags), i + padding)
    out[lo:hi] <- TRUE
  }

  out
}

test_that("classify_gazepoint_scr_intervals classifies FIR SIR and TIR", {
  dat <- data.frame(
    stimulus_onset = 0,
    peak_time = c(1.5, 5, 8, 12, NA)
  )

  out <- classify_gazepoint_scr_intervals(
    dat,
    response_time_col = "peak_time",
    stimulus_onset_col = "stimulus_onset"
  )

  expect_equal(
    out$scr_interval,
    c("FIR", "SIR", "TIR", "outside_defined_intervals", "missing_latency")
  )

  summary <- attr(out, "scr_interval_summary")
  expect_equal(summary$fir_rows, 1)
  expect_equal(summary$sir_rows, 1)
  expect_equal(summary$tir_rows, 1)
})

test_that("flag_kleckner_eda_artifacts flags range and rapid-change artifacts", {
  dat <- data.frame(
    participant = "p1",
    time = 1:6,
    GSR_US = c(1, 1.1, 1.2, 200, 1.3, NA)
  )

  out <- flag_kleckner_eda_artifacts(
    dat,
    eda_col = "GSR_US",
    time_col = "time",
    group_cols = "participant",
    transition_padding = 0
  )

  expect_true(out$kleckner_range_artifact[4])
  expect_true(out$kleckner_nonfinite[6])
  expect_true(any(out$kleckner_artifact))

  summary <- attr(out, "kleckner_artifact_summary")
  expect_equal(summary$status, "kleckner_style_artifacts_flagged")
})

#' Import Gazepoint-related streams from an LSL/XDF file
#'
#' Reads an XDF file through the optional Python `pyxdf` package via
#' `reticulate`. This supports high-end LSL workflows without making Python a
#' hard dependency of gpbiometrics.
#'
#' @param path Path to an `.xdf` file.
#' @param stream_name_pattern Regular expression used to identify Gazepoint-like
#'   streams when `include_all_streams = FALSE`.
#' @param include_all_streams Logical. If `TRUE`, return all streams.
#' @param flatten Logical. If `TRUE`, convert streams to data frames where
#'   possible.
#' @param pyxdf_module Python module name, usually `"pyxdf"`.
#'
#' @return A list with `overview`, `streams`, `header`, and `settings`.
#' @export
import_gazepoint_lsl_xdf <- function(path,
                                     stream_name_pattern = "Gazepoint|GP3|GSR|EDA|Biometric|TTL|Pupil|Gaze",
                                     include_all_streams = FALSE,
                                     flatten = TRUE,
                                     pyxdf_module = "pyxdf") {
  if (!is.character(path) || length(path) != 1) {
    stop("`path` must be a single file path.", call. = FALSE)
  }

  if (!file.exists(path)) {
    stop("File does not exist: ", path, call. = FALSE)
  }

  if (!requireNamespace("reticulate", quietly = TRUE)) {
    stop(
      "The optional package `reticulate` is required for XDF import. Install it with install.packages('reticulate').",
      call. = FALSE
    )
  }

  pyxdf <- tryCatch(
    reticulate::import(pyxdf_module, delay_load = FALSE),
    error = function(e) {
      stop(
        "Python module `", pyxdf_module, "` could not be imported. ",
        "Install it in your active Python environment with: pip install pyxdf",
        call. = FALSE
      )
    }
  )

  loaded <- pyxdf$load_xdf(path)
  streams_raw <- reticulate::py_to_r(loaded[[1]])
  header <- reticulate::py_to_r(loaded[[2]])

  stream_meta <- lapply(seq_along(streams_raw), function(i) {
    gpbiometrics_xdf_stream_meta(streams_raw[[i]], i)
  })

  meta_df <- do.call(rbind, stream_meta)

  selected <- rep(TRUE, nrow(meta_df))

  if (!isTRUE(include_all_streams)) {
    selected <- grepl(stream_name_pattern, meta_df$name, ignore.case = TRUE) |
      grepl(stream_name_pattern, meta_df$type, ignore.case = TRUE)
  }

  selected_indices <- which(selected)

  streams <- lapply(selected_indices, function(i) {
    stream <- streams_raw[[i]]
    meta <- gpbiometrics_xdf_stream_meta(stream, i)

    if (isTRUE(flatten)) {
      data <- gpbiometrics_xdf_stream_to_data_frame(stream)
    } else {
      data <- stream
    }

    list(
      meta = meta,
      data = data
    )
  })

  names(streams) <- paste0("stream_", selected_indices)

  overview <- data.frame(
    path = normalizePath(path, winslash = "/", mustWork = FALSE),
    total_streams = length(streams_raw),
    selected_streams = length(streams),
    include_all_streams = include_all_streams,
    stream_name_pattern = stream_name_pattern,
    status = if (length(streams) > 0) "xdf_import_complete" else "xdf_import_no_matching_streams",
    interpretation = paste(
      "XDF import returns synchronized LSL stream contents and metadata.",
      "Downstream alignment and resampling remain design-specific and should be verified before analysis."
    ),
    stringsAsFactors = FALSE
  )

  structure(
    list(
      overview = overview,
      stream_overview = meta_df,
      streams = streams,
      header = header,
      settings = list(
        path = path,
        stream_name_pattern = stream_name_pattern,
        include_all_streams = include_all_streams,
        flatten = flatten,
        pyxdf_module = pyxdf_module
      )
    ),
    class = c("gazepoint_lsl_xdf_import", "list")
  )
}

gpbiometrics_xdf_stream_meta <- function(stream, index) {
  info <- stream$info

  data.frame(
    stream_index = index,
    name = gpbiometrics_xdf_scalar(info$name, paste0("stream_", index)),
    type = gpbiometrics_xdf_scalar(info$type, NA_character_),
    channel_count = suppressWarnings(as.integer(gpbiometrics_xdf_scalar(info$channel_count, NA_character_))),
    nominal_srate = suppressWarnings(as.numeric(gpbiometrics_xdf_scalar(info$nominal_srate, NA_character_))),
    stringsAsFactors = FALSE
  )
}

gpbiometrics_xdf_scalar <- function(x, default = NA_character_) {
  if (is.null(x)) {
    return(default)
  }

  if (is.list(x) && length(x) > 0) {
    return(as.character(x[[1]]))
  }

  if (length(x) > 0) {
    return(as.character(x[1]))
  }

  default
}

gpbiometrics_xdf_stream_to_data_frame <- function(stream) {
  ts <- stream$time_series
  stamps <- as.numeric(stream$time_stamps)

  if (is.null(ts)) {
    return(data.frame(time_stamp = stamps))
  }

  if (is.matrix(ts) || is.data.frame(ts)) {
    out <- as.data.frame(ts, stringsAsFactors = FALSE)
    names(out) <- paste0("channel_", seq_len(ncol(out)))
    out <- cbind(time_stamp = stamps, out)
    return(out)
  }

  if (is.vector(ts) && !is.list(ts)) {
    return(data.frame(time_stamp = stamps, value = ts, stringsAsFactors = FALSE))
  }

  if (is.list(ts)) {
    values <- vapply(ts, function(x) paste(as.character(x), collapse = " | "), character(1))
    return(data.frame(time_stamp = stamps, value = values, stringsAsFactors = FALSE))
  }

  data.frame(time_stamp = stamps, stringsAsFactors = FALSE)
}

#' Model EDA events as a dependency-light point process
#'
#' Creates event-time, inter-event interval, and inverse-Gaussian-style summary
#' tables for EDA/SCR events. Events can be supplied directly through an event
#' column or derived from positive EDA-derivative bursts.
#'
#' This function is a compact point-process summary/model-preparation helper. It
#' does not fit a full latent sympathetic state-space model.
#'
#' @param dat A data frame.
#' @param eda_col Numeric EDA column.
#' @param time_col Numeric time column.
#' @param group_cols Optional grouping columns.
#' @param event_time_col Optional column of event onset times.
#' @param event_indicator_col Optional binary event indicator column.
#' @param derivative_mad_multiplier MAD multiplier for derivative-derived events.
#' @param min_event_distance_s Minimum distance between derived events.
#'
#' @return A list with `overview`, `event_table`, `interval_table`,
#'   `process_summary`, and `settings`.
#' @export
model_gazepoint_eda_point_process <- function(dat,
                                              eda_col = "GSR_US",
                                              time_col = "CNT",
                                              group_cols = NULL,
                                              event_time_col = NULL,
                                              event_indicator_col = NULL,
                                              derivative_mad_multiplier = 6,
                                              min_event_distance_s = 1) {
  if (!is.data.frame(dat)) {
    stop("`dat` must be a data frame.", call. = FALSE)
  }

  required <- c(eda_col, time_col, event_time_col, event_indicator_col)
  required <- required[!is.null(required)]

  missing_required <- setdiff(required, names(dat))
  if (length(missing_required) > 0) {
    stop("Missing required columns: ", paste(missing_required, collapse = ", "), call. = FALSE)
  }

  if (!is.numeric(dat[[eda_col]])) {
    stop("`eda_col` must identify a numeric column.", call. = FALSE)
  }

  if (!is.numeric(dat[[time_col]])) {
    stop("`time_col` must identify a numeric column.", call. = FALSE)
  }

  if (is.null(group_cols)) {
    group_cols <- character()
  }

  missing_groups <- setdiff(group_cols, names(dat))
  if (length(missing_groups) > 0) {
    stop("Missing `group_cols`: ", paste(missing_groups, collapse = ", "), call. = FALSE)
  }

  groups <- gpbiometrics_pp_split(dat, group_cols)

  event_rows <- list()
  interval_rows <- list()
  summary_rows <- list()
  event_id <- 1L
  interval_id <- 1L

  for (group_id in names(groups)) {
    idx <- groups[[group_id]]
    idx <- idx[order(dat[[time_col]][idx])]

    time <- dat[[time_col]][idx]
    eda <- dat[[eda_col]][idx]

    event_times <- gpbiometrics_eda_pp_events(
      dat = dat,
      idx = idx,
      time_col = time_col,
      eda_col = eda_col,
      event_time_col = event_time_col,
      event_indicator_col = event_indicator_col,
      derivative_mad_multiplier = derivative_mad_multiplier,
      min_event_distance_s = min_event_distance_s
    )

    event_times <- sort(unique(event_times[is.finite(event_times)]))

    for (i in seq_along(event_times)) {
      event_rows[[event_id]] <- data.frame(
        group_id = group_id,
        event_index = i,
        event_time = event_times[i],
        event_source = if (!is.null(event_time_col)) {
          "event_time_col"
        } else if (!is.null(event_indicator_col)) {
          "event_indicator_col"
        } else {
          "eda_derivative"
        },
        stringsAsFactors = FALSE
      )
      event_id <- event_id + 1L
    }

    intervals <- diff(event_times)

    if (length(intervals) > 0) {
      for (i in seq_along(intervals)) {
        interval_rows[[interval_id]] <- data.frame(
          group_id = group_id,
          interval_index = i,
          start_time = event_times[i],
          end_time = event_times[i + 1],
          inter_event_interval_s = intervals[i],
          stringsAsFactors = FALSE
        )
        interval_id <- interval_id + 1L
      }
    }

    duration <- max(time, na.rm = TRUE) - min(time, na.rm = TRUE)
    ig <- gpbiometrics_pp_inverse_gaussian_summary(intervals)

    summary_rows[[group_id]] <- data.frame(
      group_id = group_id,
      n_samples = length(idx),
      n_events = length(event_times),
      duration_s = duration,
      event_rate_per_min = if (is.finite(duration) && duration > 0) {
        length(event_times) / duration * 60
      } else {
        NA_real_
      },
      mean_inter_event_interval_s = if (length(intervals) > 0) mean(intervals) else NA_real_,
      inverse_gaussian_mu = ig$mu,
      inverse_gaussian_lambda = ig$lambda,
      interval_cv = ig$cv,
      status = if (length(event_times) >= 3) {
        "eda_point_process_summarised"
      } else {
        "insufficient_eda_events"
      },
      stringsAsFactors = FALSE
    )
  }

  event_table <- if (length(event_rows) > 0) do.call(rbind, event_rows) else data.frame()
  interval_table <- if (length(interval_rows) > 0) do.call(rbind, interval_rows) else data.frame()
  process_summary <- do.call(rbind, summary_rows)
  rownames(process_summary) <- NULL

  overview <- data.frame(
    group_count = length(groups),
    event_rows = nrow(event_table),
    interval_rows = nrow(interval_table),
    successful_groups = sum(process_summary$status == "eda_point_process_summarised"),
    problem_groups = sum(process_summary$status != "eda_point_process_summarised"),
    status = if (all(process_summary$status == "eda_point_process_summarised")) {
      "eda_point_process_complete"
    } else if (any(process_summary$status == "eda_point_process_summarised")) {
      "eda_point_process_partial"
    } else {
      "eda_point_process_insufficient_events"
    },
    interpretation = paste(
      "EDA point-process summaries describe event timing and inter-event intervals.",
      "They do not estimate latent sympathetic nerve firing or arousal states by themselves."
    ),
    stringsAsFactors = FALSE
  )

  structure(
    list(
      overview = overview,
      event_table = event_table,
      interval_table = interval_table,
      process_summary = process_summary,
      settings = list(
        eda_col = eda_col,
        time_col = time_col,
        group_cols = group_cols,
        event_time_col = event_time_col,
        event_indicator_col = event_indicator_col,
        derivative_mad_multiplier = derivative_mad_multiplier,
        min_event_distance_s = min_event_distance_s
      )
    ),
    class = c("gazepoint_eda_point_process", "list")
  )
}

#' Model heartbeats as a dependency-light point process
#'
#' Creates beat-time, interbeat interval, and inverse-Gaussian-style summary
#' tables from IBI/RR intervals. This is a compact point-process
#' model-preparation helper, not a full adaptive Bayesian heartbeat filter.
#'
#' @param dat A data frame.
#' @param ibi_col Numeric IBI/RR interval column.
#' @param time_col Optional time column.
#' @param beat_time_col Optional explicit beat-time column.
#' @param group_cols Optional grouping columns.
#' @param ibi_units `"auto"`, `"seconds"`, or `"milliseconds"`.
#'
#' @return A list with `overview`, `beat_table`, `interval_table`,
#'   `process_summary`, and `settings`.
#' @export
model_gazepoint_hr_point_process <- function(dat,
                                             ibi_col = "IBI",
                                             time_col = NULL,
                                             beat_time_col = NULL,
                                             group_cols = NULL,
                                             ibi_units = c("auto", "seconds", "milliseconds")) {
  if (!is.data.frame(dat)) {
    stop("`dat` must be a data frame.", call. = FALSE)
  }

  ibi_units <- match.arg(ibi_units)

  required <- c(ibi_col, time_col, beat_time_col)
  required <- required[!is.null(required)]

  missing_required <- setdiff(required, names(dat))
  if (length(missing_required) > 0) {
    stop("Missing required columns: ", paste(missing_required, collapse = ", "), call. = FALSE)
  }

  if (!is.numeric(dat[[ibi_col]])) {
    stop("`ibi_col` must identify a numeric column.", call. = FALSE)
  }

  if (!is.null(time_col) && !is.numeric(dat[[time_col]])) {
    stop("`time_col` must identify a numeric column.", call. = FALSE)
  }

  if (!is.null(beat_time_col) && !is.numeric(dat[[beat_time_col]])) {
    stop("`beat_time_col` must identify a numeric column.", call. = FALSE)
  }

  if (is.null(group_cols)) {
    group_cols <- character()
  }

  missing_groups <- setdiff(group_cols, names(dat))
  if (length(missing_groups) > 0) {
    stop("Missing `group_cols`: ", paste(missing_groups, collapse = ", "), call. = FALSE)
  }

  groups <- gpbiometrics_pp_split(dat, group_cols)

  beat_rows <- list()
  interval_rows <- list()
  summary_rows <- list()
  beat_id <- 1L
  interval_id <- 1L

  for (group_id in names(groups)) {
    idx <- groups[[group_id]]

    if (!is.null(time_col)) {
      idx <- idx[order(dat[[time_col]][idx])]
    }

    ibi <- dat[[ibi_col]][idx]
    ibi <- ibi[is.finite(ibi) & ibi > 0]
    ibi_s <- gpbiometrics_pp_ibi_to_seconds(ibi, ibi_units)

    if (!is.null(beat_time_col)) {
      beat_time <- dat[[beat_time_col]][idx]
      beat_time <- beat_time[is.finite(beat_time)]
      beat_time <- sort(unique(beat_time))
      intervals <- diff(beat_time)
    } else {
      beat_time <- cumsum(ibi_s)
      intervals <- ibi_s
    }

    for (i in seq_along(beat_time)) {
      beat_rows[[beat_id]] <- data.frame(
        group_id = group_id,
        beat_index = i,
        beat_time = beat_time[i],
        instantaneous_hr_bpm = if (i <= length(ibi_s) && ibi_s[i] > 0) 60 / ibi_s[i] else NA_real_,
        stringsAsFactors = FALSE
      )
      beat_id <- beat_id + 1L
    }

    intervals <- intervals[is.finite(intervals) & intervals > 0]

    for (i in seq_along(intervals)) {
      interval_rows[[interval_id]] <- data.frame(
        group_id = group_id,
        interval_index = i,
        interbeat_interval_s = intervals[i],
        stringsAsFactors = FALSE
      )
      interval_id <- interval_id + 1L
    }

    ig <- gpbiometrics_pp_inverse_gaussian_summary(intervals)

    summary_rows[[group_id]] <- data.frame(
      group_id = group_id,
      n_beats = length(beat_time),
      n_intervals = length(intervals),
      mean_interbeat_interval_s = if (length(intervals) > 0) mean(intervals) else NA_real_,
      mean_hr_bpm = if (length(intervals) > 0) 60 / mean(intervals) else NA_real_,
      inverse_gaussian_mu = ig$mu,
      inverse_gaussian_lambda = ig$lambda,
      interval_cv = ig$cv,
      status = if (length(intervals) >= 5) {
        "hr_point_process_summarised"
      } else {
        "insufficient_heartbeats"
      },
      stringsAsFactors = FALSE
    )
  }

  beat_table <- if (length(beat_rows) > 0) do.call(rbind, beat_rows) else data.frame()
  interval_table <- if (length(interval_rows) > 0) do.call(rbind, interval_rows) else data.frame()
  process_summary <- do.call(rbind, summary_rows)
  rownames(process_summary) <- NULL

  overview <- data.frame(
    group_count = length(groups),
    beat_rows = nrow(beat_table),
    interval_rows = nrow(interval_table),
    successful_groups = sum(process_summary$status == "hr_point_process_summarised"),
    problem_groups = sum(process_summary$status != "hr_point_process_summarised"),
    status = if (all(process_summary$status == "hr_point_process_summarised")) {
      "hr_point_process_complete"
    } else if (any(process_summary$status == "hr_point_process_summarised")) {
      "hr_point_process_partial"
    } else {
      "hr_point_process_insufficient_beats"
    },
    interpretation = paste(
      "HR point-process summaries describe heartbeat timing and interval distributions.",
      "They are not adaptive Bayesian heartbeat filters or clinical diagnoses by themselves."
    ),
    stringsAsFactors = FALSE
  )

  structure(
    list(
      overview = overview,
      beat_table = beat_table,
      interval_table = interval_table,
      process_summary = process_summary,
      settings = list(
        ibi_col = ibi_col,
        time_col = time_col,
        beat_time_col = beat_time_col,
        group_cols = group_cols,
        ibi_units = ibi_units
      )
    ),
    class = c("gazepoint_hr_point_process", "list")
  )
}

gpbiometrics_pp_split <- function(dat, group_cols) {
  if (length(group_cols) == 0) {
    return(list(all_rows = seq_len(nrow(dat))))
  }

  gf <- dat[group_cols]
  gf[] <- lapply(gf, function(x) {
    x <- as.character(x)
    x[is.na(x)] <- "<NA>"
    x
  })

  split(seq_len(nrow(dat)), do.call(paste, c(gf, sep = " | ")))
}

gpbiometrics_eda_pp_events <- function(dat,
                                       idx,
                                       time_col,
                                       eda_col,
                                       event_time_col = NULL,
                                       event_indicator_col = NULL,
                                       derivative_mad_multiplier = 6,
                                       min_event_distance_s = 1) {
  if (!is.null(event_time_col)) {
    return(dat[[event_time_col]][idx])
  }

  time <- dat[[time_col]][idx]

  if (!is.null(event_indicator_col)) {
    indicator <- dat[[event_indicator_col]][idx]
    event_idx <- which(is.finite(indicator) & indicator > 0)
    return(time[event_idx])
  }

  eda <- dat[[eda_col]][idx]
  finite <- is.finite(time) & is.finite(eda)

  time <- time[finite]
  eda <- eda[finite]

  if (length(time) < 5) {
    return(numeric())
  }

  dt <- diff(time)
  dx <- diff(eda)

  derivative <- dx / dt
  derivative[!is.finite(derivative)] <- NA_real_

  center <- stats::median(derivative, na.rm = TRUE)
  mad_value <- stats::mad(derivative, constant = 1, na.rm = TRUE)

  if (!is.finite(mad_value) || mad_value == 0) {
    mad_value <- .Machine$double.eps
  }

  threshold <- center + derivative_mad_multiplier * mad_value
  candidate <- which(is.finite(derivative) & derivative > threshold) + 1L

  if (length(candidate) == 0) {
    return(numeric())
  }

  selected <- candidate[1]

  if (length(candidate) > 1) {
    for (i in candidate[-1]) {
      if ((time[i] - time[selected[length(selected)]]) >= min_event_distance_s) {
        selected <- c(selected, i)
      } else if (eda[i] > eda[selected[length(selected)]]) {
        selected[length(selected)] <- i
      }
    }
  }

  time[selected]
}

gpbiometrics_pp_inverse_gaussian_summary <- function(intervals) {
  intervals <- intervals[is.finite(intervals) & intervals > 0]

  if (length(intervals) < 2) {
    return(list(mu = NA_real_, lambda = NA_real_, cv = NA_real_))
  }

  mu <- mean(intervals)
  variance <- stats::var(intervals)

  lambda <- if (is.finite(variance) && variance > 0) {
    mu^3 / variance
  } else {
    NA_real_
  }

  cv <- if (is.finite(mu) && mu > 0) {
    stats::sd(intervals) / mu
  } else {
    NA_real_
  }

  list(mu = mu, lambda = lambda, cv = cv)
}

gpbiometrics_pp_ibi_to_seconds <- function(ibi, ibi_units = "auto") {
  if (ibi_units == "milliseconds") {
    return(ibi / 1000)
  }

  if (ibi_units == "seconds") {
    return(ibi)
  }

  if (stats::median(ibi, na.rm = TRUE) > 10) {
    ibi / 1000
  } else {
    ibi
  }
}

#' Estimate lag between two Gazepoint biometric signals
#'
#' Estimates the time shift that maximizes the association between two recorded
#' biometric signals within each group. This is a conservative synchronization
#' diagnostic for inspecting whether two recorded traces show similar temporal
#' structure at different shifts. It should not be interpreted as causal timing
#' or true physiological latency unless the design includes appropriate event
#' markers and independently justified signal-processing assumptions.
#'
#' @param data A Gazepoint biometric data frame.
#' @param signal_x_col Name of the first signal column.
#' @param signal_y_col Name of the second signal column.
#' @param time_col Optional time or counter column. If `NULL`, a common
#'   Gazepoint time/counter column is detected.
#' @param group_cols Optional grouping columns, such as participant, stimulus,
#'   trial, or source file.
#' @param max_lag Maximum absolute lag to evaluate, in the same units as
#'   `time_col`.
#' @param lag_step Step size between candidate lags, in the same units as
#'   `time_col`. If `NULL`, the median positive time step is used.
#' @param method Correlation method passed to [stats::cor()].
#' @param min_complete_pairs Minimum complete aligned observations required for
#'   a candidate lag.
#' @param use_first_difference If `TRUE`, correlations are estimated on first
#'   differences rather than raw signal levels.
#'
#' @return A list with `overview`, `lag_by_group`, `lag_profile`, and `settings`.
#' @export
estimate_gazepoint_signal_lag <- function(data,
                                          signal_x_col,
                                          signal_y_col,
                                          time_col = NULL,
                                          group_cols = NULL,
                                          max_lag = 1000,
                                          lag_step = NULL,
                                          method = c("pearson", "spearman"),
                                          min_complete_pairs = 20,
                                          use_first_difference = FALSE) {
  method <- match.arg(method)

  if (!is.data.frame(data)) {
    stop("`data` must be a data frame.", call. = FALSE)
  }

  gpbiometrics_sync_validate_col_name(signal_x_col, "signal_x_col")
  gpbiometrics_sync_validate_col_name(signal_y_col, "signal_y_col")

  if (!is.numeric(max_lag) || length(max_lag) != 1 ||
      is.na(max_lag) || max_lag < 0) {
    stop("`max_lag` must be a single non-negative number.", call. = FALSE)
  }

  if (!is.null(lag_step) &&
      (!is.numeric(lag_step) || length(lag_step) != 1 ||
       is.na(lag_step) || lag_step <= 0)) {
    stop("`lag_step` must be NULL or a single positive number.", call. = FALSE)
  }

  if (!is.numeric(min_complete_pairs) || length(min_complete_pairs) != 1 ||
      is.na(min_complete_pairs) || min_complete_pairs < 2) {
    stop("`min_complete_pairs` must be a single number >= 2.", call. = FALSE)
  }

  if (!is.logical(use_first_difference) ||
      length(use_first_difference) != 1 ||
      is.na(use_first_difference)) {
    stop("`use_first_difference` must be TRUE or FALSE.", call. = FALSE)
  }

  dat <- as.data.frame(data, stringsAsFactors = FALSE)
  dat$.gpbiometrics_row_id <- seq_len(nrow(dat))
  names_dat <- names(dat)

  missing_signals <- setdiff(c(signal_x_col, signal_y_col), names_dat)

  if (length(missing_signals) > 0) {
    stop(
      "Signal columns not found in `data`: ",
      paste(missing_signals, collapse = ", "),
      call. = FALSE
    )
  }

  if (is.null(time_col)) {
    time_col <- gpbiometrics_sync_first_existing(
      names_dat,
      c(
        "time_ms", "timestamp_ms", "timestamp",
        "TIME", "Time", "time",
        "CNT", "cnt"
      )
    )
  }

  if (is.null(time_col) || !time_col %in% names_dat) {
    stop("No usable time column was found. Supply `time_col`.", call. = FALSE)
  }

  group_cols <- gpbiometrics_sync_resolve_group_cols(names_dat, group_cols)

  missing_group_cols <- setdiff(group_cols, names_dat)

  if (length(missing_group_cols) > 0) {
    stop(
      "`group_cols` not found in `data`: ",
      paste(missing_group_cols, collapse = ", "),
      call. = FALSE
    )
  }

  if (nrow(dat) == 0) {
    lag_by_group <- gpbiometrics_sync_empty_lag_summary(group_cols)
    lag_profile <- gpbiometrics_sync_empty_lag_profile(group_cols)

    return(structure(
      list(
        overview = data.frame(
          input_rows = 0L,
          group_count = 0L,
          candidate_lag_count = 0L,
          estimated_group_count = 0L,
          status = "empty_input",
          interpretation = paste(
            "No rows were available for lag estimation.",
            "No synchronization inference can be made."
          ),
          stringsAsFactors = FALSE
        ),
        lag_by_group = lag_by_group,
        lag_profile = lag_profile,
        settings = list(
          signal_x_col = signal_x_col,
          signal_y_col = signal_y_col,
          time_col = time_col,
          group_cols = group_cols,
          max_lag = max_lag,
          lag_step = lag_step,
          method = method,
          min_complete_pairs = min_complete_pairs,
          use_first_difference = use_first_difference
        )
      ),
      class = c("gazepoint_signal_lag", "list")
    ))
  }

  time_value <- suppressWarnings(as.numeric(dat[[time_col]]))

  if (is.null(lag_step)) {
    lag_step <- gpbiometrics_sync_infer_lag_step(time_value)
  }

  lag_candidates <- gpbiometrics_sync_lag_candidates(max_lag, lag_step)

  dat$.gpbiometrics_group_id <- gpbiometrics_sync_group_id(dat, group_cols)
  group_indices <- split(seq_len(nrow(dat)), dat$.gpbiometrics_group_id, drop = TRUE)

  lag_summary_parts <- list()
  lag_profile_parts <- list()

  for (group_id in names(group_indices)) {
    idx <- group_indices[[group_id]]
    group_values <- gpbiometrics_sync_group_values(dat, group_cols, idx)

    series <- gpbiometrics_sync_prepare_series(
      time = suppressWarnings(as.numeric(dat[[time_col]][idx])),
      signal_x = suppressWarnings(as.numeric(dat[[signal_x_col]][idx])),
      signal_y = suppressWarnings(as.numeric(dat[[signal_y_col]][idx])),
      use_first_difference = use_first_difference
    )

    if (nrow(series) < min_complete_pairs) {
      summary_row <- data.frame(
        group_id = group_id,
        signal_x = signal_x_col,
        signal_y = signal_y_col,
        estimated_lag = NA_real_,
        selected_correlation = NA_real_,
        abs_selected_correlation = NA_real_,
        n_complete_pairs = 0L,
        candidate_count = length(lag_candidates),
        status = "insufficient_data",
        interpretation = paste(
          "Too few usable observations were available.",
          "No lag estimate should be interpreted for this group."
        ),
        stringsAsFactors = FALSE
      )

      lag_summary_parts[[length(lag_summary_parts) + 1L]] <-
        gpbiometrics_sync_prepend_group_values(group_values, summary_row)

      next
    }

    profile_rows <- lapply(lag_candidates, function(lag) {
      y_shifted <- stats::approx(
        x = series$time,
        y = series$signal_y,
        xout = series$time + lag,
        rule = 1,
        ties = mean
      )$y

      complete <- is.finite(series$signal_x) & is.finite(y_shifted)
      n_complete <- sum(complete)

      correlation <- NA_real_
      status <- "insufficient_complete_pairs"

      if (n_complete >= min_complete_pairs) {
        x_complete <- series$signal_x[complete]
        y_complete <- y_shifted[complete]

        if (length(unique(x_complete)) < 2 ||
            length(unique(y_complete)) < 2) {
          status <- "flat_signal"
        } else {
          correlation <- suppressWarnings(stats::cor(
            x_complete,
            y_complete,
            method = method,
            use = "complete.obs"
          ))

          if (is.finite(correlation)) {
            status <- "estimated"
          } else {
            status <- "correlation_unavailable"
          }
        }
      }

      row <- data.frame(
        group_id = group_id,
        signal_x = signal_x_col,
        signal_y = signal_y_col,
        lag = lag,
        n_complete_pairs = as.integer(n_complete),
        correlation = correlation,
        abs_correlation = abs(correlation),
        status = status,
        stringsAsFactors = FALSE
      )

      gpbiometrics_sync_prepend_group_values(group_values, row)
    })

    profile <- gpbiometrics_sync_rbind(profile_rows)
    lag_profile_parts[[length(lag_profile_parts) + 1L]] <- profile

    valid_profile <- profile[
      profile$status == "estimated" & is.finite(profile$correlation),
      ,
      drop = FALSE
    ]

    if (nrow(valid_profile) == 0) {
      summary_row <- data.frame(
        group_id = group_id,
        signal_x = signal_x_col,
        signal_y = signal_y_col,
        estimated_lag = NA_real_,
        selected_correlation = NA_real_,
        abs_selected_correlation = NA_real_,
        n_complete_pairs = max(profile$n_complete_pairs, na.rm = TRUE),
        candidate_count = nrow(profile),
        status = "no_valid_lag_profile",
        interpretation = paste(
          "Candidate lags were evaluated, but no stable finite correlation was available.",
          "No lag estimate should be interpreted for this group."
        ),
        stringsAsFactors = FALSE
      )
    } else {
      best <- valid_profile[which.max(abs(valid_profile$correlation)), , drop = FALSE]

      summary_row <- data.frame(
        group_id = group_id,
        signal_x = signal_x_col,
        signal_y = signal_y_col,
        estimated_lag = best$lag[1],
        selected_correlation = best$correlation[1],
        abs_selected_correlation = abs(best$correlation[1]),
        n_complete_pairs = as.integer(best$n_complete_pairs[1]),
        candidate_count = nrow(profile),
        status = "estimated",
        interpretation = paste(
          "Estimated lag maximizes absolute signal association within the tested range.",
          "Use as a synchronization QC diagnostic only, not as evidence of causal timing or physiological latency."
        ),
        stringsAsFactors = FALSE
      )
    }

    lag_summary_parts[[length(lag_summary_parts) + 1L]] <-
      gpbiometrics_sync_prepend_group_values(group_values, summary_row)
  }

  lag_by_group <- gpbiometrics_sync_rbind(lag_summary_parts)
  lag_profile <- gpbiometrics_sync_rbind(lag_profile_parts)

  estimated_group_count <- sum(lag_by_group$status == "estimated", na.rm = TRUE)

  overview_status <- if (estimated_group_count > 0) {
    "estimated"
  } else {
    "no_valid_estimates"
  }

  overview <- data.frame(
    input_rows = nrow(dat),
    group_count = length(group_indices),
    candidate_lag_count = length(lag_candidates),
    estimated_group_count = as.integer(estimated_group_count),
    status = overview_status,
    interpretation = paste(
      "Lag estimates are conservative signal-alignment diagnostics.",
      "They do not establish true physiological latency, causality, emotion, cognition, trust, or preference."
    ),
    stringsAsFactors = FALSE
  )

  structure(
    list(
      overview = overview,
      lag_by_group = lag_by_group,
      lag_profile = lag_profile,
      settings = list(
        signal_x_col = signal_x_col,
        signal_y_col = signal_y_col,
        time_col = time_col,
        group_cols = group_cols,
        max_lag = max_lag,
        lag_step = lag_step,
        method = method,
        min_complete_pairs = min_complete_pairs,
        use_first_difference = use_first_difference
      )
    ),
    class = c("gazepoint_signal_lag", "list")
  )
}

#' Audit Gazepoint biometric synchronization drift
#'
#' Combines time-order/reset diagnostics with conservative signal-lag summaries
#' across signal pairs and groups. The helper is intended for quality control and
#' synchronization review. It does not infer emotional valence, cognitive states,
#' causal timing, or true physiological latency.
#'
#' @param data A Gazepoint biometric data frame.
#' @param time_col Optional time or counter column.
#' @param group_cols Optional grouping columns.
#' @param signal_pairs Optional two-column data frame, matrix, or list defining
#'   signal pairs. If `NULL`, pairs are formed between a reference signal and
#'   other detected biometric signals.
#' @param signal_cols Optional candidate signal columns used when `signal_pairs`
#'   is `NULL`.
#' @param reference_signal_col Optional reference signal used when `signal_pairs`
#'   is `NULL`.
#' @param max_lag Maximum absolute lag to evaluate, in the same units as
#'   `time_col`.
#' @param lag_step Step size between candidate lags. If `NULL`, the median
#'   positive time step is used.
#' @param drift_tolerance Optional threshold for the range of estimated lags
#'   across groups. If `NULL`, drift is summarized but not threshold-classified.
#' @param method Correlation method passed to [stats::cor()].
#' @param min_complete_pairs Minimum complete aligned observations required for
#'   each candidate lag.
#' @param use_first_difference If `TRUE`, lag diagnostics use first differences.
#' @param include_reset_segments If `TRUE`, reset segments from
#'   [audit_gazepoint_time_resets()] are added to grouping when available.
#'
#' @return A list with `overview`, `checks`, `time_reset_audit`,
#'   `lag_by_group`, `lag_profile`, `drift_summary`, and `settings`.
#' @export
audit_gazepoint_biometric_sync_drift <- function(data,
                                                 time_col = NULL,
                                                 group_cols = NULL,
                                                 signal_pairs = NULL,
                                                 signal_cols = NULL,
                                                 reference_signal_col = NULL,
                                                 max_lag = 1000,
                                                 lag_step = NULL,
                                                 drift_tolerance = NULL,
                                                 method = c("pearson", "spearman"),
                                                 min_complete_pairs = 20,
                                                 use_first_difference = FALSE,
                                                 include_reset_segments = TRUE) {
  method <- match.arg(method)

  if (!is.data.frame(data)) {
    stop("`data` must be a data frame.", call. = FALSE)
  }

  if (!is.null(drift_tolerance) &&
      (!is.numeric(drift_tolerance) || length(drift_tolerance) != 1 ||
       is.na(drift_tolerance) || drift_tolerance < 0)) {
    stop("`drift_tolerance` must be NULL or a single non-negative number.", call. = FALSE)
  }

  if (!is.logical(include_reset_segments) ||
      length(include_reset_segments) != 1 ||
      is.na(include_reset_segments)) {
    stop("`include_reset_segments` must be TRUE or FALSE.", call. = FALSE)
  }

  dat <- as.data.frame(data, stringsAsFactors = FALSE)
  names_dat <- names(dat)

  if (is.null(time_col)) {
    time_col <- gpbiometrics_sync_first_existing(
      names_dat,
      c(
        "time_ms", "timestamp_ms", "timestamp",
        "TIME", "Time", "time",
        "CNT", "cnt"
      )
    )
  }

  if (is.null(time_col) || !time_col %in% names_dat) {
    stop("No usable time column was found. Supply `time_col`.", call. = FALSE)
  }

  group_cols <- gpbiometrics_sync_resolve_group_cols(names_dat, group_cols)

  missing_group_cols <- setdiff(group_cols, names_dat)

  if (length(missing_group_cols) > 0) {
    stop(
      "`group_cols` not found in `data`: ",
      paste(missing_group_cols, collapse = ", "),
      call. = FALSE
    )
  }

  time_reset_audit <- tryCatch(
    audit_gazepoint_time_resets(
      dat,
      time_col = time_col,
      group_cols = group_cols,
      allow_ties = TRUE,
      split_on_negative_step = TRUE,
      return_reindexed_time = FALSE,
      min_segment_rows = 1
    ),
    error = function(e) {
      structure(
        list(
          overview = data.frame(
            status = "time_reset_audit_failed",
            detail = conditionMessage(e),
            stringsAsFactors = FALSE
          ),
          row_flags = data.frame(stringsAsFactors = FALSE),
          data_with_segments = dat
        ),
        class = c("gazepoint_time_reset_audit_failed", "list")
      )
    }
  )

  time_reset_issue_count <- gpbiometrics_sync_time_reset_issue_count(time_reset_audit)

  lag_data <- dat
  lag_group_cols <- group_cols

  if (isTRUE(include_reset_segments) &&
      is.list(time_reset_audit) &&
      is.data.frame(time_reset_audit$data_with_segments) &&
      "reset_segment_index" %in% names(time_reset_audit$data_with_segments)) {
    lag_data <- as.data.frame(time_reset_audit$data_with_segments, stringsAsFactors = FALSE)
    lag_group_cols <- unique(c(group_cols, "reset_segment_index"))
  }

  pair_table <- gpbiometrics_sync_resolve_signal_pairs(
    names_dat = names(lag_data),
    signal_pairs = signal_pairs,
    signal_cols = signal_cols,
    reference_signal_col = reference_signal_col
  )

  if (nrow(pair_table) == 0) {
    empty_lag_by_group <- gpbiometrics_sync_empty_lag_summary(lag_group_cols)
    empty_lag_profile <- gpbiometrics_sync_empty_lag_profile(lag_group_cols)
    empty_drift <- gpbiometrics_sync_empty_drift_summary()

    checks <- data.frame(
      check = c("time_order", "lag_estimation", "drift_variability"),
      status = c(
        if (is.na(time_reset_issue_count)) "not_available" else if (time_reset_issue_count > 0) "warn" else "pass",
        "warn",
        "not_available"
      ),
      detail = c(
        gpbiometrics_sync_time_reset_detail(time_reset_issue_count),
        "No usable signal pairs were available for lag estimation.",
        "Drift variability was not evaluated because no signal pairs were available."
      ),
      stringsAsFactors = FALSE
    )

    return(structure(
      list(
        overview = data.frame(
          input_rows = nrow(dat),
          signal_pair_count = 0L,
          lag_estimate_rows = 0L,
          drift_rows = 0L,
          time_reset_issue_count = time_reset_issue_count,
          drift_issue_count = 0L,
          status = "no_signal_pairs",
          interpretation = paste(
            "No signal-pair synchronization diagnostic was computed.",
            "No synchronization or drift inference can be made."
          ),
          stringsAsFactors = FALSE
        ),
        checks = checks,
        time_reset_audit = time_reset_audit,
        lag_by_group = empty_lag_by_group,
        lag_profile = empty_lag_profile,
        drift_summary = empty_drift,
        settings = list(
          time_col = time_col,
          group_cols = group_cols,
          lag_group_cols = lag_group_cols,
          signal_pairs = pair_table,
          max_lag = max_lag,
          lag_step = lag_step,
          drift_tolerance = drift_tolerance,
          method = method,
          min_complete_pairs = min_complete_pairs,
          use_first_difference = use_first_difference,
          include_reset_segments = include_reset_segments
        )
      ),
      class = c("gazepoint_biometric_sync_drift_audit", "list")
    ))
  }

  lag_results <- lapply(seq_len(nrow(pair_table)), function(i) {
    pair <- pair_table[i, , drop = FALSE]

    out <- estimate_gazepoint_signal_lag(
      data = lag_data,
      signal_x_col = pair$signal_x,
      signal_y_col = pair$signal_y,
      time_col = time_col,
      group_cols = lag_group_cols,
      max_lag = max_lag,
      lag_step = lag_step,
      method = method,
      min_complete_pairs = min_complete_pairs,
      use_first_difference = use_first_difference
    )

    pair_id <- paste(pair$signal_x, pair$signal_y, sep = "__vs__")

    out$lag_by_group$pair_id <- pair_id
    out$lag_profile$pair_id <- pair_id

    out$lag_by_group$signal_x <- pair$signal_x
    out$lag_by_group$signal_y <- pair$signal_y
    out$lag_profile$signal_x <- pair$signal_x
    out$lag_profile$signal_y <- pair$signal_y

    out
  })

  lag_by_group <- gpbiometrics_sync_rbind(lapply(lag_results, `[[`, "lag_by_group"))
  lag_profile <- gpbiometrics_sync_rbind(lapply(lag_results, `[[`, "lag_profile"))

  drift_summary <- gpbiometrics_sync_make_drift_summary(
    lag_by_group = lag_by_group,
    pair_table = pair_table,
    drift_tolerance = drift_tolerance
  )

  lag_estimate_rows <- sum(lag_by_group$status == "estimated", na.rm = TRUE)
  drift_issue_count <- sum(drift_summary$status == "drift_exceeds_tolerance", na.rm = TRUE)

  checks <- data.frame(
    check = c("time_order", "lag_estimation", "drift_variability"),
    status = c(
      if (is.na(time_reset_issue_count)) {
        "not_available"
      } else if (time_reset_issue_count > 0) {
        "warn"
      } else {
        "pass"
      },
      if (lag_estimate_rows > 0) "pass" else "warn",
      if (is.null(drift_tolerance)) {
        "info"
      } else if (drift_issue_count > 0) {
        "warn"
      } else {
        "pass"
      }
    ),
    detail = c(
      gpbiometrics_sync_time_reset_detail(time_reset_issue_count),
      if (lag_estimate_rows > 0) {
        paste(lag_estimate_rows, "group-level lag estimate(s) were available.")
      } else {
        "No valid group-level lag estimates were available."
      },
      if (is.null(drift_tolerance)) {
        "Drift range was summarized without threshold classification."
      } else if (drift_issue_count > 0) {
        paste(drift_issue_count, "signal pair(s) exceeded the drift tolerance.")
      } else {
        "No signal pair exceeded the supplied drift tolerance."
      }
    ),
    stringsAsFactors = FALSE
  )

  overall_status <- if (lag_estimate_rows == 0) {
    "insufficient_lag_evidence"
  } else if ((!is.na(time_reset_issue_count) && time_reset_issue_count > 0) ||
             drift_issue_count > 0) {
    "review_recommended"
  } else {
    "diagnostic_complete"
  }

  overview <- data.frame(
    input_rows = nrow(dat),
    signal_pair_count = nrow(pair_table),
    lag_estimate_rows = as.integer(lag_estimate_rows),
    drift_rows = nrow(drift_summary),
    time_reset_issue_count = time_reset_issue_count,
    drift_issue_count = as.integer(drift_issue_count),
    status = overall_status,
    interpretation = paste(
      "Synchronization drift output is a conservative QC diagnostic.",
      "It does not establish true physiological latency, causal timing, emotional valence, cognition, trust, preference, or evaluation."
    ),
    stringsAsFactors = FALSE
  )

  structure(
    list(
      overview = overview,
      checks = checks,
      time_reset_audit = time_reset_audit,
      lag_by_group = lag_by_group,
      lag_profile = lag_profile,
      drift_summary = drift_summary,
      settings = list(
        time_col = time_col,
        group_cols = group_cols,
        lag_group_cols = lag_group_cols,
        signal_pairs = pair_table,
        max_lag = max_lag,
        lag_step = lag_step,
        drift_tolerance = drift_tolerance,
        method = method,
        min_complete_pairs = min_complete_pairs,
        use_first_difference = use_first_difference,
        include_reset_segments = include_reset_segments
      )
    ),
    class = c("gazepoint_biometric_sync_drift_audit", "list")
  )
}

gpbiometrics_sync_validate_col_name <- function(x, arg) {
  if (!is.character(x) || length(x) != 1 || is.na(x) || !nzchar(x)) {
    stop("`", arg, "` must be a single non-empty character value.", call. = FALSE)
  }

  invisible(TRUE)
}

gpbiometrics_sync_first_existing <- function(names_dat, candidates) {
  exact <- candidates[candidates %in% names_dat]

  if (length(exact) > 0) {
    return(exact[1])
  }

  lower_names <- tolower(names_dat)
  lower_candidates <- tolower(candidates)
  idx <- match(lower_candidates, lower_names)
  idx <- idx[!is.na(idx)]

  if (length(idx) > 0) {
    return(names_dat[idx[1]])
  }

  NULL
}

gpbiometrics_sync_resolve_group_cols <- function(names_dat, group_cols) {
  if (!is.null(group_cols)) {
    group_cols <- as.character(group_cols)
    group_cols <- group_cols[!is.na(group_cols) & nzchar(group_cols)]
    return(unique(group_cols))
  }

  candidates <- c(
    "source_file",
    "source_participant",
    "participant",
    "subject",
    "subject_id",
    "USER",
    "USER_FILE",
    "MEDIA_ID",
    "MEDIA_NAME",
    "stimulus",
    "stimulus_id",
    "trial",
    "trial_id",
    "trial_global"
  )

  unique(candidates[candidates %in% names_dat])
}

gpbiometrics_sync_group_id <- function(dat, group_cols) {
  if (length(group_cols) == 0) {
    return(rep("all", nrow(dat)))
  }

  group_dat <- dat[group_cols]

  group_dat[] <- lapply(group_dat, function(x) {
    x_chr <- as.character(x)
    x_chr[is.na(x_chr)] <- "<NA>"
    x_chr
  })

  apply(group_dat, 1, paste, collapse = "||")
}

gpbiometrics_sync_group_values <- function(dat, group_cols, idx) {
  if (length(group_cols) == 0) {
    return(NULL)
  }

  out <- as.data.frame(dat[idx[1], group_cols, drop = FALSE], stringsAsFactors = FALSE)
  rownames(out) <- NULL
  out
}

gpbiometrics_sync_prepend_group_values <- function(group_values, row) {
  row <- as.data.frame(row, stringsAsFactors = FALSE)
  rownames(row) <- NULL

  if (is.null(group_values) || ncol(group_values) == 0) {
    return(row)
  }

  group_values <- as.data.frame(group_values, stringsAsFactors = FALSE)
  rownames(group_values) <- NULL

  cbind(group_values, row, stringsAsFactors = FALSE)
}

gpbiometrics_sync_rbind <- function(x) {
  x <- x[!vapply(x, is.null, logical(1))]

  if (length(x) == 0) {
    return(data.frame(stringsAsFactors = FALSE))
  }

  out <- do.call(rbind, x)
  rownames(out) <- NULL
  out
}

gpbiometrics_sync_infer_lag_step <- function(time_value) {
  time_value <- suppressWarnings(as.numeric(time_value))
  time_value <- time_value[is.finite(time_value)]

  if (length(time_value) < 2) {
    return(1)
  }

  time_value <- sort(unique(time_value))
  delta <- diff(time_value)
  delta <- delta[is.finite(delta) & delta > 0]

  if (length(delta) == 0) {
    return(1)
  }

  stats::median(delta)
}

gpbiometrics_sync_lag_candidates <- function(max_lag, lag_step) {
  if (max_lag == 0) {
    return(0)
  }

  out <- seq(-max_lag, max_lag, by = lag_step)

  if (!any(abs(out) < sqrt(.Machine$double.eps))) {
    out <- c(out, 0)
  }

  sort(unique(out))
}

gpbiometrics_sync_prepare_series <- function(time,
                                             signal_x,
                                             signal_y,
                                             use_first_difference = FALSE) {
  time <- suppressWarnings(as.numeric(time))
  signal_x <- suppressWarnings(as.numeric(signal_x))
  signal_y <- suppressWarnings(as.numeric(signal_y))

  ok_time <- is.finite(time)

  time <- time[ok_time]
  signal_x <- signal_x[ok_time]
  signal_y <- signal_y[ok_time]

  if (length(time) == 0) {
    return(data.frame(
      time = numeric(),
      signal_x = numeric(),
      signal_y = numeric(),
      stringsAsFactors = FALSE
    ))
  }

  ord <- order(time)
  time <- time[ord]
  signal_x <- signal_x[ord]
  signal_y <- signal_y[ord]

  unique_time <- sort(unique(time))

  x_mean <- vapply(unique_time, function(tt) {
    value <- signal_x[time == tt]
    value <- value[is.finite(value)]

    if (length(value) == 0) {
      NA_real_
    } else {
      mean(value)
    }
  }, numeric(1))

  y_mean <- vapply(unique_time, function(tt) {
    value <- signal_y[time == tt]
    value <- value[is.finite(value)]

    if (length(value) == 0) {
      NA_real_
    } else {
      mean(value)
    }
  }, numeric(1))

  out <- data.frame(
    time = unique_time,
    signal_x = x_mean,
    signal_y = y_mean,
    stringsAsFactors = FALSE
  )

  if (isTRUE(use_first_difference)) {
    if (nrow(out) < 3) {
      return(out[0, , drop = FALSE])
    }

    out <- data.frame(
      time = out$time[-1],
      signal_x = diff(out$signal_x),
      signal_y = diff(out$signal_y),
      stringsAsFactors = FALSE
    )
  }

  out
}

gpbiometrics_sync_empty_lag_summary <- function(group_cols = character()) {
  out <- data.frame(
    group_id = character(),
    signal_x = character(),
    signal_y = character(),
    estimated_lag = numeric(),
    selected_correlation = numeric(),
    abs_selected_correlation = numeric(),
    n_complete_pairs = integer(),
    candidate_count = integer(),
    status = character(),
    interpretation = character(),
    stringsAsFactors = FALSE
  )

  if (length(group_cols) == 0) {
    return(out)
  }

  group_data <- as.data.frame(
    stats::setNames(
      rep(list(character()), length(group_cols)),
      group_cols
    ),
    stringsAsFactors = FALSE
  )

  cbind(group_data, out, stringsAsFactors = FALSE)
}

gpbiometrics_sync_empty_lag_profile <- function(group_cols = character()) {
  out <- data.frame(
    group_id = character(),
    signal_x = character(),
    signal_y = character(),
    lag = numeric(),
    n_complete_pairs = integer(),
    correlation = numeric(),
    abs_correlation = numeric(),
    status = character(),
    stringsAsFactors = FALSE
  )

  if (length(group_cols) == 0) {
    return(out)
  }

  group_data <- as.data.frame(
    stats::setNames(
      rep(list(character()), length(group_cols)),
      group_cols
    ),
    stringsAsFactors = FALSE
  )

  cbind(group_data, out, stringsAsFactors = FALSE)
}

gpbiometrics_sync_infer_signal_cols <- function(names_dat) {
  candidates <- c(
    "GSR_US",
    "GSR_US_PHASIC",
    "GSR_US_TONIC",
    "GSR",
    "EDA",
    "eda",
    "eda_clean",
    "eda_phasic",
    "HR",
    "hr",
    "HRP",
    "hrp",
    "IBI",
    "ibi",
    "IBI_clean_ms",
    "DIAL",
    "dial"
  )

  out <- character()

  for (candidate in candidates) {
    idx <- match(tolower(candidate), tolower(names_dat))

    if (!is.na(idx)) {
      out <- c(out, names_dat[idx])
    }
  }

  unique(out)
}

gpbiometrics_sync_resolve_signal_pairs <- function(names_dat,
                                                   signal_pairs = NULL,
                                                   signal_cols = NULL,
                                                   reference_signal_col = NULL) {
  explicit_pairs <- !is.null(signal_pairs)

  if (!is.null(signal_pairs)) {
    if (is.data.frame(signal_pairs)) {
      if (all(c("signal_x", "signal_y") %in% names(signal_pairs))) {
        out <- data.frame(
          signal_x = as.character(signal_pairs$signal_x),
          signal_y = as.character(signal_pairs$signal_y),
          stringsAsFactors = FALSE
        )
      } else if (ncol(signal_pairs) >= 2) {
        out <- data.frame(
          signal_x = as.character(signal_pairs[[1]]),
          signal_y = as.character(signal_pairs[[2]]),
          stringsAsFactors = FALSE
        )
      } else {
        stop("`signal_pairs` must have at least two columns.", call. = FALSE)
      }
    } else if (is.matrix(signal_pairs) && ncol(signal_pairs) >= 2) {
      out <- data.frame(
        signal_x = as.character(signal_pairs[, 1]),
        signal_y = as.character(signal_pairs[, 2]),
        stringsAsFactors = FALSE
      )
    } else if (is.list(signal_pairs)) {
      pair_list <- lapply(signal_pairs, function(x) {
        x <- unlist(x, use.names = FALSE)

        if (length(x) < 2) {
          stop("Each element of `signal_pairs` must contain at least two signal names.", call. = FALSE)
        }

        x[1:2]
      })

      pair_matrix <- do.call(rbind, pair_list)

      out <- data.frame(
        signal_x = as.character(pair_matrix[, 1]),
        signal_y = as.character(pair_matrix[, 2]),
        stringsAsFactors = FALSE
      )
    } else {
      stop(
        "`signal_pairs` must be NULL, a two-column data frame, a matrix, or a list of pairs.",
        call. = FALSE
      )
    }
  } else {
    if (is.null(signal_cols)) {
      signal_cols <- gpbiometrics_sync_infer_signal_cols(names_dat)
    }

    signal_cols <- unique(as.character(signal_cols))
    signal_cols <- signal_cols[signal_cols %in% names_dat]

    if (length(signal_cols) < 2) {
      return(data.frame(
        signal_x = character(),
        signal_y = character(),
        stringsAsFactors = FALSE
      ))
    }

    if (is.null(reference_signal_col)) {
      preferred <- c("GSR_US", "GSR_US_PHASIC", "GSR", "EDA", "HR", "HRP", "DIAL")
      reference_signal_col <- NULL

      for (candidate in preferred) {
        idx <- match(tolower(candidate), tolower(signal_cols))

        if (!is.na(idx)) {
          reference_signal_col <- signal_cols[idx]
          break
        }
      }

      if (is.null(reference_signal_col)) {
        reference_signal_col <- signal_cols[1]
      }
    }

    if (!reference_signal_col %in% signal_cols) {
      stop("`reference_signal_col` must be included in `signal_cols`.", call. = FALSE)
    }

    other_signals <- setdiff(signal_cols, reference_signal_col)

    out <- data.frame(
      signal_x = reference_signal_col,
      signal_y = other_signals,
      stringsAsFactors = FALSE
    )
  }

  out$signal_x <- as.character(out$signal_x)
  out$signal_y <- as.character(out$signal_y)

  out <- out[
    !is.na(out$signal_x) &
      !is.na(out$signal_y) &
      nzchar(out$signal_x) &
      nzchar(out$signal_y) &
      out$signal_x != out$signal_y,
    ,
    drop = FALSE
  ]

  out <- unique(out)

  missing_cols <- setdiff(unique(c(out$signal_x, out$signal_y)), names_dat)

  if (length(missing_cols) > 0 && isTRUE(explicit_pairs)) {
    stop(
      "Signal columns in `signal_pairs` were not found in `data`: ",
      paste(missing_cols, collapse = ", "),
      call. = FALSE
    )
  }

  out <- out[out$signal_x %in% names_dat & out$signal_y %in% names_dat, , drop = FALSE]
  rownames(out) <- NULL

  out
}

gpbiometrics_sync_time_reset_issue_count <- function(time_reset_audit) {
  if (!is.list(time_reset_audit) || !is.data.frame(time_reset_audit$row_flags)) {
    return(NA_integer_)
  }

  row_flags <- time_reset_audit$row_flags

  issue_cols <- intersect(
    c("flag_nonfinite_time", "flag_negative_step", "flag_nonmonotonic"),
    names(row_flags)
  )

  if (length(issue_cols) == 0) {
    return(NA_integer_)
  }

  issue_matrix <- row_flags[issue_cols]
  issue_matrix[] <- lapply(issue_matrix, function(x) {
    x <- as.logical(x)
    x[is.na(x)] <- FALSE
    x
  })

  as.integer(sum(rowSums(issue_matrix) > 0, na.rm = TRUE))
}

gpbiometrics_sync_time_reset_detail <- function(time_reset_issue_count) {
  if (is.na(time_reset_issue_count)) {
    return("Time reset audit was not available.")
  }

  if (time_reset_issue_count > 0) {
    return(paste(
      time_reset_issue_count,
      "row(s) had nonfinite, negative-step, or nonmonotonic time flags."
    ))
  }

  "No nonfinite, negative-step, or nonmonotonic time flags were detected."
}

gpbiometrics_sync_make_drift_summary <- function(lag_by_group,
                                                 pair_table,
                                                 drift_tolerance = NULL) {
  if (!is.data.frame(lag_by_group) || nrow(pair_table) == 0) {
    return(gpbiometrics_sync_empty_drift_summary())
  }

  out <- lapply(seq_len(nrow(pair_table)), function(i) {
    signal_x <- pair_table$signal_x[i]
    signal_y <- pair_table$signal_y[i]
    pair_id <- paste(signal_x, signal_y, sep = "__vs__")

    d <- lag_by_group[
      lag_by_group$pair_id == pair_id &
        lag_by_group$status == "estimated" &
        is.finite(lag_by_group$estimated_lag),
      ,
      drop = FALSE
    ]

    n_estimates <- nrow(d)

    if (n_estimates == 0) {
      lag_min <- NA_real_
      lag_max <- NA_real_
      lag_range <- NA_real_
      lag_median <- NA_real_
      lag_sd <- NA_real_
      max_abs_lag <- NA_real_
      status <- "no_estimates"
      interpretation <- paste(
        "No valid lag estimates were available for this signal pair.",
        "No drift interpretation should be made."
      )
    } else {
      lag_min <- min(d$estimated_lag, na.rm = TRUE)
      lag_max <- max(d$estimated_lag, na.rm = TRUE)
      lag_range <- lag_max - lag_min
      lag_median <- stats::median(d$estimated_lag, na.rm = TRUE)
      lag_sd <- if (n_estimates > 1) stats::sd(d$estimated_lag, na.rm = TRUE) else NA_real_
      max_abs_lag <- max(abs(d$estimated_lag), na.rm = TRUE)

      if (n_estimates == 1) {
        status <- "single_estimate"
        interpretation <- paste(
          "Only one group-level lag estimate was available.",
          "Between-group drift variability cannot be assessed."
        )
      } else if (is.null(drift_tolerance)) {
        status <- "estimated_no_tolerance"
        interpretation <- paste(
          "Lag variability was summarized without threshold classification.",
          "Use as a descriptive synchronization QC diagnostic only."
        )
      } else if (lag_range > drift_tolerance) {
        status <- "drift_exceeds_tolerance"
        interpretation <- paste(
          "Estimated lag range exceeded the supplied tolerance.",
          "Manual synchronization review is recommended."
        )
      } else {
        status <- "drift_within_tolerance"
        interpretation <- paste(
          "Estimated lag range was within the supplied tolerance.",
          "This supports QC consistency only and does not prove true physiological synchronization."
        )
      }
    }

    data.frame(
      pair_id = pair_id,
      signal_x = signal_x,
      signal_y = signal_y,
      n_estimates = as.integer(n_estimates),
      lag_min = lag_min,
      lag_median = lag_median,
      lag_max = lag_max,
      lag_range = lag_range,
      lag_sd = lag_sd,
      max_abs_lag = max_abs_lag,
      drift_tolerance = if (is.null(drift_tolerance)) NA_real_ else drift_tolerance,
      status = status,
      interpretation = interpretation,
      stringsAsFactors = FALSE
    )
  })

  gpbiometrics_sync_rbind(out)
}

gpbiometrics_sync_empty_drift_summary <- function() {
  data.frame(
    pair_id = character(),
    signal_x = character(),
    signal_y = character(),
    n_estimates = integer(),
    lag_min = numeric(),
    lag_median = numeric(),
    lag_max = numeric(),
    lag_range = numeric(),
    lag_sd = numeric(),
    max_abs_lag = numeric(),
    drift_tolerance = numeric(),
    status = character(),
    interpretation = character(),
    stringsAsFactors = FALSE
  )
}

test_that("estimate_gazepoint_signal_lag detects a simple delayed signal", {
  n <- 120
  time_ms <- seq_len(n)

  x <- rep(0, n)
  y <- rep(0, n)

  x[30:50] <- 1
  y[35:55] <- 1

  dat <- data.frame(
    participant = "p1",
    time_ms = time_ms,
    x = x,
    y = y
  )

  out <- estimate_gazepoint_signal_lag(
    dat,
    signal_x_col = "x",
    signal_y_col = "y",
    time_col = "time_ms",
    group_cols = "participant",
    max_lag = 10,
    lag_step = 1,
    min_complete_pairs = 20
  )

  expect_s3_class(out, "gazepoint_signal_lag")
  expect_equal(out$overview$status, "estimated")
  expect_equal(nrow(out$lag_by_group), 1)
  expect_equal(out$lag_by_group$status, "estimated")
  expect_equal(out$lag_by_group$estimated_lag, 5)
  expect_true(nrow(out$lag_profile) > 0)
  expect_true(all(abs(out$lag_profile$lag) <= 10))
})

test_that("estimate_gazepoint_signal_lag handles insufficient data conservatively", {
  dat <- data.frame(
    time_ms = 1:5,
    x = c(1, 2, 3, 4, 5),
    y = c(1, 2, 3, 4, 5)
  )

  out <- estimate_gazepoint_signal_lag(
    dat,
    signal_x_col = "x",
    signal_y_col = "y",
    time_col = "time_ms",
    max_lag = 2,
    lag_step = 1,
    min_complete_pairs = 20
  )

  expect_s3_class(out, "gazepoint_signal_lag")
  expect_equal(out$overview$status, "no_valid_estimates")
  expect_equal(out$lag_by_group$status, "insufficient_data")
  expect_true(is.na(out$lag_by_group$estimated_lag))
})

test_that("audit_gazepoint_biometric_sync_drift summarizes lag variability", {
  make_group <- function(participant, delay) {
    n <- 140
    x <- rep(0, n)
    y <- rep(0, n)

    x[40:60] <- 1
    y[(40 + delay):(60 + delay)] <- 1

    data.frame(
      participant = participant,
      time_ms = seq_len(n),
      x = x,
      y = y
    )
  }

  dat <- rbind(
    make_group("p1", 2),
    make_group("p2", 6)
  )

  out <- audit_gazepoint_biometric_sync_drift(
    dat,
    time_col = "time_ms",
    group_cols = "participant",
    signal_pairs = data.frame(signal_x = "x", signal_y = "y"),
    max_lag = 8,
    lag_step = 1,
    drift_tolerance = 2,
    min_complete_pairs = 20,
    include_reset_segments = FALSE
  )

  expect_s3_class(out, "gazepoint_biometric_sync_drift_audit")
  expect_true(all(c(
    "overview",
    "checks",
    "time_reset_audit",
    "lag_by_group",
    "lag_profile",
    "drift_summary",
    "settings"
  ) %in% names(out)))

  expect_equal(out$overview$signal_pair_count, 1)
  expect_equal(out$overview$lag_estimate_rows, 2)
  expect_equal(nrow(out$drift_summary), 1)
  expect_equal(out$drift_summary$status, "drift_exceeds_tolerance")
  expect_true(out$drift_summary$lag_range > 2)
})

test_that("audit_gazepoint_biometric_sync_drift is conservative without signal pairs", {
  dat <- data.frame(
    time_ms = seq_len(20),
    x = seq_len(20)
  )

  out <- audit_gazepoint_biometric_sync_drift(
    dat,
    time_col = "time_ms",
    signal_cols = "x"
  )

  expect_s3_class(out, "gazepoint_biometric_sync_drift_audit")
  expect_equal(out$overview$status, "no_signal_pairs")
  expect_equal(out$overview$signal_pair_count, 0)
  expect_equal(nrow(out$lag_by_group), 0)
  expect_equal(nrow(out$drift_summary), 0)
})

#' Run a Gazepoint Biometrics workflow
#'
#' Runs a compact end-to-end workflow for Gazepoint Biometrics exports. The
#' workflow imports rectangular all-gaze/fixation-style CSV exports from a
#' folder, validates biometric columns, detects active channels, audits
#' missingness, signal quality, sampling/timing, optionally creates window-level
#' summaries, creates optional biometric exclusion recommendations, extracts
#' optional TTL marker events, and produces checklist and methods-text outputs.
#'
#' @param path Folder containing Gazepoint CSV exports.
#' @param group_columns Optional columns used to create multimodal window
#'   summaries, such as `c("source_participant", "MEDIA_ID")`.
#' @param recursive Should subfolders be searched?
#' @param include_fixations Should fixation files be imported? Defaults to
#'   `FALSE` because continuous biometric summaries should usually be computed
#'   from all-gaze sample-level exports rather than fixation-level exports.
#' @param include_all_gaze Should all-gaze files be imported?
#' @param include_other_csv Should other non-Data-Summary CSV files be attempted?
#' @param require_active_signal Logical. Should inactive biometric signals be
#'   flagged in validation/checklist outputs?
#' @param create_exclusion_recommendations Logical. Should window-level and
#'   participant-level keep/review/exclude recommendations be created when
#'   `group_columns` are supplied?
#' @param gsr_min_usable_pct Minimum acceptable usable percentage for GSR/EDA
#'   windows.
#' @param hr_min_usable_pct Minimum acceptable usable percentage for heart-rate
#'   windows.
#' @param dial_min_usable_pct Minimum acceptable usable percentage for
#'   engagement-dial windows.
#' @param extract_ttl_events Logical. Should TTL marker events be extracted?
#' @param ttl_event_mode TTL event extraction mode passed to
#'   `extract_gazepoint_ttl_events()`. Use `"changes"` or `"nonzero"`.
#' @param audit_sampling Logical. Should sampling/timing information be audited?
#' @param sampling_group_columns Optional columns for the sampling audit. If
#'   `NULL`, the workflow uses available file/participant/media columns.
#' @param sampling_time_column Optional time/order column for the sampling audit.
#' @param sampling_time_unit Unit of the selected time/order column. Use
#'   `"seconds"`, `"milliseconds"`, `"microseconds"`, or `"samples"`.
#' @param expected_sampling_rate_hz Optional expected sampling rate in Hz.
#'
#' @return A list with imported data, validation outputs, missingness summaries,
#'   quality audits, sampling/timing audits, optional window summaries, optional
#'   exclusion recommendations, optional TTL events, checklist, and methods
#'   text. The object has class `"gazepoint_biometrics_workflow"`.
#'
#' @export
run_gazepoint_biometrics_workflow <- function(path,
                                              group_columns = NULL,
                                              recursive = FALSE,
                                              include_fixations = FALSE,
                                              include_all_gaze = TRUE,
                                              include_other_csv = FALSE,
                                              require_active_signal = TRUE,
                                              create_exclusion_recommendations = TRUE,
                                              gsr_min_usable_pct = 50,
                                              hr_min_usable_pct = 50,
                                              dial_min_usable_pct = 50,
                                              extract_ttl_events = TRUE,
                                              ttl_event_mode = c("changes", "nonzero"),
                                              audit_sampling = TRUE,
                                              sampling_group_columns = NULL,
                                              sampling_time_column = NULL,
                                              sampling_time_unit = c(
                                                "samples",
                                                "seconds",
                                                "milliseconds",
                                                "microseconds"
                                              ),
                                              expected_sampling_rate_hz = 60) {
  ttl_event_mode <- match.arg(ttl_event_mode)
  sampling_time_unit <- match.arg(sampling_time_unit)

  data <- import_gazepoint_biometric_folder(
    path = path,
    recursive = recursive,
    include_fixations = include_fixations,
    include_all_gaze = include_all_gaze,
    include_other_csv = include_other_csv
  )

  validation <- validate_gazepoint_biometrics(
    data,
    require_active_signal = require_active_signal
  )

  missingness <- audit_gazepoint_biometric_missingness(data)

  quality <- combine_gazepoint_tables(list(
    audit_gazepoint_gsr_quality(data),
    audit_gazepoint_hr_quality(data),
    audit_gazepoint_engagement_dial(data)
  ))

  sampling <- NULL

  if (isTRUE(audit_sampling)) {
    if (is.null(sampling_group_columns)) {
      sampling_group_columns <- intersect(
        c("source_file", "source_participant", "MEDIA_ID", "MEDIA_NAME"),
        names(data)
      )
    }

    sampling <- audit_gazepoint_biometric_sampling(
      data = data,
      group_columns = sampling_group_columns,
      time_column = sampling_time_column,
      time_unit = sampling_time_unit,
      expected_rate_hz = expected_sampling_rate_hz
    )
  }

  windows <- NULL
  exclusion_recommendations <- NULL

  if (!is.null(group_columns)) {
    windows <- summarise_gazepoint_multimodal_windows(
      data = data,
      group_columns = group_columns
    )

    if (isTRUE(create_exclusion_recommendations)) {
      exclusion_recommendations <- recommend_gazepoint_biometric_exclusions(
        data = windows,
        data_is_window_summary = TRUE,
        gsr_min_usable_pct = gsr_min_usable_pct,
        hr_min_usable_pct = hr_min_usable_pct,
        dial_min_usable_pct = dial_min_usable_pct
      )
    }
  }

  ttl_events <- NULL

  if (isTRUE(extract_ttl_events)) {
    ttl_group_columns <- intersect(
      c("source_participant", "USER", "USERID", "MEDIA_ID", "MEDIA_NAME"),
      names(data)
    )

    ttl_events <- extract_gazepoint_ttl_events(
      data = data,
      group_columns = ttl_group_columns,
      mode = ttl_event_mode
    )
  }

  checklist <- create_gazepoint_biometrics_checklist(
    data,
    require_active_signal = require_active_signal
  )

  methods_text <- create_gazepoint_biometrics_methods_text(
    checklist = checklist
  )

  overview <- data.frame(
    n_rows = nrow(data),
    n_columns = ncol(data),
    source_file_count = length(unique(data$source_file)),
    has_sampling_audit = !is.null(sampling),
    sampling_group_count = ifelse(is.null(sampling), NA_integer_, nrow(sampling)),
    has_window_summaries = !is.null(windows),
    has_exclusion_recommendations = !is.null(exclusion_recommendations),
    has_ttl_events = !is.null(ttl_events),
    ttl_event_count = ifelse(is.null(ttl_events), NA_integer_, nrow(ttl_events)),
    validation_issue_count = nrow(validation$issues),
    active_signal_count = validation$overview$active_signal_count,
    stringsAsFactors = FALSE
  )

  out <- list(
    overview = overview,
    data = data,
    validation = validation,
    missingness = missingness,
    quality = quality,
    sampling = sampling,
    windows = windows,
    exclusion_recommendations = exclusion_recommendations,
    ttl_events = ttl_events,
    checklist = checklist,
    methods_text = methods_text
  )

  class(out) <- c("gazepoint_biometrics_workflow", "list")
  out
}


#' Summarise a Gazepoint Biometrics workflow object
#'
#' Creates a compact summary table from an object returned by
#' `run_gazepoint_biometrics_workflow()`.
#'
#' @param workflow A workflow object returned by
#'   `run_gazepoint_biometrics_workflow()`.
#'
#' @return A one-row data frame summarising the workflow.
#'
#' @export
summarise_gazepoint_biometrics_workflow <- function(workflow) {
  if (!inherits(workflow, "gazepoint_biometrics_workflow")) {
    stop(
      "`workflow` must be produced by run_gazepoint_biometrics_workflow().",
      call. = FALSE
    )
  }

  active_channels <- workflow$validation$active_channels

  data.frame(
    n_rows = workflow$overview$n_rows,
    n_columns = workflow$overview$n_columns,
    source_file_count = workflow$overview$source_file_count,
    validation_issue_count = workflow$overview$validation_issue_count,
    active_gsr_eda = is_signal_active(active_channels, "gsr_eda"),
    active_heart_rate = is_signal_active(active_channels, "heart_rate"),
    active_engagement_dial = is_signal_active(active_channels, "engagement_dial"),
    active_ttl_marker = is_signal_active(active_channels, "ttl_marker"),
    has_sampling_audit = workflow$overview$has_sampling_audit,
    sampling_group_count = workflow$overview$sampling_group_count,
    has_window_summaries = workflow$overview$has_window_summaries,
    has_exclusion_recommendations = workflow$overview$has_exclusion_recommendations,
    has_ttl_events = workflow$overview$has_ttl_events,
    ttl_event_count = workflow$overview$ttl_event_count,
    stringsAsFactors = FALSE
  )
}

test_that("run_gazepoint_biometrics_workflow runs end-to-end on folder exports", {
  folder <- tempfile("gazepoint_workflow_")
  dir.create(folder)

  file_one <- file.path(folder, "User 0_all_gaze.csv")
  file_two <- file.path(folder, "User 0_fixations.csv")

  writeLines(
    c(
      "USER,MEDIA_ID,CNT,GSR_US,GSRV,HR,HRV,DIAL,DIALV,TTL0,TTLV,",
      "U1,1,1,2.0,1,70,1,0.1,1,1007,1,",
      "U1,1,2,2.2,1,72,1,0.2,1,1008,1,"
    ),
    file_one,
    useBytes = TRUE
  )

  writeLines(
    c(
      "USER,MEDIA_ID,CNT,FPOGX,FPOGY,FPOGS,FPOGD,FPOGID,GSR_US,GSRV,HR,HRV,DIAL,DIALV,TTL0,TTLV,",
      "U1,1,3,0.5,0.6,0.01,0.20,1,2.4,1,74,1,0.3,1,1008,1,",
      "U1,1,4,0.6,0.7,0.03,0.30,2,2.6,1,76,1,0.4,1,1009,1,"
    ),
    file_two,
    useBytes = TRUE
  )

  workflow <- run_gazepoint_biometrics_workflow(
    path = folder,
    group_columns = c("USER", "MEDIA_ID"),
    include_fixations = TRUE
  )

  expect_s3_class(workflow, "gazepoint_biometrics_workflow")
  expect_true(is.data.frame(workflow$overview))
  expect_true(is.data.frame(workflow$data))
  expect_true(is.list(workflow$validation))
  expect_true(is.data.frame(workflow$missingness))
  expect_true(is.data.frame(workflow$quality))
  expect_true(is.data.frame(workflow$sampling))
  expect_true(is.data.frame(workflow$windows))
  expect_s3_class(
    workflow$exclusion_recommendations,
    "gazepoint_biometric_exclusion_recommendations"
  )
  expect_true(is.data.frame(workflow$ttl_events))
  expect_s3_class(workflow$checklist, "gazepoint_biometrics_checklist")
  expect_type(workflow$methods_text, "character")

  expect_equal(workflow$overview$n_rows, 4)
  expect_equal(workflow$overview$source_file_count, 2)
  expect_true(workflow$overview$has_sampling_audit)
  expect_true(workflow$overview$sampling_group_count > 0)
  expect_true(workflow$overview$has_window_summaries)
  expect_true(workflow$overview$has_exclusion_recommendations)
  expect_true(workflow$overview$has_ttl_events)
  expect_true(workflow$overview$ttl_event_count > 0)
  expect_equal(nrow(workflow$windows), 1)
  expect_true("gsr_mean_value" %in% names(workflow$windows))
})


test_that("summarise_gazepoint_biometrics_workflow summarises workflow object", {
  folder <- tempfile("gazepoint_workflow_")
  dir.create(folder)

  file_one <- file.path(folder, "User 0_all_gaze.csv")

  writeLines(
    c(
      "USER,MEDIA_ID,CNT,GSR_US,GSRV,HR,HRV,DIAL,DIALV,TTL0,TTLV,",
      "U1,1,1,2.0,1,70,1,0.1,1,1007,1,",
      "U1,1,2,2.2,1,72,1,0.2,1,1008,1,"
    ),
    file_one,
    useBytes = TRUE
  )

  workflow <- run_gazepoint_biometrics_workflow(
    path = folder,
    group_columns = c("USER", "MEDIA_ID")
  )

  summary <- summarise_gazepoint_biometrics_workflow(workflow)

  expect_true(is.data.frame(summary))
  expect_equal(nrow(summary), 1)
  expect_equal(summary$n_rows, 2)
  expect_true(summary$active_gsr_eda)
  expect_true(summary$active_heart_rate)
  expect_true(summary$active_engagement_dial)
  expect_true(summary$active_ttl_marker)
  expect_true(summary$has_sampling_audit)
  expect_true(summary$sampling_group_count > 0)
  expect_true(summary$has_window_summaries)
  expect_true(summary$has_exclusion_recommendations)
  expect_true(summary$has_ttl_events)
  expect_true(summary$ttl_event_count > 0)
})


test_that("run_gazepoint_biometrics_workflow can skip exclusion recommendations", {
  folder <- tempfile("gazepoint_workflow_")
  dir.create(folder)

  file_one <- file.path(folder, "User 0_all_gaze.csv")

  writeLines(
    c(
      "USER,MEDIA_ID,CNT,GSR_US,GSRV,HR,HRV,DIAL,DIALV,TTL0,TTLV,",
      "U1,1,1,2.0,1,70,1,0.1,1,1007,1,",
      "U1,1,2,2.2,1,72,1,0.2,1,1008,1,"
    ),
    file_one,
    useBytes = TRUE
  )

  workflow <- run_gazepoint_biometrics_workflow(
    path = folder,
    group_columns = c("USER", "MEDIA_ID"),
    create_exclusion_recommendations = FALSE
  )

  expect_true(is.data.frame(workflow$windows))
  expect_null(workflow$exclusion_recommendations)
  expect_false(workflow$overview$has_exclusion_recommendations)

  summary <- summarise_gazepoint_biometrics_workflow(workflow)

  expect_false(summary$has_exclusion_recommendations)
})


test_that("run_gazepoint_biometrics_workflow can skip TTL extraction", {
  folder <- tempfile("gazepoint_workflow_")
  dir.create(folder)

  file_one <- file.path(folder, "User 0_all_gaze.csv")

  writeLines(
    c(
      "USER,MEDIA_ID,CNT,GSR_US,GSRV,HR,HRV,DIAL,DIALV,TTL0,TTLV,",
      "U1,1,1,2.0,1,70,1,0.1,1,1007,1,",
      "U1,1,2,2.2,1,72,1,0.2,1,1008,1,"
    ),
    file_one,
    useBytes = TRUE
  )

  workflow <- run_gazepoint_biometrics_workflow(
    path = folder,
    group_columns = c("USER", "MEDIA_ID"),
    extract_ttl_events = FALSE
  )

  expect_null(workflow$ttl_events)
  expect_false(workflow$overview$has_ttl_events)
  expect_true(is.na(workflow$overview$ttl_event_count))

  summary <- summarise_gazepoint_biometrics_workflow(workflow)

  expect_false(summary$has_ttl_events)
  expect_true(is.na(summary$ttl_event_count))
})


test_that("run_gazepoint_biometrics_workflow can skip sampling audit", {
  folder <- tempfile("gazepoint_workflow_")
  dir.create(folder)

  file_one <- file.path(folder, "User 0_all_gaze.csv")

  writeLines(
    c(
      "USER,MEDIA_ID,CNT,GSR_US,GSRV,HR,HRV,DIAL,DIALV,TTL0,TTLV,",
      "U1,1,1,2.0,1,70,1,0.1,1,1007,1,",
      "U1,1,2,2.2,1,72,1,0.2,1,1008,1,"
    ),
    file_one,
    useBytes = TRUE
  )

  workflow <- run_gazepoint_biometrics_workflow(
    path = folder,
    group_columns = c("USER", "MEDIA_ID"),
    audit_sampling = FALSE
  )

  expect_null(workflow$sampling)
  expect_false(workflow$overview$has_sampling_audit)
  expect_true(is.na(workflow$overview$sampling_group_count))

  summary <- summarise_gazepoint_biometrics_workflow(workflow)

  expect_false(summary$has_sampling_audit)
  expect_true(is.na(summary$sampling_group_count))
})


test_that("summarise_gazepoint_biometrics_workflow rejects invalid object", {
  expect_error(
    summarise_gazepoint_biometrics_workflow(list()),
    "must be produced"
  )
})

#' Baseline-correct Gazepoint GSR/EDA
#'
#' Adds a baseline-corrected GSR/EDA column to a Gazepoint Biometrics table.
#' When available, `GSR_US` is used by default because it represents skin
#' conductance in microsiemens in Gazepoint exports. The baseline is estimated
#' from rows selected by `baseline_rows`, optionally within groups.
#'
#' @param data A data frame or a path to a Gazepoint CSV export.
#' @param baseline_rows Logical vector identifying baseline rows.
#' @param value_column Optional GSR/EDA value column. If `NULL`, `GSR_US` is
#'   used when present, otherwise `GSR`.
#' @param validity_column Optional validity column. Defaults to `"GSRV"`.
#' @param group_columns Optional grouping columns. When supplied, baselines are
#'   estimated separately within each group.
#' @param output_column Name of the corrected output column.
#' @param summary Baseline summary, either `"mean"` or `"median"`.
#' @param exclude_zero Should zero values be excluded from baseline estimation?
#'
#' @return A data frame with the added baseline-corrected column and a
#'   baseline-summary attribute named `"baseline_summary"`.
#'
#' @export
baseline_correct_gazepoint_gsr <- function(data,
                                           baseline_rows,
                                           value_column = NULL,
                                           validity_column = "GSRV",
                                           group_columns = NULL,
                                           output_column = NULL,
                                           summary = c("mean", "median"),
                                           exclude_zero = TRUE) {
  dat <- coerce_gazepoint_biometrics_data(data)

  if (is.null(value_column)) {
    value_column <- choose_first_present(dat, c("GSR_US", "GSR"))
  }

  if (is.null(output_column)) {
    output_column <- paste0(value_column, "_baseline_corrected")
  }

  baseline_correct_gazepoint_signal(
    data = dat,
    baseline_rows = baseline_rows,
    value_column = value_column,
    validity_column = validity_column,
    group_columns = group_columns,
    output_column = output_column,
    summary = summary,
    exclude_zero = exclude_zero
  )
}


#' Baseline-correct Gazepoint heart rate
#'
#' Adds a baseline-corrected heart-rate column to a Gazepoint Biometrics table.
#' `HRV` is treated as the heart-rate validity flag, not as a heart-rate
#' variability metric.
#'
#' @param data A data frame or a path to a Gazepoint CSV export.
#' @param baseline_rows Logical vector identifying baseline rows.
#' @param value_column Heart-rate value column. Defaults to `"HR"`.
#' @param validity_column Heart-rate validity column. Defaults to `"HRV"`.
#' @param group_columns Optional grouping columns. When supplied, baselines are
#'   estimated separately within each group.
#' @param output_column Name of the corrected output column.
#' @param summary Baseline summary, either `"mean"` or `"median"`.
#' @param exclude_zero Should zero values be excluded from baseline estimation?
#'
#' @return A data frame with the added baseline-corrected column and a
#'   baseline-summary attribute named `"baseline_summary"`.
#'
#' @export
baseline_correct_gazepoint_hr <- function(data,
                                          baseline_rows,
                                          value_column = "HR",
                                          validity_column = "HRV",
                                          group_columns = NULL,
                                          output_column = NULL,
                                          summary = c("mean", "median"),
                                          exclude_zero = TRUE) {
  dat <- coerce_gazepoint_biometrics_data(data)

  if (is.null(output_column)) {
    output_column <- paste0(value_column, "_baseline_corrected")
  }

  baseline_correct_gazepoint_signal(
    data = dat,
    baseline_rows = baseline_rows,
    value_column = value_column,
    validity_column = validity_column,
    group_columns = group_columns,
    output_column = output_column,
    summary = summary,
    exclude_zero = exclude_zero
  )
}


#' Smooth a Gazepoint biometric signal
#'
#' Adds a simple centered moving-average smoothing column to a Gazepoint
#' Biometrics table. This is intentionally conservative and dependency-free.
#' It does not replace specialised biosignal preprocessing libraries.
#'
#' @param data A data frame or a path to a Gazepoint CSV export.
#' @param value_column Column to smooth.
#' @param window Number of samples in the moving window. Must be a positive odd
#'   integer.
#' @param output_column Name of the smoothed output column.
#' @param na_rm Should missing values be ignored within the moving window?
#'
#' @return A data frame with the added smoothed column.
#'
#' @export
smooth_gazepoint_biometrics <- function(data,
                                        value_column,
                                        window = 5L,
                                        output_column = NULL,
                                        na_rm = TRUE) {
  dat <- coerce_gazepoint_biometrics_data(data)

  if (missing(value_column) || length(value_column) != 1L || !nzchar(value_column)) {
    stop("`value_column` must be a single non-empty column name.", call. = FALSE)
  }

  if (!value_column %in% names(dat)) {
    stop("`value_column` was not found in `data`: ", value_column, call. = FALSE)
  }

  if (length(window) != 1L || is.na(window) || window < 1L || window %% 2L == 0L) {
    stop("`window` must be a positive odd integer.", call. = FALSE)
  }

  if (is.null(output_column)) {
    output_column <- paste0(value_column, "_smoothed")
  }

  x <- as_numeric_safe(dat[[value_column]])

  dat[[output_column]] <- centered_moving_average(
    x = x,
    window = as.integer(window),
    na_rm = na_rm
  )

  dat
}


baseline_correct_gazepoint_signal <- function(data,
                                              baseline_rows,
                                              value_column,
                                              validity_column = NULL,
                                              group_columns = NULL,
                                              output_column,
                                              summary = c("mean", "median"),
                                              exclude_zero = TRUE) {
  summary <- match.arg(summary)

  if (is.null(value_column) || length(value_column) != 1L || is.na(value_column)) {
    stop("`value_column` could not be determined.", call. = FALSE)
  }

  if (!value_column %in% names(data)) {
    stop("`value_column` was not found in `data`: ", value_column, call. = FALSE)
  }

  if (!is.logical(baseline_rows) || length(baseline_rows) != nrow(data)) {
    stop(
      "`baseline_rows` must be a logical vector with length equal to nrow(data).",
      call. = FALSE
    )
  }

  if (!is.null(group_columns)) {
    missing_groups <- setdiff(group_columns, names(data))

    if (length(missing_groups) > 0L) {
      stop(
        "`group_columns` were not found in `data`: ",
        paste(missing_groups, collapse = ", "),
        call. = FALSE
      )
    }
  }

  x <- as_numeric_safe(data[[value_column]])

  valid <- !is.na(x)

  if (isTRUE(exclude_zero)) {
    valid <- valid & x != 0
  }

  if (!is.null(validity_column) && validity_column %in% names(data)) {
    v <- as_numeric_safe(data[[validity_column]])
    valid <- valid & !is.na(v) & v > 0
  }

  baseline_eligible <- baseline_rows & valid

  if (is.null(group_columns)) {
    baseline_value <- compute_baseline_value(x[baseline_eligible], summary)

    corrected <- x - baseline_value

    baseline_summary <- data.frame(
      group = "all",
      value_column = value_column,
      validity_column = ifelse(
        !is.null(validity_column) && validity_column %in% names(data),
        validity_column,
        NA_character_
      ),
      baseline_rows = sum(baseline_rows),
      baseline_usable_rows = sum(baseline_eligible),
      baseline_value = baseline_value,
      summary = summary,
      stringsAsFactors = FALSE
    )
  } else {
    group_key <- make_group_key(data, group_columns)
    corrected <- rep(NA_real_, length(x))

    baseline_summary <- lapply(unique(group_key), function(key) {
      in_group <- group_key == key
      eligible <- baseline_eligible & in_group
      baseline_value <- compute_baseline_value(x[eligible], summary)

      corrected[in_group] <<- x[in_group] - baseline_value

      data.frame(
        group = key,
        value_column = value_column,
        validity_column = ifelse(
          !is.null(validity_column) && validity_column %in% names(data),
          validity_column,
          NA_character_
        ),
        baseline_rows = sum(baseline_rows & in_group),
        baseline_usable_rows = sum(eligible),
        baseline_value = baseline_value,
        summary = summary,
        stringsAsFactors = FALSE
      )
    })

    baseline_summary <- do.call(rbind, baseline_summary)
    rownames(baseline_summary) <- NULL
  }

  data[[output_column]] <- corrected

  attr(data, "baseline_summary") <- baseline_summary

  data
}


compute_baseline_value <- function(x, summary) {
  x <- x[!is.na(x)]

  if (length(x) == 0L) {
    return(NA_real_)
  }

  if (identical(summary, "median")) {
    return(stats::median(x))
  }

  mean(x)
}


make_group_key <- function(data, group_columns) {
  if (length(group_columns) == 1L) {
    return(as.character(data[[group_columns]]))
  }

  do.call(
    paste,
    c(data[group_columns], sep = "\r")
  )
}


centered_moving_average <- function(x, window, na_rm = TRUE) {
  n <- length(x)
  half_window <- floor(window / 2L)
  out <- rep(NA_real_, n)

  for (i in seq_len(n)) {
    from <- max(1L, i - half_window)
    to <- min(n, i + half_window)

    values <- x[from:to]

    if (isTRUE(na_rm)) {
      values <- values[!is.na(values)]
    }

    if (length(values) == 0L || any(is.na(values))) {
      out[i] <- NA_real_
    } else {
      out[i] <- mean(values)
    }
  }

  out
}

test_that("baseline_correct_gazepoint_gsr baseline-corrects GSR_US", {
  dat <- data.frame(
    trial = c(1, 1, 1, 1),
    GSR_US = c(2.0, 2.2, 2.5, 2.7),
    GSRV = c(1, 1, 1, 1),
    baseline = c(TRUE, TRUE, FALSE, FALSE)
  )

  out <- baseline_correct_gazepoint_gsr(
    dat,
    baseline_rows = dat$baseline
  )

  expect_true("GSR_US_baseline_corrected" %in% names(out))
  expect_equal(out$GSR_US_baseline_corrected[1], -0.1, tolerance = 1e-8)
  expect_equal(out$GSR_US_baseline_corrected[2], 0.1, tolerance = 1e-8)
  expect_equal(out$GSR_US_baseline_corrected[3], 0.4, tolerance = 1e-8)

  baseline_summary <- attr(out, "baseline_summary")
  expect_true(is.data.frame(baseline_summary))
  expect_equal(baseline_summary$baseline_usable_rows, 2)
  expect_equal(baseline_summary$baseline_value, 2.1, tolerance = 1e-8)
})


test_that("baseline_correct_gazepoint_hr baseline-corrects HR", {
  dat <- data.frame(
    HR = c(70, 72, 80, 82),
    HRV = c(1, 1, 1, 1),
    baseline = c(TRUE, TRUE, FALSE, FALSE)
  )

  out <- baseline_correct_gazepoint_hr(
    dat,
    baseline_rows = dat$baseline
  )

  expect_true("HR_baseline_corrected" %in% names(out))
  expect_equal(out$HR_baseline_corrected[1], -1, tolerance = 1e-8)
  expect_equal(out$HR_baseline_corrected[2], 1, tolerance = 1e-8)
  expect_equal(out$HR_baseline_corrected[3], 9, tolerance = 1e-8)

  baseline_summary <- attr(out, "baseline_summary")
  expect_equal(baseline_summary$baseline_value, 71, tolerance = 1e-8)
})


test_that("baseline correction supports groups", {
  dat <- data.frame(
    participant = c("P1", "P1", "P1", "P2", "P2", "P2"),
    HR = c(70, 72, 80, 60, 62, 70),
    HRV = c(1, 1, 1, 1, 1, 1),
    baseline = c(TRUE, TRUE, FALSE, TRUE, TRUE, FALSE)
  )

  out <- baseline_correct_gazepoint_hr(
    dat,
    baseline_rows = dat$baseline,
    group_columns = "participant"
  )

  expect_equal(out$HR_baseline_corrected[3], 9, tolerance = 1e-8)
  expect_equal(out$HR_baseline_corrected[6], 9, tolerance = 1e-8)

  baseline_summary <- attr(out, "baseline_summary")
  expect_equal(nrow(baseline_summary), 2)
})


test_that("baseline correction excludes invalid and zero baseline rows", {
  dat <- data.frame(
    HR = c(0, 70, 90),
    HRV = c(0, 1, 1),
    baseline = c(TRUE, TRUE, FALSE)
  )

  out <- baseline_correct_gazepoint_hr(
    dat,
    baseline_rows = dat$baseline
  )

  expect_equal(out$HR_baseline_corrected[3], 20, tolerance = 1e-8)

  baseline_summary <- attr(out, "baseline_summary")
  expect_equal(baseline_summary$baseline_usable_rows, 1)
  expect_equal(baseline_summary$baseline_value, 70, tolerance = 1e-8)
})


test_that("smooth_gazepoint_biometrics computes centered moving average", {
  dat <- data.frame(
    HR = c(70, 72, 74, 76, 78)
  )

  out <- smooth_gazepoint_biometrics(
    dat,
    value_column = "HR",
    window = 3
  )

  expect_true("HR_smoothed" %in% names(out))
  expect_equal(out$HR_smoothed[1], 71, tolerance = 1e-8)
  expect_equal(out$HR_smoothed[3], 74, tolerance = 1e-8)
  expect_equal(out$HR_smoothed[5], 77, tolerance = 1e-8)
})


test_that("smooth_gazepoint_biometrics rejects even window", {
  dat <- data.frame(
    HR = c(70, 72, 74)
  )

  expect_error(
    smooth_gazepoint_biometrics(dat, value_column = "HR", window = 4),
    "positive odd integer"
  )
})

#' Convert Gazepoint GSR resistance to conductance
#'
#' Converts GSR resistance values to conductance in microSiemens. The helper is
#' intentionally conservative. If a conductance column such as `GSR_US` is
#' already present, the data are returned unchanged unless `overwrite = TRUE`.
#' If `input_unit = "auto"`, conversion is performed only when the source column
#' has a resistance-like name, such as `GSR_OHMS` or `resistance_ohms`.
#'
#' Generic `GSR` columns are not automatically assumed to be resistance because
#' Gazepoint exports and workflows may represent GSR/EDA differently. For a
#' generic `GSR` column, use `input_unit = "ohms"` or `input_unit = "kohms"`
#' only when the study documentation confirms resistance units.
#'
#' @param data A data frame.
#' @param gsr_col Optional source GSR column. If `NULL`, a resistance-like column
#'   is detected when possible.
#' @param output_col Name of the output conductance column.
#' @param input_unit Source unit. `"auto"` converts only resistance-like columns;
#'   `"ohms"` converts ohms to microSiemens; `"kohms"` converts kilo-ohms to
#'   microSiemens; `"microsiemens"` copies values to `output_col`.
#' @param overwrite Logical. If `FALSE`, an existing `output_col` is not
#'   overwritten.
#'
#' @return The input data frame with a conductance column when conversion is
#'   possible. A structured conversion summary is stored in the
#'   `gsr_conversion_summary` attribute.
#'
#' @examples
#' df <- data.frame(GSR_OHMS = c(1000000, 500000, NA))
#' convert_gazepoint_gsr_to_conductance(df)
#'
#' @export
convert_gazepoint_gsr_to_conductance <- function(data,
                                                 gsr_col = NULL,
                                                 output_col = "GSR_US",
                                                 input_unit = c("auto",
                                                                "ohms",
                                                                "kohms",
                                                                "microsiemens"),
                                                 overwrite = FALSE) {
  if (!is.data.frame(data)) {
    stop("`data` must be a data frame.", call. = FALSE)
  }

  input_unit <- match.arg(input_unit)

  if (!is.character(output_col) ||
      length(output_col) != 1L ||
      is.na(output_col) ||
      !nzchar(output_col)) {
    stop("`output_col` must be a non-empty character string.", call. = FALSE)
  }

  if (!is.logical(overwrite) || length(overwrite) != 1L || is.na(overwrite)) {
    stop("`overwrite` must be TRUE or FALSE.", call. = FALSE)
  }

  if (!is.null(gsr_col)) {
    .gpbiom_assert_columns(data, gsr_col, "gsr_col")
  }

  existing_conductance <- .gpbiom_existing_conductance_column(data)

  if (output_col %in% names(data) && !isTRUE(overwrite)) {
    summary <- .gpbiom_gsr_conversion_summary(
      status = "conductance_column_already_present",
      source_column = NA_character_,
      output_column = output_col,
      input_unit = input_unit,
      n = nrow(data),
      n_converted = sum(!is.na(data[[output_col]])),
      n_invalid = 0L,
      note = paste0(
        "`", output_col, "` already exists and `overwrite = FALSE`; ",
        "data were returned unchanged."
      )
    )

    attr(data, "gsr_conversion_summary") <- summary
    return(data)
  }

  if (!is.na(existing_conductance) &&
      !identical(existing_conductance, output_col) &&
      !isTRUE(overwrite)) {
    summary <- .gpbiom_gsr_conversion_summary(
      status = "conductance_column_already_present",
      source_column = existing_conductance,
      output_column = existing_conductance,
      input_unit = "microsiemens",
      n = nrow(data),
      n_converted = sum(!is.na(data[[existing_conductance]])),
      n_invalid = 0L,
      note = paste0(
        "A conductance-like column (`", existing_conductance,
        "`) is already present; no conversion was performed."
      )
    )

    attr(data, "gsr_conversion_summary") <- summary
    return(data)
  }

  source_col <- if (is.null(gsr_col)) {
    .gpbiom_detect_resistance_column(data)
  } else {
    gsr_col
  }

  if (is.na(source_col)) {
    summary <- .gpbiom_gsr_conversion_summary(
      status = "no_resistance_source_detected",
      source_column = NA_character_,
      output_column = output_col,
      input_unit = input_unit,
      n = nrow(data),
      n_converted = 0L,
      n_invalid = 0L,
      note = paste0(
        "No resistance-like source column was detected. Generic GSR columns ",
        "are not converted automatically."
      )
    )

    attr(data, "gsr_conversion_summary") <- summary
    return(data)
  }

  if (!is.numeric(data[[source_col]])) {
    stop("The selected GSR source column must be numeric.", call. = FALSE)
  }

  source_standard <- standardise_gazepoint_biometric_names(source_col)

  resolved_unit <- .gpbiom_resolve_gsr_input_unit(
    input_unit = input_unit,
    source_standard = source_standard
  )

  if (identical(resolved_unit, "unknown")) {
    summary <- .gpbiom_gsr_conversion_summary(
      status = "unit_not_confirmed",
      source_column = source_col,
      output_column = output_col,
      input_unit = input_unit,
      n = nrow(data),
      n_converted = 0L,
      n_invalid = 0L,
      note = paste0(
        "The source column was not clearly resistance-like. Use ",
        "`input_unit = \"ohms\"` or `input_unit = \"kohms\"` only when ",
        "documentation confirms resistance units."
      )
    )

    attr(data, "gsr_conversion_summary") <- summary
    return(data)
  }

  source_values <- data[[source_col]]
  converted <- rep(NA_real_, length(source_values))
  invalid <- rep(FALSE, length(source_values))

  if (identical(resolved_unit, "ohms")) {
    valid <- !is.na(source_values) & is.finite(source_values) & source_values > 0
    invalid <- !is.na(source_values) & (!is.finite(source_values) | source_values <= 0)
    converted[valid] <- 1000000 / source_values[valid]
  } else if (identical(resolved_unit, "kohms")) {
    valid <- !is.na(source_values) & is.finite(source_values) & source_values > 0
    invalid <- !is.na(source_values) & (!is.finite(source_values) | source_values <= 0)
    converted[valid] <- 1000 / source_values[valid]
  } else if (identical(resolved_unit, "microsiemens")) {
    valid <- !is.na(source_values) & is.finite(source_values)
    invalid <- !is.na(source_values) & !is.finite(source_values)
    converted[valid] <- source_values[valid]
  }

  data[[output_col]] <- converted

  summary <- .gpbiom_gsr_conversion_summary(
    status = "conductance_created",
    source_column = source_col,
    output_column = output_col,
    input_unit = resolved_unit,
    n = nrow(data),
    n_converted = sum(!is.na(converted)),
    n_invalid = sum(invalid),
    note = "Conductance values are reported in microSiemens."
  )

  attr(data, "gsr_conversion_summary") <- summary
  data
}


#' Summarise tonic and phasic GSR/EDA components
#'
#' Creates a simple descriptive tonic/phasic decomposition of a GSR/EDA signal.
#' The tonic component is estimated with a rolling median, and the phasic
#' component is the observed signal minus the rolling-median tonic estimate.
#'
#' This is a lightweight descriptive helper, not a full skin-conductance-response
#' deconvolution model. It should be used for quality checks, window summaries,
#' and exploratory reporting unless a study requires a specialised EDA model.
#'
#' @param data A data frame.
#' @param gsr_col Optional GSR/EDA column. If `NULL`, the function prefers
#'   `GSR_US`, then `GSR`, then other recognised GSR/EDA columns.
#' @param group_cols Optional grouping columns. Tonic/phasic values are computed
#'   separately within each group.
#' @param time_col Optional time column used to order rows within groups.
#' @param window_n Rolling-median window size in samples.
#' @param peak_threshold Optional phasic peak threshold. If `NULL`, a robust
#'   data-driven threshold is computed within each group as
#'   `median(phasic) + 2 * MAD(phasic)`.
#' @param output_prefix Prefix for generated columns.
#'
#' @return A list with `data`, `summary`, and `settings`.
#'
#' @examples
#' df <- data.frame(
#'   CNT = 1:10,
#'   GSR_US = c(1, 1.1, 1.0, 1.2, 2.0, 1.3, 1.2, 1.1, 1.0, 1.1)
#' )
#' summarise_gazepoint_gsr_tonic_phasic(df, window_n = 3)
#'
#' @export
summarise_gazepoint_gsr_tonic_phasic <- function(data,
                                                 gsr_col = NULL,
                                                 group_cols = NULL,
                                                 time_col = NULL,
                                                 window_n = 15L,
                                                 peak_threshold = NULL,
                                                 output_prefix = "gsr") {
  if (!is.data.frame(data)) {
    stop("`data` must be a data frame.", call. = FALSE)
  }

  if (!is.null(gsr_col)) {
    .gpbiom_assert_columns(data, gsr_col, "gsr_col")
  }

  if (!is.null(group_cols)) {
    .gpbiom_assert_columns(data, group_cols, "group_cols")
  }

  if (!is.null(time_col)) {
    .gpbiom_assert_columns(data, time_col, "time_col")
  }

  .gpbiom_assert_positive_integer(window_n, "window_n")

  if (!is.null(peak_threshold) &&
      (!is.numeric(peak_threshold) ||
       length(peak_threshold) != 1L ||
       is.na(peak_threshold))) {
    stop("`peak_threshold` must be NULL or a single numeric value.",
         call. = FALSE)
  }

  if (!is.character(output_prefix) ||
      length(output_prefix) != 1L ||
      is.na(output_prefix) ||
      !nzchar(output_prefix)) {
    stop("`output_prefix` must be a non-empty character string.",
         call. = FALSE)
  }

  source_col <- if (is.null(gsr_col)) {
    .gpbiom_choose_gsr_column(data)
  } else {
    gsr_col
  }

  if (is.na(source_col)) {
    stop("No GSR/EDA column was detected. Provide `gsr_col` explicitly.",
         call. = FALSE)
  }

  if (!is.numeric(data[[source_col]])) {
    stop("The selected GSR/EDA column must be numeric.", call. = FALSE)
  }

  out <- data

  tonic_col <- paste0(output_prefix, "_tonic")
  phasic_col <- paste0(output_prefix, "_phasic")
  peak_col <- paste0(output_prefix, "_phasic_peak")
  threshold_col <- paste0(output_prefix, "_phasic_peak_threshold")

  out[[tonic_col]] <- NA_real_
  out[[phasic_col]] <- NA_real_
  out[[peak_col]] <- FALSE
  out[[threshold_col]] <- NA_real_

  groups <- .gpbiom_group_indices(out, group_cols)

  for (indices in groups) {
    ordered_indices <- .gpbiom_order_indices(out, indices, time_col)
    values <- out[[source_col]][ordered_indices]

    tonic <- .gpbiom_rolling_median(values, window_n)
    phasic <- values - tonic

    threshold <- if (is.null(peak_threshold)) {
      .gpbiom_default_phasic_threshold(phasic)
    } else {
      peak_threshold
    }

    peaks <- .gpbiom_phasic_peak_flags(phasic, threshold)

    out[[tonic_col]][ordered_indices] <- tonic
    out[[phasic_col]][ordered_indices] <- phasic
    out[[peak_col]][ordered_indices] <- peaks
    out[[threshold_col]][ordered_indices] <- threshold
  }

  summary <- .gpbiom_tonic_phasic_summary(
    data = out,
    source_col = source_col,
    tonic_col = tonic_col,
    phasic_col = phasic_col,
    peak_col = peak_col,
    threshold_col = threshold_col,
    group_cols = group_cols
  )

  settings <- list(
    gsr_col = source_col,
    group_cols = group_cols,
    time_col = time_col,
    window_n = as.integer(window_n),
    peak_threshold = peak_threshold,
    output_prefix = output_prefix,
    tonic_col = tonic_col,
    phasic_col = phasic_col,
    peak_col = peak_col,
    threshold_col = threshold_col,
    note = paste0(
      "This is a descriptive rolling-median tonic/phasic decomposition, ",
      "not a full EDA deconvolution model."
    )
  )

  list(
    data = out,
    summary = summary,
    settings = settings
  )
}


.gpbiom_existing_conductance_column <- function(data) {
  mapping <- standardise_gazepoint_biometric_names(data, rename = FALSE)
  conductance <- mapping$original_name[mapping$standard_name == "GSR_US"]

  if (length(conductance) == 0L) {
    return(NA_character_)
  }

  conductance[1L]
}


.gpbiom_detect_resistance_column <- function(data) {
  mapping <- standardise_gazepoint_biometric_names(data, rename = FALSE)
  resistance <- mapping$original_name[mapping$standard_name == "GSR_OHMS"]

  if (length(resistance) == 0L) {
    return(NA_character_)
  }

  resistance[1L]
}


.gpbiom_resolve_gsr_input_unit <- function(input_unit, source_standard) {
  if (identical(input_unit, "ohms")) {
    return("ohms")
  }

  if (identical(input_unit, "kohms")) {
    return("kohms")
  }

  if (identical(input_unit, "microsiemens")) {
    return("microsiemens")
  }

  if (identical(input_unit, "auto") &&
      identical(source_standard, "GSR_OHMS")) {
    return("ohms")
  }

  if (identical(input_unit, "auto") &&
      identical(source_standard, "GSR_US")) {
    return("microsiemens")
  }

  "unknown"
}


.gpbiom_gsr_conversion_summary <- function(status,
                                           source_column,
                                           output_column,
                                           input_unit,
                                           n,
                                           n_converted,
                                           n_invalid,
                                           note) {
  data.frame(
    status = status,
    source_column = source_column,
    output_column = output_column,
    input_unit = input_unit,
    n = n,
    n_converted = n_converted,
    n_invalid = n_invalid,
    note = note,
    stringsAsFactors = FALSE
  )
}

.gpbiom_choose_gsr_column <- function(data) {
  mapping <- standardise_gazepoint_biometric_names(data, rename = FALSE)

  preferred <- c("GSR_US", "GSR", "GSR_OHMS")

  recognised_candidates <- character()

  for (standard in preferred) {
    candidate <- mapping$original_name[mapping$standard_name == standard]

    if (length(candidate) == 0L) {
      next
    }

    recognised_candidates <- c(recognised_candidates, candidate)

    numeric_candidate <- candidate[vapply(candidate, function(column) {
      is.numeric(data[[column]])
    }, logical(1))]

    if (length(numeric_candidate) > 0L) {
      return(numeric_candidate[1L])
    }
  }

  if (length(recognised_candidates) > 0L) {
    return(recognised_candidates[1L])
  }

  NA_character_
}

.gpbiom_rolling_median <- function(values, window_n) {
  values <- as.numeric(values)
  n <- length(values)

  if (n == 0L) {
    return(numeric())
  }

  half_window <- floor(window_n / 2)
  out <- rep(NA_real_, n)

  for (i in seq_len(n)) {
    start <- max(1L, i - half_window)
    end <- min(n, i + half_window)

    window_values <- values[start:end]
    window_values <- window_values[!is.na(window_values) & is.finite(window_values)]

    if (length(window_values) > 0L) {
      out[i] <- stats::median(window_values)
    }
  }

  out
}


.gpbiom_default_phasic_threshold <- function(phasic) {
  finite <- phasic[!is.na(phasic) & is.finite(phasic)]

  if (length(finite) == 0L) {
    return(NA_real_)
  }

  robust_scale <- stats::mad(finite, constant = 1.4826, na.rm = TRUE)

  if (!is.finite(robust_scale) || robust_scale == 0) {
    robust_scale <- stats::sd(finite, na.rm = TRUE)
  }

  if (!is.finite(robust_scale) || robust_scale == 0) {
    return(Inf)
  }

  stats::median(finite, na.rm = TRUE) + 2 * robust_scale
}


.gpbiom_phasic_peak_flags <- function(phasic, threshold) {
  n <- length(phasic)
  peaks <- rep(FALSE, n)

  if (n == 0L || !is.finite(threshold)) {
    return(peaks)
  }

  finite <- !is.na(phasic) & is.finite(phasic)

  if (n == 1L) {
    peaks[1L] <- finite[1L] && phasic[1L] > threshold
    return(peaks)
  }

  for (i in seq_len(n)) {
    if (!finite[i] || phasic[i] <= threshold) {
      next
    }

    left_ok <- if (i == 1L) TRUE else !finite[i - 1L] || phasic[i] >= phasic[i - 1L]
    right_ok <- if (i == n) TRUE else !finite[i + 1L] || phasic[i] >= phasic[i + 1L]

    peaks[i] <- left_ok && right_ok
  }

  peaks
}


.gpbiom_tonic_phasic_summary <- function(data,
                                         source_col,
                                         tonic_col,
                                         phasic_col,
                                         peak_col,
                                         threshold_col,
                                         group_cols) {
  if (length(group_cols) == 0L) {
    return(.gpbiom_tonic_phasic_one_summary(
      data = data,
      group = "all",
      source_col = source_col,
      tonic_col = tonic_col,
      phasic_col = phasic_col,
      peak_col = peak_col,
      threshold_col = threshold_col
    ))
  }

  split_key <- interaction(data[group_cols], drop = TRUE, lex.order = TRUE)
  indices <- split(seq_len(nrow(data)), split_key)

  rows <- lapply(names(indices), function(group_name) {
    .gpbiom_tonic_phasic_one_summary(
      data = data[indices[[group_name]], , drop = FALSE],
      group = group_name,
      source_col = source_col,
      tonic_col = tonic_col,
      phasic_col = phasic_col,
      peak_col = peak_col,
      threshold_col = threshold_col
    )
  })

  do.call(rbind, rows)
}


.gpbiom_tonic_phasic_one_summary <- function(data,
                                             group,
                                             source_col,
                                             tonic_col,
                                             phasic_col,
                                             peak_col,
                                             threshold_col) {
  signal <- data[[source_col]]
  tonic <- data[[tonic_col]]
  phasic <- data[[phasic_col]]

  finite_signal <- signal[!is.na(signal) & is.finite(signal)]
  finite_tonic <- tonic[!is.na(tonic) & is.finite(tonic)]
  finite_phasic <- phasic[!is.na(phasic) & is.finite(phasic)]

  positive_phasic <- finite_phasic[finite_phasic > 0]

  data.frame(
    group = group,
    n_rows = nrow(data),
    source_column = source_col,
    n_signal_finite = length(finite_signal),
    mean_signal = if (length(finite_signal) > 0L) mean(finite_signal) else NA_real_,
    median_signal = if (length(finite_signal) > 0L) stats::median(finite_signal) else NA_real_,
    mean_tonic = if (length(finite_tonic) > 0L) mean(finite_tonic) else NA_real_,
    median_tonic = if (length(finite_tonic) > 0L) stats::median(finite_tonic) else NA_real_,
    mean_phasic = if (length(finite_phasic) > 0L) mean(finite_phasic) else NA_real_,
    median_phasic = if (length(finite_phasic) > 0L) stats::median(finite_phasic) else NA_real_,
    max_phasic = if (length(finite_phasic) > 0L) max(finite_phasic) else NA_real_,
    min_phasic = if (length(finite_phasic) > 0L) min(finite_phasic) else NA_real_,
    positive_phasic_sum = if (length(positive_phasic) > 0L) sum(positive_phasic) else 0,
    n_phasic_peaks = sum(data[[peak_col]], na.rm = TRUE),
    peak_threshold = unique(data[[threshold_col]][!is.na(data[[threshold_col]])])[1L],
    stringsAsFactors = FALSE
  )
}

test_that("convert_gazepoint_gsr_to_conductance converts ohms automatically from resistance-like column", {
  df <- data.frame(
    GSR_OHMS = c(1000000, 500000, NA, 0, -1, Inf)
  )

  out <- convert_gazepoint_gsr_to_conductance(df)

  expect_true("GSR_US" %in% names(out))
  expect_equal(out$GSR_US[1:2], c(1, 2))
  expect_true(is.na(out$GSR_US[3]))
  expect_true(is.na(out$GSR_US[4]))
  expect_true(is.na(out$GSR_US[5]))
  expect_true(is.na(out$GSR_US[6]))

  summary <- attr(out, "gsr_conversion_summary")
  expect_s3_class(summary, "data.frame")
  expect_equal(summary$status, "conductance_created")
  expect_equal(summary$n_converted, 2)
  expect_equal(summary$n_invalid, 3)
})


test_that("convert_gazepoint_gsr_to_conductance converts kilo-ohms when explicitly requested", {
  df <- data.frame(
    GSR = c(1000, 500, NA)
  )

  out <- convert_gazepoint_gsr_to_conductance(
    df,
    gsr_col = "GSR",
    input_unit = "kohms"
  )

  expect_equal(out$GSR_US[1:2], c(1, 2))
  expect_true(is.na(out$GSR_US[3]))

  summary <- attr(out, "gsr_conversion_summary")
  expect_equal(summary$input_unit, "kohms")
  expect_equal(summary$status, "conductance_created")
})


test_that("convert_gazepoint_gsr_to_conductance does not auto-convert generic GSR", {
  df <- data.frame(
    GSR = c(1000000, 500000)
  )

  out <- convert_gazepoint_gsr_to_conductance(df, gsr_col = "GSR")

  expect_false("GSR_US" %in% names(out))
  summary <- attr(out, "gsr_conversion_summary")
  expect_equal(summary$status, "unit_not_confirmed")
})


test_that("convert_gazepoint_gsr_to_conductance leaves existing conductance unchanged by default", {
  df <- data.frame(
    GSR_US = c(1, 2, 3),
    GSR_OHMS = c(1000000, 500000, 333333.3)
  )

  out <- convert_gazepoint_gsr_to_conductance(df)

  expect_equal(out$GSR_US, c(1, 2, 3))
  summary <- attr(out, "gsr_conversion_summary")
  expect_equal(summary$status, "conductance_column_already_present")
})


test_that("convert_gazepoint_gsr_to_conductance can overwrite existing conductance", {
  df <- data.frame(
    GSR_US = c(10, 10),
    GSR_OHMS = c(1000000, 500000)
  )

  out <- convert_gazepoint_gsr_to_conductance(df, overwrite = TRUE)

  expect_equal(out$GSR_US, c(1, 2))
  summary <- attr(out, "gsr_conversion_summary")
  expect_equal(summary$status, "conductance_created")
})


test_that("convert_gazepoint_gsr_to_conductance can copy microsiemens values", {
  df <- data.frame(
    EDA = c(1.1, 1.2, NA)
  )

  out <- convert_gazepoint_gsr_to_conductance(
    df,
    gsr_col = "EDA",
    input_unit = "microsiemens",
    output_col = "conductance_us"
  )

  expect_equal(out$conductance_us, c(1.1, 1.2, NA))
  summary <- attr(out, "gsr_conversion_summary")
  expect_equal(summary$input_unit, "microsiemens")
})


test_that("convert_gazepoint_gsr_to_conductance validates arguments", {
  df <- data.frame(GSR = 1:3)

  expect_error(
    convert_gazepoint_gsr_to_conductance(1:3),
    "`data` must be"
  )

  expect_error(
    convert_gazepoint_gsr_to_conductance(df, gsr_col = "missing"),
    "not found"
  )

  expect_error(
    convert_gazepoint_gsr_to_conductance(df, output_col = ""),
    "`output_col`"
  )

  expect_error(
    convert_gazepoint_gsr_to_conductance(df, overwrite = NA),
    "`overwrite`"
  )

  expect_error(
    convert_gazepoint_gsr_to_conductance(data.frame(GSR_OHMS = letters[1:3])),
    "must be numeric"
  )
})


test_that("summarise_gazepoint_gsr_tonic_phasic creates tonic and phasic columns", {
  df <- data.frame(
    CNT = 1:10,
    GSR_US = c(1, 1.1, 1.0, 1.2, 2.0, 1.3, 1.2, 1.1, 1.0, 1.1)
  )

  out <- summarise_gazepoint_gsr_tonic_phasic(
    df,
    window_n = 3,
    peak_threshold = 0.4
  )

  expect_type(out, "list")
  expect_s3_class(out$data, "data.frame")
  expect_s3_class(out$summary, "data.frame")
  expect_true(all(c(
    "gsr_tonic",
    "gsr_phasic",
    "gsr_phasic_peak",
    "gsr_phasic_peak_threshold"
  ) %in% names(out$data)))

  expect_true(any(out$data$gsr_phasic_peak))
  expect_equal(out$settings$gsr_col, "GSR_US")
  expect_equal(out$summary$group, "all")
})


test_that("summarise_gazepoint_gsr_tonic_phasic respects groups", {
  df <- data.frame(
    USER = rep(c("P1", "P2"), each = 5),
    CNT = rep(1:5, 2),
    GSR_US = c(1, 1, 2, 1, 1, 2, 2, 3, 2, 2)
  )

  out <- summarise_gazepoint_gsr_tonic_phasic(
    df,
    group_cols = "USER",
    time_col = "CNT",
    window_n = 3,
    peak_threshold = 0.4
  )

  expect_equal(nrow(out$summary), 2)
  expect_true(all(out$summary$group %in% c("P1", "P2")))
  expect_true("gsr_phasic" %in% names(out$data))
})


test_that("summarise_gazepoint_gsr_tonic_phasic can use explicit GSR column", {
  df <- data.frame(
    custom_eda = c(1, 1.2, 1.1, 1.8, 1.1)
  )

  out <- summarise_gazepoint_gsr_tonic_phasic(
    df,
    gsr_col = "custom_eda",
    window_n = 3
  )

  expect_equal(out$settings$gsr_col, "custom_eda")
  expect_equal(out$summary$source_column, "custom_eda")
})


test_that("summarise_gazepoint_gsr_tonic_phasic handles missing values", {
  df <- data.frame(
    GSR_US = c(1, NA, 1.2, 2.0, NA, 1.1)
  )

  out <- summarise_gazepoint_gsr_tonic_phasic(
    df,
    window_n = 3,
    peak_threshold = 0.2
  )

  expect_equal(nrow(out$data), 6)
  expect_true(sum(!is.na(out$data$gsr_tonic)) >= 4)
  expect_true(sum(!is.na(out$data$gsr_phasic)) >= 4)
})


test_that("summarise_gazepoint_gsr_tonic_phasic validates arguments", {
  df <- data.frame(GSR_US = c(1, 2, 3))

  expect_error(
    summarise_gazepoint_gsr_tonic_phasic(1:3),
    "`data` must be"
  )

  expect_error(
    summarise_gazepoint_gsr_tonic_phasic(df, gsr_col = "missing"),
    "not found"
  )

  expect_error(
    summarise_gazepoint_gsr_tonic_phasic(df, window_n = 0),
    "`window_n`"
  )

  expect_error(
    summarise_gazepoint_gsr_tonic_phasic(df, peak_threshold = "high"),
    "`peak_threshold`"
  )

  expect_error(
    summarise_gazepoint_gsr_tonic_phasic(df, output_prefix = ""),
    "`output_prefix`"
  )

  expect_error(
    summarise_gazepoint_gsr_tonic_phasic(data.frame(GSR_US = letters[1:3])),
    "must be numeric"
  )
})


test_that("summarise_gazepoint_gsr_tonic_phasic requires a detectable GSR column", {
  df <- data.frame(
    HR = c(70, 71, 72)
  )

  expect_error(
    summarise_gazepoint_gsr_tonic_phasic(df),
    "No GSR/EDA column"
  )
})

#' Prepare Gazepoint EDA input for external Ledalab-style workflows
#'
#' Prepares a clean Gazepoint EDA/conductance time-series table that can be
#' exported for external Ledalab-style workflows. This function does not run
#' Ledalab and does not attempt to reproduce Ledalab internally.
#'
#' @param data A Gazepoint biometric data frame, or a list containing one.
#' @param eda_col Optional EDA/conductance column. If omitted, the function
#'   prefers `GSR_US` when available.
#' @param time_col Optional time/counter column.
#' @param group_cols Optional grouping columns.
#' @param sampling_rate Optional sampling rate in Hz, used when the time column
#'   is a sample counter.
#' @param time_unit Unit of `time_col`.
#' @param convert_resistance_to_us If `TRUE`, convert a selected resistance-like
#'   `GSR` column to microsiemens as `1,000,000 / GSR`. The default is `FALSE`
#'   because this conversion should be used only when the user has verified that
#'   `GSR` is resistance-like and `GSR_US` is unavailable.
#' @param min_finite_prop Minimum finite proportion required for a group to be
#'   labelled ready.
#' @param output_dir Optional folder where CSV files should be written.
#' @param prefix File prefix used when `output_dir` is supplied.
#'
#' @return A list with `overview`, `signal_table`, `group_summary`, `manifest`,
#'   and `settings`.
#' @export
prepare_gazepoint_ledalab_input <- function(data,
                                            eda_col = NULL,
                                            time_col = NULL,
                                            group_cols = NULL,
                                            sampling_rate = NULL,
                                            time_unit = c("auto", "ms", "seconds", "samples"),
                                            convert_resistance_to_us = FALSE,
                                            min_finite_prop = 0.50,
                                            output_dir = NULL,
                                            prefix = "gazepoint_ledalab") {
  gpbiometrics_prepare_external_eda_input(
    data = data,
    method = "ledalab",
    eda_col = eda_col,
    time_col = time_col,
    group_cols = group_cols,
    sampling_rate = sampling_rate,
    time_unit = match.arg(time_unit),
    convert_resistance_to_us = convert_resistance_to_us,
    min_finite_prop = min_finite_prop,
    output_dir = output_dir,
    prefix = prefix
  )
}

#' Prepare Gazepoint EDA input for external PsPM-style workflows
#'
#' Prepares a clean Gazepoint EDA/conductance time-series table that can be
#' exported for external PsPM-style workflows. This function does not run PsPM
#' and does not attempt to reproduce PsPM internally.
#'
#' @inheritParams prepare_gazepoint_ledalab_input
#'
#' @return A list with `overview`, `signal_table`, `group_summary`, `manifest`,
#'   and `settings`.
#' @export
prepare_gazepoint_pspm_input <- function(data,
                                         eda_col = NULL,
                                         time_col = NULL,
                                         group_cols = NULL,
                                         sampling_rate = NULL,
                                         time_unit = c("auto", "ms", "seconds", "samples"),
                                         convert_resistance_to_us = FALSE,
                                         min_finite_prop = 0.50,
                                         output_dir = NULL,
                                         prefix = "gazepoint_pspm") {
  gpbiometrics_prepare_external_eda_input(
    data = data,
    method = "pspm",
    eda_col = eda_col,
    time_col = time_col,
    group_cols = group_cols,
    sampling_rate = sampling_rate,
    time_unit = match.arg(time_unit),
    convert_resistance_to_us = convert_resistance_to_us,
    min_finite_prop = min_finite_prop,
    output_dir = output_dir,
    prefix = prefix
  )
}

#' Prepare Gazepoint EDA input for external cvxEDA-style workflows
#'
#' Prepares a clean Gazepoint EDA/conductance time-series table that can be
#' exported for external cvxEDA-style workflows. This function does not run a
#' native cvxEDA solver and does not attempt to reproduce cvxEDA internally.
#'
#' @inheritParams prepare_gazepoint_ledalab_input
#'
#' @return A list with `overview`, `signal_table`, `group_summary`, `manifest`,
#'   and `settings`.
#' @export
prepare_gazepoint_cvxeda_input <- function(data,
                                           eda_col = NULL,
                                           time_col = NULL,
                                           group_cols = NULL,
                                           sampling_rate = NULL,
                                           time_unit = c("auto", "ms", "seconds", "samples"),
                                           convert_resistance_to_us = FALSE,
                                           min_finite_prop = 0.50,
                                           output_dir = NULL,
                                           prefix = "gazepoint_cvxeda") {
  gpbiometrics_prepare_external_eda_input(
    data = data,
    method = "cvxeda",
    eda_col = eda_col,
    time_col = time_col,
    group_cols = group_cols,
    sampling_rate = sampling_rate,
    time_unit = match.arg(time_unit),
    convert_resistance_to_us = convert_resistance_to_us,
    min_finite_prop = min_finite_prop,
    output_dir = output_dir,
    prefix = prefix
  )
}

#' Classify descriptive Gazepoint EDA response patterns
#'
#' Classifies descriptive EDA response patterns from an EDA/SCR feature column.
#' The labels are intended for quality-control and descriptive reporting only.
#' They do not infer emotion, valence, stress, trust, preference, cognition, or
#' diagnosis.
#'
#' @param data A data frame containing EDA/SCR values.
#' @param response_col Optional response feature column. If omitted, the helper
#'   searches for common SCR/EDA response columns and then `GSR_US_PHASIC` or
#'   `GSR_US`.
#' @param group_cols Optional grouping columns.
#' @param summary_function Summary used within each group.
#' @param no_response_threshold Absolute response threshold for
#'   `no_detectable_response`.
#' @param low_response_threshold Upper threshold for `low_response`.
#' @param moderate_response_threshold Upper threshold for `moderate_response`.
#'
#' @return A list with `overview`, `classifications`, and `settings`.
#' @export
classify_gazepoint_eda_response_pattern <- function(data,
                                                    response_col = NULL,
                                                    group_cols = NULL,
                                                    summary_function = c("max_abs", "mean_abs", "median_abs"),
                                                    no_response_threshold = 0.01,
                                                    low_response_threshold = 0.05,
                                                    moderate_response_threshold = 0.20) {
  dat <- gpbiometrics_eda_bridge_extract_data(data)

  summary_function <- match.arg(summary_function)

  response_col <- gpbiometrics_eda_bridge_resolve_response_col(
    dat = dat,
    response_col = response_col
  )

  group_cols <- gpbiometrics_eda_bridge_resolve_group_cols(dat, group_cols)

  gpbiometrics_eda_bridge_validate_thresholds(
    no_response_threshold = no_response_threshold,
    low_response_threshold = low_response_threshold,
    moderate_response_threshold = moderate_response_threshold
  )

  split_info <- gpbiometrics_eda_bridge_split_data(dat, group_cols)

  classifications <- lapply(split_info$groups, function(group_dat) {
    response <- suppressWarnings(as.numeric(group_dat[[response_col]]))
    finite_response <- response[is.finite(response)]

    abs_response <- abs(finite_response)

    response_value <- if (length(abs_response) == 0) {
      NA_real_
    } else if (identical(summary_function, "max_abs")) {
      max(abs_response, na.rm = TRUE)
    } else if (identical(summary_function, "mean_abs")) {
      mean(abs_response, na.rm = TRUE)
    } else {
      stats::median(abs_response, na.rm = TRUE)
    }

    pattern <- gpbiometrics_eda_bridge_response_label(
      response_value = response_value,
      no_response_threshold = no_response_threshold,
      low_response_threshold = low_response_threshold,
      moderate_response_threshold = moderate_response_threshold
    )

    group_values <- gpbiometrics_eda_bridge_group_values(group_dat, group_cols)

    data.frame(
      group_values,
      group_id = gpbiometrics_eda_bridge_group_id(group_dat, group_cols),
      response_col = response_col,
      n_rows = nrow(group_dat),
      n_finite = length(finite_response),
      finite_prop = if (nrow(group_dat) > 0) length(finite_response) / nrow(group_dat) else NA_real_,
      summary_function = summary_function,
      response_value = response_value,
      response_pattern = pattern,
      status = if (length(finite_response) == 0) "fail_no_finite_response_values" else "response_pattern_classified",
      interpretation = paste(
        "This is a descriptive EDA response-pattern label only.",
        "It does not infer emotion, valence, stress, trust, preference, cognition, or diagnosis."
      ),
      stringsAsFactors = FALSE
    )
  })

  classifications <- gpbiometrics_eda_bridge_rbind(classifications)

  status <- if (nrow(classifications) == 0) {
    "no_groups_classified"
  } else if (all(classifications$status == "response_pattern_classified")) {
    "eda_response_patterns_classified"
  } else if (any(classifications$status == "response_pattern_classified")) {
    "partial_eda_response_patterns_classified"
  } else {
    "eda_response_patterns_not_classified"
  }

  overview <- data.frame(
    input_rows = nrow(dat),
    group_count = nrow(classifications),
    classified_group_count = sum(classifications$status == "response_pattern_classified"),
    response_col = response_col,
    summary_function = summary_function,
    status = status,
    interpretation = paste(
      "EDA response-pattern labels are descriptive QC/reporting aids.",
      "They do not infer emotion, valence, stress, trust, preference, cognition, or diagnosis."
    ),
    stringsAsFactors = FALSE
  )

  structure(
    list(
      overview = overview,
      classifications = classifications,
      settings = list(
        response_col = response_col,
        group_cols = group_cols,
        summary_function = summary_function,
        no_response_threshold = no_response_threshold,
        low_response_threshold = low_response_threshold,
        moderate_response_threshold = moderate_response_threshold
      )
    ),
    class = c("gazepoint_eda_response_pattern", "list")
  )
}

gpbiometrics_prepare_external_eda_input <- function(data,
                                                    method,
                                                    eda_col,
                                                    time_col,
                                                    group_cols,
                                                    sampling_rate,
                                                    time_unit,
                                                    convert_resistance_to_us,
                                                    min_finite_prop,
                                                    output_dir,
                                                    prefix) {
  dat <- gpbiometrics_eda_bridge_extract_data(data)

  eda_col <- gpbiometrics_eda_bridge_resolve_eda_col(dat, eda_col)
  time_col <- gpbiometrics_eda_bridge_resolve_time_col(dat, time_col)
  group_cols <- gpbiometrics_eda_bridge_resolve_group_cols(dat, group_cols)

  gpbiometrics_eda_bridge_validate_sampling_rate(sampling_rate)
  gpbiometrics_eda_bridge_validate_min_finite_prop(min_finite_prop)

  split_info <- gpbiometrics_eda_bridge_split_data(dat, group_cols)

  prepared <- lapply(split_info$groups, function(group_dat) {
    gpbiometrics_eda_bridge_prepare_group(
      group_dat = group_dat,
      method = method,
      eda_col = eda_col,
      time_col = time_col,
      group_cols = group_cols,
      sampling_rate = sampling_rate,
      time_unit = time_unit,
      convert_resistance_to_us = convert_resistance_to_us,
      min_finite_prop = min_finite_prop
    )
  })

  signal_table <- gpbiometrics_eda_bridge_rbind(lapply(prepared, `[[`, "signal_table"))
  group_summary <- gpbiometrics_eda_bridge_rbind(lapply(prepared, `[[`, "group_summary"))

  manifest <- gpbiometrics_eda_bridge_write_outputs(
    signal_table = signal_table,
    group_summary = group_summary,
    output_dir = output_dir,
    prefix = prefix,
    method = method
  )

  ready_group_count <- if (nrow(group_summary) == 0) {
    0L
  } else {
    sum(group_summary$status == paste0("ready_for_", method, "_input"))
  }

  status <- if (nrow(group_summary) == 0) {
    paste0(method, "_input_not_prepared")
  } else if (ready_group_count == nrow(group_summary)) {
    paste0(method, "_input_prepared")
  } else if (ready_group_count > 0) {
    paste0(method, "_input_prepared_with_review_flags")
  } else {
    paste0(method, "_input_requires_review")
  }

  overview <- data.frame(
    input_rows = nrow(dat),
    group_count = nrow(group_summary),
    ready_group_count = ready_group_count,
    method = method,
    eda_col = eda_col,
    time_col = if (is.null(time_col)) NA_character_ else time_col,
    conductance_unit = unique(signal_table$conductance_unit)[1],
    output_file_count = nrow(manifest),
    status = status,
    interpretation = paste(
      "Output is prepared for optional external EDA workflows.",
      "No external toolbox is invoked and no emotion, valence, stress, trust, preference, cognition, or diagnosis is inferred."
    ),
    stringsAsFactors = FALSE
  )

  structure(
    list(
      overview = overview,
      signal_table = signal_table,
      group_summary = group_summary,
      manifest = manifest,
      settings = list(
        method = method,
        eda_col = eda_col,
        time_col = time_col,
        group_cols = group_cols,
        sampling_rate = sampling_rate,
        time_unit = time_unit,
        convert_resistance_to_us = convert_resistance_to_us,
        min_finite_prop = min_finite_prop,
        output_dir = output_dir,
        prefix = prefix
      )
    ),
    class = c(paste0("gazepoint_", method, "_input"), "gazepoint_external_eda_input", "list")
  )
}

gpbiometrics_eda_bridge_prepare_group <- function(group_dat,
                                                  method,
                                                  eda_col,
                                                  time_col,
                                                  group_cols,
                                                  sampling_rate,
                                                  time_unit,
                                                  convert_resistance_to_us,
                                                  min_finite_prop) {
  group_id <- gpbiometrics_eda_bridge_group_id(group_dat, group_cols)
  group_values <- gpbiometrics_eda_bridge_group_values(group_dat, group_cols)

  eda_raw <- suppressWarnings(as.numeric(group_dat[[eda_col]]))

  used_conversion <- FALSE
  conductance_unit <- "as_supplied"

  if (isTRUE(convert_resistance_to_us) &&
      identical(toupper(eda_col), "GSR")) {
    eda_raw <- ifelse(is.finite(eda_raw) & eda_raw > 0, 1000000 / eda_raw, NA_real_)
    used_conversion <- TRUE
    conductance_unit <- "microsiemens_converted_from_resistance"
  } else if (toupper(eda_col) %in% c("GSR_US", "EDA_US", "CONDUCTANCE_US")) {
    conductance_unit <- "microsiemens"
  }

  timing <- gpbiometrics_eda_bridge_time_seconds(
    group_dat = group_dat,
    time_col = time_col,
    sampling_rate = sampling_rate,
    time_unit = time_unit
  )

  signal_table <- data.frame(
    group_values,
    group_id = rep(group_id, nrow(group_dat)),
    sample_index = seq_len(nrow(group_dat)),
    time_s = timing$time_s,
    conductance_us = eda_raw,
    eda_raw = suppressWarnings(as.numeric(group_dat[[eda_col]])),
    method = method,
    eda_col = eda_col,
    time_col = if (is.null(time_col)) NA_character_ else time_col,
    detected_time_unit = timing$detected_time_unit,
    conductance_unit = conductance_unit,
    used_resistance_conversion = used_conversion,
    stringsAsFactors = FALSE
  )

  if (identical(method, "cvxeda")) {
    signal_table$y <- signal_table$conductance_us
  }

  n_finite <- sum(is.finite(signal_table$conductance_us))
  finite_prop <- if (nrow(signal_table) > 0) n_finite / nrow(signal_table) else NA_real_

  status <- if (nrow(signal_table) == 0) {
    paste0("review_no_rows_for_", method)
  } else if (!is.finite(finite_prop) || finite_prop < min_finite_prop) {
    paste0("review_low_finite_signal_for_", method)
  } else if (all(is.na(signal_table$time_s))) {
    paste0("prepared_with_sample_index_only_for_", method)
  } else {
    paste0("ready_for_", method, "_input")
  }

  interpretation <- paste(
    "Prepared signal is an external-method input table only.",
    "External decomposition/model fitting remains outside gpbiometrics.",
    "No emotion, valence, stress, trust, preference, cognition, or diagnosis is inferred."
  )

  group_summary <- data.frame(
    group_values,
    group_id = group_id,
    method = method,
    eda_col = eda_col,
    n_rows = nrow(signal_table),
    n_finite = n_finite,
    finite_prop = finite_prop,
    conductance_min = suppressWarnings(min(signal_table$conductance_us, na.rm = TRUE)),
    conductance_median = suppressWarnings(stats::median(signal_table$conductance_us, na.rm = TRUE)),
    conductance_max = suppressWarnings(max(signal_table$conductance_us, na.rm = TRUE)),
    conductance_sd = suppressWarnings(stats::sd(signal_table$conductance_us, na.rm = TRUE)),
    time_col = if (is.null(time_col)) NA_character_ else time_col,
    detected_time_unit = timing$detected_time_unit,
    time_span_s = timing$time_span_s,
    median_time_step_s = timing$median_time_step_s,
    estimated_sampling_rate_hz = timing$estimated_sampling_rate_hz,
    conductance_unit = conductance_unit,
    used_resistance_conversion = used_conversion,
    status = status,
    interpretation = interpretation,
    stringsAsFactors = FALSE
  )

  finite_numeric_cols <- c(
    "conductance_min",
    "conductance_median",
    "conductance_max",
    "conductance_sd"
  )

  for (nm in finite_numeric_cols) {
    if (!is.finite(group_summary[[nm]])) {
      group_summary[[nm]] <- NA_real_
    }
  }

  list(
    signal_table = signal_table,
    group_summary = group_summary
  )
}

gpbiometrics_eda_bridge_extract_data <- function(data) {
  if (is.data.frame(data)) {
    return(data)
  }

  if (is.list(data)) {
    candidates <- c("data", "biometrics", "merged_data", "all_data", "imported_data")

    for (nm in candidates) {
      if (!is.null(data[[nm]]) && is.data.frame(data[[nm]])) {
        return(data[[nm]])
      }
    }

    is_df <- vapply(data, is.data.frame, logical(1))

    if (any(is_df)) {
      return(data[[which(is_df)[1]]])
    }
  }

  stop("`data` must be a data frame or a list containing a data frame.", call. = FALSE)
}

gpbiometrics_eda_bridge_resolve_eda_col <- function(dat, eda_col = NULL) {
  if (!is.null(eda_col)) {
    if (!is.character(eda_col) || length(eda_col) != 1 || is.na(eda_col)) {
      stop("`eda_col` must be a single column name.", call. = FALSE)
    }

    if (!eda_col %in% names(dat)) {
      stop("Column `", eda_col, "` was not found in `data`.", call. = FALSE)
    }

    return(eda_col)
  }

  candidates <- c(
    "GSR_US",
    "gsr_us",
    "EDA_US",
    "eda_us",
    "conductance_us",
    "GSR_US_PHASIC",
    "GSR_US_TONIC",
    "EDA",
    "eda",
    "GSR"
  )

  found <- intersect(candidates, names(dat))

  if (length(found) == 0) {
    stop(
      "No EDA/conductance column was detected. Supply `eda_col` explicitly.",
      call. = FALSE
    )
  }

  found[1]
}

gpbiometrics_eda_bridge_resolve_response_col <- function(dat, response_col = NULL) {
  if (!is.null(response_col)) {
    if (!is.character(response_col) || length(response_col) != 1 || is.na(response_col)) {
      stop("`response_col` must be a single column name.", call. = FALSE)
    }

    if (!response_col %in% names(dat)) {
      stop("Column `", response_col, "` was not found in `data`.", call. = FALSE)
    }

    return(response_col)
  }

  candidates <- c(
    "scr_amplitude_us",
    "SCR_amplitude_us",
    "peak_amplitude_us",
    "amplitude_us",
    "response_amplitude_us",
    "GSR_US_PHASIC",
    "GSR_US",
    "EDA",
    "eda"
  )

  found <- intersect(candidates, names(dat))

  if (length(found) == 0) {
    stop(
      "No EDA/SCR response column was detected. Supply `response_col` explicitly.",
      call. = FALSE
    )
  }

  found[1]
}

gpbiometrics_eda_bridge_resolve_time_col <- function(dat, time_col = NULL) {
  if (!is.null(time_col)) {
    if (!is.character(time_col) || length(time_col) != 1 || is.na(time_col)) {
      stop("`time_col` must be a single column name.", call. = FALSE)
    }

    if (!time_col %in% names(dat)) {
      stop("Column `", time_col, "` was not found in `data`.", call. = FALSE)
    }

    return(time_col)
  }

  candidates <- c(
    "time_s",
    "time_sec",
    "time_seconds",
    "time_ms",
    "timestamp_ms",
    "timestamp",
    "TIME",
    "Time",
    "time",
    "CNT",
    "cnt"
  )

  found <- intersect(candidates, names(dat))

  if (length(found) == 0) {
    return(NULL)
  }

  found[1]
}

gpbiometrics_eda_bridge_resolve_group_cols <- function(dat, group_cols = NULL) {
  if (is.null(group_cols)) {
    candidates <- c(
      "source_file",
      "USER_FILE",
      "participant",
      "participant_id",
      "subject",
      "subject_id",
      "MEDIA_ID",
      "MEDIA_NAME",
      "stimulus",
      "trial",
      "trial_id",
      "trial_global"
    )

    return(intersect(candidates, names(dat)))
  }

  if (!is.character(group_cols)) {
    stop("`group_cols` must be NULL or a character vector of column names.", call. = FALSE)
  }

  missing_cols <- setdiff(group_cols, names(dat))

  if (length(missing_cols) > 0) {
    stop(
      "The following `group_cols` were not found in `data`: ",
      paste(missing_cols, collapse = ", "),
      call. = FALSE
    )
  }

  group_cols
}

gpbiometrics_eda_bridge_split_data <- function(dat, group_cols) {
  if (length(group_cols) == 0) {
    return(list(groups = list(dat)))
  }

  split_key <- interaction(dat[group_cols], drop = TRUE, lex.order = TRUE)
  groups <- split(dat, split_key, drop = TRUE)

  list(groups = groups)
}

gpbiometrics_eda_bridge_group_id <- function(dat, group_cols) {
  if (length(group_cols) == 0) {
    return("all")
  }

  vals <- vapply(group_cols, function(nm) {
    value <- dat[[nm]][1]
    if (is.na(value)) "<NA>" else as.character(value)
  }, character(1))

  paste(vals, collapse = "||")
}

gpbiometrics_eda_bridge_group_values <- function(dat, group_cols) {
  if (length(group_cols) == 0) {
    return(data.frame(stringsAsFactors = FALSE))
  }

  values <- lapply(group_cols, function(nm) {
    dat[[nm]][1]
  })

  names(values) <- group_cols

  as.data.frame(
    values,
    stringsAsFactors = FALSE,
    optional = TRUE
  )
}

gpbiometrics_eda_bridge_time_seconds <- function(group_dat,
                                                 time_col,
                                                 sampling_rate,
                                                 time_unit) {
  if (is.null(time_col)) {
    time_s <- if (!is.null(sampling_rate)) {
      (seq_len(nrow(group_dat)) - 1) / sampling_rate
    } else {
      rep(NA_real_, nrow(group_dat))
    }

    return(list(
      time_s = time_s,
      detected_time_unit = if (!is.null(sampling_rate)) "sample_index" else "missing_time",
      time_span_s = if (length(time_s) > 1 && any(is.finite(time_s))) {
        max(time_s, na.rm = TRUE) - min(time_s, na.rm = TRUE)
      } else {
        NA_real_
      },
      median_time_step_s = if (length(time_s) > 1) {
        stats::median(diff(time_s), na.rm = TRUE)
      } else {
        NA_real_
      },
      estimated_sampling_rate_hz = if (!is.null(sampling_rate)) sampling_rate else NA_real_
    ))
  }

  time_raw <- suppressWarnings(as.numeric(group_dat[[time_col]]))

  detected_time_unit <- gpbiometrics_eda_bridge_detect_time_unit(
    time_raw = time_raw,
    time_col = time_col,
    time_unit = time_unit,
    sampling_rate = sampling_rate
  )

  finite_time <- time_raw[is.finite(time_raw)]
  time_origin <- if (length(finite_time) > 0) min(finite_time, na.rm = TRUE) else NA_real_

  time_s <- if (identical(detected_time_unit, "ms")) {
    (time_raw - time_origin) / 1000
  } else if (identical(detected_time_unit, "seconds")) {
    time_raw - time_origin
  } else if (identical(detected_time_unit, "samples") && !is.null(sampling_rate)) {
    (time_raw - time_origin) / sampling_rate
  } else {
    rep(NA_real_, length(time_raw))
  }

  finite_time_s <- time_s[is.finite(time_s)]
  time_step <- diff(finite_time_s)

  median_time_step_s <- if (length(time_step) > 0) {
    stats::median(time_step[time_step > 0], na.rm = TRUE)
  } else {
    NA_real_
  }

  estimated_sampling_rate_hz <- if (is.finite(median_time_step_s) && median_time_step_s > 0) {
    1 / median_time_step_s
  } else if (!is.null(sampling_rate)) {
    sampling_rate
  } else {
    NA_real_
  }

  list(
    time_s = time_s,
    detected_time_unit = detected_time_unit,
    time_span_s = if (length(finite_time_s) > 0) {
      max(finite_time_s, na.rm = TRUE) - min(finite_time_s, na.rm = TRUE)
    } else {
      NA_real_
    },
    median_time_step_s = median_time_step_s,
    estimated_sampling_rate_hz = estimated_sampling_rate_hz
  )
}

gpbiometrics_eda_bridge_detect_time_unit <- function(time_raw,
                                                     time_col,
                                                     time_unit,
                                                     sampling_rate) {
  if (!identical(time_unit, "auto")) {
    return(time_unit)
  }

  lower_name <- tolower(time_col)

  if (lower_name %in% c("cnt", "sample", "sample_index")) {
    return("samples")
  }

  if (grepl("ms|millisecond", lower_name)) {
    return("ms")
  }

  if (grepl("sec|second", lower_name)) {
    return("seconds")
  }

  finite_time <- time_raw[is.finite(time_raw)]

  if (length(finite_time) == 0) {
    return(if (!is.null(sampling_rate)) "samples" else "seconds")
  }

  range_time <- diff(range(finite_time, na.rm = TRUE))

  if (is.finite(range_time) && range_time > 1000) {
    "ms"
  } else {
    "seconds"
  }
}

gpbiometrics_eda_bridge_validate_sampling_rate <- function(sampling_rate) {
  if (is.null(sampling_rate)) {
    return(invisible(TRUE))
  }

  if (!is.numeric(sampling_rate) || length(sampling_rate) != 1 ||
      is.na(sampling_rate) || sampling_rate <= 0) {
    stop("`sampling_rate` must be NULL or a single positive number.", call. = FALSE)
  }

  invisible(TRUE)
}

gpbiometrics_eda_bridge_validate_min_finite_prop <- function(min_finite_prop) {
  if (!is.numeric(min_finite_prop) || length(min_finite_prop) != 1 ||
      is.na(min_finite_prop) || min_finite_prop < 0 || min_finite_prop > 1) {
    stop("`min_finite_prop` must be a single number between 0 and 1.", call. = FALSE)
  }

  invisible(TRUE)
}

gpbiometrics_eda_bridge_validate_thresholds <- function(no_response_threshold,
                                                        low_response_threshold,
                                                        moderate_response_threshold) {
  values <- c(no_response_threshold, low_response_threshold, moderate_response_threshold)

  if (!is.numeric(values) || any(!is.finite(values)) || any(values < 0)) {
    stop("Response thresholds must be finite non-negative numbers.", call. = FALSE)
  }

  if (!(no_response_threshold <= low_response_threshold &&
        low_response_threshold <= moderate_response_threshold)) {
    stop(
      "Thresholds must satisfy: no_response_threshold <= low_response_threshold <= moderate_response_threshold.",
      call. = FALSE
    )
  }

  invisible(TRUE)
}

gpbiometrics_eda_bridge_response_label <- function(response_value,
                                                   no_response_threshold,
                                                   low_response_threshold,
                                                   moderate_response_threshold) {
  if (!is.finite(response_value)) {
    return("unclassified_no_finite_response")
  }

  if (response_value <= no_response_threshold) {
    "no_detectable_response"
  } else if (response_value <= low_response_threshold) {
    "low_response"
  } else if (response_value <= moderate_response_threshold) {
    "moderate_response"
  } else {
    "high_response"
  }
}

gpbiometrics_eda_bridge_write_outputs <- function(signal_table,
                                                  group_summary,
                                                  output_dir,
                                                  prefix,
                                                  method) {
  manifest <- data.frame(
    file_role = character(),
    path = character(),
    stringsAsFactors = FALSE
  )

  if (is.null(output_dir)) {
    return(manifest)
  }

  if (!dir.exists(output_dir)) {
    dir.create(output_dir, recursive = TRUE, showWarnings = FALSE)
  }

  if (!dir.exists(output_dir)) {
    stop("Could not create `output_dir`: ", output_dir, call. = FALSE)
  }

  signal_path <- file.path(output_dir, paste0(prefix, "_", method, "_signal_table.csv"))
  summary_path <- file.path(output_dir, paste0(prefix, "_", method, "_group_summary.csv"))

  utils::write.csv(signal_table, signal_path, row.names = FALSE)
  utils::write.csv(group_summary, summary_path, row.names = FALSE)

  data.frame(
    file_role = c("signal_table", "group_summary"),
    path = normalizePath(c(signal_path, summary_path), winslash = "/", mustWork = FALSE),
    stringsAsFactors = FALSE
  )
}

gpbiometrics_eda_bridge_rbind <- function(x) {
  x <- x[!vapply(x, is.null, logical(1))]

  if (length(x) == 0) {
    return(data.frame())
  }

  all_names <- unique(unlist(lapply(x, names), use.names = FALSE))

  x <- lapply(x, function(dat) {
    missing_names <- setdiff(all_names, names(dat))

    for (nm in missing_names) {
      dat[[nm]] <- NA
    }

    dat[all_names]
  })

  do.call(rbind, x)
}

test_that("prepare_gazepoint_ledalab_input prepares grouped conductance tables", {
  dat <- data.frame(
    participant = rep(c("p1", "p2"), each = 20),
    CNT = rep(seq_len(20), times = 2),
    GSR_US = c(seq(1, 2, length.out = 20), seq(2, 3, length.out = 20))
  )

  out <- prepare_gazepoint_ledalab_input(
    dat,
    eda_col = "GSR_US",
    time_col = "CNT",
    group_cols = "participant",
    sampling_rate = 60
  )

  expect_s3_class(out, "gazepoint_ledalab_input")
  expect_s3_class(out, "gazepoint_external_eda_input")
  expect_equal(out$overview$status, "ledalab_input_prepared")
  expect_equal(out$overview$group_count, 2)
  expect_equal(out$overview$ready_group_count, 2)
  expect_true(is.data.frame(out$signal_table))
  expect_true(is.data.frame(out$group_summary))
  expect_true(all(c("time_s", "conductance_us", "group_id") %in% names(out$signal_table)))
  expect_true(all(is.finite(out$signal_table$time_s)))
  expect_equal(unique(out$signal_table$conductance_unit), "microsiemens")
})

test_that("prepare_gazepoint_pspm_input prepares sample-index-only data when no time column exists", {
  dat <- data.frame(
    participant = rep(c("p1", "p2"), each = 10),
    GSR_US = c(seq(1, 2, length.out = 10), seq(2, 3, length.out = 10))
  )

  out <- prepare_gazepoint_pspm_input(
    dat,
    eda_col = "GSR_US",
    group_cols = "participant",
    sampling_rate = 10
  )

  expect_s3_class(out, "gazepoint_pspm_input")
  expect_equal(out$overview$status, "pspm_input_prepared")
  expect_true(all(is.finite(out$signal_table$time_s)))
  expect_equal(unique(out$signal_table$detected_time_unit), "sample_index")
})

test_that("prepare_gazepoint_cvxeda_input adds y column for cvxEDA-style workflows", {
  dat <- data.frame(
    participant = "p1",
    time_s = seq(0, 1.9, by = 0.1),
    GSR_US = seq(1, 2, length.out = 20)
  )

  out <- prepare_gazepoint_cvxeda_input(
    dat,
    eda_col = "GSR_US",
    time_col = "time_s",
    group_cols = "participant"
  )

  expect_s3_class(out, "gazepoint_cvxeda_input")
  expect_equal(out$overview$status, "cvxeda_input_prepared")
  expect_true("y" %in% names(out$signal_table))
  expect_equal(out$signal_table$y, out$signal_table$conductance_us)
})

test_that("external EDA bridges can write optional CSV outputs", {
  dat <- data.frame(
    participant = "p1",
    CNT = seq_len(10),
    GSR_US = seq(1, 2, length.out = 10)
  )

  out_dir <- tempfile("eda_bridge_")

  out <- prepare_gazepoint_ledalab_input(
    dat,
    eda_col = "GSR_US",
    time_col = "CNT",
    group_cols = "participant",
    sampling_rate = 60,
    output_dir = out_dir,
    prefix = "test_bridge"
  )

  expect_equal(nrow(out$manifest), 2)
  expect_true(all(file.exists(out$manifest$path)))
})

test_that("external EDA bridges can conservatively convert verified resistance-like GSR", {
  dat <- data.frame(
    participant = "p1",
    CNT = seq_len(3),
    GSR = c(1000000, 500000, 250000)
  )

  out <- prepare_gazepoint_ledalab_input(
    dat,
    eda_col = "GSR",
    time_col = "CNT",
    group_cols = "participant",
    sampling_rate = 60,
    convert_resistance_to_us = TRUE
  )

  expect_equal(out$signal_table$conductance_us, c(1, 2, 4))
  expect_true(all(out$signal_table$used_resistance_conversion))
  expect_equal(
    unique(out$signal_table$conductance_unit),
    "microsiemens_converted_from_resistance"
  )
})

test_that("classify_gazepoint_eda_response_pattern gives descriptive labels only", {
  dat <- data.frame(
    participant = rep(c("none", "low", "moderate", "high"), each = 4),
    scr_amplitude_us = c(
      0, 0.002, 0.003, 0.004,
      0.02, 0.03, 0.04, 0.04,
      0.08, 0.10, 0.15, 0.18,
      0.25, 0.30, 0.40, 0.50
    )
  )

  out <- classify_gazepoint_eda_response_pattern(
    dat,
    response_col = "scr_amplitude_us",
    group_cols = "participant",
    summary_function = "max_abs",
    no_response_threshold = 0.01,
    low_response_threshold = 0.05,
    moderate_response_threshold = 0.20
  )

  expect_s3_class(out, "gazepoint_eda_response_pattern")
  expect_equal(out$overview$status, "eda_response_patterns_classified")

  patterns <- out$classifications$response_pattern
  names(patterns) <- out$classifications$participant

  expect_equal(patterns[["none"]], "no_detectable_response")
  expect_equal(patterns[["low"]], "low_response")
  expect_equal(patterns[["moderate"]], "moderate_response")
  expect_equal(patterns[["high"]], "high_response")

  expect_true(all(grepl("does not infer emotion", out$classifications$interpretation)))
})

test_that("classify_gazepoint_eda_response_pattern handles missing finite values conservatively", {
  dat <- data.frame(
    participant = "p1",
    scr_amplitude_us = c(NA_real_, NA_real_)
  )

  out <- classify_gazepoint_eda_response_pattern(
    dat,
    response_col = "scr_amplitude_us",
    group_cols = "participant"
  )

  expect_equal(out$overview$status, "eda_response_patterns_not_classified")
  expect_equal(out$classifications$response_pattern, "unclassified_no_finite_response")
  expect_equal(out$classifications$status, "fail_no_finite_response_values")
})

test_that("external EDA bridges error clearly when no EDA column is available", {
  dat <- data.frame(
    participant = "p1",
    CNT = seq_len(5),
    HR = 60 + seq_len(5)
  )

  expect_error(
    prepare_gazepoint_ledalab_input(dat, time_col = "CNT"),
    "No EDA/conductance column"
  )
})

#' Prepare Gazepoint EDA data for CTSI sparse deconvolution workflows
#'
#' Prepares signal, event, and configuration tables for downstream
#' continuous-time system identification (CTSI) sparse EDA deconvolution
#' workflows. This function does not implement the full Amin-Faghih CTSI solver
#' in R. It creates reproducible input objects and optional CSV files for
#' external CTSI implementations.
#'
#' @param dat A data frame containing EDA data.
#' @param eda_col Numeric EDA/conductance column.
#' @param time_col Numeric time column.
#' @param group_cols Optional grouping columns, such as participant/session.
#' @param event_onset_col Optional event onset column.
#' @param event_name_col Optional event/condition column.
#' @param sampling_rate Optional sampling rate in Hz.
#' @param tau0_range Candidate slow time-constant range.
#' @param tau1_range Candidate fast time-constant range.
#' @param sparsity_grid Candidate sparsity penalties.
#' @param output_dir Optional directory for CSV export.
#' @param prefix Output file prefix.
#'
#' @return A list with `overview`, `signal_table`, `event_table`,
#'   `ctsi_config`, `ctsi_notes`, `written_files`, and `settings`.
#' @export
prepare_gazepoint_ctsi_input <- function(dat,
                                         eda_col = "GSR_US",
                                         time_col = "CNT",
                                         group_cols = NULL,
                                         event_onset_col = NULL,
                                         event_name_col = NULL,
                                         sampling_rate = NULL,
                                         tau0_range = c(2, 4),
                                         tau1_range = c(0.5, 1),
                                         sparsity_grid = c(0.001, 0.01, 0.1, 1),
                                         output_dir = NULL,
                                         prefix = "gazepoint_ctsi") {
  if (!is.data.frame(dat)) {
    stop("`dat` must be a data frame.", call. = FALSE)
  }

  required <- c(eda_col, time_col, group_cols, event_onset_col, event_name_col)
  required <- required[!is.null(required)]

  missing_required <- setdiff(required, names(dat))
  if (length(missing_required) > 0) {
    stop("Missing required columns: ", paste(missing_required, collapse = ", "), call. = FALSE)
  }

  if (!is.numeric(dat[[eda_col]])) {
    stop("`eda_col` must identify a numeric column.", call. = FALSE)
  }

  if (!is.numeric(dat[[time_col]])) {
    stop("`time_col` must identify a numeric column.", call. = FALSE)
  }

  if (is.null(group_cols)) {
    group_cols <- character()
  }

  signal_table <- data.frame(
    time = dat[[time_col]],
    conductance = dat[[eda_col]],
    stringsAsFactors = FALSE
  )

  for (group_col in group_cols) {
    signal_table[[group_col]] <- dat[[group_col]]
  }

  signal_table <- signal_table[is.finite(signal_table$time), , drop = FALSE]
  signal_table <- signal_table[order(signal_table$time), , drop = FALSE]

  if (is.null(sampling_rate)) {
    sampling_rate <- gpbiometrics_ctsi_sampling_rate(signal_table$time)
  }

  event_table <- data.frame()

  if (!is.null(event_onset_col)) {
    event_table <- data.frame(
      onset = dat[[event_onset_col]],
      event_name = if (!is.null(event_name_col)) {
        as.character(dat[[event_name_col]])
      } else {
        "event"
      },
      stringsAsFactors = FALSE
    )

    for (group_col in group_cols) {
      event_table[[group_col]] <- dat[[group_col]]
    }

    event_table <- unique(event_table[is.finite(event_table$onset), , drop = FALSE])
    event_table <- event_table[order(event_table$onset), , drop = FALSE]
  }

  ctsi_config <- expand.grid(
    tau0 = seq(tau0_range[1], tau0_range[2], length.out = 5),
    tau1 = seq(tau1_range[1], tau1_range[2], length.out = 3),
    sparsity_lambda = sparsity_grid,
    KEEP.OUT.ATTRS = FALSE,
    stringsAsFactors = FALSE
  )

  ctsi_config <- ctsi_config[ctsi_config$tau0 > ctsi_config$tau1, , drop = FALSE]
  ctsi_config$sampling_rate_hz <- sampling_rate

  ctsi_notes <- c(
    "This object prepares Gazepoint EDA data for downstream CTSI sparse deconvolution workflows.",
    "It does not implement the full continuous-time system-identification solver inside R.",
    "The signal table should be checked for units, sampling rate, missingness, and filtering before external CTSI fitting.",
    "The configuration grid records candidate physiological time constants and sparsity penalties.",
    "CTSI-derived latent-driver estimates should not be interpreted as direct emotion, stress, cognition, health status, or diagnosis."
  )

  written_files <- character()

  if (!is.null(output_dir)) {
    if (!dir.exists(output_dir)) {
      dir.create(output_dir, recursive = TRUE, showWarnings = FALSE)
    }

    signal_file <- file.path(output_dir, paste0(prefix, "_signal.csv"))
    utils::write.csv(signal_table, signal_file, row.names = FALSE)
    written_files <- c(written_files, signal_file)

    config_file <- file.path(output_dir, paste0(prefix, "_config.csv"))
    utils::write.csv(ctsi_config, config_file, row.names = FALSE)
    written_files <- c(written_files, config_file)

    if (nrow(event_table) > 0) {
      event_file <- file.path(output_dir, paste0(prefix, "_events.csv"))
      utils::write.csv(event_table, event_file, row.names = FALSE)
      written_files <- c(written_files, event_file)
    }

    notes_file <- file.path(output_dir, paste0(prefix, "_notes.txt"))
    writeLines(ctsi_notes, notes_file, useBytes = TRUE)
    written_files <- c(written_files, notes_file)
  }

  overview <- data.frame(
    signal_rows = nrow(signal_table),
    event_rows = nrow(event_table),
    config_rows = nrow(ctsi_config),
    sampling_rate_hz = sampling_rate,
    output_written = length(written_files) > 0,
    status = if (nrow(signal_table) > 0 && nrow(ctsi_config) > 0) {
      "ctsi_input_prepared"
    } else {
      "ctsi_input_incomplete"
    },
    interpretation = paste(
      "Prepared tables are an interoperability bridge for CTSI sparse EDA deconvolution.",
      "They do not constitute CTSI model fitting or psychological-state inference."
    ),
    stringsAsFactors = FALSE
  )

  structure(
    list(
      overview = overview,
      signal_table = signal_table,
      event_table = event_table,
      ctsi_config = ctsi_config,
      ctsi_notes = ctsi_notes,
      written_files = written_files,
      settings = list(
        eda_col = eda_col,
        time_col = time_col,
        group_cols = group_cols,
        event_onset_col = event_onset_col,
        event_name_col = event_name_col,
        sampling_rate = sampling_rate,
        tau0_range = tau0_range,
        tau1_range = tau1_range,
        sparsity_grid = sparsity_grid,
        output_dir = output_dir,
        prefix = prefix
      )
    ),
    class = c("gazepoint_ctsi_input", "list")
  )
}

gpbiometrics_ctsi_sampling_rate <- function(time) {
  time <- time[is.finite(time)]

  if (length(time) < 3) {
    return(NA_real_)
  }

  dt <- diff(time)
  dt <- dt[is.finite(dt) & dt > 0]

  if (length(dt) == 0) {
    return(NA_real_)
  }

  median_dt <- stats::median(dt)

  if (median_dt > 10) {
    1000 / median_dt
  } else {
    1 / median_dt
  }
}

#' Synchronise Gazepoint Biometrics with gaze data
#'
#' Joins Gazepoint Biometrics data to gaze or fixation data using exact key
#' columns. This function is intentionally conservative: it does not perform
#' interpolation, nearest-neighbour matching, or automatic time shifting. For a
#' first reproducible workflow, exact joins by participant, media, trial, sample
#' counter, or fixation identifier are preferred when those fields are available.
#'
#' @param biometrics A Gazepoint Biometrics data frame.
#' @param gaze A gaze, fixation, AOI, or `gp3tools`-style summary data frame.
#' @param by Character vector of key columns used for joining.
#' @param all_x Logical. Should all rows from `gaze` be retained?
#' @param suffixes Character vector of length two used for duplicate non-key
#'   column names.
#'
#' @return A data frame with gaze rows joined to biometric columns. The returned
#'   object has class `"gazepoint_biometrics_sync"` and a `"sync_summary"`
#'   attribute.
#'
#' @export
sync_gazepoint_biometrics_with_gaze <- function(biometrics,
                                                gaze,
                                                by,
                                                all_x = TRUE,
                                                suffixes = c(".gaze", ".bio")) {
  if (!is.data.frame(biometrics)) {
    biometrics <- coerce_gazepoint_biometrics_data(biometrics)
  }

  if (!is.data.frame(gaze)) {
    stop("`gaze` must be a data frame.", call. = FALSE)
  }

  if (missing(by) || !is.character(by) || length(by) == 0L) {
    stop("`by` must be a non-empty character vector of join columns.", call. = FALSE)
  }

  missing_bio_keys <- setdiff(by, names(biometrics))
  missing_gaze_keys <- setdiff(by, names(gaze))

  if (length(missing_bio_keys) > 0L) {
    stop(
      "`by` columns were not found in `biometrics`: ",
      paste(missing_bio_keys, collapse = ", "),
      call. = FALSE
    )
  }

  if (length(missing_gaze_keys) > 0L) {
    stop(
      "`by` columns were not found in `gaze`: ",
      paste(missing_gaze_keys, collapse = ", "),
      call. = FALSE
    )
  }

  biometric_columns <- check_gazepoint_biometric_columns(biometrics)
  present_biometric_columns <- biometric_columns$column[
    biometric_columns$present &
      biometric_columns$signal %in% c(
        "gsr_eda",
        "heart_rate",
        "engagement_dial",
        "ttl_marker"
      )
  ]

  keep_bio <- unique(c(by, present_biometric_columns))
  bio_for_join <- biometrics[keep_bio]

  out <- merge(
    gaze,
    bio_for_join,
    by = by,
    all.x = all_x,
    all.y = FALSE,
    sort = FALSE,
    suffixes = suffixes
  )

  attr(out, "sync_summary") <- data.frame(
    n_gaze_rows = nrow(gaze),
    n_biometric_rows = nrow(biometrics),
    n_output_rows = nrow(out),
    join_keys = paste(by, collapse = ","),
    all_x = all_x,
    biometric_columns_joined = paste(setdiff(names(bio_for_join), by), collapse = ","),
    stringsAsFactors = FALSE
  )

  class(out) <- c("gazepoint_biometrics_sync", class(out))

  out
}


#' Join Gazepoint Biometrics to a master table
#'
#' Convenience wrapper around `sync_gazepoint_biometrics_with_gaze()` for joining
#' biometric data to a `gp3tools`-style master table or any other analysis-ready
#' gaze table.
#'
#' @param master A master gaze or analysis table.
#' @param biometrics A Gazepoint Biometrics data frame.
#' @param by Character vector of key columns used for joining.
#' @param all_x Logical. Should all rows from `master` be retained?
#'
#' @return A data frame with biometric columns joined to the master table.
#'
#' @export
join_gazepoint_biometrics_to_master <- function(master,
                                                biometrics,
                                                by,
                                                all_x = TRUE) {
  sync_gazepoint_biometrics_with_gaze(
    biometrics = biometrics,
    gaze = master,
    by = by,
    all_x = all_x
  )
}

test_that("sync_gazepoint_biometrics_with_gaze joins biometric columns to gaze data", {
  biometrics <- data.frame(
    USER = c("U1", "U1", "U2"),
    MEDIA_ID = c(1, 2, 1),
    CNT = c(10, 20, 10),
    GSR_US = c(2.0, 2.2, 1.5),
    GSRV = c(1, 1, 1),
    HR = c(75, 76, 80),
    HRV = c(1, 1, 1),
    DIAL = c(0.1, 0.2, 0.3),
    DIALV = c(1, 1, 1)
  )

  gaze <- data.frame(
    USER = c("U1", "U2"),
    MEDIA_ID = c(1, 1),
    CNT = c(10, 10),
    BPOGX = c(0.5, 0.6),
    BPOGY = c(0.4, 0.3)
  )

  out <- sync_gazepoint_biometrics_with_gaze(
    biometrics = biometrics,
    gaze = gaze,
    by = c("USER", "MEDIA_ID", "CNT")
  )

  expect_s3_class(out, "gazepoint_biometrics_sync")
  expect_equal(nrow(out), 2)
  expect_true("GSR_US" %in% names(out))
  expect_true("HR" %in% names(out))
  expect_true("DIAL" %in% names(out))
  expect_true("BPOGX" %in% names(out))

  summary <- attr(out, "sync_summary")
  expect_true(is.data.frame(summary))
  expect_equal(summary$n_gaze_rows, 2)
  expect_equal(summary$n_biometric_rows, 3)
  expect_equal(summary$n_output_rows, 2)
})


test_that("sync_gazepoint_biometrics_with_gaze preserves unmatched gaze rows when all_x is TRUE", {
  biometrics <- data.frame(
    USER = "U1",
    MEDIA_ID = 1,
    CNT = 10,
    GSR_US = 2.0,
    GSRV = 1
  )

  gaze <- data.frame(
    USER = c("U1", "U1"),
    MEDIA_ID = c(1, 1),
    CNT = c(10, 11),
    BPOGX = c(0.5, 0.6)
  )

  out <- sync_gazepoint_biometrics_with_gaze(
    biometrics = biometrics,
    gaze = gaze,
    by = c("USER", "MEDIA_ID", "CNT"),
    all_x = TRUE
  )

  expect_equal(nrow(out), 2)
  expect_true(any(is.na(out$GSR_US)))
})


test_that("join_gazepoint_biometrics_to_master wraps sync function", {
  biometrics <- data.frame(
    USER = "U1",
    MEDIA_ID = 1,
    GSR_US = 2.0,
    GSRV = 1,
    HR = 75,
    HRV = 1
  )

  master <- data.frame(
    USER = "U1",
    MEDIA_ID = 1,
    dwell_time = 1200
  )

  out <- join_gazepoint_biometrics_to_master(
    master = master,
    biometrics = biometrics,
    by = c("USER", "MEDIA_ID")
  )

  expect_s3_class(out, "gazepoint_biometrics_sync")
  expect_true("dwell_time" %in% names(out))
  expect_true("GSR_US" %in% names(out))
  expect_true("HR" %in% names(out))
})


test_that("sync_gazepoint_biometrics_with_gaze rejects missing biometric join keys", {
  biometrics <- data.frame(
    USER = "U1",
    GSR_US = 2.0
  )

  gaze <- data.frame(
    USER = "U1",
    MEDIA_ID = 1
  )

  expect_error(
    sync_gazepoint_biometrics_with_gaze(
      biometrics = biometrics,
      gaze = gaze,
      by = c("USER", "MEDIA_ID")
    ),
    "biometrics"
  )
})


test_that("sync_gazepoint_biometrics_with_gaze rejects missing gaze join keys", {
  biometrics <- data.frame(
    USER = "U1",
    MEDIA_ID = 1,
    GSR_US = 2.0
  )

  gaze <- data.frame(
    USER = "U1",
    BPOGX = 0.5
  )

  expect_error(
    sync_gazepoint_biometrics_with_gaze(
      biometrics = biometrics,
      gaze = gaze,
      by = c("USER", "MEDIA_ID")
    ),
    "gaze"
  )
})

#' Create a Gazepoint EDA analysis pipeline guide
#'
#' Creates a structured six-phase analysis pipeline for Gazepoint
#' Biometrics EDA/GSR workflows. The helper maps each phase to native
#' `gpbiometrics` functions, optional external-method bridges, and optional
#' downstream model templates for `brms` and `lme4`.
#'
#' This function does not fit statistical models, does not run external
#' software, and does not infer emotion, valence, stress, trust, preference,
#' cognition, or diagnosis. It is a reproducible planning and reporting aid.
#'
#' @param include_external_bridges Logical. If `TRUE`, include NeuroKit2,
#'   Ledalab-style, PsPM-style, cvxEDA-style, RHRV, and pyPPG bridge helpers.
#' @param include_model_templates Logical. If `TRUE`, include text templates
#'   for downstream `brms` hurdle models and `lme4` mixed-effects models.
#' @param include_reporting_guidance Logical. If `TRUE`, include reporting and
#'   interpretation guardrails.
#' @param style Output style. `"compact"` returns concise phase descriptions;
#'   `"detailed"` returns fuller phase notes.
#'
#' @return A list with `overview`, `phases`, `function_map`,
#'   `model_templates`, `reporting_guidance`, `interpretation_guardrails`, and
#'   `settings`.
#' @export
create_gazepoint_eda_analysis_pipeline <- function(include_external_bridges = TRUE,
                                                   include_model_templates = TRUE,
                                                   include_reporting_guidance = TRUE,
                                                   style = c("compact", "detailed")) {
  style <- match.arg(style)

  gpbiometrics_eda_pipeline_validate_logical(
    include_external_bridges,
    "include_external_bridges"
  )
  gpbiometrics_eda_pipeline_validate_logical(
    include_model_templates,
    "include_model_templates"
  )
  gpbiometrics_eda_pipeline_validate_logical(
    include_reporting_guidance,
    "include_reporting_guidance"
  )

  phases <- gpbiometrics_eda_pipeline_phases(style = style)
  function_map <- gpbiometrics_eda_pipeline_function_map(
    include_external_bridges = include_external_bridges
  )

  model_templates <- if (isTRUE(include_model_templates)) {
    gpbiometrics_eda_pipeline_model_templates()
  } else {
    data.frame(
      model_family = character(),
      package = character(),
      target_outcome = character(),
      template = character(),
      notes = character(),
      stringsAsFactors = FALSE
    )
  }

  reporting_guidance <- if (isTRUE(include_reporting_guidance)) {
    gpbiometrics_eda_pipeline_reporting_guidance()
  } else {
    data.frame(
      topic = character(),
      guidance = character(),
      stringsAsFactors = FALSE
    )
  }

  interpretation_guardrails <- gpbiometrics_eda_pipeline_guardrails()

  overview <- data.frame(
    phase_count = nrow(phases),
    function_rows = nrow(function_map),
    model_template_count = nrow(model_templates),
    reporting_guidance_rows = nrow(reporting_guidance),
    external_bridges_included = include_external_bridges,
    model_templates_included = include_model_templates,
    status = "eda_analysis_pipeline_created",
    interpretation = paste(
      "This object is a planning and reporting aid for Gazepoint EDA/GSR workflows.",
      "It does not fit models, run external software, or infer emotion, valence, stress, trust, preference, cognition, or diagnosis."
    ),
    stringsAsFactors = FALSE
  )

  structure(
    list(
      overview = overview,
      phases = phases,
      function_map = function_map,
      model_templates = model_templates,
      reporting_guidance = reporting_guidance,
      interpretation_guardrails = interpretation_guardrails,
      settings = list(
        include_external_bridges = include_external_bridges,
        include_model_templates = include_model_templates,
        include_reporting_guidance = include_reporting_guidance,
        style = style
      )
    ),
    class = c("gazepoint_eda_analysis_pipeline", "list")
  )
}

gpbiometrics_eda_pipeline_phases <- function(style = "compact") {
  compact <- data.frame(
    phase = 1:6,
    phase_name = c(
      "Data ingestion, auditing, and quality control",
      "Baseline preprocessing and peak detection",
      "Advanced signal decomposition through external-method bridges",
      "Synchronization and model formatting",
      "Statistical modelling outside gpbiometrics",
      "Reproducible reporting"
    ),
    purpose = c(
      "Import Gazepoint exports, detect schemas and active channels, and audit missingness, time resets, inactive signals, and signal quality.",
      "Apply conservative unit conversion, smoothing, baseline correction, descriptive SCR peak extraction, event-window summaries, and response-pattern classification.",
      "Prepare cleaned EDA data for established external tools when SCR overlap or close stimulus spacing requires deconvolution or external decomposition.",
      "Align physiological signals to TTL/event markers, audit lag and drift, and prepare model-ready repeated-measures tables.",
      "Use prepared data with external statistical packages such as brms or lme4 for zero-heavy SCR amplitude and repeated-measures SCL/EDA models.",
      "Create report bundles, report-ready tables, cautious methods text, dashboards, and reproducible analysis documentation."
    ),
    native_gpbiometrics = c(TRUE, TRUE, FALSE, TRUE, FALSE, TRUE),
    stringsAsFactors = FALSE
  )

  if (identical(style, "compact")) {
    return(compact)
  }

  compact$details <- c(
    paste(
      "Raw Gazepoint biometric exports should be audited before substantive analysis.",
      "The package supports file and folder import, Data Summary parsing, schema detection, active-channel detection, missingness checks, time-reset diagnostics, and signal-activity auditing."
    ),
    paste(
      "Preprocessing is deliberately conservative.",
      "Raw GSR conversion is user-confirmed, baseline correction and smoothing are transparent, and peak/event summaries describe physical EDA responses without psychological labels."
    ),
    paste(
      "Complex EDA decomposition is not reimplemented natively.",
      "Instead, gpbiometrics prepares clean input tables for established external workflows such as NeuroKit2, Ledalab-style, PsPM-style, and cvxEDA-style analysis."
    ),
    paste(
      "Multimodal modelling requires reliable alignment between physiology and events.",
      "The package supports TTL extraction/alignment, lag estimation, drift audits, SCR hurdle-model data preparation, and LME-ready biometric data preparation."
    ),
    paste(
      "Model fitting remains outside gpbiometrics to avoid heavy dependencies.",
      "Templates show how prepared outputs can be used with brms for hurdle-style SCR models and lme4 for mixed-effects SCL/EDA models."
    ),
    paste(
      "Reporting helpers document data quality, preprocessing, exclusions, transformations, and analysis-readiness decisions.",
      "Generated text remains cautious and avoids direct emotion, cognition, valence, stress, trust, or preference inference."
    )
  )

  compact
}

gpbiometrics_eda_pipeline_function_map <- function(include_external_bridges = TRUE) {
  native_rows <- data.frame(
    phase = c(
      rep(1L, 9),
      rep(2L, 7),
      rep(4L, 6),
      rep(6L, 6)
    ),
    function_name = c(
      "import_gazepoint_biometrics",
      "import_gazepoint_biometric_folder",
      "import_gazepoint_data_summary",
      "detect_gazepoint_biometric_schema",
      "detect_gazepoint_time_columns",
      "detect_gazepoint_biometric_timebase",
      "audit_gazepoint_time_resets",
      "audit_gazepoint_signal_activity",
      "audit_gazepoint_biometric_missingness",
      "convert_gazepoint_gsr_to_conductance",
      "smooth_gazepoint_biometrics",
      "baseline_correct_gazepoint_gsr",
      "audit_gazepoint_eda_artifacts",
      "detect_gazepoint_scr_peaks",
      "summarise_gazepoint_scr_event_windows",
      "classify_gazepoint_eda_response_pattern",
      "extract_gazepoint_ttl_events",
      "align_gazepoint_biometrics_to_ttl",
      "estimate_gazepoint_signal_lag",
      "audit_gazepoint_biometric_sync_drift",
      "prepare_gazepoint_scr_hurdle_model_data",
      "prepare_gazepoint_biometrics_lme_data",
      "export_gazepoint_biometrics_report_bundle",
      "create_gazepoint_biometrics_report_tables",
      "create_gazepoint_biometrics_methods_text",
      "plot_gazepoint_biometric_report_dashboard",
      "run_gazepoint_biometrics_real_data_readiness",
      "create_gazepoint_biometrics_feature_inventory"
    ),
    role = c(
      "Import single Gazepoint biometric export.",
      "Import a folder of Gazepoint biometric exports.",
      "Parse Gazepoint Data Summary exports.",
      "Detect Gazepoint biometric schema.",
      "Detect candidate time or counter columns.",
      "Detect biometric timebase.",
      "Audit grouped time/counter resets.",
      "Audit inactive, all-zero, or low-variance channels.",
      "Audit biometric missingness.",
      "Conservatively convert verified resistance-like GSR to conductance.",
      "Smooth selected biometric time-series columns.",
      "Apply GSR/EDA baseline correction.",
      "Flag EDA artefacts and corrupted segments.",
      "Detect descriptive SCR peaks.",
      "Summarise SCR responses in event windows.",
      "Classify descriptive EDA response patterns without psychological labels.",
      "Extract TTL/event marker rows.",
      "Align biometric data to TTL events.",
      "Estimate descriptive lag between two recorded signals.",
      "Audit synchronization drift descriptively.",
      "Prepare zero-heavy SCR response data for hurdle-style models.",
      "Prepare repeated-measures biometric data for mixed-effects models.",
      "Export report bundle outputs.",
      "Create report-ready tables.",
      "Generate cautious methods text.",
      "Create lightweight QC dashboard.",
      "Run real-data readiness checks.",
      "Summarise package feature availability."
    ),
    native_gpbiometrics = TRUE,
    external_dependency_required = FALSE,
    stringsAsFactors = FALSE
  )

  if (!isTRUE(include_external_bridges)) {
    native_rows$available <- gpbiometrics_eda_pipeline_available(native_rows$function_name)
    return(native_rows)
  }

  bridge_rows <- data.frame(
    phase = rep(3L, 5),
    function_name = c(
      "prepare_gazepoint_neurokit_eda_input",
      "prepare_gazepoint_cvxeda_input",
      "prepare_gazepoint_ledalab_input",
      "prepare_gazepoint_pspm_input",
      "run_gazepoint_neurokit_eda_crosscheck"
    ),
    role = c(
      "Prepare cleaned EDA input for optional NeuroKit2-style workflows.",
      "Prepare cleaned EDA input for optional cvxEDA-style workflows.",
      "Prepare cleaned EDA input for optional Ledalab-style workflows.",
      "Prepare cleaned EDA input for optional PsPM-style workflows.",
      "Optionally cross-check EDA processing with NeuroKit2 when available."
    ),
    native_gpbiometrics = TRUE,
    external_dependency_required = c(FALSE, FALSE, FALSE, FALSE, TRUE),
    stringsAsFactors = FALSE
  )

  out <- rbind(native_rows, bridge_rows)
  out <- out[order(out$phase, out$function_name), , drop = FALSE]
  rownames(out) <- NULL
  out$available <- gpbiometrics_eda_pipeline_available(out$function_name)

  out
}

gpbiometrics_eda_pipeline_model_templates <- function() {
  data.frame(
    model_family = c(
      "SCR occurrence and amplitude",
      "SCL or mean EDA level",
      "EDA response pattern"
    ),
    package = c("brms", "lme4", "lme4 or brms"),
    target_outcome = c(
      "Zero-heavy SCR amplitude or response magnitude",
      "Continuous skin conductance level or baseline-corrected EDA",
      "Descriptive response-pattern categories"
    ),
    template = c(
      paste(
        "brms::brm(",
        "  bf(scr_amplitude ~ condition + trial_order + (1 | participant),",
        "     hu ~ condition + trial_order + (1 | participant)),",
        "  data = scr_hurdle_data,",
        "  family = hurdle_lognormal()",
        ")",
        sep = "\n"
      ),
      paste(
        "lme4::lmer(",
        "  scl_mean ~ condition + trial_order + baseline_scl + (1 | participant),",
        "  data = lme_data",
        ")",
        sep = "\n"
      ),
      paste(
        "lme4::glmer(",
        "  response_detected ~ condition + trial_order + (1 | participant),",
        "  data = response_pattern_data,",
        "  family = binomial()",
        ")",
        sep = "\n"
      )
    ),
    notes = c(
      paste(
        "Template only. Use when zero responses are substantively meaningful.",
        "Check distributional assumptions, priors, convergence, posterior predictive checks, and trial/participant nesting."
      ),
      paste(
        "Template only. Use for continuous background arousal/SCL-style summaries.",
        "Interpret with baseline, artefact, task, and participant-level context."
      ),
      paste(
        "Template only. Response-pattern labels are descriptive and should not be interpreted as emotion, valence, stress, trust, preference, cognition, or diagnosis."
      )
    ),
    stringsAsFactors = FALSE
  )
}

gpbiometrics_eda_pipeline_reporting_guidance <- function() {
  data.frame(
    topic = c(
      "Data source",
      "Quality control",
      "Preprocessing",
      "EDA/SCR extraction",
      "External methods",
      "Synchronization",
      "Statistical modelling",
      "Interpretation"
    ),
    guidance = c(
      "Report Gazepoint export type, software/export context when known, imported files, detected channels, time columns, and timebase.",
      "Report missingness, inactive channels, all-zero or flat channels, time resets, sampling issues, artefact flags, and exclusion/review thresholds.",
      "Report conductance conversion only when unit assumptions are explicit; report smoothing and baseline-correction settings.",
      "Report SCR peak/event-window rules, thresholds, response windows, and whether zero responses were retained as informative outcomes.",
      "Report external tools as optional post-export workflows; do not imply that gpbiometrics reimplemented Ledalab, PsPM, cvxEDA, NeuroKit2, RHRV, or pyPPG.",
      "Report TTL/event alignment, lag checks, drift checks, and how timing windows were defined.",
      "Report model family, link/distribution, random effects, participant/trial nesting, covariates, convergence checks, and treatment of zero responses.",
      "Avoid direct inference from EDA to emotion, valence, trust, preference, cognition, or diagnosis without converging task, behavioural, and self-report evidence."
    ),
    stringsAsFactors = FALSE
  )
}

gpbiometrics_eda_pipeline_guardrails <- function() {
  data.frame(
    signal_or_method = c(
      "GSR/EDA",
      "SCR peaks",
      "Response patterns",
      "External decomposition",
      "Heart rate",
      "IBI-derived HRV",
      "Eye-tracking measures",
      "Timing diagnostics"
    ),
    conservative_interpretation = c(
      "Electrodermal or arousal-related signal; not emotional valence.",
      "Discrete conductance responses under specified detection thresholds; not direct evidence of emotion.",
      "Descriptive QC/reporting labels only.",
      "External method input or cross-check; not a native clone or replacement.",
      "Requires baseline, artefact, and task context.",
      "Compute only from genuine IBI/RR intervals, not raw Gazepoint HRV validity/vendor columns.",
      "Visual attention and timing indicators, not direct cognition, scrutiny, or evaluation.",
      "Synchronization QC only; not causal timing or true physiological latency."
    ),
    avoid_claiming = c(
      "Emotion, valence, stress, trust, preference, cognition, diagnosis.",
      "Emotion, valence, stress, trust, preference, cognition, diagnosis.",
      "Emotion class, valence class, trust state, diagnostic state.",
      "That gpbiometrics reimplemented Ledalab, PsPM, cvxEDA, NeuroKit2, RHRV, or pyPPG.",
      "Task engagement or affect without context.",
      "HRV from raw Gazepoint HRV columns unless independently documented.",
      "Deep cognition or evaluative processing from gaze alone.",
      "Physiological latency or causal ordering without design support."
    ),
    stringsAsFactors = FALSE
  )
}

gpbiometrics_eda_pipeline_available <- function(function_name) {
  vapply(function_name, function(fn) {
    exists(fn, mode = "function", envir = parent.frame(2), inherits = TRUE) ||
      exists(fn, mode = "function", envir = asNamespace("gpbiometrics"), inherits = FALSE)
  }, logical(1))
}

gpbiometrics_eda_pipeline_validate_logical <- function(x, name) {
  if (!is.logical(x) || length(x) != 1 || is.na(x)) {
    stop("`", name, "` must be TRUE or FALSE.", call. = FALSE)
  }

  invisible(TRUE)
}

test_that("create_gazepoint_eda_analysis_pipeline returns six phases", {
  out <- create_gazepoint_eda_analysis_pipeline()

  expect_s3_class(out, "gazepoint_eda_analysis_pipeline")
  expect_equal(out$overview$status, "eda_analysis_pipeline_created")
  expect_equal(out$overview$phase_count, 6)
  expect_equal(nrow(out$phases), 6)
  expect_true(all(c("phase", "phase_name", "purpose") %in% names(out$phases)))
  expect_true(is.data.frame(out$function_map))
  expect_true(is.data.frame(out$model_templates))
  expect_true(is.data.frame(out$reporting_guidance))
  expect_true(is.data.frame(out$interpretation_guardrails))
})

test_that("create_gazepoint_eda_analysis_pipeline includes expected phase helpers", {
  out <- create_gazepoint_eda_analysis_pipeline()

  expected <- c(
    "import_gazepoint_biometrics",
    "audit_gazepoint_time_resets",
    "audit_gazepoint_signal_activity",
    "audit_gazepoint_eda_artifacts",
    "detect_gazepoint_scr_peaks",
    "summarise_gazepoint_scr_event_windows",
    "classify_gazepoint_eda_response_pattern",
    "prepare_gazepoint_cvxeda_input",
    "align_gazepoint_biometrics_to_ttl",
    "estimate_gazepoint_signal_lag",
    "prepare_gazepoint_scr_hurdle_model_data",
    "prepare_gazepoint_biometrics_lme_data",
    "export_gazepoint_biometrics_report_bundle",
    "create_gazepoint_biometrics_methods_text"
  )

  expect_true(all(expected %in% out$function_map$function_name))
  expect_true(all(out$function_map$available))
})

test_that("create_gazepoint_eda_analysis_pipeline can omit bridges and templates", {
  out <- create_gazepoint_eda_analysis_pipeline(
    include_external_bridges = FALSE,
    include_model_templates = FALSE,
    include_reporting_guidance = FALSE
  )

  expect_false("prepare_gazepoint_cvxeda_input" %in% out$function_map$function_name)
  expect_equal(nrow(out$model_templates), 0)
  expect_equal(nrow(out$reporting_guidance), 0)
  expect_false(out$settings$include_external_bridges)
  expect_false(out$settings$include_model_templates)
  expect_false(out$settings$include_reporting_guidance)
})

test_that("create_gazepoint_eda_analysis_pipeline includes conservative model templates", {
  out <- create_gazepoint_eda_analysis_pipeline()

  expect_true(any(out$model_templates$package == "brms"))
  expect_true(any(out$model_templates$package == "lme4"))
  expect_true(any(grepl("hurdle_lognormal", out$model_templates$template)))
  expect_true(any(grepl("lme4::lmer", out$model_templates$template)))

  all_template_text <- paste(out$model_templates$notes, collapse = " ")
  expect_true(grepl("Template only", all_template_text))
})

test_that("create_gazepoint_eda_analysis_pipeline includes interpretation guardrails", {
  out <- create_gazepoint_eda_analysis_pipeline()

  expect_true(any(out$interpretation_guardrails$signal_or_method == "GSR/EDA"))
  expect_true(any(grepl("not emotional valence", out$interpretation_guardrails$conservative_interpretation)))
  expect_true(any(grepl("not direct cognition", out$interpretation_guardrails$conservative_interpretation)))
  expect_true(any(grepl("not causal timing", out$interpretation_guardrails$conservative_interpretation)))
})

test_that("create_gazepoint_eda_analysis_pipeline validates logical arguments", {
  expect_error(
    create_gazepoint_eda_analysis_pipeline(include_external_bridges = NA),
    "include_external_bridges"
  )

  expect_error(
    create_gazepoint_eda_analysis_pipeline(include_model_templates = "yes"),
    "include_model_templates"
  )

  expect_error(
    create_gazepoint_eda_analysis_pipeline(include_reporting_guidance = c(TRUE, FALSE)),
    "include_reporting_guidance"
  )
})

#' Correct EDA for ambient or body temperature
#'
#' Regresses an EDA/conductance signal on one or more continuous temperature
#' covariates and returns a temperature-adjusted residual series. The adjusted
#' signal is temperature-adjusted EDA, not "pure" cognitive or emotional EDA.
#'
#' @param dat A data frame.
#' @param eda_col Numeric EDA/conductance column.
#' @param temperature_cols One or more numeric temperature columns.
#' @param group_cols Optional grouping columns.
#' @param time_col Optional time column retained in summaries.
#' @param output_col Output residual-adjusted EDA column.
#' @param fitted_col Output fitted temperature component column.
#' @param model_by_group Logical. If `TRUE`, fit one model per group.
#' @param add_intercept_mean Logical. If `TRUE`, add the group mean EDA back to
#'   residuals so the adjusted signal remains on the original scale.
#'
#' @return A data frame with adjusted EDA columns and model-summary attributes.
#' @export
correct_gazepoint_eda_temperature <- function(dat,
                                              eda_col = "GSR_US",
                                              temperature_cols,
                                              group_cols = NULL,
                                              time_col = NULL,
                                              output_col = "eda_temperature_adjusted",
                                              fitted_col = "eda_temperature_fitted",
                                              model_by_group = TRUE,
                                              add_intercept_mean = TRUE) {
  if (!is.data.frame(dat)) {
    stop("`dat` must be a data frame.", call. = FALSE)
  }

  required <- c(eda_col, temperature_cols, group_cols, time_col)
  required <- required[!is.null(required)]

  missing_required <- setdiff(required, names(dat))
  if (length(missing_required) > 0) {
    stop("Missing required columns: ", paste(missing_required, collapse = ", "), call. = FALSE)
  }

  numeric_required <- c(eda_col, temperature_cols, time_col)
  numeric_required <- numeric_required[!is.null(numeric_required)]

  non_numeric <- numeric_required[!vapply(dat[numeric_required], is.numeric, logical(1))]
  if (length(non_numeric) > 0) {
    stop("These columns must be numeric: ", paste(non_numeric, collapse = ", "), call. = FALSE)
  }

  if (is.null(group_cols)) {
    group_cols <- character()
  }

  out <- dat
  out[[output_col]] <- NA_real_
  out[[fitted_col]] <- NA_real_
  out$eda_temperature_correction_status <- "not_processed"

  groups <- if (isTRUE(model_by_group)) {
    gpbiometrics_11y_split(dat, group_cols)
  } else {
    list(all_rows = seq_len(nrow(dat)))
  }

  summary_rows <- list()

  for (group_id in names(groups)) {
    idx <- groups[[group_id]]
    d <- out[idx, c(eda_col, temperature_cols), drop = FALSE]
    names(d) <- c(".eda", paste0(".temp", seq_along(temperature_cols)))

    complete <- stats::complete.cases(d)

    if (sum(complete) < max(5, length(temperature_cols) + 2)) {
      out$eda_temperature_correction_status[idx] <- "insufficient_complete_cases"

      summary_rows[[group_id]] <- data.frame(
        group_id = group_id,
        n_rows = length(idx),
        n_complete = sum(complete),
        r_squared = NA_real_,
        status = "insufficient_complete_cases",
        stringsAsFactors = FALSE
      )
      next
    }

    fit <- stats::lm(.eda ~ ., data = d[complete, , drop = FALSE])

    fitted_values <- rep(NA_real_, length(idx))
    fitted_values[complete] <- stats::predict(fit, newdata = d[complete, , drop = FALSE])

    residual_values <- out[[eda_col]][idx] - fitted_values

    if (isTRUE(add_intercept_mean)) {
      residual_values <- residual_values + mean(out[[eda_col]][idx][complete], na.rm = TRUE)
    }

    out[[output_col]][idx] <- residual_values
    out[[fitted_col]][idx] <- fitted_values
    out$eda_temperature_correction_status[idx] <- ifelse(
      complete,
      "temperature_adjusted",
      "incomplete_temperature_or_eda"
    )

    fit_summary <- summary(fit)

    summary_rows[[group_id]] <- data.frame(
      group_id = group_id,
      n_rows = length(idx),
      n_complete = sum(complete),
      r_squared = fit_summary$r.squared,
      adjusted_r_squared = fit_summary$adj.r.squared,
      residual_sd = stats::sd(stats::residuals(fit), na.rm = TRUE),
      status = "temperature_model_fitted",
      stringsAsFactors = FALSE
    )
  }

  model_summary <- do.call(rbind, summary_rows)
  rownames(model_summary) <- NULL

  overview <- data.frame(
    input_rows = nrow(dat),
    output_rows = nrow(out),
    group_count = length(groups),
    successful_groups = sum(model_summary$status == "temperature_model_fitted"),
    problem_groups = sum(model_summary$status != "temperature_model_fitted"),
    status = if (all(model_summary$status == "temperature_model_fitted")) {
      "eda_temperature_correction_complete"
    } else if (any(model_summary$status == "temperature_model_fitted")) {
      "eda_temperature_correction_partial"
    } else {
      "eda_temperature_correction_failed"
    },
    interpretation = paste(
      "The output is temperature-adjusted EDA.",
      "Residualisation does not make the signal purely cognitive, emotional, or sympathetic."
    ),
    stringsAsFactors = FALSE
  )

  attr(out, "eda_temperature_overview") <- overview
  attr(out, "eda_temperature_model_summary") <- model_summary
  attr(out, "eda_temperature_settings") <- list(
    eda_col = eda_col,
    temperature_cols = temperature_cols,
    group_cols = group_cols,
    time_col = time_col,
    output_col = output_col,
    fitted_col = fitted_col,
    model_by_group = model_by_group,
    add_intercept_mean = add_intercept_mean
  )

  class(out) <- unique(c("gazepoint_eda_temperature_corrected", class(out)))
  out
}

#' Extract heartbeat candidates from Gazepoint pulse using k-means
#'
#' Uses k-means clustering on the raw pulse waveform to classify likely
#' heartbeat regions and then selects local extrema as beat candidates. This is
#' a Gazepoint Biometrics-oriented fallback for difficult pulse waveforms, not
#' an ECG-equivalent R-peak detector.
#'
#' @param dat A data frame.
#' @param pulse_col Numeric pulse/PPG column.
#' @param time_col Numeric time column.
#' @param group_cols Optional grouping columns.
#' @param k Number of k-means clusters.
#' @param peak_polarity `"positive"` or `"negative"`.
#' @param min_distance_s Minimum time between selected beats.
#' @param sampling_rate Optional sampling rate in Hz.
#' @param seed Optional random seed.
#'
#' @return A list with `overview`, `beat_table`, `interval_table`,
#'   `timeseries`, and `settings`.
#' @export
extract_gazepoint_beats_kmeans <- function(dat,
                                           pulse_col = "HRP",
                                           time_col = "CNT",
                                           group_cols = NULL,
                                           k = 2,
                                           peak_polarity = c("positive", "negative"),
                                           min_distance_s = 0.30,
                                           sampling_rate = NULL,
                                           seed = NULL) {
  if (!is.data.frame(dat)) {
    stop("`dat` must be a data frame.", call. = FALSE)
  }

  peak_polarity <- match.arg(peak_polarity)

  required <- c(pulse_col, time_col, group_cols)
  required <- required[!is.null(required)]

  missing_required <- setdiff(required, names(dat))
  if (length(missing_required) > 0) {
    stop("Missing required columns: ", paste(missing_required, collapse = ", "), call. = FALSE)
  }

  if (!is.numeric(dat[[pulse_col]])) {
    stop("`pulse_col` must identify a numeric column.", call. = FALSE)
  }

  if (!is.numeric(dat[[time_col]])) {
    stop("`time_col` must identify a numeric column.", call. = FALSE)
  }

  if (!is.null(seed)) {
    set.seed(seed)
  }

  if (is.null(group_cols)) {
    group_cols <- character()
  }

  groups <- gpbiometrics_11y_split(dat, group_cols)

  beat_rows <- list()
  interval_rows <- list()
  timeseries_rows <- list()
  summary_rows <- list()
  beat_id <- 1L
  interval_id <- 1L

  for (group_id in names(groups)) {
    idx <- groups[[group_id]]
    idx <- idx[order(dat[[time_col]][idx])]

    time <- dat[[time_col]][idx]
    pulse <- dat[[pulse_col]][idx]
    finite <- is.finite(time) & is.finite(pulse)

    fs <- gpbiometrics_11y_sampling_rate(time, sampling_rate)

    status <- rep("not_beat_candidate", length(idx))
    cluster <- rep(NA_integer_, length(idx))

    if (sum(finite) < max(10, k * 3) || length(unique(pulse[finite])) < k) {
      timeseries_rows[[group_id]] <- data.frame(
        row_index = idx,
        group_id = group_id,
        time = time,
        pulse = pulse,
        beat_cluster = cluster,
        beat_candidate = FALSE,
        status = "insufficient_pulse_variability",
        stringsAsFactors = FALSE
      )

      summary_rows[[group_id]] <- data.frame(
        group_id = group_id,
        n_rows = length(idx),
        beat_count = 0L,
        mean_ibi_s = NA_real_,
        mean_hr_bpm = NA_real_,
        sampling_rate_hz = fs,
        status = "insufficient_pulse_variability",
        stringsAsFactors = FALSE
      )
      next
    }

    km <- stats::kmeans(pulse[finite], centers = k, nstart = 10)
    centers <- as.numeric(km$centers)

    beat_cluster_id <- if (peak_polarity == "positive") {
      which.max(centers)
    } else {
      which.min(centers)
    }

    cluster[finite] <- km$cluster
    candidate_local <- which(cluster == beat_cluster_id)

    selected_local <- gpbiometrics_11y_select_kmeans_beats(
      candidate = candidate_local,
      time = time,
      pulse = pulse,
      polarity = peak_polarity,
      min_distance_s = min_distance_s
    )

    beat_flag <- rep(FALSE, length(idx))
    beat_flag[selected_local] <- TRUE

    status[cluster == beat_cluster_id] <- "cluster_candidate"
    status[selected_local] <- "selected_beat"

    beat_times <- time[selected_local]
    beat_amplitudes <- pulse[selected_local]
    ibi <- diff(beat_times)

    for (i in seq_along(beat_times)) {
      beat_rows[[beat_id]] <- data.frame(
        group_id = group_id,
        beat_index = i,
        row_index = idx[selected_local[i]],
        beat_time = beat_times[i],
        pulse_amplitude = beat_amplitudes[i],
        method = "kmeans_pulse_classification",
        stringsAsFactors = FALSE
      )
      beat_id <- beat_id + 1L
    }

    if (length(ibi) > 0) {
      for (i in seq_along(ibi)) {
        interval_rows[[interval_id]] <- data.frame(
          group_id = group_id,
          interval_index = i,
          start_time = beat_times[i],
          end_time = beat_times[i + 1],
          ibi_s = ibi[i],
          hr_bpm = if (ibi[i] > 0) 60 / ibi[i] else NA_real_,
          stringsAsFactors = FALSE
        )
        interval_id <- interval_id + 1L
      }
    }

    timeseries_rows[[group_id]] <- data.frame(
      row_index = idx,
      group_id = group_id,
      time = time,
      pulse = pulse,
      beat_cluster = cluster,
      beat_candidate = beat_flag,
      status = status,
      stringsAsFactors = FALSE
    )

    summary_rows[[group_id]] <- data.frame(
      group_id = group_id,
      n_rows = length(idx),
      beat_count = length(beat_times),
      mean_ibi_s = if (length(ibi) > 0) mean(ibi, na.rm = TRUE) else NA_real_,
      mean_hr_bpm = if (length(ibi) > 0) 60 / mean(ibi, na.rm = TRUE) else NA_real_,
      sampling_rate_hz = fs,
      status = if (length(beat_times) >= 2) {
        "kmeans_beats_extracted"
      } else {
        "too_few_beats_detected"
      },
      stringsAsFactors = FALSE
    )
  }

  beat_table <- if (length(beat_rows) > 0) do.call(rbind, beat_rows) else data.frame()
  interval_table <- if (length(interval_rows) > 0) do.call(rbind, interval_rows) else data.frame()
  timeseries <- do.call(rbind, timeseries_rows)
  summary <- do.call(rbind, summary_rows)
  rownames(summary) <- NULL

  overview <- data.frame(
    group_count = length(groups),
    beat_rows = nrow(beat_table),
    interval_rows = nrow(interval_table),
    successful_groups = sum(summary$status == "kmeans_beats_extracted"),
    problem_groups = sum(summary$status != "kmeans_beats_extracted"),
    status = if (all(summary$status == "kmeans_beats_extracted")) {
      "kmeans_beats_extracted"
    } else if (any(summary$status == "kmeans_beats_extracted")) {
      "kmeans_beats_partial"
    } else {
      "kmeans_beats_failed"
    },
    interpretation = paste(
      "K-means pulse beat candidates are hardware-oriented pulse features.",
      "They are not ECG-equivalent R peaks and should be visually/QC checked before HRV inference."
    ),
    stringsAsFactors = FALSE
  )

  structure(
    list(
      overview = overview,
      beat_table = beat_table,
      interval_table = interval_table,
      timeseries = timeseries,
      summary = summary,
      settings = list(
        pulse_col = pulse_col,
        time_col = time_col,
        group_cols = group_cols,
        k = k,
        peak_polarity = peak_polarity,
        min_distance_s = min_distance_s,
        sampling_rate = sampling_rate,
        seed = seed
      )
    ),
    class = c("gazepoint_kmeans_beats", "list")
  )
}

#' Audit or trim the EDA electrode stabilization period
#'
#' Flags or removes the initial stabilization period in each recording/group.
#' This is intended to prevent early skin-electrode drift from being treated as
#' a stable physiological baseline.
#'
#' @param dat A data frame.
#' @param time_col Numeric time column.
#' @param group_cols Optional grouping columns.
#' @param stabilization_minutes Stabilization duration to flag or trim.
#' @param action `"flag"` or `"trim"`.
#' @param output_col Output logical flag column.
#' @param time_units `"auto"`, `"seconds"`, or `"milliseconds"`.
#'
#' @return A data frame with stabilization-period attributes.
#' @export
audit_gazepoint_stabilization_period <- function(dat,
                                                 time_col = "CNT",
                                                 group_cols = NULL,
                                                 stabilization_minutes = 10,
                                                 action = c("flag", "trim"),
                                                 output_col = "in_stabilization_period",
                                                 time_units = c("auto", "seconds", "milliseconds")) {
  if (!is.data.frame(dat)) {
    stop("`dat` must be a data frame.", call. = FALSE)
  }

  action <- match.arg(action)
  time_units <- match.arg(time_units)

  required <- c(time_col, group_cols)
  required <- required[!is.null(required)]

  missing_required <- setdiff(required, names(dat))
  if (length(missing_required) > 0) {
    stop("Missing required columns: ", paste(missing_required, collapse = ", "), call. = FALSE)
  }

  if (!is.numeric(dat[[time_col]])) {
    stop("`time_col` must identify a numeric column.", call. = FALSE)
  }

  if (is.null(group_cols)) {
    group_cols <- character()
  }

  out <- dat
  out[[output_col]] <- FALSE
  out$stabilization_elapsed_s <- NA_real_
  out$stabilization_audit_status <- "not_processed"

  groups <- gpbiometrics_11y_split(out, group_cols)
  summary_rows <- list()

  cutoff_s <- stabilization_minutes * 60

  for (group_id in names(groups)) {
    idx <- groups[[group_id]]
    time_raw <- out[[time_col]][idx]
    time_s <- gpbiometrics_11y_time_to_seconds(time_raw, time_units)

    start_s <- min(time_s[is.finite(time_s)], na.rm = TRUE)
    elapsed_s <- time_s - start_s

    in_stabilization <- is.finite(elapsed_s) & elapsed_s < cutoff_s

    out[[output_col]][idx] <- in_stabilization
    out$stabilization_elapsed_s[idx] <- elapsed_s
    out$stabilization_audit_status[idx] <- ifelse(
      in_stabilization,
      "within_stabilization_period",
      "after_stabilization_period"
    )

    summary_rows[[group_id]] <- data.frame(
      group_id = group_id,
      n_rows = length(idx),
      stabilization_rows = sum(in_stabilization, na.rm = TRUE),
      retained_rows_after_stabilization = sum(!in_stabilization, na.rm = TRUE),
      stabilization_minutes = stabilization_minutes,
      start_time_s = start_s,
      cutoff_time_s = start_s + cutoff_s,
      status = "stabilization_period_audited",
      stringsAsFactors = FALSE
    )
  }

  audit_summary <- do.call(rbind, summary_rows)
  rownames(audit_summary) <- NULL

  if (action == "trim") {
    out <- out[!out[[output_col]], , drop = FALSE]
  }

  overview <- data.frame(
    input_rows = nrow(dat),
    output_rows = nrow(out),
    group_count = length(groups),
    stabilization_minutes = stabilization_minutes,
    action = action,
    status = "stabilization_period_audited",
    interpretation = paste(
      "The stabilization flag identifies the initial skin-electrode adaptation period.",
      "Flagging does not prove that all early data are invalid; trimming should match the study protocol."
    ),
    stringsAsFactors = FALSE
  )

  attr(out, "stabilization_overview") <- overview
  attr(out, "stabilization_summary") <- audit_summary
  attr(out, "stabilization_settings") <- list(
    time_col = time_col,
    group_cols = group_cols,
    stabilization_minutes = stabilization_minutes,
    action = action,
    output_col = output_col,
    time_units = time_units
  )

  class(out) <- unique(c("gazepoint_stabilization_audit", class(out)))
  out
}

#' Regress stimulus luminance from pupil diameter
#'
#' Regresses continuous pupil diameter on frame-wise or sample-wise stimulus
#' luminance and returns a luminance-adjusted pupil series. This controls a
#' major visual confound but does not prove that residual pupil changes are
#' cognitive-load-only effects.
#'
#' @param dat A data frame.
#' @param pupil_col Numeric pupil column.
#' @param luminance_col Numeric luminance/brightness column.
#' @param group_cols Optional grouping columns.
#' @param time_col Optional time column.
#' @param output_col Output luminance-adjusted pupil column.
#' @param fitted_col Output fitted luminance component column.
#' @param include_quadratic Logical. If `TRUE`, include luminance squared.
#' @param model_by_group Logical. If `TRUE`, fit models per group.
#' @param add_intercept_mean Logical. If `TRUE`, add mean pupil size back to
#'   residuals.
#'
#' @return A data frame with luminance-adjusted pupil columns and attributes.
#' @export
regress_gazepoint_pupil_luminance <- function(dat,
                                              pupil_col,
                                              luminance_col,
                                              group_cols = NULL,
                                              time_col = NULL,
                                              output_col = "pupil_luminance_adjusted",
                                              fitted_col = "pupil_luminance_fitted",
                                              include_quadratic = TRUE,
                                              model_by_group = TRUE,
                                              add_intercept_mean = TRUE) {
  if (!is.data.frame(dat)) {
    stop("`dat` must be a data frame.", call. = FALSE)
  }

  required <- c(pupil_col, luminance_col, group_cols, time_col)
  required <- required[!is.null(required)]

  missing_required <- setdiff(required, names(dat))
  if (length(missing_required) > 0) {
    stop("Missing required columns: ", paste(missing_required, collapse = ", "), call. = FALSE)
  }

  numeric_required <- c(pupil_col, luminance_col, time_col)
  numeric_required <- numeric_required[!is.null(numeric_required)]

  non_numeric <- numeric_required[!vapply(dat[numeric_required], is.numeric, logical(1))]
  if (length(non_numeric) > 0) {
    stop("These columns must be numeric: ", paste(non_numeric, collapse = ", "), call. = FALSE)
  }

  if (is.null(group_cols)) {
    group_cols <- character()
  }

  out <- dat
  out[[output_col]] <- NA_real_
  out[[fitted_col]] <- NA_real_
  out$pupil_luminance_regression_status <- "not_processed"

  groups <- if (isTRUE(model_by_group)) {
    gpbiometrics_11y_split(dat, group_cols)
  } else {
    list(all_rows = seq_len(nrow(dat)))
  }

  summary_rows <- list()

  for (group_id in names(groups)) {
    idx <- groups[[group_id]]

    d <- data.frame(
      pupil = out[[pupil_col]][idx],
      luminance = out[[luminance_col]][idx],
      stringsAsFactors = FALSE
    )

    if (isTRUE(include_quadratic)) {
      d$luminance2 <- d$luminance^2
    }

    complete <- stats::complete.cases(d)

    if (sum(complete) < if (include_quadratic) 5 else 4) {
      out$pupil_luminance_regression_status[idx] <- "insufficient_complete_cases"

      summary_rows[[group_id]] <- data.frame(
        group_id = group_id,
        n_rows = length(idx),
        n_complete = sum(complete),
        r_squared = NA_real_,
        status = "insufficient_complete_cases",
        stringsAsFactors = FALSE
      )
      next
    }

    fit <- stats::lm(pupil ~ ., data = d[complete, , drop = FALSE])

    fitted_values <- rep(NA_real_, length(idx))
    fitted_values[complete] <- stats::predict(fit, newdata = d[complete, , drop = FALSE])

    residual_values <- out[[pupil_col]][idx] - fitted_values

    if (isTRUE(add_intercept_mean)) {
      residual_values <- residual_values + mean(out[[pupil_col]][idx][complete], na.rm = TRUE)
    }

    out[[output_col]][idx] <- residual_values
    out[[fitted_col]][idx] <- fitted_values
    out$pupil_luminance_regression_status[idx] <- ifelse(
      complete,
      "luminance_adjusted",
      "incomplete_pupil_or_luminance"
    )

    fit_summary <- summary(fit)

    summary_rows[[group_id]] <- data.frame(
      group_id = group_id,
      n_rows = length(idx),
      n_complete = sum(complete),
      r_squared = fit_summary$r.squared,
      adjusted_r_squared = fit_summary$adj.r.squared,
      residual_sd = stats::sd(stats::residuals(fit), na.rm = TRUE),
      status = "luminance_model_fitted",
      stringsAsFactors = FALSE
    )
  }

  model_summary <- do.call(rbind, summary_rows)
  rownames(model_summary) <- NULL

  overview <- data.frame(
    input_rows = nrow(dat),
    output_rows = nrow(out),
    group_count = length(groups),
    successful_groups = sum(model_summary$status == "luminance_model_fitted"),
    problem_groups = sum(model_summary$status != "luminance_model_fitted"),
    status = if (all(model_summary$status == "luminance_model_fitted")) {
      "pupil_luminance_regression_complete"
    } else if (any(model_summary$status == "luminance_model_fitted")) {
      "pupil_luminance_regression_partial"
    } else {
      "pupil_luminance_regression_failed"
    },
    interpretation = paste(
      "The output is luminance-adjusted pupil diameter.",
      "Residual pupil variation should not be interpreted as cognitive load alone without design and modelling support."
    ),
    stringsAsFactors = FALSE
  )

  attr(out, "pupil_luminance_overview") <- overview
  attr(out, "pupil_luminance_model_summary") <- model_summary
  attr(out, "pupil_luminance_settings") <- list(
    pupil_col = pupil_col,
    luminance_col = luminance_col,
    group_cols = group_cols,
    time_col = time_col,
    output_col = output_col,
    fitted_col = fitted_col,
    include_quadratic = include_quadratic,
    model_by_group = model_by_group,
    add_intercept_mean = add_intercept_mean
  )

  class(out) <- unique(c("gazepoint_pupil_luminance_adjusted", class(out)))
  out
}

#' Model heartbeat timing using an IPFM-style impulse train
#'
#' Builds an impulse-train representation of heartbeat timing from IBI/RR
#' intervals or beat times and computes a simple spectrum of the resulting
#' impulse train. This is an IPFM-style model-preparation helper, not a perfect
#' reconstruction of sinoatrial-node physiology.
#'
#' @param dat A data frame.
#' @param ibi_col Optional numeric IBI/RR interval column.
#' @param beat_time_col Optional explicit beat-time column.
#' @param group_cols Optional grouping columns.
#' @param ibi_units `"auto"`, `"seconds"`, or `"milliseconds"`.
#' @param output_sampling_rate Sampling rate for regular impulse train in Hz.
#' @param max_frequency Maximum frequency returned in spectrum.
#'
#' @return A list with `overview`, `beat_table`, `impulse_table`,
#'   `spectrum_table`, `summary`, and `settings`.
#' @export
model_gazepoint_hrv_ipfm <- function(dat,
                                     ibi_col = "IBI",
                                     beat_time_col = NULL,
                                     group_cols = NULL,
                                     ibi_units = c("auto", "seconds", "milliseconds"),
                                     output_sampling_rate = 4,
                                     max_frequency = 0.50) {
  if (!is.data.frame(dat)) {
    stop("`dat` must be a data frame.", call. = FALSE)
  }

  ibi_units <- match.arg(ibi_units)

  if (is.null(ibi_col) && is.null(beat_time_col)) {
    stop("Supply either `ibi_col` or `beat_time_col`.", call. = FALSE)
  }

  required <- c(ibi_col, beat_time_col, group_cols)
  required <- required[!is.null(required)]

  missing_required <- setdiff(required, names(dat))
  if (length(missing_required) > 0) {
    stop("Missing required columns: ", paste(missing_required, collapse = ", "), call. = FALSE)
  }

  if (!is.null(ibi_col) && !is.numeric(dat[[ibi_col]])) {
    stop("`ibi_col` must identify a numeric column.", call. = FALSE)
  }

  if (!is.null(beat_time_col) && !is.numeric(dat[[beat_time_col]])) {
    stop("`beat_time_col` must identify a numeric column.", call. = FALSE)
  }

  if (is.null(group_cols)) {
    group_cols <- character()
  }

  groups <- gpbiometrics_11y_split(dat, group_cols)

  beat_rows <- list()
  impulse_rows <- list()
  spectrum_rows <- list()
  summary_rows <- list()
  beat_id <- 1L
  impulse_id <- 1L
  spectrum_id <- 1L

  for (group_id in names(groups)) {
    idx <- groups[[group_id]]

    if (!is.null(beat_time_col)) {
      beat_times <- sort(unique(dat[[beat_time_col]][idx]))
      beat_times <- beat_times[is.finite(beat_times)]
    } else {
      ibi <- dat[[ibi_col]][idx]
      ibi <- ibi[is.finite(ibi) & ibi > 0]
      ibi_s <- gpbiometrics_11y_ibi_to_seconds(ibi, ibi_units)
      beat_times <- cumsum(ibi_s)
    }

    if (length(beat_times) < 3) {
      summary_rows[[group_id]] <- data.frame(
        group_id = group_id,
        beat_count = length(beat_times),
        impulse_rows = 0L,
        dominant_frequency_hz = NA_real_,
        status = "insufficient_beats",
        stringsAsFactors = FALSE
      )
      next
    }

    beat_times <- beat_times - min(beat_times, na.rm = TRUE)
    duration <- max(beat_times, na.rm = TRUE)
    grid <- seq(0, duration, by = 1 / output_sampling_rate)
    impulse <- rep(0, length(grid))

    nearest <- vapply(beat_times, function(bt) {
      which.min(abs(grid - bt))
    }, integer(1))

    impulse[nearest] <- 1

    for (i in seq_along(beat_times)) {
      beat_rows[[beat_id]] <- data.frame(
        group_id = group_id,
        beat_index = i,
        beat_time = beat_times[i],
        stringsAsFactors = FALSE
      )
      beat_id <- beat_id + 1L
    }

    for (i in seq_along(grid)) {
      impulse_rows[[impulse_id]] <- data.frame(
        group_id = group_id,
        time = grid[i],
        impulse = impulse[i],
        stringsAsFactors = FALSE
      )
      impulse_id <- impulse_id + 1L
    }

    spec <- gpbiometrics_11y_ipfm_spectrum(impulse, output_sampling_rate, max_frequency)

    if (nrow(spec) > 0) {
      for (i in seq_len(nrow(spec))) {
        spectrum_rows[[spectrum_id]] <- data.frame(
          group_id = group_id,
          frequency_hz = spec$frequency_hz[i],
          power = spec$power[i],
          stringsAsFactors = FALSE
        )
        spectrum_id <- spectrum_id + 1L
      }
    }

    dominant_frequency <- if (nrow(spec) > 0) {
      spec$frequency_hz[which.max(spec$power)]
    } else {
      NA_real_
    }

    summary_rows[[group_id]] <- data.frame(
      group_id = group_id,
      beat_count = length(beat_times),
      impulse_rows = length(grid),
      dominant_frequency_hz = dominant_frequency,
      status = "ipfm_impulse_train_created",
      stringsAsFactors = FALSE
    )
  }

  beat_table <- if (length(beat_rows) > 0) do.call(rbind, beat_rows) else data.frame()
  impulse_table <- if (length(impulse_rows) > 0) do.call(rbind, impulse_rows) else data.frame()
  spectrum_table <- if (length(spectrum_rows) > 0) do.call(rbind, spectrum_rows) else data.frame()
  summary <- do.call(rbind, summary_rows)
  rownames(summary) <- NULL

  overview <- data.frame(
    group_count = length(groups),
    beat_rows = nrow(beat_table),
    impulse_rows = nrow(impulse_table),
    spectrum_rows = nrow(spectrum_table),
    successful_groups = sum(summary$status == "ipfm_impulse_train_created"),
    problem_groups = sum(summary$status != "ipfm_impulse_train_created"),
    status = if (all(summary$status == "ipfm_impulse_train_created")) {
      "ipfm_model_created"
    } else if (any(summary$status == "ipfm_impulse_train_created")) {
      "ipfm_model_partial"
    } else {
      "ipfm_model_failed"
    },
    interpretation = paste(
      "The impulse train is an IPFM-style heartbeat timing representation.",
      "It is not a perfect reconstruction of sinoatrial-node physiology."
    ),
    stringsAsFactors = FALSE
  )

  structure(
    list(
      overview = overview,
      beat_table = beat_table,
      impulse_table = impulse_table,
      spectrum_table = spectrum_table,
      summary = summary,
      settings = list(
        ibi_col = ibi_col,
        beat_time_col = beat_time_col,
        group_cols = group_cols,
        ibi_units = ibi_units,
        output_sampling_rate = output_sampling_rate,
        max_frequency = max_frequency
      )
    ),
    class = c("gazepoint_hrv_ipfm", "list")
  )
}

gpbiometrics_11y_split <- function(dat, group_cols) {
  if (length(group_cols) == 0) {
    return(list(all_rows = seq_len(nrow(dat))))
  }

  gf <- dat[group_cols]
  gf[] <- lapply(gf, function(x) {
    x <- as.character(x)
    x[is.na(x)] <- "<NA>"
    x
  })

  split(seq_len(nrow(dat)), do.call(paste, c(gf, sep = " | ")))
}

gpbiometrics_11y_sampling_rate <- function(time, sampling_rate = NULL) {
  if (!is.null(sampling_rate)) {
    return(sampling_rate)
  }

  time <- time[is.finite(time)]

  if (length(time) < 3) {
    return(NA_real_)
  }

  dt <- diff(time)
  dt <- dt[is.finite(dt) & dt > 0]

  if (length(dt) == 0) {
    return(NA_real_)
  }

  median_dt <- stats::median(dt)

  if (median_dt > 10) {
    1000 / median_dt
  } else {
    1 / median_dt
  }
}

gpbiometrics_11y_time_to_seconds <- function(time, time_units = "auto") {
  if (time_units == "milliseconds") {
    return(time / 1000)
  }

  if (time_units == "seconds") {
    return(time)
  }

  finite_time <- time[is.finite(time)]

  if (length(finite_time) < 2) {
    return(time)
  }

  dt <- diff(finite_time)
  dt <- dt[is.finite(dt) & dt > 0]

  if (length(dt) > 0 && stats::median(dt) > 10) {
    time / 1000
  } else {
    time
  }
}

gpbiometrics_11y_ibi_to_seconds <- function(ibi, ibi_units = "auto") {
  if (ibi_units == "milliseconds") {
    return(ibi / 1000)
  }

  if (ibi_units == "seconds") {
    return(ibi)
  }

  if (stats::median(ibi, na.rm = TRUE) > 10) {
    ibi / 1000
  } else {
    ibi
  }
}

gpbiometrics_11y_select_kmeans_beats <- function(candidate,
                                                 time,
                                                 pulse,
                                                 polarity,
                                                 min_distance_s) {
  if (length(candidate) == 0) {
    return(integer())
  }

  candidate <- sort(candidate)
  gaps <- c(TRUE, diff(candidate) > 1)
  run_id <- cumsum(gaps)

  selected <- integer()

  for (run in unique(run_id)) {
    run_idx <- candidate[run_id == run]

    if (polarity == "positive") {
      local <- run_idx[which.max(pulse[run_idx])]
    } else {
      local <- run_idx[which.min(pulse[run_idx])]
    }

    selected <- c(selected, local)
  }

  selected <- selected[order(time[selected])]

  if (length(selected) <= 1) {
    return(selected)
  }

  keep <- selected[1]

  for (idx in selected[-1]) {
    last <- keep[length(keep)]

    if ((time[idx] - time[last]) >= min_distance_s) {
      keep <- c(keep, idx)
    } else {
      replace <- if (polarity == "positive") {
        pulse[idx] > pulse[last]
      } else {
        pulse[idx] < pulse[last]
      }

      if (isTRUE(replace)) {
        keep[length(keep)] <- idx
      }
    }
  }

  keep
}

gpbiometrics_11y_ipfm_spectrum <- function(impulse,
                                           sampling_rate,
                                           max_frequency = 0.50) {
  impulse <- impulse - mean(impulse, na.rm = TRUE)

  if (length(impulse) < 8 || stats::sd(impulse) == 0) {
    return(data.frame())
  }

  spec <- stats::spec.pgram(
    impulse,
    taper = 0.1,
    plot = FALSE,
    demean = TRUE,
    detrend = TRUE,
    fast = TRUE
  )

  freq <- spec$freq * sampling_rate
  power <- spec$spec

  keep <- is.finite(freq) & is.finite(power) & freq > 0 & freq <= max_frequency

  data.frame(
    frequency_hz = freq[keep],
    power = power[keep],
    stringsAsFactors = FALSE
  )
}

test_that("correct_gazepoint_eda_temperature returns temperature-adjusted EDA", {
  dat <- data.frame(
    participant = "p1",
    time = seq_len(60),
    GSR_US = 1 + 0.05 * seq_len(60) + rnorm(60, sd = 0.01),
    ambient_temp = 20 + 0.1 * seq_len(60)
  )

  out <- correct_gazepoint_eda_temperature(
    dat,
    eda_col = "GSR_US",
    temperature_cols = "ambient_temp",
    group_cols = "participant",
    time_col = "time"
  )

  expect_s3_class(out, "gazepoint_eda_temperature_corrected")
  expect_true("eda_temperature_adjusted" %in% names(out))
  expect_equal(
    attr(out, "eda_temperature_overview")$status,
    "eda_temperature_correction_complete"
  )
})

test_that("extract_gazepoint_beats_kmeans detects pulse beat candidates", {
  time <- seq(0, 20, by = 0.02)
  pulse <- rep(0, length(time))

  beat_times <- seq(1, 19, by = 1)

  for (bt in beat_times) {
    pulse <- pulse + exp(-0.5 * ((time - bt) / 0.03)^2)
  }

  dat <- data.frame(
    participant = "p1",
    time = time,
    HRP = pulse + rnorm(length(time), sd = 0.02)
  )

  out <- extract_gazepoint_beats_kmeans(
    dat,
    pulse_col = "HRP",
    time_col = "time",
    group_cols = "participant",
    min_distance_s = 0.5,
    seed = 1
  )

  expect_s3_class(out, "gazepoint_kmeans_beats")
  expect_true(nrow(out$beat_table) > 5)
  expect_true("ibi_s" %in% names(out$interval_table))
})

test_that("audit_gazepoint_stabilization_period flags initial period", {
  dat <- data.frame(
    participant = "p1",
    CNT = seq(0, 1200, by = 10),
    GSR_US = rnorm(121)
  )

  out <- audit_gazepoint_stabilization_period(
    dat,
    time_col = "CNT",
    group_cols = "participant",
    stabilization_minutes = 10,
    action = "flag"
  )

  expect_s3_class(out, "gazepoint_stabilization_audit")
  expect_true("in_stabilization_period" %in% names(out))
  expect_true(any(out$in_stabilization_period))
  expect_true(any(!out$in_stabilization_period))
})

test_that("regress_gazepoint_pupil_luminance adjusts pupil for luminance", {
  dat <- data.frame(
    participant = "p1",
    time = seq_len(80),
    pupil = 3 + 0.4 * scale(seq_len(80))[, 1] + rnorm(80, sd = 0.05),
    luminance = scale(seq_len(80))[, 1]
  )

  out <- regress_gazepoint_pupil_luminance(
    dat,
    pupil_col = "pupil",
    luminance_col = "luminance",
    group_cols = "participant",
    time_col = "time"
  )

  expect_s3_class(out, "gazepoint_pupil_luminance_adjusted")
  expect_true("pupil_luminance_adjusted" %in% names(out))
  expect_equal(
    attr(out, "pupil_luminance_overview")$status,
    "pupil_luminance_regression_complete"
  )
})

test_that("model_gazepoint_hrv_ipfm creates impulse train and spectrum", {
  dat <- data.frame(
    participant = "p1",
    IBI = rep(0.8, 60) + rnorm(60, sd = 0.02)
  )

  out <- model_gazepoint_hrv_ipfm(
    dat,
    ibi_col = "IBI",
    group_cols = "participant",
    output_sampling_rate = 4,
    max_frequency = 0.5
  )

  expect_s3_class(out, "gazepoint_hrv_ipfm")
  expect_true(nrow(out$impulse_table) > 0)
  expect_true("frequency_hz" %in% names(out$spectrum_table))
  expect_equal(out$overview$status, "ipfm_model_created")
})

#' Run a six-phase Gazepoint EDA/GSR analysis pipeline
#'
#' Runs a conservative six-phase Gazepoint EDA/GSR workflow using native
#' `gpbiometrics` helpers where possible. The function imports or accepts data,
#' audits signal quality, prepares preprocessing outputs, creates optional
#' external-method bridge inputs, prepares synchronization/model-formatting
#' outputs, attaches model templates, and generates reporting outputs.
#'
#' The function does not fit `brms` or `lme4` models, does not run external
#' software, and does not infer emotion, valence, stress, trust, preference,
#' cognition, or diagnosis.
#'
#' @param data Optional Gazepoint biometric data frame or imported object.
#' @param path Optional file or folder path. Used only when `data` is `NULL`.
#' @param eda_col Optional EDA/conductance column. If omitted, the runner
#'   prefers `GSR_US` when available.
#' @param time_col Optional time/counter column.
#' @param group_cols Optional grouping columns.
#' @param signal_cols Optional biometric signal columns for activity plots and
#'   signal audits.
#' @param sampling_rate Optional sampling rate in Hz, used when `time_col` is a
#'   sample counter.
#' @param baseline_window Optional baseline window object passed to baseline
#'   correction helpers when supported.
#' @param event_windows Optional event-window table used for SCR event-window
#'   summaries when available.
#' @param event_data Optional event table used for TTL/event alignment when
#'   available.
#' @param lag_signal_pair Optional character vector of length two giving signals
#'   for lag estimation.
#' @param convert_resistance_to_us Logical. If `TRUE`, allow conservative
#'   resistance-to-conductance conversion when the selected EDA column is `GSR`.
#' @param prepare_external_bridges Logical. If `TRUE`, prepare selected external
#'   EDA bridge inputs.
#' @param bridge_methods Character vector containing any of `"neurokit"`,
#'   `"cvxeda"`, `"ledalab"`, and `"pspm"`.
#' @param prepare_model_data Logical. If `TRUE`, attempt to create SCR hurdle
#'   and biometric LME-ready data objects.
#' @param create_reports Logical. If `TRUE`, attempt to create report outputs.
#' @param output_dir Optional output directory for report bundles or bridge
#'   files where supported.
#' @param prefix File prefix used by output-producing helpers where supported.
#' @param continue_on_error Logical. If `TRUE`, failed steps are stored in
#'   `errors` and the pipeline continues.
#'
#' @return A list with `overview`, `phases`, `errors`, `pipeline_guide`,
#'   `model_templates`, `reporting_guidance`, `interpretation_guardrails`, and
#'   `settings`.
#' @export
run_gazepoint_eda_analysis_pipeline <- function(data = NULL,
                                                path = NULL,
                                                eda_col = NULL,
                                                time_col = NULL,
                                                group_cols = NULL,
                                                signal_cols = NULL,
                                                sampling_rate = NULL,
                                                baseline_window = NULL,
                                                event_windows = NULL,
                                                event_data = NULL,
                                                lag_signal_pair = NULL,
                                                convert_resistance_to_us = FALSE,
                                                prepare_external_bridges = TRUE,
                                                bridge_methods = c("neurokit", "cvxeda", "ledalab", "pspm"),
                                                prepare_model_data = TRUE,
                                                create_reports = TRUE,
                                                output_dir = NULL,
                                                prefix = "gazepoint_eda_pipeline",
                                                continue_on_error = TRUE) {
  gpbiometrics_eda_runner_validate_logical(
    prepare_external_bridges,
    "prepare_external_bridges"
  )
  gpbiometrics_eda_runner_validate_logical(
    prepare_model_data,
    "prepare_model_data"
  )
  gpbiometrics_eda_runner_validate_logical(
    create_reports,
    "create_reports"
  )
  gpbiometrics_eda_runner_validate_logical(
    convert_resistance_to_us,
    "convert_resistance_to_us"
  )
  gpbiometrics_eda_runner_validate_logical(
    continue_on_error,
    "continue_on_error"
  )

  bridge_methods <- match.arg(
    bridge_methods,
    choices = c("neurokit", "cvxeda", "ledalab", "pspm"),
    several.ok = TRUE
  )

  if (!is.null(lag_signal_pair) &&
      (!is.character(lag_signal_pair) || length(lag_signal_pair) != 2)) {
    stop("`lag_signal_pair` must be NULL or a character vector of length two.", call. = FALSE)
  }

  errors <- data.frame(
    phase = integer(),
    step = character(),
    function_name = character(),
    message = character(),
    stringsAsFactors = FALSE
  )

  add_error <- function(phase, step, function_name, message) {
    errors <<- rbind(
      errors,
      data.frame(
        phase = phase,
        step = step,
        function_name = function_name,
        message = message,
        stringsAsFactors = FALSE
      )
    )
  }

  run_step <- function(phase, step, function_name, args = list()) {
    result <- gpbiometrics_eda_runner_call(
      function_name = function_name,
      args = args,
      continue_on_error = continue_on_error
    )

    if (inherits(result, "gazepoint_eda_pipeline_error")) {
      add_error(
        phase = phase,
        step = step,
        function_name = function_name,
        message = result$message
      )
    }

    result
  }

  skip_step <- function(reason) {
    gpbiometrics_eda_runner_skip(reason)
  }

  imported <- NULL

  if (is.null(data)) {
    if (is.null(path)) {
      stop("Supply either `data` or `path`.", call. = FALSE)
    }

    if (dir.exists(path)) {
      imported <- run_step(
        phase = 1,
        step = "folder_import",
        function_name = "import_gazepoint_biometric_folder",
        args = list(
          path = path
        )
      )
    } else {
      imported <- run_step(
        phase = 1,
        step = "file_import",
        function_name = "import_gazepoint_biometrics",
        args = list(
          file = path,
          path = path
        )
      )
    }
  } else {
    imported <- data
  }

  dat <- gpbiometrics_eda_runner_extract_data(imported)

  if (!is.data.frame(dat)) {
    stop("Could not extract a data frame from `data` or imported object.", call. = FALSE)
  }

  resolved_eda_col <- gpbiometrics_eda_runner_resolve_col(
    dat = dat,
    supplied = eda_col,
    candidates = c(
      "GSR_US",
      "GSR_US_PHASIC",
      "GSR_US_TONIC",
      "EDA_US",
      "conductance_us",
      "GSR",
      "EDA"
    ),
    label = "EDA/GSR"
  )

  resolved_time_col <- gpbiometrics_eda_runner_resolve_col(
    dat = dat,
    supplied = time_col,
    candidates = c(
      "time_s",
      "time_sec",
      "time_seconds",
      "time_ms",
      "timestamp_ms",
      "timestamp",
      "TIME",
      "Time",
      "time",
      "CNT",
      "cnt"
    ),
    label = "time/counter",
    allow_null = TRUE
  )

  resolved_group_cols <- gpbiometrics_eda_runner_resolve_group_cols(dat, group_cols)

  resolved_signal_cols <- gpbiometrics_eda_runner_resolve_signal_cols(
    dat = dat,
    signal_cols = signal_cols,
    eda_col = resolved_eda_col
  )

  phase1 <- list(
    imported = imported,
    data = dat,
    schema = run_step(
      1,
      "schema_detection",
      "detect_gazepoint_biometric_schema",
      list(data = dat)
    ),
    time_columns = run_step(
      1,
      "time_column_detection",
      "detect_gazepoint_time_columns",
      list(data = dat)
    ),
    timebase = run_step(
      1,
      "timebase_detection",
      "detect_gazepoint_biometric_timebase",
      list(data = dat, time_col = resolved_time_col, time_column = resolved_time_col)
    ),
    active_channels = run_step(
      1,
      "active_channel_detection",
      "detect_active_biometric_channels",
      list(data = dat)
    ),
    validation = run_step(
      1,
      "validation",
      "validate_gazepoint_biometrics",
      list(data = dat)
    ),
    missingness = run_step(
      1,
      "missingness_audit",
      "audit_gazepoint_biometric_missingness",
      list(data = dat, group_cols = resolved_group_cols, group_columns = resolved_group_cols)
    ),
    time_resets = run_step(
      1,
      "time_reset_audit",
      "audit_gazepoint_time_resets",
      list(data = dat, time_col = resolved_time_col, time_column = resolved_time_col, group_cols = resolved_group_cols, group_columns = resolved_group_cols)
    ),
    signal_activity = run_step(
      1,
      "signal_activity_audit",
      "audit_gazepoint_signal_activity",
      list(data = dat, signal_cols = resolved_signal_cols, signal_columns = resolved_signal_cols, group_cols = resolved_group_cols, group_columns = resolved_group_cols)
    ),
    eda_artifacts = run_step(
      1,
      "eda_artifact_audit",
      "audit_gazepoint_eda_artifacts",
      list(data = dat, eda_col = resolved_eda_col, gsr_col = resolved_eda_col, time_col = resolved_time_col, time_column = resolved_time_col, group_cols = resolved_group_cols, group_columns = resolved_group_cols)
    ),
    dashboard = run_step(
      1,
      "qc_dashboard",
      "plot_gazepoint_biometric_report_dashboard",
      list(data = dat, signal_cols = resolved_signal_cols, time_col = resolved_time_col, group_cols = resolved_group_cols, max_groups = 12)
    )
  )

  working_data <- dat

  conversion <- if (isTRUE(convert_resistance_to_us)) {
    run_step(
      2,
      "conservative_conductance_conversion",
      "convert_gazepoint_gsr_to_conductance",
      list(data = working_data, gsr_col = resolved_eda_col, eda_col = resolved_eda_col)
    )
  } else {
    skip_step("Skipped because `convert_resistance_to_us = FALSE`; no blind GSR resistance assumption was made.")
  }

  if (is.data.frame(conversion)) {
    working_data <- conversion
  } else if (is.list(conversion) && is.data.frame(conversion$data)) {
    working_data <- conversion$data
  }

  smoothed <- run_step(
    2,
    "smoothing",
    "smooth_gazepoint_biometrics",
    list(data = working_data, signal_cols = resolved_signal_cols, signal_columns = resolved_signal_cols, group_cols = resolved_group_cols, group_columns = resolved_group_cols)
  )

  if (is.data.frame(smoothed)) {
    working_data <- smoothed
  } else if (is.list(smoothed) && is.data.frame(smoothed$data)) {
    working_data <- smoothed$data
  }

  baseline_corrected <- if (!is.null(baseline_window)) {
    run_step(
      2,
      "baseline_correction",
      "baseline_correct_gazepoint_gsr",
      list(data = working_data, eda_col = resolved_eda_col, gsr_col = resolved_eda_col, baseline_window = baseline_window, group_cols = resolved_group_cols, group_columns = resolved_group_cols)
    )
  } else {
    skip_step("Skipped because `baseline_window` was not supplied.")
  }

  if (is.data.frame(baseline_corrected)) {
    working_data <- baseline_corrected
  } else if (is.list(baseline_corrected) && is.data.frame(baseline_corrected$data)) {
    working_data <- baseline_corrected$data
  }

  scr_peaks <- run_step(
    2,
    "scr_peak_detection",
    "detect_gazepoint_scr_peaks",
    list(data = working_data, eda_col = resolved_eda_col, gsr_col = resolved_eda_col, time_col = resolved_time_col, time_column = resolved_time_col, group_cols = resolved_group_cols, group_columns = resolved_group_cols, sampling_rate = sampling_rate)
  )

  scr_event_windows <- if (!is.null(event_windows)) {
    run_step(
      2,
      "scr_event_window_summary",
      "summarise_gazepoint_scr_event_windows",
      list(data = working_data, peaks = scr_peaks, event_windows = event_windows, eda_col = resolved_eda_col, time_col = resolved_time_col, group_cols = resolved_group_cols, group_columns = resolved_group_cols)
    )
  } else {
    skip_step("Skipped because `event_windows` was not supplied.")
  }

  response_patterns <- run_step(
    2,
    "eda_response_pattern_classification",
    "classify_gazepoint_eda_response_pattern",
    list(data = working_data, response_col = resolved_eda_col, group_cols = resolved_group_cols)
  )

  phase2 <- list(
    conversion = conversion,
    smoothed = smoothed,
    baseline_corrected = baseline_corrected,
    scr_peaks = scr_peaks,
    scr_event_windows = scr_event_windows,
    response_patterns = response_patterns,
    working_data = working_data
  )

  phase3 <- list()

  if (isTRUE(prepare_external_bridges)) {
    if ("neurokit" %in% bridge_methods) {
      phase3$neurokit <- run_step(
        3,
        "neurokit_bridge",
        "prepare_gazepoint_neurokit_eda_input",
        list(data = working_data, eda_col = resolved_eda_col, gsr_col = resolved_eda_col, time_col = resolved_time_col, group_cols = resolved_group_cols, sampling_rate = sampling_rate)
      )
    }

    if ("cvxeda" %in% bridge_methods) {
      phase3$cvxeda <- run_step(
        3,
        "cvxeda_bridge",
        "prepare_gazepoint_cvxeda_input",
        list(data = working_data, eda_col = resolved_eda_col, time_col = resolved_time_col, group_cols = resolved_group_cols, sampling_rate = sampling_rate, output_dir = NULL)
      )
    }

    if ("ledalab" %in% bridge_methods) {
      phase3$ledalab <- run_step(
        3,
        "ledalab_bridge",
        "prepare_gazepoint_ledalab_input",
        list(data = working_data, eda_col = resolved_eda_col, time_col = resolved_time_col, group_cols = resolved_group_cols, sampling_rate = sampling_rate, output_dir = NULL)
      )
    }

    if ("pspm" %in% bridge_methods) {
      phase3$pspm <- run_step(
        3,
        "pspm_bridge",
        "prepare_gazepoint_pspm_input",
        list(data = working_data, eda_col = resolved_eda_col, time_col = resolved_time_col, group_cols = resolved_group_cols, sampling_rate = sampling_rate, output_dir = NULL)
      )
    }
  } else {
    phase3$bridges <- skip_step("Skipped because `prepare_external_bridges = FALSE`.")
  }

  ttl_events <- run_step(
    4,
    "ttl_event_extraction",
    "extract_gazepoint_ttl_events",
    list(data = working_data, group_cols = resolved_group_cols, group_columns = resolved_group_cols, time_col = resolved_time_col, time_column = resolved_time_col)
  )

  ttl_alignment <- if (!is.null(event_data) || !inherits(ttl_events, "gazepoint_eda_pipeline_error")) {
    run_step(
      4,
      "ttl_alignment",
      "align_gazepoint_biometrics_to_ttl",
      list(data = working_data, ttl_events = ttl_events, events = event_data, event_data = event_data, time_col = resolved_time_col, time_column = resolved_time_col, group_cols = resolved_group_cols, group_columns = resolved_group_cols)
    )
  } else {
    skip_step("Skipped because no usable TTL/event data were available.")
  }

  resolved_lag_pair <- gpbiometrics_eda_runner_lag_pair(
    dat = working_data,
    lag_signal_pair = lag_signal_pair,
    eda_col = resolved_eda_col,
    signal_cols = resolved_signal_cols
  )

  signal_lag <- if (length(resolved_lag_pair) == 2) {
    run_step(
      4,
      "signal_lag_estimation",
      "estimate_gazepoint_signal_lag",
      list(data = working_data, signal_x_col = resolved_lag_pair[1], signal_y_col = resolved_lag_pair[2], time_col = resolved_time_col, group_cols = resolved_group_cols)
    )
  } else {
    skip_step("Skipped because fewer than two numeric signal columns were available for lag estimation.")
  }

  sync_drift <- run_step(
    4,
    "sync_drift_audit",
    "audit_gazepoint_biometric_sync_drift",
    list(data = working_data, time_col = resolved_time_col, group_cols = resolved_group_cols, signal_cols = resolved_signal_cols)
  )

  scr_hurdle_data <- if (isTRUE(prepare_model_data)) {
    run_step(
      4,
      "scr_hurdle_model_data",
      "prepare_gazepoint_scr_hurdle_model_data",
      list(data = working_data, scr_events = scr_event_windows, peaks = scr_peaks, group_cols = resolved_group_cols, group_columns = resolved_group_cols)
    )
  } else {
    skip_step("Skipped because `prepare_model_data = FALSE`.")
  }

  lme_data <- if (isTRUE(prepare_model_data)) {
    run_step(
      4,
      "biometrics_lme_data",
      "prepare_gazepoint_biometrics_lme_data",
      list(data = working_data, outcome_cols = resolved_signal_cols, signal_cols = resolved_signal_cols, time_col = resolved_time_col, group_cols = resolved_group_cols, group_columns = resolved_group_cols)
    )
  } else {
    skip_step("Skipped because `prepare_model_data = FALSE`.")
  }

  phase4 <- list(
    ttl_events = ttl_events,
    ttl_alignment = ttl_alignment,
    signal_lag = signal_lag,
    sync_drift = sync_drift,
    scr_hurdle_data = scr_hurdle_data,
    lme_data = lme_data
  )

  pipeline_guide <- run_step(
    5,
    "pipeline_guide",
    "create_gazepoint_eda_analysis_pipeline",
    list(include_external_bridges = prepare_external_bridges, include_model_templates = TRUE, include_reporting_guidance = TRUE, style = "detailed")
  )

  model_templates <- if (is.list(pipeline_guide) && is.data.frame(pipeline_guide$model_templates)) {
    pipeline_guide$model_templates
  } else {
    data.frame()
  }

  phase5 <- list(
    model_templates = model_templates,
    note = paste(
      "Model fitting is intentionally external to gpbiometrics.",
      "Use the prepared SCR hurdle and LME-ready tables with brms, lme4, or another appropriate modelling package."
    )
  )

  report_bundle <- if (isTRUE(create_reports) && !is.null(output_dir)) {
    run_step(
      6,
      "report_bundle",
      "export_gazepoint_biometrics_report_bundle",
      list(data = working_data, output_dir = output_dir, prefix = prefix)
    )
  } else if (isTRUE(create_reports)) {
    skip_step("Skipped report-bundle file export because `output_dir` was not supplied.")
  } else {
    skip_step("Skipped because `create_reports = FALSE`.")
  }

  report_tables <- if (isTRUE(create_reports)) {
    run_step(
      6,
      "report_tables",
      "create_gazepoint_biometrics_report_tables",
      list(data = working_data, workflow = list(data = working_data, quality = phase1, windows = phase2))
    )
  } else {
    skip_step("Skipped because `create_reports = FALSE`.")
  }

  methods_text <- if (isTRUE(create_reports)) {
    run_step(
      6,
      "methods_text",
      "create_gazepoint_biometrics_methods_text",
      list(data = working_data)
    )
  } else {
    skip_step("Skipped because `create_reports = FALSE`.")
  }

  phase6 <- list(
    report_bundle = report_bundle,
    report_tables = report_tables,
    methods_text = methods_text
  )

  phases <- list(
    phase_1_ingestion_qc = phase1,
    phase_2_preprocessing_peaks = phase2,
    phase_3_external_bridges = phase3,
    phase_4_sync_model_formatting = phase4,
    phase_5_model_templates = phase5,
    phase_6_reporting = phase6
  )

  skipped_count <- gpbiometrics_eda_runner_count_class(
    phases,
    "gazepoint_eda_pipeline_skip"
  )

  error_count <- nrow(errors)

  status <- if (error_count == 0) {
    "eda_analysis_pipeline_completed"
  } else if (isTRUE(continue_on_error)) {
    "eda_analysis_pipeline_completed_with_errors"
  } else {
    "eda_analysis_pipeline_failed"
  }

  overview <- data.frame(
    input_rows = nrow(dat),
    input_columns = ncol(dat),
    phase_count = length(phases),
    error_count = error_count,
    skipped_count = skipped_count,
    eda_col = if (is.null(resolved_eda_col)) NA_character_ else resolved_eda_col,
    time_col = if (is.null(resolved_time_col)) NA_character_ else resolved_time_col,
    group_cols = paste(resolved_group_cols, collapse = ", "),
    signal_cols = paste(resolved_signal_cols, collapse = ", "),
    external_bridges_prepared = isTRUE(prepare_external_bridges),
    model_data_requested = isTRUE(prepare_model_data),
    reports_requested = isTRUE(create_reports),
    status = status,
    interpretation = paste(
      "The pipeline organizes Gazepoint EDA/GSR preprocessing, QC, bridge preparation, model formatting, and reporting.",
      "It does not fit statistical models, run external software, or infer emotion, valence, stress, trust, preference, cognition, or diagnosis."
    ),
    stringsAsFactors = FALSE
  )

  structure(
    list(
      overview = overview,
      phases = phases,
      errors = errors,
      pipeline_guide = pipeline_guide,
      model_templates = model_templates,
      reporting_guidance = if (is.list(pipeline_guide)) pipeline_guide$reporting_guidance else data.frame(),
      interpretation_guardrails = if (is.list(pipeline_guide)) pipeline_guide$interpretation_guardrails else data.frame(),
      settings = list(
        path = path,
        eda_col = eda_col,
        resolved_eda_col = resolved_eda_col,
        time_col = time_col,
        resolved_time_col = resolved_time_col,
        group_cols = group_cols,
        resolved_group_cols = resolved_group_cols,
        signal_cols = signal_cols,
        resolved_signal_cols = resolved_signal_cols,
        sampling_rate = sampling_rate,
        baseline_window_supplied = !is.null(baseline_window),
        event_windows_supplied = !is.null(event_windows),
        event_data_supplied = !is.null(event_data),
        lag_signal_pair = lag_signal_pair,
        convert_resistance_to_us = convert_resistance_to_us,
        prepare_external_bridges = prepare_external_bridges,
        bridge_methods = bridge_methods,
        prepare_model_data = prepare_model_data,
        create_reports = create_reports,
        output_dir = output_dir,
        prefix = prefix,
        continue_on_error = continue_on_error
      )
    ),
    class = c("gazepoint_eda_analysis_pipeline_run", "list")
  )
}

gpbiometrics_eda_runner_call <- function(function_name,
                                         args = list(),
                                         continue_on_error = TRUE) {
  fn <- gpbiometrics_eda_runner_get_function(function_name)

  if (is.null(fn)) {
    msg <- paste0("Function `", function_name, "` is not available.")
    if (!isTRUE(continue_on_error)) {
      stop(msg, call. = FALSE)
    }
    return(gpbiometrics_eda_runner_error(msg))
  }

  args <- gpbiometrics_eda_runner_filter_args(fn, args)

  out <- tryCatch(
    do.call(fn, args),
    error = function(e) e
  )

  if (inherits(out, "error")) {
    if (!isTRUE(continue_on_error)) {
      stop(conditionMessage(out), call. = FALSE)
    }
    return(gpbiometrics_eda_runner_error(conditionMessage(out)))
  }

  out
}

gpbiometrics_eda_runner_get_function <- function(function_name) {
  if (exists(function_name, mode = "function", envir = asNamespace("gpbiometrics"), inherits = FALSE)) {
    return(get(function_name, mode = "function", envir = asNamespace("gpbiometrics"), inherits = FALSE))
  }

  if (exists(function_name, mode = "function", inherits = TRUE)) {
    return(get(function_name, mode = "function", inherits = TRUE))
  }

  NULL
}

gpbiometrics_eda_runner_filter_args <- function(fn, args) {
  args <- args[!vapply(args, function(x) length(x) == 0, logical(1))]

  formal_names <- names(formals(fn))

  if ("..." %in% formal_names) {
    return(args)
  }

  args[names(args) %in% formal_names]
}

gpbiometrics_eda_runner_extract_data <- function(x) {
  if (is.data.frame(x)) {
    return(x)
  }

  if (is.list(x)) {
    candidates <- c(
      "data",
      "biometrics",
      "merged_data",
      "all_data",
      "imported_data",
      "raw_data"
    )

    for (nm in candidates) {
      if (!is.null(x[[nm]]) && is.data.frame(x[[nm]])) {
        return(x[[nm]])
      }
    }

    is_df <- vapply(x, is.data.frame, logical(1))

    if (any(is_df)) {
      return(x[[which(is_df)[1]]])
    }
  }

  NULL
}

gpbiometrics_eda_runner_resolve_col <- function(dat,
                                                supplied,
                                                candidates,
                                                label,
                                                allow_null = FALSE) {
  if (!is.null(supplied)) {
    if (!is.character(supplied) || length(supplied) != 1 || is.na(supplied)) {
      stop("`", label, "` column must be a single column name.", call. = FALSE)
    }

    if (!supplied %in% names(dat)) {
      stop("Column `", supplied, "` was not found in `data`.", call. = FALSE)
    }

    return(supplied)
  }

  found <- intersect(candidates, names(dat))

  if (length(found) > 0) {
    return(found[1])
  }

  if (isTRUE(allow_null)) {
    return(NULL)
  }

  stop("No ", label, " column was detected. Supply it explicitly.", call. = FALSE)
}

gpbiometrics_eda_runner_resolve_group_cols <- function(dat, group_cols = NULL) {
  if (!is.null(group_cols)) {
    if (!is.character(group_cols)) {
      stop("`group_cols` must be NULL or a character vector.", call. = FALSE)
    }

    missing_cols <- setdiff(group_cols, names(dat))

    if (length(missing_cols) > 0) {
      stop(
        "The following `group_cols` were not found in `data`: ",
        paste(missing_cols, collapse = ", "),
        call. = FALSE
      )
    }

    return(group_cols)
  }

  candidates <- c(
    "source_file",
    "source_participant",
    "USER_FILE",
    "participant",
    "participant_id",
    "subject",
    "subject_id",
    "MEDIA_ID",
    "MEDIA_NAME",
    "stimulus",
    "trial",
    "trial_id",
    "trial_global"
  )

  intersect(candidates, names(dat))
}

gpbiometrics_eda_runner_resolve_signal_cols <- function(dat,
                                                        signal_cols = NULL,
                                                        eda_col = NULL) {
  if (!is.null(signal_cols)) {
    if (!is.character(signal_cols)) {
      stop("`signal_cols` must be NULL or a character vector.", call. = FALSE)
    }

    missing_cols <- setdiff(signal_cols, names(dat))

    if (length(missing_cols) > 0) {
      stop(
        "The following `signal_cols` were not found in `data`: ",
        paste(missing_cols, collapse = ", "),
        call. = FALSE
      )
    }

    return(signal_cols)
  }

  candidates <- c(
    eda_col,
    "GSR_US",
    "GSR_US_PHASIC",
    "GSR_US_TONIC",
    "GSR",
    "EDA",
    "HR",
    "HRP",
    "IBI",
    "DIAL"
  )

  unique(intersect(candidates[!is.na(candidates)], names(dat)))
}

gpbiometrics_eda_runner_lag_pair <- function(dat,
                                             lag_signal_pair = NULL,
                                             eda_col = NULL,
                                             signal_cols = NULL) {
  if (!is.null(lag_signal_pair)) {
    if (all(lag_signal_pair %in% names(dat))) {
      return(lag_signal_pair)
    }

    return(character())
  }

  candidates <- unique(c(
    eda_col,
    intersect(c("HR", "IBI", "DIAL", "GSR_US_PHASIC", "GSR_US_TONIC"), names(dat)),
    signal_cols
  ))

  candidates <- candidates[!is.na(candidates)]
  candidates <- candidates[candidates %in% names(dat)]
  candidates <- candidates[vapply(dat[candidates], is.numeric, logical(1))]

  if (length(candidates) >= 2) {
    return(candidates[1:2])
  }

  character()
}

gpbiometrics_eda_runner_error <- function(message) {
  structure(
    list(
      message = message,
      status = "error"
    ),
    class = c("gazepoint_eda_pipeline_error", "list")
  )
}

gpbiometrics_eda_runner_skip <- function(reason) {
  structure(
    list(
      reason = reason,
      status = "skipped"
    ),
    class = c("gazepoint_eda_pipeline_skip", "list")
  )
}

gpbiometrics_eda_runner_count_class <- function(x, class_name) {
  count <- 0L

  walk <- function(obj) {
    if (inherits(obj, class_name)) {
      count <<- count + 1L
    }

    if (is.list(obj) && !is.data.frame(obj)) {
      for (item in obj) {
        walk(item)
      }
    }

    invisible(NULL)
  }

  walk(x)

  count
}

gpbiometrics_eda_runner_validate_logical <- function(x, name) {
  if (!is.logical(x) || length(x) != 1 || is.na(x)) {
    stop("`", name, "` must be TRUE or FALSE.", call. = FALSE)
  }

  invisible(TRUE)
}

test_that("run_gazepoint_eda_analysis_pipeline returns a six-phase run object", {
  dat <- data.frame(
    source_file = "user1_all_gaze.csv",
    MEDIA_ID = 1,
    CNT = seq_len(120),
    GSR_US = 1 + sin(seq(0, 4 * pi, length.out = 120)) * 0.1,
    GSR_US_PHASIC = sin(seq(0, 8 * pi, length.out = 120)) * 0.03,
    HR = 70 + sin(seq(0, 2 * pi, length.out = 120)) * 3,
    IBI = rep(0.8, 120),
    DIAL = rep(1, 120)
  )

  out <- run_gazepoint_eda_analysis_pipeline(
    data = dat,
    eda_col = "GSR_US",
    time_col = "CNT",
    group_cols = c("source_file", "MEDIA_ID"),
    signal_cols = c("GSR_US", "GSR_US_PHASIC", "HR", "IBI", "DIAL"),
    sampling_rate = 60,
    prepare_external_bridges = FALSE,
    prepare_model_data = FALSE,
    create_reports = FALSE,
    continue_on_error = TRUE
  )

  expect_s3_class(out, "gazepoint_eda_analysis_pipeline_run")
  expect_equal(out$overview$phase_count, 6)
  expect_equal(out$overview$input_rows, 120)
  expect_equal(out$overview$eda_col, "GSR_US")
  expect_equal(out$overview$time_col, "CNT")
  expect_named(
    out$phases,
    c(
      "phase_1_ingestion_qc",
      "phase_2_preprocessing_peaks",
      "phase_3_external_bridges",
      "phase_4_sync_model_formatting",
      "phase_5_model_templates",
      "phase_6_reporting"
    )
  )
  expect_true(is.data.frame(out$model_templates))
  expect_true(is.data.frame(out$interpretation_guardrails))
})

test_that("run_gazepoint_eda_analysis_pipeline can prepare external bridge phase outputs", {
  dat <- data.frame(
    participant = "p1",
    CNT = seq_len(60),
    GSR_US = seq(1, 2, length.out = 60),
    HR = 70 + sin(seq(0, 2 * pi, length.out = 60))
  )

  out <- run_gazepoint_eda_analysis_pipeline(
    data = dat,
    eda_col = "GSR_US",
    time_col = "CNT",
    group_cols = "participant",
    signal_cols = c("GSR_US", "HR"),
    sampling_rate = 60,
    prepare_external_bridges = TRUE,
    bridge_methods = c("cvxeda", "ledalab", "pspm"),
    prepare_model_data = FALSE,
    create_reports = FALSE,
    continue_on_error = TRUE
  )

  expect_s3_class(out, "gazepoint_eda_analysis_pipeline_run")
  expect_true("cvxeda" %in% names(out$phases$phase_3_external_bridges))
  expect_true("ledalab" %in% names(out$phases$phase_3_external_bridges))
  expect_true("pspm" %in% names(out$phases$phase_3_external_bridges))

  expect_false(inherits(out$phases$phase_3_external_bridges$cvxeda, "gazepoint_eda_pipeline_error"))
  expect_false(inherits(out$phases$phase_3_external_bridges$ledalab, "gazepoint_eda_pipeline_error"))
  expect_false(inherits(out$phases$phase_3_external_bridges$pspm, "gazepoint_eda_pipeline_error"))
})

test_that("run_gazepoint_eda_analysis_pipeline records errors when continuing", {
  dat <- data.frame(
    participant = "p1",
    CNT = seq_len(10),
    HR = 70 + seq_len(10)
  )

  expect_error(
    run_gazepoint_eda_analysis_pipeline(
      data = dat,
      eda_col = "missing_gsr",
      time_col = "CNT",
      group_cols = "participant",
      continue_on_error = TRUE
    ),
    "missing_gsr"
  )
})

test_that("run_gazepoint_eda_analysis_pipeline stops on step errors when requested", {
  dat <- data.frame(
    participant = "p1",
    CNT = seq_len(20),
    GSR_US = seq(1, 2, length.out = 20)
  )

  expect_error(
    run_gazepoint_eda_analysis_pipeline(
      data = dat,
      eda_col = "GSR_US",
      time_col = "CNT",
      group_cols = "missing_group",
      continue_on_error = FALSE
    ),
    "missing_group"
  )
})

#' Run blockwise online design optimization decision support
#'
#' Provides a safe, dependency-light decision-support/simulation helper for
#' online design optimization. The function recommends the next condition by
#' combining expected model-discrimination utility with optional exploration and
#' balancing penalties.
#'
#' This function does not control stimulus presentation software and should not
#' be used as autonomous real-time experiment control without separate ethical,
#' preregistration, and software-integration review.
#'
#' @param candidate_table A data frame containing candidate conditions.
#' @param condition_col Candidate condition column.
#' @param utility_col Expected utility/model-discrimination column.
#' @param block_col Optional block column.
#' @param cost_col Optional cost or burden column subtracted from utility.
#' @param previous_assignments Optional previous condition assignments.
#' @param exploration_weight Weight for favouring under-sampled conditions.
#' @param balance_weight Weight for penalising over-sampled conditions.
#' @param maximise Logical. If `TRUE`, select highest score.
#'
#' @return A list with `overview`, `ranked_candidates`, `recommendation`,
#'   `assignment_summary`, and `settings`.
#' @export
run_gazepoint_online_design_optimization <- function(candidate_table,
                                                     condition_col = "condition",
                                                     utility_col = "expected_utility",
                                                     block_col = NULL,
                                                     cost_col = NULL,
                                                     previous_assignments = NULL,
                                                     exploration_weight = 0.10,
                                                     balance_weight = 0.10,
                                                     maximise = TRUE) {
  if (!is.data.frame(candidate_table)) {
    stop("`candidate_table` must be a data frame.", call. = FALSE)
  }

  required <- c(condition_col, utility_col, block_col, cost_col)
  required <- required[!is.null(required)]

  missing_required <- setdiff(required, names(candidate_table))

  if (length(missing_required) > 0) {
    stop("Missing required columns: ", paste(missing_required, collapse = ", "), call. = FALSE)
  }

  if (!is.numeric(candidate_table[[utility_col]])) {
    stop("`utility_col` must identify a numeric column.", call. = FALSE)
  }

  if (!is.null(cost_col) && !is.numeric(candidate_table[[cost_col]])) {
    stop("`cost_col` must identify a numeric column.", call. = FALSE)
  }

  ranked <- candidate_table
  ranked$.condition <- as.character(ranked[[condition_col]])

  if (is.null(previous_assignments)) {
    previous_assignments <- character()
  } else if (is.data.frame(previous_assignments)) {
    if (!condition_col %in% names(previous_assignments)) {
      stop("`previous_assignments` data frame must contain `condition_col`.", call. = FALSE)
    }
    previous_assignments <- as.character(previous_assignments[[condition_col]])
  } else {
    previous_assignments <- as.character(previous_assignments)
  }

  condition_counts <- table(factor(
    previous_assignments,
    levels = unique(ranked$.condition)
  ))

  assignment_summary <- data.frame(
    condition = names(condition_counts),
    previous_n = as.integer(condition_counts),
    stringsAsFactors = FALSE
  )

  ranked$.previous_n <- assignment_summary$previous_n[
    match(ranked$.condition, assignment_summary$condition)
  ]

  ranked$.previous_n[is.na(ranked$.previous_n)] <- 0L

  max_previous <- max(ranked$.previous_n, 0)
  ranked$.exploration_bonus <- exploration_weight / (ranked$.previous_n + 1)
  ranked$.balance_penalty <- balance_weight * ifelse(
    max_previous > 0,
    ranked$.previous_n / max_previous,
    0
  )

  ranked$.cost <- if (!is.null(cost_col)) {
    ranked[[cost_col]]
  } else {
    0
  }

  ranked$.optimization_score <- ranked[[utility_col]] +
    ranked$.exploration_bonus -
    ranked$.balance_penalty -
    ranked$.cost

  ranked <- ranked[order(ranked$.optimization_score, decreasing = maximise), , drop = FALSE]
  ranked$optimization_rank <- seq_len(nrow(ranked))

  recommendation <- ranked[1, , drop = FALSE]

  overview <- data.frame(
    candidate_count = nrow(candidate_table),
    previous_assignment_count = length(previous_assignments),
    recommended_condition = recommendation$.condition[1],
    recommended_score = recommendation$.optimization_score[1],
    status = "online_design_recommendation_created",
    interpretation = paste(
      "The recommendation is blockwise design-optimization decision support.",
      "It does not autonomously control an experiment and does not prove which hypothesis is true."
    ),
    stringsAsFactors = FALSE
  )

  structure(
    list(
      overview = overview,
      ranked_candidates = ranked,
      recommendation = recommendation,
      assignment_summary = assignment_summary,
      settings = list(
        condition_col = condition_col,
        utility_col = utility_col,
        block_col = block_col,
        cost_col = cost_col,
        exploration_weight = exploration_weight,
        balance_weight = balance_weight,
        maximise = maximise
      )
    ),
    class = c("gazepoint_online_design_optimization", "list")
  )
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/respiration-ceemdan-kalman.R
\name{extract_gazepoint_respiration_ceemdan}
\alias{extract_gazepoint_respiration_ceemdan}
\title{Extract respiration proxy using a CEEMDAN-style bridge}
\usage{
extract_gazepoint_respiration_ceemdan(
  dat,
  signal_col,
  time_col = "CNT",
  group_cols = NULL,
  sampling_rate = NULL,
  respiration_band = c(0.1, 0.6),
  scales = c(5, 15, 30, 60, 120),
  external_fun = NULL
)
}
\arguments{
\item{dat}{A data frame.}

\item{signal_col}{Numeric PPG/ECG-derived signal column.}

\item{time_col}{Numeric time column.}

\item{group_cols}{Optional grouping columns.}

\item{sampling_rate}{Optional sampling rate in Hz.}

\item{respiration_band}{Frequency band in Hz used to select respiration-like
components.}

\item{scales}{Moving-average scales used by the fallback decomposition.}

\item{external_fun}{Optional function with arguments \code{x}, \code{time}, and
\code{sampling_rate}, returning either a numeric vector or a list/data frame of
components.}
}
\value{
A list with \code{overview}, \code{component_table}, \code{respiration_timeseries},
\code{summary}, and \code{settings}.
}
\description{
Extracts respiration-like components from PPG or ECG-derived respiratory
proxy signals. If \code{external_fun} is supplied, it is used as the CEEMDAN
backend. Otherwise, the function uses a dependency-light multiscale
decomposition fallback and labels the result accordingly.
}
\details{
This function does not claim to reproduce full CEEMDAN unless a validated
external CEEMDAN function is supplied.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pdr-rsa.R
\name{calculate_gazepoint_rsa}
\alias{calculate_gazepoint_rsa}
\title{Calculate respiration-informed RSA proxy features}
\usage{
calculate_gazepoint_rsa(
  dat,
  ibi_col = "IBI",
  time_col = "CNT",
  group_cols = NULL,
  pdr = NULL,
  resp_rate_hz = NULL,
  respiration_band = c(0.12, 0.4),
  resample_rate = 4
)
}
\arguments{
\item{dat}{A data frame containing IBI/RR data.}

\item{ibi_col}{IBI/RR interval column.}

\item{time_col}{Time column for the IBI/RR observations.}

\item{group_cols}{Optional grouping columns.}

\item{pdr}{Optional output from \code{extract_gazepoint_pdr_signals()}.}

\item{resp_rate_hz}{Optional fixed respiration rate in Hz.}

\item{respiration_band}{Default respiration/HF band when no PDR rate is
available.}

\item{resample_rate}{Resampling rate for spectral RSA proxy calculation.}
}
\value{
A list with \code{overview}, \code{rsa_summary}, and \code{settings}.
}
\description{
Calculates dependency-light RSA proxy summaries from IBI/RR intervals and,
optionally, PPG-derived respiration features from
\code{extract_gazepoint_pdr_signals()}. This provides peak-to-trough and
Porges-Bohrer-inspired band-power proxy summaries. These outputs should be
interpreted as respiration-informed HRV/RSA features, not direct clinical
vagal-tone estimates.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/preprocess.R
\name{baseline_correct_gazepoint_gsr}
\alias{baseline_correct_gazepoint_gsr}
\title{Baseline-correct Gazepoint GSR/EDA}
\usage{
baseline_correct_gazepoint_gsr(
  data,
  baseline_rows,
  value_column = NULL,
  validity_column = "GSRV",
  group_columns = NULL,
  output_column = NULL,
  summary = c("mean", "median"),
  exclude_zero = TRUE
)
}
\arguments{
\item{data}{A data frame or a path to a Gazepoint CSV export.}

\item{baseline_rows}{Logical vector identifying baseline rows.}

\item{value_column}{Optional GSR/EDA value column. If \code{NULL}, \code{GSR_US} is
used when present, otherwise \code{GSR}.}

\item{validity_column}{Optional validity column. Defaults to \code{"GSRV"}.}

\item{group_columns}{Optional grouping columns. When supplied, baselines are
estimated separately within each group.}

\item{output_column}{Name of the corrected output column.}

\item{summary}{Baseline summary, either \code{"mean"} or \code{"median"}.}

\item{exclude_zero}{Should zero values be excluded from baseline estimation?}
}
\value{
A data frame with the added baseline-corrected column and a
baseline-summary attribute named \code{"baseline_summary"}.
}
\description{
Adds a baseline-corrected GSR/EDA column to a Gazepoint Biometrics table.
When available, \code{GSR_US} is used by default because it represents skin
conductance in microsiemens in Gazepoint exports. The baseline is estimated
from rows selected by \code{baseline_rows}, optionally within groups.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/adaptive-ema-normalization.R
\name{standardise_gazepoint_adaptive_ema}
\alias{standardise_gazepoint_adaptive_ema}
\alias{standardize_gazepoint_adaptive_ema}
\title{Adaptive EMA normalization for non-stationary EDA}
\usage{
standardise_gazepoint_adaptive_ema(
  dat,
  signal_col = "GSR_US",
  group_cols = NULL,
  time_col = NULL,
  alpha = 0.05,
  iqr_multiplier = 1.5,
  suffix = "_adaptive_ema",
  center_suffix = "_ema_center",
  scale_suffix = "_ema_scale",
  min_scale = 1e-08,
  overwrite = FALSE
)

standardize_gazepoint_adaptive_ema(
  dat,
  signal_col = "GSR_US",
  group_cols = NULL,
  time_col = NULL,
  alpha = 0.05,
  iqr_multiplier = 1.5,
  suffix = "_adaptive_ema",
  center_suffix = "_ema_center",
  scale_suffix = "_ema_scale",
  min_scale = 1e-08,
  overwrite = FALSE
)
}
\arguments{
\item{dat}{A data frame.}

\item{signal_col}{Numeric signal column.}

\item{group_cols}{Optional grouping columns.}

\item{time_col}{Optional time column used to order rows within group.}

\item{alpha}{EMA smoothing parameter in \verb{(0, 1]}.}

\item{iqr_multiplier}{IQR multiplier for outlier screening.}

\item{suffix}{Suffix for the normalized output column.}

\item{center_suffix}{Suffix for the EMA center column.}

\item{scale_suffix}{Suffix for the EMA scale column.}

\item{min_scale}{Minimum scale used to avoid division by zero.}

\item{overwrite}{Logical. If \code{FALSE}, existing output columns are protected.}
}
\value{
A data frame with adaptive normalized signal columns and attributes.
}
\description{
Applies dependency-light adaptive normalization using an exponential moving
average center and robust local scale after IQR-based outlier screening.
This preserves local dynamics more than whole-session z-scoring, but it is
still a preprocessing transformation and not an emotion/stress classifier.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/advanced-modality-bridges.R
\name{extract_gazepoint_edr_pca}
\alias{extract_gazepoint_edr_pca}
\title{Extract ECG-derived respiration using PCA}
\usage{
extract_gazepoint_edr_pca(
  dat,
  ecg_cols,
  time_col = NULL,
  group_cols = NULL,
  n_components = 1,
  scale = TRUE,
  output_prefix = "edr_pca"
)
}
\arguments{
\item{dat}{A data frame.}

\item{ecg_cols}{Numeric ECG morphology columns.}

\item{time_col}{Optional time column.}

\item{group_cols}{Optional grouping columns.}

\item{n_components}{Number of PCA components to retain.}

\item{scale}{Logical. If \code{TRUE}, scale ECG morphology columns before PCA.}

\item{output_prefix}{Prefix for PCA output columns.}
}
\value{
A list with \code{overview}, \code{edr_timeseries}, \code{component_summary}, and
\code{settings}.
}
\description{
Extracts an ECG-derived respiration proxy from beat-level ECG morphology
features using principal component analysis. This function requires
ECG-derived morphology columns, such as QRS amplitudes, widths, or sampled
beat-shape features. It is not intended for HR, IBI, or PPG-only data.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compatibility-wrappers.R
\name{summarise_gazepoint_dial_windows}
\alias{summarise_gazepoint_dial_windows}
\title{Summarise Gazepoint engagement-dial windows}
\usage{
summarise_gazepoint_dial_windows(data, ..., dial_col = NULL)
}
\arguments{
\item{data}{A data frame containing Gazepoint Biometrics engagement/dial data.}

\item{...}{Additional arguments passed to
\code{\link[=summarise_gazepoint_engagement_windows]{summarise_gazepoint_engagement_windows()}}.}

\item{dial_col}{Optional dial/engagement column. When supplied, it is mapped
to the corresponding value-column argument of the underlying helper.}
}
\value{
The output of \code{\link[=summarise_gazepoint_engagement_windows]{summarise_gazepoint_engagement_windows()}}.
}
\description{
Compatibility wrapper for \code{\link[=summarise_gazepoint_engagement_windows]{summarise_gazepoint_engagement_windows()}}. This
helper uses the term "dial" for users who refer to Gazepoint engagement-dial
or self-reported engagement streams, while delegating the calculation to the
canonical engagement-window summariser.
}
\examples{
df <- data.frame(
  USER = rep(c("P1", "P2"), each = 3),
  DIAL = c(40, 45, 50, 55, 60, 65)
)
summarise_gazepoint_dial_windows(df)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/eda-gram-plot.R
\name{plot_gazepoint_eda_gram}
\alias{plot_gazepoint_eda_gram}
\title{Plot an EDA-gram-style time-frequency representation}
\usage{
plot_gazepoint_eda_gram(
  dat,
  eda_col = "GSR_US",
  time_col = "CNT",
  group_cols = NULL,
  group_id_to_plot = NULL,
  sampling_rate = NULL,
  window_seconds = 30,
  step_seconds = 5,
  frequency_range = c(0.01, 0.5),
  frequency_bins = 64,
  log_power = TRUE,
  plot = TRUE,
  main = "EDA-gram"
)
}
\arguments{
\item{dat}{A data frame containing EDA data.}

\item{eda_col}{Numeric EDA/conductance column.}

\item{time_col}{Numeric time column.}

\item{group_cols}{Optional grouping columns.}

\item{group_id_to_plot}{Optional group ID to plot. If \code{NULL}, plots the first
available group.}

\item{sampling_rate}{Optional sampling rate in Hz. If \code{NULL}, estimated from
\code{time_col}.}

\item{window_seconds}{Sliding-window length.}

\item{step_seconds}{Sliding-window step.}

\item{frequency_range}{Frequency range shown in Hz.}

\item{frequency_bins}{Number of frequency bins.}

\item{log_power}{Logical. If \code{TRUE}, plot log1p power.}

\item{plot}{Logical. If \code{TRUE}, draw the EDA-gram.}

\item{main}{Plot title.}
}
\value{
Invisibly returns a list with \code{overview}, \code{gram_table},
\code{plot_matrix}, and \code{settings}.
}
\description{
Creates a dependency-light EDA-gram-style representation using sliding-window
spectral power. This is inspired by EDA-gram visualisations, but it does not
implement a full sparse dictionary decomposition unless such a model is
supplied externally.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/quality.R
\name{audit_gazepoint_hr_quality}
\alias{audit_gazepoint_hr_quality}
\title{Audit Gazepoint heart-rate signal quality}
\usage{
audit_gazepoint_hr_quality(
  data,
  value_column = "HR",
  validity_column = "HRV",
  min_value = 30,
  max_value = 220,
  jump_threshold = 25
)
}
\arguments{
\item{data}{A data frame or a path to a Gazepoint CSV export.}

\item{value_column}{Heart-rate value column. Defaults to \code{"HR"}.}

\item{validity_column}{Heart-rate validity column. Defaults to \code{"HRV"}.}

\item{min_value}{Minimum plausible heart rate.}

\item{max_value}{Maximum plausible heart rate.}

\item{jump_threshold}{Threshold for detecting large sample-to-sample jumps.}
}
\value{
A one-row data frame summarising signal quality.
}
\description{
Audits Gazepoint heart-rate values for missingness, inactive zero rows,
validity flags, plausible value ranges, sudden jumps, flatlining, and usable
sample coverage. \code{HRV} is treated as the heart-rate validity flag, not as a
heart-rate-variability metric.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/eda-intervals-and-kleckner.R
\name{flag_kleckner_eda_artifacts}
\alias{flag_kleckner_eda_artifacts}
\title{Flag EDA artifacts using transparent Kleckner-style heuristics}
\usage{
flag_kleckner_eda_artifacts(
  dat,
  eda_col = "GSR_US",
  time_col = NULL,
  group_cols = NULL,
  min_us = 0.01,
  max_us = 100,
  max_abs_percent_change_per_second = 20,
  transition_padding = 1,
  output_prefix = "kleckner"
)
}
\arguments{
\item{dat}{A data frame containing EDA data.}

\item{eda_col}{Conductance column in microsiemens.}

\item{time_col}{Optional time column.}

\item{group_cols}{Optional grouping columns.}

\item{min_us}{Minimum plausible conductance.}

\item{max_us}{Maximum plausible conductance.}

\item{max_abs_percent_change_per_second}{Maximum absolute percent change per
second before flagging.}

\item{transition_padding}{Number of neighbouring rows to flag around bad
samples within each group.}

\item{output_prefix}{Prefix for output columns.}
}
\value{
A data frame with artifact flag columns and summary attributes.
}
\description{
Applies simple transparent EDA artifact flags: non-finite values,
physiological range violations, rapid percent change per second, and
transitional padding around flagged samples.
}
\details{
This helper is Kleckner-style rather than a claim of exact reproduction of
every rule in a specific external implementation.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/import.R
\name{detect_active_biometric_channels}
\alias{detect_active_biometric_channels}
\title{Detect active Gazepoint biometric channels}
\usage{
detect_active_biometric_channels(data)
}
\arguments{
\item{data}{A data frame imported from a Gazepoint export.}
}
\value{
A data frame with one row per signal family.
}
\description{
Detects whether GSR/EDA, heart-rate, engagement-dial, and TTL channels are
present and whether they appear active. A channel can be present but inactive
when validity flags are zero or the signal contains only zeros or missing
values. For each signal family, \code{summary_column} identifies the primary
column used for the reported minimum and maximum values.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/eda-spectral-power.R
\name{extract_gazepoint_eda_spectral_power}
\alias{extract_gazepoint_eda_spectral_power}
\title{Extract frequency-domain EDA spectral power}
\usage{
extract_gazepoint_eda_spectral_power(
  dat,
  eda_col = "GSR_US",
  time_col = NULL,
  group_cols = NULL,
  sampling_rate = NULL,
  band = c(0.045, 0.25),
  min_samples = 32,
  detrend = TRUE
)
}
\arguments{
\item{dat}{A data frame containing EDA data.}

\item{eda_col}{EDA/conductance column.}

\item{time_col}{Optional time column.}

\item{group_cols}{Optional grouping columns.}

\item{sampling_rate}{Optional sampling rate in Hz. Required if \code{time_col}
does not allow sampling-rate estimation.}

\item{band}{Numeric vector of length two defining the frequency band in Hz.}

\item{min_samples}{Minimum finite samples per group.}

\item{detrend}{Logical. If \code{TRUE}, remove a linear trend before spectral analysis.}
}
\value{
A list with \code{overview}, \code{spectral_summary}, \code{settings}, and
interpretation text.
}
\description{
Computes power spectral density summaries for an EDA signal, including
spectral power in the EDASymp-inspired 0.045--0.25 Hz band. This is a
descriptive spectral feature and should not be interpreted as direct stress,
emotion, valence, cognition, trust, preference, or diagnosis.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compatibility-wrappers.R
\name{join_gazepoint_biometrics_to_gp3tools}
\alias{join_gazepoint_biometrics_to_gp3tools}
\title{Join Gazepoint Biometrics data to gp3tools-style eye-tracking data}
\usage{
join_gazepoint_biometrics_to_gp3tools(biometrics, gp3tools_master, ...)
}
\arguments{
\item{biometrics}{A data frame containing Gazepoint Biometrics samples or
summaries.}

\item{gp3tools_master}{A gp3tools-style master eye-tracking data frame.}

\item{...}{Additional arguments passed to
\code{\link[=join_gazepoint_biometrics_to_master]{join_gazepoint_biometrics_to_master()}}, including the required \code{by}
argument when the underlying join helper requires explicit join columns.}
}
\value{
The output of \code{\link[=join_gazepoint_biometrics_to_master]{join_gazepoint_biometrics_to_master()}}.
}
\description{
Compatibility wrapper for \code{\link[=join_gazepoint_biometrics_to_master]{join_gazepoint_biometrics_to_master()}}. This alias
is provided for users who work with a gp3tools master table and want an
explicit gp3tools-facing function name. The implementation delegates to the
canonical biometric-to-master join helper.
}
\examples{
biometrics <- data.frame(USER = rep("P1", 3), CNT = 1:3, HR = c(70, 71, 72))
master <- data.frame(USER = rep("P1", 3), CNT = 1:3, AOI = c("A", "B", "A"))
join_gazepoint_biometrics_to_gp3tools(
  biometrics,
  master,
  by = c("USER", "CNT")
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/artifact-svm-bridge.R
\name{flag_gazepoint_artifacts_svm}
\alias{flag_gazepoint_artifacts_svm}
\title{Flag EDA artifacts with a user-supplied SVM-style model}
\usage{
flag_gazepoint_artifacts_svm(
  x,
  model = NULL,
  feature_cols = NULL,
  probability_threshold = 0.5,
  ...
)
}
\arguments{
\item{x}{Either raw EDA data or output from
\code{prepare_gazepoint_artifact_svm_features()}.}

\item{model}{Optional model object or function. If \code{NULL}, features are
returned with missing artifact predictions.}

\item{feature_cols}{Optional feature columns used by the model.}

\item{probability_threshold}{Threshold for artifact probability.}

\item{...}{Passed to \code{prepare_gazepoint_artifact_svm_features()} when \code{x}
is raw data.}
}
\value{
A data frame with artifact probabilities/classes where available.
}
\description{
Applies a user-supplied model or prediction function to segment-level
artifact features. No pretrained model is bundled, so this function avoids
pretending to reproduce any proprietary or externally trained classifier.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot-contracts.R
\name{standardise_gazepoint_plot_contract}
\alias{standardise_gazepoint_plot_contract}
\title{Standardise a Gazepoint plot return contract}
\usage{
standardise_gazepoint_plot_contract(
  plot,
  plot_data = NULL,
  settings = list(),
  interpretation_notes = NULL,
  plot_type = NULL
)
}
\arguments{
\item{plot}{A ggplot object.}

\item{plot_data}{Optional data frame used to create the plot.}

\item{settings}{Optional list of plot settings.}

\item{interpretation_notes}{Optional character vector of interpretation notes.}

\item{plot_type}{Optional short plot-type label.}
}
\value{
A ggplot object with standardized attributes.
}
\description{
Adds consistent attributes to a ggplot object so plotting helpers can be
tested and reused in automated reports.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mad-artifact-typology.R
\name{flag_gazepoint_mad_artifacts}
\alias{flag_gazepoint_mad_artifacts}
\title{Flag MAD-based EDA wearable artifacts}
\usage{
flag_gazepoint_mad_artifacts(
  dat,
  eda_col = "GSR_US",
  time_col = NULL,
  group_cols = NULL,
  mad_multiplier = 8,
  flatline_tolerance = 1e-06,
  flatline_min_run = 5,
  wall_abs_change = NULL,
  output_prefix = "mad"
)
}
\arguments{
\item{dat}{A data frame.}

\item{eda_col}{Numeric EDA/conductance column.}

\item{time_col}{Optional time column for ordering within group.}

\item{group_cols}{Optional grouping columns.}

\item{mad_multiplier}{MAD multiplier used for robust thresholding.}

\item{flatline_tolerance}{Maximum absolute sample-to-sample change treated
as flatline.}

\item{flatline_min_run}{Minimum consecutive flatline samples.}

\item{wall_abs_change}{Optional absolute change threshold for wall artifacts.}

\item{output_prefix}{Prefix for output columns.}
}
\value{
A data frame with artifact flags and artifact-summary attributes.
}
\description{
Flags dependency-light, subject-specific EDA artifact categories using robust
median absolute deviation (MAD) logic. The categories are heuristic QC
labels: step artifacts, needle artifacts, flatline artifacts, and wall
artifacts.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/diagnose_workflow.R
\name{diagnose_gazepoint_biometrics_workflow}
\alias{diagnose_gazepoint_biometrics_workflow}
\title{Diagnose a Gazepoint Biometrics workflow}
\usage{
diagnose_gazepoint_biometrics_workflow(
  workflow,
  require_gsr = TRUE,
  require_hr = TRUE,
  require_dial = FALSE,
  max_exclude_window_pct = 25,
  max_review_window_pct = 25
)
}
\arguments{
\item{workflow}{A workflow object produced by
\code{run_gazepoint_biometrics_workflow()}.}

\item{require_gsr}{Should GSR/EDA be required for a pass status?}

\item{require_hr}{Should heart rate be required for a pass status?}

\item{require_dial}{Should engagement dial be required for a pass status?}

\item{max_exclude_window_pct}{Maximum acceptable percentage of excluded
windows before the workflow is marked as fail.}

\item{max_review_window_pct}{Maximum acceptable percentage of review windows
before the workflow is marked as review.}
}
\value{
A one-row data frame with diagnostic status and reasons.
}
\description{
Creates a compact diagnostic readiness summary from a Gazepoint Biometrics
workflow object. The function does not change or remove data. It returns a
one-row decision table with pass/review/fail status and concise reasons.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ttl-alignment.R
\name{align_gazepoint_biometrics_to_ttl}
\alias{align_gazepoint_biometrics_to_ttl}
\title{Align Gazepoint biometric samples to TTL events}
\usage{
align_gazepoint_biometrics_to_ttl(
  data,
  ttl_cols = NULL,
  event_col = NULL,
  ttl_valid_col = NULL,
  time_col = NULL,
  sample_col = NULL,
  group_cols = NULL,
  participant_col = NULL,
  stimulus_col = NULL,
  trial_col = NULL,
  event_value = NULL,
  valid_values = c(TRUE, 1, "1"),
  event_edge = c("rising", "change", "active"),
  pre_window_ms = 1000,
  post_window_ms = 5000,
  pre_window_samples = NULL,
  post_window_samples = NULL,
  collapse_nearby_ms = 0,
  require_valid_ttl = TRUE
)
}
\arguments{
\item{data}{A data frame containing biometric samples.}

\item{ttl_cols}{Optional TTL marker columns. If \code{NULL}, the function first
looks for \code{ttl_marker}, then raw \code{TTL0}-\code{TTL6} columns.}

\item{event_col}{Optional single user-specified event column. If supplied, it
is used instead of automatic TTL-column detection.}

\item{ttl_valid_col}{Optional TTL validity column. If \code{NULL}, the function
looks for \code{ttl_validity_flag} or \code{TTLV}.}

\item{time_col}{Optional time column. If \code{NULL}, common time-column names are
detected automatically when present.}

\item{sample_col}{Optional sample/counter column. If \code{NULL}, \code{CNT}/\code{cnt} is
used when present; otherwise row order is used.}

\item{group_cols}{Optional grouping columns. If \code{NULL}, the function uses
available participant/stimulus/trial-like columns when present.}

\item{participant_col, stimulus_col, trial_col}{Optional explicit grouping
columns to add to \code{group_cols}.}

\item{event_value}{Optional value(s) that define an active event. If \code{NULL},
non-zero numeric/logical values and non-empty character values are treated
as active.}

\item{valid_values}{Values treated as valid in the TTL validity column.}

\item{event_edge}{Event-detection rule. \code{"rising"} keeps inactive-to-active
transitions, \code{"change"} keeps changes among active event values, and
\code{"active"} keeps every active sample.}

\item{pre_window_ms, post_window_ms}{Event window in milliseconds when a usable
time column is available.}

\item{pre_window_samples, post_window_samples}{Event window in samples when no
usable time column is available. If omitted, only the event sample is kept.}

\item{collapse_nearby_ms}{Optional minimum distance between retained events
within a group, in milliseconds.}

\item{require_valid_ttl}{If \code{TRUE}, a detected TTL validity column must be
active for a row to count as an event.}
}
\value{
A list with \code{overview}, \code{events}, \code{aligned_data}, and \code{settings}.
}
\description{
Aligns biometric rows to TTL/event markers and returns event-relative time and
sample indices. The helper is conservative: TTL events are detected from
rising edges by default, validity flags are used when available, and no
physiological interpretation is added.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/time-activity-qc.R
\name{audit_gazepoint_signal_activity}
\alias{audit_gazepoint_signal_activity}
\title{Audit Gazepoint biometric signal activity}
\usage{
audit_gazepoint_signal_activity(
  data,
  signal_cols = NULL,
  group_cols = NULL,
  zero_is_inactive = TRUE,
  min_unique_nonzero = 2,
  missing_as_inactive = TRUE
)
}
\arguments{
\item{data}{A data frame containing Gazepoint biometric rows.}

\item{signal_cols}{Optional signal columns. If \code{NULL}, common Gazepoint
biometric columns are detected automatically.}

\item{group_cols}{Optional grouping columns. If \code{NULL}, available source,
participant, media, or trial columns are used.}

\item{zero_is_inactive}{Logical. If \code{TRUE}, all-zero signals are labelled as
inactive.}

\item{min_unique_nonzero}{Minimum number of distinct non-zero finite values
required for an \code{"active"} status.}

\item{missing_as_inactive}{Logical. If \code{TRUE}, all-missing signals are labelled
as insufficient/inactive.}
}
\value{
A list with \code{overview}, \code{signal_by_group}, \code{inactive_groups},
\code{inactive_signals}, and \code{settings}.
}
\description{
Screens biometric signal columns for missingness, all-zero channels, constant
values, low variation, and active signal presence within groups. This helper
is designed to identify inactive files or channels before event-level EDA,
HR, IBI, or multimodal analysis.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/classical-standardization.R
\name{standardise_gazepoint_range_correction}
\alias{standardise_gazepoint_range_correction}
\alias{standardize_gazepoint_range_correction}
\title{Standardise SCR or SCL using within-participant range correction}
\usage{
standardise_gazepoint_range_correction(
  dat,
  signal_col,
  group_col = "source_participant",
  suffix = "_Range_Corrected",
  min_valid = 2,
  zero_range_action = c("NA", "zero"),
  overwrite = FALSE
)

standardize_gazepoint_range_correction(
  dat,
  signal_col,
  group_col = "source_participant",
  suffix = "_Range_Corrected",
  min_valid = 2,
  zero_range_action = c("NA", "zero"),
  overwrite = FALSE
)
}
\arguments{
\item{dat}{A data frame containing SCR, SCL, or another biometric signal.}

\item{signal_col}{Signal column to range-correct.}

\item{group_col}{Participant or unit column.}

\item{suffix}{Suffix for the output column.}

\item{min_valid}{Minimum finite observations required within each group.}

\item{zero_range_action}{What to do when max equals min: \code{"NA"} or \code{"zero"}.}

\item{overwrite}{Logical. If \code{FALSE}, existing output columns are protected.}
}
\value{
A data frame with an added range-corrected column.
}
\description{
Adds a range-corrected signal column using \code{(x - min) / (max - min)} within
participant or another grouping unit. This expresses each value as a
proportion of the observed within-unit range.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ibi-hr-hrv-core.R
\name{compare_gazepoint_hr_ibi_consistency}
\alias{compare_gazepoint_hr_ibi_consistency}
\title{Compare Gazepoint HR and IBI-derived heart rate}
\usage{
compare_gazepoint_hr_ibi_consistency(
  data,
  hr_col = "HR",
  ibi_col = "IBI",
  time_col = NULL,
  group_cols = NULL,
  unit = c("auto", "ms", "seconds"),
  max_abs_diff_bpm = 10,
  max_rel_diff_prop = 0.15
)
}
\arguments{
\item{data}{A Gazepoint biometric data frame or \code{gazepoint_ibi_filter} object.}

\item{hr_col}{Heart-rate column in beats per minute.}

\item{ibi_col}{IBI/RR interval column.}

\item{time_col}{Optional time/counter column.}

\item{group_cols}{Optional grouping columns.}

\item{unit}{Unit of the IBI column: \code{"auto"}, \code{"ms"}, or \code{"seconds"}.}

\item{max_abs_diff_bpm}{Maximum acceptable absolute HR difference in bpm.}

\item{max_rel_diff_prop}{Maximum acceptable relative HR difference.}
}
\value{
A list with \code{overview}, \code{row_diagnostics}, \code{group_summary}, and
\code{settings}.
}
\description{
Compares recorded HR against HR derived from genuine IBI/RR intervals using
\code{60000 / IBI_ms}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/validity-dropouts.R
\name{summarise_gazepoint_biometric_validity}
\alias{summarise_gazepoint_biometric_validity}
\title{Summarise validity and availability of Gazepoint biometric signals}
\usage{
summarise_gazepoint_biometric_validity(
  data,
  signal_cols = NULL,
  validity_cols = NULL,
  group_cols = NULL,
  active_min_unique = 2L
)
}
\arguments{
\item{data}{A data frame.}

\item{signal_cols}{Optional character vector of biometric signal columns to
summarise. If \code{NULL}, common Gazepoint biometric signal columns are detected
from the column names.}

\item{validity_cols}{Optional character vector of validity-flag columns to
summarise. If \code{NULL}, common validity-like columns are detected, including
\code{HRV} when present.}

\item{group_cols}{Optional character vector of grouping columns, such as
participant, stimulus, trial, or condition columns.}

\item{active_min_unique}{Minimum number of unique finite values required for
a numeric signal to be treated as active.}
}
\value{
A list with \code{overview}, \code{signals}, \code{validity_flags},
\code{group_summary}, and \code{settings}.
}
\description{
Summarises missingness, finite numeric availability, variability, and optional
validity-flag columns for Gazepoint Biometrics data. The helper is descriptive:
it reports whether biometric signals appear available and usable, but it does
not infer emotion, valence, or HRV from ambiguous raw columns. In particular,
raw \code{HRV} columns are treated as validity/vendor flags unless the user has
independent documentation proving otherwise.
}
\examples{
df <- data.frame(
  USER = rep(c("P1", "P2"), each = 4),
  GSR = c(1, 2, NA, 4, 2, 2, 2, 2),
  HR = c(70, 71, 72, NA, 80, 81, 82, 83),
  HRV = c(1, 1, 0, 1, 1, 1, 1, 1)
)
summarise_gazepoint_biometric_validity(df, group_cols = "USER")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sync.R
\name{join_gazepoint_biometrics_to_master}
\alias{join_gazepoint_biometrics_to_master}
\title{Join Gazepoint Biometrics to a master table}
\usage{
join_gazepoint_biometrics_to_master(master, biometrics, by, all_x = TRUE)
}
\arguments{
\item{master}{A master gaze or analysis table.}

\item{biometrics}{A Gazepoint Biometrics data frame.}

\item{by}{Character vector of key columns used for joining.}

\item{all_x}{Logical. Should all rows from \code{master} be retained?}
}
\value{
A data frame with biometric columns joined to the master table.
}
\description{
Convenience wrapper around \code{sync_gazepoint_biometrics_with_gaze()} for joining
biometric data to a \code{gp3tools}-style master table or any other analysis-ready
gaze table.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/preprocess.R
\name{smooth_gazepoint_biometrics}
\alias{smooth_gazepoint_biometrics}
\title{Smooth a Gazepoint biometric signal}
\usage{
smooth_gazepoint_biometrics(
  data,
  value_column,
  window = 5L,
  output_column = NULL,
  na_rm = TRUE
)
}
\arguments{
\item{data}{A data frame or a path to a Gazepoint CSV export.}

\item{value_column}{Column to smooth.}

\item{window}{Number of samples in the moving window. Must be a positive odd
integer.}

\item{output_column}{Name of the smoothed output column.}

\item{na_rm}{Should missing values be ignored within the moving window?}
}
\value{
A data frame with the added smoothed column.
}
\description{
Adds a simple centered moving-average smoothing column to a Gazepoint
Biometrics table. This is intentionally conservative and dependency-free.
It does not replace specialised biosignal preprocessing libraries.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hrv-fuzzy-csi.R
\name{extract_gazepoint_hrv_fuzzy_csi}
\alias{extract_gazepoint_hrv_fuzzy_csi}
\title{Extract FuzzyEn and Lorenz-plot CSI HRV features}
\usage{
extract_gazepoint_hrv_fuzzy_csi(
  dat,
  ibi_col = "IBI",
  group_cols = NULL,
  m = 2,
  r_multiplier = 0.2,
  fuzzy_power = 2,
  min_intervals = 10
)
}
\arguments{
\item{dat}{A data frame.}

\item{ibi_col}{Numeric IBI/RR interval column.}

\item{group_cols}{Optional grouping columns.}

\item{m}{Embedding dimension.}

\item{r_multiplier}{Tolerance multiplier applied to within-group SD.}

\item{fuzzy_power}{Fuzzy exponential power.}

\item{min_intervals}{Minimum intervals per group.}
}
\value{
A list with \code{overview}, \code{features}, and \code{settings}.
}
\description{
Computes fuzzy entropy and Lorenz/Poincare-derived cardiac sympathetic index
style descriptors from IBI/RR intervals.
}
\details{
These outputs are nonlinear/geometric HRV descriptors. They do not infer
seizure status, diagnosis, health status, emotion, stress, or cognition.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aoi-biometrics.R
\name{prepare_gazepoint_aoi_biometrics_model_data}
\alias{prepare_gazepoint_aoi_biometrics_model_data}
\title{Prepare AOI-biometric model data}
\usage{
prepare_gazepoint_aoi_biometrics_model_data(
  x,
  outcome_col = "mean_value",
  predictor_cols = c("aoi_label", "signal"),
  factor_cols = c("aoi_label", "signal"),
  numeric_cols = NULL,
  group_cols = NULL,
  drop_missing_outcome = TRUE,
  min_rows = NULL,
  standardise_outcome = FALSE,
  standardise_within = c("signal", "all")
)
}
\arguments{
\item{x}{A \code{gazepoint_aoi_biometrics_summary} object or summary data frame.}

\item{outcome_col}{Outcome column to model.}

\item{predictor_cols}{Optional predictor columns to retain.}

\item{factor_cols}{Optional columns converted to factors.}

\item{numeric_cols}{Optional columns converted to numeric.}

\item{group_cols}{Optional grouping columns for random-effect formulas.}

\item{drop_missing_outcome}{Logical. If \code{TRUE}, rows with missing outcomes are
removed.}

\item{min_rows}{Optional minimum contributing rows required.}

\item{standardise_outcome}{Logical. If \code{TRUE}, add a z-scored outcome column.}

\item{standardise_within}{Standardization scope used when
\code{standardise_outcome = TRUE}. Use \code{"signal"} to z-score within each
biometric signal or \code{"all"} to z-score across all rows.}
}
\value{
A list with \code{overview}, \code{model_data}, \code{variable_summary},
\code{model_formulas}, and \code{settings}.
}
\description{
Converts AOI-biometric summaries into a modelling-ready table for GLM/LMM/GLMM
workflows.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/environment-confounds-and-ipfm.R
\name{model_gazepoint_hrv_ipfm}
\alias{model_gazepoint_hrv_ipfm}
\title{Model heartbeat timing using an IPFM-style impulse train}
\usage{
model_gazepoint_hrv_ipfm(
  dat,
  ibi_col = "IBI",
  beat_time_col = NULL,
  group_cols = NULL,
  ibi_units = c("auto", "seconds", "milliseconds"),
  output_sampling_rate = 4,
  max_frequency = 0.5
)
}
\arguments{
\item{dat}{A data frame.}

\item{ibi_col}{Optional numeric IBI/RR interval column.}

\item{beat_time_col}{Optional explicit beat-time column.}

\item{group_cols}{Optional grouping columns.}

\item{ibi_units}{\code{"auto"}, \code{"seconds"}, or \code{"milliseconds"}.}

\item{output_sampling_rate}{Sampling rate for regular impulse train in Hz.}

\item{max_frequency}{Maximum frequency returned in spectrum.}
}
\value{
A list with \code{overview}, \code{beat_table}, \code{impulse_table},
\code{spectrum_table}, \code{summary}, and \code{settings}.
}
\description{
Builds an impulse-train representation of heartbeat timing from IBI/RR
intervals or beat times and computes a simple spectrum of the resulting
impulse train. This is an IPFM-style model-preparation helper, not a perfect
reconstruction of sinoatrial-node physiology.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/feature-inventory-format.R
\name{format_gazepoint_biometrics_feature_inventory}
\alias{format_gazepoint_biometrics_feature_inventory}
\title{Format the gpbiometrics feature inventory for users}
\usage{
format_gazepoint_biometrics_feature_inventory(
  inventory = NULL,
  include_internal = FALSE,
  sort = TRUE
)
}
\arguments{
\item{inventory}{Optional inventory object returned by
\code{create_gazepoint_biometrics_feature_inventory()}. If \code{NULL}, a fresh
inventory is created.}

\item{include_internal}{Logical. Passed to
\code{create_gazepoint_biometrics_feature_inventory()} when \code{inventory = NULL}.}

\item{sort}{Logical. If \code{TRUE}, sort by domain, user level, and function name.}
}
\value{
A data frame with polished user-facing inventory columns.
}
\description{
Adds user-facing labels and interpretation metadata to the package feature
inventory. This helper is intentionally non-breaking: it does not replace the
core inventory object returned by \code{create_gazepoint_biometrics_feature_inventory()}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gsr-eda-helpers.R
\name{convert_gazepoint_gsr_to_conductance}
\alias{convert_gazepoint_gsr_to_conductance}
\title{Convert Gazepoint GSR resistance to conductance}
\usage{
convert_gazepoint_gsr_to_conductance(
  data,
  gsr_col = NULL,
  output_col = "GSR_US",
  input_unit = c("auto", "ohms", "kohms", "microsiemens"),
  overwrite = FALSE
)
}
\arguments{
\item{data}{A data frame.}

\item{gsr_col}{Optional source GSR column. If \code{NULL}, a resistance-like column
is detected when possible.}

\item{output_col}{Name of the output conductance column.}

\item{input_unit}{Source unit. \code{"auto"} converts only resistance-like columns;
\code{"ohms"} converts ohms to microSiemens; \code{"kohms"} converts kilo-ohms to
microSiemens; \code{"microsiemens"} copies values to \code{output_col}.}

\item{overwrite}{Logical. If \code{FALSE}, an existing \code{output_col} is not
overwritten.}
}
\value{
The input data frame with a conductance column when conversion is
possible. A structured conversion summary is stored in the
\code{gsr_conversion_summary} attribute.
}
\description{
Converts GSR resistance values to conductance in microSiemens. The helper is
intentionally conservative. If a conductance column such as \code{GSR_US} is
already present, the data are returned unchanged unless \code{overwrite = TRUE}.
If \code{input_unit = "auto"}, conversion is performed only when the source column
has a resistance-like name, such as \code{GSR_OHMS} or \code{resistance_ohms}.
}
\details{
Generic \code{GSR} columns are not automatically assumed to be resistance because
Gazepoint exports and workflows may represent GSR/EDA differently. For a
generic \code{GSR} column, use \code{input_unit = "ohms"} or \code{input_unit = "kohms"}
only when the study documentation confirms resistance units.
}
\examples{
df <- data.frame(GSR_OHMS = c(1000000, 500000, NA))
convert_gazepoint_gsr_to_conductance(df)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/advanced-modality-bridges.R
\name{extract_gazepoint_bilateral_eda_asymmetry}
\alias{extract_gazepoint_bilateral_eda_asymmetry}
\title{Extract bilateral EDA asymmetry features}
\usage{
extract_gazepoint_bilateral_eda_asymmetry(
  dat,
  left_col,
  right_col,
  time_col = NULL,
  group_cols = NULL,
  output_prefix = "beda"
)
}
\arguments{
\item{dat}{A data frame.}

\item{left_col}{Numeric left-side EDA column.}

\item{right_col}{Numeric right-side EDA column.}

\item{time_col}{Optional numeric time column for ordering and gradient
calculation.}

\item{group_cols}{Optional grouping columns.}

\item{output_prefix}{Prefix used for row-level output columns.}
}
\value{
A list with \code{overview}, \code{asymmetry_timeseries}, \code{summary}, and
\code{settings}.
}
\description{
Computes left-right electrodermal activity asymmetry descriptors from two
simultaneously recorded EDA channels. The function returns row-level
asymmetry time series and group-level summaries.
}
\details{
These descriptors quantify bilateral EDA differences only. They do not infer
hemisphere activation, amygdala activity, psychopathology, emotion, stress,
cognition, health status, or diagnosis.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ibi-hr-hrv-core.R
\name{filter_gazepoint_ibi_implausible}
\alias{filter_gazepoint_ibi_implausible}
\title{Filter implausible Gazepoint IBI values}
\usage{
filter_gazepoint_ibi_implausible(
  data,
  ibi_col = "IBI",
  time_col = NULL,
  group_cols = NULL,
  validity_col = NULL,
  unit = c("auto", "ms", "seconds"),
  min_ibi_ms = 300,
  max_ibi_ms = 2000,
  max_change_ms = 400,
  max_change_prop = 0.3,
  output_col = "IBI_clean_ms"
)
}
\arguments{
\item{data}{A Gazepoint biometric data frame.}

\item{ibi_col}{IBI/RR interval column.}

\item{time_col}{Optional time/counter column.}

\item{group_cols}{Optional grouping columns.}

\item{validity_col}{Optional validity column.}

\item{unit}{Unit of the IBI column: \code{"auto"}, \code{"ms"}, or \code{"seconds"}.}

\item{min_ibi_ms}{Minimum plausible IBI in milliseconds.}

\item{max_ibi_ms}{Maximum plausible IBI in milliseconds.}

\item{max_change_ms}{Maximum plausible absolute adjacent IBI change within
group, in milliseconds.}

\item{max_change_prop}{Maximum plausible proportional adjacent IBI change
within group.}

\item{output_col}{Name of the cleaned IBI output column.}
}
\value{
A list with \code{overview}, \code{data}, \code{row_flags}, \code{group_summary}, and
\code{settings}.
}
\description{
Flags and optionally cleans implausible inter-beat interval values. The helper
is conservative and does not remove rows; instead, it returns row-level flags
and a cleaned IBI column with implausible values set to \code{NA}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hrv-fragmentation-asymmetry.R
\name{extract_gazepoint_hrv_asymmetry}
\alias{extract_gazepoint_hrv_asymmetry}
\title{Extract heart-rate asymmetry features}
\usage{
extract_gazepoint_hrv_asymmetry(
  dat,
  ibi_col = "IBI",
  group_cols = NULL,
  zero_tolerance = 0
)
}
\arguments{
\item{dat}{A data frame containing IBI/RR intervals.}

\item{ibi_col}{Numeric IBI/RR interval column.}

\item{group_cols}{Optional grouping columns.}

\item{zero_tolerance}{Absolute change below which interval differences are
treated as zero.}
}
\value{
A list with \code{overview}, \code{features}, \code{run_table}, and \code{settings}.
}
\description{
Computes dependency-light heart-rate asymmetry descriptors from IBI/RR
intervals, including acceleration/deceleration proportions, signed run
summaries, and Guzik-style squared-difference asymmetry.
}
\details{
Positive IBI/RR differences are treated as decelerations because the heart
period lengthens. Negative IBI/RR differences are treated as accelerations.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/report_tables.R
\name{create_gazepoint_biometrics_report_tables}
\alias{create_gazepoint_biometrics_report_tables}
\title{Create Gazepoint Biometrics report tables}
\usage{
create_gazepoint_biometrics_report_tables(
  workflow = NULL,
  validation = NULL,
  quality = NULL,
  sampling = NULL,
  diagnostics = NULL,
  exclusion_recommendations = NULL,
  ttl_events = NULL,
  max_ttl_events = 20
)
}
\arguments{
\item{workflow}{Optional workflow object produced by
\code{run_gazepoint_biometrics_workflow()}.}

\item{validation}{Optional validation object produced by
\code{validate_gazepoint_biometrics()}.}

\item{quality}{Optional quality-audit table.}

\item{sampling}{Optional sampling/timing audit table produced by
\code{audit_gazepoint_biometric_sampling()}.}

\item{diagnostics}{Optional diagnostic table produced by
\code{diagnose_gazepoint_biometrics_workflow()}.}

\item{exclusion_recommendations}{Optional object produced by
\code{recommend_gazepoint_biometric_exclusions()}.}

\item{ttl_events}{Optional TTL event table produced by
\code{extract_gazepoint_ttl_events()}.}

\item{max_ttl_events}{Maximum number of TTL events to include in the compact
TTL event table.}
}
\value{
A list of report-ready tables.
}
\description{
Creates compact report-ready tables from a Gazepoint Biometrics workflow
object or from separate workflow components. The function does not write
files. It returns cleaned tables that can be printed, exported, or inserted
into reports and supplementary materials.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/qc-dashboard-plots.R
\name{plot_gazepoint_biometric_report_dashboard}
\alias{plot_gazepoint_biometric_report_dashboard}
\title{Create a lightweight Gazepoint biometric QC plot dashboard}
\usage{
plot_gazepoint_biometric_report_dashboard(
  data = NULL,
  signal_activity = NULL,
  time_resets = NULL,
  signal_cols = NULL,
  group_cols = NULL,
  time_col = NULL,
  include_signal_activity = TRUE,
  include_time_resets = TRUE,
  max_groups = 30,
  continue_on_error = TRUE,
  title_prefix = "Gazepoint biometric QC"
)
}
\arguments{
\item{data}{Optional Gazepoint biometric data frame.}

\item{signal_activity}{Optional result from
\code{\link[=audit_gazepoint_signal_activity]{audit_gazepoint_signal_activity()}}.}

\item{time_resets}{Optional result from \code{\link[=audit_gazepoint_time_resets]{audit_gazepoint_time_resets()}}.}

\item{signal_cols}{Optional signal columns used when computing signal
activity from \code{data}.}

\item{group_cols}{Optional grouping columns.}

\item{time_col}{Optional time or counter column.}

\item{include_signal_activity}{If \code{TRUE}, include a signal-activity plot.}

\item{include_time_resets}{If \code{TRUE}, include a time-reset plot.}

\item{max_groups}{Maximum number of groups to display in each plot.}

\item{continue_on_error}{If \code{TRUE}, plot failures are recorded in \code{errors}
rather than stopping the dashboard.}

\item{title_prefix}{Optional title prefix.}
}
\value{
A list with \code{overview}, \code{plots}, \code{errors}, \code{inputs}, and \code{settings}.
}
\description{
Creates a lightweight dashboard object containing QC plots for signal activity
and time-reset diagnostics. This is a structured list of ggplot objects, not a
Shiny application. The dashboard is intended for report preparation and manual
QC review.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/windows.R
\name{summarise_gazepoint_multimodal_windows}
\alias{summarise_gazepoint_multimodal_windows}
\title{Summarise Gazepoint multimodal biometric windows}
\usage{
summarise_gazepoint_multimodal_windows(
  data,
  group_columns = NULL,
  exclude_zero = TRUE
)
}
\arguments{
\item{data}{A data frame or a path to a Gazepoint CSV export.}

\item{group_columns}{Optional grouping columns defining windows, such as
\code{c("USER", "MEDIA_ID")}.}

\item{exclude_zero}{Should zero values be excluded from GSR and heart-rate
summaries?}
}
\value{
A data frame with one row per window and prefixed biometric summary
columns.
}
\description{
Creates a combined window-level summary table for GSR/EDA, heart rate, and
engagement dial. The output is suitable for later merging with eye-tracking
summaries from \code{gp3tools}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot-contracts.R
\name{get_gazepoint_plot_settings}
\alias{get_gazepoint_plot_settings}
\title{Extract stored plot settings}
\usage{
get_gazepoint_plot_settings(plot)
}
\arguments{
\item{plot}{A plot object returned by a gpbiometrics plotting helper.}
}
\value{
A list.
}
\description{
Extract stored plot settings
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/qc-dashboard-plots.R
\name{plot_gazepoint_signal_activity}
\alias{plot_gazepoint_signal_activity}
\title{Plot Gazepoint biometric signal activity}
\usage{
plot_gazepoint_signal_activity(
  data,
  signal_cols = NULL,
  group_cols = NULL,
  metric = c("active_signal", "nonzero_prop", "missing_prop", "n_unique_finite"),
  max_groups = 30,
  title = NULL
)
}
\arguments{
\item{data}{A Gazepoint biometric data frame, or an
\code{audit_gazepoint_signal_activity()} result.}

\item{signal_cols}{Optional signal columns used when \code{data} is a raw data
frame.}

\item{group_cols}{Optional grouping columns used when \code{data} is a raw data
frame.}

\item{metric}{Summary metric to plot.}

\item{max_groups}{Maximum number of groups to display.}

\item{title}{Optional plot title.}
}
\value{
A ggplot object with the package plot contract attached.
}
\description{
Plots signal-activity summaries produced by
\code{\link[=audit_gazepoint_signal_activity]{audit_gazepoint_signal_activity()}}, or computes them from a biometric data
frame. The plot is intended for quality-control review of signal
availability, missingness, zero activity, and basic activity status. It does
not infer emotion, valence, cognition, preference, trust, or physiological
diagnosis.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/environment-confounds-and-ipfm.R
\name{regress_gazepoint_pupil_luminance}
\alias{regress_gazepoint_pupil_luminance}
\title{Regress stimulus luminance from pupil diameter}
\usage{
regress_gazepoint_pupil_luminance(
  dat,
  pupil_col,
  luminance_col,
  group_cols = NULL,
  time_col = NULL,
  output_col = "pupil_luminance_adjusted",
  fitted_col = "pupil_luminance_fitted",
  include_quadratic = TRUE,
  model_by_group = TRUE,
  add_intercept_mean = TRUE
)
}
\arguments{
\item{dat}{A data frame.}

\item{pupil_col}{Numeric pupil column.}

\item{luminance_col}{Numeric luminance/brightness column.}

\item{group_cols}{Optional grouping columns.}

\item{time_col}{Optional time column.}

\item{output_col}{Output luminance-adjusted pupil column.}

\item{fitted_col}{Output fitted luminance component column.}

\item{include_quadratic}{Logical. If \code{TRUE}, include luminance squared.}

\item{model_by_group}{Logical. If \code{TRUE}, fit models per group.}

\item{add_intercept_mean}{Logical. If \code{TRUE}, add mean pupil size back to
residuals.}
}
\value{
A data frame with luminance-adjusted pupil columns and attributes.
}
\description{
Regresses continuous pupil diameter on frame-wise or sample-wise stimulus
luminance and returns a luminance-adjusted pupil series. This controls a
major visual confound but does not prove that residual pupil changes are
cognitive-load-only effects.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/windows.R
\name{summarise_gazepoint_gsr_windows}
\alias{summarise_gazepoint_gsr_windows}
\title{Summarise Gazepoint GSR/EDA windows}
\usage{
summarise_gazepoint_gsr_windows(
  data,
  group_columns = NULL,
  value_column = NULL,
  validity_column = "GSRV",
  exclude_zero = TRUE
)
}
\arguments{
\item{data}{A data frame or a path to a Gazepoint CSV export.}

\item{group_columns}{Optional grouping columns defining windows, such as
\code{c("USER", "MEDIA_ID")}.}

\item{value_column}{Optional GSR/EDA value column. If \code{NULL}, \code{GSR_US} is
used when present, otherwise \code{GSR}.}

\item{validity_column}{Optional validity column. Defaults to \code{"GSRV"}.}

\item{exclude_zero}{Should zero values be excluded from usable summaries?}
}
\value{
A data frame with one row per window.
}
\description{
Summarises Gazepoint GSR/EDA values within participant, trial, stimulus, AOI,
or other user-defined windows. When available, \code{GSR_US} is used by default
because it represents skin conductance in microsiemens in Gazepoint exports.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ac-eda-susceptance.R
\name{analyze_gazepoint_ac_susceptance}
\alias{analyze_gazepoint_ac_susceptance}
\title{Analyse AC EDA admittance and susceptance recordings}
\usage{
analyze_gazepoint_ac_susceptance(
  dat,
  conductance_col = NULL,
  susceptance_col = NULL,
  admittance_col = NULL,
  phase_col = NULL,
  frequency_col = NULL,
  time_col = NULL,
  group_cols = NULL
)
}
\arguments{
\item{dat}{A data frame.}

\item{conductance_col}{Optional real conductance component column.}

\item{susceptance_col}{Optional imaginary susceptance component column.}

\item{admittance_col}{Optional admittance magnitude column.}

\item{phase_col}{Optional phase angle column.}

\item{frequency_col}{Optional AC frequency column.}

\item{time_col}{Optional time column.}

\item{group_cols}{Optional grouping columns.}
}
\value{
A list with \code{overview}, \code{timeseries}, \code{summary}, and \code{settings}.
}
\description{
Computes summaries for specialised alternating-current EDA recordings. This
function is for true AC admittance/susceptance data, not ordinary DC
skin-conductance columns such as \code{GSR_US}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/eda-analysis-pipeline.R
\name{create_gazepoint_eda_analysis_pipeline}
\alias{create_gazepoint_eda_analysis_pipeline}
\title{Create a Gazepoint EDA analysis pipeline guide}
\usage{
create_gazepoint_eda_analysis_pipeline(
  include_external_bridges = TRUE,
  include_model_templates = TRUE,
  include_reporting_guidance = TRUE,
  style = c("compact", "detailed")
)
}
\arguments{
\item{include_external_bridges}{Logical. If \code{TRUE}, include NeuroKit2,
Ledalab-style, PsPM-style, cvxEDA-style, RHRV, and pyPPG bridge helpers.}

\item{include_model_templates}{Logical. If \code{TRUE}, include text templates
for downstream \code{brms} hurdle models and \code{lme4} mixed-effects models.}

\item{include_reporting_guidance}{Logical. If \code{TRUE}, include reporting and
interpretation guardrails.}

\item{style}{Output style. \code{"compact"} returns concise phase descriptions;
\code{"detailed"} returns fuller phase notes.}
}
\value{
A list with \code{overview}, \code{phases}, \code{function_map},
\code{model_templates}, \code{reporting_guidance}, \code{interpretation_guardrails}, and
\code{settings}.
}
\description{
Creates a structured six-phase analysis pipeline for Gazepoint
Biometrics EDA/GSR workflows. The helper maps each phase to native
\code{gpbiometrics} functions, optional external-method bridges, and optional
downstream model templates for \code{brms} and \code{lme4}.
}
\details{
This function does not fit statistical models, does not run external
software, and does not infer emotion, valence, stress, trust, preference,
cognition, or diagnosis. It is a reproducible planning and reporting aid.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/distributional-drift.R
\name{audit_gazepoint_distributional_drift}
\alias{audit_gazepoint_distributional_drift}
\title{Audit distributional drift across sessions or ordered blocks}
\usage{
audit_gazepoint_distributional_drift(
  dat,
  signal_cols,
  session_col = "session",
  participant_col = NULL,
  reference_session = NULL,
  bins = 10,
  psi_warn = 0.1,
  psi_fail = 0.25
)
}
\arguments{
\item{dat}{A data frame containing longitudinal biometric data.}

\item{signal_cols}{Numeric signal columns to audit.}

\item{session_col}{Session/block/timepoint column.}

\item{participant_col}{Optional participant column.}

\item{reference_session}{Optional reference session. If \code{NULL}, the first
ordered session is used within each participant/global group.}

\item{bins}{Number of bins for PSI.}

\item{psi_warn}{PSI threshold for warning.}

\item{psi_fail}{PSI threshold for failure.}
}
\value{
A list with \code{overview}, \code{drift_summary}, and \code{settings}.
}
\description{
Compares signal distributions across sessions/blocks using baseline-vs-current
differences, Kolmogorov-Smirnov tests, and Population Stability Index (PSI).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/full_windows.R
\name{summarise_gazepoint_full_biometric_windows}
\alias{summarise_gazepoint_full_biometric_windows}
\title{Summarise full Gazepoint biometric windows}
\usage{
summarise_gazepoint_full_biometric_windows(
  data,
  group_columns,
  include_ibi_hrv = TRUE
)
}
\arguments{
\item{data}{A data frame or a path to a Gazepoint CSV export.}

\item{group_columns}{Columns defining analysis windows, such as
\code{c("source_participant", "MEDIA_ID")}.}

\item{include_ibi_hrv}{Logical. Should IBI-derived HRV summaries be included?}
}
\value{
A data frame with one row per window and prefixed biometric summary
columns.
}
\description{
Creates a combined window-level summary table containing GSR/EDA,
heart-rate, engagement-dial, and IBI-derived HRV summaries. This function is
intended for biometric analyses where both continuous physiological values
and interbeat-interval variability features are needed.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/eda-scr-plots.R
\name{plot_gazepoint_scr_events}
\alias{plot_gazepoint_scr_events}
\title{Plot Gazepoint SCR events on an EDA signal}
\usage{
plot_gazepoint_scr_events(
  data,
  scr_peaks,
  event_windows = NULL,
  events = NULL,
  time_col = NULL,
  signal_col = NULL,
  phasic_col = NULL,
  group_cols = NULL,
  show_events = TRUE,
  max_points = 5000,
  title = NULL
)
}
\arguments{
\item{data}{Gazepoint biometric data frame.}

\item{scr_peaks}{A \code{gazepoint_scr_peak_detection} object or peak data frame.}

\item{event_windows}{Optional \code{gazepoint_scr_event_window_summary} object or
event-window data frame.}

\item{events}{Optional event table used when \code{event_windows} is not supplied.}

\item{time_col}{Optional time/counter column.}

\item{signal_col}{Optional signal column to plot.}

\item{phasic_col}{Optional preferred phasic signal column.}

\item{group_cols}{Optional grouping columns used for facets and matching.}

\item{show_events}{Logical. If \code{TRUE}, show event onsets when available.}

\item{max_points}{Maximum number of signal rows retained after downsampling.}

\item{title}{Optional plot title.}
}
\value{
A ggplot object with plot data stored in attributes.
}
\description{
Plots an EDA/GSR signal with detected SCR peak markers and optional event
onsets from SCR event-window summaries or event tables.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compatibility-aliases.R
\name{standardize_gazepoint_plot_contracts}
\alias{standardize_gazepoint_plot_contracts}
\title{Standardize Gazepoint plot return contracts}
\usage{
standardize_gazepoint_plot_contracts(
  plot,
  plot_data = NULL,
  settings = list(),
  interpretation_notes = NULL,
  plot_type = NULL
)
}
\arguments{
\item{plot}{A ggplot object, or a list of ggplot objects.}

\item{plot_data}{Optional data frame, or a list of data frames when \code{plot} is
a list.}

\item{settings}{Optional settings list, or a list of settings lists when
\code{plot} is a list.}

\item{interpretation_notes}{Optional character vector, or a list/character
vector of notes when \code{plot} is a list.}

\item{plot_type}{Optional plot-type label, or a character vector/list of
labels when \code{plot} is a list.}
}
\value{
A standardized ggplot object, or a list of standardized ggplot
objects.
}
\description{
US-spelling compatibility wrapper around
\code{\link[=standardise_gazepoint_plot_contract]{standardise_gazepoint_plot_contract()}}. The wrapper accepts either a single
ggplot object or a list of ggplot objects. For a list of plots, \code{plot_data},
\code{settings}, \code{interpretation_notes}, and \code{plot_type} may be supplied either as
single values applied to all plots or as same-length lists/vectors applied
elementwise.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pupil-and-saccade-diagnostics.R
\name{baseline_correct_gazepoint_pupil}
\alias{baseline_correct_gazepoint_pupil}
\title{Baseline-correct Gazepoint pupil size}
\usage{
baseline_correct_gazepoint_pupil(
  dat,
  pupil_col = NULL,
  time_col = "CNT",
  stimulus_onset_col = NULL,
  trial_cols = NULL,
  baseline_window = c(-240, -200),
  baseline_function = c("median", "mean"),
  correction = c("subtract", "divide"),
  suffix = "_baseline_corrected",
  min_baseline_rows = 2,
  overwrite = FALSE
)
}
\arguments{
\item{dat}{A data frame containing pupil-size data.}

\item{pupil_col}{Pupil column. If \code{NULL}, common Gazepoint pupil columns are
detected.}

\item{time_col}{Time column.}

\item{stimulus_onset_col}{Optional stimulus-onset column. If supplied,
baseline windows are interpreted relative to onset.}

\item{trial_cols}{Trial/grouping columns.}

\item{baseline_window}{Numeric vector of length two defining the baseline
window relative to stimulus onset.}

\item{baseline_function}{\code{"median"} or \code{"mean"}.}

\item{correction}{\code{"subtract"} or \code{"divide"}.}

\item{suffix}{Suffix for corrected output column.}

\item{min_baseline_rows}{Minimum finite baseline rows required.}

\item{overwrite}{Logical. If \code{FALSE}, existing output columns are protected.}
}
\value{
A data frame with a corrected pupil column and baseline attributes.
}
\description{
Applies subtractive or divisive baseline correction to pupil-size data within
trials or other grouping units.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/simulation-and-chunking.R
\name{chunk_gazepoint_biometrics}
\alias{chunk_gazepoint_biometrics}
\title{Chunk Gazepoint biometric data into fixed analysis episodes}
\usage{
chunk_gazepoint_biometrics(
  dat,
  time_col = "CNT",
  group_cols = NULL,
  chunk_seconds = 60,
  start_time = NULL,
  chunk_col = "chunk_id",
  episode_col = "episode_id",
  include_partial = FALSE
)
}
\arguments{
\item{dat}{A data frame.}

\item{time_col}{Numeric time column.}

\item{group_cols}{Optional grouping columns.}

\item{chunk_seconds}{Chunk duration in seconds.}

\item{start_time}{Optional fixed start time. If \code{NULL}, uses group minimum.}

\item{chunk_col}{Output chunk identifier column.}

\item{episode_col}{Output episode label column.}

\item{include_partial}{Logical. If \code{FALSE}, partial final chunks are marked
but not assigned as complete chunks.}
}
\value{
A data frame with chunk columns and chunk-summary attributes.
}
\description{
Adds programmatic fixed-duration chunks/episodes to continuous biometric data.
This is useful for baseline segmentation, repeated-measures feature
extraction, and analyses that do not rely on external TTL markers.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shiny-and-preregistration.R
\name{create_gazepoint_preregistration_template}
\alias{create_gazepoint_preregistration_template}
\title{Create a Gazepoint biometrics preregistration template}
\usage{
create_gazepoint_preregistration_template(
  study_title = "Gazepoint biometrics study",
  signal_standardization = c("within_participant_z", "range_correction", "none"),
  artifact_rules = c("kleckner_style", "custom", "none"),
  eda_min_us = 0.01,
  eda_max_us = 100,
  rapid_change_threshold = 20,
  output_file = NULL
)
}
\arguments{
\item{study_title}{Study title.}

\item{signal_standardization}{Standardization plan.}

\item{artifact_rules}{Artifact-rule description.}

\item{eda_min_us}{Minimum conductance threshold.}

\item{eda_max_us}{Maximum conductance threshold.}

\item{rapid_change_threshold}{Maximum absolute percent change per second.}

\item{output_file}{Optional path to write the template as a text file.}
}
\value{
A character string.
}
\description{
Creates a cautious preregistration template for Gazepoint Biometrics EDA/GSR
workflows.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gsr-units-audit.R
\name{audit_gazepoint_gsr_units}
\alias{audit_gazepoint_gsr_units}
\title{Audit likely GSR/EDA units}
\usage{
audit_gazepoint_gsr_units(
  dat,
  gsr_col = "GSR",
  convert = FALSE,
  output_col = NULL,
  resistance_to_us_factor = 1e+06
)
}
\arguments{
\item{dat}{A data frame.}

\item{gsr_col}{Name of the GSR/EDA column to audit.}

\item{convert}{Logical. If \code{TRUE}, add a conductance-converted column when
the signal is likely resistance/impedance-like.}

\item{output_col}{Output column used when \code{convert = TRUE}.}

\item{resistance_to_us_factor}{Conversion factor. For Ohms to microSiemens,
use \code{1000000 / resistance}.}
}
\value{
A list with \code{overview}, \code{diagnostics}, \code{recommendation}, and,
when requested, \code{data}.
}
\description{
Checks whether a Gazepoint GSR/EDA column looks more like conductance
in microSiemens or resistance/impedance-like values in Ohms. This is a
preprocessing safety audit, not a definitive device calibration test.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/respiration-ceemdan-kalman.R
\name{fuse_gazepoint_respiration_kalman}
\alias{fuse_gazepoint_respiration_kalman}
\title{Fuse respiration proxies using a Kalman filter}
\usage{
fuse_gazepoint_respiration_kalman(
  dat,
  primary_col,
  secondary_col,
  time_col = NULL,
  group_cols = NULL,
  process_var = 0.01,
  primary_var = 0.05,
  secondary_var = 0.05,
  output_col = "respiration_kalman_fused"
)
}
\arguments{
\item{dat}{A data frame.}

\item{primary_col}{First respiration proxy column.}

\item{secondary_col}{Second respiration proxy column.}

\item{time_col}{Optional time column.}

\item{group_cols}{Optional grouping columns.}

\item{process_var}{Process variance.}

\item{primary_var}{Measurement variance for \code{primary_col}.}

\item{secondary_var}{Measurement variance for \code{secondary_col}.}

\item{output_col}{Output fused respiration column.}
}
\value{
A data frame with fused respiration output and Kalman attributes.
}
\description{
Fuses two respiration proxy streams, such as PPG-derived respiration and
ECG-derived respiration, using a transparent one-dimensional Kalman filter.
This is a linear Kalman fusion helper. It is not an extended Kalman filter
unless the user supplies nonlinear state/measurement logic externally.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/advanced-modality-bridges.R
\name{analyze_gazepoint_skin_potential}
\alias{analyze_gazepoint_skin_potential}
\title{Analyse endosomatic skin-potential recordings}
\usage{
analyze_gazepoint_skin_potential(
  dat,
  sp_col,
  time_col,
  group_cols = NULL,
  response_direction = c("both", "positive", "negative"),
  response_threshold = NULL,
  min_response_distance_s = 1
)
}
\arguments{
\item{dat}{A data frame.}

\item{sp_col}{Numeric skin-potential column, usually in millivolts.}

\item{time_col}{Numeric time column.}

\item{group_cols}{Optional grouping columns.}

\item{response_direction}{\code{"both"}, \code{"positive"}, or \code{"negative"}.}

\item{response_threshold}{Optional absolute threshold for response detection.
If \code{NULL}, a MAD-based derivative threshold is used.}

\item{min_response_distance_s}{Minimum distance between detected responses.}
}
\value{
A list with \code{overview}, \code{level_summary}, \code{response_table},
\code{timeseries}, and \code{settings}.
}
\description{
Computes skin-potential level and skin-potential response descriptors from a
voltage-like skin-potential column. This is for endosomatic skin-potential
recordings, not standard exosomatic skin conductance.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sampling.R
\name{audit_gazepoint_biometric_sampling}
\alias{audit_gazepoint_biometric_sampling}
\title{Audit Gazepoint biometric sampling and timing}
\usage{
audit_gazepoint_biometric_sampling(
  data,
  group_columns = NULL,
  time_column = NULL,
  time_unit = c("seconds", "milliseconds", "microseconds", "samples"),
  expected_rate_hz = 60,
  tolerance_hz = 5
)
}
\arguments{
\item{data}{A data frame or a path to a Gazepoint CSV export.}

\item{group_columns}{Optional grouping columns within which timing should be
audited, such as \code{c("source_participant", "MEDIA_ID")}.}

\item{time_column}{Optional time/order column. If \code{NULL}, the function uses
the first available column among \code{TIME}, \code{TIME_TICK}, and \code{CNT}.}

\item{time_unit}{Unit of the selected time column. Use \code{"seconds"},
\code{"milliseconds"}, \code{"microseconds"}, or \code{"samples"}. When \code{"samples"} is
used, sampling-rate estimates are returned as \code{NA}.}

\item{expected_rate_hz}{Optional expected sampling rate in Hz.}

\item{tolerance_hz}{Acceptable absolute deviation from \code{expected_rate_hz}.}
}
\value{
A data frame with one row per group.
}
\description{
Audits timing or row-order information in Gazepoint Biometrics exports. The
function checks monotonicity, duplicate timestamps, nonpositive intervals,
and estimated sampling rate when the selected time column has a real time
unit. If only \code{CNT} is available, the function can still check ordering but
does not estimate a sampling rate unless \code{time_unit} is explicitly meaningful.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ibi-hr-hrv-core.R
\name{extract_gazepoint_hrv_features}
\alias{extract_gazepoint_hrv_features}
\title{Extract time-domain HRV features from Gazepoint IBI intervals}
\usage{
extract_gazepoint_hrv_features(
  data,
  ibi_col = "IBI_clean_ms",
  group_cols = NULL,
  unit = c("auto", "ms", "seconds"),
  min_intervals = 3,
  min_duration_s = 30,
  diff_threshold_ms = 50,
  collapse_repeated_intervals = TRUE,
  repeated_tolerance_ms = 1e-08
)
}
\arguments{
\item{data}{A Gazepoint biometric data frame or \code{gazepoint_ibi_filter} object.}

\item{ibi_col}{IBI/RR interval column.}

\item{group_cols}{Optional grouping columns.}

\item{unit}{Unit of the IBI column: \code{"auto"}, \code{"ms"}, or \code{"seconds"}.}

\item{min_intervals}{Minimum clean intervals required per group.}

\item{min_duration_s}{Minimum IBI-sequence duration in seconds required before
computed HRV features are treated as fully reportable. Groups below this
duration still return features but receive \code{warn_short_hrv_duration}.}

\item{diff_threshold_ms}{Threshold for NN50/pNN50.}

\item{collapse_repeated_intervals}{Logical. If \code{TRUE}, consecutive repeated
IBI values are collapsed before HRV features are computed. This is useful
for Gazepoint exports where the same IBI value may be repeated across
multiple gaze-sampling rows until a new interval is available.}

\item{repeated_tolerance_ms}{Numeric tolerance used when identifying
repeated consecutive IBI values.}
}
\value{
A list with \code{overview}, \code{features}, \code{settings}.
}
\description{
Computes simple time-domain HRV features from genuine IBI/RR intervals. This
helper does not use the Gazepoint \code{HRV} column as an HRV outcome.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/advanced-modality-bridges.R
\name{denoise_gazepoint_quantization_noise}
\alias{denoise_gazepoint_quantization_noise}
\title{Add small uniform noise to reduce quantization overlap}
\usage{
denoise_gazepoint_quantization_noise(
  dat,
  signal_cols,
  resolution,
  group_cols = NULL,
  output_suffix = "_quantization_jittered",
  seed = NULL,
  overwrite = FALSE
)
}
\arguments{
\item{dat}{A data frame.}

\item{signal_cols}{Numeric signal columns to jitter.}

\item{resolution}{Numeric scalar or named numeric vector giving measurement
resolution for each column.}

\item{group_cols}{Optional grouping columns, retained in settings.}

\item{output_suffix}{Suffix for jittered columns.}

\item{seed}{Optional random seed.}

\item{overwrite}{Logical. If \code{FALSE}, existing output columns are protected.}
}
\value{
A data frame with jittered columns and quantization-noise attributes.
}
\description{
Adds uniform white noise with magnitude tied to hardware resolution. This is
intended only for nonlinear phase-space methods that are sensitive to exact
repeated values caused by coarse interval quantization.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scr-specification-curve-plot.R
\name{plot_gazepoint_scr_specification_curve}
\alias{plot_gazepoint_scr_specification_curve}
\title{Plot an SCR specification curve}
\usage{
plot_gazepoint_scr_specification_curve(
  x,
  estimate_col = NULL,
  specification_col = "specification_id",
  add_zero_line = TRUE,
  main = "SCR specification curve"
)
}
\arguments{
\item{x}{Output from \code{run_gazepoint_scr_multiverse()} or a data frame.}

\item{estimate_col}{Column to rank and plot. Defaults to
\code{"mean_response_amplitude"} when available, otherwise \code{"response_rate"}.}

\item{specification_col}{Specification identifier column.}

\item{add_zero_line}{Logical. If \code{TRUE}, draw a horizontal zero line.}

\item{main}{Plot title.}
}
\value{
Invisibly returns a list with plot data and settings.
}
\description{
Plots a specification-curve style display from the output of
\code{run_gazepoint_scr_multiverse()} or from a compatible data frame.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/schema-helpers.R
\name{detect_gazepoint_biometric_timebase}
\alias{detect_gazepoint_biometric_timebase}
\title{Detect the likely timebase of Gazepoint biometric data}
\usage{
detect_gazepoint_biometric_timebase(data, time_col = NULL, counter_col = NULL)
}
\arguments{
\item{data}{A data frame.}

\item{time_col}{Optional explicit timing column.}

\item{counter_col}{Optional explicit counter column.}
}
\value{
A list with \code{overview}, \code{time_columns}, \code{interval_summary}, and
\code{warnings}.
}
\description{
Inspects timing and counter columns and returns a conservative summary of the
likely primary timebase. Sampling rate is estimated only when numeric timing
information is available and intervals are positive.
}
\examples{
df <- data.frame(CNT = 1:5, TIME = seq(0, by = 1 / 60, length.out = 5))
detect_gazepoint_biometric_timebase(df)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hrv-rqa-geometric.R
\name{extract_gazepoint_hrv_geometric}
\alias{extract_gazepoint_hrv_geometric}
\title{Extract geometric HRV features}
\usage{
extract_gazepoint_hrv_geometric(
  dat,
  ibi_col = "IBI",
  group_cols = NULL,
  bin_width = NULL
)
}
\arguments{
\item{dat}{A data frame.}

\item{ibi_col}{Numeric IBI/RR interval column.}

\item{group_cols}{Optional grouping columns.}

\item{bin_width}{Histogram bin width in the same units as \code{ibi_col}.}
}
\value{
A list with \code{overview}, \code{features}, and \code{settings}.
}
\description{
Computes dependency-light geometric HRV descriptors, including the HRV
triangular index and an approximate TINN-style triangular interpolation width.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/point-process-models.R
\name{model_gazepoint_hr_point_process}
\alias{model_gazepoint_hr_point_process}
\title{Model heartbeats as a dependency-light point process}
\usage{
model_gazepoint_hr_point_process(
  dat,
  ibi_col = "IBI",
  time_col = NULL,
  beat_time_col = NULL,
  group_cols = NULL,
  ibi_units = c("auto", "seconds", "milliseconds")
)
}
\arguments{
\item{dat}{A data frame.}

\item{ibi_col}{Numeric IBI/RR interval column.}

\item{time_col}{Optional time column.}

\item{beat_time_col}{Optional explicit beat-time column.}

\item{group_cols}{Optional grouping columns.}

\item{ibi_units}{\code{"auto"}, \code{"seconds"}, or \code{"milliseconds"}.}
}
\value{
A list with \code{overview}, \code{beat_table}, \code{interval_table},
\code{process_summary}, and \code{settings}.
}
\description{
Creates beat-time, interbeat interval, and inverse-Gaussian-style summary
tables from IBI/RR intervals. This is a compact point-process
model-preparation helper, not a full adaptive Bayesian heartbeat filter.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/import.R
\name{check_gazepoint_biometric_columns}
\alias{check_gazepoint_biometric_columns}
\title{Check Gazepoint biometric columns}
\usage{
check_gazepoint_biometric_columns(data)
}
\arguments{
\item{data}{A data frame imported from a Gazepoint export.}
}
\value{
A data frame describing expected columns, their signal family,
interpretation, and whether they are present.
}
\description{
Checks whether a data frame contains known Gazepoint Biometrics columns.
This function does not assume that the channels are active. It only checks
whether the expected columns are present.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/eda-analysis-pipeline-runner.R
\name{run_gazepoint_eda_analysis_pipeline}
\alias{run_gazepoint_eda_analysis_pipeline}
\title{Run a six-phase Gazepoint EDA/GSR analysis pipeline}
\usage{
run_gazepoint_eda_analysis_pipeline(
  data = NULL,
  path = NULL,
  eda_col = NULL,
  time_col = NULL,
  group_cols = NULL,
  signal_cols = NULL,
  sampling_rate = NULL,
  baseline_window = NULL,
  event_windows = NULL,
  event_data = NULL,
  lag_signal_pair = NULL,
  convert_resistance_to_us = FALSE,
  prepare_external_bridges = TRUE,
  bridge_methods = c("neurokit", "cvxeda", "ledalab", "pspm"),
  prepare_model_data = TRUE,
  create_reports = TRUE,
  output_dir = NULL,
  prefix = "gazepoint_eda_pipeline",
  continue_on_error = TRUE
)
}
\arguments{
\item{data}{Optional Gazepoint biometric data frame or imported object.}

\item{path}{Optional file or folder path. Used only when \code{data} is \code{NULL}.}

\item{eda_col}{Optional EDA/conductance column. If omitted, the runner
prefers \code{GSR_US} when available.}

\item{time_col}{Optional time/counter column.}

\item{group_cols}{Optional grouping columns.}

\item{signal_cols}{Optional biometric signal columns for activity plots and
signal audits.}

\item{sampling_rate}{Optional sampling rate in Hz, used when \code{time_col} is a
sample counter.}

\item{baseline_window}{Optional baseline window object passed to baseline
correction helpers when supported.}

\item{event_windows}{Optional event-window table used for SCR event-window
summaries when available.}

\item{event_data}{Optional event table used for TTL/event alignment when
available.}

\item{lag_signal_pair}{Optional character vector of length two giving signals
for lag estimation.}

\item{convert_resistance_to_us}{Logical. If \code{TRUE}, allow conservative
resistance-to-conductance conversion when the selected EDA column is \code{GSR}.}

\item{prepare_external_bridges}{Logical. If \code{TRUE}, prepare selected external
EDA bridge inputs.}

\item{bridge_methods}{Character vector containing any of \code{"neurokit"},
\code{"cvxeda"}, \code{"ledalab"}, and \code{"pspm"}.}

\item{prepare_model_data}{Logical. If \code{TRUE}, attempt to create SCR hurdle
and biometric LME-ready data objects.}

\item{create_reports}{Logical. If \code{TRUE}, attempt to create report outputs.}

\item{output_dir}{Optional output directory for report bundles or bridge
files where supported.}

\item{prefix}{File prefix used by output-producing helpers where supported.}

\item{continue_on_error}{Logical. If \code{TRUE}, failed steps are stored in
\code{errors} and the pipeline continues.}
}
\value{
A list with \code{overview}, \code{phases}, \code{errors}, \code{pipeline_guide},
\code{model_templates}, \code{reporting_guidance}, \code{interpretation_guardrails}, and
\code{settings}.
}
\description{
Runs a conservative six-phase Gazepoint EDA/GSR workflow using native
\code{gpbiometrics} helpers where possible. The function imports or accepts data,
audits signal quality, prepares preprocessing outputs, creates optional
external-method bridge inputs, prepares synchronization/model-formatting
outputs, attaches model templates, and generates reporting outputs.
}
\details{
The function does not fit \code{brms} or \code{lme4} models, does not run external
software, and does not infer emotion, valence, stress, trust, preference,
cognition, or diagnosis.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scr-event-windows.R
\name{summarise_gazepoint_scr_event_windows}
\alias{summarise_gazepoint_scr_event_windows}
\title{Summarise Gazepoint SCR responses in event windows}
\usage{
summarise_gazepoint_scr_event_windows(
  data = NULL,
  scr_peaks,
  events = NULL,
  time_col = NULL,
  event_time_col = NULL,
  event_id_col = NULL,
  event_label_col = NULL,
  group_cols = NULL,
  ttl_cols = NULL,
  ttl_valid_col = NULL,
  event_detection = c("rising", "active"),
  analysis_window = c(0, 6),
  response_window = c(1, 4),
  amplitude_col = "amplitude",
  peak_time_col = "peak_time",
  onset_time_col = "onset_time",
  rise_time_col = "rise_time",
  recovery_time_col = "recovery_time_after_peak",
  peak_status_col = "status",
  peak_selection = c("largest_amplitude", "first_peak"),
  collapse_simultaneous_events = FALSE
)
}
\arguments{
\item{data}{Optional Gazepoint biometric data frame. Required when \code{events}
is \code{NULL} and events should be derived from TTL columns.}

\item{scr_peaks}{A \code{gazepoint_scr_peak_detection} object returned by
\code{detect_gazepoint_scr_peaks()}, or a data frame of detected peaks.}

\item{events}{Optional event data frame. If \code{NULL}, events are derived from
TTL columns in \code{data}.}

\item{time_col}{Optional time/counter column in \code{data}. Used for TTL-derived
events.}

\item{event_time_col}{Optional event-time column in \code{events}. If \code{NULL},
common event-time column names are detected.}

\item{event_id_col}{Optional event identifier column in \code{events}.}

\item{event_label_col}{Optional event label/condition column in \code{events}.}

\item{group_cols}{Optional grouping columns used to match events and peaks.}

\item{ttl_cols}{Optional TTL marker columns used when \code{events = NULL}.}

\item{ttl_valid_col}{Optional TTL validity column. If supplied, TTL-derived
events require this column to be non-zero.}

\item{event_detection}{Event-detection rule for TTL columns. \code{"rising"}
detects rising edges; \code{"active"} treats every active TTL row as an event.}

\item{analysis_window}{Numeric length-two vector giving the event-relative
analysis window in the same units as \code{time_col} or \code{event_time_col}.}

\item{response_window}{Numeric length-two vector giving the event-relative
response window used for the binary SCR response flag.}

\item{amplitude_col}{Column in the peak table containing SCR amplitude.}

\item{peak_time_col}{Column in the peak table containing peak time.}

\item{onset_time_col}{Column in the peak table containing onset time.}

\item{rise_time_col}{Column in the peak table containing SCR rise time.}

\item{recovery_time_col}{Column in the peak table containing recovery time
after peak.}

\item{peak_status_col}{Column in the peak table containing peak status.}

\item{peak_selection}{How to choose one peak when several peaks fall in the
response window. \code{"largest_amplitude"} selects the largest response;
\code{"first_peak"} selects the earliest peak.}

\item{collapse_simultaneous_events}{Logical. If \code{TRUE}, events with the same
group and event time are collapsed into one row before matching peaks. This
is useful when Gazepoint TTL0--TTL6 channels mark the same event
simultaneously.}
}
\value{
A list with \code{overview}, \code{event_table}, \code{window_qc}, \code{events},
\code{peaks}, and \code{settings}.
}
\description{
Creates one row per event or TTL marker and links detected SCR peaks to
event-relative analysis and response windows. The helper is intended to
produce transparent event-level EDA/SCR features for downstream mixed models,
hurdle models, and reporting.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/eda-scr-plots.R
\name{plot_gazepoint_eda_decomposition}
\alias{plot_gazepoint_eda_decomposition}
\title{Plot Gazepoint EDA decomposition channels}
\usage{
plot_gazepoint_eda_decomposition(
  data,
  time_col = NULL,
  signal_cols = NULL,
  group_cols = NULL,
  standardise = FALSE,
  max_points = 5000,
  title = NULL
)
}
\arguments{
\item{data}{A Gazepoint biometric data frame or list containing a data frame.}

\item{time_col}{Optional time/counter column.}

\item{signal_cols}{Optional signal columns to plot.}

\item{group_cols}{Optional grouping columns used for facets.}

\item{standardise}{Logical. If \code{TRUE}, standardise each signal to z-scores.}

\item{max_points}{Maximum number of rows retained after simple downsampling.}

\item{title}{Optional plot title.}
}
\value{
A ggplot object with plot data stored in attributes.
}
\description{
Plots available EDA/GSR decomposition channels, typically \code{GSR_US},
\code{GSR_US_TONIC}, and \code{GSR_US_PHASIC}, as a ggplot object.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/validate.R
\name{validate_gazepoint_biometrics}
\alias{validate_gazepoint_biometrics}
\title{Validate a Gazepoint Biometrics export}
\usage{
validate_gazepoint_biometrics(data, require_active_signal = FALSE)
}
\arguments{
\item{data}{A data frame or a path to a Gazepoint CSV export.}

\item{require_active_signal}{Logical. If \code{TRUE}, validation reports a warning
when no active GSR/EDA, heart-rate, or engagement-dial channel is detected.}
}
\value{
A list with \code{overview}, \code{columns}, \code{active_channels}, and \code{issues}.
The returned object has class \code{"gazepoint_biometrics_validation"}.
}
\description{
Performs a conservative validation of a Gazepoint Biometrics table or CSV
file. The function checks whether known biometric columns are present,
whether biometric channels appear active, whether common time/synchronisation
columns are available, and whether obvious structural issues are present.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/changepoint-scr-recovery.R
\name{extract_gazepoint_scr_recovery_times}
\alias{extract_gazepoint_scr_recovery_times}
\title{Extract SCR recovery times}
\usage{
extract_gazepoint_scr_recovery_times(
  dat,
  eda_col = "GSR_US",
  time_col = "CNT",
  event_onset_col = NULL,
  group_cols = NULL,
  pre_onset_baseline_s = 2,
  peak_window_s = 5,
  recovery_window_s = 20
)
}
\arguments{
\item{dat}{A data frame.}

\item{eda_col}{Numeric EDA/conductance column.}

\item{time_col}{Numeric time column.}

\item{event_onset_col}{Optional event onset column. Finite values are treated
as event onsets.}

\item{group_cols}{Optional grouping columns.}

\item{pre_onset_baseline_s}{Baseline window before event onset.}

\item{peak_window_s}{Window after onset used to find the response peak.}

\item{recovery_window_s}{Window after peak used to find recovery.}
}
\value{
A list with \code{overview}, \code{recovery_table}, and \code{settings}.
}
\description{
Extracts 50 percent half-recovery time (\code{rec.t2}) and 63 percent recovery
time (\code{rec.tc}) for skin conductance responses from an EDA waveform and
event onsets.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/exclusions.R
\name{recommend_gazepoint_biometric_exclusions}
\alias{recommend_gazepoint_biometric_exclusions}
\title{Recommend Gazepoint biometric exclusions}
\usage{
recommend_gazepoint_biometric_exclusions(
  data,
  group_columns = NULL,
  data_is_window_summary = FALSE,
  participant_column = NULL,
  gsr_min_usable_pct = 50,
  hr_min_usable_pct = 50,
  dial_min_usable_pct = 50,
  require_gsr = TRUE,
  require_hr = TRUE,
  require_dial = FALSE
)
}
\arguments{
\item{data}{A row-level Gazepoint Biometrics data frame, a path to a
Gazepoint CSV export, or an already summarised multimodal window table.}

\item{group_columns}{Columns defining analysis windows when \code{data} is
row-level, such as \code{c("source_participant", "MEDIA_ID")}.}

\item{data_is_window_summary}{Logical. If \code{TRUE}, \code{data} is treated as an
already summarised window table.}

\item{participant_column}{Optional participant identifier column. If \code{NULL},
the function tries \code{source_participant}, \code{USER}, and \code{USERID}.}

\item{gsr_min_usable_pct}{Minimum acceptable usable percentage for GSR/EDA.}

\item{hr_min_usable_pct}{Minimum acceptable usable percentage for heart rate.}

\item{dial_min_usable_pct}{Minimum acceptable usable percentage for
engagement dial.}

\item{require_gsr}{Should low GSR/EDA coverage cause exclusion rather than
review?}

\item{require_hr}{Should low heart-rate coverage cause exclusion rather than
review?}

\item{require_dial}{Should low engagement-dial coverage cause exclusion
rather than review?}
}
\value{
A list with \code{overview}, \code{window_recommendations},
\code{participant_recommendations}, and \code{settings}.
}
\description{
Creates window-level and participant-level exclusion recommendations from
Gazepoint biometric usable-sample summaries. The function does not remove
data. It only creates transparent keep, review, or exclude recommendations.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pdr-rsa.R
\name{extract_gazepoint_pdr_signals}
\alias{extract_gazepoint_pdr_signals}
\title{Extract PPG-derived respiration proxy signals}
\usage{
extract_gazepoint_pdr_signals(
  dat,
  ppg_col = "HRP",
  time_col = "CNT",
  group_cols = NULL,
  sampling_rate = NULL,
  min_peak_distance_s = 0.3,
  smooth_window = 5,
  respiration_band = c(0.1, 0.6),
  pdr_resample_rate = 4
)
}
\arguments{
\item{dat}{A data frame containing a PPG/pulse waveform.}

\item{ppg_col}{Numeric PPG/pulse waveform column, often \code{HRP}.}

\item{time_col}{Numeric time column.}

\item{group_cols}{Optional grouping columns.}

\item{sampling_rate}{Optional sampling rate in Hz. If \code{NULL}, estimated from
\code{time_col}.}

\item{min_peak_distance_s}{Minimum plausible distance between pulse peaks.}

\item{smooth_window}{Number of samples used for simple moving-average
smoothing before peak detection.}

\item{respiration_band}{Expected respiration-frequency band in Hz.}

\item{pdr_resample_rate}{Resampling rate used for spectral estimation of PDR
proxy signals.}
}
\value{
A list with \code{overview}, \code{pulse_features}, \code{pdr_timeseries},
\code{pdr_summary}, and \code{settings}.
}
\description{
Extracts dependency-light PPG-derived respiration proxy features from a
Gazepoint pulse/PPG waveform. The function estimates respiration-modulated
pulse features such as respiratory-induced intensity variability (RIIV),
pulse amplitude variability (PAV), pulse width variability (PWV), and
pulse-rate variability (PRV).
}
\details{
These are proxy respiratory features. They should not be treated as a
replacement for a respiration belt or clinical respiratory measurement.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/eda-intervals-and-kleckner.R
\name{classify_gazepoint_scr_intervals}
\alias{classify_gazepoint_scr_intervals}
\title{Classify SCRs into FIR, SIR, and TIR latency intervals}
\usage{
classify_gazepoint_scr_intervals(
  dat,
  response_time_col = NULL,
  stimulus_onset_col = NULL,
  latency_col = NULL,
  output_col = "scr_interval",
  latency_output_col = "scr_latency_s",
  fir = c(1, 4),
  sir = c(4, 7),
  tir = c(7, 10)
)
}
\arguments{
\item{dat}{A data frame containing SCR events or peaks.}

\item{response_time_col}{Optional response/peak time column.}

\item{stimulus_onset_col}{Optional stimulus-onset column. Required when
\code{latency_col} is not supplied.}

\item{latency_col}{Optional precomputed latency column.}

\item{output_col}{Name of the output interval column.}

\item{latency_output_col}{Name of the latency output column.}

\item{fir}{Numeric vector of length two defining FIR window in seconds.}

\item{sir}{Numeric vector of length two defining SIR window in seconds.}

\item{tir}{Numeric vector of length two defining TIR window in seconds.}
}
\value{
A data frame with interval labels and latency metadata.
}
\description{
Classifies extracted SCRs into first-, second-, and third-interval response
windows based on response latency after stimulus onset.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ctsi-bridge.R
\name{prepare_gazepoint_ctsi_input}
\alias{prepare_gazepoint_ctsi_input}
\title{Prepare Gazepoint EDA data for CTSI sparse deconvolution workflows}
\usage{
prepare_gazepoint_ctsi_input(
  dat,
  eda_col = "GSR_US",
  time_col = "CNT",
  group_cols = NULL,
  event_onset_col = NULL,
  event_name_col = NULL,
  sampling_rate = NULL,
  tau0_range = c(2, 4),
  tau1_range = c(0.5, 1),
  sparsity_grid = c(0.001, 0.01, 0.1, 1),
  output_dir = NULL,
  prefix = "gazepoint_ctsi"
)
}
\arguments{
\item{dat}{A data frame containing EDA data.}

\item{eda_col}{Numeric EDA/conductance column.}

\item{time_col}{Numeric time column.}

\item{group_cols}{Optional grouping columns, such as participant/session.}

\item{event_onset_col}{Optional event onset column.}

\item{event_name_col}{Optional event/condition column.}

\item{sampling_rate}{Optional sampling rate in Hz.}

\item{tau0_range}{Candidate slow time-constant range.}

\item{tau1_range}{Candidate fast time-constant range.}

\item{sparsity_grid}{Candidate sparsity penalties.}

\item{output_dir}{Optional directory for CSV export.}

\item{prefix}{Output file prefix.}
}
\value{
A list with \code{overview}, \code{signal_table}, \code{event_table},
\code{ctsi_config}, \code{ctsi_notes}, \code{written_files}, and \code{settings}.
}
\description{
Prepares signal, event, and configuration tables for downstream
continuous-time system identification (CTSI) sparse EDA deconvolution
workflows. This function does not implement the full Amin-Faghih CTSI solver
in R. It creates reproducible input objects and optional CSV files for
external CTSI implementations.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nonlinear-complexity.R
\name{extract_gazepoint_hrv_nonlinear}
\alias{extract_gazepoint_hrv_nonlinear}
\title{Extract nonlinear HRV features from IBI/RR intervals}
\usage{
extract_gazepoint_hrv_nonlinear(
  dat,
  ibi_col = "IBI",
  group_cols = NULL,
  min_intervals = 10,
  sampen_m = 2,
  sampen_r_multiplier = 0.2,
  mse_scales = 1:5
)
}
\arguments{
\item{dat}{A data frame containing IBI/RR intervals.}

\item{ibi_col}{IBI/RR interval column.}

\item{group_cols}{Optional grouping columns.}

\item{min_intervals}{Minimum finite intervals per group.}

\item{sampen_m}{Embedding dimension for sample entropy.}

\item{sampen_r_multiplier}{Tolerance multiplier applied to the within-group SD.}

\item{mse_scales}{Integer scales used for multiscale entropy.}
}
\value{
A list with \code{overview}, \code{features}, and \code{settings}.
}
\description{
Computes dependency-light nonlinear HRV descriptors from IBI/RR intervals,
including Poincare SD1/SD2, sample entropy, approximate entropy,
multiscale entropy, and detrended fluctuation analysis.
}
\details{
These are variability and complexity descriptors. They should not be
interpreted as direct emotion, cognitive-load, health-status, or diagnostic
labels by themselves.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/online-design-optimization.R
\name{run_gazepoint_online_design_optimization}
\alias{run_gazepoint_online_design_optimization}
\title{Run blockwise online design optimization decision support}
\usage{
run_gazepoint_online_design_optimization(
  candidate_table,
  condition_col = "condition",
  utility_col = "expected_utility",
  block_col = NULL,
  cost_col = NULL,
  previous_assignments = NULL,
  exploration_weight = 0.1,
  balance_weight = 0.1,
  maximise = TRUE
)
}
\arguments{
\item{candidate_table}{A data frame containing candidate conditions.}

\item{condition_col}{Candidate condition column.}

\item{utility_col}{Expected utility/model-discrimination column.}

\item{block_col}{Optional block column.}

\item{cost_col}{Optional cost or burden column subtracted from utility.}

\item{previous_assignments}{Optional previous condition assignments.}

\item{exploration_weight}{Weight for favouring under-sampled conditions.}

\item{balance_weight}{Weight for penalising over-sampled conditions.}

\item{maximise}{Logical. If \code{TRUE}, select highest score.}
}
\value{
A list with \code{overview}, \code{ranked_candidates}, \code{recommendation},
\code{assignment_summary}, and \code{settings}.
}
\description{
Provides a safe, dependency-light decision-support/simulation helper for
online design optimization. The function recommends the next condition by
combining expected model-discrimination utility with optional exploration and
balancing penalties.
}
\details{
This function does not control stimulus presentation software and should not
be used as autonomous real-time experiment control without separate ethical,
preregistration, and software-integration review.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sync-drift-diagnostics.R
\name{audit_gazepoint_biometric_sync_drift}
\alias{audit_gazepoint_biometric_sync_drift}
\title{Audit Gazepoint biometric synchronization drift}
\usage{
audit_gazepoint_biometric_sync_drift(
  data,
  time_col = NULL,
  group_cols = NULL,
  signal_pairs = NULL,
  signal_cols = NULL,
  reference_signal_col = NULL,
  max_lag = 1000,
  lag_step = NULL,
  drift_tolerance = NULL,
  method = c("pearson", "spearman"),
  min_complete_pairs = 20,
  use_first_difference = FALSE,
  include_reset_segments = TRUE
)
}
\arguments{
\item{data}{A Gazepoint biometric data frame.}

\item{time_col}{Optional time or counter column.}

\item{group_cols}{Optional grouping columns.}

\item{signal_pairs}{Optional two-column data frame, matrix, or list defining
signal pairs. If \code{NULL}, pairs are formed between a reference signal and
other detected biometric signals.}

\item{signal_cols}{Optional candidate signal columns used when \code{signal_pairs}
is \code{NULL}.}

\item{reference_signal_col}{Optional reference signal used when \code{signal_pairs}
is \code{NULL}.}

\item{max_lag}{Maximum absolute lag to evaluate, in the same units as
\code{time_col}.}

\item{lag_step}{Step size between candidate lags. If \code{NULL}, the median
positive time step is used.}

\item{drift_tolerance}{Optional threshold for the range of estimated lags
across groups. If \code{NULL}, drift is summarized but not threshold-classified.}

\item{method}{Correlation method passed to \code{\link[stats:cor]{stats::cor()}}.}

\item{min_complete_pairs}{Minimum complete aligned observations required for
each candidate lag.}

\item{use_first_difference}{If \code{TRUE}, lag diagnostics use first differences.}

\item{include_reset_segments}{If \code{TRUE}, reset segments from
\code{\link[=audit_gazepoint_time_resets]{audit_gazepoint_time_resets()}} are added to grouping when available.}
}
\value{
A list with \code{overview}, \code{checks}, \code{time_reset_audit},
\code{lag_by_group}, \code{lag_profile}, \code{drift_summary}, and \code{settings}.
}
\description{
Combines time-order/reset diagnostics with conservative signal-lag summaries
across signal pairs and groups. The helper is intended for quality control and
synchronization review. It does not infer emotional valence, cognitive states,
causal timing, or true physiological latency.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/classical-standardization.R
\name{standardise_gazepoint_zscore}
\alias{standardise_gazepoint_zscore}
\alias{standardize_gazepoint_zscore}
\title{Standardise SCR or SCL using intra-individual z-scoring}
\usage{
standardise_gazepoint_zscore(
  dat,
  signal_col = "SCR_Amplitude",
  group_col = "source_participant",
  suffix = "_Z",
  min_valid = 2,
  overwrite = FALSE
)

standardize_gazepoint_zscore(
  dat,
  signal_col = "SCR_Amplitude",
  group_col = "source_participant",
  suffix = "_Z",
  min_valid = 2,
  overwrite = FALSE
)
}
\arguments{
\item{dat}{A data frame containing SCR, SCL, or another biometric signal.}

\item{signal_col}{Signal column to standardise.}

\item{group_col}{Participant or unit column.}

\item{suffix}{Suffix for the output column.}

\item{min_valid}{Minimum finite observations required within each group.}

\item{overwrite}{Logical. If \code{FALSE}, existing output columns are protected.}
}
\value{
A data frame with an added z-scored column.
}
\description{
Adds a within-participant z-scored version of a signal column. This is a
lightweight compatibility wrapper around the package's more general
within-unit standardisation helper.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/quality.R
\name{audit_gazepoint_gsr_quality}
\alias{audit_gazepoint_gsr_quality}
\title{Audit Gazepoint GSR/EDA signal quality}
\usage{
audit_gazepoint_gsr_quality(
  data,
  value_column = NULL,
  validity_column = "GSRV",
  min_value = 0,
  max_value = 100,
  jump_threshold = NULL
)
}
\arguments{
\item{data}{A data frame or a path to a Gazepoint CSV export.}

\item{value_column}{Optional GSR/EDA value column. If \code{NULL}, \code{GSR_US} is
used when present, otherwise \code{GSR}.}

\item{validity_column}{Optional validity column. Defaults to \code{"GSRV"}.}

\item{min_value}{Minimum plausible value.}

\item{max_value}{Maximum plausible value.}

\item{jump_threshold}{Optional threshold for detecting large sample-to-sample
jumps.}
}
\value{
A one-row data frame summarising signal quality.
}
\description{
Audits Gazepoint GSR/EDA columns for missingness, inactive zero rows,
validity flags, plausible value ranges, flatlining, and usable sample
coverage. When available, \code{GSR_US} is used by default because it represents
skin conductance in microsiemens in Gazepoint exports.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gpbiometrics-package.R
\docType{package}
\name{gpbiometrics-package}
\alias{gpbiometrics-package}
\alias{gpbiometrics}
\title{gpbiometrics: Process, Audit, Plot, and Report Gazepoint Biometrics Data}
\description{
\code{gpbiometrics} provides helper functions for importing, checking,
preprocessing, summarising, diagnosing, plotting, and reporting Gazepoint
Biometrics exports. It is designed for reproducible offline workflows
involving physiological, response-channel, event-marker, and timing data
exported alongside Gazepoint eye-tracking recordings.
}
\details{
The package supports common Gazepoint biometric channels, including GSR/EDA,
heart rate, interbeat interval or RR-style intervals, engagement dial, and
TTL markers. It provides functions for folder-level import, data-summary
import, schema and timebase detection, active-channel detection, column
validation, missingness audits, signal-quality audits, sampling and timing
audits, dropout and flatline flagging, baseline correction, smoothing, window
summaries, TTL event extraction, biometric-to-gaze synchronisation, exclusion
recommendations, workflow diagnostics, report-ready tables, cautious methods
text, lightweight report generation, and base-R diagnostic plotting.

Schema and timing helpers include
\code{standardise_gazepoint_biometric_names()},
\code{detect_gazepoint_biometric_schema()},
\code{detect_gazepoint_time_columns()}, and
\code{detect_gazepoint_biometric_timebase()}. Signal-availability and quality
helpers include \code{summarise_gazepoint_biometric_validity()},
\code{flag_gazepoint_biometric_dropouts()}, \code{audit_gazepoint_gsr_quality()},
\code{audit_gazepoint_hr_quality()}, \code{audit_gazepoint_engagement_dial()}, and
\code{audit_gazepoint_ibi_quality()}.

Preprocessing and summary helpers include
\code{baseline_correct_gazepoint_gsr()},
\code{baseline_correct_gazepoint_hr()}, \code{smooth_gazepoint_biometrics()},
\code{convert_gazepoint_gsr_to_conductance()},
\code{summarise_gazepoint_gsr_windows()},
\code{summarise_gazepoint_hr_windows()},
\code{summarise_gazepoint_engagement_windows()},
\code{summarise_gazepoint_dial_windows()},
\code{summarise_gazepoint_ibi_windows()},
\code{summarise_gazepoint_gsr_tonic_phasic()},
\code{summarise_gazepoint_multimodal_windows()},
\code{summarise_gazepoint_ibi_hrv_windows()}, and
\code{summarise_gazepoint_full_biometric_windows()}.

Synchronisation, modelling, workflow, and reporting helpers include
\code{sync_gazepoint_biometrics_with_gaze()},
\code{join_gazepoint_biometrics_to_master()},
\code{join_gazepoint_biometrics_to_gp3tools()},
\code{prepare_gazepoint_multimodal_model_data()},
\code{run_gazepoint_biometrics_workflow()},
\code{summarise_gazepoint_biometrics_workflow()},
\code{diagnose_gazepoint_biometrics_workflow()},
\code{create_gazepoint_biometrics_checklist()},
\code{create_gazepoint_biometrics_methods_text()},
\code{create_gazepoint_biometrics_report_tables()},
\code{write_gazepoint_biometrics_report_tables()}, and
\code{create_gazepoint_biometrics_report()}. Diagnostic plotting helpers include
\code{plot_gazepoint_biometric_signals()} and
\code{plot_gazepoint_biometric_quality()}.

The package treats biometric signals conservatively. GSR/EDA is handled as an
electrodermal activity or arousal-related signal rather than emotional
valence. Heart-rate summaries should be interpreted relative to baseline,
artefact handling, and task context. Raw Gazepoint \code{HRV} columns should be
treated as validity or vendor flags unless independent documentation proves
otherwise. IBI-derived HRV-style summaries should be computed only from
genuine interbeat interval or RR-style interval columns. Eye-tracking
measures, when combined with biometric data, should be interpreted as
indicators of visual attention rather than direct evidence of cognition,
scrutiny, or evaluation.
}
\section{Main workflow}{


The main workflow function is:

\code{run_gazepoint_biometrics_workflow()}

Useful follow-up helpers include:

\code{summarise_gazepoint_biometrics_workflow()},
\code{diagnose_gazepoint_biometrics_workflow()},
\code{create_gazepoint_biometrics_report_tables()},
\code{write_gazepoint_biometrics_report_tables()}, and
\code{create_gazepoint_biometrics_report()}.
}

\section{Interpretation caution}{


\code{gpbiometrics} is a preprocessing, quality-control, visualisation, and
reporting toolkit. It does not classify emotions, mental states, or cognitive
processes directly. Researchers should interpret biometric signals in
relation to experimental design, baseline periods, stimulus timing, task
demands, artefact handling, and complementary behavioural or self-report
measures.
}

\author{
\strong{Maintainer}: Stefanos Balaskas \email{s.balaskas@ac.upatras.gr}

Authors:
\itemize{
  \item Stefanos Balaskas \email{s.balaskas@ac.upatras.gr}
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/eda-tvsymp.R
\name{extract_gazepoint_eda_tvsymp}
\alias{extract_gazepoint_eda_tvsymp}
\title{Extract time-varying spectral EDA features}
\usage{
extract_gazepoint_eda_tvsymp(
  dat,
  eda_col = "GSR_US",
  time_col = "CNT",
  group_cols = NULL,
  sampling_rate = NULL,
  band = c(0.08, 0.24),
  window_seconds = 60,
  step_seconds = 5,
  min_valid_fraction = 0.7,
  normalise = TRUE
)
}
\arguments{
\item{dat}{A data frame containing EDA data.}

\item{eda_col}{Numeric EDA/conductance column.}

\item{time_col}{Numeric time column.}

\item{group_cols}{Optional grouping columns.}

\item{sampling_rate}{Optional sampling rate in Hz. If \code{NULL}, estimated from
\code{time_col}.}

\item{band}{Frequency band in Hz used for TVSymp-style power.}

\item{window_seconds}{Sliding-window length in seconds.}

\item{step_seconds}{Sliding-window step in seconds.}

\item{min_valid_fraction}{Minimum valid fraction per window.}

\item{normalise}{Logical. If \code{TRUE}, compute EDASympn-style relative band
power normalised by total positive-frequency power.}
}
\value{
A list with \code{overview}, \code{tvsymp_timeseries}, \code{summary}, and \code{settings}.
}
\description{
Computes a dependency-light approximation of TVSymp-style time-varying
spectral EDA power using sliding-window spectral analysis. The default band
is 0.08--0.24 Hz, following the TVSymp literature. This function does not
claim exact VFCDM reproduction.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scr-multiverse.R
\name{run_gazepoint_scr_multiverse}
\alias{run_gazepoint_scr_multiverse}
\title{Run a multiverse of SCR scoring specifications}
\usage{
run_gazepoint_scr_multiverse(
  dat,
  signal_col = "GSR_US",
  time_col = "time",
  trial_cols = NULL,
  condition_col = NULL,
  participant_col = NULL,
  event_time_col = NULL,
  latency_windows = list(c(1, 3), c(1, 4), c(1, 5)),
  thresholds = c(0.01, 0.05),
  baseline_methods = c("median", "mean"),
  baseline_window = c(-1, 0),
  response_metrics = c("max_minus_baseline"),
  model_function = NULL
)
}
\arguments{
\item{dat}{Sample-level EDA data.}

\item{signal_col}{Conductance/EDA signal column.}

\item{time_col}{Time column, preferably relative to stimulus onset. If
\code{event_time_col} is supplied, relative time is computed as
\code{time_col - event_time_col}.}

\item{trial_cols}{Columns identifying trials.}

\item{condition_col}{Optional experimental condition column.}

\item{participant_col}{Optional participant column.}

\item{event_time_col}{Optional event/stimulus onset time column.}

\item{latency_windows}{List of response windows in seconds.}

\item{thresholds}{SCR response thresholds.}

\item{baseline_methods}{Baseline methods: \code{"median"}, \code{"mean"}, or \code{"none"}.}

\item{baseline_window}{Baseline window in relative seconds.}

\item{response_metrics}{Response metrics: \code{"max_minus_baseline"} or
\code{"peak_to_peak"}.}

\item{model_function}{Optional function applied to each specification-level
trial summary.}
}
\value{
A list with specification grid, scored trials, optional model results,
and robustness overview.
}
\description{
Scores SCR amplitudes across multiple latency windows, thresholds, baseline
methods, and response metrics. Optionally applies a user-supplied model
function to each specification. This supports transparent sensitivity
analysis and specification-curve style reporting.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/within-unit-standardization.R
\name{standardize_gazepoint_biometrics_within_unit}
\alias{standardize_gazepoint_biometrics_within_unit}
\alias{standardise_gazepoint_biometrics_within_unit}
\title{Standardize biometric signals within participant or other analysis units}
\usage{
standardize_gazepoint_biometrics_within_unit(
  data,
  signal_cols = NULL,
  unit_cols = NULL,
  reference_col = NULL,
  reference_value = TRUE,
  suffix = "_z_within",
  center = TRUE,
  scale = TRUE,
  min_valid = 2,
  zero_sd_action = c("NA", "zero"),
  overwrite = FALSE
)

standardise_gazepoint_biometrics_within_unit(
  data,
  signal_cols = NULL,
  unit_cols = NULL,
  reference_col = NULL,
  reference_value = TRUE,
  suffix = "_z_within",
  center = TRUE,
  scale = TRUE,
  min_valid = 2,
  zero_sd_action = c("NA", "zero"),
  overwrite = FALSE
)
}
\arguments{
\item{data}{A data frame containing Gazepoint biometric data.}

\item{signal_cols}{Character vector of biometric signal columns to
standardize. If \code{NULL}, common numeric biometric columns are detected.}

\item{unit_cols}{Character vector defining the unit within which means and
standard deviations are computed. If \code{NULL}, common participant/session
columns are detected. If no columns are detected, the whole data frame is
treated as one unit.}

\item{reference_col}{Optional logical or categorical column identifying rows
used to estimate the reference mean and standard deviation. For example,
this can be a baseline-window flag. The resulting parameters are then
applied to all rows in the same unit.}

\item{reference_value}{Value in \code{reference_col} that marks reference rows.
Defaults to \code{TRUE}.}

\item{suffix}{Suffix for standardized output columns.}

\item{center}{Logical. If \code{TRUE}, subtract the within-unit reference mean.}

\item{scale}{Logical. If \code{TRUE}, divide by the within-unit reference
standard deviation.}

\item{min_valid}{Minimum number of finite reference observations required
per unit and signal.}

\item{zero_sd_action}{What to do when the within-unit standard deviation is
zero or unavailable. \code{"NA"} returns \code{NA}; \code{"zero"} returns zero for finite
centered values.}

\item{overwrite}{Logical. If \code{FALSE}, existing output columns are protected.}
}
\value{
A data frame with added standardized columns. Attributes include
\code{standardization_summary}, \code{standardization_parameters}, and \code{settings}.
}
\description{
Adds within-unit standardized biometric columns, usually within participant
or participant-by-session/stimulus groups. This is useful when the analysis
focuses on relative within-person signal change rather than absolute
between-person level differences.
}
\details{
The helper is intentionally conservative. It does not run automatically in
the main workflow and does not infer emotion, valence, stress, trust,
preference, cognition, or diagnosis. Within-unit z-scoring removes
between-unit level and scale differences and should therefore be reported
explicitly.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/windows.R
\name{summarise_gazepoint_engagement_windows}
\alias{summarise_gazepoint_engagement_windows}
\title{Summarise Gazepoint engagement-dial windows}
\usage{
summarise_gazepoint_engagement_windows(
  data,
  group_columns = NULL,
  value_column = "DIAL",
  validity_column = "DIALV",
  exclude_zero = FALSE
)
}
\arguments{
\item{data}{A data frame or a path to a Gazepoint CSV export.}

\item{group_columns}{Optional grouping columns defining windows, such as
\code{c("USER", "MEDIA_ID")}.}

\item{value_column}{Engagement-dial value column. Defaults to \code{"DIAL"}.}

\item{validity_column}{Engagement-dial validity column. Defaults to
\code{"DIALV"}.}

\item{exclude_zero}{Should zero values be excluded from usable summaries?}
}
\value{
A data frame with one row per window.
}
\description{
Summarises Gazepoint engagement-dial values within participant, trial,
stimulus, AOI, or other user-defined windows.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/environment-confounds-and-ipfm.R
\name{audit_gazepoint_stabilization_period}
\alias{audit_gazepoint_stabilization_period}
\title{Audit or trim the EDA electrode stabilization period}
\usage{
audit_gazepoint_stabilization_period(
  dat,
  time_col = "CNT",
  group_cols = NULL,
  stabilization_minutes = 10,
  action = c("flag", "trim"),
  output_col = "in_stabilization_period",
  time_units = c("auto", "seconds", "milliseconds")
)
}
\arguments{
\item{dat}{A data frame.}

\item{time_col}{Numeric time column.}

\item{group_cols}{Optional grouping columns.}

\item{stabilization_minutes}{Stabilization duration to flag or trim.}

\item{action}{\code{"flag"} or \code{"trim"}.}

\item{output_col}{Output logical flag column.}

\item{time_units}{\code{"auto"}, \code{"seconds"}, or \code{"milliseconds"}.}
}
\value{
A data frame with stabilization-period attributes.
}
\description{
Flags or removes the initial stabilization period in each recording/group.
This is intended to prevent early skin-electrode drift from being treated as
a stable physiological baseline.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/environment-confounds-and-ipfm.R
\name{correct_gazepoint_eda_temperature}
\alias{correct_gazepoint_eda_temperature}
\title{Correct EDA for ambient or body temperature}
\usage{
correct_gazepoint_eda_temperature(
  dat,
  eda_col = "GSR_US",
  temperature_cols,
  group_cols = NULL,
  time_col = NULL,
  output_col = "eda_temperature_adjusted",
  fitted_col = "eda_temperature_fitted",
  model_by_group = TRUE,
  add_intercept_mean = TRUE
)
}
\arguments{
\item{dat}{A data frame.}

\item{eda_col}{Numeric EDA/conductance column.}

\item{temperature_cols}{One or more numeric temperature columns.}

\item{group_cols}{Optional grouping columns.}

\item{time_col}{Optional time column retained in summaries.}

\item{output_col}{Output residual-adjusted EDA column.}

\item{fitted_col}{Output fitted temperature component column.}

\item{model_by_group}{Logical. If \code{TRUE}, fit one model per group.}

\item{add_intercept_mean}{Logical. If \code{TRUE}, add the group mean EDA back to
residuals so the adjusted signal remains on the original scale.}
}
\value{
A data frame with adjusted EDA columns and model-summary attributes.
}
\description{
Regresses an EDA/conductance signal on one or more continuous temperature
covariates and returns a temperature-adjusted residual series. The adjusted
signal is temperature-adjusted EDA, not "pure" cognitive or emotional EDA.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gsr-eda-helpers.R
\name{summarise_gazepoint_gsr_tonic_phasic}
\alias{summarise_gazepoint_gsr_tonic_phasic}
\title{Summarise tonic and phasic GSR/EDA components}
\usage{
summarise_gazepoint_gsr_tonic_phasic(
  data,
  gsr_col = NULL,
  group_cols = NULL,
  time_col = NULL,
  window_n = 15L,
  peak_threshold = NULL,
  output_prefix = "gsr"
)
}
\arguments{
\item{data}{A data frame.}

\item{gsr_col}{Optional GSR/EDA column. If \code{NULL}, the function prefers
\code{GSR_US}, then \code{GSR}, then other recognised GSR/EDA columns.}

\item{group_cols}{Optional grouping columns. Tonic/phasic values are computed
separately within each group.}

\item{time_col}{Optional time column used to order rows within groups.}

\item{window_n}{Rolling-median window size in samples.}

\item{peak_threshold}{Optional phasic peak threshold. If \code{NULL}, a robust
data-driven threshold is computed within each group as
\code{median(phasic) + 2 * MAD(phasic)}.}

\item{output_prefix}{Prefix for generated columns.}
}
\value{
A list with \code{data}, \code{summary}, and \code{settings}.
}
\description{
Creates a simple descriptive tonic/phasic decomposition of a GSR/EDA signal.
The tonic component is estimated with a rolling median, and the phasic
component is the observed signal minus the rolling-median tonic estimate.
}
\details{
This is a lightweight descriptive helper, not a full skin-conductance-response
deconvolution model. It should be used for quality checks, window summaries,
and exploratory reporting unless a study requires a specialised EDA model.
}
\examples{
df <- data.frame(
  CNT = 1:10,
  GSR_US = c(1, 1.1, 1.0, 1.2, 2.0, 1.3, 1.2, 1.1, 1.0, 1.1)
)
summarise_gazepoint_gsr_tonic_phasic(df, window_n = 3)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/schema-helpers.R
\name{standardise_gazepoint_biometric_names}
\alias{standardise_gazepoint_biometric_names}
\title{Standardise Gazepoint biometric column names}
\usage{
standardise_gazepoint_biometric_names(
  data,
  style = c("canonical", "snake"),
  rename = TRUE
)
}
\arguments{
\item{data}{A data frame or a character vector of column names.}

\item{style}{Naming style to return. \code{"canonical"} returns uppercase
Gazepoint-style names for recognised columns. \code{"snake"} returns lowercase
snake-case names.}

\item{rename}{Logical. If \code{data} is a data frame, should the returned data
frame have standardised names? If \code{FALSE}, a name-mapping table is returned.}
}
\value{
If \code{data} is a character vector, a character vector of standardised
names. If \code{data} is a data frame and \code{rename = TRUE}, the data frame with
standardised names. If \code{rename = FALSE}, a data frame mapping original names
to standardised names.
}
\description{
Standardises common Gazepoint Biometrics column-name variants to stable
canonical names. The helper is intentionally conservative: it recognises
common biometric, timing, marker, participant, and stimulus columns, but it
leaves unknown columns unchanged apart from optional snake-case cleaning.
}
\examples{
standardise_gazepoint_biometric_names(c("time ms", "heart rate", "eda uS"))

df <- data.frame(`time ms` = 1:3, `heart rate` = c(70, 72, 71))
names(standardise_gazepoint_biometric_names(df))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/report-bundle-readiness.R
\name{export_gazepoint_biometrics_report_bundle}
\alias{export_gazepoint_biometrics_report_bundle}
\title{Export a Gazepoint biometrics report bundle}
\usage{
export_gazepoint_biometrics_report_bundle(
  bundle = NULL,
  output_dir,
  prefix = "gpbiometrics_report",
  tables = NULL,
  text = NULL,
  plots = NULL,
  include_readme = TRUE,
  include_session_info = TRUE,
  overwrite = FALSE
)
}
\arguments{
\item{bundle}{Optional list-like object containing data frames, text, or plots.}

\item{output_dir}{Output directory.}

\item{prefix}{File prefix.}

\item{tables}{Optional named list of data frames to export as CSV files.}

\item{text}{Optional named list or character vector of text outputs to export
as TXT files.}

\item{plots}{Optional named list of ggplot objects to export as PNG files.}

\item{include_readme}{Logical. Should a README text file be written?}

\item{include_session_info}{Logical. Should session information be written?}

\item{overwrite}{Logical. Should existing files be overwritten?}
}
\value{
A list with \code{overview}, \code{manifest}, \code{output_dir}, and \code{settings}.
}
\description{
Exports selected report tables, text outputs, optional plot objects, and a
manifest to a local output directory. This helper is intended for reproducible
reporting. It does not commit files and should not be used to export private
real-data outputs into a package repository.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hrv-features.R
\name{summarise_gazepoint_hrv_features}
\alias{summarise_gazepoint_hrv_features}
\title{Summarise time-domain HRV features from Gazepoint IBI/RR intervals}
\usage{
summarise_gazepoint_hrv_features(
  data,
  ibi_col = NULL,
  group_cols = NULL,
  time_col = NULL,
  ibi_unit = c("auto", "seconds", "milliseconds"),
  min_ibi_ms = 300,
  max_ibi_ms = 2000,
  min_valid_ibi = 3L
)
}
\arguments{
\item{data}{A data frame.}

\item{ibi_col}{Optional IBI/RR interval column. If \code{NULL}, a likely IBI/RR
column is detected. The raw \code{HRV} column is never selected automatically.}

\item{group_cols}{Optional grouping columns, such as participant, stimulus,
trial, or window.}

\item{time_col}{Optional time/order column used to order IBI values within
each group before calculating successive-difference features.}

\item{ibi_unit}{Unit of the IBI/RR column. Use \code{"auto"}, \code{"seconds"}, or
\code{"milliseconds"}.}

\item{min_ibi_ms}{Minimum plausible IBI in milliseconds.}

\item{max_ibi_ms}{Maximum plausible IBI in milliseconds.}

\item{min_valid_ibi}{Minimum number of valid IBI values required before a
group is marked as having computed HRV features.}
}
\value{
A list with \code{overview}, \code{features}, and \code{settings}.
}
\description{
Computes conservative time-domain HRV-style features from genuine interbeat
interval or RR interval columns. The raw Gazepoint \code{HRV} column is not used
as an HRV metric because it is treated as a validity/vendor flag unless
independently documented otherwise.
}
\details{
The helper computes descriptive features including mean IBI, mean heart rate
derived from IBI, SDNN, RMSSD, and pNN50. It does not compute frequency-domain
HRV and does not replace specialised ECG/PPG HRV software.
}
\examples{
df <- data.frame(
  participant = "P1",
  IBI = c(0.9, 1.0, 1.1, 1.0, 0.95)
)
summarise_gazepoint_hrv_features(df, group_cols = "participant")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reporting.R
\name{create_gazepoint_biometrics_methods_text}
\alias{create_gazepoint_biometrics_methods_text}
\title{Create Gazepoint Biometrics methods text}
\usage{
create_gazepoint_biometrics_methods_text(
  checklist = NULL,
  data = NULL,
  include_cautions = TRUE
)
}
\arguments{
\item{checklist}{A checklist produced by
\code{create_gazepoint_biometrics_checklist()}. If \code{NULL}, \code{data} must be
supplied.}

\item{data}{Optional data frame or path to a Gazepoint CSV export used to
create the checklist when \code{checklist = NULL}.}

\item{include_cautions}{Logical. Should interpretation cautions be appended?}
}
\value{
A character string containing draft methods text.
}
\description{
Creates a compact draft methods paragraph describing Gazepoint Biometrics
data processing. The text is intentionally cautious and avoids making
emotional or cognitive claims from physiological or eye-tracking measures
alone.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compatibility-aliases.R
\name{prepare_gazepoint_rhrv_input}
\alias{prepare_gazepoint_rhrv_input}
\title{Prepare Gazepoint IBI/RR data for RHRV}
\usage{
prepare_gazepoint_rhrv_input(
  data,
  ibi_col = "IBI_clean_ms",
  group_cols = NULL,
  unit = c("auto", "ms", "seconds"),
  collapse_repeated_intervals = TRUE,
  repeated_tolerance_ms = 1e-08,
  min_ibi_ms = 300,
  max_ibi_ms = 2000,
  output_dir = NULL,
  prefix = "gazepoint_rhrv"
)
}
\arguments{
\item{data}{A Gazepoint biometric data frame or \code{gazepoint_ibi_filter} object.}

\item{ibi_col}{IBI/RR interval column.}

\item{group_cols}{Optional grouping columns.}

\item{unit}{Unit of the IBI column: \code{"auto"}, \code{"ms"}, or \code{"seconds"}.}

\item{collapse_repeated_intervals}{Logical. If \code{TRUE}, consecutive repeated
IBI values are collapsed before export.}

\item{repeated_tolerance_ms}{Numeric tolerance used when identifying repeated
consecutive IBI values.}

\item{min_ibi_ms}{Minimum plausible IBI in milliseconds retained for export.}

\item{max_ibi_ms}{Maximum plausible IBI in milliseconds retained for export.}

\item{output_dir}{Optional directory where per-group CSV files are written.}

\item{prefix}{File prefix used when \code{output_dir} is supplied.}
}
\value{
A list returned by \code{\link[=export_gazepoint_rhrv_input]{export_gazepoint_rhrv_input()}}.
}
\description{
Backward-compatible alias for \code{\link[=export_gazepoint_rhrv_input]{export_gazepoint_rhrv_input()}}. This helper
prepares analysis-ready inter-beat interval input for optional RHRV workflows.
It does not call RHRV and does not derive HRV from Gazepoint's raw \code{HRV}
field. HRV features should be derived from genuine IBI/RR intervals.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/import.R
\name{import_gazepoint_biometrics}
\alias{import_gazepoint_biometrics}
\title{Import a Gazepoint Biometrics export}
\usage{
import_gazepoint_biometrics(file, na = c("", "NA", "NaN"))
}
\arguments{
\item{file}{Path to a Gazepoint CSV export.}

\item{na}{Values that should be treated as missing.}
}
\value{
A data frame with Gazepoint export columns preserved. The returned
object has class \code{"gazepoint_biometrics"} and an attribute named
\code{"biometric_columns"}.
}
\description{
Reads a Gazepoint CSV export containing biometric columns such as GSR,
heart rate, interbeat interval, pulse signal, engagement dial, and TTL
synchronization fields. The function is conservative: it preserves original
column names, removes only empty trailing columns, and attaches a basic
biometric-column summary as an attribute.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sync-drift-diagnostics.R
\name{estimate_gazepoint_signal_lag}
\alias{estimate_gazepoint_signal_lag}
\title{Estimate lag between two Gazepoint biometric signals}
\usage{
estimate_gazepoint_signal_lag(
  data,
  signal_x_col,
  signal_y_col,
  time_col = NULL,
  group_cols = NULL,
  max_lag = 1000,
  lag_step = NULL,
  method = c("pearson", "spearman"),
  min_complete_pairs = 20,
  use_first_difference = FALSE
)
}
\arguments{
\item{data}{A Gazepoint biometric data frame.}

\item{signal_x_col}{Name of the first signal column.}

\item{signal_y_col}{Name of the second signal column.}

\item{time_col}{Optional time or counter column. If \code{NULL}, a common
Gazepoint time/counter column is detected.}

\item{group_cols}{Optional grouping columns, such as participant, stimulus,
trial, or source file.}

\item{max_lag}{Maximum absolute lag to evaluate, in the same units as
\code{time_col}.}

\item{lag_step}{Step size between candidate lags, in the same units as
\code{time_col}. If \code{NULL}, the median positive time step is used.}

\item{method}{Correlation method passed to \code{\link[stats:cor]{stats::cor()}}.}

\item{min_complete_pairs}{Minimum complete aligned observations required for
a candidate lag.}

\item{use_first_difference}{If \code{TRUE}, correlations are estimated on first
differences rather than raw signal levels.}
}
\value{
A list with \code{overview}, \code{lag_by_group}, \code{lag_profile}, and \code{settings}.
}
\description{
Estimates the time shift that maximizes the association between two recorded
biometric signals within each group. This is a conservative synchronization
diagnostic for inspecting whether two recorded traces show similar temporal
structure at different shifts. It should not be interpreted as causal timing
or true physiological latency unless the design includes appropriate event
markers and independently justified signal-processing assumptions.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/report_table_export.R
\name{write_gazepoint_biometrics_report_tables}
\alias{write_gazepoint_biometrics_report_tables}
\title{Write Gazepoint Biometrics report tables}
\usage{
write_gazepoint_biometrics_report_tables(
  tables,
  output_dir,
  prefix = "gazepoint_biometrics",
  overwrite = TRUE,
  include_empty_message_tables = FALSE
)
}
\arguments{
\item{tables}{A Gazepoint Biometrics workflow object, report-table object, or
named list of data frames.}

\item{output_dir}{Output directory for CSV files.}

\item{prefix}{Filename prefix.}

\item{overwrite}{Should existing files be overwritten?}

\item{include_empty_message_tables}{Should placeholder tables containing only
a \code{message} column be written?}
}
\value{
A data frame indexing written and skipped files.
}
\description{
Writes report-ready Gazepoint Biometrics tables to CSV files. The input can
be a workflow object produced by \code{run_gazepoint_biometrics_workflow()}, a
report-table object produced by \code{create_gazepoint_biometrics_report_tables()},
or a named list of data frames.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot-contracts.R
\name{check_gazepoint_plot_contract}
\alias{check_gazepoint_plot_contract}
\title{Check a Gazepoint plot return contract}
\usage{
check_gazepoint_plot_contract(
  plot,
  require_plot_data = TRUE,
  require_settings = TRUE
)
}
\arguments{
\item{plot}{A plot object.}

\item{require_plot_data}{Logical. If \code{TRUE}, \code{plot_data} must be present.}

\item{require_settings}{Logical. If \code{TRUE}, \code{settings} must be present.}
}
\value{
A list with \code{overview}, \code{checks}, \code{plot_data}, and \code{settings}.
}
\description{
Checks whether an object follows the package's plotting return convention.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sync.R
\name{sync_gazepoint_biometrics_with_gaze}
\alias{sync_gazepoint_biometrics_with_gaze}
\title{Synchronise Gazepoint Biometrics with gaze data}
\usage{
sync_gazepoint_biometrics_with_gaze(
  biometrics,
  gaze,
  by,
  all_x = TRUE,
  suffixes = c(".gaze", ".bio")
)
}
\arguments{
\item{biometrics}{A Gazepoint Biometrics data frame.}

\item{gaze}{A gaze, fixation, AOI, or \code{gp3tools}-style summary data frame.}

\item{by}{Character vector of key columns used for joining.}

\item{all_x}{Logical. Should all rows from \code{gaze} be retained?}

\item{suffixes}{Character vector of length two used for duplicate non-key
column names.}
}
\value{
A data frame with gaze rows joined to biometric columns. The returned
object has class \code{"gazepoint_biometrics_sync"} and a \code{"sync_summary"}
attribute.
}
\description{
Joins Gazepoint Biometrics data to gaze or fixation data using exact key
columns. This function is intentionally conservative: it does not perform
interpolation, nearest-neighbour matching, or automatic time shifting. For a
first reproducible workflow, exact joins by participant, media, trial, sample
counter, or fixation identifier are preferred when those fields are available.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ttl.R
\name{extract_gazepoint_ttl_events}
\alias{extract_gazepoint_ttl_events}
\title{Extract Gazepoint TTL marker events}
\usage{
extract_gazepoint_ttl_events(
  data,
  ttl_columns = NULL,
  group_columns = NULL,
  validity_column = "TTLV",
  require_validity = TRUE,
  mode = c("changes", "nonzero"),
  include_initial = TRUE
)
}
\arguments{
\item{data}{A data frame or a path to a Gazepoint CSV export.}

\item{ttl_columns}{TTL marker columns. If \code{NULL}, the function uses all
available columns from \code{TTL0} to \code{TTL6}.}

\item{group_columns}{Optional grouping columns within which TTL changes are
detected, such as \code{source_participant}, \code{USER}, or \code{MEDIA_ID}.}

\item{validity_column}{Optional TTL validity column. Defaults to \code{"TTLV"}.}

\item{require_validity}{Logical. Should rows be retained only when
\code{validity_column} is present and greater than zero? Defaults to \code{TRUE}.}

\item{mode}{Event extraction mode. \code{"changes"} returns rows where TTL values
change. \code{"nonzero"} returns all rows with nonzero TTL values.}

\item{include_initial}{Should the first valid, non-missing TTL value within
each group be treated as an event when \code{mode = "changes"}?}
}
\value{
A data frame of TTL events.
}
\description{
Extracts TTL marker events from Gazepoint Biometrics exports. The function
can return either rows where TTL marker values change or all nonzero TTL
rows. By default, rows are retained only when the TTL validity column is
present and greater than zero. This avoids treating invalid placeholder TTL
values as experimental events.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/external-interoperability.R
\name{prepare_gazepoint_neurokit_eda_input}
\alias{prepare_gazepoint_neurokit_eda_input}
\title{Prepare Gazepoint EDA input for NeuroKit2-style workflows}
\usage{
prepare_gazepoint_neurokit_eda_input(
  data,
  eda_col = "GSR_US",
  time_col = NULL,
  group_cols = NULL,
  sampling_rate = NULL,
  output_dir = NULL,
  prefix = "gazepoint_neurokit_eda"
)
}
\arguments{
\item{data}{A Gazepoint biometric data frame.}

\item{eda_col}{EDA/GSR signal column.}

\item{time_col}{Optional time/counter column.}

\item{group_cols}{Optional grouping columns.}

\item{sampling_rate}{Optional sampling rate in Hz.}

\item{output_dir}{Optional directory where per-group CSV files are written.}

\item{prefix}{File prefix used when \code{output_dir} is supplied.}
}
\value{
A list with \code{overview}, \code{eda_table}, \code{group_summary}, \code{manifest},
and \code{settings}.
}
\description{
Prepares EDA/GSR signal tables for optional external NeuroKit2 processing.
This helper does not require Python or NeuroKit2.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pyppg-hrp-support.R
\name{assess_gazepoint_hrp_waveform_quality}
\alias{assess_gazepoint_hrp_waveform_quality}
\title{Assess Gazepoint HRP waveform quality}
\usage{
assess_gazepoint_hrp_waveform_quality(
  data,
  hrp_col = NULL,
  time_col = NULL,
  group_cols = NULL,
  sampling_rate = NULL,
  time_unit = c("auto", "ms", "seconds", "samples"),
  min_rows = 20,
  min_finite_prop = 0.8,
  max_flat_prop = 0.95,
  flat_tolerance = 1e-08,
  max_gap_multiplier = 3
)
}
\arguments{
\item{data}{A Gazepoint biometric data frame or a list containing one.}

\item{hrp_col}{Optional HRP/PPG waveform column. If \code{NULL}, common column
names are detected.}

\item{time_col}{Optional time, timestamp, or sample-counter column.}

\item{group_cols}{Optional grouping columns.}

\item{sampling_rate}{Optional sampling rate in Hz.}

\item{time_unit}{Unit of \code{time_col}: \code{"auto"}, \code{"ms"}, \code{"seconds"}, or
\code{"samples"}.}

\item{min_rows}{Minimum rows required per group.}

\item{min_finite_prop}{Minimum finite waveform proportion required per group.}

\item{max_flat_prop}{Maximum allowed proportion of near-zero consecutive
differences among finite waveform values.}

\item{flat_tolerance}{Absolute difference threshold used to identify near-flat
consecutive waveform changes.}

\item{max_gap_multiplier}{Time gaps larger than this multiple of the median
positive time step are flagged.}
}
\value{
A list with \code{overview}, \code{group_quality}, \code{row_flags}, and \code{settings}.
}
\description{
Computes descriptive quality-control summaries for a Gazepoint HRP/PPG
waveform column. The output is intended for waveform availability,
missingness, flatness, and timing-gap review. It does not infer diagnosis,
emotion, valence, cognition, preference, or true physiological state.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scr-hurdle-model-data.R
\name{prepare_gazepoint_scr_hurdle_model_data}
\alias{prepare_gazepoint_scr_hurdle_model_data}
\title{Prepare Gazepoint SCR hurdle-model data}
\usage{
prepare_gazepoint_scr_hurdle_model_data(
  scr_event_windows,
  response_col = "response_flag",
  amplitude_col = "scr_amplitude",
  latency_col = "scr_latency",
  rise_time_col = "scr_rise_time",
  recovery_time_col = "scr_recovery_time",
  predictor_cols = NULL,
  factor_cols = NULL,
  numeric_cols = NULL,
  group_cols = NULL,
  event_id_col = "event_id",
  amplitude_transform = c("none", "log", "log1p"),
  amplitude_offset = 1e-06,
  drop_missing_predictors = TRUE
)
}
\arguments{
\item{scr_event_windows}{A \code{gazepoint_scr_event_window_summary} object
returned by \code{summarise_gazepoint_scr_event_windows()}, or a data frame
containing event-window SCR summaries.}

\item{response_col}{Column containing the binary SCR response flag.}

\item{amplitude_col}{Column containing SCR amplitude.}

\item{latency_col}{Optional column containing SCR latency.}

\item{rise_time_col}{Optional column containing SCR rise time.}

\item{recovery_time_col}{Optional column containing SCR recovery time.}

\item{predictor_cols}{Optional fixed-effect predictor columns to include in
generated formulas and complete-case checks.}

\item{factor_cols}{Optional columns to coerce to factors.}

\item{numeric_cols}{Optional columns to coerce to numeric.}

\item{group_cols}{Optional grouping columns retained for random effects or
clustered summaries.}

\item{event_id_col}{Optional event identifier column.}

\item{amplitude_transform}{Transformation for the positive-amplitude outcome:
\code{"none"}, \code{"log"}, or \code{"log1p"}.}

\item{amplitude_offset}{Small positive offset used when
\code{amplitude_transform = "log"}.}

\item{drop_missing_predictors}{Logical. If \code{TRUE}, model datasets are
restricted to rows complete on outcome and predictor columns.}
}
\value{
A list with \code{overview}, \code{response_model_data},
\code{amplitude_model_data}, \code{variable_summary}, \code{model_formulas}, and
\code{settings}.
}
\description{
Converts SCR event-window summaries into model-ready data for two-part
electrodermal-response analyses: a binary response/no-response component and
a positive-amplitude component among response events.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scr-peaks.R
\name{detect_gazepoint_scr_peaks}
\alias{detect_gazepoint_scr_peaks}
\title{Detect Gazepoint SCR peaks}
\usage{
detect_gazepoint_scr_peaks(
  data,
  signal_col = NULL,
  phasic_col = NULL,
  time_col = NULL,
  group_cols = NULL,
  prefer_vendor_phasic = TRUE,
  amplitude_min = 0.01,
  recovery_fraction = 0.5,
  smooth_width = 1,
  min_peak_distance = 1
)
}
\arguments{
\item{data}{A data frame containing Gazepoint biometric rows.}

\item{signal_col}{Optional conductance-like signal column, typically
\code{GSR_US}. Used when \code{phasic_col} is absent or unavailable.}

\item{phasic_col}{Optional phasic EDA signal column, typically
\code{GSR_US_PHASIC}.}

\item{time_col}{Optional time/counter column. If \code{NULL}, common Gazepoint
time columns are detected automatically.}

\item{group_cols}{Optional grouping columns. If \code{NULL}, available
source/participant/media/trial-like columns are used.}

\item{prefer_vendor_phasic}{Logical. If \code{TRUE}, prefer \code{GSR_US_PHASIC} when
available.}

\item{amplitude_min}{Minimum trough-to-peak amplitude required for a detected
SCR.}

\item{recovery_fraction}{Fraction of the peak amplitude used to define
recovery. The default \code{.5} estimates half-recovery.}

\item{smooth_width}{Optional odd integer moving-average width. Use \code{1} for no
smoothing.}

\item{min_peak_distance}{Minimum distance, in rows, allowed between retained
candidate peaks within each group. The default \code{1} preserves all local
maxima. Larger values reduce repeated detection of closely spaced local
maxima within a sustained SCR-like response.}
}
\value{
A list with \code{overview}, \code{peaks}, \code{group_summary}, \code{signal_summary},
and \code{settings}.
}
\description{
Detects candidate skin conductance responses (SCRs) from Gazepoint EDA/GSR
signals. The helper prefers a phasic channel such as \code{GSR_US_PHASIC} when
available, and otherwise falls back to a conductance-like signal such as
\code{GSR_US}. It returns explicit onset, peak, amplitude, rise-time, and
recovery-time fields for downstream event-window summaries and statistical
modelling.
}
\details{
This is a conservative R-native peak detector. It is not a replacement for
full model-based EDA decomposition tools such as Ledalab, PsPM, or cvxEDA.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hrv-rcmse.R
\name{extract_gazepoint_hrv_rcmse}
\alias{extract_gazepoint_hrv_rcmse}
\title{Extract refined composite multiscale entropy from HRV intervals}
\usage{
extract_gazepoint_hrv_rcmse(
  dat,
  ibi_col = "IBI",
  group_cols = NULL,
  scales = 1:10,
  m = 2,
  r_multiplier = 0.2,
  min_intervals = 20
)
}
\arguments{
\item{dat}{A data frame containing IBI/RR intervals.}

\item{ibi_col}{Numeric IBI/RR interval column.}

\item{group_cols}{Optional grouping columns.}

\item{scales}{Positive integer scales.}

\item{m}{Embedding dimension.}

\item{r_multiplier}{Tolerance multiplier applied to SD.}

\item{min_intervals}{Minimum intervals per group.}
}
\value{
A list with \code{overview}, \code{rcmse_by_scale}, \code{summary}, and \code{settings}.
}
\description{
Computes refined composite multiscale entropy (RCMSE) from IBI/RR intervals.
RCMSE pools template-match counts across all coarse-grained offsets at each
scale, making it more stable than ordinary MSE for shorter physiological
time series.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/simulation-and-chunking.R
\name{simulate_gazepoint_biometrics}
\alias{simulate_gazepoint_biometrics}
\title{Simulate Gazepoint-style biometric signals}
\usage{
simulate_gazepoint_biometrics(
  n_seconds = 120,
  sampling_rate = 60,
  participant_id = "sim_p1",
  scr_onsets = NULL,
  scr_rate_per_min = 4,
  pulse_rate_bpm = 72,
  respiration_rate_bpm = 15,
  eda_noise_sd = 0.01,
  ppg_noise_sd = 0.02,
  include_ttl = TRUE,
  seed = NULL
)
}
\arguments{
\item{n_seconds}{Duration in seconds.}

\item{sampling_rate}{Sampling rate in Hz.}

\item{participant_id}{Participant identifier.}

\item{scr_onsets}{Optional SCR onset times in seconds.}

\item{scr_rate_per_min}{Expected SCR events per minute when \code{scr_onsets} is
\code{NULL}.}

\item{pulse_rate_bpm}{Mean pulse rate.}

\item{respiration_rate_bpm}{Respiration-modulation rate.}

\item{eda_noise_sd}{EDA noise SD.}

\item{ppg_noise_sd}{PPG noise SD.}

\item{include_ttl}{Logical. If \code{TRUE}, adds TTL0 pulses at SCR onsets.}

\item{seed}{Optional random seed.}
}
\value{
A list with \code{overview}, \code{data}, \code{ground_truth}, and \code{settings}.
}
\description{
Generates synthetic Gazepoint-style EDA, PPG, HR, IBI, and TTL-like signals
with known ground-truth events. This is intended for teaching, examples, unit
tests, and model-validation workflows. It does not generate real participant
physiology.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aoi-biometrics.R
\name{summarise_gazepoint_aoi_biometrics}
\alias{summarise_gazepoint_aoi_biometrics}
\title{Summarise biometric signals by AOI}
\usage{
summarise_gazepoint_aoi_biometrics(
  data,
  aoi_col = "AOI",
  signal_cols = NULL,
  group_cols = NULL,
  time_col = NULL,
  valid_aoi_values = NULL,
  drop_missing_aoi = TRUE,
  min_rows = 1
)
}
\arguments{
\item{data}{A Gazepoint data frame containing AOI labels and biometric signals.}

\item{aoi_col}{AOI label column.}

\item{signal_cols}{Biometric signal columns to summarise.}

\item{group_cols}{Optional grouping columns, for example participant/media.}

\item{time_col}{Optional time/counter column.}

\item{valid_aoi_values}{Optional AOI labels to retain.}

\item{drop_missing_aoi}{Logical. If \code{TRUE}, rows with missing/blank AOI labels
are excluded.}

\item{min_rows}{Minimum rows required for a group/AOI/signal summary to be
marked as usable.}
}
\value{
A list with \code{overview}, \code{summary}, \code{signal_summary}, \code{aoi_summary},
\code{data}, and \code{settings}.
}
\description{
Summarises Gazepoint biometric channels within area-of-interest (AOI) rows.
The helper is intended for AOI-linked physiological descriptions, not for
inferring emotional valence.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/feature-inventory-format.R
\name{summarise_gazepoint_biometrics_feature_inventory}
\alias{summarise_gazepoint_biometrics_feature_inventory}
\title{Summarise the formatted gpbiometrics feature inventory}
\usage{
summarise_gazepoint_biometrics_feature_inventory(formatted_inventory = NULL)
}
\arguments{
\item{formatted_inventory}{Optional table returned by
\code{format_gazepoint_biometrics_feature_inventory()}.}
}
\value{
A list with domain, method-family, and user-level summaries.
}
\description{
Summarise the formatted gpbiometrics feature inventory
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/windows.R
\name{summarise_gazepoint_hr_windows}
\alias{summarise_gazepoint_hr_windows}
\title{Summarise Gazepoint heart-rate windows}
\usage{
summarise_gazepoint_hr_windows(
  data,
  group_columns = NULL,
  value_column = "HR",
  validity_column = "HRV",
  exclude_zero = TRUE
)
}
\arguments{
\item{data}{A data frame or a path to a Gazepoint CSV export.}

\item{group_columns}{Optional grouping columns defining windows, such as
\code{c("USER", "MEDIA_ID")}.}

\item{value_column}{Heart-rate value column. Defaults to \code{"HR"}.}

\item{validity_column}{Heart-rate validity column. Defaults to \code{"HRV"}.}

\item{exclude_zero}{Should zero values be excluded from usable summaries?}
}
\value{
A data frame with one row per window.
}
\description{
Summarises Gazepoint heart-rate values within participant, trial, stimulus,
AOI, or other user-defined windows. \code{HRV} is treated as a validity flag, not
as a heart-rate-variability metric.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/eda-wavelet-denoising.R
\name{denoise_gazepoint_eda_wavelet}
\alias{denoise_gazepoint_eda_wavelet}
\title{Denoise EDA using dependency-light Haar wavelet shrinkage}
\usage{
denoise_gazepoint_eda_wavelet(
  dat,
  eda_col = "GSR_US",
  group_cols = NULL,
  output_col = NULL,
  levels = 3,
  threshold_multiplier = 1,
  overwrite = FALSE
)
}
\arguments{
\item{dat}{A data frame containing EDA data.}

\item{eda_col}{EDA/conductance column.}

\item{group_cols}{Optional grouping columns.}

\item{output_col}{Optional output column.}

\item{levels}{Number of Haar decomposition levels.}

\item{threshold_multiplier}{Multiplier applied to the robust noise estimate.}

\item{overwrite}{Logical. If \code{FALSE}, protect existing output columns.}
}
\value{
A data frame with denoised EDA and denoising attributes.
}
\description{
Applies simple Haar wavelet soft-threshold denoising to EDA signals within
optional groups. This is a dependency-light wavelet denoising helper and
should not be described as an exact reproduction of stationary-wavelet
artifact-removal algorithms.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/schema-helpers.R
\name{detect_gazepoint_biometric_schema}
\alias{detect_gazepoint_biometric_schema}
\title{Detect the schema of Gazepoint biometric data}
\usage{
detect_gazepoint_biometric_schema(data)
}
\arguments{
\item{data}{A data frame.}
}
\value{
A list with \code{overview}, \code{columns}, \code{time_columns}, \code{timebase},
\code{name_map}, and \code{notes}.
}
\description{
Detects likely biometric, timing, marker, and identifying columns in a
Gazepoint Biometrics export. The function is deliberately descriptive. It
reports what appears to be present and active, but it does not infer emotion,
valence, or HRV from ambiguous raw columns.
}
\examples{
df <- data.frame(
  CNT = 1:5,
  TIME = seq(0, by = 1 / 60, length.out = 5),
  GSR = c(100, 101, 102, 101, 100),
  HR = c(70, 71, 72, 71, 70),
  HRV = c(1, 1, 1, 1, 1)
)
detect_gazepoint_biometric_schema(df)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nonlinear-complexity.R
\name{extract_gazepoint_eda_complexity}
\alias{extract_gazepoint_eda_complexity}
\title{Extract EDA complexity features}
\usage{
extract_gazepoint_eda_complexity(
  dat,
  eda_col = "GSR_US",
  group_cols = NULL,
  min_samples = 32,
  sampen_m = 2,
  sampen_r_multiplier = 0.2
)
}
\arguments{
\item{dat}{A data frame containing EDA data.}

\item{eda_col}{EDA/conductance column.}

\item{group_cols}{Optional grouping columns.}

\item{min_samples}{Minimum finite samples per group.}

\item{sampen_m}{Embedding dimension for sample entropy.}

\item{sampen_r_multiplier}{Tolerance multiplier applied to within-group SD.}
}
\value{
A list with \code{overview}, \code{features}, and \code{settings}.
}
\description{
Computes dependency-light EDA complexity descriptors, including sample entropy
and detrended fluctuation analysis alpha.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/report-bundle-readiness.R
\name{run_gazepoint_biometrics_real_data_readiness}
\alias{run_gazepoint_biometrics_real_data_readiness}
\title{Run a final real-data readiness gate for Gazepoint biometrics data}
\usage{
run_gazepoint_biometrics_real_data_readiness(
  data = NULL,
  workflow_result = NULL,
  min_rows = 100,
  min_active_signal_count = 1,
  max_missing_prop = 0.5,
  required_signal_cols = NULL,
  require_gsr_us_preferred = TRUE,
  require_ibi_for_hrv = FALSE,
  time_col = NULL,
  ttl_cols = NULL
)
}
\arguments{
\item{data}{A biometric data frame. If \code{NULL}, the function tries to extract a
data frame from \code{workflow_result}.}

\item{workflow_result}{Optional workflow/list object containing biometric data.}

\item{min_rows}{Minimum number of rows expected for a usable real-data check.}

\item{min_active_signal_count}{Minimum number of biometric signal columns with
at least one non-missing/non-zero value.}

\item{max_missing_prop}{Maximum acceptable missing proportion for detected
signal columns before a warning is raised.}

\item{required_signal_cols}{Optional signal columns that must be present.}

\item{require_gsr_us_preferred}{Logical. If \code{TRUE}, warns when \code{GSR_US} is
absent but \code{GSR} is present.}

\item{require_ibi_for_hrv}{Logical. If \code{TRUE}, fails when HRV is present but
IBI is absent. If \code{FALSE}, this condition is reported as a warning.}

\item{time_col}{Optional time column. If \code{NULL}, common time columns are
detected automatically.}

\item{ttl_cols}{Optional TTL marker columns. If \code{NULL}, \code{ttl_marker} or
\code{TTL0}-\code{TTL6} are detected automatically.}
}
\value{
A list with \code{overview}, \code{checks}, \code{signal_summary}, and \code{settings}.
}
\description{
Provides conservative pass/warn/fail checks before using real Gazepoint
Biometrics exports for analysis or reporting. The function checks basic row
count, signal availability, missingness, time ordering, TTL availability, and
HRV/IBI caution status. It does not certify data quality or infer emotional
states.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hrv.R
\name{summarise_gazepoint_ibi_hrv_windows}
\alias{summarise_gazepoint_ibi_hrv_windows}
\title{Summarise Gazepoint IBI-derived HRV windows}
\usage{
summarise_gazepoint_ibi_hrv_windows(
  data,
  group_columns,
  ibi_column = "IBI",
  validity_column = "HRV",
  min_ibi = 0.3,
  max_ibi = 2
)
}
\arguments{
\item{data}{A data frame or a path to a Gazepoint CSV export.}

\item{group_columns}{Columns defining analysis windows, such as
\code{c("source_participant", "MEDIA_ID")}.}

\item{ibi_column}{Interbeat-interval column. Defaults to \code{"IBI"}.}

\item{validity_column}{Optional validity column. Defaults to \code{"HRV"}.}

\item{min_ibi}{Minimum plausible IBI in seconds.}

\item{max_ibi}{Maximum plausible IBI in seconds.}
}
\value{
A data frame with one row per window and IBI-derived HRV summaries.
}
\description{
Summarises interbeat-interval (IBI) values within participant, stimulus,
trial, AOI, or other user-defined windows. This function derives simple
time-domain variability features from \code{IBI}. It does not use the Gazepoint
\code{HRV} column as a heart-rate-variability metric, because \code{HRV} is treated as
the Gazepoint heart-rate validity flag.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shiny-and-preregistration.R
\name{run_gpbiometrics_shiny}
\alias{run_gpbiometrics_shiny}
\title{Launch a lightweight gpbiometrics Shiny dashboard}
\usage{
run_gpbiometrics_shiny()
}
\value{
A Shiny application object, launched for interactive use.
}
\description{
Launches an optional local Shiny interface for inspecting a Gazepoint
Biometrics CSV file. Shiny is optional and is not required to use the package.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/validate.R
\name{audit_gazepoint_biometric_missingness}
\alias{audit_gazepoint_biometric_missingness}
\title{Audit missingness in Gazepoint biometric channels}
\usage{
audit_gazepoint_biometric_missingness(data, columns = NULL)
}
\arguments{
\item{data}{A data frame or a path to a Gazepoint CSV export.}

\item{columns}{Optional character vector of columns to audit. If \code{NULL},
known present Gazepoint biometric, TTL, and validity columns are audited.}
}
\value{
A data frame with one row per audited column.
}
\description{
Summarises missingness and zero values for Gazepoint biometric columns. This
is useful because Gazepoint exports may contain biometric columns even when a
channel was inactive or invalid during recording.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/automated-statistics.R
\name{run_gazepoint_automated_statistics}
\alias{run_gazepoint_automated_statistics}
\title{Run automated exploratory statistics for Gazepoint feature tables}
\usage{
run_gazepoint_automated_statistics(
  dat,
  outcome_cols,
  group_col,
  alpha = 0.05,
  p_adjust_method = "holm",
  normality_alpha = 0.05,
  min_group_n = 3
)
}
\arguments{
\item{dat}{A data frame.}

\item{outcome_cols}{Numeric outcome columns.}

\item{group_col}{Grouping/condition column.}

\item{alpha}{Significance level.}

\item{p_adjust_method}{P-value adjustment method.}

\item{normality_alpha}{Alpha used for Shapiro-Wilk normality screening.}

\item{min_group_n}{Minimum observations per group.}
}
\value{
A list with \code{overview}, \code{test_table}, \code{posthoc_table},
\code{normality_table}, and \code{settings}.
}
\description{
Runs simple exploratory group comparisons for numeric feature columns.
The function selects one-way ANOVA when all groups pass Shapiro-Wilk checks
and Kruskal-Wallis otherwise. It also performs pairwise post-hoc tests with
multiplicity correction.
}
\details{
This is an exploratory reporting helper. It is not a substitute for a
preregistered statistical model or expert review of study design.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/quality.R
\name{audit_gazepoint_engagement_dial}
\alias{audit_gazepoint_engagement_dial}
\title{Audit Gazepoint engagement-dial signal quality}
\usage{
audit_gazepoint_engagement_dial(
  data,
  value_column = "DIAL",
  validity_column = "DIALV",
  min_value = 0,
  max_value = 1,
  jump_threshold = NULL
)
}
\arguments{
\item{data}{A data frame or a path to a Gazepoint CSV export.}

\item{value_column}{Engagement-dial value column. Defaults to \code{"DIAL"}.}

\item{validity_column}{Engagement-dial validity column. Defaults to
\code{"DIALV"}.}

\item{min_value}{Minimum plausible dial value.}

\item{max_value}{Maximum plausible dial value.}

\item{jump_threshold}{Optional threshold for detecting large sample-to-sample
jumps.}
}
\value{
A one-row data frame summarising signal quality.
}
\description{
Audits Gazepoint engagement-dial values for missingness, inactive rows,
validity flags, plausible range, flatlining, and usable sample coverage.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/point-process-models.R
\name{model_gazepoint_eda_point_process}
\alias{model_gazepoint_eda_point_process}
\title{Model EDA events as a dependency-light point process}
\usage{
model_gazepoint_eda_point_process(
  dat,
  eda_col = "GSR_US",
  time_col = "CNT",
  group_cols = NULL,
  event_time_col = NULL,
  event_indicator_col = NULL,
  derivative_mad_multiplier = 6,
  min_event_distance_s = 1
)
}
\arguments{
\item{dat}{A data frame.}

\item{eda_col}{Numeric EDA column.}

\item{time_col}{Numeric time column.}

\item{group_cols}{Optional grouping columns.}

\item{event_time_col}{Optional column of event onset times.}

\item{event_indicator_col}{Optional binary event indicator column.}

\item{derivative_mad_multiplier}{MAD multiplier for derivative-derived events.}

\item{min_event_distance_s}{Minimum distance between derived events.}
}
\value{
A list with \code{overview}, \code{event_table}, \code{interval_table},
\code{process_summary}, and \code{settings}.
}
\description{
Creates event-time, inter-event interval, and inverse-Gaussian-style summary
tables for EDA/SCR events. Events can be supplied directly through an event
column or derived from positive EDA-derivative bursts.
}
\details{
This function is a compact point-process summary/model-preparation helper. It
does not fit a full latent sympathetic state-space model.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lsl-xdf-import.R
\name{import_gazepoint_lsl_xdf}
\alias{import_gazepoint_lsl_xdf}
\title{Import Gazepoint-related streams from an LSL/XDF file}
\usage{
import_gazepoint_lsl_xdf(
  path,
  stream_name_pattern = "Gazepoint|GP3|GSR|EDA|Biometric|TTL|Pupil|Gaze",
  include_all_streams = FALSE,
  flatten = TRUE,
  pyxdf_module = "pyxdf"
)
}
\arguments{
\item{path}{Path to an \code{.xdf} file.}

\item{stream_name_pattern}{Regular expression used to identify Gazepoint-like
streams when \code{include_all_streams = FALSE}.}

\item{include_all_streams}{Logical. If \code{TRUE}, return all streams.}

\item{flatten}{Logical. If \code{TRUE}, convert streams to data frames where
possible.}

\item{pyxdf_module}{Python module name, usually \code{"pyxdf"}.}
}
\value{
A list with \code{overview}, \code{streams}, \code{header}, and \code{settings}.
}
\description{
Reads an XDF file through the optional Python \code{pyxdf} package via
\code{reticulate}. This supports high-end LSL workflows without making Python a
hard dependency of gpbiometrics.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/external-eda-method-bridges.R
\name{classify_gazepoint_eda_response_pattern}
\alias{classify_gazepoint_eda_response_pattern}
\title{Classify descriptive Gazepoint EDA response patterns}
\usage{
classify_gazepoint_eda_response_pattern(
  data,
  response_col = NULL,
  group_cols = NULL,
  summary_function = c("max_abs", "mean_abs", "median_abs"),
  no_response_threshold = 0.01,
  low_response_threshold = 0.05,
  moderate_response_threshold = 0.2
)
}
\arguments{
\item{data}{A data frame containing EDA/SCR values.}

\item{response_col}{Optional response feature column. If omitted, the helper
searches for common SCR/EDA response columns and then \code{GSR_US_PHASIC} or
\code{GSR_US}.}

\item{group_cols}{Optional grouping columns.}

\item{summary_function}{Summary used within each group.}

\item{no_response_threshold}{Absolute response threshold for
\code{no_detectable_response}.}

\item{low_response_threshold}{Upper threshold for \code{low_response}.}

\item{moderate_response_threshold}{Upper threshold for \code{moderate_response}.}
}
\value{
A list with \code{overview}, \code{classifications}, and \code{settings}.
}
\description{
Classifies descriptive EDA response patterns from an EDA/SCR feature column.
The labels are intended for quality-control and descriptive reporting only.
They do not infer emotion, valence, stress, trust, preference, cognition, or
diagnosis.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hrv-surrogate-nonlinearity.R
\name{test_gazepoint_hrv_nonlinearity}
\alias{test_gazepoint_hrv_nonlinearity}
\title{Test HRV nonlinearity using surrogate data}
\usage{
test_gazepoint_hrv_nonlinearity(
  dat,
  ibi_col = "IBI",
  group_cols = NULL,
  metric = c("sample_entropy", "approximate_entropy", "sd1_sd2_ratio"),
  n_surrogates = 99,
  surrogate_method = c("phase_randomized", "shuffle"),
  m = 2,
  r_multiplier = 0.2,
  statistic_fun = NULL,
  seed = NULL
)
}
\arguments{
\item{dat}{A data frame containing IBI/RR intervals.}

\item{ibi_col}{Numeric IBI/RR interval column.}

\item{group_cols}{Optional grouping columns.}

\item{metric}{Nonlinear statistic to test.}

\item{n_surrogates}{Number of surrogate series per group.}

\item{surrogate_method}{\code{"phase_randomized"} or \code{"shuffle"}.}

\item{m}{Embedding dimension for entropy metrics.}

\item{r_multiplier}{Tolerance multiplier for entropy metrics.}

\item{statistic_fun}{Optional custom statistic function accepting numeric x.}

\item{seed}{Optional random seed.}
}
\value{
A list with \code{overview}, \code{results}, \code{surrogate_statistics}, and
\code{settings}.
}
\description{
Tests whether a nonlinear HRV statistic differs from surrogate RR/IBI
sequences. This is a screening tool for evidence inconsistent with a simple
linear stochastic null process. It does not prove deterministic chaos or
diagnose any condition.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hrv-rqa-geometric.R
\name{extract_gazepoint_hrv_rqa}
\alias{extract_gazepoint_hrv_rqa}
\title{Extract HRV recurrence quantification analysis features}
\usage{
extract_gazepoint_hrv_rqa(
  dat,
  ibi_col = "IBI",
  group_cols = NULL,
  embedding_dimension = 2,
  delay = 1,
  radius = NULL,
  radius_multiplier = 0.2,
  min_line_length = 2
)
}
\arguments{
\item{dat}{A data frame.}

\item{ibi_col}{Numeric IBI/RR interval column.}

\item{group_cols}{Optional grouping columns.}

\item{embedding_dimension}{Embedding dimension for phase-space reconstruction.}

\item{delay}{Delay used in embedding.}

\item{radius}{Radius for recurrence threshold. If \code{NULL}, uses
\code{radius_multiplier * SD}.}

\item{radius_multiplier}{Multiplier used when \code{radius = NULL}.}

\item{min_line_length}{Minimum diagonal/vertical line length.}
}
\value{
A list with \code{overview}, \code{features}, and \code{settings}.
}
\description{
Computes dependency-light recurrence quantification analysis (RQA) features
from IBI/RR intervals. This is intended as a compact nonlinear HRV summary
and not as a clinical diagnostic tool.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/environment-confounds-and-ipfm.R
\name{extract_gazepoint_beats_kmeans}
\alias{extract_gazepoint_beats_kmeans}
\title{Extract heartbeat candidates from Gazepoint pulse using k-means}
\usage{
extract_gazepoint_beats_kmeans(
  dat,
  pulse_col = "HRP",
  time_col = "CNT",
  group_cols = NULL,
  k = 2,
  peak_polarity = c("positive", "negative"),
  min_distance_s = 0.3,
  sampling_rate = NULL,
  seed = NULL
)
}
\arguments{
\item{dat}{A data frame.}

\item{pulse_col}{Numeric pulse/PPG column.}

\item{time_col}{Numeric time column.}

\item{group_cols}{Optional grouping columns.}

\item{k}{Number of k-means clusters.}

\item{peak_polarity}{\code{"positive"} or \code{"negative"}.}

\item{min_distance_s}{Minimum time between selected beats.}

\item{sampling_rate}{Optional sampling rate in Hz.}

\item{seed}{Optional random seed.}
}
\value{
A list with \code{overview}, \code{beat_table}, \code{interval_table},
\code{timeseries}, and \code{settings}.
}
\description{
Uses k-means clustering on the raw pulse waveform to classify likely
heartbeat regions and then selects local extrema as beat candidates. This is
a Gazepoint Biometrics-oriented fallback for difficult pulse waveforms, not
an ECG-equivalent R-peak detector.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cardiorespiratory-causality.R
\name{analyze_gazepoint_cardiorespiratory_causality}
\alias{analyze_gazepoint_cardiorespiratory_causality}
\title{Analyse cardiorespiratory Granger-style directionality}
\usage{
analyze_gazepoint_cardiorespiratory_causality(
  dat,
  respiration_col,
  cardiac_col,
  time_col = NULL,
  group_cols = NULL,
  lag_order = 3,
  min_rows = 30,
  standardise = TRUE
)
}
\arguments{
\item{dat}{A data frame.}

\item{respiration_col}{Numeric respiration proxy column.}

\item{cardiac_col}{Numeric cardiac column, such as HR, IBI, or RR.}

\item{time_col}{Optional time column for ordering.}

\item{group_cols}{Optional grouping columns.}

\item{lag_order}{VAR lag order.}

\item{min_rows}{Minimum complete rows per group.}

\item{standardise}{Logical. If \code{TRUE}, z-standardise both series per group.}
}
\value{
A list with \code{overview}, \code{causality_summary}, and \code{settings}.
}
\description{
Computes dependency-light linear Granger-style directionality tests between a
respiration proxy and heart-rate or IBI/RR signal. This estimates predictive
directionality in a VAR-style model. It does not prove physiological
causality from observational data by itself.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/eda-advanced.R
\name{detect_gazepoint_scr_events}
\alias{detect_gazepoint_scr_events}
\title{Detect SCR-like events in Gazepoint GSR/EDA data}
\usage{
detect_gazepoint_scr_events(
  data,
  phasic_col = NULL,
  signal_col = NULL,
  time_col = NULL,
  group_cols = NULL,
  threshold = NULL,
  min_peak_distance = 10L,
  window_size = 31L
)
}
\arguments{
\item{data}{A data frame.}

\item{phasic_col}{Optional phasic EDA column.}

\item{signal_col}{Optional raw/conductance EDA column used when \code{phasic_col}
is not supplied.}

\item{time_col}{Optional time/order column.}

\item{group_cols}{Optional grouping columns.}

\item{threshold}{Optional numeric detection threshold. If \code{NULL}, a robust
group-specific threshold is estimated as median plus three MADs, bounded
below by zero.}

\item{min_peak_distance}{Minimum distance between retained peaks in samples.}

\item{window_size}{Rolling-median window size used if decomposition is needed.}
}
\value{
A list with \code{overview}, \code{events}, \code{group_summary}, and \code{settings}.
}
\description{
Detects simple SCR-like peaks from a phasic EDA signal. If a phasic column is
not supplied, the function first creates a descriptive phasic component using
\code{decompose_gazepoint_eda()}. This helper is intended for exploratory quality
control and descriptive summaries, not as a replacement for specialised SCR
detection pipelines.
}
\examples{
df <- data.frame(
  CNT = 1:20,
  GSR_US_PHASIC = c(rep(0, 5), 0.2, 0.8, 0.2, rep(0, 12))
)
detect_gazepoint_scr_events(df, phasic_col = "GSR_US_PHASIC", time_col = "CNT")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/biometrics-report.R
\name{create_gazepoint_biometrics_report}
\alias{create_gazepoint_biometrics_report}
\title{Create a Gazepoint Biometrics report}
\usage{
create_gazepoint_biometrics_report(
  data = NULL,
  workflow = NULL,
  validation = NULL,
  quality = NULL,
  sampling = NULL,
  missingness = NULL,
  exclusions = NULL,
  report_tables = NULL,
  methods_text = NULL,
  checklist = NULL,
  title = "Gazepoint Biometrics report",
  subtitle = NULL,
  output_file = NULL,
  format = c("markdown", "html"),
  include_timestamp = FALSE,
  overwrite = FALSE,
  max_table_rows = 20L
)
}
\arguments{
\item{data}{Optional biometric data frame.}

\item{workflow}{Optional workflow object or workflow summary list.}

\item{validation}{Optional validation object or data frame.}

\item{quality}{Optional quality-audit object or data frame.}

\item{sampling}{Optional sampling-audit object or data frame.}

\item{missingness}{Optional missingness-audit object or data frame.}

\item{exclusions}{Optional exclusion-recommendation object or data frame.}

\item{report_tables}{Optional report-table object, data frame, or named list
of data frames.}

\item{methods_text}{Optional methods text, character vector, or object
returned by \code{create_gazepoint_biometrics_methods_text()}.}

\item{checklist}{Optional checklist object or data frame.}

\item{title}{Report title.}

\item{subtitle}{Optional report subtitle.}

\item{output_file}{Optional path to write a report file.}

\item{format}{Output format when \code{output_file} is supplied. Supported values
are \code{"markdown"} and \code{"html"}.}

\item{include_timestamp}{Logical. Should a creation timestamp be included?}

\item{overwrite}{Logical. Should an existing \code{output_file} be overwritten?}

\item{max_table_rows}{Maximum number of rows shown per table in the written
report.}
}
\value{
A list of class \code{"gazepoint_biometrics_report"} with \code{overview},
\code{sections}, \code{tables}, \code{objects}, \code{output_file}, and \code{settings}.
}
\description{
Creates a structured, manuscript-oriented report object for Gazepoint
Biometrics data, workflow outputs, quality checks, report tables, methods
text, and reporting checklists. The report is intentionally conservative:
GSR/EDA is described as electrodermal activity/arousal-related signal rather
than emotional valence; heart-rate interpretation is tied to baseline/task
context; eye-tracking is described as visual attention rather than direct
cognition; and raw \code{HRV} columns are not treated as HRV metrics.
}
\details{
The function can also write a lightweight Markdown or HTML file without
adding heavy reporting dependencies.
}
\examples{
df <- data.frame(
  CNT = 1:5,
  GSR = c(1, 1.1, 1.2, 1.1, 1),
  HR = c(70, 71, 72, 71, 70),
  DIAL = c(40, 42, 44, 43, 41)
)
report <- create_gazepoint_biometrics_report(df)
names(report)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pupil-and-saccade-diagnostics.R
\name{plot_gazepoint_saccade_main_sequence}
\alias{plot_gazepoint_saccade_main_sequence}
\title{Plot Gazepoint saccade main-sequence diagnostics}
\usage{
plot_gazepoint_saccade_main_sequence(
  dat,
  amplitude_col = NULL,
  peak_velocity_col = NULL,
  group_col = NULL,
  log_axes = TRUE,
  add_smoother = TRUE,
  main = "Gazepoint saccade main-sequence diagnostic"
)
}
\arguments{
\item{dat}{A saccade-level data frame.}

\item{amplitude_col}{Saccade amplitude column.}

\item{peak_velocity_col}{Peak velocity column.}

\item{group_col}{Optional grouping column.}

\item{log_axes}{Logical. If \code{TRUE}, use log10 axes.}

\item{add_smoother}{Logical. If \code{TRUE}, add a lowess curve.}

\item{main}{Plot title.}
}
\value{
Invisibly returns the plotted data and settings.
}
\description{
Plots saccade amplitude against peak velocity. The function expects
saccade-level amplitude and peak-velocity columns. If raw sample-level data
are supplied, users should first derive saccade-level kinematics with a
validated fixation/saccade detector.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data_summary.R
\name{import_gazepoint_data_summary}
\alias{import_gazepoint_data_summary}
\title{Import a Gazepoint Data Summary export}
\usage{
import_gazepoint_data_summary(file)
}
\arguments{
\item{file}{Path to a Gazepoint \code{Data_Summary_export_*.csv} file.}
}
\value{
A list with \code{metadata}, \code{aoi_summary}, and \code{aoi_statistics} data
frames. The returned object has class \code{"gazepoint_data_summary"}.
}
\description{
Reads a multi-section \code{Data_Summary_export_*.csv} file produced by Gazepoint
Analysis. These files are not ordinary rectangular CSV files. They contain
metadata followed by sections such as \verb{AOI Summary} and \verb{AOI Statistics (for each user)}. The latter may include AOI-level biometric summaries such as
average dial value, average GSR, average heart rate, average interbeat
interval, and pupil diameter.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/schema-helpers.R
\name{detect_gazepoint_time_columns}
\alias{detect_gazepoint_time_columns}
\title{Detect Gazepoint biometric time columns}
\usage{
detect_gazepoint_time_columns(data)
}
\arguments{
\item{data}{A data frame or a character vector of column names.}
}
\value{
A data frame with one row per detected time-related column.
}
\description{
Detects likely timing and counter columns in Gazepoint Biometrics exports.
The function reports candidate timing columns rather than assuming that any
single time variable is always present or always measured in the same unit.
}
\examples{
detect_gazepoint_time_columns(c("CNT", "TIME_MS", "GSR", "HR"))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/external-eda-method-bridges.R
\name{prepare_gazepoint_pspm_input}
\alias{prepare_gazepoint_pspm_input}
\title{Prepare Gazepoint EDA input for external PsPM-style workflows}
\usage{
prepare_gazepoint_pspm_input(
  data,
  eda_col = NULL,
  time_col = NULL,
  group_cols = NULL,
  sampling_rate = NULL,
  time_unit = c("auto", "ms", "seconds", "samples"),
  convert_resistance_to_us = FALSE,
  min_finite_prop = 0.5,
  output_dir = NULL,
  prefix = "gazepoint_pspm"
)
}
\arguments{
\item{data}{A Gazepoint biometric data frame, or a list containing one.}

\item{eda_col}{Optional EDA/conductance column. If omitted, the function
prefers \code{GSR_US} when available.}

\item{time_col}{Optional time/counter column.}

\item{group_cols}{Optional grouping columns.}

\item{sampling_rate}{Optional sampling rate in Hz, used when the time column
is a sample counter.}

\item{time_unit}{Unit of \code{time_col}.}

\item{convert_resistance_to_us}{If \code{TRUE}, convert a selected resistance-like
\code{GSR} column to microsiemens as \verb{1,000,000 / GSR}. The default is \code{FALSE}
because this conversion should be used only when the user has verified that
\code{GSR} is resistance-like and \code{GSR_US} is unavailable.}

\item{min_finite_prop}{Minimum finite proportion required for a group to be
labelled ready.}

\item{output_dir}{Optional folder where CSV files should be written.}

\item{prefix}{File prefix used when \code{output_dir} is supplied.}
}
\value{
A list with \code{overview}, \code{signal_table}, \code{group_summary}, \code{manifest},
and \code{settings}.
}
\description{
Prepares a clean Gazepoint EDA/conductance time-series table that can be
exported for external PsPM-style workflows. This function does not run PsPM
and does not attempt to reproduce PsPM internally.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/biometric-plots.R
\name{plot_gazepoint_biometric_quality}
\alias{plot_gazepoint_biometric_quality}
\title{Plot Gazepoint biometric quality indicators}
\usage{
plot_gazepoint_biometric_quality(
  data,
  quality_cols = NULL,
  signal_cols = NULL,
  time_col = NULL,
  group_col = NULL,
  dropout_prefix = "biometric_dropout",
  max_points = 5000L,
  main = NULL,
  plot = TRUE,
  ...
)
}
\arguments{
\item{data}{A data frame.}

\item{quality_cols}{Optional quality/flag columns. If \code{NULL}, likely quality
columns are detected from names and types.}

\item{signal_cols}{Optional signal columns used to derive missingness flags
when no quality columns are detected.}

\item{time_col}{Optional time/order column recorded in the returned settings.}

\item{group_col}{Optional grouping column for group-level quality summaries.}

\item{dropout_prefix}{Prefix used by dropout columns created by
\code{\link[=flag_gazepoint_biometric_dropouts]{flag_gazepoint_biometric_dropouts()}}.}

\item{max_points}{Maximum number of rows used for row-level returned plot
data.}

\item{main}{Optional plot title.}

\item{plot}{Logical. If \code{FALSE}, no plot is drawn.}

\item{...}{Additional arguments passed to \code{\link[=barplot]{barplot()}}.}
}
\value{
A list with \code{overview}, \code{quality_summary}, \code{group_summary},
\code{plot_data}, and \code{settings}.
}
\description{
Plots and summarises biometric quality indicators such as dropout flags,
validity flags, missingness flags, and quality/audit flags. When no explicit
quality columns are available, the function can derive missingness indicators
from detected biometric signal columns.
}
\examples{
df <- data.frame(
  CNT = 1:5,
  GSR = c(1, NA, 1.2, 1.1, NA),
  HR_valid = c(1, 1, 0, 1, 1)
)
plot_gazepoint_biometric_quality(df, signal_cols = "GSR", plot = FALSE)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/autoencoder-denoising-bridges.R
\name{denoise_gazepoint_ppg_autoencoder}
\alias{denoise_gazepoint_ppg_autoencoder}
\title{Denoise PPG using a user-supplied autoencoder reconstruction model}
\usage{
denoise_gazepoint_ppg_autoencoder(
  dat,
  ppg_col = "HRP",
  time_col = NULL,
  group_cols = NULL,
  model = NULL,
  window_samples = 128,
  output_col = NULL,
  overwrite = FALSE
)
}
\arguments{
\item{dat}{A data frame.}

\item{ppg_col}{Numeric PPG/pulse column.}

\item{time_col}{Optional time column.}

\item{group_cols}{Optional grouping columns.}

\item{model}{A user-supplied function or model. If \code{NULL}, the original
signal is copied and status records that no model was supplied.}

\item{window_samples}{Window length in samples.}

\item{output_col}{Optional output column.}

\item{overwrite}{Logical. If \code{FALSE}, existing output column is protected.}
}
\value{
A data frame with reconstructed signal and denoising attributes.
}
\description{
Applies a user-supplied reconstruction function or model to fixed-length PPG
windows. No pretrained neural network is bundled. This function is an
interoperability bridge for validated user-supplied autoencoders.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/time-activity-qc.R
\name{audit_gazepoint_time_resets}
\alias{audit_gazepoint_time_resets}
\title{Audit Gazepoint biometric time resets}
\usage{
audit_gazepoint_time_resets(
  data,
  time_col = NULL,
  group_cols = NULL,
  allow_ties = TRUE,
  split_on_negative_step = TRUE,
  return_reindexed_time = FALSE,
  min_segment_rows = 1
)
}
\arguments{
\item{data}{A data frame containing Gazepoint biometric rows.}

\item{time_col}{Optional time/counter column. If \code{NULL}, common Gazepoint time
columns are detected automatically.}

\item{group_cols}{Optional grouping columns. If \code{NULL}, available
source/participant/media/trial-like columns are used.}

\item{allow_ties}{Logical. If \code{TRUE}, repeated time values are not treated as
non-monotonic.}

\item{split_on_negative_step}{Logical. If \code{TRUE}, negative time steps start a
new segment within each group.}

\item{return_reindexed_time}{Logical. If \code{TRUE}, adds
\code{time_reindexed_within_segment}, starting at zero within each detected
segment.}

\item{min_segment_rows}{Minimum rows expected per segment before a segment is
flagged as short.}
}
\value{
A list with \code{overview}, \code{segment_summary}, \code{row_flags},
\code{data_with_segments}, and \code{settings}.
}
\description{
Detects negative time steps, duplicate time steps, non-finite time values, and
recording segments within grouped Gazepoint biometric exports. This helper is
intended for quality control and synchronization inspection. It does not alter
raw values unless \code{return_reindexed_time = TRUE}, in which case an additional
segment-relative time column is added.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/eda-nonresponders.R
\name{screen_gazepoint_eda_nonresponders}
\alias{screen_gazepoint_eda_nonresponders}
\title{Screen Gazepoint EDA/SCR nonresponders}
\usage{
screen_gazepoint_eda_nonresponders(
  x,
  group_cols = NULL,
  response_col = "response_flag",
  amplitude_col = "scr_amplitude",
  min_events = 1,
  min_response_events = 1,
  min_response_rate = 0.05,
  min_detected_peaks = 1
)
}
\arguments{
\item{x}{A \code{gazepoint_scr_event_window_summary},
\code{gazepoint_scr_peak_detection}, or data frame.}

\item{group_cols}{Optional grouping columns, usually participant columns.}

\item{response_col}{Binary response column for event-window data.}

\item{amplitude_col}{SCR amplitude column.}

\item{min_events}{Minimum number of events required before event-window
nonresponder classification is made.}

\item{min_response_events}{Minimum number of response events required.}

\item{min_response_rate}{Minimum response rate required.}

\item{min_detected_peaks}{Minimum detected peaks required when only peak data
are available.}
}
\value{
A list with \code{overview}, \code{group_summary}, \code{candidate_nonresponders},
and \code{settings}.
}
\description{
Screens groups, usually participants, for low or absent SCR responding using
either SCR event-window summaries or SCR peak-detection outputs. The helper
reports candidate nonresponders under explicit, user-controlled criteria.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/workflow.R
\name{summarise_gazepoint_biometrics_workflow}
\alias{summarise_gazepoint_biometrics_workflow}
\title{Summarise a Gazepoint Biometrics workflow object}
\usage{
summarise_gazepoint_biometrics_workflow(workflow)
}
\arguments{
\item{workflow}{A workflow object returned by
\code{run_gazepoint_biometrics_workflow()}.}
}
\value{
A one-row data frame summarising the workflow.
}
\description{
Creates a compact summary table from an object returned by
\code{run_gazepoint_biometrics_workflow()}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/changepoint-scr-recovery.R
\name{detect_gazepoint_doubly_stochastic_changepoints}
\alias{detect_gazepoint_doubly_stochastic_changepoints}
\title{Detect stochastic change points in noisy biometric signals}
\usage{
detect_gazepoint_doubly_stochastic_changepoints(
  dat,
  signal_col,
  time_col = "CNT",
  group_cols = NULL,
  window_seconds = 10,
  step_seconds = 2,
  threshold_mad_multiplier = 6,
  min_distance_s = 5
)
}
\arguments{
\item{dat}{A data frame.}

\item{signal_col}{Numeric signal column.}

\item{time_col}{Numeric time column.}

\item{group_cols}{Optional grouping columns.}

\item{window_seconds}{Window length in seconds.}

\item{step_seconds}{Step size in seconds.}

\item{threshold_mad_multiplier}{Robust threshold multiplier.}

\item{min_distance_s}{Minimum distance between detected change points.}
}
\value{
A list with \code{overview}, \code{score_table}, \code{changepoints}, and \code{settings}.
}
\description{
Detects abrupt changes in noisy biological time series using a dependency-light
stochastic rolling-window approximation. The score combines adjacent-window
changes in mean and variance with a robust adaptive threshold.
}
\details{
This is not a full reproduction of any specific doubly stochastic model. It
is a transparent approximation for QC and exploratory segmentation.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/artifact-svm-bridge.R
\name{prepare_gazepoint_artifact_svm_features}
\alias{prepare_gazepoint_artifact_svm_features}
\title{Prepare EDA artifact-classifier segment features}
\usage{
prepare_gazepoint_artifact_svm_features(
  dat,
  eda_col = "GSR_US",
  time_col = NULL,
  group_cols = NULL,
  segment_seconds = 5,
  samples_per_segment = NULL,
  sampling_rate = NULL
)
}
\arguments{
\item{dat}{A data frame.}

\item{eda_col}{EDA/conductance column.}

\item{time_col}{Optional time column.}

\item{group_cols}{Optional grouping columns.}

\item{segment_seconds}{Segment length in seconds when \code{time_col} is supplied.}

\item{samples_per_segment}{Segment length in samples when no usable time
column or sampling rate is available.}

\item{sampling_rate}{Optional sampling rate in Hz.}
}
\value{
A segment-level feature data frame.
}
\description{
Creates segment-level features that can be passed to a user-supplied
artifact classifier such as an SVM. No pretrained classifier is bundled.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pyppg-hrp-support.R
\name{prepare_gazepoint_pyppg_input}
\alias{prepare_gazepoint_pyppg_input}
\title{Prepare Gazepoint HRP/PPG waveform input for pyPPG}
\usage{
prepare_gazepoint_pyppg_input(
  data,
  ppg_col = NULL,
  time_col = NULL,
  group_cols = NULL,
  sampling_rate = NULL,
  time_unit = c("auto", "ms", "seconds", "samples"),
  min_finite_prop = 0.5,
  output_dir = NULL,
  prefix = "gazepoint_pyppg"
)
}
\arguments{
\item{data}{A Gazepoint biometric data frame or a list containing one.}

\item{ppg_col}{Optional HRP/PPG waveform column. If \code{NULL}, common Gazepoint
HRP/PPG column names are detected.}

\item{time_col}{Optional time, timestamp, or sample-counter column.}

\item{group_cols}{Optional grouping columns.}

\item{sampling_rate}{Optional sampling rate in Hz. Useful when \code{time_col}
is a sample counter such as \code{CNT}.}

\item{time_unit}{Unit of \code{time_col}: \code{"auto"}, \code{"ms"}, \code{"seconds"}, or
\code{"samples"}.}

\item{min_finite_prop}{Minimum finite waveform proportion used for group
readiness summaries.}

\item{output_dir}{Optional directory for CSV export. If \code{NULL}, no files are
written.}

\item{prefix}{File prefix used when \code{output_dir} is supplied.}
}
\value{
A list with \code{overview}, \code{waveform_table}, \code{group_summary},
\code{manifest}, and \code{settings}.
}
\description{
Prepares a Gazepoint heart-rate pulse waveform column, usually \code{HRP}, as a
lightweight input table for optional pyPPG workflows. This helper does not
call Python, does not require pyPPG, and does not derive HRV features. It only
prepares waveform values, timing information when available, and conservative
group-level summaries for interoperability review.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pspm-dcm-bridge.R
\name{prepare_gazepoint_pspm_dcm_input}
\alias{prepare_gazepoint_pspm_dcm_input}
\title{Prepare Gazepoint EDA data for PsPM DCM workflows}
\usage{
prepare_gazepoint_pspm_dcm_input(
  dat,
  eda_col = "GSR_US",
  time_col = "CNT",
  event_onset_col = NULL,
  event_duration_col = NULL,
  event_name_col = NULL,
  participant_col = NULL,
  session_col = NULL,
  sampling_rate = NULL,
  output_dir = NULL,
  prefix = "gazepoint_pspm_dcm"
)
}
\arguments{
\item{dat}{A data frame containing EDA data.}

\item{eda_col}{Numeric EDA/conductance column.}

\item{time_col}{Numeric time column.}

\item{event_onset_col}{Optional event onset column.}

\item{event_duration_col}{Optional event duration column.}

\item{event_name_col}{Optional event name/condition column.}

\item{participant_col}{Optional participant column.}

\item{session_col}{Optional session column.}

\item{sampling_rate}{Optional sampling rate in Hz.}

\item{output_dir}{Optional directory for CSV export.}

\item{prefix}{File prefix when \code{output_dir} is supplied.}
}
\value{
A list with \code{overview}, \code{signal_table}, \code{event_table},
\code{pspm_notes}, \code{written_files}, and \code{settings}.
}
\description{
Prepares Gazepoint EDA data and event metadata for downstream PsPM dynamic
causal modelling workflows. This function does not run PsPM or invert a DCM
model in R. It creates structured input tables and notes for MATLAB/PsPM.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ibi-helpers.R
\name{audit_gazepoint_ibi_quality}
\alias{audit_gazepoint_ibi_quality}
\title{Audit IBI/RR interval quality}
\usage{
audit_gazepoint_ibi_quality(
  data,
  ibi_col = NULL,
  group_cols = NULL,
  time_col = NULL,
  unit = c("auto", "milliseconds", "seconds"),
  min_ibi_ms = 300,
  max_ibi_ms = 2000,
  max_jump_ms = 500
)
}
\arguments{
\item{data}{A data frame.}

\item{ibi_col}{Optional IBI/RR interval column. If \code{NULL}, the function
detects recognised IBI/RR-style column names.}

\item{group_cols}{Optional grouping columns, such as participant, trial,
stimulus, condition, or window labels.}

\item{time_col}{Optional time/order column used to order samples before
successive-difference checks.}

\item{unit}{Unit of the IBI values. \code{"auto"} treats median values below 10 as
seconds and larger values as milliseconds.}

\item{min_ibi_ms}{Minimum plausible IBI in milliseconds.}

\item{max_ibi_ms}{Maximum plausible IBI in milliseconds.}

\item{max_jump_ms}{Maximum plausible absolute change between successive IBI
values within a group.}
}
\value{
A list with \code{overview}, \code{samples}, \code{group_summary}, and \code{settings}.
}
\description{
Audits inter-beat interval (IBI) or RR-interval values in Gazepoint
Biometrics-style exports. This helper deliberately does not use raw \code{HRV}
columns as HRV metrics. HRV-style summaries should be derived only from
genuine IBI/RR interval columns.
}
\examples{
df <- data.frame(
  USER = rep(c("P1", "P2"), each = 4),
  IBI = c(800, 810, 790, 805, 900, 910, 905, 920)
)
audit_gazepoint_ibi_quality(df, group_cols = "USER")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/validity-dropouts.R
\name{flag_gazepoint_biometric_dropouts}
\alias{flag_gazepoint_biometric_dropouts}
\title{Flag biometric dropouts and flatline periods}
\usage{
flag_gazepoint_biometric_dropouts(
  data,
  signal_cols = NULL,
  group_cols = NULL,
  time_col = NULL,
  min_missing_run = 5L,
  min_flatline_run = 10L,
  constant_tolerance = 0,
  prefix = "biometric_dropout"
)
}
\arguments{
\item{data}{A data frame.}

\item{signal_cols}{Optional character vector of biometric signal columns. If
\code{NULL}, common Gazepoint biometric signal columns are detected.}

\item{group_cols}{Optional grouping columns. Runs are computed separately
within each group.}

\item{time_col}{Optional time column used to order rows within each group
before run detection. If \code{NULL}, the current row order is used.}

\item{min_missing_run}{Minimum consecutive missing/non-finite samples required
to flag a missing dropout.}

\item{min_flatline_run}{Minimum consecutive unchanged finite samples required
to flag a flatline dropout.}

\item{constant_tolerance}{Numeric tolerance used when detecting unchanged
values for flatline runs.}

\item{prefix}{Prefix for generated dropout columns.}
}
\value{
The input data frame with added logical dropout columns. The attributes
\code{dropout_summary} and \code{dropout_settings} contain structured summaries.
}
\description{
Flags missing-value runs and sustained flatline runs in Gazepoint biometric
signal columns. Missing dropouts are defined as consecutive missing or
non-finite numeric samples. Flatline dropouts are defined as consecutive
finite numeric samples that remain unchanged within a tolerance.
}
\details{
The function adds row-level flags and stores a dropout summary in the returned
data frame attributes. It does not remove rows.
}
\examples{
df <- data.frame(
  CNT = 1:8,
  GSR = c(1, NA, NA, NA, 2, 2, 2, 3),
  HR = c(70, 71, 72, 73, 74, 75, 76, 77)
)
flag_gazepoint_biometric_dropouts(df, min_missing_run = 3, min_flatline_run = 3)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/eda-artifact-qc.R
\name{audit_gazepoint_eda_artifacts}
\alias{audit_gazepoint_eda_artifacts}
\title{Audit Gazepoint EDA/GSR artifacts}
\usage{
audit_gazepoint_eda_artifacts(
  data,
  signal_col = NULL,
  time_col = NULL,
  group_cols = NULL,
  prefer_gsr_us = TRUE,
  jump_threshold_sd = 6,
  slope_threshold_sd = 6,
  flat_run_length = 20,
  zero_run_length = 20,
  saturation_min = NULL,
  saturation_max = NULL,
  negative_allowed = NULL
)
}
\arguments{
\item{data}{A data frame containing Gazepoint biometric rows.}

\item{signal_col}{Optional EDA/GSR signal column. If \code{NULL}, the function
prefers \code{GSR_US} and then falls back to common Gazepoint EDA columns.}

\item{time_col}{Optional time/counter column. If \code{NULL}, common Gazepoint
time columns are detected automatically.}

\item{group_cols}{Optional grouping columns. If \code{NULL}, available
source/participant/media/trial-like columns are used.}

\item{prefer_gsr_us}{Logical. If \code{TRUE}, prefer \code{GSR_US} when \code{signal_col}
is not supplied.}

\item{jump_threshold_sd}{Robust z threshold for absolute signal jumps.}

\item{slope_threshold_sd}{Robust z threshold for absolute signal slopes.}

\item{flat_run_length}{Minimum repeated-value run length flagged as flatline.}

\item{zero_run_length}{Minimum zero-value run length flagged as zero run.}

\item{saturation_min}{Optional lower bound for acceptable signal values.}

\item{saturation_max}{Optional upper bound for acceptable signal values.}

\item{negative_allowed}{Optional logical. If \code{NULL}, negative values are
allowed for phasic component columns but not for conductance-like columns.}
}
\value{
A list with \code{overview}, \code{row_flags}, \code{artifact_runs},
\code{group_summary}, and \code{settings}.
}
\description{
Flags row-level artifacts in Gazepoint electrodermal activity signals,
preferring \code{GSR_US} conductance when available. The helper detects abrupt
jumps, abrupt slopes, flatline runs, zero runs, negative conductance values,
and optional out-of-range values. It is a conservative preprocessing/QC
helper and does not interpret EDA as emotional valence.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/external-interoperability.R
\name{run_gazepoint_neurokit_eda_crosscheck}
\alias{run_gazepoint_neurokit_eda_crosscheck}
\title{Optionally run a NeuroKit2 EDA cross-check}
\usage{
run_gazepoint_neurokit_eda_crosscheck(
  data,
  eda_col = "GSR_US",
  time_col = NULL,
  group_cols = NULL,
  sampling_rate = NULL,
  execute = FALSE,
  python = "python",
  output_dir = tempdir(),
  prefix = "gazepoint_neurokit_crosscheck",
  keep_files = FALSE
)
}
\arguments{
\item{data}{A Gazepoint biometric data frame or
\code{gazepoint_neurokit_eda_input} object.}

\item{eda_col}{EDA/GSR signal column, used when \code{data} is a data frame.}

\item{time_col}{Optional time/counter column.}

\item{group_cols}{Optional grouping columns.}

\item{sampling_rate}{Sampling rate in Hz required for NeuroKit2 execution.}

\item{execute}{Logical. If \code{FALSE}, only prepare input and return skipped
status.}

\item{python}{Python executable.}

\item{output_dir}{Directory for temporary/input/output files.}

\item{prefix}{File prefix.}

\item{keep_files}{Logical. If \code{FALSE}, temporary files produced during
execution may be removed.}
}
\value{
A list with \code{overview}, \code{prepared_input}, \code{crosscheck_summary},
\code{manifest}, and \code{settings}.
}
\description{
Optionally calls Python/NeuroKit2 on prepared EDA input. By default,
\code{execute = FALSE}, so no external dependency is required.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/qc-dashboard-plots.R
\name{plot_gazepoint_time_resets}
\alias{plot_gazepoint_time_resets}
\title{Plot Gazepoint time resets and time-order flags}
\usage{
plot_gazepoint_time_resets(
  data,
  time_col = NULL,
  group_cols = NULL,
  max_groups = 30,
  title = NULL
)
}
\arguments{
\item{data}{A Gazepoint biometric data frame, or an
\code{audit_gazepoint_time_resets()} result.}

\item{time_col}{Optional time or counter column used when \code{data} is a raw
data frame.}

\item{group_cols}{Optional grouping columns used when \code{data} is a raw data
frame.}

\item{max_groups}{Maximum number of groups to display.}

\item{title}{Optional plot title.}
}
\value{
A ggplot object with the package plot contract attached.
}
\description{
Plots row-level time/counter progression and flags from
\code{\link[=audit_gazepoint_time_resets]{audit_gazepoint_time_resets()}}, or computes them from a biometric data
frame. The plot is intended for synchronization and file-structure QC. It
does not establish causal timing or true physiological latency.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hrv-fragmentation-asymmetry.R
\name{extract_gazepoint_hrv_fragmentation}
\alias{extract_gazepoint_hrv_fragmentation}
\title{Extract heart-rate fragmentation features}
\usage{
extract_gazepoint_hrv_fragmentation(
  dat,
  ibi_col = "IBI",
  group_cols = NULL,
  zero_tolerance = 0,
  short_segment_length = 3
)
}
\arguments{
\item{dat}{A data frame containing IBI/RR intervals.}

\item{ibi_col}{Numeric IBI/RR interval column.}

\item{group_cols}{Optional grouping columns.}

\item{zero_tolerance}{Absolute change below which interval differences are
treated as zero.}

\item{short_segment_length}{Maximum segment length counted as short.}
}
\value{
A list with \code{overview}, \code{features}, and \code{settings}.
}
\description{
Computes dependency-light heart-rate fragmentation descriptors from IBI/RR
intervals. Metrics include percentage of inflection points (PIP), inverse
average segment length (IALS), percentage of short segments (PSS), percentage
of alternation segments (PAS), and long/short segment summaries.
}
\details{
These are fragmentation descriptors of interbeat interval dynamics. They
should not be interpreted as clinical diagnoses or direct autonomic-state
labels by themselves.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scr-threshold-sensitivity.R
\name{run_gazepoint_scr_threshold_sensitivity}
\alias{run_gazepoint_scr_threshold_sensitivity}
\title{Run Gazepoint SCR threshold sensitivity checks}
\usage{
run_gazepoint_scr_threshold_sensitivity(
  data,
  phasic_col = NULL,
  signal_col = NULL,
  time_col = NULL,
  group_cols = NULL,
  amplitude_min_values = c(0.005, 0.01, 0.02, 0.03),
  min_peak_distance_values = c(1, 5, 10, 20, 30),
  recovery_fraction = 0.5,
  smooth_width = 1,
  events = NULL,
  event_time_col = NULL,
  event_id_col = NULL,
  event_label_col = NULL,
  ttl_cols = NULL,
  ttl_valid_col = NULL,
  event_detection = c("rising", "active"),
  analysis_window = c(0, 6),
  response_window = c(1, 4),
  peak_selection = c("largest_amplitude", "first_peak"),
  collapse_simultaneous_events = FALSE,
  include_event_windows = TRUE,
  keep_objects = FALSE
)
}
\arguments{
\item{data}{A Gazepoint biometric data frame.}

\item{phasic_col}{Optional phasic EDA signal column, typically
\code{GSR_US_PHASIC}.}

\item{signal_col}{Optional conductance-like fallback signal column, typically
\code{GSR_US}.}

\item{time_col}{Optional time/counter column.}

\item{group_cols}{Optional grouping columns.}

\item{amplitude_min_values}{Numeric vector of SCR amplitude thresholds.}

\item{min_peak_distance_values}{Numeric vector of minimum peak distances.}

\item{recovery_fraction}{Fraction used by \code{detect_gazepoint_scr_peaks()} to
define recovery.}

\item{smooth_width}{Optional moving-average width passed to
\code{detect_gazepoint_scr_peaks()}.}

\item{events}{Optional event table passed to
\code{summarise_gazepoint_scr_event_windows()}.}

\item{event_time_col}{Optional event-time column when \code{events} is supplied.}

\item{event_id_col}{Optional event identifier column when \code{events} is supplied.}

\item{event_label_col}{Optional event label column when \code{events} is supplied.}

\item{ttl_cols}{Optional TTL columns for event derivation when \code{events} is
\code{NULL}.}

\item{ttl_valid_col}{Optional TTL validity column.}

\item{event_detection}{Event-detection rule for TTL columns.}

\item{analysis_window}{Event-relative analysis window.}

\item{response_window}{Event-relative response window.}

\item{peak_selection}{Peak-selection rule for event-window summaries.}

\item{collapse_simultaneous_events}{Logical. Passed to
\code{summarise_gazepoint_scr_event_windows()}.}

\item{include_event_windows}{Logical. If \code{TRUE}, compute event-window
summaries for each sensitivity setting.}

\item{keep_objects}{Logical. If \code{TRUE}, retain peak-detection and
event-window objects in list columns.}
}
\value{
A list with \code{overview}, \code{sensitivity_grid},
\code{peak_group_summary}, \code{event_window_summary}, optional \code{objects}, and
\code{settings}.
}
\description{
Re-runs SCR peak detection across combinations of amplitude thresholds and
minimum peak-distance settings. Optionally, it also carries each peak-detection
result through SCR event-window summaries so users can see how preprocessing
choices affect event-level response rates.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_data.R
\name{prepare_gazepoint_multimodal_model_data}
\alias{prepare_gazepoint_multimodal_model_data}
\title{Prepare Gazepoint multimodal model data}
\usage{
prepare_gazepoint_multimodal_model_data(
  biometrics,
  eye_tracking = NULL,
  group_columns = NULL,
  biometric_is_summarised = FALSE,
  by = NULL,
  all = FALSE
)
}
\arguments{
\item{biometrics}{A data frame containing row-level Gazepoint Biometrics data
or an already summarised biometric window table.}

\item{eye_tracking}{Optional eye-tracking summary table to merge with the
biometric summaries.}

\item{group_columns}{Columns defining the analysis unit, such as
\code{c("USER", "MEDIA_ID")}.}

\item{biometric_is_summarised}{Logical. If \code{FALSE}, biometric window
summaries are created using \code{\link[=summarise_gazepoint_multimodal_windows]{summarise_gazepoint_multimodal_windows()}}.
If \code{TRUE}, \code{biometrics} is treated as already summarised.}

\item{by}{Optional merge keys. If \code{NULL}, \code{group_columns} are used.}

\item{all}{Should a full outer join be used when eye-tracking data are
supplied? Defaults to \code{FALSE}, giving an inner join.}
}
\value{
A data frame with class \code{"gazepoint_multimodal_model_data"} and a
\code{"model_data_summary"} attribute.
}
\description{
Creates a model-ready table from Gazepoint biometric window summaries and,
optionally, eye-tracking summaries produced by \code{gp3tools} or another
workflow. The function is intentionally conservative: it does not fit a
model, impute missing values, or remove rows automatically.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/biometric-plots.R
\name{plot_gazepoint_biometric_signals}
\alias{plot_gazepoint_biometric_signals}
\title{Plot Gazepoint biometric signal time series}
\usage{
plot_gazepoint_biometric_signals(
  data,
  signal_cols = NULL,
  time_col = NULL,
  group_col = NULL,
  max_points = 5000L,
  standardize = FALSE,
  type = c("line", "points", "both"),
  main = NULL,
  xlab = NULL,
  ylab = NULL,
  legend = TRUE,
  plot = TRUE,
  ...
)
}
\arguments{
\item{data}{A data frame.}

\item{signal_cols}{Optional character vector of signal columns. If \code{NULL},
common Gazepoint biometric signal columns are detected.}

\item{time_col}{Optional time/order column for the x-axis. If \code{NULL}, row
number is used.}

\item{group_col}{Optional grouping column recorded in the returned overview.
The current plotting implementation overlays the selected rows rather than
faceting by group.}

\item{max_points}{Maximum number of rows to plot. Large data are evenly
downsampled for display only; returned summaries still describe the input
signal columns.}

\item{standardize}{Logical. Should each signal be z-standardised before
plotting? This is useful when signals are on different scales.}

\item{type}{Plot type: \code{"line"}, \code{"points"}, or \code{"both"}.}

\item{main}{Optional plot title.}

\item{xlab}{Optional x-axis label.}

\item{ylab}{Optional y-axis label.}

\item{legend}{Logical. Should a legend be drawn when more than one signal is
plotted?}

\item{plot}{Logical. If \code{FALSE}, no plot is drawn and only the plot object is
returned.}

\item{...}{Additional arguments passed to \code{\link[=matplot]{matplot()}}.}
}
\value{
A list with \code{overview}, \code{plot_data}, \code{signal_summary}, and \code{settings}.
}
\description{
Plots one or more Gazepoint biometric signals using base R graphics and
returns the plotted data and signal summary. The helper is intentionally
descriptive and does not infer emotional valence, cognition, or HRV from raw
biometric columns.
}
\examples{
df <- data.frame(
  CNT = 1:5,
  GSR = c(1, 1.1, 1.2, 1.1, 1),
  HR = c(70, 71, 72, 71, 70)
)
plot_gazepoint_biometric_signals(df, time_col = "CNT", plot = FALSE)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/biometrics-lme-data.R
\name{prepare_gazepoint_biometrics_lme_data}
\alias{prepare_gazepoint_biometrics_lme_data}
\title{Prepare Gazepoint biometric summaries for mixed-model analysis}
\usage{
prepare_gazepoint_biometrics_lme_data(
  data,
  outcome_col,
  fixed_effect_cols = NULL,
  condition_cols = NULL,
  covariate_cols = NULL,
  random_effect_cols = NULL,
  participant_col = NULL,
  stimulus_col = NULL,
  trial_col = NULL,
  window_col = NULL,
  baseline_col = NULL,
  baseline_correct = FALSE,
  factor_cols = NULL,
  continuous_cols = NULL,
  scale_continuous = FALSE,
  include_window = TRUE,
  drop_missing = TRUE,
  min_rows = 10
)
}
\arguments{
\item{data}{A data frame containing biometric summary rows.}

\item{outcome_col}{Name of the outcome column to analyse.}

\item{fixed_effect_cols}{Optional fixed-effect predictor columns.}

\item{condition_cols}{Optional condition/design columns to include as fixed
effects.}

\item{covariate_cols}{Optional covariate columns to include as fixed effects.}

\item{random_effect_cols}{Optional grouping columns for random intercepts.}

\item{participant_col, stimulus_col, trial_col}{Optional common grouping columns.}

\item{window_col}{Optional analysis-window column. Included as a fixed effect
when \code{include_window = TRUE}.}

\item{baseline_col}{Optional baseline column.}

\item{baseline_correct}{Logical. If \code{TRUE}, creates an outcome column equal to
\code{outcome_col - baseline_col}.}

\item{factor_cols}{Optional columns to convert to factors.}

\item{continuous_cols}{Optional numeric predictor columns to scale when
\code{scale_continuous = TRUE}.}

\item{scale_continuous}{Logical. If \code{TRUE}, creates z-scored versions of
numeric continuous predictors and uses those in the formula.}

\item{include_window}{Logical. Should \code{window_col} be included as a fixed
effect?}

\item{drop_missing}{Logical. Should incomplete model rows be removed from
\code{model_data}?}

\item{min_rows}{Minimum number of complete rows required for a \code{"ready"}
status.}
}
\value{
A list with \code{overview}, \code{data}, \code{model_data}, \code{model_formula},
\code{variable_summary}, and \code{settings}.
}
\description{
Prepares biometric window-level or event-level summaries for downstream
mixed-model analysis. This helper does not fit a model. It checks variables,
optionally baseline-corrects the selected outcome, optionally scales numeric
predictors, converts grouping/factor variables, flags complete cases, and
returns a conservative model formula.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/external-interoperability.R
\name{export_gazepoint_rhrv_input}
\alias{export_gazepoint_rhrv_input}
\title{Export Gazepoint IBI data for RHRV-style workflows}
\usage{
export_gazepoint_rhrv_input(
  data,
  ibi_col = "IBI_clean_ms",
  group_cols = NULL,
  unit = c("auto", "ms", "seconds"),
  collapse_repeated_intervals = TRUE,
  repeated_tolerance_ms = 1e-08,
  min_ibi_ms = 300,
  max_ibi_ms = 2000,
  output_dir = NULL,
  prefix = "gazepoint_rhrv"
)
}
\arguments{
\item{data}{A Gazepoint biometric data frame or \code{gazepoint_ibi_filter} object.}

\item{ibi_col}{IBI/RR interval column.}

\item{group_cols}{Optional grouping columns.}

\item{unit}{Unit of the IBI column: \code{"auto"}, \code{"ms"}, or \code{"seconds"}.}

\item{collapse_repeated_intervals}{Logical. If \code{TRUE}, consecutive repeated
IBI values are collapsed before export.}

\item{repeated_tolerance_ms}{Numeric tolerance used when identifying repeated
consecutive IBI values.}

\item{min_ibi_ms}{Minimum plausible IBI in milliseconds retained for export.}

\item{max_ibi_ms}{Maximum plausible IBI in milliseconds retained for export.}

\item{output_dir}{Optional directory where per-group CSV files are written.}

\item{prefix}{File prefix used when \code{output_dir} is supplied.}
}
\value{
A list with \code{overview}, \code{beat_table}, \code{group_summary}, \code{manifest},
and \code{settings}.
}
\description{
Prepares cleaned inter-beat interval data for external HRV workflows. The
helper writes simple beat tables with cumulative beat time and IBI/RR interval
columns. It does not use the Gazepoint \code{HRV} column.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aoi-biometrics.R
\name{plot_gazepoint_aoi_biometrics}
\alias{plot_gazepoint_aoi_biometrics}
\title{Plot AOI-linked biometric summaries}
\usage{
plot_gazepoint_aoi_biometrics(
  x,
  value_col = "mean_value",
  aoi_col = "aoi_label",
  signal_col = "signal",
  group_col = NULL,
  plot_type = c("boxplot", "point", "line"),
  title = NULL
)
}
\arguments{
\item{x}{A \code{gazepoint_aoi_biometrics_summary},
\code{gazepoint_aoi_biometrics_model_data}, or data frame.}

\item{value_col}{Value column to plot.}

\item{aoi_col}{AOI label column.}

\item{signal_col}{Signal label column.}

\item{group_col}{Optional grouping column.}

\item{plot_type}{\code{"boxplot"}, \code{"point"}, or \code{"line"}.}

\item{title}{Optional plot title.}
}
\value{
A ggplot object with plot data stored in attributes.
}
\description{
Plots AOI-biometric summary values as a ggplot object.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ibi-helpers.R
\name{summarise_gazepoint_ibi_windows}
\alias{summarise_gazepoint_ibi_windows}
\title{Summarise IBI/RR windows}
\usage{
summarise_gazepoint_ibi_windows(
  data,
  ibi_col = NULL,
  group_cols = NULL,
  time_col = NULL,
  unit = c("auto", "milliseconds", "seconds"),
  min_ibi_ms = 300,
  max_ibi_ms = 2000,
  max_jump_ms = 500,
  exclude_large_jumps = TRUE,
  min_valid_ibi = 2L
)
}
\arguments{
\item{data}{A data frame.}

\item{ibi_col}{Optional IBI/RR interval column. If \code{NULL}, the function
detects recognised IBI/RR-style column names.}

\item{group_cols}{Optional grouping columns defining windows, such as
participant, trial, stimulus, condition, or window labels.}

\item{time_col}{Optional time/order column used to order IBI values before
successive-difference metrics are computed.}

\item{unit}{Unit of the IBI values. \code{"auto"} treats median values below 10 as
seconds and larger values as milliseconds.}

\item{min_ibi_ms}{Minimum plausible IBI in milliseconds.}

\item{max_ibi_ms}{Maximum plausible IBI in milliseconds.}

\item{max_jump_ms}{Maximum plausible absolute change between successive IBI
values within a group.}

\item{exclude_large_jumps}{Logical. Should intervals flagged as large jumps be
excluded from the window summaries?}

\item{min_valid_ibi}{Minimum valid IBI count required for a window to be marked
as sufficient.}
}
\value{
A list with \code{overview}, \code{windows}, \code{samples}, and \code{settings}.
}
\description{
Computes descriptive IBI/RR interval and simple HRV-style window summaries
from genuine inter-beat interval data. The function does not use raw \code{HRV}
columns as HRV metrics. It calculates metrics such as mean IBI, mean
instantaneous heart rate, SDNN, RMSSD, pNN20, and pNN50 only from valid IBI/RR
intervals.
}
\examples{
df <- data.frame(
  USER = rep(c("P1", "P2"), each = 4),
  IBI = c(800, 810, 790, 805, 900, 910, 905, 920)
)
summarise_gazepoint_ibi_windows(df, group_cols = "USER")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cvxeda-tau-optimization.R
\name{optimize_gazepoint_cvxeda_tau}
\alias{optimize_gazepoint_cvxeda_tau}
\title{Optimise subject-specific cvxEDA slow time constant}
\usage{
optimize_gazepoint_cvxeda_tau(
  dat,
  eda_col = "GSR_US",
  time_col = "CNT",
  group_cols = NULL,
  tau0_grid = seq(2, 4, by = 0.25),
  tau1 = 0.7,
  sampling_rate = NULL,
  ridge_lambda = 0.01,
  max_irf_seconds = 20
)
}
\arguments{
\item{dat}{A data frame containing EDA data.}

\item{eda_col}{Numeric EDA/conductance column.}

\item{time_col}{Numeric time column.}

\item{group_cols}{Optional grouping columns, usually participant/session.}

\item{tau0_grid}{Candidate slow time constants.}

\item{tau1}{Fixed fast time constant.}

\item{sampling_rate}{Optional sampling rate in Hz. If \code{NULL}, estimated from
\code{time_col}.}

\item{ridge_lambda}{Small ridge penalty used in frequency-domain
deconvolution.}

\item{max_irf_seconds}{Maximum impulse-response duration.}
}
\value{
A list with \code{overview}, \code{best_tau}, \code{optimization_table}, and
\code{settings}.
}
\description{
Performs a dependency-light grid search over the slow Bateman impulse-response
time constant used in cvxEDA-style EDA decomposition workflows. The default
fast time constant is fixed at 0.7 seconds and the slow time constant is
searched between 2 and 4 seconds. This function does not run the original
cvxEDA optimisation; it provides a subject-specific tau-selection bridge for
downstream cvxEDA-style workflows.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/folder.R
\name{import_gazepoint_biometric_folder}
\alias{import_gazepoint_biometric_folder}
\title{Import a folder of Gazepoint Biometrics exports}
\usage{
import_gazepoint_biometric_folder(
  path,
  pattern = "\\\\.csv$",
  recursive = FALSE,
  include_fixations = TRUE,
  include_all_gaze = TRUE,
  include_other_csv = FALSE,
  na = c("", "NA", "NaN")
)
}
\arguments{
\item{path}{Folder containing Gazepoint CSV exports.}

\item{pattern}{Regular expression used to identify candidate CSV files.}

\item{recursive}{Should subfolders be searched?}

\item{include_fixations}{Should files with \code{"fixation"} in the file name be
included?}

\item{include_all_gaze}{Should files with \code{"all_gaze"} in the file name be
included?}

\item{include_other_csv}{Should other CSV files be attempted? The default is
\code{FALSE} to avoid accidentally trying to parse multi-section
\code{Data_Summary_export} files as rectangular data.}

\item{na}{Values that should be treated as missing.}
}
\value{
A data frame with all imported rows combined. The output includes a
\code{source_file} column and has class \code{"gazepoint_biometrics_folder"}.
}
\description{
Reads rectangular Gazepoint CSV exports from a folder and combines files that
contain at least one known Gazepoint Biometrics column. This function is
designed for all-gaze and fixation-style exports. Multi-section Gazepoint
Data Summary files should be parsed separately.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/eda-advanced.R
\name{decompose_gazepoint_eda}
\alias{decompose_gazepoint_eda}
\title{Decompose Gazepoint GSR/EDA into tonic and phasic components}
\usage{
decompose_gazepoint_eda(
  data,
  signal_col = NULL,
  tonic_col = NULL,
  phasic_col = NULL,
  time_col = NULL,
  group_cols = NULL,
  window_size = 31L,
  output_prefix = "eda",
  overwrite = FALSE
)
}
\arguments{
\item{data}{A data frame.}

\item{signal_col}{Optional GSR/EDA signal column. If \code{NULL}, a likely
conductance-like column is detected.}

\item{tonic_col}{Optional existing tonic column.}

\item{phasic_col}{Optional existing phasic column.}

\item{time_col}{Optional time/order column.}

\item{group_cols}{Optional grouping columns.}

\item{window_size}{Rolling-median window size used when existing tonic/phasic
columns are not available. Even values are increased by one.}

\item{output_prefix}{Prefix for output columns.}

\item{overwrite}{Logical. Should existing output columns be overwritten?}
}
\value{
A data frame with added tonic, phasic, and method columns. Attributes
include \code{overview} and \code{settings}.
}
\description{
Creates descriptive tonic and phasic EDA columns from Gazepoint GSR/EDA data.
If vendor-provided tonic/phasic columns such as \code{GSR_US_TONIC} and
\code{GSR_US_PHASIC} are available, they are used by default. Otherwise, a simple
rolling-median tonic estimate is used and the phasic component is calculated
as signal minus tonic. This helper is intentionally conservative and does not
replace specialised biosignal-processing software.
}
\examples{
df <- data.frame(CNT = 1:10, GSR_US = seq(1, 2, length.out = 10))
out <- decompose_gazepoint_eda(df, signal_col = "GSR_US", window_size = 3)
names(out)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/multimodal-timeline-plot.R
\name{plot_gazepoint_multimodal_timeline}
\alias{plot_gazepoint_multimodal_timeline}
\title{Plot multimodal Gazepoint biometric timelines}
\usage{
plot_gazepoint_multimodal_timeline(
  data,
  time_col = NULL,
  signal_cols = NULL,
  group_cols = NULL,
  participant_col = NULL,
  stimulus_col = NULL,
  trial_col = NULL,
  event_time_col = NULL,
  event_col = NULL,
  standardise = TRUE,
  show_event_markers = TRUE,
  title = NULL
)
}
\arguments{
\item{data}{A data frame containing biometric samples or aligned biometric
rows.}

\item{time_col}{Optional time column. If \code{NULL}, common time columns are
detected automatically.}

\item{signal_cols}{Optional biometric signal columns. If \code{NULL}, common
Gazepoint biometric columns are detected automatically.}

\item{group_cols}{Optional grouping columns used to separate trajectories.}

\item{participant_col, stimulus_col, trial_col}{Optional common grouping columns
to add to \code{group_cols}.}

\item{event_time_col}{Optional column containing event times for vertical
markers.}

\item{event_col}{Optional event/TTL indicator column used for vertical markers.}

\item{standardise}{Logical. If \code{TRUE}, signals are z-scored within channel for
visual comparison.}

\item{show_event_markers}{Logical. Should event markers be drawn when
available?}

\item{title}{Optional plot title.}
}
\value{
A ggplot object with the long plotting data stored in the \code{plot_data}
attribute and settings stored in the \code{settings} attribute.
}
\description{
Creates a conservative timeline plot for one or more biometric channels. The
plot is intended for inspection, synchronization checks, and reporting
support. It does not interpret electrodermal activity as emotional valence.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/feature-inventory.R
\name{create_gazepoint_biometrics_feature_inventory}
\alias{create_gazepoint_biometrics_feature_inventory}
\title{Create a gpbiometrics feature inventory}
\usage{
create_gazepoint_biometrics_feature_inventory(include_internal = FALSE)
}
\arguments{
\item{include_internal}{Logical. If \code{TRUE}, also checks for non-exported
internal helper names when they are included in the inventory.}
}
\value{
A list with \code{overview}, \code{inventory}, \code{domain_summary},
\code{missing_expected}, and \code{settings}.
}
\description{
Creates a structured inventory of implemented gpbiometrics helper functions.
This is useful for reporting, readiness checks, documentation, and package
development audits.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/workflow.R
\name{run_gazepoint_biometrics_workflow}
\alias{run_gazepoint_biometrics_workflow}
\title{Run a Gazepoint Biometrics workflow}
\usage{
run_gazepoint_biometrics_workflow(
  path,
  group_columns = NULL,
  recursive = FALSE,
  include_fixations = FALSE,
  include_all_gaze = TRUE,
  include_other_csv = FALSE,
  require_active_signal = TRUE,
  create_exclusion_recommendations = TRUE,
  gsr_min_usable_pct = 50,
  hr_min_usable_pct = 50,
  dial_min_usable_pct = 50,
  extract_ttl_events = TRUE,
  ttl_event_mode = c("changes", "nonzero"),
  audit_sampling = TRUE,
  sampling_group_columns = NULL,
  sampling_time_column = NULL,
  sampling_time_unit = c("samples", "seconds", "milliseconds", "microseconds"),
  expected_sampling_rate_hz = 60
)
}
\arguments{
\item{path}{Folder containing Gazepoint CSV exports.}

\item{group_columns}{Optional columns used to create multimodal window
summaries, such as \code{c("source_participant", "MEDIA_ID")}.}

\item{recursive}{Should subfolders be searched?}

\item{include_fixations}{Should fixation files be imported? Defaults to
\code{FALSE} because continuous biometric summaries should usually be computed
from all-gaze sample-level exports rather than fixation-level exports.}

\item{include_all_gaze}{Should all-gaze files be imported?}

\item{include_other_csv}{Should other non-Data-Summary CSV files be attempted?}

\item{require_active_signal}{Logical. Should inactive biometric signals be
flagged in validation/checklist outputs?}

\item{create_exclusion_recommendations}{Logical. Should window-level and
participant-level keep/review/exclude recommendations be created when
\code{group_columns} are supplied?}

\item{gsr_min_usable_pct}{Minimum acceptable usable percentage for GSR/EDA
windows.}

\item{hr_min_usable_pct}{Minimum acceptable usable percentage for heart-rate
windows.}

\item{dial_min_usable_pct}{Minimum acceptable usable percentage for
engagement-dial windows.}

\item{extract_ttl_events}{Logical. Should TTL marker events be extracted?}

\item{ttl_event_mode}{TTL event extraction mode passed to
\code{extract_gazepoint_ttl_events()}. Use \code{"changes"} or \code{"nonzero"}.}

\item{audit_sampling}{Logical. Should sampling/timing information be audited?}

\item{sampling_group_columns}{Optional columns for the sampling audit. If
\code{NULL}, the workflow uses available file/participant/media columns.}

\item{sampling_time_column}{Optional time/order column for the sampling audit.}

\item{sampling_time_unit}{Unit of the selected time/order column. Use
\code{"seconds"}, \code{"milliseconds"}, \code{"microseconds"}, or \code{"samples"}.}

\item{expected_sampling_rate_hz}{Optional expected sampling rate in Hz.}
}
\value{
A list with imported data, validation outputs, missingness summaries,
quality audits, sampling/timing audits, optional window summaries, optional
exclusion recommendations, optional TTL events, checklist, and methods
text. The object has class \code{"gazepoint_biometrics_workflow"}.
}
\description{
Runs a compact end-to-end workflow for Gazepoint Biometrics exports. The
workflow imports rectangular all-gaze/fixation-style CSV exports from a
folder, validates biometric columns, detects active channels, audits
missingness, signal quality, sampling/timing, optionally creates window-level
summaries, creates optional biometric exclusion recommendations, extracts
optional TTL marker events, and produces checklist and methods-text outputs.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shiny-annotator.R
\name{run_gpbiometrics_shiny_annotator}
\alias{run_gpbiometrics_shiny_annotator}
\title{Launch a lightweight gpbiometrics Shiny peak/artifact annotator}
\usage{
run_gpbiometrics_shiny_annotator()
}
\value{
A Shiny application object, launched for interactive use.
}
\description{
Launches a local Shiny app for manual EDA peak/artifact annotation. The app
lets users upload a Gazepoint CSV, inspect a selected signal, enter manual
peak/artifact intervals, and download annotation CSVs.
}
\details{
This is an optional local GUI helper. It does not replace automated scoring
or infer emotion, stress, cognition, trust, preference, or diagnosis.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reporting.R
\name{create_gazepoint_biometrics_checklist}
\alias{create_gazepoint_biometrics_checklist}
\title{Create a Gazepoint Biometrics reporting checklist}
\usage{
create_gazepoint_biometrics_checklist(data, require_active_signal = TRUE)
}
\arguments{
\item{data}{A data frame or a path to a Gazepoint CSV export.}

\item{require_active_signal}{Logical. Should inactive biometric channels be
flagged in the validation output?}
}
\value{
A list with \code{overview}, \code{channels}, \code{quality}, \code{missingness},
\code{validation_issues}, \code{workflow_capabilities}, \code{feature_inventory},
\code{reporting_guidance}, and \code{interpretation_cautions}.
}
\description{
Creates a compact reporting checklist for Gazepoint Biometrics exports. The
checklist summarises detected biometric channels, validation issues, signal
quality, missingness, available workflow domains, and interpretation
cautions. It is intended to support transparent manuscript reporting and
reviewer-facing methods documentation.
}
